/**********************************************************************
Copyright (c) 2003 David Jencks and others. All rights reserved.
This program and the accompanying materials are made available under
the terms of the JPOX License v1.0 which accompanies this distribution.

Contributors:
2003 Erik Bengtson - removed exist() operation
2003 Andy Jefferson - commented
2003 Andy Jefferson - added check on makePersistent that class is persistence capable
2003 Andy Jefferson - introduction of localiser
2003 Erik Bengtson - provided initial caching mechanism
2003 Erik Bengtson - added getObjectbyAID
2003 Erik Bengtson - removed unused variables
2004 Erik Bengtson - implemented evictAll    
2004 Andy Jefferson - converted to use Logger
2004 Andy Jefferson - added LifecycleListener
2004 Andy Jefferson - rewritten Cache
2005 Andy Jefferson - moved ClassLoaderResolver handling to separate class
     ...
 **********************************************************************/
package org.jpox;

import java.io.PrintWriter;
import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Modifier;
import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.jdo.AttachLifecycleListener;
import javax.jdo.ClearLifecycleListener;
import javax.jdo.CreateLifecycleListener;
import javax.jdo.DeleteLifecycleListener;
import javax.jdo.DetachLifecycleListener;
import javax.jdo.DirtyLifecycleListener;
import javax.jdo.Extent;
import javax.jdo.FetchPlan;
import javax.jdo.JDOException;
import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOFatalUserException;
import javax.jdo.JDOObjectNotFoundException;
import javax.jdo.JDOUserException;
import javax.jdo.InstanceLifecycleEvent;
import javax.jdo.InstanceLifecycleListener;
import javax.jdo.LoadLifecycleListener;
import javax.jdo.PersistenceManagerFactory;
import javax.jdo.Query;
import javax.jdo.Sequence;
import javax.jdo.SingleFieldIdentity;
import javax.jdo.StoreLifecycleListener;
import javax.jdo.spi.JDOImplHelper;
import javax.jdo.spi.PersistenceCapable;

import org.jpox.cache.Level1Cache;
import org.jpox.cache.Level2Cache;
import org.jpox.exceptions.ClassNotDetachableException;
import org.jpox.exceptions.ClassNotPersistenceCapableException;
import org.jpox.exceptions.CommitStateTransitionException;
import org.jpox.exceptions.MetaDataForPersistenceCapableClassNotReachableException;
import org.jpox.exceptions.RollbackStateTransitionException;
import org.jpox.exceptions.TransactionActiveException;
import org.jpox.exceptions.TransactionNotActiveException;
import org.jpox.metadata.ClassMetaData;
import org.jpox.metadata.FieldMetaData;
import org.jpox.metadata.IdentityType;
import org.jpox.metadata.MetaDataManager;
import org.jpox.metadata.QueryLanguage;
import org.jpox.metadata.QueryMetaData;
import org.jpox.sco.SCO;
import org.jpox.state.StateManagerImpl;
import org.jpox.store.FieldValues;
import org.jpox.store.OID;
import org.jpox.store.SCOID;
import org.jpox.store.StoreManager;
import org.jpox.store.query.JDOQLQuery;
import org.jpox.store.query.JPOXSQLQuery;
import org.jpox.store.query.SQLQuery;
import org.jpox.store.rdbms.RDBMSManagerFactory;
import org.jpox.util.AIDUtils;
import org.jpox.util.ClassUtils;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;

/**
 * This class implements the functionality needed for a persistence manager. 
 * It does not implement PersistenceManager since it would be confusing for a 
 * ManagedConnectionFactory to implement a user-app-level class.
 * 
 * @version $Revision: 1.39 $
 */
public abstract class AbstractPersistenceManager implements PersistenceManager
{
    /** Localisation utility for output messages */
    protected static final Localiser LOCALISER = Localiser.getInstance("org.jpox.Localisation");

    /** Owning PMF for this PM. */
    protected final AbstractPersistenceManagerFactory apmf;

    /** StoreManager that the PM relates to. */
    private StoreManager srm;

    /** Map of user objects attached to this PM. */
    private Map userObjectMap;

    /** User object attached to the PM. */
    private Object userObject;

    /** State variable for whether the PM is closed. */
    private boolean closed;

    /** Level 1 Cache */
    private Level1Cache cache;

    /** Cache of StateManagers enlisted in the current transaction */
    private Set txCache = new HashSet();

    /** State variable used when searching for the StateManager for an object. */
    private PersistenceCapable lookingForStateManagerFor = null;

    /** State variable used when searching for the StateManager for an object. */
    private StateManager foundStateManager = null;

    /** Current transaction */
    protected Transaction tx;

    /** Whether to ignore the cache */
    private boolean ignoreCache;

    /** Whether to detach objects on close of the PM. */
    private boolean detachOnClose;

    /** Whether to operate multithreaded */
    private boolean multithreaded;

    /** Class loader to use by default */
    private ClassLoader contextLoader;

    /** List of StateManagers for all current dirty objects managed by this PM. */
    private List dirties = new ArrayList();

    /** Current FetchPlan for the PM. */
    private FetchPlan fetchPlan;

    /** Lifecycle Listeners for events */
    protected List lifecycleListeners = new ArrayList();

    /** The ClassLoader resolver to use for class loading issues. */
    ClassLoaderResolver clr = null;

    /**
     * Constructor.
     * @param apmf Persistence Manager Factory
     * @param userName Username for the datastore
     * @param password Password for the datastore
     */
    public AbstractPersistenceManager(AbstractPersistenceManagerFactory apmf, String userName, String password)
    {
        contextLoader = Thread.currentThread().getContextClassLoader();
        clr = new JDOClassLoaderResolver(contextLoader);
        this.apmf = apmf;

        closed = false;
        if (JPOXLogger.JDO.isDebugEnabled())
        {
            JPOXLogger.JDO.debug(LOCALISER.msg("PM.Opened", this, srm));
        }

        // TODO Generalise this so we know to use RDBMSManagerFactory based on the URL
        // For example if it starts with "jdbc:" it should be RDBMS
        srm = RDBMSManagerFactory.getStoreManager(clr, apmf, userName, password);

        setIgnoreCache(apmf.getPMFConfiguration().getIgnoreCache());
        setDetachOnClose(apmf.getPMFConfiguration().getDetachOnClose());
        setMultithreaded(apmf.getPMFConfiguration().getMultithreaded());
        fetchPlan = new FetchPlanImpl();
        userObject = null;
        userObjectMap = new HashMap();

        // Set up the Level 1 Cache
        String level1Type = apmf.getPMFConfiguration().getJdoCacheLevel1Type();
        try
        {
            Class level1CacheClass = Class.forName(level1Type);
            cache = (Level1Cache) level1CacheClass.newInstance();
            JPOXLogger.CACHE.info(LOCALISER.msg("Cache.Level1.Initialised", level1Type));
        }
        catch (Exception e)
        {
            String fallbackType = "org.jpox.cache.WeakRefCache";
            JPOXLogger.CACHE.info(LOCALISER.msg("Cache.Level1.FailedRetrying", level1Type, fallbackType));
            level1Type = fallbackType;
            try
            {
                Class level1CacheClass = Class.forName(level1Type);
                cache = (Level1Cache) level1CacheClass.newInstance();
                JPOXLogger.CACHE.info(LOCALISER.msg("Cache.Level1.Initialised", level1Type));
            }
            catch (Exception e2)
            {
                JPOXLogger.CACHE.info(LOCALISER.msg("Cache.Level1.InitialisationFailed"));
            }
        }
    }

    /**
     * Accessor for the ClassLoaderResolver
     * @return the ClassLoaderResolver
     */
    public ClassLoaderResolver getClassLoaderResolver()
    {
        return clr;
    }

    /**
     * Accessor for the Store Manager.
     * @return StoreManager
     */
    public StoreManager getStoreManager()
    {
        return srm;
    }

    /**
     * Accessor for a Connection for this datastore.
     * @param forWriting Whether the connection can be used for writes
     * @return The Connection to the datastore
     * @throws SQLException Thrown if an error occurs in the creation of the
     * connection.
     */
    public Connection getConnection(boolean forWriting) 
    throws SQLException
    {
        return tx.getConnection(forWriting);
    }

    /**
     * Accessor for the connection to the datasource.
     * @param forWriting Whether the connection is read-only or read-write.
     * @param useOpenedConnection whether to use an already opened connection, if one is opened. 
     * @return The Connection
     * @throws SQLException When a problem occurs getting the connection.
     */
    public Connection getConnection(boolean forWriting, boolean useOpenedConnection) 
    throws SQLException
    {
        return tx.getConnection(forWriting, useOpenedConnection);
    }

    /**
     * Checks whether a connection to the datasource is open.
     * @return true if the Connection is open
     * @throws SQLException Thrown if an error occurs checking the connection
     */
    public boolean isConnectionOpen()
    {
        return tx.isConnectionOpen();
    }

    /**
     * Method to return a Connection for this datastore to the pool.
     * @param conn The Connection
     * @throws SQLException Thrown if an error occurs in the release of the
     * connection.
     */
    public synchronized void releaseConnection(Connection conn) throws SQLException
    {
        tx.releaseConnection(conn);
    }

    /**
     * Method to enlist the specified StateManager in the current transaction.
     * @param sm The StateManager
     */
    public synchronized void enlistInTransaction(StateManager sm)
    {
        assertActiveTransaction();
        if (JPOXLogger.TRANSACTION.isDebugEnabled())
        {
            JPOXLogger.TRANSACTION.debug(LOCALISER.msg("Transaction.ObjectEnlistedInCache", sm.getInternalObjectId().toString()));
        }
        if (!txCache.add(sm))
        {
            throw new JDOFatalInternalException(LOCALISER.msg("PM.ObjectAlreadyTransactional", sm.getInternalObjectId()));
        }
    }

    /**
     * Method to evict the specified StateManager from the current transaction.
     * @param sm The StateManager
     */
    public synchronized void evictFromTransaction(StateManager sm)
    {
        if (JPOXLogger.TRANSACTION.isDebugEnabled())
        {
            JPOXLogger.TRANSACTION.debug(LOCALISER.msg("Transaction.ObjectEvictedFromCache", sm.getInternalObjectId().toString()));
        }

        if (!txCache.remove(sm))
        {
            throw new JDOFatalInternalException(LOCALISER.msg("PM.ObjectNotTransactional", sm.getInternalObjectId()));
        }
    }

    /**
     * Method to remove the specified StateManager.
     * @param sm The StateManager
     */
    public synchronized void removeStateManager(StateManager sm)
    {
        PersistenceCapable pc = sm.getObject();

        // Remove from the Level 1 Cache
        removeObjectFromCache(pc, sm.getInternalObjectId(), true, false);

        // Remove it from any transaction
        txCache.remove(sm);
    }

    /**
     * Method to add the specified StateManager.
     * This has the effect of adding the object to the respective caches
     * @param sm The StateManager
     */
    public synchronized void addStateManager(StateManager sm)
    {
        // Add to the Level 1/2 Caches
        PersistenceCapable pc = sm.getObject();
        putObjectIntoCache(pc, true, true);
    }

    /**
     * Method to assert if this Persistence Manager is open. Throws a
     * JDOFatalUserException if it has been closed.
     */
    private void assertIsOpen()
    {
        if (isClosed())
        {
            throw new JDOFatalUserException(LOCALISER.msg("PM.IsClosed"));
        }
    }

    /**
     * Utility method to check if the specified class has reachable metadata.
     * @param cls The class to check
     * @return Whether the class has reachable metadata.
     */
    private boolean hasMetaDataForPersistenceCapableClass(Class cls)
    {
        return cls != null && getMetaDataManager().getMetaDataForClass(cls) != null;
    }

    /**
     * Method to assert if the specified class is Persistence Capable. Throws a
     * ClassNotPersistenceCapableException if not capable
     * @param cls The class to check
     */
    private void assertPersistenceCapableClass(Class cls)
    {
        if (!ClassUtils.isPersistenceCapableClass(cls))
        {
            throw new ClassNotPersistenceCapableException(cls.getName());
        }
        if (!hasMetaDataForPersistenceCapableClass(cls))
        {
            throw new MetaDataForPersistenceCapableClassNotReachableException(cls.getName());
        }
    }

    /**
     * Method to assert if the specified object is Persistence Capable. Throws a
     * ClassNotPersistenceCapableException if not capable
     * @param object The object to check
     */
    private void assertPersistenceCapable(Object object)
    {
        if (!ClassUtils.isPersistenceCapable(object))
        {
            throw new ClassNotPersistenceCapableException(object.getClass().getName());
        }
        if (!hasMetaDataForPersistenceCapableClass(object.getClass()))
        {
            throw new MetaDataForPersistenceCapableClassNotReachableException(object.getClass().getName());
        }
    }

    /**
     * Method to assert if the specified object is Detachable. Throws a
     * ClassNotDetachableException if not capable
     * @param object The object to check
     */
    private void assertDetachable(Object object)
    {
        if (!ClassUtils.isDetachable(object))
        {
            throw new ClassNotDetachableException(object.getClass().getName());
        }
    }

    /**
     * Method to assert if the current transaction is active. Throws a
     * TransactionNotActiveException if not active
     */
    private void assertActiveTransaction()
    {
        if (!tx.isActive())
        {
            throw new TransactionNotActiveException();
        }
    }

    /**
     * Accessor for whether this Persistence Manager is closed.
     * 
     * @return Whether this manager is closed.
     */
    public boolean isClosed()
    {
        return closed;
    }

    /**
     * Method to close the Persistence Manager.
     */
    public synchronized void close()
    {
        if (closed)
        {
            throw new JDOUserException(LOCALISER.msg("PM.IsClosed"));
        }
        if (tx.isActive())
        {
            throw new TransactionActiveException();
        }

        contextLoader = null;
        userObject = null;
        userObjectMap = null;

        // If the user requires detachment on closure, go through the managed objects and detach the lot.
        if (detachOnClose)
        {
            // TODO Detach the object graph
        }

        // Disconnect all StateManagers for objects managed by this PM
        ArrayList stateManagersToDisconnect = new ArrayList();
        Iterator iter = cache.values().iterator();
        while (iter.hasNext())
        {
            PersistenceCapable pc = (PersistenceCapable) iter.next();
            StateManagerImpl sm = (StateManagerImpl)findStateManager(pc);
            if (sm != null)
            {
                stateManagersToDisconnect.add(sm);
            }
            else
            {
                // Should never happen, but provides some defense should the situation arise.
                JPOXLogger.CACHE.warn("Level 1 Cache contained object with id " + pc.jdoGetObjectId() + " but no StateManager was found for this object");
            }
        }

        Iterator smIter = stateManagersToDisconnect.iterator();
        while (smIter.hasNext())
        {
            StateManagerImpl sm = (StateManagerImpl)smIter.next();
            sm.disconnect();
        }
        cache.clear();
        JPOXLogger.CACHE.info(LOCALISER.msg("Cache.Level1.Cleared"));

        // Release this PM from the owning PMF
        apmf.releasePersistenceManager(getPMHandle());

        txCache.clear();
        closed = true;
        srm = null;
        tx = null;
        dirties.clear();
        if (JPOXLogger.JDO.isDebugEnabled())
        {
            JPOXLogger.JDO.debug(LOCALISER.msg("PM.Closed", this));
        }
    }

    /**
     * Accessor for the current transaction.
     * @return The transaction
     */
    public javax.jdo.Transaction currentTransaction()
    {
        assertIsOpen();
        return tx;
    }

    // ----------------------------- Eviction --------------------------------------------

    /**
     * Method to evict an object (internally).
     * Evicts the objects from L1 and L2 caches.
     * @param obj The object
     */
    private void internalEvict(Object obj)
    {
        if (obj == null)
        {
            return;
        }
        assertPersistenceCapable(obj);
        PersistenceCapable pc = (PersistenceCapable) obj;
        StateManager sm = findStateManager(pc);
        if (sm == null)
        {
            throw new JDOUserException(LOCALISER.msg("PM.DifferentOwner", pc.jdoGetObjectId()));
        }
        sm.evict();

        // Evict from L1 and L2
        removeObjectFromCache(pc, pc.jdoGetObjectId(), true, true);
    }

    /**
     * Method to evict an object.
     * @param pc The object
     */
    public synchronized void evict(Object pc)
    {
        assertIsOpen();
        internalEvict(pc);
    }

    /**
     * Method to evict an array of objects.
     * @param pcs The objects
     */
    public synchronized void evictAll(Object[] pcs)
    {
        evictAll(Arrays.asList(pcs));
    }

    /**
     * Method to evict a collection of objects. Throws a JDOUserException if
     * some instances could not be evicted.
     * @param pcs The objects
     */
    public synchronized void evictAll(Collection pcs)
    {
        assertIsOpen();
        ArrayList failures = new ArrayList();
        Iterator i = pcs.iterator();
        while (i.hasNext())
        {
            try
            {
                internalEvict(i.next());
            }
            catch (RuntimeException e)
            {
                failures.add(e);
            }
        }
        if (!failures.isEmpty())
        {
            throw new JDOUserException(LOCALISER.msg("PM.EvictError"), (Exception[]) failures.toArray(new Exception[failures.size()]));
        }
    }

    /**
     * Method to evict all current objects.
     */
    public synchronized void evictAll()
    {
        assertIsOpen();

        // All persistent non-transactional instances should be evicted here,
        // but we don't support those yet.
        ArrayList stateManagersToEvict = new ArrayList();
        Iterator iter = cache.values().iterator();
        while (iter.hasNext())
        {
            PersistenceCapable pc = (PersistenceCapable) iter.next();
            StateManagerImpl sm = (StateManagerImpl)findStateManager(pc);
            stateManagersToEvict.add(sm);
        }

        // Evict StateManagers and remove objects from cache
        // Performed in separate loop to avoid ConcurrentModificationException
        Iterator smIter = stateManagersToEvict.iterator();
        while (smIter.hasNext())
        {
            StateManagerImpl sm = (StateManagerImpl)smIter.next();
            PersistenceCapable pc = sm.getObject();
            sm.evict();

            // Evict from L1 and L2
            removeObjectFromCache(pc, pc.jdoGetObjectId(), true, true);
        }
    }

    // --------------------------------- Refresh ----------------------------------------

    /**
     * Method to do a refresh of an object (internally), updating it from
     * its datastore representation. Also updates the object in the L1/L2 caches.
     * @param obj The Object
     */
    private void internalRefresh(Object obj)
    {
        if (obj == null)
        {
            return;
        }
        PersistenceCapable pc = (PersistenceCapable) obj;
        StateManager sm = findStateManager(pc);
        if (sm == null)
        {
            throw new JDOUserException(LOCALISER.msg("PM.DifferentOwner", pc.jdoGetObjectId()));
        }
        sm.refresh();

        // Refresh L2 cache object since it is a copy of this object
        putObjectIntoCache(pc, false, true);
    }

    /**
     * Method to do a refresh of an object.
     * @param pc The Object
     */
    public synchronized void refresh(Object pc)
    {
        assertIsOpen();
        internalRefresh(pc);
    }

    /**
     * Method to do a refresh of an array of objects.
     * @param pcs The Objects
     */
    public synchronized void refreshAll(Object[] pcs)
    {
        refreshAll(Arrays.asList(pcs));
    }

    /**
     * Method to do a refresh of a collection of objects. Throws a
     * JDOUserException if instances could not be refreshed.
     * @param pcs The Objects
     */
    public synchronized void refreshAll(Collection pcs)
    {
        assertIsOpen();
        ArrayList failures = new ArrayList();
        Iterator iter = pcs.iterator();
        while (iter.hasNext())
        {
            try
            {
                internalRefresh(iter.next());
            }
            catch (RuntimeException e)
            {
                failures.add(e);
            }
        }
        if (!failures.isEmpty())
        {
            throw new JDOUserException(LOCALISER.msg("PM.RefreshError"), (Exception[]) failures.toArray(new Exception[failures.size()]));
        }
    }

    /**
     * Method to do a refresh of all objects. Throws a JDOUserException if
     * instances could not be refreshed.
     */
    public synchronized void refreshAll()
    {
        assertIsOpen();
        ArrayList failures = new ArrayList();
        Iterator iter = tx.isActive() ? new ArrayList(txCache).iterator() : new ArrayList(cache.values()).iterator();
        while (iter.hasNext())
        {
            try
            {
                PersistenceCapable pc = (PersistenceCapable)iter.next();
                StateManagerImpl sm = (StateManagerImpl)findStateManager(pc);
                sm.refresh();

                // Refresh L2 cache object since it is a copy of this object
                putObjectIntoCache(pc, false, true);
            }
            catch (RuntimeException e)
            {
                failures.add(e);
            }
        }
        if (!failures.isEmpty())
        {
            throw new JDOUserException(LOCALISER.msg("PM.RefreshError"), (Exception[]) failures.toArray(new Exception[failures.size()]));
        }
    }

    /**
     * Method to do a refresh of objects that failed verification in the
     * exception.
     * @param exc The JDO exception containing the objects that failed
     * @since 1.1
     */
    public synchronized void refreshAll(JDOException exc)
    {
        Object obj = exc.getFailedObject();
        if (obj != null)
        {
            refresh(obj);
        }

        Throwable[] nested_excs = exc.getNestedExceptions();
        if (nested_excs != null)
        {
            for (int i = 0; i < nested_excs.length; i++)
            {
                if (nested_excs[i] instanceof JDOException)
                {
                    refreshAll((JDOException) nested_excs[i]);
                }
            }
        }
    }

    // ------------------------------- Retrieve ------------------------------------------

    /**
     * Method to retrieve an object (internally).
     * @param obj The object
     * @param dfgOnly Whether to use default-fetch-group only
     */
    private void internalRetrieve(Object obj, boolean dfgOnly)
    {
        if (obj == null)
        {
            return;
        }
        PersistenceCapable pc = (PersistenceCapable) obj;
        StateManager sm = findStateManager(pc);
        if (sm == null)
        {
            throw new JDOUserException(LOCALISER.msg("PM.DifferentOwner", pc.jdoGetObjectId()));
        }
        sm.retrieve(dfgOnly);
    }

    /**
     * Method to retrieve an object.
     * @param pc The object
     */
    public synchronized void retrieve(Object pc)
    {
        assertIsOpen();
        internalRetrieve(pc, false);
    }

    /**
     * Method to retrieve an array of objects.
     * @param pcs The objects
     */
    public synchronized void retrieveAll(Object[] pcs)
    {
        retrieveAll(Arrays.asList(pcs), false);
    }

    /**
     * Retrieve field values of instances from the store. This tells the
     * <code>PersistenceManager</code> that the application intends to use the
     * instances, and their field values should be retrieved. The fields in the
     * default fetch group must be retrieved, and the implementation might
     * retrieve more fields than the default fetch group.
     * <P>
     * The <code>PersistenceManager</code> might use policy information about
     * the class to retrieve associated instances.
     * @param pcs the instances
     * @param dfgOnly whether to retrieve only the default-fetch-group fields
     * @since JDO1.0.1
     */
    public void retrieveAll(Object[] pcs, boolean dfgOnly)
    {
        retrieveAll(Arrays.asList(pcs), dfgOnly);
    }

    /**
     * Retrieve field values of instances from the store. This tells the
     * <code>PersistenceManager</code> that the application intends to use the
     * instances, and their field values should be retrieved. The fields in the
     * default fetch group must be retrieved, and the implementation might
     * retrieve more fields than the default fetch group.
     * <P>
     * The <code>PersistenceManager</code> might use policy information about
     * the class to retrieve associated instances.
     * @param pcs the instances
     * @param dfgOnly whether to retrieve only the default-fetch-group fields
     * @since JDO1.0.1
     */
    public void retrieveAll(Collection pcs, boolean dfgOnly)
    {
        assertIsOpen();
        ArrayList failures = new ArrayList();
        Iterator i = pcs.iterator();
        while (i.hasNext())
        {
            try
            {
                internalRetrieve(i.next(), dfgOnly);
            }
            catch (RuntimeException e)
            {
                failures.add(e);
            }
        }
        if (!failures.isEmpty())
        {
            throw new JDOUserException(LOCALISER.msg("PM.RetrieveError"), (Exception[]) failures.toArray(new Exception[failures.size()]));
        }
    }

    /**
     * Method to retrieve a collection of objects. Throws a JDOUserException if
     * instances could not be retrieved.
     * @param pcs The objects
     */
    public synchronized void retrieveAll(Collection pcs)
    {
        retrieveAll(pcs.toArray(), false);
    }

    // ---------------------------------- Make Persistent ---------------------------------------

    /**
     * Method to make an object persistent (internally). Throws a
     * JDOUserException if the object is managed by a different PM.
     * @param obj The object
     * @param preInsertChanges Any changes to make before inserting
     */
    private void internalMakePersistent(Object obj, FieldValues preInsertChanges)
    {
        if (obj == null)
        {
            return;
        }
        // Check the class is persistable
        assertPersistenceCapable(obj);
        PersistenceCapable pc = (PersistenceCapable) obj;
        javax.jdo.PersistenceManager pm = pc.jdoGetPersistenceManager();
        if (JPOXLogger.JDO.isDebugEnabled())
        {
            try
            {
                JPOXLogger.JDO.debug(LOCALISER.msg("PM.MakePersistent", pc));
            }
            catch (Exception e)
            {
                // we catch any exception that can be raised inside the toString
                // method in the pc (PersistenceCapable) instance
                JPOXLogger.JDO.debug(LOCALISER.msg("PM.MakePersistent", pc.getClass().getName()));
            }
        }
        if (pm != null)
        {
            if (!(pm instanceof PersistenceManager) || ((PersistenceManager) pm).getPMHandle() != this.getPMHandle())
            {
                throw new JDOUserException(LOCALISER.msg("PM.DifferentOwner", obj));
            }
        }
        if (pc.jdoIsTransactional() && !pc.jdoIsPersistent())
        {
            StateManager sm = findStateManager(pc);
            if (sm == null)
            {
                throw new JDOUserException(LOCALISER.msg("PM.DifferentOwner", pc.jdoGetObjectId()));
            }
            sm.makePersistentTransactionalTransient();
        }
        // TODO This should be reinstated when attachCopy will correctly persist non-persisted objects.
        // TODO Required by 5.5.8 of JDO 2 spec
        /*else if (pc.jdoIsDetached())
        {
            throw new JDOUserException(LOCALISER.msg("PM.MakePersistentOnDetachedInstances", pc.jdoGetObjectId()));
        }*/
        else if (!pc.jdoIsPersistent())
        {
            if (this.multithreaded)
            {
                synchronized (pc)
                {
                    StateManager sm = findStateManager(pc);
                    if (sm == null)
                    {
                        sm = new StateManagerImpl(this, pc, preInsertChanges);
                        sm.makePersistent();
                    }
                    else
                    {
                        sm.makePersistent();
                    }
                }
            }
            else
            {
                StateManager sm = findStateManager(pc);
                if (sm == null)
                {
                    sm = new StateManagerImpl(this, pc, preInsertChanges);
                    sm.makePersistent();
                }
                else
                {
                    sm.makePersistent();
                }
            }
        }
        else if (pc.jdoIsPersistent() && pc.jdoGetObjectId() == null)
        {
            // Embedded objects have a StateManager yet have no identity.
            // They can be persisted in their own right, so we allow this.
            if (this.multithreaded)
            {
                synchronized (pc)
                {
                    StateManager sm = findStateManager(pc);
                    sm.makePersistent();
                }
            }
            else
            {
                StateManager sm = findStateManager(pc);
                sm.makePersistent();
            }
        }
    }

    /**
     * Method to make persistent an object.
     * @param obj The object
     */
    public synchronized void makePersistent(Object obj)
    {
        assertIsOpen();
        assertActiveTransaction();
        internalMakePersistent(obj, null);
    }

    /**
     * Method to make persistent an object.
     * @param obj The object
     * @param preInsertChanges Any changes to make before the insert.
     */
    public synchronized void makePersistent(Object obj, FieldValues preInsertChanges)
    {
        assertIsOpen();
        assertActiveTransaction();
        internalMakePersistent(obj, preInsertChanges);
    }

    /**
     * Method to make persistent an array of objects.
     * @param pcs The objects
     */
    public synchronized void makePersistentAll(Object[] pcs)
    {
        makePersistentAll(Arrays.asList(pcs));
    }

    /**
     * Method to make persistent a collection of objects. Throws a
     * JDOUserException if objects could not be made persistent.
     * @param pcs The objects
     */
    public synchronized void makePersistentAll(Collection pcs)
    {
        assertIsOpen();
        assertActiveTransaction();
        ArrayList failures = new ArrayList();
        Iterator i = pcs.iterator();
        while (i.hasNext())
        {
            try
            {
                internalMakePersistent(i.next(), null);
            }
            catch (RuntimeException e)
            {
                failures.add(e);
            }
        }
        if (!failures.isEmpty())
        {
            throw new JDOUserException(LOCALISER.msg("PM.MakePersistentError"), (Exception[]) failures.toArray(new Exception[failures.size()]));
        }
    }

    /**
     * Tests whether this object is being inserted.
     * @param pc the object to verify the status
     * @return true if this instance is inserting.
     */
    public boolean isInserting(PersistenceCapable pc)
    {
        StateManager sm = findStateManager(pc);
        if (sm == null)
        {
            return false;
        }
        return sm.isInserting();
    }
    
    // ------------------------------- Delete Persistent ------------------------------------------

    /**
     * Method to delete an object from the datastore (internally).
     * @param obj The object
     */
    private void internalDeletePersistent(Object obj)
    {
        if (obj == null)
        {
            return;
        }
        
        PersistenceCapable pc = (PersistenceCapable) obj;
        if (JPOXLogger.JDO.isDebugEnabled())
        {
            try
            {
                JPOXLogger.JDO.debug(LOCALISER.msg("PM.DeletePersistent", pc));
            }
            catch (Exception e)
            {
                // we catch any exception that can be raised inside the toString
                // method in the pc (PersistenceCapable) instance
            }
        }
        if (!pc.jdoIsPersistent() && pc.jdoIsTransactional())
        {
            throw new JDOUserException(LOCALISER.msg("PM.DeletePersistentOnTransientInstances"));
        }
        if (pc.jdoIsDetached())
        {
            throw new JDOUserException(LOCALISER.msg("PM.DeletePersistentOnDetachedInstances", pc.jdoGetObjectId()));
        }
        StateManager sm = findStateManager(pc);
        if (sm == null)
        {
            throw new JDOUserException(LOCALISER.msg("PM.DifferentOwner", pc.jdoGetObjectId()));
        }
        sm.deletePersistent();
    }

    /**
     * Method to delete an object from the datastore.
     * @param obj The object
     */
    public synchronized void deletePersistent(Object obj)
    {
        assertIsOpen();
        assertActiveTransaction();
        internalDeletePersistent(obj);
    }

    /**
     * Method to delete an array of objects from the datastore.
     * @param pcs The objects
     */
    public synchronized void deletePersistentAll(Object[] pcs)
    {
        deletePersistentAll(Arrays.asList(pcs));
    }

    /**
     * Method to delete a collection of objects from the datastore. Throws a
     * JDOUserException if objects could not be deleted.
     * @param pcs The objects
     */
    public synchronized void deletePersistentAll(Collection pcs)
    {
        assertIsOpen();
        assertActiveTransaction();
        ArrayList failures = new ArrayList();
        Iterator i = pcs.iterator();
        while (i.hasNext())
        {
            try
            {
                internalDeletePersistent(i.next());
            }
            catch (RuntimeException e)
            {
                failures.add(e);
            }
        }
        if (!failures.isEmpty())
        {
            throw new JDOUserException(LOCALISER.msg("PM.DeletePersistentError"), (Exception[]) failures.toArray(new Exception[failures.size()]));
        }
    }

    // -------------------------------- Make Transient -----------------------------------------

    /**
     * Method to make transient an object (internally).
     * @param obj The object
     */
    private void internalMakeTransient(Object obj)
    {
        if (obj == null)
        {
            return;
        }
        assertPersistenceCapable(obj);
        PersistenceCapable pc = (PersistenceCapable) obj;
        if (JPOXLogger.JDO.isDebugEnabled())
        {
            try
            {
                JPOXLogger.JDO.debug(LOCALISER.msg("PM.MakeTransient", pc));
            }
            catch (Exception e)
            {
                // we catch any exception that can be raised inside the toString
                // method in the pc (PersistenceCapable) instance
            }
        }
        if (pc.jdoIsPersistent())
        {
            StateManager sm = findStateManager(pc);
            sm.makeTransient();
        }
    }

    /**
     * Method to make transient an object.
     * @param pc The object
     */
    public synchronized void makeTransient(Object pc)
    {
        assertIsOpen();
        internalMakeTransient(pc);
    }

    /**
     * Method to make transient an array of objects.
     * @param pcs The objects
     */
    public synchronized void makeTransientAll(Object[] pcs)
    {
        makeTransientAll(Arrays.asList(pcs));
    }

    /**
     * Method to make transient a collection of objects. Throws a
     * JDOUserException if objects could not be made transient.
     * 
     * @param pcs
     *            The objects
     */
    public synchronized void makeTransientAll(Collection pcs)
    {
        assertIsOpen();
        ArrayList failures = new ArrayList();
        Iterator i = pcs.iterator();
        while (i.hasNext())
        {
            try
            {
                internalMakeTransient(i.next());
            }
            catch (RuntimeException e)
            {
                failures.add(e);
            }
        }
        if (!failures.isEmpty())
        {
            throw new JDOUserException(LOCALISER.msg("PM.MakeTransientError"), (Exception[]) failures.toArray(new Exception[failures.size()]));
        }
    }

    // ----------------------------------- Make Transactional --------------------------------------

    /**
     * Method to make transactional an object (internally).
     * @param obj The object
     */
    private void internalMakeTransactional(Object obj)
    {
        if (obj == null)
        {
            return;
        }
        assertPersistenceCapable(obj);
        PersistenceCapable pc = (PersistenceCapable) obj;
        if (pc.jdoIsPersistent())
        {
            assertActiveTransaction();
        }
        StateManager sm = findStateManager(pc);
        if (sm == null)
        {
            sm = new StateManagerImpl(this, pc);
        }
        sm.makeTransactional();
    }

    /**
     * Method to make transactional an object.
     * @param pc The object
     */
    public synchronized void makeTransactional(Object pc)
    {
        assertIsOpen();
        internalMakeTransactional(pc);
    }

    /**
     * Method to make transactional an array of objects.
     * @param pcs The objects
     */
    public synchronized void makeTransactionalAll(Object[] pcs)
    {
        makeTransactionalAll(Arrays.asList(pcs));
    }

    /**
     * Method to make transactional a collection of objects. Throws a
     * JDOUserException if objects could not be made transactional.
     * @param pcs The objects
     */
    public synchronized void makeTransactionalAll(Collection pcs)
    {
        assertIsOpen();
        assertActiveTransaction();
        ArrayList failures = new ArrayList();
        Iterator i = pcs.iterator();
        while (i.hasNext())
        {
            try
            {
                internalMakeTransactional(i.next());
            }
            catch (RuntimeException e)
            {
                failures.add(e);
            }
        }
        if (!failures.isEmpty())
        {
            throw new JDOUserException(LOCALISER.msg("PM.MakeTransactionalError"), (Exception[]) failures.toArray(new Exception[failures.size()]));
        }
    }

    // ------------------------------ Make NonTransactional -------------------------------------------

    /**
     * Method to make nontransactional an object (internally).
     * @param obj The object
     */
    private void internalMakeNontransactional(Object obj)
    {
        if (obj == null)
        {
            return;
        }
        PersistenceCapable pc = (PersistenceCapable) obj;
        if (!pc.jdoIsPersistent() && pc.jdoIsTransactional() && pc.jdoIsDirty())
        {
            throw new JDOUserException(LOCALISER.msg("PM.MakeNonTransactionalOnTransientDirtyInstances"));
        }
        StateManager sm = findStateManager(pc);
        sm.makeNontransactional();
    }

    /**
     * Method to make nontransactional an object.
     * @param pc The object
     */
    public synchronized void makeNontransactional(Object pc)
    {
        assertIsOpen();
        // for non transactional read, tx might be not active
        // TODO add verification if is non transactional read = true
        if (tx.isActive())
        {
            assertActiveTransaction();
        }
        // if !transactional and !persistent
        if (!((PersistenceCapable) pc).jdoIsTransactional() && !((PersistenceCapable) pc).jdoIsPersistent())
        {
            throw new JDOUserException(LOCALISER.msg("PM.MakeNontransactionalOnNontransactionalTransientInstances"));
        }
        // if !transactional and persistent, do nothing
        if (!((PersistenceCapable) pc).jdoIsTransactional() && ((PersistenceCapable) pc).jdoIsPersistent())
        {
            return;
        }
        internalMakeNontransactional(pc);
    }

    /**
     * Method to make nontransactional an array of objects.
     * @param pcs The objects.
     */
    public synchronized void makeNontransactionalAll(Object[] pcs)
    {
        makeNontransactionalAll(Arrays.asList(pcs));
    }

    /**
     * Method to make nontransactional a collection of objects.
     * @param pcs The objects.
     */
    public synchronized void makeNontransactionalAll(Collection pcs)
    {
        assertIsOpen();
        assertActiveTransaction();
        ArrayList failures = new ArrayList();
        Iterator i = pcs.iterator();
        while (i.hasNext())
        {
            try
            {
                internalMakeNontransactional(i.next());
            }
            catch (RuntimeException e)
            {
                failures.add(e);
            }
        }
        if (!failures.isEmpty())
        {
            throw new JDOUserException(LOCALISER.msg("PM.MakeNonTransactionalError"), (Exception[]) failures.toArray(new Exception[failures.size()]));
        }
    }

    // ------------------------- Attach/Detach instances -----------------------

    /**
     * Method to detach a persistent object.
     * If the object is of class that is not detachable, a ClassNotDetachableException
     * will be thrown
     * @param pc The object
     * @return The detached object
     */
    public synchronized Object detachCopy(Object pc)
    {
        assertIsOpen();
        assertPersistenceCapable(pc);
        assertDetachable(pc);
        StateManager sm = findStateManager((PersistenceCapable) pc);
        if (sm == null)
        {
            throw new JDOUserException(LOCALISER.msg("PM.DifferentOwner", ((PersistenceCapable) pc).jdoGetObjectId()));
        }
        return sm.detachCopy();
    }

    /**
     * Method to detach an array of objects.
     * If any of the objects are of class that is not detachable, a ClassNotDetachableException
     * will be thrown
     * @param pcs The objects
     */
    public synchronized Object[] detachCopyAll(Object[] pcs)
    {
        Collection copies = detachCopyAll(Arrays.asList(pcs));
        return copies.toArray();
    }

    /**
     * Method to detach a collection of objects.
     * If any of the objects are of class that is not detachable, a ClassNotDetachableException
     * will be thrown
     * @param pcs The objects
     */
    public synchronized Collection detachCopyAll(Collection pcs)
    {
        // Check for detachability before starting
        List notDetachableExceptions = null;
        for (Iterator it = pcs.iterator(); it.hasNext();)
        {
            try
            {
                assertDetachable(it.next());
            }
            catch (ClassNotDetachableException cnde)
            {
                // Gather any non-detachable exceptions
                if (notDetachableExceptions == null)
                {
                    notDetachableExceptions = new ArrayList();
                }
                notDetachableExceptions.add(cnde);
            }
        }

        // Throw an exception for any non-detachable objects
        if (notDetachableExceptions != null && notDetachableExceptions.size() > 1)
        {
            ClassNotDetachableException[] exceptions = new ClassNotDetachableException[notDetachableExceptions.size()];
            Iterator iter = notDetachableExceptions.iterator();
            int exceptionNo = 0;
            while (iter.hasNext())
            {
                exceptions[exceptionNo++] = (ClassNotDetachableException)iter.next();
            }
            throw new ClassNotDetachableException(exceptions);
        }
        else if (notDetachableExceptions != null && notDetachableExceptions.size() == 1)
        {
            throw (ClassNotDetachableException)notDetachableExceptions.get(0);
        }

        // Detach the objects
        List detacheds = new ArrayList();
        for (Iterator it = pcs.iterator(); it.hasNext();)
        {
            detacheds.add(detachCopy(it.next()));
        }

        return detacheds;
    }

    /**
     * Method to attach a persistent object.
     * If the object is of class that is not detachable, a ClassNotDetachableException
     * will be thrown
     * @param pc The object
     * @param makeTransactional Whether to make it transactional
     * @return THe attached object
     */
    public synchronized Object attachCopy(Object pc, boolean makeTransactional)
    {
        assertIsOpen();
        assertPersistenceCapable(pc);
        assertDetachable(pc);

        // Check if object has been persisted before. If not, persist it
        Object id = ((PersistenceCapable) pc).jdoGetObjectId();
        if (id == null)
        {
            makePersistent(pc);
            // Should really return a copy ?
            return pc;
        }

        // Check if is already has StateManager - hence can't be attached
        if (((PersistenceCapable) pc).jdoIsPersistent())
        {
            return (PersistenceCapable) pc;
        }

        // Create a copy and attach it
        PersistenceCapable pcTarget = (PersistenceCapable) getObjectById(((PersistenceCapable) pc).jdoGetObjectId(), false);
        findStateManager(pcTarget).attachCopy((PersistenceCapable) pc, makeTransactional);

        return pcTarget;
    }

    /**
     * Method to attach an array of objects.
     * If any of the objects are of class that is not detachable, a ClassNotDetachableException
     * will be thrown
     * @param pcs The objects
     */
    public synchronized Object[] attachCopyAll(Object[] pcs, boolean makeTransactional)
    {
        Collection copies = attachCopyAll(Arrays.asList(pcs), makeTransactional);
        return copies.toArray();
    }

    /**
     * Method to attach a collection of objects.
     * If any of the objects are of class that is not detachable, a ClassNotDetachableException
     * will be thrown
     * @param pcs The objects
     */
    public synchronized Collection attachCopyAll(Collection pcs, boolean makeTransactional)
    {
        // Check for detachability before starting
        List notDetachableExceptions = null;
        for (Iterator it = pcs.iterator(); it.hasNext();)
        {
            try
            {
                assertDetachable(it.next());
            }
            catch (ClassNotDetachableException cnde)
            {
                // Gather any non-detachable exceptions
                if (notDetachableExceptions == null)
                {
                    notDetachableExceptions = new ArrayList();
                }
                notDetachableExceptions.add(cnde);
            }
        }

        // Throw an exception for any non-detachable objects
        if (notDetachableExceptions != null && notDetachableExceptions.size() > 1)
        {
            ClassNotDetachableException[] exceptions = new ClassNotDetachableException[notDetachableExceptions.size()];
            Iterator iter = notDetachableExceptions.iterator();
            int exceptionNo = 0;
            while (iter.hasNext())
            {
                exceptions[exceptionNo++] = (ClassNotDetachableException)iter.next();
            }
            throw new ClassNotDetachableException(exceptions);
        }
        else if (notDetachableExceptions != null && notDetachableExceptions.size() == 1)
        {
            throw (ClassNotDetachableException)notDetachableExceptions.get(0);
        }

        // Attach the objects
        List attacheds = new ArrayList();
        for (Iterator it = pcs.iterator(); it.hasNext();)
        {
            attacheds.add(attachCopy(it.next(), makeTransactional));
        }
        return attacheds;
    }

    // --------------------------------- Queries ----------------------------------------

    /**
     * Construct an empty query instance.
     * @return The query
     */
    public synchronized Query newQuery()
    {
        return newQuery("javax.jdo.query.JDOQL", null);
    }

    /**
     * Construct a query instance from another query. The parameter might be a
     * serialized/restored Query instance from the same JDO vendor but a
     * different execution environment, or the parameter might be currently
     * bound to a PersistenceManager from the same JDO vendor. Any of the
     * elements Class, Filter, IgnoreCache flag, Import declarations, Variable
     * declarations, Parameter declarations, and Ordering from the parameter
     * Query are copied to the new Query instance, but a candidate Collection or
     * Extent element is discarded.
     * @param obj The object to use in the query
     * @return The query
     */
    public synchronized Query newQuery(Object obj)
    {
        return newQuery("javax.jdo.query.JDOQL", obj);
    }

    /**
     * Construct a query instance using the specified Single-String query.
     * @param query The single-string query
     * @return The Query
     */
    public synchronized Query newQuery(String query)
    {
        return newQuery("javax.jdo.query.JDOQL", query);
    }

    /**
     * Construct a query instance using the specified language and the specified
     * query. The query instance will be of a class defined by the query
     * language.
     * @param language The language parameter for the JDO Query language. This
     * is by default "javax.jdo.query.JDOQL", but in JDO 2.0 can also be
     * "javax.jdo.query.SQL", or vendor provided languages.
     * @param query The query object
     * @return The query
     */
    public synchronized Query newQuery(String language, Object query)
    {
        // This is the only query construction that performs a create
        assertIsOpen();

        Query q = null;
        PersistenceManager pm = getPMHandle();
        if (language == null)
        {
            language = "javax.jdo.query.JDOQL";
        }
        
        // Check that our store supports the language
        if (!srm.supportsQueryLanguage(language))
        {
            throw new JDOUserException("PM.Query.LanguageNotSupportedByStore");
        }

        if (language.equals("javax.jdo.query.JDOQL"))
        {
            if (query != null && !(query instanceof JDOQLQuery) && !(query instanceof String))
            {
                throw new JDOUserException(LOCALISER.msg("PM.Query.InvalidArgument", query, JDOQLQuery.class.getName()));
            }

            if (query == null)
            {
                q = new JDOQLQuery(pm);
            }
            // Build from existing query
            else if (query instanceof JDOQLQuery)
            {
                q = new JDOQLQuery(pm, (JDOQLQuery)query);
            }
            // Single query string (JDO 2)
            else if (query instanceof String)
            {
                q = new JDOQLQuery(pm, (String)query);
            }
        }
        else if (language.equals("javax.jdo.query.JPOXSQL"))
        {
            if (query == null || !(query instanceof String))
            {
                throw new JDOUserException(LOCALISER.msg("PM.Query.InvalidArgumentString", query));
            }
            q = new JPOXSQLQuery(pm, (String)query);
        }
        else if (language.equals("javax.jdo.query.SQL"))
        {
            if (query == null || !(query instanceof String))
            {
                throw new JDOUserException(LOCALISER.msg("PM.Query.InvalidArgumentString", query));
            }
            q = new SQLQuery(pm, (String)query);
        }
        else
        {
            throw new JDOUserException(LOCALISER.msg("PM.Query.LanguageUnknown", language));
        }
        return q;
    }

    /**
     * Construct a query instance with the candidate class specified.
     * @param cls The class to query
     * @return The query
     */
    public synchronized Query newQuery(Class cls)
    {
        Query query = newQuery();
        query.setClass(cls);
        return query;
    }

    /**
     * Construct a query instance with the candidate Extent specified; the
     * candidate class is taken from the Extent.
     * @param cln The extent to query
     * @return The query
     */
    public synchronized Query newQuery(Extent cln)
    {
        Query query = newQuery();
        query.setClass(cln.getCandidateClass());
        query.setCandidates(cln);
        return query;
    }

    /**
     * Construct a query instance with the candidate class and candidate
     * Collection specified.
     * @param cls The class to query
     * @param cln The collection
     * @return The query
     */
    public synchronized Query newQuery(Class cls, Collection cln)
    {
        Query query = newQuery();
        query.setClass(cls);
        query.setCandidates(cln);
        return query;
    }

    /**
     * Construct a query instance with the candidate class and filter specified.
     * @param cls The class to query
     * @param filter A filter to apply
     * @return The query
     */
    public synchronized Query newQuery(Class cls, String filter)
    {
        Query query = newQuery();
        query.setClass(cls);
        query.setFilter(filter);
        return query;
    }

    /**
     * Construct a query instance with the candidate class, the candidate
     * Collection, and filter specified.
     * @param cls The class to query
     * @param cln A collection
     * @param filter A filter to apply
     * @return The query
     */
    public synchronized Query newQuery(Class cls, Collection cln, String filter)
    {
        Query query = newQuery();
        query.setClass(cls);
        query.setCandidates(cln);
        query.setFilter(filter);
        return query;
    }

    /**
     * Construct a query instance with the candidate Extent and filter
     * specified. The candidate class is taken from the Extent.
     * @param cln The extent to query
     * @param filter A filter to apply
     * @return The query
     */
    public synchronized Query newQuery(Extent cln, String filter)
    {
        Query query = newQuery();
        query.setClass(cln.getCandidateClass());
        query.setCandidates(cln);
        query.setFilter(filter);
        return query;
    }

    /**
     * Construct a query instance with the candidate class and the query name.
     * @param cls The class to query
     * @param queryName Name of the query.
     * @return The query
     */
    public synchronized Query newNamedQuery(Class cls, String queryName)
    {
        assertIsOpen();

        // Throw exception on incomplete input
        if (queryName == null)
        {
            throw new JDOUserException(LOCALISER.msg("PM.Query.NamedQueryNotFound", queryName, cls));
        }

        // Find the Query for the specified class
        QueryMetaData qmd = null;
        if (cls != null)
        {
            qmd = getMetaDataManager().getMetaDataForQuery(cls, queryName);
        }
        else
        {
            qmd = getMetaDataManager().getMetaDataForQuery(null, queryName);
        }
        if (qmd == null)
        {
            throw new JDOUserException(LOCALISER.msg("PM.Query.NamedQueryNotFound", queryName, cls));
        }

        // Create the Query
        Query query = newQuery(qmd.getLanguage().toString(), qmd.getQuery());
        if (cls != null)
        {
            query.setClass(cls);
        }
        if (qmd.getLanguage() == QueryLanguage.JPOXSQL)
        {
            // Apply any imports specified
            if (qmd.hasExtension("imports"))
            {
                query.declareImports(qmd.getValueForExtension("imports"));
            }
        }

        return query;
    }

    // ------------------------------- Extents ------------------------------------------

    /**
     * Extents are collections of datastore objects managed by the datastore,
     * not by explicit user operations on collections. Extent capability is a
     * boolean property of classes that are persistence capable. If an instance
     * of a class that has a managed extent is made persistent via reachability,
     * the instance is put into the extent implicitly.
     * @param pcClass The class to query
     * @param subclasses Whether to include subclasses in the query.
     * @return returns an Extent that contains all of the instances in the
     * parameter class, and if the subclasses flag is true, all of the instances
     * of the parameter class and its subclasses.
     */
    public synchronized Extent getExtent(Class pcClass, boolean subclasses)
    {
        assertIsOpen();
        assertPersistenceCapableClass(pcClass);
        return srm.getExtent(getPMHandle(), pcClass, subclasses);
    }
    
    /**
     * Extents are collections of datastore objects managed by the datastore,
     * not by explicit user operations on collections. Extent capability is a
     * boolean property of classes that are persistence capable. If an instance
     * of a class that has a managed extent is made persistent via reachability,
     * the instance is put into the extent implicitly.
     * @param pcClass The class to query
     * @return returns an Extent that contains all of the instances in the
     * parameter class, and all of the instances of the parameter class and its
     * subclasses.
     * @since 1.1
     */
    public synchronized Extent getExtent(Class pcClass)
    {
        return getExtent(pcClass, true);
    }

    // ----------------------------- New Instances ----------------------------------

    /**
     * Accessor for a new instance of an object.
     * @param type Type of object (the class).
     * @param parameterTypes Classes of params for the constructor
     * @param parameters The parameters for the constructor
     * @return The object
     */
    private static Object newInstance(Class type, Class[] parameterTypes, Object[] parameters)
    {
        Object obj;
        try
        {
            Constructor ctor = type.getConstructor(parameterTypes);
            obj = ctor.newInstance(parameters);
        }
        catch (NoSuchMethodException e)
        {
            throw new JDOFatalInternalException(LOCALISER.msg("PM.NewInstance.ClassMissingConstructorError", type.getName()), new Exception[]{e});
        }
        catch (IllegalAccessException e)
        {
            throw new JDOFatalInternalException(LOCALISER.msg("PM.NewInstance.ClassAccessError", type.getName()), new Exception[]{e});
        }
        catch (InstantiationException e)
        {
            throw new JDOFatalInternalException(LOCALISER.msg("PM.NewInstance.ClassInstantiationError", type.getName()), new Exception[]{e});
        }
        catch (InvocationTargetException e)
        {
            Throwable t = e.getTargetException();
            if (t instanceof RuntimeException)
            {
                throw (RuntimeException) t;
            }
            else if (t instanceof Error)
            {
                throw (Error) t;
            }
            else
            {
                throw new JDOFatalInternalException(LOCALISER.msg("PM.NewInstance.ConstructorError", type.getName(), t));
            }
        }
        return obj;
    }

    /**
     * Accessor for a new Second-Class Object instance.
     * @param type The class of the object
     * @param owner The owner Object of the SCO
     * @param fmd The Field MetaData for the related field.
     * @return The Second-Class Object
     */
    public SCO newSCOInstance(Class type, Object owner, FieldMetaData fmd)
    {
        assertIsOpen();
        if (!fmd.getType().isAssignableFrom(type))
        {
            throw new JDOUserException(LOCALISER.msg("PM.NewSCOInstance.FieldTypeInvalid", type.getName(), fmd.getName(), fmd.getType()));
        }
        Class newType = TypeManager.getTypeManager().getSecondClassWrapper(type.getName());
        if (newType == null)
        {
            throw new JDOUserException(LOCALISER.msg("PM.NewSCOInstance.ClassInvalid", type.getName()));
        }
        SCO obj = (SCO) newInstance(newType, new Class[]{Object.class, String.class}, new Object[]{owner, fmd.getName()});
        return obj;
    }

    /**
     * Accessor for a new Second-Class Object instance.
     * @param type The class of the object
     * @param owner The owner Object of the SCO
     * @param fieldName Field name for the related field.
     * @return The Second-Class Object
     */
    public Object newSCOInstance(Class type, Object owner, String fieldName)
    {
        if (Collection.class.isAssignableFrom(type))
        {
            return newCollectionInstance(type, owner, fieldName, Object.class, true, null, null, null);
        }
        if (Map.class.isAssignableFrom(type))
        {
            return newMapInstance(type, owner, fieldName, Object.class, Object.class, true, null, null, null);
        }
        StateManager sm = findStateManager((PersistenceCapable) owner);
        ClassMetaData cmd = getMetaDataManager().getMetaDataForClass(owner.getClass());
        int fieldNumber = cmd.getFieldNumberAbsolute(fieldName);
        FieldMetaData fmd = cmd.getManagedFieldAbsolute(fieldNumber);
        Object obj = newSCOInstance(type, owner, fmd);
        sm.replaceField(fieldNumber, obj);
        sm.makeDirty(fieldNumber);
        return obj;
    }

    /**
     * Accessor for a new collection instance.
     * @param type Class name of the collection.
     * @param sm StateManager for the owning object.
     * @param fmd Field MetaData for the collection field
     * @param elementType Class name of element
     * @param allowNulls Whether to allow nulls
     * @return The collection.
     */
    private Collection newCollectionInstance(Class type, StateManager sm, FieldMetaData fmd, Class elementType, boolean allowNulls)
    {
        assertIsOpen();
        if (!fmd.getType().isAssignableFrom(type))
        {
            throw new JDOUserException(LOCALISER.msg("PM.NewSCOInstance.FieldTypeInvalid", type.getName(), fmd.getName(), fmd.getTypeName()));
        }

        Class newType = TypeManager.getTypeManager().getSecondClassWrapper(type.getName());
        if (newType == null)
        {
            throw new JDOUserException(LOCALISER.msg("PM.NewSCOInstance.ClassInvalid", type.getName()));
        }

        Collection obj = (Collection) newInstance(newType, new Class[]{StateManager.class, String.class, Class.class, boolean.class}, 
            new Object[]{sm, fmd.getName(), elementType, ClassUtils.booleanValueOf(allowNulls)});
        return obj;
    }

    /**
     * Accessor for a new collection instance.
     * @param type Class name of the collection.
     * @param owner The owning object
     * @param fieldName Field name for the collection field
     * @param elementType Class name of element
     * @param allowNulls Whether to allow nulls
     * @param initialCapacity Initial capacity of the collection
     * @param loadFactor Load factor ?
     * @param initialContents Initial contents of the collection.
     * @return The collection.
     */
    public Collection newCollectionInstance(Class type, Object owner, String fieldName, Class elementType, boolean allowNulls, Integer initialCapacity,
            Float loadFactor, Collection initialContents)
    {
        StateManager sm = findStateManager((PersistenceCapable) owner);
        ClassMetaData cmd = getMetaDataManager().getMetaDataForClass(owner.getClass());
        int fieldNumber = cmd.getFieldNumberAbsolute(fieldName);
        FieldMetaData fmd = cmd.getManagedFieldAbsolute(fieldNumber);
        Collection obj = newCollectionInstance(type, sm, fmd, elementType, allowNulls);
        if (initialContents != null)
        {
            obj.clear();
            obj.addAll(initialContents);
        }
        sm.replaceField(fieldNumber, obj);
        sm.makeDirty(fieldNumber);
        return obj;
    }

    /**
     * Accessor for a new map instance.
     * @param type Class name of the map
     * @param sm StateManager for the owning object.
     * @param fmd Field MetaData for the map field
     * @param keyType Class name of keys in the Map
     * @param valueType Class name of values in the Map
     * @param allowNulls Whether to allow nulls
     * @return The Map
     */
    private Map newMapInstance(Class type, StateManager sm, FieldMetaData fmd, Class keyType, Class valueType, boolean allowNulls)
    {
        assertIsOpen();
        if (!fmd.getType().isAssignableFrom(type))
        {
            throw new JDOUserException(LOCALISER.msg("PM.NewSCOInstance.FieldTypeInvalid", type.getName(), fmd.getName(), fmd.getTypeName()));
        }
        Class newType = TypeManager.getTypeManager().getSecondClassWrapper(type.getName());
        if (newType == null)
        {
            throw new JDOUserException(LOCALISER.msg("PM.NewSCOInstance.ClassInvalid", type.getName()));
        }
        Map obj = (Map) newInstance(newType, new Class[]{StateManager.class, String.class, Class.class, Class.class, boolean.class}, new Object[]{sm,
                fmd.getName(), keyType, valueType, ClassUtils.booleanValueOf(allowNulls)});
        return obj;
    }

    /**
     * Accessor for a new map instance.
     * @param type Class name of the map
     * @param owner The owning object
     * @param fieldName Field name for the map field
     * @param keyType Class name of keys in the Map
     * @param valueType Class name of values in the Map
     * @param allowNulls Whether to allow nulls
     * @param initialCapacity Initial capacity of the Map
     * @param loadFactor Load factor ?
     * @param initialContents Initial contents of the Map
     * @return The Map
     */
    public Map newMapInstance(Class type, Object owner, String fieldName, Class keyType, Class valueType, boolean allowNulls, Integer initialCapacity,
            Float loadFactor, Map initialContents)
    {
        StateManager sm = findStateManager((PersistenceCapable) owner);
        ClassMetaData cmd = getMetaDataManager().getMetaDataForClass(owner.getClass());
        int fieldNumber = cmd.getFieldNumberAbsolute(fieldName);
        FieldMetaData fmd = cmd.getManagedFieldAbsolute(fieldNumber);
        Map obj = newMapInstance(type, sm, fmd, keyType, valueType, allowNulls);
        if (initialContents != null)
        {
            obj.clear();
            obj.putAll(initialContents);
        }
        sm.replaceField(fieldNumber, obj);
        sm.makeDirty(fieldNumber);
        return obj;
    }

    /**
     * Method to generate an instance of an interface or abstract class.
     * @param pc The class of the interface or abstract class defined in MetaData
     * @return The instance
     */
    public Object newInstance(Class pc)
    {
        assertIsOpen();
        // TODO Implement newInstance(Class)
        throw new JDOUserException("newInstance(Class pc) not yet supported");
    }

    /**
     * Accessor for the StateManager of an object given the object id.
     * @param id Id of the object.
     * @return The StateManager
     */
    public synchronized StateManager getStateManagerById(Object id)
    {
        assertIsOpen();
        PersistenceCapable pc = getObjectFromCache(id);
        return findStateManager(pc);
    }

    /**
     * Accessor for the StateManager of an object given the object AID.
     * @param pcClass <description>
     * @param fv <description>
     * @param ignoreCache true if it must ignore the cache
     * @param checkInheritance Whether look to the database to determine which
     * class this object is. This parameter is a hint. Set false, if it's
     * already determined the correct pcClass for this pc "object" in a certain
     * level in the hierarchy. Set to true and it will look to the database.
     * @return Object
     */
    public synchronized Object getObjectByAID(Class pcClass, FieldValues fv, boolean ignoreCache, boolean checkInheritance)
    {
        assertIsOpen();
        PersistenceCapable pc;
        StateManager sm = new StateManagerImpl(this, pcClass, fv, checkInheritance);
        if (!ignoreCache)
        {
            Object oid = ((StateManagerImpl) sm).getInternalObjectId();
            pc = getObjectFromCache(oid);
            if (pc != null)
            {
                return pc;
            }
        }

        pc = sm.getObject();
        putObjectIntoCache(pc, true, true);

        return pc;
    }

    /**
     * Accessor for an object given the object id.
     * @param id Id of the object.
     * @param fv Field values for the object
     * @param clazz Class name of the object
     * @param ignoreCache true if it must ignore the cache
     * @return The Object
     */
    public synchronized Object getObjectById(Object id, FieldValues fv, Class clazz, boolean ignoreCache)
    {
        assertIsOpen();

        PersistenceCapable pc = null;
        if (!ignoreCache)
        {
            pc = getObjectFromCache(id);
        }

        if (pc == null)
        {
            StateManager sm = new StateManagerImpl(this, clazz, id, fv);
            //TODO verify it

            pc = sm.getObject();
            putObjectIntoCache(pc, true, true);
        }

        return pc;
    }

    /**
     * Accessor for an object given the object id.
     * @param id Id of the object.
     * @return The Object
     * @since 1.1
     */
    public synchronized Object getObjectById(Object id)
    {
        return getObjectById(id, true);
    }

    /**
     * Accessor for an object given the object id.
     * @param id Id of the object.
     * @param validate Whether to validate the object state
     * @return The Object
     */
    public synchronized Object getObjectById(Object id, boolean validate)
    {
        assertIsOpen();
        if (id == null)
        {
            throw new JDOUserException(LOCALISER.msg("PM.GetObjectByIdNullId"));
        }

        // Use inheritance checks whatever the identity type. Really we should be able to
        // assume that SingleFieldIdentity and datastore identity cases are of the class defined
        // by the identity itself since it contains the class name.
        return getObjectById(id, validate, true);
    }

    /**
     * Accessor for an object given the object id.
     * @param id Id of the object.
     * @param validate Whether to validate the object state
     * @param checkInheritance Whether look to the database to determine which
     * class this object is. This parameter is a hint. Set false, if it's
     * already determined the correct pcClass for this pc "object" in a certain
     * level in the hierarchy. Set to true and it will look to the database.
     * @return The Object with this id
     */
    public synchronized Object getObjectById(Object id, boolean validate, boolean checkInheritance)
    {
        assertIsOpen();
        if (id == null)
        {
            throw new JDOUserException(LOCALISER.msg("PM.GetObjectByIdNullId"));
        }

        // try to find object in cache
        PersistenceCapable pc = getObjectFromCache(id);
        StateManagerImpl sm = null;
        if (pc == null)
        {
            String className;
            if (checkInheritance)
            {
	            // if not found in cache, verify if correct class inheritance level
	            // is set in Datastore identity
	            className = srm.getClassNameForObjectID(this, id);

	            // className is null when id class exists, and object has been
	            // validated and doesn't exist.
	            if (className == null)
	            {
	                throw new JDOObjectNotFoundException(LOCALISER.msg("PM.ObjectDoesntExist"), id);
	            }
	            if (id instanceof OID)
	            {
	                ((OID) id).setPcClass(className);

	                // try again to read object from cache
	                pc = getObjectFromCache(id);
	            }
            }
            else if (id instanceof OID)
            {
                className = ((OID)id).getPcClass();
            }
            else if (id instanceof SingleFieldIdentity)
            {
                className = ((SingleFieldIdentity)id).getTargetClassName();
            }
            else
            {
                // Should never come here since the only non-inheritance cases will be
                // OID or SingleFieldIdentity
                return null;
            }

            // not found in cache, will get a Hollow instance
            if (pc == null)
            {
                try
                {
                    Class pcClass = clr.classForName(className, (id instanceof OID) ? null : id.getClass().getClassLoader());
                    sm = new StateManagerImpl(this, pcClass, id);
                    pc = sm.getObject();

                    putObjectIntoCache(pc, true, true);
                }
                catch (ClassNotFoundException e)
                {
                    JPOXLogger.JDO.warn(LOCALISER.msg("PM.GetObjectByIdClassNotFound", id));
                    throw new JDOUserException(LOCALISER.msg("PM.GetObjectByIdClassNotFound", id), e);
                }
            }
        }

        if (validate)
        {
            if (sm == null)
            {
                sm = (StateManagerImpl)findStateManager(pc);
            }
            sm.validate();
        }

        return pc;
    }

    /**
     * Accessor for the objects given the object ids.
     * @param oids Ids of the objects.
     * @param validate Whether to validate the object state
     * @return The Objects with these ids (in the same order)
     * @since 1.1
     */
    public Collection getObjectsById(Collection oids, boolean validate)
    {
        assertIsOpen();
        if (oids == null || oids.size() == 0)
        {
            throw new JDOUserException(LOCALISER.msg("PM.GetObjectsByIdNullId"));
        }

        Collection objects = new HashSet(oids.size());
        Iterator iter = oids.iterator();
        while (iter.hasNext())
        {
            Object oid = iter.next();
            // Just use getObjectById to get the object
            objects.add(getObjectById(oid, validate));
        }
        return objects;
    }

    /**
     * Accessor for the objects given the object ids.
     * @param oids Ids of the objects.
     * @param validate Whether to validate the object state
     * @return The Objects with these ids (in the same order)
     * @since 1.1
     */
    public Object[] getObjectsById(Object[] oids, boolean validate)
    {
        assertIsOpen();
        if (oids == null)
        {
            throw new JDOUserException(LOCALISER.msg("PM.GetObjectsByIdNullId"));
        }
        Object[] objects = new Object[oids.length];
        for (int i = 0; i < oids.length; i++)
        {
            // Just use getObjectById to get the object
            objects[i] = getObjectById(oids[i], validate);
        }
        return objects;
    }

    /**
     * Accessor for the objects given the object ids, validating the objects.
     * @param oids Ids of the objects.
     * @return The Objects with these ids (in the same order)
     * @since 1.1
     */
    public Collection getObjectsById(Collection oids)
    {
        return getObjectsById(oids, true);
    }

    /**
     * Accessor for the objects given the object ids, validating the objects.
     * @param oids Ids of the objects.
     * @return The Objects with these ids (in the same order)
     * @since 1.1
     */
    public Object[] getObjectsById(Object[] oids)
    {
        return getObjectsById(oids, true);
    }

    /**
     * Convenience method that exactly matches the behavior of calling 
     * pm.getObjectById (pm.newObjectIdInstance (cls, key), true).
     * @param cls Class of the PersistenceCapable
     * @param key Value of the key field for SingleFieldIdentity
     * @return The object for this id.
     * @since 1.1
     */
    public Object getObjectById (Class cls, Object key)
    {
        return getObjectById(newObjectIdInstance (cls, key), true);
    }

    /**
     * Accessor for an object given the object id.
     * @param id Id of the object.
     * @param fv Field values for the object
     * @return The Object
     */
    public synchronized Object getObjectById(Object id, FieldValues fv)
    {
        assertIsOpen();

        PersistenceCapable pc = getObjectFromCache(id);
        if (pc == null)
        {
            String className = srm.getClassNameForObjectID(this, id);
            // className is null when id class exists, and object has been
            // validated and doesn't exist.
            if (className == null)
            {
                throw new JDOObjectNotFoundException(LOCALISER.msg("PM.ObjectDoesntExist"), id);
            }

            if (id instanceof OID)
            {
                ((OID) id).setPcClass(className);

                // try again to read object from cache
                pc = getObjectFromCache(id);
            }
            // not found in cache, will get a Hollow instance
            if (pc == null)
            {
                try
                {
                    Class pcClass = clr.classForName(className, id instanceof OID ? null : id.getClass().getClassLoader());
                    StateManagerImpl sm = new StateManagerImpl(this, pcClass, id, fv);
                    pc = sm.getObject();
                    putObjectIntoCache(pc, true, true);
                }
                catch (ClassNotFoundException e)
                {
                    JPOXLogger.JDO.warn(LOCALISER.msg("PM.GetObjectByIdClassNotFound", id));
                    throw new JDOUserException(LOCALISER.msg("PM.GetObjectByIdClassNotFound", id), e);
                }
            }
        }
        return pc;
    }

    /**
     * This method returns an object id instance corresponding to the pcClass and key arguments. 
     * A String argument might have been the result of executing toString on an object id instance.
     * The key argument is the value of the key field for single field identity. This method 
     * is portable for datastore identity and application identity.
     * @param pcClass Class of the PersistenceCapable to create the OID for.
     * @param key Value of the key for SingleFieldIdentity
     * @return The new object-id instance
     */
    public Object newObjectIdInstance (Class pcClass, Object key)
    {
        assertIsOpen();
        if (pcClass == null)
        {
            throw new JDOUserException(LOCALISER.msg("PM.ObjectId.ClassIsNull"));
        }
        assertPersistenceCapableClass(pcClass);

        ClassMetaData cmd = getMetaDataManager().getMetaDataForClass(pcClass);
        if (cmd == null)
        {
            throw new JDOFatalInternalException(LOCALISER.msg("PM.ObjectId.ClassHasNoMetaData", pcClass.getName()));
        }

        // If the class is not yet managed, manage it
        if (!srm.managesClass(pcClass.getName()))
        {
            srm.addClass(pcClass.getName());
        }

        Object id = null;
        if (AIDUtils.isSingleFieldIdentityClass(cmd.getObjectidClass()))
        {
            // Single Field Identity
            Class idType = clr.classForName(cmd.getObjectidClass());
            id = AIDUtils.getNewSingleFieldIdentity(idType, pcClass, key);
        }
        else if (key instanceof java.lang.String)
        {
            // String-based PK (datastore identity or application identity)
            id = newObjectIdInstance(pcClass, (String)key);
        }
        else
        {
            // Key is not a string, and is not SingleFieldIdentity
            throw new JDOUserException(LOCALISER.msg("PM.ObjectId.KeyValueNotSupported", pcClass.getName(), key.getClass().getName()));
        }

        return id;
    }

    /**
     * This method returns an object id instance corresponding to the Class and
     * String arguments. The String argument might have been the result of
     * executing toString on an object id instance.
     * @param pcClass PersistenceCapable Class
     * @param str String identifier
     * @return The object id
     */
    public Object newObjectIdInstance(Class pcClass, String str)
    {
        ClassMetaData cmd = getMetaDataManager().getMetaDataForClass(pcClass);
        if (cmd == null)
        {
            return null;
        }

        // If the class is not yet managed, manage it
        if (!srm.managesClass(pcClass.getName()))
        {
            srm.addClass(pcClass.getName());
        }

        // Return the new OID instance
        if (cmd.getIdentityType() == IdentityType.APPLICATION)
        {
            if (Modifier.isAbstract(pcClass.getModifiers()) && cmd.getObjectidClass() != null) 
            {
                try
                {
                    Constructor c = clr.classForName(cmd.getObjectidClass()).getDeclaredConstructor(new Class[] {java.lang.String.class});
                    Object id = c.newInstance(new Object[] {str});
                    return id;
                }
                catch(Exception e) 
                {
                    String msg = LOCALISER.msg("PM.NewObjectIdInstance.ConstructionError", cmd.getObjectidClass(), cmd.getFullClassName());
                    JPOXLogger.JDO.error(msg);
                    JPOXLogger.JDO.error(e);

                    throw new JDOUserException(msg);
                }
            }
            else
            {
                return JDOImplHelper.getInstance().newObjectIdInstance(pcClass, str);
            }
        }
        else
        {
            return new OID(str);
        }
    }

    /**
     * Accessor for an object id given the object.
     * @param pc The object
     * @return The Object id
     */
    public Object getObjectId(Object pc)
    {
        assertIsOpen();
        if (pc != null && pc instanceof PersistenceCapable)
        {
            PersistenceCapable p = (PersistenceCapable) pc;
            if (p.jdoIsPersistent())
            {
                return p.jdoGetObjectId();
            }
        }
        return null;
    }

    /**
     * Accessor for the object id of a transactional object given the object.
     * @param pc The object
     * @return The Object id
     */
    public Object getTransactionalObjectId(Object pc)
    {
        assertIsOpen();
        return ((PersistenceCapable) pc).jdoGetTransactionalObjectId();
    }

    /**
     * Acessor for the current FetchPlan
     * @return FetchPlan
     * @since 1.1
     */
    public FetchPlan getFetchPlan()
    {
        return fetchPlan;
    }

    /**
     * Method to put a user object into the PersistenceManager. This is so that
     * multiple users can each have a user object for example. <I>The parameter
     * is not inspected or used in any way by the JDO implementation. </I>
     * @param key The key to store the user object under
     * @param value The object to store
     * @return The previous value for this key
     * @since 1.1
     */
    public synchronized Object putUserObject(Object key, Object value)
    {
        assertIsOpen();
        if (key == null)
        {
            return null;
        }
        if (value == null)
        {
            // Remove the object
            return userObjectMap.remove(key);
        }
        else
        {
            // Put the object
            return userObjectMap.put(key, value);
        }
    }

    /**
     * Method to get a user object from the PersistenceManager. This is for user
     * objects which are stored under a key. <I>The parameter is not inspected
     * or used in any way by the JDO implementation. </I>
     * @param key The key to store the user object under
     * @return The user object for that key
     * @since 1.1
     */
    public synchronized Object getUserObject(Object key)
    {
        assertIsOpen();
        if (key == null)
        {
            return null;
        }
        return userObjectMap.get(key);
    }

    /**
     * Method to remove a user object from the PersistenceManager. This is for
     * user objects which are stored under a key. <I>The parameter is not
     * inspected or used in any way by the JDO implementation. </I>
     * @param key The key whose uder object is to be removed.
     * @return The user object that was removed
     * @since 1.1
     */
    public synchronized Object removeUserObject(Object key)
    {
        assertIsOpen();
        if (key == null)
        {
            return null;
        }
        return userObjectMap.remove(key);
    }

    /**
     * The application might manage PersistenceManager instances by using an
     * associated object for bookkeeping purposes. These methods allow the user
     * to manage the associated object. <I>The parameter is not inspected or
     * used in any way by the JDO implementation. </I>
     * @param userObject The object
     */
    public synchronized void setUserObject(Object userObject)
    {
        assertIsOpen();
        this.userObject = userObject;
    }

    /**
     * The application might manage PersistenceManager instances by using an
     * associated object for bookkeeping purposes. These methods allow the user
     * to manage the associated object. <I>The parameter is not inspected or
     * used in any way by the JDO implementation. </I>
     * @return The user object
     */
    public synchronized Object getUserObject()
    {
        assertIsOpen();
        return userObject;
    }

    /**
     * Accessor for the PersistenceManager Factory.
     * @return The PersistenceManagerFactory
     */
    public AbstractPersistenceManagerFactory getAbstractPersistenceManagerFactory()
    {
        return apmf;
    }

    /**
     * Accessor for the PersistenceManager Factory.
     * @return The PersistenceManagerFactory
     */
    public PersistenceManagerFactory getPersistenceManagerFactory()
    {
        // To be implemented by subclasses
        throw new JDOException(LOCALISER.msg("PM.GetPersistenceManagerFactoryNotImplemented"));
    }

    /**
     * Accessor for the class of the object id given the class of object.
     * @param cls The class name of the object
     * @return The class name of the object id
     */
    public Class getObjectIdClass(Class cls)
    {
        assertIsOpen();
        if (!ClassUtils.isPersistenceCapableClass(cls) || !hasMetaDataForPersistenceCapableClass(cls))
        {
            return null;
        }
        ClassMetaData cmd = getMetaDataManager().getMetaDataForClass(cls);
        if (cmd.getIdentityType() == IdentityType.DATASTORE)
        {
            return OID.class;
        }
        else if (cmd.getIdentityType() == IdentityType.APPLICATION)
        {
            try
            {
                return this.getClassLoaderResolver().classForName(getMetaDataManager().getMetaDataForClass(cls).getObjectidClass(), null);
            }
            catch (ClassNotFoundException e)
            {
                String msg = LOCALISER.msg("PM.ObjectIdClassNotFound", cls.getName());
                JPOXLogger.GENERAL.error(msg);
                throw new JDOException(msg);
            }
        }
        else
        {
            if (cmd.isRequiresExtent())
            {
                return OID.class;
            }
            else
            {
                return SCOID.class;
            }
        }
    }

    /**
     * Mutator for whether the Persistence Manager is multithreaded.
     * @param flag Whether to run multithreaded.
     */
    public void setMultithreaded(boolean flag)
    {
        assertIsOpen();
        this.multithreaded = flag;
    }

    /**
     * Accessor for whether the Persistence Manager is multithreaded.
     * @return Whether to run multithreaded.
     */
    public boolean getMultithreaded()
    {
        assertIsOpen();
        return this.multithreaded;
    }

    /**
     * Mutator for whether to ignore the cache.
     * @param flag Whether to ignore the cache.
     */
    public void setIgnoreCache(boolean flag)
    {
        assertIsOpen();
        this.ignoreCache = flag;
    }

    /**
     * Accessor for whether to ignore the cache.
     * @return Whether to ignore the cache.
     */
    public boolean getIgnoreCache()
    {
        assertIsOpen();
        return ignoreCache;
    }
    
    /**
     * Mutator for whether to detach objects on close of the PM.
     * @param flag Whether to detach on close.
     */
    public void setDetachOnClose(boolean flag)
    {
        assertIsOpen();
        this.detachOnClose = flag;
    }

    /**
     * Accessor for whether to detach objects on close of the PM.
     * @return Whether to detach on close.
     */
    public boolean getDetachOnClose()
    {
        assertIsOpen();
        return detachOnClose;
    }

    /**
     * Method to find the StateManager for an object.
     * @param pc The object
     * @return The StateManager, null if StateManager not found. see JDO spec
     * for the calling behavior when null is returned
     */
    public synchronized StateManager findStateManager(PersistenceCapable pc)
    {
        StateManager sm = null;
        PersistenceCapable previousLookingFor = lookingForStateManagerFor;
        StateManager previousFound = foundStateManager;
        try
        {
            lookingForStateManagerFor = pc;
            foundStateManager = null;
            // this method will call the
            // StateManager.getPersistenceManager(PersistenceCapable pc)
            // StateManager.getPersistenceManager will call hereIsStateManager
            // that will set the foundStateManager
            PersistenceManager pm = (PersistenceManager) pc.jdoGetPersistenceManager();
            if (pm != null && this.getPMHandle() != pm.getPMHandle())
            {
                throw new JDOUserException(LOCALISER.msg("PM.DifferentOwner", pc.jdoGetObjectId()));
            }
            sm = foundStateManager;
        }
        finally
        {
            lookingForStateManagerFor = previousLookingFor;
            foundStateManager = previousFound;
        }
        return sm;
    }

    /**
     * Method to add the StateManager for an object to this PersistenceManager's
     * list.
     * @param sm The StateManager
     * @param pc The object managed by the StateManager
     */
    public synchronized void hereIsStateManager(StateManager sm, PersistenceCapable pc)
    {
        if (lookingForStateManagerFor == pc)
        {
            foundStateManager = sm;
        }
    }

    /**
     * Method to mark an object (StateManager) as dirty.
     * @param sm The StateManager
     */
    public synchronized void clearDirty(StateManager sm)
    {
        dirties.remove(sm);
    }

    /**
     * Method to mark an object (StateManager) as dirty.
     * @param sm The StateManager
     */
    public synchronized void markDirty(StateManager sm)
    {
        if (!this.isDelayDatastoreOperationsEnabled() && dirties.size() > 0 && !dirties.contains(sm))
        {
            flush();
        }
        if (!dirties.contains(sm))
        {
            dirties.add(sm);
        }
    }

    /**
     * Method to mark as clean the StateManager.
     */
    public synchronized void clearDirty()
    {
        dirties.clear();
    }

    /**
     * This method flushes all dirty, new, and deleted instances to the
     * datastore. It has no effect if a transaction is not active. If a
     * datastore transaction is active, this method synchronizes the cache with
     * the datastore and reports any exceptions. If an optimistic transaction is
     * active, this method obtains a datastore connection and synchronizes the
     * cache with the datastore using this connection. The connection obtained
     * by this method is held until the end of the transaction.
     *  
     */
    public synchronized void flush()
    {
        assertIsOpen();
        List toFlush = new ArrayList(dirties);
        clearDirty();
        for (int i = 0; i < toFlush.size(); i++)
        {
            StateManager sm = (StateManager) toFlush.get(i);
            sm.flush();
        }
    }

    /**
     * This method validates the cache with the datastore. It has no effect if a transaction 
     * is not active. If a datastore transaction is active, this method verifies the consistency 
     * of instances in the cache against the datastore. An implementation might flush instances 
     * as if flush() were called, but it is not required to do so. 
     * If an optimistic transaction is active, this method obtains a datastore connection 
     * and verifies the consistency of the instances in the cache against the datastore. If any 
     * inconsistencies are detected, a JDOOptimisticVerificationException is thrown. This 
     * exception contains a nested JDOOptimisticVerificationException for each object 
     * that failed the consistency check. No datastore resources acquired during the execution 
     * of this method are held beyond the scope of this method.
     */
    public void checkConsistency()
    {
        assertIsOpen();

        // If transaction is not active do nothing
        if (!tx.isActive())
        {
            return;
        }

        // TODO Implement checkConsistency()
        throw new JDOUserException("checkConsistency() not yet implemented");
    }

    /**
     * Commit any changes made to objects managed by the persistence manager to
     * the database.
     */
    public synchronized void postCommit()
    {
        ArrayList failures = new ArrayList();
        try
        {
            StateManager[] sms = (StateManager[]) txCache.toArray(new StateManager[txCache.size()]);
            for (int i = 0; i < sms.length; ++i)
            {
                try
                {
                    sms[i].postCommit(currentTransaction());
                }
                catch (RuntimeException e)
                {
                    failures.add(e);
                }
            }
        }
        finally
        {
            txCache.clear();
        }
        if (!failures.isEmpty())
        {
            throw new CommitStateTransitionException((Exception[]) failures.toArray(new Exception[failures.size()]));
        }
    }

    /**
     * Rollback any changes made to objects managed by the persistence manager
     * to the database.
     */
    public synchronized void preRollback()
    {
        ArrayList failures = new ArrayList();
        try
        {
            StateManager[] sms = (StateManager[]) txCache.toArray(new StateManager[txCache.size()]);
            for (int i = 0; i < sms.length; ++i)
            {
                try
                {
                    sms[i].preRollback();
                }
                catch (RuntimeException e)
                {
                    failures.add(e);
                }
            }
            clearDirty();
        }
        finally
        {
            txCache.clear();
        }
        if (!failures.isEmpty())
        {
            throw new RollbackStateTransitionException((Exception[]) failures.toArray(new Exception[failures.size()]));
        }
    }

    /**
     * Accessor for a handle to this PersistenceManager.
     * 
     * @return The handle.
     */
    public abstract PersistenceManager getPMHandle();

    /**
     * Whether the datastore operations are delayed until commit. In optimistic
     * transactions this is automatically enabled.
     * 
     * @return true if datastore operations are delayed until commit
     */
    public boolean isDelayDatastoreOperationsEnabled()
    {
        if (tx.getOptimistic())
        {
            return true;
        }
        else
        {
            return getPMFContext().getPmfConfiguration().getDelayDatastoreOperationsEnabled();
        }
    }

    public void dump(Object obj, PrintWriter out)
    {
        findStateManager((PersistenceCapable) obj).dump(out);
    }

    /**
     * Gets the context in which this PersistenceManager is running
     * @return Returns the pmfContext.
     */
    public PMFContext getPMFContext()
    {
        return apmf.getPMFContext();
    }

    /**
     * Accessor for the MetaDataManager for this PMF. This is used as the
     * interface to MetaData in the PM/PMF.
     * @return Returns the MetaDataManager.
     */
    public MetaDataManager getMetaDataManager()
    {
        return apmf.getPMFContext().getMetaDataManager();
    }

    // ------------------------------------- Sequence Management --------------------------------------

    /**
     * Method to retrieve a sequence by name. JDO spec section 12.14
     * If the named sequence is not known, throws a JDOUserException.
     * @param sequenceName Name of the sequence
     * @return The sequence
     */
    public Sequence getSequence(String sequenceName)
    {
        assertIsOpen();

        // TODO Implement getSequence(String)
        throw new JDOUserException("getSequence() not currently supported");
    }

    // -------------------------------------- Cache Management ---------------------------------------

    /**
     * Convenience method to add an object to the cache(s).
     * @param pc The Persistence Capable object
     * @param level1 Whether to put it in the L1 cache
     * @param level2 Whether to put it in the L2 cache
     */
    public synchronized void putObjectIntoCache(PersistenceCapable pc, boolean level1, boolean level2)
    {
        if (pc == null || pc.jdoGetObjectId() == null)
        {
            JPOXLogger.CACHE.warn(LOCALISER.msg("Cache.Level1.AddingNullObjectError"));
            return;
        }

        // Put into Level 1 Cache
        if (level1)
        {
            if (JPOXLogger.CACHE.isDebugEnabled())
            {
                JPOXLogger.CACHE.debug(LOCALISER.msg("Cache.Level1.AddingObject", pc.jdoGetObjectId()));
            }
            cache.put(pc.jdoGetObjectId(), pc);
        }

        // Put into Level 2 Cache
        if (level2 && this.apmf.getJdoCacheLevel2())
        {
            if (JPOXLogger.CACHE.isDebugEnabled())
            {
                JPOXLogger.CACHE.debug(LOCALISER.msg("Cache.Level2.AddingObject", pc.jdoGetObjectId()));
            }
            // Store a copy with no connected StateManager
            PersistenceCapable pcCopy = ((StateManagerImpl)findStateManager(pc)).getPCCopyWithoutSM();
            ((Level2Cache)this.apmf.getDataStoreCache()).put(pc.jdoGetObjectId(), pcCopy);
        }
    }

    /**
     * Convenience method to evict an object from the cache(s).
     * @param pc The PersistenceCapable object
     * @param id The PersistenceCapable object id
     * @param level1 Whether to evict from the L1 cache
     * @param level2 Whether to evict from the L2 cache
     */
    public synchronized void removeObjectFromCache(PersistenceCapable pc, Object id, boolean level1, boolean level2)
    {
        // Evict from the L1 cache
        if (level1 && id != null)
        {
            if (JPOXLogger.CACHE.isDebugEnabled())
            {
                JPOXLogger.CACHE.debug(LOCALISER.msg("Cache.Level1.RemovingObject", id, "" + cache.size()));
            }
            Object pcRemoved = cache.remove(id);
            if (pcRemoved == null && JPOXLogger.CACHE.isDebugEnabled())
            {
                // For some reason the object isn't in the L1 cache - garbage collected maybe ?
                JPOXLogger.CACHE.debug(LOCALISER.msg("Cache.Level1.RemovalOfObjectFailed", id));
            }
        }

        // Evict from the L2 cache
        if (level2 && id != null && apmf.getJdoCacheLevel2())
        {
            Level2Cache l2Cache = ((Level2Cache)apmf.getDataStoreCache());
            if (JPOXLogger.CACHE.isDebugEnabled())
            {
                JPOXLogger.CACHE.debug(LOCALISER.msg("Cache.Level2.RemovingObject", id, "" + l2Cache.getSize()));
            }
            l2Cache.evict(pc);
        }
    }

    /**
     * Convenience method to access an object in the cache.
     * Firstly looks in the L1 cache for this PM, and if not found looks in the L2 cache.
     * @param id Id of the object
     * @return Persistence Capable object (with connected StateManager).
     */
    public synchronized PersistenceCapable getObjectFromCache(Object id)
    {
        PersistenceCapable pc = null;

        // Try Level 1 first
        pc = (PersistenceCapable)cache.get(id);
        if (pc != null)
        {
            if (JPOXLogger.CACHE.isDebugEnabled())
            {
                JPOXLogger.CACHE.debug(LOCALISER.msg("Cache.Level1.RetrievalOfObject", id, "" + cache.size()));
            }
            return pc;
        }
        else
        {
            if (JPOXLogger.CACHE.isDebugEnabled())
            {
                JPOXLogger.CACHE.debug(LOCALISER.msg("Cache.Level1.RetrievalFailed", id, "" + cache.size()));
            }
        }

        // Try Level 2 since not in Level 1
        if (pc == null && apmf.getJdoCacheLevel2())
        {
            Level2Cache l2Cache = ((Level2Cache)apmf.getDataStoreCache());
            pc = l2Cache.get(id);

            // Copy the cached object and connect to a StateManager, with the same object id
            if (pc != null)
            {
                StateManagerImpl sm = new StateManagerImpl(this, id, pc);
                pc = sm.getObject();
                if (JPOXLogger.CACHE.isDebugEnabled())
                {
                    JPOXLogger.CACHE.debug(LOCALISER.msg("Cache.Level2.RetrievalOfObject", id, "" + l2Cache.getSize()));
                }
            }
            else
            {
                if (JPOXLogger.CACHE.isDebugEnabled())
                {
                    JPOXLogger.CACHE.debug(LOCALISER.msg("Cache.Level2.RetrievalFailed", id, "" + l2Cache.getSize()));
                }
            }
        }

        return null;
    }

    // ------------------------------------- Lifecycle Listeners --------------------------------------
    
    /**
     * JDO 2.0 spec 12.15 "LifecycleListeners".
     * (non-Javadoc)
     * @see javax.jdo.PersistenceManager#addLifecycleListener(javax.jdo.InstanceLifecycleListener, java.lang.Class[])
     * @since 1.1
     */
    public void addLifecycleListener(InstanceLifecycleListener listener, Class[] classes)
    {
        assertIsOpen();
        if (listener == null)
        {
            return;
        }

        lifecycleListeners.add(new LifecycleListenerSpecification(listener, classes));
    }

    /**
     * JDO 2.0 spec 12.15 "LifecycleListeners".
     * (non-Javadoc)
     * @see javax.jdo.PersistenceManager#removeLifecycleListener(javax.jdo.InstanceLifecycleListener)
     * @since 1.1
     */
    public void removeLifecycleListener(InstanceLifecycleListener listener)
    {
        assertIsOpen();
        if (listener != null)
        {
            Iterator iter = lifecycleListeners.iterator();
            while (iter.hasNext())
            {
                LifecycleListenerSpecification classListener = (LifecycleListenerSpecification) iter.next();
                if (classListener.getListener() == listener)
                {
                    iter.remove();
                }
            }
        }
    }

    /**
     * Utility to notify the lifecycle listeners for an event on an object.
     * JDO 2.0 spec 12.15 "LifecycleListeners".
     * @param obj The object that had the operation performed on it
     * @param lifecycleEventType The lifecycle event
     * @since 1.1
     */
    public void notifyLifecycleListenersForEvent(Object obj, int lifecycleEventType)
    {
        if (obj == null)
        {
            return;
        }

        Iterator iter = lifecycleListeners.iterator();
        while (iter.hasNext())
        {
            LifecycleListenerSpecification classListener = (LifecycleListenerSpecification) iter.next();

            // Check if the passed object relates to this listener
            boolean forClass = classListener.forClass(obj.getClass());
            if (forClass)
            {
                InstanceLifecycleListener listener=classListener.getListener();
                if (listener instanceof DeleteLifecycleListener &&
                        lifecycleEventType == LifecycleListenerSpecification.EVENT_PRE_DELETE)
                {
                    ((DeleteLifecycleListener)listener).preDelete(new InstanceLifecycleEvent(obj, InstanceLifecycleEvent.DELETE));
                }
                else if (listener instanceof DeleteLifecycleListener &&
                        lifecycleEventType == LifecycleListenerSpecification.EVENT_POST_DELETE)
                {
                    ((DeleteLifecycleListener)listener).postDelete(new InstanceLifecycleEvent(obj, InstanceLifecycleEvent.DELETE));
                }
                else if (listener instanceof LoadLifecycleListener &&
                        lifecycleEventType == LifecycleListenerSpecification.EVENT_POST_LOAD)
                {
                    ((LoadLifecycleListener)listener).postLoad(new InstanceLifecycleEvent(obj, InstanceLifecycleEvent.LOAD));
                }
                else if (listener instanceof StoreLifecycleListener &&
                        lifecycleEventType == LifecycleListenerSpecification.EVENT_PRE_STORE)
                {
                    ((StoreLifecycleListener)listener).preStore(new InstanceLifecycleEvent(obj, InstanceLifecycleEvent.STORE));
                }
                else if (listener instanceof StoreLifecycleListener &&
                        lifecycleEventType == LifecycleListenerSpecification.EVENT_POST_STORE)
                {
                    ((StoreLifecycleListener)listener).postStore(new InstanceLifecycleEvent(obj, InstanceLifecycleEvent.STORE));
                }
                else if (listener instanceof CreateLifecycleListener &&
                        lifecycleEventType == LifecycleListenerSpecification.EVENT_POST_CREATE)
                {
                    ((CreateLifecycleListener)listener).postCreate(new InstanceLifecycleEvent(obj, InstanceLifecycleEvent.CREATE));
                }
                else if (listener instanceof ClearLifecycleListener &&
                        lifecycleEventType == LifecycleListenerSpecification.EVENT_PRE_CLEAR)
                {
                    ((ClearLifecycleListener)listener).preClear(new InstanceLifecycleEvent(obj, InstanceLifecycleEvent.CLEAR));
                }
                else if (listener instanceof ClearLifecycleListener &&
                        lifecycleEventType == LifecycleListenerSpecification.EVENT_POST_CLEAR)
                {
                    ((ClearLifecycleListener)listener).postClear(new InstanceLifecycleEvent(obj, InstanceLifecycleEvent.CLEAR));
                }
                else if (listener instanceof DirtyLifecycleListener &&
                        lifecycleEventType == LifecycleListenerSpecification.EVENT_PRE_DIRTY)
                {
                    ((DirtyLifecycleListener)listener).preDirty(new InstanceLifecycleEvent(obj, InstanceLifecycleEvent.DIRTY));
                }
                else if (listener instanceof DirtyLifecycleListener &&
                        lifecycleEventType == LifecycleListenerSpecification.EVENT_POST_DIRTY)
                {
                    ((DirtyLifecycleListener)listener).postDirty(new InstanceLifecycleEvent(obj, InstanceLifecycleEvent.DIRTY));
                }
                else if (listener instanceof DetachLifecycleListener &&
                        lifecycleEventType == LifecycleListenerSpecification.EVENT_PRE_DETACH)
                {
                    ((DetachLifecycleListener)listener).preDetach(new InstanceLifecycleEvent(obj, InstanceLifecycleEvent.DETACH));
                }
                else if (listener instanceof DetachLifecycleListener &&
                        lifecycleEventType == LifecycleListenerSpecification.EVENT_POST_DETACH)
                {
                    ((DetachLifecycleListener)listener).postDetach(new InstanceLifecycleEvent(obj, InstanceLifecycleEvent.DETACH));
                }
                else if (listener instanceof AttachLifecycleListener &&
                        lifecycleEventType == LifecycleListenerSpecification.EVENT_PRE_ATTACH)
                {
                    ((AttachLifecycleListener)listener).preAttach(new InstanceLifecycleEvent(obj, InstanceLifecycleEvent.ATTACH));
                }
                else if (listener instanceof AttachLifecycleListener &&
                        lifecycleEventType == LifecycleListenerSpecification.EVENT_POST_ATTACH)
                {
                    ((AttachLifecycleListener)listener).postAttach(new InstanceLifecycleEvent(obj, InstanceLifecycleEvent.ATTACH));
                }
            }
        }
    }
}