/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others. All rights reserved.
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2002 Kelly Grizzle (TJDO)
2003 Erik Bengtson - refactored the persistent id generator System property
2003 Erik Bengtson - added PMFConfiguration, PMFContext
2003 Andy Jefferson - commented, and added ApplicationId to supported options
2003 Andy Jefferson - introduction of localiser
2004 Andy Jefferson - update to getProperties() method
2004 Andy Jefferson - added LifecycleListener
2004 Andy Jefferson - added Level 2 Cache
    ...
**********************************************************************/
package org.jpox;

import java.lang.reflect.Constructor;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Collections;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Properties;
import java.util.ResourceBundle;
import java.util.Set;

import javax.jdo.DataStoreCache;
import javax.jdo.JDOFatalUserException;
import javax.jdo.JDOUserException;
import javax.jdo.InstanceLifecycleListener;
import javax.jdo.spi.JDOPermission;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.sql.DataSource;

import org.jpox.cache.Level2Cache;
import org.jpox.cache.NullLevel2Cache;
import org.jpox.exceptions.ConnectionFactoryNotFoundException;
import org.jpox.exceptions.UnsupportedConnectionFactoryException;
import org.jpox.util.JPOXLogger;

/**
 * A factory used to obtain {@link javax.jdo.PersistenceManager} instances.
 * The factory is configurable up to a point when it is frozen. Thereafter
 * nothing can be changed.
 *
 * @version $Revision: 1.14 $
 */
public abstract class AbstractPersistenceManagerFactory extends PMFConfiguration
{
    private static final String VERSION_NUMBER_PROPERTY = "VersionNumber";
    private static final String VENDOR_NAME_PROPERTY = "VendorName";
    private static final String VENDOR_NAME="JPOX";

    /** Version of JPOX being used. Read in at startup from properties. */
    private static String jpoxVersion=null;

    /** Vendor of this version of JPOX. */
    private static String jpoxVendor=null;

    /** Transactional data source to use. */
    private transient DataSource tds = null;

    /** Non-transactional data source (or just the transactional variant) */
    private transient DataSource ntds = null;

    /** The cache of PM's */
    private Set pmCache;

    /** The context that this PMF uses. */
    protected PMFContext pmfContext;

    /** Level 2 Cache */
    protected Level2Cache cache;

    /** Lifecycle Listeners */
    protected List lifecycleListeners = new ArrayList();

    /** Whether the PMF is closed */
    private boolean closed;

    /**
     * Constructor for the PMF.
     */
    public AbstractPersistenceManagerFactory()
    {
        super();
        pmfContext = new PMFContext(this);

        // Check if JDO 2 classes are available.
        // This may be the cause of errors if the user puts a JDO 1 jar in front of
        // the JDO 2 classes.
        // TODO Remove/Update this when we have the real JDO 2 jar
        Class pc=PersistenceManager.class;
        boolean found=false;
        Method[] methods=pc.getMethods();
        for (int i=0;i<methods.length;i++)
        {
            Method m=methods[i];
            if (m.getName().equals("detachCopy"))
            {
                found = true;
                break;
            }
        }
        if (!found)
        {
            throw new JDOFatalUserException(LOCALISER.msg("PMF.JDO2Inaccessible"));
        }
    }

    /**
     * Looks up a DataSource object in JNDI.
     *
     * @param name The JNDI name of the DataSource.
     * @return  The DataSource object.
     * @exception ConnectionFactoryNotFoundException If a JNDI lookup failure occurs.
     * @exception UnsupportedConnectionFactoryException If the object is not a javax.sql.DataSource.
     */
    private DataSource lookupDataSource(String name)
    {
        Object obj;

        try
        {
            obj = new InitialContext().lookup(name);
        }
        catch (NamingException e)
        {
            throw new ConnectionFactoryNotFoundException(name, e);
        }

        if (!(obj instanceof DataSource))
        {
            throw new UnsupportedConnectionFactoryException(obj);
        }

        return (DataSource)obj;
    }

    /**
     * Freezes the current configuration.
     *
     * @exception UnsupportedConnectionFactoryException
     *      If the configured connection factory is not a javax.sql.DataSource.
     */
    protected void freezeConfiguration()
    {
        if (configurable)
        {
            if (getConnectionFactory() != null)
            {
                if (!(getConnectionFactory() instanceof DataSource))
                {
                    throw new UnsupportedConnectionFactoryException(getConnectionFactory());
                }

                tds = (DataSource)getConnectionFactory();
            }
            else if (getConnectionFactoryName() != null)
            {
                tds = lookupDataSource(getConnectionFactoryName());
            }
            else
            {
                tds = new DriverManagerDataSource(getConnectionDriverName(), getConnectionURL());
            }

            if (getConnectionFactory2() != null)
            {
                if (!(getConnectionFactory2() instanceof DataSource))
                {
                    throw new UnsupportedConnectionFactoryException(getConnectionFactory2());
                }

                tds = (DataSource)getConnectionFactory2();
            }
            else if (getConnectionFactory2Name() != null)
            {
                ntds = lookupDataSource(getConnectionFactory2Name());
            }
            else
            {
                ntds = tds;
            }

            // Set up the Level 2 Cache
            if (getJdoCacheLevel2())
            {
                String level2Type = getJdoCacheLevel2Type();
                try
                {
                    cache = createCache(level2Type);
                    JPOXLogger.CACHE.info(LOCALISER.msg("Cache.Level2.Initialised", level2Type));
                }
                catch (Exception e)
                {
                    String fallbackType = "org.jpox.cache.DefaultLevel2Cache";
                    JPOXLogger.CACHE.info(LOCALISER.msg("Cache.Level2.FailedRetrying", level2Type, fallbackType));
                    level2Type = fallbackType;
                    try
                    {
                        cache = createCache(level2Type);
                        JPOXLogger.CACHE.info(LOCALISER.msg("Cache.Level2.Initialised", level2Type));
                    }
                    catch (Exception e2)
                    {
                        JPOXLogger.CACHE.info(LOCALISER.msg("Cache.Level2.InitialisationFailed"));
                    }
                }
            }
            else
            {
                cache = new NullLevel2Cache();
            }

            configurable = false;

            // Log the configuration
            JPOXLogger.JDO.info(LOCALISER.msg("PMF.Version",getVendorName(),getVersionNumber()));
            JPOXLogger.JDO.info(LOCALISER.msg("PMF.Initialised", getConnectionURL(),getConnectionDriverName(), getConnectionUserName()));
            JPOXLogger.JDO.debug("================ PMF Configuration ================");
            JPOXLogger.JDO.debug("Options : " + (getMultithreaded() ? "multithreaded" : "singlethreaded") +
                    (getRetainValues() ? ", retainValues" : "") +
                    (getRestoreValues() ? ", restoreValues" : "") +
                    (getNontransactionalRead() ? ", nontransactionalRead" : "") +
                    (getNontransactionalWrite() ? ", nontransactionalWrite" : "") +
                    (getIgnoreCache() ? ", ignoreCache" : "") +
                    (getReadOnlySchema() ? ", READONLY" : "") +
                    (getFixedSchema() ? ", FIXED" : "") +
                    (getUseUpdateLock() ? ", useUpdateLock" : "") +
                    (getAutoCreateSchema() ? ", autoCreateSchema" : "") +
                    (getAutoCreateTables() ? ", autoCreateTables" : "") +
                    (getAutoCreateColumns() ? ", autoCreateColumns" : "") +
                    (getAutoCreateConstraints() ? ", autoCreateConstraints" : "") +
                    (getValidateTables() ? ", validateTables" : "") +
                    (getValidateConstraints() ? ", validateConstraints" : "") +
                    ", foreignKeyCreateMode=" + getForeignKeyCreateMode() +
                    ", jdoMetadataFileSuffix=" + getJdoMetaDataFileExtension() + 
                    ", ormMetadataFileSuffix=" + getOrmMetaDataFileExtension() +
                    ", mappingName=" + getMapping());
            JPOXLogger.JDO.debug("Types Control : stringDefaultLength=" + getStringDefaultLength() + ", dateTimezone=" + getDateTimezone());
            JPOXLogger.JDO.debug("Schema Identifiers : defaultCatalogName=" + getDefaultCatalogName() + ", defaultSchemaName=" + getDefaultSchemaName() + 
                ", fullyQualified=" + getFullyQualifiedNames());
            JPOXLogger.JDO.debug("Initialisation : mechanism=" + getAutoStartMechanism() + ", mode=" + getAutoStartMechanismMode());
            JPOXLogger.JDO.debug("Transactions : " + (getOptimistic() ? "optimistic" : "pessimistic") +
                ", isolationLevel=" + getTransactionIsolation() +
                ", poidIsolationLevel=" + getPoidTransactionIsolationLevel() + ", poidTransactionAttribute=" + getPoidTransactionAttribute());

            String cache_level = "1";
            if (getJdoCacheLevel2())
            {
                cache_level = "1/2";
            }
            JPOXLogger.JDO.debug("Cache : level=" + cache_level + 
                ", cachedCollections=" + getJdoCacheCollections() + 
                ", level1Type=" + getJdoCacheLevel1Type() + 
                (getJdoCacheLevel2() ? (", level2Type=" + getJdoCacheLevel2Type()) : ""));
            JPOXLogger.JDO.debug("Queries : fetchSize=" + getQueryFetchSize() + ", fetchDirection=" + getQueryFetchDirection() + 
                ", resultSetType=" + getQueryResultSetType() + ", resultSetConcurrency=" + getQueryResultSetConcurrency());
            JPOXLogger.JDO.debug("===================================================");
        }
    }
    
    /**
     * Utility to create an instance of the level 2 cache using reflection.
     * @param level2Type The level 2 type
     * @return The Level 2 Cache
     * @throws Exception Thrown if an error occurs
     */
    private Level2Cache createCache(String level2Type)
    throws Exception
    {
        Properties cacheProps = new Properties();
        if (getJdoCacheLevel2CacheName() != null)
        {
            cacheProps.put("CacheName", getJdoCacheLevel2CacheName());
        }

        Class level2CacheClass = Class.forName(level2Type);
        Class[] ctrArgsClasses = new Class[] {Properties.class};
        Object[] ctrArgs = new Object[] {cacheProps};

        Constructor ctr = level2CacheClass.getConstructor(ctrArgsClasses);
        cache = (Level2Cache)ctr.newInstance(ctrArgs);
        
        return cache;
    }

    /**
     * Returns the data source to be used for transactional access.
     *
     * @return  The data source to be used for transactional access.
     */
    public synchronized DataSource getTransactionalDataSource()
    {
        freezeConfiguration();

        return tds;
    }

    /**
     * Returns the data source to be used for non-transactional access.
     *
     * @return  The data source to be used for non-transactional access.
     */
    public synchronized DataSource getNontransactionalDataSource()
    {
        freezeConfiguration();

        return ntds;
    }

    /**
     * Return non-configurable properties of this PersistenceManagerFactory.
     * Properties with keys VendorName and VersionNumber are required.  Other
     * keys are optional.
     * @return the non-configurable properties of this
     * PersistenceManagerFactory.
     */
    public Properties getProperties()
    {
        Properties props = new Properties();

        props.setProperty(VENDOR_NAME_PROPERTY,getVendorName());
        props.setProperty(VERSION_NUMBER_PROPERTY,getVersionNumber());

        return props;
    }

    /**
     * Utility to get the version of JPOX.
     * @return Version number for JPOX.
     **/
    public static String getVersionNumber()
    {
        if (jpoxVersion != null)
        {
            return jpoxVersion;
        }
        
        String version="Unknown";
        try
        {
            ResourceBundle bundle=ResourceBundle.getBundle("org.jpox.JPOXVersion");
            try
            {
                version = bundle.getString("jpox.version");
            }
            catch (Exception e1)
            {
            }
        }
        catch (Exception e)
        {
        }
        
        return jpoxVersion = version;
    }

    /**
     * Utility to get the vendor of JPOX.
     * @return Vendor name for JPOX.
     **/
    public static String getVendorName()
    {
        if (jpoxVendor != null)
        {
            return jpoxVendor;
        }
        
        String vendor=VENDOR_NAME;
        try
        {
            ResourceBundle bundle=ResourceBundle.getBundle("org.jpox.JPOXVersion");
            try
            {
                vendor = bundle.getString("jpox.vendor");
            }
            catch (Exception e1)
            {
            }
        }
        catch (Exception e)
        {
        }

        return jpoxVendor = vendor;
    }

    /**
     * The application can determine from the results of this method which
     * optional features, and which query languages are supported by the JDO
     * implementation. Se esection 11.6 of the JDO 2 specification.
     * @return A Collection of String representing the supported options.
     */
    public Collection supportedOptions()
    {
        return Collections.unmodifiableList(Arrays.asList(OPTION_ARRAY));
    }

    /**
     * The JDO spec optional features that JPOX supports.
     * See JDO 2.0 spec section 11.6 for the full list of possibilities.
     **/
    private static final String[] OPTION_ARRAY =
    {
        "javax.jdo.option.TransientTransactional",
        "javax.jdo.option.NontransactionalRead",
        "javax.jdo.option.RetainValues",
        "javax.jdo.option.Optimistic",
        "javax.jdo.option.ApplicationIdentity",
        "javax.jdo.option.DatastoreIdentity",
        "javax.jdo.option.ArrayList",
        "javax.jdo.option.LinkedList",
        "javax.jdo.option.Vector",
        "javax.jdo.option.List",
        "javax.jdo.option.Array",
        "javax.jdo.option.NullCollection",
        "javax.jdo.query.JDOQL",
        "javax.jdo.query.SQL",
        "javax.jdo.query.JPOXSQL",
        "javax.jdo.option.BinaryCompatibility",
        "javax.jdo.option.GetDataStoreConnection",
        "javax.jdo.option.UnconstrainedQueryVariables"
    };

    /**
     * Return the Set of PersistenceManagers actually in cache
     * @return Set this contains instances of PersistenceManagers in the cache
     */
    private Set getPmCache()
    {
        if (pmCache == null)
        {
            pmCache = new HashSet();
        }
        return pmCache;
    }

    /**
     * Try to get a PersistenceManager from cache, otherwise creates a new
     * @param userName  the user name for the connection
     * @param password  the password for the connection
     * @return Returns a PersistenceManager.
     */
    private synchronized PersistenceManager getPersistenceManagerFromCache(String userName, String password)
    {
        PersistenceManager pm = new PersistenceManagerImpl(this, userName, password);   
        getPmCache().add(pm);
        return pm;
    }

    /**
     * Test if we have active transactions on the PersistenceManagers
     * @return boolean returns true if we have active transactions
     */
    private synchronized boolean hasActiveTransactions()
    {
        Iterator pms = getPmCache().iterator();
        while (pms.hasNext())
        {
            AbstractPersistenceManager pm = (AbstractPersistenceManager)pms.next();
            // if pm not is closed and we have active transactions
            if (!pm.isClosed() && pm.currentTransaction().isActive())
            {
                return true;
            }
        }
        return false;
    }

    /**
     * Get a PersistenceManager from cache.
     * @param userName  the user name for the connection
     * @param password  the password for the connection
     * @return Returns a PersistenceManager.
     */ 
    protected synchronized PersistenceManager getPersistenceManagerInternal(String userName, String password)
    {
        return getPersistenceManagerFromCache(userName,password);
    }

    /**
     * Remove a PersistenceManager from the cache
     * Only the PersistenceManager is allowed to call this method
     * @param pm  the PersistenceManager to be removed from cache
     */
    public void releasePersistenceManager(PersistenceManager pm)
    {
        getPmCache().remove(pm);
    }

    /**
     * Close this PersistenceManagerFactory. Check for
     * JDOPermission("closePersistenceManagerFactory") and if not authorized,
     * throw SecurityException.
     * <P>If the authorization check succeeds, check to see that all
     * PersistenceManager instances obtained from this PersistenceManagerFactory
     * have no active transactions. If any PersistenceManager instances have
     * an active transaction, throw a JDOUserException, with one nested
     * JDOUserException for each PersistenceManager with an active Transaction.
     * <P>If there are no active transactions, then close all PersistenceManager
     * instances obtained from this PersistenceManagerFactory, mark this
     * PersistenceManagerFactory as closed, disallow getPersistenceManager
     * methods, and allow all other get methods. If a set method or
     * getPersistenceManager method is called after close, then
     * JDOUserException is thrown.
     * 
     * @see javax.jdo.PersistenceManagerFactory#close()
     */
    public void close()
    {
        SecurityManager secmgr = System.getSecurityManager();
        if (secmgr != null)
        {
            // checkPermission will throw SecurityException if not authorized
            secmgr.checkPermission(JDOPermission.CLOSE_PERSISTENCE_MANAGER_FACTORY);
        }
        // if we have active transactions, cant close PMF
        if (hasActiveTransactions())
        {
            throw new JDOUserException(LOCALISER.msg("PMF.CloseError"));
        }
        // iterate though the list of pms to release resources
        Iterator pms = getPmCache().iterator();
        while (pms.hasNext())
        {
            ((AbstractPersistenceManager)pms.next()).close();
        }

        // Clear the Cache
        if (getJdoCacheLevel2() && cache != null)
        {
            cache.clear();
            JPOXLogger.CACHE.info(LOCALISER.msg("Cache.Level2.Cleared"));
        }

        closed = true;
    }

    /**
     * Utility to return whether the PMF is closed or not.
     * @return Whether it is closed.
     */
    public boolean isClosed()
    {
        return closed;
    }

    /**
     * Gets the context for this PersistenceManagerFactory
     * @return Returns the pmfContext.
     */
    public PMFContext getPMFContext()
    {
        return pmfContext;
    }

    /**
     * Accessor for the PMFConfiguration.
     * @return Returns the PMF config.
     */
    public PMFConfiguration getPMFConfiguration()
    {
        return this;
    }

    /**
     * Accessor for the DataStore (level 2) Cache
     * @return The datastore cache
     * @since 1.1
     */
    public DataStoreCache getDataStoreCache()
    {
        freezeConfiguration();

        return cache;
    }

    // -------------------------------- Lifecycle Listeners -------------------------------

    /**
     * Method to add lifecycle listeners for particular classes.
     * Adds the listener to all PMs already created.
     * @param listener The listener
     * @param classes The classes that the listener relates to
     * @since 1.1
     */
    public void addLifecycleListener(InstanceLifecycleListener listener, Class[] classes)
    {
        if (listener == null)
        {
            return;
        }

        // Add to any PMs
        Iterator pms = getPmCache().iterator();
        while (pms.hasNext())
        {
            AbstractPersistenceManager pm = (AbstractPersistenceManager)pms.next();
            pm.addLifecycleListener(listener, classes);
        }

        // Add to the PMF
        lifecycleListeners.add(new LifecycleListenerSpecification(listener, classes));
    }

    /**
     * Method to remove a lifecycle listener. Removes the listener from all PM's as well.
     * @param listener The Listener
     * @since 1.1
     */
    public void removeLifecycleListener(InstanceLifecycleListener listener)
    {
        if (listener != null)
        {
            // Remove from any PMs
            Iterator pms = getPmCache().iterator();
            while (pms.hasNext())
            {
                AbstractPersistenceManager pm = (AbstractPersistenceManager)pms.next();
                pm.removeLifecycleListener(listener);
            }

            // Remove from the PMF
            Iterator iter = lifecycleListeners.iterator();
            while (iter.hasNext())
            {
                LifecycleListenerSpecification classListener = (LifecycleListenerSpecification) iter.next();
                if (classListener.getListener() == listener)
                {
                    iter.remove();
                }
            }
        }
    }
}