/**********************************************************************
Copyright (c) 2002 Kelly Grizzle and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - commented
    ...
**********************************************************************/
package org.jpox;

import java.io.PrintWriter;
import java.sql.Connection;
import java.sql.DriverManager;
import java.sql.SQLException;

import javax.jdo.JDOFatalUserException;
import javax.sql.DataSource;

import org.jpox.util.Localiser;

/**
 * Wrapper to the JDBC DataSource class. Provides checking for driver class
 * existence, and utility methods for obtaining a connection.
 * <P>
 * It should be noted that setting the log writer and login timeout will apply
 * to DriverManager and NOT to the Data Source on its own. If you have 2 or
 * more DataSource's they will have THE SAME log writer and login timeout. 
 * </P> 
 *
 * @version $Revision: 1.1 $
 **/ 
public class DriverManagerDataSource implements DataSource
{
    private static final Localiser LOCALISER=Localiser.getInstance("org.jpox.Localisation");

    private final String driverName;
    private final String url;

    /** Constructor.
     * @param   driverName  Class name of the JDBC driver.
     * @param   url         URL of the data source.
     **/
    public DriverManagerDataSource(String driverName, String url)
    {
        this.driverName = driverName;
        this.url        = url;

        if (driverName != null)
        {
            try
            {
                Class.forName(driverName).newInstance();
            }
            catch (Exception e)
            {
                throw new JDOFatalUserException(LOCALISER.msg("Driver.InvalidClass",driverName), e);
            }
        }
    }

    /** Accessor for a JDBC connection for this data source.
     * @return  The connection
     * @throws  SQLException    Thrown when an error occurs obtaining the
     * connection. */
    public Connection getConnection() throws SQLException
    {
        return DriverManager.getConnection(url);
    }

    /** Accessor for a JDBC connection for this data source, specifying
     * username and password.
     * @param   userName    User name for the data source
     * @param   password    Password for the data source
     * @return  The connection
     * @throws  SQLException    Thrown when an error occurs obtaining the
     * connection. */
    public Connection getConnection(String userName, String password)
    throws SQLException
    {
        return DriverManager.getConnection(url, userName, password);
    }

    /** Accessor for the LogWriter of the driver manager.
     * @return  The Log Writer
     */
    public PrintWriter getLogWriter()
    {
        return DriverManager.getLogWriter();
    }

    /** Mutator for the LogWriter of the driver manager.
     * @param   out The Log Writer
     */
    public void setLogWriter(PrintWriter out)
    {
        DriverManager.setLogWriter(out);
    }

    /** Accessor for the Login timeout for the driver manager.
     * @return  The login timeout (seconds)
     */
    public int getLoginTimeout()
    {
        return DriverManager.getLoginTimeout();
    }

    /** Mutator for the Login timeout for the driver manager.
     * @param  seconds  The login timeout (seconds)
     */
    public void setLoginTimeout(int seconds)
    {
        DriverManager.setLoginTimeout(seconds);
    }

    /** Equality operator.
     * @param   obj The object to compare against.
     * @return  Whether the objects are equal. */
    public boolean equals(Object obj)
    {
        if (obj == this)
        {
            return true;
        }

        if (!(obj instanceof DriverManagerDataSource))
        {
            return false;
        }

        DriverManagerDataSource dmds = (DriverManagerDataSource)obj;

        if (driverName == null)
        {
            if (dmds.driverName != null)
            {
                return false;
            }
        }
        else if (!driverName.equals(dmds.driverName))
        {
            return false;
        }

        if (url == null)
        {
            if (dmds.url != null)
            {
                return false;
            }
        }
        else if (!url.equals(dmds.url))
        {
            return false;
        }

        return true;
    }

    /** Hashcode operator.
     * @return  The Hashcode for this object. */
    public int hashCode()
    {
        return (driverName == null ? 0 : driverName.hashCode()) ^
               (url        == null ? 0 : url.hashCode());
    }
}
