/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved.
This program and the accompanying materials are made available under
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
    ...
**********************************************************************/
package org.jpox;

import java.io.Serializable;
import java.util.BitSet;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;

import javax.jdo.FetchPlan;
import javax.jdo.JDOFatalUserException;

import org.jpox.metadata.ClassMetaData;
import org.jpox.metadata.FetchGroupMetaData;
import org.jpox.metadata.FieldPersistenceModifier;

/**
 * This class implements the FetchPlan interface. In addition to the FetchPlan
 * interface implementation, this class updates and provides for each class
 * managed the fields correspondent to the actual fetch plan. A fetch group
 * defines a particular loaded state for an object graph. It specifies fields to
 * be loaded for all of the instances in the graph. Fetch groups are activated
 * using methods on an interface called FetchPlan. FetchPlan interface is
 * provided to user and enables him add, remove or reset the active groups in
 * the fetch plan
 * @since 1.1
 * @version $Revision: 1.7 $
 */
public class FetchPlanImpl implements FetchPlan, Serializable
{
    final Set groups = new HashSet();

    int fetchSize = 0;

    final Map managedClass = new HashMap();

    /**
     * Constructor. Starts with the default fetch group.
     */
    public FetchPlanImpl()
    {
        super();
        groups.add(FetchPlan.DEFAULT);
    }

    /**
     * Manage the list of active fields for the FetchPlan
     */
    public class FetchPlanForClass
    {
        final ClassMetaData cmd;

        int[] fieldsInActualFetchPlan;

        boolean dirty = true;

        /**
         * Constructor
         * @param cmd the ClassMetaData
         */
        public FetchPlanForClass(final ClassMetaData cmd)
        {
            super();
            this.cmd = cmd;
        }

        void markDirty()
        {
            dirty = true;
        }

        /**
         * Get all fields in the actual fetch plan
         * @return an array with the absolute position of the fields
         */
        public int[] getFieldsInActualFetchPlan()
        {
            if (dirty)
            {
                dirty = false;
                BitSet fieldsNumber = getFieldsInActualFetchPlanByBitSet();
                int countFieldsInFP = 0;
                for (int i = 0; i < fieldsNumber.length(); i++)
                {
                    if (fieldsNumber.get(i))
                    {
                        countFieldsInFP++;
                    }
                }

                fieldsInActualFetchPlan = new int[countFieldsInFP];
                int nextField = 0;
                for (int i = 0; i < fieldsNumber.length(); i++)
                {
                    if (fieldsNumber.get(i))
                    {
                        fieldsInActualFetchPlan[nextField++] = i;
                    }
                }
            }
            return fieldsInActualFetchPlan;
        }

        /**
         * Get all fields in the actual fetch plan
         * @return an BitSet with the bits set in the absolute position of the
         * fields
         */
        public BitSet getFieldsInActualFetchPlanByBitSet()
        {
            return getFieldsInActualFetchPlanByBitSet(cmd);
        }

        /**
         * Get all fields in the actual fetch plan for this class and super
         * classes
         * @param cmd this ClassMetaData
         * @return an BitSet with the bits set in the absolute position of the
         * fields
         */
        private BitSet getFieldsInActualFetchPlanByBitSet(ClassMetaData cmd)
        {
            BitSet bitSet = getFetchPlanForClass(cmd).getFieldsAbsoluteNumber(cmd.getFetchGroupMetaData());
            if (cmd.getPersistenceCapableSuperclass() != null)
            {
                manageFetchPlanForClass(cmd.getSuperclassMetaData());
                bitSet.or(getFetchPlanForClass(cmd.getSuperclassMetaData()).getFieldsInActualFetchPlanByBitSet(cmd.getSuperclassMetaData()));
            }
            return bitSet;
        }

        /**
         * Get the absolute number of the fields for an array of Fetch Group
         * @param fgmds The Fetch Groups
         * @return
         */
        private BitSet getFieldsAbsoluteNumber(FetchGroupMetaData[] fgmds)
        {
            BitSet fieldsNumber = new BitSet(0);
            if (fgmds != null)
            {
                for (int i = 0; i < fgmds.length; i++)
                {
                    if (groups.contains(fgmds[i].getName()))
                    {
                        fieldsNumber.or(getFieldsAbsoluteNumberInFetchGroup(fgmds[i]));
                    }
                }
            }
            if (groups.contains(FetchPlan.DEFAULT))
            {
                setDefaultFieldNumbers(fieldsNumber);
            }
            if (groups.contains(FetchPlan.ALL))
            {
                setAllFieldNumbers(fieldsNumber);
            }
            if (groups.contains(FetchPlan.NONE))
            {
                setNoneFieldNumbers(fieldsNumber);
            }
            return fieldsNumber;
        }

        /**
         * Sets the given list of field numbers to include all the fields
         * defined in the DEFAULT FetchPlan.
         * @param fieldsNumber list of field numbers
         */
        private void setDefaultFieldNumbers(BitSet fieldsNumber)
        {
            for (int i = 0; i < cmd.getDefaultFetchGroupFieldNumbers().length; i++)
            {
                fieldsNumber.set(cmd.getDefaultFetchGroupFieldNumbers()[i]);
            }
        }

        /**
         * Sets the given list of field numbers to include all the fields
         * defined in the ALL FetchPlan.
         * @param fieldsNumber list of field numbers
         */
        private void setAllFieldNumbers(BitSet fieldsNumber)
        {
            for (int i = 0; i < cmd.getNoOfManagedFields(); i++)
            {
                if (cmd.getManagedField(i).getPersistenceModifier() != FieldPersistenceModifier.NONE)
                {
                    fieldsNumber.set(cmd.getAbsoluteFieldNumberForRelativeFieldNumber(i));
                }
            }
        }

        /**
         * Sets the given list of field numbers to include all the fields
         * defined in the NONE FetchPlan.
         * @param fieldsNumber list of field numbers
         */
        private void setNoneFieldNumbers(BitSet fieldsNumber)
        {
            for (int i = 0; i < cmd.getNoOfManagedFields(); i++)
            {
                if (cmd.getManagedFieldAbsolute(cmd.getAbsoluteFieldNumberForRelativeFieldNumber(i)).isPrimaryKey())
                {
                    fieldsNumber.set(cmd.getAbsoluteFieldNumberForRelativeFieldNumber(i));
                }
            }
        }

        /**
         * Get the absolute field number for a particular Fetch Group
         * @param fgmd The Fetch Group
         * @return a list of field numbers
         */
        private BitSet getFieldsAbsoluteNumberInFetchGroup(FetchGroupMetaData fgmd)
        {
            BitSet fieldsNumber = new BitSet(0);
            for (int i = 0; i < fgmd.getFieldMetaData().length; i++)
            {
                int fieldNumber = getFieldNumber(cmd, fgmd.getFieldMetaData()[i].getName());
                if (fieldNumber == -1)
                {
                    throw new JDOFatalUserException(
                            "Field " + fgmd.getFieldMetaData()[i].getName() + " declared in the fetch-group " + fgmd.getName() + " was not found in class " + cmd
                                    .getFullClassName());
                }
                fieldsNumber.set(fieldNumber);
            }
            // fields in nested fetch-groups
            for (int i = 0; i < fgmd.getFetchGroupMetaData().length; i++)
            {
                String nestedGroupName = fgmd.getFetchGroupMetaData()[i].getName();
                if (nestedGroupName.equals(FetchPlan.DEFAULT)) 
                {
                    setDefaultFieldNumbers(fieldsNumber);
                }
                else if(nestedGroupName.equals(FetchPlan.ALL)) 
                {
                    setAllFieldNumbers(fieldsNumber);
                }
                else if(nestedGroupName.equals(FetchPlan.NONE)) 
                {
                    setNoneFieldNumbers(fieldsNumber);
                }
                else
                {
                    FetchGroupMetaData nestedFGMD = null;
                    for (int j = 0; j < cmd.getFetchGroupMetaData().length; j++)
                    {
                        if (cmd.getFetchGroupMetaData()[j].getName().equals(nestedGroupName))
                        {
                            nestedFGMD = cmd.getFetchGroupMetaData()[j];
                        }
                    }
                    if (nestedFGMD == null)
                    {
                        throw new JDOFatalUserException("Fetch-group " + fgmd.getFetchGroupMetaData()[i].getName() +
                            " nested in the fetch-group " + fgmd.getName() + " was not found in class " + cmd.getFullClassName());
                    }
                    fieldsNumber.or(getFieldsAbsoluteNumberInFetchGroup(nestedFGMD));
                }
            }
            return fieldsNumber;
        }

        /**
         * gets the field number for a fieldname in this or superclasses
         * @param cmd this classMetaData
         * @param fieldName the field Name
         * @return the field number. -1 if not found
         */
        private int getFieldNumber(ClassMetaData cmd, String fieldName)
        {
            int fieldNumber = cmd.getFieldNumberAbsolute(fieldName);
            if (fieldNumber == -1 && cmd.getPersistenceCapableSuperclass() != null)
            {
                fieldNumber = getFieldNumber(cmd.getSuperclassMetaData(), fieldName);
            }
            return fieldNumber;
        }

        /**
         * Get all the fetch groups this field is included
         * @param fgmds The Fetch Groups
         * @param absoluteFieldNumber the field absolute number
         * @return The Fetch Groups
         */
        private Set getFetchGroupsForFieldAbsoluteNumber(FetchGroupMetaData[] fgmds, int absoluteFieldNumber)
        {
            Set fetchGroups = new HashSet();
            if (fgmds != null)
            {
                for (int i = 0; i < fgmds.length; i++)
                {
                    for (int j = 0; j < fgmds[i].getFieldMetaData().length; j++)
                    {
                        if (fgmds[i].getFieldMetaData()[j].getName().equals(cmd.getManagedFieldAbsolute(absoluteFieldNumber).getName()))
                        {
                            fetchGroups.add(fgmds[i]);
                        }
                    }
                    for (int j = 0; j < fgmds[i].getFetchGroupMetaData().length; j++)
                    {
                        fetchGroups.addAll(getFetchGroupsForFieldAbsoluteNumber(fgmds[i].getFetchGroupMetaData(), absoluteFieldNumber));
                    }
                }
            }
            return fetchGroups;
        }

        /**
         * Whether to call the post load or not. Checks if fields in actual
         * FetchPlan where not previouly loaded and the post-load is enabled in
         * the metadata
         * @param loadedFields alredy loaded fields
         * @return if is to call the postLoad
         */
        public boolean isToCallPostLoadFetchPlan(boolean[] loadedFields)
        {
            for (int i = 0; i < getFieldsInActualFetchPlan().length; i++)
            {
                //if field in actual fetch plan was not previously loaded
                if (!loadedFields[getFieldsInActualFetchPlan()[i]])
                {
                    //to call jdoPostLoad, field must be in default-fetch-group
                    // when DFG is active
                    if (cmd.getManagedFieldAbsolute(getFieldsInActualFetchPlan()[i]).isDefaultFetchGroup() && groups.contains(FetchPlan.DEFAULT))
                    {
                        return true;
                    }
                    //or, field must be in a fetch-group which has post-load
                    // set to true
                    else
                    {
                        Set fetchGroups = getFetchGroupsForFieldAbsoluteNumber(cmd.getFetchGroupMetaData(), getFieldsInActualFetchPlan()[i]);
                        for (Iterator it = fetchGroups.iterator(); it.hasNext();)
                        {
                            FetchGroupMetaData fgmd = (FetchGroupMetaData) it.next();
                            if (fgmd.getPostLoad().booleanValue())
                            {
                                return true;
                            }
                        }
                    }

                }
            }
            return false;
        }
    }

    /**
     * Manage the fetch plan for the class
     * @param cmd ClassMetaData for the class to manage
     * @return the FetchPlanForClass
     */
    public FetchPlanForClass manageFetchPlanForClass(ClassMetaData cmd)
    {
        FetchPlanForClass fetchPlanForClass = (FetchPlanForClass) managedClass.get(cmd);
        if (fetchPlanForClass == null)
        {
            fetchPlanForClass = new FetchPlanForClass(cmd);
            managedClass.put(cmd, fetchPlanForClass);
        }
        return fetchPlanForClass;
    }

    /**
     * Mark all managed fetch plans to be dirty, so they fields active needs to
     * be recomputed.
     */
    private void markDirty()
    {
        if (groups.size() < 1)
        {
            groups.add(FetchPlan.DEFAULT);
        }
        Iterator it = managedClass.values().iterator();
        while (it.hasNext())
        {
            ((FetchPlanForClass) it.next()).markDirty();
        }
    }

    /**
     * Manage the fetch plan for the class
     * @param cmd the ClassMetaData
     */
    public FetchPlanForClass getFetchPlanForClass(ClassMetaData cmd)
    {
        if (managedClass.get(cmd) != null)
        {
            return (FetchPlanForClass) managedClass.get(cmd);
        }
        return null;
    }

    /*
     * (non-Javadoc)
     * @see javax.jdo.FetchPlan#addGroup(java.lang.String)
     */
    public FetchPlan addGroup(String fetchGroupName)
    {
        if (fetchGroupName != null)
        {
            if (groups.add(fetchGroupName))
            {
                markDirty();
            }
        }
        return this;
    }

    /*
     * (non-Javadoc)
     * @see javax.jdo.FetchPlan#removeGroup(java.lang.String)
     */
    public FetchPlan removeGroup(String fetchGroupName)
    {
        if (groups.remove(fetchGroupName))
        {
            markDirty();
        }
        return this;
    }

    /**
     * Method to clear the current groups, including the DFG.
     * @return The FetchPlan
     */
    public FetchPlan clearGroups()
    {
        groups.clear();
        markDirty();
        return this;
    }

    /**
     * Performs same action as clearGroups and this is deprecated
     * @deprecated Use clearGroups() instead
     */
    public FetchPlan resetGroups()
    {
        groups.clear();
        markDirty();
        return this;
    }

    /*
     * (non-Javadoc)
     * @see javax.jdo.FetchPlan#getGroups()
     */
    public Collection getGroups()
    {
        return groups;
    }

    /*
     * (non-Javadoc)
     * @see javax.jdo.FetchPlan#setGroups(java.util.Collection)
     */
    public FetchPlan setGroups(Collection fetchGroupNames)
    {
        groups.clear();
        if (fetchGroupNames != null)
        {
            groups.addAll(fetchGroupNames);
        }
        markDirty();
        return this;
    }

    /**
     * Method to set the groups using an array.
     * @param fetchGroupNames Names of the fetch groups
     * @return The Fetch Plan
     */
    public FetchPlan setGroups(String[] fetchGroupNames)
    {
        groups.clear();
        if (fetchGroupNames != null && fetchGroupNames.length > 0)
        {
            for (int i=0;i<fetchGroupNames.length;i++)
            {
                groups.add(fetchGroupNames[i]);
            }
        }
        markDirty();
        return this;
    }

    /**
     * Method to set the fetch group.
     * @param fetchGroupName Name of the fetch group
     * @return The Fetch Plan
     */
    public FetchPlan setGroup(String fetchGroupName)
    {
        groups.clear();
        if (fetchGroupName != null)
        {
            groups.add(fetchGroupName);
        }
        markDirty();
        return this;
    }

    /*
     * (non-Javadoc)
     * @see javax.jdo.FetchPlan#setFetchSize(int)
     */
    public FetchPlan setFetchSize(int fetchSize)
    {
        if (fetchSize != FETCH_SIZE_GREEDY && fetchSize != FETCH_SIZE_OPTIMAL && fetchSize < 0)
        {
            // Invalid fetch size so just return
            return this;
        }
        this.fetchSize = fetchSize;
        return this;
    }

    /*
     * (non-Javadoc)
     * @see javax.jdo.FetchPlan#getFetchSize()
     */
    public int getFetchSize()
    {
        return fetchSize;
    }

    /*
     * (non-Javadoc)
     * @see java.lang.Object#toString()
     */
    public String toString()
    {
        return groups.toString();
    }
    
    /**
     * Returns a copy of this FetchPlan with all settings initialized
     * @return the FetchPlan
     */
    public FetchPlan getCopy()
    {
        FetchPlanImpl fp = new FetchPlanImpl();
        fp.groups.addAll(this.groups);
        fp.managedClass.putAll(this.managedClass);
        fp.fetchSize = this.fetchSize;
        return fp;
    }
}