/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved.
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2004 Erik Bengtson - added extra classForName method
    ...
**********************************************************************/
package org.jpox;

import javax.jdo.JDOException;

import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;

/**
 * A basic implementation of a ClassLoaderResolver for JDO. A ClassLoaderResolver
 * provides a series of methods for resolving classes from their names. 
 * It supports up to 3 class loaders.
 *
 * <H3>JDO2 Spec section 12.5</H3>
 * The class loaders will be used in this order:
 * <OL>
 * <LI>The loader that loaded the class or instance referred to in the API that caused this class to be loaded.</LI>
 * <UL>
 * <LI>In case of query, this is the loader of the candidate class, or the loader of the object passed to the newQuery</LI>
 * <LI>In case of navigation from a persistent instance, this is the loader of the class of the instance.</LI>
 * <LI>In the case of getExtent with subclasses, this is the loader of the candidate class.</LI>
 * <LI>In the case of getObjectById, this is the loader of the object id instance.</LI>
 * <LI>Other cases do not have an explicit loader.</LI>
 * </UL>
 * <LI>The loader returned in the current context by Thread.getContextClassLoader().</LI>
 * <LI>The loader returned by Thread.getContextClassLoader() at the time of PersistenceManagerFactory.getPersistenceManager().</LI>
 * </OL>
 *
 * @version $Revision: 1.2 $
 */
public class JDOClassLoaderResolver implements ClassLoaderResolver
{
    /** Localisation utility for output messages */
    protected static final Localiser LOCALISER = Localiser.getInstance("org.jpox.Localisation");

    /** ClassLoader initialised by the PM creation. */
    protected final ClassLoader pmContextLoader;

    /**
     * Constructor for PersistenceManager cases.
     * @param pmLoader Loader from PM initialisation time.
     */
    public JDOClassLoaderResolver(ClassLoader pmLoader)
    {
        pmContextLoader = pmLoader;
    }

    /**
     * Constructor for non-PersistenceManager cases so there is no
     * PM context loader.
     */
    public JDOClassLoaderResolver()
    {
        pmContextLoader = null;
    }

    /**
     * JDO's Class Loading mechanism (Spec 1.0.1 Chapter 12.5).
     * Try 3 loaders, starting with user-supplied loader, then try
     * the current thread's loader, and finally try the PM context
     * loader.
     * @param name Name of the Class to be loaded
     * @param primary primary ClassLoader to use (or null)
     * @return The class given the name, using the required loader.
     */
    public Class classForName(String name, ClassLoader primary)
    throws ClassNotFoundException
    {
        // Try the supplied loader first
        Class cls = classOrNull(name, primary);

        // Try the loader for the current thread
        if (cls == null)
        {
            cls = classOrNull(name, Thread.currentThread().getContextClassLoader());
        }

        // Try the loader for the PM context
        if (cls == null)
        {
            cls = classOrNull(name, pmContextLoader);
        }

        // Fallback to the default loader for Class - is this required ?
        if (cls == null)
        {
            cls = Class.forName(name);
        }

        return cls;
    }

    /**
     * JDO's Class Loading mechanism (Spec 1.0.1 Chapter 12.5)
     * @param name Name of the Class to be loaded
     * @param primary the primary ClassLoader to use (or null)
     * @param initialize whether to initialize the class or not.
     * @return The Class given the name, using the specified ClassLoader
     */
    public Class classForName(String name, ClassLoader primary, boolean initialize)
    throws ClassNotFoundException
    {
        Class cls = classForName(name, primary);
        if (cls != null)
        {
            // We need to initialise the class, so use the specific Class method
            Class.forName(cls.getName(), true, cls.getClassLoader());
        }
        return cls;
    }

    /**
     * JDO's Class Loading mechanism (Spec 1.0.1 Chapter 12.5)
     * @param name Name of the Class to be loaded
     * @return The class given the name, using the required loader.
     */
    public Class classForName(String name)
    {
        try
        {
            return classForName(name, null);
        }
        catch (ClassNotFoundException e)
        {
            String msg = LOCALISER.msg("ClassLoader.ClassNotFound", name);
            JPOXLogger.GENERAL.error(msg);
            throw new JDOException(msg);
        }
    }

    /**
     * JDO's Class Loading mechanism (Spec 1.0.1 Chapter 12.5)
     * @param name Name of the Class to be loaded
     * @param initialize whether to initialize the class or not.
     * @return The Class given the name, using the specified ClassLoader
     */
    public Class classForName(String name, boolean initialize)
    {
        try
        {
            return classForName(name, null, initialize);
        }
        catch (ClassNotFoundException e)
        {
            String msg = LOCALISER.msg("ClassLoader.ClassNotFound", name);
            JPOXLogger.GENERAL.error(msg);
            throw new JDOException(msg);
        }
    }

    /**
     * Utility to check the assignability of 2 classes in accordance with JDO's
     * Class Loading mechanism. This will check
     * <I>class_1.isAssignableFrom(class_2);</I>
     * @param class_name_1 Name of first class
     * @param class_name_2 Name of second class
     * @return Whether Class 2 is assignable from Class 1
     */
    public boolean isAssignableFrom(String class_name_1, String class_name_2)
    {
        if (class_name_1 == null || class_name_2 == null)
        {
            return false;
        }

        Class class_1 = classForName(class_name_1);
        Class class_2 = classForName(class_name_2);

        return class_1.isAssignableFrom(class_2);
    }

    /**
     * Utility to check the assignability of 2 classes in accordance with JDO's
     * Class Loading mechanism. This will check
     * <I>class_1.isAssignableFrom(class_2);</I>
     * @param class_name_1 Name of first class
     * @param class_2 Second class
     * @return Whether Class 2 is assignable from Class 1
     */
    public boolean isAssignableFrom(String class_name_1, Class class_2)
    {
        if (class_name_1 == null || class_2 == null)
        {
            return false;
        }

        try
        {
            Class class_1 = null;
            if (class_2.getClassLoader() != null)
            {
                // Use class_2's loader if possible
                class_1 = class_2.getClassLoader().loadClass(class_name_1);
            }
            else
            {
                // Use the boot class loader
                class_1 = Class.forName(class_name_1);
            }
            return class_1.isAssignableFrom(class_2);
        }
        catch (Exception e)
        {
            return false;
        }
    }

    /**
     * Utility to check the assignability of 2 classes in accordance with JDO's
     * Class Loading mechanism. This will check
     * <I>class_1.isAssignableFrom(class_2);</I>
     * @param class_1 First class
     * @param class_name_2 Name of second class
     * @return Whether Class 2 is assignable from Class 1
     */
    public boolean isAssignableFrom(Class class_1, String class_name_2)
    {
        if (class_1 == null || class_name_2 == null)
        {
            return false;
        }

        try
        {
            Class class_2 = null;
            if (class_1.getClassLoader() != null)
            {
                // Use class_1's loader if possible
                class_2 = class_1.getClassLoader().loadClass(class_name_2);
            }
            else
            {
                // Use the boot class loader
                class_2 = Class.forName(class_name_2);
            }
            return class_1.isAssignableFrom(class_2);
        }
        catch (Exception e)
        {
            return false;
        }
    }

    /**
     * Returns the Class
     * @param name the class name
     * @param loader the ClassLoader
     * @return Class the class loaded; null if not found
     */
    private Class classOrNull(String name, ClassLoader loader)
    {
        try
        {
            return loader == null ? null : loader.loadClass(name);
        }
        catch (ClassNotFoundException cnfe)
        {
            return null;
        }
    }
}