/**********************************************************************
Copyright (c) 2005 Erik Bengtson and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
    ...
**********************************************************************/
package org.jpox;

import java.sql.CallableStatement;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.sql.SQLWarning;
import java.sql.Savepoint;
import java.sql.Statement;
import java.util.Map;

import javax.jdo.JDOConnection;
import javax.jdo.JDOUserException;


final public class JDOConnectionImpl implements Connection, JDOConnection
{
    /** the connection **/
    private final Connection conn;
    
    /** run onClose on close call */
    private final Runnable onClose;
    
    /** whether this connection is available to the developer */
    private boolean isAvailable = true;

    /**
     * @param conn the connection
     * @param onClose the Runnable operation that will be called at user call to close()
     */
    public JDOConnectionImpl(Connection conn, Runnable onClose)
    {
        this.conn = conn;
        this.onClose = onClose;
    }

    public int getHoldability() throws SQLException
    {
        assertAvailable();
        return this.conn.getHoldability();
    }

    public int getTransactionIsolation() throws SQLException
    {
        assertAvailable();
        return this.conn.getTransactionIsolation();
    }

    public void clearWarnings() throws SQLException
    {
        assertAvailable();
        this.conn.clearWarnings();
    }

    public void close() throws SQLException
    {
        assertAvailable();
        isAvailable = false;
        onClose.run();
    }

    public void commit() throws SQLException
    {
        throw new JDOUserException("Invalid call to commit.");	
    }

    public void rollback() throws SQLException
    {
        throw new JDOUserException("Invalid call to roolback.");	
    }

    public boolean getAutoCommit() throws SQLException
    {
        assertAvailable();
        return this.conn.getAutoCommit();
    }

    public boolean isClosed() throws SQLException
    {
        assertAvailable();
        return this.conn.isClosed();
    }

    public boolean isReadOnly() throws SQLException
    {
        assertAvailable();
        return this.conn.isReadOnly();
    }

    public void setHoldability(int arg0) throws SQLException
    {
        throw new JDOUserException("Invalid call to setHoldability.");	
    }

    public void setTransactionIsolation(int level) throws SQLException
    {
        throw new JDOUserException("Invalid call to setTransactionIsolation.");	
    }

    public void setAutoCommit(boolean autoCommit) throws SQLException
    {
        throw new JDOUserException("Invalid call to setAutoCommit.");	
    }

    public void setReadOnly(boolean readOnly) throws SQLException
    {
        throw new JDOUserException("Invalid call to setReadOnly.");	
    }

    public String getCatalog() throws SQLException
    {
        assertAvailable();
        return this.conn.getCatalog();
    }

    public void setCatalog(String catalog) throws SQLException
    {
        throw new JDOUserException("Invalid call to setCatalog.");	
    }

    public DatabaseMetaData getMetaData() throws SQLException
    {
        throw new JDOUserException("Invalid call to getMetaData.");	
    }

    public SQLWarning getWarnings() throws SQLException
    {
        assertAvailable();
        return this.conn.getWarnings();
    }

    public Savepoint setSavepoint() throws SQLException
    {
        throw new JDOUserException("Invalid call to setSavepoint.");	
    }

    public void releaseSavepoint(Savepoint arg0) throws SQLException
    {
        throw new JDOUserException("Invalid call to releaseSavepoint.");	
    }

    public void rollback(Savepoint arg0) throws SQLException
    {
        throw new JDOUserException("Invalid call to rollback.");	
    }

    public Statement createStatement() throws SQLException
    {
        assertAvailable();
        return this.conn.createStatement();
    }

    public Statement createStatement(int resultSetType, int resultSetConcurrency) throws SQLException
    {
        assertAvailable();
        return this.conn.createStatement(resultSetType,resultSetConcurrency);
    }

    public Statement createStatement(int arg0, int arg1, int arg2) throws SQLException
    {
        assertAvailable();
        return this.conn.createStatement(arg0,arg1,arg2);
    }

    public Map getTypeMap() throws SQLException
    {
        assertAvailable();
        return this.conn.getTypeMap();
    }

    public void setTypeMap(Map map) throws SQLException
    {
        throw new JDOUserException("Invalid call to setTypeMap.");	
    }

    public String nativeSQL(String sql) throws SQLException
    {
        assertAvailable();
        return this.conn.nativeSQL(sql);
    }

    public CallableStatement prepareCall(String sql) throws SQLException
    {
        assertAvailable();
        return this.conn.prepareCall(sql);
    }

    public CallableStatement prepareCall(String sql, int resultSetType, int resultSetConcurrency) throws SQLException
    {
        assertAvailable();
        return this.conn.prepareCall(sql,resultSetType,resultSetConcurrency);
    }

    public CallableStatement prepareCall(String arg0, int arg1, int arg2, int arg3) throws SQLException
    {
        assertAvailable();
        return this.conn.prepareCall(arg0,arg1,arg2,arg3);
    }

    public PreparedStatement prepareStatement(String sql) throws SQLException
    {
        assertAvailable();
        return this.conn.prepareStatement(sql);
    }

    public PreparedStatement prepareStatement(String arg0, int arg1) throws SQLException
    {
        assertAvailable();
        return this.conn.prepareStatement(arg0,arg1);
    }

    public PreparedStatement prepareStatement(String sql, int resultSetType, int resultSetConcurrency) throws SQLException
    {
        assertAvailable();
        return this.conn.prepareStatement(sql,resultSetType,resultSetConcurrency);
    }

    public PreparedStatement prepareStatement(String arg0, int arg1, int arg2, int arg3) throws SQLException
    {
        assertAvailable();
        return this.conn.prepareStatement(arg0,arg1,arg2,arg3);
    }

    public PreparedStatement prepareStatement(String arg0, int[] arg1) throws SQLException
    {
        assertAvailable();
        return this.conn.prepareStatement(arg0,arg1);
    }

    public Savepoint setSavepoint(String arg0) throws SQLException
    {
        throw new JDOUserException("Invalid call to setSavepoint.");	
    }

    public PreparedStatement prepareStatement(String arg0, String[] arg1) throws SQLException
    {
        assertAvailable();
        return this.conn.prepareStatement(arg0,arg1);
    }
    
    public void assertAvailable()
    {
        if( !isAvailable )
        {
            throw new JDOUserException("JDOConnection is no longer available.");	
        }
    }

    /* (non-Javadoc)
     * @see javax.jdo.JDOConnection#getNativeConnection()
     */
    public Object getNativeConnection()
    {
        return this;
    }
}