/**********************************************************************
Copyright (c) 2002 Kelly Grizzle and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - commented
2003 Erik Bengtson - added Optimistic Transactions
2003 Andy Jefferson - converted to use Logger
2004 Andy Jefferson - addition of rollbackOnly
2004 Andy Jefferson - updated useUpdate treatment to ignore when on Optimistic
    ...
**********************************************************************/
package org.jpox;

import java.sql.Connection;
import java.sql.SQLException;

import javax.jdo.JDOConnection;
import javax.jdo.JDODataStoreException;
import javax.jdo.JDOFatalDataStoreException;
import javax.jdo.JDOUnsupportedOptionException;
import javax.jdo.JDOUserException;
import javax.jdo.PersistenceManager;
import javax.sql.DataSource;
import javax.transaction.Status;
import javax.transaction.Synchronization;

import org.jpox.exceptions.ConnectionInUseException;
import org.jpox.exceptions.TransactionActiveException;
import org.jpox.exceptions.TransactionNotActiveException;
import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.store.rdbms.adapter.DatabaseAdapter;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;


/**
 * Implementation of a non managed transaction.
 * Provides the begin, commit and rollback methods and, in addition,
 * manages a connection to the datastore.
 *
 * @version $Revision: 1.6 $
 **/
class NonmanagedTransaction implements Transaction
{
    protected static final Localiser LOCALISER=Localiser.getInstance("org.jpox.Localisation");

    /** Whether we lock objects for update */
    private boolean useUpdateLockProperty=false;

    private final PersistenceManagerImpl pm;
    private final DatabaseAdapter dba;
    private final DataSource tds;
    private final DataSource ntds;
    private final String userName;
    private final String password;

    private boolean retainValues;
    private boolean restoreValues;
    private boolean optimistic;
    private boolean nontransactionalRead;
    private boolean nontransactionalWrite;
    private int isolationLevel;

    /**
     * Whether the transaction is marked for rollback. JDO 2.0 section 13.4.5
     */
    private boolean rollbackOnly=false;

    private Synchronization sync = null;
    private Connection conn = null;
    private boolean active = false;

    /**
     * optimistic transaction normally release connection after reading
     * Connection are pinned (hold opened during transaction) when 
     * ignoreCache == false 
     */
    private boolean pinnedConnection = false;

    private boolean isJDOConnectionInUse = false;

    /**
     * Constructor.
     * @param pm The PersistenceManager
     * @param userName The username for the datasource
     * @param password The password for the datasource
     **/
    public NonmanagedTransaction(PersistenceManagerImpl pm, String userName, String password)
    {
        this.pm = pm;
        this.dba = ((RDBMSManager)pm.getStoreManager()).getDatabaseAdapter();

        PersistenceManagerFactoryImpl pmf = (PersistenceManagerFactoryImpl)pm.getPersistenceManagerFactory();

        this.tds = pmf.getTransactionalDataSource();
        this.ntds = pmf.getNontransactionalDataSource();
        this.userName = userName;
        this.password = password;

        this.useUpdateLockProperty = pmf.getUseUpdateLock();

        setRetainValues(pmf.getRetainValues());
        setRestoreValues(pmf.getRestoreValues());
        setOptimistic(pmf.getOptimistic());
        setNontransactionalRead(pmf.getNontransactionalRead());
        setNontransactionalWrite(pmf.getNontransactionalWrite());
        setTransactionIsolation(pmf.getTransactionIsolation());
    }

    /**
     * Accessor for the Persistence Manager.
     * @return Persistence Manager
     **/
    public PersistenceManager getPersistenceManager()
    {
        return pm;
    }

    /**
     * Accessor for the connection to the datasource.
     * If this method is called while a datastore transaction is active,
     * the object returned will be enlisted in the current transaction.
     * If called in an optimistic transaction or outside an active
     * transaction, the object returned will not be enlisted in any transaction.
     * 
     * The object must be returned to the JDO implementation prior to calling
     * any JDO method or performing any action on any persistent instance that
     * might require the JDO implementation to use a connection. If the object
     * has not been returned and the JDO implementation needs a connection,
     * a JDOUserException is thrown. The object is returned to the JDO
     * implementation by calling the standard method on the object.
     * 
     * @return The Connection
     * @throws SQLException
     */
    public synchronized JDOConnection getJDOConnection()
    {
        assertJDOConnectionNotInUse();
        
        final Connection nativeConnection;
        if( conn != null )
        {
            isJDOConnectionInUse = true;
            nativeConnection = conn;
        }
        else
        {
	        try
	        {
	            nativeConnection = dba.getConnection(ntds, userName, password, getTransactionIsolation());
	        }
	        catch (SQLException e)
	        {
	            throw new JDODataStoreException(e.getMessage());
	        }
        }
        return new JDOConnectionImpl(nativeConnection, new Runnable()
        {
            public void run()
            {
                isJDOConnectionInUse = false;
            }
        });
    }

    /**
     * Checks whether a connection to the datasource is open.
     * @return true if the Connection is open
     * @throws SQLException
     */
    public synchronized boolean isConnectionOpen()
    {
        return conn != null;
    }

    /**
     * Accessor for the connection to the datasource.
     * @param forWriting Whether the connection is read-only or read-write.
     * @return The Connection
     * @throws SQLException
     */
    public synchronized Connection getConnection(boolean forWriting)
    throws SQLException
    {
        return getConnection(forWriting, false);
    }

    /**
     * Accessor for the connection to the datasource.
     * @param forWriting Whether the connection is read-only or read-write.
     * @param useOpenedConnection whether to use an already opened connection, if one is opened
     * @return The Connection
     * @throws SQLException
     */
    public synchronized Connection getConnection(boolean forWriting, boolean useOpenedConnection)
    throws SQLException
    {
        assertJDOConnectionNotInUse();
        
        if (active)
        {
            // Before performing any (more) SQL operations in this transaction,
            // flush out any updates pending in the persistence manager.
            if (!getPersistenceManager().getIgnoreCache() || forWriting)
            {
                if (optimistic && !pinnedConnection)
                {
                    pinnedConnection = true;
                    conn = dba.getConnection(ntds, userName, password, getTransactionIsolation());
                    if (JPOXLogger.TRANSACTION.isDebugEnabled())
                    {
                        JPOXLogger.TRANSACTION.debug(LOCALISER.msg("Transaction.Optimistic.ConnectionRetrieved", conn));
                    }
                }
                pm.flush();
            }
            if (optimistic && !pinnedConnection)
            {
                conn = dba.getConnection(ntds, userName, password, getTransactionIsolation());
                if (JPOXLogger.TRANSACTION.isDebugEnabled())
                {
                    JPOXLogger.TRANSACTION.debug(LOCALISER.msg("Transaction.Optimistic.ConnectionRetrieved", conn));
                }
            }
        }
        else
        {
            if (!nontransactionalRead || (forWriting && !nontransactionalWrite))
            {
                throw new TransactionNotActiveException();
            }
            if (conn != null && !useOpenedConnection)
            {
                throw new ConnectionInUseException();
            }
            else if (conn == null)
            {
                conn = dba.getConnection(ntds, userName, password, getTransactionIsolation());
                if (JPOXLogger.TRANSACTION.isDebugEnabled())
                {
                    JPOXLogger.TRANSACTION.debug(LOCALISER.msg(
                        optimistic ? "Transaction.Optimistic.ConnectionRetrieved" : "Transaction.Pessimistic.ConnectionRetrieved", conn));
                }
            }
        }

        return conn;
    }

    /**
     * Method to release the Connection.
     * @param conn The Connection to release.
     * @throws SQLException Thrown when an error occurs releasing the
     * connection.
     **/
    public synchronized void releaseConnection(Connection conn)
    throws SQLException
    {
        assertJDOConnectionNotInUse();

        if (conn == this.conn)
        {
            if (!active || (optimistic && !pinnedConnection))
            {
                if(conn != null)
                {
                    conn.commit();
                }
                closeConnection();
            }
        }
    }

    /**
     * Accessor for whether to use an update lock when fetching data.
     * @return Whether to use an update lock.
     **/
    public boolean useUpdateLockOnFetch()
    {
        // If we aren't active, or if doing optimistic transaction then don't lock
        if (!active || optimistic)
        {
            return false;
        }
        else
        {
            if (isolationLevel == Connection.TRANSACTION_REPEATABLE_READ ||
            	isolationLevel == Connection.TRANSACTION_SERIALIZABLE)
            {
                return false;
            }
            else
            {
                // TODO Extend the code here to allow multiple locking strategies
                // so that we can just lock the first query in a transaction for example
                return useUpdateLockProperty;
            }
        }
    }

    /**
     * Method to close the Connection.
     * @throws SQLException Thrown when an error occurs closing the connection.
     **/
    private void closeConnection() 
    throws SQLException
    {
        try
        {
            if (conn != null)
            {
                if (JPOXLogger.TRANSACTION.isDebugEnabled())
                {
                    JPOXLogger.TRANSACTION.debug(LOCALISER.msg(
                        optimistic ? "Transaction.Optimistic.ConnectionClosed" : "Transaction.Pessimistic.ConnectionClosed", conn));
                }
                dba.closeConnection(conn);
            }
        }
        finally
        {
            pinnedConnection = false;
            conn = null;
        }
    }

    /**
     * Asserts that the transaction is not in use.
     **/
    private void assertNotInUse()
    {
        if (active)
        {
            throw new TransactionActiveException();
        }

        if (conn != null)
        {
            throw new ConnectionInUseException();
        }
    }

    /**
     * Method to start the transaction.
     **/
    public synchronized void begin()
    {
        assertNotInUse();

        try
        {
            if (!optimistic)
            {
                conn = dba.getConnection(tds, userName, password, getTransactionIsolation());
                if (JPOXLogger.TRANSACTION.isDebugEnabled())
                {
                    JPOXLogger.TRANSACTION.debug(LOCALISER.msg("Transaction.Pessimistic.ConnectionRetrieved", conn));
                }
            }

            active = true;

            if (JPOXLogger.TRANSACTION.isDebugEnabled())
            {
                JPOXLogger.TRANSACTION.debug(LOCALISER.msg("Transaction.StartedForConnection",conn));
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("Transaction.FailedToStart"),e);
        }
    }

    /**
     * Method to commit the transaction.
     **/
    public synchronized void commit()
    {
        if (!active)
        {
            throw new TransactionNotActiveException();
        }

        assertJDOConnectionNotInUse();

        // JDO 2.0 section 13.4.5 rollbackOnly functionality
        // It isn't clear from the spec if we are expected to do the rollback
        // here. The spec simply says that we throw an exception. This is
        // assumed as meaning that the users code will catch the exception and
        // call rollback themselves. i.e we don't need to close the DB
        // connection or set "active" to false. 
        if (rollbackOnly)
        {
            // Throw an exception since can only exit via rollback
            if (JPOXLogger.TRANSACTION.isDebugEnabled())
            {
                JPOXLogger.TRANSACTION.debug(LOCALISER.msg("Transaction.CommitWithRollbackOnly"));
            }

            throw new JDOFatalDataStoreException(LOCALISER.msg("Transaction.CommitWithRollbackOnly"));
        }

        try
        {
            boolean success = false;

            try
            {
                try
                {
                    if (JPOXLogger.TRANSACTION.isDebugEnabled())
                    {
                        JPOXLogger.TRANSACTION.debug(LOCALISER.msg("Transaction.CommitForConnection",conn));
                    }
                    
                    pm.flush();

                    if (sync != null)
                    {
                        sync.beforeCompletion();
                    }

                    //optimistic transactions that don't have dirty
                    if (conn != null)
                    {
                        conn.commit();
                    }

                    success = true;
                }
                finally
                {
                    if (!success)
                    {
                        try
                        {
                            pm.preRollback();
                        }
                        finally
                        {
                            if (conn != null)
                            {
                                conn.rollback();
                            }
                        }
                    }
                }
            }
            finally
            {
                try
                {
                    active = false;
                    closeConnection();
                }
                finally
                {
                    try
                    {
                        if (success)
                        {
                            pm.postCommit();
                        }
                    }
                    finally
                    {
                        if (sync != null)
                        {
                            if (success)
                            {
                                sync.afterCompletion(Status.STATUS_COMMITTED);
                            }
                            else
                            {
                                sync.afterCompletion(Status.STATUS_ROLLEDBACK);
                            }
                        }
                    }
                }
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("Transaction.FailedToCommit"),e);
        }
    }

    /**
     * Method to rollback the transaction.
     **/
    public synchronized void rollback()
    {
        if (!active)
        {
            throw new TransactionNotActiveException();
        }
        
        assertJDOConnectionNotInUse();

        try
        {
            try
            {
                try
                {
                    if (JPOXLogger.TRANSACTION.isDebugEnabled())
                    {
                        JPOXLogger.TRANSACTION.debug(LOCALISER.msg("Transaction.RollbackForConnection",conn));
                    }

                    pm.preRollback();
                }
                finally
                {
                    if (conn != null)
                    {
                        conn.rollback();
                    }
                }
            }
            finally
            {
                try
                {
                    active = false;
                    closeConnection();
                }
                finally
                {
                    if (sync != null)
                    {
                        sync.afterCompletion(Status.STATUS_ROLLEDBACK);
                    }
                }
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("Transaction.FailedToRollback"),e);
        }

        // Reset rollbackOnly flag
        rollbackOnly = false;
    }

    /**
     * Accessor for whether the transaction is active.
     * @return Whether the transaction is active.
     **/
    public boolean isActive()
    {
        return active;
    }

    /**
     * Accessor for the "rollback only" flag.
     * @return The rollback only flag
     * @since 1.1
     */
    public boolean getRollbackOnly()
    {
        return rollbackOnly;
    }

    /**
     * Mutator for the "rollback only" flag. Sets the transaction as for
     * rollback only.
     * @since 1.1
     */
    public void setRollbackOnly()
    {
        // Only apply to active transactions
        if (active)
        {
            rollbackOnly = true;
        }
    }

    /**
     * Mutator to set whether to allow nontransactional reads.
     * @param nontransactionalRead Whether to allow nontransactional reads 
     **/ 
    public synchronized void setNontransactionalRead(boolean nontransactionalRead)
    {
		//assertInUse();

        this.nontransactionalRead = nontransactionalRead;
    }

    /**
     * Accessor for whether to allow nontransactional reads.
     * @return Whether to allow nontransactional reads.
     **/
    public boolean getNontransactionalRead()
    {
        return nontransactionalRead;
    }

    /**
     * Mutator to set whether to allow nontransactional writes.
     * @param nontransactionalWrite Whether to allow nontransactional writes. 
     **/ 
    public synchronized void setNontransactionalWrite(boolean nontransactionalWrite)
    {
        assertNotInUse();

        if (nontransactionalWrite)
        {
            throw new JDOUnsupportedOptionException(LOCALISER.msg("Transaction.NontransactionalWriteUnsupported"));
        }

        this.nontransactionalWrite = nontransactionalWrite;
    }

    /**
     * Accessor for whether to allow nontransactional writes.
     * @return Whether to allow nontransactional writes.
     **/
    public boolean getNontransactionalWrite()
    {
        return nontransactionalWrite;
    }

    /**
     * Mutator for whether to retain values at commit.
     * @param retainValues Whether to retain values at commit.
     **/
    public synchronized void setRetainValues(boolean retainValues)
    {
        assertNotInUse();

        this.retainValues = retainValues;

        if (retainValues)
        {
            nontransactionalRead = true;
        }
    }

    /**
     * Accessor for whether to retain values at commit.
     * @return Whether to retain values at commit.
     **/
    public boolean getRetainValues()
    {
        return retainValues;
    }

    /**
     * Mutator for whether to restore values at rollback.
     * @param restoreValues Whether to restore values at rollback.
     **/ 
    public synchronized void setRestoreValues(boolean restoreValues)
    {
        assertNotInUse();

        this.restoreValues = restoreValues;
    }

    /**
     * Accessor for whether to restore values at rollback.
     * @return Whether to restore values at rollback.
     **/
    public boolean getRestoreValues()
    {
        return restoreValues;
    }

    /**
     * Mutator for whether to support optimistic transactions.
     * @param optimistic Whether to support optimistic transactions.
     **/ 
    public synchronized void setOptimistic(boolean optimistic)
    {
        assertNotInUse();

        this.optimistic = optimistic;
    }

    /**
     * Accessor for whether to support optimistic transactions.
     * @return Whether to return optimistic transactions.
     **/
    public boolean getOptimistic()
    {
        return optimistic;
    }

    /**
     * Mutator for the transaction isolation level.
     * @param isolationLevel The isolation level required.
     **/
    public synchronized void setTransactionIsolation(int isolationLevel)
    {
        assertNotInUse();

        switch (isolationLevel)
        {
            case Connection.TRANSACTION_READ_UNCOMMITTED:
            case Connection.TRANSACTION_READ_COMMITTED:
            case Connection.TRANSACTION_REPEATABLE_READ:
            case Connection.TRANSACTION_SERIALIZABLE:
                break;

            default:
                throw new IllegalArgumentException(LOCALISER.msg("Transaction.IsolationLevelInvalid",isolationLevel));
        }

        this.isolationLevel = isolationLevel;
    }

    /**
     * Accessor for the transaction isolation level.
     * @return The transaction isolation level.
     **/
    public int getTransactionIsolation()
    {
        return isolationLevel;
    }

    /**
     * Mutator for the synchronization object to be notified on transaction
     * completion.
     * @param sync The synchronization object to be notified on transaction
     * completion
     **/
    public synchronized void setSynchronization(Synchronization sync)
    {
        this.sync = sync;
    }

    /**
     * Accessor for the synchronization object to be notified on transaction
     * completion.
     * @return The synchronization instance ot be notified on transaction
     * completion.
     **/
    public Synchronization getSynchronization()
    {
        return sync;
    }
    
    /**
     * Asserts that's the connection is not hold by the developer's application
     * @throws JDOUserException if hold by developer
     */
    private void assertJDOConnectionNotInUse()
    {
        if (isJDOConnectionInUse)
        {
            throw new JDOUserException("The JDOConnection was acquired by the developer and must be closed before using the JDO API.");
        }
    }
}