/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2004 Andy Jefferson - added description of addition process.
2004 Andy Jefferson - added constructor, and try-catch on initialisation
2004 Andy Jefferson - added various new properties
    ...
**********************************************************************/
package org.jpox;

import java.sql.Connection;
import java.util.TimeZone;

import javax.jdo.JDOUnsupportedOptionException;
import javax.jdo.JDOUserException;

import org.jpox.store.AutoStartMechanism;
import org.jpox.store.Dictionary;
import org.jpox.store.XMLAutoStarter;
import org.jpox.store.rdbms.SchemaAutoStarter;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;
import org.jpox.util.StringUtils;

/**
 * Contains the user configuration for the PMF. This is extended by
 * AbstractPersistenceManagerFactory, which in turn is extended by
 * PersistenceManagerFactoryimpl and ManagedConnectionFactoryImpl. If you add a
 * setter/getter in here, there is an initialisation block in
 * PersistenceManagerFactoryImpl that will likely need the parameter adding.
 * @version $Revision: 1.15 $
 */
public abstract class PMFConfiguration
{
    /** Localisation of messages. */
    protected static final Localiser LOCALISER=Localiser.getInstance("org.jpox.Localisation");

    /** Datastore driver name property. */
    public static final String JDO_DATASTORE_DRIVERNAME_PROPERTY = "javax.jdo.option.ConnectionDriverName";

    /** Datastore URL property. */
    public static final String JDO_DATASTORE_URL_PROPERTY = "javax.jdo.option.ConnectionURL";

    /** Datastore connection username property. */
    public static final String JDO_DATASTORE_USERNAME_PROPERTY = "javax.jdo.option.ConnectionUserName";

    /** Datastore connection password property. */
    public static final String JDO_DATASTORE_PASSWORD_PROPERTY = "javax.jdo.option.ConnectionPassword";

    /** Mapping suffix property. */
    public static final String JDO_MAPPING_PROPERTY = "javax.jdo.option.Mapping";

    /**
     * Whether to check if the table/view exists.
     */
    public static final String CHECK_EXIST_TABLES_VIEWS_PROPERTY = "org.jpox.rdbms.CheckExistTablesOrViews";

    /**
     * The system property that selects the default value for the
     * ValidateTables setting.
     */
    public static final String VALIDATE_TABLES_PROPERTY = "org.jpox.validateTables";

    /**
     * The system property that selects the default value for the
     * ValidateConstraints setting.
     */
    public static final String VALIDATE_CONSTRAINTS_PROPERTY = "org.jpox.validateConstraints";

    /**
     * The system property that selects the default value for the
     * AutoCreateSchema setting.
     */
    public static final String AUTO_CREATE_SCHEMA_PROPERTY = "org.jpox.autoCreateSchema";

    /**
     * The system property that selects the default value for the
     * AutoCreateTables setting.
     */
    public static final String AUTO_CREATE_TABLES_PROPERTY = "org.jpox.autoCreateTables";
    
    /**
     * The system property that selects the default value for the
     * AutoCreateColumns setting.
     */
    public static final String AUTO_CREATE_COLUMNS_PROPERTY = "org.jpox.autoCreateColumns";

    /**
     * The system property that selects the default value for the
     * AutoCreateConstraints setting.
     */
    public static final String AUTO_CREATE_CONSTRAINTS_PROPERTY = "org.jpox.autoCreateConstraints";

    /**
     * Whether the datastore schema is read-only and so writes should not be performed.
     */
    public static final String READ_ONLY_SCHEMA_PROPERTY = "org.jpox.readOnlySchema";

    /**
     * Whether the datastore schema is read-only and so writes should not be performed.
     */
    public static final String FIXED_SCHEMA_PROPERTY = "org.jpox.fixedSchema";

    /** The system property that defines the mode of creation of foreign keys. */
    public static final String FOREIGN_KEY_CREATE_MODE_PROPERTY = "org.jpox.foreignKeyCreateMode";

    /** Auto-creation mode, including users MetaData specifications. */
    public static final String FOREIGN_KEY_AUTO_MODE="AUTO";

    /** Strict mode, only using users MetaData specifications. */
    public static final String FOREIGN_KEY_STRICT_MODE="STRICT";

    /**
     * The system property that selects the default value for the
     * transaction isolation level.  This is the string
     * "org.jpox.transactionIsolation".
     */
    public static final String TRANSACTION_ISOLATION_PROPERTY = "org.jpox.transactionIsolation";

    /**
     * The system property defines the default generator id class
     * to be used for generating ids. This is the string
     * "org.jpox.poid.generatorClass".
     */
    public static final String GENERATOR_ID_CLASS_PROPERTY = "org.jpox.poid.generatorClass";

    /**
     * The system property that selects the default value for the
     * transaction isolation level when opening a connection for the Poid generator.
     * This is the string "org.jpox.poid.transactionIsolation".
     */
    public static final String POID_TRANSACTION_ISOLATION_PROPERTY = "org.jpox.poid.transactionIsolation";

    /**
     * Whether to use the PM connection or open a new connection
     * This is the string "transactionAttribute".
     */
    public static final String POID_TRANSACTION_ATTRIBUTE_PROPERTY = "org.jpox.poid.transactionAttribute";
    
    /**
     * The system property defining whether to use the Schema Table
     * for registration of supported classes.
     */
    public static final String AUTO_START_MECHANISM_PROPERTY = "org.jpox.autoStartMechanism";

    /**
     * The system property defining whether to handle the startup in a
     * quiet way, or raise exceptions.
     */
    public static final String AUTO_START_MECHANISM_MODE_PROPERTY = "org.jpox.autoStartMechanismMode";

    /**
     * Name of the file used when using the XML AutoStartMechanism.
     */
    public static final String AUTO_START_MECHANISM_XML_FILE_PROPERTY = "org.jpox.autoStartMechanismXmlFile";

    /**
     * The system property defining the type of Level 1 Cache to use.
     */
    public static final String JDO_CACHE_LEVEL_1_TYPE_PROPERTY = "org.jpox.cache.level1.type";

    /**
     * The system property defining whether to use a Level 2 Cache..
     */
    public static final String JDO_CACHE_LEVEL_2_PROPERTY = "org.jpox.cache.level2";
    
    /**
     * The system property defining the type of Level 2 Cache to use.
     */
    public static final String JDO_CACHE_LEVEL_2_TYPE_PROPERTY = "org.jpox.cache.level2.type";

    /**
     * The system property defining the symbolic name of the cache to use for Level 2.
     */
    public static final String JDO_CACHE_LEVEL_2_CACHE_NAME_PROPERTY = "org.jpox.cache.level2.cacheName";

    /**
     * The system property defining whether to cache collections.
     */
    public static final String JDO_CACHE_COLLECTIONS_PROPERTY = "org.jpox.cache.collections";

    /**
     * The file extension for the JDO metadata files.
     */
    public static final String JDO_METADATA_FILE_EXTENSION_PROPERTY = "org.jpox.metadata.jdoFileExtension";

    /**
     * The file extension for the ORM metadata files.
     */
    public static final String ORM_METADATA_FILE_EXTENSION_PROPERTY = "org.jpox.metadata.ormFileExtension";

    /**
     * Default catalog name to use in any table access.
     */
    public static final String DEFAULT_CATALOG_NAME_PROPERTY = "org.jpox.identifier.defaultCatalogName";

    /**
     * Default schema name to use in any table access.
     */
    public static final String DEFAULT_SCHEMA_NAME_PROPERTY = "org.jpox.identifier.defaultSchemaName";

    /**
     * use fully qualified for identifiers if tables
     */
    public static final String FULLY_QUALIFIED_NAMES_PROPERTY = "org.jpox.identifier.fullyQualifiedNames";

    /**
     * Default schema name to use in any table access.
     */
    public static final String IDENTIFIER_NAMING_STRATEGY_PROPERTY = "org.jpox.identifier.namingStrategy";

    /**
     * Whether to use "SELECT ... FOR UPDATE" to fetch objects when the
     * transaction isolation is READ COMMITTED or lower.
     */
    public static final String USE_UPDATE_LOCK_PROPERTY = "org.jpox.useUpdateLock";

    /**
     * The default strategy for inheritance, allowing change of behaviour from
     * the JDO2 standard.
     */
    public static final String DEFAULT_INHERITANCE_STRATEGY_PROPERTY = "org.jpox.rdbms.defaultInheritanceStrategy";

    /**
     * Whether the datastore operations are delayed until commit.
     * In optimistic transactions this is automatically enabled.
     */
    public static final String DELAY_DATASTORE_OPERATIONS_UNTIL_COMMIT_PROPERTY = "org.jpox.delayDatastoreOperationsUntilCommit";

    /**
     * Whether to add unique constraints to the element table for a map inverse field
     */
    public static final String ADD_UNIQUE_CONSTRAINT_MAP_INVERSE_PROPERTY = "org.jpox.rdbms.uniqueConstraints.mapInverse";

    /** Number of rows of a result set that should be fetched when more rows are needed. */
    public static final String QUERY_FETCH_SIZE_PROPERTY = "org.jpox.query.fetchSize";

    /** Direction in which a result set will be navigated. */
    public static final String QUERY_FETCH_DIRECTION_PROPERTY = "org.jpox.query.fetchDirection";

    /** Type of result set to create. */
    public static final String QUERY_RESULT_SET_TYPE_PROPERTY = "org.jpox.query.resultSetType";

    /** Concurrency of the result set */
    public static final String QUERY_RESULT_SET_CONCURRENCY_PROPERTY = "org.jpox.query.resultSetConcurrency";

    /** Default string length when storing a String type in the datastore and no length given. */
    public static final String STRING_DEFAULT_LENGTH_PROPERTY = "org.jpox.rdbms.stringDefaultLength";

    /** Timezone name to assume all dates are in in the datastore. */
    public static final String DATE_TIMEZONE_PROPERTY = "org.jpox.rdbms.dateTimezone";

    public TimeZone DATE_TIMEZONE;

    private String driverName = null;
    private String url = null;
    private String userName = null;
    private String password = null;

    private Object connectionFactory = null;
    private String connectionFactoryName = null;

    private Object connectionFactory2 = null;
    private String connectionFactory2Name = null;

    private boolean multithreaded = false;
    private boolean optimistic = false;
    private boolean retainValues = false;
    private boolean restoreValues = false;
    private boolean nontransactionalRead = false;
    private boolean nontransactionalWrite = false;
    private boolean ignoreCache = false;
    private boolean detachOnClose = false;
    private String defaultCatalogName = null;
    private String defaultSchemaName = null;
    private boolean fullyQualifiedNames = false;
    private String identifierNamingStrategy = "ClassFieldName";
    private boolean useUpdateLock = false;
    private boolean delayDatastoreOperationsEnabled = false;
    private boolean uniqueConstraintsMapInverse = true;
    private String defaultInheritanceStrategy = "JDO2";

    // Query
    private int queryFetchSize = 0;
    private String queryFetchDirection = "forward";
    private String queryResultSetType = "forward-only";
    private String queryResultSetConcurrency = "read-only";

    // Types
    private int stringDefaultLength = 255;
    private String dateTimezone = "UTC";

    private boolean checkExistTablesOrViews=true;
    private boolean validateTables=true;
    private boolean validateConstraints=true;
    private boolean autoCreateSchema=false;
    private boolean autoCreateTables=false;
    private boolean autoCreateColumns=false;
    private boolean autoCreateConstraints=false;
    private String  autoStartMechanism=SchemaAutoStarter.SCHEMA_TABLE;
    private String  autoStartMechanismXmlFile="jpoxAutoStart.xml";
    private String  autoStartMechanismMode=AutoStartMechanism.MODE_QUIET;
    private boolean readOnlySchema=false;
    private boolean fixedSchema=false;
    private String foreignKeyCreateMode=FOREIGN_KEY_AUTO_MODE;

    private int     isolationLevel=Connection.TRANSACTION_READ_COMMITTED;
    private int     poidIsolationLevel=Connection.TRANSACTION_READ_COMMITTED;
    private String  poidTransactionAttribute="New";
    private String  persistentIdGenerator="org.jpox.poid.DefaultIdentifierGenerator";

    private boolean jdoCacheCollections=true;
    private String jdoCacheLevel1Type="org.jpox.cache.WeakRefCache";
    private boolean jdoCacheLevel2=false;
    private String jdoCacheLevel2Type="org.jpox.cache.DefaultLevel2Cache";
    private String jdoCacheLevel2CacheName=null;

    private String jdoMetaDataFileExtension="jdo";
    private String ormMetaDataFileExtension="orm";
    private String mapping=null;

    private Dictionary dictionary = Dictionary.getDefaultDictionary();

    protected transient boolean configurable=true;

    /**
     * Constructor. 
     * Initialises all properties using the System property values. This behaviour
     * is not specified by the JDO spec(s) but is a nice to have capability allowing
     * users to define command line args to set up their PMF.
     */
    public PMFConfiguration()
    {
        try
        {
            // Schema Management
            validateTables = new Boolean(System.getProperty(VALIDATE_TABLES_PROPERTY, "true")).booleanValue();
            validateConstraints = new Boolean(System.getProperty(VALIDATE_CONSTRAINTS_PROPERTY, "true")).booleanValue();
            autoCreateSchema = new Boolean(System.getProperty(AUTO_CREATE_SCHEMA_PROPERTY, "false")).booleanValue();
            autoCreateTables = new Boolean(System.getProperty(AUTO_CREATE_TABLES_PROPERTY, "false")).booleanValue();
            autoCreateColumns = new Boolean(System.getProperty(AUTO_CREATE_COLUMNS_PROPERTY, "false")).booleanValue();
            autoCreateConstraints = new Boolean(System.getProperty(AUTO_CREATE_CONSTRAINTS_PROPERTY, "false")).booleanValue();
            readOnlySchema = new Boolean(System.getProperty(READ_ONLY_SCHEMA_PROPERTY, "false")).booleanValue();
            fixedSchema = new Boolean(System.getProperty(FIXED_SCHEMA_PROPERTY, "false")).booleanValue();
            defaultCatalogName = System.getProperty(DEFAULT_CATALOG_NAME_PROPERTY);
            defaultSchemaName = System.getProperty(DEFAULT_SCHEMA_NAME_PROPERTY);
            if (readOnlySchema || fixedSchema)
            {
                if (autoCreateSchema)
                {
                    JPOXLogger.JDO.warn(LOCALISER.msg("PMF.PropertyUnusableForSchema","autoCreateSchema"));
                    autoCreateSchema = false;
                }
                if (autoCreateTables)
                {
                    JPOXLogger.JDO.warn(LOCALISER.msg("PMF.PropertyUnusableForSchema","autoCreateTables"));
                    autoCreateTables = false;
                }
                if (autoCreateColumns)
                {
                    JPOXLogger.JDO.warn(LOCALISER.msg("PMF.PropertyUnusableForSchema","autoCreateColumns"));
                    autoCreateColumns = false;
                }
                if (autoCreateConstraints)
                {
                    JPOXLogger.JDO.warn(LOCALISER.msg("PMF.PropertyUnusableForSchema","autoCreateConstraints"));
                    autoCreateConstraints = false;
                }
            }
            String fkCreateMode = System.getProperty(FOREIGN_KEY_CREATE_MODE_PROPERTY);
            if (fkCreateMode != null && fkCreateMode.equals("" + FOREIGN_KEY_AUTO_MODE))
            {
                foreignKeyCreateMode = FOREIGN_KEY_AUTO_MODE;
            }
            else if (fkCreateMode != null && fkCreateMode.equals("" + FOREIGN_KEY_STRICT_MODE))
            {
                foreignKeyCreateMode = FOREIGN_KEY_STRICT_MODE;
            }
            else
            {
                foreignKeyCreateMode = FOREIGN_KEY_AUTO_MODE;
            }

            // JDO operation
            defaultInheritanceStrategy = System.getProperty(DEFAULT_INHERITANCE_STRATEGY_PROPERTY);
            persistentIdGenerator = System.getProperty(GENERATOR_ID_CLASS_PROPERTY, "org.jpox.poid.DefaultIdentifierGenerator");
            autoStartMechanism = System.getProperty(AUTO_START_MECHANISM_PROPERTY, SchemaAutoStarter.SCHEMA_TABLE);
            autoStartMechanismXmlFile = System.getProperty(AUTO_START_MECHANISM_XML_FILE_PROPERTY, "jpoxAutoStart.xml");
            autoStartMechanismMode = System.getProperty(AUTO_START_MECHANISM_MODE_PROPERTY, AutoStartMechanism.MODE_QUIET);
            poidTransactionAttribute = System.getProperty(POID_TRANSACTION_ATTRIBUTE_PROPERTY, "New");

            if (System.getProperty(Dictionary.STORE_DICTIONARY_CASE_IDENTIFIER) != null)
            {
                dictionary.setCaseIdentifier(Dictionary.CaseIdentifier.getCaseIdentifier(System.getProperty(Dictionary.STORE_DICTIONARY_CASE_IDENTIFIER)));
            }
            
            String extension = System.getProperty(JDO_METADATA_FILE_EXTENSION_PROPERTY);
            if (!StringUtils.isWhitespace(extension))
            {
                jdoMetaDataFileExtension = extension;
            }
            extension = System.getProperty(ORM_METADATA_FILE_EXTENSION_PROPERTY);
            if (!StringUtils.isWhitespace(extension))
            {
                ormMetaDataFileExtension = extension;
            }

            // Caching
            jdoCacheCollections = new Boolean(System.getProperty(JDO_CACHE_COLLECTIONS_PROPERTY,"true")).booleanValue();
            jdoCacheLevel1Type = System.getProperty(JDO_CACHE_LEVEL_1_TYPE_PROPERTY, "org.jpox.cache.WeakRefCache");
            jdoCacheLevel2 = new Boolean(System.getProperty(JDO_CACHE_LEVEL_2_PROPERTY,"false")).booleanValue();
            jdoCacheLevel2Type = System.getProperty(JDO_CACHE_LEVEL_2_TYPE_PROPERTY, "org.jpox.cache.DefaultLevel2Cache");
            jdoCacheLevel2CacheName = System.getProperty(JDO_CACHE_LEVEL_2_CACHE_NAME_PROPERTY);

            // Queries
            String queryFetchSizeString = System.getProperty(QUERY_FETCH_SIZE_PROPERTY);
            if (queryFetchSizeString != null)
            {
                try
                {
                    queryFetchSize = (new Integer(queryFetchSizeString)).intValue();
                }
                catch (NumberFormatException nfe)
                {
                    // Do nothing and use default.
                }
            }
            queryFetchDirection = System.getProperty(QUERY_FETCH_DIRECTION_PROPERTY, "forward");
            queryResultSetType = System.getProperty(QUERY_RESULT_SET_TYPE_PROPERTY, "forward-only");
            queryResultSetConcurrency = System.getProperty(QUERY_RESULT_SET_CONCURRENCY_PROPERTY, "read-only");

            // Types
            String stringLengthTemp = System.getProperty(STRING_DEFAULT_LENGTH_PROPERTY);
            if (stringLengthTemp != null)
            {
                try
                {
                    stringDefaultLength = (new Integer(stringLengthTemp)).intValue();
                }
                catch (NumberFormatException nfe)
                {
                    stringDefaultLength = 255;
                }
            }
            dateTimezone = System.getProperty(DATE_TIMEZONE_PROPERTY, "UTC");
            DATE_TIMEZONE = TimeZone.getTimeZone(dateTimezone);
            if (DATE_TIMEZONE == null)
            {
                DATE_TIMEZONE = TimeZone.getTimeZone("UTC");
            }
        }
        catch (Exception e)
        {
            // May arrive here if the server throws a SecurityException
            JPOXLogger.GENERAL.error(e);
        }
    }

    /**
     * Equality operator.
     * @param obj Object to compare against.
     * @return Whether the objects are equal.
     **/
    public synchronized boolean equals(Object obj)
    {
        if (obj == this)
        {
            return true;
        }
        if (!(obj instanceof PMFConfiguration))
        {
            return false;
        }

        PMFConfiguration pmfConfiguration = (PMFConfiguration)obj;

        boolean isEquals = true;
        isEquals = isEquals & equalsVariable(driverName,              pmfConfiguration.driverName);
        isEquals = isEquals & equalsVariable(url,                     pmfConfiguration.url);
        isEquals = isEquals & equalsVariable(userName,                pmfConfiguration.userName);
        isEquals = isEquals & equalsVariable(password,                pmfConfiguration.password);
        isEquals = isEquals & equalsVariable(connectionFactory,       pmfConfiguration.connectionFactory);
        isEquals = isEquals & equalsVariable(connectionFactoryName,   pmfConfiguration.connectionFactoryName);
        isEquals = isEquals & equalsVariable(connectionFactory2,      pmfConfiguration.connectionFactory2);
        isEquals = isEquals & equalsVariable(connectionFactory2Name,  pmfConfiguration.connectionFactory2Name);
        isEquals = isEquals & equalsVariable(mapping, pmfConfiguration.mapping);

        isEquals = isEquals & equalsVariable(jdoCacheLevel1Type, pmfConfiguration.jdoCacheLevel1Type);
        isEquals = isEquals & equalsVariable(jdoCacheLevel2Type, pmfConfiguration.jdoCacheLevel2Type);
        isEquals = isEquals & equalsVariable(jdoCacheLevel2CacheName, pmfConfiguration.jdoCacheLevel2CacheName);
        isEquals = isEquals & (jdoCacheLevel2 == pmfConfiguration.jdoCacheLevel2);
        isEquals = isEquals & (jdoCacheCollections == pmfConfiguration.jdoCacheCollections);

        isEquals = isEquals & equalsVariable(connectionFactoryName,   pmfConfiguration.connectionFactoryName);
        isEquals = isEquals & equalsVariable(connectionFactoryName,   pmfConfiguration.connectionFactoryName);
        isEquals = isEquals & equalsVariable(connectionFactoryName,   pmfConfiguration.connectionFactoryName);
        isEquals = isEquals & equalsVariable(persistentIdGenerator,   pmfConfiguration.persistentIdGenerator);
        isEquals = isEquals & equalsVariable(autoStartMechanism,      pmfConfiguration.autoStartMechanism);
        isEquals = isEquals & equalsVariable(autoStartMechanismXmlFile, pmfConfiguration.autoStartMechanismXmlFile);
        isEquals = isEquals & equalsVariable(autoStartMechanismMode,  pmfConfiguration.autoStartMechanismMode);
        isEquals = isEquals & equalsVariable(dictionary,              pmfConfiguration.dictionary);
        isEquals = isEquals & equalsVariable(jdoMetaDataFileExtension, pmfConfiguration.jdoMetaDataFileExtension);
        isEquals = isEquals & equalsVariable(ormMetaDataFileExtension, pmfConfiguration.ormMetaDataFileExtension);
        isEquals = isEquals & equalsVariable(poidTransactionAttribute,pmfConfiguration.poidTransactionAttribute);
        isEquals = isEquals & equalsVariable(defaultCatalogName,   pmfConfiguration.defaultCatalogName);
        isEquals = isEquals & equalsVariable(defaultSchemaName,   pmfConfiguration.defaultSchemaName);
        isEquals = isEquals & (checkExistTablesOrViews == pmfConfiguration.checkExistTablesOrViews);
        isEquals = isEquals & (fullyQualifiedNames == pmfConfiguration.fullyQualifiedNames);
        isEquals = isEquals & equalsVariable(identifierNamingStrategy, pmfConfiguration.identifierNamingStrategy);
        isEquals = isEquals & equalsVariable(defaultInheritanceStrategy, pmfConfiguration.defaultInheritanceStrategy);
        isEquals = isEquals & (stringDefaultLength == pmfConfiguration.stringDefaultLength);
        isEquals = isEquals & equalsVariable(dateTimezone, pmfConfiguration.dateTimezone);

        isEquals = isEquals & (delayDatastoreOperationsEnabled == pmfConfiguration.delayDatastoreOperationsEnabled);
        isEquals = isEquals & (isolationLevel == pmfConfiguration.isolationLevel);
        isEquals = isEquals & (poidIsolationLevel == pmfConfiguration.poidIsolationLevel);
        isEquals = isEquals & (multithreaded == pmfConfiguration.multithreaded);
        isEquals = isEquals & (optimistic == pmfConfiguration.optimistic);
        isEquals = isEquals & (retainValues == pmfConfiguration.retainValues);
        isEquals = isEquals & (restoreValues == pmfConfiguration.restoreValues);
        isEquals = isEquals & (nontransactionalRead == pmfConfiguration.nontransactionalRead);
        isEquals = isEquals & (nontransactionalWrite == pmfConfiguration.nontransactionalWrite);
        isEquals = isEquals & (ignoreCache == pmfConfiguration.ignoreCache);
        isEquals = isEquals & (detachOnClose == pmfConfiguration.detachOnClose);
        isEquals = isEquals & (validateTables == pmfConfiguration.validateTables);
        isEquals = isEquals & (validateConstraints == pmfConfiguration.validateConstraints);
        isEquals = isEquals & (autoCreateSchema == pmfConfiguration.autoCreateSchema);
        isEquals = isEquals & (autoCreateTables == pmfConfiguration.autoCreateTables);
        isEquals = isEquals & (autoCreateColumns == pmfConfiguration.autoCreateColumns);
        isEquals = isEquals & (autoCreateConstraints == pmfConfiguration.autoCreateConstraints);
        isEquals = isEquals & (readOnlySchema == pmfConfiguration.readOnlySchema);
        isEquals = isEquals & (fixedSchema == pmfConfiguration.fixedSchema);
        isEquals = isEquals & (foreignKeyCreateMode == pmfConfiguration.foreignKeyCreateMode);
        isEquals = isEquals & (uniqueConstraintsMapInverse == pmfConfiguration.uniqueConstraintsMapInverse);

        return isEquals;
    }

    /**
     * Equals two variables testing for nulls
     * @param var1 the first variable
     * @param var2 the second variable
     * @return true if equals
     */    
    private boolean equalsVariable(Object var1, Object var2)
    {
        if (var1 == null)
        {
            if (var2 != null)
            {
                return false;
            }
        }
        else if (!var1.equals(var2))
        {
            return false;
        }
        return true;
    }

    /**
     * Set the user name for the data store connection.
     * @param userName the user name for the data store connection.
     */
    public synchronized void setConnectionUserName(String userName)
    {
        assertConfigurable();
        this.userName = userName;
    }

    /**
     * Get the user name for the data store connection.
     * @return the user name for the data store connection.
     */
    public String getConnectionUserName()
    {
        return userName;
    }

    /**
     * Set the password for the data store connection.
     * @param password the password for the data store connection.
     */
    public synchronized void setConnectionPassword(String password)
    {
        assertConfigurable();
        this.password = password;
    }

    /**
     * Set the URL for the data store connection.
     * @param url the URL for the data store connection.
     */
    public synchronized void setConnectionURL(String url)
    {
        assertConfigurable();
        this.url = url;
    }

    /**
     * Get the URL for the data store connection.
     * @return the URL for the data store connection.
     */
    public String getConnectionURL()
    {
        return url;
    }

    /**
     * Set the driver name for the data store connection.
     * @param driverName the driver name for the data store connection.
     */
    public synchronized void setConnectionDriverName(String driverName)
    {
        assertConfigurable();
        this.driverName = driverName;
    }

    /**
     * Get the driver name for the data store connection.
     * @return the driver name for the data store connection.
     */
    public String getConnectionDriverName()
    {
        return driverName;
    }
 
    /**
     * Get the password for the data store connection.
     * @return the password for the data store connection.
     */
    public String getConnectionPassword()
    {
        return password;
    }    

    /**
     * Set the name for the data store connection factory.
     * @param connectionFactoryName name of the data store connection factory.
     */
    public synchronized void setConnectionFactoryName(String connectionFactoryName)
    {
        assertConfigurable();
        this.connectionFactoryName = connectionFactoryName;
    }

    /**
     * Get the name for the data store connection factory.
     * @return the name of the data store connection factory.
     */
    public String getConnectionFactoryName()
    {
        return connectionFactoryName;
    }

    /**
     * Set the data store connection factory.  JDO implementations
     * will support specific connection factories.  The connection
     * factory interfaces are not part of the JDO specification.
     * @param connectionFactory the data store connection factory.
     */
    public synchronized void setConnectionFactory(Object connectionFactory)
    {
        assertConfigurable();
        this.connectionFactory = connectionFactory;
    }

    /**
     * Get the data store connection factory.
     * @return the data store connection factory.
     */
    public Object getConnectionFactory()
    {
        return connectionFactory;
    }

    /**
     * Set the name for the second data store connection factory.  This is
     * needed for managed environments to get nontransactional connections for
     * optimistic transactions.
     * @param connectionFactoryName name of the data store connection factory.
     */
    public synchronized void setConnectionFactory2Name(String connectionFactoryName)
    {
        assertConfigurable();
        this.connectionFactory2Name = connectionFactoryName;
    }

    /**
     * Get the name for the second data store connection factory.  This is
     * needed for managed environments to get nontransactional connections for
     * optimistic transactions.
     * @return the name of the data store connection factory.
     */
    public String getConnectionFactory2Name()
    {
        return connectionFactory2Name;
    }

    /**
     * Set the second data store connection factory.  This is
     * needed for managed environments to get nontransactional connections for
     * optimistic transactions.  JDO implementations
     * will support specific connection factories.  The connection
     * factory interfaces are not part of the JDO specification.
     * @param connectionFactory the data store connection factory.
     */
    public synchronized void setConnectionFactory2(Object connectionFactory)
    {
        assertConfigurable();
        this.connectionFactory2 = connectionFactory;
    }

    /**
     * Get the second data store connection factory.  This is
     * needed for managed environments to get nontransactional connections for
     * optimistic transactions.
     * @return the data store connection factory.
     */
    public Object getConnectionFactory2()
    {
        return connectionFactory2;
    }

    /**
     * Set the default Multithreaded setting for all <tt>PersistenceManager</tt>
     * instances obtained from this factory.
     * @param flag the default Multithreaded setting.
     */
    public synchronized void setMultithreaded(boolean flag)
    {
        assertConfigurable();
        multithreaded = flag;
    }

    /**
     * Get the default Multithreaded setting for all <tt>PersistenceManager</tt>
     * instances obtained from this factory.
     * @return the default Multithreaded setting.
     */
    public boolean getMultithreaded()
    {
        return multithreaded;
    }

    /**
     * Set the default Optimistic setting for all <tt>PersistenceManager</tt>
     * instances obtained from this factory.
     * @param flag the default Optimistic setting.
     */
    public synchronized void setOptimistic(boolean flag)
    {
        assertConfigurable();
        optimistic = flag;
    }

    /**
     * Get the default Optimistic setting for all <tt>PersistenceManager</tt>
     * instances obtained from this factory.
     * @return the default Optimistic setting.
     */
    public boolean getOptimistic()
    {
        return optimistic;
    }

    /**
     * Set the default RetainValues setting for all <tt>PersistenceManager</tt>
     * instances obtained from this factory.
     * @param flag the default RetainValues setting.
     */
    public synchronized void setRetainValues(boolean flag)
    {
        assertConfigurable();
        retainValues = flag;
    }

    /**
     * Get the default RetainValues setting for all <tt>PersistenceManager</tt>
     * instances obtained from this factory.
     * @return the default RetainValues setting.
     */
    public boolean getRetainValues()
    {
        return retainValues;
    }

    /**
     * Set the default RestoreValues setting for all <tt>PersistenceManager</tt>
     * instances obtained from this factory.
     * @param flag the default RestoreValues setting.
     */
    public synchronized void setRestoreValues(boolean flag)
    {
        assertConfigurable();
        restoreValues = flag;
    }

    /**
     * Get the default RestoreValues setting for all <tt>PersistenceManager</tt>
     * instances obtained from this factory.
     * @return the default RestoreValues setting.
     */
    public boolean getRestoreValues()
    {
        return restoreValues;
    }

    /**
     * Set the default NontransactionalRead setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param flag the default NontransactionalRead setting.
     */
    public synchronized void setNontransactionalRead(boolean flag)
    {
        assertConfigurable();
        nontransactionalRead = flag;
    }

    /**
     * Get the default NontransactionalRead setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @return the default NontransactionalRead setting.
     */
    public boolean getNontransactionalRead()
    {
        return nontransactionalRead;
    }

    /**
     * Set the default NontransactionalWrite setting for all 
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param flag the default NontransactionalWrite setting.
     */
    public synchronized void setNontransactionalWrite(boolean flag)
    {
        if (flag)
        {
            throw new JDOUnsupportedOptionException("nontransactionalWrite is not supported by JPOX");
        }
        assertConfigurable();
        nontransactionalWrite = flag;
    }

    /**
     * Get the default NontransactionalWrite setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @return the default NontransactionalWrite setting.
     */
    public boolean getNontransactionalWrite()
    {
        return nontransactionalWrite;
    }

    /**
     * Set the default IgnoreCache setting for all <tt>PersistenceManager</tt>
     * instances obtained from this factory.
     * @param flag the default IgnoreCache setting.
     */
    public synchronized void setIgnoreCache(boolean flag)
    {
        assertConfigurable();
        ignoreCache = flag;
    }

    /**
     * Get the default IgnoreCache setting for all <tt>PersistenceManager</tt>
     * instances obtained from this factory.
     * @return the IgnoreCache setting.
     */
    public boolean getIgnoreCache()
    {
        return ignoreCache;
    }

    /**
     * Set the default DetachOnClose setting for all <tt>PersistenceManager</tt>
     * instances obtained from this factory.
     * @param flag the default DetachOnClose setting.
     */
    public synchronized void setDetachOnClose(boolean flag)
    {
        assertConfigurable();
        detachOnClose = flag;
    }

    /**
     * Get the default DetachOnClose setting for all <tt>PersistenceManager</tt>
     * instances obtained from this factory.
     * @return the DetachOnClose setting.
     */
    public boolean getDetachOnClose()
    {
        return detachOnClose;
    }

    /**
     * Set the default PersistentIdGenerator class name for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param generatorClassName the default persistent Id Generator class name.
     */
    public synchronized void setPersistentIdGenerator(String generatorClassName)
    {
        assertConfigurable();
        persistentIdGenerator = generatorClassName;
    }

    /**
     * Get the default PersistentIdGenerator class name for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @return the default persistent Id Generator class name .
     */
    public String getPersistentIdGenerator()
    {
        return persistentIdGenerator;
    }
    
    /**
     * Set the default CheckExistTablesOrViews setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param flag the default CheckExistTablesOrViews setting.
     */
    public synchronized void setCheckExistTablesOrViews(boolean flag)
    {
        assertConfigurable();
        checkExistTablesOrViews = flag;
    }

    /**
     * Get the default CheckExistTablesOrViews setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @return the default CheckExistTablesOrViews setting.
     */
    public boolean getCheckExistTablesOrViews()
    {
        return checkExistTablesOrViews;
    }
    
    /**
     * Set the default ValidateTables setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param flag the default ValidateTables setting.
     */
    public synchronized void setValidateTables(boolean flag)
    {
        assertConfigurable();
        validateTables = flag;
    }

    /**
     * Get the default ValidateTables setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @return the default ValidateTables setting.
     */
    public boolean getValidateTables()
    {
        return validateTables;
    }

    /**
     * Set the default ValidateConstraints setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param flag the default ValidateConstraints setting.
     */
    public synchronized void setValidateConstraints(boolean flag)
    {
        assertConfigurable();
        validateConstraints = flag;
    }

    /**
     * Get the default ValidateConstraints setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @return the default ValidateConstraints setting.
     */
    public boolean getValidateConstraints()
    {
        return validateConstraints;
    }
    
    /**
     * Set whether this datastore schema is readOnly
     * @param flag the readOnly setting.
     */
    public synchronized void setReadOnlySchema(boolean flag)
    {
        assertConfigurable();
        readOnlySchema = flag;
        if (flag)
        {
            setAutoCreateConstraints(false);
            setAutoCreateSchema(false);
            setAutoCreateTables(false);
            setAutoCreateColumns(false);
        }
    }

    /**
     * Get the readOnlySchema setting for the datastore managed by this PMF.
     * @return the readOnly setting.
     */
    public boolean getReadOnlySchema()
    {
        return readOnlySchema;
    }

    /**
     * Set whether this datastore schema is fixed
     * @param flag the fixed setting.
     */
    public synchronized void setFixedSchema(boolean flag)
    {
        assertConfigurable();
        fixedSchema = flag;
        if (flag)
        {
            setAutoCreateConstraints(false);
            setAutoCreateSchema(false);
            setAutoCreateTables(false);
            setAutoCreateColumns(false);
        }
    }

    /**
     * Get the fixedSchema setting for the datastore managed by this PMF.
     * @return the fixed setting.
     */
    public boolean getFixedSchema()
    {
        return fixedSchema;
    }

    /**
     * Mutator for the foreign key creation mode
     * @param mode the FK create mode
     */
    public synchronized void setForeignKeyCreateMode(String mode)
    {
        assertConfigurable();
        if (mode.equals(FOREIGN_KEY_AUTO_MODE) || mode.equals(FOREIGN_KEY_STRICT_MODE))
        {
            foreignKeyCreateMode = mode;
        }
    }

    /**
     * Accessor for the FK creation mode.
     * @return the FK creation mode
     */
    public String getForeignKeyCreateMode()
    {
        return foreignKeyCreateMode;
    }

    /**
     * Set the default UniqueConstraintsMapInverse setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param flag the default UniqueConstraintsMapInverse setting.
     */
    public synchronized void setUniqueConstraintsMapInverse(boolean flag)
    {
        assertConfigurable();
        uniqueConstraintsMapInverse = flag;
    }

    /**
     * Get the default UniqueConstraintsMapInverse setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @return the default UniqueConstraintsMapInverse setting.
     */
    public boolean getUniqueConstraintsMapInverse()
    {
        return uniqueConstraintsMapInverse;
    }

    /**
     * Get the default catalog name
     * @return the default catalog name
     */
    public String getDefaultCatalogName()
    {
        return defaultCatalogName;
    }

    /**
     * Set the default catalog name
     * @param name the default catalog name
     */
    public synchronized void setDefaultCatalogName(String name)
    {
        assertConfigurable();
        defaultCatalogName = name;
    }
 
    /**
     * Get the default schema name
     * @return the default schema name
     */
    public String getDefaultSchemaName()
    {
        return defaultSchemaName;
    }

    /**
     * Set the default schema name
     * @param name the default schema name
     */
    public synchronized void setDefaultSchemaName(String name)
    {
        assertConfigurable();
        defaultSchemaName = name;
    }
 
    /**
     * Get the fully qualified for identifiers
     * @return the default fullyQualifiedNames setting.
     */
    public boolean getFullyQualifiedNames()
    {
        return fullyQualifiedNames;
    }

    /**
     * Set the fully qualified for identifiers
     * @param flag the fullyQualifiedNames setting.
     */
    public synchronized void setFullyQualifiedNames(boolean flag)
    {
        assertConfigurable();
        fullyQualifiedNames = flag;
    }

    /**
     * Get the naming strategy
     * @return the naming strategy for identifiers
     */
    public String getIdentifierNamingStrategy()
    {
        return identifierNamingStrategy;
    }

    /**
     * Set the identifier naming strategy
     * @param strategy the identifier naming strategy
     */
    public synchronized void setIdentifierNamingStrategy(String strategy)
    {
        assertConfigurable();
        if (!identifierNamingStrategy.equals("ClassFieldName"))
        {
            throw new IllegalArgumentException(LOCALISER.msg("PMF.PropertyValueInvalid", IDENTIFIER_NAMING_STRATEGY_PROPERTY, strategy));
        }
        identifierNamingStrategy = strategy;
    }

    /**
     * Get the use update lock flag for fetches.
     * @return the default useUpdateLock setting.
     */
    public boolean getUseUpdateLock()
    {
        return useUpdateLock;
    }

    /**
     * Set whether to use the update lock for fetches.
     * @param flag the useUpdateLock setting.
     */
    public synchronized void setUseUpdateLock(boolean flag)
    {
        assertConfigurable();
        useUpdateLock = flag;
    }
 
    /**
     * Whether the datastore operations are delayed until commit.
     * In optimistic transactions this is automatically enabled.
     * @return true if datastore operations are delayed until commit
     */
    public boolean getDelayDatastoreOperationsEnabled()
    {
        return delayDatastoreOperationsEnabled;
    }

    /**
     * Whether the datastore operations are delayed until commit.
     * In optimistic transactions this is automatically enabled.
     * @param flag true if datastore operations should be delayed until commit
     */
    public synchronized void setDelayDatastoreOperationsEnabled(boolean flag)
    {
        assertConfigurable();
        delayDatastoreOperationsEnabled = flag;
    }

    /**
     * Accessor for the default inheritance strategy to use.
     * @return the default inheritance strategy
     */
    public String getDefaultInheritanceStrategy()
    {
        return defaultInheritanceStrategy;
    }

    /**
     * Mutator for the default inheritance strategy.
     * @param strategy the default inheritance strategy
     */
    public synchronized void setDefaultInheritanceStrategy(String strategy)
    {
        assertConfigurable();
        if (strategy == null)
        {
            return;
        }
        else if (!strategy.equalsIgnoreCase("JDO2") && !strategy.equalsIgnoreCase("JPOX"))
        {
            // Illegal value
            return;
        }
        defaultInheritanceStrategy = strategy;
    }

    /**
     * Set the default AutoCreateSchema setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param flag the default AutoCreateSchema setting.
     */
    public synchronized void setAutoCreateSchema(boolean flag)
    {
        assertConfigurable();
        autoCreateSchema = flag;
        if ((readOnlySchema || fixedSchema) && flag)
        {
            autoCreateSchema = false;
            JPOXLogger.JDO.warn(LOCALISER.msg("PMF.PropertyUnusableForSchema","autoCreateSchema"));
        }
    }

    /**
     * Get the default AutoCreateSchema setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @return the default AutoCreateSchema setting.
     */
    public boolean getAutoCreateSchema()
    {
        return autoCreateSchema;
    }

    /**
     * Set the default AutoCreateTables setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param flag the default AutoCreateTables setting.
     */
    public synchronized void setAutoCreateTables(boolean flag)
    {
        assertConfigurable();
        autoCreateTables = flag;
        if ((readOnlySchema || fixedSchema) && flag)
        {
            autoCreateTables = false;
            JPOXLogger.JDO.warn(LOCALISER.msg("PMF.PropertyUnusableForSchema","autoCreateTables"));
        }
    }

    /**
     * Get the default AutoCreateTables setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @return the default AutoCreateTables setting.
     */
    public boolean getAutoCreateTables()
    {
        return autoCreateTables;
    }
    
    /**
     * Set the default AutoCreateColumns setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param flag the default AutoCreateColumns setting.
     */
    public synchronized void setAutoCreateColumns(boolean flag)
    {
        assertConfigurable();
        autoCreateColumns = flag;
        if ((readOnlySchema || fixedSchema) && flag)
        {
            autoCreateTables = false;
            JPOXLogger.JDO.warn(LOCALISER.msg("PMF.PropertyUnusableForSchema","autoCreateColumns"));
        }
    }

    /**
     * Get the default AutoCreateColumns setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @return the default AutoCreateColumns setting.
     */
    public boolean getAutoCreateColumns()
    {
        return autoCreateColumns;
    }

    /**
     * Set the default AutoCreateConstraints setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param flag the default AutoCreateConstraints setting.
     */
    public synchronized void setAutoCreateConstraints(boolean flag)
    {
        assertConfigurable();
        autoCreateConstraints = flag;
        if ((readOnlySchema || fixedSchema) && flag)
        {
            autoCreateConstraints = false;
            JPOXLogger.JDO.warn(LOCALISER.msg("PMF.PropertyUnusableForSchema","autoCreateConstraints"));
        }
    }

    /**
     * Get the default AutoCreateConstraints setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @return the default AutoCreateConstraints setting.
     */
    public boolean getAutoCreateConstraints()
    {
        return autoCreateConstraints;
    }

    /**
     * Accessor for the Auto-Start Mechanism for management of which classes
     * are supported in this data store. Currently supports 2 possible values
     * for this parameter "None" and "SchemaTable" (default).
     * @return the Auto-Start Mechanism setting.
     */
    public String getAutoStartMechanism()
    {
        if (autoStartMechanism == null)
        {
            return AutoStartMechanism.NONE;
        }
        
        if (autoStartMechanism.equalsIgnoreCase(AutoStartMechanism.NONE))
        {
            return AutoStartMechanism.NONE;
        }
        else if (autoStartMechanism.equalsIgnoreCase(XMLAutoStarter.XML_STARTER))
        {
            return XMLAutoStarter.XML_STARTER;
        }
        else    // Default
        {
            return SchemaAutoStarter.SCHEMA_TABLE;
        }
    }

    /**
     * Set the default AutoStartMechanism setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     *
     * @param mechanism None        = no auto start
     *                  SchemaTable = creates a schema table (default)
     * 				    XML         = use XML file
     */
    public synchronized void setAutoStartMechanism(String mechanism)
    {
        assertConfigurable();
        if (!mechanism.equals(XMLAutoStarter.XML_STARTER) &&
            !mechanism.equals(SchemaAutoStarter.SCHEMA_TABLE) &&
            !mechanism.equals(AutoStartMechanism.NONE))
        {
            throw new IllegalArgumentException(LOCALISER.msg("PMF.PropertyValueInvalid", AUTO_START_MECHANISM_PROPERTY, mechanism));
        }
        autoStartMechanism = mechanism;
    }

    /**
     * Accessor for the name of the XML file to use when using the XML AutoStarter
     * @return The XML file name
     */
    public String getAutoStartMechanismXmlFile()
    {
        return autoStartMechanismXmlFile;
    }
    
    /**
     * Mutator for the XML filename when using the XML AutoStarter.
     * @param xmlFile Name of the XML file
     */
    public synchronized void setAutoStartMechanismXmlFile(String xmlFile)
    {
        autoStartMechanismXmlFile = xmlFile;
    }
    
    /**
     * Accessor for the Auto-Start Mechanism Mode for how the auto start
     * mechanism operates. Currently supports 3 possible values for this
     * parameter "Checked", "Ignored", and "Quiet" (default).
     * @return the Auto-Start Mechanism Mode setting.
     */
    public String getAutoStartMechanismMode()
    {
        if (autoStartMechanismMode != null &&
            autoStartMechanismMode.equalsIgnoreCase(AutoStartMechanism.MODE_CHECKED))
        {
            return AutoStartMechanism.MODE_CHECKED;
        }
        else if (autoStartMechanismMode != null &&
                 autoStartMechanismMode.equalsIgnoreCase(AutoStartMechanism.MODE_IGNORED))
        {
            return AutoStartMechanism.MODE_IGNORED;
        }
        else    // Default
        {
            return AutoStartMechanism.MODE_QUIET;
        }
    }

    /**
     * Set the default AutoStartMechanismMode setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param mode Checked = raise exceptions on errors
     *             Ignored = proceed without corrections
     *             Quiet   = correct errors and proceed (default)
     */
    public synchronized void setAutoStartMechanismMode(String mode)
    {
        assertConfigurable();
        if (!mode.equals(AutoStartMechanism.MODE_CHECKED) &&
            !mode.equals(AutoStartMechanism.MODE_IGNORED) &&
            !mode.equals(AutoStartMechanism.MODE_QUIET))
        {
            throw new IllegalArgumentException(LOCALISER.msg("PMF.PropertyValueInvalid", AUTO_START_MECHANISM_MODE_PROPERTY, mode));
        }
        autoStartMechanismMode = mode;
    }

    /**
     * Get the default transaction isolation level for all
     * <tt>PoidGenerator</tt> instances.
     *
     * @return the default transaction isolation level for 
     *         <tt>PoidGenerator</tt> instances.
     */
    public int getPoidTransactionIsolationLevel()
    {
        return poidIsolationLevel;
    }
 
    /**
     * Set the default transaction isolation level for all
     * <tt>PoidGenerator</tt> instances.
     * @param isolationLevelName
     *      One of the values "read uncommitted", "read committed", "repeatable
     *      read", or "serializable".  The name is case-insensitive, and spaces
     *      and underscores are equivalent.
     */
    protected synchronized void setPoidTransactionIsolation(String isolationLevelName)
    {
        String iln = isolationLevelName.trim().replace(' ', '_').toUpperCase();
        if (iln.equals("READ_UNCOMMITTED"))
        {
            poidIsolationLevel = Connection.TRANSACTION_READ_UNCOMMITTED;
        }
        else if (iln.equals("READ_COMMITTED"))
        {
            poidIsolationLevel = Connection.TRANSACTION_READ_COMMITTED;
        }
        else if (iln.equals("REPEATABLE_READ"))
        {
            poidIsolationLevel = Connection.TRANSACTION_REPEATABLE_READ;
        }
        else if (iln.equals("SERIALIZABLE"))
        {
            poidIsolationLevel = Connection.TRANSACTION_SERIALIZABLE;
        }
        else
        {
            throw new IllegalArgumentException(LOCALISER.msg("PMF.PropertyValueInvalid", POID_TRANSACTION_ISOLATION_PROPERTY, isolationLevelName));
        }
    }

    /**
     * Set the default transaction isolation level for all
     * <tt>PoidGenerator</tt> instances.
     * @param isolationLevel the default transaction isolation level.
     */
    public synchronized void setPoidTransactionIsolation(int isolationLevel)
    {
        switch (isolationLevel)
        {
            case Connection.TRANSACTION_READ_UNCOMMITTED:
            case Connection.TRANSACTION_READ_COMMITTED:
            case Connection.TRANSACTION_REPEATABLE_READ:
            case Connection.TRANSACTION_SERIALIZABLE:
                break;

            default:
                throw new IllegalArgumentException(LOCALISER.msg("PMF.PropertyValueInvalid", POID_TRANSACTION_ISOLATION_PROPERTY, "" + isolationLevel));
        }

        this.poidIsolationLevel = isolationLevel;
    }

    /**
     * Whether to use the PM connection or open a new connection
     * @return Returns the poidTransactionAttribute.
     */
    public String getPoidTransactionAttribute()
    {
        return poidTransactionAttribute;
    }

    /**
     * Set whether to use the PM connection or open a new connection
     * @param poidTransactionAttribute The poidTransactionAttribute to set.
     */
    public void setPoidTransactionAttribute(String poidTransactionAttribute)
    {
        this.poidTransactionAttribute = poidTransactionAttribute;
    }

    /**
     * Accessor for the fetch size to use for ResultSet's.
     * @return The fetch size
     */
    public int getQueryFetchSize()
    {
        return queryFetchSize;
    }

    /**
     * Set the fetch size to use for ResultSet's.
     * @param size The fetch size to use
     */
    public void setQueryFetchSize(int size)
    {
        this.queryFetchSize = size;
    }

    /**
     * Accessor for the fetch direction to use for ResultSet's.
     * @return The fetch direction.
     */
    public String getQueryFetchDirection()
    {
        return queryFetchDirection;
    }

    /**
     * Set the fetch direction to use for ResultSet's.
     * @param dir The fetch direction to use
     */
    public void setQueryFetchDirection(String dir)
    {
        if (dir == null ||
            !dir.equals("forward") ||
            !dir.equals("reverse") ||
            !dir.equals("unknown"))
        {
            return;
        }
        queryFetchDirection = dir;
    }

    /**
     * Accessor for the type of ResultSet
     * @return The ResultSet type.
     */
    public String getQueryResultSetType()
    {
        return queryResultSetType;
    }

    /**
     * Set the type of the ResultSet
     * @param type The ResultSet type
     */
    public void setQueryResultSetType(String type)
    {
        if (type == null ||
            !type.equals("forward-only") ||
            !type.equals("scroll-sensitive") ||
            !type.equals("scroll-insensitive"))
        {
            return;
        }
        queryResultSetType = type;
    }

    /**
     * Accessor for the concurrency of ResultSet
     * @return The ResultSet concurrency.
     */
    public String getQueryResultSetConcurrency()
    {
        return queryResultSetConcurrency;
    }

    /**
     * Set the concurrency of the ResultSet
     * @param concur The ResultSet concurrency
     */
    public void setQueryResultSetConcurrency(String concur)
    {
        if (concur == null ||
            !concur.equals("read-only") ||
            !concur.equals("updateable"))
        {
            return;
        }
        queryResultSetConcurrency = concur;
    }

    /**
     * Accessor for the default max length for strings.
     * @return The default length for strings
     */
    public int getStringDefaultLength()
    {
        return stringDefaultLength;
    }

    /**
     * Set the default length for strings.
     * @param len Default length for strings
     */
    public void setStringDefaultLength(int len)
    {
        this.stringDefaultLength = len;
    }

    /**
     * Accessor for the timezone in which all dates are stored in the datastore.
     * @return The timezone for all dates in the datastore
     */
    public String getDateTimezone()
    {
        return dateTimezone;
    }

    /**
     * Set the timezone to use for dates in the datastore.
     * @param timezone The timezone to use for all dates in the datastore
     */
    public void setDateTimezone(String timezone)
    {
        this.dateTimezone = timezone;
        DATE_TIMEZONE = TimeZone.getTimeZone(timezone);
        if (DATE_TIMEZONE == null)
        {
            DATE_TIMEZONE = TimeZone.getTimeZone("UTC");
        }
    }

    /**
     * Accessor for whether to cache collections.
     * @return whether to cache collections.
     */
    public boolean getJdoCacheCollections()
    {
        return jdoCacheCollections;
    }

    /**
     * Set whether to cache collections for this PMF.
     * @param cache Whether to cache collections.
     */
    public synchronized void setJdoCacheCollections(boolean cache)
    {
        assertConfigurable();
        jdoCacheCollections = cache;
    }
    
    /**
     * Accessor for the JDO Level 1 Cache Type
     * @return the JDO Level 1 Cache Type
     */
    public String getJdoCacheLevel1Type()
    {
        return jdoCacheLevel1Type;
    }

    /**
     * Set the default Level 1 Cache Type for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param type the Level 1 Cache Type
     */
    public synchronized void setJdoCacheLevel1Type(String type)
    {
        assertConfigurable();
        jdoCacheLevel1Type = type;
    }
    
    /**
     * Accessor for whether to use a level 2 Cache.
     * @return whether to use a level 2 Cache.
     */
    public boolean getJdoCacheLevel2()
    {
        return jdoCacheLevel2;
    }

    /**
     * Set whether to use a Level 2 Cache for this PMF.
     * @param cache Whether to use a level 2 Cache
     */
    public synchronized void setJdoCacheLevel2(boolean cache)
    {
        assertConfigurable();
        jdoCacheLevel2 = cache;
    }

    /**
     * Accessor for the JDO Level 2 Cache Type
     * @return the JDO Level 2 Cache Type
     */
    public String getJdoCacheLevel2Type()
    {
        return jdoCacheLevel2Type;
    }

    /**
     * Set the default Level 2 Cache Type for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param type the Level 2 Cache Type
     */
    public synchronized void setJdoCacheLevel2Type(String type)
    {
        assertConfigurable();
        jdoCacheLevel2Type = type;
    }

    /**
     * Accessor for the JDO Level 2 Cache Name
     * @return the JDO Level 2 Cache Name
     */
    public String getJdoCacheLevel2CacheName()
    {
        return jdoCacheLevel2CacheName;
    }

    /**
     * Set the default Level 2 Cache Name for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param name Name of the cache
     */
    public synchronized void setJdoCacheLevel2CacheName(String name)
    {
        assertConfigurable();
        jdoCacheLevel2CacheName = name;
    }

    /**
     * Accessor for the Store Dictionary Case Identifier
     * @return the Store Dictionary Case Identifier
     */
    public String getStoreDictionaryCaseIdentifier()
    {
        return dictionary.getCaseIdentifier().toString();
    }

    /**
     * Accessor for the Store Dictionary Case Identifier
     * @param value the Store Dictionary Case Identifier
     */
    public synchronized void setStoreDictionaryCaseIdentifier(String value)
    {
        assertConfigurable();
        dictionary.setCaseIdentifier(Dictionary.CaseIdentifier.getCaseIdentifier(value));
    }   
 
    /**
     * Set the default transaction isolation level for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param isolationLevelName
     *      One of the values "read uncommitted", "read committed", "repeatable
     *      read", or "serializable".  The name is case-insensitive, and spaces
     *      and underscores are equivalent.
     */
    protected synchronized void setTransactionIsolation(String isolationLevelName)
    {
        assertConfigurable();

        String iln = isolationLevelName.trim().replace(' ', '_').toUpperCase();
        if (iln.equals("READ_UNCOMMITTED"))
        {
            isolationLevel = Connection.TRANSACTION_READ_UNCOMMITTED;
        }
        else if (iln.equals("READ_COMMITTED"))
        {
            isolationLevel = Connection.TRANSACTION_READ_COMMITTED;
        }
        else if (iln.equals("REPEATABLE_READ"))
        {
            isolationLevel = Connection.TRANSACTION_REPEATABLE_READ;
        }
        else if (iln.equals("SERIALIZABLE"))
        {
            isolationLevel = Connection.TRANSACTION_SERIALIZABLE;
        }
        else
        {
            throw new IllegalArgumentException(LOCALISER.msg("PMF.PropertyValueInvalid", TRANSACTION_ISOLATION_PROPERTY, isolationLevelName));
        }
    }

    /**
     * Set the default transaction isolation level for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param isolationLevel the default transaction isolation level.
     */
    public synchronized void setTransactionIsolation(int isolationLevel)
    {
        assertConfigurable();

        switch (isolationLevel)
        {
            case Connection.TRANSACTION_READ_UNCOMMITTED:
            case Connection.TRANSACTION_READ_COMMITTED:
            case Connection.TRANSACTION_REPEATABLE_READ:
            case Connection.TRANSACTION_SERIALIZABLE:
                break;

            default:
                throw new IllegalArgumentException(LOCALISER.msg("PMF.PropertyValueInvalid", TRANSACTION_ISOLATION_PROPERTY, "" + isolationLevel));
        }

        this.isolationLevel = isolationLevel;
    }

    /**
     * Get the default transaction isolation level for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @return the default transaction isolation level.
     */
    public int getTransactionIsolation()
    {
        return isolationLevel;
    }

    /**
     * Accessor for the dictionary. 
     * @return Returns the dictionary.
     */
    public Dictionary getDictionary()
    {
        return dictionary;
    }

    /**
     * Accessor for the JDO MetaData file extension. 
     * @return Returns the JDO metadata File Extension.
     */
    public String getJdoMetaDataFileExtension()
    {
        return jdoMetaDataFileExtension;
    }
 
    /**
     * Mutator for the JDO MetaData file extension. 
     * @param metadataFileExtension The JDO metadata File Extension to set.
     */
    public void setJdoMetaDataFileExtension(String metadataFileExtension)
    {
        this.jdoMetaDataFileExtension = metadataFileExtension;
    }

    /**
     * Accessor for the ORM MetaData file extension. 
     * @return Returns the ORM metadata File Extension.
     */
    public String getOrmMetaDataFileExtension()
    {
        return ormMetaDataFileExtension;
    }
 
    /**
     * Mutator for the ORM MetaData file extension. 
     * @param metadataFileExtension The ORM metadata File Extension to set.
     */
    public void setOrmMetaDataFileExtension(String metadataFileExtension)
    {
        this.ormMetaDataFileExtension = metadataFileExtension;
    }

    /**
     * Set the name for any mapping, used in searching for ORM/Query metadata
     * files.
     * @param mapping the mapping name
     */
    public synchronized void setMapping(String mapping)
    {
        assertConfigurable();
        this.mapping = mapping;
    }

    /**
     * Get the name for any mapping, used in retrieving metadata files for
     * ORM/Query data.
     * @return the name for the mapping.
     */
    public String getMapping()
    {
        return mapping;
    }
    
    /**
     * Asserts that a change to a configuration property is allowed.
     */
    protected void assertConfigurable()
    {
        if (!configurable)
        {
            throw new JDOUserException(LOCALISER.msg("PMF.NotConfigurableAfterGeneration"));
        }
    }
}