/**********************************************************************
Copyright (c) 2002 Kelly Grizzle and others. All rights reserved.
This program and the accompanying materials are made available under
the terms of the JPOX License v1.0 which accompanies this distribution.

Contributors:
2002 Mike Martin - unknown changes
2003 Erik Bengtson - refactored the persistent id generator System property
2003 Andy Jefferson - introduction of LOCALISER, and comments
2003 Erik Bengtson - removed unused variables
2004 Andy Jefferson - converted to use Logger
2004 Andy Jefferson - removed autoCreate properties
2004 Andy Jefferson - added readOnly
 	...
 **********************************************************************/
package org.jpox;

import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.ObjectOutputStream;
import java.util.Enumeration;
import java.util.HashMap;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;

import javax.jdo.PersistenceManager;
import javax.jdo.PersistenceManagerFactory;
import javax.naming.Context;
import javax.naming.Name;
import javax.naming.Reference;
import javax.naming.Referenceable;
import javax.naming.StringRefAddr;
import javax.naming.spi.ObjectFactory;

import org.jpox.store.Dictionary;
import org.jpox.util.JPOXLogger;

/**
 * A factory used to obtain {@link javax.jdo.PersistenceManager} instances.
 *
 * @version $Revision: 1.12 $
 */
public class PersistenceManagerFactoryImpl extends AbstractPersistenceManagerFactory 
    implements PersistenceManagerFactory, ObjectFactory, Referenceable
{
    /**
     * This is an interface that allows setting properties on a
     * PersistenceManagerFactory with Strings.
     */
    public interface PropertySetter
    {
        /**
         * Set this property on the given PersistenceManagerFactoryImpl to
         * the String value.
         * @param configuration    The PMFConfiguration to set the value on.
         * @param value            The String value to set.
         */
        void set(PMFConfiguration configuration, String value);
    }

    /**
     * A Map mapping property name to the corresponding PropertySetter.
     */
    private static final Map PROPERTY_SETTERS = initPropertySetters();

    /**
     * Return a new PersistenceManagerFactoryImpl with options set according
     * to the given Properties.
     * @param props The Properties to initialize the PersistenceManagerFactory with.
     * @return  A PersistenceManagerFactoryImpl with options set according to
     *          the given Properties.
     * @see javax.jdo.JDOHelper#getPersistenceManagerFactory(java.util.Properties)
     * @exception UnsupportedConnectionFactoryException Thrown if the
     * configured connection factory is not a javax.sql.DataSource 
     */
    public synchronized static PersistenceManagerFactory getPersistenceManagerFactory(Properties props)
    {
        final PersistenceManagerFactoryImpl pmf = new PersistenceManagerFactoryImpl();

        // Read properties.
        pmf.setOptions(props);

        // Freeze the PMF config now that we are handing out PM's
        // see JDO 1.0.1 section 11.7. The returned object from this method has to be read-only
        pmf.freezeConfiguration();
        
        return pmf;
    }

    /**
     * Initialize the PROPERTY_SETTERS Map.
     * @return  The PROPERTY_SETTERS Map.
     */
    private static final Map initPropertySetters()
    {
        final Map map = new HashMap();
        
        // JDO Standard properties
        map.put("javax.jdo.option.Optimistic", new PropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setOptimistic(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put("javax.jdo.option.RetainValues", new PropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setRetainValues(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put("javax.jdo.option.RestoreValues", new PropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setRestoreValues(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put("javax.jdo.option.IgnoreCache", new PropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setIgnoreCache(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put("javax.jdo.option.DetachOnClose", new PropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    configuration.setDetachOnClose(Boolean.valueOf(s).booleanValue());
                }
            });
        map.put("javax.jdo.option.NontransactionalRead", new PropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setNontransactionalRead(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put("javax.jdo.option.NontransactionalWrite", new PropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setNontransactionalWrite(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put("javax.jdo.option.Multithreaded", new PropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setMultithreaded(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put(PMFConfiguration.JDO_DATASTORE_USERNAME_PROPERTY, new PropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setConnectionUserName(s);
            }
        });
        map.put(PMFConfiguration.JDO_DATASTORE_PASSWORD_PROPERTY, new PropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setConnectionPassword(s);
            }
        });
        map.put(PMFConfiguration.JDO_DATASTORE_DRIVERNAME_PROPERTY, new PropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setConnectionDriverName(s);
            }
        });
        map.put(PMFConfiguration.JDO_DATASTORE_URL_PROPERTY, new PropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setConnectionURL(s);
            }
        });
        map.put("javax.jdo.option.ConnectionFactoryName", new PropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setConnectionFactoryName(s);
            }
        });
        map.put("javax.jdo.option.ConnectionFactory2Name", new PropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setConnectionFactory2Name(s);
            }
        });
        map.put(PMFConfiguration.JDO_MAPPING_PROPERTY, new PropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setMapping(s);
            }
        });

        // JPOX Extension properties
        map.put(PMFConfiguration.GENERATOR_ID_CLASS_PROPERTY, new PropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setPersistentIdGenerator(s);
            }
        });
        map.put(PMFConfiguration.CHECK_EXIST_TABLES_VIEWS_PROPERTY, new PropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setCheckExistTablesOrViews(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put(PMFConfiguration.VALIDATE_TABLES_PROPERTY, new PropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setValidateTables(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put(PMFConfiguration.ADD_UNIQUE_CONSTRAINT_MAP_INVERSE_PROPERTY, new PropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setUniqueConstraintsMapInverse(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put(PMFConfiguration.VALIDATE_CONSTRAINTS_PROPERTY, new PropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setValidateConstraints(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put(PMFConfiguration.AUTO_CREATE_SCHEMA_PROPERTY, new PropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setAutoCreateSchema(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put(PMFConfiguration.AUTO_CREATE_TABLES_PROPERTY, new PropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setAutoCreateTables(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put(PMFConfiguration.AUTO_CREATE_COLUMNS_PROPERTY, new PropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    configuration.setAutoCreateColumns(Boolean.valueOf(s).booleanValue());
                }
            });
        map.put(PMFConfiguration.AUTO_CREATE_CONSTRAINTS_PROPERTY, new PropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setAutoCreateConstraints(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put(PMFConfiguration.AUTO_START_MECHANISM_PROPERTY, new PropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setAutoStartMechanism(s);
            }
        });
        map.put(PMFConfiguration.AUTO_START_MECHANISM_XML_FILE_PROPERTY, new PropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setAutoStartMechanismXmlFile(s);
            }
        });
        map.put(PMFConfiguration.AUTO_START_MECHANISM_MODE_PROPERTY, new PropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setAutoStartMechanismMode(s);
            }
        });

        map.put(PMFConfiguration.DEFAULT_CATALOG_NAME_PROPERTY, new PropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setDefaultCatalogName(s);
            }
        });
        map.put(PMFConfiguration.DEFAULT_SCHEMA_NAME_PROPERTY, new PropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setDefaultSchemaName(s);
            }
        });
        map.put(PMFConfiguration.FULLY_QUALIFIED_NAMES_PROPERTY, new PropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setFullyQualifiedNames(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put(PMFConfiguration.IDENTIFIER_NAMING_STRATEGY_PROPERTY, new PropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    configuration.setIdentifierNamingStrategy(s);
                }
            });
        map.put(PMFConfiguration.READ_ONLY_SCHEMA_PROPERTY, new PropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    configuration.setReadOnlySchema(Boolean.valueOf(s).booleanValue());
                }
            });
        map.put(PMFConfiguration.FIXED_SCHEMA_PROPERTY, new PropertySetter()
            {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setFixedSchema(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put(PMFConfiguration.FOREIGN_KEY_CREATE_MODE_PROPERTY, new PropertySetter()
            {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setForeignKeyCreateMode(s);
            }
        });

        map.put(PMFConfiguration.DEFAULT_INHERITANCE_STRATEGY_PROPERTY, new PropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    configuration.setDefaultInheritanceStrategy(s);
                }
            });
        map.put(PMFConfiguration.USE_UPDATE_LOCK_PROPERTY, new PropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setUseUpdateLock(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put(PMFConfiguration.DELAY_DATASTORE_OPERATIONS_UNTIL_COMMIT_PROPERTY, new PropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setDelayDatastoreOperationsEnabled(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put(PMFConfiguration.TRANSACTION_ISOLATION_PROPERTY, new PropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setTransactionIsolation(s);
            }
        });
        map.put(PMFConfiguration.POID_TRANSACTION_ATTRIBUTE_PROPERTY, new PropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setPoidTransactionAttribute(s);
            }
        });
        map.put(PMFConfiguration.POID_TRANSACTION_ISOLATION_PROPERTY, new PropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setPoidTransactionIsolation(s);
            }
        });

        // Query
        map.put(PMFConfiguration.QUERY_FETCH_SIZE_PROPERTY, new PropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    try
                    {
                        int len = (new Integer(s)).intValue();
                        configuration.setQueryFetchSize(len);
                    }
                    catch (NumberFormatException nfe)
                    {
                        // Do nothing
                    }
                }
            });
        map.put(PMFConfiguration.QUERY_FETCH_DIRECTION_PROPERTY, new PropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    configuration.setQueryFetchDirection(s);
                }
            });
        map.put(PMFConfiguration.QUERY_RESULT_SET_TYPE_PROPERTY, new PropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    configuration.setQueryResultSetType(s);
                }
            });
        map.put(PMFConfiguration.QUERY_RESULT_SET_CONCURRENCY_PROPERTY, new PropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    configuration.setQueryResultSetConcurrency(s);
                }
            });

        // Types
        map.put(PMFConfiguration.STRING_DEFAULT_LENGTH_PROPERTY, new PropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    try
                    {
                        int len = (new Integer(s)).intValue();
                        configuration.setStringDefaultLength(len);
                    }
                    catch (NumberFormatException nfe)
                    {
                        // Do nothing
                    }
                }
            });
        map.put(PMFConfiguration.DATE_TIMEZONE_PROPERTY, new PropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    configuration.setDateTimezone(s);
                }
            });

        // Cache
        map.put(PMFConfiguration.JDO_CACHE_COLLECTIONS_PROPERTY, new PropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setJdoCacheCollections(Boolean.valueOf(s).booleanValue());
            }
        });
        map.put(PMFConfiguration.JDO_CACHE_LEVEL_1_TYPE_PROPERTY, new PropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    configuration.setJdoCacheLevel1Type(s);
                }
            });
        map.put(PMFConfiguration.JDO_CACHE_LEVEL_2_PROPERTY, new PropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    configuration.setJdoCacheLevel2(Boolean.valueOf(s).booleanValue());
                }
            });
        map.put(PMFConfiguration.JDO_CACHE_LEVEL_2_TYPE_PROPERTY, new PropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    configuration.setJdoCacheLevel2Type(s);
                }
            });
        map.put(PMFConfiguration.JDO_CACHE_LEVEL_2_CACHE_NAME_PROPERTY, new PropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    configuration.setJdoCacheLevel2CacheName(s);
                }
            });

        map.put(Dictionary.STORE_DICTIONARY_CASE_IDENTIFIER, new PropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setStoreDictionaryCaseIdentifier(s);
            }
        });
        map.put(PMFConfiguration.JDO_METADATA_FILE_EXTENSION_PROPERTY, new PropertySetter()
        {
            public void set(PMFConfiguration configuration, String s)
            {
                configuration.setJdoMetaDataFileExtension(s);
            }
        });
        map.put(PMFConfiguration.ORM_METADATA_FILE_EXTENSION_PROPERTY, new PropertySetter()
            {
                public void set(PMFConfiguration configuration, String s)
                {
                    configuration.setOrmMetaDataFileExtension(s);
                }
            });

        return map;
    }

    /**
     * Constructs a new PersistenceManagerFactoryImpl.
     */
    public PersistenceManagerFactoryImpl()
    {
        String tip = null;
        try
        {
            tip = System.getProperty(PMFConfiguration.TRANSACTION_ISOLATION_PROPERTY);
        }
        catch (SecurityException e)
        {
            JPOXLogger.GENERAL.warn(e.getMessage(), e);
        }

        if (tip != null)
        {
            try
            {
                setTransactionIsolation(tip);
            }
            catch (IllegalArgumentException e)
            {
                JPOXLogger.TRANSACTION.warn(LOCALISER.msg("PMF.TransactionIsolationInvalid", PMFConfiguration.TRANSACTION_ISOLATION_PROPERTY, tip));
            }
        }
    }

    /**
     * Set the options for this PersistenceManagerFactory based on the
     * given Properties.
     * @param props  The Properties to set the options from.
     */
    private void setOptions(Properties props)
    {
        for (Enumeration e = props.propertyNames(); e.hasMoreElements();)
        {
            String prop = (String) e.nextElement();

            PropertySetter setter = (PropertySetter) PROPERTY_SETTERS.get(prop);
            if (prop != null && setter != null)
            {
                String value = props.getProperty(prop);
                setter.set(this.getPMFConfiguration(), value);
            }
            else if (prop != null && setter == null)
            {
                // Omit the ones that we dont have setters for
                if (!prop.equals("javax.jdo.PersistenceManagerFactoryClass"))
                {
                    JPOXLogger.JDO.warn(LOCALISER.msg("PMF.UnknownProperty",prop));
                }
            }
        }
    }

    /**
     * Get an instance of <tt>PersistenceManager</tt> from this factory.  The
     * instance has default values for options.
     *
     * <p>After the first use of getPersistenceManager, no "set" methods will
     * succeed.</p>
     *
     * @return a <tt>PersistenceManager</tt> instance with default options.
     * @exception UnsupportedConnectionFactoryException Thrown if the
     * configured connection factory is not a javax.sql.DataSource 
     */
    public synchronized PersistenceManager getPersistenceManager()
    {
        // Freeze the PMF config now that we are handing out PM's
        freezeConfiguration();

        return getPersistenceManager(this.getPMFConfiguration().getConnectionUserName(), this.getPMFConfiguration().getConnectionPassword());
    }

    /**
     * Get an instance of <tt>PersistenceManager</tt> from this factory.  The
     * instance has default values for options.  The parameters userid and
     * password are used when obtaining datastore connections from the
     * connection pool.
     *
     * <p>After the first use of getPersistenceManager, no "set" methods will
     * succeed.</p>
     *
     * @param userName  the user name for the connection
     * @param password  the password for the connection
     *
     * @return a <tt>PersistenceManager</tt> instance with default options.
     * @exception UnsupportedConnectionFactoryException Thrown if the
     * configured connection factory is not a javax.sql.DataSource 
     */
    public synchronized PersistenceManager getPersistenceManager(String userName, String password)
    {
        // Freeze the PMF config now that we are handing out PM's
        freezeConfiguration();

        PersistenceManager pm = new PersistenceManagerImpl(this, userName, password);

        // Add the lifecycle listeners to the PM
        Iterator listenerIter = lifecycleListeners.iterator();
        while (listenerIter.hasNext())
        {
            LifecycleListenerSpecification listener = (LifecycleListenerSpecification) listenerIter.next();
            pm.addLifecycleListener(listener.getListener(), listener.getClasses());
        }

        return pm;
    }

    /**
     * Equality operator.
     * @param obj Object to compare against
     * @return Whether the objects are the same.
     **/
    public synchronized boolean equals(Object obj)
    {
        if (obj == this)
        {
            return true;
        }

        if (!(obj instanceof PersistenceManagerFactoryImpl))
        {
            return false;
        }

        return super.equals(obj);
    }

    /**
     * Uses StringRefAddr's to store the information
     * @param obj The object
     * @param name Name
     * @param ctx Context
     * @param env properties
     * @return The instance
     * @throws Exception
     */
    public Object getObjectInstance(Object obj, Name name, Context ctx, Hashtable env) throws Exception
    {
        PersistenceManagerFactoryImpl rc = null;

        if (obj instanceof Reference)
        {
            Reference ref = (Reference) obj;
            if (ref.getClassName().equals(PersistenceManagerFactoryImpl.class.getName()))
            {
                Properties p = new Properties();
                for (Enumeration e = ref.getAll(); e.hasMoreElements();)
                {
                    StringRefAddr sra = (StringRefAddr) e.nextElement();
                    p.setProperty(sra.getType(), (String) sra.getContent());
                }

                rc = new PersistenceManagerFactoryImpl();
                rc.setOptions(p);
            }
        }
        return rc;
    }

    /**
     * Uses StringRefAddr's to store the information
     * @return The reference
     */
    public Reference getReference()
    {
        Reference rc = null;
        ByteArrayOutputStream baos = new ByteArrayOutputStream();

        try
        {
            ObjectOutputStream oos = new ObjectOutputStream(baos);
            oos.writeObject(this);
            rc = new Reference(PersistenceManagerFactoryImpl.class.getName(), PersistenceManagerFactoryImpl.class.getName(), null);
            
            Properties p = new Properties();
            for (Enumeration e = p.propertyNames(); e.hasMoreElements();)
            {
                String key = (String) e.nextElement();
                String value = p.getProperty(key);
                rc.add(new StringRefAddr(key, value));
            }
        }
        catch (IOException ex)
        {
            // do something
        }
        return rc;
    }
}