/**********************************************************************
Copyright (c) 2003 Andy Jefferson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2004 Joerg Van Frantzius - changes to support a form of DDL output
2004 Erik Bengtson - dbinfo() mode
2004 Andy Jefferson - added "mapping" property to allow ORM files
    ...
**********************************************************************/
package org.jpox;

import java.io.File;
import java.io.FileWriter;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Properties;
import java.util.regex.Pattern;

import javax.jdo.JDOHelper;
import javax.jdo.PersistenceManagerFactory;

import org.jpox.metadata.MetaDataUtils;
import org.jpox.store.StoreManager;
import org.jpox.store.StoreManagerFactory;
import org.jpox.store.rdbms.adapter.DatabaseAdapter;
import org.jpox.store.rdbms.table.AbstractTable;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;

/**
 * Provides user-domain utilities for the maintenace of schemas.
 * These utilities cover the creation of a set of tables representing classes
 * specified in input JDO meta-data files. Additionally the tables representing
 * these classes can be deleted. 
 *
 * @version $Revision: 1.11 $
 */
public class SchemaTool
{
    private static final Localiser LOCALISER=Localiser.getInstance("org.jpox.Localisation");

    /**
     * The file we dump any DDL SQL to, also meaning we are not supposed to actually
     * create any tables in the DB
     */
    private static FileWriter ddlFileWriter = null;

    /** Method to handle the creation of a schema's tables.
     * @param jdo_files Array of JDO files defining the classes to be
     *      persisted (and hence the tables to be created).
     * @param verbose Verbose output
     * @return Return code
     * @throws Exception Thrown when either an error occurs parsing the
     *                   MetaData, or the DB definition is not defined.
     **/
    public static int createSchemaTables(String[] jdo_files,boolean verbose)
    throws Exception
    {
        // Build up set of classes
        Map  classToMetaDataMap=new HashMap();
        List class_names=null;
        try
        {
            class_names = MetaDataUtils.getOrderedClassNamesFromFileList(
                            jdo_files,classToMetaDataMap);
        }
        catch (Exception e)
        {
            String msg = LOCALISER.msg("SchemaTool.MetaDataParseError", e);
            JPOXLogger.SCHEMATOOL.error(msg);
            System.err.println(msg);

            // Add the exception to the log
            JPOXLogger.SCHEMATOOL.error(e);

            throw new Exception(msg, e);
        }
        if (class_names != null && class_names.size() > 0)
        {
            String[] class_name_array=new String[class_names.size()];
            Iterator class_names_iter=class_names.iterator();
            int i=0;
            while (class_names_iter.hasNext())
            {
                class_name_array[i++] = (String)class_names_iter.next();
            }

            // Set up properties for the data store
            Properties props=null;
            try
            {
                props = getPropertiesForDatastore(verbose);
                addProperty(props, PMFConfiguration.AUTO_CREATE_TABLES_PROPERTY, "true", verbose);

                // Respect any setting of AUTO_CREATE_CONSTRAINTS in case they dont want the constraints
                String autoCreateConstraints = System.getProperty(PMFConfiguration.AUTO_CREATE_CONSTRAINTS_PROPERTY);
                if (autoCreateConstraints != null)
                {
                    addProperty(props, PMFConfiguration.AUTO_CREATE_CONSTRAINTS_PROPERTY, autoCreateConstraints, verbose);
                }
                else
                {
                    addProperty(props, PMFConfiguration.AUTO_CREATE_CONSTRAINTS_PROPERTY, "true", verbose);
                }

                String autoCreateColumns = System.getProperty(PMFConfiguration.AUTO_CREATE_COLUMNS_PROPERTY);
                if (autoCreateColumns != null)
                {
                    addProperty(props, PMFConfiguration.AUTO_CREATE_COLUMNS_PROPERTY, autoCreateColumns, verbose);
                }
                else
                {
                    addProperty(props, PMFConfiguration.AUTO_CREATE_COLUMNS_PROPERTY, "false", verbose);
                }

                // Respect any setting of VALIDATE_TABLES
                String validateTables = System.getProperty(PMFConfiguration.VALIDATE_TABLES_PROPERTY);
                if (validateTables != null)
                {
                    addProperty(props, PMFConfiguration.VALIDATE_TABLES_PROPERTY, validateTables, verbose);
                }
                else
                {
                    addProperty(props, PMFConfiguration.VALIDATE_TABLES_PROPERTY, "false", verbose);
                }

                // Respect any setting of VALIDATE_CONSTRAINTS
                String validateConstraints = System.getProperty(PMFConfiguration.VALIDATE_CONSTRAINTS_PROPERTY);
                if (validateConstraints != null)
                {
                    addProperty(props, PMFConfiguration.VALIDATE_CONSTRAINTS_PROPERTY, validateConstraints, verbose);
                }
                else
                {
                    addProperty(props, PMFConfiguration.VALIDATE_CONSTRAINTS_PROPERTY, "false", verbose);
                }
            }
            catch (Exception e)
            {
                System.err.println(e.getMessage());
                throw e;
            }

            if (ddlFileWriter != null)
            {
                // the tables must not be created in the DB, so do not validate
                // (SQL is being dumped to a file)
                addProperty(props, PMFConfiguration.VALIDATE_CONSTRAINTS_PROPERTY, "false", verbose);
                addProperty(props, PMFConfiguration.VALIDATE_TABLES_PROPERTY, "false", verbose);
            }
            if (verbose)
            {
                System.out.println("");
            }

            // Create a PersistenceManager for this store and create the tables
            PersistenceManagerFactory pmf = JDOHelper.getPersistenceManagerFactory(props);
            StoreManager store_mgr = StoreManagerFactory.getStoreManager(pmf.getPersistenceManager());
            store_mgr.addClasses(class_name_array);

            return 0;
        }
        else
        {
            String msg = LOCALISER.msg("SchemaTool.NoClassesError");
            JPOXLogger.SCHEMATOOL.error(msg);
            System.err.println(msg);

            throw new Exception(msg);
        }
    }

    /** Method to handle the deletion of a schema's tables.
     * @param jdo_files Array of JDO files defining the classes to be
     *      unpersisted (and hence the tables to be deleted).
     * @param verbose Verbose output
     * @return Return code
     * @throws Exception Thrown when either an error occurs parsing the
     *                   MetaData, or the DB definition is not defined. 
     **/
    public static int deleteSchemaTables(String[] jdo_files,boolean verbose)
    throws Exception
    {
        // Build up set of classes
        Map  classToMetaDataMap=new HashMap();
        List class_names=null;
        try
        {
            class_names = MetaDataUtils.getOrderedClassNamesFromFileList(
                            jdo_files,classToMetaDataMap);
        }
        catch (Exception e)
        {
            String msg = LOCALISER.msg("SchemaTool.MetaDataParseError");
            JPOXLogger.SCHEMATOOL.error(msg);
            System.err.println(msg);
            
            // Add the exception to the log
            JPOXLogger.SCHEMATOOL.error(e);

            throw new Exception(msg, e);
        }
        if (class_names != null && class_names.size() > 0)
        {
            String[] class_name_array=new String[class_names.size()];
            Iterator class_names_iter=class_names.iterator();
            int i=0;
            while (class_names_iter.hasNext())
            {
                class_name_array[i++] = (String)class_names_iter.next();
            }

            // Set up properties for the data store
            Properties props=null;
            try
            {
                props = getPropertiesForDatastore(verbose);
                if (verbose)
                {
                    System.out.println("");
                }
            }
            catch (Exception e)
            {
                JPOXLogger.SCHEMATOOL.error(e.getMessage());
                System.err.println(e.getMessage());

                throw e;
            }

            // Create a PersistenceManager for this store and delete the tables
            PersistenceManagerFactory pmf = JDOHelper.getPersistenceManagerFactory(props);
            StoreManager store_mgr = StoreManagerFactory.getStoreManager(pmf.getPersistenceManager());
            store_mgr.addClasses(class_name_array); // Add them to mgr first
            store_mgr.removeAllClasses();

            return 0;
        }
        else
        {
            String msg = LOCALISER.msg("SchemaTool.NoClassesError");
            JPOXLogger.SCHEMATOOL.info(msg);
            System.err.println(msg);

            throw new Exception(msg);
        }
    }

    /** Method to handle the validation of a schema's tables.
     * @param jdo_files Array of JDO files defining the classes to be validated
     * @param verbose Verbose output
     * @return  Return code
     * @throws Exception Thrown when either an error occurs parsing the
     *                   MetaData, or the DB definition is not defined. 
     **/
    public static int validateSchemaTables(String[] jdo_files,boolean verbose)
    throws Exception
    {
        // Build up set of classes
        Map  classToMetaDataMap=new HashMap();
        List class_names=null;
        try
        {
            class_names = MetaDataUtils.getOrderedClassNamesFromFileList(
                            jdo_files,classToMetaDataMap);
        }
        catch (Exception e)
        {
            String msg = LOCALISER.msg("SchemaTool.MetaDataParseError", e);
            JPOXLogger.SCHEMATOOL.error(msg);
            System.err.println(msg);

            // Log the exception
            JPOXLogger.SCHEMATOOL.error(e);

            throw new Exception(msg, e);
        }
        if (class_names != null && class_names.size() > 0)
        {
            String[] class_name_array=new String[class_names.size()];
            Iterator class_names_iter=class_names.iterator();
            int i=0;
            while (class_names_iter.hasNext())
            {
                class_name_array[i++] = (String)class_names_iter.next();
            }

            // Set up properties for the data store
            Properties props=null;
            try
            {
                props = getPropertiesForDatastore(verbose);
                addProperty(props, PMFConfiguration.VALIDATE_TABLES_PROPERTY, "true", verbose);
                addProperty(props, PMFConfiguration.VALIDATE_CONSTRAINTS_PROPERTY," true", verbose);
                if (verbose)
                {
                    System.out.println("");
                }
            }
            catch (Exception e)
            {
                JPOXLogger.SCHEMATOOL.error(e.getMessage());
                System.err.println(e.getMessage());

                throw e;
            }

            // Create a PersistenceManager for this store and validate
            PersistenceManagerFactory pmf = JDOHelper.getPersistenceManagerFactory(props);
            StoreManager store_mgr = StoreManagerFactory.getStoreManager(pmf.getPersistenceManager());
            store_mgr.addClasses(class_name_array); // This will do the validation

            return 0;
        }
        else
        {
            String msg = LOCALISER.msg("SchemaTool.NoClassesError");
            JPOXLogger.SCHEMATOOL.error(msg);
            System.err.println(msg);

            throw new Exception(msg);
        }
    }

    /** Method to output the database information.
     * @param verbose Verbose output
     * @return  Return code
     * @throws Exception Thrown when the DB definition is not defined. 
     **/
    public static int outputDBInfo(boolean verbose)
    throws Exception
    {
        // Set up properties for the data store
        Properties props=null;
        try
        {
            props = getPropertiesForDatastore(verbose);
            if (verbose)
            {
                System.out.println("");
            }
        }
        catch (Exception e)
        {
            System.err.println(e.getMessage());
            throw e;
        }

        // Create a PersistenceManager for this store and validate
        PersistenceManagerFactory pmf = JDOHelper.getPersistenceManagerFactory(props);
        StoreManager store_mgr = StoreManagerFactory.getStoreManager(pmf.getPersistenceManager());
        DatabaseAdapter dba = (DatabaseAdapter) store_mgr.getDatastoreAdapter();
        JPOXLogger.SCHEMATOOL.info(dba.toString());
        System.out.println(dba.toString());
        System.out.println();

        String msg = LOCALISER.msg("SchemaTool.DatabaseTypeInfo");
        JPOXLogger.SCHEMATOOL.info(msg);
        System.out.println(msg);
        Iterator iteTypeInfos = dba.iteratorTypeInfo();
        while (iteTypeInfos.hasNext())
        {
            System.out.println(iteTypeInfos.next());
        }
        System.out.println("");

        msg = LOCALISER.msg("SchemaTool.DatabaseKeywords");
        JPOXLogger.SCHEMATOOL.info(msg);
        System.out.println(msg);
        Iterator iteReservedWords = dba.iteratorReservedWords();
        while (iteReservedWords.hasNext())
        {
            System.out.println(iteReservedWords.next());
        }
        System.out.println("");

        return 0;
    }

    /**
     * Utility to set up the properties defining the datastore. Uses the
     * DatabaseProperties object to obtain them from System.properties.
     * @param verbose Verbose output
     * @return The properties definition for the datastore.
     * @throws Exception Thrown when the datastore is not fully specified.
     */
    public static Properties getPropertiesForDatastore(boolean verbose)
    throws Exception
    {
        Properties  props=new Properties();

        if (System.getProperty(PMFConfiguration.JDO_DATASTORE_DRIVERNAME_PROPERTY) == null ||
            System.getProperty(PMFConfiguration.JDO_DATASTORE_URL_PROPERTY) == null)
        {
            throw new Exception(LOCALISER.msg("SchemaTool.DatabaseDriverNotSpecifiedError"));
        }

        // Add the properties necessary
        if (verbose)
        {
            String msg = LOCALISER.msg("SchemaTool.Properties");
            JPOXLogger.SCHEMATOOL.info(msg);
            System.out.println(msg);
        }

        addProperty(props, "javax.jdo.PersistenceManagerFactoryClass", "org.jpox.PersistenceManagerFactoryImpl", verbose);
        addProperty(props, PMFConfiguration.JDO_DATASTORE_DRIVERNAME_PROPERTY,
            System.getProperty(PMFConfiguration.JDO_DATASTORE_DRIVERNAME_PROPERTY), verbose);
        addProperty(props, PMFConfiguration.JDO_DATASTORE_URL_PROPERTY,
            System.getProperty(PMFConfiguration.JDO_DATASTORE_URL_PROPERTY), verbose);
        addProperty(props, PMFConfiguration.JDO_DATASTORE_USERNAME_PROPERTY,
            System.getProperty(PMFConfiguration.JDO_DATASTORE_USERNAME_PROPERTY), verbose);
        addProperty(props, PMFConfiguration.JDO_DATASTORE_PASSWORD_PROPERTY,
            System.getProperty(PMFConfiguration.JDO_DATASTORE_PASSWORD_PROPERTY), false);
        addProperty(props, PMFConfiguration.JDO_MAPPING_PROPERTY,
            System.getProperty(PMFConfiguration.JDO_MAPPING_PROPERTY), verbose);

        return props;
    }

    /**
     * Convenience utility to add a property to the specified properties list.
     * @param props The current properties list to add it to
     * @param name Name of the property
     * @param value Value of the property
     * @param verbose Whether to give verbose logging info
     */
    private static void addProperty(Properties props, String name, String value, boolean verbose)
    {
        if (name == null || value == null)
        {
            return;
        }

        props.setProperty(name, value);

        if (verbose)
        {
            String msg = LOCALISER.msg("SchemaTool.Property", name, value);
            JPOXLogger.SCHEMATOOL.info(msg);
            System.out.println(msg);
        }
    }

    /**
     * Entry method when invoked from the command line.
     *
     * @param args
     *   List of options for processing by the available methods in this class.
     *   -create requests that DB tables be created for classes specified by the
     *   passed JDO files.
     * @throws Exception
     */
    public static void main(String[] args) throws Exception
    {
        boolean create_flag=true;
        boolean delete_flag=false;
        boolean validate_flag=false;
        boolean dbinfo_flag=false;
        boolean verbose=false;
        String ddlFileName=null;

        ArrayList files=new ArrayList();
        for (int i=0;i<args.length;i++)
        {
            if (args[i].equals("-create"))
            {
                dbinfo_flag = false;
                create_flag = true;
                delete_flag = false;
                validate_flag = false;
            }
            else if (args[i].equals("-delete"))
            {
                dbinfo_flag = false;
                delete_flag = true;
                create_flag = false;
                validate_flag = false;
            }
            else if (args[i].equals("-validate"))
            {
                dbinfo_flag = false;
                validate_flag = true;
                create_flag = false;
                delete_flag = false;
            }
            else if (args[i].equals("-dbinfo"))
            {
                dbinfo_flag = true;
                validate_flag = false;
                create_flag = false;
                delete_flag = false;
            }            
            else if (args[i].equals("-v"))
            {
                verbose = true;
            }
            else if (args[i].equals("-help"))
            {
                System.out.println(LOCALISER.msg("SchemaTool.HelpTitle"));
                System.out.println(LOCALISER.msg("SchemaTool.HelpCommand"));
                System.out.println(LOCALISER.msg("SchemaTool.HelpOptions"));
                System.out.println(LOCALISER.msg("SchemaTool.HelpOption.Create"));
                System.out.println(LOCALISER.msg("SchemaTool.HelpOption.Delete"));
                System.out.println(LOCALISER.msg("SchemaTool.HelpOption.Validate"));
                System.out.println(LOCALISER.msg("SchemaTool.HelpOption.DumpDdl"));
                System.out.println(LOCALISER.msg("SchemaTool.HelpOption.Help"));
                System.out.println(LOCALISER.msg("SchemaTool.HelpJDOFiles"));
            }
            else if (args[i].equals("-dumpDdl"))
            {
                i++;
                ddlFileName = args[i];
            }
            else
            {
                files.add(args[i]);
            }
        }

        // Mode of operation
        if (create_flag)
        {
            System.out.println(LOCALISER.msg("SchemaTool.Title.SchemaCreation", PersistenceManagerFactoryImpl.getVersionNumber()));
        }
        else if (delete_flag)
        {
            System.out.println(LOCALISER.msg("SchemaTool.Title.SchemaDeletion",PersistenceManagerFactoryImpl.getVersionNumber()));
        }
        else if (validate_flag)
        {
            System.out.println(LOCALISER.msg("SchemaTool.Title.SchemaValidation", PersistenceManagerFactoryImpl.getVersionNumber()));
        }
        else if (dbinfo_flag)
        {
            System.out.println(LOCALISER.msg("SchemaTool.Title.DBInfo", PersistenceManagerFactoryImpl.getVersionNumber()));
        }
        System.out.println();

        if (verbose)
        {
            // Classpath
            String msg = LOCALISER.msg("SchemaTool.Classpath");
            JPOXLogger.ENHANCER.info(msg);
            System.out.println(msg);
            String[] cp = Pattern.compile(File.pathSeparator).split(System.getProperty("java.class.path"));
            for (int i = 0; i < cp.length; ++i)
            {
                String entry = LOCALISER.msg("SchemaTool.Classpath.Entry", cp[i]);
                JPOXLogger.ENHANCER.info(entry);
                System.out.println(entry);
            }
            System.out.println();

            System.out.println(LOCALISER.msg("SchemaTool.InputJdoFiles"));
            Iterator files_iter=files.iterator();
            while (files_iter.hasNext())
            {
                String jdo_file=(String)files_iter.next();
                System.out.println(">> " + jdo_file);
            }
            System.out.println();
        }

        if (ddlFileName != null)
        {
            if (verbose)
            {
                System.out.println(LOCALISER.msg("SchemaTool.OutputDDL", ddlFileName));
                System.out.println();
            }
            File ddlFile = new File(ddlFileName);
            if (ddlFile.exists())
            {
                ddlFile.delete();
            }
            ddlFile.createNewFile();
            ddlFileWriter = new FileWriter(ddlFile);
            AbstractTable.setDdlDumpWriter(ddlFileWriter);
        }
        
        // Perform the required action
        int code=0;
        String[] jdo_files=null;
        try
        {
            if (files != null && files.size() > 0)
            {
                jdo_files = (String[])files.toArray(new String[files.size()]);
                if (create_flag)
                {
                    code = createSchemaTables(jdo_files,verbose);
                }
                else if (delete_flag)
                {
                    code = deleteSchemaTables(jdo_files,verbose);
                }
                else if (validate_flag)
                {
                    code = validateSchemaTables(jdo_files,verbose);
                }
                else if (dbinfo_flag)
                {
                    code = outputDBInfo(verbose);
                }
                if (code == 0)
                {
                    String msg = LOCALISER.msg("SchemaTool.Success");
                    JPOXLogger.SCHEMATOOL.info(msg);
                    System.out.println(msg);
                }
            }
            else
            {
                String msg = LOCALISER.msg("SchemaTool.NoMetaDataFilesError");
                JPOXLogger.SCHEMATOOL.error(msg);
                System.err.println(msg);
                code = 2;
            }
        }
        catch (Exception e)
        {
            String msg = LOCALISER.msg("SchemaTool.Error", e.getMessage());
            JPOXLogger.SCHEMATOOL.error(msg);
            System.out.println(msg);

            // Log the exception
            JPOXLogger.SCHEMATOOL.error(e);

            code = 3;
        }

        if (ddlFileWriter != null)
        {
            ddlFileWriter.close();
        }

        System.exit(code);
    }
}