/**********************************************************************
Copyright (c) 2003 Andy Jefferson and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
2004 Erik Bengtson - added interfaces methods
    ...
**********************************************************************/
package org.jpox;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import org.jpox.sco.SCOList;
import org.jpox.sco.SCOMap;
import org.jpox.util.JPOXLogger;
import org.jpox.util.JavaUtils;
import org.jpox.util.Localiser;

/**
 * Manager of Type information in JPOX. Acts as a registry of type information
 * both first-class (FCO) and second-class (SCO). Provides global settings 
 * for types in JPOX. 
 * <P>MappingManager provides similar type of functionality, but more on a type by type basis.
 * TypeManager should ultimately be reduced in complexity so that it simply stores the types that
 * JPOX handles and not much more, and MappingManager stores the more detailed mapping info.
 *
 * @version $Revision: 1.7 $ 
 */
public class TypeManager
{
    private static final Localiser LOCALISER=Localiser.getInstance("org.jpox.Localisation");

    /** The types, keyed by the class name. */
    HashMap types=new HashMap();

    /** Singleton instance */
    static TypeManager typeMgr=new TypeManager();

    /**
     * Static initialisation block to load up support for type mappings.
     * This adds entries for the majority of types that JDO requires, or allows
     * as optional. It utilises the type mappings stored under store which are,
     * strictly speaking, for RDBMS stores. If we ever support other store
     * types we will have to change this so we have types as store-based.
     **/
    static
    {
        // "primitives"
        typeMgr.addType(boolean.class,org.jpox.store.mapping.BooleanMapping.class,null);
        typeMgr.addType(byte.class,org.jpox.store.mapping.ByteMapping.class,null);
        typeMgr.addType(char.class,org.jpox.store.mapping.CharacterMapping.class,null);
        typeMgr.addType(short.class,org.jpox.store.mapping.ShortMapping.class,null);
        typeMgr.addType(int.class,org.jpox.store.mapping.IntegerMapping.class,null);
        typeMgr.addType(long.class,org.jpox.store.mapping.LongMapping.class,null);
        typeMgr.addType(float.class,org.jpox.store.mapping.FloatMapping.class,null);
        typeMgr.addType(double.class,org.jpox.store.mapping.DoubleMapping.class,null);

        // "simple" objects
        typeMgr.addType(Boolean.class,org.jpox.store.mapping.BooleanMapping.class,null);
        typeMgr.addType(Byte.class,org.jpox.store.mapping.ByteMapping.class,null);
        typeMgr.addType(Character.class,org.jpox.store.mapping.CharacterMapping.class,null);
        typeMgr.addType(Short.class,org.jpox.store.mapping.ShortMapping.class,null);
        typeMgr.addType(Integer.class,org.jpox.store.mapping.IntegerMapping.class,null);
        typeMgr.addType(Long.class,org.jpox.store.mapping.LongMapping.class,null);
        typeMgr.addType(Float.class,org.jpox.store.mapping.FloatMapping.class,null);
        typeMgr.addType(Double.class,org.jpox.store.mapping.DoubleMapping.class,null);
        typeMgr.addType(String.class,org.jpox.store.mapping.StringMapping.class,null);
        typeMgr.addType(java.math.BigDecimal.class,org.jpox.store.mapping.BigDecimalMapping.class,null);
        typeMgr.addType(java.math.BigInteger.class,org.jpox.store.mapping.BigIntegerMapping.class,null);
        typeMgr.addType(java.util.Date.class,org.jpox.store.mapping.DateMapping.class,org.jpox.sco.Date.class);
        typeMgr.addType(java.util.Locale.class,org.jpox.store.mapping.LocaleMapping.class,null);
        typeMgr.addType(java.util.TimeZone.class,org.jpox.store.mapping.TimeZoneMapping.class,null);
        typeMgr.addType(java.sql.Date.class,org.jpox.store.mapping.SqlDateMapping.class,org.jpox.sco.SqlDate.class);
        typeMgr.addType(java.sql.Time.class,org.jpox.store.mapping.SqlTimeMapping.class,org.jpox.sco.SqlTime.class);
        typeMgr.addType(java.sql.Timestamp.class,org.jpox.store.mapping.SqlTimestampMapping.class,org.jpox.sco.SqlTimestamp.class);

        typeMgr.addType(Number.class,null,null);
        typeMgr.addType(java.lang.Object.class,org.jpox.store.mapping.ObjectMapping.class,null);

        // array types
        typeMgr.addType(boolean[].class,org.jpox.store.mapping.BooleanArrayMapping.class,null);
        typeMgr.addType(byte[].class,org.jpox.store.mapping.ByteArrayMapping.class,null);
        typeMgr.addType(char[].class,org.jpox.store.mapping.CharArrayMapping.class,null);
        typeMgr.addType(short[].class,org.jpox.store.mapping.ShortArrayMapping.class,null);
        typeMgr.addType(int[].class,org.jpox.store.mapping.IntArrayMapping.class,null);
        typeMgr.addType(long[].class,org.jpox.store.mapping.LongArrayMapping.class,null);
        typeMgr.addType(float[].class,org.jpox.store.mapping.FloatArrayMapping.class,null);
        typeMgr.addType(double[].class,org.jpox.store.mapping.DoubleArrayMapping.class,null);

        // container objects
        typeMgr.addType(java.util.ArrayList.class,org.jpox.store.mapping.ArrayListMapping.class,org.jpox.sco.ArrayList.class);
        typeMgr.addType(java.util.Collection.class,org.jpox.store.mapping.CollectionMapping.class,org.jpox.sco.Collection.class);
        typeMgr.addType(java.util.HashMap.class,org.jpox.store.mapping.HashMapMapping.class,org.jpox.sco.HashMap.class);
        typeMgr.addType(java.util.HashSet.class,org.jpox.store.mapping.HashSetMapping.class,org.jpox.sco.HashSet.class);
        typeMgr.addType(java.util.Hashtable.class,org.jpox.store.mapping.HashtableMapping.class,org.jpox.sco.Hashtable.class);
        typeMgr.addType(java.util.LinkedList.class,org.jpox.store.mapping.LinkedListMapping.class,org.jpox.sco.LinkedList.class);
        typeMgr.addType(java.util.List.class,org.jpox.store.mapping.ListMapping.class,org.jpox.sco.List.class);
        typeMgr.addType(java.util.Map.class,org.jpox.store.mapping.MapMapping.class,org.jpox.sco.Map.class);
        typeMgr.addType(java.util.Set.class,org.jpox.store.mapping.SetMapping.class,org.jpox.sco.Set.class);
        typeMgr.addType(java.util.Stack.class,org.jpox.store.mapping.StackMapping.class,org.jpox.sco.Stack.class);
        typeMgr.addType(java.util.TreeSet.class,org.jpox.store.mapping.TreeSetMapping.class,org.jpox.sco.TreeSet.class);
        typeMgr.addType(java.util.Vector.class,org.jpox.store.mapping.VectorMapping.class,org.jpox.sco.Vector.class);

        // JPOX objects
        typeMgr.addType(org.jpox.store.OID.class,org.jpox.store.mapping.OIDMapping.class,null);

        // JDK1.4 additions
        if (JavaUtils.isJRE1_4OrAbove())
        {
            typeMgr.addType(java.util.Currency.class,org.jpox.store.mapping.CurrencyMapping.class,null);
            typeMgr.addType(java.util.LinkedHashSet.class, org.jpox.store.mapping.LinkedHashSetMapping.class, org.jpox.sco.LinkedHashSet.class);
            typeMgr.addType(java.util.LinkedHashMap.class, org.jpox.store.mapping.LinkedHashMapMapping.class, org.jpox.sco.LinkedHashMap.class);
        }

        // JDK1.5 additions
    }

    /**
     * Accessor for the TypeManager for JPOX (singleton)
     * @return The manager of type information
     */
    public static TypeManager getTypeManager()
    {
        return typeMgr;
    }

    /**
     * Protected constructor to prevent outside instantiation
     */
    protected TypeManager()
    {
    }

    /**
     * Method to add support for a class (to some degree). The type can be
     * supported either first-class, or second-class, or both.
     * @param c The class
     * @param sco_mapping_type The second class mapping type
     * @param sco_wrapper_type The second class wrapper type
     */
    public void addType(Class c,Class sco_mapping_type,Class sco_wrapper_type)
    {
        if (c == null)
        {
            return;
        }

        JPOXType type=new JPOXType(c,sco_mapping_type,sco_wrapper_type);
        types.put(c.getName(),type);
        if (JPOXLogger.GENERAL.isDebugEnabled())
        {
            JPOXLogger.GENERAL.debug(LOCALISER.msg("Type.AddingSupport",type));
        }
    }

    /**
     * Accessor for whether an array type is supported.
     * @param type Array type
     * @return Whether it is a supported array type.
     */
    public boolean isSupportedArrayType(Class type)
    {
        if (type == null)
        {
            return false;
        }

        if (!type.isArray())
        {
            throw new IllegalArgumentException("Type " + type.getName() + " is not an array type");
        }

        // Check our types and see if we have this one
        Set keys=types.keySet();
        Iterator iter=keys.iterator();
        while (iter.hasNext())
        {
            String className=(String)iter.next();
            if (className.equals(type.getName()))
            {
                return true;
            }
        }

        return false;
    }

    /**
     * Accessor for whether a class is a "Simple" interface.
     * This distinguishes it from implicitly supported interfaces (Collection, Set, etc).
     * @param c The class
     * @return Whether it is a simple interface
     **/
    public boolean isSimpleInterface(Class c)
    {
        boolean is_simple_interface=true;

        if (!c.isInterface())
        {
            is_simple_interface = false;
        }
        else if (isSupportedType(c.getName()))
        {
            is_simple_interface = false;
        }

        return is_simple_interface;
    }
    
    /**
     * Accessor for whether a class is supported.
     * @param class_name The class name
     * @return Whether the class is supported (to some degree)
     */
    public boolean isSupportedType(String class_name)
    {
        if (class_name == null)
        {
            return false;
        }

        JPOXType type=getType(class_name, null);
        if (type == null)
        {
            return false;
        }
        else
        {
            return true;
        }
    }

    /**
     * Accessor for whether a class is supported.
     * @param clazz The class
     * @return Whether the class is supported (to some degree)
     */
    public boolean isSupportedType(Class clazz)
    {
        if (clazz == null)
        {
            return false;
        }

        JPOXType type=getType(clazz);
        if (type == null)
        {
            return false;
        }
        else
        {
            return true;
        }
    }
    
    /**
     * Accessor for whether a class is supported.
     * @param class_name The class name
     * @return Whether the class is supported (to some degree)
     */
    public boolean isSupportedType(String class_name, ClassLoaderResolver clr)
    {
        if (class_name == null)
        {
            return false;
        }

        JPOXType type=getType(class_name, clr);
        if (type == null)
        {
            return false;
        }
        else
        {
            return true;
        }
    }
    
    /**
     * Accessor for the Supported Types.
     * @return Set of supported types (containing the Class objects).
     **/
    public Set getSupportedTypes()
    {
        Set keys=types.keySet();

        return new HashSet(keys);
    }

    /**
     * Accessor for whether a class is a supported second class mutable type
     * @param class_name The class name
     * @return Whether it is a second class mutable type
     */
    public boolean isSecondClassMutableType(String class_name)
    {
        return (getSecondClassWrapper(class_name) != null);
    }

    /**
     * Accessor for whether the type is a List type.
     * @param c The class
     * @return if c is a List type
     */
    public boolean isSCOList(Class c)
    {
        if (c == null)
        {
            return false;
        }
        if (isSimpleInterface(c))
        {
        	return false;
        }

        JPOXType type=getType(c.getName(), null);
        if (type != null)
        {
            // Check for Collection interfaces
            Class[] interfaces=type.getSCOMutableWrapperType().getInterfaces();
            if (interfaces != null && interfaces.length > 0)
            {
                for (int i=0;i<interfaces.length;i++)
                {
                    if (SCOList.class.getName().equals(interfaces[i].getName()))
                    {
                        return true;
                    }
                }
            }            
        }
        return false;
    }

    /**
     * Accessor for whether the type is a Map type.
     * @param c The class
     * @return true if c is a Map type
     */
    public boolean isSCOMap(Class c)
    {
        if (c == null)
        {
            return false;
        }
        if (isSimpleInterface(c))
        {
        	return false;
        }

        JPOXType type=getType(c.getName(), null);
        if (type != null)
        {
            // Check for Collection interfaces
            Class[] interfaces=type.getSCOMutableWrapperType().getInterfaces();
            if (interfaces != null && interfaces.length > 0)
            {
                for (int i=0;i<interfaces.length;i++)
                {
                    if (SCOMap.class.getName().equals(interfaces[i].getName()))
                    {
                        return true;
                    }
                }
            }            
        }
        return false;
    }
    
    /**
     * Accessor for the Second Class Wrapper class for the supplied class.
     * A type will have a SCO wrapper if it is SCO supported and is mutable.
     * @param class_name The class name
     * @return The second class wrapper
     */
    public Class getSecondClassWrapper(String class_name)
    {
        if (class_name == null)
        {
            return null;
        }

        JPOXType type=getType(class_name, null);
        if (type == null)
        {
            return null;
        }
        else
        {
            return type.getSCOMutableWrapperType();
        }
    }

    /**
     * Utility class to retrieve a supported type
     * @param class_name The class name
     * @param clr The ClassLoaderResolver
     * @return The internal type information for the class
     */
    protected JPOXType getType(String class_name, ClassLoaderResolver clr)
    {
        if (class_name == null)
        {
            return null;
        }

        Object obj=types.get(class_name);
        if (obj != null)
        {
            return (JPOXType)obj;
        }
        else
        {
            if( clr != null )
            {
                /*
                 * This enable mapping to subclass of our supported types.
                 * For instance you have a persistent class with a field of a class extending the java.util.ArrayList.
                 * JPOX is able to persisting this field using the same semantics of
                 * java.util.ArrayList or java.util.List 
                 */                
                Iterator it = types.values().iterator();
                while( it.hasNext() )
                {
                    JPOXType type = (JPOXType)it.next();
                    if( !type.c.getName().equals(Object.class.getName()) )
                    {
	                    if( type.c.isAssignableFrom(clr.classForName(class_name)) )
	                    {
	                        return type;
	                    }
                    }
                }
            }
            return null;
        }
    }
    
    /**
     * Utility class to retrieve a supported type
     * @param c The class
     * @return The internal type information for the class
     */
    protected JPOXType getType(Class c)
    {
        if (c == null)
        {
            return null;
        }

        Object obj=types.get(c.getName());
        if (obj != null)
        {
            return (JPOXType)obj;
        }
        else
        {
            /*
             * This enable mapping to subclass of our supported types.
             * For instance you have a persistent class with a field of a class extending the java.util.ArrayList.
             * JPOX is able to persisting this field using the same semantics of
             * java.util.ArrayList or java.util.List 
             */
            Iterator it = types.values().iterator();
            while( it.hasNext() )
            {
                JPOXType type = (JPOXType)it.next();
                if( !type.c.getName().equals(Object.class.getName()) )
                {
                    if( type.c.isAssignableFrom(c) )
                    {
                        return type;
                    }
                }
            }
            return null;
        }
    }    

    /**
     * Inner class defining a supported type. The type is supported to some
     * degree by JPOX, maybe FCO, SCO, or DFG.
     */
    static class JPOXType
    {
        Class c;
        Class scoMappingType;
        Class scoMutableWrapperType;

        public JPOXType(Class c,
                        Class sco_mapping_type,
                        Class sco_wrapper_type)
        {
            this.c = c;
            this.scoMappingType = sco_mapping_type;
            this.scoMutableWrapperType = sco_wrapper_type;
        }

        public Class getSCOMappingType()
        {
            return scoMappingType;
        }

        public Class getSCOMutableWrapperType()
        {
            return scoMutableWrapperType;
        }

        public String toString()
        {
            StringBuffer str=new StringBuffer(c.getName() + " [");
            if (scoMappingType != null)
            {
                str.append(" SCO : mapping=" + scoMappingType);
            }
            if (scoMutableWrapperType != null)
            {
                str.append(" SCO : mutable wrapper=" + scoMutableWrapperType);
            }
            str.append("]");
            return str.toString();
        }
    }
}