/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved.
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution.

Contributors:
    ...
**********************************************************************/
package org.jpox.cache;

import java.lang.ref.Reference;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;
import java.util.Map;
import java.util.Properties;

import javax.jdo.spi.PersistenceCapable;

import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;
import org.jpox.util.WeakValueMap;

/**
 * Default implementation of a Level 2 cache for JPOX.
 * Operates with 2 maps internally. One stores all pinned objects that have been
 * selected to be retained by user's application. The other stores all other objects.
 * This second map is the default location where objects are placed when being added here.
 * The second (unpinned) map stores weak references meaning that they can get garbage
 * collected as necessary by the JVM.
 * <P>
 * All mutating methods, and the get method have been synchronized to prevent conflicts.
 *
 * @version $Revision: 1.5 $
 */
public class DefaultLevel2Cache implements Level2Cache
{
    /**
     * Localised messages
     */
    private static Localiser LOCALISER = Localiser.getInstance("org.jpox.cache.Localisation");

    /**
     * Pinned instances
     */
    private Map pinnedCache = new HashMap();

    /**
     * Unpinned instances.
     */
    private WeakValueMap unpinnedCache = new WeakValueMap();

    /**
     * Constructor.
     * @param props Any properties to control the cache
     */
    public DefaultLevel2Cache(Properties props)
    {
        // Do nothing with the properties
    }

    /**
     * Method to evict an object from the cache.
     * @param pc The PersistenceCapable to evict
     */
    public synchronized void evict(Object pc)
    {
        if (pc == null)
        {
            return;
        }

        unpinnedCache.remove(((PersistenceCapable)pc).jdoGetObjectId());
        pinnedCache.remove(((PersistenceCapable)pc).jdoGetObjectId());
    }

    /**
     * Method to evict all objects of the given types from the cache.
     * @param pcClass The class to evict
     * @param subclasses Whether to also evict subclasses
     */
    public synchronized void evictAll(Class pcClass, boolean subclasses)
    {
        if (pcClass == null)
        {
            return;
        }

        Collection pinnedObjects = pinnedCache.values();
        Iterator pinnedIter = pinnedObjects.iterator();
        while (pinnedIter.hasNext())
        {
            Object obj = pinnedIter.next();
            if ((subclasses && pcClass.isInstance(obj)) ||
                pcClass.getName().equals(obj.getClass().getName()))
            {
                evict(obj);
            }
        }
        
        Collection unpinnedObjects = unpinnedCache.values();
        Iterator unpinnedIter = unpinnedObjects.iterator();
        while (unpinnedIter.hasNext())
        {
            Object obj = unpinnedIter.next();
            if ((subclasses && pcClass.isInstance(obj)) ||
                pcClass.getName().equals(obj.getClass().getName()))
            {
                evict(obj);
            }
        }
    }

    /**
     * Method to evict all of the specified objects
     * @param pcs The objects to evict
     */
    public synchronized void evictAll(Collection pcs)
    {
        if (pcs == null)
        {
            return;
        }

        Iterator iter = pcs.iterator();
        while (iter.hasNext())
        {
            evict(iter.next());
        }
    }

    /**
     * Method to evict all of the specified objects.
     * @param pcs The objects to evict
     */
    public synchronized void evictAll(Object[] pcs)
    {
        if (pcs == null)
        {
            return;
        }

        for (int i=0;i<pcs.length;i++)
        {
            evict(pcs[i]);
        }
    }

    /**
     * Method to pin an object to the cache.
     * @param oid The id of the object to pin
     */
    public synchronized void pin(Object oid)
    {
        if (oid == null)
        {
            return;
        }

        Object pc = unpinnedCache.get(oid);
        if (pc != null)
        {
            pinnedCache.put(oid, pc);
            unpinnedCache.remove(oid);
        }
    }

    /**
     * Method to pin all objects of the given types.
     * @param cls The class
     * @param subs Whether to include subclasses
     */
    public synchronized void pinAll(Class cls, boolean subs)
    {
        if (cls == null)
        {
            return;
        }

        Collection unpinnedObjects = unpinnedCache.values();
        Iterator unpinnedIter = unpinnedObjects.iterator();
        while (unpinnedIter.hasNext())
        {
            Object obj = unpinnedIter.next();
            if ((subs && cls.isInstance(obj)) ||
                cls.getName().equals(obj.getClass().getName()))
            {
                pin(obj);
            }
        }
    }

    /**
     * Method to pin all of the supplied objects
     * @param oids The Object ids to pin
     */
    public synchronized void pinAll(Collection oids)
    {
        if (oids == null)
        {
            return;
        }

        Iterator iter = oids.iterator();
        while (iter.hasNext())
        {
            pin(iter.next());
        }
    }

    /**
     * Method to pin all of the supplied objects
     * @param oids The object ids to pin
     */
    public synchronized void pinAll(Object[] oids)
    {
        if (oids == null)
        {
            return;
        }

        for (int i=0;i<oids.length;i++)
        {
            pin(oids[i]);
        }
    }

    /**
     * Method to unpin an object
     * @param oid The object id
     */
    public synchronized void unpin(Object oid)
    {
        if (oid == null)
        {
            return;
        }

        Object pc = pinnedCache.get(oid);
        if (pc != null)
        {
            unpinnedCache.put(oid, pc);
            pinnedCache.remove(oid);
        }
    }

    /**
     * Method to unpin all objects of the specified types.
     * @param cls Base class
     * @param subs Whether to include subclasses
     */
    public synchronized void unpinAll(Class cls, boolean subs)
    {
        if (cls == null)
        {
            return;
        }

        Collection pinnedObjects = pinnedCache.values();
        Iterator pinnedIter = pinnedObjects.iterator();
        while (pinnedIter.hasNext())
        {
            Object obj = pinnedIter.next();
            if ((subs && cls.isInstance(obj)) ||
                cls.getName().equals(obj.getClass().getName()))
            {
                unpin(obj);
            }
        }
    }

    /**
     * Method to unpin all of the supplied objects
     * @param oids The object ids to unpin
     */
    public synchronized void unpinAll(Collection oids)
    {
        if (oids == null)
        {
            return;
        }

        Iterator iter = oids.iterator();
        while (iter.hasNext())
        {
            unpin(iter.next());
        }
    }

    /**
     * Method to unpin all of the specified objects
     * @param oids The object ids to unpin
     */
    public synchronized void unpinAll(Object[] oids)
    {
        if (oids == null)
        {
            return;
        }

        for (int i=0;i<oids.length;i++)
        {
            unpin(oids[i]);
        }
    }

    /**
     * Method to clear the cache.
     */
    public void clear()
    {
        pinnedCache.clear();
        unpinnedCache.clear();
    }

    /**
     * Accessor for an object from the cache.
     * The returned object will not have a StateManager connected. This is
     * because data stored in the Level 2 cache is StateManager and
     * PersistenceManager independent.
     * @param oid The Object ID
     * @return The PersistenceCapable object
     */
    public synchronized PersistenceCapable get(Object oid)
    {
        if (oid == null)
        {
            return null;
        }

        PersistenceCapable pc = (PersistenceCapable)pinnedCache.get(oid);
        if (pc != null)
        {
            return pc;
        }
        pc = (PersistenceCapable)unpinnedCache.get(oid);
        return pc;
    }

    /**
     * Accessor for the number of pinned objects in the cache.
     * @return Number of pinned objects
     */
    public int getNumberOfPinnedObjects()
    {
        return pinnedCache.size();
    }

    /**
     * Accessor for the number of unpinned objects in the cache.
     * @return Number of unpinned objects
     */
    public int getNumberOfUnpinnedObjects()
    {
        return unpinnedCache.size();
    }

    /**
     * Accessor for the total number of objects in the L2 cache.
     * @return Number of objects
     */
    public int getSize()
    {
        return getNumberOfPinnedObjects() + getNumberOfUnpinnedObjects();
    }

    /**
     * Method to put an object in the cache. Note that the pc object being
     * passed in must NOT have a StateManager connected. Data stored in
     * the Level 2 cache has to be independent of PersistenceManager and
     * StateManager.
     * @param oid The Object id for this object
     * @param pc The PersistenceCapable object
     * @return The value previously associated with this oid
     */
    public synchronized PersistenceCapable put(Object oid, PersistenceCapable pc)
    {
        if (oid == null || pc == null)
        {
            JPOXLogger.CACHE.warn(LOCALISER.msg("Cache.CantStoreNullObject"));
            return null;
        }

        // Check if the object is disconnected from its StateManager/PM
        if (pc.jdoGetPersistenceManager() != null)
        {
            JPOXLogger.CACHE.error(LOCALISER.msg("Cache.ObjectConnectedError", oid));
            return null;
        }

        // Update the pinned cache if object is already there
        Object obj = null;
        if (pinnedCache.get(oid) != null)
        {
            obj = pinnedCache.put(oid, pc);
            if (obj != null)
            {
                return (PersistenceCapable)obj;
            }
        }
        // Update the unpinned cache otherwise
        else
        {
            obj = unpinnedCache.put(oid, pc);
            if (obj != null)
            {
                // Values in a WeakRefCache are of type Reference
                Reference ref = (Reference)obj;
                return (PersistenceCapable)ref.get();
            }
        }

        return null;
    }

    /**
     * Method to check if an object with the specified id is in the cache
     * @param oid The object ID
     * @return Whether it is present
     */
    public boolean containsOid(Object oid)
    {
        return (pinnedCache.containsKey(oid) || unpinnedCache.containsKey(oid));
    }

    /**
     * Accessor for whether the specified persistence capable exists in the cache.
     * @param pc The persistence capable
     * @return Whether it is present.
     */
    public boolean containsObject(PersistenceCapable pc)
    {
        if (pc == null)
        {
            return false;
        }

        Object oid = pc.jdoGetObjectId();
        return containsOid(oid);
    }

    /**
     * Accessor for whether the cache is empty.
     * @return Whether it is empty.
     */
    public boolean isEmpty()
    {
        return (pinnedCache.isEmpty() && unpinnedCache.isEmpty());
    }
}