/**********************************************************************
Copyright (c) 2003 Erik Bengtson and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
    ...
**********************************************************************/
package org.jpox.driver;

import java.sql.Connection;
import java.sql.Driver;
import java.sql.DriverManager;
import java.sql.DriverPropertyInfo;
import java.sql.SQLException;
import java.util.Properties;

import javax.naming.Context;
import javax.naming.InitialContext;
import javax.naming.NamingException;
import javax.sql.DataSource;

/**
 * This class is an Driver Adapter to be used by jpox when dealing
 * with datasource connections
 *
 * @version $Revision: 1.1 $
 */
public class JPOXDriver implements Driver
{
    static
    {
        // Register the JPOXDriver with DriverManager
        try
        {
            JPOXDriver driver = new JPOXDriver();
            DriverManager.registerDriver(driver);
        }
        catch (SQLException ex)
        {
            throw new RuntimeException("JPOX JDBC driver can't be registered: " + ex);
        }
    }

    private static String scheme = "jpox:";
    private static final int MAJOR_VERSION = 1;
    private static final int MINOR_VERSION = 1;

    private Context ctx = null;

    /**
     * @param url
     * @see java.sql.Driver#acceptsURL(String)
     * @return boolean
     */
    public boolean acceptsURL(String url)
    {
        if (url == null)
        {
            return false;
        }
        return url.startsWith(scheme);
    }

    /**
     * This method create a connection
     * @param url The URL to connect to
     * @param info The properties for the connection
     * @see java.sql.Driver#connect(String, Properties)
     * @return The Connection
     * @throws SQLException
     */
    public Connection connect(String url, Properties info) throws SQLException
    {
        if (!acceptsURL(url))
        {
            return null;
        }

        String jndiName = url.substring(scheme.length());
        try
        {
            // locate data source
            DataSource dataSource = (DataSource) getContext().lookup(jndiName);

            if (dataSource == null)
            {
                throw new NamingException(jndiName + " not found in this context");
            }

            String user = (String) info.get("user");
            String password = (String) info.get("password");

            if (user != null && password != null)
            {
                return dataSource.getConnection(user, password);
            }

            return dataSource.getConnection();
        }
        catch (NamingException ex)
        {
            throw new SQLException("Can't get datasource for: " + jndiName + ". " + ex);
        }
    }

    /**
     * @see java.sql.Driver#getMajorVersion()
     * @return The Major version
     */
    public int getMajorVersion()
    {
        return MAJOR_VERSION;
    }

    /**
     * @see java.sql.Driver#getMinorVersion()
     * @return The Minor version
     */
    public int getMinorVersion()
    {
        return MINOR_VERSION;
    }

    /**
     * @param url The URL
     * @param info The properties
     * @see java.sql.Driver#getPropertyInfo(String, Properties)
     * @return DriverPropertyInfo[]
     */
    public DriverPropertyInfo[] getPropertyInfo(String url, Properties info)
    {
        return new DriverPropertyInfo[] { new DriverPropertyInfo("scheme", scheme)};
    }

    /**
     * @see java.sql.Driver#jdbcCompliant()
     * @return boolean Whether it is JDBC compliant
     */
    public boolean jdbcCompliant()
    {
        return false;
    }

    /**
     * Accessor for the Context
     * @return The context
     * @throws SQLException Thrown if an error occurs retrieving the context
     **/ 
    private Context getContext() throws SQLException
    {
        if (this.ctx == null)
        {
            try
            {
                this.ctx = new InitialContext();
            }
            catch (NamingException e)
            {
                throw new SQLException("Can't get Initial Context.");
            }
        }
        return this.ctx;
    }

    /**
     * This method permits you to set the context from within you will obtain
     * the data source. Normally, this method will be used when you are running
     * in a non managed context (outside the server)
     * If you don't set the context, the InitialContext() will be used to do
     * the lookup for the data source
     * @param ctx The Context
     */
    public void setContext(Context ctx)
    {
        this.ctx = ctx;
    }
}
