/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0which accompanies this distribution.

Contributors:
2004 Andy Jefferson - added javadocs.
2004 Andy Jefferson - added unique, indexed
2005 Andy Jefferson - changed foreignKey attr to delete-action
    ...
**********************************************************************/
package org.jpox.metadata;

import java.util.ArrayList;
import java.util.List;

/**
 * This element specifies the mapping for the element component of arrays and
 * collections.
 * 
 * If only one column is mapped, and no additional information is needed for the
 * column, then the column attribute can be used. Otherwise, the column
 * element(s) are used.
 * 
 * The serialized attribute specifies that the key values are to be serialized
 * into the named column.
 * 
 * The foreign-key attribute specifies the name of a foreign key to be
 * generated.
 * 
 * @since 1.1
 * @version $Revision: 1.6 $
 */
public abstract class AbstractElementMetaData extends ExtendableMetaData
{
    /**
     * Columns ColumnMetaData
     */
    protected final List columns = new ArrayList();
    
    /**
     * IndexMetaData
     */
    protected IndexMetaData indexMetaData;

    /**
     * The indexing value
     */
    protected IndexedValue indexed=null;

    /**
     * UniqueMetaData.
     */
    protected UniqueMetaData uniqueMetaData;

    /**
     * Whether to add a unique constraint
     */
    protected final boolean uniqueConstraint;

    /**
     * ForeignKeyMetaData
     */
    protected ForeignKeyMetaData foreignKeyMetaData;

    /**
     * EmbeddedMetaData
     */
    protected EmbeddedMetaData embeddedMetaData;

    /**
     * column tag value
     */
    protected String column;

    /**
     * embedded tag value.
     */
    protected Boolean embedded;

    /**
     * serialized tag value.
     */
    protected Boolean serialized;

    // -------------------------------------------------------------------------
    // Fields below here are not represented in the output MetaData. They are
    // for use internally in the operation of the JDO system. The majority are
    // for convenience to save iterating through the fields since the fields
    // are fixed once initialised.

    protected ColumnMetaData columnMetaData[];

    /**
     * Constructor. 
     * @param column Name of column
     * @param embedded Whether it is embedded
     * @param serialized Whether it is serialized
     * @param deleteAction attribute delete-action value
     * @param updateAction attribute update-action value
     * @param indexed The indexing value
     * @param unique Whether to add a unique constraint
     */
    public AbstractElementMetaData(String column,
                                   String embedded,
                                   String serialized,
                                   String deleteAction,
                                   String updateAction,
                                   String indexed,
                                   String unique)
    {
        super(null);    // Ignore parent

        this.column = column;

        if (deleteAction != null || updateAction != null)
        {
            foreignKeyMetaData = new ForeignKeyMetaData(null, deleteAction, updateAction, null, null, null);
        }

        // "embedded"
        if (embedded != null)
        {
            if (embedded.equalsIgnoreCase("true"))
            {
                this.embedded = Boolean.TRUE;
            }
            else if (embedded.equalsIgnoreCase("false"))
            {
                this.embedded = Boolean.FALSE;
            }
        }

        // "serialized"
        if (serialized != null)
        {
            if (serialized.equalsIgnoreCase("true"))
            {
                this.serialized = Boolean.TRUE;
            }
            else if (serialized.equalsIgnoreCase("false"))
            {
                this.serialized = Boolean.FALSE;
            }
        }

        this.indexed = IndexedValue.getIndexedValue(indexed);

        if (unique != null && unique.equalsIgnoreCase("true"))
        {
            uniqueConstraint = true;
        }
        else
        {
            uniqueConstraint = false;
        }
    }

    /**
     * Method to initialise the object, creating any convenience arrays needed.
     * Initialises all sub-objects. 
     */
    public void initialise()
    {
        // Cater for user specifying column name, or columns
        if (columns.size() == 0 && column != null)
        {
            columnMetaData = new ColumnMetaData[1];
            columnMetaData[0] = new ColumnMetaData(this, column);
        }
        else
        {
            columnMetaData = new ColumnMetaData[columns.size()];
        	for (int i=0; i<columnMetaData.length; i++)
        	{
            	columnMetaData[i] = (ColumnMetaData) columns.get(i);
            	columnMetaData[i].initialise();
        	}
        }

        // Interpret the "indexed" value to create our IndexMetaData where it wasn't specified that way
        if (indexMetaData == null && columnMetaData != null && indexed != null && indexed != IndexedValue.FALSE)
        {
            indexMetaData = new IndexMetaData(null, null, (indexed == IndexedValue.UNIQUE) ? "true" : "false");
            for (int i=0;i<columnMetaData.length;i++)
            {
                indexMetaData.addColumn(columnMetaData[i]);
            }
        }
        if (indexMetaData != null)
        {
            indexMetaData.initialise();
        }

        if (uniqueMetaData == null && uniqueConstraint)
        {
            uniqueMetaData = new UniqueMetaData(null, column, null);
            for (int i=0;i<columnMetaData.length;i++)
            {
                uniqueMetaData.addColumn(columnMetaData[i]);
            }
        }
        if (uniqueMetaData != null)
        {
            uniqueMetaData.initialise();
        }

        if (foreignKeyMetaData != null)
        {
            foreignKeyMetaData.initialise();
        }

        if (embeddedMetaData != null)
        {
            embeddedMetaData.initialise();
        }

        setInitialised();
    }

    // -------------------------------- Accessors ------------------------------
 
    /**
     * Accessor for the embedded tag value
     * @return embedded tag value
     */
    public boolean isEmbedded()
    {
        if (embedded == null)
        {
            return false;
        }
        else
        {
            return embedded.booleanValue();
        }
    }
    
    /**
     * Accessor for the serialized tag value
     * @return serialized tag value
     */
    public boolean isSerialized()
    {
        if (serialized == null)
        {
            return false;
        }
        else
        {
            return serialized.booleanValue();
        }
    }
    
    /**
     * Accessor for column
     * @return Returns the column.
     */
    public final String getColumn()
    {
        return column;
    }

    /**
     * Accessor for columnMetaData
     * @return Returns the columnMetaData.
     */
    public final ColumnMetaData[] getColumnMetaData()
    {
        return columnMetaData;
    }

    /**
     * Accessor for embeddedMetaData
     * @return Returns the embeddedMetaData.
     */
    public final EmbeddedMetaData getEmbeddedMetaData()
    {
        return embeddedMetaData;
    }

    /**
     * Accessor for foreignKeyMetaData
     * @return Returns the foreignKeyMetaData.
     */
    public final ForeignKeyMetaData getForeignKeyMetaData()
    {
        return foreignKeyMetaData;
    }

    /**
     * Accessor for indexMetaData
     * @return Returns the indexMetaData.
     */
    public final IndexMetaData getIndexMetaData()
    {
        return indexMetaData;
    }

    /**
     * Accessor for uniqueMetaData
     * @return Returns the uniqueMetaData.
     */
    public final UniqueMetaData getUniqueMetaData()
    {
        return uniqueMetaData;
    }

    // ------------------------------ Mutators ---------------------------------
 
    /**
     * Add a new ColumnMetaData element
     * @param colmd The Column MetaData 
     */
    public void addColumn(ColumnMetaData colmd)
    {
        columns.add(colmd);
    }

    /**
     * Mutator for the Embedded MetaData 
     * @param embeddedMetaData The embeddedMetaData to set.
     */
    public final void setEmbeddedMetaData(EmbeddedMetaData embeddedMetaData)
    {
        this.embeddedMetaData = embeddedMetaData;
    }
 
    /**
     * Mutator for the Foreign Key MetaData 
     * @param foreignKeyMetaData The foreignKeyMetaData to set.
     */
    public final void setForeignKeyMetaData(
            ForeignKeyMetaData foreignKeyMetaData)
    {
        this.foreignKeyMetaData = foreignKeyMetaData;
    }

    /**
     * Mutator for the Index MetaData 
     * @param indexMetaData The indexMetaData to set.
     */
    public final void setIndexMetaData(IndexMetaData indexMetaData)
    {
        this.indexMetaData = indexMetaData;
    }

    /**
     * Mutator for the Unique MetaData 
     * @param uniqueMetaData The uniqueMetaData to set.
     */
    public final void setUniqueMetaData(UniqueMetaData uniqueMetaData)
    {
        this.uniqueMetaData = uniqueMetaData;
    }
}