/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution.

Contributors:
2004 Kikuchi Kousuke - org.jpox.enhancer.conf.JDOConfigArray
    ...
**********************************************************************/
package org.jpox.metadata;

import java.util.List;
import java.util.Set;

import org.jpox.ClassLoaderResolver;

/**
 * Representation of the Meta-Data for an Array.
 * <P>Represents the following MetaData Element 
 * <PRE>
 * &lt;!ELEMENT array (extension)*&gt;
 * &lt;!ATTLIST array embedded-element (true|false) #IMPLIED&gt;
 * &lt;!ATTLIST array dependent-element (true|false) #IMPLIED&gt;
 * </PRE>
 *
 * @since 1.1 
 * @version $Revision: 1.2 $
 */
public class ArrayMetaData extends ContainerMetaData
{
    /**
     * embedded-element value.
     */
    protected Boolean embeddedElement;

    /**
     * Constructor.
     * @param parent The parent Field
     * @param embeddedElement The embedded element
     * TODO dependentElement
     */
    public ArrayMetaData(FieldMetaData parent,
                         String embeddedElement)
    {
        super(parent);

        if (embeddedElement != null)
        {
            this.embeddedElement = new Boolean(embeddedElement);
        }
    }

    /**
     * Method to populate any defaults, and check the validity of the MetaData.
     * @param loader ClassLoader to use in loading any classes 
     */
    public void populate(ClassLoader loader)
    {
        // Check the field type and see if it is an array type
        Class field_type=getFieldMetaData().getType();
        if (!field_type.isArray())
        {
            throw new InvalidMetaDataException(LOCALISER,
                    "MetaData.Array.IsNotArrayError",
                    getFieldName(),getFieldMetaData().getClassName(false));
        }

        // "embedded-element"
        if (embeddedElement == null)
        {
            Class component_type=field_type.getComponentType();
            if (component_type.isInterface())
            {
                embeddedElement = Boolean.TRUE;
            }
            else
            {
                embeddedElement = Boolean.FALSE;
            }
        }

        setPopulated();
    }

    // --------------------------------- Accessors -----------------------------

    /**
     * Accessor for the embedded-element value
     * @return embedded-element value
     */
    public boolean isEmbeddedElement()
    {
        if (embeddedElement == null)
        {
            return false;
        }
        else
        {
            return embeddedElement.booleanValue();
        }
    }

    // --------------------------------- Mutators ------------------------------

    // --------------------------------- Utilities -----------------------------

    /**
     * Accessor for all ClassMetaData referenced by this array.
     * @param orderedCMDs List of ordered ClassMetaData objects (added to).
     * @param referencedCMDs Set of all ClassMetaData objects (added to).
     * @param dba_vendor_id Vendor ID of the DBA. Used for view addition.
     * @param clr the ClassLoaderResolver 
     **/
    void getReferencedClassMetaData(final List orderedCMDs,
                                    final Set referencedCMDs,
                                    final String dba_vendor_id,
                                    final ClassLoaderResolver clr)
    {
        ClassMetaData element_cmd=getMetaDataManager().getMetaDataForClass(getFieldMetaData().getType().getComponentType());
        if (element_cmd != null)
        {
            element_cmd.getReferencedClassMetaData(orderedCMDs,referencedCMDs,dba_vendor_id,clr);
        }
    }

    /**
     * Returns a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        return toString("","");
    }

    /**
     * Returns a string representation of the object.
     * @param prefix The prefix string 
     * @param indent The indent string 
     * @return a string representation of the object.
     */
    public String toString(String prefix,String indent)
    {
        StringBuffer sb = new StringBuffer();
        sb.append(prefix).append("<array \"").append(embeddedElement).append("\">\n");

        // Add extensions
        sb.append(super.toString(prefix + indent,indent));

        sb.append(prefix).append("</array>\n");
        return sb.toString();
    }
}