/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution.

Contributors:
2004 Kikuchi Kousuke - org.jpox.enhancer.conf.JDOConfigClass
2004 Marco Schulze (NightLabs) - changed the behaviour to warn only if
        an inherited class declares an own objectid and it equals the
        one of the superclass.
2004 Andy Jefferson - Added discriminator/inheritance checks
2004 Erik Bengtson - changes for application identity
2004 Andy Jefferson - moved PK class checks out into JDOUtils
    ...
**********************************************************************/
package org.jpox.metadata;

import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.jdo.FetchPlan;
import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOFatalUserException;
import javax.jdo.JDOUserException;

import org.jpox.ClassLoaderResolver;
import org.jpox.TypeManager;
import org.jpox.util.ClassUtils;
import org.jpox.util.JDOUtils;
import org.jpox.util.JPOXLogger;
import org.jpox.util.StringUtils;
import org.jpox.util.MacroString;
import org.jpox.util.ViewUtils;

/**
 * Representation of the MetaData of a class. Has a parent PackageMetaData that
 * can contain the metadata for several classes.
 *
 * <H3>Lifecycle state</H3>
 * This object supports 3 lifecycle states. The first is the raw
 * constructed object which represents pure MetaData (maybe from a MetaData
 * file). The second is a "populated" object which represents MetaData for a
 * Class with the metadata aligned to be appropriate for that Class.
 * The third is "initialised" once the internal arrays are created.
 * This object, once populated, will represent ALL fields in the class
 * (including static, final and transient fields).
 *
 * <H3>Fields</H3>
 * This object keeps a list of FieldMetaData objects for the fields of this
 * class (stored as "fields"). In addition it has an array of FieldMetaData
 * objects representing those that are actually managed by JDO
 * ("managedFields"). This second set does not contain things like static, final
 * or transient fields since JDO doesn't support those yet.
 *
 * <H3>Numbering of fields</H3>
 * Fields of the class are numbered in 2 ways. The first way is the numbering
 * within a class. In a class, the field 'id's will start at 0. If a class is
 * inherited, it will also have a second numbering for its fields - the
 * "absolute" numbering. With "absolute" numbering, the fields start at the
 * first field in the root superclass which has absolute number 0, and they are
 * numbered from there, navigating down the hierarchy. In terms of what is
 * stored in the records, the FieldMetaData stores fieldId as the first
 * method (relative to the class it is in). The "absolute" numbering is 
 * always derived from this and the inheritance hierarchy.
 *
 * <H3>MetaData Element</H3>
 * The MetaData Element represented here is as follows
 * <PRE>
 * &lt;!ELEMENT class (datastore-identity?, implements*, inheritance?, version?, join*,
 *      foreign-key*, index*, unique*, field*, unmapped?, query*, fetch-group*, extension*)&gt;
 * &lt;!ATTLIST class name CDATA #REQUIRED&gt;
 * &lt;!ATTLIST class identity-type (application|datastore|nondurable) #IMPLIED&gt;
 * &lt;!ATTLIST class table CDATA #IMPLIED&gt;
 * &lt;!ATTLIST class persistence-capable-superclass CDATA #IMPLIED&gt;
 * &lt;!ATTLIST class objectid-class CDATA #IMPLIED&gt;
 * &lt;!ATTLIST class requires-extent (true|false) 'true'&gt;
 * &lt;!ATTLIST class detachable (true|false) 'true'&gt;
 * &lt;!ATTLIST class embedded-only (true|false) #IMPLIED&gt;
 * &lt;!ATTLIST class persistence-modifier 
 *      (persistence-capable|persistence-aware|non-persistent) #IMPLIED&gt;
 * </PRE>
 *
 * @since 1.1
 * @version $Revision: 1.32 $
 */
public class ClassMetaData extends ExtendableMetaData
{
    /**
     * List of named queries
     */
    protected List queries = new ArrayList();

    /**
     * List of joins
     */
    protected List joins = new ArrayList();

    /**
     * List of fetch-group
     */
    protected List fetchGroups = new ArrayList();

    /**
     * List of foreign-key
     */
    protected List foreignKeys = new ArrayList();

    /**
     * List of indexes
     */
    protected List indexes = new ArrayList();

    /**
     * List of uniqueConstraints
     */
    protected List uniqueConstraints = new ArrayList();

    /**
     * List of implements
     */
    protected List implementations = new ArrayList();

    /**
     * List of fields. This includes static, final, transient ... all fields.
     * But doesn't include anything for superclasses.
     */
    protected List fields = new ArrayList();

    /**
     * Identity-type tag value.
     */
    protected IdentityType identityType = IdentityType.DATASTORE;

    /**
     * Class name
     */
    protected final String name;

    /**
     * objectid-class tag value.
     **/
    protected String objectidClass;

    /**
     * persistence-capable-superclas tag value.
     */
    protected String persistenceCapableSuperclass;

    /**
     * requires-extent tag value.
     */
    protected boolean requiresExtent = true;

    /**
     * embedded-only tag value.
     */
    protected boolean embeddedOnly = false;

    /**
     * detachable tag value.
     */
    protected boolean detachable = false;

    /**
     * persistence-modifier tag value.
     */
    protected ClassPersistenceModifier persistenceModifier = null;

    /**
     * Table name (O/R mapping)
     */
    protected String table;

    /**
     * Full class name. (e.g org.jpox.MyClass)
     */
    protected final String fullClassName;

    /**
     * Version MetaData
     */
    protected VersionMetaData versionMetaData;

    /**
     * Identity MetaData
     */
    protected IdentityMetaData identityMetaData;

    /**
     * Inheritance MetaData
     */
    protected InheritanceMetaData inheritanceMetaData;

    /**
     * Unmapped MetaData
     */
    protected UnmappedMetaData unmappedMetaData;

    // -------------------------------------------------------------------------
    // Fields below here are not represented in the output MetaData. They are
    // for use internally in the operation of the JDO system. The majority are
    // for convenience to save iterating through the fields since the fields
    // are fixed once initialised.

    /**
     * ImplementsMetaData
     */
    protected ImplementsMetaData[] implementsMetaData;

    /**
     * JoinMetaData
     */
    protected JoinMetaData[] joinMetaData;

    /**
     * IndexMetaData
     */
    protected IndexMetaData[] indexMetaData;

    /**
     * ForeignKeyMetaData
     */
    protected ForeignKeyMetaData[] foreignKeyMetaData;

    /**
     * UniqueMetaData
     */
    protected UniqueMetaData[] uniqueMetaData;

    /**
     * A contained fetch-group element indicates that the named group is to be
     * included in the group being defined. Nested fetch group elements are
     * limited to only the name attribute.
     */
    protected FetchGroupMetaData[] fetchGroupMetaData;    

    /**
     * JDO managed fields. Contains the same FieldMetaData objects as are in
     * "fields".
     */
    protected FieldMetaData[] managedFields;

    /**
     * Field numbers mapped by the name of the field.
     */
    protected Map fieldNumbersByName;

    /**
     * Field numbers of all fields (inc superclasses).
     */
    protected int[] allFieldNumbers;

    /**
     * Field numbers of the primary key fields (inc superclasses).
     */
    protected int[] primaryKeyFieldNumbers;

    /**
     * Field numbers of the non-primary key fields (inc superclasses).
     */
    protected int[] nonPrimaryKeyFieldNumbers;
    
    /**
     * Field numbers of the non-primary key fields (inc superclasses).
     */
    protected boolean[] nonPrimaryKeyFieldFlags;    

    /**
     * Field numbers of the default fetch group fields (inc superclasses).
     */
    protected int[] defaultFetchGroupFieldNumbers;

    /**
     * Flags of the default fetch group state for all fields.
     */
    protected boolean[] defaultFetchGroupFieldFlags;

    /**
     * Field numbers of the second class mutable fields (inc superclasses).
     */
    protected int[] secondClassMutableFieldNumbers;

    /**
     * Flags of the second class mutable state for all fields.
     */
    protected boolean[] secondClassMutableFieldFlags;

    /**
     * No of managed fields in superclasses, that are inherited by this class.
     */
    protected int noOfInheritedManagedFields=0;

    /**
     * Class MetaData for the Persistence-Capable superclass (if any)
     */
    protected ClassMetaData pcSuperclassMetaData=null;

    // ----------------------------- Constructors ------------------------------
 
    /**
     * Constructor.
     * Takes the basic string information found in the MetaData file.
     * @param parent The package to which this class belongs
     * @param name Name of class
     * @param identityType identity-type flag
     * @param objectidClass Primary key class name
     * @param requiresExtent Whether the class requires an extent
     * @param detachable Whether the class can be detached
     * @param embeddedOnly embedded-only tag
     * @param persistenceCapableSuperclass Name of PC superclass
     * @param table RDBMS table to store the class in
     */
    public ClassMetaData(final PackageMetaData parent,
                         final String name,
                         final String identityType,
                         final String objectidClass,
                         final String requiresExtent,
                         final String detachable,
                         final String embeddedOnly,
                         final String modifier,
                         final String persistenceCapableSuperclass,
                         final String table)
    {
        super(parent);

        this.name = name;
        if (StringUtils.isWhitespace(name))
        {
            throw new InvalidMetaDataException(LOCALISER,
                "MetaData.Class.NameNotSpecified.Error",parent.name);
        }

        this.fullClassName = ClassUtils.createFullClassName(parent.name, name);

        // "objectid-class"
        if (!StringUtils.isWhitespace(objectidClass))
        {
            this.objectidClass = ClassUtils.createFullClassName(parent.name, objectidClass);
        }
        else
        {
            this.objectidClass = null;
        }

        // "identity-type"
        this.identityType = IdentityType.getIdentityType(identityType);

        // "requires-extent"
        if (requiresExtent != null && requiresExtent.equalsIgnoreCase("false"))
        {
            this.requiresExtent = false;
        }
        else
        {
            this.requiresExtent = true;
        }

        // "persistence-modifier"
        persistenceModifier = ClassPersistenceModifier.getClassPersistenceModifier(modifier);

        // "detachable"
        if (detachable != null && detachable.equalsIgnoreCase("false"))
        {
            this.detachable = false;
        }
        else
        {
            this.detachable = true;
        }

        // "embedded-only"
        if (embeddedOnly != null && embeddedOnly.equalsIgnoreCase("true"))
        {
            this.embeddedOnly = true;
        }
        else
        {
            this.embeddedOnly = false;
        }

        // "persistence-capable-superclass"
        if (!StringUtils.isWhitespace(persistenceCapableSuperclass))
        {
            this.persistenceCapableSuperclass =
                ClassUtils.createFullClassName(parent.name, persistenceCapableSuperclass);
        }

        // "table"
        this.table = table;
    }
    
    /**
     * Method to merge in the associated metadata from the ORM definition
     * @param ormCmd The ORM Class definition
     */
    public void mergeORMData(ClassMetaData ormCmd)
    {
        if (ormCmd == null)
        {
            return;
        }
        
        if (isInitialised() || isPopulated())
        {
            throw new JDOFatalInternalException(LOCALISER.msg("MetaData.Class.AlreadyPopulatedError",name));
        }
        
        // Update the O/R mapping for this class
        table = ormCmd.getTable();
        inheritanceMetaData = ormCmd.getInheritanceMetaData();
        identityMetaData = ormCmd.getIdentityMetaData();
        versionMetaData = ormCmd.getVersionMetaData();

        // Queries
        queries.clear();
        queries.addAll(ormCmd.queries);

        // Joins
        joins.clear();
        joins.addAll(ormCmd.joins);

        // Indexes
        indexes.clear();
        indexes.addAll(ormCmd.indexes);

        // Foreign-Keys
        foreignKeys.clear();
        foreignKeys.addAll(ormCmd.foreignKeys);
        
        // Unique Constraints
        uniqueConstraints.clear();
        uniqueConstraints.addAll(ormCmd.uniqueConstraints);

        // Add on any fields that weren't defined previously
        for (int i=0;i<ormCmd.getNoOfFields();i++)
        {
            FieldMetaData ormFmd=ormCmd.getField(i);
            if (!hasField(ormFmd.getName()))
            {
                //creates a new fmd with this as parent and add to fields
                FieldMetaData fmd = new FieldMetaData(this,ormFmd.getName());
                fmd.mergeORMData(ormFmd);
                addField(fmd);
            }
        }
        
        // Update the fields
        Iterator iter=fields.iterator();
        while (iter.hasNext())
        {
            FieldMetaData fmd=(FieldMetaData)iter.next();
            fmd.mergeORMData(ormCmd.getField(fmd.getName()));
        }
    }

    /**
     * Method to provide the details of the class being represented by this
     * MetaData. This can be used to firstly provide defaults for attributes
     * that aren't specified in the MetaData, and secondly to report any errors
     * with attributes that have been specifed that are inconsistent with the
     * class being represented.
     * <P>
     * One possible use of this method would be to take a basic ClassMetaData
     * for a class and call this, passing in the users class. This would then
     * add FieldMetaData for all fields in this class providing defaults for
     * all of these.
     *
     * @param loader ClassLoader to use in loading any classes
     */
    public synchronized void populate(ClassLoader loader)
    {
        if (isInitialised() || isPopulated())
        {
            JPOXLogger.METADATA.error(LOCALISER.msg("MetaData.Class.AlreadyPopulatedError",name));
            throw new JDOFatalInternalException(LOCALISER.msg("MetaData.Class.AlreadyPopulatedError",fullClassName));
        }

        // No class loader, so use System
        if (loader == null)
        {
            JPOXLogger.METADATA.warn(LOCALISER.msg("MetaData.Class.UsingSystemClassLoader",fullClassName));
            loader = ClassLoader.getSystemClassLoader();
        }

        if (JPOXLogger.METADATA.isDebugEnabled())
        {
            JPOXLogger.METADATA.debug(LOCALISER.msg("MetaData.Class.PopulatingMetaData",fullClassName));
        }

        // Load the class we are modelling
        Class cls=null;
        try
        {
            cls = loader.loadClass(fullClassName);
        }
        catch (ClassNotFoundException cnfe)
        {
            JPOXLogger.METADATA.error(LOCALISER.msg("MetaData.Class.ClassNotFoundError",name));
            throw new InvalidMetaDataException(LOCALISER,
                "MetaData.Class.ClassNotFoundError",
                fullClassName);
        }

        getMetaDataManager().addORMDataToClass(cls);

        // If a class is an inner class and is non-static it is invalid
        if (ClassUtils.isInnerClass(fullClassName) && !Modifier.isStatic(cls.getModifiers()) &&
            persistenceModifier == ClassPersistenceModifier.PERSISTENCE_CAPABLE)
        {
            throw new InvalidMetaDataException(LOCALISER, "MetaData.Class.NonStaticInnerClassNotPersistable", fullClassName);
        }

        // The persistence-capable-superclass attribute is deprecated for JDO2.0. The
        // attribute will be ignored so metadata files from previous releases can be used.
        // The persistenceCapableSuperclass we use will be retrieved from the class itself.
        String realPcSuperclassName = null;
        Collection superclasses = ClassUtils.getSuperclasses(cls);
        MetaDataManager mgr = getPackageMetaData().getFileMetaData().getMetaDataManager();
        if (mgr == null)
        {
            mgr = new MetaDataManager();
        }
        if (!superclasses.isEmpty())
        {
            Iterator superclassesIter = superclasses.iterator();
            while (superclassesIter.hasNext())
            {
                Class superclass = (Class) superclassesIter.next();
                ClassMetaData superCmd = mgr.getMetaDataForClass(superclass);
                if (superCmd != null && superCmd.getPersistenceModifier() == ClassPersistenceModifier.PERSISTENCE_CAPABLE)
                {
                    realPcSuperclassName = superclass.getName();
                    break;
                }
            }
        }
        if (persistenceCapableSuperclass != null && !persistenceCapableSuperclass.equals(realPcSuperclassName))
        {
            // Valifies the defined persistenceCapableSuperclass
            try
            {
                ClassMetaData superCmd = mgr.getMetaDataForClass(loader.loadClass(persistenceCapableSuperclass));
                if (superCmd == null || superCmd.getPersistenceModifier() != ClassPersistenceModifier.PERSISTENCE_CAPABLE)
                {
                    throw new InvalidMetaDataException(LOCALISER, "MetaData.Class.PersistenceCapableSuperclass.IsNotPCError", name,
                        persistenceCapableSuperclass);
                }
            }
            catch (ClassNotFoundException cnfe)
            {
                throw new InvalidMetaDataException(LOCALISER, "MetaData.Class.PersistenceCapableSuperclass.ClassNotFoundError", name,
                    persistenceCapableSuperclass);
            }
            if (realPcSuperclassName != null)
            {
                throw new InvalidMetaDataException(LOCALISER, "MetaData.Class.PersistenceCapableSuperclass.InvalidDefinition", name, realPcSuperclassName,
                    persistenceCapableSuperclass);
            }
            else
            {
                // The defined persistenceCapableSuperclass could NOT have been a superclass
                // of this class, otherwise it would have been equal to the realPcSuperclassName.
                throw new InvalidMetaDataException(LOCALISER, "MetaData.Class.PersistenceCapableSuperclass.IsNotSuperclassError", name,
                    persistenceCapableSuperclass);
            }
        }
        if (persistenceCapableSuperclass == null && realPcSuperclassName != null)
        {
            persistenceCapableSuperclass = realPcSuperclassName;
            JPOXLogger.METADATA.info(LOCALISER.msg("MetaData.Class.PersistenceCapableSuperclass.Set", name, persistenceCapableSuperclass));
        }

        // "persistence-capable-superclass"
        // Check that the class can be loaded, and is a true superclass 
        if (persistenceCapableSuperclass != null)
        {
            Class pcsc=null;
            try
            {
                // Load using same class loader as this class
                pcsc = loader.loadClass(persistenceCapableSuperclass);
            }
            catch (ClassNotFoundException cnfe)
            {
                throw new InvalidMetaDataException(LOCALISER,
                    "MetaData.Class.PersistenceCapableSuperclass.NotFoundError",
                    name,persistenceCapableSuperclass);
            }

            if (persistenceCapableSuperclass.equals(fullClassName) ||
                !(pcsc.isAssignableFrom(cls)))
            {
                throw new InvalidMetaDataException(LOCALISER,
                    "MetaData.Class.PersistenceCapableSuperclass.IsNotSuperclassError",
                    name,persistenceCapableSuperclass);
            }

            // Retrieve the Meta-Data for the superclass
            if (mgr != null)
            {
                // JPOX normal operation will have a MetaDataManager and so we
                // can ask that for the MetaData of the superclass.
                pcSuperclassMetaData = mgr.getMetaDataForClass(pcsc);
                if (pcSuperclassMetaData == null)
                {
                    // Superclass isn't persistence capable since no MetaData
                    // could be found.
                    throw new InvalidMetaDataException(LOCALISER,
                        "MetaData.Class.PersistenceCapableSuperclass.IsNotPCError",
                        name,persistenceCapableSuperclass);
                }
            }
            else
            {
                // The enhancer doesn't need MetaDataManager so we just
                // navigate up to the FileMetaData and find it.
                // NOTE : assumes that the class is specified in the same file 
                String superclass_pkg_name=persistenceCapableSuperclass.substring(0,persistenceCapableSuperclass.lastIndexOf('.'));
                PackageMetaData pmd=getPackageMetaData().getFileMetaData().getPackage(superclass_pkg_name);
                if (pmd != null)
                {
                    String superclass_class_name=persistenceCapableSuperclass.substring(persistenceCapableSuperclass.lastIndexOf('.')+1);
                    pcSuperclassMetaData = pmd.getClass(superclass_class_name);
                }
            }
            if (pcSuperclassMetaData == null)
            {
                mgr = new MetaDataManager();
                pcSuperclassMetaData = mgr.getMetaDataForClass(pcsc);
            }            
            if (pcSuperclassMetaData == null)
            {
                throw new InvalidMetaDataException(LOCALISER,
                    "MetaData.Class.PersistenceCapableSuperclass.NoSuperclassMetaDataError",
                                        name, persistenceCapableSuperclass);
            }
            if (!pcSuperclassMetaData.isPopulated() &&
                !pcSuperclassMetaData.isInitialised())
            {
                pcSuperclassMetaData.populate(loader);
            }
            if (!pcSuperclassMetaData.isInitialised())
            {
                pcSuperclassMetaData.initialise();
            }
            
            // Class has superclass, yet has objectid-class defined! ops, this might result in user errors
            if (objectidClass != null)
            {
                String superObjectIdClass = pcSuperclassMetaData.getObjectidClass();
                if (superObjectIdClass == null || !objectidClass.equals(superObjectIdClass))
                {
                    throw new InvalidMetaDataException(LOCALISER,
                      "MetaData.Class.PersistenceCapableSuperclass.ObjectIdClassDiffersFromSuperclassError",
                                        name, persistenceCapableSuperclass);
                }
                // by default users should only specify the object-id class in the root persistent class
                JPOXLogger.METADATA.warn(LOCALISER.msg("MetaData.Class.PersistenceCapableSuperclass.ObjectIdClassSpecifiedWarning",
                    name, persistenceCapableSuperclass));
            } 
            else
            {
                // get the objectid class from superclass
                this.objectidClass = pcSuperclassMetaData.getObjectidClass();
            }

            if (this.identityType == null)
            {
                this.identityType = pcSuperclassMetaData.getIdentityType();
            }

            //this identityType must be the same as  the one in the superclass
            if (!this.identityType.equals(pcSuperclassMetaData.getIdentityType()))
            {
                throw new InvalidMetaDataException(LOCALISER,
                        "MetaData.Class.IdentityTypeNotEqualsSuperClassIdentityType",
                        fullClassName);                
            }
        }
        else
        {
            // Check that none of our real superclasses is PersistenceCapable since this
            // class hasn't declared a PC superclass
            if (!superclasses.isEmpty())
            {
                Iterator superclassesIter = superclasses.iterator();
                while (superclassesIter.hasNext())
                {
                    Class superclass = (Class)superclassesIter.next();
                    ClassMetaData superCmd = mgr.getMetaDataForClass(superclass);
                    if (superCmd != null && superCmd.getPersistenceModifier() == ClassPersistenceModifier.PERSISTENCE_CAPABLE)
                    {
                        throw new InvalidMetaDataException(LOCALISER, "MetaData.Class.NoSuperclassYetSuperclassIsPC",
                            fullClassName, superclass.getName());
                    }
                }
            }
        }

        if (this.identityType == null)
        {
            this.identityType = (this.objectidClass == null) ?
                IdentityType.DATASTORE : IdentityType.APPLICATION;
        }

        // Add fields for the class that aren't in the FieldMetaData.
        // We use Reflection here since JDOImplHelper would only give use info
        // for enhanced files (and the enhancer needs unenhanced as well). 
        // NOTE 1 : We ignore fields in superclasses
        // NOTE 2 : We ignore "enhanced" fields (added by the JDO enhancer)
        // NOTE 3 : We ignore inner class fields (containing "$") 
        // NOTE 4 : We sort the fields into ascending alphabetical order
        Collections.sort(fields);
        try
        {
            // Get all (reflected) fields in the populating class
            Field[] cls_fields=cls.getDeclaredFields();
            for (int i=0;i<cls_fields.length;i++)
            {
                // Limit to fields in this class, that aren't enhanced fields
                // that aren't inner class fields, and that aren't static
                if (cls_fields[i].getDeclaringClass().getName().equals(fullClassName) &&
                    !cls_fields[i].getName().startsWith("jdo") &&
                    !ClassUtils.isInnerClass(cls_fields[i].getName()) &&
                    !Modifier.isStatic(cls_fields[i].getModifiers()))
                {
                    // Find if there is a FieldMetaData for this field.
                    // This is possible as FieldMetaData implements Comparable
                    if (Collections.binarySearch(fields,cls_fields[i].getName()) < 0)
                    {
                        // Add a default FieldMetaData for this field.
                        JPOXLogger.METADATA.debug(LOCALISER.msg("MetaData.Class.AddingUnspecifiedField",cls_fields[i].getName(),name));
                        FieldMetaData fmd=newDefaultedField(cls_fields[i].getName());
                        fields.add(fmd);
                        Collections.sort(fields);
                    }
                }
            }
        }
        catch (Exception e)
        {
            JPOXLogger.METADATA.error(e.getMessage(), e);
            throw new RuntimeException(e.getMessage());
        }
        Collections.sort(fields);

        // Populate the FieldMetaData with their real field values
        // This will populate any containers in these fields also
        Iterator fields_iter=fields.iterator();
        while (fields_iter.hasNext())
        {
            FieldMetaData fmd=(FieldMetaData)fields_iter.next();

            Field cls_field=null;
            try
            {
                cls_field = cls.getDeclaredField(fmd.getName());
            }
            catch (Exception e)
            {
                // MetaData field doesn't exist in the class!
                throw new InvalidMetaDataException(LOCALISER,
                    "MetaData.Class.DeclaredFieldNotFoundError",
                    fullClassName,fmd.getFullFieldName());
            }
            fmd.populate(loader,cls_field);
        }

        // Update "objectid-class" if required yet not specified
        if (objectidClass == null &&
            persistenceCapableSuperclass == null &&
            identityType == IdentityType.APPLICATION)
        {
            int no_of_pk_fields = 0;
            FieldMetaData fmd_pk = null;
            Iterator fields_pk_iter=fields.iterator();
            while (fields_pk_iter.hasNext())
            {
                FieldMetaData fmd=(FieldMetaData)fields_pk_iter.next();
                if (fmd.isPrimaryKey())
                {
                    fmd_pk = fmd;
                    no_of_pk_fields++;
                }
            }

            // Check for more than 1 field
            if (no_of_pk_fields > 1 || no_of_pk_fields == 0)
            {
                JPOXLogger.METADATA.error(LOCALISER.msg("MetaData.Class.SingleFieldIdentityInvalid", fullClassName, "" + no_of_pk_fields));
                throw new InvalidMetaDataException(LOCALISER, "MetaData.Class.SingleFieldIdentityInvalid", fullClassName, "" + no_of_pk_fields);
            }

            // Assign associated SingleField identity class
            Class pk_type = fmd_pk.getType();
            if (Byte.class.isAssignableFrom(pk_type) ||
                byte.class.isAssignableFrom(pk_type))
            {
                objectidClass = "javax.jdo.ByteIdentity";
            }
            else if (Character.class.isAssignableFrom(pk_type) ||
                     char.class.isAssignableFrom(pk_type))
            {
                objectidClass = "javax.jdo.CharIdentity";
            }
            else if (Integer.class.isAssignableFrom(pk_type) ||
                     int.class.isAssignableFrom(pk_type))
            {
                objectidClass = "javax.jdo.IntIdentity";
            }
            else if (Long.class.isAssignableFrom(pk_type) ||
                     long.class.isAssignableFrom(pk_type))
            {
                objectidClass = "javax.jdo.LongIdentity";
            }
            else if (Short.class.isAssignableFrom(pk_type) ||
                     short.class.isAssignableFrom(pk_type))
            {
                objectidClass = "javax.jdo.ShortIdentity";
            }
            else if (String.class.isAssignableFrom(pk_type))
            {
                objectidClass = "javax.jdo.StringIdentity";
            }
            else
            {
                JPOXLogger.METADATA.error(LOCALISER.msg("MetaData.Class.SingleFieldIdentityInvalidType", fullClassName, pk_type.getName()));
                throw new InvalidMetaDataException(LOCALISER, "MetaData.Class.SingleFieldIdentityInvalidType", fullClassName, pk_type.getName());
            }
            JPOXLogger.METADATA.info(LOCALISER.msg("MetaData.Class.SingleFieldIdentityUsed", fullClassName, objectidClass));
        }

        // Check no of primary key fields (inc superclass)
        int no_of_pk_fields=0;
        if (pcSuperclassMetaData != null)
        {
            no_of_pk_fields = pcSuperclassMetaData.getNoOfPrimaryKeyFields();
        }
        else
        {
            Iterator fields_pk_iter=fields.iterator();
            while (fields_pk_iter.hasNext())
            {
                FieldMetaData fmd=(FieldMetaData)fields_pk_iter.next();
                if (fmd.isPrimaryKey())
                {
                    no_of_pk_fields++;
                }
            }
        }
        if (no_of_pk_fields == 0 &&
            identityType == IdentityType.APPLICATION)
        {
            // No primary key fields found (even in superclasses)
            throw new InvalidMetaDataException(LOCALISER,
                "MetaData.Class.PrimaryKeyNotSpecifiedError",
                fullClassName,objectidClass);
        }

        // Update "objectid-class" since has been specified.
        // Only check at the least derived pc class
        if (objectidClass != null && getPersistenceCapableSuperclass() == null)
        {
            try
            {
                // Load the class, using the same class loader as this class
                Class obj_cls=loader.loadClass(objectidClass);

                JDOUtils.isValidPrimaryKeyClass(obj_cls, this, no_of_pk_fields);
            }
            catch (ClassNotFoundException cnfe)
            {
                // ObjectIdClass not found
                throw new InvalidMetaDataException(LOCALISER,
                    "MetaData.Class.ObjectIdClass.ClassNotFoundError",
                    fullClassName,objectidClass);
            }
        }

        // Check that the inheritance strategy is credible
        if (inheritanceMetaData != null)
        {
            // Check validity of inheritance strategy and discriminator
            if (inheritanceMetaData.getStrategyValue() == InheritanceStrategy.SUPERCLASS_TABLE)
            {
                ClassMetaData superCmd = getClassManagingTableForClass(this);
                if (superCmd == null)
                {
                    // We need a superclass table yet there is no superclass with its own table!
                    throw new InvalidMetaDataException(LOCALISER, "MetaData.Class.Inheritance.NoSuperclassError", fullClassName);
                }

                DiscriminatorMetaData superDismd = superCmd.getInheritanceMetaData().getDiscriminatorMetaData();
                if (superDismd == null)
                {
                    // If we are using "superclass-table" then the superclass should have specified the discriminator.
                    throw new InvalidMetaDataException(LOCALISER, "MetaData.Class.Inheritance.NoSuperclassDiscriminator", 
                        fullClassName, superCmd.getFullClassName());
                }

                DiscriminatorMetaData dismd = inheritanceMetaData.getDiscriminatorMetaData();
                if (superDismd.getStrategy() == DiscriminatorStrategy.VALUE_MAP &&
                    (dismd == null || dismd.getValue() == null))
                {
                    // If we are using "superclass-table" and the discriminator uses "value-map" then we must specify a value
                    throw new InvalidMetaDataException(LOCALISER, "MetaData.Class.Inheritance.NoDiscriminatorValue",
                        fullClassName, superCmd.getFullClassName(), superDismd.getColumnName());
                }
            }
        }

        // Impose a default inheritance strategy to match the JDO 2 spec
        if (inheritanceMetaData == null)
        {
            if (pcSuperclassMetaData != null)
            {
                if (getMetaDataManager().getDefaultInheritanceStrategy().equalsIgnoreCase("JPOX"))
                {
                    // JPOX 1.0 behaviour
                    inheritanceMetaData = new InheritanceMetaData(this, InheritanceStrategy.NEW_TABLE.toString());
                }
                else
                {
                    // JDO2 behaviour
                    inheritanceMetaData = new InheritanceMetaData(this, InheritanceStrategy.SUPERCLASS_TABLE.toString());
                }
            }
            else
            {
                inheritanceMetaData = new InheritanceMetaData(this,InheritanceStrategy.NEW_TABLE.toString());
            }
        }
        
        // JPOX doesnt allow "no-table" classes to have "container" fields since this would break
        // the assumptions about foreign keys in 1-N's
        if (inheritanceMetaData.getStrategyValue() == InheritanceStrategy.NO_TABLE)
        {
            fields_iter = fields.iterator();
            while (fields_iter.hasNext())
            {
                FieldMetaData fmd = (FieldMetaData)fields_iter.next();
                if (fmd.hasContainer())
                {
                    throw new InvalidMetaDataException(LOCALISER, "MetaData.Class.ContainerClassWithNoTableInvalid", fullClassName);
                }
            }
        }

        // "use-poid-generator" was used in JPOX 1.0, and 1.1 until beta-2. Now replaced by "value-strategy".
        if (hasExtension("use-poid-generator"))
        {
            throw new InvalidMetaDataException(LOCALISER, "MetaData.Class.UsePoidGeneratorDeprecated", fullClassName);
        }

        setPopulated();
    }

    /**
     * Utility to navigate up to superclasses to find the next class with its own table.
     * @param cmd ClassMetaData of the class
     * @return The ClassMetaData of the class managing its own table
     */
    private static ClassMetaData getClassManagingTableForClass(ClassMetaData cmd)
    {
        if (cmd == null)
        {
            return null;
        }

        InheritanceMetaData imd = cmd.getInheritanceMetaData();
        if (imd == null)
        {
            return cmd;
        }
        else if (imd.getStrategyValue() == InheritanceStrategy.NEW_TABLE)
        {
            return cmd;
        }
        else if (imd.getStrategyValue() == InheritanceStrategy.SUPERCLASS_TABLE)
        {
            return getClassManagingTableForClass(cmd.getSuperclassMetaData());
        }
        return null;
    }

    /**
     * Method to initialise the object, creating internal convenience arrays.
     * Initialises all sub-objects. If populate() is going to be used it should
     * be used BEFORE calling this method.
     */
    public synchronized void initialise()
    {
        if (JPOXLogger.METADATA.isDebugEnabled())
        {
            JPOXLogger.METADATA.debug(LOCALISER.msg("MetaData.Class.InitialisingMetaData",fullClassName));
        }        
        // Calculate the no of managed fields so we can set up arrays
        Iterator fields_iter=fields.iterator();
        int no_of_managed_fields=0;
        while (fields_iter.hasNext())
        {
            FieldMetaData fmd=(FieldMetaData)fields_iter.next();

            // Initialise the FieldMetaData (and its sub-objects)
            fmd.initialise();
            
            if (fmd.isJdoField())
            {
                no_of_managed_fields++;
            }
        }

        // Generate an array of JDO managed fields (subset of "fields")
        managedFields = new FieldMetaData[no_of_managed_fields]; 
        fields_iter=fields.iterator();
        int field_id=0;
        fieldNumbersByName = new HashMap();
        while (fields_iter.hasNext())
        {
            FieldMetaData fmd=(FieldMetaData)fields_iter.next();

            if (fmd.isJdoField())
            {
                fmd.setFieldId(field_id);
                managedFields[field_id] = fmd;
                fieldNumbersByName.put(fmd.getName(),new Integer(field_id));
                field_id++;
            }
        }

        if (pcSuperclassMetaData != null)
        {
            if (!pcSuperclassMetaData.isInitialised())
            {
                pcSuperclassMetaData.initialise();
            }
            noOfInheritedManagedFields = pcSuperclassMetaData.getNoOfInheritedManagedFields() + pcSuperclassMetaData.getNoOfManagedFields();
        }

        // Set up the various convenience arrays of field numbers
        int total_field_count=noOfInheritedManagedFields + managedFields.length;
        allFieldNumbers = new int[total_field_count];
        defaultFetchGroupFieldFlags = new boolean[total_field_count];
        secondClassMutableFieldFlags = new boolean[total_field_count];
        nonPrimaryKeyFieldFlags = new boolean[total_field_count];
        int pk_field_count=0;
        int dfg_field_count=0;
        int scm_field_count=0;
        for (int i=0;i<total_field_count;i++)
        {
            allFieldNumbers[i] = i;

            FieldMetaData fmd=getManagedFieldAbsoluteInternal(i);
            if (fmd.isPrimaryKey())
            {
                pk_field_count++;
            }
            else
            {
                nonPrimaryKeyFieldFlags[i] = true;
            }
            if (fmd.isDefaultFetchGroup())
            {
                defaultFetchGroupFieldFlags[i] = true;
                dfg_field_count++;
            }
            if (TypeManager.getTypeManager().isSecondClassMutableType(fmd.getTypeName()))
            {
                secondClassMutableFieldFlags[i] = true;
                scm_field_count++;
            }
        }
 
        if (pk_field_count > 0 &&
                identityType != IdentityType.APPLICATION)
        {
            // primary key fields found, but not using application identity
            throw new InvalidMetaDataException(LOCALISER,
                "MetaData.Class.PrimaryKeyFieldsFoundButNotUsingApplicationIdentity",
                fullClassName,new Integer(pk_field_count),identityType);
        }        
        else if (pk_field_count > 0)
        {
            primaryKeyFieldNumbers = new int[pk_field_count];
            for (int i=0,pk_num=0;i<total_field_count;i++)
            {
                FieldMetaData fmd=getManagedFieldAbsoluteInternal(i);
                if (fmd.isPrimaryKey())
                {
                    primaryKeyFieldNumbers[pk_num++] = i;
                }
            }
        }
        else if (pk_field_count == 0 &&
            identityType == IdentityType.APPLICATION)
        {
            // No primary key fields found
            throw new InvalidMetaDataException(LOCALISER,
                "MetaData.Class.PrimaryKeyNotSpecifiedError",
                fullClassName,objectidClass);
        }

        nonPrimaryKeyFieldNumbers = new int[total_field_count-pk_field_count];
        for (int i=0,npkf=0;i<total_field_count;i++)
        {
            FieldMetaData fmd = getManagedFieldAbsoluteInternal(i);
            if (!fmd.isPrimaryKey())
            {
                nonPrimaryKeyFieldNumbers[npkf++] = i;
            }
        }

        defaultFetchGroupFieldNumbers = new int[dfg_field_count];
        secondClassMutableFieldNumbers = new int[scm_field_count];
        for (int i=0,dfg_num=0,scm_num=0;i<total_field_count;i++)
        {
            if (defaultFetchGroupFieldFlags[i])
            {
                defaultFetchGroupFieldNumbers[dfg_num++] = i;
            }
            if (secondClassMutableFieldFlags[i])
            {
                secondClassMutableFieldNumbers[scm_num++] = i;
            }
        }

        // Initialise any sub-objects
        implementsMetaData = new ImplementsMetaData[implementations.size()];
        for (int i=0; i<implementations.size(); i++)
        {
            implementsMetaData[i] = (ImplementsMetaData) implementations.get(i);
            implementsMetaData[i].initialise();
        }
        joinMetaData = new JoinMetaData[joins.size()];
        for (int i=0; i<joinMetaData.length; i++)
        {
            joinMetaData[i] = (JoinMetaData) joins.get(i);
            joinMetaData[i].initialise();
        }
        indexMetaData = new IndexMetaData[indexes.size()];
        for (int i=0; i<indexMetaData.length; i++)
        {
            indexMetaData[i] = (IndexMetaData) indexes.get(i);
            indexMetaData[i].initialise();
        }
        foreignKeyMetaData = new ForeignKeyMetaData[foreignKeys.size()];
        for (int i=0; i<foreignKeyMetaData.length; i++)
        {
            foreignKeyMetaData[i] = (ForeignKeyMetaData) foreignKeys.get(i);
            foreignKeyMetaData[i].initialise();
        }
        uniqueMetaData = new UniqueMetaData[uniqueConstraints.size()];
        for (int i=0; i<uniqueMetaData.length; i++)
        {
            uniqueMetaData[i] = (UniqueMetaData) uniqueConstraints.get(i);
            uniqueMetaData[i].initialise();
        }

        boolean foundValuesFetchGroup = false;
        for (int i=0; i<fetchGroups.size(); i++)
        {
            FetchGroupMetaData fgmd = (FetchGroupMetaData) fetchGroups.get(i);
            if (fgmd.getName().equals(FetchPlan.VALUES))
            {
                foundValuesFetchGroup = true;
            }
        }        
        //if not defined a VALUES fetch-group, creates one
        if (!foundValuesFetchGroup)
        {
            FetchGroupMetaData fgmd = new FetchGroupMetaData(this,null,Boolean.FALSE.toString(), FetchPlan.VALUES);
            for (int i=0; i<managedFields.length; i++)
            {
                FieldMetaData fmd = managedFields[i];
                if (fmd.isPrimaryKey() || (fmd.getPersistenceModifier() == FieldPersistenceModifier.PERSISTENT && MetaDataUtils.isDefaultFetchGroup(fmd.getType())))
                {
                    fgmd.addField(new FieldMetaData(fgmd,fmd.getName()));
                }
            }
            addFetchGroup(fgmd);
        }
        fetchGroupMetaData = new FetchGroupMetaData[fetchGroups.size()];
        for (int i=0; i<fetchGroupMetaData.length; i++)
        {
            fetchGroupMetaData[i] = (FetchGroupMetaData) fetchGroups.get(i);
            fetchGroupMetaData[i].initialise();
            if (fetchGroupMetaData[i].getName().equals(FetchPlan.VALUES))
            {
                foundValuesFetchGroup = true;
            }
        }         
        
        // If using datastore id and user hasn't provided the identity element,
        // add a defaulted one
        if (identityType == IdentityType.DATASTORE && identityMetaData == null)
        {
            identityMetaData = new IdentityMetaData(this,null,null,null);
        }

        if (versionMetaData != null)
        {
            versionMetaData.initialise();
        }
        if (identityMetaData != null)
        {
            identityMetaData.initialise();
        }
        if (inheritanceMetaData != null)
        {
            inheritanceMetaData.initialise();
        }

        setInitialised();
    }

    /**
     * Method to check whether the Meta-Data has been initialised.
     * @throws JDOFatalInternalException Thrown if the Meta-Data hasn't been
     *                                   initialised. 
     **/
    private void checkInitialised()
    throws JDOFatalInternalException
    {
        if (!isInitialised())
        {
            throw new JDOFatalInternalException(LOCALISER.msg("MetaData.Class.NotInitialisedError",fullClassName));
        }
    }

    /**
     * Utility to add a defaulted FieldMetaData to the class. Provided as a
     * method since then any derived classes can override it (e.g
     * JDOConfigClass can create a JDOConfigField)
     * @param name name of field
     */
    protected FieldMetaData newDefaultedField(String name)
    {
        return new FieldMetaData(this, name);
    }

    // ------------------------------ Accessors --------------------------------
 
    /**
     * Accessor for the number of named queries.
     * @return no of named queries
     */
    public int getNoOfQueries()
    {
        return queries.size();
    }

    /**
     * Accessor for the metadata of a named query.
     * @param index query index
     * @return Meta-Data for the named query.
     */
    public QueryMetaData getQuery(int index)
    {
        if (index < 0 || index >= queries.size())
        {
            return null;
        }
        return (QueryMetaData)queries.get(index);
    }

    /**
     * Accessor for versionMetaData
     * @return Returns the versionMetaData.
     */
    public final VersionMetaData getVersionMetaData()
    {
        return versionMetaData;
    }
    
    /**
     * Accessor for joinMetaData
     * @return Returns the joinMetaData.
     */
    public final JoinMetaData[] getJoinMetaData()
    {
        return joinMetaData;
    }

    /**
     * Accessor for Unmapped MetaData
     * @return Returns the Unmapped MetaData.
     */
    public final UnmappedMetaData getUnmappedMetaData()
    {
        return unmappedMetaData;
    }

    /**
     * Accessor for the full class name.
     * @return full class name.
     */
    public String getFullClassName()
    {
        return fullClassName;
    }

    /**
     * Accessor for the number of fields.
     * This is the total number of fields (inc static, final etc) in this class 
     * @return no of fields
     */
    public int getNoOfFields()
    {
        return fields.size();
    }

    /**
     * Accessor for fetchGroupMetaData
     * @return Returns the fetchGroupMetaData.
     */
    public final FetchGroupMetaData[] getFetchGroupMetaData()
    {
        return fetchGroupMetaData;
    }
    
    /**
     * Accessor for the metadata of a field. Does not include superclasses
     * @param index field index relative to this class only starting from 0
     * @return Meta-Data for the field.
     */
    public FieldMetaData getField(int index)
    {
        if (index < 0 || index >= fields.size())
        {
            return null;
        }
        return (FieldMetaData)fields.get(index);
    }

    /**
     * Accessor for the identity-type. 
     * @return identity-type tag value
     */
    public IdentityType getIdentityType()
    {
        return identityType;
    }

    /**
     * Accessor for indexMetaData
     * @return Returns the indexMetaData.
     */
    public final IndexMetaData[] getIndexMetaData()
    {
        return indexMetaData;
    }

    /**
     * Accessor for foreignKeyMetaData
     * @return Returns the foreignKeyMetaData.
     */
    public final ForeignKeyMetaData[] getForeignKeyMetaData()
    {
        return foreignKeyMetaData;
    }

    /**
     * Accessor for UniqueMetaData
     * @return Returns the UniqueMetaData.
     */
    public final UniqueMetaData[] getUniqueMetaData()
    {
        return uniqueMetaData;
    }

    /**
     * Accessor for the persistence-modifier
     * @return persistence-modifier tag value
     */
    public ClassPersistenceModifier getPersistenceModifier()
    {
        return persistenceModifier;
    }

    /**
     * Accessor for the class name.
     * @return class name
     */
    public String getName()
    {
        return name;
    }

    /**
     * Accessor for the objectid-class tag value.
     * @return objectid-class tag value
     */
    public String getObjectidClass()
    {
        return objectidClass;
    }

    /**
     * Accessor for the persistence-capable-superclass tag value.
     * @return persistence-capable-superclass tag value
     */
    public String getPersistenceCapableSuperclass()
    {
        if (StringUtils.isWhitespace(persistenceCapableSuperclass))
        {
            return null;
        }
        return persistenceCapableSuperclass;
    }

    /**
     * Accessor for the table name.
     * @return table name
     */
    public String getTable()
    {
        return table;
    }

    /**
     * Check if the argument cmd is the same as this or a descedent.
     * @param cmd the ClassMetaData to be verify if this is an ancestor
     * @return true if the argument is a child or same as this
     */
    public boolean isSameOrAncestorOf(ClassMetaData cmd)
    {
        checkInitialised();
        
        if( cmd == null )
        {
            return false;
        }
        
        if( getFullClassName().equals(cmd.getFullClassName()) )
        {
            return true;
        }
        
        ClassMetaData parent = cmd.getSuperclassMetaData();
        while( parent != null )
        {
            if( getFullClassName().equals(parent.getFullClassName()) )
            {
                return true;
            }
            parent = parent.getSuperclassMetaData();
        }
        return false;
    }
    
    /**
     * Accessor for the requires-extent tag value.
     * @return requires-extent tag value
     */
    public boolean isRequiresExtent()
    {
        return requiresExtent;
    }

    /**
     * Accessor for the detachable tag value.
     * @return detachable tag value
     */
    public boolean isDetachable()
    {
        return detachable;
    }

    /**
     * Accessor for the embedded-only tag value.
     * @return embedded-only tag value
     */
    public boolean isEmbeddedOnly()
    {
        return embeddedOnly;
    }

    /**
     * Accessor for identityMetaData
     * @return Returns the identityMetaData.
     */
    public final IdentityMetaData getIdentityMetaData()
    {
        return identityMetaData;
    }

    /**
     * Accessor for inheritanceMetaData
     * @return Returns the inheritanceMetaData.
     */
    public final InheritanceMetaData getInheritanceMetaData()
    {
        return inheritanceMetaData;
    }

    /**
     * Convenience accessor for the discriminator strategy applying to this class.
     * This is specified on the class managing the table if at all.
     * @return The discriminator strategy
     */
    public final DiscriminatorStrategy getDiscriminatorStrategy()
    {
        if (inheritanceMetaData == null)
        {
            return null;
        }
        else if (inheritanceMetaData.getStrategyValue() == InheritanceStrategy.NEW_TABLE)
        {
            if (inheritanceMetaData.getDiscriminatorMetaData() != null)
            {
                return inheritanceMetaData.getDiscriminatorMetaData().getStrategy();
            }
            else
            {
                // Not specified.
                return null;
            }
        }
        else if (getSuperclassMetaData() != null)
        {
            return getSuperclassMetaData().getDiscriminatorStrategy();
        }
        else
        {
            // Unknown!
            return null;
        }
    }

    /**
     * Convenience accessor for the MetaData Manager in use.
     * @return MetaDataManager.
     */
    public MetaDataManager getMetaDataManager()
    {
        if (parent != null && parent.getParent() != null)
        {
            return getPackageMetaData().getFileMetaData().getMetaDataManager();
        }
        return null;
    }

    /**
     * Convenience accessor for the parent Package MetaData.
     * @return MetaData for parent package.
     */
    public PackageMetaData getPackageMetaData()
    {
        if (parent != null)
        {
            return (PackageMetaData)parent;
        }
        return null;
    }

    /**
     * Convenience accessor for the package name.
     * @return package name.
     */
    public String getPackageName()
    {
        return getPackageMetaData().getName();
    }

    /**
     * Method to check if a field exists in this classes definition.
     * @param fieldName Nme of field
     * @return return true if exists.
     */
    public boolean hasField(String fieldName)
    {
        Iterator iter=fields.iterator();
        while (iter.hasNext())
        {
            FieldMetaData field=(FieldMetaData)iter.next();
            if (field.getName().equals(fieldName))
            {
                return true;
            }
        }
        return false;
    }

    /**
     * Accessor for the Meta-Data for a field. Include superclasses
     * @param name the name of the field
     * @return Meta-Data for the field.
     */
    public FieldMetaData getField(String name)
    {
        if (name == null)
        {
            return null;
        }

        Iterator iter=fields.iterator();
        while (iter.hasNext())
        {
            FieldMetaData fmd = (FieldMetaData)iter.next();
            if (fmd.getName().equals(name))
            {
                return fmd;
            }
        }

        // Check superclass for the field with this name
        if (pcSuperclassMetaData != null)
        {
            return pcSuperclassMetaData.getField(name);
        }
        return null;
    }

    /**
     * Accessor for the Meta-Data for the superclass of this class.
     * @return MetaData of the superclass
     */
    public ClassMetaData getSuperclassMetaData()
    {
        // checkInitialised();

        return pcSuperclassMetaData;
    }

    /**
     * Convenience accessor for the ClassMetaData of the base object in this hierarchy.
     * @return The ClassMetaData for the base object.
     */
    public ClassMetaData getBaseclassMetaData()
    {
        if (pcSuperclassMetaData != null)
        {
            return pcSuperclassMetaData.getBaseclassMetaData();
        }
        else
        {
            return this;
        }
    }

    /**
     * Accessor for the number of JDO managed fields. 
     * The number of fields is for this class only.
     * @return no of JDO managed fields in this class
     */
    public int getNoOfManagedFields()
    {
        // checkInitialised();

        if (managedFields == null)
        {
            return 0;
        }
        return managedFields.length;
    }

    /**
     * Accessor for the managed fields in this class (not including superclass).
     * @return The JDO managed fields in this class
     */
    public FieldMetaData[] getManagedFields()
    {
        checkInitialised();

        return managedFields;
    }

    /**
     * Accessor for a managed field in this class. The position is relative to
     * the first field in this class (ignores superclasses).
     * @param position The position of the managed field. 0 = first in the class
     * @return The JDO managed field at that position
     */
    public FieldMetaData getManagedField(int position)
    {
        checkInitialised();

        if (managedFields == null)
        {
            return null;
        }
        if (position < 0 || position >= managedFields.length)
        {
            return null;
        }

        return managedFields[position];
    }

    /**
     * Accessor for the number of inherited managed fields in superclasses.
     * @return No of inherited managed fields in superclasses.
     */
    public int getNoOfInheritedManagedFields()
    {
        checkInitialised();

        return noOfInheritedManagedFields;
    }

    /**
     * Accessor for a managed field including superclass fields.
     * @param abs_position The position of the managed field including the
     * superclass. Fields are numbered from 0 in the root superclass.
     * @return The JDO managed field at this "absolute" position.
     */
    public FieldMetaData getManagedFieldAbsolute(int abs_position)
    {
        checkInitialised();

        return getManagedFieldAbsoluteInternal(abs_position);
    }

    /**
     * Internal method to get the field for an absolute field number.
     * @param abs_position The position of the managed field including the
     * superclass. Fields are numbered from 0 in the root superclass.
     * @return The JDO managed field at this "absolute" position.
     */
    protected FieldMetaData getManagedFieldAbsoluteInternal(int abs_position)
    {
        // If the field is in a superclass, go there
        if (abs_position < noOfInheritedManagedFields)
        {
            if (pcSuperclassMetaData == null)
            {
                return null;
            }
            else
            {
                return pcSuperclassMetaData.getManagedFieldAbsoluteInternal(abs_position);
            }
        }
        // If the field is in this class, return it
        else if (abs_position - noOfInheritedManagedFields >= managedFields.length)
        {
            return null;
        }
        else
        {
            return managedFields[abs_position - noOfInheritedManagedFields];
        }
    }
    
    /**
     * Accessor for the Absolute field number for a relative field number
     * @param relativeFieldNo The relative field number
     * @return The absolute field number
     */
    public int getAbsoluteFieldNumberForRelativeFieldNumber(int relativeFieldNo)
    {
        return noOfInheritedManagedFields + relativeFieldNo;
    }

    /**
     * Accessor for the position of the field with the specified name.
     * The returned position is relative to this class only.
     * @param field_name Name of the field
     * @return Position of the field in this class.
     */
    public int getFieldNumber(String field_name)
    {
        checkInitialised();

        if (field_name == null)
        {
            return -1;
        }

        Integer i=(Integer)fieldNumbersByName.get(field_name);

        return i == null ? -1 : i.intValue();
    }

    /**
     * Accessor for the absolute position of the field with the specified name.
     * The absolute position has origin of the root superclass, starting at 0.
     * @param field_name Name of the field
     * @return Absolute position of the field.
     */
    public int getFieldNumberAbsolute(String field_name)
    {
        checkInitialised();

        if (field_name == null)
        {
            return -1;
        }

        int i=getFieldNumber(field_name);
        if (i < 0)
        {
            if (pcSuperclassMetaData != null)
            {
                i = pcSuperclassMetaData.getFieldNumberAbsolute(field_name);
            }
        }
        else
        {
            i += noOfInheritedManagedFields;
        }

        return i;
    }

    /**
     * Accessor for the number of primary key fields.
     * @return no of primary key fields
     */
    public int getNoOfPrimaryKeyFields()
    {
        if (primaryKeyFieldNumbers == null)
        {
            return 0;
        }
        else
        {
            return primaryKeyFieldNumbers.length;
        }
    }

    /**
     * Accessor for all field numbers. These are absolute numbers and include
     * superclasses and are really just 0, 1, 2, ... n.
     * @return The field numbers of all (managed) fields.
     **/
    public int[] getAllFieldNumbers()
    {
        checkInitialised();
        return allFieldNumbers;
    }
 
    /**
     * Accessor for the field numbers of the primary key fields. These are
     * absolute numbers (including superclasses).
     * @return The field numbers of the primary key fields
     **/
    public int[] getPrimaryKeyFieldNumbers()
    {
        checkInitialised();
        return primaryKeyFieldNumbers;
    }
 
    /**
     * Accessor for the field numbers of the non primary key fields (inc
     * superclass fields).
     * @return The field numbers
     **/
    public int[] getNonPrimaryKeyFieldNumbers()
    {
        checkInitialised();
        return nonPrimaryKeyFieldNumbers;
    }    

    /**
     * Accessor for the flags of the non primary key fields (inc
     * superclass fields).
     * @return The flags whether the field is non primary key
     **/
    public boolean[] getNonPrimaryKeyFieldFlags()
    {
        checkInitialised();
        return nonPrimaryKeyFieldFlags;
    } 
    
    /**
     * Accessor for the absolute field numbers of the default fetch group
     * fields (inc superclasses).
     * @return The field numbers of the DFG fields (inc superclasses).
     **/
    public int[] getDefaultFetchGroupFieldNumbers()
    {
        checkInitialised();
        return defaultFetchGroupFieldNumbers;
    }

    /**
     * Accessor for the flags of the default fetch group fields (inc superclass
     * fields).
     * @return The flags whether the field is in the DFG
     **/
    public boolean[] getDefaultFetchGroupFieldFlags()
    {
        checkInitialised();
        return defaultFetchGroupFieldFlags;
    }

    /**
     * Accessor for the absolute field numbers of the second class mutable
     * fields
     * @return The field numbers of the second class mutable fields (inc
     *         superclasses).
     **/
    public int[] getSecondClassMutableFieldNumbers()
    {
        checkInitialised();
        return secondClassMutableFieldNumbers;
    }

    /**
     * Accessor for the flags of the second class mutable fields (inc
     * superclass fields).
     * @return The flags whether the field is second class mutable
     **/
    public boolean[] getSecondClassMutableFieldFlags()
    {
        checkInitialised();
        return secondClassMutableFieldFlags;
    }

    // ------------------------------- Mutators --------------------------------
 
    /**
     * Method to add a field to this class. Rejects the addition of duplicate
     * named fields.
     * @param fmd Meta-Data for the field.
     */
    public void addField(FieldMetaData fmd)
    {
        if (fmd == null)
        {
            return;
        }

        if (isInitialised())
        {
            throw new RuntimeException("adding field???"+fmd.getName());
        }
        Iterator iter=fields.iterator();
        while (iter.hasNext())
        {
            FieldMetaData md = (FieldMetaData)iter.next();
            if (fmd.getName().equals(md.getName()))
            {
                throw new RuntimeException(LOCALISER.msg("MetaData.Class.DuplicateFieldError",fullClassName,fmd.getName()));
            }
        }
        fields.add(fmd);
    }

    /**
     * Method to add a named query to this class. Rejects the addition of
     * duplicate named queries.
     * @param qmd Meta-Data for the query.
     */
    public void addQuery(QueryMetaData qmd)
    {
        if (qmd == null)
        {
            return;
        }

        queries.add(qmd);
    }

    /**
     * Method to add an index to this class.
     * @param idxmd Meta-Data for the index.
     */
    public void addIndex(IndexMetaData idxmd)
    {
        if (idxmd == null)
        {
            return;
        }

        if (isInitialised())
        {
            throw new RuntimeException("Already initialised");
        }
        indexes.add(idxmd);
    }

    /**
     * Method to add an foreign-key to this class.
     * @param fkmd Meta-Data for the foreign-key.
     */
    public void addForeignKey(ForeignKeyMetaData fkmd)
    {
        if (fkmd == null)
        {
            return;
        }

        if (isInitialised())
        {
            throw new RuntimeException("Already initialised");
        }
        foreignKeys.add(fkmd);
    }

    /**
     * Method to add a unique constraint to this class.
     * @param unimd Meta-Data for the unique constraint.
     */
    public void addUniqueConstraint(UniqueMetaData unimd)
    {
        if (unimd == null)
        {
            return;
        }

        if (isInitialised())
        {
            throw new RuntimeException("Already initialised");
        }
        uniqueConstraints.add(unimd);
    }

    /**
     * Method to add a fetch-group to this class.
     * @param fgmd Meta-Data for the fetch-group.
     */
    public void addFetchGroup(FetchGroupMetaData fgmd)
    {
        if (fgmd == null)
        {
            return;
        }

        if (isInitialised())
        {
            throw new RuntimeException("Already initialised");
        }
        fetchGroups.add(fgmd);
    }
    
    /**
     * Method to add a join to this class.
     * Rejects the addition of duplicate named fields.
     * @param jnmd Meta-Data for the join.
     */
    public void addJoin(JoinMetaData jnmd)
    {
        if (jnmd == null)
        {
            return;
        }

        if (isInitialised())
        {
            throw new RuntimeException("Already initialised");
        }
        joins.add(jnmd);
    }

    /**
     * Method to add an implements to this class.
     * @param implmd Meta-Data for the implements
     */
    public void addImplements(ImplementsMetaData implmd)
    {
        if (implmd == null)
        {
            return;
        }

        if (isInitialised())
        {
            throw new RuntimeException("Already initialised");
        }
        implementations.add(implmd);
    }

    /**
     * Mutator for the Unmapped MetaData.
     * @param unmappedMetaData The UnmappedMetaData to set.
     */
    public final void setUnmappedMetaData(UnmappedMetaData unmappedMetaData)
    {
        this.unmappedMetaData = unmappedMetaData;
    }

    /**
     * Mutator for the Version MetaData.
     * @param versionMetaData The versionMetaData to set.
     */
    public final void setVersionMetaData(VersionMetaData versionMetaData)
    {
        this.versionMetaData = versionMetaData;
    }

    /**
     * Mutator for the identity MetaData.
     * @param identityMetaData The identityMetaData to set.
     */
    public final void setIdentityMetaData(IdentityMetaData identityMetaData)
    {
        this.identityMetaData = identityMetaData;
    }

    /**
     * Mutator for the inheritance MetaData.
     * @param inheritanceMetaData The inheritanceMetaData to set.
     */
    public final void setInheritanceMetaData(InheritanceMetaData inheritanceMetaData)
    {
        this.inheritanceMetaData = inheritanceMetaData;
    }

    /**
     * Mutator for the table name.
     * @param table_name Name of the table
     */
    public final void setTable(String table_name)
    {
        this.table = table_name;
    }

    // ------------------------------ Utilities --------------------------------
 
    /**
     * Method to return the ClassMetaData records for classes referenced
     * by this object. This adds the entries to orderedCMDs ordered by
     * dependency, and to referencedCMDs for fast lookups.
     * <p>
     * Uses recursion to add all referenced ClassMetaData for any fields,
     * objectid classes, superclasses, and JPOX extension RDBMS "views".
     * </p>
     * <p>
     * This is the entry point for this process, and provides the core of the
     * "persistence-by-reachability" concept.
     * </p> 
     * @param orderedCMDs List of ordered ClassMetaData objects (added to).
     * @param referencedCMDs Set of all ClassMetaData objects (added to).
     * @param dba_vendor_id Vendor ID of the DBA. Used for view addition.
     * @param clr the ClassLoaderResolver
     */
    void getReferencedClassMetaData(final List orderedCMDs,
                                    final Set referencedCMDs,
                                    final String dba_vendor_id,
                                    final ClassLoaderResolver clr)
    {
        Map viewReferences = new HashMap();
        getReferencedClassMetaData(orderedCMDs,referencedCMDs,dba_vendor_id,viewReferences,clr);
    }

    /**
     * Method to return the ClassMetaData for classes referenced by this
     * object. This method does the actual work of addition.
     * @param orderedCMDs List of ordered ClassMetaData objects (added to).
     * @param referencedCMDs Set of all ClassMetaData objects (added to).
     * @param dba_vendor_id Vendor ID of the DBA. Used for view addition.
     * @param viewReferences Map, mapping class name to set of referenced class
     *                       for all views.
     * @param clr the ClassLoaderResolver
     */
    private void getReferencedClassMetaData(final List orderedCMDs,
                                            final Set referencedCMDs,
                                            final String dba_vendor_id,
                                            final Map viewReferences,
                                            final ClassLoaderResolver clr)
    {
        // Recursively call getReferencedClassMetaData(...) before adding them
        // to the orderedCmds and referenced. This will ensure that any
        // classes with dependencies on them are put in the orderedCmds List
        // in the correct order.
        if (!referencedCMDs.contains(this))
        {
            // Go ahead and add this class to the referenced Set, it will
            // get added to the orderedCMDs List after all classes that this
            // depends on have been added.
            referencedCMDs.add(this);

            for (int i=0;i<managedFields.length;i++)
            {
                FieldMetaData fmd=managedFields[i];
                fmd.getReferencedClassMetaData(orderedCMDs,referencedCMDs,dba_vendor_id,clr);
            }

            // Add on any superclass
            if (persistenceCapableSuperclass != null)
            {
                ClassMetaData super_cmd=getSuperclassMetaData();
                super_cmd.getReferencedClassMetaData(orderedCMDs,referencedCMDs,dba_vendor_id,clr);
            }

            // Add on any objectid class
            if (objectidClass != null)
            {
                ClassMetaData id_cmd=getMetaDataManager().getMetaDataForClass(objectidClass,clr);
                if (id_cmd != null)
                {
                    id_cmd.getReferencedClassMetaData(orderedCMDs,referencedCMDs,dba_vendor_id,clr);
                }
            }

            // Add on any views defined for this class
            String view_defn_key="view-definition";
            if (dba_vendor_id != null)
            {
                view_defn_key += "-" + dba_vendor_id;
            }
            if (hasExtension(view_defn_key))
            {
                MacroString viewDef = new MacroString(fullClassName,
                           getValueForExtension("view-imports"),
                           getValueForExtension(view_defn_key));
                viewDef.substituteMacros(new MacroString.MacroHandler()
                    {
                        public void onIdentifierMacro(MacroString.IdentifierMacro im)
                        {
                            addViewReference(viewReferences,im.className);
                            ClassMetaData view_cmd=getMetaDataManager().getMetaDataForClass(im.className,clr);
                            view_cmd.getReferencedClassMetaData(orderedCMDs,referencedCMDs,dba_vendor_id,viewReferences,clr);
                        }

                        public void onParameterMacro(MacroString.ParameterMacro pm)
                        {
                            throw new JDOUserException("Parameter macros not allowed in view definitions: " + pm);
                        }
                    },clr);
            }

            orderedCMDs.add(this);
        }
    }

    /**
     * Method to add a reference for views. Check the view references for
     * circular dependencies.  If there are any circular dependencies, throw
     * a JDOUserException.
     * @param viewReferences The Map of classname to Set of referenced Classes
     *                       to add the reference to.
     * @param referenced_name Class name of the referenced class
     * @throws JDOFatalUserException Thrown if a circular reference is found
     */
    private void addViewReference(Map viewReferences,String referenced_name)
    throws JDOFatalUserException
    {
        if (fullClassName.equals(referenced_name))
        {
            // Add this reference to the Map.
            Set referencedSet = (Set)viewReferences.get(referenced_name);
            if (referencedSet == null)
            {
                referencedSet = new HashSet();
                viewReferences.put(fullClassName,referencedSet);
            }
            referencedSet.add(referenced_name);

            // Check to see if there is a circular dependency.  This will
            // be true if the referenced class references this class.
            ViewUtils.checkForCircularViewReferences(viewReferences,fullClassName,referenced_name,null);
        }
    }

    /**
     * Returns a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        return toString("","");
    }

    /**
     * Returns a string representation of the object.
     * This can be used as part of a facility to output a MetaData file. 
     * @param prefix prefix string
     * @param indent indent string
     * @return a string representation of the object.
     */
    public String toString(String prefix,String indent)
    {
        StringBuffer sb = new StringBuffer();
        sb.append(prefix).append("<class name=\"" + name + "\"\n");
        if (identityType != null)
        {
            sb.append(prefix).append("       identity-type=\"" + identityType + "\"\n");
        }
        if (objectidClass != null)
        {
            sb.append(prefix).append("       objectid-class=\"" + objectidClass + "\"\n");
        }
        if (persistenceCapableSuperclass != null)
        {
            sb.append(prefix).append("       persistence-capable-superclass=\"" + persistenceCapableSuperclass + "\"\n");
        }
        if (!requiresExtent)
        {
            sb.append(prefix).append("       requires-extent=\"" + requiresExtent + "\"");
        }
        if (detachable)
        {
            sb.append(prefix).append("       detachable=\"" + detachable + "\"\n");
        }
        if (embeddedOnly)
        {
            sb.append(prefix).append("       embedded-only=\"" + embeddedOnly + "\"\n");
        }
        if (persistenceModifier != null)
        {
            sb.append(prefix).append("       persistence-modifier=\"" + persistenceModifier + "\"\n");
        }
        if (table != null)
        {
            sb.append(prefix).append("       table=\"" + table + "\"\n");
        }
        sb.append(">\n");

        // Identity
        if (identityMetaData != null)
        {
            sb.append(identityMetaData.toString(prefix + indent,indent));
        }

        // TODO Implements
 
        // Inheritance
        if (inheritanceMetaData != null)
        {
            sb.append(inheritanceMetaData.toString(prefix + indent,indent));
        }

        // Add Version
        if (versionMetaData != null)
        {
            sb.append(versionMetaData.toString(prefix + indent,indent));
        }

        // Add joins
        for (int i=0; i<joins.size(); i++)
        {
            JoinMetaData jmd = (JoinMetaData)joins.get(i);
            sb.append(jmd.toString(prefix + indent,indent));
        }

        // Add foreign-keys
        for (int i=0; i<foreignKeys.size(); i++)
        {
            ForeignKeyMetaData fkmd = (ForeignKeyMetaData)foreignKeys.get(i);
            sb.append(fkmd.toString(prefix + indent,indent));
        }
 
        // Add indexes
        for (int i=0; i<indexes.size(); i++)
        {
            IndexMetaData imd = (IndexMetaData)indexes.get(i);
            sb.append(imd.toString(prefix + indent,indent));
        }
        
        // Add unique constraints
        for (int i=0; i<uniqueConstraints.size(); i++)
        {
            UniqueMetaData unimd = (UniqueMetaData)uniqueConstraints.get(i);
            sb.append(unimd.toString(prefix + indent,indent));
        }

        // Add fields
        for (int i=0; i<fields.size(); i++)
        {
            FieldMetaData f = (FieldMetaData)fields.get(i);
            sb.append(f.toString(prefix + indent,indent));
        }

        // Unmapped
        if (unmappedMetaData != null)
        {
            sb.append(unmappedMetaData.toString(prefix + indent,indent));
        }

        // Add queries
        for (int i=0;i<queries.size();i++)
        {
            QueryMetaData q = (QueryMetaData)queries.get(i);
            sb.append(q.toString(prefix + indent,indent));
        }

        // Add fetch-groups
        for (int i=0; i<fetchGroups.size(); i++)
        {
            FetchGroupMetaData fgmd = (FetchGroupMetaData)fetchGroups.get(i);
            sb.append(fgmd.toString(prefix + indent,indent));
        }

        // Add extensions
        sb.append(super.toString(prefix + indent,indent)); 

        sb.append(prefix + "</class>\n");
        return sb.toString();
    }
}