/**********************************************************************
Copyright (c) 04-Jun-2004 Andy Jefferson and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
    Kikuchi Kousuke - org.jpox.enhancer.conf.JDOConfigCollection
    Erik Bengtson - add dependent elements
    ...
**********************************************************************/
package org.jpox.metadata;

import java.util.List;
import java.util.Set;
import javax.jdo.spi.PersistenceCapable;
import org.jpox.ClassLoaderResolver;
import org.jpox.util.ClassUtils;
import org.jpox.util.JPOXLogger;

/**
 * Representation of the MetaData of a collection.
 *
 * <H3>MetaData Element</H3>
 * Represents the following MetaData Element 
 * <PRE>
 * &lt;!ELEMENT collection (extension)*&gt;
 * &lt;!ATTLIST collection element-type CDATA #IMPLIED&gt;
 * &lt;!ATTLIST collection embedded-element (true|false) #IMPLIED&gt;
 * &lt;!ATTLIST collection dependent-element (true|false) #IMPLIED&gt;
 * </PRE>
 *
 * @since 1.1 
 * @version $Revision: 1.3 $
 */
public class CollectionMetaData extends ContainerMetaData
{
    /**
     * Element-type tag value.
     */
    protected String elementType = "java.lang.Object";

    /**
     * Embedded-element tag value.
     */
    protected Boolean embeddedElement;
    
    /**
     * dependent-element tag value.
     * 
     * The dependent-element attribute indicates that the collection's element
     * contains a reference that is to be deleted if the referring instance is
     * deleted.
     */
    protected Boolean dependentElement;    

    /**
     * Constructor.
     * @param parent The parent Field element
     * @param elementType The type of element
     * @param embeddedElement Whether elements are embedded 
     * @param dependentElement attribute dependent-element value
     */
    public CollectionMetaData(FieldMetaData parent,
                              String elementType,
                              String embeddedElement,
                              String dependentElement)
    {
        super(parent);

        this.elementType = ClassUtils.createFullClassName(parent.getClassMetaData().getPackageName(), elementType);

        if (embeddedElement != null)
        {
            this.embeddedElement = new Boolean(embeddedElement);
        }
 
        // "dependent-element"
        if (dependentElement != null)
        {
            if (dependentElement.equalsIgnoreCase("true"))
            {
                this.dependentElement = Boolean.TRUE;
            }
            else if (dependentElement.equalsIgnoreCase("false"))
            {
                this.dependentElement = Boolean.FALSE;
            }
        }
    }

    /**
     * Method to populate any defaults, and check the validity of the MetaData.
     * @param loader ClassLoader to use for any loading operations
     */
    public void populate(ClassLoader loader)
    {
        // Check the field type and see if it is castable to a Collection
        Class field_type=getFieldMetaData().getType();
        if (!java.util.Collection.class.isAssignableFrom(field_type))
        {
            throw new InvalidMetaDataException(LOCALISER,
                    "MetaData.Collection.IsNotCollectionError",
                    getFieldName(),getFieldMetaData().getClassName(false));
        }

        // "element-type"
        if (elementType == null)
        {
            throw new InvalidMetaDataException(LOCALISER,
                    "MetaData.Collection.ElementTypeNotSpecifiedError",
                    getFieldName(),getFieldMetaData().getClassName(false));
        }

        // "embedded-element"
        if (embeddedElement == null)
        {
            // Load the element type
            Class element_type_class=MetaDataUtils.resolvedClassForName(elementType,
                    getFieldMetaData().getClassMetaData().getPackageName(),loader); 
            if (element_type_class == null)
            {
                throw new InvalidMetaDataException(LOCALISER,
                        "MetaData.Collection.ElementTypeNotFoundError",
                        getFieldName(),getFieldMetaData().getClassName(false),
                        elementType);
            }
            else
            {
                if (!element_type_class.getName().equals(elementType))
                {
                    // Update to the fully-qualified name
                    JPOXLogger.METADATA.warn(LOCALISER.msg("MetaData.Collection.ElementTypeResolved",getFieldName(),getFieldMetaData().getClassName(false),elementType,element_type_class.getName())); 
                    elementType = element_type_class.getName();
                }
            }

            // Assign default for "embedded-element" based on 18.13.1 of JDO 2 spec
            if (PersistenceCapable.class.isAssignableFrom(element_type_class) ||
                Object.class.isAssignableFrom(element_type_class) ||
                element_type_class.isInterface())
            {
                embeddedElement = Boolean.FALSE;
            }
            else
            {
                embeddedElement = Boolean.TRUE;
            }
        }

        setPopulated();
    }

    // ------------------------------ Accessors --------------------------------
 
    /**
     * Accessor for the element-type tag value
     * @return element-type tag value
     */
    public String getElementType()
    {
        return elementType;
    }

    /**
     * Accessor for the embedded-element tag value
     * @return embedded-element tag value
     */
    public boolean isEmbeddedElement()
    {
        if (embeddedElement == null)
        {
            return false;
        }
        else
        {
            return embeddedElement.booleanValue();
        }
    }

    /**
     * Accessor for The dependent-element attribute indicates that the
     * collections element contains a reference that is to be deleted if the
     * referring instance is deleted.
     * 
     * @return dependent-element tag value
     */
    public boolean isDependentElement()
    {
        if (dependentElement == null)
        {
            return false;
        }
        else
        {
            return dependentElement.booleanValue();
        }
    }
    
    // ------------------------------- Mutators --------------------------------
 
    // ------------------------------- Utilities -------------------------------

    /**
     * Accessor for all ClassMetaData referenced by this array.
     * @param orderedCMDs List of ordered ClassMetaData objects (added to).
     * @param referencedCMDs Set of all ClassMetaData objects (added to).
     * @param dba_vendor_id Vendor ID of the DBA. Used for view addition.
     * @param clr the ClassLoaderResolver 
     **/
    void getReferencedClassMetaData(final List orderedCMDs,
                                    final Set referencedCMDs,
                                    final String dba_vendor_id,
                                    final ClassLoaderResolver clr)
    { 
        ClassMetaData element_cmd=getMetaDataManager().getMetaDataForClass(elementType,clr);
        if (element_cmd != null)
        {
            element_cmd.getReferencedClassMetaData(orderedCMDs,referencedCMDs,dba_vendor_id,clr);
        }
    }

    /**
     * Returns a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        return toString("","");
    }
 
    /**
     * Returns a string representation of the object.
     * @param prefix prefix string
     * @param indent indent string
     * @return a string representation of the object.
     */
    public String toString(String prefix,String indent)
    {
        StringBuffer sb = new StringBuffer();
        if (embeddedElement != null)
        {
            sb.append(prefix).append("<collection element-type=\"").append(elementType).append("\" embedded-element=\"").append(embeddedElement).append("\">\n");
        }
        else
        {
            sb.append(prefix).append("<collection element-type=\"").append(elementType).append("\">\n");
        }

        // Add extensions
        sb.append(super.toString(prefix + indent,indent)); 
 
        sb.append(prefix).append("</collection>\n");
        return sb.toString();
    }
}
