/**********************************************************************
Copyright (c) 11-Jun-2004 Andy Jefferson and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
    ...
**********************************************************************/
package org.jpox.metadata;

/**
 * Representation of the Meta-Data for a column mapping of a field.
 *
 * <PRE>
 * &lt;!ELEMENT column (extension*)?&gt;
 * &lt;!ATTLIST column name CDATA #IMPLIED&gt;
 * &lt;!ATTLIST column target CDATA #IMPLIED&gt;
 * &lt;!ATTLIST column target-field CDATA #IMPLIED&gt;
 * &lt;!ATTLIST column jdbc-type CDATA #IMPLIED&gt;
 * &lt;!ATTLIST column sql-type CDATA #IMPLIED&gt;
 * &lt;!ATTLIST column length CDATA #IMPLIED&gt;
 * &lt;!ATTLIST column scale CDATA #IMPLIED&gt;
 * &lt;!ATTLIST column allows-null CDATA #IMPLIED&gt;
 * &lt;!ATTLIST column default-value CDATA #IMPLIED&gt;
 * </PRE>
 *
 * @since 1.1
 * @version $Revision: 1.4 $
 */
public class ColumnMetaData extends ExtendableMetaData
{
    /**
     * column name value.
     */
    protected String name;

    /**
     * target tag value.
     */
    protected String target;

    /**
     * target-field tag value.
     */
    protected String targetField;

    /**
     * jdbc-type tag value.
     */
    protected String jdbcType;

    /**
     * sql-type tag value.
     */
    protected String sqlType;

    /**
     * length tag value.
     */
    protected Integer length;

    /**
     * scale tag value.
     */
    protected Integer scale;

    /**
     * precision tag value.
     */
    protected Integer precision;
    
    /**
     * allows-null tag value.
     */
    protected Boolean allowsNull;

    /**
     * column default value (when constructing the table with this column)
     */
    protected String defaultValue;

    /**
     * Convenience constructor specifying just the column name, and the parent metadata component.
     * Assigns nulls to other parameters.
     * @param parent Parent MetaData component
     * @param name Name of the column
     */
    public ColumnMetaData(final ExtendableMetaData parent, final String name)
    {
        this(parent, name, null, null, null, null, null, null, null, null, null);
    }

    /**
     * Constructor.
     * @param parent parent MetaData instance
     * @param name field name 
     * @param target target
     * @param targetField target field
     * @param jdbcType JDBC Type to use
     * @param sqlType SQL Type to use
     * @param length length of field
     * @param scale scale of field
     * @param precision precision of field
     * @param allowsNull Whether nulls are allowed
     * @param defaultValue The default value for the column
     */
    public ColumnMetaData(ExtendableMetaData parent,
                          final String name,
                          final String target,
                          final String targetField,
                          final String jdbcType,
                          final String sqlType,
                          final String length,
                          final String scale,
                          final String precision,
                          final String allowsNull,
                          final String defaultValue)
    {
        super(parent);

        this.name = name;

        this.target = target;
        this.targetField = targetField;
        this.jdbcType = jdbcType;
        this.sqlType = sqlType;

        if (length != null && length.length() > 0)
        {
            this.length = Integer.valueOf(length);
        }

        if (scale != null && scale.length() > 0)
        {
            this.scale = Integer.valueOf(scale);
        }

        if (precision != null && precision.length() > 0)
        {
            this.precision = Integer.valueOf(precision);
        }

        if (allowsNull != null)
        {
            if (allowsNull.equalsIgnoreCase("true"))
            {
                this.allowsNull = Boolean.TRUE;
            }
            else if (allowsNull.equalsIgnoreCase("false"))
            {
                this.allowsNull = Boolean.FALSE;
            }
        }
        
        this.defaultValue = defaultValue;
    }

    /**
     * Accessor for the name
     * @return name
     */
    public String getName()
    {
        return name;
    }

    /**
     * Accessor for the target tag value
     * @return target tag value
     */
    public String getTarget()
    {
        return target;
    }

    /**
     * Accessor for the target-field tag value
     * @return target-field tag value
     */
    public String getTargetField()
    {
        return targetField;
    }

    /**
     * Accessor for the jdbc-type tag value
     * @return jdbc-type tag value
     */
    public String getJdbcType()
    {
        return jdbcType;
    }

    /**
     * Accessor for the sql-type tag value
     * @return sql-type tag value
     */
    public String getSqlType()
    {
        return sqlType;
    }

    /**
     * Accessor for the length tag value
     * @return length tag value
     */
    public Integer getLength()
    {
        return length;
    }

    /**
     * Accessor for the scale tag value
     * @return scale tag value
     */
    public Integer getScale()
    {
        return scale;
    }

    /**
     * Accessor for the precision tag value
     * @return precision tag value
     */
    public Integer getPrecision()
    {
        return precision;
    }
 
    /**
     * Accessor for whether the nulls allowed flag has been set.
     * @return Whether it was set.
     */
    public boolean isAllowsNullSet()
    {
        return (allowsNull != null);
    }

    /**
     * Accessor for the nulls-allowed tag value
     * @return nulls-allowed tag value
     */
    public boolean isAllowsNull()
    {
        if (allowsNull == null)
        {
            return false;
        }
        else
        {
            return allowsNull.booleanValue();
        }
    }

    /**
     * Accessor for the default value
     * @return default value
     */
    public String getDefaultValue()
    {
        return defaultValue;
    }

    // ------------------------------- Mutators --------------------------------
 
    /**
     * Mutator for the precision. 
     * @param precision The precision to set.
     */
    public final void setPrecision(Integer precision)
    {
        this.precision = precision;
    }

    /**
     * Mutator for the precision.
     * @param precision The precision to set.
     */
    public final void setPrecision(int precision)
    {
        this.precision = new Integer(precision);
    }

    /**
     * Mutator for the scale 
     * @param scale The scale to set.
     */
    public final void setScale(Integer scale)
    {
        this.scale = scale;
    }

    /**
     * Mutator for the scale
     * @param scale The scale to set.
     */
    public final void setScale(int scale)
    {
        this.scale = new Integer(scale);
    }

    /**
     * Mutator for the JDBC type
     * @param jdbcType The jdbcType to set.
     */
    public final void setJdbcType(String jdbcType)
    {
        this.jdbcType = jdbcType;
    }

    /**
     * Mutator for the length 
     * @param length The length to set.
     */
    public final void setLength(Integer length)
    {
        this.length = length;
    }

    /**
     * Mutator for the length 
     * @param length The length to set.
     */
    public final void setLength(int length)
    {
        this.length = new Integer(length);
    }    

    /**
     * Mutator for the name 
     * @param name The name to set.
     */
    public final void setName(String name)
    {
        this.name = name;
    }

    /**
     * Mutator for whether nulls are allowed.
     * @param allowsNull The allowsNull to set.
     */
    public final void setAllowsNull(Boolean allowsNull)
    {
        this.allowsNull = allowsNull;
    }

    /**
     * Mutator for the SQL type
     * @param sqlType The sqlType to set.
     */
    public final void setSqlType(String sqlType)
    {
        this.sqlType = sqlType;
    }

    // ------------------------------- Utilities -------------------------------
 
    /**
     * Accessor for a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        return toString("","");
    }

    /**
     * Returns a string representation of the object using a prefix
     * @param prefix prefix string
     * @param indent indent string
     * @return a string representation of the object.
     */
    public String toString(String prefix,String indent)
    {
        boolean first_param=true;

        StringBuffer sb = new StringBuffer();
        sb.append(prefix).append("<column");
        if (name != null)
        {
            if (!first_param)
            {
                sb.append("\n").append(prefix).append("        name=\"" + name + "\"");
            }
            else
            {
                sb.append(" name=\"" + name + "\"");
            }
            first_param = false;
        }
        if (target != null)
        {
            if (!first_param)
            {
                sb.append("\n").append(prefix).append("        target=\"" + target + "\"");
            }
            else
            {
                sb.append(" target=\"" + target + "\"");
            }
            first_param = false;
        }
        if (targetField != null)
        {
            if (!first_param)
            {
                sb.append("\n").append(prefix).append("        target-field=\"" + targetField + "\"");
            }
            else
            {
                sb.append(" target-field=\"" + targetField + "\"");
            }
            first_param = false;
        }
        if (jdbcType != null)
        {
            if (!first_param)
            {
                sb.append("\n").append(prefix).append("        jdbc-type=\"" + jdbcType + "\"");
            }
            else
            {
                sb.append(" jdbc-type=\"" + jdbcType + "\"");
            }
            first_param = false;
        }
        if (sqlType != null)
        {
            if (!first_param)
            {
                sb.append("\n").append(prefix).append("        sql-type=\"" + sqlType + "\"");
            }
            else
            {
                sb.append(" sql-type=\"" + sqlType + "\"");
            }
            first_param = false;
        }
        if (allowsNull != null)
        {
            if (!first_param)
            {
                sb.append("\n").append(prefix).append("        allows-null=\"" + allowsNull + "\"");
            }
            else
            {
                sb.append(" allowsNull=\"" + allowsNull + "\"");
            }
            first_param = false;
        }
        if (length != null)
        {
            if (!first_param)
            {
                sb.append("\n").append(prefix).append("        length=\"" + length + "\"");
            }
            else
            {
                sb.append(" length=\"" + length + "\"");
            }
            first_param = false;
        }
        if (scale != null)
        {
            if (!first_param)
            {
                sb.append("\n").append(prefix).append("        scale=\"" + scale + "\"");
            }
            else
            {
                sb.append(" scale=\"" + scale + "\"");
            }
            first_param = false;
        }
        if (precision != null)
        {
            if (!first_param)
            {
                sb.append("\n").append(prefix).append("        precision=\"" + precision + "\"");
            }
            else
            {
                sb.append(" precision=\"" + precision + "\"");
            }
            first_param = false;
        }
        if (defaultValue != null)
        {
            if (!first_param)
            {
                sb.append("\n").append(prefix).append("        default-value=\"" + defaultValue + "\"");
            }
            else
            {
                sb.append(" default-value=\"" + defaultValue + "\"");
            }
            first_param = false;
        }
        sb.append(">\n");

        // Add extensions
        sb.append(super.toString(prefix + indent,indent));

        sb.append(prefix).append("</column>\n");

        return sb.toString();
    }
}