/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2004 Andy Jefferson - added initialise() method
    ...
**********************************************************************/
package org.jpox.metadata;

/**
 * Representation of a discriminator in an inheritance strategy.
 * 
 * <H3>MetaData Element</H3> 
 * The MetaData element represented here is as follows 
 * <PRE>
 * &lt;!ELEMENT discriminator (extension*, column*, index?, extension*)&gt;
 * &lt;!ATTLIST discriminator column CDATA #IMPLIED&gt;
 * &lt;!ATTLIST discriminator value CDATA #IMPLIED&gt;
 * &lt;!ATTLIST discriminator strategy CDATA #IMPLIED&gt;
 * &lt;!ATTLIST discriminator indexed (true|false|unique) #IMPLIED&gt;
 * </PRE>
 *
 * @since 1.1
 * @version $Revision: 1.5 $
 */
public class DiscriminatorMetaData extends ExtendableMetaData
{
    /**
     * strategy tag value.
     */
    protected DiscriminatorStrategy strategy=null;

    /**
     * Column name of discriminator
     */
    protected String columnName=null;

    /**
     * Value for discriminator column
     */
    protected String value=null;

    /**
     * Discriminator column
     */
    protected ColumnMetaData columnMetaData=null;

    /**
     * IndexMetaData.
     */
    protected IndexMetaData indexMetaData;

    /**
     * The indexing value
     */
    protected IndexedValue indexed=null;

    /**
     * Constructor.
     * @param parent parent InheritanceMetaData instance
     * @param columnName Name of the column
     * @param value Value for discriminator column
     * @param strategy The strategy
     */
    public DiscriminatorMetaData(InheritanceMetaData parent,
            					 final String columnName,
            					 final String value,
                               	 final String strategy,
                               	 final String indexed)
    {
        super(parent);

        this.columnName = columnName;
        this.value = value;
        this.strategy = DiscriminatorStrategy.getDiscriminatorStrategy(strategy);
        this.indexed = IndexedValue.getIndexedValue(indexed);
    }

    /**
     * Initialisation method. This should be called AFTER using the populate
     * method if you are going to use populate. It creates the internal
     * convenience arrays etc needed for normal operation.
     */
    public void initialise()
    {
        // Interpret the "indexed" value to create our IndexMetaData where it wasn't specified that way
        if (indexMetaData == null && columnMetaData != null && indexed != null && indexed != IndexedValue.FALSE)
        {
            indexMetaData = new IndexMetaData(null, null, (indexed == IndexedValue.UNIQUE) ? "true" : "false");
            // Note we only support a single column discriminator here
            indexMetaData.addColumn(columnMetaData);
        }
        if (indexMetaData != null)
        {
            indexMetaData.initialise();
        }

        if (columnMetaData == null)
        {
            if (columnName == null)
            {
                // No column name specified, so use "DISCRIMINATOR"
                columnMetaData = new ColumnMetaData(this, "DISCRIMINATOR");
            }
            else
            {
                columnMetaData = new ColumnMetaData(this, columnName);
            }
        }

        setInitialised();
    }

    // --------------------------- Accessors/Mutators ---------------------------

    /**
     * Accessor for column MetaData.
     * @return Returns the column MetaData.
     */
    public ColumnMetaData getColumnMetaData()
    {
        return columnMetaData;
    }

    /**
     * Mutator for column MetaData.
     * @param columnMetaData The column MetaData to set.
     */
    public void setColumnMetaData(ColumnMetaData columnMetaData)
    {
        this.columnMetaData = columnMetaData;
    }

    /**
     * Accessor for indexMetaData
     * @return Returns the indexMetaData.
     */
    public final IndexMetaData getIndexMetaData()
    {
        return indexMetaData;
    }

    /**
     * Mutator for the index MetaData 
     * @param indexMetaData The indexMetaData to set.
     */
    public final void setIndexMetaData(IndexMetaData indexMetaData)
    {
        this.indexMetaData = indexMetaData;
    }

    /**
     * Accessor for value.
     * @return Returns the value.
     */
    public String getValue()
    {
        return value;
    }

    /**
     * Accessor for columnName.
     * @return Returns the columnName.
     */
    public String getColumnName()
    {
        // Return the column variant if specified, otherwise the name field
        if (columnMetaData != null && columnMetaData.getName() != null)
        {
            return columnMetaData.getName();
        }
        return columnName;
    }

    /**
     * Accessor for strategy.
     * @return Returns the strategy.
     */
    public final DiscriminatorStrategy getStrategy()
    {
        return strategy;
    }

    /**
     * Accessor for indexed value.
     * @return Returns the indexed value.
     */
    public final IndexedValue getIndexedValue()
    {
        return indexed;
    }

    /**
     * Mutator for columnName.
     * @param columnName The columnName to set.
     */
    public void setColumnName(String columnName)
    {
        this.columnName = columnName;
    }

    //  ----------------------------- Utilities ---------------------------------
    
    /**
     * Accessor for a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        return toString("","");
    }

    /**
     * Returns a string representation of the object using a prefix
     * @param prefix prefix string
     * @param indent indent string
     * @return a string representation of the object.
     */
    public String toString(String prefix,String indent)
    {
        StringBuffer sb = new StringBuffer();
        sb.append(prefix).append("<discriminator" + 
            strategy != null ? (" strategy=\"" + strategy.toString() + "\"") : "" +
            columnName != null ? (" column=\"" + columnName + "\"") : "" +
            value != null ? (" value=\"" + value + "\"") : "" +
            indexed != null ? (" indexed=\"" + indexed.toString() + "\"") : "" +
            ">\n");

        // Column MetaData
        if (columnMetaData != null)
        {
            sb.append(columnMetaData.toString(prefix + indent,indent));
        }

        // Add index
        if (indexMetaData != null)
        {
            sb.append(indexMetaData.toString(prefix + indent,indent));
        }

        // Add extensions
        sb.append(super.toString(prefix + indent,indent));

        sb.append(prefix).append("</discriminator>\n");

        return sb.toString();
    }
}