/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2004 Andy Jefferson - toString(), MetaData, javadocs
2004 Andy Jefferson - nullIndicatorColumn/Value, ownerField
    ...
**********************************************************************/
package org.jpox.metadata;

import java.lang.reflect.Field;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.jpox.util.JPOXLogger;

/**
 * This element specifies the mapping for an embedded type. It contains multiple
 * field elements, one for each field in the type.
 * 
 * The null-indicator-column optionally identifies the name of the column used
 * to indicate whether the embedded instance is null. By default, if the value
 * of this column is null, then the embedded instance is null. This column might
 * be mapped to a field of the embedded instance but might be a synthetic column
 * for the sole purpose of indicating a null reference.
 * 
 * The null-indicator-value specifies the value to indicate that the embedded
 * instance is null. This is only used for non-nullable columns.
 * 
 * If null-indicator-column is omitted, then the embedded instance is assumed
 * always to exist.
 *
 * <H3>MetaData Element</H3>
 * The MetaData element represented here is as follows
 * <PRE>
 * &lt;!ELEMENT embedded (extension*, field*, extension*)&gt;
 * <!ATTLIST embedded owner-field CDATA>
 * <!ATTLIST embedded null-indicator-column CDATA #IMPLIED>
 * <!ATTLIST embedded null-indicator-value CDATA #IMPLIED>
 * </PRE>
 *
 * @since 1.1
 * @version $Revision: 1.8 $
 */
public class EmbeddedMetaData extends ExtendableMetaData
{
    /** Fields of the embedded object. */
    protected final List fields = new ArrayList();

    /** Name of the owner field in the embedded object. */
    protected final String ownerField;

    /** Name of a column used for determining if the embedded object is null */
    protected final String nullIndicatorColumn;

    /** Value in the null column indicating that the embedded object is null */
    protected final String nullIndicatorValue;

    // -------------------------------------------------------------------------
    // Fields below here are not represented in the output MetaData. They are
    // for use internally in the operation of the JDO system. The majority are
    // for convenience to save iterating through the fields since the fields
    // are fixed once initialised.

    protected FieldMetaData fieldMetaData[];

    /**
     * Constructor.
     * @param parent Owning metadata
     * @param ownerField The owner field
     * @param nullColumn Column to use for determining null embedded object
     * @param nullValue Value of column to use for determining null embedded object
     */
    public EmbeddedMetaData(ExtendableMetaData parent,
                            String ownerField,
                            String nullColumn,
                            String nullValue)
    {
        super(parent);
        this.ownerField = ownerField;
        nullIndicatorColumn = nullColumn;
        nullIndicatorValue = nullValue;
    }

    public void populate(ClassLoader loader)
    {
        // Find the class that the embedded fields apply to
        FieldMetaData fmd = (FieldMetaData)getParent();
        ClassMetaData cmd = fmd.getClassMetaData();
        ClassMetaData embCmd = cmd.getMetaDataManager().getMetaDataForClass(fmd.getType());
        if (embCmd == null)
        {
            // Check for embedded collection or map
            if (fmd.getCollection() != null)
            {
                try
                {
                    Class cls = loader.loadClass(fmd.getCollection().getElementType());
                    embCmd = cmd.getMetaDataManager().getMetaDataForClass(cls);
                }
                catch (ClassNotFoundException cnfe)
                {
                    // Should be caught by populating the Collection
                }
            }
            else if (fmd.getMap() != null)
            {
                try
                {
                    Class cls = loader.loadClass(fmd.getMap().getValueType());
                    embCmd = cmd.getMetaDataManager().getMetaDataForClass(cls);
                }
                catch (ClassNotFoundException cnfe)
                {
                    // Should be caught by populating the Map
                }
            }
        }
        if (embCmd == null)
        {
            JPOXLogger.METADATA.error(LOCALISER.msg("MetaData.Embedded.UnsupportedType", fmd.getFullFieldName(), fmd.getTypeName()));
            throw new InvalidMetaDataException(LOCALISER,
                "MetaData.Embedded.UnsupportedType", fmd.getFullFieldName(), fmd.getTypeName());
        }

        // Check that the specified fields exist in the Embedded class
        Class embCls = null;
        try
        {
            embCls = loader.loadClass(embCmd.getFullClassName());
        }
        catch (ClassNotFoundException cnfe)
        {
            // Do nothing
        }

        Iterator fieldsIter = fields.iterator();
        while (fieldsIter.hasNext())
        {
            FieldMetaData fieldFmd = (FieldMetaData)fieldsIter.next();
            Field cls_field = null;
            try
            {
                cls_field = embCls.getDeclaredField(fieldFmd.getName());
            }
            catch (Exception e)
            {
                // MetaData field doesn't exist in the class!
                throw new InvalidMetaDataException(LOCALISER,
                    "MetaData.Class.DeclaredFieldNotFoundError",
                    embCls.getName(), fieldFmd.getFullFieldName());
            }
            fieldFmd.populate(loader, cls_field);
        }
    }

    /**
     * Method to initialise the object, creating all internal convenience
     * arrays.
     */
    public void initialise()
    {
        fieldMetaData = new FieldMetaData[fields.size()];
        for (int i=0; i<fieldMetaData.length; i++)
        {
            fieldMetaData[i] = (FieldMetaData) fields.get(i);
            fieldMetaData[i].initialise();
        }

        setInitialised();
    }

    // ----------------------------- Accessors ---------------------------------

    /**
     * Accessor for fieldMetaData
     * @return Returns the fieldMetaData.
     */
    public final FieldMetaData[] getFieldMetaData()
    {
        return fieldMetaData;
    }

    /**
     * Accessor for the owner field to contain the FK back to the owner
     * @return The owner field
     */
    public final String getOwnerField()
    {
        return ownerField;
    }

    /**
     * Accessor for the column to check for null values.
     * @return The column name
     */
    public final String getNullIndicatorColumn()
    {
        return nullIndicatorColumn;
    }

    /**
     * Accessor for the value of the null indicator column when null
     * @return The value
     */
    public final String getNullIndicatorValue()
    {
        return nullIndicatorValue;
    }

    // ----------------------------- Mutators ----------------------------------

    /**
     * Method to add a field to the embedded definition.
     * Rejects the addition of duplicate named fields.
     * @param fmd Meta-Data for the field.
     */
    public void addField(FieldMetaData fmd)
    {
        if (fmd == null)
        {
            return;
        }

        if (isInitialised())
        {
            throw new RuntimeException(LOCALISER.msg("MetaData.Field.AlreadyInitialisedError", 
                fmd.getName(), fmd.getClassMetaData().getFullClassName()));
        }
        Iterator iter=fields.iterator();
        while (iter.hasNext())
        {
            FieldMetaData md = (FieldMetaData)iter.next();
            if (fmd.getName().equals(md.getName()))
            {
                throw new RuntimeException(LOCALISER.msg("MetaData.Field.DuplicateField", 
                    fmd.getName(), fmd.getClassMetaData().getFullClassName()));
            }
        }
        fields.add(fmd);
    }

    // ------------------------------- Utilities -------------------------------

    /**
     * Accessor for a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        return toString("","");
    }

    /**
     * Returns a string representation of the object using a prefix
     * This can be used as part of a facility to output a MetaData file. 
     * @param prefix prefix string
     * @param indent indent string
     * @return a string representation of the object.
     */
    public String toString(String prefix,String indent)
    {
        // Field needs outputting so generate metadata
        StringBuffer sb = new StringBuffer();
        sb.append(prefix).append("<embedded");
        if (ownerField != null)
        {
            sb.append(" owner-field=\"" + ownerField + "\"");
        }
        if (nullIndicatorColumn != null)
        {
            sb.append(" null-indicator-column=\"" + nullIndicatorColumn + "\"");
        }
        if (nullIndicatorValue != null)
        {
            sb.append(" null-indicator-value=\"" + nullIndicatorValue + "\"");
        }
        sb.append(">\n");

        // Add fields
        for (int i=0; i<fields.size(); i++)
        {
            FieldMetaData f = (FieldMetaData)fields.get(i);
            sb.append(f.toString(prefix + indent,indent));
        }

        // Add extensions
        sb.append(super.toString(prefix + indent,indent));

        sb.append(prefix + "</embedded>\n");
        return sb.toString();
    }
}