/**********************************************************************
Copyright (c) 04-Jun-2004 Andy Jefferson and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
    Kikuchi Kousuke - org.jpox.enhancer.conf.HasExtension
    ...
**********************************************************************/
package org.jpox.metadata;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import org.jpox.util.StringUtils;

/**
 * Base class of any extendable meta-data. In JDO typically all types of
 * meta-data are extendable so extend this.
 *
 * @since 1.1
 * @version $Revision: 1.1 $
 */
public class ExtendableMetaData extends MetaData
{
    private static final String JPOX_VENDOR_NAME="jpox";

    /**
     * List of extensions.
     */
    protected List extensions = new ArrayList();

    /**
     * default constructor
     */
    public ExtendableMetaData(MetaData parent)
    {
        super(parent);
    }

    // -------------------------------- Accessors ------------------------------
 
    /**
     * Accessor for the number of extensions.
     * @return extension length
     */
    public int getNoOfExtensions()
    {
        return extensions.size();
    }

    /**
     * Accessor for the extensions.
     * @return The extensions.
     **/
    public ExtensionMetaData[] getExtensions()
    {
        if (extensions.size() == 0)
        {
            return null;
        }
        ExtensionMetaData result[] = new ExtensionMetaData[extensions.size()];

        Iterator iter=extensions.iterator();
        int i=0;
        while (iter.hasNext())
        {
            result[i++] = (ExtensionMetaData)iter.next();
        }
        return result;
    }

    /**
     * Accessor for an extension
     * @param i index number
     * @return ExtensionMetaData instance
     */
    public ExtensionMetaData getExtension(int i)
    {
        return (ExtensionMetaData)extensions.get(i);
    }

    /**
     * Accessor for whether an extension exists (for JPOX).
     * @param key The key of the extension
     * @return Whether the extension exists
     */
    public boolean hasExtension(String key)
    {
        if (key == null)
        {
            return false;
        }

        Iterator iter=extensions.iterator();
        while (iter.hasNext())
        {
            ExtensionMetaData ex = (ExtensionMetaData)iter.next();
            if (ex.getKey().equals(key) &&
                ex.getVendorName().equals(JPOX_VENDOR_NAME))
            {
                return true;
            }
        }
        return false;
    }

    /**
     * Accessor for the value of a particular extension (for JPOX).
     * @param key The key of the extension
     * @return The value of the extension (null if not existing)
     */
    public String getValueForExtension(String key)
    {
        if (key == null)
        {
            return null;
        }

        Iterator iter=extensions.iterator();
        while (iter.hasNext())
        {
            ExtensionMetaData ex = (ExtensionMetaData)iter.next();
            if (ex.getKey().equals(key) &&
                ex.getVendorName().equals(JPOX_VENDOR_NAME))
            {
                return ex.getValue();
            }
        }
        return null;
    }

    /**
     * Accessor for the value of a particular extension (for JPOX), but
     * splitting it into separate parts. This is for extension tags that have a
     * value as comma separated.
     * @param key The key of the extension
     * @return The value(s) of the extension (null if not existing)
     */
    public String[] getValuesForExtension(String key)
    {
        if (key == null)
        {
            return null;
        }

        Iterator iter=extensions.iterator();
        while (iter.hasNext())
        {
            ExtensionMetaData ex = (ExtensionMetaData)iter.next();
            if (ex.getKey().equals(key) &&
                ex.getVendorName().equals(JPOX_VENDOR_NAME))
            {
                String[] values=StringUtils.split(ex.getValue(),",");

                // Remove any whitespace around the values
                if (values != null)
                {
                    for (int i=0;i<values.length;i++)
                    {
                        values[i] = values[i].trim();
                    }
                }
                return values;
            }
        }
        return null;
    }

    // -------------------------------- Mutators -------------------------------
 
    /**
     * Append new Extension (for the specified vendor).
     * Will throw an InvalidMetaDataException if the input is invalid.
     * @param vendor vendor-name tag value
     * @param key key tag value
     * @param value value tag value
     */
    public void addExtension(String vendor, String key, String value)
    {
        if (vendor == null)
        {
            throw new InvalidMetaDataException(LOCALISER, "MetaData.Extendable.Invalid", vendor, key, value);
        }
        else if (vendor.equalsIgnoreCase("JPOX") && (key == null || value == null))
        {
            throw new InvalidMetaDataException(LOCALISER,"MetaData.Extendable.Invalid", vendor, key, value);
        }
        extensions.add(new ExtensionMetaData(vendor, key, value));
    }

    /**
     * Append new Extension (for JPOX).
     * Will throw an InvalidMetaDataException if the input is invalid.
     * @param key key tag value
     * @param value value tag value
     */
    public void addExtension(String key, String value)
    {
        if (key == null || value == null)
        {
            throw new InvalidMetaDataException(LOCALISER,"MetaData.Extendable.Invalid",JPOX_VENDOR_NAME,key,value);
        }
        extensions.add(new ExtensionMetaData(JPOX_VENDOR_NAME, key, value));
    }

    /**
     * Method to remove a (JPOX) extension.
     * Restricted to package scope so we protect usage.
     * @param key Key of the tag
     */
    void removeExtension(String key)
    {
        Iterator iter=extensions.iterator();
        while (iter.hasNext())
        {
            ExtensionMetaData ex = (ExtensionMetaData)iter.next();
            if (ex.getVendorName().equals(JPOX_VENDOR_NAME) &&
                ex.getKey().equals(key))
            {
                iter.remove();
                break;
            }
        }
    }

    // -------------------------------- Utilities ------------------------------
 
    /**
     * Returns a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        if (extensions.size() == 0)
        {
            return "";
        }

        StringBuffer sb = new StringBuffer();
        Iterator iter=extensions.iterator();
        while (iter.hasNext())
        {
            ExtensionMetaData ex = (ExtensionMetaData)iter.next();
            sb.append(ex.toString()).append("\n");
        }
        return sb.toString();
    }

    /**
     * Returns a string representation of the object.
     * @param prefix prefix string
     * @param indent indent string
     * @return a string representation of the object.
     */
    public String toString(String prefix,String indent)
    {
        if (extensions.size() == 0)
        {
            return "";
        }

        StringBuffer sb = new StringBuffer();
        Iterator iter=extensions.iterator();
        while (iter.hasNext())
        {
            ExtensionMetaData ex = (ExtensionMetaData)iter.next();
            sb.append(prefix).append(ex.toString()).append("\n");
        }
        return sb.toString();
    }
}
