/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
2004 Andy Jefferson - changed to extend MetaData
2004 Andy Jefferson - added toString()
    ...
**********************************************************************/
package org.jpox.metadata;

import java.util.ArrayList;
import java.util.List;

/**
 * A fetch group defines a particular loaded state for an object graph. It
 * specifies fields to be loaded for all of the instances in the graph.
 * 
 * <H3>MetaData Element</H3>
 * The MetaData Element represented here is as follows
 * <PRE>
 * &lt;!ELEMENT fetch-group (fetch-group|field)*&gt;
 * &lt;!ATTLIST fetch-group name CDATA #REQUIRED&gt;
 * &lt;!ATTLIST fetch-group post-load (true|false) #IMPLIED&gt;
 * </PRE>
 * 
 * @since 1.1
 * @version $Revision: 1.2 $
 */
public class FetchGroupMetaData extends MetaData
{
    /**
     * A contained fetch-group element indicates that the named group is to be
     * included in the group being defined. Nested fetch group elements are
     * limited to only the name attribute.
     */
    protected final List fetchGroups = new ArrayList();    

    /**
     * fields declared to be in this fetch group
     */
    protected final List fields = new ArrayList();    

    /**
     * Recursive fetch group references are controlled by the depth attribute. A
     * depth of 0 (the default) will fetch the whole graph of instances
     * reachable from this field.
     */
    int depth = 0;

    /**
     * The post-load attribute on the fetch-group element indicates whether the
     * jdoPost-Load callback will be made when the fetch group is loaded. It
     * defaults to false, for all fetch groups except the default fetch group,
     * on which it defaults to true.
     */
    Boolean postLoad;

    /**
     * The name attribute on a field element contained within a fetch-group
     * element is the name of field in the enclosing class or a dot-separated
     * expression identifying a field reachable from the class by navigating a
     * reference, collection or map. For maps of persistencecapable classes
     * "#key" or "#value" may be appended to the name of the map field to
     * navigate the key or value respectively (e.g. to include a field of the
     * key class or value class in the fetch group).
     * 
     * For collection and arrays of persistence-capable classes, "#element" may
     * be appended to the name of the field to navigate the element. This is
     * optional; if omitted for collections and arrays, #element is assumed.
     */
    final String name;

    // -------------------------------------------------------------------------
    // Fields below here are not represented in the output MetaData. They are
    // for use internally in the operation of the JDO system. The majority are
    // for convenience to save iterating through the fields since the fields
    // are fixed once populated.

    /**
     * A contained fetch-group element indicates that the named group is to be
     * included in the group being defined. Nested fetch group elements are
     * limited to only the name attribute.
     */
    protected FetchGroupMetaData[] fetchGroupMetaData;

    /**
     * fields declared to be in this fetch group
     */
    protected FieldMetaData[] fieldMetaData;
    
    /**
     * Constructor.
     * @param parent The parent MetaData
     */
    public FetchGroupMetaData(
            MetaData parent,
            String depth,
            String postLoad,
            String name)
    {
        super(parent);
        
        // "postLoad"
        if (postLoad != null)
        {
            if (postLoad.equalsIgnoreCase("true"))
            {
                this.postLoad = Boolean.TRUE;
            }
            else if (postLoad.equalsIgnoreCase("false"))
            {
                this.postLoad = Boolean.FALSE;
            }
        }

        // "depth"
        if (depth != null)
        {
            this.depth = Integer.valueOf(depth).intValue();
        }
        
        // "name"
        this.name = name;
    }
    
    /**
     * Initialisation method. This should be called AFTER using the populate
     * method if you are going to use populate. It creates the internal
     * convenience arrays etc needed for normal operation.
     */
    public void initialise()
    {
        // "postLoad"
        if (postLoad == null)
        {
            this.postLoad = Boolean.FALSE;
        }
        
        if (fieldMetaData == null)
        {
            fieldMetaData = new FieldMetaData[0];
        }
        if (fetchGroupMetaData == null)
        {
            fetchGroupMetaData = new FetchGroupMetaData[0];
        }
        for (int i=0; i<fetchGroupMetaData.length; i++)
        {
            fetchGroupMetaData[i].initialise();
        }
        super.initialise();
    }
    
    /**
     * Method to provide the details of the field being represented by this
     * MetaData. This can be used to firstly provide defaults for attributes
     * that aren't specified in the MetaData, and secondly to report any errors
     * with attributes that have been specifed that are inconsistent with the
     * field being represented.
     */
    public void populate()
    {
        if (postLoad == null)
        {
            this.postLoad = Boolean.FALSE;
        }
        super.populate();
    }
 
    /**
     * Accessor for depth
     * @return Returns the depth.
     */
    public final int getDepth()
    {
        return depth;
    }

    /**
     * Accessor for fields
     * @return Returns the fields.
     */
    public final List getFields()
    {
        return fields;
    }

    /**
     * Accessor for name
     * @return Returns the name.
     */
    public final String getName()
    {
        return name;
    }

    /**
     * Accessor for postLoad
     * @return Returns the postLoad.
     */
    public final Boolean getPostLoad()
    {
        return postLoad;
    }
    
    /**
     * Accessor for fetchGroupMetaData
     * @return Returns the fetchGroupMetaData.
     */
    public final FetchGroupMetaData[] getFetchGroupMetaData()
    {
        return fetchGroupMetaData;
    }

    /**
     * Accessor for fieldMetaData
     * @return Returns the fieldMetaData.
     */
    public final FieldMetaData[] getFieldMetaData()
    {
        return fieldMetaData;
    }
    
    // ------------------------------ Mutators ---------------------------------
    
    /**
     * Add a new FetchGroupMetaData
     * @param fgmd the fetch group
     */
    public void addFetchGroup(FetchGroupMetaData fgmd)
    {
        fetchGroups.add(fgmd);
        fetchGroupMetaData = new FetchGroupMetaData[fetchGroups.size()];
        for (int i=0; i<fetchGroupMetaData.length; i++)
        {
            fetchGroupMetaData[i] = (FetchGroupMetaData) fetchGroups.get(i);
        }
    }

    /**
     * Add a new FieldMetaData
     * @param fmd the field
     */
    public void addField(FieldMetaData fmd)
    {
        fields.add(fmd);
        fieldMetaData = new FieldMetaData[fields.size()];
        for (int i=0; i<fieldMetaData.length; i++)
        {
            fieldMetaData[i] = (FieldMetaData) fields.get(i);
        }
    }

    // ----------------------------- Utilities ------------------------------------

    /**
     * Returns a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        return toString("","");
    }

    /**
     * Returns a string representation of the object.
     * This can be used as part of a facility to output a MetaData file. 
     * @param prefix prefix string
     * @param indent indent string
     * @return a string representation of the object.
     */
    public String toString(String prefix, String indent)
    {
        StringBuffer sb = new StringBuffer();
        sb.append(prefix).append("<fetch-group name=\"" + name + "\"\n");

        // Add fetch-groups
        for (int i=0; i<fetchGroups.size(); i++)
        {
            FetchGroupMetaData fgmd = (FetchGroupMetaData)fetchGroups.get(i);
            sb.append(fgmd.toString(prefix + indent,indent));
        }

        // Add fields
        for (int i=0; i<fields.size(); i++)
        {
            FieldMetaData f = (FieldMetaData)fields.get(i);
            sb.append(f.toString(prefix + indent,indent));
        }

        sb.append(prefix + "</fetch-group>\n");
        return sb.toString();
    }
}