/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Kikuchi Kousuke - org.jpox.enhancer.conf.JDOConfigField
    ...
**********************************************************************/
package org.jpox.metadata;

import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.util.ArrayList;
import java.util.List;
import java.util.Set;

import javax.jdo.JDOFatalInternalException;

import org.jpox.ClassLoaderResolver;
import org.jpox.TypeManager;
import org.jpox.util.ClassUtils;
import org.jpox.util.JPOXLogger;
import org.jpox.util.StringUtils;

/**
 * Representation of the Meta-Data for a field of a class.
 *
 * <H3>Lifecycle state</H3>
 * An object of this type has 2 lifecycle states. The first is the raw
 * constructed object which represents pure MetaData (maybe from a MetaData
 * file). The second is a "populated" object which represents MetaData for a
 * Field of a class with the metadata aligned to be appropriate for that Field. 
 *
 * <H3>Containers</H3>
 * Each field can represent a container. The container can be an array, a
 * Collection or a Map. The field type must be of the correct type to represent
 * these.
 *
 * <H3>JDO Management</H3>
 * Each field can be managed by JDO or not. The class provides a method for
 * identifying if a field is managed by JDO (<I>isJdoField()</I>). If a field
 * is managed by JDO, it will have a field "id" (within its class). In a class
 * the field "id" will start at 0 (for the first field, in alphabetical order).
 *
 * <H3>MetaData Element</H3> 
 * The MetaData element represented here is as follows
 * <PRE> 
 * &lt;!ELEMENT field (extension*, (collection|map|array|(column*))?, join?, element?, 
 *      key?, value?, fetch-group*, order?, embedded?, index?, unique?, foreign-key?,
 *      delete-action?, extension*)?&gt;
 * &lt;!ATTLIST field name CDATA #REQUIRED&gt;
 * &lt;!ATTLIST field persistence-modifier (persistent|transactional|none)
 *      #IMPLIED&gt;
 * &lt;!ATTLIST field table CDATA #IMPLIED&gt;
 * &lt;!ATTLIST field null-value (exception|default|none) 'none'&gt;
 * &lt;!ATTLIST field default-fetch-group (true|false) #IMPLIED&gt;
 * &lt;!ATTLIST field embedded (true|false) #IMPLIED&gt;
 * &lt;!ATTLIST field serialized (true|false) #IMPLIED&gt;
 * &lt;!ATTLIST field dependent (true|false) #IMPLIED&gt;
 * &lt;!ATTLIST field indexed (true|false|unique) #IMPLIED&gt;
 * &lt;!ATTLIST field unique (true|false) #IMPLIED&gt;
 * &lt;!ATTLIST field foreign-key CDATA #IMPLIED&gt;
 * &lt;!ATTLIST field load-fetch-group CDATA #IMPLIED&gt;
 * &lt;!ATTLIST field fetch-depth CDATA #IMPLIED&gt;
 * &lt;!ATTLIST field primary-key (true|false) 'false'&gt;
 * &lt;!ATTLIST field mapped-by CDATA #IMPLIED&gt;
 * &lt;!ATTLIST field value-strategy CDATA #IMPLIED&gt;
 * &lt;!ATTLIST field delete-action (restrict|cascade|null|default|none) #IMPLIED&gt;
 * &lt;!ATTLIST field sequence CDATA #IMPLIED&gt;
 * </PRE>
 *
 * @since 1.1 
 * @version $Revision: 1.18 $
 */
public class FieldMetaData extends ExtendableMetaData implements Comparable
{
    /**
     * Meta-Data of any container.
     */
    protected ContainerMetaData container;

    /**
     * EmbeddedMetaData.
     */
    protected EmbeddedMetaData embeddedMetaData;

    /**
     * JoinMetaData.
     */
    protected JoinMetaData joinMetaData;

    /**
     * ElementMetaData.
     */
    protected ElementMetaData elementMetaData;

    /**
     * KeyMetaData.
     */
    protected KeyMetaData keyMetaData;

    /**
     * ValueMetaData.
     */
    protected ValueMetaData valueMetaData;

    /**
     * IndexMetaData.
     */
    protected IndexMetaData indexMetaData;

    /**
     * The indexing value
     */
    protected IndexedValue indexed=null;

    /**
     * UniqueMetaData.
     */
    protected UniqueMetaData uniqueMetaData;

    /**
     * Whether to add a unique constraint
     */
    protected final boolean uniqueConstraint;

    /**
     * OrderMetaData.
     */
    protected OrderMetaData orderMetaData;

    /**
     * ForeignKeyMetaData.
     */
    protected ForeignKeyMetaData foreignKeyMetaData;

    /**
     * Columns ColumnMetaData
     */
    protected final List columns = new ArrayList();
    
    /**
     * default-fetch-group tag value. 
     */
    protected Boolean defaultFetchGroup;

    /**
     * column tag value. 
     */
    protected String column;
    
    /**
     * mapped-by tag value. 
     */
    protected String mappedBy;
    
    /**
     * embedded tag value.
     */
    protected Boolean embedded;

    /**
     * dependent tag value.
     * 
     * The dependent attribute indicates that the field contains a reference
     * that is to be deleted from the datastore if the referring instance in
     * which the field is declared is deleted, or if the referring field is
     * nullified.
     */
    protected Boolean dependent;
    
    /**
     * serialized tag value.
     */
    protected Boolean serialized;

    /**
     * load-fetch-group value
     */
    protected final String loadFetchGroup;

    /**
     * fetch-depth value
     */
    protected int fetchDepth=0;

    /**
     * Field name.
     */
    protected final String name;

    /**
     * null-value tag value.
     * default is NONE
     */
    protected NullValue nullValue=NullValue.NONE;

    /**
     * persistence-modifier tag value.
     */
    protected FieldPersistenceModifier persistenceModifier=FieldPersistenceModifier.DEFAULT;

    /**
     * primary key tag value.
     */
    protected Boolean primaryKey;

    /**
     * Table name for this field.
     */
    protected String table;

    /**
     * The value-strategy attribute specifies the strategy used to generate
     * values for the field. This attribute has the same values and meaning as
     * the strategy attribute in datastoreidentity.
     */
    protected IdentityStrategy valueStrategy;
    
    /**
     * If the value-strategy is sequence, the sequence attribute specifies the
     * name of the sequence to use to automatically generate a value for the
     * field.
     */
    protected String sequence;
    
    /**
     * Field being represented by this MetaData.
     */
    protected Field field;

    /**
     * Id of the field in its class (only for fields managed by JDO).
     * If the value is -1, the field is NOT managed by JDO or the object hasn't
     * been populated.
     */
    protected int fieldId=-1;

    // -------------------------------------------------------------------------
    // Fields below here are not represented in the output MetaData. They are
    // for use internally in the operation of the JDO system. The majority are
    // for convenience to save iterating through the fields since the fields
    // are fixed once populated.

    /**
     * Contains the metadata for column
     */
    protected ColumnMetaData[] columnMetaData;

    /**
     * Convenience constructor taking defaults
     * @param parent Parent component
     * @param name Name of the field
     */
    public FieldMetaData(MetaData parent,
                         final String name)
    {
        this(parent, name, null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, null, null);
    }

    /**
     * Constructor. Saves the MetaData with the specified values. The object is
     * then in an "unpopulated" state. It can become "populated" by calling the
     * <B>populate()</B> method which compares it against the field it is to
     * represent and updates any unset attributes and flags up any errors.
     * @param parent parent MetaData instance
     * @param name field name 
     * @param pk attribute primary-key value
     * @param modifier attribute persistence-modifier value
     * @param defaultFetchGroup attribute default-fetch-group value
     * @param nullValue attribute null-value value 
     * @param embedded attribute embedded value
     * @param serialized attribute serialized value
     * @param dependent attribute dependent value
     * @param mappedBy attribute mapped-by value
     * @param column attribute column value
     * @param table attribute table value
     * @param deleteAction attribute delete-action value
     * @param indexed Whether this is indexed
     * @param unique Apply a unique constraint
     * @param fetchDepth The depth of fetch to use
     * @param loadFetchGroup Name of the additional fetch group to use when loading
     * @param valueStrategy attribute value-strategy value
     * @param sequence attribute sequence value
     */
    public FieldMetaData(MetaData parent,
                         final String name,
                         final String pk,
                         final String modifier,
                         final String defaultFetchGroup,
                         final String nullValue,
                         final String embedded,
                         final String serialized,
                         final String dependent,
                         final String mappedBy,
                         final String column,
                         final String table,
                         final String deleteAction,
                         final String indexed,
                         final String unique,
                         final String fetchDepth,
                         final String loadFetchGroup,
                         final String valueStrategy,
                         final String sequence)
    {
        super(parent);

        this.name = name;

        // "primary-key"
        if (!StringUtils.isWhitespace(pk) && pk.equalsIgnoreCase("true"))
        {
            this.primaryKey = Boolean.TRUE;
            this.defaultFetchGroup = Boolean.TRUE;
        }
        else
        {
            this.primaryKey = Boolean.FALSE;
        }

        this.column = column;

        this.mappedBy = mappedBy;

        // "dependent"
        if (dependent != null)
        {
            if (dependent.equalsIgnoreCase("true"))
            {
                this.dependent = Boolean.TRUE;
            }
            else if (dependent.equalsIgnoreCase("false"))
            {
                this.dependent = Boolean.FALSE;
            }
        }

        // "embedded"
        if (embedded != null)
        {
            if (embedded.equalsIgnoreCase("true"))
            {
                this.embedded = Boolean.TRUE;
            }
            else if (embedded.equalsIgnoreCase("false"))
            {
                this.embedded = Boolean.FALSE;
            }
        }

        // "serialized"
        if (serialized != null)
        {
            if (serialized.equalsIgnoreCase("true"))
            {
                this.serialized = Boolean.TRUE;
            }
            else if (serialized.equalsIgnoreCase("false"))
            {
                this.serialized = Boolean.FALSE;
            }
        }

        // "null-value"
        this.nullValue = NullValue.getNullValue(nullValue);

        // "persistence-modifier"
        if (modifier != null)
        { 
            persistenceModifier = FieldPersistenceModifier.getFieldPersistenceModifier(modifier);
        }

        // "default-fetch-group"
        if (defaultFetchGroup != null)
        {
            if (defaultFetchGroup.equalsIgnoreCase("true"))
            {
                this.defaultFetchGroup = Boolean.TRUE;
            }
            else if (defaultFetchGroup.equalsIgnoreCase("false"))
            {
                this.defaultFetchGroup = Boolean.FALSE;
            }
        }

        // "table"
        this.table = table;

        if (deleteAction != null)
        {
            foreignKeyMetaData = new ForeignKeyMetaData(null, deleteAction, null, null, null, null);
        }

        this.indexed = IndexedValue.getIndexedValue(indexed);

        if (unique != null && unique.equalsIgnoreCase("true"))
        {
            uniqueConstraint = true;
        }
        else
        {
            uniqueConstraint = false;
        }

        // Fetch Group
        this.loadFetchGroup = loadFetchGroup;
        if (fetchDepth != null)
        {
            try
            {
                this.fetchDepth = (new Integer(fetchDepth)).intValue();
            }
            catch (Exception e)
            {
                this.fetchDepth = 0;
            }
        }

        this.valueStrategy = valueStrategy == null ? null : IdentityStrategy.getIdentityStrategy(valueStrategy);
        this.sequence = sequence;
    }

    /**
     * Method to merge in the associated metadata from the ORM definition
     * @param ormFmd The ORM Field definition
     */
    public void mergeORMData(FieldMetaData ormFmd)
    {
        if (ormFmd == null)
        {
            return;
        }

        if (isInitialised() || isPopulated())
        {
            throw new JDOFatalInternalException(LOCALISER.msg("MetaData.Field.AlreadyPopulatedError",name,ormFmd.getClassMetaData().getFullClassName()));
        }

        // Update our O/R mapping details
        primaryKey = ClassUtils.booleanValueOf(ormFmd.isPrimaryKey());
        table = ormFmd.getTable();
        mappedBy = ormFmd.getMappedBy();
        valueStrategy = ormFmd.getValueStrategy();
        sequence = ormFmd.getSequence();

        joinMetaData = ormFmd.getJoinMetaData();
        embeddedMetaData = ormFmd.getEmbeddedMetaData();
        elementMetaData = ormFmd.getElementMetaData();
        keyMetaData = ormFmd.getKeyMetaData();
        valueMetaData = ormFmd.getValueMetaData();
        orderMetaData = ormFmd.getOrderMetaData();
        foreignKeyMetaData = ormFmd.getForeignKeyMetaData();
        indexed = ormFmd.indexed;
        indexMetaData = ormFmd.getIndexMetaData();
        uniqueMetaData = ormFmd.getUniqueMetaData();

        // Add the columns (assumes none existing)
        // TODO Allow for others being specified in the basic file
        columns.clear();
        ColumnMetaData[] ormColumns=ormFmd.getColumnMetaData();
        if (ormColumns != null)
        {
            for (int i = 0; i < ormColumns.length; i++)
            {
                columns.add(ormColumns[i]);
            }
        }
    }

    /**
     * Method to provide the details of the field being represented by this
     * MetaData. This can be used to firstly provide defaults for attributes
     * that aren't specified in the MetaData, and secondly to report any errors
     * with attributes that have been specifed that are inconsistent with the
     * field being represented.
     * @param loader ClassLoader to use for any class loading 
     * @param field The (Reflection) field being represented
     */
    public synchronized void populate(ClassLoader loader,Field field)
    {
        if (isPopulated())
        {
            JPOXLogger.METADATA.error(LOCALISER.msg("MetaData.Field.AlreadyPopulatedError",name,getClassName(false)));
            throw new InvalidMetaDataException(LOCALISER,
                "MetaData.Field.AlreadyPopulatedError",
                name,getClassName(false));
        }

        if (field == null)
        {
            JPOXLogger.METADATA.error(LOCALISER.msg("MetaData.Field.PopulateWithNullFieldError",name,getClassName(false)));
            throw new InvalidMetaDataException(LOCALISER,
                "MetaData.Field.PopulateWithNullError",
                name,getClassName(false));
        }

        // No class loader, so use System
        if (loader == null)
        {
            JPOXLogger.METADATA.warn(LOCALISER.msg("MetaData.Field.UsingSystemClassLoader",name,getClassName(true)));
            loader = ClassLoader.getSystemClassLoader();
        }

        this.field = field;
        Class field_type=field.getType();

        // Update "embedded" based on type
        if (primaryKey == Boolean.FALSE)
        {
            Class element_type=field_type;
            if (element_type.isArray())
            {
                element_type = element_type.getComponentType();
                if (MetaDataUtils.isDefaultEmbeddedType(element_type))
                {
                    embedded = Boolean.TRUE;
                }
            }
            else if (MetaDataUtils.isDefaultEmbeddedType(element_type))
            {
                embedded = Boolean.TRUE;
            }
        }
        if (embedded == null)
        {
            embedded = Boolean.FALSE;
        }

        // Update "persistence-modifier" according to type etc
        if (FieldPersistenceModifier.DEFAULT.equals(persistenceModifier))
        {
            persistenceModifier =
                MetaDataUtils.getDefaultFieldPersistenceModifier(
                        field_type,field.getModifiers(),
                        isFieldTypePersistenceCapable());
        }
        // TODO If this field is NONE in superclass, make it NONE here too

        // Check whether the type is supported by JPOX
        if (!persistenceModifier.equals(FieldPersistenceModifier.NONE) &&
            field_type.isArray())
        {
            if (!TypeManager.getTypeManager().isSupportedArrayType(field_type))
            {
                JPOXLogger.METADATA.warn(LOCALISER.msg("MetaData.Field.InvalidArrayType",name,getClassName(false)));
                // This should be an error, but we want the enhancer to pass
                // this, yet the JPOX runtime fail it.
//              throw new InvalidMetaDataException(LOCALISER,
//                  "MetaData.Field.InvalidArrayType",
//                  name,getClassName(false));
            }
        }

        // Update "default-fetch-group" according to type
        if (defaultFetchGroup == null &&
            persistenceModifier.equals(FieldPersistenceModifier.NONE))
        {
            defaultFetchGroup = Boolean.FALSE;
        }
        else if (defaultFetchGroup == null &&
                persistenceModifier.equals(FieldPersistenceModifier.TRANSACTIONAL))
        {
            defaultFetchGroup = Boolean.FALSE;
        }
        else if (defaultFetchGroup == null)
        {
            defaultFetchGroup = Boolean.FALSE;
            if (!primaryKey.equals(Boolean.TRUE) &&
                MetaDataUtils.isDefaultFetchGroup(field_type))
            {
                defaultFetchGroup = Boolean.TRUE;
            }
        }

        // Field is not specified as "persistent" yet has DFG or primary-key !
        if (persistenceModifier.equals(FieldPersistenceModifier.TRANSACTIONAL) ||
            persistenceModifier.equals(FieldPersistenceModifier.NONE))
        {
            if (defaultFetchGroup == Boolean.TRUE ||
                primaryKey == Boolean.TRUE)
            {
                throw new InvalidMetaDataException(LOCALISER,
                    "MetaData.Field.NotPersistentWithIllegalAttributesError",
                    name,getClassName(false),this.getField().getType().getName(),
                    persistenceModifier.toString());
            }
        }

        // Populate any embedded object
        if (embeddedMetaData != null)
        {
            embeddedMetaData.populate(loader);
        }

        // Populate any attached container
        if (container != null)
        {
            if (container instanceof CollectionMetaData)
            {
                ((CollectionMetaData)container).populate(loader);
            }
            else if (container instanceof MapMetaData)
            {
                ((MapMetaData)container).populate(loader);
            }
            else if (container instanceof ArrayMetaData)
            {
                ((ArrayMetaData)container).populate(loader);
            }
        }

        setPopulated();
    }

    /**
     * Initialisation method. This should be called AFTER using the populate
     * method if you are going to use populate. It creates the internal
     * convenience arrays etc needed for normal operation.
     */
    public synchronized void initialise()
    {
        // Cater for user specifying column name, or columns
        if (columns.size() == 0 && column != null)
        {
            columnMetaData = new ColumnMetaData[1];
            columnMetaData[0] = new ColumnMetaData(this, column);
        }
        else
        {
            columnMetaData = new ColumnMetaData[columns.size()];
            for (int i=0; i<columnMetaData.length; i++)
            {
                columnMetaData[i] = (ColumnMetaData) columns.get(i);
                columnMetaData[i].initialise();
            }
        }

        // Initialise all sub-objects
        if (container != null)
        {
            container.initialise();
        }
        if (embeddedMetaData != null)
        {
            embeddedMetaData.initialise();
        }
        if (joinMetaData != null)
        {
            joinMetaData.initialise();
        }
        if (elementMetaData != null)
        {
            elementMetaData.initialise();
        }
        if (keyMetaData != null)
        {
            keyMetaData.initialise();
        }
        if (valueMetaData != null)
        {
            valueMetaData.initialise();
        }

        // Interpret the "indexed" value to create our IndexMetaData where it wasn't specified that way
        if (indexMetaData == null && columnMetaData != null && indexed != null && indexed != IndexedValue.FALSE)
        {
            indexMetaData = new IndexMetaData(null, null, (indexed == IndexedValue.UNIQUE) ? "true" : "false");
            for (int i=0;i<columnMetaData.length;i++)
            {
                indexMetaData.addColumn(columnMetaData[i]);
            }
        }
        if (indexMetaData != null)
        {
            indexMetaData.initialise();
        }
        if (uniqueMetaData == null && uniqueConstraint)
        {
            uniqueMetaData = new UniqueMetaData(null, column, null);
            for (int i=0;i<columnMetaData.length;i++)
            {
                uniqueMetaData.addColumn(columnMetaData[i]);
            }
        }
        if (uniqueMetaData != null)
        {
            uniqueMetaData.initialise();
        }

        if (foreignKeyMetaData != null)
        {
            foreignKeyMetaData.initialise();
        }
        if (orderMetaData != null)
        {
            orderMetaData.initialise();
        }

        setInitialised();
    }

    /**
     * Utility to return if this field is PersistenceCapable.
     * Here we just check for the presence of a JDO field on the (enhanced)
     * class.
     * @return Whether the field type is persistence capable.
     */
    protected boolean isFieldTypePersistenceCapable()
    {
        Class type=field.getType();
        try
        {
            type.getDeclaredMethod("jdoProvideField",new Class[] {int.class});
            return true;
        }
        catch (NoSuchMethodException nsme)
        {
            return false;
        }
    }

    /**
     * Convenience method to return if this field represents a static field.
     * When the object is not "populated" always returns false; 
     * @return Whether the field is static
     */
    public boolean isStatic()
    {
        if (!isPopulated())
        {
            return false;
        }
        return Modifier.isStatic(field.getModifiers());
    }

    /**
     * Convenience method to return if this field represents a final field.
     * When the object is not "populated" always returns false; 
     * @return Whether the field is final
     */
    public boolean isFinal()
    {
        if (!isPopulated())
        {
            return false;
        }
        return Modifier.isFinal(field.getModifiers());
    }

    /**
     * Convenience method to return if this field represents a transient field.
     * When the object is not "populated" always returns false; 
     * @return Whether the field is transient
     */
    public boolean isTransient()
    {
        if (!isPopulated())
        {
            return false;
        }
        return Modifier.isTransient(field.getModifiers());
    }

    // ------------------------------- Accessors -------------------------------

    /**
     * The value-strategy attribute specifies the strategy used to generate
     * values for the field. This attribute has the same values and meaning as
     * the strategy attribute in datastoreidentity.
     * @return the value strategy
     */
    public IdentityStrategy getValueStrategy()
    {
        return valueStrategy;
    }
    
    /**
     * If the value-strategy is sequence, the sequence attribute specifies the
     * name of the sequence to use to automatically generate a value for the
     * field.
     * @return the sequence name 
     */
    public String getSequence()
    {
        return sequence;
    }
    
    /**
     * Accessor for the (additional) fetch group for loading this field
     * @return Returns the load fetch group
     */
    public String getLoadFetchGroup()
    {
        return loadFetchGroup;
    }

    /**
     * Accessor for the depth of the fetch
     * @return Returns the depth of the fetch
     */
    public int getFetchDepth()
    {
        return fetchDepth;
    }

    /**
     * Convenience accessor for the MetaData of the parent class.
     * @return Returns the MetaData of the parent class.
     */
    public ClassMetaData getClassMetaData()
    {
        if (parent == null)
        {
            return null;
        }
        else if (parent instanceof ClassMetaData)
        {
            return (ClassMetaData)parent;
        }
        else if (parent instanceof EmbeddedMetaData)
        {
            // This is giving the class that it is contained in, but maybe should give the embedded type
            return (ClassMetaData)((EmbeddedMetaData)parent).getParent().getParent();
        }
        return null;
    }

    /**
     * Convenience to return the class name that this a field of.
     * @param fully_qualified Whether the name should be fully qualified.
     * @return Class name
     */
    public String getClassName(boolean fully_qualified)
    {
        if (parent == null)
        {
            return null;
        }
        else if (parent instanceof ClassMetaData)
        {
            ClassMetaData cmd=(ClassMetaData)parent;
            if (fully_qualified)
            {
                return cmd.getFullClassName();
            }
            else
            {
                return cmd.getName();
            }
        }
        else if (parent instanceof EmbeddedMetaData)
        {
            // This returns the type of the embedded field
            FieldMetaData fmd = (FieldMetaData)((EmbeddedMetaData)parent).getParent();
            return fmd.getTypeName();
        }
        return null;
    }

    /**
     * Accessor for orderMetaData
     * @return Returns the orderMetaData.
     */
    public final OrderMetaData getOrderMetaData()
    {
        return orderMetaData;
    }
    
    /**
     * Accessor for the field name 
     * @return field name
     */
    public String getName()
    {
        return name;
    }

    /**
     * Accessor for the full field name. This prepends the class name. 
     * @return full field name.
     */
    public String getFullFieldName()
    {
        return getClassName(true) + "." + name;
    }

    /**
     * Accessor for the null-value tag
     * @return null-value tag
     */
    public NullValue getNullValue()
    {
        return nullValue;
    }

    /**
     * Accessor for the persistence-modifier tag value
     * @return persistence-modifier tag value
     */
    public FieldPersistenceModifier getPersistenceModifier()
    {
        return persistenceModifier;
    }

    /**
     * Accessor for the default-fetch-group tag value
     * @return default-fetch-group tag value
     */
    public boolean isDefaultFetchGroup()
    {
        if (defaultFetchGroup == null)
        {
            return false;
        }
        else
        {
            return defaultFetchGroup.booleanValue();
        }
    }
    
    /**
     * Accessor for the dependent attribute indicates that the field contains a
     * reference that is to be deleted from the datastore if the referring
     * instance in which the field is declared is deleted, or if the referring
     * field is nullified.
     * 
     * @return dependent tag value
     */
    public boolean isDependent()
    {
        if (dependent == null &&
            (this.getCollection() != null || this.getMap() != null || this.getArray() !=null ))
        {
            return true;
        }
        else if (dependent == null)
        {
            return false;
        }
        else
        {
            return dependent.booleanValue();
        }
    }

    /**
     * Accessor for the embedded tag value.
     * This value is a hint only to the implementation so will not mean
     * that the type is certainly embedded.
     * @return embedded tag value
     */
    public boolean isEmbedded()
    {
        if (embedded == null)
        {
            return false;
        }
        else
        {
            return embedded.booleanValue();
        }
    }
    
    /**
     * Accessor for the serialized tag value
     * @return serialized tag value
     */
    public boolean isSerialized()
    {
        if (serialized == null)
        {
            return false;
        }
        else
        {
            return serialized.booleanValue();
        }
    }

    /**
     * Accessor for the primary-key tag value.
     * @return primary-key tag value.
     */
    public boolean isPrimaryKey()
    {
        if (primaryKey == null)
        {
            return false;
        }
        else
        {
            return primaryKey.booleanValue();
        }
    }

    /**
     * Accessor for the table name 
     * @return table name
     */
    public String getTable()
    {
        return table;
    }

    /**
     * Accessor for the field id
     * @return field id
     */
    public int getFieldId()
    {
        return fieldId;
    }

    /**
     * Accessor for the field
     * @return Reflection field
     */
    public Field getField()
    {
        return field;
    }

    /**
     * Accessor for the field id
     * @return field id
     */
    public int getAbsoluteFieldNumber()
    {
        return fieldId + getClassMetaData().getNoOfInheritedManagedFields();
    }
    
    /**
     * Accessor for the field type
     * @return Reflection field type
     */
    public Class getType()
    {
        if (field != null)
        {
            return field.getType();
        }
        return null;
    }

    /**
     * Accessor for the field type name
     * @return Reflection field type name
     */
    public String getTypeName()
    {
        if (field != null)
        {
            return field.getType().getName();
        }
        return null;
    }

    /**
     * Accessor for the container for this field.
     * @return The MetaData of the container for this field.
     **/
    public ContainerMetaData getContainer()
    {
        return container;
    }

    /**
     * Accessor for an array container for this field. Returns null if no array
     * attached.
     * @return The MetaData of the container for this field if an array
     **/
    public ArrayMetaData getArray()
    {
        if (container != null && container instanceof ArrayMetaData)
        {
            return (ArrayMetaData)container;
        }
        return null;
    }

    /**
     * Accessor for a collection container for this field. Returns null if no
     * collection attached.
     * @return The MetaData of the container for this field if a Collection.
     **/
    public CollectionMetaData getCollection()
    {
        if (container != null && container instanceof CollectionMetaData)
        {
            return (CollectionMetaData)container;
        }
        return null;
    }

    /**
     * Accessor for the column 
     * @return Returns the column.
     */
    public final String getColumn()
    {
        return column;
    }

    /**
     * Accessor for a map container for this field. Returns null if no map
     * attached.
     * @return The MetaData of the container for this field if a Map.
     **/
    public MapMetaData getMap()
    {
        if (container != null && container instanceof MapMetaData)
        {
            return (MapMetaData)container;
        }
        return null;
    }

    /**
     * Accessor for mappedBy
     * @return Returns the mappedBy.
     */
    public final String getMappedBy()
    {
        return mappedBy;
    }

    /**
     * Acessor for the columns
     * @return Returns the columnMetaData.
     */
    public final ColumnMetaData[] getColumnMetaData()
    {
        return columnMetaData;
    }

    /**
     * Accessor for elementMetaData
     * @return Returns the elementMetaData.
     */
    public final ElementMetaData getElementMetaData()
    {
        return elementMetaData;
    }

    /**
     * Accessor for keyMetaData
     * @return Returns the keyMetaData.
     */
    public final KeyMetaData getKeyMetaData()
    {
        return keyMetaData;
    }

    /**
     * Accessor for valueMetaData
     * @return Returns the valueMetaData.
     */
    public final ValueMetaData getValueMetaData()
    {
        return valueMetaData;
    }

    /**
     * Accessor for embeddedMetaData
     * @return Returns the embeddedMetaData.
     */
    public final EmbeddedMetaData getEmbeddedMetaData()
    {
        return embeddedMetaData;
    }

    /**
     * Accessor for foreignKeyMetaData
     * @return Returns the foreignKeyMetaData.
     */
    public final ForeignKeyMetaData getForeignKeyMetaData()
    {
        return foreignKeyMetaData;
    }

    /**
     * Accessor for indexMetaData
     * @return Returns the indexMetaData.
     */
    public final IndexMetaData getIndexMetaData()
    {
        return indexMetaData;
    }

    /**
     * Accessor for uniqueMetaData
     * @return Returns the uniqueMetaData.
     */
    public final UniqueMetaData getUniqueMetaData()
    {
        return uniqueMetaData;
    }

    /**
     * Accessor for joinMetaData
     * @return Returns the joinMetaData.
     */
    public final JoinMetaData getJoinMetaData()
    {
        return joinMetaData;
    }

    /**
     * Add a new ColumnMetaData element
     */
    public void addColumn(ColumnMetaData colmd)
    {
        columns.add(colmd);
        columnMetaData = new ColumnMetaData[columns.size()];
        for (int i=0; i<columnMetaData.length; i++)
        {
            columnMetaData[i] = (ColumnMetaData) columns.get(i);
        }
    }

    /**
     * Accessor for whether the field has a container.
     * @return Whether it represents a container.
     */
    public boolean hasContainer()
    {
        return (container != null);
    }

    /**
     * Accessor for whether the field has an array
     * @return return true if has array
     */
    public boolean hasArray()
    {
        if (container == null)
        {
            return false;
        }
        return (container instanceof ArrayMetaData);
    }

    /**
     * Accessor for whether the field has a collection
     * @return return true if has collection
     */
    public boolean hasCollection()
    {
        if (container == null)
        {
            return false;
        }
        return (container instanceof CollectionMetaData);
    }

    /**
     * Accessor for whether the field has a map.
     * @return return true if has map
     */
    public boolean hasMap()
    {
        if (container == null)
        {
            return false;
        }
        return (container instanceof MapMetaData);
    }

    /**
     * Accessor for whether the field is to be managed by JDO.
     * @return Whether it is JDO managed
     */
    public boolean isJdoField()
    {
        if (isPopulated())
        {
            if (isStatic() || isFinal())
            {
                return false;
            }
        }

        if (persistenceModifier == null)
        {
            return false;
        }
        else if (persistenceModifier.equals(FieldPersistenceModifier.NONE))
        {
            return false;
        }
        return true;
    }

    // ------------------------------ Mutators ---------------------------------

    /**
     * The value-strategy attribute specifies the strategy used to generate
     * values for the field. This attribute has the same values and meaning as
     * the strategy attribute in datastoreidentity.
     * @param valueStrategy the value strategy
     */
    public void setValueStrategy(IdentityStrategy valueStrategy)
    {
        this.valueStrategy = valueStrategy;
    }
    
    /**
     * If the value-strategy is sequence, the sequence attribute specifies the
     * name of the sequence to use to automatically generate a value for the
     * field.
     * @param sequence the sequence name 
     */
    public void setSequence(String sequence)
    {
        this.sequence = sequence;
    }

    /**
     * Mutator for dependent attribute.
     * @param dependent Whether it is dependent.
     */
    public void setDependent(boolean dependent)
    {
        this.dependent = new Boolean(dependent);
    }

    /**
     * Mutator for mappedBy
     * @param mappedby The mapped-by field.
     */
    public void setMappedBy(String mappedby)
    {
        mappedBy = mappedby;
    }

    /**
     * Method to set the container for this field (if this field represents a
     * container (collection, map, array).
     * @param conmd The MetaData of the container for this field.
     **/
    public void setContainer(ContainerMetaData conmd)
    {
        container = conmd;
    }

    /**
     * Mutator for the element MetaData 
     * @param elementMetaData The elementMetaData to set.
     */
    public final void setElementMetaData(ElementMetaData elementMetaData)
    {
        this.elementMetaData = elementMetaData;
    }

    /**
     * Mutator for the key MetaData 
     * @param keyMetaData The keyMetaData to set.
     */
    public final void setKeyMetaData(KeyMetaData keyMetaData)
    {
        this.keyMetaData = keyMetaData;
    }

    /**
     * Mutator for the order MetaData 
     * @param orderMetaData The orderMetaData to set.
     */
    public final void setOrderMetaData(OrderMetaData orderMetaData)
    {
        this.orderMetaData = orderMetaData;
    }

    /**
     * Mutator for the value MetaData 
     * @param valueMetaData The valueMetaData to set.
     */
    public final void setValueMetaData(ValueMetaData valueMetaData)
    {
        this.valueMetaData = valueMetaData;
    }

    /**
     * Mutator for the embedded MetaData 
     * @param embeddedMetaData The embeddedMetaData to set.
     */
    public final void setEmbeddedMetaData(EmbeddedMetaData embeddedMetaData)
    {
        this.embeddedMetaData = embeddedMetaData;
    }

    /**
     * Mutator for the foreignKey MetaData 
     * @param foreignKeyMetaData The foreignKeyMetaData to set.
     */
    public final void setForeignKeyMetaData(
            ForeignKeyMetaData foreignKeyMetaData)
    {
        this.foreignKeyMetaData = foreignKeyMetaData;
    }

    /**
     * Mutator for the index MetaData 
     * @param indexMetaData The indexMetaData to set.
     */
    public final void setIndexMetaData(IndexMetaData indexMetaData)
    {
        this.indexMetaData = indexMetaData;
    }

    /**
     * Mutator for the unique MetaData 
     * @param uniqueMetaData The uniqueMetaData to set.
     */
    public final void setUniqueMetaData(UniqueMetaData uniqueMetaData)
    {
        this.uniqueMetaData = uniqueMetaData;
    }

    /**
     * Mutator for the join MetaData 
     * @param joinMetaData The joinMetaData to set.
     */
    public final void setJoinMetaData(JoinMetaData joinMetaData)
    {
        this.joinMetaData = joinMetaData;
    }

    /**
     * Mutator for the field id. Given package access since updated by
     * ClassMetaData typically.
     * @param field_id Id of the field
     */
    void setFieldId(int field_id)
    {
        fieldId = field_id;
    }

    /**
     * Mutator for the table name
     * @param table The table name
     */
    public void setTable(String table)
    {
        this.table = table;
    }

    // ------------------------------ Utilities --------------------------------

    /**
     * Accessor for all ClassMetaData referenced by this Field.
     * <p>
     * Part of the "persistence-by-reachability" concept. 
     * @param orderedCMDs List of ordered ClassMetaData objects (added to).
     * @param referencedCMDs Set of referenced ClassMetaData objects (added to)
     * @param dba_vendor_id Vendor id of the DBA
     * @param clr the ClassLoaderResolver
     **/
    void getReferencedClassMetaData(final List orderedCMDs,
                                    final Set referencedCMDs,
                                    final String dba_vendor_id,
                                    final ClassLoaderResolver clr)
    {
        ClassMetaData type_cmd=getClassMetaData().getMetaDataManager().getMetaDataForClass(field.getType());
        if (type_cmd != null)
        {
            type_cmd.getReferencedClassMetaData(orderedCMDs,referencedCMDs,dba_vendor_id,clr);
        }

        if (container != null)
        {
            if (container instanceof CollectionMetaData)
            {
                ((CollectionMetaData)container).getReferencedClassMetaData(orderedCMDs,referencedCMDs,dba_vendor_id,clr);
            }
            else if (container instanceof MapMetaData)
            {
                ((MapMetaData)container).getReferencedClassMetaData(orderedCMDs,referencedCMDs,dba_vendor_id,clr);
            }
            else if (container instanceof ArrayMetaData)
            {
                ((ArrayMetaData)container).getReferencedClassMetaData(orderedCMDs,referencedCMDs,dba_vendor_id,clr);
            }
        }
    }

    /**
     * Accessor for a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        return toString("","");
    }

    /**
     * Returns a string representation of the object using a prefix
     * This can be used as part of a facility to output a MetaData file. 
     * @param prefix prefix string
     * @param indent indent string
     * @return a string representation of the object.
     */
    public String toString(String prefix,String indent)
    {
        // If this field is static or final, don't bother with MetaData since
        // JDO will ignore it anway.
        if (isStatic() || isFinal())
        {
            return "";
        }

        // Field needs outputting so generate metadata
        StringBuffer sb = new StringBuffer();
        sb.append(prefix).append("<field name=\"" + name + "\"");
        if (persistenceModifier != null &&
            !StringUtils.isWhitespace(persistenceModifier.toString()))
        {
            sb.append("\n").append(prefix).append("       persistence-modifier=\"" + persistenceModifier + "\"");
        }
        if (!StringUtils.isWhitespace(table))
        {
            sb.append("\n").append(prefix).append("       table=\"" + table + "\"");
        }
        if (primaryKey.booleanValue())
        {
            sb.append("\n").append(prefix).append("       primary-key=\"" + primaryKey + "\"");
        }
        sb.append("\n").append(prefix).append("       null-value=\"" + nullValue + "\"");
        if (defaultFetchGroup != null &&
            !StringUtils.isWhitespace(defaultFetchGroup.toString()))
        {
            sb.append("\n").append(prefix).append("       default-fetch-group=\"" + defaultFetchGroup + "\"");
        }
        if (embedded != null &&
            !StringUtils.isWhitespace(embedded.toString()))
        {
            sb.append("\n").append(prefix).append("       embedded=\"" + embedded + "\"");
        }
        if (serialized != null &&
            !StringUtils.isWhitespace(serialized.toString()))
        {
            sb.append("\n").append(prefix).append("       serialized=\"" + serialized + "\"");
        }
        if (dependent != null)
        {
            sb.append("\n").append(prefix).append("       dependent=\"" + dependent + "\"");
        }
        if (mappedBy != null)
        {
            sb.append("\n").append(prefix).append("       mapped-by=\"" + mappedBy + "\"");
        }
        if (loadFetchGroup != null)
        {
            sb.append("\n").append(prefix).append("       load-fetch-group=\"" + loadFetchGroup + "\"");
        }
        if (fetchDepth != 0)
        {
            sb.append("\n").append(prefix).append("       fetch-depth=\"" + fetchDepth + "\"");
        }
        if (valueStrategy != null)
        {
            sb.append("\n").append(prefix).append("       value-strategy=\"" + valueStrategy + "\"");
        }
        if (sequence != null)
        {
            sb.append("\n").append(prefix).append("       sequence=\"" + sequence + "\"");
        }
        sb.append(">\n");

        // Add field containers
        if (container != null)
        {
            if (container instanceof CollectionMetaData)
            {
                CollectionMetaData c = (CollectionMetaData)container;
                sb.append(c.toString(prefix + indent,indent));
            }
            else if (container instanceof ArrayMetaData)
            {
                ArrayMetaData c = (ArrayMetaData)container;
                sb.append(c.toString(prefix + indent,indent));
            }
            else if (container instanceof MapMetaData)
            {
                MapMetaData c = (MapMetaData)container;
                sb.append(c.toString(prefix + indent,indent));
            }
        }

        // Add columns
        for (int i=0; i<columns.size(); i++)
        {
            ColumnMetaData c = (ColumnMetaData)columns.get(i);
            sb.append(c.toString(prefix + indent,indent));
        }

        // Add join
        if (joinMetaData != null)
        {
            sb.append(joinMetaData.toString(prefix + indent,indent));
        }

        // Add element
        if (elementMetaData != null)
        {
            sb.append(elementMetaData.toString(prefix + indent,indent));
        }

        // Add key
        if (keyMetaData != null)
        {
            sb.append(keyMetaData.toString(prefix + indent,indent));
        }

        // Add value
        if (valueMetaData != null)
        {
            sb.append(valueMetaData.toString(prefix + indent,indent));
        }

        // TODO Add fetch-groups

        // Add order
        if (orderMetaData != null)
        {
            sb.append(orderMetaData.toString(prefix + indent,indent));
        }

        // Add embedded
        if (embeddedMetaData != null)
        {
            sb.append(embeddedMetaData.toString(prefix + indent,indent));
        }

        // Add index
        if (indexMetaData != null)
        {
            sb.append(indexMetaData.toString(prefix + indent,indent));
        }

        // Add unique
        if (uniqueMetaData != null)
        {
            sb.append(uniqueMetaData.toString(prefix + indent,indent));
        }

        // Add foreign-key
        if (foreignKeyMetaData != null)
        {
            sb.append(foreignKeyMetaData.toString(prefix + indent,indent));
        }

        // Add extensions
        sb.append(super.toString(prefix + indent,indent));

        sb.append(prefix).append("</field>\n");
        return sb.toString();
    }

    /**
     * Comparator method. This allows the ClassMetaData to search for a
     * FieldMetaData with a particular name.
     * @param o The object to compare against
     * @return The comparison result
     */ 
    public int compareTo(Object o)
    {
        if (o instanceof FieldMetaData)
        {
            FieldMetaData c = (FieldMetaData)o;
            return this.name.compareTo(c.name);
        }
        else if (o instanceof String)
        {
            return this.name.compareTo((String)o);
        }
        else if (o == null)
        {
            throw new ClassCastException("object is null");
        }
        throw new ClassCastException(this.getClass().getName() + " != " + o.getClass().getName());
    }
}