/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved. 
This program and the accompanying materials are made available under
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2004 Kikuchi Kousuke - org.jpox.enhancer.conf.JDOConfig
    ...
**********************************************************************/
package org.jpox.metadata;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;


/**
 * Representation of a Meta-Data file. 
 * Contains a list of package meta-data.
 *
 * <H3>MetaData Element</H3>
 * The MetaData Element represented here is as follows 
 * <PRE>
 * &lt;!ELEMENT jdo ((package)+, (query)*, (extension)*)&gt;
 * </PRE>
 *
 * @since 1.1 
 * @version $Revision: 1.5 $
 */
public class FileMetaData extends ExtendableMetaData
{
    /**
     * Manager for this MetaData object.
     */
    protected final MetaDataManager metaDataManager;

    /**
     * Name of file
     */
    protected String filename;

    /**
     * List of named queries
     */
    protected List queries = new ArrayList();

    /**
     * list of packages
     */
    protected List packages = new ArrayList();

    /**
     * Constructor.
     * @param filename The file where this is stored (or null).
     * @param metaDataManager the manager for this MetaData object.
     */
    public FileMetaData(String filename,MetaDataManager metaDataManager)
    {
        super(null);

        this.filename = filename;
        this.metaDataManager = metaDataManager;
    }

    // ------------------------------ Accessors --------------------------------
 
    /**
     * Accessor for the number of named queries.
     * @return no of named queries
     */
    public int getNoOfQueries()
    {
        return queries.size();
    }

    /**
     * Accessor for the metadata of a named query.
     * @param index query index
     * @return Meta-Data for the named query.
     */
    public QueryMetaData getQuery(int index)
    {
        if (index < 0 || index >= queries.size())
        {
            return null;
        }
        return (QueryMetaData)queries.get(index);
    }

    /**
     * Accessor for the MetaDataManager for this file.
     * @return The MetaDataManager for this file.
     */
    public MetaDataManager getMetaDataManager()
    {
        return metaDataManager;
    }

    /**
     * Accessor for the filename
     * @return The filename of this MetaData file.
     */
    public String getFilename()
    {
        return filename;
    }

    /**
     * Accessor for the number of packages.
     * @return no of packages.
     */
    public int getNoOfPackages()
    {
        return packages.size();
    }

    /**
     * Accessor for the meta-data of a package.
     * @param i index number
     * @return Meta-Data for a package.
     */
    public PackageMetaData getPackage(int i)
    {
        return (PackageMetaData)packages.get(i);
    }

    /**
     * Accessor for the Meta-Data of a package with a given name.
     * @param name Name of the package
     * @return Meta-Data for the package
     */
    public PackageMetaData getPackage(String name)
    {
        Iterator iter=packages.iterator();
        while (iter.hasNext())
        {
            PackageMetaData p = (PackageMetaData)iter.next();
            if (p.name.equals(name))
            {
                return p;
            }
        }
        return null;
    }

    /**
     * Utility method to check if the MetaData for a class is contained in this
     * file.
     * @param pkg_name Name of package
     * @param class_name Name of class
     * @return The MetaData for the class
     */
    public ClassMetaData getClass(String pkg_name,String class_name)
    {
        if (pkg_name == null || class_name == null)
        {
            return null;
        }

        PackageMetaData pmd=getPackage(pkg_name);
        if (pmd != null)
        {
            return pmd.getClass(class_name);
        }

        return null;
    }

    // -------------------------------- Mutators -------------------------------
 
    /**
     * Method to add a named query to this class. Rejects the addition of
     * duplicate named queries.
     * @param qmd Meta-Data for the query.
     */
    public void addQuery(QueryMetaData qmd)
    {
        if (qmd == null)
        {
            return;
        }

        queries.add(qmd);
    }

    /**
     * Mutator for the filename for this MetaData file.
     * @param filename The filename of this MetaData file.
     */
    public void setFilename(String filename)
    {
        this.filename = filename;
    }

    /**
     * Method to add a package
     * @param pkg The PackageMetaData to add.
     **/
    public void addPackage(PackageMetaData pkg)
    {
        if (pkg == null)
        {
            return;
        }
        Iterator iter=packages.iterator();
        while (iter.hasNext())
        {
            PackageMetaData p=(PackageMetaData)iter.next();
            // Check if already exists
            if (pkg.getName().equals(p.getName()))
            {
                return;
            }
        }
        packages.add(pkg);
    }

    // -------------------------------- Utilities ------------------------------
 
    /**
     * Returns a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        return toString("    ");
    }

    /**
     * Returns a string representation of the object.
     * @param indent The indent
     * @return a string representation of the object.
     */
    public String toString(String indent)
    {
        if (indent == null)
        {
            indent = "";
        }

        StringBuffer sb = new StringBuffer();
        sb.append("<jdo>\n");

        // Add packages
        Iterator iter=packages.iterator();
        while (iter.hasNext())
        {
            sb.append(((PackageMetaData)iter.next()).toString(indent,indent));
        }

        // Add queries
        for (int i=0;i<queries.size();i++)
        {
            QueryMetaData q = (QueryMetaData)queries.get(i);
            sb.append(q.toString(indent,indent));
        }

        // Add extensions
        sb.append(super.toString(indent,indent)); 

        sb.append("</jdo>");
        return sb.toString();
    }
}
