/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved.
This program and the accompanying materials are made available under
the terms of the JPOX License v1.0 which accompanies this distribution.

Contributors:
    ...
**********************************************************************/
package org.jpox.metadata;

/**
 * Foreign keys represent a consistency constraint in the database that must be
 * maintained. The user can specify by the value of the delete-action attribute
 * what happens if the target row of a foreign key is deleted.
 * 
 * <B>See jdo specification 2.0 15.5</B>
 * @since 1.1
 * @version $Revision: 1.2 $
 */
public class ForeignKeyDeleteAction extends ForeignKeyAction
{
    /**
     * delete-action="cascade". The database will automatically delete all rows
     * that refer to the row being deleted
     */
    public static final ForeignKeyDeleteAction CASCADE = new ForeignKeyDeleteAction(1);

    /**
     * delete-action="restrict". The user is required to explicitly make the
     * relationship valid by application code
     */
    public static final ForeignKeyDeleteAction RESTRICT = new ForeignKeyDeleteAction(2);

    /**
     * delete-action="null". The database will automatically nullify the columns
     * in all rows that refer to the row being deleted
     */
    public static final ForeignKeyDeleteAction NULL = new ForeignKeyDeleteAction(3);

    /**
     * delete-action="default". The database will automatically set the columns
     * in all rows that refer to the row being deleted to their default value
     */
    public static final ForeignKeyDeleteAction DEFAULT = new ForeignKeyDeleteAction(4);

    /**
     * delete-action="none". No foreign-key should be created.
     */
    public static final ForeignKeyDeleteAction NONE = new ForeignKeyDeleteAction(5);

    /**
     * the default action
     */
    public static final ForeignKeyDeleteAction DEFAULT_ACTION = new ForeignKeyDeleteAction(-1);

    /**
     * constructor
     * @param i type id
     */
    private ForeignKeyDeleteAction(int i)
    {
        super(i);
    }

    /**
     * Returns a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        switch (getType())
        {
            case 1 :
                return "cascade";
            case 2 :
                return "restrict";
            case 3 :
                return "null";
            case 4 :
                return "default";
            case 5 :
                return "none";
        }
        return "";
    }

    /**
     * Return ForeignKeyDeleteAction from String.
     * @param value delete-action attribute value
     * @return Instance of ForeignKeyDeleteAction. 
     *         If value invalid, return null.
     */
    public static ForeignKeyDeleteAction getForeignKeyDeleteAction(final String value)
    {
        if (value == null)
        {
            return null;
        }
        else if (ForeignKeyDeleteAction.CASCADE.toString().equalsIgnoreCase(value))
        {
            return ForeignKeyDeleteAction.CASCADE;
        }
        else if (ForeignKeyDeleteAction.DEFAULT.toString().equalsIgnoreCase(value))
        {
            return ForeignKeyDeleteAction.DEFAULT;
        }
        else if (ForeignKeyDeleteAction.NULL.toString().equalsIgnoreCase(value))
        {
            return ForeignKeyDeleteAction.NULL;
        }
        else if (ForeignKeyDeleteAction.RESTRICT.toString().equalsIgnoreCase(value))
        {
            return ForeignKeyDeleteAction.RESTRICT;
        }
        else if (ForeignKeyDeleteAction.NONE.toString().equalsIgnoreCase(value))
        {
            return ForeignKeyDeleteAction.NONE;
        }
        return null;
    }
}