/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2004 Andy Jefferson- added toString(), "column", MetaData, javadocs
    ...
**********************************************************************/
package org.jpox.metadata;

import java.util.ArrayList;
import java.util.List;

/**
 * Foreign keys in metadata serve two quite different purposes. First, when
 * generating schema, the foreign key element identifies foreign keys to be
 * generated. Second, when using the database, foreign key elements identify
 * foreign keys that are assumed to exist in the database. This is important for
 * the runtime to properly order insert, update, and delete statements to avoid
 * constraint violations. A foreign-key element can be contained by a field,
 * element, key, value, or join element, if all of the columns mapped are to be
 * part of the same foreign key. A foreign-key element can be contained within a
 * class element. In this case, the column elements are mapped elsewhere, and
 * the column elements contained in the foreign-key element have only the column
 * name.
 * 
 * <B>See jdo specification 2.0 15.5</B>
 *
 * <H3>MetaData Element</H3> 
 * The MetaData element represented here is as follows 
 * <PRE>
 * &lt;!ELEMENT foreign-key (extension*, (column|field|property)*, extension*)&gt;
 * &lt;!ATTLIST foreign-key table CDATA #IMPLIED&gt;
 * &lt;!ATTLIST foreign-key deferred (true|false) #IMPLIED&gt;
 * &lt;!ATTLIST foreign-key delete-action (cascade|restrict|null|default) #IMPLIED&gt;
 * &lt;!ATTLIST foreign-key update-action (cascade|restrict|null|default) #IMPLIED&gt;
 * &lt;!ATTLIST foreign-key unique (true|false) #IMPLIED&gt;
 * &lt;!ATTLIST foreign-key name CDATA #IMPLIED&gt;
 * </PRE>
 *
 * @since 1.1
 * @author Erik Bengtson
 * @version $Revision: 1.7 $
 */
public class ForeignKeyMetaData extends ExtendableMetaData
{
    /**
     * The deferred attribute specifies whether the foreign key constraint is
     * defined to be checked only at commit time.
     */
    protected boolean deferred;

    /**
     * Foreign keys represent a consistency constraint in the database that must
     * be maintained. The user can specify by the value of the delete-action
     * attribute what happens if the target row of a foreign key is deleted.
     */
    protected ForeignKeyDeleteAction deleteAction;

    /**
     * Foreign keys represent a consistency constraint in the database that must
     * be maintained. The user can specify by the update-action attribute what
     * happens if the target row of a foreign key is updated.
     */
    protected ForeignKeyUpdateAction updateAction;

    /**
     * The unique attribute specifies whether the foreign key constraint is
     * defined to be a unique constraint as well. This is most often used with
     * one-to-one mappings.
     */
    protected boolean unique;

    /**
     * the foreign-key table name. Name of the table to which this applies (null implies the enclosing class' table).
     */
    protected String table;

    /**
     * the foreign-key name
     */
    protected String name;

    /**
     * The fields for this foreign key.
     */
    protected final List fields = new ArrayList();

    /**
     * The columns for this foreign key.
     */
    protected final List columns = new ArrayList();

    // -------------------------------------------------------------------------
    // Fields below here are not represented in the output MetaData. They are
    // for use internally in the operation of the JDO system. The majority are
    // for convenience to save iterating through the fields since the fields
    // are fixed once initialised.

    /**
     * Contains the metadata for fields
     */
    protected FieldMetaData[] fieldMetaData;

    /**
     * Contains the metadata for columns
     */
    protected ColumnMetaData[] columnMetaData;

    /**
     * Constructor. 
     * @param deferred Whether the key is deferred
     * @param delete_action Action to perform on deletion
     * @param update_action Action to perform on update
     * @param table Table to apply the FK to
     * @param unique Whether the key is unique
     * @param name Name of the foreign key
     */
    public ForeignKeyMetaData(final String deferred,
                              final String delete_action,
                              final String update_action,
                              final String table,
                              final String unique,
                              final String name)
    {
        super(null);    // Ignore parent

        if (deferred != null)
        {
            if (deferred.equalsIgnoreCase("true"))
            {
                this.deferred = true;
            }
            else if (deferred.equalsIgnoreCase("false"))
            {
                this.deferred = false;
            }
        }

        this.deleteAction = ForeignKeyDeleteAction.getForeignKeyDeleteAction(delete_action);
        this.updateAction = ForeignKeyUpdateAction.getForeignKeyUpdateAction(update_action);

        this.table = table;
        if (unique != null)
        {
            if (unique.equalsIgnoreCase("true"))
            {
                this.unique = true;
            }
            else if (unique.equalsIgnoreCase("false"))
            {
                this.unique = false;
            }
        }
        this.name = name;
    }

    /**
     * Method to initialise the object, setting up all internal arrays.
     * Initialises all sub-objects.
     */
    public void initialise()
    {
        // Set up the columnMetaData
        if (fields.size() == 0)
        {
            fieldMetaData = null;
        }
        else
        {
            fieldMetaData = new FieldMetaData[fields.size()];
            for (int i=0; i<fieldMetaData.length; i++)
            {
                fieldMetaData[i] = (FieldMetaData) fields.get(i);
                fieldMetaData[i].initialise();
            }
        }

        // Set up the columnMetaData
        if (columns.size() == 0)
        {
            columnMetaData = null;
        }
        else
        {
            columnMetaData = new ColumnMetaData[columns.size()];
            for (int i=0; i<columnMetaData.length; i++)
            {
                columnMetaData[i] = (ColumnMetaData) columns.get(i);
                columnMetaData[i].initialise();
            }
        }

        setInitialised();
    }

    // ---------------------------- Mutators -----------------------------------

    /**
     * Add a new FieldMetaData element
     * @param fmd MetaData for the field
     */
    public void addField(FieldMetaData fmd)
    {
        fields.add(fmd);
    }

    /**
     * Add a new ColumnMetaData element
     * @param colmd MetaData for the column
     */
    public void addColumn(ColumnMetaData colmd)
    {
        columns.add(colmd);
    }

    // ---------------------------- Accessors ----------------------------------

    /**
     * Accessor for fieldMetaData
     * @return Returns the fieldMetaData.
     */
    public final FieldMetaData[] getFieldMetaData()
    {
        return fieldMetaData;
    }

    /**
     * Accessor for columnMetaData
     * @return Returns the columnMetaData.
     */
    public final ColumnMetaData[] getColumnMetaData()
    {
        return columnMetaData;
    }

    /**
     * Accessor for whether the FK is deferred
     * @return Returns the deferred.
     */
    public final boolean isDeferred()
    {
        return deferred;
    }

    /**
     * Accessor for the delete action
     * @return Returns the deleteAction.
     */
    public final ForeignKeyDeleteAction getDeleteAction()
    {
        return deleteAction;
    }

    /**
     * Accessor for the name of the foreign key
     * @return Returns the name.
     */
    public final String getName()
    {
        return name;
    }

    /**
     * Accessor for the name of the table
     * @return Returns the table.
     */
    public final String getTable()
    {
        return table;
    }

    /**
     * Accessor for the unique flag 
     * @return Returns the unique.
     */
    public final boolean isUnique()
    {
        return unique;
    }

    /**
     * Accessor for the update action
     * @return Returns the updateAction.
     */
    public final ForeignKeyUpdateAction getUpdateAction()
    {
        return updateAction;
    }

    // ---------------------------- Utilities ----------------------------------

    /**
     * Accessor for a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        return toString("","");
    }

    /**
     * Returns a string representation of the object using a prefix
     * This can be used as part of a facility to output a MetaData file. 
     * @param prefix prefix string
     * @param indent indent string
     * @return a string representation of the object.
     */
    public String toString(String prefix,String indent)
    {
        // Field needs outputting so generate metadata
        StringBuffer sb = new StringBuffer();
        sb.append(prefix).append("<foreign-key deferred=\"" + deferred + "\"\n");
        sb.append(prefix).append("       unique=\"" + unique + "\"");
        if (updateAction != null)
        {
            sb.append("\n").append(prefix).append("       update-action=\"" + updateAction + "\"");
        }
        if (deleteAction != null)
        {
            sb.append("\n").append(prefix).append("       delete-action=\"" + deleteAction + "\"");
        }
        if (table != null)
        {
            sb.append("\n").append(prefix).append("       table=\"" + table + "\"");
        }
        if (name != null)
        {
            sb.append("\n").append(prefix).append("       name=\"" + name + "\"");
        }
        sb.append(">\n");

        // Add fields
        for (int i=0;i<fields.size();i++)
        {
            FieldMetaData fmd=(FieldMetaData)fields.get(i);
            sb.append(fmd.toString(prefix + indent,indent));
        }

        // Add columns
        for (int i=0;i<columns.size();i++)
        {
            ColumnMetaData colmd=(ColumnMetaData)columns.get(i);
            sb.append(colmd.toString(prefix + indent,indent));
        }

        // Add extensions
        sb.append(super.toString(prefix + indent,indent));

        sb.append(prefix).append("</foreign-key>\n");
        return sb.toString();
    }
}