/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
    ...
**********************************************************************/
package org.jpox.metadata;

/**
 * Foreign keys represent a consistency constraint in the database that must be
 * maintained. The user can specify by the update-action attribute what happens
 * if the target row of a foreign key is updated
 * 
 * <B>See jdo specification 2.0 15.5</B>
 * @since 1.1
 * @version $Revision: 1.2 $
 */
public class ForeignKeyUpdateAction extends ForeignKeyAction
{
    /**
     * update-action="cascade". The database will automatically update all rows
     * that refer to the row being updated
     */
    public static final ForeignKeyUpdateAction CASCADE = new ForeignKeyUpdateAction(1);

    /**
     * update-action="restrict". The user is required to explicitly make the
     * relationship valid by application code
     */
    public static final ForeignKeyUpdateAction RESTRICT = new ForeignKeyUpdateAction(2);

    /**
     * update-action="null". The database will automatically nullify the columns
     * in all rows that refer to the row being updated
     */
    public static final ForeignKeyUpdateAction NULL = new ForeignKeyUpdateAction(3);

    /**
     * update-action="default". The database will automatically set the columns
     * in all rows that refer to the row being updated to their default value
     */
    public static final ForeignKeyUpdateAction DEFAULT = new ForeignKeyUpdateAction(4);

    /**
     * the default action
     */
    public static final ForeignKeyUpdateAction DEFAULT_ACTION = new ForeignKeyUpdateAction(-1);

    /**
     * constructor
     * @param i type id
     */
    private ForeignKeyUpdateAction(int i)
    {
        super(i);
    }

    /**
     * Returns a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        switch (getType())
        {
            case 1 :
                return "cascade";
            case 2 :
                return "restrict";
            case 3 :
                return "null";
            case 4 :
                return "default";
        }
        return "";
    }

    /**
     * Return ForeignKeyUpdateAction from String.
     * @param value update-action attribute value
     * @return Instance of ForeignKeyUpdateAction. 
     *         If value invalid, return null.
     */
    public static ForeignKeyUpdateAction getForeignKeyUpdateAction(final String value)
    {
        if (value == null)
        {
            return null;
        }
        else if (ForeignKeyUpdateAction.CASCADE.toString().equalsIgnoreCase(value))
        {
            return ForeignKeyUpdateAction.CASCADE;
        }
        else if (ForeignKeyUpdateAction.RESTRICT.toString().equalsIgnoreCase(value))
        {
            return ForeignKeyUpdateAction.RESTRICT;
        }
        else if (ForeignKeyUpdateAction.NULL.toString().equalsIgnoreCase(value))
        {
            return ForeignKeyUpdateAction.NULL;
        }
        else if (ForeignKeyUpdateAction.DEFAULT.toString().equalsIgnoreCase(value))
        {
            return ForeignKeyUpdateAction.DEFAULT;
        }
        return null;
    }
}