/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
    ...
**********************************************************************/
package org.jpox.metadata;

/**
 * Representation of the values for identity "strategy".
 *
 * @since 1.1 
 * @version $Revision: 1.7 $
 */
public class IdentityStrategy
{
    /**
     * strategy="native"
     * 
     * The value "native" allows the JDO implementation to pick the most
     * suitable strategy based on the underlying database.
     */
    public static final IdentityStrategy NATIVE=new IdentityStrategy(1);

    /**
     * strategy="sequence"
     * 
     * The value "sequence" specifies that a named database sequence is used to
     * generate key values for the table. If sequence is used, then the
     * sequence-name attribute is required.
     */
    public static final IdentityStrategy SEQUENCE=new IdentityStrategy(2);

    /**
     * strategy="autoassign"
     * 
     * The value "autoassignt" specifies that the column identified as the key
     * column is managed by the database to automatically increment key values.
     */
    public static final IdentityStrategy AUTOASSIGN=new IdentityStrategy(3);

    /**
     * strategy="identity"
     * 
     * The value "identity" specifies that the column
     * identified as the key column is managed by the database as an identity
     * type.
     */
    public static final IdentityStrategy IDENTITY=new IdentityStrategy(4);

    /**
     * strategy="increment"
     * 
     * The value "increment" specifies a strategy that simply finds the largest
     * key already in the database and increments the key value for new
     * instances. It can be used with integral column types when the JDO
     * application is the only database user inserting new instances.
     */
    public static final IdentityStrategy INCREMENT=new IdentityStrategy(5);

    /**
     * strategy="uuid-string"
     * 
     * The value "uuid-string" specifies a strategy that generates a 128-bit
     * UUID unique within a network (the IP address of the machine running the
     * application is part of the id) and represents the result as a
     * 16-character String.
     */
    public static final IdentityStrategy UUIDSTRING=new IdentityStrategy(6);

    /**
     * strategy="uuid-hex"
     * 
     * The value "uuid-hex" specifies a strategy that generates a 128-bit UUID
     * unique within a network (the IP address of the machine running the
     * application is part of the id) and represents the result as a
     * 32-character String.
     */
    public static final IdentityStrategy UUIDHEX=new IdentityStrategy(7);

    /**
     * strategy="max"
     * 
     * The value "max" specifies a strategy that simply finds the maximum
     * for a column and increments the value by 1. Similar to the "identity"
     * strategy.
     */
    public static final IdentityStrategy MAX=new IdentityStrategy(8);

    /**
     * strategy="auid"
     * 
     * The value "auid" specifies a strategy that creates a pure UUID
     * identity using the OpenGroup specification for UUID. The identities
     * will have 36 characters
     */
    public static final IdentityStrategy AUID=new IdentityStrategy(9);

    /**
     * strategy="custom"
     * 
     * The value "custom" specifies a strategy that uses a user own generator class.
     */
    public static final IdentityStrategy CUSTOM=new IdentityStrategy(10);

    /**
     * The type id.
     */
    private final int typeId;

    /**
     * constructor
     * @param i type id
     */
    private IdentityStrategy(int i)
    {
        this.typeId = i;
    }

    /**
     * Indicates whether some other object is "equal to" this one.
     * @param o the reference object with which to compare. 
     * @return true if this object is the same as the obj argument;
     *         false otherwise.
     */
    public boolean equals(Object o)
    {
        if (o instanceof IdentityStrategy)
        {
            return ((IdentityStrategy)o).typeId == typeId;
        }
        return false;
    }

    /**
     * Returns a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        switch (typeId)
        {
            case 1 :
                return "native";
            case 2 :
                return "sequence";
            case 3 :
                return "autoassign";
            case 4 :
                return "identity";
            case 5 :
                return "increment";
            case 6 :
                return "uuid-string";
            case 7 :
                return "uuid-hex";
            case 8 :
                return "max";
            case 9 :
                return "auid";
            case 10 :
                return "custom";
        }
        return "";
    }

    /**
     * Accessor for the type.
     * @return Type
     **/
    public int getType()
    {
        return typeId;
    }

    /**
     * Gets an IdentityStrategy for the given value argument.
     * @param value the String representation of IdentityStrategy
     * @return the IdentityStrategy corresponding to the value argument. NATIVE
     * IdentityStrategy is returned if the value argument is null or no
     * corresponding strategy was found
     */
    public static IdentityStrategy getIdentityStrategy(final String value)
    {
        if (value == null)
        {
            return IdentityStrategy.NATIVE;
        }
        else if (IdentityStrategy.NATIVE.toString().equals(value))
        {
            return IdentityStrategy.NATIVE;
        }
        else if (IdentityStrategy.SEQUENCE.toString().equals(value))
        {
            return IdentityStrategy.SEQUENCE;
        }
        else if (IdentityStrategy.AUTOASSIGN.toString().equals(value))
        {
            return IdentityStrategy.AUTOASSIGN;
        }
        else if (IdentityStrategy.IDENTITY.toString().equals(value))
        {
            return IdentityStrategy.IDENTITY;
        }
        else if (IdentityStrategy.INCREMENT.toString().equals(value))
        {
            return IdentityStrategy.INCREMENT;
        }
        else if (IdentityStrategy.UUIDSTRING.toString().equals(value))
        {
            return IdentityStrategy.UUIDSTRING;
        }
        else if (IdentityStrategy.UUIDHEX.toString().equals(value))
        {
            return IdentityStrategy.UUIDHEX;
        }
        // JPOX extensions below here
        else if (IdentityStrategy.MAX.toString().equals(value))
        {
            return IdentityStrategy.MAX;
        }
        else if (IdentityStrategy.AUID.toString().equals(value))
        {
            return IdentityStrategy.AUID;
        }
        else if (IdentityStrategy.CUSTOM.toString().equals(value))
        {
            return IdentityStrategy.CUSTOM;
        }
        return IdentityStrategy.NATIVE;
    }
}