/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2004 Andy Jefferson - toString(), "column", javadocs, initialise()
    ...
**********************************************************************/
package org.jpox.metadata;

import java.util.ArrayList;
import java.util.List;

/**
 * For schema generation, it might be useful to specify that a column or columns
 * be indexed, and to provide the name of the index. For this purpose, an index
 * element can be contained within a field, element, key, value, or join
 * element, and this indicates that the column(s) associated with the referenced
 * element should be indexed. Indexes can also be specified at the class level,
 * by including index elements containing column elements. In this case, the
 * column elements are mapped elsewhere, and the column elements contain only
 * the column name.
 *
 * <H3>MetaData Element</H3>
 * The MetaData Element represented here is as follows 
 * <PRE>
 * &lt;!ELEMENT index (extension*, (column|field|property)*, extension*)&gt;
 * &lt;!ATTLIST index name CDATA #IMPLIED&gt;
 * &lt;!ATTLIST index table CDATA #IMPLIED&gt;
 * &lt;!ATTLIST index unique (true|false) `false'&gt;
 * </PRE>
 *
 * @since 1.1
 * @version $Revision: 1.6 $
 */
public class IndexMetaData extends ExtendableMetaData
{
    /**
     * the index name
     */
    final String name;

    /**
     * the table name. If not specified, then applies to the table of the containing class.
     */
    final String table;

    /**
     * You can use UNIQUE constraints to ensure that no duplicate values are
     * entered in specific columns that do not participate in a primary key.
     * Although both a UNIQUE constraint and a PRIMARY KEY constraint enforce
     * uniqueness, use a UNIQUE constraint instead of a PRIMARY KEY constraint
     * when you want to enforce the uniqueness of:
     * <ul>
     * <li>
     * A column, or combination of columns, that is not the primary key.
     * Multiple UNIQUE constraints can be defined on a table, whereas only one
     * PRIMARY KEY constraint can be defined on a table.
     * </li>
     * <li>
     * A column that allows null values. UNIQUE constraints can be defined on
     * columns that allow null values, whereas PRIMARY KEY constraints can be
     * defined only on columns that do not allow null values.
     * </li>
     * </ul>
     * A UNIQUE constraint can also be referenced by a FOREIGN KEY constraint.
     */
    Boolean unique;

    /**
     * the field elements to be included in the index. Can be empty.
     */
    final List fields = new ArrayList();

    /**
     * the columns elements to be included in the index. Suitable to be empty
     * when this metadata is contained within a field, element, key, value, or
     * join elements
     */
    final List columns = new ArrayList();
    
    // -------------------------------------------------------------------------
    // Fields below here are not represented in the output MetaData. They are
    // for use internally in the operation of the JDO system. The majority are
    // for convenience to save iterating through the fields since the fields
    // are fixed once initialised.

    /**
     * Contains the metadata for fields
     */
    protected FieldMetaData[] fieldMetaData;

    /**
     * Contains the metadata for columns
     */
    protected ColumnMetaData[] columnMetaData;

    /**
     * Constructor. 
     * @param name Name of index
     * @param table Name of the table
     * @param unique Whether it is unique.
     */
    public IndexMetaData(final String name, String table, String unique)
    {
        super(null);    // Ignore parent

        this.name = name;
        this.table = table;
        if (unique != null)
        {
            if (unique.equalsIgnoreCase("true"))
            {
                this.unique = Boolean.TRUE;
            }
            else if (unique.equalsIgnoreCase("false"))
            {
                this.unique = Boolean.FALSE;
            }
        }
    }

    /**
     * Method to initialise the object, creating internal convenience arrays.
     * Initialise all sub-objects.
     */
    public void initialise()
    {
        // Set up the fieldMetaData
        if (fields.size() == 0)
        {
            fieldMetaData = null;
        }
        else
        {
            fieldMetaData = new FieldMetaData[fields.size()];
            for (int i=0; i<fieldMetaData.length; i++)
            {
                fieldMetaData[i] = (FieldMetaData) fields.get(i);
                fieldMetaData[i].initialise();
            }
        }

        // Set up the columnMetaData
        if (columns.size() == 0)
        {
            columnMetaData = null;
        }
        else
        {
            columnMetaData = new ColumnMetaData[columns.size()];
            for (int i=0; i<columnMetaData.length; i++)
            {
                columnMetaData[i] = (ColumnMetaData) columns.get(i);
                columnMetaData[i].initialise();
            }
        }

        setInitialised();
    }

    // -------------------------------- Mutators -------------------------------

    /**
     * Add a new FieldMetaData element
     */
    public void addField(FieldMetaData fmd)
    {
        fields.add(fmd);
    }

    /**
     * Accessor for fieldMetaData
     * @return Returns the fieldMetaData.
     */
    public final FieldMetaData[] getFieldMetaData()
    {
        return fieldMetaData;
    }

    /**
     * Add a new ColumnMetaData element
     */
    public void addColumn(ColumnMetaData colmd)
    {
        columns.add(colmd);
    }

    /**
     * Accessor for columnMetaData
     * @return Returns the columnMetaData.
     */
    public final ColumnMetaData[] getColumnMetaData()
    {
        return columnMetaData;
    }

    /**
     * Accessor for the name of the index.
     * @return Returns the name.
     */
    public final String getName()
    {
        return name;
    }

    /**
     * Accessor for the table
     * @return Returns the table..
     */
    public final String getTable()
    {
        return table;
    }

    /**
     * Accessor for whether the index is unique
     * @return Returns whether the index is unique.
     */
    public final Boolean isUnique()
    {
        return unique;
    }

    // -------------------------------- Utilities ------------------------------

    /**
     * Returns a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        return toString("","");
    }

    /**
     * Returns a string representation of the object.
     * This can be used as part of a facility to output a MetaData file. 
     * @param prefix prefix string
     * @param indent indent string
     * @return a string representation of the object.
     */
    public String toString(String prefix,String indent)
    {
        StringBuffer sb = new StringBuffer();
        sb.append(prefix).append("<index unique=\"" + unique + "\"");
        if (table != null)
        {
            sb.append(" table=\"" + table + "\"");
        }
        sb.append(name != null ? (" name=\"" + name + "\">\n") : ">\n");

        // Add fields
        for (int i=0;i<fields.size();i++)
        {
            FieldMetaData fmd = (FieldMetaData)fields.get(i);
            sb.append(fmd.toString(prefix + indent,indent));
        }

        // Add columns
        for (int i=0;i<columns.size();i++)
        {
            ColumnMetaData colmd = (ColumnMetaData)columns.get(i);
            sb.append(colmd.toString(prefix + indent,indent));
        }

        // Add extensions
        sb.append(super.toString(prefix + indent,indent));

        sb.append(prefix).append("</index>\n");
        return sb.toString();
    }
}