/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
2004 Andy Jefferson - added initialise() method
    ...
**********************************************************************/
package org.jpox.metadata;

/**
 * Representation of the Meta-Data defining inherited classes.
 *
 * <H3>MetaData Element</H3> 
 * The MetaData element represented here is as follows 
 * <PRE>
 * &lt;!ELEMENT inheritance (extension*, join?, discriminator?, extension*)?&gt;
 * &lt;!ATTLIST inheritance strategy CDATA #IMPLIED&gt;
 * </PRE> 
 *
 * @since 1.1 
 * @version $Revision: 1.3 $
 */
public class InheritanceMetaData extends ExtendableMetaData
{
    /**
     * strategy tag value.
     */
    protected InheritanceStrategy strategy=null;

    /**
     * JoinMetaData element.
     */
    protected JoinMetaData joinMetaData;

    /**
     * DiscriminatorMetaData element.
     */
    protected DiscriminatorMetaData discriminatorMetaData;

    /**
     * Constructor.
     * @param parent parent ClassMetaData instance
     * @param strategy Inheritance strategy
     */
    public InheritanceMetaData(ClassMetaData parent,
                               final String strategy)
    {
        super(parent);

        this.strategy = InheritanceStrategy.getInheritanceStrategy(strategy);
    }

    /**
     * Method to initialise the object, creating internal convenience arrays.
     * Initialises all sub-objects.
     */
    public void initialise()
    {
        if (joinMetaData != null)
        {
            joinMetaData.initialise();
        }

        if (discriminatorMetaData != null)
        {
            discriminatorMetaData.initialise();
        }

        setInitialised();
    }

    // ----------------------------- Accessors ---------------------------------

    /**
     * Accessor for the strategy tag value
     * @return strategy tag value
     */
    public InheritanceStrategy getStrategyValue()
    {
        return strategy;
    }

    /**
     * Accessor for the Join MetaData.
     * @return Returns the joinMetaData.
     */
    public JoinMetaData getJoinMetaData()
    {
        return joinMetaData;
    }

    /**
     * Mutator for the Join MetaData.
     * @param joinMetaData The joinMetaData to set.
     */
    public void setJoinMetaData(JoinMetaData joinMetaData)
    {
        this.joinMetaData = joinMetaData;
    }

    /**
     * Accessor for the Discrimintor MetaData.
     * @return Returns the Discrimintor MetaData.
     */
    public DiscriminatorMetaData getDiscriminatorMetaData()
    {
        return discriminatorMetaData;
    }

    /**
     * Mutator for the Discriminator MetaData.
     * @param discriminatorMetaData The discriminatorMetaData to set.
     */
    public void setDiscriminatorMetaData(DiscriminatorMetaData discriminatorMetaData)
    {
        this.discriminatorMetaData = discriminatorMetaData;
    }

    // ----------------------------- Utilities ---------------------------------

    /**
     * Accessor for a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        return toString("","");
    }

    /**
     * Returns a string representation of the object using a prefix
     * @param prefix prefix string
     * @param indent indent string
     * @return a string representation of the object.
     */
    public String toString(String prefix,String indent)
    {
        StringBuffer sb = new StringBuffer();
        sb.append(prefix).append("<inheritance strategy=\"" + strategy + "\">\n");

        // Add join
        if (joinMetaData != null)
        {
            sb.append(joinMetaData.toString(prefix + indent,indent));
        }

        // Add discriminator
        if (discriminatorMetaData != null)
        {
            sb.append(discriminatorMetaData.toString(prefix + indent,indent));
        }

        // Add extensions
        sb.append(super.toString(prefix + indent,indent));

        sb.append(prefix).append("</inheritance>\n");

        return sb.toString();
    }
}