/**********************************************************************
Copyright (c) 06-Jul-2004 Andy Jefferson and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution.

Contributors:
    ...
**********************************************************************/
package org.jpox.metadata;

import java.util.ArrayList;
import java.util.List;

import org.jpox.util.ClassUtils;
import org.jpox.util.StringUtils;

/**
 * Representation of the MetaData of an interface.
 *
 * <H3>MetaData Element</H3>
 * The MetaData Element represented here is as follows
 * <PRE>
 * &lt;!ELEMENT interface ((property)*, (extension*))&gt;
 * &lt;!ATTLIST interface name CDATA #REQUIRED&gt;
 * &lt;!ATTLIST interface requires-extent (true|false) `true'&gt;
 * </PRE> 
 *
 * @since 1.1
 * @version $Revision: 1.3 $
 */
public class InterfaceMetaData extends ExtendableMetaData
{
    /**
     * List of properties
     */
    protected List properties = new ArrayList();

    /**
     * Name of interface.
     */
    protected final String name;

    /**
     * Whether the interface requires an extent.
     */
    protected boolean requiresExtent=true;

    // -------------------------------------------------------------------------
    // Fields below here are not represented in the output MetaData. They are
    // for use internally in the operation of the JDO system. The majority are
    // for convenience to save iterating through the fields since the fields
    // are fixed once initialised.
 
    // ----------------------------- Constructors ------------------------------
 
    /**
     * Full name of interface.
     */
    protected String fullInterfaceName;

    /**
     * Constructor.
     * Takes the basic string information found in the MetaData file.
     * @param parent The package to which this class belongs
     * @param name Name of class
     * @param requiresExtent Whether the class requires an extent
     */
    public InterfaceMetaData(final PackageMetaData parent,
                         final String name,
                         final String requiresExtent)
    {
        super(parent);

        this.name = name;
        if (StringUtils.isWhitespace(name))
        {
            throw new InvalidMetaDataException(LOCALISER,
                "MetaData.Interface.NameNotSpecified.Error",parent.name);
        }

        this.fullInterfaceName = ClassUtils.createFullClassName(parent.name, name);

        // "requires-extent"
        if (requiresExtent != null && requiresExtent.equalsIgnoreCase("false"))
        {
            this.requiresExtent = false;
        }
        else
        {
            this.requiresExtent = true;
        }
    }

    /**
     * Method to initialise the object, creating internal convenience arrays.
     * Initialises all sub-objects. If populate() is going to be used it should
     * be used BEFORE calling this method.
     */
    public void initialise()
    {
        setInitialised();
    }

    // ------------------------------ Accessors --------------------------------
 
    /**
     * Accessor for the full interface name.
     * @return full interface name.
     */
    public String getFullInterfaceName()
    {
        return fullInterfaceName;
    }

    /**
     * Accessor for the number of properties
     * @return no of properties
     */
    public int getNoOfProperties()
    {
        return properties.size();
    }

    /**
     * Accessor for the interface name.
     * @return interface name
     */
    public String getName()
    {
        return name;
    }

    /**
     * Accessor for the requires-extent tag value.
     * @return requires-extent tag value
     */
    public boolean isRequiresExtent()
    {
        return requiresExtent;
    }

    /**
     * Convenience accessor for the parent Package MetaData.
     * @return MetaData for parent package.
     */
    public PackageMetaData getPackageMetaData()
    {
        if (parent != null)
        {
            return (PackageMetaData)parent;
        }
        return null;
    }

    /**
     * Convenience accessor for the package name.
     * @return package name.
     */
    public String getPackageName()
    {
        return getPackageMetaData().getName();
    }

    // ------------------------------- Mutators --------------------------------
 
    /**
     * Method to add a property to this interface.
     * @param pmd Property
     */
    public void addProperty(PropertyMetaData pmd)
    {
        if (pmd == null)
        {
            return;
        }
        
        properties.add(pmd);
    }

    // ------------------------------ Utilities --------------------------------
 
    /**
     * Returns a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        return toString("","");
    }

    /**
     * Returns a string representation of the object.
     * This can be used as part of a facility to output a MetaData file. 
     * @param prefix prefix string
     * @param indent indent string
     * @return a string representation of the object.
     */
    public String toString(String prefix,String indent)
    {
        StringBuffer sb = new StringBuffer();
        sb.append(prefix).append("<interface name=\"" + name + "\"\n");
        sb.append(prefix).append("     requires-extent=\"" + requiresExtent + "\">\n");

        // Add properties
        for (int i=0;i<properties.size();i++)
        {
            PropertyMetaData pmd = (PropertyMetaData)properties.get(i);
            sb.append(pmd.toString(prefix + indent, indent));
        }

        // Add extensions
        sb.append(super.toString(prefix + indent, indent)); 

        sb.append(prefix + "</interface>\n");
        return sb.toString();
    }
}
