/******************************************************************
 * Copyright (c) 03-Oct-2004 Andy Jefferson and others.
 * All rights reserved. This program and the accompanying materials
 * are made available under the terms of the JPOX License v1.0
 * which accompanies this distribution. 
 *
 * Contributors:
 *     ...
 *****************************************************************/
package org.jpox.metadata;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.net.URL;

import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;
import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

/**
 * Implementation of an entity resolver for JDO metadata files.
 * Handles JDO 1, JDO 2, and JDO 2 ORM entity resolution.
 * 
 * @version $Revision: 1.5 $
 */
public class JDOEntityResolver implements EntityResolver
{
    protected static Localiser LOCALISER=Localiser.getInstance("org.jpox.metadata.Localisation");
    
    /**
     * Public Key for JDO 1.0.1
     */
    private static final String PUBLIC_ID_KEY_JDO_1_0="-//Sun Microsystems, Inc.//DTD Java Data Objects Metadata 1.0//EN";

    /**
     * Public Key for JDO 2.0
     */
    private static final String PUBLIC_ID_KEY_JDO_2_0="-//Sun Microsystems, Inc.//DTD Java Data Objects Metadata 2.0//EN";
    
    /**
     * Public Key for ORM 2.0
     */
    private static final String PUBLIC_ID_KEY_ORM="-//Sun Microsystems, Inc.//DTD Java Data Objects Mapping Metadata 2.0//EN";

    /**
     * Method to resolve XML entities. Checks for PUBLIC and SYSTEM entities
     * valid for JDO 1.0 and JDO 2.0. Optimises access using local versions of
     * the JDO 1 and 2 DTD's.
     * @param publicId The public id.
     * @param systemId The system id.
     * @return Input Source for the URI.
     * @see org.xml.sax.EntityResolver#resolveEntity(java.lang.String,java.lang.String)
     */
    public InputSource resolveEntity(String publicId, String systemId)
    throws SAXException
    {
        try
        {
            if (JPOXLogger.METADATA.isDebugEnabled())
            {
                JPOXLogger.METADATA.debug(LOCALISER.msg("MetaData.Parser.ResolveEntity", publicId, systemId));
            }

            // Use publicId if possible
            if (publicId != null) 
            {
                if (publicId.equals(PUBLIC_ID_KEY_JDO_1_0))
                {
                    return getLocalInputSource("/org/jpox/jdo_1_0.dtd");
                }
                else if (publicId.equals(PUBLIC_ID_KEY_JDO_2_0))
                {
                    return getLocalInputSource("/org/jpox/jdo_2_0.dtd");
                }
                else if (publicId.equals(PUBLIC_ID_KEY_ORM))
                {
                    return getLocalInputSource("/org/jpox/jdo_orm_2_0.dtd");
                }
            }
            // Use systemId if possible
            if (systemId != null)
            {
                if (systemId.equals(PUBLIC_ID_KEY_JDO_1_0))
                {
                    return getLocalInputSource("/org/jpox/jdo_1_0.dtd");
                }
                else if (systemId.equals(PUBLIC_ID_KEY_JDO_2_0))
                {
                    return getLocalInputSource("/org/jpox/jdo_2_0.dtd");
                }
                else if (systemId.equals("file:/javax/jdo/jdo.dtd"))
                {
                    return getLocalInputSource("/org/jpox/jdo_2_0.dtd");
                }
                else if (systemId.equals("file:/javax/jdo/jdo_orm.dtd"))
                {
                    // Deprecated. Used in JPOX up to and including beta-1. Remove before full 1.1 release
                    return getLocalInputSource("/org/jpox/jdo_orm_2_0.dtd");
                }
                else if (systemId.equals("file:/javax/jdo/orm.dtd"))
                {
                    return getLocalInputSource("/org/jpox/jdo_orm_2_0.dtd");
                }
                else if (systemId.startsWith("file://"))
                {
                    // This is used to load from a file on the system, not on the CLASSPATH
                    String localPath = systemId.substring(7);
                    File file = new File(localPath);
                    if (file.exists())
                    {
                        FileInputStream in = new FileInputStream(file);
                        return new InputSource(in);
                    }
                    return null;
                }
                else if (systemId.startsWith("file:"))
                {
                    // Try to get the local file using CLASSPATH
                    return getLocalInputSource(systemId.substring(5));
                }
                else if (systemId.startsWith("http:"))
                {
                    // Try to get the URL and open its stream
                    try
                    {
                        URL system_url = new URL(systemId);
                        InputStream url_stream = system_url.openStream();
                        return new InputSource(url_stream);
                    }
                    catch (Exception e)
                    {
                        JPOXLogger.METADATA.error(e);
                    }
                }
            }

            JPOXLogger.METADATA.error(this.getClass().getName() + ".resolveEntity(\"" + publicId + "\", \"" + systemId + "\") could not be found!");
            return null;
        }
        catch (Exception e)
        {
        	JPOXLogger.METADATA.error(this.getClass().getName() + ".resolveEntity(\"" + publicId + "\", \"" + systemId + "\") failed!", e);
            throw new SAXException(e.getMessage(), e);
        }
    }

    /**
     * Accessor for the input source for a path.
     * @param localPath The local path
     * @return The input source
     * @throws FileNotFoundException
     */
    private InputSource getLocalInputSource(String localPath)
    	throws FileNotFoundException
    {
    	JPOXLogger.METADATA.debug(this.getClass().getName() + ".getLocalInputSource(\"" + localPath + "\")");

    	InputStream in = getClass().getResourceAsStream(localPath);
    	if (in == null)
    	{
    		JPOXLogger.METADATA.fatal("local resource \"" + localPath + "\" does not exist!!!");
    		throw new FileNotFoundException("Unable to load resource: " + localPath);
    	}
    	return new InputSource(new InputStreamReader(in));
    }
}