/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
2004 Andy Jefferson - addition of table name
    ...
**********************************************************************/
package org.jpox.metadata;

import java.util.ArrayList;
import java.util.List;

/**
 * Secondary tables and join tables are mapped using a join condition that
 * associates a column or columns in the secondary or join table with a column
 * or columns in the primary table, typically the primary tables primary key
 * columns. Column elements used for relationship mapping or join conditions
 * specify the column name and optionally the target column name. The target
 * column name is the name of the column in the associated table corresponding
 * to the named column. The target column name is optional when the target
 * column is the single primary key column of the associated table.
 * 
 * <H3>MetaData Element</H3> 
 * The MetaData element represented here is as follows 
 * <PRE> 
 * &lt;!ELEMENT join (column*, index?, foreign-key?, extension*)&gt;
 * &lt;!ATTLIST join table CDATA #IMPLIED&gt;
 * &lt;!ATTLIST join column CDATA #IMPLIED&gt;
 * &lt;!ATTLIST join outer (true|false) `false'&gt;
 * &lt;!ATTLIST join foreign-key CDATA #IMPLIED&gt;
 * </PRE>
 *
 * @since 1.1
 * @version $Revision: 1.3 $
 */
public class JoinMetaData extends ExtendableMetaData
{
    /**
     * the foreign-key element.
     */
    ForeignKeyMetaData foreignKeyMetaData;

    /**
     * the index element.
     */
    IndexMetaData indexMetaData;
    
    /**
     * column elements
     */
    final List columns = new ArrayList();
       
    /**
     * if is outer join. 
     * Outer joins return all elements from at least one of the sides joined
     */
    Boolean outer;

    /**
     * the table name.
     */
    String table;

    /**
     * the column name.
     */
    String column;

    /**
     * the foreign-key column name.
     */
    String foreignKey;
       
    // -------------------------------------------------------------------------
    // Fields below here are not represented in the output MetaData. They are
    // for use internally in the operation of the JDO system. The majority are
    // for convenience to save iterating through the fields since the fields
    // are fixed once initialised.

    /**
     * Contains the metadata for column
     */
    protected ColumnMetaData[] columnMetaData;
 
    /**
     * Constructor for the Join. 
     * @param table The table name
     * @param column The column name
     * @param outer Whether it is an outer join
     * @param foreignKey
     */
    public JoinMetaData(final String table,
                        final String column,
                        final String outer,
                        final String foreignKey)
    {
        super(null);    // Ignore parent

        this.table = table;
        this.column = column;

        if (outer != null)
        {
            if (outer.equalsIgnoreCase("true"))
            {
                this.outer = Boolean.TRUE;
            }
            else if (outer.equalsIgnoreCase("false"))
            {
                this.outer = Boolean.FALSE;
            }
        }

        this.foreignKey = foreignKey;
    }

    /**
     * Method to initialise the object, creating internal convenience arrays.
     * Initialises all sub-objects.
     */
    public void initialise()
    {
        // Cater for user specifying column name, or columns
        if (columns.size() == 0 && column != null)
        {
            columnMetaData = new ColumnMetaData[1];
            columnMetaData[0] = new ColumnMetaData(this, column);
        }
        else
        {
            columnMetaData = new ColumnMetaData[columns.size()];
            for (int i=0; i<columnMetaData.length; i++)
            {
                columnMetaData[i] = (ColumnMetaData) columns.get(i);
                columnMetaData[i].initialise();
            }
        }

        if (foreignKeyMetaData != null)
        {
            foreignKeyMetaData.initialise();
        }
        if (indexMetaData != null)
        {
            indexMetaData.initialise();
        }

        setInitialised();
    }

    // ------------------------------- Accessors -------------------------------
 
    /**
     * Add a new ColumnMetaData element
     */
    public void addColumn(ColumnMetaData colmd)
    {
        columns.add(colmd);
    }

    /**
     * Accessor for the table name 
     * @return Returns the table name
     */
    public final String getTable()
    {
        return column;
    }

    /**
     * Accessor for the column name 
     * @return Returns the column.
     */
    public final String getColumn()
    {
        return column;
    }

    /**
     * Accessor for columnMetaData
     * @return Returns the columnMetaData.
     */
    public final ColumnMetaData[] getColumnMetaData()
    {
        return columnMetaData;
    }

    /**
     * @return Returns the foreignKey.
     */
    public final String getForeignKey()
    {
        return foreignKey;
    }

    /**
     * @return Returns the outer.
     */
    public final boolean isOuter()
    {
        if (outer == null)
        {
            return false;
        }
        else
        {
            return outer.booleanValue();
        }
    }

    /**
     * @return Returns the indexMetaData.
     */
    public final IndexMetaData getIndexMetaData()
    {
        return indexMetaData;
    }

    /**
     * @return Returns the foreignKeyMetaData.
     */
    public final ForeignKeyMetaData getForeignKeyMetaData()
    {
        return foreignKeyMetaData;
    }

    // ----------------------------- Mutators ----------------------------------
 
    /**
     * @param foreignKeyMetaData The foreignKeyMetaData to set.
     */
    public final void setForeignKeyMetaData(
            ForeignKeyMetaData foreignKeyMetaData)
    {
        this.foreignKeyMetaData = foreignKeyMetaData;
    }

    /**
     * @param indexMetaData The indexMetaData to set.
     */
    public final void setIndexMetaData(IndexMetaData indexMetaData)
    {
        this.indexMetaData = indexMetaData;
    }

    // ----------------------------- Utilities ---------------------------------
 
    /**
     * Accessor for a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        return toString("","");
    }

    /**
     * Returns a string representation of the object using a prefix
     * @param prefix prefix string
     * @param indent indent string
     * @return a string representation of the object.
     */
    public String toString(String prefix,String indent)
    {
        StringBuffer sb = new StringBuffer();
        sb.append(prefix).append("<join");
        boolean first_param=true;
        if (table != null)
        {
            if (first_param)
            {
                sb.append(" table=\"" + table + "\"");
            }
            else
            {
                sb.append("\n").append(prefix).append("     table=\"" + table + "\"");
            }
            first_param = false;
        }
        if (column != null)
        {
            if (first_param)
            {
                sb.append(" column=\"" + column + "\"");
            }
            else
            {
                sb.append("\n").append(prefix).append("     column=\"" + column + "\"");
            }
            first_param = false;
        }
        if (outer != null)
        {
            if (first_param)
            {
                sb.append(" outer=\"" + outer + "\"");
            }
            else
            {
                sb.append("\n").append(prefix).append("     outer=\"" + outer + "\"");
            }
            first_param = false;
        }
        if (foreignKey != null)
        {
            if (first_param)
            {
                sb.append(" foreign-key=\"" + foreignKey + "\"");
            }
            else
            {
                sb.append("\n").append(prefix).append("     foreign-key=\"" + foreignKey + "\"");
            }
            first_param = false;
        }
        sb.append(">\n");

        // Add extensions
        sb.append(super.toString(prefix + indent,indent));

        sb.append(prefix).append("</join>\n");
        return sb.toString();
    }
}