/**********************************************************************
Copyright (c) 04-Jun-2004 Andy Jefferson and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
    Kikuchi Kousuke - org.jpox.enhancer.conf.JDOConfigMap
    Erik Bengtson - add dependent keys and values
    ...
**********************************************************************/
package org.jpox.metadata;

import java.util.List;
import java.util.Set;

import javax.jdo.spi.PersistenceCapable;

import org.jpox.ClassLoaderResolver;
import org.jpox.util.JPOXLogger;
import org.jpox.util.StringUtils;

/**
 * Representation of the Meta-Data for a Map.
 * 
 * <H3>MetaData Element</H3>
 * Represents the following MetaData Element
 * <PRE>
 * &lt;!ELEMENT map (extension)*&gt;
 * &lt;!ATTLIST map key-type CDATA #IMPLIED&gt;
 * &lt;!ATTLIST map embedded-key (true|false) #IMPLIED&gt;
 * &lt;!ATTLIST map value-type CDATA #IMPLIED&gt;
 * &lt;!ATTLIST map embedded-value (true|false) #IMPLIED&gt;
 * &lt;!ATTLIST map dependent-value (true|false) #IMPLIED&gt;
 * &lt;!ATTLIST map dependent-key (true|false) #IMPLIED&gt;
 * </PRE>
 *
 * @since 1.1 
 * @version $Revision: 1.3 $
 */
public class MapMetaData extends ContainerMetaData
{
    /**
     * embedded-key tag value.
     */
    protected Boolean embeddedKey;

    /**
     * embedded-value tag value.
     */
    protected Boolean embeddedValue;

    /**
     * key-type tag  value.
     */
    protected String keyType = "java.lang.Object";

    /**
     * value-type  tag value.
     */
    protected String valueType = "java.lang.Object";

    /**
     * dependent-value tag value.
     * 
     * The dependent-value attribute indicates that the collection's value
     * contains references that are to be deleted if the referring instance is
     * deleted.
     */
    protected Boolean dependentValue; 

    /**
     * dependent-key tag value.
     * 
     * The dependent-key attribute indicates that the collection's key contains
     * references that are to be deleted if the referring instance is deleted.
     */
    protected Boolean dependentKey; 
    
    /**
     * Constructor.
     * @param parent parent Field
     * @param keyType key-type tag value
     * @param embeddedKey embedded-key tag value
     * @param valueType value-type  tag value
     * @param embeddedValue embedded-value tag value
     * @param dependentValue dependent-value tag value
     * @param dependentKey dependent-key tag value
     */
    public MapMetaData(FieldMetaData parent,
                       String keyType,
                       String embeddedKey,
                       String valueType,
                       String embeddedValue,
                       String dependentValue,
                       String dependentKey)
    {
        super(parent);

        if (!StringUtils.isWhitespace(keyType))
        {
            this.keyType = keyType;
        }
        if (!StringUtils.isWhitespace(valueType))
        {
            this.valueType = valueType;
        }
        if (embeddedKey != null)
        {
            this.embeddedKey = new Boolean(embeddedKey);
        }
        if (embeddedValue != null)
        {
            this.embeddedValue = new Boolean(embeddedValue);
        }
        // "dependent-value"
        if (dependentValue != null)
        {
            if (dependentValue.equalsIgnoreCase("true"))
            {
                this.dependentValue = Boolean.TRUE;
            }
            else if (dependentValue.equalsIgnoreCase("false"))
            {
                this.dependentValue = Boolean.FALSE;
            }
        }
        // "dependent-key"
        if (dependentKey != null)
        {
            if (dependentKey.equalsIgnoreCase("true"))
            {
                this.dependentKey = Boolean.TRUE;
            }
            else if (dependentKey.equalsIgnoreCase("false"))
            {
                this.dependentKey = Boolean.FALSE;
            }
        }
    }

    /**
     * Method to populate any defaults, and check the validity of the MetaData.
     * @param loader ClassLoader to use for loading any key/value types 
     */
    public void populate(ClassLoader loader)
    {
        // Check the field type and see if it is castable to a Map
        Class field_type=getFieldMetaData().getType();
        if (!java.util.Map.class.isAssignableFrom(field_type))
        {
            throw new InvalidMetaDataException(LOCALISER,
                    "MetaData.Map.IsNotMapError",
                    getFieldName(),getFieldMetaData().getClassName(false));
        }

        // "key-type"
        if (keyType == null)
        {
            throw new InvalidMetaDataException(LOCALISER,
                    "MetaData.Map.KeyTypeNotSpecified.Error",
                    getFieldName(),getFieldMetaData().getClassName(false));
        }

        // "embedded-key"
        if (embeddedKey == null)
        {
            // Load the key type
            Class key_type_class=MetaDataUtils.resolvedClassForName(keyType,
                    getFieldMetaData().getClassMetaData().getPackageName(),loader); 
            if (key_type_class == null)
            {
                throw new InvalidMetaDataException(LOCALISER,
                        "MetaData.Map.KeyTypeNotFound.Error",
                        getFieldName(),getFieldMetaData().getClassName(false),
                        keyType);
            }
            else
            {
                if (!key_type_class.getName().equals(keyType))
                {
                    // Update to the fully-qualified name
                    JPOXLogger.METADATA.warn(LOCALISER.msg("MetaData.Map.KeyTypeResolved",getFieldName(),getFieldMetaData().getClassName(false),keyType,key_type_class.getName()));
                    keyType = key_type_class.getName();
                }
            }

            // Assign default for "embedded-key" based on 18.13.2 of JDO 2 spec
            if (PersistenceCapable.class.isAssignableFrom(key_type_class) ||
                Object.class.isAssignableFrom(key_type_class) ||
                key_type_class.isInterface())
            {
                embeddedKey = Boolean.TRUE;
            }
            else
            {
                embeddedKey = Boolean.FALSE;
            }
        }

        // "value-type"
        if (valueType == null)
        {
            throw new InvalidMetaDataException(LOCALISER,
                    "MetaData.Map.ValueTypeNotSpecified.Error",
                    getFieldName(),getFieldMetaData().getClassName(false));
        }

        // "embedded-value"
        if (embeddedValue == null)
        {
            // Load the value type
            Class value_type_class=MetaDataUtils.resolvedClassForName(valueType,
                    getFieldMetaData().getClassMetaData().getPackageName(),loader); 
            if (value_type_class == null)
            {
                throw new InvalidMetaDataException(LOCALISER,
                        "MetaData.Map.ValueTypeNotFound.Error",
                        getFieldMetaData().getName(),
                        getFieldMetaData().getClassName(false),
                        valueType);
            }
            else
            {
                if (!value_type_class.getName().equals(valueType))
                {
                    // Update to the fully-qualified name
                    JPOXLogger.METADATA.warn(LOCALISER.msg("MetaData.Map.ValueTypeResolved",getFieldName(),getFieldMetaData().getClassName(false),valueType,value_type_class.getName()));
                    valueType = value_type_class.getName();
                }
            }

            // Assign default for "embedded-key" based on 18.13.2 of JDO 2 spec
            if (PersistenceCapable.class.isAssignableFrom(value_type_class) ||
                Object.class.isAssignableFrom(value_type_class) ||
                value_type_class.isInterface())
            {
                embeddedValue = Boolean.TRUE;
            }
            else
            {
                embeddedValue = Boolean.FALSE;
            }
        }

        setPopulated();
    }

    // ----------------------------- Accessors ---------------------------------
 
    /**
     * Accessor for the key-type tag value.
     * @return key-type tag value
     */
    public String getKeyType()
    {
        return keyType;
    }

    /**
     * Accessor for the value-type tag value.
     * @return value-type tag value
     */
    public String getValueType()
    {
        return valueType;
    }

    /**
     * Accessor for the embedded-key tag value.
     * @return embedded-key tag value
     */
    public boolean isEmbeddedKey()
    {
        if (embeddedKey == null)
        {
            return false;
        }
        else
        {
            return embeddedKey.booleanValue();
        }
    }

    /**
     * Accessor for the embedded-value tag value.
     * @return embedded-value tag value
     */
    public boolean isEmbeddedValue()
    {
        if (embeddedValue == null)
        {
            return false;
        }
        else
        {
            return embeddedValue.booleanValue();
        }
    }

    /**
     * Accessor for the dependent-key attribute indicates that the collection's
     * key contains references that are to be deleted if the referring instance
     * is deleted.
     * 
     * @return dependent-key tag value
     */
    public boolean isDependentKey()
    {
        if (dependentKey == null)
        {
            return false;
        }
        else
        {
            return dependentKey.booleanValue();
        }
    }
    
    /**
     * Accessor for the dependent-value attribute indicates that the
     * collection's value contains references that are to be deleted if the
     * referring instance is deleted.
     * 
     * @return dependent-value tag value
     */
    public boolean isDependentValue()
    {
        if (dependentValue == null)
        {
            return false;
        }
        else
        {
            return dependentValue.booleanValue();
        }
    }
    
    // ----------------------------- Mutators ----------------------------------
 
    // ----------------------------- Utilities ---------------------------------
 
    /**
     * Accessor for all ClassMetaData referenced by this array.
     * @param orderedCMDs List of ordered ClassMetaData objects (added to).
     * @param referencedCMDs Set of all ClassMetaData objects (added to).
     * @param dba_vendor_id Vendor ID of the DBA. Used for view addition.
     * @param clr the ClassLoaderResolver 
     **/
    void getReferencedClassMetaData(final List orderedCMDs,
                                    final Set referencedCMDs,
                                    final String dba_vendor_id,
                                    final ClassLoaderResolver clr) 
    {
        ClassMetaData key_cmd=getMetaDataManager().getMetaDataForClass(keyType,clr);
        if (key_cmd != null)
        {
            key_cmd.getReferencedClassMetaData(orderedCMDs,referencedCMDs,dba_vendor_id,clr);
        }

        ClassMetaData value_cmd=getMetaDataManager().getMetaDataForClass(valueType,clr);
        if (value_cmd != null)
        {
            value_cmd.getReferencedClassMetaData(orderedCMDs,referencedCMDs,dba_vendor_id,clr);
        }
    }

    /**
     * Returns a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        return toString("","");
    }

    /**
     * Returns a string representation of the object.
     * @param prefix prefix string
     * @param indent indent string
     * @return a string representation of the object.
     */
    public String toString(String prefix,String indent)
    {
        StringBuffer sb = new StringBuffer();
        if (embeddedKey != null)
        {
            sb.append(prefix).append("<map key-type=\"").append(keyType).append("\" embedded-key=\"").append(embeddedKey).append("\"\n");
        }
        else
        {
            sb.append(prefix).append("<map key-type=\"").append(keyType).append("\"\n");
        }
        if (embeddedValue != null)
        {
            sb.append(prefix).append("     value-type=\"").append(valueType).append("\" embedded-value=\"").append(embeddedValue).append("\">\n");
        }
        else
        {
            sb.append(prefix).append("     value-type=\"").append(valueType).append("\">\n");
        }

        // Add extensions
        sb.append(super.toString(prefix + indent,indent));

        sb.append(prefix).append("</map>\n");
        return sb.toString();
    }
}
