/**********************************************************************
Copyright (c) Jun 21, 2004 Erik Bengtson and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
    Andy Jefferson - addition of states
    ...
**********************************************************************/
package org.jpox.metadata;

import org.jpox.util.Localiser;

/**
 * Base class for all MetaData. Provides state management for the MetaData.
 * The states represent the lifecycle of a MetaData object. The lifecycle goes
 * as follows :
 * <OL>
 * <LI>MetaData object is created (values passed in from a parsed file, or
 * manually generated)</LI>
 * <LI>MetaData object is populated (maybe pass in a class that it represents,
 * creating any additional information that wasn't in the initial data).</LI>
 * <LI>MetaData object is initialised (any internal arrays are set up, and
 * additions of data is blocked from this point).
 * <LI>MetaData object is added to with runtime information like actual column
 * names and types in use.</LI> 
 * </OL>
 *
 * @version $Revision: 1.3 $
 */
public class MetaData
{
    protected static Localiser LOCALISER=Localiser.getInstance("org.jpox.metadata.Localisation");

    /**
     * State representing the start state of MetaData, representing the initial
     * values passed in.
     */
    public static final int METADATA_CREATED_STATE=0;

    /**
     * State reflecting that MetaData object has been populated with the real
     * class definition, adding in extra fields that weren't in the startup.
     */
    public static final int METADATA_POPULATED_STATE=1;

    /**
     * State reflecting that MetaData object has been initialised, setting up
     * any internal information that is required.
     */
    public static final int METADATA_INITIALISED_STATE=2;

    /**
     * State reflecting that MetaData object has been modified with usage
     * information (e.g actual column names for the fields).
     */
    public static final int METADATA_USED_STATE=3;

    /**
     * State indicator for this MetaData object.
     */
    protected int metaDataState=METADATA_CREATED_STATE;

    /**
     * Parent MetaData object.
     */
    protected final MetaData parent;

    /**
     * Constructor. Taking the parent MetaData object (if any).
     * @param parent The parent MetaData object.
     */
    public MetaData(MetaData parent)
    {
        this.parent = parent;
    }

    // -------------------------------- Mutators -------------------------------
 
    /**
     * Method to populate the object.
     * The state changes to "POPULATED" after this call.
     */
    public void populate()
    {
        setPopulated();
    }

    /**
     * Method to initialise the object.
     * The state changes to "INITIALISED" after this call.
     */
    public void initialise()
    {
        setInitialised();
    }

    /**
     * Utility to set the state as initialised.
     */
    public void setInitialised()
    {
        metaDataState = METADATA_INITIALISED_STATE;
    }

    /**
     * Utility to set the state as populated.
     */
    public void setPopulated()
    {
        metaDataState = METADATA_POPULATED_STATE;
    }

    /**
     * Utility to set the state as used.
     */
    public void setUsed()
    {
        metaDataState = METADATA_USED_STATE;
    }

    // -------------------------------- Accessors ------------------------------
 
    /**
     * Accessor for the parent MetaData object.
     * @return Parent MetaData object.
     */
    public MetaData getParent()
    {
        return parent;
    }

    /**
     * Accessor for whether the object state is "created".
     * @return Whether it is in "created" state
     */
    public synchronized boolean isCreated()
    {
        return (metaDataState == METADATA_CREATED_STATE);
    }

    /**
     * Accessor for whether the object state is "populated".
     * @return Whether it is populated.
     */
    public synchronized boolean isPopulated()
    {
        return (metaDataState == METADATA_POPULATED_STATE);
    }

    /**
     * Accessor for whether the object state is "initialised".
     * @return Whether it is initialised
     */
    public synchronized boolean isInitialised()
    {
        return (metaDataState == METADATA_INITIALISED_STATE);
    }

    /**
     * Accessor for whether the object state is "used".
     * @return Whether it is used.
     */
    public synchronized boolean isUsed()
    {
        return (metaDataState == METADATA_USED_STATE);
    }
}
