/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
    ...
**********************************************************************/
package org.jpox.metadata;

import java.net.URL;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.StringTokenizer;

import javax.jdo.JDOUserException;

import org.jpox.ClassLoaderResolver;
import org.jpox.TypeManager;
import org.jpox.util.ClassUtils;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;
import org.jpox.util.StringUtils;

/**
 * Manager of MetaData information in JPOX. Acts as a registry of JDO metadata
 * so that metadata files don't need to be parsed multiple times. MetaData is
 * stored as a FileMetaData, which contains PackageMetaData, which contains
 * ClassMetaData, and so on. This maps exactly to the users model of their
 * metadata. The users access point is <B>getMetaDataForClass()</B> which will
 * check the known classes without metadata, then check the existing registered
 * metdata, then check the valid locations for metdata files. This way, the
 * metadata is managed from this single point.
 * <P>
 * Maintains a list of all classes that have been checked for MetaData and
 * don't have any available. This avoids the needs to look up MetaData multiple
 * times finding the same result. Currently this list is for all ClassMetaData 
 * objects keyed by the class name. If necessary we could store them keyed by
 * file name, but not currently necessary.
 * </P>
 * <P>
 * Each PMFContext typically will have its own MetaDataManager so allowing
 * Meta-Data to be for different datastores. In addition, each PMF can allow
 * MetaData files to use a particular suffix, hence we allow the JDO file
 * suffix to be specifiable at construction.
 * </P>
 * TODO Add a cache of MetaDataManagers and add a check of the other managers
 * when a class is required and we don't have the file (to save the parse). 
 *
 * @version $Revision: 1.15 $ 
 */
public class MetaDataManager
{
    private static final Localiser LOCALISER=Localiser.getInstance("org.jpox.metadata.Localisation");

    /**
     * Suffix for JDO MetaData files.
     */
    protected String jdoFileSuffix="jdo";

    /**
     * Suffix for ORM MetaData files.
     */
    protected String ormFileSuffix="orm";
    
    /**
     * Name for the ORM mapping (used to define the filename of the ORM files).
     * If this is null we only use the jdo metadata file.
     */
    protected String ormMappingName=null;

    /**
     * Default for the inheritance strategy to use. Null implies use JDO2's strategy.
     */
    protected String defaultInheritanceStrategy="JDO2";

    /**
     * Map of ClassMetaData, keyed by the class name. Uses the same objects as
     * created during the parse process, and stored above.
     */
    protected Map classMetaDataByClass=new HashMap();

    /**
     * Map of QueryMetaData, keyed by the class name and query name.
     * Uses the same objects as created during the parse process.
     */
    protected Map queryMetaDataByClassQuery=new HashMap();

    protected Map ormClassMetaDataByClass=new HashMap();
    
    /**
     * Classes without MetaData. Stores the class names.
     */
    protected Collection classesWithoutMetaData=new HashSet();

    protected Map fileMetaDataByURL=new HashMap();

    // -------------------------------------------------------------------------

    /**
     * Constructor.
     */
    public MetaDataManager()
    {
        // Register all of the types managed by the TypeManager as not needing
        // MetaData.
        Set supported_classes=TypeManager.getTypeManager().getSupportedTypes();
        Iterator iter=supported_classes.iterator();
        while (iter.hasNext())
        {
            String s=(String)iter.next();
            addClassWithoutMetaData(s);
        }
    }

    /**
     * Constructor, allowing specification of the suffix used for JDO files.
     * @param jdo_file_suffix Suffix to be used on all JDO MetaData files.
     * @param orm_file_suffix Suffix to be used on all ORM MetaData files.
     * @param orm_mapping_name Name of the ORM mapping
     * @param default_inheritance_strategy Default Strategy for inheritance
     */
    public MetaDataManager(String jdo_file_suffix,
                           String orm_file_suffix,
                           String orm_mapping_name,
                           String default_inheritance_strategy)
    {
        this();

        // Only accept a suffix if it isn't null
        if (!StringUtils.isWhitespace(jdo_file_suffix))
        {
            jdoFileSuffix = jdo_file_suffix;
        }

        if (!StringUtils.isWhitespace(orm_file_suffix))
        {
            ormFileSuffix = orm_file_suffix;
        }

        if (!StringUtils.isWhitespace(orm_mapping_name))
        {
            ormMappingName = orm_mapping_name;
        }

        if (!StringUtils.isWhitespace(default_inheritance_strategy))
        {
            defaultInheritanceStrategy = default_inheritance_strategy;
        }
    }

    /**
     * Accessor for the suffix used when looking for JDO MetaData files.
     * @return JDO MetaData file suffix.
     */
    public String getJdoFileSuffix()
    {
        return jdoFileSuffix;
    }
    
    /**
     * Accessor for the suffix used when looking for ORM MetaData files.
     * @return ORM MetaData file suffix.
     */
    public String getOrmFileSuffix()
    {
        return ormFileSuffix;
    }

    /**
     * Accessor for the default inheritance strategy to use.
     * @return Default inheritance strategy to use.
     */
    public String getDefaultInheritanceStrategy()
    {
        return defaultInheritanceStrategy;
    }

    /**
     * Accessor for the MetaData for a class given the name and a loader.
     * If the class can't be loaded, null will be returned. 
     * @param class_name Name of the class to find MetaData for
     * @param clr ClassLoader resolver for use in loading the class.
     * @return The ClassMetaData for this class (or null if not found)
     **/
    public ClassMetaData getMetaDataForClass(String class_name,
                                             ClassLoaderResolver clr)
    {
        if (class_name == null)
        {
            return null;
        }

        // Check if this class has no MetaData before instantiating its class
        if (isClassWithoutMetaData(class_name))
        {
            return null;
        }

        Class c=null;
        try
        {
            if (clr == null)
            {
                c = Class.forName(class_name);
            }
            else
            {
                c = clr.classForName(class_name,null);
            }
        }
        catch (ClassNotFoundException cnfe)
        {
        }

        if (c == null)
        {
            return null;
        }

        return getMetaDataForClass(c);
    }

    /**
     * Accessor for the MetaData for a class.
     * Runs through the following process
     * <OL>
     * <LI>Checks if the class is known not to have metata</LI>
     * <LI>Check if we have metadata for the class in one of the files that has
     * been parsed.</LI>
     * <LI>If we have metadata, check that it is initialised</LI>
     * <LI>If we don't have metadata, find the file for this class.</LI>
     * <LI>If we cant find a file for it, add it to the list of classes known
     * to have no metadata</LI>
     * </OL>
     * @param c The class to find MetaData for
     * @return The ClassMetaData for this class (or null if not found)
     **/
    public ClassMetaData getMetaDataForClass(Class c)
    {
        if (c == null)
        {
            return null;
        }

        // If we know that this class has no MetaData don't bother searching
        if (isClassWithoutMetaData(c.getName()))
        {
            return null;
        }

        ClassMetaData the_cmd=null; 

        // Check if we have the MetaData already and it is initialised
        Object obj=classMetaDataByClass.get(c.getName());
        if (obj != null)
        {
            the_cmd = (ClassMetaData)obj;

            // Initialise this classes metadata if necessary
            if( !the_cmd.isInitialised() )
            {
                // Initialise the MetaData for the class
                initialiseMetaDataForClass(the_cmd, c);
            }

            return the_cmd;
        }

        // Search the valid locations for MetaData for a file with this class
        StringTokenizer token = new StringTokenizer(jdoFileSuffix,",");
        URL file_url = null;
        while( token.hasMoreTokens() && file_url == null)
        {
	        file_url=MetaDataUtils.getMetaDataFileForClass(c, null, token.nextToken());
        }
        if (file_url != null)
        {
            FileMetaData fmd = (FileMetaData) fileMetaDataByURL.get(file_url.toString());

            if( fmd == null )
            {
	            // Since we have no MetaData for the class, we haven't parsed this before
	            // so parse the file, validating it against the DTD
	            fmd = parseFile(file_url,true);
	            fileMetaDataByURL.put(file_url.toString(),fmd);
            }
            
            // Add all queries for this file
            // Store queries against "queryname"
            for (int i=0;i<fmd.getNoOfQueries();i++)
            {
                QueryMetaData qmd=fmd.getQuery(i);
                queryMetaDataByClassQuery.put(qmd.getName(),qmd);
            }

            // Add all classes and queries into the respective lookup maps
            for (int i = 0; i < fmd.getNoOfPackages(); i++)
            {
                PackageMetaData pmd = fmd.getPackage(i);
                for (int j = 0; j < pmd.getNoOfClasses(); j++)
                {
                    ClassMetaData cmd = pmd.getClass(j);
                    classMetaDataByClass.put(cmd.getFullClassName(), cmd);

                    // Store queries against "classname_queryname"
                    for (int k = 0; k < cmd.getNoOfQueries(); k++)
                    {
                        QueryMetaData qmd = cmd.getQuery(k);
                        queryMetaDataByClassQuery.put(cmd.getFullClassName() + "_" + qmd.getName(), qmd);
                    }

                    // Store the classes MetaData while we have it
                    if (cmd.getFullClassName().equals(c.getName()))
                    {
                        the_cmd = cmd;
                    }
                }
            }
            
            // We've found the class in this file so initialise the class's MetaData
            if (the_cmd != null)
            {
                // Initialise the MetaData for the class
                initialiseMetaDataForClass(the_cmd, c);

                return the_cmd;
            }
        }

        // Not found, so add to known classes without MetaData
        addClassWithoutMetaData(c.getName());

        return null;
    }

    /**
     * Utility to parse a file.
     * This is plit out as a separate method to allow overriding of the parser
     * used to read the XML file. The enhancer, for example, uses its own
     * extension to MetaDataParser.
     * @param file_url URL of the file
     * @param validate Whether to validate the file
     * @return The FileMetaData for this file
     */
    protected FileMetaData parseFile(URL file_url, boolean validate)
    {
        return MetaDataParser.parseMetaData(file_url,true,this);
    }

    /**
     * Load up and add the OR mapping info for the specified class to the stored MetaData.
     * @param c The class
     */
    public void addORMDataToClass(Class c)
    {
        // Get the basic classMetaData for this class
        ClassMetaData cmd = (ClassMetaData)classMetaDataByClass.get(c.getName());
        
        // See if we already have a file with the ORM metadata for this class
        ClassMetaData orm_cmd = (ClassMetaData)ormClassMetaDataByClass.get(c.getName());
        if (orm_cmd != null)
        {
            cmd.mergeORMData(orm_cmd);
            
            // Remove it from the map since now used.
            ormClassMetaDataByClass.remove(c.getName());
            
            return;
        }
        
        // Check for an ORM MetaData file for this class
        URL orm_file_url = MetaDataUtils.getMetaDataFileForClass(c, ormMappingName, ormFileSuffix);
        if (orm_file_url != null)
        {
            FileMetaData fmd_orm = (FileMetaData) fileMetaDataByURL.get(orm_file_url.toString());

            if( fmd_orm == null )
            {
	            // Since we have no MetaData for the class, we haven't parsed this before
	            // so parse the file, validating it against the DTD
                fmd_orm = parseFile(orm_file_url,true);
	            fileMetaDataByURL.put(orm_file_url.toString(),fmd_orm);
            }
            
            // Check all ORM ClassMetaData specified in this file
            for (int i=0;i<fmd_orm.getNoOfPackages();i++)
            {
                PackageMetaData pmd_orm = fmd_orm.getPackage(i);
                for (int j=0;j<pmd_orm.getNoOfClasses();j++)
                {
                    ClassMetaData cmd_orm = pmd_orm.getClass(j);
                    
                    // Use the data if for the required class
                    if (cmd_orm.getFullClassName().equals(c.getName()))
                    {
                        cmd.mergeORMData(cmd_orm);
                    }
                    // Else save it for later
                    else
                    {
                        ormClassMetaDataByClass.put(cmd_orm.getFullClassName(),cmd_orm);
                    }
                }
            }
        }
    }

    /**
     * Accessor for the MetaData for an interface. Since interfaces don't
     * currently have their own MetaData, this finds an implementation of this
     * interface and returns the MetaData for that class.
     * @param intf The interface to find MetaData for
     * @param value Object of a class, to return if possible (null=ignore)
     * @return The ClassMetaData for an implementation of the interface
     */
    public ClassMetaData getMetaDataForInterface(Class intf,
                                                 Object value)
    {
        if (intf == null)
        {
            return null;
        }

        ClassMetaData cmd=null;
        if (intf.isInterface())
        {
            // Search for the package/class required
            Set classMetaDataClasses=classMetaDataByClass.keySet();
            Iterator classMetaDataClassesIter=classMetaDataClasses.iterator();
            
            while (classMetaDataClassesIter.hasNext())
            {
                String class_name = (String)classMetaDataClassesIter.next();
                ClassMetaData cmd_cls = (ClassMetaData)classMetaDataByClass.get(class_name);

                try
                {
                    // Check if class is implementation of "intf"
                    Class cls = intf.getClassLoader().loadClass(class_name);
                    if (intf.isAssignableFrom(cls))
                    {
                        // Find the base class that is an implementation
                        cmd = cmd_cls;

                        if (value != null && cmd.getFullClassName().equals(value.getClass().getName()))
                        {
                            return cmd;
                        }

                        ClassMetaData cmd_superclass = cmd.getSuperclassMetaData();
                        while (cmd_superclass != null)
                        {
                            // TODO Check if superclass an implement'n
                            cmd = cmd_superclass;

                            if (value != null && cmd.getFullClassName().equals(value.getClass().getName()))
                            {
                                break;
                            }

                            // Go to next superclass
                            cmd_superclass = cmd_superclass.getSuperclassMetaData();
                            if (cmd_superclass == null)
                            {
                                break;
                            }
                        }
                    }
                }
                catch (Exception e)
                {
                }
            }
        }
        else
        {
            // TODO Return the CMD for this class
        }

        return cmd;
    }

    /**
     * Accessor for the MetaData for a field of a class. Utilises
     * getMetaDataForClass, and then finds the relevant field.
     * @param class_name The name of the class owning the field
     * @param field_name The name of the field to find MetaData for
     * @param clr ClassLoader resolver for any loading of classes
     * @return The FieldMetaData for this field (or null if not found)
     */
    public FieldMetaData getMetaDataForField(String class_name,
                                             String field_name,
                                             ClassLoaderResolver clr)
    {
        if (class_name == null || field_name == null)
        {
            return null;
        }

        ClassMetaData cmd=getMetaDataForClass(class_name,clr);
        if (cmd == null)
        {
            return null;
        }
        else
        {
            return cmd.getField(field_name);
        }
    }

    /**
     * Accessor for the MetaData for a field of a class.
     * Utilises getMetaDataForClass, and then finds the relevant field.
     * @param c The class owning the field
     * @param field_name The name of the field to find MetaData for
     * @return The FieldMetaData for this field (or null if not found)
     **/
    public FieldMetaData getMetaDataForField(Class c,String field_name)
    {
        if (c == null || field_name == null)
        {
            return null;
        }

        ClassMetaData cmd=getMetaDataForClass(c);
        if (cmd == null)
        {
            return null;
        }
        else
        {
            return cmd.getField(field_name);
        }
    }

    /**
     * Accessor for the MetaData for a named query for a class.
     * If the class is not specified, searches for the query with this name for
     * any class.
     * @param cls The class which has the query defined for it
     * @param query_name Name of the query
     * @return The QueryMetaData for the query for this class
     **/
    public QueryMetaData getMetaDataForQuery(Class cls,String query_name)
    {
        if (query_name == null)
        {
            return null;
        }

        String query_key=query_name;
        if (cls != null)
        {
            query_key = cls.getName() + "_" + query_name;
        }
        Object obj=queryMetaDataByClassQuery.get(query_key);
        if (obj != null)
        {
            return (QueryMetaData)obj;
        }
        else
        {
            // If the class hasn't been loaded, load it
            ClassMetaData cmd=getMetaDataForClass(cls);
            if (cmd == null)
            {
                return null;
            }
            else
            {
                obj = queryMetaDataByClassQuery.get(query_key);
                if (obj != null)
                {
                    return (QueryMetaData)obj;
                }
                else
                {
                    return null;
                }
            }
        }
    }

    // ------------------------------- Utilities -------------------------------

    /**
     * Utility to initialise the MetaData for a class, using the specified
     * class. This assigns defaults to tags that haven't been assigned.
     * If the class that is being used to populate the MetaData is not
     * enhanced, this will throw a JDOUserException informing them of this. 
     * @param cmd The classes metadata
     * @param cls The class to use as a basis for initialisation
     */
    protected void initialiseMetaDataForClass(ClassMetaData cmd,
                                              Class cls)
    {
        if (cmd.getPersistenceModifier() == ClassPersistenceModifier.PERSISTENCE_CAPABLE &&
            !ClassUtils.isPersistenceCapableClass(cls))
        {
            throw new JDOUserException(LOCALISER.msg("MetaData.Class.NotEnhancedError", cls.getName()));
        }
        if (!cmd.isPopulated() && !cmd.isInitialised())
        {
            cmd.populate(cls.getClassLoader());
        }
        if (!cmd.isInitialised())
        {
            cmd.initialise();
        }
    }

    /**
     * Accessor for whether a class doesn't have MetaData.
     * @param class_name Name of the class
     * @return Whether it has no metadata.
     */
    public boolean isClassWithoutMetaData(String class_name)
    {
        if (class_name == null)
        {
            return true;
        }

        // Standard Java classes have no MetaData
        if (class_name.startsWith("java.lang") ||
            class_name.startsWith("java.io") ||
            class_name.startsWith("java.net") ||
            class_name.startsWith("java.util") ||
            class_name.startsWith("java.text") ||
            class_name.equals("javax.jdo.LongIdentity") ||
            class_name.equals("javax.jdo.IntIdentity") ||
            class_name.equals("javax.jdo.StringIdentity") ||
            class_name.equals("javax.jdo.CharIdentity") ||
            class_name.equals("javax.jdo.ByteIdentity") ||
            class_name.equals("javax.jdo.ShortIdentity"))
        {
            return true;
        }

        return classesWithoutMetaData.contains(class_name);
    }

    /**
     * Method to add a class that doesn't have MetaData.
     * @param class_name Name of class that has no metadata.
     */
    protected void addClassWithoutMetaData(String class_name)
    {
        if (class_name == null)
        {
            return;
        }

        if (JPOXLogger.METADATA.isDebugEnabled())
        {
            JPOXLogger.METADATA.debug(LOCALISER.msg("MetaData.RegisterClassWithoutMetaData",class_name)); 
        }
        classesWithoutMetaData.add(class_name);
    }

    // ------------------------------ Utilities --------------------------------

    /**
     * Utility to return all ClassMetaData that is referenced from the supplier
     * class.
     * @param cmd The origin class's MetaData.
     * @param dba_vendor_id The Vendor id of the database adapter in use.
     *                      (Used in handling "views" support)
     * @param clr ClassLoader resolver for loading any classes.
     * @return List of ClassMetaData referenced by the origin
     */
    public List getReferencedClassMetaData(ClassMetaData cmd,
                                           String dba_vendor_id,
                                           ClassLoaderResolver clr)
    {
        if (cmd == null)
        {
            return null;
        }

        List orderedCMDs=new ArrayList();
        Set referencedCMDs=new HashSet();

        // Use the ClassMetaData to tell us about its classes
        cmd.getReferencedClassMetaData(orderedCMDs,referencedCMDs,dba_vendor_id,clr);

        return orderedCMDs;
    }
}