/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
    ...
**********************************************************************/
package org.jpox.metadata;

import java.io.FileWriter;
import java.io.IOException;
import java.net.URL;
import java.util.ArrayList;
import java.util.Iterator;

import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;

/**
 * Utility to migrate MetaData from the JDO 1.0 standard (where many things are
 * in extension tags) to the JDO 2.0 standard.
 * <P>
 * Takes a list of MetaData files, and options for the generation.
 * </P>
 * TODO Add option on which DTD to use
 * TODO Add option on output location for updated file
 *
 * @version $Revision: 1.10 $ 
 */
public class MetaDataMigrator
{
    private static final Localiser LOCALISER=Localiser.getInstance("org.jpox.metadata.Localisation");

    /**
     * Main entry point for the migrator.
     * @param args The command line arguments.
     * @throws Exception Thrown when an error occurs. 
     */
    public static void main(String[] args)
    throws Exception
    {
        int return_code=0;

        // Print opening title
        System.out.println(LOCALISER.msg("MetaData.Migrator.Title"));
        System.out.println(LOCALISER.msg("MetaData.Migrator.Title.2"));

        // Process the command line arguments
        boolean verbose=false;
        ArrayList files=new ArrayList();
        String indent_value="4";
        for (int i=0;i<args.length;i++)
        {
            if (args[i].equals("-v"))
            {
                verbose = true;
            }
            else if (args[i].equals("-indent"))
            {
                indent_value = args[++i];
            }
            else if (args[i].equals("-help"))
            {
                System.out.println(LOCALISER.msg("MetaData.Migrator.Help.Usage"));
                System.out.println(LOCALISER.msg("MetaData.Migrator.Help.Options.0"));
                System.out.println(LOCALISER.msg("MetaData.Migrator.Help.Options.1"));
                System.out.println(LOCALISER.msg("MetaData.Migrator.Help.Options.2"));
                System.out.println(LOCALISER.msg("MetaData.Migrator.Help.Options.3"));
                System.exit(0);
            }
            else
            {
                files.add(args[i]);
            }
        }

        // Set up the indent string
        int indent_size=4;
        try
        {
            indent_size = (new Integer(indent_value)).intValue();
        }
        catch (Exception e)
        {
            indent_size = 4;
        }

        StringBuffer indent_str=new StringBuffer("");
        for (int i=0;i<indent_size;i++)
        {
            indent_str.append(" ");
        }

        // Print input details
        if (verbose)
        {
            System.out.println(LOCALISER.msg("MetaData.Migrator.IndentSize",indent_size));
        }

        // Check for erroneous or incomplete input
        if (files.isEmpty())
        {
            String error_msg=LOCALISER.msg("MetaData.Migrator.NoFiles.Error");
            System.err.println(error_msg);
            JPOXLogger.METADATA.error(error_msg);
        }

        // Process the files.
        Iterator iter=files.iterator();
        while (iter.hasNext())
        {
            String file=(String)iter.next();
            if (verbose)
            {
                System.out.print(LOCALISER.msg("MetaData.Migrator.ProcessingFile",file));
                if (JPOXLogger.METADATA.isDebugEnabled())
                {
                    JPOXLogger.METADATA.debug(LOCALISER.msg("MetaData.Migrator.ProcessingFile",file));
                }
            }

            // Parse the MetaData file
            URL file_url=new URL("file:" + file);
            FileMetaData filemd=MetaDataParser.parseMetaData(file_url,false,new MetaDataManager()); 

            // Convert the MetaData
            convertFileMetaData(filemd); 

            // Output the new file
            FileWriter writer=null; 
            try
            {
                writer = new FileWriter(file);

                // Write JDO DTD header
                // TODO Give options for which DTD to use 
                writer.write("<?xml version=\"1.0\"?>\n"); 
                writer.write("<!DOCTYPE jdo SYSTEM \"file:/javax/jdo/jdo.dtd\">\n");
 
                // Write the MetaData XML
                writer.write(filemd.toString(indent_str.toString()));
                if (JPOXLogger.METADATA.isDebugEnabled())
                {
                    JPOXLogger.METADATA.debug("Written file " + file);
                }
            }
            catch (IOException ioe)
            {
                JPOXLogger.METADATA.error("Error writing file " + file + " : " + ioe);
            }
            finally
            {
                if (writer != null)
                {
                    writer.close();
                }
            }

            if (verbose)
            {
                System.out.println(LOCALISER.msg("MetaData.Migrator.ProcessedFile"));
            }
        }

        System.exit(return_code);
    }

    /**
     * Method to convert JDO 1 style extension tags to JDO 2 tags.
     * @param filemd The File MetaData.
     */
    protected static void convertFileMetaData(FileMetaData filemd)
    {
        if (filemd == null)
        {
            return;
        }

        for (int i=0;i<filemd.getNoOfPackages();i++)
        {
            PackageMetaData pmd=filemd.getPackage(i);
            for (int j=0;j<pmd.getNoOfClasses();j++)
            {
                ClassMetaData cmd=pmd.getClass(j);

                // JPOX 1.0 used "new-table" always
                cmd.setInheritanceMetaData(new InheritanceMetaData(cmd,"new-table"));

                // Migrate "table-name"
                if (cmd.hasExtension("table-name"))
                {
                    cmd.setTable(cmd.getValueForExtension("table-name"));
                    cmd.removeExtension("table-name");
                    if (JPOXLogger.METADATA.isDebugEnabled())
                    {
                        JPOXLogger.METADATA.debug(LOCALISER.msg("MetaData.Migrator.ChangeExtensionTagToAttribute","table-name","table"));
                    }
                }

                JPOXLogger.METADATA.info("identity-type for class is " + cmd.getIdentityType());
                // Migrate any datastore identity tags
                if (cmd.getIdentityType() == null ||
                    cmd.getIdentityType() == IdentityType.DATASTORE)
                {
                    IdentityMetaData imd=null;

                    if (cmd.hasExtension("key-auto-increment"))
                    {
                        imd = new IdentityMetaData(cmd,null,"autoincrement",null);
                        if (JPOXLogger.METADATA.isDebugEnabled())
                        {
                            JPOXLogger.METADATA.debug(LOCALISER.msg("MetaData.Migrator.ChangeExtensionTagToElement","key-auto-increment","datastore-identity"));
                        }
                    }
                    else if (cmd.hasExtension("poid-class-generator") ||
                             cmd.hasExtension("use-poid-generator"))
                    {
                        String poid_class=cmd.getValueForExtension("poid-class-generator");
                        if (poid_class == null)
                        {
                        }
                        else if (poid_class.equals("org.jpox.poid.SequenceTablePoidGenerator"))
                        {
                            imd = new IdentityMetaData(cmd,null,"native",null);
                        }
                        else if (poid_class.equals("org.jpox.poid.MaxPoidGenerator"))
                        {
                            imd = new IdentityMetaData(cmd,null,"increment",null);
                        }
                        else if (poid_class.equals("org.jpox.poid.OracleSequencePoidGenerator") ||
                                 poid_class.equals("org.jpox.poid.PostgresqlSequencePoidGenerator") ||
                                 poid_class.equals("org.jpox.poid.SAPDBSequencePoidGenerator"))
                        {
                            String sequence=cmd.getValueForExtension("sequence-name");
                            imd = new IdentityMetaData(cmd,null,"sequence",sequence);
                        }
                        if (JPOXLogger.METADATA.isDebugEnabled())
                        {
                            JPOXLogger.METADATA.debug(LOCALISER.msg("MetaData.Migrator.ChangeExtensionTagToElement","poid-class-generator","datastore-identity"));
                        }
                    }
                    if (imd != null)
                    {
                        cmd.setIdentityMetaData(imd);
                    }

                    // Migrate PK column name
                    if (cmd.hasExtension("key-column-name"))
                    {
                        imd = cmd.getIdentityMetaData();
                        if (imd == null)
                        {
                            imd = new IdentityMetaData(cmd,null,null,null);
                            cmd.setIdentityMetaData(imd);
                        }

                        imd.addColumn(new ColumnMetaData(imd,cmd.getValueForExtension("key-column-name")));
                        if (JPOXLogger.METADATA.isDebugEnabled())
                        {
                            JPOXLogger.METADATA.debug(LOCALISER.msg("MetaData.Migrator.ChangeExtensionTagToElement","key-column-name","datastore-identity -> column"));
                        }
                    }

                    cmd.removeExtension("key-auto-increment");
                    cmd.removeExtension("poid-class-generator");
                    cmd.removeExtension("use-poid-generator");
                    cmd.removeExtension("sequence-name");
                    cmd.removeExtension("key-column-name");
                }

                // Migrate "optimistic-mode", "optimistic-column-name"
                if (cmd.hasExtension("optimistic-mode") ||
                    cmd.hasExtension("optimistic-column-name"))
                {
                    VersionMetaData vmd=cmd.getVersionMetaData();
                    if (vmd == null)
                    {
                        String mode_value=cmd.getValueForExtension("optimistic-mode");
                        String strategy=null;
                        if (mode_value == null)
                        {
                        }
                        else if (mode_value.toLowerCase().equals("version"))
                        {
                            strategy = "version-number";
                        }
                        else if (mode_value.toLowerCase().equals("timestamp"))
                        {
                            strategy = "timestamp";
                        }
                        vmd = new VersionMetaData(strategy, cmd.getValueForExtension("optimistic-column-name"), null);
                        cmd.setVersionMetaData(vmd);
                    }

                    cmd.removeExtension("optimistic-mode");
                    cmd.removeExtension("optimistic-column-name");
                }

                // Migrate field "column-name", "length", "scale", "precision"
                for (int k=0;k<cmd.getNoOfFields();k++)
                {
                    FieldMetaData fmd=cmd.getField(k);

                    ColumnMetaData[] colmds=fmd.getColumnMetaData();
                    ColumnMetaData colmd=null;
                    if (colmds != null && colmds.length > 0)
                    {
                        colmd = colmds[0];
                    }

                    // Migrate "column-name"
                    if (fmd.hasExtension("column-name"))
                    {
                        if (colmd == null)
                        {
                            colmd = new ColumnMetaData(fmd,null);
                            fmd.addColumn(colmd);
                        }
                        colmd.setName(fmd.getValueForExtension("column-name"));

                        fmd.removeExtension("column-name");
                        if (JPOXLogger.METADATA.isDebugEnabled())
                        {
                            JPOXLogger.METADATA.debug(LOCALISER.msg("MetaData.Migrator.ChangeExtensionTagToElement","column-name","column -> name"));
                        }
                    }

                    // Migrate "length" tag
                    if (fmd.hasExtension("length"))
                    {
                        if (colmd == null)
                        {
                            colmd = new ColumnMetaData(fmd,null);
                            fmd.addColumn(colmd);
                        }

                        String length_value=fmd.getValueForExtension("length");
                        if (length_value.startsWith("max"))
                        {
                            colmd.setLength(Integer.parseInt(length_value.substring(3).trim()));
                            colmd.setJdbcType("VARCHAR");
                        }
                        else if (length_value.toLowerCase().equals("unlimited"))
                        {
                            colmd.setJdbcType("BLOB");
                        }
                        else
                        {
                            colmd.setLength(Integer.parseInt(length_value));
                            colmd.setJdbcType("CHAR");
                        }

                        fmd.removeExtension("length");
                        if (JPOXLogger.METADATA.isDebugEnabled())
                        {
                            JPOXLogger.METADATA.debug(LOCALISER.msg("MetaData.Migrator.ChangeExtensionTagToElement","length","column -> length"));
                        }
                    }

                    // Migrate "scale" tag
                    if (fmd.hasExtension("scale"))
                    {
                        if (colmd == null)
                        {
                            colmd = new ColumnMetaData(fmd,null);
                            fmd.addColumn(colmd);
                        }

                        String scale_value=fmd.getValueForExtension("scale");
                        colmd.setScale(Integer.parseInt(scale_value));

                        fmd.removeExtension("scale");
                        if (JPOXLogger.METADATA.isDebugEnabled())
                        {
                            JPOXLogger.METADATA.debug(LOCALISER.msg("MetaData.Migrator.ChangeExtensionTagToElement","scale","column -> scale"));
                        }
                    }

                    // Migrate "precision" tag
                    if (fmd.hasExtension("precision"))
                    {
                        if (colmd == null)
                        {
                            colmd = new ColumnMetaData(fmd,null);
                            fmd.addColumn(colmd);
                        }

                        String precision_value=fmd.getValueForExtension("precision");
                        if (precision_value.startsWith("min"))
                        {
                            colmd.setLength(Integer.parseInt(precision_value.substring(3).trim()));
                            colmd.setJdbcType("DECIMAL");
                        }
                        else
                        {
                            colmd.setPrecision(Integer.parseInt(precision_value));
                        }

                        fmd.removeExtension("precision");
                        if (JPOXLogger.METADATA.isDebugEnabled())
                        {
                            JPOXLogger.METADATA.debug(LOCALISER.msg("MetaData.Migrator.ChangeExtensionTagToElement","precision","column -> precision"));
                        }
                    }

                    if (fmd.hasExtension("key-auto-increment"))
                    {
                        if(fmd.getValueForExtension("key-auto-increment").equals("true"))
                        {
                            fmd.setValueStrategy(IdentityStrategy.AUTOASSIGN);
                        }
                        fmd.removeExtension("key-auto-increment");
                        if (JPOXLogger.METADATA.isDebugEnabled())
                        {
                            JPOXLogger.METADATA.debug(LOCALISER.msg("MetaData.Migrator.ChangeExtensionTagToElement","key-auto-increment","value-strategy"));
                        }
                    }
                    
                    // Process collection tags
                    CollectionMetaData collmd=fmd.getCollection();
                    if (collmd != null)
                    {
                        // Normal Relationship
                        if (!collmd.hasExtension("owner-field"))
                        {
                            if (fmd.getJoinMetaData() == null)
                            {
                                fmd.setJoinMetaData(new JoinMetaData(null,null,null,null));
                            }
                        }
                        // Inverse Bidirectional Relationship
                        else
                        {
                            String owner_field_value=collmd.getValueForExtension("owner-field");
                            if (fmd.getMappedBy() == null)
                            {
                                fmd.setMappedBy(owner_field_value);
                            }

                            collmd.removeExtension("owner-field");
                        }

                        // Normal relationship with specified table-name
                        if (collmd.hasExtension("table-name"))
                        {
                            fmd.setTable(collmd.getValueForExtension("table-name"));
                            collmd.removeExtension("table-name");
                        }

                        // Normal relationship with specified owner-column-name
                        if (collmd.hasExtension("owner-column-name"))
                        {
                            JoinMetaData jmd=fmd.getJoinMetaData();
                            if (jmd == null)
                            {
                                jmd = new JoinMetaData(null,null,null,null);
                                fmd.setJoinMetaData(jmd);
                            }
                            jmd.addColumn(new ColumnMetaData(jmd,collmd.getValueForExtension("owner-column-name")));
                            collmd.removeExtension("owner-column-name");
                        }
                        
                        // Normal relationship with specified element-column-name
                        if (collmd.hasExtension("element-column-name"))
                        {
                            ElementMetaData emd=fmd.getElementMetaData();
                            if (emd == null)
                            {
                                emd = new ElementMetaData(null, null, null, null, null, null, null);
                                fmd.setElementMetaData(emd);
                            }
                            emd.addColumn(new ColumnMetaData(emd,collmd.getValueForExtension("element-column-name")));
                            collmd.removeExtension("element-column-name");
                        }

                        // Cascade Delete
                        String clear_on_delete_value=collmd.getValueForExtension("clear-on-delete");
                        if (clear_on_delete_value == null ||
                            clear_on_delete_value.equalsIgnoreCase("true"))
                        {
                            fmd.setDependent(true);
                        }
                        else
                        {
                            fmd.setDependent(false);
                        }
                        collmd.removeExtension("clear-on-delete");
                    }

                    // Process map tags
                    MapMetaData mapmd=fmd.getMap();
                    if (mapmd != null)
                    {
                        // Normal Relationship
                        if (!mapmd.hasExtension("owner-field"))
                        {
                            if (fmd.getJoinMetaData() == null)
                            {
                                fmd.setJoinMetaData(new JoinMetaData(null,null,null,null));
                            }
                        }
                        // Inverse Bidirectional Relationship
                        else
                        {
                            String owner_field_value=mapmd.getValueForExtension("owner-field");
                            if (fmd.getMappedBy() == null)
                            {
                                fmd.setMappedBy(owner_field_value);
                            }

                            mapmd.removeExtension("owner-field");
                        }
                    }

                    // Inverse Bidirectional Relationship
                    if (fmd.hasExtension("collection-field"))
                    {
                        fmd.removeExtension("collection-field");
                    }
                    if (fmd.hasExtension("list-field"))
                    {
                        fmd.removeExtension("list-field");
                    }
                    if (fmd.hasExtension("map-field"))
                    {
                        fmd.removeExtension("map-field");
                    }
                }
            }
        }
    }
}