/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved. 
This program and the accompanying materials are made available 
under the terms of the JPOX License v1.0 which accompanies this distribution.

Contributors:
2004 Kikuchi Kousuke - org.jpox.enhancer.conf.MetaDataParser
2004 Andy Jefferson - added class and field custom object creation facility
2004 Andy Jefferson - updated resolve entity to use Ralf Ullrich suggestion
2004 Marco Schulze (NightLabs.de) - added safety checks for missing local dtd files
2004 Marco Schulze (NightLabs.de) - added special handling of SAXException
    ...
**********************************************************************/
package org.jpox.metadata;

import java.io.File;
import java.io.FileInputStream;
import java.io.InputStream;
import java.net.URL;
import java.util.Stack;

import javax.jdo.JDOException;
import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;

import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.SAXParseException;
import org.xml.sax.helpers.DefaultHandler;

import org.jpox.util.Localiser;
import org.jpox.util.JPOXLogger;

/**
 * Class to parse a MetaData file and return a FileMetaData object.
 * This class is invoked by either of
 * <PRE>
 * MetaDataParser.parseMetaData(url,validate);
 * MetaDataParser.parseMetaData(input_str,validate);
 * </PRE>
 * returning the FileMetaData. This class simply constructs the MetaData
 * representation mirroring what is in the MetaData file. It has no knowledge
 * of the class(es) that it represents, simply the information in the MetaData
 * file. The knowledge of the classes is imposed on the representation at a
 * later stage where necessary.
 * <P>Operates the parse process using a Stack. MetaData components are added
 * to the stack as they are encountered and created. They are then popped off
 * the stack when the end element is encountered.</P>
 *
 * @since 1.1
 * @version $Revision: 1.30 $
 */
public class MetaDataParser extends DefaultHandler
{
    protected static Localiser LOCALISER=Localiser.getInstance("org.jpox.metadata.Localisation");

    private JDOEntityResolver entityResolver=new JDOEntityResolver();
    
    /**
     * Method to parse a JDO MetaData file given the URL of the file.
     * @param url Url of the file
     * @param validate Whether to validate against the DTD as well
     * @param mdm the metadata manager
     * @return The MetaData for this file
     * @throws JDOException thrown if error occurred
     */
    public static FileMetaData parseMetaData(URL url,boolean validate,MetaDataManager mdm)
    {
        if (url == null)
        {
            String msg=LOCALISER.msg("MetaData.Parser.NullFile");
            JPOXLogger.METADATA.error(msg);
            throw new JDOException(msg);
        }

        JPOXLogger.METADATA.info(LOCALISER.msg("MetaData.Parser.ParseFile",url.getFile()));
        try
        {
            InputStream in = null;
            try
            {
                in = url.openStream();
            }
            catch (Exception ignore)
            {
            }
            if (in == null)
            {
                try
                {
                    in = new FileInputStream(new File(url.getFile()));
                }
                catch (Exception ignore)
                {
                }
            }
            if (in == null)
            {
                JPOXLogger.METADATA.error(LOCALISER.msg("MetaData.Parser.CannotReadFile.Error",url.getFile()));
                throw new JDOException(LOCALISER.msg("MetaData.Parser.CannotReadFile.Error",url.getFile()));
            }

            // Parse the file, and set its filename
            FileMetaData filemd=parseMetaData(in,validate,mdm);
            filemd.setFilename(url.getFile());

            return filemd;
        }
        catch (Exception e)
        {
            String message = LOCALISER.msg("MetaData.Parser.CannotReadFile.Error", url.getFile());
            JPOXLogger.METADATA.error(message, e);

            if (e instanceof SAXException)
            {
                SAXException se = (SAXException)e;
                Throwable cause = se.getException();
                if (cause == null)
                {
          	        cause = e.getCause();
                }
                JPOXLogger.METADATA.error(LOCALISER.msg("MetaData.Parser.ErrorCause",url.getFile(),cause));
            }

            throw new JDOException(message);
        }
    }

    /**
     * Method to parse a JDO MetaData file given an InputStream.
     * @param in input stream
     * @param validate Whether to validate the MetaData
     * @param mdm the metadata manager
     * @return The MetaData for this file
     * @throws JDOException thrown if error occurred
     */
    public static FileMetaData parseMetaData(InputStream in,boolean validate,MetaDataManager mdm)
    {
        if (in == null)
        {
            throw new NullPointerException("input stream is null");
        }
        try
        {
            SAXParserFactory factory = SAXParserFactory.newInstance();
            factory.setValidating(validate);
            SAXParser parser = factory.newSAXParser();
            MetaDataParser handler = new MetaDataParser(mdm);

            parser.parse(in, handler);
            return handler.getMetaData();
        }
        catch (InvalidMetaDataException e)
        {
            throw e;
        }
        catch (Exception e)
        {
            String message = LOCALISER.msg("MetaData.Parser.CannotReadFile.Error",
                            "<input stream> " + e.getMessage());
            JPOXLogger.METADATA.error(message, e);
            throw new JDOException(message);
        }
        finally
        {
            try
            {
                in.close();
            }
            catch (Exception ignore)
            {
                // Do nothing
            }
        }
    }

    /**
     * parser buffer
     */
    private StringBuffer charactersBuffer = new StringBuffer();

    /**
     * Stack of meta-data elements. The top of the stack is always the element
     * being process at the moment. The elements are not the XML Element type
     * but are things like ClassMetaData, PackageMetaData etc.
     */
    protected Stack stack = new Stack();

    /**
     * Latest element. This only keeps track of the element just started.
     * Its useful where we have an element with no subelements, where it is set
     * in the startElement() and used in characters() so we can extract the
     * value.
     */
    protected String latestElement = null;

    /**
     * The MetaData for the file.
     */
    protected final FileMetaData metadata;

    /**
     * Constructor. Protected to prevent instantiation.
     */
    private MetaDataParser()
    {
        metadata = null;
    }


    /**
     * Constructor. Protected to prevent instantiation.
     * @param mdm the metadata manager
     */
    protected MetaDataParser(MetaDataManager mdm)
    {
        metadata = new FileMetaData(null,mdm);
        pushStack(metadata);
    }   
    
    /**
     * Accessor for the MetaData for this file.
     * @return The MetaData.
     */
    protected FileMetaData getMetaData()
    {
        return metadata;
    }

    /**
     * Handler method called at the start of the XML document.
     */
    public void startDocument()
    throws SAXException
    {
        super.startDocument();
    }

    /**
     * Handler method called at the end of the XML document.
     */
    public void endDocument()
    throws SAXException
    {
        super.endDocument();
    }

    /**
     * Utility to create a new class component.
     * @param pmd The parent PackageMetaData
     * @param attrs The attributes
     * @return The ClassMetaData
     */
    protected ClassMetaData newClassObject(PackageMetaData pmd,Attributes attrs)
    {
        ClassMetaData cmd = new ClassMetaData(pmd,
                            getAttr(attrs,"name"),
                            getAttr(attrs,"identity-type"),
                            getAttr(attrs,"objectid-class"),
                            getAttr(attrs,"requires-extent"),
                            getAttr(attrs,"detachable"),
                            getAttr(attrs,"embedded-only"),
                            getAttr(attrs,"persistence-modifier"),
                            getAttr(attrs,"persistence-capable-superclass"),
                            getAttr(attrs,"table"));
        return cmd;
    }

    /**
     * Utility to create a new field component.
     * @param md The parent MetaData
     * @param attrs The attributes
     * @return The FieldMetaData
     */
    protected FieldMetaData newFieldObject(MetaData md,Attributes attrs)
    {
        FieldMetaData fmd = new FieldMetaData(md,
                            getAttr(attrs,"name"),
                            getAttr(attrs,"primary-key"),
                            getAttr(attrs,"persistence-modifier"),
                            getAttr(attrs,"default-fetch-group"),
                            getAttr(attrs,"null-value"),
                            getAttr(attrs,"embedded"),
                            getAttr(attrs,"serialized"),
                            getAttr(attrs,"dependent"),
                            getAttr(attrs,"mapped-by"),
                            getAttr(attrs,"column"),
                            getAttr(attrs,"table"),
                            getAttr(attrs,"delete-action"),
                            getAttr(attrs,"indexed"),
                            getAttr(attrs,"unique"),
                            getAttr(attrs,"fetch-depth"),
                            getAttr(attrs,"load-fetch-group"),
                            getAttr(attrs,"value-strategy"),
                            getAttr(attrs,"sequence"));
        return fmd;
    }

    /**
     * Handler method called at the start of an element.
     * @param uri URI of the tag
     * @param localName Local name
     * @param qName Element name
     * @param attrs Attributes for this element 
     */
    public void startElement(String uri, 
                             String localName,
                             String qName,
                             Attributes attrs)
    throws SAXException 
    {
        latestElement = qName;

        if (JPOXLogger.METADATA.isDebugEnabled())
        {
            StringBuffer sb = new StringBuffer();
            sb.append(qName + " (stack size=" + stack.size() + ") Attributes[");
            for (int i=0; i<attrs.getLength(); i++)
            {
                if (i != 0)
                {
                    sb.append(" ,");
                }
                sb.append(attrs.getQName(i)).append("=\"").append(attrs.getValue(i)).append("\"");
            }
            sb.append("]");
            JPOXLogger.METADATA.debug(LOCALISER.msg("MetaData.Parser.StartElement",sb.toString()));
        }

        // New JDO file
        if (qName.equals("jdo"))
        {
            // Do nothing
        }
        // New JDO O/R mapping file
        else if (qName.equals("orm"))
        {
            // Do nothing
        }
        // New package for this file
        else if (qName.equals("package"))
        {
            FileMetaData filemd = (FileMetaData)getStack();
            PackageMetaData pmd = new PackageMetaData(filemd,attrs.getValue("name"));
            metadata.addPackage(pmd);
            pushStack(pmd);
        }
        // New class for this package
        else if (qName.equals("class"))
        {
            PackageMetaData pmd=(PackageMetaData)getStack();
            ClassMetaData cmd=newClassObject(pmd,attrs);
            pmd.addClass(cmd);

            pushStack(cmd);
        }
        // New interface for this package
        else if (qName.equals("interface"))
        {
            PackageMetaData pmd = (PackageMetaData)getStack();
            InterfaceMetaData imd = new InterfaceMetaData(pmd,
                                              getAttr(attrs,"name"),
                                              getAttr(attrs,"requires-extent"));
            pmd.addInterface(imd);
            pushStack(imd);
        }
        // New implements for this class
        else if (qName.equals("implements"))
        {
            ClassMetaData cmd = (ClassMetaData)getStack();
            ImplementsMetaData imd = 
                new ImplementsMetaData(cmd, getAttr(attrs,"name"));
            cmd.addImplements(imd);
            pushStack(imd);
        }
        // New property for this interface
        else if (qName.equals("property"))
        {
            InterfaceMetaData intmd = (InterfaceMetaData)getStack();
            PropertyMetaData propmd = new PropertyMetaData(intmd,
                                              getAttr(attrs,"name"),
                                              getAttr(attrs,"column"));
            intmd.addProperty(propmd);
            pushStack(propmd);
        }
        // New datastore-identity for this class
        else if (qName.equals("datastore-identity"))
        {
            ClassMetaData cmd = (ClassMetaData)getStack();
            IdentityMetaData imd = new IdentityMetaData(cmd,
                            getAttr(attrs,"column"),
                            getAttr(attrs,"strategy"),
                            getAttr(attrs,"sequence"));
            cmd.setIdentityMetaData(imd);
            pushStack(imd);
        }
        // New inheritance for this class
        else if (qName.equals("inheritance"))
        {
            ClassMetaData cmd = (ClassMetaData)getStack();
            InheritanceMetaData imd = new InheritanceMetaData(cmd,
                            getAttr(attrs,"strategy"));
            cmd.setInheritanceMetaData(imd);
            pushStack(imd);
        }
        // Discriminator for this inheritance
        else if (qName.equals("discriminator"))
        {
            InheritanceMetaData imd = (InheritanceMetaData)getStack();
            DiscriminatorMetaData dismd = new DiscriminatorMetaData(imd,
                            getAttr(attrs,"column"),
                            getAttr(attrs,"value"),
                            getAttr(attrs,"strategy"),
                            getAttr(attrs,"indexed"));
            imd.setDiscriminatorMetaData(dismd);
            pushStack(dismd);
        }
        // New query for this class
        else if (qName.equals("query"))
        {
            ExtendableMetaData emd = (ExtendableMetaData)getStack();
            QueryMetaData qmd = new QueryMetaData(emd,
                            getAttr(attrs,"name"),
                            getAttr(attrs,"language"),
                            getAttr(attrs,"unmodifiable"));
            if (emd instanceof ClassMetaData)
            {
                ClassMetaData cmd = (ClassMetaData)emd;
                cmd.addQuery(qmd);
            }
            else if (emd instanceof FileMetaData)
            {
                FileMetaData fmd = (FileMetaData)emd;
                fmd.addQuery(qmd);
            }
            pushStack(qmd);
        }
        // Unmapped section for this class
        else if (qName.equals("unmapped"))
        {
            ExtendableMetaData emd = (ExtendableMetaData)getStack();
            UnmappedMetaData unmd = new UnmappedMetaData(getAttr(attrs,"column"));
            if (emd instanceof ClassMetaData)
            {
                ClassMetaData cmd = (ClassMetaData)emd;
                cmd.setUnmappedMetaData(unmd);
            }
            pushStack(unmd);
        }
        // New field for this class
        else if (qName.equals("field"))
        {
            MetaData md = getStack();
            FieldMetaData fmd = newFieldObject(md,attrs);
            if (md instanceof ClassMetaData)
            {
                ClassMetaData cmd = (ClassMetaData)md;
                cmd.addField(fmd);
            }
            else if (md instanceof FetchGroupMetaData)
            {
                FetchGroupMetaData fgmd = (FetchGroupMetaData)md;
                fgmd.addField(fmd);
            }
            else if (md instanceof EmbeddedMetaData)
            {
                EmbeddedMetaData emd = (EmbeddedMetaData)md;
                emd.addField(fmd);
            }
            else if (md instanceof ForeignKeyMetaData)
            {
                ForeignKeyMetaData fkmd = (ForeignKeyMetaData)md;
                fkmd.addField(fmd);
            }
            else if (md instanceof IndexMetaData)
            {
                IndexMetaData imd = (IndexMetaData)md;
                imd.addField(fmd);
            }
            else if (md instanceof UniqueMetaData)
            {
                UniqueMetaData umd = (UniqueMetaData)md;
                umd.addField(fmd);
            }
            pushStack(fmd);
        }
        // New join
        else if (qName.equals("join"))
        {
            MetaData parent = getStack();
            JoinMetaData jnmd = new JoinMetaData(getAttr(attrs,"table"),
                                               getAttr(attrs,"column"),
                                               getAttr(attrs,"outer"),
                                               getAttr(attrs,"foreign-key"));
            if (parent instanceof FieldMetaData)
            {
                FieldMetaData fmd = (FieldMetaData)parent;
                fmd.setJoinMetaData(jnmd);
            }
            else if (parent instanceof ClassMetaData)
            {
                ClassMetaData cmd = (ClassMetaData)parent;
                cmd.addJoin(jnmd);
            }
            else if (parent instanceof InheritanceMetaData)
            {
                InheritanceMetaData imd = (InheritanceMetaData)parent;
                imd.setJoinMetaData(jnmd);
            }
            pushStack(jnmd);
        }
        // New map container for this field
        else if (qName.equals("map"))
        {
            FieldMetaData fmd = (FieldMetaData)getStack();
            MapMetaData map = new MapMetaData(fmd,
                            getAttr(attrs,"key-type"),
                            getAttr(attrs,"embedded-key"),
                            getAttr(attrs,"value-type"),
                            getAttr(attrs,"embedded-value"),
                            getAttr(attrs,"dependent-value"),
                            getAttr(attrs,"dependent-key"));
            fmd.setContainer(map);
            pushStack(map);
        }
        // New array container for this field
        else if (qName.equals("array"))
        {
            FieldMetaData fmd = (FieldMetaData)getStack();
            ArrayMetaData amd = new ArrayMetaData(fmd,
                            getAttr(attrs,"embedded-element"));
            fmd.setContainer(amd);
            pushStack(amd);
        }
        // New collection container for this field
        else if (qName.equals("collection"))
        {
            FieldMetaData fmd = (FieldMetaData)getStack();
            CollectionMetaData colmd = new CollectionMetaData(fmd,
                            getAttr(attrs,"element-type"),
                            getAttr(attrs,"embedded-element"),
                            getAttr(attrs,"dependent-element"));
            fmd.setContainer(colmd);
            pushStack(colmd);
        }
        // New column
        else if (qName.equals("column"))
        {
            ExtendableMetaData extmd = (ExtendableMetaData)getStack();
            ColumnMetaData clnmd = new ColumnMetaData(extmd,
                            getAttr(attrs,"name"),
                            getAttr(attrs,"target"),
                            getAttr(attrs,"target-field"),
                            getAttr(attrs,"jdbc-type"),
                            getAttr(attrs,"sql-type"),
                            getAttr(attrs,"length"),
                            getAttr(attrs,"scale"),
                            getAttr(attrs,"precision"),
                            getAttr(attrs,"allows-null"),
                            getAttr(attrs,"default-value"));
            if (extmd instanceof FieldMetaData)
            {
                FieldMetaData fmd = (FieldMetaData)extmd;
                fmd.addColumn(clnmd);
            }
            else if (extmd instanceof AbstractElementMetaData)
            {
                AbstractElementMetaData elemd = (AbstractElementMetaData)extmd;
                elemd.addColumn(clnmd);
            }
            else if (extmd instanceof JoinMetaData)
            {
                JoinMetaData jnmd = (JoinMetaData)extmd;
                jnmd.addColumn(clnmd);
            }
            else if (extmd instanceof IdentityMetaData)
            {
                IdentityMetaData idmd = (IdentityMetaData)extmd;
                idmd.addColumn(clnmd);
            }
            else if (extmd instanceof ForeignKeyMetaData)
            {
                ForeignKeyMetaData fkmd = (ForeignKeyMetaData)extmd;
                fkmd.addColumn(clnmd);
            }
            else if (extmd instanceof IndexMetaData)
            {
                IndexMetaData idxmd = (IndexMetaData)extmd;
                idxmd.addColumn(clnmd);
            }
            else if (extmd instanceof UniqueMetaData)
            {
                UniqueMetaData unimd = (UniqueMetaData)extmd;
                unimd.addColumn(clnmd);
            }
            else if (extmd instanceof OrderMetaData)
            {
                OrderMetaData ormd = (OrderMetaData)extmd;
                ormd.addColumn(clnmd);
            }
            else if (extmd instanceof DiscriminatorMetaData)
            {
                DiscriminatorMetaData dismd = (DiscriminatorMetaData)extmd;
                dismd.setColumnMetaData(clnmd);
            }
            else if (extmd instanceof VersionMetaData)
            {
                VersionMetaData vermd = (VersionMetaData)extmd;
                vermd.addColumn(clnmd);
            }
            pushStack(clnmd);
        }
        // New element
        else if (qName.equals("element"))
        {
            ElementMetaData elemd = new ElementMetaData(
                            getAttr(attrs,"column"),
                            getAttr(attrs,"embedded"),
                            getAttr(attrs,"serialized"),
                            getAttr(attrs,"delete-action"),
                            getAttr(attrs,"update-action"),
                            getAttr(attrs,"indexed"),
                            getAttr(attrs,"unique"));
            FieldMetaData fmd = (FieldMetaData)getStack();
            fmd.setElementMetaData(elemd);
            pushStack(elemd);
        }
        // New key
        else if (qName.equals("key"))
        {
            KeyMetaData keymd = new KeyMetaData(
                            getAttr(attrs,"column"),
                            getAttr(attrs,"embedded"),
                            getAttr(attrs,"serialized"),
                            getAttr(attrs,"delete-action"),
                            getAttr(attrs,"indexed"),
                            getAttr(attrs,"unique"));
            FieldMetaData fmd = (FieldMetaData)getStack();
            fmd.setKeyMetaData(keymd);
            pushStack(keymd);
        }
        // New fetch-group
        else if (qName.equals("fetch-group"))
        {
            MetaData md = getStack();
            FetchGroupMetaData fgmd=new FetchGroupMetaData(md,
                            getAttr(attrs,"depth"),
                            getAttr(attrs,"post-load"),
                            getAttr(attrs,"name"));
            if (md instanceof FetchGroupMetaData)
            {
                FetchGroupMetaData fgmdParent = (FetchGroupMetaData)md;
                fgmdParent.addFetchGroup(fgmd);
            }
            else if (md instanceof ClassMetaData)
            {
                ClassMetaData cmd = (ClassMetaData)md;
                cmd.addFetchGroup(fgmd);
            }
            pushStack(fgmd);
        }        
        // New value
        else if (qName.equals("value"))
        {
            ValueMetaData valuemd = new ValueMetaData(
                            getAttr(attrs,"column"),
                            getAttr(attrs,"embedded"),
                            getAttr(attrs,"serialized"),
                            getAttr(attrs,"delete-action"),
                            getAttr(attrs,"indexed"),
                            getAttr(attrs,"unique"));
            FieldMetaData fmd=(FieldMetaData)getStack();
            fmd.setValueMetaData(valuemd);
            pushStack(valuemd);
        }
        // New extension for this tag
        else if (qName.equals("extension"))
        {
            ExtendableMetaData emd = (ExtendableMetaData)getStack();
            emd.addExtension(getAttr(attrs,"vendor-name"),
                            getAttr(attrs,"key"),
                            getAttr(attrs,"value"));
        }
        // New version
        else if (qName.equals("version"))
        {
            ClassMetaData cmd = (ClassMetaData)getStack();
            VersionMetaData vermd = new VersionMetaData(
                            getAttr(attrs,"strategy"),
                            getAttr(attrs,"column"),
                            getAttr(attrs,"indexed"));
            cmd.setVersionMetaData(vermd);
            pushStack(vermd);
        }
        // New index
        else if (qName.equals("index"))
        {
            ExtendableMetaData extmd = (ExtendableMetaData)getStack();
            IndexMetaData idxmd = new IndexMetaData(
                            getAttr(attrs,"name"),
                            getAttr(attrs,"table"),
                            getAttr(attrs,"unique"));
            if (extmd instanceof FieldMetaData)
            {
                FieldMetaData fmd = (FieldMetaData)extmd;
                fmd.setIndexMetaData(idxmd);
            }
            else if (extmd instanceof ClassMetaData)
            {
                ClassMetaData cmd = (ClassMetaData)extmd;
                cmd.addIndex(idxmd);
            }
            else if (extmd instanceof OrderMetaData)
            {
                OrderMetaData omd = (OrderMetaData)extmd;
                omd.setIndexMetaData(idxmd);
            }
            else if (extmd instanceof VersionMetaData)
            {
                VersionMetaData vermd = (VersionMetaData)extmd;
                vermd.setIndexMetaData(idxmd);
            }
            else if (extmd instanceof DiscriminatorMetaData)
            {
                DiscriminatorMetaData dismd = (DiscriminatorMetaData)extmd;
                dismd.setIndexMetaData(idxmd);
            }
            pushStack(idxmd);
        }
        // New unique constraint
        else if (qName.equals("unique"))
        {
            ExtendableMetaData extmd = (ExtendableMetaData)getStack();
            UniqueMetaData unimd = new UniqueMetaData(
                            getAttr(attrs,"name"),
                            getAttr(attrs,"table"),
                            getAttr(attrs,"deferred"));
            if (extmd instanceof ClassMetaData)
            {
                ClassMetaData cmd = (ClassMetaData)extmd;
                cmd.addUniqueConstraint(unimd);
            }
            else if (extmd instanceof FieldMetaData)
            {
                FieldMetaData fmd = (FieldMetaData)extmd;
                fmd.setUniqueMetaData(unimd);
            }
/*            else if (extmd instanceof JoinMetaData)
            {
                JoinMetaData jmd = (JoinMetaData)extmd;
                jmd.setUniqueMetaData(unimd);
            }*/
            else if (extmd instanceof KeyMetaData)
            {
                KeyMetaData kmd = (KeyMetaData)extmd;
                kmd.setUniqueMetaData(unimd);
            }
            else if (extmd instanceof ValueMetaData)
            {
                ValueMetaData vmd = (ValueMetaData)extmd;
                vmd.setUniqueMetaData(unimd);
            }
            else if (extmd instanceof ElementMetaData)
            {
                ElementMetaData emd = (ElementMetaData)extmd;
                emd.setUniqueMetaData(unimd);
            }
            pushStack(unimd);
        }
        // New foreign key
        else if (qName.equals("foreign-key"))
        {
            ExtendableMetaData extmd=(ExtendableMetaData)getStack();
            ForeignKeyMetaData fkmd=new ForeignKeyMetaData(
                            getAttr(attrs,"deferred"),
                            getAttr(attrs,"delete-action"),
                            getAttr(attrs,"update-action"),
                            getAttr(attrs,"table"),
                            getAttr(attrs,"unique"),
                            getAttr(attrs,"name"));
            if (extmd instanceof FieldMetaData)
            {
                FieldMetaData fmd = (FieldMetaData)extmd;
                fmd.setForeignKeyMetaData(fkmd);
            }
            else if (extmd instanceof ClassMetaData)
            {
                ClassMetaData cmd = (ClassMetaData)extmd;
                cmd.addForeignKey(fkmd);
            }
            pushStack(fkmd);
        }
        // New order
        else if (qName.equals("order"))
        {
            OrderMetaData ormd = new OrderMetaData(getAttr(attrs,"column"),
                                                   getAttr(attrs,"indexed"));
            FieldMetaData fmd = (FieldMetaData)getStack();
            fmd.setOrderMetaData(ormd);
            pushStack(ormd);
        }
        // Embedded
        else if (qName.equals("embedded"))
        {
            ExtendableMetaData extmd=(ExtendableMetaData)getStack();
            EmbeddedMetaData emd = new EmbeddedMetaData(extmd,
                                                        getAttr(attrs,"owner-field"),
                                                        getAttr(attrs,"null-indicator-column"),
                                                        getAttr(attrs,"null-indicator-value"));
            if (extmd instanceof FieldMetaData)
            {
                FieldMetaData fmd = (FieldMetaData)extmd;
                fmd.setEmbeddedMetaData(emd);
            }
            else if (extmd instanceof KeyMetaData)
            {
                KeyMetaData kmd = (KeyMetaData)extmd;
                kmd.setEmbeddedMetaData(emd);
            }
            else if (extmd instanceof ValueMetaData)
            {
                ValueMetaData vmd = (ValueMetaData)extmd;
                vmd.setEmbeddedMetaData(emd);
            }
            else if (extmd instanceof ElementMetaData)
            {
                ElementMetaData elmd = (ElementMetaData)extmd;
                elmd.setEmbeddedMetaData(emd);
            }
            pushStack(emd);
        }
        else
        {
            String message = LOCALISER.msg("MetaData.Parser.UnknownTag.Error",qName);
            JPOXLogger.METADATA.error(message);
            throw new RuntimeException(message);
        }
    }

    /**
     * String for a named query. Built up by characters() method.
     */
    private StringBuffer queryString = new StringBuffer();

    /**
     * Handler method called at the end of an element.
     * @param uri URI of the tag
     * @param localName local name
     * @param qName Name of element just ending
     */
    public void endElement(String uri, String localName, String qName)
    throws SAXException
    {
        if (JPOXLogger.METADATA.isDebugEnabled())
        {
            JPOXLogger.METADATA.debug(LOCALISER.msg("MetaData.Parser.EndElement",qName + " (stack size=" + stack.size()));
        }

        // Save the Query string
        if (qName.equals("query"))
        {
            ((QueryMetaData)getStack()).setQuery(queryString.toString().trim());
            queryString = new StringBuffer();
        }

        // Pop the tag
        // If startElement pushes an element onto the stack need a remove
        // here for that type
        if (qName.equals("package") ||
            qName.equals("class") ||
            qName.equals("interface") ||
            qName.equals("implements") ||
            qName.equals("property") ||
            qName.equals("datastore-identity") ||
            qName.equals("inheritance") ||
            qName.equals("version") ||
            qName.equals("unmapped") ||
            qName.equals("query") ||
            qName.equals("field") ||
            qName.equals("map") ||
            qName.equals("element") ||
            qName.equals("embedded") ||
            qName.equals("key") ||
            qName.equals("value") ||
            qName.equals("array") ||
            qName.equals("collection") ||
            qName.equals("join") ||
            qName.equals("index") ||
            qName.equals("unique") ||
            qName.equals("foreign-key") ||
            qName.equals("order") ||
            qName.equals("fetch-group") ||
            qName.equals("column") ||
            qName.equals("discriminator"))
        {
            popStack();
        }
    }

    /**
     * Parser error method. If any syntactical errors are encountered on
     * validation they will appear here. Just points the user to the
     * line/column of their Meta-Data file for now.
     * @param e Parse Exception 
     * @see org.xml.sax.ErrorHandler#error(org.xml.sax.SAXParseException)
     */
    public void error(SAXParseException e)
    throws SAXException
    {
        JPOXLogger.METADATA.error(LOCALISER.msg("MetaData.Parser.Error",
                                "" + e.getLineNumber(),
                                "" + e.getColumnNumber(),
                                e.getMessage()));
    }

    /**
     * Notification handler for the data inside an element.
     * @param ch The characters
     * @param start The start position in the character array.
     * @param length The length of the string.
     */
    public void characters(char[] ch, int start, int length)
    throws SAXException
    {
        // Add to the buffer
        charactersBuffer.append(ch, start, length);

        // Extract the value of any element that we use the value
        MetaData element = getStack();
        String element_value = new String(ch,start,length);
        if (element instanceof QueryMetaData)
        {
            if (latestElement.equals("query"))
            {
                queryString.append(element_value);
            }
        }
    }

    public String getString()
    {
        String result = charactersBuffer.toString();
        charactersBuffer = new StringBuffer();
        return result;
    }

    /**
     * Accessor for the current MetaData component.
     * @return The current MetaData component.
     */
    protected MetaData getStack()
    {
        Object o = stack.lastElement();
        if (o != null)
        {
            return (MetaData)o;
        }
        return null;
    }

    /**
     * Method to remove the current MetaData component from the Stack.
     * @return Latest MetaData component.
     */
    protected MetaData popStack()
    {
        return (MetaData)(stack.pop());
    }

    /**
     * Method to add a MetaData component to the Stack.
     * @param md The component to add.
     */
    protected void pushStack(MetaData md)
    {
        stack.push(md);
    }
    
    /**
     * Method to resolve XML entities. Checks for PUBLIC and SYSTEM entities
     * valid for JDO 1.0 and JDO 2.0. Optimises access using local versions of
     * the JDO 1 and 2 DTD's.
     * @param publicId The public id.
     * @param systemId The system id.
     * @return Input Source for the URI.
     * @see org.xml.sax.EntityResolver#resolveEntity(java.lang.String,java.lang.String)
     */
    public InputSource resolveEntity(String publicId, String systemId)
    throws SAXException
    {
        // Delegate to the entity resolver.
        InputSource source = entityResolver.resolveEntity(publicId, systemId);
        if (source == null)
        {
            // TODO Cater better here for this situation
            // Would have been nice to just call super.resolveEntity() but sadly SUN have
            // screwed the method by adding a new checked exception on it in JDK 1.5 meaning
            // that we either compile in 1.4 or in 1.5.
        }
        return source;
    }

    /**
     * Accessor for an attribute out of the attributes for an element. Allows
     * the specification of a default if no value is provided.
     * @param attrs The attributes
     * @param key Key for the attribute to return
     * @param defaultValue A default to impose if no value.
     * @return Value for the attribute with the specified key.
     */
    protected String getAttr(Attributes attrs, String key, String defaultValue)
    {
        String result = attrs.getValue(key);
        if (result == null)
        {
            return defaultValue;
        }
        else if (result.length() == 0)
        {
            return defaultValue;
        }
        return result;
    }

    /**
     * Accessor for an attribute out of the attributes for an element.
     * @param attrs The attributes
     * @param key Key for the attribute to return
     * @return Value for the attribute with the specified key.
     */
    protected String getAttr(Attributes attrs, String key)
    {
        return getAttr(attrs, key, null);
    }
}