/**********************************************************************
Copyright (c) 04-Jun-2004 Andy Jefferson and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
    ...
**********************************************************************/
package org.jpox.metadata;

import java.lang.reflect.Modifier;
import java.net.URL;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collections;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.jdo.JDOFatalInternalException;

import org.w3c.dom.Element;
import org.w3c.dom.NodeList;

import org.jpox.util.ClassUtils;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;

/**
 * Utilities needed for the processing of MetaData.
 *
 * @version $Revision: 1.4 $
 */
public class MetaDataUtils
{
    protected static final Localiser LOCALISER=Localiser.getInstance("org.jpox.metadata.Localisation");

    private static final String[] DEFAULT_PERSISTENT_TYPE;
    private static final String[] DEFAULT_PERSISTENT_ARRAY_TYPE;
    private static final String[] DEFAULT_FETCH_GROUP_TRUE;

    // Initialisation block to define the defaults/values for JDO tags
    // TODO Merge this with TypeManager.
    static
    {
        String persistentType[] =
            new String[]
            {
                String.class.getName(),
                java.lang.Number.class.getName(),
                Boolean.class.getName(),
                Byte.class.getName(),
                Character.class.getName(),
                Short.class.getName(),
                Integer.class.getName(),
                Float.class.getName(),
                Long.class.getName(),
                Double.class.getName(),
                java.math.BigDecimal.class.getName(),
                java.math.BigInteger.class.getName(),
                java.util.Date.class.getName(),
                java.util.Locale.class.getName(),
                java.util.Currency.class.getName(),
                java.util.ArrayList.class.getName(),
                java.util.HashMap.class.getName(),
                java.util.HashSet.class.getName(),
                java.util.Hashtable.class.getName(),
                java.util.LinkedList.class.getName(),
                java.util.TreeMap.class.getName(),
                java.util.TreeSet.class.getName(),
                java.util.Vector.class.getName(),
                java.util.Collection.class.getName(),
                java.util.Set.class.getName(),
                java.util.List.class.getName(),
                java.util.Map.class.getName()
            };
        Arrays.sort(persistentType);
        DEFAULT_PERSISTENT_TYPE = persistentType;

        String persistentArrayType[] =
            new String[]
            {
                "Z",
                "B",
                "C",
                "S",
                "I",
                "F",
                "J",
                "D",
                Boolean.TYPE.getName(),
                Byte.TYPE.getName(),
                Character.TYPE.getName(),
                Short.TYPE.getName(),
                Integer.TYPE.getName(),
                Float.TYPE.getName(),
                Long.TYPE.getName(),
                Double.TYPE.getName(),
                Boolean.class.getName(),
                Byte.class.getName(),
                Character.class.getName(),
                Short.class.getName(),
                Integer.class.getName(),
                Float.class.getName(),
                Long.class.getName(),
                Double.class.getName(),
                String.class.getName(),
                java.util.HashMap.class.getName(),
                java.util.HashSet.class.getName(),
                java.util.Hashtable.class.getName(),
                java.util.LinkedList.class.getName(),
                java.util.TreeMap.class.getName(),
                java.util.TreeSet.class.getName(),
                java.util.Vector.class.getName(),
                java.util.Collection.class.getName(),
                java.util.Set.class.getName(),
                java.util.List.class.getName(),
                java.util.Map.class.getName(),
            };
        Arrays.sort(persistentArrayType);
        DEFAULT_PERSISTENT_ARRAY_TYPE = persistentArrayType;

        String defaultFetchGroupTrue[] =
            new String[]
            {
                java.util.Date.class.getName(),
                Boolean.class.getName(),
                Byte.class.getName(),
                Short.class.getName(),
                Integer.class.getName(),
                Long.class.getName(),
                Character.class.getName(),
                Float.class.getName(),
                Double.class.getName(),
                String.class.getName(),
                Number.class.getName(),
                java.math.BigDecimal.class.getName(),
                java.math.BigInteger.class.getName()
            };
        Arrays.sort(defaultFetchGroupTrue);
        DEFAULT_FETCH_GROUP_TRUE = defaultFetchGroupTrue;
    }

    /**
     * Accessor for the default "persistence-modifier" for a field given the
     * class, its modifier and whether it is a PersistentCapable class.
     * @param c The class
     * @param modifier The modifiers for the field
     * @param isPCclass Whether it is persistence capable.
     * @return The default field PersistenceModifier.
     */
    public static final FieldPersistenceModifier 
            getDefaultFieldPersistenceModifier(Class c,
                                               int modifier,
                                               boolean isPCclass)
    {
        // Static, Transient and Final fields are all not persistent
        if (Modifier.isStatic(modifier) ||
            Modifier.isTransient(modifier) ||
            Modifier.isFinal(modifier))
        {
            return FieldPersistenceModifier.NONE;
        }

        // All PC class fields are persistent by default
        if (isPCclass)
        {
            return FieldPersistenceModifier.PERSISTENT;
        }

        if (c == null)
        {
            throw new RuntimeException("class is null");
        }

        // All primitive fields are persistent by default
        if (c.isPrimitive())
        {
            return FieldPersistenceModifier.PERSISTENT;
        }

        // Check based on the type
        String name = c.getName();
        if (c.isArray())
        {
            while (name.startsWith("["))
            {
                name = name.substring(1, name.length());
            }
            if (name.startsWith("L"))
            {
                name = name.substring(1, name.length());
            }
            if (name.lastIndexOf(';') >= 0)
            {
                name = name.substring(0, name.length() - 1);
            }

            if (Arrays.binarySearch(DEFAULT_PERSISTENT_ARRAY_TYPE, name) >= 0)
            {
                return FieldPersistenceModifier.PERSISTENT;
            }
        }

        if (Arrays.binarySearch(DEFAULT_PERSISTENT_TYPE, name) >= 0)
        {
            return FieldPersistenceModifier.PERSISTENT;
        }

        return FieldPersistenceModifier.NONE;
    }

    /**
     * Accessor for whether the class is by default in the default fetch group.
     * @param c The class
     * @return return true if is part of default fetch group by default
     */
    public static boolean isDefaultFetchGroup(Class c)
    {
        if (c.isPrimitive())
        {
            return true;
        }
        else if (Arrays.binarySearch(DEFAULT_FETCH_GROUP_TRUE,c.getName()) >= 0)
        {
            return true;
        }

        return false;
    }
    
    /**
     * Utility to return whether a type defaults to embedded or not in JDO.
     * See JDO 2 Section 18.13 for definition. This only provides for the
     * basic types, and ignores arrays/collections.
     * @param type The type
     * @return Whether it defaults to embedded or not.
     */
    public static boolean isDefaultEmbeddedType(Class type)
    {
        if (type == null)
        {
            return false;
        }

        // These types default to embedded
        if (type.isPrimitive() ||
            type.getName().equals("java.util.Date") ||
            type.getName().equals("java.util.Locale") ||
            type.getName().equals("java.util.Currency") ||
            type.getName().equals("java.math.BigDecimal") ||
            type.getName().equals("java.math.BigInteger") ||
            type.getName().equals("java.lang.String") ||
            type.getName().equals("java.lang.Number") ||
            type.getName().equals("java.lang.Boolean") ||
            type.getName().equals("java.lang.Byte") ||
            type.getName().equals("java.lang.Character") ||
            type.getName().equals("java.lang.Double") ||
            type.getName().equals("java.lang.Float") ||
            type.getName().equals("java.lang.Integer") ||
            type.getName().equals("java.lang.Long") ||
            type.getName().equals("java.lang.Short"))
        {
            return true;
        }

        return false;
    }

    /**
     * Utility to find the full class name of a class specified in MetaData.
     * In MetaData files people can specify things like "element" and just
     * specified the last part of the fully-qualified name. This tries to
     * resolve the name to a particular package, trying a few likely places. 
     * @param class_name Name of the class that was supplied
     * @param pkg_name Package name of the context
     * @param loader Class loader to use in checking 
     * @return Full class name
     */
    public static Class resolvedClassForName(String class_name,
                                             String pkg_name,
                                             ClassLoader loader)
    {
        if (class_name == null)
        {
            return null;
        }
        if (loader == null)
        {
            return null;
        }

        // List of locations to check
        String[] names=
            {
                class_name,
                pkg_name + '.' + class_name,
                "java.lang." + class_name,
                "java.util." + class_name,
                "java.math." + class_name
            };

        for (int i=0;i<names.length;i++)
        {
            try
            {
                Class type=loader.loadClass(names[i]);
                if (type != null)
                {
                    return type;
                }
            }
            catch (ClassNotFoundException cnfe)
            {
            }
        }

        // Not found
        return null;
    }

    // Parameters used in the definition of MetaData file location
    private static final char CLASS_SEPARATOR = '.';
    private static final char PATH_SEPARATOR = '/';
    private static final char EXTENSION_SEPARATOR = '.';
    private static final String METADATA_PACKAGE = "package";
    private static final String METADATA_LOCATION_METAINF = "/META-INF/" + METADATA_PACKAGE;
    private static final String METADATA_LOCATION_WEBINF = "/WEB-INF/" + METADATA_PACKAGE;

    /**
     * Utility to return a list of classes declared in provided metadata
     * files. The list of classes are ordered so that all superclasses occur
     * before their subclasses. Additionally populates the provided map
     * with entries mapping the class names to the JDO MetaData file name that
     * the class was found in.
     * @param metaDataFileNames The list of JDO metadata file names.
     * @param classToMetaDataMap A Map that will be populated mapping class
     *   name to the MetaData file name. This argument can be null.
     * @return An ordered list of class name strings.
     * @throws Exception
     */
    public static List getOrderedClassNamesFromFileList(
                                    String[] metaDataFileNames,
                                    Map classToMetaDataMap)
    throws Exception
    {
        HashMap superclassesByClass = new HashMap();
        final DocumentBuilderFactory factory = DocumentBuilderFactory.newInstance();
        final DocumentBuilder db = factory.newDocumentBuilder();
        db.setEntityResolver(new JDOEntityResolver());

        for (int i=0;i<metaDataFileNames.length;++i)
        {
            Element docElement=db.parse(new URL("file:" + metaDataFileNames[i]).openStream()).getDocumentElement();

            NodeList nodes=docElement.getElementsByTagName("class");
            for (int j=0;j<nodes.getLength();++j)
            {
                Element clsElement = (Element)nodes.item(j);
                String superclassName = clsElement.getAttribute("persistence-capable-superclass");

                if (superclassName.length() == 0)
                {
                    superclassName = null;
                }

                String packageName = ((Element)clsElement.getParentNode()).getAttribute("name");
                String className = ClassUtils.createFullClassName(packageName, clsElement.getAttribute("name"));
                
                superclassesByClass.put(className,superclassName);
                if (classToMetaDataMap != null)
                {
                    classToMetaDataMap.put(className,metaDataFileNames[i]);
                }
            }
        }

        int numClasses = superclassesByClass.size();
        ArrayList orderedNames = new ArrayList(numClasses);
        ArrayList sortedNames = new ArrayList(numClasses);

        sortedNames.addAll(superclassesByClass.keySet());
        Collections.sort(sortedNames);
        Iterator i=sortedNames.iterator();
        while (i.hasNext())
        {
            addClassToOrderedList(orderedNames,(String)i.next(),superclassesByClass);
        }

        return orderedNames;
    }

    /**
     * Utility to add a class to a list of class names. Recursive to include
     * superclasses.
     * @param names List of classnames (updated by this operation)
     * @param className Class name to add
     * @param superclassesByClass Map of superclasses
     * @throws Exception
     */
    private static void addClassToOrderedList(List names,
                                              String className,
                                              Map superclassesByClass)
    throws Exception
    {
        String superclassName=(String)superclassesByClass.get(className);
        if (superclassName != null)
        {
            addClassToOrderedList(names,superclassName,superclassesByClass);
        }

        if (!names.contains(className))
        {
            names.add(className);
        }
    }
    
    /**
     * Method to find the Meta-Data file for a specified class.
     * Supports the JDO 1.0.0 and JDO 1.0.1 definitions of file location,
     * namely
     * <UL>
     * <LI>META-INF/package.jdo (1.0.1)</LI>
     * <LI>WEB-INF/package.jdo (1.0.1)</LI>
     * <LI>package.jdo (1.0.1)</LI>
     * <LI>com/package.jdo (1.0.1)</LI>
     * <LI>com/xyz/package.jdo (1.0.1)</LI>
     * <LI>com/xyz/MyClass.jdo (1.0.0, 1.0.1)</LI>
     * <LI>com.jdo (1.0.0)</LI>
     * <LI>com/xyz.jdo (1.0.0)</LI>
     * </UL>
     * Checks the locations one-by-one, and checks for existence of the
     * specified class in the file.
     * <P>
     * Allows 2 variations on the naming above. The first is a modifier which
     * caters for a JDO 2.0 requirement whereby the user can specify a modifier
     * such as "mysql", which would mean that this should search for filenames
     * "package-mysql.jdo". The second variation is the suffix of the file.
     * This is "jdo" by default, but JDO 2.0 has situations where "orm", or
     * "jdoquery" are required as a suffix.
     * </P> 
     * @param pc_class The class to retrieve the Meta-Data file for
     * @param metadata_file_modifier Any modifier for the filename
     * @param metadata_file_extension File extension of MetaData files ("jdo")
     * @return URL of the Meta-Data file.
     **/
    public static URL getMetaDataFileForClass(Class pc_class,
                                              String metadata_file_modifier,
                                              String metadata_file_extension)
    {
        // Construct the suffix. This part is added on to any "package" naming.
        String suffix=null;
        if (metadata_file_extension == null)
        {
            metadata_file_extension = "jdo";
        }
        if (metadata_file_modifier != null)
        {
            suffix = "-" + metadata_file_modifier + EXTENSION_SEPARATOR + metadata_file_extension;
        }
        else
        {
            suffix = EXTENSION_SEPARATOR + metadata_file_extension;
        }

        URL url=null;
        
        // Try META-INF
        String location=METADATA_LOCATION_METAINF + suffix;
        url = pc_class.getResource(location);
        if (url != null)
        {
            Element class_element=MetaDataUtils.getMetaDataClassElementFromFile(url, pc_class);
            if (class_element == null)
            {
                url = null;
            }
        }

        if (url == null)
        {
            if (JPOXLogger.METADATA.isDebugEnabled())
            {
                JPOXLogger.METADATA.debug(LOCALISER.msg("MetaData.File.NotFoundForClassAtLocationError", pc_class.getName(), location));
            }

            // Try WEB-INF
            location = METADATA_LOCATION_WEBINF + suffix;
            url = pc_class.getResource(location);
            if (url != null)
            {
                Element class_element=MetaDataUtils.getMetaDataClassElementFromFile(url, pc_class);
                if (class_element == null)
                {
                    url = null;
                }
            }

            if (url == null)
            {
                if (JPOXLogger.METADATA.isDebugEnabled())
                {
                    JPOXLogger.METADATA.debug(LOCALISER.msg("MetaData.File.NotFoundForClassAtLocationError", pc_class.getName(), location));
                }

                // This searches through /package.jdo, /com/package.jdo,
                // /com/xyz/package.jdo, /com/xyz/MyClass.jdo
                url = getURLForMetaDataFile(pc_class,null,suffix);
            }
        }

        // Give success message as appropriate
        if (url == null)
        {
            JPOXLogger.METADATA.debug(LOCALISER.msg("MetaData.File.NotFoundForClassError", pc_class.getName()));
        }
        else
        {
            JPOXLogger.METADATA.debug(LOCALISER.msg("MetaData.File.FoundClass", pc_class.getName(),url));
        }

        return url;
    }

    /**
     * Utility to find the MetaData file for a class.
     * Works by stepping along the package chain until it finds a "live" one.
     * @param pc_class Class name to find MetaData for.
     * @param name  File name to try
     * @param suffix Suffix for the filenames (after "package").
     * @return url The URL (null if not found)
     **/
    private static URL getURLForMetaDataFile(Class pc_class,
                                             String name,
                                             String suffix)
    {
        String class_name=pc_class.getName();
        String location=name;

        // Try location
        if (name != null)
        {
            int start_posn=class_name.indexOf(name);
            if (start_posn != 0)
            {
                // Error in input so start at top-level
                name = null;
            }
            else if (name.equals(class_name))
            {
                // JDO 1.0.0, 1.0.1 class location
                location = PATH_SEPARATOR +
                    name.replace(CLASS_SEPARATOR, PATH_SEPARATOR) + suffix;
            }
            else
            {
                // JDO 1.0.1 package location
                location = PATH_SEPARATOR +
                    name.replace(CLASS_SEPARATOR, PATH_SEPARATOR) +
                    PATH_SEPARATOR +
                    METADATA_PACKAGE + suffix;
            }
        }

        if (name == null)
        {
            location = PATH_SEPARATOR + METADATA_PACKAGE + suffix;
        }
        URL url=pc_class.getResource(location);
        if (url == null)
        {
            if (JPOXLogger.METADATA.isDebugEnabled())
            {
                JPOXLogger.METADATA.debug(LOCALISER.msg("MetaData.File.NotFoundForClassAtLocationError", class_name, location));
            }
        }
 
        // Try again with JDO 1.0.0 package location. e.g. /com/xyz.jdo
        if (url == null)
        {
            if (name != null)
            {
                // Class name specified so try its MetaData file
                location = PATH_SEPARATOR +
                    name.replace(CLASS_SEPARATOR, PATH_SEPARATOR) + suffix;
                url = pc_class.getResource(location);
                if (url == null)
                {
                    if (JPOXLogger.METADATA.isDebugEnabled())
                    {
                        JPOXLogger.METADATA.debug(LOCALISER.msg("MetaData.File.NotFoundForClassAtLocationError", class_name, location));
                    }
                }
            }
        }

        // Found a file, so check existence of the class in the file
        if (url != null)
        {
            Element class_element=MetaDataUtils.getMetaDataClassElementFromFile(url, pc_class);
            if (class_element == null)
            {
                url = null;
                if (JPOXLogger.METADATA.isDebugEnabled())
                {
                    JPOXLogger.METADATA.debug(LOCALISER.msg("MetaData.File.NotFoundForClassAtLocationError", class_name, location));
                }
            }
        }
 
        // Nothing found so move down the tree
        if (url == null)
        {
            // Started from top - so find top package name
            if (name == null)
            {
                int dot_posn=class_name.indexOf(CLASS_SEPARATOR);
                if (dot_posn < 0)
                {
                    // Class is at top-level and we havent found the MetaData
                    return null;
                }
                else
                {
                    // Take first package
                    name = class_name.substring(0, dot_posn);
                }
            }
            // Already tried the class itself and not found it so give up
            else if (name.equals(class_name))
            {
                return null;
            }
            // Moving through the packages
            else
            {
                int dot_posn=class_name.indexOf(CLASS_SEPARATOR, name.length()+1);
                if (dot_posn < 0)
                {
                    // Reached end, so use class name
                    name = class_name;
                }
                else
                {
                    name = class_name.substring(0, dot_posn);
                }
            }

            return getURLForMetaDataFile(pc_class,name,suffix);
        }
        else
        {
            return url;
        }
    }

    /**
     * Utility method to return a particular class Element in the specified
     * Meta-Data file.
     * @param url The URL of the Meta-Data file.
     * @param c The class to find
     * @return The "class" Element.
     * @throws JDOFatalInternalException Thrown if an error in the parse 
     **/
    public static Element getMetaDataClassElementFromFile(URL url, Class c)
    throws JDOFatalInternalException
    {
        Element docElement;

        try
        {
            final DocumentBuilder builder = 
                DocumentBuilderFactory.newInstance().newDocumentBuilder();
            builder.setEntityResolver(new JDOEntityResolver());

            docElement = builder.parse(url.openStream()).getDocumentElement();
        }
        catch (Exception e)
        {
            throw new JDOFatalInternalException(LOCALISER.msg("MetaData.File.ParseError",url),e);
        }

        // Loop through the nodes and find the appropriate class
        NodeList nodes = docElement.getElementsByTagName("class");
        for (int i = 0; i < nodes.getLength(); ++i)
        {
            Element clsElement = (Element)nodes.item(i);
            String packageName = ((Element)clsElement.getParentNode()).getAttribute("name");
            String className = ClassUtils.createFullClassName(packageName, clsElement.getAttribute("name"));

            if (className.equals(c.getName()))
            {
                return clsElement;
            }
        }

        return null;
    }
}
