/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
2004 Andy Jefferson - toString(), MetaData, javadocs
2004 Andy Jefferson - added index, and indexed
    ...
**********************************************************************/
package org.jpox.metadata;

import java.util.ArrayList;
import java.util.List;

/**
 * Representation of Order MetaData - the ordering of the elements
 * of a List.
 *
 * <H3>MetaData Element</H3> 
 * The MetaData element represented here is as follows 
 * <PRE>
 * &lt;!ELEMENT order (extension*, column*, index?, extension*)&gt;
 * &lt;!ATTLIST order column CDATA #IMPLIED&gt;
 * &lt;!ATTLIST order indexed (true|false|unique) #IMPLIED&gt;
 * </PRE>
 *
 * @version $Revision: 1.2 $
 */
public class OrderMetaData extends ExtendableMetaData
{
    /**
     * the column name
     */
    final String column;

    /**
     * the columns elements 
     */
    final List columns = new ArrayList();

    /**
     * IndexMetaData.
     */
    protected IndexMetaData indexMetaData;

    /**
     * The indexing value
     */
    protected IndexedValue indexed=null;

    // -------------------------------------------------------------------------
    // Fields below here are not represented in the output MetaData. They are
    // for use internally in the operation of the JDO system. The majority are
    // for convenience to save iterating through the fields since the fields
    // are fixed once initialised.

    /**
     * Contains the metadata for column
     */
    protected ColumnMetaData[] columnMetaData;
 
    /**
     * Constructor. 
     * @param column Name of column
     * @param indexed The indexed value
     */
    public OrderMetaData(final String column,
                         final String indexed)
    {
        super(null);    // Ignore parent
        this.column = column;
        this.indexed = IndexedValue.getIndexedValue(indexed);
    }

    /**
     * Method to initialise the object, creating internal convenience arrays.
     * Initialises all sub-objects.
     */
    public void initialise()
    {
        columnMetaData = new ColumnMetaData[columns.size()];
        for (int i=0; i<columnMetaData.length; i++)
        {
            columnMetaData[i] = (ColumnMetaData) columns.get(i);
            columnMetaData[i].initialise();
        }

        // Interpret the "indexed" value to create our IndexMetaData where it wasn't specified that way
        if (indexMetaData == null && columnMetaData != null && indexed != null && indexed != IndexedValue.FALSE)
        {
            indexMetaData = new IndexMetaData(null, null, (indexed == IndexedValue.UNIQUE) ? "true" : "false");
            for (int i=0;i<columnMetaData.length;i++)
            {
                indexMetaData.addColumn(columnMetaData[i]);
            }
        }
        if (indexMetaData != null)
        {
            indexMetaData.initialise();
        }

        setInitialised();
    }

    // --------------------------------- Mutators ------------------------------

    /**
     * Add a new ColumnMetaData element
     * @param colmd The Column MetaData 
     */
    public void addColumn(ColumnMetaData colmd)
    {
        columns.add(colmd);
    }

    /**
     * Mutator for the index MetaData 
     * @param indexMetaData The indexMetaData to set.
     */
    public final void setIndexMetaData(IndexMetaData indexMetaData)
    {
        this.indexMetaData = indexMetaData;
    }

    // ----------------------------- Accessors ---------------------------------

    /**
     * Accessor for the Column MetaData for the columns
     * @return Returns the columnMetaData.
     */
    public final ColumnMetaData[] getColumnMetaData()
    {
        return columnMetaData;
    }

    /**
     * Accessor for the column name 
     * @return Returns the column.
     */
    public final String getColumn()
    {
        return column;
    }

    /**
     * Accessor for indexMetaData
     * @return Returns the indexMetaData.
     */
    public final IndexMetaData getIndexMetaData()
    {
        return indexMetaData;
    }

    // ------------------------------ Utilities --------------------------------

    /**
     * Accessor for a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        return toString("","");
    }

    /**
     * Returns a string representation of the object using a prefix
     * This can be used as part of a facility to output a MetaData file. 
     * @param prefix prefix string
     * @param indent indent string
     * @return a string representation of the object.
     */
    public String toString(String prefix,String indent)
    {
        // Field needs outputting so generate metadata
        StringBuffer sb = new StringBuffer();
        sb.append(prefix).append("<order");
        if (column != null)
        {
            sb.append(" column=\"" + column + "\"");
        }
        if (indexed != null)
        {
            sb.append(" indexed=\"" + indexed.toString() + "\"");
        }
        sb.append(">\n");

        // Add columns
        for (int i=0; i<columns.size(); i++)
        {
            ColumnMetaData c = (ColumnMetaData)columns.get(i);
            sb.append(c.toString(prefix + indent,indent));
        }

        // Add index
        if (indexMetaData != null)
        {
            sb.append(indexMetaData.toString(prefix + indent,indent));
        }

        // Add extensions
        sb.append(super.toString(prefix + indent,indent));

        sb.append(prefix).append("</order>\n");
        return sb.toString();
    }
}