/**********************************************************************
Copyright (c) 04-Jun-2004 Andy Jefferson and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution.

Contributors:
    Kikuchi Kousuke - org.jpox.enhancer.conf.JDOConfigPackage
    ...
**********************************************************************/
package org.jpox.metadata;

import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

/**
 * Representation of the Meta-Data for a package.
 *
 * <H3>MetaData Element</H3>
 * The MetaData Element represented here is as follows
 * <PRE>
 * &lt;!ELEMENT package (extension*, (interface|class|sequence)*, extension*)&gt;
 * &lt;!ATTLIST package name CDATA #REQUIRED&gt;
 * </PRE>
 *
 * @since 1.1
 * @version $Revision: 1.2 $
 */
public class PackageMetaData extends ExtendableMetaData
{
    /**
     * List of interfaces
     */
    protected List interfaces = new ArrayList();

    /**
     * List of classes.
     */
    protected List classes = new ArrayList();

    /**
     * Package name
     */
    protected final String name;

    /**
     * Constructor.
     * @param parent The FileMetaData owner
     * @param name Name of package
     */
    public PackageMetaData(FileMetaData parent, final String name)
    {
        super(parent);

        this.name = name;
    }

    // -------------------------------- Accessors ------------------------------

    /**
     * Accessor for the parent FileMetaData.
     * @return File MetaData.
     */
    public FileMetaData getFileMetaData()
    {
        if (parent != null)
        {
            return (FileMetaData)parent;
        }
        return null;
    }

    /**
     * Accessor for the name of the package
     * @return package name
     */
    public String getName()
    {
        return name;
    }

    /**
     * Accessor for the number of interfaces.
     * @return Number of interfaces.
     */
    public int getNoOfInterfaces()
    {
        return interfaces.size();
    }
 
    /**
     * Accessor for the Meta-Data of a interface in this package.
     * @param i interface index
     * @return Meta-Data for the interface
     */
    public InterfaceMetaData getInterface(int i)
    {
        return (InterfaceMetaData)interfaces.get(i);
    }

    /**
     * Accessor for the Meta-Data of an interface with the specified name.
     * @param name the name of the interface
     * @return Meta-Data for the interface
     */
    public InterfaceMetaData getInterface(String name)
    {
        Iterator iter=interfaces.iterator();
        while (iter.hasNext())
        {
            InterfaceMetaData imd = (InterfaceMetaData)iter.next();
            if (imd.getName().equals(name))
            {
                return imd;
            }
        }
        return null;
    }

    /**
     * Accessor for the number of classes.
     * @return Number of classes.
     */
    public int getNoOfClasses()
    {
        return classes.size();
    }

    /**
     * Accessor for the Meta-Data of a class in this package.
     * @param i class index
     * @return Meta-Data for the class
     */
    public ClassMetaData getClass(int i)
    {
        return (ClassMetaData)classes.get(i);
    }

    /**
     * Accessor for the Meta-Data of a class with the specified name.
     * @param name the name of the class
     * @return Meta-Data for the class.
     */
    public ClassMetaData getClass(String name)
    {
        Iterator iter=classes.iterator();
        while (iter.hasNext())
        {
            ClassMetaData cmd = (ClassMetaData)iter.next();
            if (cmd.getName().equals(name))
            {
                return cmd;
            }
        }
        return null;
    }

    // -------------------------------- Mutators -------------------------------

    /**
     * Method to add a class Meta-Data to the package.
     * @param cmd Meta-Data for the class
     */
    public void addClass(ClassMetaData cmd)
    {
        if (cmd == null)
        {
            return;
        }
        classes.add(cmd);
    }

    /**
     * Method to add a interface Meta-Data to the package.
     * @param imd Meta-Data for the interface
     */
    public void addInterface(InterfaceMetaData imd)
    {
        if (imd == null)
        {
            return;
        }
        interfaces.add(imd);
    }

    // ------------------------------ Utilities --------------------------------
 
    /**
     * Returns a string representation of the object.
     * @param prefix prefix string
     * @param indent indent string
     * @return a string representation of the object.
     */
    public String toString(String prefix,String indent)
    {
        StringBuffer sb = new StringBuffer();
        sb.append(prefix).append("<package name=\"" + name + "\">\n");

        // Add interfaces
        Iterator int_iter=interfaces.iterator();
        while (int_iter.hasNext())
        {
            sb.append(((InterfaceMetaData)int_iter.next()).toString(prefix + indent,indent));
        }
 
        // Add classes
        Iterator cls_iter=classes.iterator();
        while (cls_iter.hasNext())
        {
            sb.append(((ClassMetaData)cls_iter.next()).toString(prefix + indent,indent));
        }
 
        // Add extensions
        sb.append(super.toString(prefix + indent,indent));

        sb.append(prefix).append("</package>\n");
        return sb.toString();
    }
}
