/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution.

Contributors:
    ...
**********************************************************************/
package org.jpox.metadata;

import org.jpox.util.StringUtils;

/**
 * Representation of the MetaData of a named Query.
 * <br>
 * <H3>MetaData Element</H3>
 * The MetaData Element represented here is as follows
 * <PRE>
 * <!ELEMENT query (extension*)>
 * <!ATTLIST query name CDATA #IMPLIED>
 * <!ATTLIST query language CDATA #IMPLIED>
 * <!ATTLIST query unmodifiable (true|false) 'true'>
 * </PRE>
 *
 * @since 1.1
 * @version $Revision: 1.4 $
 */
public class QueryMetaData extends ExtendableMetaData
{
    /**
     * Query name
     */
    protected final String name;

    /**
     * Language tag value.
     **/
    protected QueryLanguage language;

    /**
     * unmodifiable tag value.
     */
    protected boolean unmodifiable=false;

    /**
     * The single string query
     */
    protected String query;

    /**
     * Constructor.
     */
    public QueryMetaData(final ExtendableMetaData parent,
                         final String name,
                         final String language,
                         final String unmodifiable)
    {
        super(parent);

        this.name = name;
        if (StringUtils.isWhitespace(name))
        {
            throw new InvalidMetaDataException(LOCALISER,
                "MetaData.Query.QueryNameNotSpecifiedError");
        }

        // include-subclasses
        if (unmodifiable != null && unmodifiable.equalsIgnoreCase("false"))
        {
            this.unmodifiable = false;
        }
        else
        {
            this.unmodifiable = true;
        }

        // language
        this.language = QueryLanguage.getQueryLanguage(language);
    }

    /**
     * Accessor for the class name.
     * @return class name
     */
    public String getName()
    {
        return name;
    }

    /**
     * Accessor for the language
     * @return language tag value
     */
    public QueryLanguage getLanguage()
    {
        return language;
    }

    /**
     * Accessor for the unmodifiable tag value.
     * @return unmodifiable tag value
     */
    public boolean isUnmodifiable()
    {
        return unmodifiable;
    }

    /**
     * Accessor for the query
     * @return The query
     */
    public String getQuery()
    {
        return query;
    }

    /**
     * Mutator for the query
     * @param query The query
     */
    public void setQuery(String query)
    {
        this.query = query;
    }

    /**
     * Returns a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        return toString("","");
    }

    /**
     * Returns a string representation of the object.
     * @param prefix prefix string
     * @param indent indent string
     * @return a string representation of the object.
     */
    public String toString(String prefix,String indent)
    {
        StringBuffer sb = new StringBuffer();
        sb.append(prefix).append("<query name=\"" + name + "\"\n");
        sb.append(prefix).append("       language=\"" + language + "\"\n");
        sb.append(prefix).append("       unmodifiable=\"" + unmodifiable + "\">\n");
        sb.append(prefix).append(query).append("\n");

        // Add extensions
        sb.append(super.toString(prefix + indent,indent));

        sb.append(prefix + "</query>\n");
        return sb.toString();
    }
}