/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved.
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution.

Contributors:
    ...
**********************************************************************/
package org.jpox.metadata;

import java.util.ArrayList;
import java.util.List;

/**
 * MetaData representing a unique constraint.
 *
 * <H3>MetaData Element</H3>
 * The MetaData Element represented here is as follows 
 * <PRE>
 * &lt;!ELEMENT unique (extension*, (column|field|property)*, extension*)&gt;
 * &lt;!ATTLIST unique name CDATA #IMPLIED&gt;
 * &lt;!ATTLIST unique table CDATA #IMPLIED&gt;
 * &lt;!ATTLIST unique deferred (true|false) 'false'&gt;
 * </PRE>
 * @since 1.1
 * @version $Revision: 1.3 $
 */
public class UniqueMetaData extends ExtendableMetaData
{
    /**
     * the index name
     */
    final String name;

    /**
     * the table name. If not specified, then applies to the table of the containing class.
     */
    final String table;

    /**
     * Whether the unique is initially deferred.
     */
    final boolean deferred;

    /**
     * the fields elements to be included in the index. Can be empty.
     */
    final List fields = new ArrayList();

    /**
     * the columns elements to be included in the index. Suitable to be empty
     * when this metadata is contained within a field, element, key, value, or
     * join elements
     */
    final List columns = new ArrayList();
    
    // -------------------------------------------------------------------------
    // Fields below here are not represented in the output MetaData. They are
    // for use internally in the operation of the JDO system. The majority are
    // for convenience to save iterating through the fields since the fields
    // are fixed once initialised.

    /**
     * Contains the metadata for field
     */
    protected FieldMetaData[] fieldMetaData;

    /**
     * Contains the metadata for column
     */
    protected ColumnMetaData[] columnMetaData;

    /**
     * Constructor. 
     * @param name Name of unique constraint
     * @param table Name of the table
     * @param deferredValue Whether the unique is deferred initially
     */
    public UniqueMetaData(final String name, final String table, final String deferredValue)
    {
        super(null);    // Ignore parent

        this.name = name;
        this.table = table;
        if (deferredValue != null && deferredValue.equalsIgnoreCase("true"))
        {
            this.deferred = true;
        }
        else
        {
            this.deferred = false;
        }
    }

    /**
     * Method to initialise the object, creating internal convenience arrays.
     * Initialise all sub-objects.
     */
    public void initialise()
    {
        // Set up the fieldMetaData
        if (fields.size() == 0)
        {
            fieldMetaData = null;
        }
        else
        {
            fieldMetaData = new FieldMetaData[fields.size()];
            for (int i=0; i<fieldMetaData.length; i++)
            {
                fieldMetaData[i] = (FieldMetaData) fields.get(i);
                fieldMetaData[i].initialise();
            }
        }

        // Set up the columnMetaData
        if (columns.size() == 0)
        {
            columnMetaData = null;
        }
        else
        {
            columnMetaData = new ColumnMetaData[columns.size()];
            for (int i=0; i<columnMetaData.length; i++)
            {
                columnMetaData[i] = (ColumnMetaData) columns.get(i);
                columnMetaData[i].initialise();
            }
        }

        setInitialised();
    }

    // -------------------------------- Mutators -------------------------------

    /**
     * Add a new Field.
     */
    public void addField(FieldMetaData fmd)
    {
        fields.add(fmd);
    }

    /**
     * Accessor for fieldMetaData
     * @return Returns the fieldMetaData.
     */
    public final FieldMetaData[] getFieldMetaData()
    {
        return fieldMetaData;
    }

    /**
     * Add a new ColumnMetaData element
     */
    public void addColumn(ColumnMetaData colmd)
    {
        columns.add(colmd);
    }

    /**
     * Accessor for columnMetaData
     * @return Returns the columnMetaData.
     */
    public final ColumnMetaData[] getColumnMetaData()
    {
        return columnMetaData;
    }

    /**
     * Accessor for the name of the unique constraint
     * @return Returns the name.
     */
    public final String getName()
    {
        return name;
    }

    /**
     * Accessor for the table.
     * @return Returns the table..
     */
    public final String getTable()
    {
        return table;
    }

    /**
     * Accessor for whether the unique constraint is deferred.
     * @return Returns whether the unique constraint is deferred
     */
    public final boolean isDeferred()
    {
        return deferred;
    }

    // -------------------------------- Utilities ------------------------------

    /**
     * Returns a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        return toString("","");
    }

    /**
     * Returns a string representation of the object.
     * This can be used as part of a facility to output a MetaData file. 
     * @param prefix prefix string
     * @param indent indent string
     * @return a string representation of the object.
     */
    public String toString(String prefix,String indent)
    {
        StringBuffer sb = new StringBuffer();
        sb.append(prefix).append("<unique");
        if (table != null)
        {
            sb.append(" table=\"" + table + "\"");
        }
        if (deferred)
        {
            sb.append(" deferred=\"true\"");
        }
        sb.append(name != null ? (" name=\"" + name + "\">\n") : ">\n");

        // Add fields
        for (int i=0;i<fields.size();i++)
        {
            FieldMetaData fmd = (FieldMetaData)fields.get(i);
            sb.append(fmd.toString(prefix + indent,indent));
        }

        // Add columns
        for (int i=0;i<columns.size();i++)
        {
            ColumnMetaData colmd = (ColumnMetaData)columns.get(i);
            sb.append(colmd.toString(prefix + indent,indent));
        }

        // Add extensions
        sb.append(super.toString(prefix + indent,indent));

        sb.append(prefix).append("</unique>\n");
        return sb.toString();
    }
}