/**********************************************************************
Copyright (c) 2005 Andy Jefferson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
    ...
**********************************************************************/
package org.jpox.metadata;

import java.util.ArrayList;
import java.util.List;

/**
 * Datastore tables can have many columns. Not all of these columns are
 * necessarily mapped via an O/R mapper. This element allows for specification
 * of columns that aren't mapped and that a default value may need supplying on
 * any INSERTs.
 * 
 * <H3>MetaData Element</H3> 
 * The MetaData element represented here is as follows 
 * <PRE>
 * &lt;!ELEMENT unmapped (extension*, column*, extension*)&gt;
 * &lt;!ATTLIST unmapped column CDATA #IMPLIED&gt;
 * </PRE>
 *
 * @since 1.1
 * @version $Revision: 1.1 $
 */
public class UnmappedMetaData extends ExtendableMetaData
{
    /**
     * column elements
     */
    final List columns = new ArrayList();

    /**
     * the column name.
     */
    String column;

    // -------------------------------------------------------------------------
    // Fields below here are not represented in the output MetaData. They are
    // for use internally in the operation of the JDO system. The majority are
    // for convenience to save iterating through the fields since the fields
    // are fixed once initialised.

    /**
     * Contains the metadata for column
     */
    protected ColumnMetaData[] columnMetaData;

    /**
     * Constructor for the unmapped definition.
     * @param column The column name (when only a single column)
     */
    public UnmappedMetaData(final String column)
    {
        super(null);    // Ignore parent

        this.column = column;
    }

    /**
     * Method to initialise the object, creating internal convenience arrays.
     * Initialises all sub-objects.
     */
    public void initialise()
    {
        // Cater for user specifying column name, or columns
        if (columns.size() == 0 && column != null)
        {
            columnMetaData = new ColumnMetaData[1];
            columnMetaData[0] = new ColumnMetaData(this, column);
        }
        else
        {
            columnMetaData = new ColumnMetaData[columns.size()];
            for (int i=0; i<columnMetaData.length; i++)
            {
                columnMetaData[i] = (ColumnMetaData) columns.get(i);
                columnMetaData[i].initialise();
            }
        }

        setInitialised();
    }

    /**
     * Add a new ColumnMetaData element
     */
    public void addColumn(ColumnMetaData colmd)
    {
        columns.add(colmd);
    }

    /**
     * Accessor for columnMetaData
     * @return Returns the columnMetaData.
     */
    public final ColumnMetaData[] getColumnMetaData()
    {
        return columnMetaData;
    }

    // ----------------------------- Utilities ---------------------------------

    /**
     * Accessor for a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        return toString("","");
    }

    /**
     * Returns a string representation of the object using a prefix
     * @param prefix prefix string
     * @param indent indent string
     * @return a string representation of the object.
     */
    public String toString(String prefix,String indent)
    {
        StringBuffer sb = new StringBuffer();
        sb.append(prefix).append("<unmapped>");

        // Add columns
        for (int i=0; i<columns.size(); i++)
        {
            ColumnMetaData c = (ColumnMetaData)columns.get(i);
            sb.append(c.toString(prefix + indent,indent));
        }

        // Add extensions
        sb.append(super.toString(prefix + indent,indent));

        sb.append(prefix).append("</unmapped>\n");
        return sb.toString();
    }
}