/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved. 
This program and the accompanying materials are made available under
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2004 Andy Jefferson - added toString(), MetaData docs, javadocs.
2004 Andy Jefferson - added unique, indexed
    ...
**********************************************************************/
package org.jpox.metadata;

/**
 * This element specifies the mapping for the element component of arrays and
 * collections.
 * 
 * If only one column is mapped, and no additional information is needed for the
 * column, then the column attribute can be used. Otherwise, the column
 * element(s) are used.
 * 
 * The serialized attribute specifies that the key values are to be serialized
 * into the named column.
 * 
 * The foreign-key attribute specifies the name of a foreign key to be
 * generated.
 *
 * <H3>MetaData Element</H3> 
 * The MetaData element represented here is as follows
 * <PRE> 
 * &lt;!ELEMENT value (extension*,embedded?,column*,foreign-key?,index?,unique?,extension*)&gt;
 * &lt;!ATTLIST value column CDATA #IMPLIED&gt;
 * &lt;!ATTLIST value serialized (true|false) #IMPLIED&gt;
 * &lt;!ATTLIST value delete-action (restrict|cascade|null|default|none) #IMPLIED&gt;
 * &lt;!ATTLIST value foreign-key CDATA #IMPLIED&gt;
 * &lt;!ATTLIST value indexed (true|false|unique) #IMPLIED&gt;
 * &lt;!ATTLIST value unique (true|false) #IMPLIED&gt;
 * </PRE> 
 * 
 * @since 1.1
 * @version $Revision: 1.3 $
 */
public class ValueMetaData extends AbstractElementMetaData
{
    /**
     * Constructor.
     * @param column The column name
     * @param embedded Whether it is embedded
     * @param serialized Whether it is serialized
     * @param deleteAction attribute delete-action value
     * @param indexed Whether to index this
     * @param unique Whether to add a unique constraint
     */
    public ValueMetaData(String column,
                         String embedded,
                         String serialized,
                         String deleteAction,
                         String indexed,
                         String unique)
    {
        super(column, embedded, serialized, deleteAction, null, indexed, unique);
    }

    // ------------------------------- Utilities -------------------------------

    /**
     * Accessor for a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        return toString("","");
    }

    /**
     * Returns a string representation of the object using a prefix
     * This can be used as part of a facility to output a MetaData file. 
     * @param prefix prefix string
     * @param indent indent string
     * @return a string representation of the object.
     */
    public String toString(String prefix,String indent)
    {
        // Field needs outputting so generate metadata
        StringBuffer sb = new StringBuffer();
        sb.append(prefix).append("<value serialized=\"" + serialized + "\"");
        if (column != null)
        {
            sb.append("\n");
            sb.append(prefix).append("       column=\"" + column + "\"");
        }
        sb.append(">\n");

        // Add columns
        for (int i=0;i<columns.size();i++)
        {
            ColumnMetaData colmd=(ColumnMetaData)columns.get(i);
            sb.append(colmd.toString(prefix + indent,indent));
        }

        // Add index metadata
        if (indexMetaData != null)
        {
            sb.append(indexMetaData.toString(prefix + indent,indent));
        }

        // Add unique metadata
        if (uniqueMetaData != null)
        {
            sb.append(uniqueMetaData.toString(prefix + indent,indent));
        }

        // Add embedded metadata
        if (embeddedMetaData != null)
        {
            sb.append(embeddedMetaData.toString(prefix + indent,indent));
        }

        // Add foreign-key metadata
        if (foreignKeyMetaData != null)
        {
            sb.append(foreignKeyMetaData.toString(prefix + indent,indent));
        }

        // Add extensions
        sb.append(super.toString(prefix + indent,indent));

        sb.append(prefix).append("</value>\n");
        return sb.toString();
    }
}