/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved. 
This program and the accompanying materials are made available under
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2004 Andy Jefferson - added toString(), MetaData docs, javadocs.
2004 Andy Jefferson - added multiple column options
    ...
**********************************************************************/
package org.jpox.metadata;

import java.util.ArrayList;
import java.util.List;

/**
 * Three common strategies for versioning instances are supported by standard
 * metadata. These include state-comparison, timestamp, and version-number.
 * 
 * State-comparison involves comparing the values in specific columns to
 * determine if the database row was changed.
 * 
 * Timestamp involves comparing the value in a date-time column in the table.
 * The first time in a transaction the row is updated, the timestamp value is
 * updated to the current time.
 * 
 * Version-number involves comparing the value in a numeric column in the table.
 * The first time in a transaction the row is updated, the version-number column
 * value is incremented.
 * 
 * <H3>MetaData Element</H3> 
 * The MetaData element represented here is as follows 
 * <PRE>
 * &lt;!ELEMENT version (extension*, column*, index?, extension*)&gt;
 * &lt;!ATTLIST version strategy CDATA #IMPLIED&gt;
 * &lt;!ATTLIST version column CDATA #IMPLIED&gt;
 * &lt;!ATTLIST version indexed (true|false|unique) #IMPLIED&gt;
 * </PRE>
 * @version $Revision: 1.5 $
 */
public class VersionMetaData extends ExtendableMetaData
{
    /**
     * strategy
     */
    private final VersionStrategy versionStrategy;
    
    /**
     * column name
     */
    private String column;

    /**
     * Columns ColumnMetaData
     */
    protected final List columns = new ArrayList();

    /**
     * IndexMetaData.
     */
    protected IndexMetaData indexMetaData;

    /**
     * Indexed value
     */
    protected IndexedValue indexed = null;

    // -------------------------------------------------------------------------
    // Fields below here are not represented in the output MetaData. They are
    // for use internally in the operation of the JDO system. The majority are
    // for convenience to save iterating through the fields since the fields
    // are fixed once populated.

    /**
     * Contains the metadata for column
     */
    protected ColumnMetaData[] columnMetaData;

    /**
     * Constructor. 
     * @param versionStrategy Strategy for versioning
     * @param column Column name
     */
    public VersionMetaData(final String versionStrategy,
                           final String column,
                           final String indexed)
    {
        super(null);    // Don't bother with parent

        if (versionStrategy == null ||
            VersionStrategy.getVersionStrategy(versionStrategy) == null)
        {
            throw new RuntimeException(LOCALISER.msg("MetaData.Version.StrategyNotSpecifiedError"));
        }
        this.versionStrategy = VersionStrategy.getVersionStrategy(versionStrategy);
        this.column = column;
        this.indexed = IndexedValue.getIndexedValue(indexed);
    }

    /**
     * Initialisation method. This should be called AFTER using the populate
     * method if you are going to use populate. It creates the internal
     * convenience arrays etc needed for normal operation.
     */
    public void initialise()
    {
        // Interpret the "indexed" value to create our IndexMetaData where it wasn't specified that way
        if (indexMetaData == null && columnMetaData != null && indexed != null && indexed != IndexedValue.FALSE)
        {
            indexMetaData = new IndexMetaData(null, null, (indexed == IndexedValue.UNIQUE) ? "true" : "false");
            for (int i=0;i<columnMetaData.length;i++)
            {
                indexMetaData.addColumn(columnMetaData[i]);
            }
        }
        if (indexMetaData != null)
        {
            indexMetaData.initialise();
        }

        // Cater for user specifying column name, or columns
        if (columns.size() == 0 && column != null)
        {
            columnMetaData = new ColumnMetaData[1];
            columnMetaData[0] = new ColumnMetaData(this, column);
        }
        else
        {
            columnMetaData = new ColumnMetaData[columns.size()];
            for (int i=0; i<columnMetaData.length; i++)
            {
                columnMetaData[i] = (ColumnMetaData) columns.get(i);
                columnMetaData[i].initialise();
            }
        }
    }

    // ------------------------------ Accessors --------------------------------

    /**
     * Acessor for the columns
     * @return Returns the columnMetaData.
     */
    public final ColumnMetaData[] getColumnMetaData()
    {
        return columnMetaData;
    }

    /**
     * Accessor for versionStrategy
     * @return Returns the versionStrategy.
     */
    public final VersionStrategy getVersionStrategy()
    {
        return versionStrategy;
    }

    /**
     * Accessor for indexMetaData
     * @return Returns the indexMetaData.
     */
    public final IndexMetaData getIndexMetaData()
    {
        return indexMetaData;
    }

    // ------------------------------ Mutators ---------------------------------

    /**
     * Add a new ColumnMetaData element
     */
    public void addColumn(ColumnMetaData colmd)
    {
        columns.add(colmd);
        columnMetaData = new ColumnMetaData[columns.size()];
        for (int i=0; i<columnMetaData.length; i++)
        {
            columnMetaData[i] = (ColumnMetaData) columns.get(i);
        }
    }

    /**
     * Mutator for the index MetaData 
     * @param indexMetaData The indexMetaData to set.
     */
    public final void setIndexMetaData(IndexMetaData indexMetaData)
    {
        this.indexMetaData = indexMetaData;
    }

    // ------------------------------ Utilities --------------------------------

    /**
     * Accessor for a string representation of the object.
     * @return a string representation of the object.
     */
    public String toString()
    {
        return toString("","");
    }

    /**
     * Returns a string representation of the object using a prefix
     * This can be used as part of a facility to output a MetaData file. 
     * @param prefix prefix string
     * @param indent indent string
     * @return a string representation of the object.
     */
    public String toString(String prefix,String indent)
    {
        // Field needs outputting so generate metadata
        StringBuffer sb = new StringBuffer();
        sb.append(prefix).append("<version strategy=\"" + versionStrategy.toString() + "\"" +
            indexed != null ? (" indexed=\"" + indexed.toString() + "\"") : "" +
            ">\n");

        // Add columns
        for (int i=0; i<columns.size(); i++)
        {
            ColumnMetaData c = (ColumnMetaData)columns.get(i);
            sb.append(c.toString(prefix + indent,indent));
        }

        // Add index
        if (indexMetaData != null)
        {
            sb.append(indexMetaData.toString(prefix + indent,indent));
        }

        // Add extensions
        sb.append(super.toString(prefix + indent,indent));

        sb.append(prefix).append("</version>\n");
        return sb.toString();
    }
}