/**********************************************************************
Copyright (c) 2004 Ralf Ullrich and others. All rights reserved. 
This program and the accompanying materials are made available under
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
...
**********************************************************************/
package org.jpox.poid;

import java.sql.Connection;
import java.util.Properties;

/**
 * This generator uses a Java implementation of DCE UUIDs to create unique
 * identifiers without the overhead of additional database transactions or even
 * an open database connection. The identifiers are Strings of the form
 * "LLLLLLLL-MMMM-HHHH-CCCC-NNNNNNNNNNNN" where 'L', 'M', 'H', 'C' and 'N' are
 * the DCE UUID fields named time low, time mid, time high, clock sequence and
 * node.
 * 
 * This generator can be used in concurrent applications. It is especially
 * useful in situations where large numbers of transactions within a certain
 * amount of time have to be made, and the additional overhead of synchronizing
 * the concurrent creation of unique identifiers through the database would
 * break performance limits.
 * 
 * There are no properties for this PoidGenerator.
 * 
 * Note: Due to limitations of the available Java API there is a chance of less
 * than 1:2^62 that two concurrently running JVMs will produce the same
 * identifiers, which is in practical terms never, because your database server
 * will have crashed a million times before this happens.
 * 
 * @version $Revision: 1.1 $
 */
public class AUIDPoidGenerator implements PoidGenerator
{
    /**
     * AUIDPoidGenerator will reserve only one Poid to the PoidBlock.
     * 
     * @see org.jpox.poid.PoidGenerator#reserveBlock()
     * @return PoidBlock
     */
    public PoidBlock reserveBlock()
    {
        return new PoidBlock(new Object[]{new AUID().toString()});
    }

    /**
     * AUIDPoidGenerator reserves <code>size</code> Poids to the PoidBlock.
     * 
     * @param size
     *            The block size
     * @see org.jpox.poid.PoidGenerator#reserveBlock(long)
     * @return The reserved block
     */
    public PoidBlock reserveBlock(long size)
    {
        Object[] ids = new Object[(int) size];
        for (int i = 0; i < size; i++)
        {
            ids[i] = new AUID().toString();
        }
        return new PoidBlock(ids);
    }

    /**
     * Returns false, because AUIDPoidGenerator does not need a connection.
     * 
     * @see org.jpox.poid.PoidGenerator#requiresConnection()
     * @return always false.
     */
    public boolean requiresConnection()
    {
        return false;
    }

    /**
     * Does nothing, because AUIDPoidGenerator does not need a connection.
     * 
     * @see org.jpox.poid.PoidGenerator#setConnection(Connection)
     * @param conn
     *            The Connection to the datastore
     */
    public void setConnection(Connection conn)
    {
        // nothing to do here
    }

    /**
     * Does nothing, because AUIDPoidGenerator has no properties.
     * 
     * @see org.jpox.poid.PoidGenerator#setGeneratorProperties(Properties)
     * @param properties
     *            The properties
     */
    public void setGeneratorProperties(Properties properties)
    {
        // no properties needed
    }
}
