/**********************************************************************
Copyright (c) 2005 Erik Bengtson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution.

Contributors:
    ...
**********************************************************************/
package org.jpox.poid;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;

/**
 * This generator Datastore UUID-HEX. 
 *
 * @version $Revision: 1.1 $ 
 */
public final class DatastoreUUIDHexPoidGenerator 
    implements PoidGenerator, PoidRepositoryCreator, PoidJPOXPlugin
{
    protected static final Localiser LOCALISER=Localiser.getInstance("org.jpox.poid.Localisation");
    protected static final Localiser LOCALISER_RDBMS=Localiser.getInstance("org.jpox.store.rdbms.Localisation");

    private Connection connection;
    private RDBMSManager storeMgr;
 
    private int keyCacheSize = 10;
 
    /**
     * PoidGenerator will reserve a block of ids to the PoidBlock.
     * @return The reserved block
     */
    public PoidBlock reserveBlock()
    {
        PreparedStatement ps = null;
        ResultSet rs = null;
        List oid = new ArrayList();
        try
        {
            // Find the next ID from the database
            String stmt=storeMgr.getDatabaseAdapter().getSelectNewUUIDStmt();
            if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
            {
                JPOXLogger.RDBMS_SQL.debug(stmt);
            }
            ps = connection.prepareStatement(stmt);

            for (int i=1; i<keyCacheSize; i++)
            {
	            long startTime = System.currentTimeMillis();
	            rs = ps.executeQuery();
	            if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
	            {
	                JPOXLogger.RDBMS_SQL.debug(LOCALISER_RDBMS.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
	            }
	 
	            String nextId;
	            if (rs.next())
	            {
	                nextId = rs.getString(1);
	                oid.add(nextId);
	            }
            }
            return new PoidBlock(oid);
        }
        catch (SQLException e)
        {
            throw new PoidException(LOCALISER.msg("PoidGenerator.IDAllocationError",e.getMessage()));
        }
        finally
        {
            try
            {
                if (rs != null)
                {
                    rs.close();
                }
                if (ps != null)
                {
                    ps.close();
                }
            }
            catch (SQLException e)
            {
                // non-recoverable error
            }
        }
    }

    /**
     * Reserve a block of ids.
     * @param size Block size
     * @return The reserved block
     */
    public PoidBlock reserveBlock(long size)
    {
        return reserveBlock();
    }

    /**
     * PoidManager will check if PoidGenerator needs a connection.
     * @see org.jpox.poid.PoidGenerator#requiresConnection()
     * @return true SequencePoidGenerator needs a connection
     */
    public boolean requiresConnection()
    {
        return true;
    }

    /**
     * PoidManager will set a connection.
     * @param conn The Connection to the datastore
     * @see org.jpox.poid.PoidGenerator#setConnection(Connection)
     */
    public void setConnection(Connection conn)
    {
        this.connection = conn;
    }

    /**
     * Configure the properties PoidGenerator needs.
     * @param properties The properties
     * @see org.jpox.poid.PoidGenerator#setGeneratorProperties(Properties)
     */
    public void setGeneratorProperties(Properties properties)
    {
        if (properties.get("key-cache-size") != null)
        {
            try
            {
                keyCacheSize = Integer.parseInt((String)properties.get("key-cache-size"));
            }
            catch (Exception e)
            {
                JPOXLogger.RDBMS_SQL.debug(LOCALISER.msg("PoidGenerator.InvalidKeyCacheSize",properties.get("key-cache-size")));
            }
        }
    }

    /**
     * Method to create the repository.
     * @return Whether it was created successfully. 
     * @see org.jpox.poid.PoidRepositoryCreator#createRepository()
     */
    public boolean createRepository()
    {
        return true;
    }

    /**
     * Mutator for the StoreManager.
     * @param storeMgr The Store Manager 
     * @see org.jpox.poid.PoidJPOXPlugin#setStoreManager(org.jpox.store.rdbms.RDBMSManager)
     */
    public void setStoreManager(RDBMSManager storeMgr)
    {
        this.storeMgr = storeMgr;
    }
}