/**********************************************************************
Copyright (c) 2003 Erik Bengtson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - coding standards
2004 Andy Jefferson - change from Integer to Long
2004 Andy Jefferson - removed last-id option since was dangerous in concurrent apps
    ...
**********************************************************************/
package org.jpox.poid;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Properties;

import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;

/**
 * This generator for Relational Databases uses the "select max(column) from
 * table" strategy. The block Poid size is limited to 1. MaxPoidGenerator works
 * with numbers, so clients using this Poid must cast the ID to Long
 * <P>
 * Properties required for this PoidGenerator:
 * <UL>
 * <LI>column-name</LI>
 * <LI>table-name</LI>
 * </UL>
 * </P> 
 *
 * @version $Revision: 1.3 $ 
 */
public class MaxPoidGenerator implements PoidGenerator
{
    protected static final Localiser LOCALISER=Localiser.getInstance("org.jpox.store.rdbms.Localisation");

    private Connection connection;
    private Properties properties;
    
    /**
     * MaxPoidGenerator will reserve only one Poid to the PoidBlock.
     * @see org.jpox.poid.PoidGenerator#reserveBlock()
     * @return PoidBlock
     */
    public PoidBlock reserveBlock()
    {
        // search an Id in the database
        PreparedStatement ps = null;
        ResultSet rs = null;
        try
        {
            String stmt=getStatement();
            ps = connection.prepareStatement(stmt);
            if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
            {
                JPOXLogger.RDBMS_SQL.debug(stmt);
            }

            long startTime = System.currentTimeMillis();
            rs = ps.executeQuery();
            if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
            {
                JPOXLogger.RDBMS_SQL.debug(LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
            }
 
            if (!rs.next())
            {
                return new PoidBlock(new Object[] { new Long(1) });
            }
 
            return new PoidBlock(new Object[] { new Long(rs.getLong(1) + 1)});
        }
        catch (SQLException e)
        {
            //TODO adds a message correspondent to the exception. 
            //we need to work to create user friendly messages
            throw new PoidException(e.getMessage());
        }
        finally
        {
            try
            {
                if (rs != null)
                {
                    rs.close();
                }
                if (ps != null)
                {
                    ps.close();
                }
            }
            catch (SQLException e)
            {
                // no recoverable error
            }           
        }
    }

    /**
     * @param size The block size
     * @see org.jpox.poid.PoidGenerator#reserveBlock(long)
     * @return The reserved block
     */
    public PoidBlock reserveBlock(long size)
    {
        return reserveBlock();
    }

    /**
     * PoidManager will check if MaxPoidGenerator needs a connection.
     * @see org.jpox.poid.PoidGenerator#requiresConnection()
     * @return Whether MaxPoidGenerator needs a connection
     */
    public boolean requiresConnection()
    {
        return true;
    }

    /**
     * PoidManager will set a connection.
     * @param conn The Connection to the datastore
     * @see org.jpox.poid.PoidGenerator#setConnection(Connection)
     */
    public void setConnection(Connection conn)
    {
        this.connection = conn;
    }

    /**
     * Configure the properties MaxPoidGenerator needs.
     * Properties required for this PoidGenerator: COLUMN_NAME, TABLE_NAME
     *
     * @param properties The properties
     * @see org.jpox.poid.PoidGenerator#setGeneratorProperties(Properties)
     */
    public void setGeneratorProperties(Properties properties)
    {
        this.properties = properties;
    }

    /**
     * return the SQL statement.
     * @return statement
     */
    private String getStatement()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("SELECT ");
        stmt.append("max(");
        stmt.append(properties.get("column-name"));
        stmt.append(") ");
        stmt.append("from ");
        stmt.append(properties.get("table-name"));  
        return stmt.toString();
    } 
}