/**********************************************************************
Copyright (c) 2003 Erik Bengtson and others. All rights reserved.
This program and the accompanying materials are made available under
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - coding standards
2004 Andy Jefferson - fixed creation of repository when block failed
2004 Andy Jefferson - removed MetaData requirement
    ...
**********************************************************************/
package org.jpox.poid;

import java.util.Enumeration;
import java.util.Properties;

import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;

/**
 * PoidManager takes care of keeping a block of ids for a given PoidGenerator,
 * when the block is empty, the PoidGenerator is called to give a new block
 * of ids.
 * <P>
 * PoidManager has a 1-1 relationship with PoidGenerators
 *
 * @version $Revision $ 
 */
public class PoidManager
{
    /** Localisation of messages */
    protected static final Localiser LOCALISER=Localiser.getInstance("org.jpox.poid.Localisation");

    private PoidBlock poidBlock;

    private Properties properties;

    private String poidGeneratorName;

    private PoidJDBCConnectionProvider jdbcConnectionProvider;

    private RDBMSManager storeMgr;

    /**
     * Get Poid from the PoidBlock.
     * 
     * @return Poid the Poid
     */
    public Poid getPoid()
    {
        PoidBlock poidBlock = getPoidBlock();
        Poid poid = poidBlock.next();
 
        return poid;
    }

    /**
     * Get the PoidBlock for an type
     * 
     * @return PoidBlock the PoidBlock
     */
    private void obtainPoidBlock()
    {
        PoidGenerator poidGenerator = PoidFactory.getPoidGenerator(getPoidGeneratorName());
        poidGenerator.setGeneratorProperties(getProperties());
        if (poidGenerator instanceof PoidJPOXPlugin)
        {
            ((PoidJPOXPlugin)poidGenerator).setStoreManager(this.storeMgr);
        }

        // Try getting the block
        boolean repository_exists=true;
        try
        {            
            if (poidGenerator.requiresConnection())
            {
                poidGenerator.setConnection(getJDBCConnectionProvider().getPoidJDBCConnection());
            }
            try
            {
                poidBlock = poidGenerator.reserveBlock();
            }
            catch (PoidException poidex)
            {
                JPOXLogger.RDBMS.info(LOCALISER.msg("PoidManager.BlockAllocationError", poidex.getMessage()));

                // attempt to obtain the block of unique identifiers is invalid
                if (poidGenerator instanceof PoidRepositoryCreator)
                {
                    repository_exists = false;
                }
                else
                {
                    throw poidex;
                }
            }
        }
        finally
        {
            if (poidGenerator.requiresConnection())
            {
                getJDBCConnectionProvider().releaseConnection();
            }
        }

        // If repository didn't exist, try creating it and then get block
        if (!repository_exists)
        {
            try
            {
                if (poidGenerator.requiresConnection())
                {
                    poidGenerator.setConnection(getJDBCConnectionProvider().getPoidJDBCConnection());
                }

                JPOXLogger.RDBMS.info(LOCALISER.msg("PoidManager.CreatingRepository"));
                if (!((PoidRepositoryCreator)poidGenerator).createRepository())
                {
                    throw new PoidException(LOCALISER.msg("PoidManager.RepositoryOfIdsInvalid"));
                }
                else
                {
                    poidBlock = poidGenerator.reserveBlock();
                }
            }
            finally
            {
                if (poidGenerator.requiresConnection())
                {
                    getJDBCConnectionProvider().releaseConnection();
                }
            }
        }
    }

    /**
     * Accessor for the properties.
     * @return The Properties
     */
    private Properties getProperties()
    {
        return properties;
    }

    /**
     * Sets the properties.
     * @param properties The properties to set
     */
    public void setProperties(Properties properties)
    {
        this.properties = properties;
    }

    /**
     * Returns the poidGeneratorName.
     * @return The Poid Generator name
     */
    private String getPoidGeneratorName()
    {
        return poidGeneratorName;
    }

    /**
     * Sets the poidGeneratorName.
     * @param poidGeneratorName The poidGeneratorName to set
     */
    public void setPoidGeneratorName(String poidGeneratorName)
    {
        this.poidGeneratorName = poidGeneratorName;
    }

    /**
     * Returns the poidBlock.
     * @return The POID Block
     */
    private PoidBlock getPoidBlock()
    {
        if (poidBlock == null || !poidBlock.hasNext())
        {
            obtainPoidBlock();
        }
        return poidBlock;
    }

    /**
     * Accessor for the HashCode
     * @return an HashCode
     */
    public int hashCode()
    {
        Properties p = new Properties();
        p.putAll(getProperties());
        int hashCode = (getPoidGeneratorName().hashCode() >> 32);
        Enumeration elements = p.elements();
        while (elements.hasMoreElements())
        {
            hashCode = hashCode + (elements.nextElement().hashCode() >> 32);
        }
        return hashCode;
    }
 
    /**
     * Accessor for the connection provider. 
     * @return The connection provider.
     */
    protected PoidJDBCConnectionProvider getJDBCConnectionProvider()
    {
        return jdbcConnectionProvider;
    }

    /**
     * Mutator for the connection provider. 
     * @param provider The connection provider.
     */
    public void setJDBCConnectionProvider(PoidJDBCConnectionProvider provider)
    {
        jdbcConnectionProvider = provider;
    }

    /**
     * Accessor for the Store Manager
     * @param storeMgr The storeMgr to set.
     */
    public void setStoreManager(RDBMSManager storeMgr)
    {
        this.storeMgr = storeMgr;
    }
}