/**********************************************************************
Copyright (c) 2003 Erik Bengtson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution.

Contributors:
2003 Andy Jefferson - coding standards
2004 Andy Jefferson - changed to provide Sequence generator for all databases
2004 Andy Jefferson - removed the MetaData requirement
    ...
**********************************************************************/
package org.jpox.poid;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;

/**
 * This generator Sequence uses the "select <sequence>.nextval from dual", 
 * "SELECT nextval('<sequence>')" strategy. 
 * It's recommended to use this generator in databases supporting sequences.
 * SequencePoidGenerator works with numbers, so clients using this Poid must cast the ID to Long
 * <P>
 * <B>Required properties</B>
 * <UL>
 * <LI><U>sequence-name</U> - the sequence name</LI>
 * </UL>
 * </P>
 * <P>
 * <B>Optional properties</B>
 * <UL>
 * <LI><U>key-cache-size</U> - number of unique identifiers to cache</LI>
 * <LI><U>key-min-value</U> - determines the minimum value a sequence can generate</LI>
 * <LI><U>key-max-value</U> - determines the maximum value a sequence can generate</LI>
 * <LI><U>key-start-with</U> - the initial value for the sequence</LI>
 * <LI><U>key-increment-by</U> - specifies which value is added to the current
 * sequence value to create a new value (default is 1)</LI>
 * <LI><U>key-database-cache-size</U> - specifies how many sequence numbers are to be
 * preallocated and stored in memory for faster access</LI>
 * </UL>
 *
 * @version $Revision: 1.3 $ 
 */
public final class SequencePoidGenerator 
    implements PoidGenerator, PoidRepositoryCreator, PoidJPOXPlugin
{
    protected static final Localiser LOCALISER=Localiser.getInstance("org.jpox.poid.Localisation");
    protected static final Localiser LOCALISER_RDBMS=Localiser.getInstance("org.jpox.store.rdbms.Localisation");

    private Connection connection;
    private Properties properties;
    private RDBMSManager storeMgr;
 
    private int keyCacheSize = 1;
 
    /**
     * SequencePoidGenerator will reserve only one Poid to the PoidBlock.
     * @return The reserved block
     */
    public PoidBlock reserveBlock()
    {
        PreparedStatement ps = null;
        ResultSet rs = null;
        List oid = new ArrayList();
        try
        {
            // Find the next ID from the database
            String stmt=storeMgr.getDatabaseAdapter().getSequenceNextStmt((String)properties.get("sequence-name"));
            if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
            {
                JPOXLogger.RDBMS_SQL.debug(stmt);
            }
            ps = connection.prepareStatement(stmt);

            long startTime = System.currentTimeMillis();
            rs = ps.executeQuery();
            if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
            {
                JPOXLogger.RDBMS_SQL.debug(LOCALISER_RDBMS.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
            }
 
            Long nextId = new Long(0);
            if (rs.next())
            {
                nextId = new Long(rs.getLong(1));
                oid.add(nextId);
            }
            for (int i=1; i<keyCacheSize; i++)
            {
                // keyCacheSize must match key-increment-by otherwise it will
                // cause duplicates keys
                nextId = new Long(nextId.longValue()+1);
                oid.add(nextId);
            }
            return new PoidBlock(oid);
        }
        catch (SQLException e)
        {
            throw new PoidException(LOCALISER.msg("SequencePoidGenerator.CantObtainSequenceId",e.getMessage()));
        }
        finally
        {
            try
            {
                if (rs != null)
                {
                    rs.close();
                }
                if (ps != null)
                {
                    ps.close();
                }
            }
            catch (SQLException e)
            {
                // non-recoverable error
            }
        }
    }

    /**
     * Reserve a block of ids.
     * @param size Block size
     * @return The reserved block
     */
    public PoidBlock reserveBlock(long size)
    {
        return reserveBlock();
    }

    /**
     * PoidManager will check if SequencePoidGenerator needs a connection.
     * @see org.jpox.poid.PoidGenerator#requiresConnection()
     * @return true SequencePoidGenerator needs a connection
     */
    public boolean requiresConnection()
    {
        return true;
    }

    /**
     * PoidManager will set a connection.
     * @param conn The Connection to the datastore
     * @see org.jpox.poid.PoidGenerator#setConnection(Connection)
     */
    public void setConnection(Connection conn)
    {
        this.connection = conn;
    }

    /**
     * Configure the properties SequencePoidGenerator needs.
     * Properties required for this PoidGenerator: sequence-name
     * @param properties The properties
     * @see org.jpox.poid.PoidGenerator#setGeneratorProperties(Properties)
     */
    public void setGeneratorProperties(Properties properties)
    {
        this.properties = properties;
        if (properties.get("key-cache-size") != null)
        {
            try
            {
                keyCacheSize = Integer.parseInt((String)properties.get("key-cache-size"));
            }
            catch (Exception e)
            {
                throw new PoidException(LOCALISER.msg("PoidGenerator.InvalidKeyCacheSize",properties.get("key-cache-size")));
            }
        }
    }

    /**
     * Method to create the sequence.
     * @return Whether it was created successfully. 
     * @see org.jpox.poid.PoidRepositoryCreator#createRepository()
     */
    public boolean createRepository()
    {
        Statement ps = null;
        String stmt=storeMgr.getDatabaseAdapter().getSequenceCreateStmt(
                    (String)properties.get("sequence-name"),
                    (String) properties.get("key-min-value"),
                    (String) properties.get("key-max-value"),
                    (String) properties.get("key-start-with"),
                    (String) properties.get("key-increment-by"),
                    (String) properties.get("key-database-cache-size"));
        try
        {
            if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
            {
                JPOXLogger.RDBMS_SQL.debug(stmt);
            }
            ps = connection.createStatement();

            long startTime = System.currentTimeMillis();
            ps.execute(stmt);
            if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
            {
                JPOXLogger.RDBMS_SQL.debug(LOCALISER_RDBMS.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
            }
        }
        catch (SQLException e)
        {
            JPOXLogger.RDBMS_SQL.error(e);
            throw new PoidException(LOCALISER.msg("SequencePoidGenerator.CantCreateSequence",e.getMessage()) + stmt);
        }
        finally
        {
            try
            {
                if (ps != null)
                {
                    ps.close();
                }
            }
            catch (SQLException e)
            {
                // non-recoverable error
            }           
        }
        return true;
    }

    /**
     * Mutator for the StoreManager.
     * @param storeMgr The Store Manager 
     * @see org.jpox.poid.PoidJPOXPlugin#setStoreManager(org.jpox.store.rdbms.RDBMSManager)
     */
    public void setStoreManager(RDBMSManager storeMgr)
    {
        this.storeMgr = storeMgr;
    }
}