/**********************************************************************
Copyright (c) 2003 Erik Bengtson and others. All rights reserved. 
This program and the accompanying materials are made available under
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2004 Andy Jefferson - removed MetaData requirement
    ...
**********************************************************************/
package org.jpox.poid;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;
import java.util.Properties;

import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.store.rdbms.table.SequenceTable;
import org.jpox.util.Localiser;

/**
 * Identity generator for RDBMS databases that generates ids using a sequence table in the database.
 * This generator is coupled to JPOX core and can't be used in standalone mode.
 * <P>
 * <B>Required properties</B>
 * <UL>
 * <LI><U>table-name</U> - the sequence name</LI>
 * </UL>
 * </P>
 * <P>
 * <B>Optional properties</B>
 * <UL>
 * <LI><U>key-cache-size</U> - number of unique identifiers to cache (defaults to 5)</LI>
 * <LI><U>sequence-table-name</U> - the table name for sequences (defaults to SEQUENCE_TABLE)</LI>
 * <LI><U>sequence-name-column-name</U> - the name for the column that represent sequence names</LI>  
 * <LI><U>sequence-nextval-column-name</U> - the name for the column that represent incrementing sequence values</LI>
 * </UL>
 * 
 * @version $Revision: 1.6 $ 
 */
public final class SequenceTablePoidGenerator 
    implements PoidGenerator, PoidRepositoryCreator, PoidJPOXPlugin
{
    /** Localisation of messages */
    protected static final Localiser LOCALISER=Localiser.getInstance("org.jpox.poid.Localisation");

    private Connection connection;

    /** Properties controlling this generator */
    private Properties properties;

    /** Store manager used by this generator */
    private RDBMSManager storeMgr;

    /** Size of cache of identities. */
    private int keyCacheSize = 5;

    /**
     * SequenceTablePoidGenerator reserves one or more unique identifiers holded by the PoidBlock.
     * @see org.jpox.poid.PoidGenerator#reserveBlock()
     * @return The reserved block
     */
    public PoidBlock reserveBlock()
    {
        // search for an ID in the database
        List oid = new ArrayList();
        try
        {
            SequenceTable st = new SequenceTable(storeMgr, properties);
            st.initialize();
            Long nextId =st.getNextVal((String)properties.get("table-name"),connection,keyCacheSize);
            for (int i=0; i<keyCacheSize; i++)
            {
                oid.add(nextId);
                nextId = new Long(nextId.longValue()+1);
            }
            return new PoidBlock(oid);
        }
        catch (SQLException e)
        {
            throw new PoidException(LOCALISER.msg("SequencePoidGenerator.CantObtainSequenceId",e.getMessage()));
        }
    }

    /**
     * @param size Block size
     * @see org.jpox.poid.PoidGenerator#reserveBlock(long)
     * @return The reserved block
     */
    public PoidBlock reserveBlock(long size)
    {
        return reserveBlock();
    }

    /**
     * PoidManager will check if SequencePoidGenerator needs a connection.
     * @see org.jpox.poid.PoidGenerator#requiresConnection()
     * @return true SequenceTablePoidGenerator needs a connection
     */
    public boolean requiresConnection()
    {
        return true;
    }

    /**
     * PoidManager will set a connection.
     * @param conn The Connection to the datastore
     * @see org.jpox.poid.PoidGenerator#setConnection(Connection)
     */
    public void setConnection(Connection conn)
    {
        this.connection = conn;
    }

    /**
     * Configure the properties SequenceTablePoidGenerator needs.
     * This generator supports the following properties
     * <UL>
     * <LI>table-name</LI>
     * <LI>key-cache-size</LI>
     * <LI>sequence-table-name</LI>
     * <LI>sequence-name-column-name</LI>
     * <LI>sequence-nextval-column-name</LI>
     * </UL>
     * @param properties The properties
     */
    public void setGeneratorProperties(Properties properties)
    {
        this.properties = properties;
        if (properties.get("key-cache-size") != null)
        {
        	try
			{
        		keyCacheSize = Integer.parseInt((String)properties.get("key-cache-size"));
			}
        	catch( Exception e )
			{
        		throw new PoidException(LOCALISER.msg("SequencePoidGenerator.InvalidKeyCacheSize",properties.get("key-cache-size")));
			}
        }
    }

	/**
	 * Method to create the repository for ids to be stored.
	 * @return Whether it was created successfully.
	 */
	public boolean createRepository()
	{
        try
        {
            SequenceTable st = new SequenceTable(storeMgr, properties);
            st.initialize();
            st.create(connection);
            return true;
        }
        catch (SQLException e)
        {
    		throw new PoidException(e.toString());
        }
	}

    /**
     * Mutator for the StoreManager for this sequence table
     * @param storeMgr The Store Manager
     */
    public void setStoreManager(RDBMSManager storeMgr)
    {
        this.storeMgr = storeMgr;
    }
}