/**********************************************************************
Copyright (c) 2005 Andy Jefferson and others. All rights reserved.
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution.

Contributors:
    ...
**********************************************************************/
package org.jpox.poid;

import java.net.InetAddress;
import java.sql.Connection;
import java.util.Properties;

/**
 * Poid generator for a UUID hexadecimal format.
 * Results in Strings of length 32 characters, containing the IP address of the local machine
 * as per the JDO2 spec section 18.6.1.
 * 
 * @version $Revision: 1.3 $
 */
public class UUIDHexPoidGenerator implements PoidGenerator
{
    /**
     * Method to reserve a POID to the PoidBlock.
     * @return PoidBlock
     */
    public PoidBlock reserveBlock()
    {
        return new PoidBlock(new Object[]{getIdentifier()});
    }

    /**
     * Method to reserve "size" POIDs to the PoidBlock.
     * @param size The block size
     * @return The reserved block
     */
    public PoidBlock reserveBlock(long size)
    {
        Object[] ids = new Object[(int) size];
        for (int i = 0; i < size; i++)
        {
            ids[i] = getIdentifier();
        }
        return new PoidBlock(ids);
    }

    /**
     * Accessor for whether this Poid Generator needs a datastore connection.
     * @return always false.
     */
    public boolean requiresConnection()
    {
        return false;
    }

    /**
     * Mutator for the datastore connection to use.
     * Does nothing.
     * @param conn The Connection to the datastore
     */
    public void setConnection(Connection conn)
    {
    }

    /**
     * Mutator for any optional properties.
     * Does nothing.
     * @param properties The properties
     */
    public void setGeneratorProperties(Properties properties)
    {
    }

    /**
     * Create an identifier with the form "IIIIIIIIJJJJJJJJHHHHLLLLLLLLCCCC".
     * Where IIIIIIII is the IP address, JJJJJJJJ is something unique across JVMs,
     * HHHH is the High Time, LLLLLLLL is the low time, and CCCC is a count.
     * @return The identifier
     */
    private String getIdentifier()
    {
        StringBuffer str = new StringBuffer(32);

        str.append(toHex(IP_ADDRESS));
        str.append(toHex(JVM_UNIQUE));
        short timeHigh = (short) (System.currentTimeMillis() >>> 32);
        str.append(toHex(timeHigh));
        int timeLow = (int) System.currentTimeMillis();
        str.append(toHex(timeLow));
        str.append(toHex(getCount()));

        return str.toString();
    }

    /** IP Address of local machine. */
    private static final int IP_ADDRESS;

    static 
    {
        // Calculate the IP address of this machine for use in the UUID
        int ipAddr = 0;
        try
        {
            byte[] ipAddrBytes = InetAddress.getLocalHost().getAddress();
            for (int i=0;i<4;i++)
            {
                ipAddr = (ipAddr << 8) - Byte.MIN_VALUE + ipAddrBytes[i];
            }
        }
        catch (Exception e)
        {
            ipAddr = 0;
        }

        IP_ADDRESS = ipAddr;
    }

    /** Unique value across JVMs on this machine. */
    private static final int JVM_UNIQUE = (int) (System.currentTimeMillis() >>> 8);

    private static short counter = (short) 0;

    /**
     * Simple counter for identities.
     */
    protected short getCount() 
    {
        synchronized(UUIDHexPoidGenerator.class) 
        {
            if (counter < 0)
            {
                counter = 0;
            }
            return counter++;
        }
    }

    /**
     * Method to format an integer into an 8 character string
     * @param value Integer value
     * @return The string value
     */
    protected String toHex(int value) 
    {
        StringBuffer str = new StringBuffer("00000000");
        String hexstr = Integer.toHexString(value);
        str.replace(8 - hexstr.length(), 8, hexstr);
        return str.toString();
    }

    /**
     * Method to format a short into a 4 character string.
     * @param value Short value
     * @return The string value
     */
    protected String toHex(short value)
    {
        StringBuffer str = new StringBuffer("0000");
        String hexstr = Integer.toHexString(value);
        str.replace(4 - hexstr.length(), 4, hexstr);
        return str.toString();
    }
}