/**********************************************************************
Copyright (c) 2003 David Jencks and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - coding standards
    ...
**********************************************************************/
package org.jpox.resource;

import java.io.PrintWriter;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;

import javax.jdo.JDOConnection;
import javax.jdo.JDOException;
import javax.jdo.JDOUnsupportedOptionException;
import javax.jdo.Transaction;
import javax.resource.ResourceException;
import javax.resource.spi.ConnectionEvent;
import javax.resource.spi.ConnectionEventListener;
import javax.resource.spi.ConnectionRequestInfo;
import javax.resource.spi.LocalTransaction;
import javax.resource.spi.ManagedConnection;
import javax.resource.spi.ManagedConnectionMetaData;
import javax.resource.spi.security.PasswordCredential;
import javax.security.auth.Subject;
import javax.transaction.xa.XAResource;
import javax.resource.spi.ManagedConnectionFactory;

import org.jpox.AbstractPersistenceManager;
import org.jpox.PersistenceManager;
import org.jpox.util.JPOXLogger;

/**
 * Implementation of ManagedConnection persistence manager
 *
 * @version $Revision: 1.2 $
 */
public class ManagedConnectionImpl extends AbstractPersistenceManager 
    implements LocalTransaction, ManagedConnection
{
    private final PasswordCredential credential;

    private final List handles = new ArrayList();

    private final Collection cels = new ArrayList();

    public ManagedConnectionImpl(ManagedConnectionFactory mcf, PasswordCredential credential)
    throws ResourceException
    {
        super((ManagedConnectionFactoryImpl)mcf, 
            credential == null ? null : credential.getUserName(),  
            credential == null ? null : new String(credential.getPassword()));
        this.credential = credential;
        tx = new ManagedTransaction(this, credential);
    }

    PasswordCredential getPasswordCredential()
    {
        return credential;
    }

    ManagedConnectionFactoryImpl getManagedConnectionFactory()
    {
        return (ManagedConnectionFactoryImpl)apmf;
    }

    // implementation of javax.resource.spi.ManagedConnection interface

    /**
     * Destroy method
     * @exception javax.resource.ResourceException <description>
     */
    public void destroy()
    throws ResourceException
    {
        if (tx != null) 
        {
            try 
            {
                ((ManagedTransaction)tx).closeConnection();
            }
            catch (SQLException e)
            {
                throw new ResourceException("Problem destroying ManagedConnection: " + e);
            }
        }
        close();
    }

    /**
     * Cleanup method
     * @exception javax.resource.ResourceException <description>
     */
    public synchronized void cleanup()
    throws ResourceException
    {
        for (Iterator i = handles.iterator(); i.hasNext();)
        {
            ((PersistenceManagerImpl)i.next()).setManagedConnection(null);
        }
        handles.clear();
    }

    /**
     * Accessor for the connection
     * @param subject <description>
     * @param cri <description>
     * @return <description>
     * @exception javax.resource.ResourceException <description>
     */
    public Object getConnection(Subject subject, ConnectionRequestInfo cri)
    throws ResourceException
    {
        PasswordCredential pc = getManagedConnectionFactory().getPasswordCredential(subject);
        if (credential != null && !credential.equals(pc))
        {
            throw new ResourceException("wrong subject");
        }
        
        Object o = new PersistenceManagerImpl(getManagedConnectionFactory(), this);
        handles.add(0, o);
        return o;
    }

    /**
     * Accessor for the Log
     * @return The Log writer
     * @exception javax.resource.ResourceException <description>
     */
    public PrintWriter getLogWriter()
    throws ResourceException
    {
        // TODO: implement this javax.resource.spi.ManagedConnection method
        return null;
    }

    /**
     * Mutator for the Log
     * @param writer PrintWriter to use for Log
     * @exception javax.resource.ResourceException <description>
     */
    public void setLogWriter(PrintWriter writer)
    throws ResourceException
    {
        // TODO: implement this javax.resource.spi.ManagedConnection method
    }

    /**
     * Mutator to add a connection listener
     * @param cel <description>
     */
    public void addConnectionEventListener(ConnectionEventListener cel)
    {
        synchronized (cels)
        {
            cels.add(cel);
        }
    }

    /**
     * Mutator to remove a connection listener
     * @param cel <description>
     */
    public void removeConnectionEventListener(ConnectionEventListener cel)
    {
        synchronized (cels)
        {
            cels.remove(cel);
        }
    }

    /**
     * Mutator to associate a connection
     * @param c <description>
     * @exception javax.resource.ResourceException <description>
     */
    public void associateConnection(Object c)
    throws ResourceException
    {
        if (!(c instanceof PersistenceManagerImpl)) 
        {
            throw new ResourceException("wrong Connection type!");
        }
        ((PersistenceManagerImpl)c).setManagedConnection(this);
        handles.add(0, c);
    }

    /**
     * Accessor for the local transaction
     * @return <description>
     * @exception javax.resource.ResourceException <description>
     */
    public LocalTransaction getLocalTransaction()
    throws ResourceException
    {
        return this;
    }

    /**
     * Accessor for the JDO transaction
     * @return The JDO transaction
     **/
    public Transaction getJdoTransaction() 
    {
        return tx;
    }

    /**
     * Accessor for the connection MetaData
     * @return <description>
     * @exception javax.resource.ResourceException <description>
     */
    public ManagedConnectionMetaData getMetaData()
    throws ResourceException
    {
        throw new ResourceException("Not Yet Implemented");
    }

    /**
     * Accessor for the XA resource
     * @return <description>
     * @exception javax.resource.ResourceException <description>
     */
    public XAResource getXAResource()
    throws ResourceException
    {
        throw new ResourceException("Not Yet Implemented");
    }

    // implementation of javax.resource.spi.LocalTransaction interface

    /**
     * Method to start the Transaction
     * @exception javax.resource.ResourceException <description>
     */
    public void begin()
    throws ResourceException
    {
        try 
        {
            tx.begin();
        }
        catch (JDOException e)
        {
            throw new ResourceException("JDOException: " + e);
        }
    }

    /**
     * Method to commit the Transaction
     * @exception javax.resource.ResourceException <description>
     */
    public void commit()
    throws ResourceException
    {
        try 
        {
            tx.commit();
        }
        catch (JDOException e)
        {
            if (JPOXLogger.JDO.isInfoEnabled())
            {
                JPOXLogger.JDO.info("Exception during commit: ", e);
            }
            throw new ResourceException("JDOException: " + e);
        }
    }

    /**
     * Method to rollback the Transaction
     * @exception javax.resource.ResourceException <description>
     */
    public void rollback()
    throws ResourceException
    {
        try 
        {
            tx.rollback();
        }
        catch (JDOException e)
        {
            if (JPOXLogger.JDO.isInfoEnabled())
            {
                JPOXLogger.JDO.info("Exception during rollback: ", e);
            }
            throw new ResourceException("JDOException: " + e);
        }
    }

    //ConnectionEvent management

    void notifyClosed(PersistenceManagerImpl handle)
    {
        ConnectionEvent ce = new ConnectionEvent(this, ConnectionEvent.CONNECTION_CLOSED, null);
        ce.setConnectionHandle(handle);
        Collection localCels = null;
        synchronized (cels)
        {
            localCels = new ArrayList(cels);
        }

        for (Iterator i = localCels.iterator(); i.hasNext(); )
        {
            ((ConnectionEventListener)i.next()).connectionClosed(ce);

        }
        handles.remove(handle);
    }

    void notifyTxBegin(PersistenceManagerImpl handle)
    {
        ConnectionEvent ce = new ConnectionEvent(this, ConnectionEvent.LOCAL_TRANSACTION_STARTED, null);
        ce.setConnectionHandle(handle);
        Collection localCels = null;
        synchronized (cels)
        {
            localCels = new ArrayList(cels);
        }

        for (Iterator i = localCels.iterator(); i.hasNext(); )
        {
            ((ConnectionEventListener)i.next()).localTransactionStarted(ce);

        }
    }

    void notifyTxCommit(PersistenceManagerImpl handle)
    {
        ConnectionEvent ce = new ConnectionEvent(this, ConnectionEvent.LOCAL_TRANSACTION_COMMITTED, null);
        ce.setConnectionHandle(handle);
        Collection localCels = null;
        synchronized (cels)
        {
            localCels = new ArrayList(cels);
        }
        for (Iterator i = localCels.iterator(); i.hasNext(); )
        {
            ((ConnectionEventListener)i.next()).localTransactionCommitted(ce);

        }
    }

    void notifyTxRollback(PersistenceManagerImpl handle)
    {
        ConnectionEvent ce = new ConnectionEvent(this, ConnectionEvent.LOCAL_TRANSACTION_ROLLEDBACK, null);
        ce.setConnectionHandle(handle);
        Collection localCels = null;
        synchronized (cels)
        {
            localCels = new ArrayList(cels);
        }

        for (Iterator i = localCels.iterator(); i.hasNext(); )
        {
            ((ConnectionEventListener)i.next()).localTransactionRolledback(ce);

        }
    }

    /**
     * Accessor for the Persistence Manager handle
     * @return The PM handle
     **/
    public PersistenceManager getPMHandle()
    {
        if (handles.size() != 1) 
        {
            if (JPOXLogger.JDO.isInfoEnabled())
            {
                JPOXLogger.JDO.info(LOCALISER.msg("PM.HandleInvalid", "" + handles.size()));
            }
        }
        return (PersistenceManager)handles.iterator().next();
    }

    /**
     * Accessor for whether this PM and the passed on are the same.
     * @param pm The other PM
     * @return Whether they are the same
     **/
    public boolean isSamePM(javax.jdo.PersistenceManager pm)
    {
        return handles.contains(pm);
    }

    /* (non-Javadoc)
     * @see javax.jdo.PersistenceManager#getDataStoreConnection()
     */
    public JDOConnection getDataStoreConnection()
    {
        // TODO Unsupported feature
        throw new JDOUnsupportedOptionException("Not Yet Implemented");
    }
}