/**********************************************************************
Copyright (c) 2003 David Jencks and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - coding standards
2003 Erik Bengtson - optimistic transactions
2004 Andy Jefferson - addition of rollbackOnly
    ...
**********************************************************************/
package org.jpox.resource;

import java.sql.Connection;
import java.sql.SQLException;

import javax.jdo.JDOConnection;
import javax.jdo.JDODataStoreException;
import javax.jdo.JDOFatalDataStoreException;
import javax.jdo.JDOUnsupportedOptionException;
import javax.jdo.JDOUserException;
import javax.jdo.PersistenceManager;
import javax.resource.ResourceException;
import javax.resource.spi.security.PasswordCredential;
import javax.sql.DataSource;
import javax.transaction.Status;
import javax.transaction.Synchronization;

import org.jpox.AbstractPersistenceManagerFactory;
import org.jpox.JDOConnectionImpl;
import org.jpox.Transaction;
import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.util.JPOXLogger;

/**
 * The class <code>ManagedTransaction</code> is an internal jdo transaction
 * class used by ManagedConnectionImpl to perform the tx operations. 
 * It should not be used by clients. The connection handle provides a wrapper
 * to this: the ManagedConnectionImpl wraps this to provide a 
 * javax.resource.spi.LocalTransaction that converts jdo exceptions
 * to ResourceExceptions.
 *
 * @version $Revision: 1.2 $
 */
class ManagedTransaction implements Transaction
{
    private boolean useUpdateLockProperty=false; 

    private final ManagedConnectionImpl mc;
    private final org.jpox.store.rdbms.adapter.DatabaseAdapter dba;
    private final DataSource tds;

    private boolean retainValues;
    private boolean restoreValues;    
    private boolean optimistic;
    private boolean nontransactionalRead;
    private boolean nontransactionalWrite;
    private int isolationLevel;
    private boolean useUpdateLock;

    /**
     * Whether the transaction is marked for rollback. JDO 2.0 section 13.4.5
     */
    private boolean rollbackOnly=false;

    private Synchronization sync = null;
    private Connection conn = null;
    private boolean isJDOConnectionInUse = false;
    private boolean active = false;

    /**
     * Constructor.
     * @param mc The connection
     * @param credential Password credentials
     * @throws ResourceException If an error occurs
     */
    public ManagedTransaction(ManagedConnectionImpl mc, PasswordCredential credential)
    throws ResourceException
    {
        this.mc = mc;
        this.dba = ((RDBMSManager)(mc.getStoreManager())).getDatabaseAdapter();

        AbstractPersistenceManagerFactory apmf = mc.getAbstractPersistenceManagerFactory();

        this.tds = apmf.getTransactionalDataSource();

        this.useUpdateLockProperty = apmf.getUseUpdateLock();

        setRetainValues(apmf.getRetainValues());
        setOptimistic(apmf.getOptimistic());
        setNontransactionalRead(apmf.getNontransactionalRead());
        setNontransactionalWrite(apmf.getNontransactionalWrite());
        setTransactionIsolation(apmf.getTransactionIsolation());
        try
        {
            conn = dba.getConnection(tds,  credential == null ? null : credential.getUserName(),  credential == null ? null : new String(credential.getPassword()), isolationLevel);
        }
        catch (SQLException sqle)
        {
            throw new ResourceException("Couldn't get connection: " + sqle);
        }
    }

    /**
     * Accessor for the persistence manager
     * @return The Persistence Manager.
     */
    public PersistenceManager getPersistenceManager()
    {
        return mc.getPMHandle();
    }

    /**
     * Method to get the connection to the datastore.
     * @param forWriting Whether the connection is for writing
     * @return The Connection
     * @throws SQLException Thrown if an error occurs
     */
    public synchronized Connection getConnection(boolean forWriting)
    throws SQLException
    {
        assertJDOConnectionNotInUse();
        
        if (active)
        {
            /*
             * Before performing any (more) SQL operations in this transaction,
             * flush out any updates pending in the persistence manager.
             */
            mc.flush();
        }

        return conn;
    }

    /**
     * Accessor for the connection to the datasource.
     * @param forWriting Whether the connection is read-only or read-write.
     * @param useOpenedConnection whether to use an already opened connection, if one is opened. 
     * @return The Connection
     * @throws SQLException
     */    
    public Connection getConnection(boolean forWriting,boolean useOpenedConnection) throws SQLException
    {
        assertJDOConnectionNotInUse();

        return conn;
    }
    
    /**
     * Checks whether a connection to the datasource is open.
     * @return true if the Connection is open
     * @throws SQLException
     */
    public boolean isConnectionOpen()
    {
        return conn != null;
    }
    
    /**
     * Method to release the connection to the datastore.
     * @param conn The connection
     * @throws SQLException Thrown if an error occurs
     */
    public synchronized void releaseConnection(Connection conn)
    throws SQLException
    {
        /*
        if (conn == this.conn)
        {
            if (!active)
                closeConnection();
        }
        */
    }

    public boolean useUpdateLockOnFetch()
    {
        return active ? useUpdateLock : false;
    }

    /**
     * Internal method to close the connection to the datastore.
     * @throws SQLException Thrown if an error occurs
     */
    void closeConnection()
    throws SQLException
    {
        try
        {
            dba.closeConnection(conn);
        }
        finally
        {
            conn = null;
        }
    }

    /**
     * Asserts that the transaction is not in use.
     **/
    private void assertNotInUse()
    {
        if (active)
        {
            throw new org.jpox.exceptions.TransactionActiveException();
        }
    }

    /**
     * Method to start the Transaction
     **/
    public synchronized void begin()
    {
        assertNotInUse();

        active = true;

        if (JPOXLogger.TRANSACTION.isDebugEnabled())
        {
            JPOXLogger.TRANSACTION.debug("Transaction begun, connection = " + conn);
        }
    }

    /**
     * Method to commit the Transaction
     **/
    public synchronized void commit()
    {
        if (!active)
        {
            throw new org.jpox.exceptions.TransactionActiveException();
        }

        assertJDOConnectionNotInUse();
        
        // JDO 2.0 section 13.4.5 rollbackOnly functionality
        // It isn't clear from the spec if we are expected to do the rollback
        // here. The spec simply says that we throw an exception. This is
        // assumed as meaning that the users code will catch the exception and
        // call rollback themselves. i.e we don't need to close the DB
        // connection or set "active" to false. 
        if (rollbackOnly)
        {
            // Throw an exception since can only exit via rollback
            if (JPOXLogger.TRANSACTION.isDebugEnabled())
            {
                JPOXLogger.TRANSACTION.debug("Attempt to commit transaction with rollbackOnly enabled. Throwing exception.");
            }

            throw new JDOFatalDataStoreException("Transaction has rollbackOnly enabled. Commit cannot be performed in this state.");
        }

        try
        {
            boolean success = false;

            try
            {
                try
                {
                    if (JPOXLogger.TRANSACTION.isDebugEnabled())
                    {
                        JPOXLogger.TRANSACTION.debug("Transaction committing, connection = " + conn);
                    }

                    mc.flush();

                    if (sync != null)
                    {
                        sync.beforeCompletion();
                    }

                    conn.commit();

                    success = true;
                }
                finally
                {
                    if (!success)
                    {
                        try
                        {
                            mc.preRollback();
                        }
                        finally
                        {
                            conn.rollback();
                        }
                    }
                }
            }
            finally
            {
                try
                {
                    active = false;
                    //closeConnection();
                }
                finally
                {
                    try
                    {
                        if (success)
                        {
                            mc.postCommit();
                        }
                    }
                    finally
                    {
                        if (sync != null)
                        {
                            if (success)
                            {
                                sync.afterCompletion(Status.STATUS_COMMITTED);
                            }
                            else
                            {
                                sync.afterCompletion(Status.STATUS_ROLLEDBACK);
                            }
                        }
                    }
                }
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException("Failed committing transaction", new Exception[] { e });
        }
    }

    /**
     * Method to rollback the Transaction
     **/
    public synchronized void rollback()
    {
        if (!active)
        {
            throw new org.jpox.exceptions.TransactionNotActiveException();
        }

        assertJDOConnectionNotInUse();

        try
        {
            try
            {
                try
                {
                    if (JPOXLogger.TRANSACTION.isDebugEnabled())
                    {
                        JPOXLogger.TRANSACTION.debug("Transaction rolling back, connection = " + conn);
                    }

                    mc.preRollback();
                }
                finally
                {
                    conn.rollback();
                }
            }
            finally
            {
                try
                {
                    active = false;
                    //closeConnection();
                }
                finally
                {
                    if (sync != null)
                    {
                        sync.afterCompletion(Status.STATUS_ROLLEDBACK);
                    }
                }
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException("Failed rolling back transaction", new Exception[] { e });
        }

        // Reset rollbackOnly flag
        rollbackOnly = false;
    }

    /**
     * Accessor for whether the connection is active.
     **/
    public boolean isActive()
    {
        return active;
    }

    /**
     * Accessor for the "rollback only" flag.
     * @return The rollback only flag
     * @since 1.1
     */
    public boolean getRollbackOnly()
    {
        return rollbackOnly;
    }

    /**
     * Mutator for the "rollback only" flag. Sets the transaction as for
     * rollback only.
     * @since 1.1
     */
    public void setRollbackOnly()
    {
        // Only apply to active transactions
        if (active)
        {
            rollbackOnly = true;
        }
    }

    public synchronized void setNontransactionalRead(boolean nontransactionalRead)
    {
        //assertInUse();
        this.nontransactionalRead = nontransactionalRead;
    }

    public boolean getNontransactionalRead()
    {
        return nontransactionalRead;
    }

    public synchronized void setNontransactionalWrite(boolean nontransactionalWrite)
    {
        assertNotInUse();

        if (nontransactionalWrite)
        {
            throw new JDOUnsupportedOptionException("Non-transactional write mode not (yet) supported");
        }

        this.nontransactionalWrite = nontransactionalWrite;
    }

    public boolean getNontransactionalWrite()
    {
        return nontransactionalWrite;
    }

    public synchronized void setRetainValues(boolean retainValues)
    {
        assertNotInUse();

        this.retainValues = retainValues;

        if (retainValues)
        {
            nontransactionalRead = true;
        }
    }

    public boolean getRetainValues()
    {
        return retainValues;
    }

    /**
     * Set the default RestoreValues setting for all
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @param flag the RestoreValues setting.
     */
    public synchronized void setRestoreValues(boolean flag)
    {
        this.restoreValues = flag;
    }

    /**
     * Get the default RestoreValues setting for all 
     * <tt>PersistenceManager</tt> instances obtained from this factory.
     * @return the RestoreValues setting.
     */
    public boolean getRestoreValues()
    {
        return this.restoreValues;
    }

    /**
     * Mutator for the optimistic transaction setting.
     * @param optimistic The optimistic transaction setting.
     **/
    public synchronized void setOptimistic(boolean optimistic)
    {
        assertNotInUse();

        this.optimistic = optimistic;
    }

    /**
     * Accessor for the Optimistic setting
     * @return Whether optimistic transactions are in operation.
     **/
    public boolean getOptimistic()
    {
        return optimistic;
    }

    /**
     * Mutator for the transaction isolation level.
     * @param isolationLevel The transaction isolation level.
     **/
    public synchronized void setTransactionIsolation(int isolationLevel)
    {
        assertNotInUse();

        switch (isolationLevel)
        {
            case Connection.TRANSACTION_READ_UNCOMMITTED:
            case Connection.TRANSACTION_READ_COMMITTED:
                useUpdateLock = useUpdateLockProperty;
                break;

            case Connection.TRANSACTION_REPEATABLE_READ:
            case Connection.TRANSACTION_SERIALIZABLE:
                useUpdateLock = false;
                break;

            default:
                throw new IllegalArgumentException("Illegal isolation level: " + isolationLevel);
        }

        this.isolationLevel = isolationLevel;
    }

    /**
     * Accessor for the Transaction Isolation level.
     * @return The transaction isolation level.
     **/
    public int getTransactionIsolation()
    {
        return isolationLevel;
    }

    /**
     * Mutator for the synchronisation
     * @param sync The Synchronisation
     **/
    public synchronized void setSynchronization(Synchronization sync)
    {
        assertNotInUse();

        this.sync = sync;
    }

    /**
     * Accessor for the Synchronisation.
     * @return The synchronisation.
     **/
    public Synchronization getSynchronization()
    {
        return sync;
    }

    /**
     * Accessor for the connection to the datasource.
     * If this method is called while a datastore transaction is active,
     * the object returned will be enlisted in the current transaction.
     * If called in an optimistic transaction or outside an active
     * transaction, the object returned will not be enlisted in any transaction.
     * 
     * The object must be returned to the JDO implementation prior to calling
     * any JDO method or performing any action on any persistent instance that
     * might require the JDO implementation to use a connection. If the object
     * has not been returned and the JDO implementation needs a connection,
     * a JDOUserException is thrown. The object is returned to the JDO
     * implementation by calling the standard method on the object.
     * 
     * @return The Connection
     * @throws SQLException
     */
    public synchronized JDOConnection getJDOConnection()
    {
        assertJDOConnectionNotInUse();
        
        isJDOConnectionInUse = true;
        return new JDOConnectionImpl(conn, new Runnable()
        {
            public void run()
            {
                isJDOConnectionInUse = false;
            }
        });
    }
    
    /**
     * Asserts that's the connection is not hold by the developer's application
     * @throws JDOUserException if hold by developer
     */
    private void assertJDOConnectionNotInUse()
    {
        if (isJDOConnectionInUse)
        {
            throw new JDOUserException("The JDOConnection was acquired by the developer and must be closed before using the JDO API.");
        }
    }    
}
