/**********************************************************************
Copyright (c) 2003 David Jencks and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Erik Bengtson - added getObjectbyAID
2004 Andy Jefferson - coding standards and javadocs
    ...
**********************************************************************/
package org.jpox.resource;

import java.io.PrintWriter;
import java.sql.SQLException;
import java.util.Collection;

import javax.jdo.Extent;
import javax.jdo.JDOException;
import javax.jdo.spi.PersistenceCapable;
import javax.jdo.FetchPlan;
import javax.jdo.JDOConnection;
import javax.jdo.InstanceLifecycleListener;
import javax.jdo.PersistenceManagerFactory;
import javax.jdo.Query;
import javax.jdo.Sequence;
import javax.jdo.Transaction;
import javax.resource.ResourceException;
import javax.resource.cci.Connection;
import javax.resource.cci.ConnectionMetaData;
import javax.resource.cci.Interaction;
import javax.resource.cci.LocalTransaction;
import javax.resource.cci.ResultSetInfo;

import org.jpox.ClassLoaderResolver;
import org.jpox.PMFContext;
import org.jpox.PersistenceManager;
import org.jpox.sco.SCO;
import org.jpox.StateManager;
import org.jpox.metadata.FieldMetaData;
import org.jpox.metadata.MetaDataManager;
import org.jpox.store.FieldValues;
import org.jpox.store.StoreManager;

/**
 * Implementation of the PersistenceManager for use in JCA environments.
 *
 * @version $Revision: 1.11 $
 */
public class PersistenceManagerImpl
    implements Connection, PersistenceManager
{
    private final JdoTransactionHandle tx;

    ManagedConnectionImpl mc;

    private PersistenceManagerFactory pmf;

    private boolean closed;

    /**
     * Method to dump to output an object for diagnostic purposes. 
     * @param obj The object to dump
     * @param out The PrintWriter to use 
     * @see org.jpox.PersistenceManager#dump(java.lang.Object, java.io.PrintWriter)
     */
    public void dump(Object obj, PrintWriter out)
    {
        mc.dump(obj,out);
    }

    /**
     * Accessor for an object given the object id.
     * @param id Id of the object.
     * @param fv Field values for the object
     * @param clazz Class name of the object
     * @param ignoreCache true if it must ignore the cache
     * @return The Object with this id
     */
    public Object getObjectById(Object id, FieldValues fv, Class clazz, boolean ignoreCache)
    {
        return mc.getObjectById(id, fv, clazz, ignoreCache);
    }

    /**
     * Accessor for the StateManager of an object given the object AID.
     * @param pcClass <description>
     * @param fv <description>
     * @param ignoreCache true if it must ignore the cache*
     * @param checkInheritance Whether look to the database to determine which
     * class this object is. This parameter is a hint. Set false, if it's
     * already determined the correct pcClass for this pc "object" in a certain
     * level in the hierarchy. Set to true and it will look to the database.
     * @return Object with this AID
     */
    public synchronized Object getObjectByAID(Class pcClass, FieldValues fv,boolean ignoreCache, boolean checkInheritance)
    {
        return mc.getObjectByAID(pcClass,fv,ignoreCache,checkInheritance);
    }
 
    /**
     * Constructor.
     * @param mcf The connection factory.
     * @param mc The connection implementation.
     **/
    public PersistenceManagerImpl(final ManagedConnectionFactoryImpl mcf, final ManagedConnectionImpl mc) 
    {
        setManagedConnection(mc);
        tx = new JdoTransactionHandle(this);
    }

    /**
     * Mutator for the connection implementation.
     * @param mc The managed connection implementation
     **/
    void setManagedConnection(final ManagedConnectionImpl mc)
    {
        this.mc = mc;
    }

    /**
     * Mutator for the persistence manager factory.
     * @param pmf The persistence manager factory
     **/
    void setPersistenceManagerFactory(final PersistenceManagerFactory pmf)
    {
        this.pmf = pmf;
    }

    // implementation of javax.resource.cci.Connection interface

    /**
     * Accessor for the local transaction.
     * @return The local transaction
     * @exception javax.resource.ResourceException Thrown when an error occurs
     **/
    public LocalTransaction getLocalTransaction()
    throws ResourceException
    {
        return tx;
    }

    /**
     * Accessor for Meta-Data.
     * @return The Connection Meta-Data.
     * @exception javax.resource.ResourceException <description>
     **/
    public ConnectionMetaData getMetaData() throws ResourceException
    {
        throw new ResourceException("Not Yet Implemented");
    }

    /**
     *
     * @return <description>
     * @exception javax.resource.ResourceException <description>
     **/
    public Interaction createInteraction() throws ResourceException
    {
        throw new ResourceException("Not Yet Implemented");
    }

    /**
     *
     * @return <description>
     * @exception javax.resource.ResourceException <description>
     **/
    public ResultSetInfo getResultSetInfo() throws ResourceException
    {
        throw new ResourceException("Not Yet Implemented");
    }

    // implementation of javax.jdo.PersistenceManager interface

    /**
     * Method to close the Persistence Manager.
     **/
    public void close()
    {
        if (mc != null) 
        {
            mc.flush();
            mc.notifyClosed(this);
        }
        closed = true;
 
        setManagedConnection(null);
    }

    /**
     * Accessor for whether the Persistence manager is closed.
     * @return Whether the Persistence Manager is closed.
     **/
    public boolean isClosed()
    {
        return closed;
    }

	/**
	 * Whether the datastore operations are delayed until commit.
     * In optimistic transactions this is automatically enabled.
	 * @return true if datastore operations are delayed until commit
     */
    public boolean isDelayDatastoreOperationsEnabled()
    {
        return mc.isDelayDatastoreOperationsEnabled();
    }
    
    /**
     * Method to refresh an object 
     * @param o Object to refresh
     **/
    public void refresh(Object o)
    {
        checkStatus();
        mc.refresh(o);
    }

    /**
     * Method to retrieve an object 
     * @param o Object to retrieve
     **/
    public void retrieve(Object o)
    {
        checkStatus();
        mc.retrieve(o);
    }

    /**
     * Accessor for the current transaction 
     * @return The current transaction
     **/
    public Transaction currentTransaction()
    {
        return tx;
    }

    /**
     * Method to evict an object 
     * @param o Object to evict
     **/
    public void evict(Object o)
    {
        checkStatus();
        mc.evict(o);
    }

    /**
     * Method to evict an array of objects 
     * @param os Array of objects to evict
     **/
    public void evictAll(Object[] os)
    {
        checkStatus();
        mc.evictAll(os);
    }

    /**
     * Method to evict all of the specified objects from the PM.
     * @param os Collection of objects to evict
     **/
    public void evictAll(Collection os)
    {
        checkStatus();
        mc.evictAll(os);
    }

    /**
     * Method to evict all objects from the PM.
     **/
    public void evictAll()
    {
       checkStatus();
       mc.evictAll();
    }

    /**
     * Method to refresh an array of objects.
     * @param os Array of objects to refresh
     **/
    public void refreshAll(Object[] os)
    {
        checkStatus();
        mc.refreshAll(os);
    }

    /**
     * Method to refresh a collection of objects.
     * @param os Collection of objects to refresh
     */
    public void refreshAll(Collection os)
    {
        checkStatus();
        mc.refreshAll(os);
    }

    /**
     * Method to refresh objects that failed verification in a JDOException
     * @param exc Exception containing objects that failed verification
     */
    public void refreshAll(JDOException exc)
    {
        checkStatus();
        mc.refreshAll(exc);
    }

    /**
     * Method to refresh all objects in the Persistence Manager.
     */
    public void refreshAll()
    {
        checkStatus();
        mc.refreshAll();
    }

    /**
     * Accessor for a new query.
     * @return The query.
     */
    public Query newQuery()
    {
        checkStatus();
        return mc.newQuery();
    }

    /**
     * Accessor for a new query.
     * @param obj object
     * @return The new query
     */
    public Query newQuery(Object obj)
    {
        checkStatus();
        return mc.newQuery(obj);
    }

    /**
     * Accessor for a single-string query.
     * @param query The single-string query
     * @return The Query
     */
    public Query newQuery(String query)
    {
        checkStatus();
        return mc.newQuery(query);
    }

    /**
     * Accessor for a new query in the specified query language etc.
     * @param language The query language
     * @param query The query definition
     * @return The new query
     */
    public Query newQuery(String language, Object query)
    {
        checkStatus();
        return mc.newQuery(language, query);
    }

    /**
     * Accessor for a new query using this candidate class.
     * @param cls The candidate class to use.
     * @return The new query
     */
    public Query newQuery(Class cls)
    {
        checkStatus();
        return mc.newQuery(cls);
    }

    /**
     * Accessor for a new query based on the provided extent.
     * @param ext The extent to use
     * @return The new query
     */
    public Query newQuery(Extent ext)
    {
        checkStatus();
        return mc.newQuery(ext);
    }

    /**
     * Accessor for a new query.
     * @param cls The candidate class.
     * @param cln The collection to use.
     * @return The new query
     */
    public Query newQuery(Class cls, Collection cln)
    {
        checkStatus();
        return mc.newQuery(cls, cln);
    }

    /**
     * Accessor for a new query.
     * @param cls The candidate class
     * @param filter The filter to use
     * @return The new query
     */
    public Query newQuery(Class cls, String filter)
    {
        checkStatus();
        return mc.newQuery(cls, filter);
    }

    /**
     * Accessor for a new query.
     * @param cls The candidate class
     * @param cln The collection
     * @param filter The filter to use
     * @return The new query
     */
    public Query newQuery(Class cls, Collection cln, String filter)
    {
        checkStatus();
        return mc.newQuery(cls, cln, filter);
    }

    /**
     * Accessor for a new query.
     * @param cln The collection
     * @param filter The filter to use
     * @return The new query
     */
    public Query newQuery(Extent cln, String filter)
    {
        checkStatus();
        return mc.newQuery(cln, filter);
    }

    /**
     * Accessor for a new named query.
     * @param cls The candidate class
     * @param queryName The name of the query
     * @return The new query
     */
    public Query newNamedQuery(Class cls, String queryName)
    {
        checkStatus();
        return mc.newNamedQuery(cls, queryName);
    }

    /**
     * Accessor for the extent of a candidate class.
     * @param cls The candidate class
     * @param subclasses Whether to include subclasses
     * @return The extent for these classes
     */
    public Extent getExtent(Class cls, boolean subclasses)
    {
        checkStatus();
        return mc.getExtent(cls, subclasses);
    }

    /**
     * Accessor for the extent of a candidate class.
     * @param cls The candidate class
     * @return The extent for these classes
     * @since 1.1
     */
    public Extent getExtent(Class cls)
    {
        checkStatus();
        return mc.getExtent(cls);
    }
    
    /**
     * Acessor for the current FetchPlan
     * @return FetchPlan
     * @since 1.1
     */
    public FetchPlan getFetchPlan()
    {
        checkStatus();
        return mc.getFetchPlan();
    }

    /**
     * Accessor for an object given its id.
     * @param id Id of the object.
     * @return The object with this id
     */
    public Object getObjectById(Object id)
    {
        return getObjectById(id, true);
    }

    /**
     * Accessor for an object given its id.
     * @param id Id of the object.
     * @param validate Whether to validate the object before returning it
     * @return The object with this id
     */
    public Object getObjectById(Object id, boolean validate)
    {
        checkStatus();
        return mc.getObjectById(id, validate);
    }

    public Collection getObjectsById (Collection oids)
    {
        checkStatus();
        return mc.getObjectsById(oids);
    }

    public Object[] getObjectsById (Object[] oids)
    {
        checkStatus();
        return mc.getObjectsById(oids);
    }

    public Collection getObjectsById (Collection oids, boolean validate)
    {
        checkStatus();
        return mc.getObjectsById(oids, validate);
    }

    public Object[] getObjectsById (Object[] oids, boolean validate)
    {
        checkStatus();
        return mc.getObjectsById(oids, validate);
    }

    public Object getObjectById (Class cls, Object key)
    {
        checkStatus();
        return mc.getObjectById(cls, key);
    }

    public Object newObjectIdInstance (Class pcClass, Object key)
    {
        checkStatus();
        return mc.newObjectIdInstance(pcClass, key);
    }

    /**
     * Accessor for an object given the object id.
     * @param id Id of the object.
     * @param validate Whether to validate the object state
     * @param checkInheritance Whether look to the database to determine which
     * class this object is. This parameter is a hint. Set false, if it's
     * already determined the correct pcClass for this pc "object" in a certain
     * level in the hierarchy. Set to true and it will look to the database.
     * @return The Object
     */
    public Object getObjectById(Object id, boolean validate, boolean checkInheritance)
    {
        checkStatus();
        return mc.getObjectById(id, validate, checkInheritance);
    }
    
    /**
     * Accessor for the id of an object.
     * @param pc The object
     * @return The objects id
     */
    public Object getObjectId(Object pc)
    {
        checkStatus();
        return mc.getObjectId(pc);
    }

    /**
     *
     * @param pc <description>
     * @return <description>
     */
    public Object getTransactionalObjectId(Object pc)
    {
        checkStatus();
        return mc.getTransactionalObjectId(pc);
    }

    /**
     *
     * @param clazz <description>
     * @param str <description>
     * @return <description>
     */
    public Object newObjectIdInstance(Class clazz, String str)
    {
        checkStatus();
        return mc.newObjectIdInstance(clazz, str);
    }

    /**
     * Method to create an instance of an interface or abstract class
     * @param persistenceCapable interface/abstract class declared in metadata
     * @return Instance of the interface / abstract class
     */
    public Object newInstance(Class persistenceCapable)
    {
        checkStatus();
        return mc.newInstance(persistenceCapable);
    }

    /**
     * Method to make an object persistent.
     * @param pc The object to persist
     */
    public void makePersistent(Object pc)
    {
        checkStatus();
        mc.makePersistent(pc);
    }

    public synchronized void makePersistent(Object obj, FieldValues preInsertChanges) 
    {
      checkStatus();
      mc.makePersistent(obj,preInsertChanges);
    } 
    
    /**
     * Make an array of Persistent Capable objects persistent 
     * @param pcs Array of Persistent Capable objects
     */
    public void makePersistentAll(Object[] pcs)
    {
        checkStatus();
        mc.makePersistentAll(pcs);
    }

    /**
     * Make a Collection of Persistence Capable objects persistent 
     * @param pcs Collection of Persistence Capable objects
     */
    public void makePersistentAll(Collection pcs)
    {
        checkStatus();
        mc.makePersistentAll(pcs);
    }

    /**
     * Delete a Persistence Capable object 
     * @param pc Persistence Capable object
     */
    public void deletePersistent(Object pc)
    {
        checkStatus();
        mc.deletePersistent(pc);
    }

    /**
     * Delete an array of Persistence Capable objects 
     * @param pcs Array of Persistence Capable objects
     */
    public void deletePersistentAll(Object[] pcs)
    {
        checkStatus();
        mc.deletePersistentAll(pcs);
    }

    /**
     * Delete a Collection of Persistence Capable objects 
     * @param pcs Collection of Persistence Capable objects
     */
    public void deletePersistentAll(Collection pcs)
    {
        checkStatus();
        mc.deletePersistentAll(pcs);
    }

    /**
     * Make a Persistence-Capable object transient 
     * @param pc Persistence-Capable object
     */
    public void makeTransient(Object pc)
    {
        checkStatus();
        mc.makeTransient(pc);
    }

    /**
     * Make an array of Persistence-Capable objects transient 
     * @param pcs Array of Persistence-Capable objects
     */
    public void makeTransientAll(Object[] pcs)
    {
        checkStatus();
        mc.makeTransientAll(pcs);
    }

    /**
     * Make a Collection of Persistence-Capable objects transient 
     * @param pcs Collection of Persistence-Capable objects
     */
    public void makeTransientAll(Collection pcs)
    {
        checkStatus();
        mc.makeTransientAll(pcs);
    }

    /**
     * Make a Persistence-Capable object transient 
     * @param pc Persistence-Capable object
     */
    public void makeTransactional(Object pc)
    {
        checkStatus();
        mc.makeTransactional(pc);
    }

    /**
     * Make an array of Persistence-Capable objects transactional 
     * @param pcs Array of Persistence-Capable objects
     */
    public void makeTransactionalAll(Object[] pcs)
    {
        checkStatus();
        mc.makeTransactionalAll(pcs);
    }

    /**
     * Make a collection of PersistenceCapable objects transactional
     * @param pcs Collection of Persistence-Capable objects
     */
    public void makeTransactionalAll(Collection pcs)
    {
        checkStatus();
        mc.makeTransactionalAll(pcs);
    }

    /**
     * Make a Persistence-Capable object non-transactional 
     * @param pc Persistence-Capable object
     */
    public void makeNontransactional(Object pc)
    {
        checkStatus();
        mc.makeNontransactional(pc);
    }

    /**
     * Make an array of Persistence-Capable objects non-transactional 
     * @param pcs Array of Persistence-Capable objects
     */
    public void makeNontransactionalAll(Object[] pcs)
    {
        checkStatus();
        mc.makeNontransactionalAll(pcs);
    }

    /**
     * Make a Collection of Persistence-Capable objects non-transactional 
     * @param pcs Collection of Persistence-Capable objects
     */
    public void makeNontransactionalAll(Collection pcs)
    {
        checkStatus();
        mc.makeNontransactionalAll(pcs);
    }

    /**
     * Method to detach a persistent object.
     * @param pc The object
     **/ 
    public synchronized Object detachCopy(Object pc)
    {
        checkStatus();
        return mc.detachCopy(pc);
    }

    /**
     * Method to detach an array of objects.
     * @param pcs The objects
     **/ 
    public synchronized Object[] detachCopyAll(Object[] pcs)
    {
        checkStatus();
        return mc.detachCopyAll(pcs);
    }

    /**
     * Method to detach a collection of objects.
     * @param pcs The objects
     **/ 
    public synchronized Collection detachCopyAll(Collection pcs)
    {
        checkStatus();
        return mc.detachCopyAll(pcs);
    }

    /**
     * Method to attach a persistent object.
     * @param pc The object
     **/ 
    public synchronized Object attachCopy(Object pc,boolean makeTransactional)
    {
        checkStatus();
        return mc.attachCopy(pc,makeTransactional);
    }

    /**
     * Method to attach an array of objects.
     * @param pcs The objects
     **/ 
    public synchronized Object[] attachCopyAll(Object[] pcs,boolean makeTransactional)
    {
        checkStatus();
        return mc.attachCopyAll(pcs,makeTransactional);
    }

    /**
     * Method to attach a collection of objects.
     * @param pcs The objects
     **/ 
    public synchronized Collection attachCopyAll(Collection pcs,boolean makeTransactional)
    {
        checkStatus();
        return mc.attachCopyAll(pcs,makeTransactional);
    }

    /**
     * Retrieve a Collection of Persistence-Capable objects 
     * @param pcs Collection of Persistence-Capable objects
     */
    public void retrieveAll(Collection pcs)
    {
        checkStatus();
        mc.retrieveAll(pcs);
    }

    /**
     * Retrieve an array of Persistence-Capable objects 
     * @param pcs Array of Persistence-Capable objects
     */
    public void retrieveAll(Object[] pcs)
    {
        checkStatus();
        mc.retrieveAll(pcs);
    }

    /**
     * Retrieve field values of instances from the store. This tells
     * the <code>PersistenceManager</code> that the application intends to use
     * the instances, and their field values should be retrieved. The fields
     * in the default fetch group must be retrieved, and the implementation
     * might retrieve more fields than the default-fetch-group.
     * <P>The <code>PersistenceManager</code> might use policy information
     * about the class to retrieve associated instances.
     * @param pcs the instances
     * @param dfgOnly whether to retrieve only the default-fetch-group fields
     * @since JDO 1.0.1
     */
    public void retrieveAll (Object[] pcs, boolean dfgOnly)
    {
        mc.retrieveAll(pcs,dfgOnly);
    }
 
    /** Retrieve field values of instances from the store.  This tells
     * the <code>PersistenceManager</code> that the application intends to use
     * the instances, and their field values should be retrieved. The fields
     * in the default fetch group must be retrieved, and the implementation
     * might retrieve more fields than the default-fetch-group.
     * <P>The <code>PersistenceManager</code> might use policy information
     * about the class to retrieve associated instances.
     * @param pcs the instances
     * @param dfgOnly whether to retrieve only the default-fetch-group fields
     * @since JDO 1.0.1
     */
    public void retrieveAll (Collection pcs, boolean dfgOnly)
    {
        mc.retrieveAll(pcs,dfgOnly);
    }

    /**
     * Method to put a user object into the PersistenceManager. This is so that
     * multiple users can each have a user object for example. <I>The parameter
     * is not inspected or used in any way by the JDO implementation. </I>
     * @param key The key to store the user object under
     * @param value The object to store
     * @return The previous value for this key
     * @since 1.1
     */
    public synchronized Object putUserObject(Object key, Object value)
    {
        checkStatus();
        return mc.putUserObject(key, value);
    }

    /**
     * Method to get a user object from the PersistenceManager. This is for user
     * objects which are stored under a key. <I>The parameter is not inspected
     * or used in any way by the JDO implementation. </I>
     * @param key The key to store the user object under
     * @return The user object for that key
     * @since 1.1
     */
    public synchronized Object getUserObject(Object key)
    {
        checkStatus();
        return mc.getUserObject(key);
    }

    /**
     * Method to remove a user object from the PersistenceManager. This is for
     * user objects which are stored under a key. <I>The parameter is not
     * inspected or used in any way by the JDO implementation. </I>
     * @param key The key whose uder object is to be removed.
     * @return The user object that was removed
     * @since 1.1
     */
    public synchronized Object removeUserObject(Object key)
    {
        checkStatus();
        return mc.removeUserObject(key);
    }

    /**
     * The application might manage PersistenceManager instances by using an
     * associated object for bookkeeping purposes. These methods allow the user
     * to manage the associated object.
     * The parameter is not inspected or used in any way by the JDO
     * implementation.
     *
     * @param obj User Object
     */
    public void setUserObject(Object obj)
    {
        checkStatus();
        mc.setUserObject(obj);
    }

    /**
     * The application might manage PersistenceManager instances by using an
     * associated object for bookkeeping purposes. These methods allow the user
     * to manage the associated object.
     * The parameter is not inspected or used in any way by the JDO
     * implementation.*
     * 
     * @return User object
     */
    public Object getUserObject()
    {
        checkStatus();
        return mc.getUserObject();
    }

    /**
     * Retrieve the PersistenceManagerFactory for this manager 
     * @return The PersistenceManagerFactory
     */
    public PersistenceManagerFactory getPersistenceManagerFactory()
    {
        return pmf;
    }

    /**
     * Retrieve the class for the objectid 
     * @param clazz The class to retrieve
     * @return The Class of the ObjectId
     */
    public Class getObjectIdClass(Class clazz)
    {
        checkStatus();
        return mc.getObjectIdClass(clazz);
    }

    /**
     * Mutator for the multithreaded capability of the manager 
     * @param multithreaded Whether to run multithreaded or not
     */
    public void setMultithreaded(boolean multithreaded)
    {
        checkStatus();
        mc.setMultithreaded(multithreaded);
    }

    /**
     * Accessor for the multithreaded capability of the manager 
     * @return Whether the PersistenceManager is multithreaded or not
     */
    public boolean getMultithreaded()
    {
        return false;
    }

    /**
     * Mutator for whether to ignore the cache or not 
     * @param ignore Whether to ignore the cache or not
     */
    public void setIgnoreCache(boolean ignore)
    {
        checkStatus();
        mc.setIgnoreCache(ignore);
    }

    /**
     * Accessor for whether to ignore the cache or not 
     * @return Whether to ignore the cache or not
     */
    public boolean getIgnoreCache()
    {
        return false;
    }

    /**
     * Accessor for the Store Manager.
     * @return The Store Manager.
     **/ 
    public StoreManager getStoreManager()
    {
        checkStatus();
        return mc.getStoreManager();
    }

    /**
     * Accessor for the MetaData Manager.
     * @return The MetaData Manager.
     **/ 
    public MetaDataManager getMetaDataManager()
    {
        checkStatus();
        return mc.getMetaDataManager();
    }

    /**
     * Accessor for a connection to the datastore.
     * @param forWriting Whether it is for writing to
     * @return The Connection.
     * @throws SQLException Thrown when an error occurs in obtaining it.
     **/
    public java.sql.Connection getConnection(boolean forWriting)
    throws SQLException
    {
        checkStatus();
        return mc.getConnection(forWriting);
    }

    /**
     * Accessor for the connection to the datasource.
     * @param forWriting Whether the connection is read-only or read-write.
     * @param useOpenedConnection whether to use an already opened connection, if one is opened. 
     * @return The Connection
     * @throws SQLException
     */
    public java.sql.Connection getConnection(boolean forWriting, boolean useOpenedConnection)
    throws SQLException
    {
        checkStatus();
        return mc.getConnection(forWriting, useOpenedConnection);
    }

    /**
     * Checks whether a connection to the datasource is open.
     * @return true if the Connection is open
     * @throws SQLException
     */
    public boolean isConnectionOpen()
    {
        checkStatus();
        return mc.isConnectionOpen();
    }
    
    /**
     * Method to release a connection.
     * @param conn The Connection to release.
     * @throws SQLException Thrown when an error occurs
     **/
    public void releaseConnection(java.sql.Connection conn) 
    throws SQLException
    {
        checkStatus();
        mc.releaseConnection(conn);
    }

    public void enlistInTransaction(StateManager sm)
    {
        checkStatus();
        mc.enlistInTransaction(sm);
    }

    public void evictFromTransaction(StateManager sm)
    {
        checkStatus();
        mc.evictFromTransaction(sm);
    }

    public void removeStateManager(StateManager sm)
    {
        checkStatus();
        mc.removeStateManager(sm);
    }

    public Object getObjectById(Object id, FieldValues fv)
    {
        checkStatus();
        return mc.getObjectById(id, fv);
    }

    public SCO newSCOInstance(Class type, Object owner, FieldMetaData fmd)
    {
        checkStatus();
        return mc.newSCOInstance(type, owner, fmd);
    }

    public StateManager findStateManager(PersistenceCapable pc)
    {
        checkStatus();
        return mc.findStateManager(pc);
    }

    public void hereIsStateManager(StateManager sm, PersistenceCapable pc)
    {
        checkStatus();
        mc.hereIsStateManager(sm, pc);
    }

    public void markDirty(StateManager sm)
    {
        checkStatus();
        mc.markDirty(sm);
    }

    /**
     * This method flushes all dirty, new, and deleted instances to the datastore. It has no effect
     * if a transaction is not active.
     * If a datastore transaction is active, this method synchronizes the cache with the datastore
     * and reports any exceptions.
     * If an optimistic transaction is active, this method obtains a datastore connection and synchronizes
     * the cache with the datastore using this connection. The connection obtained by
     * this method is held until the end of the transaction.
     */
    public void flush()
    {
        checkStatus();
        mc.flush();
    }

    /**
     * Method to check the consistency of the cache.
     */
    public void checkConsistency()
    {
        checkStatus();
        mc.checkConsistency();
    }

    /**
     * Tests whether this object is being inserted.
     * @param pc the object to verify the status
     * @return true if this instance is inserting.
     */
    public boolean isInserting(PersistenceCapable pc)
    {
        checkStatus();
        return mc.isInserting(pc);
    }
    
    public void clearDirty(StateManager sm)
    {
        checkStatus();
        mc.clearDirty(sm);
    }    

    /**
     * Method to check the status.
     **/
    private void checkStatus()
    {
        if (closed || mc == null) 
        {
            throw new JDOException("Invalid state, closed or no mc");
        }
    }

    /**
     * Accessor for the ClassLoaderResolver
     * @return the ClassLoaderResolver
     */
    public ClassLoaderResolver getClassLoaderResolver()
    {
        return mc.getClassLoaderResolver();
    }

    /**
     * Accessor for a handle to the Persistence Manager
     * @return The PM handle.
     **/
    public PersistenceManager getPMHandle()
    {
        return mc.getPMHandle();
    }

    public void addStateManager(StateManager sm)
    {
        mc.addStateManager(sm);
    }

    public StateManager getStateManagerById(Object myID)
    {
        return mc.getStateManagerById(myID);
    }

    /**
     * Gets the context which this PersistenceManager is running
     * @return Returns the pmfContext.
     */
    public PMFContext getPMFContext()
    {
        return ((PersistenceManagerFactoryImpl)pmf).getAbstractPersistenceManagerFactory().getPMFContext();
    }

    /**
     * (non-Javadoc)
     * @see javax.jdo.PersistenceManager#getDataStoreConnection()
     * @since 1.1
     */
    public JDOConnection getDataStoreConnection()
    {
        checkStatus();
        return mc.getDataStoreConnection();
    }

    /**
     * Accessor for a Sequence.
     * @param sequenceName Name of the sequence
     * @return The sequence
     */
    public Sequence getSequence(String sequenceName)
    {
        checkStatus();
        return mc.getSequence(sequenceName);
    }

    /**
     * (non-Javadoc)
     * @see javax.jdo.PersistenceManager#addLifecycleListener(javax.jdo.InstanceLifecycleListener, java.lang.Class[])
     * @since 1.1
     */
    public void addLifecycleListener(InstanceLifecycleListener listener, Class[] classes)
    {
        checkStatus();
        mc.addLifecycleListener(listener, classes);
    }
    
    /**
     * (non-Javadoc)
     * @see javax.jdo.PersistenceManager#removeLifecycleListener(javax.jdo.InstanceLifecycleListener)
     * @since 1.1
     */
    public void removeLifecycleListener(InstanceLifecycleListener listener)
    {
        checkStatus();
        mc.removeLifecycleListener(listener);
    }
}