/**********************************************************************
Copyright (c) 2003 Mike Martin and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - rewritten to always have delegate present
2004 Andy Jefferson - rewritten to allow delegate caching
    ...
**********************************************************************/
package org.jpox.sco;

import java.io.ObjectStreamException;
import java.util.AbstractSet;
import java.util.HashSet;
import java.util.Iterator;

import javax.jdo.spi.PersistenceCapable;

import org.jpox.StateManager;
import org.jpox.sco.exceptions.IncompatibleElementTypeException;
import org.jpox.sco.exceptions.IncompatibleFieldTypeException;
import org.jpox.sco.exceptions.NullsNotAllowedException;
import org.jpox.sco.exceptions.QueryUnownedSCOException;
import org.jpox.store.DatastoreClass;
import org.jpox.store.StoreManager;
import org.jpox.store.mapping.CollectionMapping;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.query.Query;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.query.Queryable;
import org.jpox.store.scostore.SetStore;
import org.jpox.util.ClassUtils;

/** 
 * A mutable second-class Collection object, representing a field in a class,
 * and containing elements of a particular type.
 * Uses a "delegate" as a local store for the Collection. Uses a "backing store"
 * (SetStore) to represent the datastore. The "delegate" is updated with the
 * "backing store" information at necessary intervals. 
 *
 * <H3>Modes of Operation</H3>
 * The user can operate the collection in 2 modes.
 * The <B>cached</B> mode will use an internal cache of the elements (in the
 * "delegate") reading them at the first opportunity and then using the cache
 * thereafter.
 * The <B>non-cached</B> mode will just go direct to the "backing store" each
 * call.
 *
 * <H3>Mutators</H3>
 * When the backing store is present any updates are passed direct to the
 * datastore as well as to the "delegate". If the "backing store" isn't present
 * the changes are made to the "delegate" only.
 *
 * <H3>Accessors</H3>
 * When any accessor method is invoked, it typically checks whether the
 * collection has been loaded from its backing store and does this as
 * necessary. Some methods (<B>size()</B>, <B>contains()</B>) just check if
 * everything is loaded and use the delegate if possible, otherwise going
 * direct to the datastore.
 *
 * @version $Revision: 1.5 $ 
 */
public class Collection extends AbstractSet 
    implements SCO, Cloneable, Queryable, java.io.Serializable
{
    protected Object owner;
    protected StateManager ownerSM;
    protected String fieldName;
    protected int fieldNumber;
    protected Class elementType;
    protected boolean allowNulls;

    /**
     * The "backing store"
     */
    protected SetStore setStore;

    /**
     * The internal "delegate".
     */
    protected HashSet delegate;

    /**
     * Whether to use "delegate" caching.
     */
    protected boolean useCache=true;

    /**
     * Status flag whether the collection is loaded into the cache.
     */
    protected boolean isCacheLoaded=false;

    /**
     * Default Constructor.
     */
    public Collection()
    {
    }
    
    /**
     * Constructor. Called from CollectionMapping. 
     * @param ownerSM The State Manager for this collection.
     * @param fieldName Name of the field
     **/
    public Collection(StateManager ownerSM, String fieldName)
    {
        this(ownerSM, fieldName, null, false);
    }

    /**
     * Constructor. Called from AbstractPersistenceManager.newSCOInstance().
     * @param ownerSM The State Manager for this collection.
     * @param fieldName Name of the field
     * @param elementType Type of the contained elements
     * @param allowNulls Whether to allow nulls
     **/
    public Collection(StateManager ownerSM,
                      String fieldName,
                      Class elementType,
                      boolean allowNulls)
    {
        init2(ownerSM, fieldName, elementType,allowNulls);
    }

    /**
     * Constructor. Called via Map SCO for handling the entrySet.
     * @param ownerSM The State Manager for this collection.
     * @param fieldName Name of the field
     * @param allowNulls Whether to allow nulls
     * @param setStore The backing store
     **/
    protected Collection(StateManager ownerSM,
                         String fieldName,
                         boolean allowNulls,
                         SetStore setStore)
    {
        init(ownerSM, fieldName, setStore, allowNulls);
    }

    /**
     * Second Initialisation method, overrideable by Set.
     **/
    protected void init2(StateManager ownerSM,
                         String fieldName,
                         Class elementType,
                         boolean allowNulls)
    {
        Class c = ownerSM.getObject().getClass();
        StoreManager storeMgr = ownerSM.getStoreManager();
        DatastoreClass ownerTable = storeMgr.getDatastoreClass(c.getName());
        JavaTypeMapping m = ownerTable.getFieldMapping(fieldName);
        if (!(m instanceof CollectionMapping))
        {
            throw new IncompatibleFieldTypeException(ownerSM, fieldName, Collection.class.getName(), ownerSM.getMetaDataManager().getMetaDataForField(c, fieldName).getTypeName());
        }

        init(ownerSM,fieldName,((CollectionMapping)m).getSetStore(),allowNulls);

        if (elementType != null &&
            !this.elementType.isAssignableFrom(elementType))
        {
            throw new IncompatibleElementTypeException(ownerSM, fieldName, this.elementType.getName(), elementType.getName());
        }
    }

    /**
     * Initialisation method.
     * @param ownerSM The owner StateManager
     * @param fieldName The name of the field of the SCO.
     * @param setStore The backing store for this SCO.
     * @param allowNulls Whether to allow null elements.
     */
    protected void init(StateManager ownerSM,
                        String fieldName,
                        SetStore setStore,
                        boolean allowNulls)
    {
        this.owner = ownerSM.getObject();
        this.ownerSM = ownerSM;
        this.fieldName = fieldName;
        this.fieldNumber = ownerSM.getMetaDataManager().getMetaDataForClass(owner.getClass()).getFieldNumberAbsolute(fieldName);
        this.setStore = setStore;
        this.elementType = ownerSM.getPersistenceManager().getClassLoaderResolver().classForName(setStore.getElementType());
        this.allowNulls = allowNulls;

        // Set caching policy
        useCache = SCOUtils.useContainerCache(ownerSM, fieldName); 

        // Set up our "delegate"
        this.delegate = new HashSet();

        // If we aren't caching, just load up what we have now.
        if (!useCache)
        {
            Iterator iter=setStore.iterator(ownerSM);
            while (iter.hasNext())
            {
                delegate.add(iter.next());
            }
        }
    }

    // ----------------------- Implementation of SCO methods -------------------

    /**
     * Mutator for setting using an object.
     * @param o The object to set from
     */
    public void setValueFrom(Object o)
    {
        java.util.Collection c = (java.util.Collection)o;

        clear();
        addAll(c);
    }

    /**
     * Accessor for the field name.
     * @return The field name
     */
    public String getFieldName()
    {
        return fieldName;
    }

    /**
     * Accessor for the owner object.
     * @return The owner object
     */
    public Object getOwner()
    {
        return owner;
    }

    /**
     * Method to unset the owner and field information.
     */
    public synchronized void unsetOwner()
    {
        if (ownerSM != null)
        {
            owner = null;
            ownerSM = null;
            fieldName = null;
            setStore = null;
        }
    }

    /**
     * Utility to apply updates
     **/
    public void applyUpdates()
    {
        /*
         * We are never dirty, since all changes are immediately delegated to
         * the underlying SetStore.
         */
        return;
    }

    /**
     * Utility to mark the object as dirty
     **/
    public void makeDirty()
    {
        /*
         * Although we are never really "dirty", the owning object must be
         * marked dirty so that the proper state change occurs and its
         * jdoPreStore() gets called properly.
         */
        if (ownerSM != null)
        {
            ownerSM.makeDirty(fieldNumber);
        }
    }

    public Object detachCopy(StateManager sm, Object value)
    {
        java.util.Collection c = (java.util.Collection)value;

        if (value == null)
        {
        	// accept null collections...
			return new java.util.HashSet();
        }

        java.util.Collection detached = new java.util.HashSet();
        for( Iterator it = c.iterator(); it.hasNext(); )
        {
            Object object = it.next();
            if( object instanceof PersistenceCapable )
            {
                detached.add(sm.getPersistenceManager().detachCopy(object));
            }
            else
            {
                detached.add(object);
            }
        }        
        return detached;
    }
    
    public Object attachCopy(StateManager sm, Object value, boolean makeTransactional)
    {
        java.util.Collection c = (java.util.Collection)value;

        if (value == null)
        {
			return null;
        }

        java.util.Collection attached = new java.util.HashSet();
        for( Iterator it = c.iterator(); it.hasNext(); )
        {
            Object element = it.next();
            if (!ClassUtils.isPersistenceCapable(element))
            {
                attached.add(element);
            }
            else
            {
                attached.add(sm.getPersistenceManager().attachCopy(element,makeTransactional));
            }
        }        
        return attached;
    }    
    
    // ------------------------ Query Statement methods ------------------------

    /**
     * Method to generate a QueryStatement for the SCO.
     * @return The QueryStatement
     */
    public synchronized QueryStatement newQueryStatement()
    {
        return newQueryStatement(elementType);
    }

    /**
     * Method to return a QueryStatement, using the specified candidate class.
     * @param candidate_class the candidate class
     * @return The QueryStatement
     */
    public synchronized QueryStatement newQueryStatement(Class candidate_class)
    {
        if (setStore == null)
        {
            throw new QueryUnownedSCOException();
        }

        return setStore.newQueryStatement(ownerSM, candidate_class.getName());
    }

    /**
     * Method to return a ResultObjectFactory for the SCO.
     * @param stmt The QueryStatement
     * @param ignoreCache Whether to ignore the cache
     * @param resultClass Whether to create objects of a particular class
     * @return The ResultObjectFactory
     */
    public synchronized Query.ResultObjectFactory newResultObjectFactory(QueryStatement stmt,boolean ignoreCache, Class resultClass)
    {
        if (setStore == null)
        {
            throw new QueryUnownedSCOException();
        }

        return setStore.newResultObjectFactory(ownerSM, stmt,ignoreCache);
    }

    // ---------------- Implementation of Collection methods -------------------
 
    /**
     * Creates and returns a copy of this object.
     * <P>
     * Mutable second-class Objects are required to provide a public
     * clone method in order to allow for copying PersistenceCapable
     * objects. In contrast to Object.clone(), this method must not throw a
     * CloneNotSupportedException.
     * @return A clone of the object
     */
    public Object clone()
    {
        Object obj=null;
        try
        {
            obj = super.clone();
            ((Set)obj).unsetOwner();
        }
        catch (CloneNotSupportedException cnse)
        {
            // Can't happen
        }

        return obj;
    }

    /**
     * Accessor for whether an element is contained in the Collection.
     * @param element The element
     * @return Whether the element is contained here
     **/
    public synchronized boolean contains(Object element)
    {
        if (useCache && isCacheLoaded)
        {
            // If the "delegate" is already loaded, use it
            return delegate.contains(element);
        }
        else if (setStore != null)
        {
            return setStore.contains(ownerSM,element);
        }

        return delegate.contains(element);
    }

    /**
     * Accessor for whether a collection of elements are contained here.
     * @param c The collection of elements.
     * @return Whether they are contained.
     **/
    public synchronized boolean containsAll(java.util.Collection c)
    {
        if (useCache)
        {
            loadFromStore();
        }
        else if (setStore != null)
        {
            java.util.HashSet h=new java.util.HashSet(c);
            Iterator iter=iterator();
            while (iter.hasNext())
            {
                h.remove(iter.next());
            }

            return h.isEmpty();
        }

        return delegate.containsAll(c);
    }

    /**
     * Equality operator.
     * @param o The object to compare against.
     * @return Whether this object is the same.
     **/
    public synchronized boolean equals(Object o)
    {
        if (useCache)
        {
            loadFromStore();
        }

        if (o == this)
        {
            return true;
        }
        if (!(o instanceof java.util.Collection))
        {
            return false;
        }
        java.util.Collection c=(java.util.Collection)o;

        return c.size() == size() && containsAll(c);
    }

    /**
     * Hashcode operator.
     * @return The Hash code.
     **/
    public synchronized int hashCode()
    {
        if (useCache)
        {
            loadFromStore();
        }
        return delegate.hashCode();
    }

    /**
     * Accessor for whether the Collection is empty.
     * @return Whether it is empty.
     **/
    public synchronized boolean isEmpty()
    {
        return (size() == 0);
    }

    /**
     * Accessor for an iterator for the Collection.
     * @return The iterator
     **/
    public synchronized Iterator iterator()
    {
        // Populate the cache if necessary
        if (useCache)
        {
            loadFromStore();
        }
        return new SCOCollectionIterator(this, ownerSM, delegate, setStore, useCache);
    }

    /**
     * Accessor for the size of the Collection.
     * @return The size
     **/
    public synchronized int size()
    {
        if (useCache && isCacheLoaded)
        {
            // If the "delegate" is already loaded, use it
            return delegate.size();
        }
        else if (setStore != null)
        {
            return setStore.size(ownerSM);
        }

        return delegate.size();
    }

    /**
     * Method to return the Collection as an array.
     * @return The array
     **/
    public synchronized Object[] toArray()
    {
        if (useCache)
        {
            loadFromStore();
        }
        return delegate.toArray();
    }

    /**
     * Method to return the Collection as an array.
     * @param a The array to write the results to
     * @return The array
     **/
    public synchronized Object[] toArray(Object a[])
    {
        if (useCache)
        {
            loadFromStore();
        }
        return delegate.toArray(a);
    }

    /**
     * Method to return the Collection as a String.
     * @return The string form
     **/
    public String toString()
    {
        StringBuffer s = new StringBuffer("[");
        int i=0;
        Iterator iter=iterator();
        while (iter.hasNext())
        {
            if (i > 0)
            {
                s.append(',');
            }
            s.append(iter.next());
            i++;
        }
        s.append("]");

        return s.toString();
    }

    // ----------------------------- Mutator methods ---------------------------

    /**
     * Method to add an element to the Collection.
     * @param element The element to add
     * @return Whether it was added successfully.
     **/
    public synchronized boolean add(Object element)
    {
        // Reject inappropriate elements
        if (element == null && !allowNulls)
        {
            throw new NullsNotAllowedException(ownerSM, fieldName);
        }

        makeDirty();

        if (useCache)
        {
            loadFromStore();
        }

        if (setStore != null)
        {
            setStore.add(ownerSM,element);
        }
        return delegate.add(element);
    }

    /**
     * Method to add a collection of elements.
     * @param c The collection of elements to add.
     * @return Whether they were added successfully.
     **/
    public synchronized boolean addAll(java.util.Collection c)
    {
        makeDirty();

        if (useCache)
        {
            loadFromStore();
        }

        if (setStore != null)
        {
            setStore.addAll(ownerSM,c);
        }
        return delegate.addAll(c);
    }

    /**
     * Method to clear the Collection.
     **/
    public synchronized void clear()
    {
        makeDirty();

        if (setStore != null)
        {
            setStore.clear(ownerSM);
        }
        delegate.clear();
    }

    /**
     * Method to remove an element from the Collection.
     * @param element The Element to remove
     * @return Whether it was removed successfully.
     **/
    public synchronized boolean remove(Object element)
    {
        makeDirty();

        if (useCache)
        {
            loadFromStore();
        }

        if (setStore != null)
        {
            setStore.remove(ownerSM, element);
        }
        return delegate.remove(element);
    }

    /**
     * Method to remove a Collection of elements.
     * @param c The collection to remove
     * @return Whether they were removed successfully.
     **/
    public synchronized boolean removeAll(java.util.Collection c)
    {
        makeDirty();
 
        if (useCache)
        {
            loadFromStore();
        }

        if (setStore != null)
        {
            setStore.removeAll(ownerSM,c);
        }
        return delegate.removeAll(c);
    }

    /**
     * Method to retain a Collection of elements (and remove all others).
     * @param c The collection to retain
     * @return Whether they were retained successfully.
     **/
    public synchronized boolean retainAll(java.util.Collection c)
    {
        makeDirty();

        if (useCache)
        {
            loadFromStore();
        }
        if (setStore != null)
        {
            Iterator iter=iterator();
            while (iter.hasNext())
            {
                if (!c.contains(iter.next()))
                {
                    iter.remove();
                }
            }
        }

        return delegate.retainAll(c);
    }

    /**
     * The writeReplace method is called when ObjectOutputStream is preparing
     * to write the object to the stream. The ObjectOutputStream checks whether
     * the class defines the writeReplace method. If the method is defined, the
     * writeReplace method is called to allow the object to designate its
     * replacement in the stream. The object returned should be either of the
     * same type as the object passed in or an object that when read and
     * resolved will result in an object of a type that is compatible with all
     * references to the object.
     *
     * @return the replaced object
     * @throws ObjectStreamException
     */
    protected Object writeReplace() throws ObjectStreamException
    {
        return new java.util.HashSet(this);
    }

    /**
     * Method to load all elements from the "backing store" where appropriate.
     */
    protected void loadFromStore()
    {
        if (setStore != null && !isCacheLoaded)
        {
            delegate.clear();
            Iterator iter=setStore.iterator(ownerSM);
            while (iter.hasNext())
            {
                delegate.add(iter.next());
            }

            isCacheLoaded = true;
        }
    }
}
