/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
2004 Andy Jefferson - added Queryable implementation
2004 Andy Jefferson - added caching capability
    ...
**********************************************************************/
package org.jpox.sco;

import java.io.ObjectStreamException;
import java.util.Collection;
import java.util.Iterator;

import javax.jdo.spi.PersistenceCapable;

import org.jpox.StateManager;
import org.jpox.sco.exceptions.IncompatibleElementTypeException;
import org.jpox.sco.exceptions.IncompatibleFieldTypeException;
import org.jpox.sco.exceptions.NullsNotAllowedException;
import org.jpox.sco.exceptions.QueryUnownedSCOException;
import org.jpox.store.DatastoreClass;
import org.jpox.store.StoreManager;
import org.jpox.store.mapping.HashMapMapping;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.query.Query;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.query.Queryable;
import org.jpox.store.scostore.MapStore;
import org.jpox.store.scostore.SetStore;
import org.jpox.util.ClassUtils;

/**
 * A mutable second-class HashMap object. Backed by a MapStore object.
 * 
 * @version $Revision: 1.7 $
 */
public class HashMap extends java.util.HashMap
    implements SCO, Cloneable, Queryable, SCOMap
{
    private transient Object owner;
    private transient StateManager ownerSM;
    private transient String fieldName;
    private transient int fieldNumber;
    private transient Class keyType;
    private transient Class valueType;
    private transient boolean allowNulls;

    /**
     * The "backing store"
     */
    protected MapStore mapStore;

    /**
     * The internal "delegate".
     */
    protected java.util.HashMap delegate;

    /**
     * Whether to use "delegate" caching.
     */
    protected boolean useCache=true;

    /**
     * Status flag whether the collection is loaded into the cache.
     */
    protected boolean isCacheLoaded=false;

    /**
     * Default Constructor.
     */
    public HashMap()
    {
    }
    
    public HashMap(StateManager ownerSM, String fieldName)
    {
        this(ownerSM, fieldName, null, null, false);
    }

    public HashMap(StateManager ownerSM,
                   String fieldName,
                   Class keyType,
                   Class valueType,
                   boolean allowNulls)
    {
        super(0);
        Class c = ownerSM.getObject().getClass();
        StoreManager storeMgr = ownerSM.getStoreManager();
        DatastoreClass ownerTable = storeMgr.getDatastoreClass(c.getName());
        JavaTypeMapping m = ownerTable.getFieldMapping(fieldName);
        if (!(m instanceof HashMapMapping))
        {
            throw new IncompatibleFieldTypeException(ownerSM, fieldName, java.util.HashMap.class.getName(), ownerSM.getMetaDataManager().getMetaDataForField(c, fieldName).getTypeName());
        }

        init(ownerSM, fieldName, ((HashMapMapping)m).getMapStore(), allowNulls);

        if (keyType != null &&
            !this.keyType.isAssignableFrom(keyType))
        {
            throw new IncompatibleElementTypeException(ownerSM, fieldName, this.keyType.getName(), keyType.getName());
        }

        if (valueType != null &&
            !this.valueType.isAssignableFrom(valueType))
        {
            throw new IncompatibleElementTypeException(ownerSM, fieldName, this.valueType.getName(), valueType.getName());
        }
    }

    HashMap(StateManager ownerSM, 
            String fieldName, 
            boolean allowNulls, 
            MapStore mapStore)
    {
        super(0);
        init(ownerSM, fieldName, mapStore, allowNulls);
    }

    /**
     * Initialisation method.
     * @param ownerSM State Manager for this HashMap.
     * @param fieldName name of field in owner object
     * @param mapStore Backing store to use
     * @param allowNulls Whether to allow nulls.
     **/
    private void init(StateManager ownerSM, 
                      String fieldName, 
                      MapStore mapStore, 
                      boolean allowNulls)
    {
        this.owner = ownerSM.getObject();
        this.ownerSM = ownerSM;
        this.fieldName = fieldName;
        this.fieldNumber = ownerSM.getMetaDataManager().getMetaDataForClass(owner.getClass()).getFieldNumberAbsolute(fieldName);
        this.mapStore = mapStore;
        this.keyType = ownerSM.getPersistenceManager().getClassLoaderResolver().classForName(mapStore.getKeyType());
        this.valueType = ownerSM.getPersistenceManager().getClassLoaderResolver().classForName(mapStore.getValueType());
        this.allowNulls = allowNulls;

        // Set caching policy
        useCache = SCOUtils.useContainerCache(ownerSM, fieldName); 

        // Set up our "delegate"
        this.delegate = new java.util.HashMap();

        // If we aren't caching, just load up what we have now.
        if (!useCache)
        {
            SetStore keystore=mapStore.keySetStore();
            Iterator iter=keystore.iterator(ownerSM);
            while (iter.hasNext())
            {
                Object key=iter.next();
                delegate.put(key,mapStore.get(ownerSM,key));
            }
        }
    }

    /**
     * Mutator to set the value using an object.
     * @param o Object to set value using.
     **/
    public void setValueFrom(Object o)
    {
        java.util.Map m = (java.util.Map)o;

        clear();
        putAll(m);
    }

    /**
     * Accessor for the field name that this HashMap relates to.
     * @return The field name
     **/
    public String getFieldName()
    {
        return fieldName;
    }

    /**
     * Accessor for the owner that this HashMap relates to.
     * @return The owner
     **/
    public Object getOwner()
    {
        return owner;
    }

    /**
     * Method to unset the owner and field details.
     **/
    public synchronized void unsetOwner()
    {
        if (ownerSM != null)
        {
            owner = null;
            ownerSM = null;
            fieldName = null;
            mapStore = null;
        }
    }

    /**
     * Utility to apply updates
     **/
    public void applyUpdates()
    {
        // We are never dirty, since all changes are immediately delegated to
        // the underlying MapStore.
        return;
    }

    /**
     * Utility to mark the object as dirty
     **/
    public void makeDirty()
    {
        // Although we are never really "dirty", the owning object must be
        // marked dirty so that the proper state change occurs and its
        // jdoPreStore() gets called properly.
        if (ownerSM != null)
        {
            ownerSM.makeDirty(fieldNumber);
        }
    }

    public Object detachCopy(StateManager sm, Object value)
    {
        java.util.Map c = (java.util.Map) value;

        if (value == null)
        {
            // accept null collections...
            return new java.util.HashSet();
        }

        java.util.Map detached = new java.util.HashMap();
        for (Iterator it = c.entrySet().iterator(); it.hasNext();)
        {
            Map.Entry entry = (Map.Entry) it.next();
            Object val = entry.getValue();
            Object key = entry.getKey();
            if (val instanceof PersistenceCapable)
            {
                if (val instanceof PersistenceCapable)
                {
                    val = sm.getPersistenceManager().detachCopy(val);
                }
            }
            if (key instanceof PersistenceCapable)
            {
                if (key instanceof PersistenceCapable)
                {
                    key = sm.getPersistenceManager().detachCopy(key);
                }
            }
            detached.put(key, val);
        }
        return detached;
    }

    public Object attachCopy(StateManager sm, Object value, boolean makeTransactional)
    {
        java.util.Map c = (java.util.Map) value;

        if (value == null)
        {
            return null;
        }

        java.util.Map attached = new java.util.HashMap();
        for (Iterator it = c.entrySet().iterator(); it.hasNext();)
        {
            Map.Entry entry = (Map.Entry) it.next();
            Object val = entry.getValue();
            Object key = entry.getKey();
            if (val instanceof PersistenceCapable)
            {
                if (ClassUtils.isPersistenceCapable(val))
                {
                    val = sm.getPersistenceManager().attachCopy(val, makeTransactional);
                }
            }
            if (key instanceof PersistenceCapable)
            {
                if (ClassUtils.isPersistenceCapable(key))
                {
                    key = sm.getPersistenceManager().attachCopy(key, makeTransactional);
                }
            }
            attached.put(key, val);
        }
        return attached;
    }
 
    // ------------------------- Queryable Methods -----------------------------

    /**
     * Method to return a new Query statement for the Map.
     * @return The Query Statement.
     **/
    public synchronized QueryStatement newQueryStatement()
    {
        return newQueryStatement(valueType);
    }

    /**
     * Method to return a new Query statement for the Map given the candidate
     * class.
     * @param candidate_class value class.
     * @return The Query Statement
     **/
    public synchronized QueryStatement newQueryStatement(Class candidate_class)
    {
        if (mapStore == null)
        {
            throw new QueryUnownedSCOException();
        }

        return mapStore.newQueryStatement(ownerSM, candidate_class.getName());
    }

    /**
     * Method to return a new result object factory for processing of Query
     * statements.
     * @param stmt The Query Statement.
     * @param ignoreCache Whether to ignore the cache.
     * @param resultClass Whether to create objects of a particular class
     * @return The result object factory.
     **/
    public synchronized Query.ResultObjectFactory newResultObjectFactory(
                                        QueryStatement stmt,
                                        boolean ignoreCache,
                                        Class resultClass)
    {
        if (mapStore == null)
        {
            throw new QueryUnownedSCOException();
        }

        return mapStore.newResultObjectFactory(ownerSM, stmt,ignoreCache);
    }

    // ------------------ Implementation of HashMap methods --------------------
 
    /**
     * Creates and returns a copy of this object.
     *
     * <P>Mutable second-class Objects are required to provide a public
     * clone method in order to allow for copying PersistenceCapable
     * objects. In contrast to Object.clone(), this method must not throw a
     * CloneNotSupportedException.
     * @return The cloned object
     */
    public Object clone()
    {
        Object obj = super.clone();
        ((HashMap)obj).unsetOwner();

        return obj;
    }

    /**
     * Method to return if the map contains this key
     * @param key The key
     * @return Whether it is contained
     **/
    public boolean containsKey(Object key)
    {
        if (useCache && isCacheLoaded)
        {
            // If the "delegate" is already loaded, use it
            return delegate.containsKey(key);
        }
        else if (mapStore != null)
        {
            return mapStore.containsKey(ownerSM, key);
        }

        return delegate.containsKey(key);
    }

    /**
     * Method to return if the map contains this value.
     * @param value The value
     * @return Whether it is contained
     **/
    public boolean containsValue(Object value)
    {
        if (useCache && isCacheLoaded)
        {
            // If the "delegate" is already loaded, use it
            return delegate.containsValue(value);
        }
        else if (mapStore != null)
        {
            return mapStore.containsValue(ownerSM, value);
        }

        return delegate.containsValue(value);
    }

    /**
     * Accessor for the set of entries in the Map.
     * @return Set of entries
     **/
    public java.util.Set entrySet()
    {
        if (useCache)
        {
            loadFromStore();
        }
        else if (mapStore != null)
        {
            return new Set(ownerSM, fieldName, false, mapStore.entrySetStore());
        }

        return delegate.entrySet();
    }

    /**
     * Method to check the equality of this map, and another.
     * @param o The map to compare against.
     * @return Whether they are equal.
     **/
    public synchronized boolean equals(Object o)
    {
        if (useCache)
        {
            loadFromStore();
        }

        if (o == this)
        {
            return true;
        }
        if (!(o instanceof java.util.HashMap))
        {
            return false;
        }
        java.util.HashMap m = (java.util.HashMap)o;

        return entrySet().equals(m.entrySet());
    }

    /**
     * Accessor for the value stored against a key.
     * @param key The key
     * @return The value.
     **/
    public Object get(Object key)
    {
        if (useCache)
        {
            loadFromStore();
        }
        else if (mapStore != null)
        {
            return mapStore.get(ownerSM, key);
        }

        return delegate.get(key);
    }

    /**
     * Method to generate a hashcode for this Map.
     * @return The hashcode.
     **/
    public synchronized int hashCode()
    {
        if (useCache)
        {
            loadFromStore();
        }
        else if (mapStore != null)
        {
            int h = 0;
            Iterator i = entrySet().iterator();
            while (i.hasNext())
            {
                h += i.next().hashCode();
            }

            return h;
        }
        return delegate.hashCode();
    }

    /**
     * Method to return if the Map is empty.
     * @return Whether it is empty.
     **/
    public boolean isEmpty()
    {
        return size() == 0;
    }

    /**
     * Accessor for the set of keys in the Map.
     * @return Set of keys.
     **/
    public java.util.Set keySet()
    {
        if (useCache)
        {
            loadFromStore();
        }
        else if (mapStore != null)
        {
            return new Set(ownerSM, fieldName, false, mapStore.keySetStore());
        }

        return delegate.keySet();
    }

    /**
     * Method to return the size of the Map.
     * @return The size
     **/
    public int size()
    {
        if (useCache && isCacheLoaded)
        {
            // If the "delegate" is already loaded, use it
            return delegate.size();
        }
        else if (mapStore != null)
        {
            return mapStore.entrySetStore().size(ownerSM);
        }

        return delegate.size();
    }

    /**
     * Accessor for the set of values in the Map.
     * @return Set of values.
     **/
    public Collection values()
    {
        if (useCache)
        {
            loadFromStore();
        }
        else if (mapStore != null)
        {
            return new Set(ownerSM, fieldName, true, mapStore.valueSetStore());
        }

        return delegate.values();
    }

    // --------------------------- Mutator methods -----------------------------
 
    /**
     * Method to clear the HashMap.
     **/
    public void clear()
    {
        makeDirty();

        if (mapStore != null)
        {
            mapStore.clear(ownerSM);
        }
        delegate.clear();
    }
 
    /**
     * Method to add a value against a key to the HashMap.
     * @param key The key
     * @param value The value
     * @return The previous value for the specified key.
     **/
    public Object put(Object key,Object value)
    {
        // Reject inappropriate elements
        if (value == null && !allowNulls)
        {
            throw new NullsNotAllowedException(ownerSM, fieldName);
        }

        makeDirty();

        if (useCache)
        {
            loadFromStore();
        }

        if (mapStore != null)
        {
            mapStore.put(ownerSM, key, value);
        }
        return delegate.put(key, value);
    }

    /**
     * Method to add the specified Map's values under their keys here.
     * @param m The map
     * @return The previous value for the specified key.
     **/
    public void putAll(java.util.Map m)
    {
        makeDirty();

        if (useCache)
        {
            loadFromStore();
        }

        if (mapStore != null)
        {
            mapStore.putAll(ownerSM, m);
        }
        delegate.putAll(m);
    }

    /**
     * Method to remove the value for a key from the HashMap.
     * @param key The key to remove
     * @return The value that was removed from this key.
     **/
    public Object remove(Object key)
    {
        makeDirty();

        if (useCache)
        {
            loadFromStore();
        }

        if (mapStore != null)
        {
            mapStore.remove(ownerSM, key);
        }
        return delegate.remove(key);
    }

    /**
     * The writeReplace method is called when ObjectOutputStream is preparing
     * to write the object to the stream. The ObjectOutputStream checks whether
     * the class defines the writeReplace method. If the method is defined, the
     * writeReplace method is called to allow the object to designate its
     * replacement in the stream. The object returned should be either of the
     * same type as the object passed in or an object that when read and
     * resolved will result in an object of a type that is compatible with all
     * references to the object.
     * 
     * @return the replaced object
     * @throws ObjectStreamException
     */
    protected Object writeReplace() throws ObjectStreamException
    {
        return new java.util.HashMap(this);
    }

    /**
     * Method to load all elements from the "backing store" where appropriate.
     */
    protected void loadFromStore()
    {
        if (mapStore != null && !isCacheLoaded)
        {
            delegate.clear();

            SetStore keystore=mapStore.keySetStore();
            Iterator iter=keystore.iterator(ownerSM);
            while (iter.hasNext())
            {
                Object key=iter.next();
                delegate.put(key,mapStore.get(ownerSM,key));
            }

            isCacheLoaded = true;
        }
    }
}
