/**********************************************************************
Copyright (c) 2003 Mike Martin and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - commented
2004 Andy Jefferson - addition of toArray, contains, etc methods
2004 Andy Jefferson - enabled ability to use cache
    ...
**********************************************************************/
package org.jpox.sco;

import java.io.ObjectStreamException;
import java.util.Collection;
import java.util.Iterator;

import javax.jdo.spi.PersistenceCapable;

import org.jpox.StateManager;
import org.jpox.sco.exceptions.IncompatibleElementTypeException;
import org.jpox.sco.exceptions.IncompatibleFieldTypeException;
import org.jpox.sco.exceptions.NullsNotAllowedException;
import org.jpox.sco.exceptions.QueryUnownedSCOException;
import org.jpox.store.DatastoreClass;
import org.jpox.store.StoreManager;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.query.Query;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.query.Queryable;
import org.jpox.store.scostore.SetStore;
import org.jpox.util.ClassUtils;

/**
 * A mutable second-class HashSet object. Backed by a SetStore object.
 * Allows caching in the delegate of the elements of the HashSet. This can be
 * enabled globally for the PMF, or individually for this particular HashSet. 
 * 
 * @version $Revision: 1.6 $
 */
public class HashSet extends java.util.HashSet 
    implements SCO, Cloneable, Queryable
{
    private transient Object owner;
    private transient StateManager ownerSM;
    private transient String fieldName;
    private transient int fieldNumber;
    private transient Class elementType;
    private transient boolean allowNulls;

    /**
     * The "backing store"
     */
    protected SetStore setStore;

    /**
     * The internal "delegate".
     */
    protected java.util.HashSet delegate;

    /**
     * Whether to use caching.
     */
    protected boolean useCache=true;

    /**
     * Status flag whether the collection is loaded into the cache.
     */
    protected boolean isCacheLoaded=false;
    
    /**
     * Default Constructor.
     */
    public HashSet()
    {
    }
 
    /**
     * Constructor, using the StateManager of the "owner" and the field name.
     * 
     * @param ownerSM The owner StateManager
     * @param fieldName The name of the field of the SCO.
     */
    public HashSet(StateManager ownerSM, String fieldName)
    {
        this(ownerSM, fieldName, null, false);
    }

    /**
     * Constructor, using the StateManager of the "owner", the field name, 
     * the type of the element and whether null elements are allowed.
     * 
     * @param ownerSM The owner StateManager
     * @param fieldName The name of the field of the SCO.
     * @param elementType The type of the element objects.
     * @param allowNulls Whether to allow null elements.
     */
    public HashSet(StateManager ownerSM,
                   String fieldName, 
                   Class elementType,
                   boolean allowNulls)
    {
        Class c = ownerSM.getObject().getClass();
        StoreManager storeMgr = ownerSM.getStoreManager();
        DatastoreClass ownerTable = storeMgr.getDatastoreClass(c.getName());
        JavaTypeMapping m = ownerTable.getFieldMapping(fieldName);
        if (!(m instanceof org.jpox.store.mapping.HashSetMapping))
        {
            throw new IncompatibleFieldTypeException(ownerSM, fieldName, java.util.HashSet.class.getName(), ownerSM.getMetaDataManager().getMetaDataForField(c, fieldName).getTypeName());
        }

        init(ownerSM, fieldName, ((org.jpox.store.mapping.HashSetMapping)m).getSetStore(), allowNulls);

        if (elementType != null &&
            !this.elementType.isAssignableFrom(elementType))
        {
            throw new IncompatibleElementTypeException(ownerSM, fieldName, this.elementType.getName(), elementType.getName());
        }
    }

    HashSet(StateManager ownerSM, String fieldName, boolean allowNulls, SetStore setStore)
    {
        init(ownerSM, fieldName, setStore, allowNulls);
    }

    /**
     * Initialisation method.
     *
     * @param ownerSM The owner StateManager
     * @param fieldName The name of the field of the SCO.
     * @param setStore The backing store for this SCO.
     * @param allowNulls Whether to allow null elements.
     */
    private void init(StateManager ownerSM, String fieldName,
                      SetStore setStore, boolean allowNulls)
    {
        this.owner = ownerSM.getObject();
        this.ownerSM = ownerSM;
        this.fieldName = fieldName;
        this.fieldNumber = ownerSM.getMetaDataManager().getMetaDataForClass(owner.getClass()).getFieldNumberAbsolute(fieldName);
        this.setStore = setStore;
        this.elementType = ownerSM.getPersistenceManager().getClassLoaderResolver().classForName(setStore.getElementType());
        this.allowNulls = allowNulls;

        // Set caching policy
        useCache = SCOUtils.useContainerCache(ownerSM, fieldName);

        // Set up our "delegate"
        this.delegate = new java.util.HashSet();

        // If we aren't caching, just load up what we have now.
        if (!useCache)
        {
            Iterator iter=setStore.iterator(ownerSM);
            while (iter.hasNext())
            {
                delegate.add(iter.next());
            }
        }
    }

    // ----------------------- Implementation of SCO methods -------------------
 
    /**
     * Mutator for setting using an object.
     * @param o  The object to set from
     */
    public void setValueFrom(Object o)
    {
        Collection c = (Collection)o;

        clear();
        addAll(c);
    }

    /**
     * Accessor for the field name.
     * @return The field name
     */
    public String getFieldName()
    {
        return fieldName;
    }

    /**
     * Accessor for the owner object.
     * @return The owner object
     */
    public Object getOwner()
    {
        return owner;
    }

    /**
     * Method to unset the owner and field information.
     */
    public synchronized void unsetOwner()
    {
        if (ownerSM != null)
        {
            owner = null;
            ownerSM = null;
            fieldName = null;
            setStore = null;
        }
    }

    /**
     * Utility to apply updates
     **/
    public void applyUpdates()
    {
        /*
         * We are never dirty, since all changes are immediately delegated to
         * the underlying SetStore.
         */
        return;
    }

    /**
     * Utility to mark the object as dirty
     **/
    public void makeDirty()
    {
        /*
         * Although we are never really "dirty", the owning object must be
         * marked dirty so that the proper state change occurs and its
         * jdoPreStore() gets called properly.
         */
        if (ownerSM != null)
        {
            ownerSM.makeDirty(fieldNumber);
        }
    }

    public Object detachCopy(StateManager sm, Object value)
    {
        java.util.Collection c = (java.util.Collection)value;

        if (value == null)
        {
        	// accept null collections...
			return new java.util.HashSet();
        }

        java.util.Collection detached = new java.util.HashSet();
        for (Iterator it = c.iterator(); it.hasNext();)
        {
            Object object = it.next();
            if (object instanceof PersistenceCapable)
            {
                detached.add(sm.getPersistenceManager().detachCopy(object));
            }
            else
            {
                detached.add(object);
            }
        }        
        return detached;
    }
    
    public Object attachCopy(StateManager sm, Object value, boolean makeTransactional)
    {
        java.util.Collection c = (java.util.Collection)value;

        if (value == null)
        {
			return null;
        }

        java.util.Collection attached = new java.util.HashSet();
        for (Iterator it = c.iterator(); it.hasNext();)
        {
            Object element = it.next();
            if (!ClassUtils.isPersistenceCapable(element))
            {
                attached.add(element);
            }
            else
            {
                attached.add(sm.getPersistenceManager().attachCopy(element,makeTransactional));
            }
        }
        return attached;
    }
    
    // ------------------------ Query Statement methods ------------------------
    /**
     * Method to generate a QueryStatement for the SCO.
     * @return The QueryStatement
     **/
    public synchronized QueryStatement newQueryStatement()
    {
        return newQueryStatement(elementType);
    }

    /**
     * Method to return a QueryStatement, using the specified candidate class.
     * @param candidate_class the candidate class
     * @return The QueryStatement
     **/
    public synchronized QueryStatement newQueryStatement(Class candidate_class)
    {
        if (setStore == null)
        {
            throw new QueryUnownedSCOException();
        }

        return setStore.newQueryStatement(ownerSM, candidate_class.getName());
    }

    /**
     * Method to return a ResultObjectFactory for the SCO.
     * @param stmt The QueryStatement
     * @param ignoreCache Whether to ignore the cache
     * @param resultClass Whether to create objects of a particular class
     * @return The ResultObjectFactory
     **/
    public synchronized Query.ResultObjectFactory newResultObjectFactory(QueryStatement stmt,boolean ignoreCache, Class resultClass)
    {
        if (setStore == null)
        {
            throw new QueryUnownedSCOException();
        }

        return setStore.newResultObjectFactory(ownerSM, stmt,ignoreCache);
    }

    // ------------------ Implementation of HashSet methods --------------------

    /**
     * Creates and returns a copy of this object.
     * @return The cloned object
     */
    public Object clone()
    {
        Object obj = super.clone();
        ((HashSet)obj).unsetOwner();

        return obj;
    }

    /**
     * Accessor for whether an element is contained in this Set.
     * @param element The element
     * @return Whether it is contained.
     **/
    public boolean contains(Object element)
    {
        if (useCache && isCacheLoaded)
        {
            // If the "delegate" is already loaded, use it
            return delegate.contains(element);
        }
        else if (setStore != null)
        {
            return setStore.contains(ownerSM,element);
        }

        return delegate.contains(element);
    }

    /**
     * Accessor for whether a collection is contained in this Set.
     * @param c The collection
     * @return Whether it is contained.
     **/
    public synchronized boolean containsAll(java.util.Collection c)
    {
        if (useCache)
        {
            loadFromStore();
        }
        else if (setStore != null)
        {
            java.util.HashSet h=new java.util.HashSet(c);
            Iterator iter=iterator();
            while (iter.hasNext())
            {
                h.remove(iter.next());
            }

            return h.isEmpty();
        }

        return delegate.containsAll(c);
    }

    /**
     * Equality operator.
     * @param o The object to compare against.
     * @return Whether this object is the same.
     **/
    public synchronized boolean equals(Object o)
    {
        if (useCache)
        {
            loadFromStore();
        }

        if (o == this)
        {
            return true;
        }
        if (!(o instanceof java.util.HashSet))
        {
            return false;
        }
        java.util.HashSet c = (java.util.HashSet)o;

        return c.size() == size() && containsAll(c);
    }

    /**
     * Hashcode operator.
     * @return The Hash code.
     **/
    public synchronized int hashCode()
    {
        if (useCache)
        {
            loadFromStore();
        }
        return delegate.hashCode();
    }

    /**
     * Accessor for whether the HashSet is empty.
     * @return Whether it is empty.
     **/
    public boolean isEmpty()
    {
        return size() == 0;
    }

    /**
     * Accessor for an iterator for the Set.
     * @return The iterator
     **/
    public Iterator iterator()
    {
        // Populate the cache if necessary
        if (useCache)
        {
            loadFromStore();
        }
        return new SCOCollectionIterator(this, ownerSM, delegate, setStore, useCache);
    }

    /**
     * Accessor for the size of the HashSet.
     * @return The size.
     **/
    public int size()
    {
        if (useCache && isCacheLoaded)
        {
            // If the "delegate" is already loaded, use it
            return delegate.size();
        }
        else if (setStore != null)
        {
            return setStore.size(ownerSM);
        }

        return delegate.size();
    }

    /**
     * Method to return the list as an array.
     * @return The array
     **/
    public Object[] toArray()
    {
        if (useCache)
        {
            loadFromStore();
        }
        return delegate.toArray();
    }

    /**
     * Method to return the list as an array.
     * @param a The runtime types of the array being defined by this param
     * @return The array
     **/
    public Object[] toArray(Object a[])
    {
        if (useCache)
        {
            loadFromStore();
        }
        return delegate.toArray(a);
    }
 
    // ------------------------------ Mutator methods --------------------------

    /**
     * Method to add an element to the HashSet.
     * @param element The new element
     * @return Whether it was added ok.
     **/
    public boolean add(Object element)
    {
        // Reject inappropriate elements
        if (element == null && !allowNulls)
        {
            throw new NullsNotAllowedException(ownerSM, fieldName);
        }

        makeDirty();

        if (useCache)
        {
            loadFromStore();
        }

        if (setStore != null)
        {
            setStore.add(ownerSM,element);
        }
        return delegate.add(element);
    }

    /**
     * Method to add a collection to the HashSet.
     * @param c The collection
     * @return Whether it was added ok.
     **/
    public boolean addAll(Collection c)
    {
        makeDirty();

        if (useCache)
        {
            loadFromStore();
        }

        if (setStore != null)
        {
            setStore.addAll(ownerSM,c);
        }
        return delegate.addAll(c);
    }

    /**
     * Method to clear the HashSet
     **/
    public void clear()
    {
        makeDirty();

        if (setStore != null)
        {
            setStore.clear(ownerSM);
        }
        delegate.clear();
    }

    /**
     * Method to remove an element from the HashSet.
     * @param element The element
     * @return Whether it was removed ok.
     **/
    public boolean remove(Object element)
    {
        makeDirty();

        if (useCache)
        {
            loadFromStore();
        }

        if (setStore != null)
        {
            setStore.remove(ownerSM, element);
        }
        return delegate.remove(element);
    }

    /**
     * Method to remove all elements from the collection from the HashSet.
     * @param c The collection of elements to remove 
     * @return Whether it was removed ok.
     **/
    public boolean removeAll(java.util.Collection c)
    {
        makeDirty();
 
        if (useCache)
        {
            loadFromStore();
        }

        if (setStore != null)
        {
            setStore.removeAll(ownerSM,c);
        }
        return delegate.removeAll(c);
    }

    /**
     * Method to remove all elements except those in the collection.
     * @param c The collection
     * @return Whether it was successful.
     **/
    public boolean retainAll(Collection c)
    {
        makeDirty();

        if (useCache)
        {
            loadFromStore();
        }
        if (setStore != null)
        {
            Iterator iter=iterator();
            while (iter.hasNext())
            {
                if (!c.contains(iter.next()))
                {
                    iter.remove();
                }
            }
        }

        return delegate.retainAll(c);
    }

    /**
     * The writeReplace method is called when ObjectOutputStream is preparing
     * to write the object to the stream. The ObjectOutputStream checks
     * whether the class defines the writeReplace method. If the method is
     * defined, the writeReplace method is called to allow the object to
     * designate its replacement in the stream. The object returned should be
     * either of the same type as the object passed in or an object that when
     * read and resolved will result in an object of a type that is compatible
     * with all references to the object.
     * 
     * @return the replaced object
     * @throws ObjectStreamException
     */
    protected Object writeReplace() throws ObjectStreamException
    {
        return new java.util.HashSet(this);
    }

    /**
     * Method to load all elements from the "backing store" where appropriate.
     */
    protected void loadFromStore()
    {
        if (setStore != null && !isCacheLoaded)
        {
            delegate.clear();
            Iterator iter=setStore.iterator(ownerSM);
            while (iter.hasNext())
            {
                delegate.add(iter.next());
            }

            isCacheLoaded = true;
        }
    }
}
