/**********************************************************************
Copyright (c) 2003 David Jencks and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - updated to make correct use of makeDirty() etc.
2004 Andy Jefferson - rewritten to always have delegate.
2004 Andy Jefferson - rewritten to allow caching
    ...
**********************************************************************/
package org.jpox.sco;

import java.io.ObjectStreamException;
import java.util.AbstractList;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.ListIterator;

import javax.jdo.spi.PersistenceCapable;

import org.jpox.StateManager;
import org.jpox.sco.exceptions.IncompatibleElementTypeException;
import org.jpox.sco.exceptions.IncompatibleFieldTypeException;
import org.jpox.sco.exceptions.NullsNotAllowedException;
import org.jpox.sco.exceptions.QueryUnownedSCOException;
import org.jpox.store.DatastoreClass;
import org.jpox.store.StoreManager;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.query.Query;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.query.Queryable;
import org.jpox.store.scostore.ListStore;
import org.jpox.util.ClassUtils;

/**
 * A mutable second-class List object, representing a field in a class,
 * and containing elements of a particular type.
 * Uses a "delegate" as a local store for the List. Uses a "backing store"
 * (ListStore) to represent the datastore. The "delegate" is updated with the
 * "backing store" information at necessary intervals. 
 *
 * <H3>Modes of Operation</H3>
 * The user can operate the collection in 2 modes.
 * The <B>cached</B> mode will use an internal cache of the elements (in the
 * "delegate") reading them at the first opportunity and then using the cache
 * thereafter.
 * The <B>non-cached</B> mode will just go direct to the "backing store" each
 * call.
 *
 * <H3>Mutators</H3>
 * When the backing store is present any updates are passed direct to the
 * datastore as well as to the "delegate". If the "backing store" isn't present
 * the changes are made to the "delegate" only.
 *
 * <H3>Accessors</H3>
 * When any accessor method is invoked, it typically checks whether the
 * collection has been loaded from its backing store and does this as
 * necessary. Some methods (<B>size()</B>, <B>contains()</B>) just check if
 * everything is loaded and use the delegate if possible, otherwise going
 * direct to the datastore.
 *
 * @version $Revision: 1.4 $
 */
public class List extends AbstractList
    implements SCO, Cloneable, Queryable, SCOList, java.io.Serializable
{
    private StateManager ownerSM;
    private Object owner;
    private String fieldName;
    private int fieldNumber;
    private Class elementType;
    private boolean allowNulls;

    /**
     * The "backing store"
     */
    private ListStore listStore;

    /**
     * The internal "delegate".
     */
    private ArrayList delegate;

    /**
     * Whether to use "delegate" caching.
     */
    protected boolean useCache=true;

    /**
     * Status flag whether the collection is loaded into the cache.
     */
    protected boolean isCacheLoaded=false;

    /**
     * Default Constructor.
     */
    public List()
    {
    }
    
    /**
     * Constructor, using the StateManager of the "owner" and the field name.
     * @param ownerSM The owner StateManager
     * @param fieldName The name of the field of the SCO.
     */
    public List(StateManager ownerSM, String fieldName)
    {
        this(ownerSM, fieldName, null, false);
    }

    /**
     * Constructor, using the StateManager of the "owner", the field name, 
     * the type of the element and whether null elements are allowed.
     * @param ownerSM The owner StateManager
     * @param fieldName The name of the field of the SCO.
     * @param elementType The type of the element objects.
     * @param allowNulls Whether to allow null elements.
     */
    public List(StateManager ownerSM,
                String fieldName,
                Class elementType,
                boolean allowNulls)
    {
        Class c = ownerSM.getObject().getClass();
        StoreManager storeMgr = ownerSM.getStoreManager();
        DatastoreClass ownerTable = storeMgr.getDatastoreClass(c.getName());
        JavaTypeMapping m = ownerTable.getFieldMapping(fieldName);
        if (!(m instanceof org.jpox.store.mapping.ListMapping))
        {
            throw new IncompatibleFieldTypeException(ownerSM, fieldName, Collection.class.getName(), ownerSM.getMetaDataManager().getMetaDataForField(c,fieldName).getTypeName());
        }

        init(ownerSM, fieldName, ((org.jpox.store.mapping.ListMapping)m).getListStore(), allowNulls);

        if (elementType != null &&
            !this.elementType.isAssignableFrom(elementType))
        {
            throw new IncompatibleElementTypeException(ownerSM, fieldName, this.elementType.getName(), elementType.getName());
        }
    }

    /**
     * Initialisation method.
     * @param ownerSM The owner StateManager
     * @param fieldName The name of the field of the SCO.
     * @param listStore The backing store for this SCO.
     * @param allowNulls Whether to allow null elements.
     */
    private void init(StateManager ownerSM, 
                      String fieldName,
                      ListStore listStore,
                      boolean allowNulls)
    {
        this.owner = ownerSM.getObject();
        this.ownerSM = ownerSM;
        this.fieldName = fieldName;
        this.fieldNumber = ownerSM.getMetaDataManager().getMetaDataForClass(owner.getClass()).getFieldNumberAbsolute(fieldName);
        this.listStore = listStore;
        this.elementType = ownerSM.getPersistenceManager().getClassLoaderResolver().classForName(listStore.getElementType());
        this.allowNulls = allowNulls;

        // Set caching policy
        useCache = SCOUtils.useContainerCache(ownerSM, fieldName);

        // Set up our delegate
        this.delegate = new ArrayList();

        // Copy in existing elements
        if (!useCache)
        { 
            Iterator iter=listStore.iterator(ownerSM);
            while (iter.hasNext())
            {
                delegate.add(iter.next());
            }
        }
    }

    /**
     * Mutator for setting using an object.
     * @param o The object to set from
     */
    public void setValueFrom(Object o)
    {
        Collection c = (Collection)o;

        clear();
        addAll(c);
    }

    /**
     * Accessor for the field name.
     * @return The field name
     */
    public String getFieldName()
    {
        return fieldName;
    }

    /**
     * Accessor for the owner object.
     * @return The owner object
     */
    public Object getOwner()
    {
        return owner;
    }

    /**
     * Method to unset the owner and field information.
     */
    public synchronized void unsetOwner()
    {
        if (ownerSM != null)
        {
            owner = null;
            ownerSM = null;
            fieldName = null;
            listStore = null;
        }
    }

    /**
     * Utility to apply updates
     **/
    public void applyUpdates()
    {
        /*
         * We are never dirty, since all changes are immediately delegated to
         * the underlying ListStore.
         */
        return;
    }

    /**
     * Utility to mark the object as dirty
     **/
    public void makeDirty()
    {
        /*
         * Although we are never really "dirty", the owning object must be
         * marked dirty so that the proper state change occurs and its
         * jdoPreStore() gets called properly.
         */
        if (ownerSM != null)
        {
            ownerSM.makeDirty(fieldNumber);
        }
    }

    public Object detachCopy(StateManager sm, Object value)
    {
        java.util.Collection c = (java.util.Collection)value;

        if (value == null)
        {
        	// accept null collections...
			return new java.util.HashSet();
        }

        java.util.Collection detached = new java.util.ArrayList();
        Object[] values = ((java.util.List)c).toArray();
        for (int i=0; i<values.length; i++)
        {
            if (values[i] == null)
            {
                detached.add(null);
            }
            else
            {
                Object object = values[i];
                if (object instanceof PersistenceCapable)
                {
                    detached.add(sm.getPersistenceManager().detachCopy(object));
                }
                else
                {
                    detached.add(object);
                }
            }
        }         
        return detached;
    }
    
    public Object attachCopy(StateManager sm, Object value, boolean makeTransactional)
    {
        java.util.Collection c = (java.util.Collection)value;

        if (value == null)
        {
			return null;
        }

        java.util.Collection attached = new java.util.ArrayList();
        Object[] values = ((java.util.List)c).toArray();
        for (int i=0; i<values.length; i++)
        {
            if (values[i] == null)
            {
                attached.add(null);
            }
            else
            {
                if (!ClassUtils.isPersistenceCapable(values[i]))
                {
                    attached.add(values[i]);
                }
                else
                {
                    attached.add(sm.getPersistenceManager().attachCopy(values[i],makeTransactional));
                }
            }
        }         
        return attached;
    }
    
    // ------------------------ Query Statement methods ------------------------

    /**
     * Method to generate a QueryStatement for the SCO.
     * @return The QueryStatement
     */
    public synchronized QueryStatement newQueryStatement()
    {
        return newQueryStatement(elementType);
    }

    /**
     * Method to return a QueryStatement, using the specified candidate class.
     * @param candidate_class The candidate class
     * @return The QueryStatement
     */
    public synchronized QueryStatement newQueryStatement(Class candidate_class)
    {
        if (listStore == null)
        {
            throw new QueryUnownedSCOException();
        }

        return listStore.newQueryStatement(ownerSM, candidate_class.getName());
    }

    /**
     * Method to return a ResultObjectFactory for the SCO.
     * @param stmt The QueryStatement
     * @param ignoreCache Whether to ignore the cache
     * @param resultClass Whether to create objects of a particular class
     * @return The ResultObjectFactory
     */
    public synchronized Query.ResultObjectFactory newResultObjectFactory(QueryStatement stmt,boolean ignoreCache,Class resultClass)
    {
        if (listStore == null)
        {
            throw new QueryUnownedSCOException();
        }

        return listStore.newResultObjectFactory(ownerSM, stmt,ignoreCache);
    }

    // ------------------- Implementation of List methods ----------------------
 
    /**
     * Creates and returns a copy of this object.
     * <P>Mutable second-class Objects are required to provide a public
     * clone method in order to allow for copying PersistenceCapable
     * objects. In contrast to Object.clone(), this method must not throw a
     * CloneNotSupportedException.
     * @return The cloned object
     */
    public Object clone()
    {
        Object obj=null;
        try
        {
            obj = super.clone();
            ((List)obj).unsetOwner();
        }
        catch (CloneNotSupportedException cnse)
        {
            // Can't happen
        }

        return obj;
    }

    /**
     * Accessor for whether an element is contained in the List.
     * @param element The element
     * @return Whether the element is contained here
     **/
    public synchronized boolean contains(Object element)
    {
        if (useCache && isCacheLoaded)
        {
            // If the "delegate" is already loaded, use it
            return delegate.contains(element);
        }
        else if (listStore != null)
        {
            return listStore.contains(ownerSM,element);
        }

        return delegate.contains(element);
    }

    /**
     * Equality operator.
     * @param o The object to compare against.
     * @return Whether this object is the same.
     **/
    public synchronized boolean equals(Object o)
    {
        if (useCache)
        {
            loadFromStore();
        }

        if (o == this)
        {
            return true;
        }
        if (!(o instanceof java.util.List))
        {
            return false;
        }
        java.util.List l=(java.util.List)o;

        return l.size() == size() && containsAll(l);
    }

    /**
     * Method to retrieve an element no.
     * @param index The item to retrieve
     * @return The element at that position.
     **/
    public Object get(int index)
    {
        if (useCache)
        {
            loadFromStore();
        }
        else if (listStore != null)
        {
            return listStore.get(ownerSM, index);
        }

        return delegate.get(index);
    }

    /**
     * Hashcode operator.
     * @return The Hash code.
     **/
    public synchronized int hashCode()
    {
        if (useCache)
        {
            loadFromStore();
        }
        return delegate.hashCode();
    }

    /**
     * Method to the position of an element.
     * @param element The element.
     * @return The position.
     **/
    public int indexOf(Object element)
    {
        if (useCache)
        {
            loadFromStore();
        }
        else if (listStore != null)
        {
            return listStore.indexOf(ownerSM, element);
        }
 
        return delegate.indexOf(element);
    }

    /**
     * Accessor for whether the List is empty.
     * @return Whether it is empty.
     **/
    public synchronized boolean isEmpty()
    {
        return size() == 0;
    }

    /**
     * Method to retrieve the last position of the element.
     * @param element The element
     * @return The last position of this element in the List.
     **/
    public int lastIndexOf(Object element)
    {
        if (useCache)
        {
            loadFromStore();
        }
        else if (listStore != null)
        {
            return listStore.lastIndexOf(ownerSM, element);
        }
 
        return delegate.lastIndexOf(element);
    }

    /**
     * Accessor for an iterator for the List
     * @return The iterator
     **/
    public synchronized Iterator iterator()
    {
        // Populate the cache if necessary
        if (useCache)
        {
            loadFromStore();
        }

        return new SCOListIterator(this, ownerSM, delegate, listStore, useCache, -1);
    }

    /**
     * Method to retrieve a List iterator for the list.
     * @return The iterator
     **/
    public ListIterator listIterator()
    {
        // Populate the cache if necessary
        if (useCache)
        {
            loadFromStore();
        }

        return new SCOListIterator(this, ownerSM, delegate, listStore, useCache, -1);
    }

    /**
     * Method to retrieve a List iterator for the list from the index.
     * @param index The start point 
     * @return The iterator
     **/
    public ListIterator listIterator(int index)
    {
        // Populate the cache if necessary
        if (useCache)
        {
            loadFromStore();
        }

        return new SCOListIterator(this, ownerSM, delegate, listStore, useCache, index);
    }

    /**
     * Accessor for the size of the List
     * @return The size
     **/
    public synchronized int size()
    {
        if (useCache && isCacheLoaded)
        {
            // If the "delegate" is already loaded, use it
            return delegate.size();
        }
        else if (listStore != null)
        {
            return listStore.size(ownerSM);
        }

        return delegate.size();
    }

    /**
     * Accessor for the subList of elements between from and to of the List
     * @param from Start index (inclusive)
     * @param to End index (exclusive) 
     * @return The subList
     **/
    public synchronized java.util.List subList(int from,int to)
    {
        if (useCache)
        {
            loadFromStore();
        }
        else if (listStore != null)
        {
            return listStore.subList(ownerSM,from,to);
        }

        return delegate.subList(from,to);
    }

    /**
     * Method to return the List as an array.
     * @return The array
     **/
    public synchronized Object[] toArray()
    {
        if (useCache)
        {
            loadFromStore();
        }
        return delegate.toArray();
    }

    /**
     * Method to return the List as an array.
     * @param a The array to copy to
     * @return The array
     **/
    public synchronized Object[] toArray(Object a[])
    {
        if (useCache)
        {
            loadFromStore();
        }
        return delegate.toArray(a);
    }

    // ----------------------------- Mutators methods --------------------------
 
    /**
     * Method to add an element to the List
     * @param element The element to add
     * @return Whether it was added successfully.
     **/
    public synchronized boolean add(Object element)
    {
        // Reject inappropriate elements
        if (element == null && !allowNulls)
        {
            throw new NullsNotAllowedException(ownerSM, fieldName);
        }

        makeDirty();

        if (useCache)
        {
            loadFromStore();
        }

        if (listStore != null)
        {
            listStore.add(ownerSM,element);
        }
        return delegate.add(element);
    }

    /**
     * Method to add an element to the List at a position.
     * @param element The element to add
     * @param index The position
     **/
    public void add(int index, Object element)
    {
        // Reject inappropriate elements
        if (element == null && !allowNulls)
        {
            throw new NullsNotAllowedException(ownerSM, fieldName);
        }

        makeDirty();

        if (useCache)
        {
            loadFromStore();
        }

        if (listStore != null)
        {
            listStore.add(ownerSM,index,element);
        }
        delegate.add(index,element);
    }

    /**
     * Method to add a collection of elements.
     * @param elements The collection of elements to add.
     * @param index The position to add them
     * @return Whether they were added successfully.
     **/
    public boolean addAll(int index, Collection elements)
    {
        makeDirty();

        if (useCache)
        {
            loadFromStore();
        }

        if (listStore != null)
        {
            listStore.addAll(ownerSM,index,elements);
        }
        return delegate.addAll(index,elements);
    }

    /**
     * Method to clear the List
     **/
    public synchronized void clear()
    {
        makeDirty();

        if (listStore != null)
        {
            listStore.clear(ownerSM);
        }
        delegate.clear();
    }

    /**
     * Method to remove an element from the List
     * @param element The Element to remove
     * @return Whether it was removed successfully.
     **/
    public synchronized boolean remove(Object element)
    {
        makeDirty();

        if (useCache)
        {
            loadFromStore();
        }

        if (listStore != null)
        {
            listStore.remove(ownerSM, element);
        }
        return delegate.remove(element);
    }

    /**
     * Method to remove an element from the ArrayList.
     * @param index The element position.
     * @return The object that was removed
     */
    public Object remove(int index)
    {
        makeDirty();
 
        if (useCache)
        {
            loadFromStore();
        }

        if (listStore != null)
        {
            listStore.remove(ownerSM, index);
        }
        return delegate.remove(index);
    }

    // This seems not to be necessary ... AbstractList.removeAll calls other
    // methods on the class to remove the elements.
    /**
     * Method to remove a collection of elements from the List.
     * @param elements Collection of elements to remove
     * @return Whether it was successful.
     */
    public boolean removeAll(Collection elements)
    {
        makeDirty();

        if (useCache)
        {
            loadFromStore();
        }

        if (listStore != null)
        {
            listStore.removeAll(ownerSM, elements);
        }
        return delegate.removeAll(elements);
    }

    /**
     * Method to set the element at a position in the ArrayList.
     * @param index The position
     * @param element The new element
     * @return The element previously at that position
     */
    public Object set(int index, Object element)
    {
        // Reject inappropriate elements
        if (element == null && !allowNulls)
        {
            throw new NullsNotAllowedException(ownerSM, fieldName);
        }

        makeDirty();

        if (useCache)
        {
            loadFromStore();
        }

        if (listStore != null)
        {
            listStore.set(ownerSM, index, element);
        }
        return delegate.set(index, element);
    }
 
    /**
     * The writeReplace method is called when ObjectOutputStream is preparing
     * to write the object to the stream. The ObjectOutputStream checks whether
     * the class defines the writeReplace method. If the method is defined, the
     * writeReplace method is called to allow the object to designate its 
     * replacement in the stream. The object returned should be either of the
     * same type as the object passed in or an object that when read and
     * resolved will result in an object of a type that is compatible with all
     * references to the object.
     * 
     * @return the replaced object
     * @throws ObjectStreamException
     */
    protected Object writeReplace() throws ObjectStreamException
    {
        return new java.util.ArrayList(this);
    }     

    /**
     * Method to load all elements from the "backing store" where appropriate.
     */
    protected void loadFromStore()
    {
        if (listStore != null && !isCacheLoaded)
        {
            delegate.clear();
            Iterator iter=listStore.iterator(ownerSM);
            while (iter.hasNext())
            {
                delegate.add(iter.next());
            }

            isCacheLoaded = true;
        }
    }
}
