/**********************************************************************
Copyright (c) 2002 Mike Martin and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
2003 Erik Bengtson  - removed ununsed imports
2003 Andy Jefferson - added missing methods
2003 Andy Jefferson - added Queryable implementation
2004 Andy Jefferson - rewritten to always have delegate populated
2004 Andy Jefferson - changed to allow caching
    ...
**********************************************************************/
package org.jpox.sco;

import java.io.ObjectStreamException;
import java.util.AbstractMap;
import java.util.Collection;
import java.util.HashMap;
import java.util.Iterator;

import javax.jdo.spi.PersistenceCapable;

import org.jpox.StateManager;
import org.jpox.sco.exceptions.IncompatibleFieldTypeException;
import org.jpox.sco.exceptions.IncompatibleKeyTypeException;
import org.jpox.sco.exceptions.IncompatibleValueTypeException;
import org.jpox.sco.exceptions.NullsNotAllowedException;
import org.jpox.sco.exceptions.QueryUnownedSCOException;
import org.jpox.store.DatastoreClass;
import org.jpox.store.StoreManager;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.query.Query;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.query.Queryable;
import org.jpox.store.scostore.MapStore;
import org.jpox.store.scostore.SetStore;
import org.jpox.util.ClassUtils;

/**
 * A mutable second-class Map object. 
 * Uses a "delegate" as a local store for the Collection. Uses a "backing store"
 * (SetStore) to represent the datastore. The "delegate" is updated with the
 * "backing store" information at necessary intervals. 
 *
 * <H3>Modes of Operation</H3>
 * The user can operate the collection in 2 modes.
 * The <B>cached</B> mode will use an internal cache of the elements (in the
 * "delegate") reading them at the first opportunity and then using the cache
 * thereafter.
 * The <B>non-cached</B> mode will just go direct to the "backing store" each
 * call.
 *
 * <H3>Mutators</H3>
 * When the backing store is present any updates are passed direct to the
 * datastore as well as to the "delegate". If the "backing store" isn't present
 * the changes are made to the "delegate" only.
 *
 * <H3>Accessors</H3>
 * When any accessor method is invoked, it typically checks whether the
 * collection has been loaded from its backing store and does this as
 * necessary. Some methods (<B>size()</B>, <B>containsKey()</B>) just check if
 * everything is loaded and use the delegate if possible, otherwise going
 * direct to the datastore.
 *
 * @version $Revision: 1.6 $
 */
public class Map extends AbstractMap
    implements java.util.Map, SCO, Cloneable, Queryable, java.io.Serializable, SCOMap
{
    private Object owner;
    private StateManager ownerSM;
    private String fieldName;
    private int fieldNumber;
    private Class keyType;
    private Class valueType;
    private boolean allowNulls;

    /**
     * The "backing store"
     */
    protected MapStore mapStore;

    /**
     * The internal "delegate".
     */
    protected HashMap delegate;

    /**
     * Whether to use "delegate" caching.
     */
    protected boolean useCache=true;

    /**
     * Status flag whether the collection is loaded into the cache.
     */
    protected boolean isCacheLoaded=false;

    /**
     * Default Constructor.
     */
    public Map()
    {
    }
    
    /**
     * Constructor, using the StateManager of the "owner" and the field name.
     * @param ownerSM The owner StateManager
     * @param fieldName The name of the field of the SCO.
     */
    public Map(StateManager ownerSM, String fieldName)
    {
        this(ownerSM, fieldName, null, null, true);
    }

    /**
     * Constructor, using the StateManager of the "owner", the field name, 
     * the type of the key/value and whether null values are allowed.
     * @param ownerSM The owner StateManager
     * @param fieldName The name of the field of the SCO.
     * @param keyType The type of the key objects.
     * @param valueType The type of the value objects.
     * @param allowNulls Whether to allow null elements.
     */
    public Map(StateManager ownerSM,
               String fieldName,
               Class keyType,
               Class valueType,
               boolean allowNulls)
    {
        Class c = ownerSM.getObject().getClass();
        StoreManager storeMgr = ownerSM.getStoreManager();
        DatastoreClass ownerTable = storeMgr.getDatastoreClass(c.getName());
        JavaTypeMapping m = ownerTable.getFieldMapping(fieldName);
        if (!(m instanceof org.jpox.store.mapping.MapMapping))
        {
            throw new IncompatibleFieldTypeException(ownerSM, fieldName, java.util.Map.class.getName(), ownerSM.getMetaDataManager().getMetaDataForField(c,fieldName).getTypeName());
        }

        init(ownerSM, fieldName,((org.jpox.store.mapping.MapMapping)m).getMapStore(),allowNulls);

        if (keyType != null &&
            !this.keyType.isAssignableFrom(keyType))
        {
            throw new IncompatibleKeyTypeException(ownerSM, fieldName, this.keyType.getName(), keyType.getName());
        }

        if (valueType != null &&
            !this.valueType.isAssignableFrom(valueType))
        {
            throw new IncompatibleValueTypeException(ownerSM, fieldName, this.valueType.getName(), valueType.getName());
        }
    }

    /**
     * Initialisation method.
     * @param ownerSM The owner StateManager
     * @param fieldName The name of the field of the SCO.
     * @param mapStore The backing store for this SCO.
     * @param allowNulls Whether to allow null elements.
     */
    private void init(StateManager ownerSM, 
                      String fieldName,
                      MapStore mapStore,
                      boolean allowNulls)
    {
        this.owner = ownerSM.getObject();
        this.ownerSM = ownerSM;
        this.fieldName = fieldName;
        this.fieldNumber = ownerSM.getMetaDataManager().getMetaDataForClass(owner.getClass()).getFieldNumberAbsolute(fieldName);
        this.mapStore = mapStore;
        this.keyType = ownerSM.getPersistenceManager().getClassLoaderResolver().classForName(mapStore.getKeyType());
        this.valueType = ownerSM.getPersistenceManager().getClassLoaderResolver().classForName(mapStore.getValueType());
        this.allowNulls = allowNulls;

        // Set caching policy
        useCache = SCOUtils.useContainerCache(ownerSM, fieldName); 

        // Set up our "delegate"
        this.delegate = new HashMap();

        // If we aren't caching, just load up what we have now.
        if (!useCache)
        {
            SetStore keystore=mapStore.keySetStore();
            Iterator iter=keystore.iterator(ownerSM);
            while (iter.hasNext())
            {
                Object key=iter.next();
                delegate.put(key,mapStore.get(ownerSM,key));
            }
        }
    }

    // ------------------------ Implementation of SCO methods ------------------

    /**
     * Mutator for setting using an object.
     * @param o  The object to set from
     */
    public synchronized void setValueFrom(Object o)
    {
        java.util.Map m = (java.util.Map)o;

        clear();
        putAll(m);
    }

    /**
     * Accessor for the field name.
     * @return The field name.
     **/
    public String getFieldName()
    {
        return fieldName;
    }

    /**
     * Accessor for the owner.
     * @return The owner.
     **/
    public Object getOwner()
    {
        return owner;
    }

    /**
     * Method to unset the owner and field information.
     */
    public synchronized void unsetOwner()
    {
        if (ownerSM != null)
        {
            owner = null;
            ownerSM = null;
            fieldName = null;
            mapStore = null;
        }
    }

    /**
     * Utility to apply updates
     **/
    public void applyUpdates()
    {
        /*
         * We are never dirty, since all changes are immediately delegated to
         * the underlying MapStore.
         */
        return;
    }

    /**
     * Utility to mark the object as dirty
     **/
    public void makeDirty()
    {
        /*
         * Although we are never really "dirty", the owning object must be
         * marked dirty so that the proper state change occurs and its
         * jdoPreStore() gets called properly.
         */
        if (ownerSM != null)
        {
            ownerSM.makeDirty(fieldNumber);
        }
    }

    public Object detachCopy(StateManager sm, Object value)
    {
        java.util.Map c = (java.util.Map)value;

        if (value == null)
        {
            // accept null collections...
            return new java.util.HashSet();
        }

        java.util.Map detached = new java.util.HashMap();
        for (Iterator it = c.entrySet().iterator(); it.hasNext();)
        {
            Map.Entry entry = (Map.Entry) it.next();
            Object val = entry.getValue();
            Object key = entry.getKey();
            if (val instanceof PersistenceCapable)
            {
                if (val instanceof PersistenceCapable)
                {
                    val = sm.getPersistenceManager().detachCopy(val);
                }
            }
            if (key instanceof PersistenceCapable)
            {
                if (key instanceof PersistenceCapable)
                {
                    key = sm.getPersistenceManager().detachCopy(key);
                }
            }
            detached.put(key, val);
        }
        return detached;
    }

    public Object attachCopy(StateManager sm, Object value, boolean makeTransactional)
    {
        java.util.Map c = (java.util.Map) value;

        if (value == null)
        {
            return null;
        }

        java.util.Map attached = new java.util.HashMap();
        for (Iterator it = c.entrySet().iterator(); it.hasNext();)
        {
            Map.Entry entry = (Map.Entry) it.next();
            Object val = entry.getValue();
            Object key = entry.getKey();
            if (val instanceof PersistenceCapable)
            {
                if (ClassUtils.isPersistenceCapable(val))
                {
                    val = sm.getPersistenceManager().attachCopy(val, makeTransactional);
                }
            }
            if (key instanceof PersistenceCapable)
            {
                if (ClassUtils.isPersistenceCapable(key))
                {
                    key = sm.getPersistenceManager().attachCopy(key, makeTransactional);
                }
            }
            attached.put(key, val);
        }
        return attached;
    }
    
    // ------------------------- Queryable Methods -----------------------------

    /**
     * Method to return a new Query statement for the Map.
     * @return The Query Statement.
     **/
    public synchronized QueryStatement newQueryStatement()
    {
        return newQueryStatement(valueType);
    }

    /**
     * Method to return a new Query statement for the Map given the candidate
     * class.
     * @param candidate_class the value class.
     * @return The Query Statement
     **/
    public synchronized QueryStatement newQueryStatement(Class candidate_class)
    {
        if (mapStore == null)
        {
            throw new QueryUnownedSCOException();
        }

        return mapStore.newQueryStatement(ownerSM, candidate_class.getName());
    }

    /**
     * Method to return a new result object factory for processing of Query
     * statements.
     * @param stmt The Query Statement.
     * @param ignoreCache Whether to ignore the cache.
     * @param resultClass Whether to create objects of a particular class
     * @return The result object factory.
     **/
    public synchronized Query.ResultObjectFactory newResultObjectFactory(
                                        QueryStatement stmt,
                                        boolean ignoreCache,
                                        Class resultClass)
    {
        if (mapStore == null)
        {
            throw new QueryUnownedSCOException();
        }

        return mapStore.newResultObjectFactory(ownerSM, stmt,ignoreCache);
    }

    // -------------------- Implementation of Map Methods ----------------------
 
    /**
     * Creates and returns a copy of this object.
     *
     * <P>Mutable second-class Objects are required to provide a public
     * clone method in order to allow for copying PersistenceCapable
     * objects. In contrast to Object.clone(), this method must not throw a
     * CloneNotSupportedException.
     * @return Clone of the object
     */
    public Object clone()
    {
        Object obj = null;

        try
        {
            obj = super.clone();
            ((Map)obj).unsetOwner();
        }
        catch (CloneNotSupportedException e)
        {
            // Can't happen
        }

        return obj;
    }

    /**
     * Utility to check if a key is contained in the Map.
     * @param key The key to check
     * @return Whether it is contained
     **/
    public synchronized boolean containsKey(Object key)
    {
        if (useCache && isCacheLoaded)
        {
            // If the "delegate" is already loaded, use it
            return delegate.containsKey(key);
        }
        else if (mapStore != null)
        {
            return mapStore.containsKey(ownerSM, key);
        }

        return delegate.containsKey(key);
    }

    /**
     * Utility to check if a value is contained in the Map.
     * @param value The value to check
     * @return Whether it is contained
     **/
    public synchronized boolean containsValue(Object value)
    {
        if (useCache && isCacheLoaded)
        {
            // If the "delegate" is already loaded, use it
            return delegate.containsValue(value);
        }
        else if (mapStore != null)
        {
            return mapStore.containsValue(ownerSM, value);
        }

        return delegate.containsValue(value);
    }

    /**
     * Accessor for the set of entries in the Map.
     * @return Set of entries
     **/
    public synchronized java.util.Set entrySet()
    {
        if (useCache)
        {
            loadFromStore();
        }
        else if (mapStore != null)
        {
            return new Set(ownerSM, fieldName, false, mapStore.entrySetStore());
        }

        return delegate.entrySet();
    }
 
    /**
     * Method to check the equality of this map, and another.
     * @param o The map to compare against.
     * @return Whether they are equal.
     **/
    public synchronized boolean equals(Object o)
    {
        if (useCache)
        {
            loadFromStore();
        }

        if (o == this)
        {
            return true;
        }
        if (!(o instanceof java.util.Map))
        {
            return false;
        }
        java.util.Map m = (java.util.Map)o;

        return entrySet().equals(m.entrySet());
    }

    /**
     * Accessor for the value stored against a key.
     * @param key The key
     * @return The value.
     **/
    public synchronized Object get(Object key)
    {
        if (useCache)
        {
            loadFromStore();
        }
        else if (mapStore != null)
        {
            return mapStore.get(ownerSM, key);
        }

        return delegate.get(key);
    }

    /**
     * Method to generate a hashcode for this Map.
     * @return The hashcode.
     **/
    public synchronized int hashCode()
    {
        if (useCache)
        {
            loadFromStore();
        }
        else if (mapStore != null)
        {
            int h = 0;
            Iterator i = entrySet().iterator();
            while (i.hasNext())
            {
                h += i.next().hashCode();
            }

            return h;
        }
        return delegate.hashCode();
    }

    /**
     * Method to return if the Map is empty.
     * @return Whether it is empty.
     **/
    public synchronized boolean isEmpty()
    {
        return size() == 0;
    }

    /**
     * Accessor for the set of keys in the Map.
     * @return Set of keys.
     **/
    public synchronized java.util.Set keySet()
    {
        if (useCache)
        {
            loadFromStore();
        }
        else if (mapStore != null)
        {
            return new Set(ownerSM, fieldName, false, mapStore.keySetStore());
        }

        return delegate.keySet();
    }

    /**
     * Method to return the size of the Map.
     * @return The size
     **/
    public synchronized int size()
    {
        if (useCache && isCacheLoaded)
        {
            // If the "delegate" is already loaded, use it
            return delegate.size();
        }
        else if (mapStore != null)
        {
            return mapStore.entrySetStore().size(ownerSM);
        }

        return delegate.size();
    }

    /**
     * Accessor for the set of values in the Map.
     * @return Set of values.
     **/
    public synchronized Collection values()
    {
        if (useCache)
        {
            loadFromStore();
        }
        else if (mapStore != null)
        {
            return new Set(ownerSM, fieldName, true, mapStore.valueSetStore());
        }

        return delegate.values();
    }

    /**
     * Method to return a string form of this Map.
     * @return String form of this Map.
     **/
    public String toString()
    {
        StringBuffer s = new StringBuffer("{");

        Iterator i = entrySet().iterator();
        boolean hasNext = i.hasNext();

        while (hasNext)
        {
            Entry e = (Entry) i.next();
            Object key = e.getKey();
            Object val = e.getValue();
            s.append(key).append('=').append(val);
            hasNext = i.hasNext();

            if (hasNext)
            {
                s.append(',');
            }
        }

        s.append("}");

        return s.toString();
    }

    // -------------------------- Mutator methods ------------------------------
 
    /**
     * Method to clear the Map.
     **/
    public synchronized void clear()
    {
        makeDirty();

        if (mapStore != null)
        {
            mapStore.clear(ownerSM);
        }
        delegate.clear();
    }

    /**
     * Method to add a value to the Map.
     * @param key The key for the value.
     * @param value The value
     * @return The previous value against this key (if any).
     **/
    public synchronized Object put(Object key, Object value)
    {
        // Reject inappropriate elements
        if (value == null && !allowNulls)
        {
            throw new NullsNotAllowedException(ownerSM, fieldName);
        }

        makeDirty();

        if (useCache)
        {
            loadFromStore();
        }

        if (mapStore != null)
        {
            mapStore.put(ownerSM, key, value);
        }
        return delegate.put(key, value);
    }

    /**
     * Method to add a Map of values to this map.
     * @param m The Map to add
     **/
    public synchronized void putAll(java.util.Map m)
    {
        makeDirty();

        if (useCache)
        {
            loadFromStore();
        }

        if (mapStore != null)
        {
            mapStore.putAll(ownerSM, m);
        }
        delegate.putAll(m);
    }

    /**
     * Method to remove a value from the Map.
     * @param key The key for the value.
     * @return The value removed.
     **/
    public synchronized Object remove(Object key)
    {
        makeDirty();

        if (useCache)
        {
            loadFromStore();
        }

        if (mapStore != null)
        {
            mapStore.remove(ownerSM, key);
        }
        return delegate.remove(key);
    }

    /**
     * The writeReplace method is called when ObjectOutputStream is preparing
     * to write the object to the stream. The ObjectOutputStream checks whether
     * the class defines the writeReplace method. If the method is defined, the
     * writeReplace method is called to allow the object to designate its
     * replacement in the stream. The object returned should be either of the
     * same type as the object passed in or an object that when read and
     * resolved will result in an object of a type that is compatible with all
     * references to the object.
     * 
     * @return the replaced object
     * @throws ObjectStreamException
     */
    protected Object writeReplace() throws ObjectStreamException
    {
        return new java.util.HashMap(this);
    }

    /**
     * Method to load all elements from the "backing store" where appropriate.
     */
    protected void loadFromStore()
    {
        if (mapStore != null && !isCacheLoaded)
        {
            delegate.clear();

            SetStore keystore=mapStore.keySetStore();
            Iterator iter=keystore.iterator(ownerSM);
            while (iter.hasNext())
            {
                Object key=iter.next();
                delegate.put(key,mapStore.get(ownerSM,key));
            }

            isCacheLoaded = true;
        }
    }
}
