/**********************************************************************
Copyright (c) 2003 Mike Martin and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
2004 Andy Jefferson - rewritten to always have delegate present
    ...
**********************************************************************/
package org.jpox.sco;

import java.io.ObjectStreamException;
import java.util.Collection;

import org.jpox.StateManager;
import org.jpox.sco.exceptions.IncompatibleElementTypeException;
import org.jpox.sco.exceptions.IncompatibleFieldTypeException;
import org.jpox.store.DatastoreClass;
import org.jpox.store.StoreManager;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.scostore.SetStore;

/**
 * A mutable second-class Set object.
 * Extends Collection, so please refer to the description there for its
 * behaviour and performance.
 *
 * @version $Revision: 1.3 $
 */
public class Set extends org.jpox.sco.Collection
    implements java.util.Set, java.io.Serializable
{
    /**
     * Default Constructor.
     */
    public Set()
    {
    }
    
    /**
     * Constructor.
     * @param ownerSM The State Manager for this collection.
     * @param fieldName Name of the field
     **/
    public Set(StateManager ownerSM, String fieldName)
    {
        this(ownerSM, fieldName, null, false);
    }

    /**
     * Constructor.
     * @param ownerSM The State Manager for this collection.
     * @param fieldName Name of the field
     * @param elementType Type of the contained elements
     * @param allowNulls Whether to allow nulls 
     **/
    public Set(StateManager ownerSM,
               String fieldName,
               Class elementType,
               boolean allowNulls)
    {
        super(ownerSM,fieldName,elementType,allowNulls);
    }

    /**
     * Constructor. Called via Map SCO for handling entrySet.
     * @param ownerSM The State Manager for this collection.
     * @param fieldName Name of the field
     * @param allowNulls Whether to allow nulls
     * @param setStore The backing store
     **/
    protected Set(StateManager ownerSM,
                  String fieldName,
                  boolean allowNulls,
                  SetStore setStore)
    {
        super(ownerSM, fieldName, allowNulls, setStore);
    }

    /**
     * Second Initialisation method, overrides the one in Collection.
     **/
    protected void init2(StateManager ownerSM,
                         String fieldName,
                         Class elementType,
                         boolean allowNulls)
    {
        Class c = ownerSM.getObject().getClass();
        StoreManager storeMgr = ownerSM.getStoreManager();
        DatastoreClass ownerTable = storeMgr.getDatastoreClass(c.getName());
        JavaTypeMapping m = ownerTable.getFieldMapping(fieldName);
        if (!(m instanceof org.jpox.store.mapping.SetMapping))
        {
            throw new IncompatibleFieldTypeException(ownerSM, fieldName, Collection.class.getName(), ownerSM.getMetaDataManager().getMetaDataForField(c,fieldName).getTypeName());
        }
 
        init(ownerSM, fieldName, ((org.jpox.store.mapping.SetMapping)m).getSetStore(), allowNulls);
 
        if (elementType != null &&
            !this.elementType.isAssignableFrom(elementType))
        {
            throw new IncompatibleElementTypeException(ownerSM, fieldName, this.elementType.getName(), elementType.getName());
        }
    }

    /**
     * Equality operator.
     * @param o The object to compare with
     * @return Whether they are equals
     **/
    public synchronized boolean equals(Object o)
    {
        if (useCache)
        {
            loadFromStore();
        }

        if (o == this)
        {
            return true;
        }
        if (!(o instanceof java.util.Set))
        {
            return false;
        }
        java.util.Set c = (java.util.Set)o;

        return c.size() == size() && containsAll(c);
    }
 
    /**
     * The writeReplace method is called when ObjectOutputStream is preparing
     * to write the object to the stream. The ObjectOutputStream checks
     * whether the class defines the writeReplace method. If the method is
     * defined, the writeReplace method is called to allow the object to
     * designate its replacement in the stream. The object returned should be
     * either of the same type as the object passed in or an object that when
     * read and resolved will result in an object of a type that is compatible
     * with all references to the object.
     *
     * @return the replaced object
     * @throws ObjectStreamException
     */
    protected Object writeReplace() throws ObjectStreamException
    {
        return new java.util.HashSet(this);
    }
}
