/**********************************************************************
Copyright (c) 2002 Mike Martin and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
    ...
**********************************************************************/
package org.jpox.sco;

import java.io.ObjectStreamException;

import javax.jdo.JDOHelper;
import javax.jdo.spi.PersistenceCapable;

import org.jpox.StateManager;

/**
 * A mutable second-class SQL date object.
 * 
 * @version $Revision: 1.1 $
 */
public class SqlDate extends java.sql.Date implements SCO
{
    private transient PersistenceCapable owner;
    private transient String fieldName;

    /**
     * Default Constructor.
     */
    public SqlDate()
    {
        super(0);        
    }
    
    /**
     * Creates a <tt>SqlDate</tt> object that represents the time at which
     * it was allocated. Assigns owning object and field name.
     *
     * @param owner the owning object
     * @param fieldName the owning field name
     */
    public SqlDate(Object owner, String fieldName)
    {
        super(0);

        if (owner instanceof PersistenceCapable)
        {
            this.owner = (PersistenceCapable)owner;
            this.fieldName = fieldName;
        }
    }

    /**
     * Creates a <tt>SqlDate</tt> object that represents the same time as the
     * given <tt>java.sql.Date</tt>. Assigns owning object and field name.
     *
     * @param owner     the owning object
     * @param fieldName the owning field name
     * @param date      the initial date value
     */
    public SqlDate(Object owner, String fieldName, java.sql.Date date)
    {
        this(owner, fieldName);

        setValueFrom(date);
    }

    /**
     * Creates and returns a copy of this object.
     *
     * <P>Mutable second-class Objects are required to provide a public
     * clone method in order to allow for copying PersistenceCapable
     * objects. In contrast to Object.clone(), this method must not throw a
     * CloneNotSupportedException.
     * @return Clone of the object
     */
    public Object clone()
    {
        Object obj = super.clone();

        ((SqlDate)obj).unsetOwner();

        return obj;
    }

    /**
     * Mutator for the time.
     * @param time The time (millisecs)
     **/
    public void setTime(long time)
    {
        super.setTime(time);
        makeDirty();
    }

    /**
     * Sets the year of this <tt>Date</tt> object to be the specified 
     * value plus 1900. This <code>Date</code> object is modified so 
     * that it represents a point in time within the specified year, 
     * with the month, date, hour, minute, and second the same as 
     * before, as interpreted in the local time zone. (Of course, if 
     * the date was February 29, for example, and the year is set to a 
     * non-leap year, then the new date will be treated as if it were 
     * on March 1.)
     *
     * @param   year    the year value.
     * @see     java.util.Calendar
     * @deprecated As of JDK version 1.1,
     * replaced by <code>Calendar.set(Calendar.YEAR, year + 1900)</code>.
     */
    public void setYear(int year)
    {
        super.setYear(year);
        makeDirty();
    }

    /**
     * Sets the month of this date to the specified value. This 
     * <tt>Date</tt> object is modified so that it represents a point 
     * in time within the specified month, with the year, date, hour, 
     * minute, and second the same as before, as interpreted in the 
     * local time zone. If the date was October 31, for example, and 
     * the month is set to June, then the new date will be treated as 
     * if it were on July 1, because June has only 30 days.
     *
     * @param   month   the month value between 0-11.
     * @see     java.util.Calendar
     * @deprecated As of JDK version 1.1,
     * replaced by <code>Calendar.set(Calendar.MONTH, int month)</code>.
     */
    public void setMonth(int month)
    {
        super.setMonth(month);
        makeDirty();
    }

    /**
     * Sets the day of the month of this <tt>Date</tt> object to the 
     * specified value. This <tt>Date</tt> object is modified so that 
     * it represents a point in time within the specified day of the 
     * month, with the year, month, hour, minute, and second the same 
     * as before, as interpreted in the local time zone. If the date 
     * was April 30, for example, and the date is set to 31, then it 
     * will be treated as if it were on May 1, because April has only 
     * 30 days.
     *
     * @param   date   the day of the month value between 1-31.
     * @see     java.util.Calendar
     * @deprecated As of JDK version 1.1,
     * replaced by <code>Calendar.set(Calendar.DAY_OF_MONTH, int date)</code>.
     */
    public void setDate(int date)
    {
        super.setDate(date);
        makeDirty();
    }

    /**
     * Mutator for the value, using an object.
     * @param o The Object
     **/
    public void setValueFrom(Object o)
    {
        super.setTime(((java.sql.Date)o).getTime());
    }

    /**
     * Utility to unset the owner.
     **/
    public void unsetOwner()
    {
        owner = null;
        fieldName = null;
    }

    /**
     * Accessor for the owner.
     * @return The owner 
     **/
    public Object getOwner()
    {
        return owner;
    }

    /**
     * Accessor for the field name
     * @return The field name
     **/
    public String getFieldName()
    {
        return fieldName;
    }

    /**
     * Utility to apply updates
     **/
    public void applyUpdates()
    {
        return;
    }

    /**
     * Utility to mark the object as dirty
     **/
    public void makeDirty()
    {
        if (owner != null)
        {
            JDOHelper.makeDirty(owner, fieldName);
        }
    }
    
    public Object detachCopy(StateManager sm, Object value)
    {
        if (value == null)
        {
			return null;
        }
        
        return new java.sql.Date(((java.sql.Date)value).getTime());
    }
    
    public Object attachCopy(StateManager sm, Object value, boolean makeTransactional)
    {
        if (value == null)
        {
			return null;
        }
        return new java.sql.Date(((java.sql.Date)value).getTime());
    }    
    
	/**
	 * The writeReplace method is called when ObjectOutputStream is preparing to write the object to the stream. The
	 * ObjectOutputStream checks whether the class defines the writeReplace method. If the method is defined, the
	 * writeReplace method is called to allow the object to designate its replacement in the stream. The object returned
	 * should be either of the same type as the object passed in or an object that when read and resolved will result in
	 * an object of a type that is compatible with all references to the object.
	 * 
	 * @return the replaced object
	 * @throws ObjectStreamException
	 */
	protected Object writeReplace() throws ObjectStreamException
	{
		return new java.sql.Date(this.getTime());
	}     
}
