/**********************************************************************
Copyright (c) 2003 Andy Jefferson and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
    ...
**********************************************************************/
package org.jpox.sco;

import java.io.ObjectStreamException;

import javax.jdo.JDOHelper;
import javax.jdo.spi.PersistenceCapable;

import org.jpox.StateManager;

/**
 * A mutable second-class SQLTime object.
 * 
 * @version $Revision: 1.1 $
 */
public class SqlTime extends java.sql.Time implements SCO
{
    private transient PersistenceCapable owner;
    private transient String fieldName;

    /**
     * Default Constructor.
     */
    public SqlTime()
    {
        super(0);        
    }
    
    /**
     * Creates a <tt>SqlTime</tt> object that represents the time at which
     * it was allocated. Assigns owning object and field name.
     *
     * @param owner the owning object
     * @param fieldName the owning field name
     */
    public SqlTime(Object owner, String fieldName)
    {
        super(0);

        if (owner instanceof PersistenceCapable)
        {
            this.owner = (PersistenceCapable)owner;
            this.fieldName = fieldName;
        }
    }

    /**
     * Creates a <tt>SqlTime</tt> object that represents the same time as the
     * given <tt>java.sql.Time</tt>. Assigns owning object and field name.
     *
     * @param owner     the owning object
     * @param fieldName the owning field name
     * @param time      the initial time value
     */
    public SqlTime(Object owner, String fieldName, java.sql.Time time)
    {
        this(owner, fieldName);

        setValueFrom(time);
    }

    /**
     * Creates and returns a copy of this object.
     *
     * <P>Mutable second-class Objects are required to provide a public
     * clone method in order to allow for copying PersistenceCapable
     * objects. In contrast to Object.clone(), this method must not throw a
     * CloneNotSupportedException.
     * @return Clone of the object
     */
    public Object clone()
    {
        Object obj = super.clone();

        ((SqlTime)obj).unsetOwner();

        return obj;
    }

    /**
     * Sets the time of this <tt>Time</tt> object to the specified value.
     * This <tt>Time</tt> object is modified so that it represents a point in
     * time with the hour, minute, second as specified.
     *
     * @param   time   millisecs since 1 Jan 1970, 00:00:00 GMT
     * @see     java.util.Calendar
     */
    public void setTime(long time)
    {
        super.setTime(time);
        makeDirty();
    }

    /**
     * Mutator for the value, using an object.
     * @param o The Object
     **/
    public void setValueFrom(Object o)
    {
        super.setTime(((java.sql.Time)o).getTime());
    }

    /**
     * Utility to unset the owner.
     **/
    public void unsetOwner()
    {
        owner = null;
        fieldName = null;
    }

    /**
     * Accessor for the owner.
     * @return The owner 
     **/
    public Object getOwner()
    {
        return owner;
    }

    /**
     * Accessor for the field name
     * @return The field name
     **/
    public String getFieldName()
    {
        return fieldName;
    }

    /**
     * Utility to apply updates
     **/
    public void applyUpdates()
    {
        return;
    }

    /**
     * Utility to mark the object as dirty
     **/
    public void makeDirty()
    {
        if (owner != null)
        {
            JDOHelper.makeDirty(owner, fieldName);
        }
    }
    
    public Object detachCopy(StateManager sm, Object value)
    {
        if (value == null)
        {
			return null;
        }
        
        return new java.sql.Time(((java.sql.Time)value).getTime());
    }
    
    public Object attachCopy(StateManager sm, Object value, boolean makeTransactional)
    {
        if (value == null)
        {
			return null;
        }
        return new java.sql.Time(((java.sql.Time)value).getTime());
    }
    
	/**
	 * The writeReplace method is called when ObjectOutputStream is preparing to write the object to the stream. The
	 * ObjectOutputStream checks whether the class defines the writeReplace method. If the method is defined, the
	 * writeReplace method is called to allow the object to designate its replacement in the stream. The object returned
	 * should be either of the same type as the object passed in or an object that when read and resolved will result in
	 * an object of a type that is compatible with all references to the object.
	 * 
	 * @return the replaced object
	 * @throws ObjectStreamException
	 */
	protected Object writeReplace() throws ObjectStreamException
	{
		return new java.sql.Time(this.getTime());
	}     
}
