/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2005 Andy Jefferson - cater for object not being persistent
    ...
**********************************************************************/
package org.jpox.state;

import javax.jdo.spi.PersistenceCapable;

import org.jpox.StateManager;
import org.jpox.sco.SCO;
import org.jpox.store.fieldmanager.AbstractFieldManager;

/**
 * Utility class to handle the attach of fields.
 * The attachment process has 2 distinct cases to cater for.
 * <OL>
 * <LI>The object was detached, has been updated, and needs reattaching.</LI>
 * <LI>The object was detached from a different datastore, and is being attached here.</LI>
 * </OL>
 * In the first case, the fields which are dirty have their values (and dirty flags) updated.
 * In the second case, all fields have their fields (and dirty flags) updated.
 * 
 * @version $Revision: 1.4 $
 */
public class AttachFieldManager extends AbstractFieldManager
{
    private final StateManager smSource;

    private final boolean makeTransactional;

    private final boolean[] secondClassMutableFields;

    private final boolean dirtyFields[];

    private final boolean persistent;

    public AttachFieldManager(StateManager smSource, 
                              boolean secondClassMutableFields[], 
                              boolean dirtyFields[], 
                              boolean makeTransactional, 
                              boolean persistent)
    {
        this.smSource = smSource;
        this.makeTransactional = makeTransactional;
        this.secondClassMutableFields = secondClassMutableFields;
        this.dirtyFields = dirtyFields;
        this.persistent = persistent;
    }

    public void storeObjectField(int fieldNumber, Object value)
    {
        if (value == null)
        {
            if (dirtyFields[fieldNumber] || !persistent)
            {
                smSource.makeDirty(fieldNumber);
                smSource.replaceField(fieldNumber, null);
            }
        }
        else if (secondClassMutableFields[fieldNumber])
        {
            if (dirtyFields[fieldNumber] || !persistent)
            {
                smSource.makeDirty(fieldNumber);
            }
            smSource.replaceField(fieldNumber, ((SCO) smSource.newSCOEmptyInstance(fieldNumber)).attachCopy(smSource, value, makeTransactional));
        }
        else if (value instanceof PersistenceCapable)
        {
            PersistenceCapable pc = (PersistenceCapable) value;
            if (dirtyFields[fieldNumber] || !persistent)
            {
                smSource.makeDirty(fieldNumber);
            }
            Object pcObj = smSource.getPersistenceManager().attachCopy(pc, makeTransactional);

            // Reinstate the dirty field since attach of the other end of an inverse may reset the flag
            if (dirtyFields[fieldNumber] || !persistent)
            {
                smSource.makeDirty(fieldNumber);
            }
            smSource.replaceField(fieldNumber, pcObj);
        }
        else
        {
            if (dirtyFields[fieldNumber] || !persistent)
            {
                smSource.makeDirty(fieldNumber);
                smSource.replaceField(fieldNumber, value);
            }
        }
    }

    /*
     * (non-Javadoc)
     * @see javax.jdo.spi.PersistenceCapable.ObjectIdFieldConsumer#storeBooleanField(int,
     * boolean)
     */
    public void storeBooleanField(int fieldNumber, boolean value)
    {
        if (dirtyFields[fieldNumber] || !persistent)
        {
            smSource.makeDirty(fieldNumber);
            SingleFieldValue sfv = new SingleFieldValue();
            sfv.storeBooleanField(fieldNumber, value);
            smSource.replaceFields(new int[]{fieldNumber}, sfv);
        }
    }

    /*
     * (non-Javadoc)
     * @see javax.jdo.spi.PersistenceCapable.ObjectIdFieldConsumer#storeByteField(int,
     * byte)
     */
    public void storeByteField(int fieldNumber, byte value)
    {
        if (dirtyFields[fieldNumber] || !persistent)
        {
            smSource.makeDirty(fieldNumber);
            SingleFieldValue sfv = new SingleFieldValue();
            sfv.storeByteField(fieldNumber, value);
            smSource.replaceFields(new int[]{fieldNumber}, sfv);
        }
    }

    /*
     * (non-Javadoc)
     * @see javax.jdo.spi.PersistenceCapable.ObjectIdFieldConsumer#storeCharField(int,
     * char)
     */
    public void storeCharField(int fieldNumber, char value)
    {
        if (dirtyFields[fieldNumber] || !persistent)
        {
            smSource.makeDirty(fieldNumber);
            SingleFieldValue sfv = new SingleFieldValue();
            sfv.storeCharField(fieldNumber, value);
            smSource.replaceFields(new int[]{fieldNumber}, sfv);
        }
    }

    /*
     * (non-Javadoc)
     * @see javax.jdo.spi.PersistenceCapable.ObjectIdFieldConsumer#storeDoubleField(int,
     * double)
     */
    public void storeDoubleField(int fieldNumber, double value)
    {
        if (dirtyFields[fieldNumber] || !persistent)
        {
            smSource.makeDirty(fieldNumber);
            SingleFieldValue sfv = new SingleFieldValue();
            sfv.storeDoubleField(fieldNumber, value);
            smSource.replaceFields(new int[]{fieldNumber}, sfv);
        }
    }

    /*
     * (non-Javadoc)
     * @see javax.jdo.spi.PersistenceCapable.ObjectIdFieldConsumer#storeFloatField(int,
     * float)
     */
    public void storeFloatField(int fieldNumber, float value)
    {
        if (dirtyFields[fieldNumber] || !persistent)
        {
            smSource.makeDirty(fieldNumber);
            SingleFieldValue sfv = new SingleFieldValue();
            sfv.storeFloatField(fieldNumber, value);
            smSource.replaceFields(new int[]{fieldNumber}, sfv);
        }
    }

    /*
     * (non-Javadoc)
     * @see javax.jdo.spi.PersistenceCapable.ObjectIdFieldConsumer#storeIntField(int,
     * int)
     */
    public void storeIntField(int fieldNumber, int value)
    {
        if (dirtyFields[fieldNumber] || !persistent)
        {
            smSource.makeDirty(fieldNumber);
            SingleFieldValue sfv = new SingleFieldValue();
            sfv.storeIntField(fieldNumber, value);
            smSource.replaceFields(new int[]{fieldNumber}, sfv);
        }
    }

    /*
     * (non-Javadoc)
     * @see javax.jdo.spi.PersistenceCapable.ObjectIdFieldConsumer#storeLongField(int,
     * long)
     */
    public void storeLongField(int fieldNumber, long value)
    {
        if (dirtyFields[fieldNumber] || !persistent)
        {
            smSource.makeDirty(fieldNumber);
            SingleFieldValue sfv = new SingleFieldValue();
            sfv.storeLongField(fieldNumber, value);
            smSource.replaceFields(new int[]{fieldNumber}, sfv);
        }
    }

    /*
     * (non-Javadoc)
     * @see javax.jdo.spi.PersistenceCapable.ObjectIdFieldConsumer#storeShortField(int,
     * short)
     */
    public void storeShortField(int fieldNumber, short value)
    {
        if (dirtyFields[fieldNumber] || !persistent)
        {
            smSource.makeDirty(fieldNumber);
            SingleFieldValue sfv = new SingleFieldValue();
            sfv.storeShortField(fieldNumber, value);
            smSource.replaceFields(new int[]{fieldNumber}, sfv);
        }
    }

    /*
     * (non-Javadoc)
     * @see javax.jdo.spi.PersistenceCapable.ObjectIdFieldConsumer#storeStringField(int,
     * java.lang.String)
     */
    public void storeStringField(int fieldNumber, String value)
    {
        if (dirtyFields[fieldNumber] || !persistent)
        {
            smSource.makeDirty(fieldNumber);
            SingleFieldValue sfv = new SingleFieldValue();
            sfv.storeStringField(fieldNumber, value);
            smSource.replaceFields(new int[]{fieldNumber}, sfv);
        }
    }
}