/**********************************************************************
Copyright (c) 2005 Andy Jefferson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
    ...
**********************************************************************/
package org.jpox.state;

import java.util.Collection;
import java.util.Map;

import javax.jdo.spi.PersistenceCapable;

import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.store.fieldmanager.AbstractFieldManager;
import org.jpox.util.JPOXLogger;

/**
 * Utility class to handle the deletion of objects and the effect on (relationship) fields.
 * This operates in 2 modes.
 * <ol> 
 * <li>Run through the fields of an object being deleted and notify its PC
 * objects to removes all references to this.</li>
 * <li>Run through all fields of this object and null out or remove all references
 * to the object being deleted.</li>
 * </ol>
 * 
 * @version $Revision: 1.2 $
 */
public class DeleteFieldManager extends AbstractFieldManager
{
    private final StateManager smSource;
    private final PersistenceCapable pcDeleting;

    /**
     * Constructor to be called when needing to detect all PC fields that may
     * refer to this object, and notify them to removes their refs.
     * @param smSource The State Manager for this object
     */
    public DeleteFieldManager(StateManager smSource)
    {
        this.smSource = smSource;
        pcDeleting = null;
    }

    /**
     * Constructor to be called when needing to remove all references to the 
     * specified PersistenceCapable object since it is being deleted.
     * @param pc The PersistenceCapable to remove refs to
     */
    public DeleteFieldManager(StateManager smSource, PersistenceCapable pc)
    {
        this.smSource = smSource;
        pcDeleting = pc;
    }

    /*
     * (non-Javadoc)
     * @see javax.jdo.spi.PersistenceCapable.ObjectIdFieldConsumer#storeObjectField(int,
     * Object)
     */
    public void storeObjectField(int fieldNumber, Object value)
    {
        if (value == null)
        {
            // Do nothing
        }
        else if (value instanceof Collection)
        {
            if (pcDeleting != null)
            {
                // Remove the object from the Collection (if it is contained)
                Collection coll = (Collection)value;
                boolean removed = coll.remove(pcDeleting);
                if (JPOXLogger.JDO.isDebugEnabled() && removed)
                {
                    JPOXLogger.JDO.debug("Removed object with id " + pcDeleting.jdoGetObjectId() + " from collection " + value + " since it has been deleted");
                }
            }
        }
        else if (value instanceof Map)
        {
            if (pcDeleting != null)
            {
                // TODO Remove the object from the Map (if it is contained)
            }
        }
        else if (value instanceof PersistenceCapable)
        {
            PersistenceCapable pc = (PersistenceCapable)value;
            if (pcDeleting != null && value == pcDeleting)
            {
                // TODO Null out the link to our deleted object (1-1 bidirectional)
                if (JPOXLogger.JDO.isDebugEnabled())
                {
                    JPOXLogger.JDO.warn("Not Yet Implemented : Need to null out the reference to object with id " + pcDeleting.jdoGetObjectId() + " in object " + smSource.getObject().jdoGetObjectId());
                }
            }
            else
            {
                // Check if the PM of the related SM is still around
                if (pc.jdoGetPersistenceManager() == null)
                {
                    JPOXLogger.JDO.warn("Unable to notify object with id " + pc.jdoGetObjectId() + " since persistence manager is not set");
                    return;
                }

                // Notify the associated StateManager if any that all refs to this object need removing
                StateManagerImpl relatedSM = (StateManagerImpl)((PersistenceManager)pc.jdoGetPersistenceManager()).findStateManager(pc);
                if (relatedSM != null)
                {
                    relatedSM.removeReferencesToObject(smSource.getObject());
                }
            }
        }
    }

    /*
     * (non-Javadoc)
     * @see javax.jdo.spi.PersistenceCapable.ObjectIdFieldConsumer#storeBooleanField(int,
     * boolean)
     */
    public void storeBooleanField(int fieldNumber, boolean value)
    {
    }

    /*
     * (non-Javadoc)
     * @see javax.jdo.spi.PersistenceCapable.ObjectIdFieldConsumer#storeByteField(int,
     * byte)
     */
    public void storeByteField(int fieldNumber, byte value)
    {
    }

    /*
     * (non-Javadoc)
     * @see javax.jdo.spi.PersistenceCapable.ObjectIdFieldConsumer#storeCharField(int,
     * char)
     */
    public void storeCharField(int fieldNumber, char value)
    {
    }

    /*
     * (non-Javadoc)
     * @see javax.jdo.spi.PersistenceCapable.ObjectIdFieldConsumer#storeDoubleField(int,
     * double)
     */
    public void storeDoubleField(int fieldNumber, double value)
    {
    }

    /*
     * (non-Javadoc)
     * @see javax.jdo.spi.PersistenceCapable.ObjectIdFieldConsumer#storeFloatField(int,
     * float)
     */
    public void storeFloatField(int fieldNumber, float value)
    {
    }

    /*
     * (non-Javadoc)
     * @see javax.jdo.spi.PersistenceCapable.ObjectIdFieldConsumer#storeIntField(int,
     * int)
     */
    public void storeIntField(int fieldNumber, int value)
    {
    }

    /*
     * (non-Javadoc)
     * @see javax.jdo.spi.PersistenceCapable.ObjectIdFieldConsumer#storeLongField(int,
     * long)
     */
    public void storeLongField(int fieldNumber, long value)
    {
    }

    /*
     * (non-Javadoc)
     * @see javax.jdo.spi.PersistenceCapable.ObjectIdFieldConsumer#storeShortField(int,
     * short)
     */
    public void storeShortField(int fieldNumber, short value)
    {
    }

    /*
     * (non-Javadoc)
     * @see javax.jdo.spi.PersistenceCapable.ObjectIdFieldConsumer#storeStringField(int,
     * java.lang.String)
     */
    public void storeStringField(int fieldNumber, String value)
    {
    }
}