/**********************************************************************
Copyright (c) 2002 Kelly Grizzle (TJDO) and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - commented
2003 Andy Jefferson - added localiser and Logger
2005 Andy Jefferson - added Detached clean/dirty states
    ...
**********************************************************************/
package org.jpox.state;

import javax.jdo.Transaction;

import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;

/**
 * Base Class representing the life cycle state. Implemented for individual
 * states.
 *
 * @version $Revision: 1.3 $
 **/
public abstract class LifeCycleState
{
    protected static final Localiser LOCALISER=Localiser.getInstance("org.jpox.state.Localisation");

    public static final int TRANSIENT       = 0;
    public static final int P_NEW           = 1;
    public static final int P_CLEAN         = 2;
    public static final int P_DIRTY         = 3;
    public static final int HOLLOW          = 4;
    public static final int T_CLEAN         = 5;
    public static final int T_DIRTY         = 6;
    public static final int P_NEW_DELETED   = 7;
    public static final int P_DELETED       = 8;
    public static final int P_NONTRANS      = 9;
    public static final int DETACHED_CLEAN  = 10;
    public static final int DETACHED_DIRTY  = 11;

    public static final int TOTAL           = 12;
    public static final int ILLEGAL_STATE   = 12;

    // These are state flags that are set to required valuers in specific state 
    protected boolean isDirty;
    protected boolean isNew;
    protected boolean isDeleted;
    protected boolean isTransactional;
    protected boolean isPersistent;

    protected int stateType;

    private static LifeCycleState states[];

    static
    {
        states = new LifeCycleState[TOTAL];

        states[HOLLOW]         = new Hollow();
        states[P_CLEAN]        = new PersistentClean();
        states[P_DIRTY]        = new PersistentDirty();
        states[P_NEW]          = new PersistentNew();
        states[P_NEW_DELETED]  = new PersistentNewDeleted();
        states[P_DELETED]      = new PersistentDeleted();
        states[P_NONTRANS]     = new PersistentNontransactional();
        states[T_CLEAN]        = new TransientClean();
        states[T_DIRTY]        = new TransientDirty();
        states[DETACHED_CLEAN] = new DetachedClean();
        states[DETACHED_DIRTY] = new DetachedDirty();
        states[TRANSIENT]      = null;
    }

    /**
     * Returns the LifeCycleState for the state constant.
     * @param stateType the type as integer
     * @return the type as LifeCycleState object
     */
    public static LifeCycleState getLifeCycleState(int stateType)
    {
        return states[stateType];
    }

    /**
     * Returns the type of the life cycle state
     * @return the type of this life cycle state
     */
    public int stateType()
    {
        return stateType;
    }

    /**
     * Utility to change state to a new state.
     * @param sm The state manager.
     * @param newStateType The new state
     * @return new LifeCycle state.
     **/
    protected LifeCycleState changeState(StateManagerImpl sm, int newStateType)
    {
        LifeCycleState newState = states[newStateType];

        if (JPOXLogger.JDO.isDebugEnabled())
        {
            JPOXLogger.JDO_LIFECYCLE.debug("\"" + sm.getInternalObjectId().toString() + "\" : " + this + "->" + newState);
        }

        if (isTransactional)
        {
            if (newState == null || !newState.isTransactional)
            {
                sm.evictFromTransaction();
            }
        }
        else
        {
            if (newState != null && newState.isTransactional)
            {
                sm.enlistInTransaction();
            }
        }

        if (newState == null)
        {
            sm.disconnect();
        }

        return newState;
    }
    
    /**
     * Utility to change state to a new state.
     *
     * @param sm The state manager.
     * @param newStateType The new state
     * @return new LifeCycle state.
     **/
    protected LifeCycleState changeTransientState(StateManagerImpl sm, int newStateType)
    {
        LifeCycleState newState = states[newStateType];
        
        //TODO enlist in transaction better
        try 
        {
            sm.enlistInTransaction();
        }
        catch (Exception e)
        {
            //sm is already enlisted
        }

        return newState;
    }

    /**
     * Method to transition to persistent state.
     * @param sm StateManager.
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionMakePersistent(StateManagerImpl sm)
    {
        return this;
    }

    /**
     * Method to transition to delete persistent state.
     * @param sm StateManager.
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionDeletePersistent(StateManagerImpl sm)
    {
        return this;
    }

    /**
     * Method to transition to transactional state.
     * @param sm StateManager.
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionMakeTransactional(StateManagerImpl sm)
    {
        return this;
    }

    /**
     * Method to transition to nontransactional state.
     * @param sm StateManager.
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionMakeNontransactional(StateManagerImpl sm)
    {
        return this;
    }

    /**
     * Method to transition to transient state.
     * @param sm StateManager.
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionMakeTransient(StateManagerImpl sm)
    {
        return this;
    }

    /**
     * Method to transition to commit state.
     * @param sm StateManager.
     * @param tx the Transaction been committed.
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionCommit(StateManagerImpl sm,Transaction tx)
    {
        return this;
    }

    /**
     * Method to transition to rollback state.
     * @param sm StateManager.
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionRollback(StateManagerImpl sm)
    {
        return this;
    }

    /**
     * Method to transition to refresh state.
     * @param sm StateManager.
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionRefresh(StateManagerImpl sm)
    {
        return this;
    }

    /**
     * Method to transition to evict state.
     * @param sm StateManager.
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionEvict(StateManagerImpl sm)
    {
        return this;
    }

    /**
     * Method to transition to read-field state.
     * @param sm StateManager.
     * @param isLoaded if the field was previously loaded 
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionReadField(StateManagerImpl sm, boolean isLoaded)
    {
        return this;
    }

    /**
     * Method to transition to write-field state.
     * @param sm StateManager.
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionWriteField(StateManagerImpl sm)
    {
        return this;
    }

    /**
     * Method to transition to retrieve state.
     * @param sm StateManager.
     * @param dfgOnly only DefaultFetchGroup (not used !)
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionRetrieve(StateManagerImpl sm, boolean dfgOnly)
    {
        return this;
    }

    /**
     * Return whether the object is dirty, ie has been changed
     * (created, updated, deleted) in this Tx.
     * @return Whether the object is dirty. 
     */
    public boolean isDirty()
    {
        return isDirty;
    }
   
    /**
     * Return whether the object was newly created.
     * @return Whether the object is new. 
     */
    public boolean isNew()
    {
        return isNew;
    }

    /**
     * Return whether the object is deleted.
     * @return Whether the object is deleted. 
     */
    public boolean isDeleted()
    {
        return isDeleted;
    }

    /**
     * Return whether the object is transactional.
     * @return Whether the object is transactional. 
     */
    public boolean isTransactional()
    {
        return isTransactional;
    }
    
    /**
     * Return whether the object is persistent.
     * @return Whether the object is persistent. 
     */
    public boolean isPersistent()
    {
        return isPersistent;
    }
}