/**********************************************************************
Copyright (c) 2002 Kelly Grizzle and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
2002 Mike Martin - unknown changes
2003 Andy Jefferson - commented
    ...
**********************************************************************/
package org.jpox.state;

import javax.jdo.Transaction;

/**
 * Class representing the life cycle state of PersistentClean.
 *
 * @version $Revision: 1.1 $
 **/
class PersistentClean extends LifeCycleState
{
    /** Protected Constructor to prevent external instantiation. */
    protected PersistentClean()
    {
		isPersistent = true;        	
        isDirty = false;
        isNew = false;
        isDeleted = false;
        isTransactional = true;
        
        stateType = P_CLEAN;
    }

    /**
     * Method to transition to delete persistent.
     * @param sm StateManager.
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionDeletePersistent(StateManagerImpl sm)
    {
        return changeState(sm, P_DELETED);
    }

    /**
     * Method to transition to nontransactional.
     * @param sm StateManager.
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionMakeNontransactional(StateManagerImpl sm)
    {
        sm.clearSavedFields();
        return changeState(sm, P_NONTRANS);
    }

    /**
     * Method to transition to transient.
     * @param sm StateManager.
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionMakeTransient(StateManagerImpl sm)
    {
        return changeState(sm, TRANSIENT);
    }

    /**
     * Method to transition to commit state.
     * @param sm StateManager.
     * @param tx the Transaction been committed.
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionCommit(StateManagerImpl sm, Transaction tx)
    {
        sm.clearSavedFields();

        if (tx.getRetainValues())
        {
            return changeState(sm, P_NONTRANS);
        }
        else
        {
            sm.clearNonPrimaryKeyFields();
            return changeState(sm, HOLLOW);
        }
    }

    /**
     * Method to transition to rollback state.
     * @param sm StateManager.
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionRollback(StateManagerImpl sm)
    {
        Transaction tx = sm.getPersistenceManager().currentTransaction();
        
        return transitionCommit(sm, tx);
    }

    /**
     * Method to transition to evict state.
     * @param sm StateManager.
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionEvict(StateManagerImpl sm)
    {
        sm.clearNonPrimaryKeyFields();
        sm.clearSavedFields();
        return changeState(sm, HOLLOW);
    }

    /**
     * Method to transition to write-field state.
     * @param sm StateManager.
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionWriteField(StateManagerImpl sm)
    {
        Transaction tx = sm.getPersistenceManager().currentTransaction();
        if (tx.getRestoreValues())
        {
            sm.saveFields();
        }

        return changeState(sm, P_DIRTY);
    }

	/**
	 * Method to transition to refresh state.
	 * @param sm StateManager.
	 * @return new LifeCycle state.
	 **/
	public LifeCycleState transitionRefresh(StateManagerImpl sm)
	{
		sm.clearSavedFields();
		sm.refreshLoadedFields();
        Transaction tx = sm.getPersistenceManager().currentTransaction();
		if (tx.isActive())
		{
			return changeState(sm,P_CLEAN);
		}
		return changeState(sm,P_NONTRANS);      
	}
	
    /**
     * Method to transition to retrieve state.
     * @param sm StateManager.
	 * @param dfgOnly only DefaultFetchGroup
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionRetrieve(StateManagerImpl sm, boolean dfgOnly)
    {
		if (dfgOnly)
        {
			sm.loadDFGFields();
        }
		else
        {
			sm.loadUnloadedFieldsInFetchPlan();
        }
        return this;
    }
    
    /**
     * Method to return a string version of this object.
     * @return The string "P_CLEAN".
     **/
    public String toString()
    {
        return "P_CLEAN";
    }
}
