/**********************************************************************
Copyright (c) 2002 Kelly Grizzle and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
2002 Mike Martin - unknown changes
2003 Andy Jefferson - commented
2003 Erik Bengtson - optimistic transaction
    ...
**********************************************************************/
package org.jpox.state;

import javax.jdo.Transaction;

import org.jpox.state.exceptions.IllegalStateTransitionException;

/**
 * Class representing the life cycle state of PersistentNontransactional
 *
 * @version $Revision: 1.1 $
 **/
class PersistentNontransactional extends LifeCycleState
{
    /** Protected Constructor to prevent external instantiation. */
    protected PersistentNontransactional()
    {
		isPersistent = true;    	
        isDirty = false;
        isNew = false;
        isDeleted = false;
        isTransactional = false;

        stateType = P_NONTRANS;
    }

    /**
     * Method to transition to delete persistent.
     * @param sm StateManager.
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionDeletePersistent(StateManagerImpl sm)
    {
        return changeState(sm, P_DELETED);
    }

    /**
     * Method to transition to transactional.
     * @param sm StateManager.
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionMakeTransactional(StateManagerImpl sm)
    {
        sm.refreshLoadedFields();
        return changeState(sm, P_CLEAN);
    }

    /**
     * Method to transition to transient.
     * @param sm StateManager.
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionMakeTransient(StateManagerImpl sm)
    {
        return changeState(sm, TRANSIENT);
    }

    /**
     * Method to transition to commit state.
     * @param sm StateManager.
     * @param tx the Transaction been committed.
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionCommit(StateManagerImpl sm, Transaction tx)
    {
        throw new IllegalStateTransitionException(this, "commit", sm);
    }

    /**
     * Method to transition to rollback state.
     * @param sm StateManager.
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionRollback(StateManagerImpl sm)
    {
        throw new IllegalStateTransitionException(this, "rollback", sm);
    }

    /**
     * Method to transition to refresh state.
     * @param sm StateManager.
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionRefresh(StateManagerImpl sm)
    {
        sm.refreshLoadedFields();
        return this;
    }
    
    /**
     * Method to transition to evict state.
     * @param sm StateManager.
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionEvict(StateManagerImpl sm)
    {
        sm.clearNonPrimaryKeyFields();
        sm.clearSavedFields();
        return changeState(sm, HOLLOW);
    }

    /**
     * Method to transition to read-field state.
     * @param sm StateManager.
     * @param isLoaded if the field was previously loaded.
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionReadField(StateManagerImpl sm, boolean isLoaded)
    {
        Transaction tx = sm.getPersistenceManager().currentTransaction();
        if (tx.isActive() && ! tx.getOptimistic())
        {
            // Save the fields for rollback.
            sm.saveFields();
            sm.refreshLoadedFields();
            return changeState(sm, P_CLEAN);
        }
        else
        {
            return this;
        }
    }

    /**
     * Method to transition to write-field state.
     * @param sm StateManager.
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionWriteField(StateManagerImpl sm)
    {
        Transaction tx = sm.getPersistenceManager().currentTransaction();
        if (tx.isActive())
        {
            // Save the fields for rollback.
            sm.saveFields();
            sm.refreshLoadedFields();
            return changeState(sm, P_DIRTY);
        }
        else
        {
            return this;
        }
    }

    /**
     * Method to transition to retrieve state.
     * @param sm StateManager.
	 * @param dfgOnly only DefaultFetchGroup
     * @return new LifeCycle state.
     **/
    public LifeCycleState transitionRetrieve(StateManagerImpl sm, boolean dfgOnly)
    {
        Transaction tx = sm.getPersistenceManager().currentTransaction();
        if (tx.isActive() && !tx.getOptimistic())
        {
            // Save the fields for rollback.
            sm.saveFields();
    		if (dfgOnly)
            {
    			sm.loadDFGFields();
            }
    		else
            {
    			sm.loadUnloadedFieldsInFetchPlan();
            }             
            return changeState(sm, P_CLEAN);
        }
        else if (tx.isActive() && tx.getOptimistic())
        {
            // Save the fields for rollback.
            sm.saveFields();
    		if (dfgOnly)
            {
    			sm.loadDFGFields();
            }
    		else
            {
    			sm.loadUnloadedFieldsInFetchPlan();
            }
    		return this;
        }
        else
        {
    		if (dfgOnly)
            {
    			sm.loadDFGFields();
            }
    		else
            {
    			sm.loadUnloadedFieldsInFetchPlan();
            } 
    		return this;
        }
    }

    /**
     * Method to return a string version of this object.
     * @return The string "P_NONTRANS".
     **/
    public String toString()
    {
        return "P_NONTRANS";
    }
}
