/**********************************************************************
Copyright (c) 2002 Kelly Grizzle and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
2003 Erik Bengtson - removed exist() operation
2003 Andy Jefferson - added localiser
2003 Erik Bengtson - added new constructor for App ID
2003 Erik Bengtson - fixed loadDefaultFetchGroup to call jdoPostLoad
2003 Erik Bengtson - fixed evict to call jdoPreClear
2004 Andy Jefferson - converted to use Logger
2004 Andy Jefferson - reordered methods to put in categories, split String utilities across to StringUtils.
2004 Andy Jefferson - added Lifecycle Listener callbacks
2004 Andy Jefferson - removed JDK 1.4 methods so that we support 1.3 also
 	...
 **********************************************************************/
package org.jpox.state;

import java.io.PrintWriter;
import java.security.AccessController;
import java.security.PrivilegedAction;
import java.util.ArrayList;
import java.util.BitSet;
import java.util.Iterator;
import java.util.List;

import javax.jdo.ClearCallback;
import javax.jdo.DeleteCallback;
import javax.jdo.JDOException;
import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOFatalUserException;
import javax.jdo.JDOObjectNotFoundException;
import javax.jdo.JDOUnsupportedOptionException;
import javax.jdo.JDOUserException;
import javax.jdo.LoadCallback;
import javax.jdo.StoreCallback;
import javax.jdo.Transaction;
import javax.jdo.spi.Detachable;
import javax.jdo.spi.JDOImplHelper;
import javax.jdo.spi.PersistenceCapable;
import javax.jdo.spi.StateManager;

import org.jpox.AbstractPersistenceManager;
import org.jpox.ClassLoaderResolver;
import org.jpox.FetchPlanImpl;
import org.jpox.LifecycleListenerSpecification;
import org.jpox.PersistenceManager;
import org.jpox.TypeManager;
import org.jpox.metadata.ClassMetaData;
import org.jpox.metadata.FieldMetaData;
import org.jpox.metadata.IdentityStrategy;
import org.jpox.metadata.IdentityType;
import org.jpox.metadata.MetaDataManager;
import org.jpox.sco.SCO;
import org.jpox.sco.UnsetOwners;
import org.jpox.store.FieldValues;
import org.jpox.store.OID;
import org.jpox.store.StoreManager;
import org.jpox.store.fieldmanager.FieldManager;
import org.jpox.store.rdbms.table.ClassTable;
import org.jpox.util.AIDUtils;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;
import org.jpox.util.StringUtils;
import org.jpox.util.TypeConversionHelper;

/**
 * This class implements the StateManager. It is implemented here as one
 * StateManager per Object so adds on functionality particular to each object.
 * All PersistenceCapable objects will have a StateManager when they have
 * had communication with the PersistenceManager. They will typically always have
 * an identity also. The exception to that is for embedded objects.
 * 
 * <H3>Embedded Objects</H3>
 * An object that is being embedded in an owning object will NOT have an identity
 * unless the object is subject to a makePersistent() call also. When an object
 * is embedded and a field is changed, the field will NOT be marked as dirty (unless
 * it is also an object in its own right with an identity). When a field is changed
 * any owning objects are updated so that they can update their tables accordingly.
 * 
 * @version $Revision: 1.34 $
 */
public final class StateManagerImpl implements org.jpox.StateManager
{
    protected static final Localiser LOCALISER = Localiser.getInstance("org.jpox.state.Localisation");

    private static final JDOImplHelper HELPER;

    private static final StateFieldManager HOLLOWFIELDMANAGER = new StateFieldManager();

    /**
     * the PersistenceManager for this StateManager
     */
    private PersistenceManager myPM;

    /**
     * The PersistenceCapable instance managed by this StateManager
     */
    protected PersistenceCapable myPC;

    /**
     * The object identity in the datastore
     */
    private Object myID;

    /**
     * The actual LifeCycleState for the PersistenceCapable instance
     */
    private LifeCycleState myLC;

    private byte jdoFlags;

    /**
     * an image of the PersistenceCapable instance at the moment the instance
     * is enlisted in the transaction 
     */
    private PersistenceCapable savedImage = null;

    /**
     * flags of the PersistenceCapable instance at the moment the instance
     * is enlisted in the transaction 
     */
    private byte savedFlags;

    /**
     * The loaded fields of the PersistenceCapable instance at the moment the instance
     * is enlisted in the transaction 
     */
    private boolean[] savedLoadedFields = null;

    /**
     * the metadata for the class
     */
    private ClassMetaData cmd;

    /**
     * the StoreManager
     */
    private StoreManager srm;

    /**
     * number of managed fields from this class plus inherited classes
     */
    private int fieldCount;

    /**
     * if the PersistenceCapable instance is dirty
     */
    private boolean dirty = false;

    /**
     * the dirty fields of the PersistenceCapable instance
     */
    private boolean[] dirtyFields;

    /**
     * the loaded fields of the PersistenceCapable instance
     */
    private boolean[] loadedFields;

    private boolean[] defaultFetchGroupFields;

    private boolean[] secondClassMutableFields;
    
    /**
     * Field numbers of the non-primary key fields (inc superclasses).
     */
    private boolean[] nonPrimaryKeyFields;    

    private int[] allFieldNumbers;

    private int[] allNonPrimaryKeyFieldNumbers;

    private int[] defaultFetchGroupFieldNumbers;

    private int[] secondClassMutableFieldNumbers;

    /**
     * if the PersistenceCapable instance is new and was flushed to the datastore
     */
    private boolean flushedNew = false;

    /**
     * version field for optimistic transactions
     */
    private Object optimisticDatastoreVersion;

    /**
     * version field for optimistic transactions.
     * This holds a version number after a insert/update, but not yet committed.
     */
    private Object optimisticTransactionalVersion;

    /**
     * the current FieldManager
     */
    private FieldManager currFM = null;

    /**
     * restore values at StateManager. If true, overwrites the restore values
     * at tx level
     */
    private boolean restoreValues = false;

    /**
     * the fetch plan
     */
    private FetchPlanImpl.FetchPlanForClass myFP;

    /**
     * loaded fields in detached instances
     */
    private BitSet jdoLoadedFields;

    /**
     * modified fields in detached instances
     */
    private BitSet jdoModifiedFields;

    /*
     * state for transitions of activities
     */
    private ActivityState inserting = ActivityState.NONE;

    private boolean flushing = false;

    private boolean changingState = false;

    private boolean postLoadPending = false;

    private boolean disconnecting = false;

    private boolean deleting = false;
    
    private boolean removingReference = false;

    /** Flag to indicate if the object is in the process of being detached. */
    private boolean detaching = false;
    private PersistenceCapable detachingPC = null;

    private boolean attaching = false;

    private List activityListeners = new ArrayList();

    /** List of owners when embedded. */
    private List embeddedOwners = null;

    static
    {
        HELPER = (JDOImplHelper) AccessController.doPrivileged(new PrivilegedAction()
        {
            public Object run()
            {
                try
                {
                    return JDOImplHelper.getInstance();
                }
                catch (SecurityException e)
                {
                    throw new JDOFatalUserException(LOCALISER.msg("StateManager.SecurityProblem"), e);
                }
            }
        });
    }

    // ------------------------------ Constructors -----------------------------

    /**
     * Constructs a state manager to manage a hollow instance having the
     * given object ID.  This constructor is used for creating new instances of
     * existing persistent objects (i.e. via {@link PersistenceManager#getObjectById}).
     * @param pm the persistence manager controlling this state manager.
     * @param pcClass the class of the new instance to be created.
     * @param id the JDO identity of the object.
     */
    public StateManagerImpl(PersistenceManager pm, Class pcClass, Object id)
    {
        myPM = pm;
        srm = myPM.getStoreManager();
        myID = id;
        myLC = LifeCycleState.getLifeCycleState(LifeCycleState.HOLLOW);
        jdoFlags = PersistenceCapable.LOAD_REQUIRED;

        // TODO interfaces
        // in this case JPOX supports only one level for interfaces
        // in case of many levels, JPOX supports only one branch
        if (pcClass.isInterface())
        {
            cmd = pm.getMetaDataManager().getMetaDataForInterface(pcClass, null);
        }
        else
        {
            try
            {
                // calling the classloaderresolver will make sure the class is initialized
                cmd = pm.getMetaDataManager().getMetaDataForClass(
                    pm.getClassLoaderResolver().classForName(pcClass.getName(), pm.getClass().getClassLoader(), true));
            }
            catch (ClassNotFoundException e)
            {
                throw new JDOFatalUserException(LOCALISER.msg("StateManager.ClassNotFoundError", pcClass.getName()));
            }
        }

        if (cmd == null)
        {
            throw new JDOFatalUserException(LOCALISER.msg("StateManager.NotPersistableClassError", pcClass));
        }

        myFP = ((FetchPlanImpl) myPM.getFetchPlan()).manageFetchPlanForClass(cmd);

        initialize();

        // if it is a datastore ID, dont copy key fields to Pc class
        ClassLoaderResolver clr = myPM.getClassLoaderResolver();
        if (id instanceof OID)
        {
            myPC = HELPER.newInstance(clr.classForName(cmd.getFullClassName()), this);
        }
        else
        {
            //this will create a new pc class, and copy the key class to fields
            myPC = HELPER.newInstance(clr.classForName(cmd.getFullClassName()), this, myID);
        }
    }

    /**
     * Constructs a state manager to manage a PersistenceCapable instance that will
     * be embedded into another PersistenceCapable object. The instance will not be
     * assigned an identity in the process since it is effectively a SCO.
     * @param pc The PersistenceCapable to manage
     * @param pm The persistence manager controlling this state manager.
     */
    public StateManagerImpl(PersistenceCapable pc, PersistenceManager pm)
    {
        myPC = pc;
        myPM = pm;
        srm = myPM.getStoreManager();
        myID = null; // It is embedded at this point so dont need an ID since we're not persisting it
        myLC = LifeCycleState.getLifeCycleState(LifeCycleState.P_NEW);
        jdoFlags = PersistenceCapable.LOAD_REQUIRED;
        replaceStateManager(this);

        cmd = pm.getMetaDataManager().getMetaDataForClass(pc.getClass().getName(), pm.getClassLoaderResolver());
        if (cmd == null)
        {
            throw new JDOFatalUserException(LOCALISER.msg("StateManager.NotPersistableClassError", pc.getClass().getName()));
        }

        myFP = ((FetchPlanImpl) myPM.getFetchPlan()).manageFetchPlanForClass(cmd);

        initialize();

        // Mark all fields as loaded since we are using the passed PersistenceCapable
        for (int i=0;i<loadedFields.length;i++)
        {
            loadedFields[i] = true;
        }
    }

    /**
     * Constructs a state manager to manage a hollow instance having the given
     * FieldValues. This constructor is used for creating new instances of
     * existing persistent objects using application identity.
     * {@link PersistenceManager#getObjectByAID}
     * @param pm the persistence manager controlling this state manager.
     * @param pcClass the class of the new instance to be created.
     * @param fv the initial field values of the object.
     * @param checkInheritance Whether look to the database to determine which
     * class this object is. This parameter is a hint. Set false, if it's
     * already determined the correct pcClass for this pc "object" in a certain
     * level in the hierarchy. Set to true and it will look to the database.
     */
    public StateManagerImpl(PersistenceManager pm, Class pcClass, FieldValues fv, boolean checkInheritance)
    {
        myPM = pm;
        srm = myPM.getStoreManager();
        myLC = LifeCycleState.getLifeCycleState(LifeCycleState.HOLLOW);
        jdoFlags = PersistenceCapable.LOAD_REQUIRED;

        //calling the classloaderresolver will make sure the class is initialized
        cmd = pm.getMetaDataManager().getMetaDataForClass(pm.getClassLoaderResolver().classForName(pcClass.getName(), true));

        if (cmd == null)
        {
            throw new JDOFatalUserException(LOCALISER.msg("StateManager.NotPersistableClassError", pcClass));
        }

        myFP = ((FetchPlanImpl) myPM.getFetchPlan()).manageFetchPlanForClass(cmd);

        if (cmd.getIdentityType() != IdentityType.APPLICATION)
        {
            throw new JDOFatalUserException("This constructor is only for objects using application identity.");
        }
        initialize();

        ClassLoaderResolver clr = myPM.getClassLoaderResolver();
        myPC = HELPER.newInstance(clr.classForName(cmd.getFullClassName()), this);
        if (myPC == null)
        {
            throw new JDOFatalUserException(LOCALISER.msg("StateManager.ClassNotRegistered", cmd.getFullClassName()));
        }

        fv.fetchFields(this); //at minimum primary key fields are loaded here
        myID = myPC.jdoNewObjectIdInstance();

        if( !AIDUtils.isSingleFieldIdentityClass(cmd.getObjectidClass()) )
        {
            myPC.jdoCopyKeyFieldsToObjectId(myID);
        }

        ///-------------------------- in inheritance cases, verify to which hierarchy level this id 
        if (checkInheritance)
        {
            String className = srm.getClassNameForObjectID(myPM, myID);
            // className is null when id class exists, and object has been validated and doesn't exist.
            if (className == null)
            {
                throw new JDOObjectNotFoundException(LOCALISER.msg("StateManager.ObjectDoesntExist"), myID);
            }
            else if (!cmd.getFullClassName().equals(className))
            {
                try
                {
                    pcClass = myPM.getClassLoaderResolver().classForName(className, myID.getClass().getClassLoader());
                }
                catch (ClassNotFoundException e)
                {
                    JPOXLogger.JDO.warn(LOCALISER.msg("StateManager.GetObjectByIdClassNotFound", myID));
                    throw new JDOUserException(LOCALISER.msg("StateManager.GetObjectByIdClassNotFound", myID), e);
                }
                //calling the classloaderresolver will make sure the class is initialized
                cmd = pm.getMetaDataManager().getMetaDataForClass(pm.getClassLoaderResolver().classForName(pcClass.getName(), true));
                if (cmd == null)
                {
                    throw new JDOFatalUserException(LOCALISER.msg("StateManager.NotPersistableClassError", pcClass));
                }

                myFP = ((FetchPlanImpl) myPM.getFetchPlan()).manageFetchPlanForClass(cmd);

                if (cmd.getIdentityType() != IdentityType.APPLICATION)
                {
                    throw new JDOFatalUserException("This constructor is only for objects using application identity.");
                }
                initialize();

                myPC = HELPER.newInstance(clr.classForName(cmd.getFullClassName()), this);
                if (myPC == null)
                {
                    throw new JDOFatalUserException(LOCALISER.msg("StateManager.ClassNotRegistered", cmd.getFullClassName()));
                }

                fv.fetchFields(this);
                
                myID = myPC.jdoNewObjectIdInstance();

                if( !AIDUtils.isSingleFieldIdentityClass(cmd.getObjectidClass()) )
                {
                    myPC.jdoCopyKeyFieldsToObjectId(myID);
                }                
            }
        }
        //---------------

        transitionReadField();

        /*
         * Need to do something about this eventually.  It's not a good idea to
         * be calling user code (jdoPostLoad()) when we haven't even finished
         * constructing the StateManager.
         */
        if (isDFGLoaded())
        {
            postLoad();
        }
    }

    /**
     * Constructs a state manager to manage a recently populated hollow instance having the
     * given object ID and the given field values.  This constructor is used for
     * creating new instances of persistent objects obtained e.g. via a Query or
     * backed by a view (i.e. via {@link PersistenceManager#getObjectById(Object,FieldValues)}).
     * @param pm the persistence manager controlling this state manager.
     * @param pcClass the class of the new instance to be created.
     * @param id the JDO identity of the object.
     * @param fv the initial field values of the object.
     */
    public StateManagerImpl(PersistenceManager pm, Class pcClass, Object id, FieldValues fv)
    {
        this(pm, pcClass, id);

        transitionReadField();

        fv.fetchFields(this);

        /*
         * Need to do something about this eventually.  It's not a good idea to
         * be calling user code (jdoPostLoad()) when we haven't even finished
         * constructing the StateManager.
         */
        if (isDFGLoaded())
        {
            postLoad();
        }
    }

    /**
     * Constructs a state manager to manage a transient instance that is 
     * becoming newly persistent.  A new object ID for the
     * instance is obtained from the store manager and the object is inserted
     * in the data store.
     * <p>This constructor is used for assigning state managers to existing
     * instances that are transitioning to a persistent state (i.e. via
     * {@link PersistenceManager#makePersistent(Object)}).
     *
     * @param pm the persistence manager controlling this state manager.
     * @param pc the instance being make persistent.
     * @param preInsertChanges Any changes to make before inserting
     */
    public StateManagerImpl(PersistenceManager pm, PersistenceCapable pc, FieldValues preInsertChanges)
    {
        Class pcClass = pc.getClass();

        myPM = pm;
        srm = myPM.getStoreManager();
        myPC = pc;
        myLC = LifeCycleState.getLifeCycleState(LifeCycleState.P_NEW);
        jdoFlags = PersistenceCapable.READ_OK;

        if (pcClass.isInterface())
        {
            cmd = pm.getMetaDataManager().getMetaDataForInterface(pcClass, null);
        }
        else
        {
            //calling the classloaderresolver will make sure the class is initialized
            cmd = pm.getMetaDataManager().getMetaDataForClass(pm.getClassLoaderResolver().classForName(pcClass.getName(), true));
        }

        if (cmd == null)
        {
            throw new JDOFatalUserException(LOCALISER.msg("StateManager.NotPersistableClassError", pcClass.getName()));
        }

        myFP = ((FetchPlanImpl) myPM.getFetchPlan()).manageFetchPlanForClass(cmd);

        initialize();

        for (int i = 0; i < fieldCount; ++i)
        {
            loadedFields[i] = true;
        }

        replaceStateManager(this);
        myPC.jdoReplaceFlags();

        saveFields();

        // Populate all fields that have "value-strategy" and are not datastore populated
        populateStrategyFields();

        // Set the identity
        setIdentity();

        if (preInsertChanges != null)
        {
            preInsertChanges.fetchFields(this);
        }

        if (cmd.getIdentityType() == IdentityType.APPLICATION)
        {
            //load key fields from Application Id instance to PC instance

            //if a primary key field is of type PersistenceCapable, it must first be persistent
            for (int fieldNumber = 0; fieldNumber < allFieldNumbers.length; fieldNumber++)
            {
                FieldMetaData fmd = cmd.getManagedFieldAbsolute(fieldNumber);
                if (fmd.isPrimaryKey())
                {
                    if (pm.getMetaDataManager().getMetaDataForClass(fmd.getType()) != null)
                    {
                        FieldManager oldFM = currFM;
                        currFM = new SingleFieldValue();
                        myPC.jdoProvideField(fieldNumber);
                        PersistenceCapable pkFieldPC = (PersistenceCapable) ((SingleFieldValue) currFM).fetchObjectField(fieldNumber);
                        if (pkFieldPC == null)
                        {
                            throw new JDOUserException(LOCALISER.msg("StateManager.PrimaryKeyFieldIsNull",fmd.getFullFieldName()));
                        }
                        if (!pkFieldPC.jdoIsPersistent())
                        {
                            myPM.makePersistent(pkFieldPC);
                        }
                        currFM = oldFM;
                    }
                }
            }
            if (!AIDUtils.isSingleFieldIdentityClass(cmd.getObjectidClass()))
            {
                myPC.jdoCopyKeyFieldsToObjectId(myID);
            }
        }

        myPM.enlistInTransaction(this);
    }

    /**
     * Constructs a state manager to manage a Transactional Transient instance.
     * A new object ID for the instance is obtained from the store manager and
     * the object is inserted in the data store.
     * <p>
     * This constructor is used for assigning state managers to Transient
     * instances that are transitioning to a transient clean state (i.e. via
     * {@link PersistenceManager#makeTransactional(Object)}).
     * @param pm the persistence manager controlling this state manager.
     * @param pc the instance being make persistent.
     */
    public StateManagerImpl(PersistenceManager pm, PersistenceCapable pc)
    {
        Class pcClass = pc.getClass();

        myPM = pm;
        srm = myPM.getStoreManager();
        myPC = pc;
        myLC = null;
        jdoFlags = PersistenceCapable.READ_OK;

        // Retrieve the MetaData for the class
        if (pcClass.isInterface())
        {
            cmd = pm.getMetaDataManager().getMetaDataForInterface(pcClass, pc);
        }
        else
        {
            //calling the classloaderresolver will make sure the class is initialized
            cmd = pm.getMetaDataManager().getMetaDataForClass(pm.getClassLoaderResolver().classForName(pcClass.getName(), true));
        }
        if (cmd == null)
        {
            throw new JDOFatalUserException(LOCALISER.msg("StateManager.NotPersistableClassError", pcClass.getName()));
        }

        myFP = ((FetchPlanImpl) myPM.getFetchPlan()).manageFetchPlanForClass(cmd);

        initialize();

        for (int i = 0; i < fieldCount; ++i)
        {
            loadedFields[i] = true;
        }

        myPC.jdoReplaceFlags();

        // Populate all fields that have "value-strategy" and are not datastore populated
        populateStrategyFields();

        // Set the identity
        setIdentity();

        // for non transactional read, tx might be not active
        // TODO add verification if is non transactional read = true
        if (myPM.currentTransaction().isActive())
        {
            myPM.enlistInTransaction(this);
        }
    }

    /**
     * Constructor for creating SM instances to manage PersistenceCapable
     * objects in detached state
     * @since 1.1
     * @param pc the detach object.
     * @param cmd the ClassMetaData
     * @param id the JDO identity of the object.
     */
    public StateManagerImpl(PersistenceCapable pc, ClassMetaData cmd, Object id)
    {
        this.cmd = cmd;
        myID = id;
        myPC = pc;
        myFP = null;
        myPM = null;
        srm = null;
        //myLC = LifeCycleState.getLifeCycleState(LifeCycleState.DETACHED_CLEAN);

        jdoFlags = PersistenceCapable.DETACHED;
        myPC.jdoReplaceFlags();
        initialize();
    }

    /**
     * Constructor to create a StateManager for a PersistenceCapable object, assigning the
     * specified id to the object. This is used when getting objects out of the L2 Cache,
     * where they have no StateManager assigned, and returning them as associated with a
     * particular PM.
     * @param pm PersistenceManager managing this object
     * @param id Id to assign to the PersistenceCapable object
     * @param pc The PersistenceCapable object
     */
    public StateManagerImpl(PersistenceManager pm, Object id, PersistenceCapable pc)
    {
        // Create a new copy of the input object type
        this(pm, pc.getClass(), id);

        transitionReadField();

        // Copy the fields from the input object
        // Note : we need to change the input object to be connected to this SM to do this
        pc.jdoReplaceStateManager(this);
        myPC.jdoCopyFields(pc, allFieldNumbers);

        // Disconnect the input object
        disconnectClone(pc);

        if (isDFGLoaded())
        {
            postLoad();
        }
    }

    /**
     * Convenience method to populate all fields in the PC object that have "value-strategy" specified
     * and that aren't datastore attributed. This applies not just to PK fields (where it is most
     * useful to use value-strategy) but also to any other field. Fields are populated only if they are null
     * This is called once on a PC object, when makePersistent is called.
     */
    private void populateStrategyFields()
    {
        int totalFieldCount = cmd.getNoOfInheritedManagedFields() + cmd.getNoOfManagedFields();
        ClassTable table = (ClassTable)srm.getDatastoreClass(cmd.getFullClassName());

        for (int fieldNumber =0; fieldNumber < totalFieldCount; fieldNumber++)
        {
            FieldMetaData fmd = cmd.getManagedFieldAbsolute(fieldNumber);
            IdentityStrategy strategy = fmd.getValueStrategy();

            // Check for the strategy, and if it is a datastore attributed strategy
            if (strategy != null && !StoreManager.isStrategyDatastoreAttributed(strategy))
            {
               // if( this.provideField(fieldNumber) == null )
               // {
	                ClassTable fieldTable = table.getBaseTableWithField(fmd);
	                Object obj = srm.getStrategyValue(myPM, fieldTable, cmd, fieldNumber, strategy, fmd.getSequence(),
	                    fmd.getExtensions());
	                this.replaceField(fieldNumber, obj);
               // }
            }
        }
    }

    /**
     * Utility to set the identity for the PersistenceCapable object.
     * Will only create the id instance if it is not attributed (in some way) by the datastore.
     * So, for example, autoassign cases will not gain their identity here. In such cases the identity
     * is only created when they are inserted into the datastore.
     */
    private void setIdentity()
    {
        if (cmd.getIdentityType() == IdentityType.APPLICATION)
        {
            boolean isObjectIDDatastoreAttributed = false;
            int totalFieldCount = cmd.getNoOfInheritedManagedFields() + cmd.getNoOfManagedFields();
            for (int fieldNumber =0; fieldNumber < totalFieldCount; fieldNumber++)
            {
                FieldMetaData fmd=cmd.getManagedFieldAbsolute(fieldNumber);
                if (fmd.isPrimaryKey() && StoreManager.isStrategyDatastoreAttributed(fmd.getValueStrategy()))
                {
                    isObjectIDDatastoreAttributed = true;
                    break;
                }
            }
            if (!isObjectIDDatastoreAttributed || !AIDUtils.isSingleFieldIdentityClass(cmd.getObjectidClass()))
            {
                myID = srm.newObjectID(this, cmd.getFullClassName(), myPC);
            }
        }
        else
        {
            if (cmd.getIdentityMetaData() == null || !StoreManager.isStrategyDatastoreAttributed(cmd.getIdentityMetaData().getStrategyValue()))
            {
                myID = srm.newObjectID(this, cmd.getFullClassName(), myPC);
            }
        }
    }

    /**
     * Convenience method to get a copy of the PersistenceCapable object used here, but
     * without a StateManager connected. Note that the id will be null on the returned object.
     * @return Copy of the PersistenceCapable
     */
    public PersistenceCapable getPCCopyWithoutSM()
    {
        PersistenceCapable pcCopy = myPC.jdoNewInstance(this, myPC.jdoGetObjectId());
        pcCopy.jdoCopyFields(myPC, allFieldNumbers);

        // Note that this removes the object id
        disconnectClone(pcCopy);

        return pcCopy;
    }

    /**
     * Utility method to initialise the state information.
     **/
    private void initialize()
    {
        fieldCount = cmd.getNoOfInheritedManagedFields() + cmd.getNoOfManagedFields();

        dirtyFields = new boolean[fieldCount];
        loadedFields = new boolean[fieldCount];
        defaultFetchGroupFields = cmd.getDefaultFetchGroupFieldFlags();
        secondClassMutableFields = cmd.getSecondClassMutableFieldFlags();
        nonPrimaryKeyFields = cmd.getNonPrimaryKeyFieldFlags();

        allFieldNumbers = cmd.getAllFieldNumbers();
        allNonPrimaryKeyFieldNumbers = cmd.getNonPrimaryKeyFieldNumbers();

        defaultFetchGroupFieldNumbers = cmd.getDefaultFetchGroupFieldNumbers();
        secondClassMutableFieldNumbers = cmd.getSecondClassMutableFieldNumbers();
    }

    /**
     * Utility to update our object to use a different state manager.
     * @param sm The new state manager.
     **/
    private void replaceStateManager(StateManager sm)
    {
        try
        {
            myPC.jdoReplaceStateManager(sm);
        }
        catch (SecurityException e)
        {
            throw new JDOFatalUserException(LOCALISER.msg("StateManager.SecurityProblem"), e);
        }
    }

    void enlistInTransaction()
    {
        myPM.enlistInTransaction(this);

        if (jdoFlags == PersistenceCapable.LOAD_REQUIRED && isDFGLoaded())
        {
            /*
             * A transactional object whose DFG fields are loaded does not need
             * to contact us in order to read those fields.
             */
            jdoFlags = PersistenceCapable.READ_OK;
            myPC.jdoReplaceFlags();
        }
    }

    void evictFromTransaction()
    {
        myPM.evictFromTransaction(this);

        /*
         * A non-transactional object needs to contact us on any field read no
         * matter what fields are loaded.
         */
        jdoFlags = PersistenceCapable.LOAD_REQUIRED;
        myPC.jdoReplaceFlags();
    }

    /**
     * Method to save all fields of the object.
     **/
    void saveFields()
    {
        savedImage = myPC.jdoNewInstance(this);
        savedImage.jdoCopyFields(myPC, allFieldNumbers);
        savedFlags = jdoFlags;
        savedLoadedFields = (boolean[]) loadedFields.clone();
    }

    /**
     * Method to restore all fields of the object.
     **/
    void restoreFields()
    {
        if (savedImage != null)
        {
            loadedFields = savedLoadedFields;
            jdoFlags = savedFlags;
            myPC.jdoReplaceFlags();
            myPC.jdoCopyFields(savedImage, allFieldNumbers);

            clearDirtyFlags();
            clearSavedFields();
        }
    }

    /**
     * Method to clear all fields of the object.
     **/
    void clearFields()
    {
        try
        {
            ((AbstractPersistenceManager) myPM).notifyLifecycleListenersForEvent(myPC, LifecycleListenerSpecification.EVENT_PRE_CLEAR);

            if (myPC instanceof ClearCallback)
            {
                ((ClearCallback) myPC).jdoPreClear();
            }

        }
        finally
        {
            replaceFields(allFieldNumbers, HOLLOWFIELDMANAGER);
            clearLoadedFlags();
            clearDirtyFlags();
            ((AbstractPersistenceManager) myPM).notifyLifecycleListenersForEvent(myPC, LifecycleListenerSpecification.EVENT_POST_CLEAR);
        }
    }

    /**
     * Method to clear all fields that are not part of the primary key of the object.
     **/
    void clearNonPrimaryKeyFields()
    {
        try
        {
            ((AbstractPersistenceManager) myPM).notifyLifecycleListenersForEvent(myPC, LifecycleListenerSpecification.EVENT_PRE_CLEAR);
            
            if (myPC instanceof ClearCallback)
            {
                ((ClearCallback) myPC).jdoPreClear();
            }
        }
        finally
        {
            replaceFields(allNonPrimaryKeyFieldNumbers, HOLLOWFIELDMANAGER);
            clearLoadedFlags();
            clearDirtyFlags();
            ((AbstractPersistenceManager) myPM).notifyLifecycleListenersForEvent(myPC, LifecycleListenerSpecification.EVENT_POST_CLEAR);
        }
    }

    /**
     * Method to clear all saved fields on the object.
     **/
    void clearSavedFields()
    {
        savedImage = null;
        savedFlags = 0;
        savedLoadedFields = null;
    }

    /**
     * Method to clear all loaded flags on the object.
     **/
    private void clearLoadedFlags()
    {
        jdoFlags = PersistenceCapable.LOAD_REQUIRED;
        myPC.jdoReplaceFlags();
        clearFlags(loadedFields);
    }

    /**
     * Method to clear all dirty flags on the object.
     **/
    private void clearDirtyFlags()
    {
        dirty = false;
        clearFlags(dirtyFields);
    }

    /**
     * Marks the given field dirty.
     * @param field The no of field to mark as dirty. 
     */
    public void makeDirty(int field)
    {
        boolean wasDirty = preWriteField(field);
        postWriteField(wasDirty);
    }

    /**
     * Marks the given field dirty for issuing an update after the insert.
     * @param pc The PersistenceCapable
     * @param fieldNumber The no of field to mark as dirty. 
     */
    public void updateFieldAfterInsert(PersistenceCapable pc, int fieldNumber)
    {
        ((StateManagerImpl) myPM.findStateManager(pc)).addActivityListener(this,ActivityState.INSERTING_CALLBACKS);

        dirty = true;
        dirtyFields[fieldNumber] = true;
    }
    
    /**
     * Mark the associated PersistenceCapable field dirty.
     *
     * @param pc the calling PersistenceCapable instance
     * @param fieldName the name of the field
     */
    public void makeDirty(PersistenceCapable pc, String fieldName)
    {
        if (!disconnectClone(pc))
        {
            int fieldNumber = cmd.getFieldNumberAbsolute(fieldName);

            if (fieldNumber == -1)
            {
                throw new JDOUserException(LOCALISER.msg("StateManager.InvalidFieldForClass", fieldName, cmd.getFullClassName()));
            }
            
            makeDirty(fieldNumber);
        }
    }

    // -------------------------- Accessor Methods -----------------------------

    /**
     * Accessor for the StoreManager used for this object.
     * @return The StoreManager.
     **/
    public StoreManager getStoreManager()
    {
        return srm;
    }

    /**
     * Accessor for the ClassMetaData for this object.
     * @return The ClassMetaData.
     **/
    public ClassMetaData getClassMetaData()
    {
        return cmd;
    }

    /**
     * Accessor for the MetaDataManager to use for this object.
     * Simply a wrapper accessor method. 
     * @return The MetaDataManager.
     **/
    public MetaDataManager getMetaDataManager()
    {
        return myPM.getMetaDataManager();
    }

    /**
     * Accessor for the PersistenceManager that owns this instance.
     * @param pc The PersistenceCapable instance
     * @return The PersistenceManager that owns this instance
     */
    public javax.jdo.PersistenceManager getPersistenceManager(PersistenceCapable pc)
    {
        PersistenceManager pm = this.myPM;
        //in identifying relationships, jdoCopyKeyFieldsFromId will call
        //this method, and at this moment, myPC in statemanager is null

        //if not identifying relationship, do the default check of disconnectClone:
        //"this.disconnectClone(pc)"        
        if (myPC != null && this.disconnectClone(pc))
        {
            return null;
        }
        else if (pm == null)
        {
            return null;
        }        
        else
        {
            this.myPM.hereIsStateManager(this, myPC);
            return pm.getPMHandle();
        }
    }

    /**
     * Accessor for the PersistenceManager for this object.
     * @return The PersistenceManager.
     **/
    public PersistenceManager getPersistenceManager()
    {
        if (myPM == null)
        {
            return null;
        }
        else
        {
            return myPM.getPMHandle();
        }
    }

    /**
     * Accessor for the Persistent Capable object.
     * @return The PersistentCapable object
     **/
    public PersistenceCapable getObject()
    {
        return myPC;
    }

    /**
     * Accessor for the Restore Values flag 
     * @return Whether to restore values
     */
    public boolean isRestoreValues()
    {
        return restoreValues;
    }

    /**
     * Mutator for the Restore Values flag 
     * @param restore_values Whether to restore values
     */
    private void setRestoreValues(boolean restore_values)
    {
        restoreValues = restore_values;
    }

    /**
     * Return the object representing the JDO identity of the calling instance.
     *
     * According to the JDO specification, if the JDO identity is being changed
     * in the current transaction, this method returns the JDO identify as of
     * the beginning of the transaction.
     *
     * @param pc the calling PersistenceCapable instance
     * @return the object representing the JDO identity of the calling instance
     */
    public Object getObjectId(PersistenceCapable pc)
    {
        if (disconnectClone(pc))
        {
            return null;
        }
        else
        {
            return getExternalObjectId(pc);
        }
    }

    /**
     * Accessor for the internal object id of the object we are managing. 
     * @return The internal object id 
     */
    public Object getInternalObjectId()
    {
        if (myID == null)
        {
            return "<unset-id>";
        }
        if (cmd.getIdentityType() == IdentityType.APPLICATION)
        {
            return myID;
        }
        else
        {
            if (myID instanceof OID)
            {
                return ((OID) myID).getInternalObjectId();
            }
            else
            {
                return (new OID(cmd.getFullClassName(), myID)).getInternalObjectId();
            }
        }
    }

	/**
     * If the id is obtained after inserting the object into the database, set
     * new a new id for persistent classes (for example, increment).
     * @param id the id received from the datastore
     */
    public void setPostStoreNewObjectId(Object id)
    {
        JPOXLogger.JDO.info(">> setPostStoreNewObjectId id of type " + id.getClass().getName() + " id=" + id);
        if (cmd.getIdentityType() == IdentityType.APPLICATION)
        {
            try
            {
                myID = null;

                int fieldNumber = 0;
                for (fieldNumber =0; fieldNumber < fieldCount; fieldNumber++)
                {
                    FieldMetaData fmd=cmd.getManagedFieldAbsolute(fieldNumber);
                    if (fmd.isPrimaryKey() &&
                            StoreManager.isStrategyDatastoreAttributed(fmd.getValueStrategy()))
                    {
                        //replace the value of the id, but before convert the value to the field type if needed
                        replaceField(fieldNumber,TypeConversionHelper.convertTo(id,fmd.getType()));
                    }
                }
            }
            catch (Exception e)
            {
                JPOXLogger.RDBMS.error(e);
            }
            finally
            {
                myID = AIDUtils.getNewObjectId(getObject(),cmd);
            }
        }
        else
        {
            myID = new OID(cmd.getFullClassName(), id);
        }
        JPOXLogger.JDO.debug(">> setPostStoreNewObjectId id=" + myID);
    }

    /**
     * return a copy from the object Id
     */
    public Object getExternalObjectId(PersistenceCapable pc)
    {
        if (cmd.getIdentityType() == IdentityType.APPLICATION)
        {
            if (!flushing)
            {
                if (myLC instanceof PersistentNew && !flushedNew && inserting == ActivityState.NONE)
                {
                    int totalFieldCount = cmd.getNoOfInheritedManagedFields() + cmd.getNoOfManagedFields();
                    for (int fieldNumber = 0; fieldNumber < totalFieldCount; fieldNumber++)
                    {
                        FieldMetaData fmd = cmd.getManagedFieldAbsolute(fieldNumber);
                        if (fmd.isPrimaryKey())
                        {
                            if (StoreManager.isStrategyDatastoreAttributed(fmd.getValueStrategy()))
                            {
                                flush();
                                break;
                            }
                        }
                    }
                }
            }

            return AIDUtils.getNewObjectId(myPC, cmd);
        }
        if (cmd.getIdentityType() == IdentityType.DATASTORE)
        {
            if (!flushing)
            {
                if (myLC instanceof PersistentNew && !flushedNew && inserting == ActivityState.NONE)
                {
                    if (StoreManager.isStrategyDatastoreAttributed(cmd.getIdentityMetaData().getStrategyValue()))
                    {
                        flush();
                    }
                }
            }
        }
        if (myID == null)
        {
            return null;
        }

        if (myID instanceof org.jpox.store.OID)
        {
            return ((org.jpox.store.OID) myID).getNewObjectIdCopy();
        }
        else
        {
            return myID;
        }
    }

    /**
     * Replace the current value of jdoStateManager.
     *
     * <P>This method is called by the PersistenceCapable whenever
     * jdoReplaceStateManager is called and there is already
     * an owning StateManager.  This is a security precaution
     * to ensure that the owning StateManager is the only
     * source of any change to its reference in the PersistenceCapable.</p>
     *
     * @return the new value for the jdoStateManager
     * @param pc the calling PersistenceCapable instance
     * @param sm the proposed new value for the jdoStateManager
     */
    public StateManager replacingStateManager(PersistenceCapable pc, StateManager sm)
    {
        if (jdoFlags == PersistenceCapable.DETACHED)
        {
            return sm;
        }

        if (myLC == null)
        {
            throw new JDOFatalInternalException("Null LifeCycleState");
        }

        if (pc == myPC)
        {
            //TODO check if we are really in transition to a transient instance
            if (sm == null)
            {
                return null;
            }
            if (sm == this)
            {
                return this;
            }

            if (this.myPM.getPMHandle() == ((StateManagerImpl) sm).myPM.getPMHandle())
            {
                // This is a race condition when makePersistent or
                // makeTransactional is called on the same PC instance for the
                // same PM. It has been already set to this SM - just 
                // disconnect the other one. Return this SM so it won't be
                // replaced.
                ((StateManagerImpl) sm).disconnect();
                return this;
            }

            if (sm != null)
            {
                throw new JDOFatalInternalException(LOCALISER.msg("StateManager.StateManagerChangeError"));
            }
            if (!disconnecting)
            {
                throw new JDOFatalInternalException(LOCALISER.msg("StateManager.StateManagerDisconnectError"));
            }

            if (JPOXLogger.JDO.isDebugEnabled())
            {
                JPOXLogger.JDO.debug(LOCALISER.msg("StateManager.ClearingStateManager", StringUtils.toJVMIDString(pc)));
            }

            return null;
        }
        else if (pc == savedImage)
        {
            return null;
        }
        else
        {
            return sm;
        }
    }

    /**
     * Return the object representing the JDO identity
     * of the calling instance.  If the JDO identity is being changed in
     * the current transaction, this method returns the current identity as
     * changed in the transaction.
     *
     * @param pc the calling PersistenceCapable instance
     * @return the object representing the JDO identity of the calling instance
     */
    public Object getTransactionalObjectId(PersistenceCapable pc)
    {
        return getObjectId(pc);
    }

    /**
     * Tests whether this object is dirty.
     *
     * Instances that have been modified, deleted, or newly
     * made persistent in the current transaction return true.
     *
     * <P>Transient nontransactional instances return false (JDO spec), but the
     * JPOX implementation does not currently support the transient
     * transactional state.
     *
     * @see PersistenceCapable#jdoMakeDirty(String fieldName)
     * @param pc the calling PersistenceCapable instance
     * @return true if this instance has been modified in current transaction.
     */
    public boolean isDirty(PersistenceCapable pc)
    {
        if (disconnectClone(pc))
        {
            return false;
        }
        else
        {
            return myLC.isDirty();
        }
    }

    /**
     * Tests whether this object is transactional.
     *
     * Instances that respect transaction boundaries return true.  These
     * instances include transient instances made transactional as a result of
     * being the target of a makeTransactional method call; newly made
     * persistent or deleted persistent instances; persistent instances read
     * in data store transactions; and persistent instances modified in
     * optimistic transactions.
     * <P>
     * Transient nontransactional instances return false.
     *
     * @param pc the calling PersistenceCapable instance
     * @return true if this instance is transactional.
     */
    public boolean isTransactional(PersistenceCapable pc)
    {
        if (disconnectClone(pc))
        {
            return false;
        }
        else
        {
            return myLC.isTransactional();
        }
    }

    /**
     * Tests whether this object is persistent.
     * Instances whose state is stored in the data store return true.
     * <P>
     * Transient instances return false.
     *
     * @see PersistenceManager#makePersistent(Object pc)
     * @param pc the calling PersistenceCapable instance
     * @return true if this instance is persistent.
     */
    public boolean isPersistent(PersistenceCapable pc)
    {
        if (disconnectClone(pc))
        {
            return false;
        }
        else
        {
            return myLC.isPersistent();
        }
    }

    /**
     * Tests whether this object has been newly made persistent.
     * Instances that have been made persistent in the current transaction
     * return true.
     * <P>
     * Transient instances return false.
     *
     * @see PersistenceManager#makePersistent(Object pc)
     * @param pc the calling PersistenceCapable instance
     * @return true if this instance was made persistent
     * in the current transaction.
     */
    public boolean isNew(PersistenceCapable pc)
    {
        if (disconnectClone(pc))
        {
            return false;
        }
        else
        {
            return myLC.isNew();
        }
    }

    /**
     * Tests whether this object has been deleted.
     * Instances that have been deleted in the current transaction return true.
     *
     * <P>Transient instances return false.
     *
     * @see PersistenceManager#deletePersistent(Object pc)
     * @param pc the calling PersistenceCapable instance
     * @return true if this instance was deleted
     * in the current transaction.
     */
    public boolean isDeleted(PersistenceCapable pc)
    {
        if (disconnectClone(pc))
        {
            return false;
        }
        else
        {
            return myLC.isDeleted();
        }
    }

    // --------------------------- Load Field Methods --------------------------

    /**
     * Accessor for whether the default fetch group fields are loaded.
     * @return Whether the default fetch group are loaded.
     **/
    private boolean isDFGLoaded()
    {
        for (int i = 0; i < defaultFetchGroupFieldNumbers.length; ++i)
        {
            if (!loadedFields[defaultFetchGroupFieldNumbers[i]])
            {
                return false;
            }
        }

        return true;
    }

    /**
     * Fetchs from the database all fields in the default fetch group not
     * already loaded.  Assumes myLC is already set to the state we're destined
     * for.
     */
    public void loadDFGFields()
    {
        int[] fieldNumbers = getFlagsSetTo(loadedFields, defaultFetchGroupFieldNumbers, false);

        if (fieldNumbers != null)
        {
            srm.fetch(this, fieldNumbers);
            postLoad();
        }
    }

    /**
     * Fetch a given field from the database.  Do NOT call with a default fetch
     * group field.
     *
     * @param fieldNumber   the field number of the field to fetch.
     */
    private void loadNonDFGField(int fieldNumber)
    {
        srm.fetch(this, new int[]{fieldNumber});
    }

    /**
     * Fetchs from the database all fields in the actual fetch plan not currently loaded.
     * Called by life-cycle transitions.
     * 
     * @since 1.1
     */
    void loadUnloadedFieldsInFetchPlan()
    {
        int[] fieldNumbers = getFlagsSetTo(loadedFields, myFP.getFieldsInActualFetchPlan(), false);

        if (fieldNumbers != null)
        {
            boolean callPostLoad = myFP.isToCallPostLoadFetchPlan(this.loadedFields);

            srm.fetch(this, fieldNumbers);

            if (callPostLoad)
            {
                postLoad();
            }
        }
    }

    /**
     * Fetchs from the database all fields in the actual fetch plan.
     * Called by life-cycle transitions.
     * 
     * @since 1.1
     */
    void loadFieldsInFetchPlan()
    {
        int[] fieldNumbers = myFP.getFieldsInActualFetchPlan();

        if (fieldNumbers != null)
        {
            boolean callPostLoad = myFP.isToCallPostLoadFetchPlan(this.loadedFields);

            srm.fetch(this, fieldNumbers);

            if (callPostLoad)
            {
                postLoad();
            }
        }
    }

    /**
     * Refreshes from the database all fields currently loaded.
     * Called by life-cycle transitions.
     */
    void refreshLoadedFields()
    {
        int[] fieldNumbers = getFlagsSetTo(loadedFields, myFP.getFieldsInActualFetchPlan(), true);

        if (fieldNumbers != null)
        {
            clearDirtyFlags();
            clearFlags(loadedFields);

            boolean callPostLoad = myFP.isToCallPostLoadFetchPlan(this.loadedFields);

            srm.fetch(this, fieldNumbers);

            if (callPostLoad)
            {
                postLoad();
            }
        }
    }

    /**
     * Loads all of the default fetch group fields into the instance.
     * Called by pm.getObjectById() when validate == true.
     */
    public void loadDefaultFetchGroup()
    {
        transitionReadField();
        loadDFGFields();
        if (myPC instanceof LoadCallback)
        {
            ((LoadCallback) myPC).jdoPostLoad();
        }
    }

    /**
     * Return true if the field is cached in the calling instance.
     * <P>In the JPOX implementation of this method, isLoaded() will always
     * return true. If the field is not loaded, it will be loaded as a side
     * effect of the call to this method. If it is in the default fetch group,
     * the default fetch group, including this field, will be loaded.
     *
     * @param pc the calling PersistenceCapable instance
     * @param field the absolute field number
     * @return always returns true (this implementation)
     */
    public boolean isLoaded(PersistenceCapable pc, int field)
    {
        if (disconnectClone(pc))
        {
            return true;
        }
        else
        {
            transitionReadField(loadedFields[field]);

            if (!loadedFields[field])
            {
                if (defaultFetchGroupFields[field])
                {
                    loadDFGFields();
                }
                else
                {
                    loadNonDFGField(field);
                }
            }

            return true;
        }
    }

    // ---------------------- Field Accessor/Mutator Methods -------------------

    /**
     * This method is called by the associated PersistenceCapable if the
     * value for the specified field is not cached (StateManager.isLoaded()
     * fails). In this implementation of the StateManager, isLoaded() has a
     * side effect of loading unloaded information and will always return true.
     * As such, this method should never be called.
     */
    public boolean getBooleanField(PersistenceCapable pc, int field, boolean currentValue)
    {
        throw new JDOUnsupportedOptionException("Method not supported");
    }

    /**
     * This method is called by the associated PersistenceCapable if the
     * value for the specified field is not cached (StateManager.isLoaded()
     * fails). In this implementation of the StateManager, isLoaded() has a
     * side effect of loading unloaded information and will always return true.
     * As such, this method should never be called.
     */
    public byte getByteField(PersistenceCapable pc, int field, byte currentValue)
    {
        throw new JDOUnsupportedOptionException("Method not supported");
    }

    /**
     * This method is called by the associated PersistenceCapable if the
     * value for the specified field is not cached (StateManager.isLoaded()
     * fails). In this implementation of the StateManager, isLoaded() has a
     * side effect of loading unloaded information and will always return true.
     * As such, this method should never be called.
     */
    public char getCharField(PersistenceCapable pc, int field, char currentValue)
    {
        throw new JDOUnsupportedOptionException("Method not supported");
    }

    /**
     * This method is called by the associated PersistenceCapable if the
     * value for the specified field is not cached (StateManager.isLoaded()
     * fails). In this implementation of the StateManager, isLoaded() has a
     * side effect of loading unloaded information and will always return true.
     * As such, this method should never be called.
     */

    public double getDoubleField(PersistenceCapable pc, int field, double currentValue)
    {
        throw new JDOUnsupportedOptionException(LOCALISER.msg("StateManager.MethodNotSupported"));
    }

    /**
     * This method is called by the associated PersistenceCapable if the
     * value for the specified field is not cached (StateManager.isLoaded()
     * fails). In this implementation of the StateManager, isLoaded() has a
     * side effect of loading unloaded information and will always return true.
     * As such, this method should never be called.
     */
    public float getFloatField(PersistenceCapable pc, int field, float currentValue)
    {
        throw new JDOUnsupportedOptionException(LOCALISER.msg("StateManager.MethodNotSupported"));
    }

    /**
     * This method is called by the associated PersistenceCapable if the
     * value for the specified field is not cached (StateManager.isLoaded()
     * fails). In this implementation of the StateManager, isLoaded() has a
     * side effect of loading unloaded information and will always return true.
     * As such, this method should never be called.
     */
    public int getIntField(PersistenceCapable pc, int field, int currentValue)
    {
        throw new JDOUnsupportedOptionException(LOCALISER.msg("StateManager.MethodNotSupported"));
    }

    /**
     * This method is called by the associated PersistenceCapable if the
     * value for the specified field is not cached (StateManager.isLoaded()
     * fails). In this implementation of the StateManager, isLoaded() has a
     * side effect of loading unloaded information and will always return true.
     * As such, this method should never be called.
     */
    public long getLongField(PersistenceCapable pc, int field, long currentValue)
    {
        throw new JDOUnsupportedOptionException(LOCALISER.msg("StateManager.MethodNotSupported"));
    }

    /**
     * This method is called by the associated PersistenceCapable if the
     * value for the specified field is not cached (StateManager.isLoaded()
     * fails). In this implementation of the StateManager, isLoaded() has a
     * side effect of loading unloaded information and will always return true.
     * As such, this method should never be called.
     */
    public short getShortField(PersistenceCapable pc, int field, short currentValue)
    {
        throw new JDOUnsupportedOptionException(LOCALISER.msg("StateManager.MethodNotSupported"));
    }

    /**
     * This method is called by the associated PersistenceCapable if the
     * value for the specified field is not cached (StateManager.isLoaded()
     * fails). In this implementation of the StateManager, isLoaded() has a
     * side effect of loading unloaded information and will always return true.
     * As such, this method should never be called.
     */
    public String getStringField(PersistenceCapable pc, int field, String currentValue)
    {
        throw new JDOUnsupportedOptionException(LOCALISER.msg("StateManager.MethodNotSupported"));
    }

    /**
     * This method is called by the associated PersistenceCapable if the
     * value for the specified field is not cached (StateManager.isLoaded()
     * fails). In this implementation of the StateManager, isLoaded() has a
     * side effect of loading unloaded information and will always return true.
     * As such, this method should never be called.
     */
    public Object getObjectField(PersistenceCapable pc, int field, Object currentValue)
    {
        throw new JDOUnsupportedOptionException(LOCALISER.msg("StateManager.MethodNotSupported"));
    }

    /**
     * Called by the various setXXXField() methods once it's established that
     * the field really deserves to be written.
     * Makes the state transition and replaces the field value in the PC
     * instance.
     */
    private void writeField(int field, Object newValue)
    {
        replaceField(field, newValue);
    }

    /**
     * This method is called by the associated PersistenceCapable when the
     * corresponding mutator method (setXXX()) is called on the
     * PersistenceCapable.
     *
     * @param pc the calling PersistenceCapable instance
     * @param field the field number
     * @param currentValue the current value of the field
     * @param newValue the new value for the field
     */
    public void setBooleanField(PersistenceCapable pc, int field, boolean currentValue, boolean newValue)
    {
        if (pc != myPC)
        {
            replaceField(pc, field, newValue ? Boolean.TRUE : Boolean.FALSE);
            disconnectClone(pc);
        }
        else if (myID != null)
        {
            boolean wasDirty = preWriteField(field);
            if (!loadedFields[field] || currentValue != newValue)
            {
                writeField(field, newValue ? Boolean.TRUE : Boolean.FALSE);
            }
            postWriteField(wasDirty);
        }
        else
        {
            replaceField(field, newValue ? Boolean.TRUE : Boolean.FALSE);
        }
    }

    /**
     * This method is called by the associated PersistenceCapable when the
     * corresponding mutator method (setXXX()) is called on the
     * PersistenceCapable.
     *
     * @param pc the calling PersistenceCapable instance
     * @param field the field number
     * @param currentValue the current value of the field
     * @param newValue the new value for the field
     */
    public void setByteField(PersistenceCapable pc, int field, byte currentValue, byte newValue)
    {
        if (pc != myPC)
        {
            replaceField(pc, field, new Byte(newValue));
            disconnectClone(pc);
        }
        else
        {
            boolean wasDirty = preWriteField(field);
            if (!loadedFields[field] || currentValue != newValue)
            {
                writeField(field, new Byte(newValue));
            }
            postWriteField(wasDirty);
        }
    }

    /**
     * This method is called by the associated PersistenceCapable when the
     * corresponding mutator method (setXXX()) is called on the
     * PersistenceCapable.
     *
     * @param pc the calling PersistenceCapable instance
     * @param field the field number
     * @param currentValue the current value of the field
     * @param newValue the new value for the field
     */
    public void setCharField(PersistenceCapable pc, int field, char currentValue, char newValue)
    {
        if (pc != myPC)
        {
            replaceField(pc, field, new Character(newValue));
            disconnectClone(pc);
        }
        else if (myID != null)
        {
            boolean wasDirty = preWriteField(field);
            if (!loadedFields[field] || currentValue != newValue)
            {
                writeField(field, new Character(newValue));
            }
            postWriteField(wasDirty);
        }
        else
        {
            replaceField(field, new Character(newValue));
        }
    }

    /**
     * This method is called by the associated PersistenceCapable when the
     * corresponding mutator method (setXXX()) is called on the
     * PersistenceCapable.
     *
     * @param pc the calling PersistenceCapable instance
     * @param field the field number
     * @param currentValue the current value of the field
     * @param newValue the new value for the field
     */
    public void setDoubleField(PersistenceCapable pc, int field, double currentValue, double newValue)
    {
        if (pc != myPC)
        {
            replaceField(pc, field, new Double(newValue));
            disconnectClone(pc);
        }
        else if (myID != null)
        {
            boolean wasDirty = preWriteField(field);
            if (!loadedFields[field] || currentValue != newValue)
            {
                writeField(field, new Double(newValue));
            }
            postWriteField(wasDirty);
        }
        else
        {
            replaceField(field, new Double(newValue));
        }
    }

    /**
     * This method is called by the associated PersistenceCapable when the
     * corresponding mutator method (setXXX()) is called on the
     * PersistenceCapable.
     *
     * @param pc the calling PersistenceCapable instance
     * @param field the field number
     * @param currentValue the current value of the field
     * @param newValue the new value for the field
     */
    public void setFloatField(PersistenceCapable pc, int field, float currentValue, float newValue)
    {
        if (pc != myPC)
        {
            replaceField(pc, field, new Float(newValue));
            disconnectClone(pc);
        }
        else if (myID != null)
        {
            boolean wasDirty = preWriteField(field);
            if (!loadedFields[field] || currentValue != newValue)
            {
                writeField(field, new Float(newValue));
            }
            postWriteField(wasDirty);
        }
        else
        {
            replaceField(field, new Float(newValue));
        }
    }

    /**
     * This method is called by the associated PersistenceCapable when the
     * corresponding mutator method (setXXX()) is called on the
     * PersistenceCapable.
     *
     * @param pc the calling PersistenceCapable instance
     * @param field the field number
     * @param currentValue the current value of the field
     * @param newValue the new value for the field
     */
    public void setIntField(PersistenceCapable pc, int field, int currentValue, int newValue)
    {
        if (pc != myPC)
        {
            replaceField(pc, field, new Integer(newValue));
            disconnectClone(pc);
        }
        else if (myID != null)
        {
            boolean wasDirty = preWriteField(field);
            if (!loadedFields[field] || currentValue != newValue)
            {
                writeField(field, new Integer(newValue));
            }
            postWriteField(wasDirty);
        }
        else
        {
            replaceField(field, new Integer(newValue));
        }
    }

    /**
     * This method is called by the associated PersistenceCapable when the
     * corresponding mutator method (setXXX()) is called on the
     * PersistenceCapable.
     *
     * @param pc the calling PersistenceCapable instance
     * @param field the field number
     * @param currentValue the current value of the field
     * @param newValue the new value for the field
     */
    public void setLongField(PersistenceCapable pc, int field, long currentValue, long newValue)
    {
        if (pc != myPC)
        {
            replaceField(pc, field, new Long(newValue));
            disconnectClone(pc);
        }
        else if (myID != null)
        {
            boolean wasDirty = preWriteField(field);
            if (!loadedFields[field] || currentValue != newValue)
            {
                writeField(field, new Long(newValue));
            }
            postWriteField(wasDirty);
        }
        else
        {
            replaceField(field, new Long(newValue));
        }
    }

    /**
     * This method is called by the associated PersistenceCapable when the
     * corresponding mutator method (setXXX()) is called on the
     * PersistenceCapable.
     *
     * @param pc the calling PersistenceCapable instance
     * @param field the field number
     * @param currentValue the current value of the field
     * @param newValue the new value for the field
     */
    public void setShortField(PersistenceCapable pc, int field, short currentValue, short newValue)
    {
        if (pc != myPC)
        {
            replaceField(pc, field, new Short(newValue));
            disconnectClone(pc);
        }
        else if (myID != null)
        {
            boolean wasDirty = preWriteField(field);
            if (!loadedFields[field] || currentValue != newValue)
            {
                writeField(field, new Short(newValue));
            }
            postWriteField(wasDirty);
        }
        else
        {
            replaceField(field, new Short(newValue));
        }
    }

    /**
     * This method is called by the associated PersistenceCapable when the
     * corresponding mutator method (setXXX()) is called on the
     * PersistenceCapable.
     *
     * @param pc the calling PersistenceCapable instance
     * @param field the field number
     * @param currentValue the current value of the field
     * @param newValue the new value for the field
     */
    public void setStringField(PersistenceCapable pc, int field, String currentValue, String newValue)
    {
        if (pc != myPC)
        {
            replaceField(pc, field, newValue);
            disconnectClone(pc);
        }
        else if (myID != null)
        {
            boolean wasDirty = preWriteField(field);
            if (!loadedFields[field] || !equals(currentValue, newValue))
            {
                writeField(field, newValue);
            }
            postWriteField(wasDirty);
        }
        else
        {
            replaceField(field, newValue);
        }
    }

    /**
     * This method is called by the associated PersistenceCapable when the
     * corresponding mutator method (setXXX()) is called on the
     * PersistenceCapable.
     *
     * @param pc the calling PersistenceCapable instance
     * @param field the field number
     * @param currentValue the current value of the field
     * @param newValue the new value for the field
     */
    public void setObjectField(PersistenceCapable pc, int field, Object currentValue, Object newValue)
    {
        if (pc != myPC)
        {
            replaceField(pc, field, newValue);
            disconnectClone(pc);
        }
        else if (myID != null)
        {
            boolean wasDirty = preWriteField(field);
            if (!loadedFields[field] || currentValue != newValue)
            {
                if (currentValue instanceof SCO)
                {
                    ((SCO) currentValue).unsetOwner();
                }

                if (newValue instanceof SCO)
                {
                    SCO sco = (SCO) newValue;
                    Object owner = sco.getOwner();

                    if (owner != null)
                    {
                        throw new JDOUserException(LOCALISER.msg("StateManager.SecondClassObjectShareError", sco.getFieldName(), owner));
                    }
                }

                writeField(field, newValue);
            }
            postWriteField(wasDirty);
        }
        else
        {
            replaceField(field, newValue);
        }
    }

    /**
     * This method is called from the associated PersistenceCapable when its
     * PersistenceCapable.jdoProvideFields() method is invoked. Its purpose is
     * to provide the value of the specified field to the StateManager.
     *
     * @param pc   the calling PersistenceCapable instance
     * @param field   the field number
     * @param currentValue   the current value of the field
     */
    public void providedBooleanField(PersistenceCapable pc, int field, boolean currentValue)
    {
        currFM.storeBooleanField(field, currentValue);
    }

    /**
     * This method is called from the associated PersistenceCapable when its
     * PersistenceCapable.jdoProvideFields() method is invoked. Its purpose is
     * to provide the value of the specified field to the StateManager.
     *
     * @param pc   the calling PersistenceCapable instance
     * @param field   the field number
     * @param currentValue   the current value of the field
     */
    public void providedByteField(PersistenceCapable pc, int field, byte currentValue)
    {
        currFM.storeByteField(field, currentValue);
    }

    /**
     * This method is called from the associated PersistenceCapable when its
     * PersistenceCapable.jdoProvideFields() method is invoked. Its purpose is
     * to provide the value of the specified field to the StateManager.
     *
     * @param pc   the calling PersistenceCapable instance
     * @param field   the field number
     * @param currentValue   the current value of the field
     */
    public void providedCharField(PersistenceCapable pc, int field, char currentValue)
    {
        currFM.storeCharField(field, currentValue);
    }

    /**
     * This method is called from the associated PersistenceCapable when its
     * PersistenceCapable.jdoProvideFields() method is invoked. Its purpose is
     * to provide the value of the specified field to the StateManager.
     *
     * @param pc   the calling PersistenceCapable instance
     * @param field   the field number
     * @param currentValue   the current value of the field
     */
    public void providedDoubleField(PersistenceCapable pc, int field, double currentValue)
    {
        currFM.storeDoubleField(field, currentValue);
    }

    /**
     * This method is called from the associated PersistenceCapable when its
     * PersistenceCapable.jdoProvideFields() method is invoked. Its purpose is
     * to provide the value of the specified field to the StateManager.
     *
     * @param pc   the calling PersistenceCapable instance
     * @param field   the field number
     * @param currentValue   the current value of the field
     */
    public void providedFloatField(PersistenceCapable pc, int field, float currentValue)
    {
        currFM.storeFloatField(field, currentValue);
    }

    /**
     * This method is called from the associated PersistenceCapable when its
     * PersistenceCapable.jdoProvideFields() method is invoked. Its purpose is
     * to provide the value of the specified field to the StateManager.
     *
     * @param pc   the calling PersistenceCapable instance
     * @param field   the field number
     * @param currentValue   the current value of the field
     */
    public void providedIntField(PersistenceCapable pc, int field, int currentValue)
    {
        currFM.storeIntField(field, currentValue);
    }

    /**
     * This method is called from the associated PersistenceCapable when its
     * PersistenceCapable.jdoProvideFields() method is invoked. Its purpose is
     * to provide the value of the specified field to the StateManager.
     *
     * @param pc   the calling PersistenceCapable instance
     * @param field   the field number
     * @param currentValue   the current value of the field
     */
    public void providedLongField(PersistenceCapable pc, int field, long currentValue)
    {
        currFM.storeLongField(field, currentValue);
    }

    /**
     * This method is called from the associated PersistenceCapable when its
     * PersistenceCapable.jdoProvideFields() method is invoked. Its purpose is
     * to provide the value of the specified field to the StateManager.
     *
     * @param pc   the calling PersistenceCapable instance
     * @param field   the field number
     * @param currentValue   the current value of the field
     */
    public void providedShortField(PersistenceCapable pc, int field, short currentValue)
    {
        currFM.storeShortField(field, currentValue);
    }

    /**
     * This method is called from the associated PersistenceCapable when its
     * PersistenceCapable.jdoProvideFields() method is invoked. Its purpose is
     * to provide the value of the specified field to the StateManager.
     *
     * @param pc   the calling PersistenceCapable instance
     * @param field   the field number
     * @param currentValue   the current value of the field
     */
    public void providedStringField(PersistenceCapable pc, int field, String currentValue)
    {
        currFM.storeStringField(field, currentValue);
    }

    /**
     * This method is called from the associated PersistenceCapable when its
     * PersistenceCapable.jdoProvideFields() method is invoked. Its purpose is
     * to provide the value of the specified field to the StateManager.
     *
     * @param pc   the calling PersistenceCapable instance
     * @param i    the field number
     * @param currentValue   the current value of the field
     */
    public void providedObjectField(PersistenceCapable pc, int i, Object currentValue)
    {
        currFM.storeObjectField(i, currentValue);
    }

    /**
     * The StateManager uses this method to supply the value of jdoFlags to the
     * associated PersistenceCapable instance.
     *
     * @param pc
     *   the calling PersistenceCapable instance
     * @return
     *   the value of jdoFlags to be stored in the PersistenceCapable instance
     */
    public byte replacingFlags(PersistenceCapable pc)
    {
        // If this is a clone, return READ_WRITE_OK.
        if (pc != myPC)
        {
            return PersistenceCapable.READ_WRITE_OK;
        }
        else
        {
            return jdoFlags;
        }
    }

    /**
     * This method is invoked by the PersistenceCapable object's
     * jdoReplaceField() method to refresh the value of a boolean field.
     *
     * @param pc the calling PersistenceCapable instance
     * @param field the field number
     * @return the new value for the field
     */
    public boolean replacingBooleanField(PersistenceCapable pc, int field)
    {
        boolean value = currFM.fetchBooleanField(field);
        loadedFields[field] = true;

        return value;
    }

    /**
     * This method is invoked by the PersistenceCapable object's
     * jdoReplaceField() method to refresh the value of a byte field.
     *
     * @param obj the calling PersistenceCapable instance
     * @param field the field number
     * @return the new value for the field
     */
    public byte replacingByteField(PersistenceCapable obj, int field)
    {
        byte value = currFM.fetchByteField(field);
        loadedFields[field] = true;

        return value;
    }

    /**
     * This method is invoked by the PersistenceCapable object's
     * jdoReplaceField() method to refresh the value of a char field.
     *
     * @param obj the calling PersistenceCapable instance
     * @param field the field number
     * @return the new value for the field
     */
    public char replacingCharField(PersistenceCapable obj, int field)
    {
        char value = currFM.fetchCharField(field);
        loadedFields[field] = true;

        return value;
    }

    /**
     * This method is invoked by the PersistenceCapable object's
     * jdoReplaceField() method to refresh the value of a double field.
     *
     * @param obj the calling PersistenceCapable instance
     * @param field the field number
     * @return the new value for the field
     */
    public double replacingDoubleField(PersistenceCapable obj, int field)
    {
        double value = currFM.fetchDoubleField(field);
        loadedFields[field] = true;

        return value;
    }

    /**
     * This method is invoked by the PersistenceCapable object's
     * jdoReplaceField() method to refresh the value of a float field.
     *
     * @param obj the calling PersistenceCapable instance
     * @param field the field number
     * @return the new value for the field
     */
    public float replacingFloatField(PersistenceCapable obj, int field)
    {
        float value = currFM.fetchFloatField(field);
        loadedFields[field] = true;

        return value;
    }

    /**
     * This method is invoked by the PersistenceCapable object's
     * jdoReplaceField() method to refresh the value of a int field.
     *
     * @param obj the calling PersistenceCapable instance
     * @param field the field number
     * @return the new value for the field
     */
    public int replacingIntField(PersistenceCapable obj, int field)
    {
        int value = currFM.fetchIntField(field);
        loadedFields[field] = true;

        return value;
    }

    /**
     * This method is invoked by the PersistenceCapable object's
     * jdoReplaceField() method to refresh the value of a long field.
     *
     * @param obj the calling PersistenceCapable instance
     * @param field the field number
     * @return the new value for the field
     */
    public long replacingLongField(PersistenceCapable obj, int field)
    {
        long value = currFM.fetchLongField(field);
        loadedFields[field] = true;

        return value;
    }

    /**
     * This method is invoked by the PersistenceCapable object's
     * jdoReplaceField() method to refresh the value of a short field.
     *
     * @param obj the calling PersistenceCapable instance
     * @param field the field number
     * @return the new value for the field
     */
    public short replacingShortField(PersistenceCapable obj, int field)
    {
        short value = currFM.fetchShortField(field);
        loadedFields[field] = true;

        return value;
    }

    /**
     * This method is invoked by the PersistenceCapable object's
     * jdoReplaceField() method to refresh the value of a String field.
     *
     * @param obj the calling PersistenceCapable instance
     * @param field the field number
     * @return the new value for the field
     */
    public String replacingStringField(PersistenceCapable obj, int field)
    {
        String value = currFM.fetchStringField(field);
        loadedFields[field] = true;

        return value;
    }

    /**
     * This method is invoked by the PersistenceCapable object's
     * jdoReplaceField() method to refresh the value of an Object field.
     *
     * @param obj the calling PersistenceCapable instance
     * @param field the field number
     * @return the new value for the field
     */
    public Object replacingObjectField(PersistenceCapable obj, int field)
    {
        Object value = currFM.fetchObjectField(field);
        loadedFields[field] = true;

        return value;
    }

    public synchronized Object provideField(int fieldNumber)
    {
        return provideField(myPC, fieldNumber);
    }

    public synchronized void replaceField(int fieldNumber, Object value)
    {
        replaceField(myPC, fieldNumber, value);
    }

    private synchronized Object provideField(PersistenceCapable pc, int fieldNumber)
    {
        Object obj;

        FieldManager prevFM = currFM;
        currFM = new SingleFieldValue();

        try
        {
            pc.jdoProvideField(fieldNumber);
            obj = currFM.fetchObjectField(fieldNumber);
        }
        finally
        {
            currFM = prevFM;
        }

        return obj;
    }

    private synchronized void replaceField(PersistenceCapable pc, int fieldNumber, Object value)
    {
        FieldManager prevFM = currFM;
        currFM = new SingleFieldValue();

        try
        {
            currFM.storeObjectField(fieldNumber, value);
            pc.jdoReplaceField(fieldNumber);
        }
        finally
        {
            currFM = prevFM;
        }

        // Notify any owners that have this object embedded
        if (embeddedOwners != null)
        {
            Iterator ownerIter = embeddedOwners.iterator();
            while (ownerIter.hasNext())
            {
                EmbeddedOwnerRelation owner = (EmbeddedOwnerRelation)ownerIter.next();

                boolean wasDirty = ((StateManagerImpl)owner.ownerSM).preWriteField(owner.fieldNumber);
                ((StateManagerImpl)owner.ownerSM).replaceField(owner.fieldNumber, pc);
                ((StateManagerImpl)owner.ownerSM).postWriteField(wasDirty);
            }
        }
    }

    /**
     * Called from the StoreManager after StoreManager.update() is called to
     * obtain updated values from the PersistenceCapable associated with this
     * StateManager.
     *
     * @param fieldNumbers
     *   An array of field numbers to be updated by the Store
     * @param fm
     *   The updated values are stored in this object. This object is only valid
     *   for the duration of this call.
     */
    public synchronized void provideFields(int fieldNumbers[], FieldManager fm)
    {
        FieldManager prevFM = currFM;
        currFM = fm;

        try
        {
            // This will respond by calling this.providedXXXFields() with the value of the field
            myPC.jdoProvideFields(fieldNumbers);
        }
        finally
        {
            currFM = prevFM;
        }
    }

    /**
     * Called from the StoreManager to refresh data in the PersistenceCapable
     * object associated with this StateManager.
     *
     * @param fieldNumbers
     *   An array of field numbers to be refreshed by the Store
     * @param fm
     *   The updated values are stored in this object. This object is only valid
     *   for the duration of this call.
     */
    public synchronized void replaceFields(int fieldNumbers[], FieldManager fm)
    {
        FieldManager prevFM = currFM;
        currFM = fm;

        try
        {
            myPC.jdoReplaceFields(fieldNumbers);
        }
        finally
        {
            currFM = prevFM;
        }
    }

    /**
     * Method to register an owner StateManager with this embedded object.
     * @param ownerSM The owning State Manager.
     * @param fieldNumber The field number that the embedded object is stored as
     */
    protected void addEmbeddedOwner(StateManager ownerSM, int fieldNumber)
    {
        if (ownerSM == null)
        {
            return;
        }

        if (embeddedOwners == null)
        {
            embeddedOwners = new ArrayList();
        }
        embeddedOwners.add(new EmbeddedOwnerRelation(ownerSM, fieldNumber));
    }

    /**
     * Wrapper class storing the owning state manager, and the field of the
     * PC managed by the owning state manager where this object is embedded.
     * @version $Revision: 1.34 $
     */
    private class EmbeddedOwnerRelation
    {
        StateManager ownerSM;
        int fieldNumber;
        
        public EmbeddedOwnerRelation(StateManager ownerSM, int fieldNumber)
        {
            this.ownerSM = ownerSM;
            this.fieldNumber = fieldNumber;
        }
    }

    public Object newSCOInstance(int fieldNumber, Object value)
    {
        if (value instanceof PersistenceCapable)
        {
            FieldMetaData fmd = cmd.getManagedFieldAbsolute(fieldNumber);
            if (fmd.getEmbeddedMetaData() != null)
            {
                StateManagerImpl subSM = (StateManagerImpl)myPM.findStateManager((PersistenceCapable)value);
                if (subSM != null)
                {
                    subSM.addEmbeddedOwner(this, fieldNumber);
                }
                // Assign the embedded object "owner-field" if specified
                if (fmd.getEmbeddedMetaData().getOwnerField() != null)
                {
                    int ownerAbsFieldNum = subSM.cmd.getFieldNumberAbsolute(fmd.getEmbeddedMetaData().getOwnerField());
                    if (ownerAbsFieldNum >= 0)
                    {
                        subSM.replaceField(ownerAbsFieldNum, myPC);
                    }
                }
            }
        }
        if (secondClassMutableFields[fieldNumber] && value != null)
        {
            if (!(value instanceof SCO) || myPC != ((SCO) value).getOwner())
            {
                FieldMetaData fmd = cmd.getManagedFieldAbsolute(fieldNumber);
                SCO newValue = myPM.newSCOInstance(fmd.getType(), myPC, fmd);
                newValue.setValueFrom(value);

                value = newValue;
            }
        }

        return value;
    }

    public void replaceSCOField(int fieldNumber, Object value)
    {
        Object newValue = newSCOInstance(fieldNumber, value);

        if (newValue != value)
        {
            replaceField(fieldNumber, newValue);
        }
    }

    public Object newSCOEmptyInstance(int fieldNumber)
    {
        FieldMetaData fmd = cmd.getManagedFieldAbsolute(fieldNumber);

        Class newType = TypeManager.getTypeManager().getSecondClassWrapper(fmd.getType().getName());
        try
        {
            return newType.newInstance();
        }
        catch (InstantiationException e)
        {
            e.printStackTrace();
        }
        catch (IllegalAccessException e)
        {
            e.printStackTrace();
        }
        return null;
    }

    /**
     * Method to change the object state to read-field.
     * 
     * @param isLoaded if the field was previously loaded
     **/
    private void transitionReadField(boolean isLoaded)
    {
        if (myLC == null)
        {
            return;
        }
        preStateChange();
        try
        {
            myLC = myLC.transitionReadField(this, isLoaded);
        }
        finally
        {
            postStateChange();
        }
    }

    /**
     * Method to change the object state to read-field.
     * 
     **/
    private void transitionReadField()
    {
        transitionReadField(false);
    }

    /**
     * Method to change the object state to write-field.
     **/
    private void transitionWriteField()
    {
        preStateChange();
        try
        {
            myLC = myLC.transitionWriteField(this);
        }
        finally
        {
            postStateChange();
        }
    }

    /**
     * Accessor the value in the datastore for the version column
     * @return Returns the optimisticDatastoreVersion.
     */
    public Object getOptimisticDatastoreVersion()
    {
        return optimisticDatastoreVersion;
    }

    /**
     * Sets the value for the version column in a transaction not yet committed
     * @param optimisticTransactionalVersion
     */
    public void setOptimisticTransactionalVersion(Object optimisticTransactionalVersion)
    {
        this.optimisticTransactionalVersion = optimisticTransactionalVersion;
    }

    /**
     * Sets the value for the version column in the datastore
     * @param optimisticDatastoreVersion
     */
    public void setOptimisticDatastoreVersion(Object optimisticDatastoreVersion)
    {
        this.optimisticDatastoreVersion = optimisticDatastoreVersion;
    }

    /**
     * Accessor for the highest field number in this class
     * @return The highest field number
     */
    public int getHighestFieldNumber()
    {
        return cmd.getNoOfManagedFields() + cmd.getNoOfInheritedManagedFields();
    }

    // ------------------------- Lifecycle Methods -----------------------------

    /**
     * Method to make the object persistent.
     **/
    public void makePersistent()
    {
        // Call any lifecycle listeners waiting for this event.
        ((AbstractPersistenceManager) myPM).notifyLifecycleListenersForEvent(myPC, LifecycleListenerSpecification.EVENT_POST_CREATE);

        // If this is an embedded object becoming persistent in its own right, assign an identity.
        if (myID == null)
        {
            setIdentity();
        }

        dirty = true;
        
        if (myPM.isDelayDatastoreOperationsEnabled())
        {
            myPM.markDirty(this);
        }
        else
        {
            // Call any lifecycle listeners waiting for this event.
            ((AbstractPersistenceManager) myPM).notifyLifecycleListenersForEvent(myPC, LifecycleListenerSpecification.EVENT_PRE_STORE);

            internalMakePersistent();

            // Call any lifecycle listeners waiting for this event.
            ((AbstractPersistenceManager) myPM).notifyLifecycleListenersForEvent(myPC, LifecycleListenerSpecification.EVENT_POST_STORE);            
        }

        registerTransactional();
    }

    /**
     * Method to make the object persistent.
     **/
    public void internalMakePersistent()
    {
        inserting = ActivityState.INSERTING;

        try
        {
            //in InstanceLifecycleEvents this object could get dirty if a field is changed in preStore or
            //postCreate, we clear dirty flags to make sure this object will not be flushed again
            clearDirtyFlags();

            if (myPC instanceof StoreCallback)
            {
                ((StoreCallback) myPC).jdoPreStore();
            }

            srm.insert(this);
            inserting = ActivityState.NONE;
            flushedNew = true;
        }
        finally
        {
            inserting = ActivityState.NONE;
            flushedNew = true;
        }
    }

    /**
     * Tests whether this object is being inserted.
     *
     * @return true if this instance is inserting.
     */
    public boolean isInserting()
    {
        return (inserting == ActivityState.INSERTING);
    }
    
    /**
     * Change the activity state.
     * @param activityState the new state
     */
    public void changeActivityState(ActivityState activityState)
    {
        inserting = activityState;
        if (activityState == ActivityState.INSERTING_CALLBACKS)
        {
            List listenersToFlush = new ArrayList(activityListeners);
            activityListeners.clear();
            for (int i=0; i<listenersToFlush.size(); i++)
            {
                ((org.jpox.StateManager)listenersToFlush.get(i)).flush();
            }
        }
    }    
    
    /**
     * Add activity state listeners
     * @param sm the state manager
     * @param activityState the ActivityState
     */
    public void addActivityListener(StateManager sm,ActivityState activityState)
    {
        activityListeners.add(sm);
    }
    
    /**
     * Makes Transactional Transient instances persistent
     */
    public void makePersistentTransactionalTransient()
    {
        preStateChange();
        try
        {
            if (myLC.isTransactional && !myLC.isPersistent)
            {
                // make the transient instance persistent in the datastore, if is transactional and !persistent 
                makePersistent();
                myLC = myLC.transitionMakePersistent(this);
            }
        }
        finally
        {
            postStateChange();
        }

    }

    /**
     * Method to change the object state to transactional.
     **/
    public void makeTransactional()
    {
        preStateChange();
        try
        {
            if (myLC == null)
            {
                initializeSM(LifeCycleState.T_CLEAN);
                setRestoreValues(true);
            }
            else
            {
                myLC = myLC.transitionMakeTransactional(this);
            }
        }
        finally
        {
            postStateChange();
        }
    }

    /**
     * Method to change the object state to nontransactional.
     **/
    public void makeNontransactional()
    {
        preStateChange();
        try
        {
            myLC = myLC.transitionMakeNontransactional(this);
        }
        finally
        {
            postStateChange();
        }
    }

    /**
     * Method to change the object state to transient.
     **/
    public void makeTransient()
    {
        retrieve(true);

        /*
         * Call unsetOwner() on all loaded SCO fields.
         */
        int[] fieldNumbers = getFlagsSetTo(loadedFields, secondClassMutableFieldNumbers, true);

        if (fieldNumbers != null)
        {
            provideFields(fieldNumbers, new UnsetOwners());
        }
        preStateChange();
        try
        {
            myLC = myLC.transitionMakeTransient(this);
        }
        finally
        {
            postStateChange();
        }
    }

    /**
     * Method to make detached copy of this instance
     * @since 1.1
     * @return the detached PersistenceCapable instance
     */
    public Object detachCopy()
    {
        if (myLC.isDeleted())
        {
            throw new JDOUserException("Can't detach PersistenceCapable object class " + this.cmd.getFullClassName() + " in delete state. : " + myID);
        }
        if (myPC.jdoIsDetached())
        {
            throw new JDOUserException("PersistenceCapable object class " + this.cmd.getFullClassName() + " is already detached.");
        }
        if (dirty)
        {
            myPM.flush();
        }

        // Object in the process of detaching (recursive) so return the object which will be the detached object
        if (detaching)
        {
            return detachingPC;
        }

        // Call any lifecycle listeners waiting for this event.
        ((AbstractPersistenceManager) myPM).notifyLifecycleListenersForEvent(myPC, LifecycleListenerSpecification.EVENT_PRE_DETACH);

        detaching = true;
        loadUnloadedFieldsInFetchPlan();
        PersistenceCapable detachedPC = myPC.jdoNewInstance(this);
        detachingPC = detachedPC;

        StateManagerImpl smDetachedPC = new StateManagerImpl(detachedPC, this.cmd, getExternalObjectId(myPC));
        detachedPC.jdoReplaceStateManager(smDetachedPC);

        // This will cause the detach of any other fields in the FetchPlan.
        internalDetachCopy(this, smDetachedPC, myFP.getFieldsInActualFetchPlan());

        detachedPC.jdoReplaceFlags();
        detachedPC.jdoReplaceObjectId(getObjectId(myPC));
        ((Detachable)detachedPC).jdoReplaceLoadedFieldList();
        ((Detachable)detachedPC).jdoReplaceModifiedFieldList();
        detachedPC.jdoReplaceStateManager(null);
        detaching = false;
        detachingPC = null;

        // Call any lifecycle listeners waiting for this event.
        ((AbstractPersistenceManager) myPM).notifyLifecycleListenersForEvent(detachedPC, LifecycleListenerSpecification.EVENT_POST_DETACH);

        return detachedPC;
    }

    /**
     * Method to detach an object.
     * @since 1.1
     * @param sm StateManager for the object to detach.
     * @param smDetached StateManager for the object detached.
     * @param fieldNumbers the fields to detach.
     */
    public void internalDetachCopy(org.jpox.StateManager sm, org.jpox.StateManager smDetached, int[] fieldNumbers)
    {
        smDetached.replaceFields(fieldNumbers, new DetachFieldManager(sm, secondClassMutableFields));
    }

    /**
     * Attach a persistent object.
     * @since 1.1
     * @param sm StateManager for the object to detach.
     * @param smAttached StateManager for the object attached.
     * @param dirtyFields the dirty fields to attach.
     * @param makeTransactional whether to make the attached object
     * transactional
     */
    public void internalAttachCopy(org.jpox.StateManager sm, org.jpox.StateManager smAttached, boolean[] dirtyFields, boolean makeTransactional, boolean persistent)
    {
        smAttached.provideFields(allFieldNumbers, new AttachFieldManager(sm, secondClassMutableFields, dirtyFields, makeTransactional, persistent));
    }

    /**
     * Method to attach the detached PersistenceCapable instance
     * @since 1.1
     * @param detachedPC the detached PersistenceCapable instance to be attached
     * @param makeTransactional whether to make this as transactional
     */
    public Object attachCopy(PersistenceCapable detachedPC, boolean makeTransactional)
    {
        if (attaching)
        {
            return this;
        }
        attaching = true;

        try
        {
            // Check if the object is already persisted
            boolean persistent = false;
            try
            {
                locate();
                persistent = true;
            }
            catch (JDOObjectNotFoundException onfe)
            {
                // Not currently present!
            }

            // Call any lifecycle listeners waiting for this event.
            ((AbstractPersistenceManager) myPM).notifyLifecycleListenersForEvent(detachedPC, LifecycleListenerSpecification.EVENT_PRE_ATTACH);

            if (detachedPC.jdoIsDeleted())
            {
                myLC = myLC.transitionDeletePersistent(this);
            }
            if (myLC == LifeCycleState.getLifeCycleState(LifeCycleState.HOLLOW))
            {
                myLC = LifeCycleState.getLifeCycleState(LifeCycleState.P_CLEAN);
            }

            StateManagerImpl smDetachedPC = new StateManagerImpl(detachedPC, this.cmd, getExternalObjectId(detachedPC));
            detachedPC.jdoReplaceStateManager(smDetachedPC);
            ((Detachable) detachedPC).jdoProvideLoadedFieldList();
            ((Detachable) detachedPC).jdoProvideModifiedFieldList();
            internalAttachCopy(this, smDetachedPC, smDetachedPC.dirtyFields, makeTransactional, persistent);
            detachedPC.jdoReplaceStateManager(null);

            // Object is not yet persisted! so make it persistent
            if (!persistent)
            {
                makePersistent();
            }

            // notify we are changing the objects
            myLC = myLC.transitionWriteField(this);
            if (makeTransactional)
            {
                myLC = myLC.transitionMakeTransactional(this);
            }

            ((AbstractPersistenceManager) myPM).notifyLifecycleListenersForEvent(myPC, LifecycleListenerSpecification.EVENT_POST_ATTACH);
        }
        finally
        {
            attaching = false;
        }
        return myPC;
    }

    /**
     * Method to delete the object from persistence.
     */
    public void deletePersistent()
    {
        if (!myLC.isDeleted())
        {
            // Call any lifecycle listeners waiting for this event.
            ((AbstractPersistenceManager) myPM).notifyLifecycleListenersForEvent(myPC, LifecycleListenerSpecification.EVENT_PRE_DELETE);

            if (myPC instanceof DeleteCallback)
            {
                ((DeleteCallback) myPC).jdoPreDelete();
            }

            // Perform any cascade on dependent fields
            srm.deleteDependent(this, allFieldNumbers);

            // Update any relationships for fields of this object that aren't dependent
            //provideFields(allFieldNumbers, new DeleteFieldManager(this));

            dirty = true;
            preStateChange();
            try
            {
                myLC = myLC.transitionDeletePersistent(this);
            }
            finally
            {
                postStateChange();
            }
            clearLoadedFlags();
            if (myPM.isDelayDatastoreOperationsEnabled())
            {
                myPM.markDirty(this);
            }
            else
            {
                internalDeletePersistent();
            }

            // Call any lifecycle listeners waiting for this event.
            ((AbstractPersistenceManager) myPM).notifyLifecycleListenersForEvent(myPC, LifecycleListenerSpecification.EVENT_POST_DELETE);
        }
    }

    /**
     * Method to delete the object from persistence.
     */
    public void internalDeletePersistent()
    {
        if (deleting)
        {
            throw new JDOUserException(LOCALISER.msg("StateManager.DeleteRecursiveError"));
        }

        deleting = true;

        try
        {
            if (dirty)
            {
                clearDirtyFlags();
                /*
                 * Clear the PM's knowledge of our being dirty. This will
                 * call our flush() method, which will do nothing.
                 */
                myPM.flush();
            }

            srm.delete(this);
        }
        finally
        {
            deleting = false;
        }
    }

    /**
     * Convenience method to update the object to remove any references to the passed
     * persistence capable object since it is being deleted. Provides a level of
     * managed relationships.
     * @param pc The Object being deleted
     */
    public void removeReferencesToObject(PersistenceCapable pc)
    {
        // Use DeleteFieldManager to check all fields for the object being deleted
        if( removingReference )
        {
            return;
        }
        try
        {
            removingReference = true;
            provideFields(allFieldNumbers, new DeleteFieldManager(this, pc));
        }
        finally
        {
            removingReference = false;
        }
    }

    /**
     * Validates whether the persistence capable instance exists in the
     * datastore. If the instance does not exist in the datastore, this method
     * will fail raising a javax.jdo.JDOObjectNotFoundException.
     */
    public void validate()
    {
        if( !myLC.isTransactional )
        {
	        int[] fieldNumbers = getFlagsSetTo(loadedFields, defaultFetchGroupFieldNumbers, false);
	        if( fieldNumbers != null )
	        {
	            fieldNumbers = getFlagsSetTo(nonPrimaryKeyFields, fieldNumbers, true);
	        }
	        if( fieldNumbers != null )
	        {
	            fieldNumbers = getFlagsSetTo(secondClassMutableFields, fieldNumbers, false);
	        }
	        if( fieldNumbers != null && fieldNumbers.length > 0 )
	        {
	            loadDefaultFetchGroup();
	        }
	        else
	        {
	            locate();
	            transitionReadField();
	        }
        }
    }
    
    /**
     * Locates this object in the datastore. If the instance does not exist in
     * the datastore, this method will fail raising a
     * javax.jdo.JDOObjectNotFoundException.
     */
    private void locate()
    {
        srm.locate(this);
    }
    
    // ------------------------- Object State Methods --------------------------

    /**
     * Method to change the object state to evicted.
     **/
    public void evict()
    {
        if (myLC != LifeCycleState.getLifeCycleState(LifeCycleState.P_CLEAN) &&
            myLC != LifeCycleState.getLifeCycleState(LifeCycleState.P_NONTRANS) )
        {
            return;
        }
            
        preStateChange();
        try
        {
            try
            {
                ((AbstractPersistenceManager) myPM).notifyLifecycleListenersForEvent(myPC, LifecycleListenerSpecification.EVENT_PRE_CLEAR);

                if (myPC instanceof ClearCallback)
                {
                    ((ClearCallback) myPC).jdoPreClear();
                }

                ((AbstractPersistenceManager) myPM).notifyLifecycleListenersForEvent(myPC, LifecycleListenerSpecification.EVENT_POST_CLEAR);
            }
            finally
            {
                myLC = myLC.transitionEvict(this);
            }
        }
        finally
        {
            postStateChange();
        }
    }

    /**
     * Method to refresh the object.
     **/
    public void refresh()
    {
        preStateChange();
        try
        {
            myLC = myLC.transitionRefresh(this);
        }
        finally
        {
            postStateChange();
        }
    }

    /**
     * Method to retrieve the object.
     * @param dfgOnly Only to use default-fetch-group
     **/
    public void retrieve(boolean dfgOnly)
    {
        preStateChange();
        try
        {
            myLC = myLC.transitionRetrieve(this, dfgOnly);
        }
        finally
        {
            postStateChange();
        }
    }

    // --------------------------- Process Methods -----------------------------

    /**
     * Method called before a change in state.
     **/
    private void preStateChange()
    {
        changingState = true;
    }

    /**
     * Method called after a change in state.
     **/
    private void postStateChange()
    {
        changingState = false;

        if (postLoadPending)
        {
            postLoadPending = false;
            postLoad();
        }
    }

    /**
     * Method called before a write of the specified field.
     * @param field The field to write
     * @return true if the field was already dirty before
     **/
    private boolean preWriteField(int field)
    {
        boolean wasDirty = dirty;
        /*
         * If we're writing a field in the process of inserting it must be due 
         * to jdoPreStore().  We haven't actually done the INSERT yet so we 
         * don't want to mark anything as dirty, which would make us want to do 
         * an UPDATE later. 
         */
        if (inserting == ActivityState.NONE)
        {
            dirty = true;
            if (!wasDirty) // (only do it for first dirty event).
            {
                // Call any lifecycle listeners waiting for this event
                ((AbstractPersistenceManager) myPM).notifyLifecycleListenersForEvent(myPC, LifecycleListenerSpecification.EVENT_PRE_DIRTY);
            }

            transitionWriteField();

            dirty = true;
            dirtyFields[field] = true;
            loadedFields[field] = true;
        }
        return wasDirty;
    }

    /**
     * Method called after the write of a field.
     * @param wasDirty whether before writing this field the pc was dirty
     */
    private void postWriteField(boolean wasDirty)
    {
        if (dirty && !wasDirty) // (only do it for first dirty event).
        {
            // Call any lifecycle listeners waiting for this event
            ((AbstractPersistenceManager) myPM).notifyLifecycleListenersForEvent(myPC, LifecycleListenerSpecification.EVENT_POST_DIRTY);
        }
        
        /*
         * If we're writing a field in the middle of flushing or inserting it
         * must be due to jdoPreStore(). To avoid an infinite recursion we don't
         * notify the PM or call flush(). if not transactional transient
         */
        if (inserting == ActivityState.NONE && !flushing && !(myLC.isTransactional && !myLC.isPersistent))
        {
            if (myPM.currentTransaction().isActive())
            {
                myPM.markDirty(this);
            }
            else
            {
                flush();
            }
        }
    }

    /**
     * Called whenever the default fetch group fields have all been loaded.
     * Updates jdoFlags and calls jdoPostLoad() as appropriate.
     * <p>
     * If it's called in the midst of a life-cycle transition both actions will
     * be deferred until the transition is complete.
     * <em>This deferral is important</em>. Without it, we could enter user
     * code (jdoPostLoad()) while still making a state transition, and that way
     * lies madness.
     * <p>
     * As an example, consider a jdoPostLoad() that calls other enhanced methods
     * that read fields (jdoPostLoad() itself is not enhanced). A P_NONTRANS
     * object accessed within a transaction would produce the following infinite
     * loop:
     * <p>
     * <blockquote>
     * 
     * <pre>
     * 
     *  isLoaded()
     *  transitionReadField()
     *  refreshLoadedFields()
     *  jdoPostLoad()
     *  isLoaded()
     *  ...
     *  
     * </pre>
     * 
     * </blockquote>
     * <p>
     * because the transition from P_NONTRANS to P_CLEAN can never be completed.
     */
    private void postLoad()
    {
        if (changingState)
        {
            postLoadPending = true;
        }
        else
        {
            /*
             * A transactional object whose DFG fields are loaded does not need
             * to contact us in order to read those fields, so we can safely set
             * READ_OK.
             *
             * A non-transactional object needs to notify us on all field reads
             * so that we can decide whether or not any transition should occur,
             * so we leave the flags at LOAD_REQUIRED.
             */

            //if (jdoFlags == PersistenceCapable.LOAD_REQUIRED)
            //{
//            if (cmd.getIdentityType() == IdentityType.APPLICATION)
//            {
//                myPC.jdoCopyKeyFieldsToObjectId(myID);
//            }
            //}

            if (jdoFlags == PersistenceCapable.LOAD_REQUIRED && myLC.isTransactional())
            {
                jdoFlags = PersistenceCapable.READ_OK;
                myPC.jdoReplaceFlags();
            }

            if (myPC instanceof LoadCallback)
            {
                ((LoadCallback) myPC).jdoPostLoad();
            }

            // Call any lifecycle listeners waiting for this event.
            ((AbstractPersistenceManager)myPM).notifyLifecycleListenersForEvent(myPC, LifecycleListenerSpecification.EVENT_POST_LOAD);

        }

    }

    /**
     * This method is invoked when a commit is performed in a Transaction
     * involving the PersistenceCapable managed by this StateManager
     */
    public void postCommit(Transaction tx)
    {
        preStateChange();
        this.optimisticDatastoreVersion = this.optimisticTransactionalVersion;
        try
        {
            myLC = myLC.transitionCommit(this, tx);
        }
        finally
        {
            postStateChange();
        }
    }

    /**
     * This method is invoked when a rollback is performed in a Transaction
     * involving the PersistenceCapable managed by this StateManager.
     */
    public void preRollback()
    {
        preStateChange();
        try
        {
            myPM.clearDirty(this);
            myLC = myLC.transitionRollback(this);
        }
        finally
        {
            postStateChange();
        }
    }

    /**
     * Guarantee that the serializable transactional and persistent fields
     * are loaded into the instance.  This method is called by the generated
     * jdoPreSerialize method prior to serialization of the instance.
     *
     * @param pc the calling PersistenceCapable instance
     */
    public void preSerialize(PersistenceCapable pc)
    {
        if (disconnectClone(pc))
        {
            return;
        }

        retrieve(false);
    }

    /**
     * Flushes any dirty fields to the data store.
     *
     * <p>Note that "dirty" in this case is not equated to being in the P_DIRTY
     * state.  The P_DIRTY state means that at least one field in the object has
     * been written by the user during the current transaction, whereas for the
     * purposes of this method, a field is "dirty" if it's been written by the
     * user but not yet updated in the data store.  The difference is, it's
     * possible for an object's state to be P_DIRTY, yet have no "dirty" fields
     * because flush() has been called at least once during the transaction.
     */
    public void flush()
    {
        if (dirty)
        {
            if (flushing)
            {
                // In the case of persisting a new object using autoincrement id within an optimistic
                // transaction, flush() will initially be called at the point of recognising that the
                // id is generated in the datastore, and will then be called again at the point of doing
                // the InsertRequest for the object itself. Just return since we are flushing right now
                return;

                // throw new JDOFatalInternalException(LOCALISER.msg("StateManager.FlushRecursiveError"));
            }

            flushing = true;
            try
            {
                if (myLC instanceof PersistentNew && !flushedNew)
                {
                    // Call any lifecycle listeners waiting for this event.
                    ((AbstractPersistenceManager)myPM).notifyLifecycleListenersForEvent(myPC, LifecycleListenerSpecification.EVENT_PRE_STORE);

                    internalMakePersistent();
                    dirty = false;

                    // Call any lifecycle listeners waiting for this event.
                    ((AbstractPersistenceManager)myPM).notifyLifecycleListenersForEvent(myPC, LifecycleListenerSpecification.EVENT_POST_STORE);                    
                }
                else if (myLC instanceof PersistentDeleted)
                {
                    // Call any lifecycle listeners waiting for this event.
                    ((AbstractPersistenceManager)myPM).notifyLifecycleListenersForEvent(myPC, LifecycleListenerSpecification.EVENT_PRE_DELETE);
                    
                    internalDeletePersistent();

                    // Call any lifecycle listeners waiting for this event.
                    ((AbstractPersistenceManager)myPM).notifyLifecycleListenersForEvent(myPC, LifecycleListenerSpecification.EVENT_POST_DELETE);                    
                }
                else if (myLC instanceof PersistentNewDeleted)
                {
                    if (flushedNew)
                    {
                        // Call any lifecycle listeners waiting for this event.
                        ((AbstractPersistenceManager)myPM).notifyLifecycleListenersForEvent(myPC, LifecycleListenerSpecification.EVENT_PRE_DELETE);

                        internalDeletePersistent();

                        // Call any lifecycle listeners waiting for this event.
                        ((AbstractPersistenceManager)myPM).notifyLifecycleListenersForEvent(myPC, LifecycleListenerSpecification.EVENT_POST_DELETE);                    
                    }
                    else
                    {
                        dirty = false;
                    }
                }
                else
                {
                    // Call any lifecycle listeners waiting for this event.
                    ((AbstractPersistenceManager)myPM).notifyLifecycleListenersForEvent(myPC, LifecycleListenerSpecification.EVENT_PRE_STORE);

                    if (myPC instanceof StoreCallback && !deleting)
                    {
                        ((StoreCallback) myPC).jdoPreStore();
                    }

                    int[] dirtyFieldNumbers = getFlagsSetTo(dirtyFields, true);
                    if (dirtyFieldNumbers == null)
                    {
                        throw new JDOFatalInternalException(LOCALISER.msg("StateManager.InconsistentDirtyStateError"));
                    }
                    srm.update(this, dirtyFieldNumbers);

                    clearDirtyFlags();
                    
                    // Call any lifecycle listeners waiting for this event.
                    ((AbstractPersistenceManager)myPM).notifyLifecycleListenersForEvent(myPC, LifecycleListenerSpecification.EVENT_POST_STORE);
                    
                }

            }
            finally
            {
                flushing = false;
            }

        }
    }

    /**
     * Initialize SM reference in PC and Oid
     * @param newState The new StateManager state 
     **/
    private void initializeSM(int newState)
    {
        final StateManagerImpl thisSM = this;
        myLC = LifeCycleState.getLifeCycleState(newState);

        try
        {
            if (myLC.isPersistent())
            {
                myPM.addStateManager(this);
            }

            // Everything OK so far. Now we can set SM reference in PC 
            // It can be done only after myLC is set to deligate validation
            // to the LC and objectId verified for uniqueness
            AccessController.doPrivileged(
            // Need to have privileges to perform jdoReplaceStateManager.
                    new PrivilegedAction()
                    {
                        public Object run()
                        {
                            try
                            {
                                myPC.jdoReplaceStateManager(thisSM);
                                return null;
                            }
                            catch (SecurityException e)
                            {
                                throw new JDOFatalUserException("EXC_CannotSetStateManager", e);
                            }
                        }
                    });

        }
        catch (SecurityException e)
        {
            throw new JDOUserException(e.getMessage());

        }
        catch (JDOException jdoException)
        {
            if (myPM.getStateManagerById(myID) == this)
            {
                myPM.removeStateManager(this);
            }
            throw jdoException;
        }
    }

    /**
     * Method to disconnect any cloned persistence capable objects from their
     * StateManager.
     * @param pc The PersistenceCapable object
     * @return Whether the object was disconnected.
     **/
    private boolean disconnectClone(PersistenceCapable pc)
    {
        if (pc != myPC)
        {
            if (JPOXLogger.JDO.isDebugEnabled())
            {
                JPOXLogger.JDO.debug(LOCALISER.msg("StateManager.DisconnectClone", StringUtils.toJVMIDString(pc), this));
            }

            // Reset jdoFlags in the clone to PersistenceCapable.READ_WRITE_OK 
            // and clear its state manager.
            pc.jdoReplaceFlags();
            pc.jdoReplaceStateManager(null);
            return true;
        }
        else
        {
            return false;
        }
    }

    /**
     * Disconnect the StateManager from the PersistenceManager
     **/
    public void disconnect()
    {
        if (JPOXLogger.JDO.isDebugEnabled())
        {
            JPOXLogger.JDO.debug(LOCALISER.msg("StateManager.Disconnect", StringUtils.toJVMIDString(myPC), this));
        }

        // Call unsetOwner() on all loaded SCO fields.
        int[] fieldNumbers = getFlagsSetTo(loadedFields, secondClassMutableFieldNumbers, true);
        if (fieldNumbers != null)
        {
            provideFields(fieldNumbers, new UnsetOwners());
        }

        myPM.removeStateManager(this);
        jdoFlags = PersistenceCapable.READ_WRITE_OK;
        myPC.jdoReplaceFlags();

        disconnecting = true;
        try
        {
            replaceStateManager(null);
        }
        finally
        {
            disconnecting = false;
        }

        clearSavedFields();
        myPM = null;
        myFP = null;
        myPC = null;
        myID = null;
        myLC = null;
        cmd = null;
        srm = null;
    }

    /**
     * Registers the pc class in the cache
     */
    public void registerTransactional()
    {
        myPM.addStateManager(this);
    }

    // ------------------------------ Detach Methods ---------------------------

    /* (non-Javadoc)
     * @see javax.jdo.spi.StateManager#providedModifiedFields(java.util.BitSet)
     */
    public void providedModifiedFields(BitSet modifiedFields)
    {
        jdoModifiedFields = modifiedFields;
        for (int i = 0; i < dirtyFields.length; i++)
        {
            dirtyFields[i] = jdoModifiedFields.get(i);
        }
    }

    /* (non-Javadoc)
     * @see javax.jdo.spi.StateManager#providedLoadedFields(java.util.BitSet)
     */
    public void providedLoadedFields(BitSet loadedFields)
    {
        jdoLoadedFields = loadedFields;
        for (int i = 0; i < this.loadedFields.length; i++)
        {
            this.loadedFields[i] = jdoLoadedFields.get(i);
        }
    }

    /* (non-Javadoc)
     * @see javax.jdo.spi.StateManager#replacingModifiedFields(java.util.BitSet)
     */
    public BitSet replacingModifiedFields(BitSet currentModifiedFields)
    {
        if (currentModifiedFields == null)
        {
            jdoModifiedFields = new BitSet();
            for (int i = 0; i < dirtyFields.length; i++)
            {
                if (dirtyFields[i])
                {
                    jdoModifiedFields.set(i);
                }
                else
                {
                    jdoModifiedFields.clear(i);
                }
            }
        }
        else
        {
            jdoModifiedFields = currentModifiedFields;
        }
        return jdoModifiedFields;
    }

    /* (non-Javadoc)
     * @see javax.jdo.spi.StateManager#replacingLoadedFields(java.util.BitSet)
     */
    public BitSet replacingLoadedFields(BitSet currentLoadedFields)
    {
        if (currentLoadedFields == null)
        {
            jdoLoadedFields = new BitSet();
            for (int i = 0; i < loadedFields.length; i++)
            {
                if (loadedFields[i])
                {
                    jdoLoadedFields.set(i);
                }
                else
                {
                    jdoLoadedFields.clear(i);
                }
            }
        }
        else
        {
            jdoLoadedFields = currentLoadedFields;
        }
        return jdoLoadedFields;

    }

    // ------------------------------ Helper Methods ---------------------------

    /**
     * Compares two objects for equality, where one or both of the object
     * references may be null.
     *
     * @return  <code>true</code> if the objects are both <code>null</code> or
     *          compare equal according to their equals() method,
     *          <code>false</code> otherwise.
     */
    private static boolean equals(Object o1, Object o2)
    {
        return o1 == null ? (o2 == null) : o1.equals(o2);
    }

    /**
     * Utility to clear the supplied flags.
     * @param flags
     **/
    private static void clearFlags(boolean[] flags)
    {
        for (int i = 0; i < flags.length; i++)
        {
            flags[i] = false;
        }
    }

    /**
     * Returns an array of integers containing the indices of all elements in
     * <tt>flags</tt> that are in the <tt>state</tt> passed as argument.
     * @param flags Array of flags (true or false)
     * @param state The state to search (true or false)
     * @return The settings of the flags
     */
    private static int[] getFlagsSetTo(boolean[] flags, boolean state)
    {
        int[] temp = new int[flags.length];
        int j = 0;

        for (int i = 0; i < flags.length; i++)
        {
            if (flags[i] == state)
            {
                temp[j++] = i;
            }
        }

        if (j != 0)
        {
            int[] fieldNumbers = new int[j];
            System.arraycopy(temp, 0, fieldNumbers, 0, j);

            return fieldNumbers;
        }
        else
        {
            return null;
        }
    }

    /**
     * Returns an array of integers containing the indices of all elements in
     * <tt>flags</tt> whose index occurs in <tt>indices</tt> and whose value is
     * <tt>state</tt>.
     */
    private static int[] getFlagsSetTo(boolean[] flags, int[] indices, boolean state)
    {
        int[] temp = new int[indices.length];
        int j = 0;

        for (int i = 0; i < indices.length; i++)
        {
            if (flags[indices[i]] == state)
            {
                temp[j++] = indices[i];
            }
        }

        if (j != 0)
        {
            int[] fieldNumbers = new int[j];
            System.arraycopy(temp, 0, fieldNumbers, 0, j);

            return fieldNumbers;
        }
        else
        {
            return null;
        }
    }

    /**
     * Utility to take a peek at a field in the PersistenceCapable object.
     * @param obj The PersistenceCapable object
     * @param fieldName The field to peek at
     * @return The value of the field.
     **/
    private static Object peekField(Object obj, String fieldName)
    {
        try
        {
            /*
             * This doesn't work due to security problems but you get the idea.
             * I'm trying to get field values directly without going through
             * the provideField machinery.
             */
            Object value = obj.getClass().getDeclaredField(fieldName).get(obj);
            if (value instanceof PersistenceCapable)
            {
                return StringUtils.toJVMIDString(value);
            }
            else
            {
                return value;
            }
        }
        catch (Exception e)
        {
            return e.toString();
        }
    }

    /**
     * Method to dump a PersistenceCapable object to the specified
     * PrintWriter.
     * @param pc The PersistenceCapable object
     * @param out The PrintWriter
     **/
    private static void dumpPC(PersistenceCapable pc, PrintWriter out)
    {
        out.println(StringUtils.toJVMIDString(pc));

        if (pc == null)
        {
            return;
        }

        out.print("jdoStateManager = " + peekField(pc, "jdoStateManager"));
        out.print("jdoFlags = ");
        Object flagsObj = peekField(pc, "jdoFlags");
        if (flagsObj instanceof Byte)
        {
            out.println(StringUtils.jdoFlagsToString(((Byte) flagsObj).byteValue()));
        }
        else
        {
            out.println(flagsObj);
        }

        Class c = pc.getClass();

        do
        {
            String[] fieldNames = HELPER.getFieldNames(c);

            for (int i = 0; i < fieldNames.length; ++i)
            {
                out.print(fieldNames[i]);
                out.print(" = ");
                out.println(peekField(pc, fieldNames[i]));
            }

            c = c.getSuperclass();
        }
        while (c != null && PersistenceCapable.class.isAssignableFrom(c));
    }

    /**
     * Utility to dump the contents of the StateManager.
     *
     * @param out PrintWriter to dump to
     **/
    public void dump(PrintWriter out)
    {
        out.println("myPM = " + myPM);
        out.println("myID = " + myID);
        out.println("myLC = " + myLC);
        out.println("cmd = " + cmd);
        out.println("srm = " + srm);
        out.println("fieldCount = " + fieldCount);
        out.println("dirty = " + dirty);
        out.println("flushing = " + flushing);
        out.println("changingState = " + changingState);
        out.println("postLoadPending = " + postLoadPending);
        out.println("disconnecting = " + disconnecting);
        out.println("dirtyFields = " + StringUtils.booleanArrayToString(dirtyFields));
        out.println("defaultFetchGroupFields = " + StringUtils.booleanArrayToString(defaultFetchGroupFields));
        out.println("secondClassMutableFields = " + StringUtils.booleanArrayToString(secondClassMutableFields));
        out.println("allFieldNumbers = " + StringUtils.intArrayToString(allFieldNumbers));
        out.println("defaultFetchGroupFieldNumbers = " + StringUtils.intArrayToString(defaultFetchGroupFieldNumbers));
        out.println("secondClassMutableFieldNumbers = " + StringUtils.intArrayToString(secondClassMutableFieldNumbers));

        out.println();
        out.println("jdoFlags = " + StringUtils.jdoFlagsToString(jdoFlags));
        out.println("loadedFields = " + StringUtils.booleanArrayToString(loadedFields));
        out.print("myPC = ");
        dumpPC(myPC, out);

        out.println();
        out.println("savedFlags = " + StringUtils.jdoFlagsToString(savedFlags));
        out.println("savedLoadedFields = " + StringUtils.booleanArrayToString(savedLoadedFields));

        out.print("savedImage = ");
        dumpPC(savedImage, out);
    }
}
