/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution.

Contributors:
2004 Erik Bengtson - AdapterTime methods
   ...
**********************************************************************/
package org.jpox.store;

import java.sql.Timestamp;
import java.util.HashSet;

import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.MappingManager;
import org.jpox.util.Localiser;

/**
 * Generalised datastore representation. Each datastore is assumed to have a
 * "name", and a "version" (of the form major.minor.revision) and will store
 * "identifiers". In addition, each field will have a Mapping to the datastore
 * <P>
 * This will typically be extended to provide e.g RDBMS connection using JDBC,
 * or XML file connection using a DOM parser.
 * </P>
 * @version $Revision: 1.4 $
 */
public abstract class DatastoreAdapter
{
    protected static final Localiser LOCALISER_BASE = Localiser.getInstance("org.jpox.store.Localisation");

    /**
     * The set of reserved keywords for this datastore.
     */
    protected final HashSet reservedKeywords = new HashSet();

    /**
     * The name of the underlying datastore.
     */
    protected String datastoreProductName;

    /**
     * The version number of the underlying datastore as a string.
     */
    protected String datastoreProductVersion;

    /**
     * The major version number of the underlying datastore.
     */
    protected int datastoreMajorVersion;

    /**
     * The minor version number of the underlying datastore.
     */
    protected int datastoreMinorVersion;

    /**
     * The revision version number of the underlying datastore.
     */
    protected int datastoreRevisionVersion = 0;

    /**
     * Whether the datastore stores all identifiers in lower-case.
     */
    protected boolean storesLowerCaseIdentifiers = false;

    /**
     * Whether the datastore stores all identifiers in mixed-case.
     */
    protected boolean storesMixedCaseIdentifiers = true;

    /**
     * Whether the datastore stores all identifiers in upper-case.
     */
    protected boolean storesUpperCaseIdentifiers = false;

    /**
     * Whether the datastore stores all identifiers in lower-case.
     */
    protected boolean storesLowerCaseQuotedIdentifiers = false;

    /**
     * Whether the datastore stores all identifiers in mixed-case.
     */
    protected boolean storesMixedCaseQuotedIdentifiers = true;

    /**
     * Whether the datastore stores all identifiers in upper-case.
     */
    protected boolean storesUpperCaseQuotedIdentifiers = false;

    /**
     * The String used to quote identifiers.
     */
    protected String identifierQuoteString;

    /**
     * The dictionary to use for identifiers etc.
     */
    protected Dictionary dictionary;

    /**
     * Manager for the mapping between Java and datastore types.
     */
    protected MappingManager mappingManager;

    /**
     * Accessor for the Mapping Manager for field mapping management.
     * @return The Mapping Manager.
     */
    public abstract MappingManager getMappingManager();

    /**
     * Accessor for the Vendor ID for this datastore.
     * @return Vendor id for this datastore
     */
    public String getVendorID()
    {
        return null;
    }

    /**
     * Method to check if a word is reserved for this datastore.
     * @param word The word
     * @return Whether it is reserved
     */
    public boolean isReservedKeyword(String word)
    {
        return reservedKeywords.contains(word);
    }

    /**
     * Accessor for whether the datastore stores lowercase identifiers.
     * @return Whether the datastore stores lowercase identifiers.
     */
    public boolean storesLowerCaseIdentifiers()
    {
        return storesLowerCaseIdentifiers;
    }

    /**
     * Accessor for whether the datastore stores mixedcase identifiers.
     * @return Whether the datastore stores mixedcase identifiers.
     */
    public boolean storesMixedCaseIdentifiers()
    {
        return storesMixedCaseIdentifiers;
    }

    /**
     * Accessor for whether the datastore stores uppercase identifiers.
     * @return Whether the datastore stores uppercase identifiers.
     */
    public boolean storesUpperCaseIdentifiers()
    {
        return storesUpperCaseIdentifiers;
    }

    /**
     * Accessor for whether the datastore stores lowercase identifiers.
     * @return Whether the datastore stores lowercase identifiers.
     */
    public boolean storesLowerCaseQuotedIdentifiers()
    {
        return storesLowerCaseQuotedIdentifiers;
    }

    /**
     * Accessor for whether the datastore stores mixedcase identifiers.
     * @return Whether the datastore stores mixedcase identifiers.
     */
    public boolean storesMixedCaseQuotedIdentifiers()
    {
        return storesMixedCaseQuotedIdentifiers;
    }

    /**
     * Accessor for whether the datastore stores uppercase identifiers.
     * @return Whether the datastore stores uppercase identifiers.
     */
    public boolean storesUpperCaseQuotedIdentifiers()
    {
        return storesUpperCaseQuotedIdentifiers;
    }

    /**
     * Method to return an identifier in the datastore adapters required case
     * (with quotes if necessary).
     * @param identifierName The name
     * @return The identifier in the correct case and quotes
     */
    public String getIdentifierInAdapterCase(String identifierName)
    {
        if (storesUpperCaseIdentifiers)
        {
            return identifierName.toUpperCase();
        }
        else if (storesLowerCaseIdentifiers)
        {
            return identifierName.toLowerCase();
        }
        else if (storesMixedCaseIdentifiers)
        {
            return identifierName;
        }
        else if (storesLowerCaseQuotedIdentifiers)
        {
            return identifierQuoteString + identifierName.toLowerCase() + identifierQuoteString;
        }
        else if (storesMixedCaseQuotedIdentifiers)
        {
            return identifierQuoteString + identifierName + identifierQuoteString;
        }
        else if (storesUpperCaseQuotedIdentifiers)
        {
            return identifierQuoteString + identifierName.toUpperCase() + identifierQuoteString;
        }
        return identifierName;
    }

    /**
     * Accessor for the mapping for the specified class.
     * @param c Class to query
     * @return The mapping for the class.
     */
    public abstract JavaTypeMapping getMapping(Class c);

    /**
     * Accessor for the mapping for the specified class.
     * @param c Class to query
     * @param storeMgr The Store Manager
     * @return The mapping for the class.
     */
    public abstract JavaTypeMapping getMapping(Class c, StoreManager storeMgr);

    /**
     * Accessor for whether the class is an embedded type.
     * @param c Class to query
     * @return Whether the class is an embedded type
     **/
    public abstract boolean isEmbeddedType(Class c);
    
    /**
     * Accessor for the Dictionary.
     * @return Returns the dictionary.
     */
    public Dictionary getDictionary()
    {
        return dictionary;
    }

    /**
     * Mutator for the Dictionary.
     * @param dictionary The dictionary to set.
     */
    public void setDictionary(Dictionary dictionary)
    {
        this.dictionary = dictionary;
    }

    /**
     * Utility to return the adapter time in case there are rounding issues with
     * millisecs etc.
     * @param time The timestamp
     * @return The time in millisecs
     */
    public long getAdapterTime(Timestamp time)
    {
        long timestamp = getTime(time.getTime(), time.getNanos());
        int ms = getMiliseconds(time.getNanos());

        return timestamp + ms;
    }

    protected long getTime(long time, long nanos)
    {
        if (nanos < 0)
        {
            return (((time / 1000) - 1) * 1000);
        }
        return (time / 1000) * 1000;
    }

    protected int getMiliseconds(long nanos)
    {
        return (int) (nanos / 1000000);
    }

    // ------------------------------ Utilities --------------------------------

    /**
     * Method to return this object as a string.
     * @return String version of this object.
     */
    public String toString()
    {
        StringBuffer str = new StringBuffer();
        String className = getClass().getName();
        String name = className.substring(className.lastIndexOf('.') + 1);

        str.append(name + " : " + datastoreProductName + " version=" + datastoreProductVersion + 
            ", major=" + datastoreMajorVersion + ", minor=" + datastoreMinorVersion + ", revision=" + datastoreRevisionVersion + "\n");
        str.append("Identifier Names supported : " +
            (storesLowerCaseIdentifiers ? "lowercase " : "") +
            (storesMixedCaseIdentifiers ? "MixedCase " : "") +
            (storesUpperCaseIdentifiers ? "UPPERCASE " : "") +
            (storesLowerCaseQuotedIdentifiers ? "\"lowercase\" " : "") +
            (storesMixedCaseQuotedIdentifiers ? "\"MixedCase\" " : "") +
            (storesUpperCaseQuotedIdentifiers ? "\"UPPERCASE\" " : ""));

        return str.toString();
    }
}