/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
    ...
**********************************************************************/
package org.jpox.store;

import org.jpox.StateManager;
import org.jpox.metadata.FieldMetaData;
import org.jpox.store.mapping.JavaTypeMapping;

/**
 * Representation of a Java class in a datastore.
 * In the case of RDBMS this will be a table. 
 * In the case of a file-based structure this may be a directory.
 * In the case of an XML-based structure this may be an element.
 *
 * @version $Revision: 1.4 $
 **/
public interface DatastoreClass extends DatastoreContainerObject
{
    /**
     * Accessor for the class represented.
     * @return Name of the class
     **/
    String getType();

    /**
     * Accessor for the name of the datastore class (table).
     * @return The name
     */
    String toString();

    /**
     * Accessor for a mapping for the ID (PersistenceCapable) for this object.
     * @return The (PersistenceCapable) mapping.
     **/
    JavaTypeMapping getIDMapping();

    /**
     * Accessor for a mapping for the datastore ID (OID) for this object.
     * @return The (OID) mapping.
     **/
    JavaTypeMapping getDataStoreObjectIdMapping();

    /**
     * Accessor for the mapping for the specified FieldMetaData.
     * @param fmd the FieldMetaData
     * @return The Mapping for the field.
     */
    JavaTypeMapping getFieldMapping(FieldMetaData fmd);

    /**
     * Accessor for the mapping for the specified field name.
     * @param fieldName Name of field
     * @return The Mapping for the field.
     */
    JavaTypeMapping getFieldMapping(String fieldName);

    /**
     * Method to insert an object for this class.
     * Will insert any superclass records as well (recursive).
     * @param sm StateManager for the object to insert.
     **/
    void insert(StateManager sm);
    
    /**
     * Method to fetch an object for this class.
     * Will fetch any superclass records as well (recursive).
     * @param sm StateManager for the object to fetch.
     * @param fieldNumbers The field numbers to be fetched
     */
    void fetch(StateManager sm, int[] fieldNumbers);

    /**
     * Method to update an object for this class.
     * Will update any superclass records as well (recursive).
     * @param sm StateManager for the object to update.
     * @param fieldNumbers the field numbers to be updated
     */
    void update(StateManager sm, int[] fieldNumbers);

    /**
     * Method to delete an object for this class.
     * Will delete any superclass records as well (recursive).
     * @param sm StateManager for the object to delete.
     **/
    void delete(StateManager sm);

    /**
     * Locates this object in the datastore. If the instance does not exist in
     * the datastore, this method will fail raising a
     * javax.jdo.JDOObjectNotFoundException.
     */
    void locate(StateManager sm);

    /**
     * Delete dependent objects of a persistent object.
     * @since 1.1
     * @param sm StateManager for the object being deleted.
     * @param fieldNumbers the fields to delete.
     */
    void deleteDependent(StateManager sm,int[] fieldNumbers);

    /**
     * Acessor for the FieldMetaData of a field.
     * @param fieldName the field name
     * @return the FieldMetaData
     */
    FieldMetaData getFieldMetaData(String fieldName);

    void prepareDatastoreMapping(JavaTypeMapping m, FieldMetaData fmd);
}