/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved.
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
    ...
**********************************************************************/
package org.jpox.store;

import java.io.IOException;
import java.util.Properties;

import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;

/**
 * Contains the dictionary configuration for identifiers.
 *
 * @version $Revision: 1.1 $ 
 */
public class Dictionary
{
    private static final Localiser LOCALISER = Localiser.getInstance("org.jpox.Localisation");

	public static final String STORE_DICTIONARY_CASE_IDENTIFIER = "org.jpox.store.Dictionary.CaseIdentifier";
    
    private CaseIdentifier caseIdentifier;
    private Naming generalNaming;
    
    public static class Naming
    {
        /**
         * separator between words. 
         * e.g. _ is the separator for: WORDA_WORDB_WORDC
         */
        private String wordSeparator;
        private String prefix;
        private String suffix;

        /**
         * @return Returns the prefix.
         */
        public String getPrefix()
        {
            return prefix;
        }

        /**
         * @param prefix
         *            The prefix to set.
         */
        public void setPrefix(String prefix)
        {
            this.prefix = prefix;
        }

        /**
         * @return Returns the suffix.
         */
        public String getSuffix()
        {
            return suffix;
        }

        /**
         * @param suffix
         *            The suffix to set.
         */
        public void setSuffix(String suffix)
        {
            this.suffix = suffix;
        }

        /**
         * @return Returns the wordSeparator.
         */
        public String getWordSeparator()
        {
            return wordSeparator;
        }

        /**
         * @param wordSeparator
         *            The wordSeparator to set.
         */
        public void setWordSeparator(String wordSeparator)
        {
            this.wordSeparator = wordSeparator;
        }
    }

    public static class CaseIdentifier
    {
        private static final String CASE_PRESERVE = "PRESERVECASE";
        private static final String CASE_UPPER = "UPPERCASE";
        private static final String CASE_LOWER = "LOWERCASE";
        private static final String CASE_SENTENCE = "SENTENCECASE";
        private static final String CASE_CAPITALIZE = "CAPITALIZECASE";
        
        public static final CaseIdentifier PRESERVECASE = new CaseIdentifier(CASE_PRESERVE);
        public static final CaseIdentifier UPPERCASE = new CaseIdentifier(CASE_UPPER);
        public static final CaseIdentifier LOWERCASE = new CaseIdentifier(CASE_LOWER);
        public static final CaseIdentifier SENTENCECASE = new CaseIdentifier(CASE_SENTENCE);
        public static final CaseIdentifier CAPITALIZECASE = new CaseIdentifier(CASE_CAPITALIZE);
        
        private final String type;

        /**
         * Constructor - private to prevent instantiation.
         * 
         * @param type
         *            The type of case.
         */
        private CaseIdentifier(String type)
        {
            this.type = type;
        }

        /**
         * Accessor for the CaseIdentifier instance by a key (string).
         * @param key the key for the CaseIdentifier
         * @return CaseIdentifier
         */
        public static CaseIdentifier getCaseIdentifier(String key)
        {
            if (key.trim().toUpperCase().equals(CASE_UPPER))
            {
                return Dictionary.CaseIdentifier.UPPERCASE;
            }
            else if (key.trim().toUpperCase().equals(CASE_LOWER))
            {
                return Dictionary.CaseIdentifier.LOWERCASE;
            }
            else if (key.trim().toUpperCase().equals(CASE_SENTENCE))
            {
                return Dictionary.CaseIdentifier.SENTENCECASE;
            }
            else if (key.trim().toUpperCase().equals(CASE_CAPITALIZE))
            {
                return Dictionary.CaseIdentifier.CAPITALIZECASE;
            }
            else
            {
                return Dictionary.CaseIdentifier.PRESERVECASE;
            }
        }

        public String toString()
        {
        	return type;
        }
        
        public boolean equals(Object object)
        {
            if (object == this)
            {
                return true;
            }

            if (!(object instanceof CaseIdentifier))
            {
                return false;
            }

            CaseIdentifier caseIdentifier = (CaseIdentifier)object;
            
            return caseIdentifier.type.equals(this.type);
        }
    }

    /**
     * @return Returns the caseIdentifier.
     */
    public CaseIdentifier getCaseIdentifier()
    {
        return caseIdentifier;
    }

    /**
     * @param caseIdentifier
     *            The caseIdentifier to set.
     */
    public void setCaseIdentifier(CaseIdentifier caseIdentifier)
    {
        this.caseIdentifier = caseIdentifier;
    }

    /**
     * @return Returns the generalNaming.
     */
    public Naming getGeneralNaming()
    {
        return generalNaming;
    }

    /**
     * @param generalNaming
     *            The generalNaming to set.
     */
    public void setGeneralNaming(Naming generalNaming)
    {
        this.generalNaming = generalNaming;
    }

    public boolean equals(Object object)
    {
        if (object == this)
        {
            return true;
        }

        if (!(object instanceof Dictionary))
        {
            return false;
        }

        Dictionary dictionary = (Dictionary)object;

        boolean isEquals = true;
        
        isEquals = isEquals & equalsVariable(generalNaming,  dictionary.generalNaming);
        isEquals = isEquals & equalsVariable(caseIdentifier, dictionary.caseIdentifier);
        
        return isEquals;
    }
    
    /**
     * Equals two variables testing for nulls
     * @param var1 the first variable
     * @param var2 the second variable
     * @return true if equals
     */
    private boolean equalsVariable(Object var1, Object var2)
    {
		if (var1 == null)
		{
			if (var2 != null)
			{
				return false;
			}
		}
		else if (!var1.equals(var2))
		{
			return false;
		}
		return true;
    }
    
    public static Dictionary getDefaultDictionary()
    {
        Dictionary dictionary = new Dictionary();
        Properties properties = new Properties();
        try
        {
            properties.load(Dictionary.class.getResourceAsStream("Dictionary.properties"));
            String property = properties.getProperty(STORE_DICTIONARY_CASE_IDENTIFIER);
            dictionary.setCaseIdentifier(CaseIdentifier.getCaseIdentifier(property));
        }
        catch (IOException e)
        {
            JPOXLogger.GENERAL.warn(LOCALISER.msg("Dictionary.ErrorReadingDefaultDictionary"));
        }

        return dictionary;
    }
}
