/**********************************************************************
Copyright (c) 2002 Kelly Grizzle (TJDO) and others. All rights reserved.
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Erik Bengtson - Refactored OID
2003 Andy Jefferson - fixed OID(String)
2003 Andy Jefferson - coding standards
2004 Andy Jefferson - fixes to allow full use of long or String OIDs
    ...
**********************************************************************/
package org.jpox.store;

import java.sql.Types;

import org.jpox.util.Localiser;

/**
 * An object identifier. OIDs are normally used as object identifiers for
 * persistent objects that use datastore identity. They're also used for
 * view objects, which actually use non-datastore identity. The behaviour of
 * this class is governed by JDO spec 5.4.3.
 *
 * @version $Revision: 1.3 $ 
 */
public class OID implements java.io.Serializable
{
    private static final Localiser LOCALISER=Localiser.getInstance("org.jpox.store.Localisation");

    /** Separator to use between fields. */
    private static final String oidSeparator = "[OID]";

    // JDO spec 5.4.3 - serializable fields required to be public.

    /** The identity. Will be a Long or String. */
    public Object oid;

    /** The PersistenceCapable class name */
    public String pcClass;

    /** The type of the OID. Will be Long or String. */
    public int oidType;

    /**
    * Creates an OID with the no value.
    * Required by the JDO spec
    */
    public OID()
    {
        oid = null;
        pcClass = null;
        oidType = -1;
    }

    /**
     * Create a string datastore identity 
     * @param pcClass The PersistenceCapable class that this represents
     * @param object The value
     */
    public OID(String pcClass, Object object)
    {
        this.pcClass = pcClass;
        this.oid = object;
        
        /**
         * we support both Numeric and String datastore identity
         */
        int type;
        try
        {
            Long.getLong(object.toString());
            type = Types.INTEGER;
        }
        catch (Exception e)
        {
            type = Types.VARCHAR;
        }
        this.oidType = type;
    }

    /**
     * Method to change the identity to be an Integer.
     * @return The OID
     */
    public OID turnDataStoreIdentityAsNumeric()
    {
        this.oidType = Types.INTEGER;
        return this;
    }

    /**
     * Method to change the identity to be a String.
     * @return The OID
     */
    public OID turnDataStoreIdentityAsString()
    {
        this.oidType = Types.VARCHAR;
        return this;
    }

    /**
     * Constructs an OID from its string representation that is
     * consistent with the output of toString().
     * @param str the string representation of an OID.
     * @exception IllegalArgumentException
     *      if the given string representation is not valid.
     * @see #toString
     */
    public OID(String str)
    throws IllegalArgumentException
    {
        if (str.length() < 2)
        {
            throw new IllegalArgumentException(LOCALISER.msg("OID.InvalidValue",str));
        }

        int start = 0;
        int end = str.indexOf(oidSeparator, start);
        this.oid = str.substring(start, end);
        start = end + oidSeparator.length();
        end = str.indexOf(oidSeparator, start);
        this.oidType = Integer.parseInt(str.substring(start, end));
        start = end + oidSeparator.length();
        this.pcClass = str.substring(start, str.length());
    }

    /**
     * Returns copy of the requested oid to be accessed by the user.
     * @return Copy of the OID.
     */
    public Object getNewObjectIdCopy()
    {
        OID oid = new OID();
        oid.oid = this.oid;
        oid.oidType = this.oidType;
        oid.pcClass = this.pcClass;
        return oid;
    }

    /**
     * Returns the requested oid to be used internally.
     * @return The OID
     */
    public Object getInternalObjectId()
    {
        return this;
    }
    
    /**
    * Provides the OID in a form that can be used by the database as a key.
    * @return The key value, including class name etc.
    */
    public String keyValue()
    {
        return String.valueOf(oid);
    }
    
    /**
     * Returns the oid value on the datastore
     * @return the oid value on the datastore 
     */
    public Object getValue()
    {
        return oid;
    }
    
    /**
     * Equality operator.
     * @param obj Object to compare against
     * @return Whether they are equal
     */
    public boolean equals(Object obj)
    {
        if (obj == this)
        {
            return true;
        }

        if (!(obj instanceof OID))
        {
            return false;
        }

        if (!this.toString().equals(((OID) obj).toString()))
        {
            return false;
        }
        return true;
    }

    /**
     * Accessor for the hashcode
     * @return Hashcode for this object
     */
    public int hashCode()
    {
        int hash;
        hash = (pcClass.toString() + "-" + String.valueOf(oid)).hashCode();
        return hash;
    }

    /**
     * Returns the string representation of the OID.
     * Will be a string such as "1[OID]4[OID]org.jpox.samples.MyClass"
     * where
     * <UL>
     * <LI>1 is the identity "id"</LI>
     * <LI>4 is the type</LI>
     * <LI>class name is the name of the PersistenceCapable class that it represents</LI>
     * </UL>
     * @return the string representation of the OID.
     */
    public String toString()
    {
        StringBuffer s = new StringBuffer();
        s.append(this.oid.toString());
        s.append(oidSeparator);
        s.append(Integer.toString(this.getOIDType()));
        s.append(oidSeparator);
        s.append(this.pcClass);
        return s.toString();
    }

    /**
     * Accessor for the OID type 
     * @return the OID type
     */
    public int getOIDType()
    {
        return oidType;
    }

    /**
     * Accessor for the PC class name 
     * @return the PC Class
     */
    public String getPcClass()
    {
        return pcClass;
    }

    /**
     * Mutator for the PC Class 
     * @param pc_class_name The PC Class
     */
    public void setPcClass(String pc_class_name)
    {
        pcClass = pc_class_name;
    }

    /**
     * Accessor for whether this is application identity 
     * @return Whether this is for application identity (false)
     */
    public boolean isApplicationIdentity()
    {
        return false;
    }

    /**
     * Accessor for whether this is datastore identity 
     * @return Whether this is for datastore identity (true)
     */
    public boolean isDataStoreIdentity()
    {
        return true;
    }
}