/**********************************************************************
Copyright (c) 2002 Kelly Grizzle (TJDO) and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution.

Contributors:
2002 Mike Martin (TJDO)
2003 Andy Jefferson - coding standards
2005 Andy Jefferson - added support for result set type/concurrency
    ...
**********************************************************************/
package org.jpox.store;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collections;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import javax.jdo.JDOFatalInternalException;

import org.jpox.PersistenceManager;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.Mappings;
import org.jpox.store.query.QueryStatement;
import org.jpox.util.Localiser;

/**
 * Representation of a statement.
 *
 * @version $Revision: 1.3 $
 **/
public class StatementText
{
    private static final Localiser LOCALISER=Localiser.getInstance("org.jpox.store.Localisation");

    private static int nextParamID = 0;

    private StringBuffer sql;
    private ArrayList parameterNames          = null;
    private HashMap   parameterMappingsByName = null;
    private HashMap   parameterValuesByName   = null;
    private HashSet   referencedColumns       = null;
    private boolean	  encloseWithInParentheses = false;
    private String	  postpend;

    /**
     * Constructor
     **/
    public StatementText()
    {
        sql = new StringBuffer();
    }

    public StatementText(String initialSql)
    {
        sql = new StringBuffer(initialSql);
    }

    /**
     * Method to initialise the statement.
     **/
    private void initParameters()
    {
        if (parameterNames == null)
        {
            parameterNames          = new ArrayList();
            parameterMappingsByName = new HashMap();
            parameterValuesByName   = new HashMap();
        }
    }

    private void initReferencedColumns()
    {
        if (referencedColumns == null)
        {
            referencedColumns = new HashSet();
        }
    }

    /**
     * Whether to enclose this statement within parentheses
     */
    public void encloseWithInParentheses()
    {
        this.encloseWithInParentheses = true;
    }
    
    public Set getReferencedColumns()
    {
        if (referencedColumns == null)
        {
            return Collections.EMPTY_SET;
        }
        else
        {
            return Collections.unmodifiableSet(referencedColumns);
        }
    }

    public StatementText prepend(char c)
    {
        sql.insert(0, c);
        return this;
    }

    public StatementText prepend(String s)
    {
        sql.insert(0, s);
        return this;
    }
    
    public StatementText postpend(String s)
    {
        postpend = s;
        return this;
    }    

    public StatementText append(char c)
    {
        sql.append(c);
        return this;
    }

    public StatementText append(String s)
    {
        sql.append(s);
        return this;
    }

    public StatementText append(QueryStatement.QueryExpression qsc)
    {
        sql.append(qsc);

        initReferencedColumns();
        referencedColumns.add(qsc);
        return this;
    }

    public StatementText append(StatementText st)
    {
        sql.append(st);

        if (st.parameterNames != null)
        {
            initParameters();
            parameterNames.addAll(st.parameterNames);
            parameterMappingsByName.putAll(st.parameterMappingsByName);
            parameterValuesByName.putAll(st.parameterValuesByName);
        }

        if (st.referencedColumns != null)
        {
            initReferencedColumns();
            referencedColumns.addAll(st.referencedColumns);
        }

        return this;
    }

    public StatementText append(ScalarExpression expr)
    {
        return append(expr.toStatementText());
    }

    public StatementText append(Object o)
    {
        sql.append(o);
        return this;
    }

    public String appendParameter(JavaTypeMapping mapping, Object value)
    {
        String name = "param-" + nextParamID++;

        sql.append('?');

        initParameters();
        parameterNames.add(name);
        parameterMappingsByName.put(name, mapping);
        parameterValuesByName.put(name, value);

        return name;
    }

    public void setParameterValue(String name, Object value)
    {
        initParameters();

        if (!parameterValuesByName.containsKey(name))
        {
            throw new JDOFatalInternalException(LOCALISER.msg("StatementText.InvalidParameter",name));
        }

        parameterValuesByName.put(name, value);
    }

    public PreparedStatement prepareStatement(PersistenceManager pm,
                                              Connection conn)
    throws SQLException
    {
        String stmtText = toString();

        PreparedStatement ps = null;

        // Apply any non-standard result set definition if required
        String rsTypeString = pm.getPMFContext().getPmfConfiguration().getQueryResultSetType();
        String rsConcurrencyString = pm.getPMFContext().getPmfConfiguration().getQueryResultSetConcurrency();
        if (!rsTypeString.equals("forward-only") || !rsConcurrencyString.equals("read-only"))
        {
            int rsType = ResultSet.TYPE_FORWARD_ONLY;
            if (rsTypeString.equals("scroll-sensitive"))
            {
                rsType = ResultSet.TYPE_SCROLL_SENSITIVE;
            }
            else if (rsTypeString.equals("scroll-insensitive"))
            {
                rsType = ResultSet.TYPE_SCROLL_INSENSITIVE;
            }
            int rsConcurrency = ResultSet.CONCUR_READ_ONLY;
            if (rsConcurrencyString.equals("updateable"))
            {
                rsConcurrency = ResultSet.CONCUR_UPDATABLE;
            }

            ps = conn.prepareStatement(stmtText, rsType, rsConcurrency);
        }
        else
        {
            ps = conn.prepareStatement(stmtText);
        }

        boolean done = false;

        try
        {
            setParameters(pm, ps);
            done = true;
        }
        finally
        {
            if (!done)
            {
                ps.close();
            }
        }

        return ps;
    }

    private void setParameters(PersistenceManager pm, PreparedStatement ps)
    {
        if (parameterNames != null)
        {
            Iterator i = parameterNames.iterator();
            int stmtParamNum = 1;

            while (i.hasNext())
            {
                Object name = i.next();
                JavaTypeMapping mapping = (JavaTypeMapping)parameterMappingsByName.get(name);
                Object value = parameterValuesByName.get(name);
                
                mapping.setObject(pm, ps, Mappings.getParametersIndex(stmtParamNum,mapping), value);
                if (mapping.getNumberOfDatastoreFields() > 0)
                {
                    stmtParamNum += mapping.getNumberOfDatastoreFields();
                }
                else
                {
                    stmtParamNum += 1;
                }
            }
        }
    }

    /**
     * Accessor for the SQL of the statement.
     * @return The SQL text
     **/
    public String toString()
    {
        if (encloseWithInParentheses)
        {
            return "(" + sql.toString() + ")" + (postpend == null ? "" : postpend);
        }
        else
        {
            return sql.toString() + (postpend == null ? "" : postpend);
        }
    }
}