/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution.

Contributors:
    ...
**********************************************************************/
package org.jpox.store;

import javax.jdo.JDOFatalUserException;

import org.jpox.metadata.ClassMetaData;
import org.jpox.metadata.ExtendableMetaData;
import org.jpox.metadata.FieldMetaData;
import org.jpox.util.Localiser;

/**
 * Representation of a class/field managed by the StoreManager.
 * <H3>Type of data</H3>
 * There are 2 constructors, one for specifying a class (FCO) being managed, and one
 * for specifying a field (SCO) being managed. The "type" is set accordingly to represent
 * which one is being used. 
 * <H3>Ownership of Tables</H3>
 * In JPOX, a class can have a table assigned where the class is not the "owner" of the table.
 * This happens when the class has its fields persisted in the table of another class
 * - so they share the table.
 *
 * @version $Revision: 1.2 $
 */
public class StoreData
{
    /** Localiser for messages. */
    protected static final Localiser LOCALISER=Localiser.getInstance("org.jpox.store.Localisation");

    /** First class object (FCO) type */
    public static final int FCO_TYPE = 1;

    /** Second class object (SCO) type */
    public static final int SCO_TYPE = 2;

    /** Name of the class/field. */
    protected final String name;

    /** Name of the datastore object where it is stored. */
    protected String tableName;

    /** Whether this class is the owner of the table. */
    protected final boolean isTableOwner;

    /** Type of data being stored. */
    protected final int type;

    /** MetaData for this object. */
    protected final ExtendableMetaData metaData;

    /**
     * Constructor. To be used when creating for the start mechanism.
     * @param name Name of the class/field
     * @param tableName Name of the table associated
     * @param tableOwner Whether this is the owner
     * @param type The type (FCO/SCO)
     */
    public StoreData(String name, String tableName, boolean tableOwner, int type)
    {
        this.metaData = null;
        this.name = name;
        this.tableName = tableName;
        this.isTableOwner = tableOwner;
        this.type = type;
    }

    /**
     * Constructor for FCO data.
     * @param cmd MetaData for the class.
     * @param tableName Name of the table where the class is stored.
     * @param tableOwner Whether the class is the owner of the table.
     */
    public StoreData(ClassMetaData cmd, String tableName, boolean tableOwner)
    {
        if (cmd == null)
        {
            throw new JDOFatalUserException(LOCALISER.msg("StoreData.MetaDataNullError", tableName));
        }
        this.name = cmd.getFullClassName();
        this.tableName = tableName;
        this.isTableOwner = tableOwner;
        this.metaData = cmd;
        this.type = FCO_TYPE;
    }

    /**
     * Constructor for SCO data.
     * @param fmd MetaData for the field.
     * @param tableName Name of the table where the field is stored.
     */
    public StoreData(FieldMetaData fmd, String tableName)
    {
        if (fmd == null)
        {
            throw new JDOFatalUserException(LOCALISER.msg("StoreData.MetaDataNullError", tableName));
        }
        this.name = fmd.getFullFieldName();
        this.tableName = tableName;
        this.isTableOwner = true; // Always the owner of the table
        this.metaData = fmd;
        type = SCO_TYPE;
    }

    /**
     * Accessor for class/field name.
     * @return Returns the class/field name.
     */
    public String getName()
    {
        return name;
    }

    /**
     * Accessor for tableName.
     * @return Returns the tableName.
     */
    public String getTableName()
    {
        return tableName;
    }

    /**
     * Accessor for whether this class is the owner of the table.
     * @return Whether it owns the table
     */
    public boolean isTableOwner()
    {
        return isTableOwner;
    }

    /**
     * Accessor for whether this has a table representation.
     * @return Whether it has a table
     */
    public boolean hasTable()
    {
        return (tableName != null);
    }

    /**
     * Accessor for whether this represents FCO data
     * @return Whether it is FCO
     */
    public boolean isFCO()
    {
        return type == FCO_TYPE;
    }

    /**
     * Accessor for whether this represents SCO data.
     * @return Whether it is SCO.
     */
    public boolean isSCO()
    {
        return type == SCO_TYPE;
    }

    /**
     * Accessor for type.
     * @return Returns the type.
     */
    public int getType()
    {
        return type;
    }
    
    /**
     * Accessor for any superclass for this class, if this is a class.
     * @return The superclass (if exists). If this is a field, just return null.
     **/
    public String getSuperclass()
    {
        if (metaData instanceof ClassMetaData)
        {
            return ((ClassMetaData)metaData).getPersistenceCapableSuperclass();
        }
        else
        {
            return null;
        }
    }

    /**
     * Accessor for class/field meta data.
     * @return Returns the class/field meta data.
     */
    public ExtendableMetaData getMetaData()
    {
        return metaData;
    }

    /**
     * Method to return this class/field managed object as a string.
     * @return String version of this class/field managed object.
     **/
    public String toString()
    {
        if (metaData instanceof ClassMetaData)
        {
            ClassMetaData cmd = (ClassMetaData)metaData;
            if (tableName != null)
            {
                return LOCALISER.msg("StoreData.DataForClass", name, tableName,
                    cmd.getInheritanceMetaData().getStrategyValue().toString());
            }
            else
            {
                return LOCALISER.msg("StoreData.DataForClass", name, "(none)",
                    cmd.getInheritanceMetaData().getStrategyValue().toString());
            }
        }
        else if (metaData instanceof FieldMetaData)
        {
            return LOCALISER.msg("StoreData.DataForField", name, tableName);
        }
        else
        {
            return null;
        }
    }
}