/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
    ...
**********************************************************************/
package org.jpox.store;

import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Map;
import java.util.Set;
import java.util.WeakHashMap;

import javax.jdo.Extent;
import javax.jdo.spi.PersistenceCapable;
import javax.sql.DataSource;

import org.jpox.AbstractPersistenceManagerFactory;
import org.jpox.ClassLoaderResolver;
import org.jpox.PMFContext;
import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.metadata.ClassMetaData;
import org.jpox.metadata.ExtendableMetaData;
import org.jpox.metadata.ExtensionMetaData;
import org.jpox.metadata.FieldMetaData;
import org.jpox.metadata.IdentityStrategy;
import org.jpox.metadata.IdentityType;
import org.jpox.metadata.InheritanceStrategy;
import org.jpox.metadata.MetaDataManager;
import org.jpox.store.exceptions.DatastoreInitialisationException;
import org.jpox.store.mapping.MappingManager;
import org.jpox.store.scostore.Store;
import org.jpox.util.AIDUtils;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;
import org.jpox.util.MacroString;
import org.jpox.util.MultiMap;

/**
 * An abstract representation of a Store Manager.
 * Manages the persistence of objects to the store.
 * Will be implemented for the type of datastore (RDBMS, ODBMS, OLAP) in question. 
 * The store manager's responsibilities include:
 * <ul>
 * <li>Creating and/or validating datastore tables according to the persistent classes being 
 * accessed by the application.</li>
 * <li>Serving as the primary intermediary between StateManagers and the database.</li>
 * <li>Serving as the base Extent and Query factory.</li>
 * </ul>
 * <p>
 * A store manager's knowledge of its contents is typically not complete. It knows about
 * the classes that it has encountered in its lifetime. The PersistenceManager can make the
 * StoreManager aware of a class, and can check if the StoreManager knows about a particular class.
 * The Auto-Start mechanism provides a way of inheriting knowledge from the last time the store
 * was used.
 *
 * @version $Revision: 1.10 $
 */
public abstract class StoreManager
{
    protected static final Localiser LOCALISER = Localiser.getInstance("org.jpox.store.Localisation");
    
    /**
     * Datasource
     */
    protected final DataSource ds;

    /**
     * Adapter for the datastore being used.
     */
    protected DatastoreAdapter dba;

    /**
     * Username for the datastore being used.
     */
    protected final String userName;

    /**
     * Password for the datastore being used.
     */
    protected final String password;

    /**
     * Whether this datastore schema is read only.
     */
    protected final boolean readOnlySchema;
    
    /**
     * Whether this datastore schema is fixed (no mods to table definitions allowed).
     */
    protected final boolean fixedSchema;
    
    /**
     * Auto-Start mechanism to use.
     */
    protected AutoStartMechanism starter = null;

    /**
     * Whether the AutoStart mechanism is initialised
     */
    protected boolean starterInitialised = false;

    /**
     * Dictionary to use for identifiers etc.
     */
    protected final Dictionary dictionary;

    /**
     * Resolver for all class loading issues.
     */
    protected final ClassLoaderResolver clr;

    /**
     * PMF context.
     */
    protected final PMFContext pmfContext;

    /**
     * Whether to auto create any tables.
     */
    protected final boolean autoCreateTables;

    /**
     * Whether to auto create any columns that are missing.
     */
    protected final boolean autoCreateColumns;

    /**
     * Whether to auto create any constraints
     */
    protected final boolean autoCreateConstraints;

    /**
     * Whether to validate any tables
     */
    protected final boolean validateTables;

    /**
     * Whether to validate any constraints
     */
    protected final boolean validateConstraints;
    
    /**
     * Whether to check if table/view exists
     */
    protected final boolean checkExistTablesOrViews;    

    /**
     * The default id generator class name for the store manager.
     * Each class can have it's own Id generator defined in the meta-data file.
     * @see org.jpox.poid.PoidGenerator
     */
    protected final String defaultPoidGeneratorClass;

    /** Cache of POID managers for the store. */
    protected final Map poidManagers = new HashMap();

    /** Map of all managed store data, keyed by the class/field name. */
    protected Map storeDataByClass = new WeakHashMap();

    /** Map of all managed store data using Application Identity, keyed by the app id PK class */
    protected MultiMap storeDataByAppIdClass = new MultiMap();

    /**
     * Constructor for a new StoreManager. Stores the basic information
     * required for the datastore management.
     * @param clr the ClassLoaderResolver
     * @param pmf The corresponding PersistenceManagerFactory. This factory's
     *            non-transactional data source will be used to get database
     *            connections as needed to perform management functions.
     * @param userName The database user name.
     * @param password The database user's password.
     * @see StoreManagerFactory
     */
    protected StoreManager(ClassLoaderResolver clr, 
                           AbstractPersistenceManagerFactory pmf, 
                           String userName,
                           String password)
    {
        ds = pmf.getNontransactionalDataSource();
        this.userName = userName;
        this.password = password;
        this.dictionary = pmf.getPMFConfiguration().getDictionary();
        this.clr = clr;
        this.pmfContext = pmf.getPMFContext();
        this.readOnlySchema = pmf.getReadOnlySchema();
        this.fixedSchema = pmf.getFixedSchema();

        this.defaultPoidGeneratorClass = pmf.getPMFConfiguration().getPersistentIdGenerator();

        autoCreateColumns = pmf.getPMFConfiguration().getAutoCreateColumns();
        if (pmf.getPMFConfiguration().getAutoCreateSchema())
        {
            autoCreateTables = true;
            autoCreateConstraints = true;
        }
        else
        {
            autoCreateTables = pmf.getPMFConfiguration().getAutoCreateTables();
            autoCreateConstraints = pmf.getPMFConfiguration().getAutoCreateConstraints();
        }
        validateTables = pmf.getPMFConfiguration().getValidateTables();
        validateConstraints = pmf.getPMFConfiguration().getValidateConstraints();
        checkExistTablesOrViews = pmf.getPMFConfiguration().getCheckExistTablesOrViews();
    }

    // -------------------------------- Management of Classes --------------------------------

    /**
     * Method to register some data with the store.
     * This will also register the data with the starter process.
     * @param data The StoreData to add
     */
    protected void registerStoreData(StoreData data)
    {
        if (data.isFCO())
        {
            // Index any classes by the class name
            storeDataByClass.put(data.getName(), data);

            // If it's a class, using APPLICATION identity and is users own AID then store the PK class.
            // We don't need SingleFieldIdentity in here since they define the class being used
            ClassMetaData cmd = (ClassMetaData)data.getMetaData();
            if (cmd.getIdentityType() == IdentityType.APPLICATION &&
                !AIDUtils.isSingleFieldIdentityClass(cmd.getObjectidClass()))
            {
                storeDataByAppIdClass.put(cmd.getObjectidClass(), data);
            }
        }
        else
        {
            // Index any fields by the class name of the field
            storeDataByClass.put(((FieldMetaData)data.getMetaData()).getField(), data);
        }

        if (JPOXLogger.RDBMS_SCHEMA.isInfoEnabled())
        {
            JPOXLogger.RDBMS_SCHEMA.info(LOCALISER.msg("StoreManager.AddStoreData", data));
        }

        // Keep the AutoStarter in step with our managed classes/fields
        if (starter != null && starterInitialised)
        {
            starter.addClass(data);
        }
    }

    /**
     * Method to deregister all existing store data so that we are managing nothing.
     */
    protected void deregisterAllStoreData()
    {
        storeDataByClass.clear();
        storeDataByAppIdClass.clear();
        if (JPOXLogger.RDBMS_SCHEMA.isInfoEnabled())
        {
            JPOXLogger.RDBMS_SCHEMA.info(LOCALISER.msg("StoreManager.RemoveStoreData"));
        }

        // Keep the AutoStarter in step with our managed classes/fields
        clearAutoStarter();
    }

    /**
     * Utility to return the StoreData for a table name.
     * @param tableName Name of the table
     * @return The StoreData for this table (if managed).
     */
    public synchronized StoreData getStoreDataForTable(String tableName)
    {
        Iterator iterator = storeDataByClass.values().iterator();
        while (iterator.hasNext())
        {
            StoreData data = (StoreData) iterator.next();
            if (data.getTableName() != null && data.getTableName().equals(tableName))
            {
                return data;
            }
        }
        return null;
    }

    // -------------------------------- Auto Starter ------------------------------------

    /**
     * Method to initialise the auto-start mechanism, loading up the classes
     * from its store into memory so that we start from where we got to last
     * time.
     */
    protected void initialiseAutoStart()
    throws DatastoreInitialisationException
    {
        if (starter == null || starterInitialised)
        {
            return;
        }

        try
        {
            if (!starter.isOpen())
            {
                starter.open();
            }
            Collection existing_data = starter.getAllClassData();
            if (existing_data != null)
            {
                Iterator existing_data_iter = existing_data.iterator();
                while (existing_data_iter.hasNext())
                {
                    // Add classes one-by-one
                    // TODO : Do this in one call.
                    // Consider sorting the classes so that we load superclasses first
                    StoreData data = (StoreData) existing_data_iter.next();
                    if (data.isFCO())
                    {
                        try
                        {
                            String the_class = data.getName();
                            if (JPOXLogger.JDO.isInfoEnabled())
                            {
                                JPOXLogger.JDO.info(LOCALISER.msg("StoreManager.AddPreviousStoreData", data.getName()));
                            }

                            // Catch classes that don't exist (e.g in use by a different app)
                            if (clr.classForName(the_class) != null)
                            {
                                addClass(the_class);
                            }
                        }
                        catch (Exception e)
                        {
                            // Error, so act according to the "autoStartMechanismMode"
                            if (starter.getMode().equals(AutoStartMechanism.MODE_CHECKED))
                            {
                                String msg = LOCALISER.msg("AutoStarter.InitialisationError", data.getName());
                                JPOXLogger.JDO.error(e);
                                JPOXLogger.JDO.error(msg);
                                throw new DatastoreInitialisationException(msg);
                            }
                            else if (starter.getMode().equals(AutoStartMechanism.MODE_IGNORED))
                            {
                                JPOXLogger.JDO.warn(e);
                                JPOXLogger.JDO.warn(LOCALISER.msg("AutoStarter.InitialisationError", data.getName()));
                            }
                            else if (starter.getMode().equals(AutoStartMechanism.MODE_QUIET))
                            {
                                JPOXLogger.JDO.warn(e);
                                JPOXLogger.JDO.warn(LOCALISER.msg("AutoStarter.DeleteClass", data.getName()));
                                starter.deleteClass(data.getName());
                            }
                        }
                    }
                }
            }
        }
        finally
        {
            if (starter.isOpen())
            {
                starter.close();
            }
        }
        
        starterInitialised = true;
    }

    /**
     * Method to clear the Auto-Starter status.
     */
    protected void clearAutoStarter()
    {
        // AutoStarter - Delete all supported classes
        if (starter != null)
        {
            try
            {
                if (!starter.isOpen())
                {
                    starter.open();
                }
                starter.deleteAllClasses();
            }
            finally
            {
                if (starter.isOpen())
                {
                    starter.close();
                }
            }
        }
    }

    // ------------------------------- Class Management -----------------------------------

    /**
     * Accessor for whether the specified class is managed currently
     * @param className The name of the class
     * @return Whether it is managed
     */
    public boolean managesClass(String className)
    {
        return storeDataByClass.containsKey(className);
    }
    
    /**
     * Method to add a class to the managed list for this datastore manager.
     * @param className Name of the class
     */
    public abstract void addClass(String className);
    
    /**
     * Add classes to the persistence model for the datastore.
     * In the case of an RDBMS, this will map these classes to RDBMS tables.
     * In the case of XML, this will map these classes to XML datafiles. 
     * <p>
     * This method is primarily useful for applications that wish to perform all
     * of their datastore initialization up front, rather than wait for the JPOX
     * runtime to do it on-demand.
     * @param classes The class(es) to be added.
     * @exception org.jpox.store.exceptions.SchemaValidationException
     *      If there is some mismatch between the current schema contents and
     *      those necessary to enable persistence of the given classes.
     */
    public abstract void addClasses(String[] classes);

    /**
     * Drops all tables in the schema.
     * This empties the schema of all database objects managed by JPOX.
     * All objects not managed by JPOX are left untouched.
     */
    public abstract void removeAllClasses();

    /**
     * Method to return the class(es) that has a table managing the persistence of
     * the fields of the supplied class. For the 3 inheritance strategies, the following
     * occurs :-
     * <UL>
     * <LI>new-table : will return the same ClassMetaData</LI>
     * <LI>no-table : will return all subclasses that have a table managing its fields</LI>
     * <LI>superclass-table : will return the next superclass that has a table</LI>
     * </UL> 
     * @param cmd The supplied class.
     * @return The ClassMetaData's managing the fields of the supplied class
     */
    public ClassMetaData[] getClassesManagingTableForClass(ClassMetaData cmd)
    {
        // Null input, so just return null;
        if (cmd == null)
        {
            return null;
        }

        // This class manages a table so return the classes metadata.
        if (cmd.getInheritanceMetaData().getStrategyValue() == InheritanceStrategy.NEW_TABLE)
        {
            return new ClassMetaData[] {cmd};
        }
        else if (cmd.getInheritanceMetaData().getStrategyValue() == InheritanceStrategy.NO_TABLE)
        {
            // Class managed by subclass so find classes which are subclass of this class.
            HashSet managingClasses=new HashSet();
            Set managedClasses=storeDataByClass.keySet();
            Iterator managedClassesIter = managedClasses.iterator();
            while (managedClassesIter.hasNext())
            {
                StoreData data = (StoreData)storeDataByClass.get(managedClassesIter.next());
                if (data.isFCO() && ((ClassMetaData)data.getMetaData()).getSuperclassMetaData() != null &&
                    ((ClassMetaData)data.getMetaData()).getSuperclassMetaData().getFullClassName() == cmd.getFullClassName())
                {
                    ClassMetaData[] superCmds = getClassesManagingTableForClass((ClassMetaData)data.getMetaData());
                    if (superCmds != null)
                    {
                        for (int i=0;i<superCmds.length;i++)
                        {
                            managingClasses.add(superCmds[i]);
                        }
                    }
                }
            }
            
            Iterator managingClassesIter = managingClasses.iterator();
            ClassMetaData managingCmds[] = new ClassMetaData[managingClasses.size()];
            int i=0;
            while (managingClassesIter.hasNext())
            {
                managingCmds[i++] = (ClassMetaData)(managingClassesIter.next());
            }
            
            return managingCmds;
        }
        else if (cmd.getInheritanceMetaData().getStrategyValue() == InheritanceStrategy.SUPERCLASS_TABLE)
        {
            // Fields managed by superclass, so recurse to that
            return getClassesManagingTableForClass(cmd.getSuperclassMetaData());
        }
        return null;
    }
    
    // ------------------------------ PersistenceManager interface -----------------------------------

    /**
     * Interface to getting an Extent for a class.
     * @param pm The Persistence Manager
     * @param c  The class requiring the Extent
     * @param subclasses Whether to include subclasses of 'c'
     * @return The Extent.
     */
    public abstract Extent getExtent(org.jpox.PersistenceManager pm,Class c,boolean subclasses);

    /**
     * Accessor for the query languages supported by this store
     * @param language The language name
     * @return Whether it is supported.
     */
    public abstract boolean supportsQueryLanguage(String language);

    /**
     * Returns the class corresponding to the given object JDO ID.
     * If the object is an OID, return the PersistenceCapable class.
     * If the object is a SCOID, return the SCO class.
     * If the object is a PersistenceCapable class, return the class.
     * If the object is an AppID PK, return the PersistenceCapable class. 
     * @param pm Persistence Manager
     * @param id The JDO identity of some object.
     * @return   For datastore identity, return the class of the corresponding object.
     *           For application identity, return the class of the corresponding object or null if object does not exist.
     * @exception ClassCastException
     *           If the type of ID is not recognized 
     *           ({@link OID} or {@link SCOID}).
     */
    public abstract String getClassNameForObjectID(org.jpox.PersistenceManager pm,Object id);

    /**
     * Check if the strategy is attributed by the database when the
     * PersistenceCapable object is inserted into the database
     * 
     * @param identityStrategy the identityStrategy
     * @return if the object for the strategy is attributed by the database
     */
    public static boolean isStrategyDatastoreAttributed(IdentityStrategy identityStrategy)
    {
        if (identityStrategy == null)
        {
            return false;
        }

        // "autoassign" and "identity" are processed in the datastore.
        // They are treated the same in the database adapter since they relate to the 
        // same functionality - a keyword to autoincrement a column.
        if (identityStrategy == IdentityStrategy.AUTOASSIGN ||
            identityStrategy == IdentityStrategy.IDENTITY)
        {
            return true;
        }

        return false;
    }

    /**
     * Method to retrieve the value for a strategy for a particular field.
     * @param pm The Persistence Manager
     * @param table The datastore table
     * @param cmd ClassMetaData for the class
     * @param absoluteFieldNumber The field number
     * @param strategy The strategy
     * @param sequence The sequence name if strategy is sequence-based
     * @param extensions Any extensions to control the value generation process
     * @return The value
     */
    public abstract Object getStrategyValue(PersistenceManager pm,
                                            DatastoreClass table,
                                            ClassMetaData cmd,
                                            int absoluteFieldNumber,
                                            IdentityStrategy strategy,
                                            String sequence,
                                            ExtensionMetaData[] extensions);

    /**
     * Returns a new, unique ID for an object of the given class.
     * @param className Name of the class of the object.
     * @param pc The persistenceCapable object. Useful for AppID
     * @return A new object ID.
     */
    public abstract Object newObjectID(StateManager sm,String className,PersistenceCapable pc);

    /**
     * Inserts a persistent object into the database.
     * @param sm The state manager of the object to be inserted.
     */
    public abstract void insert(StateManager sm);

    /**
     * Fetches a persistent object from the database.
     * @param sm The state manager of the object to be fetched.
     * @param fieldNumbers The numbers of the fields to be fetched.
     */
    public abstract void fetch(StateManager sm,int[] fieldNumbers);

    /**
     * Updates a persistent object in the database.
     * @param sm The state manager of the object to be updated.
     * @param fieldNumbers The numbers of the fields to be updated.
     */
    public abstract void update(StateManager sm,int[] fieldNumbers);

    /**
     * Deletes a persistent object from the database.
     * @param sm The state manager of the object to be deleted.
     */
    public abstract void delete(StateManager sm);

    /**
     * Locates this object in the datastore. If the instance does not exist in
     * the datastore, this method will fail raising a
     * javax.jdo.JDOObjectNotFoundException.
     */
    public abstract void locate(StateManager sm);

    /**
     * Delete dependent objects of a persistent object.
     * @since 1.1
     * @param sm StateManager for the object being deleted.
     * @param fieldNumbers the fields to delete.
     */
    public abstract void deleteDependent(StateManager sm,int[] fieldNumbers);

    /**
     * Called by Mapping objects to request the creation of a DatastoreObject (table).
     * @param fmd The field metadata describing the field.
     */
    public abstract DatastoreContainerObject newJoinDatastoreContainerObject(FieldMetaData fmd);
    
    /**
     * Called by Mapping objects to get Store objects
     * @param table The join table
     * @param fmd The field metadata describing the field.
     */
    public abstract Store newStore(DatastoreObject table, FieldMetaData fmd);

    /**
     * Utility to return the Classes that are known subclasses to the provided
     * class.
     * @param className Class for which we search for subclasses.
     * @param include_subclasses Whether to include subclasses of subclasses etc
     * @return Set of classes that are subclasses of the input class.
     */
    public HashSet getSubClassesForClass(String className, boolean include_subclasses)
    {
        HashSet subclasses = new HashSet();
        Iterator iter = storeDataByClass.values().iterator();
        StoreData data = null;
        ExtendableMetaData md = null;

        while (iter.hasNext())
        {
            data = (StoreData) iter.next();
            md = data.getMetaData();
            String superclassName = data.getSuperclass();
            if (superclassName != null && superclassName.equals(className) && md instanceof ClassMetaData)
            {
                String subclassName = ((ClassMetaData) md).getFullClassName();

                // If we don't have this class yet, add it
                if (!subclasses.contains(subclassName))
                {
                    subclasses.add(subclassName);
                    // Check all subclasses of this class if required
                    if (include_subclasses)
                    {
                        HashSet subsubclasses = getSubClassesForClass(subclassName, include_subclasses);
                        subclasses.addAll(subsubclasses);
                    }
                }
            }
        }

        return subclasses;
    }

    // ------------------------------- Accessors ------------------------------------

    /**
     * Accessor for the ClassLoader resolver. 
     * @return Returns the ClassLoaderResolver.
     */
    public ClassLoaderResolver getClassLoaderResolver()
    {
        return clr;
    }

    /**
     * Accessor for the context in which this RDBMSManager is running
     * @return Returns the pmfContext.
     */   
    public PMFContext getPMFContext()
    {
        return pmfContext;
    }
 
    /**
     * Gets the MetaDataManager to use for this store.
     * @return Returns the MetaDataManager.
     */   
    public MetaDataManager getMetaDataManager()
    {
        return pmfContext.getMetaDataManager();
    }

    /**
     * Gets the MappingManager to use for this store.
     * @return Returns the MappingManager.
     */
    public MappingManager getMappingManager()
    {
        return dba.getMappingManager();
    }

    /**
     * Gets the DatastoreAdapter to use for this store.
     * @return Returns the DatastoreAdapter
     */   
    public DatastoreAdapter getDatastoreAdapter()
    {
        return dba;
    }
    
    /**
     * Gets the DatastoreClass for a given classname
     * @param className the class name with datastore representation
     * @return Returns the DatastoreClass
     */
    public abstract DatastoreClass getDatastoreClass(String className);
    
    /**
     * Resolves an identifier macro. The public fields <var>clazz </var>,
     * <var>fieldName </var>, and <var>subfieldName </var> of the given macro
     * are taken as inputs, and the public <var>value </var> field is set to the
     * SQL identifier of the corresponding database table or column.
     * @param im The macro to resolve.
     */
    public abstract void resolveIdentifierMacro(MacroString.IdentifierMacro im);
    
}