/******************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved. 
This program and the accompanying materials are made available under
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
    ...
*****************************************************************/
package org.jpox.store;

import java.io.FileNotFoundException;
import java.io.InputStream;
import java.io.InputStreamReader;

import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;
import org.xml.sax.EntityResolver;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;

/**
 * Implementation of an entity resolver for JPOX Auto Starter files.
 * 
 * @version $Revision: 1.1 $
 */
public class XMLAutoStarterEntityResolver implements EntityResolver
{
    protected static Localiser LOCALISER=Localiser.getInstance("org.jpox.store.Localisation");
    
    /**
     * Public Key
     */
    public static final String PUBLIC_ID_KEY="-//Java Persistent Objects//DTD JPOX Auto Starter Metadata 1.0//EN";

    /**
     * Method to resolve XML entities. Checks for PUBLIC and SYSTEM entities.
     * Optimises access using local versions.
     * @param publicId The public id.
     * @param systemId The system id.
     * @return Input Source for the URI.
     * @see org.xml.sax.EntityResolver#resolveEntity(java.lang.String,java.lang.String)
     */
    public InputSource resolveEntity(String publicId, String systemId)
    throws SAXException
    {
        try
        {
            if (JPOXLogger.METADATA.isDebugEnabled())
            {
                JPOXLogger.METADATA.debug(LOCALISER.msg("XMLAutoStarter.ResolveEntity", publicId, systemId));
            }

            // Use publicId if possible
            if (publicId != null) 
            {
                if (publicId.equals(PUBLIC_ID_KEY))
                {
                    return getLocalInputSource("/org/jpox/jpox_1_0.dtd");
                }
            }
            // Else use systemId if possible
            else if (systemId != null)
            {
                if (systemId.equals(PUBLIC_ID_KEY))
                {
                    return getLocalInputSource("/org/jpox/jpox_1_0.dtd");
                }
                else if (systemId.equals("file:/org/jpox/jpox_1_0.dtd"))
                {
                    return getLocalInputSource("/org/jpox/jpox_1_0.dtd");
                }
                else if (systemId.startsWith("file:"))
                {
                    // Try to get the local file
                    return getLocalInputSource(systemId.substring(5));
                }
            }

            JPOXLogger.METADATA.error(this.getClass().getName() + ".resolveEntity(\"" + publicId + "\", \"" + systemId + "\") could not be found!");
            return null;
        }
        catch (Exception e)
        {
        	JPOXLogger.METADATA.error(this.getClass().getName() + ".resolveEntity(\"" + publicId + "\", \"" + systemId + "\") failed!", e);
            throw new SAXException(e.getMessage(), e);
        }
    }
    
    /**
     * Accessor for the input source for a path.
     * @param localPath The local path
     * @return The input source
     * @throws FileNotFoundException
     */
    private InputSource getLocalInputSource(String localPath)
    	throws FileNotFoundException
    {
    	JPOXLogger.METADATA.debug(this.getClass().getName()+".getLocalInputSource(\""+localPath+"\")");

    	InputStream in = getClass().getResourceAsStream(localPath);
    	if (in == null)
    	{
    		JPOXLogger.METADATA.fatal("local resource \""+localPath+"\" does not exist!!!");
    		throw new FileNotFoundException("Unable to load resource: "+localPath);
    	}
    	return new InputSource(new InputStreamReader(in));
    }
}