/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - coding standards
2003 Andy Jefferson - renamed from CandidateSetExpression
    ...
**********************************************************************/
package org.jpox.store.expression;

import javax.jdo.JDOUserException;

import org.jpox.ClassLoaderResolver;
import org.jpox.store.DatastoreAdapter;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.StatementText;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.rdbms.sqlidentifier.TableIdentifier;
import org.jpox.store.scostore.CollectionStore;

/**
 * An expression that represents some collection field in a query candidate
 * class, or a collection field in an object linked from the candidate class
 * by navigation.
 * <p>
 * When navigated through using contains(expr), the elements of the collection
 * are relationally joined onto the query statement.
 * </p>
 * <p>
 * As this is a Collection Expression it works equally for Sets and Lists,
 * and we use CollectionStore as the backing store interface, so that
 * ListStore and SetStore are equally applicable.
 *
 * @version $Revision: 1.7 $
 */
public class CollectionExpression extends ScalarExpression
{
    private final CollectionStore collStore;
    private final String fieldName;
    private final DatastoreAdapter dba;

    /**
     * Constructor.
     * @param qs The Query Statement
     * @param mapping The java field mapping
     * @param te The Table Expression
     * @param collStore the backing store.
     * @param fieldName Name of the field for the collection.
     **/
    public CollectionExpression(QueryStatement qs, 
                                JavaTypeMapping mapping,
                                TableExpression te,
                                CollectionStore collStore,
                                String fieldName)
    {
        super(qs);

        this.mapping = mapping;
        this.collStore = collStore;
        this.fieldName = fieldName;
        this.te = te;

        dba = qs.getStoreManager().getDatastoreAdapter();
    }
    
    /**
     * Executed when the contains() method is found in a query filter.
     * @param expr  The ScalarExpression passed as a parameter to contains().
     * @return  The BooleanExpression resulting from the contains() method.
     */
    public BooleanExpression containsMethod(ScalarExpression expr)
    {
        // ct... = "collection table"
        // et... = "element table"
        String ctJavaName = te.getRangeVariable().getNewJavaName(fieldName);
   
        if (expr instanceof UnboundVariable)
        {
            UnboundVariable var = (UnboundVariable)expr;
            //ctJavaName += + '.' + var.getVariableName();
            //String etJavaName = qs.getDefaultTableExpression().getRangeVariable().getJavaName() + '.' + var.getVariableName();
            String etJavaName = "UNBOUND" + '.' + var.getVariableName();

            TableIdentifier ctRangeVar = new TableIdentifier(dba, ctJavaName);
            TableIdentifier etRangeVar = new TableIdentifier(dba, etJavaName);

            var.bindTo(collStore.joinElementsTo(qs,
                							   mapping,
                							   te,
                                               ctRangeVar,
                                               var.getVariableType(),
                                               expr.te,
                                               expr.te == null ? etRangeVar : expr.te.getRangeVariable()));
            //TODO so,etimes we want distinct results, but we don't have enough columns in the select to make the distinction 
            qs.setDistinctResults(true);

            //TODO use appropriate mapping
            return new BooleanLiteral(qs, mapping, true);
        }
        else
        {
            TableIdentifier ctRangeVar = new TableIdentifier(dba, ctJavaName);
            DatastoreIdentifier etRangeVar;             
            if( expr.te == null ) // literals
            {
                int n = 0;

                do
                {
                    String etJavaName = ctJavaName + '.' + (++n);
                    etRangeVar = new TableIdentifier(dba, etJavaName);
                } while (qs.getTableExpression(etRangeVar) != null);
            }
            else
            {
                etRangeVar = expr.te.getRangeVariable();
            }

            ClassLoaderResolver clr=qs.getStoreManager().getClassLoaderResolver();
            ScalarExpression joinExpr =
               collStore.joinElementsTo(qs,
                                        mapping,
                                        te,
                                        ctRangeVar,
                                        clr.classForName(collStore.getElementType()),
                                        expr.te,
                                        etRangeVar);
            //TODO so,etimes we want distinct results, but we don't have enough columns in the select to make the distinction 
            qs.setDistinctResults(true);

            return joinExpr.eq(expr);
        }
    }

    /**
     * Return the BooleanExpression for a query filter in the form 
     * "collection.isEmpty()".
     * 
     * @return  The BooleanExpression for a query filter in the form 
     * "collection.isEmpty()".
     */
    public BooleanExpression isEmptyMethod()
    {
        // ct... = "collection table"
        String ctJavaName = te.getRangeVariable().getNewJavaName(fieldName);
        TableIdentifier ctRangeVar = new TableIdentifier(dba, ctJavaName);

        return new ExistsExpression(qs, collStore.getExistsSubquery(mapping, te, ctRangeVar), false);
    }

    public StatementText toStatementText()
    {
        throw new JDOUserException("Cannot reference collection object directly: field name = " + fieldName);
    }
}
