/**********************************************************************
Copyright (c) 2002 Kelly Grizzle (TJDO) and others.
All rights reserved.   This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
2003 Erik Bengtson - removed ununsed imports
2003 Andy Jefferson - coding standards
2004 Andy Jefferson - renamed from SetLiteral
    ...
**********************************************************************/
package org.jpox.store.expression;

import java.util.Collection;
import java.util.Iterator;

import org.jpox.store.DatastoreAdapter;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.query.QueryStatement;

/**
 * An SQL expression that will test if a column of a table falls within the
 * given Collection of values.  This is used for Querys where a transient
 * Collection is passed in as a parameter.
 *
 * @version $Revision: 1.4 $
 */
public class CollectionLiteral extends ScalarExpression
{
    private final boolean isEmpty;
    private final boolean containsNull;
    private final DatastoreAdapter dba;

    /**
     * Constructor.
     * @param qs     The QueryStatement the CollectionLiteral will be used in.
     * @param value  The transient Collection that is the value.
     */
    public CollectionLiteral(QueryStatement qs, JavaTypeMapping mapping, Collection value)
    {
        super(qs);

        this.mapping = mapping;
        containsNull = (value != null) && value.contains(null);
        dba = qs.getStoreManager().getDatastoreAdapter();

        /*
         * We'll consider the Collection to be empty if it is null, is really
         * empty, or only contains null. 
         * If it contains null we need a special case when creating the SQL.
         */
        isEmpty =
            (value == null) ||
            (value.isEmpty()) ||
            (value.size() == 1 && containsNull);

        // If the Collection is empty, don't build the list of SQLExpressions.
        if (!isEmpty)
        {
            st.append("(");

            boolean hadPrev = false;

            for (Iterator it=value.iterator(); it.hasNext();)
            {
                Object current = it.next();
                if (current != null)
                {
                    JavaTypeMapping m = dba.getMapping(current.getClass(), qs.getStoreManager());
                    ScalarExpression expr = m.newLiteral(qs,current);

                    // Append the SQLExpression (should be a literal) for the
                    // current element.
                    st.append(hadPrev ? "," : "");
                    st.append(expr);

                    hadPrev = true;
                }
            }

            st.append(")");
        }
    }

    /**
     * Method to check the containing of an element.
     * Return the BooleanExpression that results from 
     * CollectionLiteral.contains(SQLExpression).
     * 
     * @param expr The SQLExpression that is checked for membership in the
     *             Collection
     * @return The BooleanExpression that results from 
     *         CollectionLiteral.contains(SQLExpression).
     */
    public BooleanExpression containsMethod(ScalarExpression expr)
    {
        // TODO use appropriate mapping
        BooleanExpression returnVal = isEmpty ? new BooleanLiteral(qs, mapping, false) : expr.in(this);

        // The expression FIELD IN ('Foo','Bar',NULL) does not work as we need
        // it to. To get around this, if there is a null in the Collection
        // parameter, we will explicitly check if expr is NULL.
        if (containsNull)
        {
            returnVal = returnVal.ior(expr.eq(new NullLiteral(qs)));
        }

        return returnVal;
    }

    /**
     * Method to check for emptiness of the collection.
     * @return The BooleanExpression.
     **/
    public BooleanExpression isEmptyMethod()
    {
        // TODO use appropriate mapping
        return new BooleanLiteral(qs, mapping, isEmpty);
    }
}