/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
2004 Erik Bengtson - added containsEntry()
2004 Erik Bengtson - added getMethod()
    ...
**********************************************************************/
package org.jpox.store.expression;

import javax.jdo.JDOUserException;

import org.jpox.ClassLoaderResolver;
import org.jpox.store.DatastoreAdapter;
import org.jpox.store.StatementText;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.rdbms.sqlidentifier.TableIdentifier;
import org.jpox.store.scostore.MapStore;

/**
 * An expression that represents some Map field in a query candidate
 * class, or a Map field in an object linked from the candidate class
 * by navigation.
 * <p>
 * When navigated through using containsKey(expr) the keys of the Map are
 * relationally joined onto the query statement. When navigated through using
 * containsValue(expr) the values of the Map are relationally joined onto the
 * query statement. These 2 methods are required for JDO 2.0, whilst the
 * isEmpty() and contains() are JDO 1.0.1. containsEntry() is a JPOX extension.
 * </p>
 *
 * @version $Revision: 1.7 $
 */
public class MapExpression extends ScalarExpression
{
    private final TableExpression te;
    private final MapStore mapStore;
    private final String fieldName;
    private final DatastoreAdapter dba;

    /**
     * Constructor.
     * @param qs The Query Statement
     * @param mapping The java field mapping
     * @param te The Table Expression
     * @param mapStore the backing store.
     * @param fieldName Name of the field for the map.
     **/
    public MapExpression(QueryStatement qs, 
                         JavaTypeMapping mapping,
                         TableExpression te,
                         MapStore mapStore,
                         String fieldName)
    {
        super(qs);

        this.mapping  = mapping;
        this.te = te;
        this.mapStore  = mapStore;
        this.fieldName = fieldName;

        dba = qs.getStoreManager().getDatastoreAdapter();
    }

    /**
     * Executed when the containsKey() method is found in a query filter.
     * This is added in JDO 2.0. 
     * @param expr The ScalarExpression param for map.containsKey(...).
     * @return The BooleanExpression resulting from map.containsKey().
     **/
    public BooleanExpression containsKeyMethod(ScalarExpression expr)
    {
        // mt... = "map table"
        // vt... = "value table"
        String mtJavaName = te.getRangeVariable().getNewJavaName(fieldName);

        if (expr instanceof UnboundVariable)
        {
            UnboundVariable var = (UnboundVariable)expr;
            String vtJavaName = "UNBOUND" + '.' + var.getVariableName();

            TableIdentifier mtRangeVar = new TableIdentifier(dba, mtJavaName);
            TableIdentifier vtRangeVar = new TableIdentifier(dba, vtJavaName);

            var.bindTo(mapStore.joinKeysTo(qs,
                                           mapping,
                                           te,
                                           mtRangeVar,
                                           var.getVariableType(),
                                           vtRangeVar));
            qs.setDistinctResults(true);

            return new BooleanLiteral(qs, mapping, true);
        }
        else
        {
            TableIdentifier mtRangeVar = new TableIdentifier(dba, mtJavaName);
            TableIdentifier vtRangeVar;
            int n = 0;

            do
            {
                String vtJavaName = mtJavaName + '.' + (++n);
                vtRangeVar = new TableIdentifier(dba, vtJavaName);
            } while (qs.getTableExpression(vtRangeVar) != null);

            ClassLoaderResolver clr=qs.getStoreManager().getClassLoaderResolver();
            ScalarExpression joinKeysExpr =
                    mapStore.joinKeysTo(qs,
                                        mapping,
                                        te,
                                        mtRangeVar,
                                        clr.classForName(mapStore.getKeyType()),
                                        vtRangeVar);
            qs.setDistinctResults(true);
            return joinKeysExpr.eq(expr);
        }
    }

    /**
     * Executed when the containsEntry() method is found in a query filter.
     * This is a JPOX extension to JDOQL 1.0.1/2.0.
     * @param keyExpr The ScalarExpression param for map.containsEntry(...).
     * @param valueExpr The ScalarExpression param for map.containsEntry(...).
     * @return The BooleanExpression resulting from map.containsEntry().
     **/
    public BooleanExpression containsEntryMethod(ScalarExpression keyExpr,
                                                 ScalarExpression valueExpr)
    {
        // mt... = "map table"
        // kt... = "key table"
        // vt... = "value table"
        String mtJavaName = te.getRangeVariable().getNewJavaName(fieldName);
        TableIdentifier mtRangeVar = new TableIdentifier(dba, mtJavaName);

        // obtains a variable name for key
        Class keyType;
        TableIdentifier ktRangeVar;
        UnboundVariable keyVar = null;
        ClassLoaderResolver clr=qs.getStoreManager().getClassLoaderResolver();
        int nKey = 0;
        if (keyExpr instanceof UnboundVariable)
        {
            keyVar = (UnboundVariable)keyExpr;
            String ktJavaName = "UNBOUND" + '.' + keyVar.getVariableName();
            keyType = keyVar.getVariableType();
            ktRangeVar = new TableIdentifier(dba, ktJavaName);
        }
        else
        {
            keyType = clr.classForName(mapStore.getKeyType());
            do
            {
                String ktJavaName = mtJavaName + '.' + (++nKey);
                ktRangeVar = new TableIdentifier(dba, ktJavaName);
            } while (qs.getTableExpression(ktRangeVar) != null);
        }

        // obtains a variable name for value
        Class valueType;
        TableIdentifier vtRangeVar;
        UnboundVariable valueVar = null;
        int nValue = 0;
        if (valueExpr instanceof UnboundVariable)
        {
            valueVar = (UnboundVariable)valueExpr;
            String vtJavaName = "UNBOUND" + '.' + valueVar.getVariableName();
            valueType = valueVar.getVariableType();
            vtRangeVar = new TableIdentifier(dba, vtJavaName);
        }
        else
        {
            valueType = clr.classForName(mapStore.getValueType());
            do
            {
                String vtJavaName = mtJavaName + '.' + (++nValue);
                vtRangeVar = new TableIdentifier(dba, vtJavaName);
            } while (qs.getTableExpression(vtRangeVar) != null);
        }

        ScalarExpression[] qclKeyValues = 
            mapStore.joinKeysValuesTo(qs,
                                mapping,
                                te,
                                mtRangeVar,
                                keyType,
                                valueType,
                                ktRangeVar,
                                vtRangeVar);
        if (keyExpr instanceof UnboundVariable)
        {
            keyVar.bindTo(qclKeyValues[0]);
        }
        if (valueExpr instanceof UnboundVariable)
        {
            valueVar.bindTo(qclKeyValues[1]);
        }
        
        qs.setDistinctResults(true);

        BooleanExpression kExpr = null;
        BooleanExpression vExpr = null;

        if ((keyExpr instanceof UnboundVariable) &&
            (valueExpr instanceof UnboundVariable))
        {
            return new BooleanLiteral(qs, mapping, true);
        }
        else if (!(keyExpr instanceof UnboundVariable) &&
                 !(valueExpr instanceof UnboundVariable))
        {
            kExpr = keyExpr.eq(qclKeyValues[0]);
            vExpr = valueExpr.eq(qclKeyValues[1]);
            return kExpr.and(vExpr);
        }
        else if (kExpr != null)
        {
            return kExpr;
        }
        else if (vExpr != null)
        {
            return vExpr;
        }

        //will never get here
        return new BooleanLiteral(qs, mapping, true);
    }

    /**
     * Executed when a contains() method is found in a query filter.
     * This simply uses the containsValueMethod() to check against values.
     * @param expr  The ScalarExpression param for map.contains(...).
     * @return  The BooleanExpression resulting from map.contains().
     **/
    public BooleanExpression containsMethod(ScalarExpression expr)
    {
        return containsValueMethod(expr);
    }

    /**
     * Executed when the containsValue() method is found in a query filter.
     * This is added in JDO 2.0.
     * @param expr  The ScalarExpression param for map.containsValue(...).
     * @return  The BooleanExpression resulting from map.containsValue().
     **/
    public BooleanExpression containsValueMethod(ScalarExpression expr)
    {
        // mt... = "map table"
        // vt... = "value table"
        String mtJavaName = te.getRangeVariable().getNewJavaName(fieldName);

        if (expr instanceof UnboundVariable)
        {
            UnboundVariable var = (UnboundVariable)expr;
            String vtJavaName = "UNBOUND" + '.' + var.getVariableName();

            TableIdentifier mtRangeVar = new TableIdentifier(dba, mtJavaName);
            TableIdentifier vtRangeVar = new TableIdentifier(dba, vtJavaName);

            var.bindTo(mapStore.joinValuesTo(qs,
                                             mapping,
                                             te,
                                             mtRangeVar,
                                             var.getVariableType(),
                                             vtRangeVar));
            qs.setDistinctResults(true);

            return new BooleanLiteral(qs, mapping, true);
        }
        else
        {
            TableIdentifier mtRangeVar = new TableIdentifier(dba, mtJavaName);
            TableIdentifier vtRangeVar;
            int n = 0;

            do
            {
                String vtJavaName = mtJavaName + '.' + (++n);
                vtRangeVar = new TableIdentifier(dba, vtJavaName);
            } while (qs.getTableExpression(vtRangeVar) != null);

            ClassLoaderResolver clr=qs.getStoreManager().getClassLoaderResolver();
            ScalarExpression joinValueExpr =
                mapStore.joinValuesTo(qs,
                                      mapping,
                                      te,
                                      mtRangeVar,
                                      clr.classForName(mapStore.getValueType()),
                                      vtRangeVar);
            qs.setDistinctResults(true);
            return joinValueExpr.eq(expr);
        }
    }

    /**
     * Return the BooleanExpression for a query filter in the form 
     * "map.isEmpty()".
     * @return  The BooleanExpression for a query filter in the form 
     *          "map.isEmpty()".
     **/
    public BooleanExpression isEmptyMethod()
    {
        // mt... = "map table"
        String mtJavaName = te.getRangeVariable().getNewJavaName(fieldName);
        TableIdentifier mtRangeVar = new TableIdentifier(dba, mtJavaName);

        return new ExistsExpression(qs, mapStore.getExistsSubquery(mapping, te, mtRangeVar), false);
    }

    /**
     * Method to get a value from the Map for a key
     * Currently not supported.
     * @return The statement
     **/
    public ScalarExpression getMethod(ScalarExpression expr)
    {
        // mt... = "map table"
        // vt... = "value table"
        String mtJavaName = te.getRangeVariable().getNewJavaName(fieldName);

        TableIdentifier mtRangeVar;
        TableIdentifier ktRangeVar;
        TableIdentifier vtRangeVar;
        int n = 0;
        if (expr instanceof UnboundVariable)
        {
            UnboundVariable var = (UnboundVariable)expr;
            String ktJavaName = "UNBOUND" + '.' + var.getVariableName();
            String vtJavaName = "UNBOUNDVALUE" + '.' + var.getVariableName();

            mtRangeVar = new TableIdentifier(dba, mtJavaName);
            ktRangeVar = new TableIdentifier(dba, ktJavaName);
            vtRangeVar = new TableIdentifier(dba, vtJavaName);

        }
        else
        {
            mtRangeVar = new TableIdentifier(dba, mtJavaName);

            do
            {
                String ktJavaName = mtJavaName + '.' + (++n);
                ktRangeVar = new TableIdentifier(dba, ktJavaName);
            } while (qs.getTableExpression(ktRangeVar) != null);

            n = 0;

            do
            {
                String vtJavaName = mtJavaName + '.' + (++n);
                vtRangeVar = new TableIdentifier(dba, vtJavaName);
            } while (qs.getTableExpression(vtRangeVar) != null);
        }
            
        ClassLoaderResolver clr=qs.getStoreManager().getClassLoaderResolver();
        ScalarExpression[] joinKeysExpr =
            mapStore.joinKeysToGet(qs,
                mapping,
                te,
                mtRangeVar,
                clr.classForName(mapStore.getKeyType()),
                ktRangeVar,
                vtRangeVar);
        if (expr instanceof UnboundVariable)
        {
            UnboundVariable var = (UnboundVariable)expr;
            var.bindTo(joinKeysExpr[0]);
        }
        qs.setDistinctResults(true);
        return new ObjectExpression(qs,joinKeysExpr[1],joinKeysExpr[0].eq(expr));
    }
    
    /**
     * Method to return the statement text. Throws an exception since this
     * object is inaccessible directly.
     * @return The statement
     **/
    public StatementText toStatementText()
    {
        throw new JDOUserException("Cannot reference Map object directly: field name = " + fieldName);
    }
}
