/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
    ...
**********************************************************************/
package org.jpox.store.expression;

import java.util.Collection;
import java.util.Iterator;
import java.util.Map;

import org.jpox.store.DatastoreAdapter;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.query.QueryStatement;

/**
 * An SQL expression that will test if a column of a table falls within the
 * given Map's values.  This is used for Querys where a transient Map is passed
 * in as a parameter.
 *
 * @version $Revision: 1.4 $
 */
public class MapValueLiteral extends ScalarExpression
{
    private final boolean isEmpty;
    private final boolean containsNull;
    private final DatastoreAdapter dba;

    /**
     * Constructor.
     * @param qs  The QueryStatement the MapValueLiteral will be used in.
     * @param map The transient Map that is the value.
     */
    public MapValueLiteral(QueryStatement qs, JavaTypeMapping mapping, Map map)
    {
        super(qs);
        this.mapping = mapping;
        containsNull = (map != null) && map.containsValue(null);
        dba = qs.getStoreManager().getDatastoreAdapter();

        // We'll consider the Map to be empty if it is null, is really
        // empty, or only contains null. 
        // If it contains null we need a special case when creating the SQL.
        isEmpty =
            (map == null) ||
            (map.isEmpty()) ||
            (map.size() == 1 && containsNull);

        // If the Map is empty, don't build the list of SQLExpressions.
        if (!isEmpty)
        {
            st.append("(");

            boolean hadPrev = false;

            Collection values=map.values();
            for (Iterator it=values.iterator(); it.hasNext();)
            {
                Object current = it.next();
                if (null != current)
                {
                    JavaTypeMapping m = dba.getMapping(current.getClass(), qs.getStoreManager());
                    ScalarExpression expr = m.newLiteral(qs, current);

                    // Append the SQLExpression (should be a literal) for the
                    // current element.
                    st.append(hadPrev ? "," : "");
                    st.append(expr);

                    hadPrev = true;
                }
            }

            st.append(")");
        }
    }

    /**
     * Method to check the containing of a key in the Map.
     * Return the BooleanExpression that results from 
     * MapValueLiteral.contains(SQLExpression).
     * @param expr The SQLExpression that is checked for membership in the Map.
     * @return The BooleanExpression that results from 
     *         MapValueLiteral.contains(SQLExpression).
     */
    public BooleanExpression containsMethod(ScalarExpression expr)
    {
        BooleanExpression returnVal = isEmpty ? new BooleanLiteral(qs, mapping, false) : expr.in(this);

        // The expression FIELD IN ('Foo','Bar',NULL) does not work as we need
        // it to. To get around this, if there is a null in the Map
        // parameter, we will explicitly check if expr is NULL.
        if (containsNull)
        {
            returnVal = returnVal.ior(expr.eq(new NullLiteral(qs)));
        }

        return returnVal;
    }

    /**
     * Method to check for emptiness of the collection.
     * @return The BooleanExpression.
     **/
    public BooleanExpression isEmptyMethod()
    {
        return new BooleanLiteral(qs, mapping, isEmpty);
    }
}
