/**********************************************************************
Copyright (c) 2003 Erik Bengtsonn and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - coding standards
    ...
**********************************************************************/
package org.jpox.store.expression;

import org.jpox.exceptions.ClassNotPersistenceCapableException;
import org.jpox.store.DatastoreAdapter;
import org.jpox.store.DatastoreClass;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.rdbms.sqlidentifier.TableIdentifier;

/**
 * Representation of an Object expression in a Query.
 *
 * @version $Revision: 1.12 $
 **/
public class ObjectExpression extends ScalarExpression
{
    protected ScalarExpression conditionExpr;
    private Class castType;
    private String fieldName;
    
    protected ObjectExpression(QueryStatement qs)
    {
        super(qs);
    }

    public ObjectExpression(QueryStatement qs, JavaTypeMapping mapping, TableExpression te)
    {
        super(qs, mapping, te);
    }

    public ObjectExpression(QueryStatement qs, ScalarExpression expr, ScalarExpression conditionExpr)
    {
        super(qs);
        this.mapping = expr.mapping;
        this.st.append(expr.st.toString());
        this.expressionList = expr.expressionList;
        this.conditionExpr = conditionExpr;
    }
    
    public ObjectExpression(QueryStatement qs, JavaTypeMapping mapping, TableExpression te, String outerJoinSuffix)
    {
        this(qs, mapping, te);
        st.append(outerJoinSuffix);
    }
    
    public ScalarExpression cast(Class castType)
    {
        ObjectExpression objectCast;
        TableExpression te = qs.getTableExpression(this.qs.getStoreManager().getDatastoreClass(castType.getName()).getName());
        DatastoreClass dc = this.qs.getStoreManager().getDatastoreClass(castType.getName());
        if( te == null )
        {
            String jtJavaName = this.te.getRangeVariable().getJavaName();
            
            if (castType != null && !castType.equals(mapping.getType()))
            {
                String castTypeName = castType.getName();
                jtJavaName = this.te.getRangeVariable().getNewJavaName(castTypeName.substring(castTypeName.lastIndexOf('.') + 1));
            }        

            TableIdentifier jtRangeVar = new TableIdentifier(qs.getStoreManager().getDatastoreAdapter(), jtJavaName);
            TableExpression jtTblExpr = qs.getTableExpression(jtRangeVar);

            if (jtTblExpr == null)
            {
                jtTblExpr = qs.newTableExpression(dc, jtRangeVar);
            }    
            te = jtTblExpr;
            qs.leftOuterJoin(this,dc.getIDMapping().newScalarExpression(qs,jtTblExpr), jtTblExpr, true);
        }
        objectCast = new ObjectExpression(qs,dc.getIDMapping(),te);
        objectCast.conditionExpr = this.conditionExpr;
        return objectCast;
    }
    

    public BooleanExpression eq(ScalarExpression expr)
    {
        BooleanExpression bExpr = null;
        if (expr instanceof NullLiteral)
        {
            for( int i=0; i<this.expressionList.size(); i++ )
            {
                if( bExpr == null )
                {
                    bExpr = expr.eq(this.expressionList.getExpression(i));
                }
                else
                {
                    bExpr = bExpr.and(expr.eq(this.expressionList.getExpression(i)));
                }
            }
        }
        else if (expr instanceof ObjectLiteral)
        {
            bExpr = expr.eq(this);
        }	        
        else if (expr instanceof ObjectExpression)
        {
            for( int i=0; i<this.expressionList.size(); i++ )
            {
                ScalarExpression source = this.expressionList.getExpression(i);
                ScalarExpression target = expr.expressionList.getExpression(i);
                if( bExpr == null )
                {
                    bExpr = source.eq(target);
                }
                else
                {
                    bExpr = bExpr.and(source.eq(target));
                }
            }
        }
        else if (expr instanceof UnboundVariable)
        {
            bExpr = expr.eq(this);
        }
        else if (expr instanceof BooleanBitColumnExpression)
        {
            bExpr = null;
        }
        else
        {
            bExpr = super.eq(expr);
        }
        if( conditionExpr != null)
        {
            return new BooleanExpression(conditionExpr, OP_AND, bExpr);
        }
        return bExpr;
    }

    public BooleanExpression noteq(ScalarExpression expr)
    {
        BooleanExpression bExpr = null;
        if (expr instanceof NullLiteral)
        {
            for( int i=0; i<this.expressionList.size(); i++ )
            {
                if( bExpr == null )
                {
                    bExpr = expr.eq(this.expressionList.getExpression(i));
                }
                else
                {
                    bExpr = bExpr.and(expr.eq(this.expressionList.getExpression(i)));
                }
            }
            bExpr = new BooleanExpression(OP_NOT, bExpr.encloseWithInParentheses());
        }
        else if (expr instanceof ObjectLiteral)
        {
            bExpr = expr.noteq(this);
        }
        else if (expr instanceof ObjectExpression)
        {
            for( int i=0; i<this.expressionList.size(); i++ )
            {
                ScalarExpression source = this.expressionList.getExpression(i);
                ScalarExpression target = expr.expressionList.getExpression(i);
                if( bExpr == null )
                {
                    bExpr = source.eq(target);
                }
                else
                {
                    bExpr = bExpr.and(source.eq(target));
                }
            }
            bExpr = new BooleanExpression(OP_NOT, bExpr.encloseWithInParentheses());
        }
        else if (expr instanceof UnboundVariable)
        {
            bExpr = expr.noteq(this);
        }       
        else if (expr instanceof BooleanBitColumnExpression)
        {
            if( conditionExpr != null )
            {
                bExpr = new BooleanExpression(ScalarExpression.OP_NOT, conditionExpr);
            }
        }        
        else
        {
            bExpr = super.noteq(expr);
        }
        if( conditionExpr != null)
        {
            return new BooleanExpression(conditionExpr, OP_AND, bExpr);
        }
        return bExpr;
    }
    
    public BooleanExpression in(ScalarExpression expr) 
    {
        return new BooleanExpression(this, OP_IN, expr);
    }

    /**
     * Access a normal field or a collection field in a PersistenceCapable class
     * @param subfieldName the field to be accessed
     * @return The field expression
     */
    public ScalarExpression accessField(String subfieldName, boolean innerJoin)
    {
        DatastoreClass table;
        
        try
        {
            table = qs.getStoreManager().getDatastoreClass(mapping.getType().getName());
        }
        catch( ClassNotPersistenceCapableException cnpce )
        {
            return te.newFieldExpression(subfieldName);
        }
        
        DatastoreAdapter dba = qs.getStoreManager().getDatastoreAdapter();
        // jt... = "joined table"
        String jtJavaName = te.getRangeVariable().getJavaName();

        if(fieldName != null)
        {
            jtJavaName += '.' + fieldName;
        }
        if (!subfieldName.equals("this"))
        {
            jtJavaName += '.' + subfieldName;
        }

        if (castType != null && !castType.equals(mapping.getType()))
        {
            String castTypeName = castType.getName();
            jtJavaName += '.' + castTypeName.substring(castTypeName.lastIndexOf('.') + 1);
        }        
        
        TableIdentifier jtRangeVar = new TableIdentifier(dba, jtJavaName);
        TableExpression jtTblExpr = qs.getTableExpression(jtRangeVar);

        if (jtTblExpr == null)
        {
            DatastoreClass jt = table;
            jtTblExpr = qs.newTableExpression(jt, jtRangeVar);

            ScalarExpression jtExpr = jt.getIDMapping().newScalarExpression(qs,jtTblExpr);
            ScalarExpression expr = mapping.newScalarExpression(qs,te);

            if (mapping.isNullable())
            {
                if( innerJoin )
                {
                    qs.innerJoin(jtExpr, expr, jtTblExpr, true);
                }
                else
                {
                    qs.leftOuterJoin(jtExpr, expr, jtTblExpr, true);
                }
            }
            else
            {
                qs.innerJoin(jtExpr, expr, jtTblExpr, true);
            }
        }

        return jtTblExpr.newFieldExpression(subfieldName);
    }
    
    
    /**
     * Set the field which this expression was created from
     * @param fieldName The fieldName to set.
     */
    public void setFieldName(String fieldName)
    {
        this.fieldName = fieldName;
    }
}
