/**********************************************************************
Copyright (c) 2003 Erik Bengtson and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - coding standards
    ...
**********************************************************************/
package org.jpox.store.expression;

import java.lang.reflect.Field;
import java.security.AccessController;
import java.security.PrivilegedAction;

import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOHelper;
import javax.jdo.JDOUserException;
import javax.jdo.spi.PersistenceCapable;

import org.jpox.PersistenceManager;
import org.jpox.metadata.ClassMetaData;
import org.jpox.metadata.FieldMetaData;
import org.jpox.metadata.IdentityType;
import org.jpox.state.StateManagerImpl;
import org.jpox.store.DatastoreAdapter;
import org.jpox.store.OID;
import org.jpox.store.fieldmanager.FieldManager;
import org.jpox.store.fieldmanager.SimpleFieldManager;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.query.QueryStatement;

/**
 * Representation of an Object literal in a query.
 *
 * @version $Revision: 1.5 $
 **/
public class ObjectLiteral extends ObjectExpression implements Literal
{
    private Object value;
    
    // passed by parameters or variable in a query
    private Class mappingClass;
    
    public ObjectLiteral(QueryStatement qs, JavaTypeMapping m, Object value)
    {
        super(qs);

        st.appendParameter(m, value);
        this.mapping = m;
        this.value = value;
    }
    
    public ObjectLiteral(QueryStatement qs, JavaTypeMapping m, Object value, Class mappingClass)
    {
        super(qs);

        st.appendParameter(m, value);
        this.value = value;
        this.mapping = m;
        this.mappingClass = mappingClass;
    }    

    public Object getValue()
    {
        return value;
    }
    
    public BooleanExpression eq(ScalarExpression expr)
    {
        if (expr instanceof ObjectLiteral)
        {
            return new BooleanLiteral(qs, mapping, value.equals(((ObjectLiteral)expr).value));
        }
        else if (expr instanceof BooleanBitColumnExpression)
        {
            return null;
        }
        else if (expr instanceof ObjectExpression)
        {
            BooleanExpression bExpr = null;
            if( value == null )
            {
                bExpr = expr.eq(new NullLiteral(qs));
            }
            else
            {
	            if( value instanceof OID )
	            {
	                DatastoreAdapter dba = qs.getStoreManager().getDatastoreAdapter();	                
                    JavaTypeMapping m = dba.getMapping(((OID)value).getValue().getClass(), qs.getStoreManager());
                    ScalarExpression oidExpr =  m.newLiteral(qs,((OID)value).getValue());
                    bExpr = expr.expressionList.getExpression(0).eq(oidExpr);                        
	            }
	            else
	            {
	                for( int i=0; i<expr.expressionList.size(); i++ )
	                {
	                    ScalarExpression source = expr.expressionList.getExpression(i);
	                    ClassMetaData cmd = qs.getStoreManager().getMetaDataManager().getMetaDataForClass(value.getClass());
	                    if( cmd == null )
	                    {
	                        //if cmd == null, the value is probably not PersistenceCapable, thus the ClassMetaData is null
	                        bExpr = new BooleanLiteral(qs, mapping, false).eq(new BooleanLiteral(qs, mapping, true));
	                    }
	                    else
	                    {
		                    if( cmd.getIdentityType() == IdentityType.APPLICATION )
		                    {
		                        FieldMetaData fmd = cmd.getManagedFieldAbsolute(cmd.getPrimaryKeyFieldNumbers()[i]);

			                    Object fieldValue = getFieldValue(fmd,value);
			                    JavaTypeMapping mapping = qs.getStoreManager().getDatastoreAdapter().getMapping(fieldValue.getClass(), qs.getStoreManager());                
			                    ScalarExpression target = mapping.newLiteral(qs,fieldValue);
			                    if( bExpr == null )
			                    {
			                        bExpr = source.eq(target);
			                    }
			                    else
			                    {
			                        bExpr = bExpr.and(source.eq(target));
			                    }
		                    }
		                    else
		                    {
		                        DatastoreAdapter dba = qs.getStoreManager().getDatastoreAdapter();	                
		                        JavaTypeMapping m = dba.getMapping(((OID)JDOHelper.getObjectId(value)).getValue().getClass(), qs.getStoreManager());
		                        ScalarExpression oidExpr =  m.newLiteral(qs,((OID)JDOHelper.getObjectId(value)).getValue());
		                        bExpr = expr.expressionList.getExpression(0).eq(oidExpr);                        
		                    }
	                    }
	                }                
	                if( ((ObjectExpression)expr).conditionExpr != null)
	                {
	                    bExpr = new BooleanExpression(((ObjectExpression)expr).conditionExpr, OP_AND, bExpr);
	                }
	            }
            }
            return bExpr;
        }
        else
        {
            return super.eq(expr);
        }
    }

    public BooleanExpression noteq(ScalarExpression expr)
    {
        if (expr instanceof ObjectLiteral)
        {
            return new BooleanLiteral(qs, mapping, !value.equals(((ObjectLiteral)expr).value));
        }
        else if (expr instanceof ObjectExpression)
        {
            BooleanExpression bExpr = null;
            if( value == null )
            {
                bExpr = expr.eq(new NullLiteral(qs));
            }
            else
            {
	            if( value instanceof OID )
	            {
	                DatastoreAdapter dba = qs.getStoreManager().getDatastoreAdapter();	                
                    JavaTypeMapping m = dba.getMapping(((OID)value).getValue().getClass(), qs.getStoreManager());
                    ScalarExpression oidExpr =  m.newLiteral(qs,((OID)value).getValue());
                    bExpr = expr.expressionList.getExpression(0).eq(oidExpr);                        
	            }
	            else
	            {
	                for( int i=0; i<expr.expressionList.size(); i++ )
	                {
	                    ScalarExpression source = expr.expressionList.getExpression(i);
	                    ClassMetaData cmd = qs.getStoreManager().getMetaDataManager().getMetaDataForClass(value.getClass());
	                    if( cmd == null )
	                    {
	                        //if cmd == null, the value is probably not PersistenceCapable, thus the ClassMetaData is null
	                        bExpr = new BooleanLiteral(qs, mapping, false).eq(new BooleanLiteral(qs, mapping, true));
	                    }
	                    else
	                    {
		                    if( cmd.getIdentityType() == IdentityType.APPLICATION )
		                    {
		                        FieldMetaData fmd = cmd.getManagedFieldAbsolute(cmd.getPrimaryKeyFieldNumbers()[i]);

			                    Object fieldValue = getFieldValue(fmd,value);
			                    JavaTypeMapping mapping = qs.getStoreManager().getDatastoreAdapter().getMapping(fieldValue.getClass(), qs.getStoreManager());                
			                    ScalarExpression target = mapping.newLiteral(qs,fieldValue);
			                    if( bExpr == null )
			                    {
			                        bExpr = source.eq(target);
			                    }
			                    else
			                    {
			                        bExpr = bExpr.and(source.eq(target));
			                    }
		                    }
		                    else
		                    {
		                        DatastoreAdapter dba = qs.getStoreManager().getDatastoreAdapter();	                
		                        JavaTypeMapping m = dba.getMapping(((OID)JDOHelper.getObjectId(value)).getValue().getClass(), qs.getStoreManager());
		                        ScalarExpression oidExpr =  m.newLiteral(qs,((OID)JDOHelper.getObjectId(value)).getValue());
		                        bExpr = expr.expressionList.getExpression(0).eq(oidExpr);                        
		                    }
	                    }
	                }                
	                if( ((ObjectExpression)expr).conditionExpr != null)
	                {
	                    bExpr = new BooleanExpression(((ObjectExpression)expr).conditionExpr, OP_AND, bExpr);
	                }
	            }
            }
	        bExpr = new BooleanExpression(this, OP_NOTEQ, bExpr);
            return bExpr;
        }
        else
        {
            return super.noteq(expr);
        }
    }

    /**
     * Reads a field of an object passed by parameter to the query
     * @param subfieldName
     * @return a new ScalarExpression from the field value 
     */    
    public ScalarExpression accessField(String subfieldName, boolean innerJoin)
    {
        Object fieldValue = null;
        try
        {
            
            FieldMetaData fmd = qs.getStoreManager().getMetaDataManager().getMetaDataForClass(this.mappingClass).getField(subfieldName);
            fieldValue = getFieldValue(fmd,value);
        }
        catch (Exception e)
        {
            fieldValue = getFieldValue(value, subfieldName);
        }               
        try
        {
            if (fieldValue == null)
            {
                return new NullLiteral(this.qs);
            }
            DatastoreAdapter dba = qs.getStoreManager().getDatastoreAdapter();
            JavaTypeMapping mapping;
            if (mappingClass != null && subfieldName == null)
            {
                mapping = dba.getMapping(mappingClass, qs.getStoreManager());
            }
            else
            {
                mapping = dba.getMapping(fieldValue.getClass(), qs.getStoreManager());                
            }
            return mapping.newLiteral(qs,fieldValue);
        }
        catch (SecurityException e1)
        {
            throw new JDOUserException("Cannot access field: "+subfieldName);
        }
        catch (IllegalArgumentException e)
        {
            throw new JDOUserException("Cannot access field: "+subfieldName);
        }
        catch (Exception e)
        {
            throw new JDOUserException("Cannot access field: "+subfieldName+" " +e);
        }       
    }

    public String toString()
    {
        return super.toString() + " = " + value.toString();
    }
    
    /**
     * Helper method to retrieve the java.lang.reflect.Field
     * @param clazz the Class instance of the declaring class or interface
     * @param fieldName the field name
     * @return The field
     */
    private Field getDeclaredFieldPrivileged(final Class clazz, 
                                             final String fieldName)
    {
        if ((clazz == null) || (fieldName == null))
        {
            return null;
        }

        return (Field) AccessController.doPrivileged(
            new PrivilegedAction()
            {
                public Object run ()
                {
                    Class seekingClass = clazz;
                    do
                    {
	                    try
	                    {
	                        return seekingClass.getDeclaredField(fieldName);
	                    }
	                    catch (SecurityException ex)
	                    {
	                        throw new JDOFatalInternalException("CannotGetDeclaredField");
	                    }
	                    catch (NoSuchFieldException ex)
	                    {
	                        // do nothing, we will return null later if no 
	                        // field is found in this class or superclasses
	                    }
	                    catch (LinkageError ex)
	                    {
	                        throw new JDOFatalInternalException("ClassLoadingError");
	                    }
	                    seekingClass = seekingClass.getSuperclass();
                    } while(seekingClass != null);
                    
                    //no field found
                    return null;
                }
            }
            );
    }    

    /**
     * Get the field value of a managed field. 
     * @param fmd the FieldMetaData
     * @param object the pc object
     * @return The field value
     */
    private Object getFieldValue(FieldMetaData fmd, Object object)
    {
        PersistenceManager pm = (PersistenceManager) JDOHelper.getPersistenceManager(object);
        if( pm == null )
        {
            return getFieldValue(object, fmd.getName());
        }
        StateManagerImpl sm = (StateManagerImpl) pm.findStateManager((PersistenceCapable)object);

        FieldManager fm = new SimpleFieldManager();

        if( !fmd.isPrimaryKey() )
        {
            // we expect that primary key field are non null
            sm.isLoaded((PersistenceCapable) object, fmd.getAbsoluteFieldNumber());
        }
        sm.provideFields(new int[] {fmd.getAbsoluteFieldNumber()},fm);

        return fm.fetchObjectField(fmd.getAbsoluteFieldNumber());
    }
    
    private Object getFieldValue(Object object, String fieldName)
    {
        Object fieldValue;
        final Field field = getDeclaredFieldPrivileged(object.getClass(), fieldName);
        if (field == null)
        {
            throw new JDOUserException("Cannot access field: "+fieldName);
        }
        
        try
        {
            // if the field is not accessible, try to set the accessible flag.
            if (!field.isAccessible())
            {
                try
                {
                    AccessController.doPrivileged(
                        new PrivilegedAction()
                        {
                            public Object run ()
                            {
                                field.setAccessible(true);
                                return null;
                            }
                        });
                }
                catch (SecurityException ex)
                {
                    throw new JDOFatalInternalException("Cannot access field: "+fieldName);
                }
            }               
            fieldValue = field.get(object);
        }
        catch (IllegalArgumentException e2)
        {
            throw new JDOUserException("Cannot access field: "+fieldName);
        }
        catch (IllegalAccessException e2)
        {
            throw new JDOUserException("Cannot access field: "+fieldName);
        }
        return fieldValue;       
    }
}
