/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - commented
    ...
**********************************************************************/
package org.jpox.store.expression;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;

import javax.jdo.JDOException;
import javax.jdo.JDOUserException;

import org.jpox.store.DatastoreField;
import org.jpox.store.StatementText;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.query.QueryStatement.QueryExpression;
import org.jpox.store.rdbms.adapter.DatabaseAdapter;

/**
 * Representation of a Scalar expression in a Query.
 *
 * @version $Revision: 1.7 $
 **/
public abstract class ScalarExpression
{
    /**
     * Inner class representing an Operator
     **/
    protected static class Operator
    {
        protected final String sql;
        protected final int precedence;

        public Operator(String sql, int precedence)
        {
            this.sql = sql;
            this.precedence = precedence;
        }

        public String toString()
        {
            return sql;
        }
    }

    /**
     * Inner class for "Monadic" operator
     **/
    protected static class MonadicOperator extends Operator
    {
        public MonadicOperator(String sql, int precedence)
        {
            super(sql, precedence);
        }

        public boolean isHigherThan(Operator op)
        {
            if (op == null)
            {
                return false;
            }
            else
            {
                return precedence > op.precedence;
            }
        }
    }

    /**
     * Inner class for "Dyadic" operator
     **/
    protected static class DyadicOperator extends Operator
    {
        private final boolean isAssociative;

        public DyadicOperator(String sql, int precedence, boolean isAssociative)
        {
            super(" " + sql + " ", precedence);
            this.isAssociative = isAssociative;
        }

        public boolean isHigherThanLeftSide(Operator op)
        {
            if (op == null)
            {
                return false;
            }
            else
            {
                return precedence > op.precedence;
            }
        }

        public boolean isHigherThanRightSide(Operator op)
        {
            if (op == null)
            {
                return false;
            }
            else if (precedence == op.precedence)
            {
                return !isAssociative;
            }
            else
            {
                return precedence > op.precedence;
            }
        }
    }

    public static final DyadicOperator  OP_OR      = new DyadicOperator ("OR",     0, true);
    public static final DyadicOperator  OP_AND     = new DyadicOperator ("AND",    1, true);
    public static final MonadicOperator OP_NOT     = new MonadicOperator("NOT ",   2);
    public static final DyadicOperator  OP_EQ      = new DyadicOperator ("=",      3, false);
    public static final DyadicOperator  OP_NOTEQ   = new DyadicOperator ("<>",     3, false);
    public static final DyadicOperator  OP_LT      = new DyadicOperator ("<",      3, false);
    public static final DyadicOperator  OP_LTEQ    = new DyadicOperator ("<=",     3, false);
    public static final DyadicOperator  OP_GT      = new DyadicOperator (">",      3, false);
    public static final DyadicOperator  OP_GTEQ    = new DyadicOperator (">=",     3, false);
    public static final DyadicOperator  OP_LIKE    = new DyadicOperator ("LIKE",   3, false);
    public static final DyadicOperator  OP_BETWEEN = new DyadicOperator ("BETWEEN",3, false);
    public static final DyadicOperator  OP_IS      = new DyadicOperator ("IS",     3, false);
    public static final DyadicOperator  OP_ISNOT   = new DyadicOperator ("IS NOT", 3, false);
    public static final DyadicOperator  OP_IN      = new DyadicOperator ("IN"    , 3, false);
    public static final DyadicOperator  OP_ADD     = new DyadicOperator ("+",      4, true);
    public static final DyadicOperator  OP_SUB     = new DyadicOperator ("-",      4, false);
    public static final DyadicOperator  OP_CONCAT  = new DyadicOperator ("||",     4, true);
    public static final DyadicOperator  OP_MUL     = new DyadicOperator ("*",      5, true);
    public static final DyadicOperator  OP_DIV     = new DyadicOperator ("/",      5, false);
    public static final MonadicOperator OP_NEG     = new MonadicOperator("-",      6);

    protected final QueryStatement qs;
    protected TableExpression te;
    protected final StatementText st = new StatementText();
    protected Operator lowestOperator = null;
    
    private String alias;
    private String nonAliasExpression;
    
    /**
     * result expressions in Query needs to know the mapping 
     */
    protected JavaTypeMapping mapping; 

    protected ExpressionList expressionList = new ExpressionList();
    
    protected ScalarExpression(QueryStatement qs)
    {
        this.qs = qs;
    }

    protected ScalarExpression(QueryStatement qs, JavaTypeMapping mapping, TableExpression te)
    {
        this(qs);
        this.mapping = mapping;
        this.te = te;
        for (int i = 0; i < mapping.getNumberOfDatastoreFields(); i++)
        {
            expressionList.addExpression(new FieldExpression(qs,mapping.getDataStoreMapping(i).getDatastoreField(),te));
        }        
        st.append(expressionList.toString());
    }

    protected ScalarExpression(String functionName, List args)
    {
        st.append(functionName).append('(');

        ListIterator i = args.listIterator();
        ScalarExpression arg = (ScalarExpression)i.next();
        st.append(arg);
        qs = arg.qs;

        while (i.hasNext())
        {
            arg = (ScalarExpression)i.next();
            st.append(',').append(arg);
        }

        st.append(')');
    }

    protected ScalarExpression(MonadicOperator op, ScalarExpression operand)
    {
        st.append(op);

        if (op.isHigherThan(operand.lowestOperator))
        {
            st.append('(').append(operand).append(')');
        }
        else
        {
            st.append(operand);
        }

        qs = operand.qs;
        lowestOperator = op;
    }

    protected ScalarExpression(ScalarExpression operand1, DyadicOperator op, ScalarExpression operand2)
    {
        qs = operand1.qs;
        lowestOperator = op;

        if( op == ScalarExpression.OP_CONCAT )
        {
            //enclose within parentheses to make sure the concat expression is first evaluated
            st.append(((DatabaseAdapter) qs.getStoreManager().getDatastoreAdapter()).concatenateExpressions(operand1, operand2).encloseWithInParentheses());
        }
        else
        {
	        if (op.isHigherThanLeftSide(operand1.lowestOperator))
	        {
	            st.append('(').append(operand1).append(')');
	        }
	        else
	        {
	            st.append(operand1);
	        }
	
	        st.append(op);
	
	        if (op.isHigherThanRightSide(operand2.lowestOperator))
	        {
	            st.append('(').append(operand2).append(')');
	        }
	        else
	        {
	            st.append(operand2);
	        }
	        /*
	         * some databases use a ESCAPE expression with LIKE when using an escaped pattern 
	         */       
	        if( op == ScalarExpression.OP_LIKE && ((DatabaseAdapter) operand1.qs.getStoreManager().getDatastoreAdapter()).supportsEscapeExpressionInLikePredicate())
	        {
	            st.append(' ');
	            st.append(((DatabaseAdapter) operand1.qs.getStoreManager().getDatastoreAdapter()).getEscapePatternExpression());
	            st.append(' ');
	        }
        }
    }

    public QueryStatement getQueryStatement()
    {
        return qs;
    }

    /**
     * @return Returns the te.
     */
    public TableExpression getTe()
    {
        return te;
    }
    
    public BooleanExpression and(ScalarExpression expr)
    {
        throw new IllegalOperationException(this, "&&", expr);
    }

    public BooleanExpression eor(ScalarExpression expr)
    {
        throw new IllegalOperationException(this, "^", expr);
    }

    public BooleanExpression ior(ScalarExpression expr)
    {
        throw new IllegalOperationException(this, "||", expr);
    }

    public BooleanExpression not()
    {
        throw new IllegalOperationException("!", this);
    }

    public BooleanExpression eq(ScalarExpression expr)
    {
        throw new IllegalOperationException(this, "==", expr);
    }

    public BooleanExpression noteq(ScalarExpression expr)
    {
        throw new IllegalOperationException(this, "!=", expr);
    }

    public BooleanExpression lt(ScalarExpression expr)
    {
        throw new IllegalOperationException(this, "<", expr);
    }

    public BooleanExpression lteq(ScalarExpression expr)
    {
        throw new IllegalOperationException(this, "<=", expr);
    }

    public BooleanExpression gt(ScalarExpression expr)
    {
        throw new IllegalOperationException(this, ">", expr);
    }

    public BooleanExpression gteq(ScalarExpression expr)
    {
        throw new IllegalOperationException(this, ">=", expr);
    }

    public BooleanExpression in(ScalarExpression expr)
    {
        throw new IllegalOperationException(this, "in", expr);
    }

    public ScalarExpression add(ScalarExpression expr)
    {
        throw new IllegalOperationException(this, "+", expr);
    }

    public ScalarExpression sub(ScalarExpression expr)
    {
        throw new IllegalOperationException(this, "-", expr);
    }

    public ScalarExpression mul(ScalarExpression expr)
    {
        throw new IllegalOperationException(this, "*", expr);
    }

    public ScalarExpression div(ScalarExpression expr)
    {
        throw new IllegalOperationException(this, "/", expr);
    }

    public ScalarExpression mod(ScalarExpression expr)
    {
        throw new IllegalOperationException(this, "%", expr);
    }

    public ScalarExpression neg()
    {
        throw new IllegalOperationException("-", this);
    }

    public ScalarExpression com()
    {
        throw new IllegalOperationException("~", this);
    }

    public ScalarExpression cast(Class type)
    {
        throw new IllegalOperationException("cast to " + type.getName(), this);
    }

    public ScalarExpression accessField(String fieldName, boolean innerJoin)
    {
        throw new IllegalOperationException("access field " + fieldName, this);
    }

    /**
     * "as" Alias
     * @param alias the alias
     * @return this
     */
    public ScalarExpression as(String alias)
    {
        if( this.alias == null )
        {
            this.alias = alias;
            nonAliasExpression = st.toString();
            st.postpend(" AS " + alias);
        }
        return this;
    }
    
    /**
     * @return Returns the alias.
     */
    public String getAlias()
    {
        return alias;
    }
    
    /**
     * @return Returns the nonAliasExpression.
     */
    public String getNonAliasExpression()
    {
        return nonAliasExpression;
    }
    
    /*
     * This can be enlarged as needed, but for the present there are no methods
     * accepting more than two arguments.
     */

    private static final Class[][] PARAM_TYPE_SIGNATURES =
        {
            null,
            new Class[] { ScalarExpression.class },
            new Class[] { ScalarExpression.class, ScalarExpression.class }
        };

    public ScalarExpression callMethod(String methodName, List arguments)
    {
        ScalarExpression expr;

        try
        {
            int numArgs = arguments.size();

            if (numArgs >= PARAM_TYPE_SIGNATURES.length)
            {
                throw new NoSuchMethodException(methodName + typeList(arguments));
            }

            Method m = getClass().getMethod(methodName + "Method", PARAM_TYPE_SIGNATURES[numArgs]);

            expr = (ScalarExpression)m.invoke(this, arguments.toArray());
        }
        catch (InvocationTargetException e)
        {
            Throwable t = e.getTargetException();

            if (t instanceof Error)
            {
                throw (Error)t;
            }
            if (t instanceof JDOException)
            {
                throw (JDOException)t;
            }
            else
            {
                throw new MethodInvocationException(methodName, arguments, t);
            }
        }
        catch (Exception e)
        {
            throw new MethodInvocationException(methodName, arguments, e);
        }

        return expr;
    }

    /**
     * Whether to enclose this expression within parentheses
     */
    public ScalarExpression encloseWithInParentheses()
    {
        st.encloseWithInParentheses();
        return this;
    }
    
    public StatementText toStatementText()
    {
        return st;
    }

    public String toString()
    {
        String value;

        try
        {
            value = " \"" + toStatementText().toString() + '"';
        }
        catch (JDOException e)
        {
            value = "";
        }

        String className = getClass().getName();

        return className.substring(className.lastIndexOf('.') + 1) + value;
    }

    private static String typeList(List exprs)
    {
        StringBuffer s = new StringBuffer("(");
        Iterator i = exprs.iterator();

        while (i.hasNext())
        {
            s.append(i.next()).append(i.hasNext() ? ',' : ')');
        }

        return s.toString();
    }

    /**
     * @return Returns the mapping.
     */
    public JavaTypeMapping getMapping()
    {
        return mapping;
    }

    /**
     * Ordered list of expressions comma separated
     **/
    public class ExpressionList
    {
        private List expressions = new ArrayList();

        /**
         * Add a new Expression to the list
         * @param expression in RDBMS datastore expressions can be columns
         */
        public void addExpression(ScalarExpression expression)
        {
            expressions.add(expression);
        }

        /**
         * Gets the expression in the index
         * @param index the position for the expression in the list of expressions
         * @return the expression in the index
         */
        public ScalarExpression getExpression(int index)
        {
            return (ScalarExpression) expressions.get(index);
        }
    
        public int size()
        {
            return expressions.size();
        }
        
        /**
         * Returns an string with a serie of expression comma separated.
         * BNF notation: [ expr [, expr] ] 
         * @return the expression list comma separated 
         */
        public String toString()
        {
            StringBuffer expr = new StringBuffer();
            for( int i=0; i<expressions.size(); i++)
            {
                expr.append(getExpression(i).toString());
                if( i<(expressions.size()-1))
                {
                    expr.append(',');
                }
            }
            return expr.toString();
        }
        
        public ScalarExpression[] toArray()
        {
            return (ScalarExpression[]) expressions.toArray(new ScalarExpression[expressions.size()]);
        }
    }
    
    /**
     * Returns the expression list
     * @return The expressions list
     */
    public ExpressionList getExpressionList()
    {
        return expressionList;
    }
    
    public static class FieldExpression extends ScalarExpression
    {
        DatastoreField field;
        public FieldExpression(QueryStatement qs, DatastoreField field, TableExpression te)
        {
            super(qs);
            this.field = field;
            this.te = te;
            st.append(te.referenceColumn(field));
        }
        
        public BooleanExpression eq(ScalarExpression expr)
        {
            return new BooleanExpression(expr,OP_EQ,this);
        }

        public BooleanExpression noteq(ScalarExpression expr)
        {
            return new BooleanExpression(expr,OP_NOTEQ,this);
        }
        
        public int hashCode()
        {
            return te.hashCode() ^ field.hashCode();
        }

        public boolean equals(Object o)
        {
            if (o == this)
            {
                return true;
            }

            if (!(o instanceof QueryExpression))
            {
                return false;
            }

            QueryExpression qsc = (QueryExpression)o;

            return te.equals(qsc.te) && field.equals(qsc.datastoreField);
        }

        public String toString()
        {
            return te.referenceColumn(field);
        }        
    }    
    
    /**
     * Inner class representing an illegal argument error.
     **/
    public static class IllegalArgumentTypeException extends IllegalArgumentException
    {
        public IllegalArgumentTypeException(ScalarExpression arg)
        {
            super("Illegal argument type: " + arg);
        }
    }

    /**
     * Inner class representing a method invocation error.
     **/
    public static class MethodInvocationException extends JDOUserException
    {
        public MethodInvocationException(String methodName, List arguments, Throwable t)
        {
            super("Exception occurred invoking method " + methodName + typeList(arguments), t);
        }
    }

    /**
     * Inner class representing an illegal operation.
     **/
    public static class IllegalOperationException extends JDOUserException
    {
        public IllegalOperationException(String operation, ScalarExpression operand)
        {
            super("Cannot perform operation \"" + operation + "\" on " + operand);
        }

        public IllegalOperationException(ScalarExpression operand1, String operation, ScalarExpression operand2)
        {
            super("Cannot perform operation \"" + operation + "\" on " + operand1 + " and " + operand2);
        }
    }
}
