/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others. All rights reserved.
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - coding standards
2005 Andy Jefferson - added contrib from Guillaume Bodet for escape of single-quote
    ...
**********************************************************************/
package org.jpox.store.expression;

import java.math.BigInteger;

import org.jpox.store.DatastoreAdapter;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.query.QueryStatement;
import org.jpox.util.StringUtils;

/**
 * Representation of a string literal.
 *
 * @version $Revision: 1.5 $
 **/
public class StringLiteral extends StringExpression implements Literal
{
    private final String value;

    public StringLiteral(QueryStatement qs, JavaTypeMapping mapping, char value)
    {
        super(qs);
        this.mapping = mapping;
        this.value = String.valueOf(value);

        // Escape any single-quotes
        st.append('\'').append(StringUtils.replaceAll(this.value, "'", "''")).append('\'');
    }

    public StringLiteral(QueryStatement qs, JavaTypeMapping mapping, String value)
    {
        this(qs,mapping,value,false);
        this.mapping = mapping;
    }

    public StringLiteral(QueryStatement qs, JavaTypeMapping mapping, String value,boolean useParameter)
    {
        super(qs);

        this.mapping = mapping;
        this.value = value;
        if (useParameter)
        {
            DatastoreAdapter dba = qs.getStoreManager().getDatastoreAdapter();
	        st.appendParameter(dba.getMapping(String.class, qs.getStoreManager()), value);
        }
        else
        {
            // Escape any single-quotes
            st.append('\'').append(StringUtils.replaceAll(value, "'", "''")).append('\'');
        }
    }

    public Object getValue()
    {
        return value;
    }

    public BooleanExpression eq(ScalarExpression expr)
    {
        if (expr instanceof StringLiteral)
        {
            return new BooleanLiteral(qs, mapping, value.equals(((StringLiteral)expr).value));
        }
        else
        {
            return super.eq(expr);
        }
    }

    public BooleanExpression noteq(ScalarExpression expr)
    {
        if (expr instanceof StringLiteral)
        {
            return new BooleanLiteral(qs, mapping, !value.equals(((StringLiteral)expr).value));
        }
        else
        {
            return super.noteq(expr);
        }
    }

    public BooleanExpression lt(ScalarExpression expr)
    {
        if (expr instanceof StringLiteral)
        {
            return new BooleanLiteral(qs, mapping, value.compareTo(((StringLiteral)expr).value) < 0);
        }
        else
        {
            return super.lt(expr);
        }
    }

    public BooleanExpression lteq(ScalarExpression expr)
    {
        if (expr instanceof StringLiteral)
        {
            return new BooleanLiteral(qs, mapping, value.compareTo(((StringLiteral)expr).value) <= 0);
        }
        else
        {
            return super.lteq(expr);
        }
    }

    public BooleanExpression gt(ScalarExpression expr)
    {
        if (expr instanceof StringLiteral)
        {
            return new BooleanLiteral(qs, mapping, value.compareTo(((StringLiteral)expr).value) > 0);
        }
        else
        {
            return super.gt(expr);
        }
    }

    public BooleanExpression gteq(ScalarExpression expr)
    {
        if (expr instanceof StringLiteral)
        {
            return new BooleanLiteral(qs, mapping, value.compareTo(((StringLiteral)expr).value) >= 0);
        }
        else
        {
            return super.gteq(expr);
        }
    }

    public ScalarExpression add(ScalarExpression expr)
    {
        if (expr instanceof StringLiteral)
        {
            return new StringLiteral(qs, mapping, value.concat(((StringLiteral)expr).value));
        }
        else if (expr instanceof IntegerLiteral || expr instanceof FloatingPointLiteral || expr instanceof BooleanLiteral )
        {
            return new StringLiteral(qs, mapping, value.concat(((Literal)expr).getValue().toString()));
        }        
        else
        {
            return super.add(expr);
        }
    }

    public BooleanExpression endsWithMethod(ScalarExpression str)
    {
        if (str instanceof StringLiteral)
        {
            return new BooleanLiteral(qs, mapping, value.endsWith(((StringLiteral)str).value));
        }
        else
        {
            return super.endsWithMethod(str);
        }
    }

    public NumericExpression indexOfMethod(ScalarExpression str)
    {
        if (str instanceof StringLiteral)
        {
            return new IntegerLiteral(qs, mapping, BigInteger.valueOf(value.indexOf(((StringLiteral)str).value)));
        }
        else
        {
            return super.indexOfMethod(str);
        }
    }

    public NumericExpression lengthMethod()
    {
        return new IntegerLiteral(qs, mapping, BigInteger.valueOf(value.length()));
    }

    public BooleanExpression startsWithMethod(ScalarExpression str)
    {
        if (str instanceof StringLiteral)
        {
            return new BooleanLiteral(qs, mapping, value.startsWith(((StringLiteral)str).value));
        }
        else
        {
            return super.startsWithMethod(str);
        }
    }

    public BooleanExpression startsWithMethod(ScalarExpression str, ScalarExpression toffset)
    {
        if (str instanceof StringLiteral && toffset instanceof IntegerLiteral)
        {
            return new BooleanLiteral(qs, mapping, value.startsWith(((StringLiteral)str).value,
                            ((BigInteger) ((IntegerLiteral)toffset).getValue()).intValue()));
        }
        else
        {
            return super.startsWithMethod(str, toffset);
        }
    }

    public StringExpression substringMethod(NumericExpression begin)
    {
        if (begin instanceof IntegerLiteral)
        {
            return new StringLiteral(qs, mapping, value.substring(((BigInteger)((IntegerLiteral)begin).getValue()).intValue()));
        }
        else
        {
            return super.substringMethod(begin);
        }
    }

    public StringExpression substringMethod(NumericExpression begin, NumericExpression end)
    {
        if (begin instanceof IntegerLiteral && end instanceof IntegerLiteral)
        {
            return new StringLiteral(qs, mapping, value.substring(((BigInteger)((IntegerLiteral)begin).getValue()).intValue(),
                									     ((BigInteger)((IntegerLiteral)end).getValue()).intValue()));
        }
        else
        {
            return super.substringMethod(begin, end);
        }
    }

    public StringExpression toLowerCase()
    {
        return new StringLiteral(qs, mapping, value.toLowerCase());
    }

    public StringExpression toUpperCase()
    {
        return new StringLiteral(qs, mapping, value.toUpperCase());
    }

    public StringExpression trim()
    {
        return new StringLiteral(qs, mapping, value.trim());
    }
}