/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
    ...
**********************************************************************/
package org.jpox.store.extent;

import java.util.HashMap;
import java.util.Iterator;

import javax.jdo.Extent;
import javax.jdo.FetchPlan;
import javax.jdo.JDOFatalUserException;

import org.jpox.FetchPlanImpl;
import org.jpox.PersistenceManager;
import org.jpox.metadata.ClassMetaData;
import org.jpox.store.StoreManager;
import org.jpox.store.query.Query;
import org.jpox.store.query.QueryResult;
import org.jpox.util.Localiser;

/**
 * Abstract representation of an Extent for stores. Individual stores
 * should extend this to implement Extent's for their purpose.
 * 
 * @version $Revision: 1.1 $
 */
public abstract class AbstractExtent implements Extent
{
    protected static final Localiser LOCALISER = Localiser.getInstance("org.jpox.store.extent.Localisation");

    protected final PersistenceManager pm;

    protected final String className;

    protected final boolean subclasses;

    protected final ClassMetaData cmd;
    
    protected final StoreManager storeMgr;

    protected final Query query;

    protected HashMap queryResultsByIterator = new HashMap();

    protected final FetchPlanImpl fetchPlan;
    
    /**
     * Constructor.
     * @param pm Persistence Manager
     * @param className Name of the candidate class
     * @param subclasses Whether to include subclasses
     */
    public AbstractExtent(PersistenceManager pm, String className, boolean subclasses)
    {
        this.pm = pm;
        this.className = className;
        this.subclasses = subclasses;
        storeMgr = pm.getStoreManager();

        if (className == null)
        {
            throw new JDOFatalUserException(LOCALISER.msg("Extent.ClassNameNotSpecified"));
        }
        
        // Find the MetaData for this class
        cmd = storeMgr.getMetaDataManager().getMetaDataForClass(getCandidateClass());
        if (cmd == null)
        {
            throw new JDOFatalUserException(LOCALISER.msg("Extent.ClassHasNoMetaData", className));
        }

        this.fetchPlan = (FetchPlanImpl) ((FetchPlanImpl)pm.getFetchPlan()).getCopy();

        query = (Query) pm.newQuery();
        
        Class candidateClass = getCandidateClass();
        query.setClass(candidateClass);
        query.setCandidates(this);
    }

    /**
     * Returns whether this Extent was defined to contain subclasses.
     * @return true if this Extent was defined to include subclasses.
     */
    public boolean hasSubclasses()
    {
        return subclasses;
    }

    /**
     * An Extent contains all instances of a particular Class in the data
     * store; this method returns the Class of the instances
     * @return the Class of instances of this Extent
     */
    public Class getCandidateClass()
    {
        return storeMgr.getClassLoaderResolver().classForName(className);
    }

    /**
     * An Extent is managed by a PersistenceManager; this method gives access
     * to the owning PersistenceManager.
     * @return the owning PersistenceManager
     */
    public javax.jdo.PersistenceManager getPersistenceManager()
    {
        return pm;
    }

    /**
     * Returns an iterator over all the instances in the Extent.
     * @return an iterator over all the instances in the Extent.
     */
    public Iterator iterator()
    {
        QueryResult qr = (QueryResult)query.execute();
        Iterator iter = qr.iterator();

        queryResultsByIterator.put(iter, qr);

        return iter;
    }
    
    /**
     * Close an Iterator associated with this Extent instance.  Iterators closed
     * by this method will return false to hasNext() and will throw
     * NoSuchElementException on next().  The Extent instance can still be used
     * as a parameter of Query.setCandidates, and to get an Iterator.
     *
     * @param iter an iterator obtained by the method iterator() on this Extent
     * instance.
     */
    public void close(Iterator iter)
    {
        QueryResult qr = (QueryResult) queryResultsByIterator.remove(iter);

        qr.close();
    }

    /**
     * Close all Iterators associated with this Extent instance.  Iterators
     * closed by this method will return false to hasNext() and will throw
     * NoSuchElementException on next().  The Extent instance can still be used
     * as a parameter of Query.setCandidates, and to get an Iterator.
     */
    public void closeAll()
    {
        Iterator iter = queryResultsByIterator.values().iterator();
        while (iter.hasNext())
        {
            QueryResult qr = (QueryResult) iter.next();

            qr.close();
            iter.remove();
        }
    }

    /**
     * This method retrieves the fetch plan associated with the Extent. It
     * always returns the identical instance for the same Extent instance. Any
     * change made to the fetch plan affects subsequent instance retrievals via
     * next(). Only instances not already in memory are affected by the fetch
     * plan. Fetch plan is described in Section 12.7.
     * @return the FetchPlan
     */
    public FetchPlan getFetchPlan()
    {
        return fetchPlan;
    }
    
    /**
     * Stringifier method.
     * @return Stringified form of this object
     */
    public String toString()
    {
        return LOCALISER.msg("Extent.ClassExtentInfo", className, "" + subclasses);
    }
}