/**********************************************************************
Copyright (c) 2002 David Jencks (TJDO) and others. All rights reserved.
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Erik Bengtson - removed unused import and variable
2003 Andy Jefferson - added localiser
    ...
**********************************************************************/

package org.jpox.store.mapping;

import java.util.Collection;
import java.util.List;

import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOUserException;

import org.jpox.StateManager;
import org.jpox.metadata.CollectionMetaData;
import org.jpox.metadata.FieldMetaData;
import org.jpox.sco.SCO;
import org.jpox.store.DatastoreAdapter;
import org.jpox.store.DatastoreClass;
import org.jpox.store.DatastoreObject;
import org.jpox.store.StoreManager;
import org.jpox.store.exceptions.NoDatastoreMappingException;
import org.jpox.store.expression.CollectionExpression;
import org.jpox.store.expression.CollectionLiteral;
import org.jpox.store.expression.CollectionSubqueryExpression;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.TableExpression;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.query.Queryable;
import org.jpox.store.scostore.ListStore;

/**
 * Abstract representation of a list.
 *
 * @version $Revision: 1.4 $
 **/
abstract class AbstractListMapping extends JavaTypeMapping implements MappingCallbacks
{
    protected final String fieldName;
    protected final StoreManager storeMgr;
    protected final DatastoreClass ownerTable;

    private DatastoreObject listTable;
    protected ListStore listStore;

    /**
     * Constructor taking the database adapter representing the datastore being
     * used and the class of the mapping.
     *
     * @param dba Database adapter
     * @param type class mapped to the DB
     **/
    public AbstractListMapping(DatastoreAdapter dba, Class type)
    {
        super(null, dba, type);
        this.fieldName           = null;
        this.storeMgr            = null;
        this.ownerTable          = null;
    }
   
    public AbstractListMapping(DatastoreAdapter dba, FieldMetaData fmd, DatastoreClass table)
    {
        super(fmd, dba, fmd.getType());

        this.fieldName = fmd.getName();
        this.storeMgr = table.getStoreManager();
        this.ownerTable = table;

        CollectionMetaData colmd = fmd.getCollection();
        if (colmd == null)
        {
            throw new JDOUserException(LOCALISER.msg("AbstractListMapping.CollectionMetaDataError",fmd));
        }

        listTable = storeMgr.newJoinDatastoreContainerObject(fmd);
        listStore = null;
    }

    public synchronized ListStore getListStore()
    {
        if (listStore == null)
        {
            listStore = (ListStore) storeMgr.newStore(listTable,fmd);
        }

        return listStore;
    }

    public int getNumberOfDatastoreFields()
    {
		return ownerTable.getIDMapping().getNumberOfDatastoreFields();
    }
    
    /**
     * Accessor for a datastore mapping
     * @param index The id of the mapping
     * @return The datastore mapping
     */
    public DatastoreMapping getDataStoreMapping(int index)
    {
        throw new NoDatastoreMappingException(fieldName);
    }
    
    /**
     * Accessor for the datastore mappings for this java type
     * @return The datastore mapping(s)
     */
    public List getDataStoreMappings()
    {
        throw new NoDatastoreMappingException(fieldName);
    }
   
    public boolean includeInFetchStatement()
    {
        return false;
    }
    
    /**
     * This mapping is included in the update statement.
     * @return Whether to include in update statement
     */
    public boolean includeInUpdateStatement()
    {
        return false;
    }

    /**
     * This mapping is included in the insert statement.
     * @return Whether to include in insert statement
     */
    public boolean includeInInsertStatement()
    {
        return false;
    }    

    protected abstract List newList(StateManager sm, String fieldName);

    public void deleteDependent(StateManager sm)
    {
        //makes sure field is loaded
        sm.isLoaded(sm.getObject(),fmd.getAbsoluteFieldNumber());
        Collection value = (Collection)sm.provideField(fmd.getAbsoluteFieldNumber());

        if( value != null && !value.isEmpty() )
        {
	        // if to delete elements
	        if ( fmd.getCollection().isDependentElement())
	        {
		        Object[] values = ((List)value).toArray();
		        //clear the collection
		        preDelete(sm);
		        for( int i=0; i<values.length; i++ )
		        {
		            sm.getPersistenceManager().deletePersistent(values[i]);
		        }
	        }
	        else
	        {
		        //clear the collection
		        preDelete(sm);
	        }
        }
    }
    
    public void postInsert(StateManager sm)
    {
        Collection value = (Collection)sm.provideField(fmd.getAbsoluteFieldNumber());

        //if (value == null)
        //    throw new JDOUserException("List fields cannot be null");

        getListStore().addAll(sm, value);

        sm.replaceField(fmd.getAbsoluteFieldNumber(), newList(sm, fieldName));
    }

    public void postFetch(StateManager sm)
    {
        sm.replaceField(fmd.getAbsoluteFieldNumber(), newList(sm, fieldName));
    }

    public void postUpdate(StateManager sm)
    {
        Collection value = (Collection)sm.provideField(fmd.getAbsoluteFieldNumber());

        if (value == null)
        {
            throw new JDOUserException(LOCALISER.msg("AbstractListMapping.NullValueError"));
        }

        if (value instanceof SCO)
        {
            SCO sco = (SCO)value;

            if (sm.getObject() == sco.getOwner() &&
                fieldName.equals(sco.getFieldName()))
            {
                sco.applyUpdates();
                return;
            }

            if (sco.getOwner() != null)
            {
                throw new JDOFatalInternalException(LOCALISER.msg("AbstractListMapping.WrongOwnerError"));
            }
        }

        getListStore().clear(sm);
        getListStore().addAll(sm, value);

        sm.replaceField(fmd.getAbsoluteFieldNumber(), newList(sm, fieldName));
    }

    /**
     * Method to be called before any delete of the owner class element,
     * if the field in the owner is dependent
     *
     * @param sm StateManager of the owner
     **/
    public void preDelete(StateManager sm)
    {
        if (fmd.isDependent())
        {
            getListStore().clear(sm);
        }
    }

    public boolean equals(Object obj)
    {
        if (obj == this)
        {
            return true;
        }

        if (!obj.getClass().equals(getClass()))
        {
            return false;
        }

        AbstractListMapping sm = (AbstractListMapping)obj;

        return fmd.equals(sm.fmd) && storeMgr.equals(sm.storeMgr);
    }

    public int hashCode()
    {
        return fmd.hashCode() ^ storeMgr.hashCode();
    }

    public Object getSampleValue()
    {
        return null;
    }

    public ScalarExpression newLiteral(QueryStatement qs, Object value)
    {
        if (value instanceof Queryable)
        {
            return new CollectionSubqueryExpression(qs, ((Queryable)value).newQueryStatement());
        }
        else
        {
            return new CollectionLiteral(qs, this, (Collection)value);
        }
    }    

    public ScalarExpression newScalarExpression(QueryStatement qs, TableExpression te)
    {
		return new CollectionExpression(qs,
                        ownerTable.getIDMapping(),
                        te,
                        getListStore(),
                        fieldName);
    }
}
