/******************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved. 
This program and the accompanying materials are made available under
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
    ...
*****************************************************************/
package org.jpox.store.mapping;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedHashMap;
import java.util.LinkedHashSet;
import java.util.LinkedList;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.Stack;
import java.util.Vector;

import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOUnsupportedOptionException;
import javax.jdo.spi.PersistenceCapable;

import org.jpox.TypeManager;
import org.jpox.metadata.FieldMetaData;
import org.jpox.store.DatastoreAdapter;
import org.jpox.store.DatastoreClass;
import org.jpox.store.OID;
import org.jpox.util.JPOXLogger;
import org.jpox.util.JavaUtils;
import org.jpox.util.Localiser;
import org.jpox.util.MultiMap;

/**
 * Base implementation of a MappingManager. Datastores should extend this to add their own specifics, and to
 * add the initialisation of their supported types.
 * <P>
 * The idea behind a MappingManager is that at the Java side we have a series of Java type mappings,
 * and at the datastore side we have a series of datastore type mappings. We need a link between the two
 * to say that "this Java type can map to any of these 3 datastore types, and by default use this one".
 * 
 * @version $Revision: 1.8 $
 */
public abstract class AbstractMappingManager implements MappingManager
{
	protected static final Localiser LOCALISER=Localiser.getInstance("org.jpox.store.mapping.Localisation");

    Map javaTypeMappings = new MultiMap();

    /**
     * Constructor.
     */
    public AbstractMappingManager()
    {
        initialise();
    }
    
    /**
     * Method to load up the types for this manager.
     */
    public void initialise()
    {
        // Primitive types
        registerJavaTypeMapping(boolean.class.getName(), BooleanMapping.class, true);
        registerJavaTypeMapping(byte.class.getName(), ByteMapping.class, true);
        registerJavaTypeMapping(char.class.getName(), CharacterMapping.class, true);
        registerJavaTypeMapping(double.class.getName(), DoubleMapping.class, true);
        registerJavaTypeMapping(float.class.getName(), FloatMapping.class, true);
        registerJavaTypeMapping(int.class.getName(), IntegerMapping.class, true);
        registerJavaTypeMapping(long.class.getName(), LongMapping.class, true);
        registerJavaTypeMapping(short.class.getName(), ShortMapping.class, true);

        // Primitive wrapper types
        registerJavaTypeMapping(Boolean.class.getName(), BooleanMapping.class, true);
        registerJavaTypeMapping(Byte.class.getName(), ByteMapping.class, true);
        registerJavaTypeMapping(Character.class.getName(), CharacterMapping.class, true);
        registerJavaTypeMapping(Double.class.getName(), DoubleMapping.class, true);
        registerJavaTypeMapping(Float.class.getName(), FloatMapping.class, true);
        registerJavaTypeMapping(Integer.class.getName(), IntegerMapping.class, true);
        registerJavaTypeMapping(Long.class.getName(), LongMapping.class, true);
        registerJavaTypeMapping(Short.class.getName(), ShortMapping.class, true);
        
        // java.math types
        registerJavaTypeMapping(BigDecimal.class.getName(), BigDecimalMapping.class, true);
        registerJavaTypeMapping(BigInteger.class.getName(), BigIntegerMapping.class, true);

        // java.util types
        registerJavaTypeMapping(String.class.getName(), StringMapping.class, true);
        registerJavaTypeMapping(java.util.Date.class.getName(), DateMapping.class, true);
        registerJavaTypeMapping(java.util.Locale.class.getName(), LocaleMapping.class, true);
        registerJavaTypeMapping(java.util.TimeZone.class.getName(), TimeZoneMapping.class, true);

        // java.sql types
        registerJavaTypeMapping(java.sql.Date.class.getName(), SqlDateMapping.class, true);
        registerJavaTypeMapping(java.sql.Time.class.getName(), SqlTimeMapping.class, true);
        registerJavaTypeMapping(java.sql.Timestamp.class.getName(), SqlTimestampMapping.class, true);

        // other JDK types
        registerJavaTypeMapping(Object.class.getName(), ObjectMapping.class, true);

        // Non JDK types
        registerJavaTypeMapping(OID.class.getName(), OIDMapping.class, true);

        // Array types
        registerJavaTypeMapping(boolean[].class.getName(), BooleanArrayMapping.class, true);
        registerJavaTypeMapping(byte[].class.getName(), ByteArrayMapping.class, true);
        registerJavaTypeMapping(char[].class.getName(), CharArrayMapping.class, true);
        registerJavaTypeMapping(double[].class.getName(), DoubleArrayMapping.class, true);
        registerJavaTypeMapping(float[].class.getName(), FloatArrayMapping.class, true);
        registerJavaTypeMapping(int[].class.getName(), IntArrayMapping.class, true);
        registerJavaTypeMapping(long[].class.getName(), LongArrayMapping.class, true);
        registerJavaTypeMapping(short[].class.getName(), ShortArrayMapping.class, true);

        // Container types
        registerJavaTypeMapping(ArrayList.class.getName(), ArrayListMapping.class, true);
        registerJavaTypeMapping(Collection.class.getName(), CollectionMapping.class, true);
        registerJavaTypeMapping(HashMap.class.getName(), HashMapMapping.class, true);
        registerJavaTypeMapping(HashSet.class.getName(), HashSetMapping.class, true);
        registerJavaTypeMapping(Hashtable.class.getName(), HashtableMapping.class, true);
        registerJavaTypeMapping(LinkedList.class.getName(), LinkedListMapping.class, true);
        registerJavaTypeMapping(List.class.getName(), ListMapping.class, true);
        registerJavaTypeMapping(Map.class.getName(), MapMapping.class, true);
        registerJavaTypeMapping(Set.class.getName(), SetMapping.class, true);
        registerJavaTypeMapping(Stack.class.getName(), StackMapping.class, true);
        registerJavaTypeMapping(Vector.class.getName(), VectorMapping.class, true);

        // JPOX proxy container types
        registerJavaTypeMapping(org.jpox.sco.ArrayList.class.getName(), ArrayListMapping.class, true);
        registerJavaTypeMapping(org.jpox.sco.Collection.class.getName(), CollectionMapping.class, true);
        registerJavaTypeMapping(org.jpox.sco.HashMap.class.getName(), HashMapMapping.class, true);
        registerJavaTypeMapping(org.jpox.sco.HashSet.class.getName(), HashSetMapping.class, true);
        registerJavaTypeMapping(org.jpox.sco.Hashtable.class.getName(), HashtableMapping.class, true);
        registerJavaTypeMapping(org.jpox.sco.LinkedList.class.getName(), LinkedListMapping.class, true);
        registerJavaTypeMapping(org.jpox.sco.List.class.getName(), ListMapping.class, true);
        registerJavaTypeMapping(org.jpox.sco.Map.class.getName(), MapMapping.class, true);
        registerJavaTypeMapping(org.jpox.sco.Set.class.getName(), SetMapping.class, true);
        registerJavaTypeMapping(org.jpox.sco.Stack.class.getName(), StackMapping.class, true);
        registerJavaTypeMapping(org.jpox.sco.TreeSet.class.getName(), TreeSetMapping.class, true);
        registerJavaTypeMapping(org.jpox.sco.Vector.class.getName(), VectorMapping.class, true);

        // JDK 1.4 additions
        if (JavaUtils.isJRE1_4OrAbove())
        {
            registerJavaTypeMapping(java.util.Currency.class.getName(), CurrencyMapping.class, true);
            registerJavaTypeMapping(LinkedHashSet.class.getName(), LinkedHashSetMapping.class, true);
            registerJavaTypeMapping(LinkedHashMap.class.getName(), LinkedHashMapMapping.class, true);
            registerJavaTypeMapping(org.jpox.sco.LinkedHashSet.class.getName(), LinkedHashSetMapping.class, true);
            registerJavaTypeMapping(org.jpox.sco.LinkedHashMap.class.getName(), LinkedHashMapMapping.class, true);
        }

        // JDK 1.5 additions
    }

    /**
     * Method to register a java type mapping
     * @param javaTypeName Name of the java type
     * @param javaMappingType Mapping type to use for thie java type
     * @param isDefault Whether it is the default mapping for this java type
     */
    public void registerJavaTypeMapping(String javaTypeName, Class javaMappingType, boolean isDefault)
    {
        TypeMapping mapping = new TypeMapping(javaMappingType, isDefault);
        javaTypeMappings.put(javaTypeName, mapping);
    }

    /**
     * Accessor for the mapping class for the specified class.
     * @param c The class to retrieve the mapping class for.
     * @return The mapping class for this class
     */
    public Class getMappingClass(Class c)
    {
        return getMappingClass(c, null);
    }

    /**
     * Accessor for the mapping class for the specified class.
     * Provides special handling for interface types and for classes that are
     * being embedded in a field (detected using the FieldMetaData argument).
     * Refers others to its mapping manager lookup.
     * @param c Class to query
     * @param fmd The FieldMetaData for the field
     * @return The mapping class for the class
     **/
    public Class getMappingClass(Class c, FieldMetaData fmd)
    {
        Class mappingClass = null;

        // Is the field embedded ?
        if (fmd != null && fmd.getEmbeddedMetaData() != null)
        {
            if (PersistenceCapable.class.isAssignableFrom(c) && c != java.lang.Object.class)
            {
                // Embedded "simple" PC object
                mappingClass = EmbeddedPCMapping.class;
            }
            else if (Collection.class.isAssignableFrom(c) || Map.class.isAssignableFrom(c))
            {
                JPOXLogger.JDO.warn(LOCALISER.msg("MappingManager.EmbeddedContainersNotYetSupported", fmd.getFullFieldName()));
            }
            else
            {
                JPOXLogger.JDO.warn(LOCALISER.msg("MappingManager.NoMappingForEmbeddedJavaType", fmd.getFullFieldName(), c.getName()));
            }
        }

        // Is the class an interface
        if (mappingClass == null && TypeManager.getTypeManager().isSimpleInterface(c))
        {
        	mappingClass = InterfaceMapping.class;
        }

        // Try the default mapping
        if (mappingClass == null)
        {
        	mappingClass = getDefaultJavaTypeMapping(c.getName());
        }

        // If it is a 'supported' type yet no FCO mapping then error
        if (mappingClass == null &&
            TypeManager.getTypeManager().isSupportedType(c.getName()))
        {
            throw new JDOUnsupportedOptionException(LOCALISER.msg("MappingManager.NoMappingForJavaTypeField", fmd.getFullFieldName(), c.getName()));
        }

        if (mappingClass == null)
        {
        	// Check for PersistenceCapable
            if (PersistenceCapable.class.isAssignableFrom(c) && c != java.lang.Object.class)
            {
                mappingClass = PersistenceCapableMapping.class;
            }
            // Else it is an Object mapped type (serialised)
            else
            {
                mappingClass = ObjectMapping.class;
            }
        }

        return mappingClass;
    }

    /**
     * Accessor for the mapping for the field of the specified table.
     * @param table Table to query
     * @param fmd FieldMetaData for the field to map
     * @param dba Datastore adapter
     * @return The mapping for the field.
     */
    public JavaTypeMapping getMapping(DatastoreClass table, FieldMetaData fmd, DatastoreAdapter dba)
    {
        Class mc = getMappingClass(fmd.getType(), fmd);

        JavaTypeMapping m = MappingFactory.createMapping(mc, dba, fmd, table);
        if (m == null)
        {
            throw new JDOFatalInternalException(LOCALISER.msg("Mapping.ConstructorMissing", mc.getName()));
        }

        return m;
    }

    /**
     * Method to return the default java type mapping class for a specified java type name.
     * @param javaTypeName Name of the java type
     * @return The mapping class to use (by default)
     */
    public Class getDefaultJavaTypeMapping(String javaTypeName)
    {
        Collection coll = (Collection) javaTypeMappings.get(javaTypeName);
        if (coll == null)
        {
        	JPOXLogger.JDO.debug(LOCALISER.msg("MappingManager.NoMappingForJavaType",javaTypeName));
        	return null;
        }

        Iterator iter = coll.iterator();
        while (iter.hasNext())
        {
            TypeMapping map = (TypeMapping)iter.next();
            if (map.isDefault())
            {
                return map.getMappingType();
            }
        }
        return null;
    }

    protected class TypeMapping
    {
        Class mappingType;
        
        boolean isDefault;
        
        public TypeMapping(Class mappingType, boolean isDefault)
        {
            this.mappingType = mappingType;
            this.isDefault = isDefault;
        }
        
        /**
         * @return Returns the isDefault.
         */
        public boolean isDefault()
        {
            return isDefault;
        }
        
        /**
         * Mutator for whether this is the default datastore mapping for the mapping
         * @param isDefault Whether it is the default.
         */
        public void setDefault(boolean isDefault)
        {
            this.isDefault = isDefault;
        }
        
        /**
         * @return Returns the mappingType.
         */
        public Class getMappingType()
        {
            return mappingType;
        }
    }
}