/**********************************************************************
Copyright (c) 2002 Kelly Grizzle (TJDO) and others. All rights reserved.
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2002 Mike Martin (TJDO)
2003 Andy Jefferson - Added javadoc comments
2003 Erik Bengtson - corrected javadoc comments
2003 Erik Bengtson - removed unused import and variable
2003 Andy Jefferson - Added localiser
    ...
**********************************************************************/
package org.jpox.store.mapping;

import java.util.Collection;
import java.util.List;

import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOUserException;

import org.jpox.StateManager;
import org.jpox.metadata.CollectionMetaData;
import org.jpox.metadata.FieldMetaData;
import org.jpox.sco.SCO;
import org.jpox.store.DatastoreAdapter;
import org.jpox.store.DatastoreClass;
import org.jpox.store.DatastoreObject;
import org.jpox.store.StoreManager;
import org.jpox.store.exceptions.NoDatastoreMappingException;
import org.jpox.store.expression.CollectionExpression;
import org.jpox.store.expression.CollectionLiteral;
import org.jpox.store.expression.CollectionSubqueryExpression;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.TableExpression;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.query.Queryable;
import org.jpox.store.scostore.SetStore;

/**
 * Representation of the mapping between an owner table and a field.
 *
 * @version $Revision: 1.4 $
 **/
abstract class AbstractSetMapping extends JavaTypeMapping implements MappingCallbacks
{
    protected final String fieldName;
    protected final StoreManager storeMgr;
    protected final DatastoreClass ownerTable;

    private DatastoreObject setTable;
    protected SetStore setStore;

    /**
     * Constructor taking the database adapter representing the datastore being
     * used and the class of the mapping.
     * @param dba Database adapter
     * @param type class mapped to the DB
     **/
    public AbstractSetMapping(DatastoreAdapter dba, Class type)
    {
        super(null, dba, type);
        this.fieldName           = null;
        this.storeMgr            = null;
        this.ownerTable          = null;
    }

    /**
     * Constructor, taking the table being mapped, and the field number to be
     * mapped.
     * @param dba Datastore adapter
     * @param fmd FieldMetaData for field to be mapped
     * @param table Table to be mapped
     **/
    public AbstractSetMapping(DatastoreAdapter dba, FieldMetaData fmd, DatastoreClass table)
    {
        super(fmd, dba, fmd.getType());

        this.fieldName = fmd.getName();
        this.storeMgr = table.getStoreManager();
        this.ownerTable = table;

        CollectionMetaData colmd = fmd.getCollection();
        if (colmd == null)
        {
            throw new JDOUserException(LOCALISER.msg("AbstractSetMapping.CollectionMetaDataError",fmd));
        }

        setTable = storeMgr.newJoinDatastoreContainerObject(fmd);
        setStore = null;
    }

    public int getNumberOfDatastoreFields()
    {
		return ownerTable.getIDMapping().getNumberOfDatastoreFields();
    }
    
    /**
     * Accessor for the Store representing this mapping.
     *
     * @return The store
     **/
    public synchronized SetStore getSetStore()
    {
        if (setStore == null)
        {
            setStore = (SetStore) storeMgr.newStore(setTable,fmd);
        }

        return setStore;
    }

    /**
     * Accessor for a datastore mapping
     * @param index The id of the mapping
     * @return The datastore mapping
     */
    public DatastoreMapping getDataStoreMapping(int index)
    {
        throw new NoDatastoreMappingException(fieldName);
    }
    
    /**
     * Accessor for the datastore mappings for this java type
     * @return The datastore mapping(s)
     */
    public List getDataStoreMappings()
    {
        throw new NoDatastoreMappingException(fieldName);
    }

    public boolean includeInFetchStatement()
    {
        return false;
    }

    /**
     * This mapping is included in the update statement.
     * @return Whether to include in update statement
     */
    public boolean includeInUpdateStatement()
    {
        return false;
    }

    /**
     * This mapping is included in the insert statement.
     * @return Whether to include in insert statement
     */
    public boolean includeInInsertStatement()
    {
        return false;
    }
    
    protected abstract Collection newSet(StateManager sm, String fieldName);

    /**
     * Method to be called after the insert of the owner class element.
     *
     * @param sm StateManager of the Set (?)
     **/
    public void postInsert(StateManager sm)
    {
        Collection value = (Collection)sm.provideField(fmd.getAbsoluteFieldNumber());

        if (value == null)
        {
        	// accept null collections...
			sm.replaceField(fmd.getAbsoluteFieldNumber(),  newSet(sm, fieldName));
        	return;
            //throw new JDOUserException("Set fields cannot be null");
        }

        getSetStore().addAll(sm, value);

        sm.replaceField(fmd.getAbsoluteFieldNumber(), newSet(sm, fieldName));
    }
   
    public void deleteDependent(StateManager sm)
    {
        //makes sure field is loaded
        sm.isLoaded(sm.getObject(),fmd.getAbsoluteFieldNumber());
        Collection value = (Collection)sm.provideField(fmd.getAbsoluteFieldNumber());

        if( value != null && !value.isEmpty() )
        {
	        // if to delete elements
	        if (fmd.getCollection().isDependentElement())
	        {
		        Object[] values = value.toArray();
		        //clear the collection
		        preDelete(sm);
		        for( int i=0; i<values.length; i++ )
		        {
		            sm.getPersistenceManager().deletePersistent(values[i]);
		        }
	        }
	        else
	        {
		        //clear the collection
		        preDelete(sm);
	        }
        }
    }

    /**
     * Method to be called after any fetch of the owner class element.
     *
     * @param sm StateManager of the Set (?)
     **/
    public void postFetch(StateManager sm)
    {
        sm.replaceField(fmd.getAbsoluteFieldNumber(), newSet(sm, fieldName));
    }

    /**
     * Method to be called after any update of the owner class element.
     *
     * @param sm StateManager of the Set (?)
     **/
    public void postUpdate(StateManager sm)
    {
        Collection value = (Collection)sm.provideField(fmd.getAbsoluteFieldNumber());

		if (value == null)
		{
			// accept null collections...
			getSetStore().clear(sm);
			sm.replaceField(fmd.getAbsoluteFieldNumber(),  newSet(sm, fieldName));
			return;
			//throw new JDOUserException("Set fields cannot be null");
		}

        if (value instanceof SCO)
        {
            SCO sco = (SCO)value;

            if (sm.getObject() == sco.getOwner() && fieldName.equals(sco.getFieldName()))
            {
                sco.applyUpdates();
                return;
            }

            if (sco.getOwner() != null)
            {
                throw new JDOFatalInternalException(LOCALISER.msg("AbstractSetMapping.WrongOwnerError"));
            }
        }

        getSetStore().clear(sm);
        getSetStore().addAll(sm, value);

        sm.replaceField(fmd.getAbsoluteFieldNumber(), newSet(sm, fieldName));
    }

    /**
     * Method to be called before any delete of the owner class element,
     * if the field in the owner is dependent
     *
     * @param sm StateManager of the owner
     **/
    public void preDelete(StateManager sm)
    {
        if (fmd.isDependent())
        {
            getSetStore().clear(sm);
        }
    }

    public boolean equals(Object obj)
    {
        if (obj == this)
        {
            return true;
        }

        if (!obj.getClass().equals(getClass()))
        {
            return false;
        }

        AbstractSetMapping sm = (AbstractSetMapping)obj;

        return fmd.equals(sm.fmd) && storeMgr.equals(sm.storeMgr);
    }

    public int hashCode()
    {
        return fmd.hashCode() ^ storeMgr.hashCode();
    }

	public Object getSampleValue()
	{
		return null;
	}

    public ScalarExpression newLiteral(QueryStatement qs, Object value)
    {
        if (value instanceof Queryable)
        {
            return new CollectionSubqueryExpression(qs, ((Queryable)value).newQueryStatement());
        }
        else
        {
            return new CollectionLiteral(qs, this, (Collection)value);
        }
    }

    public ScalarExpression newScalarExpression(QueryStatement qs, TableExpression te)
    {
        return new CollectionExpression(qs,
                        ownerTable.getIDMapping(),
                        te,
                        getSetStore(),
                        fieldName);
    }
}
