/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution.

Contributors:
2005 Andy Jefferson - added capability to have String or long type discriminators
    ...
**********************************************************************/
package org.jpox.store.mapping;

import org.jpox.PersistenceManager;
import org.jpox.metadata.DiscriminatorMetaData;
import org.jpox.store.DatastoreAdapter;
import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.DatastoreField;
import org.jpox.store.Role;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.TableExpression;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.rdbms.sqlidentifier.SQLIdentifier;

/**
 * Mapping for a discriminator column in a table used in inheritance.
 * The discriminator column is, by default, a String type, typically VARCHAR.
 * It can however be "long" based if the user specifies INTEGER, BIGINT, or NUMERIC
 * as the jdbc-type. In the latter case we make the necessary conversions between
 * value types in this mapping class.
 * 
 * @version $Revision: 1.9 $
 */
public class DiscriminatorMapping extends SingleFieldMapping
{
    private final SingleFieldMapping delegate;

    /**
     * Constructor.
     * @param dba Datastore Adapter
     * @param table Datastore table
     * @param type Type of field
     */
    public DiscriminatorMapping(DatastoreAdapter dba, DatastoreContainerObject table, Class type)
    {
        super(dba, type);
        delegate = (SingleFieldMapping) dba.getMapping(type);

        DiscriminatorMetaData dismd = table.getDiscriminatorMetaData();
        DatastoreField column = table.addDatastoreField(type,
            new SQLIdentifier(getDatabaseAdapter(), dismd.getColumnMetaData().getName(), dismd.getColumnMetaData().getName()),
            Role.CUSTOM, this, dismd.getColumnMetaData());

        table.getStoreManager().getMappingManager().getDatastoreMapping(delegate, table.getStoreManager(), column, type);
    }

    /**
     * Accessor for the type represented here, returning the class itself
     * @return This class.
     */
    public Class getJavaType()
    {
        return DiscriminatorMapping.class;
    }

    /**
     * Accessor for a sample value for this type.
     * @return Sample value
     */
    public Object getSampleValue()
    {
        return delegate.getSampleValue();
    }

    /**
     * Accessor for a new literal for this mapping.
     * @param qs The QueryStatement
     * @param value The value of the object
     * @return The new literal
     */
    public ScalarExpression newLiteral(QueryStatement qs, Object value)
    {
        Object valueObj = value;
        if (value instanceof java.lang.String && type == long.class)
        {
            valueObj = new Integer((String)value);
        }
        return delegate.newLiteral(qs, valueObj);
    }

    /**
     * Accessor for a new scalar expression including this mapping.
     * @param qs The QueryStatement
     * @param te The table Expression
     * @return The new scalar expression
     */
    public ScalarExpression newScalarExpression(QueryStatement qs, TableExpression te)
    {
        return delegate.newScalarExpression(qs, te);
    }

    /**
     * Mutator for the object in this column
     * @param pm The Persistence Manager
     * @param preparedStatement The statement
     * @param exprIndex The indexes
     * @param value The value to set it to
     */
    public void setObject(PersistenceManager pm, Object preparedStatement, int[] exprIndex, Object value)
    {
        Object valueObj = value;
        if (value instanceof java.lang.String && type == long.class)
        {
            valueObj = new Integer((String)value);
        }
        delegate.setObject(pm, preparedStatement, exprIndex, valueObj);
    }

    /**
     * Accessor for the object in this column
     * @param pm The Persistence manager
     * @param resultSet The ResultSet to get the value from
     * @param exprIndex The indexes
     * @return The object
     */
    public Object getObject(PersistenceManager pm, Object resultSet, int[] exprIndex)
    {
        Object value = delegate.getObject(pm, resultSet, exprIndex);
        Object valueObj = value;
        if (value instanceof java.lang.String && type == long.class)
        {
            valueObj = new Integer((String)value);
        }
        return valueObj;
    }

    /**
     * Accessor for the number of datastore fields.
     * @return Number of datastore fields
     */
    public int getNumberOfDatastoreFields()
    {
        return delegate.getNumberOfDatastoreFields();
    }

    /**
     * Accessor for a datastore mapping
     * @param index Index of the mapping
     * @return The datastore mapping.
     */
    public DatastoreMapping getDataStoreMapping(int index)
    {
        return delegate.getDataStoreMapping(index);
    }

    /**
     * Mutator to add a datastore mapping
     * @param datastoreMapping Datastore mapping
     */
    public void addDataStoreMapping(DatastoreMapping datastoreMapping)
    {
        datastoreMappings.add(datastoreMapping);
    }
}