/**********************************************************************
Copyright (c) 2005 Andy Jefferson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
    ...
**********************************************************************/
package org.jpox.store.mapping;

import java.util.ArrayList;
import java.util.List;

import javax.jdo.JDOFatalInternalException;
import javax.jdo.spi.PersistenceCapable;

import org.jpox.PersistenceManager;
import org.jpox.metadata.ClassMetaData;
import org.jpox.metadata.FieldMetaData;
import org.jpox.state.StateManagerImpl;
import org.jpox.store.DatastoreAdapter;
import org.jpox.store.DatastoreClass;
import org.jpox.store.expression.ObjectExpression;
import org.jpox.store.expression.ObjectLiteral;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.TableExpression;
import org.jpox.store.query.QueryStatement;

/**
 * Mapping for a PersistenceCapable object embedded within another
 * PersistenceCapable object. Provides mapping for a single Java type
 * (the PC type) to multiple datastore columns.
 *
 * @version $Revision: 1.8 $
 **/
public class EmbeddedPCMapping extends SingleFieldMapping implements SimpleDatastoreRepresentation
{
    private List javaTypeMappings;

    /**
     * Constructor.
     * @param dba Datastore Adapter
     * @param type Type represented
     */
    public EmbeddedPCMapping(DatastoreAdapter dba, Class type)
    {
        super(dba, type);
    }

    /**
     * Constructor.
     * @param dba Datastore Adapter
     * @param fmd FieldMetaData for the field
     * @param table Table for persisting this field
     */
    public EmbeddedPCMapping(DatastoreAdapter dba, FieldMetaData fmd, DatastoreClass table)
    {
        super(dba, fmd, table);
    }

    /**
     * Add a new JavaTypeMapping to manage.
     * @param mapping the JavaTypeMapping
     */
    public void addJavaTypeMapping(JavaTypeMapping mapping)
    {
        if (javaTypeMappings == null)
        {
            javaTypeMappings = new ArrayList();
        }
        if (mapping == null)
        {
            throw new JDOFatalInternalException("mapping argument in PersistenceCapableMapping.addJavaTypeMapping is null");
        }
        javaTypeMappings.add(mapping);
    }

    /**
     * Mutator for the embedded object in the datastore.
     * @param pm The Persistence Manager managing this object
     * @param ps The Prepared Statement
     * @param param Param numbers in the PreparedStatement for the fields of this object
     * @param value The embedded object to use
     */
    public void setObject(PersistenceManager pm, Object ps, int[] param, Object value)
    {
        if (value == null)
        {
            int n = 0;
            String nullColumn = fmd.getEmbeddedMetaData().getNullIndicatorColumn();
            String nullValue = fmd.getEmbeddedMetaData().getNullIndicatorValue();
            for (int i=0; i<javaTypeMappings.size(); i++)
            {
                JavaTypeMapping mapping = ((JavaTypeMapping)javaTypeMappings.get(i));
                int[] posMapping = new int[mapping.getNumberOfDatastoreFields()];
                for (int j=0; j<posMapping.length; j++)
                {
                    posMapping[j] = param[n++];
                }

                // Null out this field unless it is the null-indicator column and has a value
                // in which case apply the required value
                if (nullColumn != null && nullValue != null &&
                    mapping.getFieldMetaData().getColumnMetaData().length > 0 &&
                    mapping.getFieldMetaData().getColumnMetaData()[0].getName().equals(nullColumn))
                {
                    // Try to cater for user having an integer based column and value
                    if (mapping instanceof IntegerMapping ||
                        mapping instanceof BigIntegerMapping ||
                        mapping instanceof LongMapping ||
                        mapping instanceof ShortMapping)
                    {
                        Object convertedValue = null;
                        try
                        {
                            if (mapping instanceof IntegerMapping ||
                                mapping instanceof ShortMapping)
                            {
                                convertedValue = new Integer(nullValue);
                            }
                            else if (mapping instanceof LongMapping ||
                                mapping instanceof BigIntegerMapping)
                            {
                                convertedValue = new Long(nullValue);
                            }
                        }
                        catch (Exception e)
                        {
                        }
                        mapping.setObject(pm, ps, posMapping, convertedValue);
                    }
                    else
                    {
                        mapping.setObject(pm, ps, posMapping, nullValue);
                    }
                }
                else
                {
                    mapping.setObject(pm, ps, posMapping, null);
                }
            }
        }
        else
        {
            if (!(value instanceof PersistenceCapable))
            {
                throw new JDOFatalInternalException(LOCALISER.msg("Mapping.ObjectIsNotPersistenceCapable",value.getClass(),value));
            }

            PersistenceCapable pc = (PersistenceCapable) value;
            StateManagerImpl embSM = (StateManagerImpl)pm.findStateManager(pc);
            if (embSM == null || pc.jdoGetPersistenceManager() == null)
            {
                // Assign a StateManager to manage our embedded object
                embSM = new StateManagerImpl(pc, pm);
            }

            ClassMetaData embCmd = pm.getMetaDataManager().getMetaDataForClass(pc.getClass().getName(),pm.getClassLoaderResolver());
            int n = 0;
            for (int i=0; i<javaTypeMappings.size(); i++)
            {
                JavaTypeMapping mapping = ((JavaTypeMapping)javaTypeMappings.get(i));
                int[] posMapping = new int[mapping.getNumberOfDatastoreFields()];
                for (int j=0; j<posMapping.length; j++)
                {
                    posMapping[j] = param[n++];
                }

                // Retrieve value of field from Embedded StateManager
                int embAbsFieldNum = embCmd.getFieldNumberAbsolute(mapping.getFieldMetaData().getName());
                Object fieldValue = embSM.provideField(embAbsFieldNum);

                mapping.setObject(pm, ps, posMapping, fieldValue);
            }
        }
    }

    /**
     * Accessor for the embedded object from the result set
     * @param pm Persistence Manager managing this object
     * @param rs The ResultSet
     * @param param Array of param numbers in the ResultSet for the fields of this object
     * @return The embedded object
     */
    public Object getObject(PersistenceManager pm, Object rs, int[] param)
    {
        Object value = null;

        // Create a PersistenceCapable to put the values into
        StateManagerImpl embSM = new StateManagerImpl(pm, fmd.getType(), (Object)null);
        value = embSM.getObject();

        String nullColumn = fmd.getEmbeddedMetaData().getNullIndicatorColumn();
        String nullValue = fmd.getEmbeddedMetaData().getNullIndicatorValue();
        ClassMetaData embCmd = pm.getMetaDataManager().getMetaDataForClass(fmd.getType());
        int n = 0;
        for (int i=0; i<javaTypeMappings.size(); i++)
        {
            JavaTypeMapping mapping = ((JavaTypeMapping)javaTypeMappings.get(i));
            int[] posMapping = new int[mapping.getNumberOfDatastoreFields()];
            for (int j=0; j<posMapping.length; j++)
            {
                posMapping[j] = param[n++];
            }

            // Retrieve value of field from Embedded StateManager
            int embAbsFieldNum = embCmd.getFieldNumberAbsolute(mapping.getFieldMetaData().getName());

            // Get the value from the ResultSet, and plug it into the PersistenceCapable if it is set
            Object fieldValue = mapping.getObject(pm, rs, posMapping);
            if (fieldValue != null)
            {
                embSM.replaceField(embAbsFieldNum, fieldValue);
            }

            // Check for null values
            if (nullColumn != null &&
                mapping.getFieldMetaData().getColumnMetaData()[0].getName().equals(nullColumn))
            {
                if ((nullValue == null && fieldValue == null) ||
                    (nullValue != null && fieldValue.toString().equals(nullValue)))
                {
                    value = null;
                    break;
                }
            }
        }

        return value;
    }

    public Object getSampleValue()
    {
        return "NO SAMPLE AVAILABLE";
    }

    /**
     * Accessor for the Java type being represented here.
     * @return The Java type
     */
    public Class getJavaType()
    {
        return fmd.getType();
    }

    public ScalarExpression newLiteral(QueryStatement qs, Object value)
    {
       ScalarExpression expr = new ObjectLiteral(qs, this, value, this.type);
       return expr;        
    }

    public ScalarExpression newScalarExpression(QueryStatement qs, TableExpression te)
    {
        ScalarExpression expr = new ObjectExpression(qs, this, te);
        return expr;        
    }
}