/**********************************************************************
Copyright (c) 01-Jan-2004 Barry Haddow and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution.

Contributors:
    Erik Bengtson - nearly all
    Marco Schulze - replaced catch(NotPersistenceCapableException ...) 
                    by advance-check via TypeManager.isSupportedType(...) 
***********************************************************************/
package org.jpox.store.mapping;

import java.util.ArrayList;
import java.util.List;

import org.jpox.PersistenceManager;
import org.jpox.metadata.FieldMetaData;
import org.jpox.store.DatastoreAdapter;
import org.jpox.store.DatastoreClass;
import org.jpox.store.OID;
import org.jpox.store.exceptions.NotYetFlushedException;
import org.jpox.store.exceptions.NullValueException;
import org.jpox.store.expression.InterfaceExpression;
import org.jpox.store.expression.ObjectLiteral;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.TableExpression;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.rdbms.table.ColumnCreator;
import org.jpox.store.rdbms.table.Table;

/**
 * Mapping for Interfaces. A column is created for each possible 
 * implementation of the interface. 
 *
 * @version $Revision: 1.3 $ 
 */
public class InterfaceMapping extends SingleFieldMapping implements SimpleDatastoreRepresentation
{
    private List javaTypeMappings;
    public InterfaceMapping(DatastoreAdapter dba, Class type)
    {
        super(dba, type);
    }
    
    public InterfaceMapping(DatastoreAdapter dba, FieldMetaData fmd, DatastoreClass table)
    {
        super(dba, fmd, table, false);
        createColumns(table, fmd);
    }

    protected void createColumns(DatastoreClass table, FieldMetaData fmd)
    {
        ColumnCreator.createColumnsForInterface(this, (Table)table, fmd);
    }    
   
	public void addJavaTypeMapping(JavaTypeMapping mapping)
	{
	    if(  javaTypeMappings == null )
	    {
	        javaTypeMappings = new ArrayList();
	    }
	    javaTypeMappings.add(mapping);	    
	}
	
	public List getJavaTypeMapping()
	{
	    return javaTypeMappings;	    
	}
	
    public int getNumberOfDatastoreFields()
    {
        int number = 0;
        for( int i=0; i<javaTypeMappings.size(); i++)
        {
            number += ((JavaTypeMapping)javaTypeMappings.get(i)).getNumberOfDatastoreFields();
        }
        return number;
    }
	/* (non-Javadoc)
	 * @see org.jpox.store.mapping.JavaTypeMapping#getDataStoreMapping(int)
	 */
	public DatastoreMapping getDataStoreMapping(int index)
	{
        List datastoreMappings = new ArrayList();
        for( int i=0; i<javaTypeMappings.size(); i++)
        {
            int number = ((JavaTypeMapping)javaTypeMappings.get(i)).getNumberOfDatastoreFields();
            for( int j=0; j<number; j++)
            {
                datastoreMappings.add(((JavaTypeMapping)javaTypeMappings.get(i)).getDataStoreMapping(j));
            }
        }
	    return (DatastoreMapping) datastoreMappings.get(index);
	}  
	
	/**
	 * @throws NotYetFlushedException
	 */
    public void setObject(PersistenceManager pm, Object ps, 
            int[] pos, Object value)
    {
        int n=0;
        boolean foundClassAssignableFromValue = false;
        NotYetFlushedException notYetFlushed = null;
        for( int i=0; i<javaTypeMappings.size(); i++)
        {
            JavaTypeMapping mapping = ((JavaTypeMapping)javaTypeMappings.get(i));
            int[] posMapping = new int[mapping.getNumberOfDatastoreFields()];
            for( int j=0; j<posMapping.length; j++)
            {
                posMapping[j] = pos[n++];
            }
            try
            {
                if( value != null && mapping.getDataStoreMapping(0).getDatastoreField().getType().isAssignableFrom(value.getClass()) )
                {
                    foundClassAssignableFromValue = true;
                    mapping.setObject(pm, ps, posMapping, value);
                }
                else
                {
                    mapping.setObject(pm, ps, posMapping, null);
                }
            }
            catch( NotYetFlushedException e )
            {
                notYetFlushed = e;
            }
        }
        if( notYetFlushed != null)
        {
            throw notYetFlushed;
        }
        if( value != null && !foundClassAssignableFromValue )
        {
            //as required by the JDO spec, a ClassCastException is thrown
            throw new ClassCastException("Can't assign class "+value.getClass().getName()+(fmd == null ? "" : " to field "+fmd.getFullFieldName()));
        }        
    }
    
    public Object getObject(PersistenceManager pm, final Object rs, int[] pos)
    {
        int n=0;
        for( int i=0; i<javaTypeMappings.size(); i++)
        {
            JavaTypeMapping mapping = ((JavaTypeMapping)javaTypeMappings.get(i));
            int[] posMapping = new int[mapping.getNumberOfDatastoreFields()];
            for( int j=0; j<posMapping.length; j++)
            {
                posMapping[j] = pos[n++];
            }
            Object value = null;
            try
            {
                value = mapping.getObject(pm, rs, posMapping);
            }
            catch( NullValueException e )
            {
                //expected if implementation object is null and has primitive fields in the primary key
            }
            if( value != null )
            {
                if( value instanceof OID )
                {
                    ((OID)value).setPcClass(mapping.getDataStoreMapping(0).getDatastoreField().getType().getName());
                    return pm.getObjectById(value,false);
                }
                return value;
            }
        }
        return null;
    }

    /* (non-Javadoc)
     * @see org.jpox.store.mapping.Mapping#getSampleValue()
     */
    public Object getSampleValue()
    {
        throw new UnsupportedOperationException("Not yet implemented");
    }

    /* (non-Javadoc)
     * @see org.jpox.store.mapping.Mapping#newLiteral(org.jpox.store.QueryStatement, java.lang.Object)
     */
    public ScalarExpression newLiteral(
        QueryStatement qs,
        Object value)
    {
        ScalarExpression expr = new ObjectLiteral(qs, this, value,value.getClass());
        return expr;
    }

    public ScalarExpression newScalarExpression(QueryStatement qs, TableExpression te)
    {
        ScalarExpression expr = new InterfaceExpression(qs,this,te);
        return expr; 
    }

    /* (non-Javadoc)
     * @see org.jpox.store.mapping.JavaTypeMapping#getJavaType()
     */
    public Class getJavaType()
    {
        return null;
    }

}
