/**********************************************************************
Copyright (c) 14-Sep-2004 Erik Bengtson and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
    ...
**********************************************************************/
package org.jpox.store.mapping;

import java.util.ArrayList;
import java.util.List;

import javax.jdo.JDOFatalInternalException;

import org.jpox.PersistenceManager;
import org.jpox.metadata.FieldMetaData;
import org.jpox.store.DatastoreAdapter;
import org.jpox.store.DatastoreClass;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.TableExpression;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.rdbms.adapter.DatabaseAdapter;
import org.jpox.util.Localiser;

/**
 * Representation of the mapping of a Java type.
 * The java type maps to one or more datastore mappings. This means that a field
 * in a java class can be mapped to many columns in a table (in an RDBMS).
 *
 * @version $Revision: 1.3 $
 **/
public abstract class JavaTypeMapping
{
    protected static final Localiser LOCALISER=Localiser.getInstance("org.jpox.store.mapping.Localisation");

    protected final FieldMetaData fmd;

    protected List datastoreMappings = new ArrayList();

    /**
     * Adapter for the datastore to map to.
     */
    protected final DatastoreAdapter dba;

    /**
     * Type being mapped.
     */
    protected final Class type;

    /**
     * Create a new Mapping with the given DatastoreAdapter for the given type.
     * @param fmd FieldMetaData for the field to be mapped
     * @param dba The Datastore Adapter that this Mapping should use.
     * @param type The Class that this mapping maps to the database.
     */
    protected JavaTypeMapping(FieldMetaData fmd, DatastoreAdapter dba, Class type)
    {
        this.dba = dba;
        this.type = type;
        this.fmd = fmd;
    }

    /**
     * Accessor for whether this mapping is nullable
     * @return Whether it is nullable
     */
    public boolean isNullable()
    {
        for (int i=0; i<datastoreMappings.size(); i++)
        {
            if (!((DatastoreMapping)datastoreMappings.get(i)).isNullable())
            {
                return false;
            }
        }
        return true;
    }
 
    /**
     * Accessor for the datastore mappings for this java type
     * @return The datastore mapping(s)
     */
    public List getDataStoreMappings()
    {
        return datastoreMappings;
    }
    
    /**
     * Accessor for the java type being mapped
     * @return The java type
     */
    public abstract Class getJavaType();
    
    /**
     * Accessor for the datastore class (e.g SQL Table) that this is mapped into
     * @return The datastore class
     */
    public abstract DatastoreClass getDatastoreClass();

    /**
     * Accessor for a datastore mapping
     * @param index The id of the mapping
     * @return The datastore mapping
     */
    public DatastoreMapping getDataStoreMapping(int index)
    {
        return (DatastoreMapping) datastoreMappings.get(index);
    }
    
    /**
     * Method to add a datastore mapping
     * @param datastoreMapping The datastore mapping
     */
    public void addDataStoreMapping(DatastoreMapping datastoreMapping)
    {
        datastoreMappings.add(datastoreMapping);
    }   
    
    /**
     * Acessor for the number of datastore fields (e.g. RDBMS number of columns)
     * @return the number of datastore fields
     */
    public int getNumberOfDatastoreFields()
    {
        return datastoreMappings.size();
    }
 
    /**
     * Convenience to access the Datastore adapter as a DatabaseAdapter.
     * @return The adapter in use
     */
    protected DatabaseAdapter getDatabaseAdapter()
    {
        return (DatabaseAdapter)dba;
    }

    /**
     * Accessor for the FieldMetaData of the field to be mapped
     * @return Returns the FieldMetaData.
     */
    public FieldMetaData getFieldMetaData()
    {
        return fmd;
    }
    
    /**
     * Return the Class that this Mapping maps to the database.
     * @return The Class that this Mapping maps to the database.
     */
    public Class getType()
    {
        return type;
    }
   
    /**
     * Return a sample value of the mapping type to be used for internal
     * evaluation of type and conversion.
     * @return The sample value. 
     */
    public abstract Object getSampleValue();

    /**
     * Whether this mapping is included in the fetch statement.
     * @return Whether to include in fetch statement
     */
    public abstract boolean includeInFetchStatement();

    /**
     * This mapping is included in the update statement.
     * @return Whether to include in update statement
     */
    public abstract boolean includeInUpdateStatement();

    /**
     * This mapping is included in the insert statement.
     * @return Whether to include in insert statement
     */
    public abstract boolean includeInInsertStatement();
    
    /**
     * Creates a literal from an value.
     * A string literal is enclosed in single quotes. for example: "literal".
     * A string literal that includes a single quote is represented by two 
     * single quotes. for example: "literal''s".
     * An exact numeric literal is a numeric value without a decimal point,
     * such as 57, -957, +62.
     * An approximate numeric literal is a numeric value in scientific notation,
     * such as 7E3, -57.9E2, or a numeric value with a decimal, such as 7.,
     * -95.7, +6.2.
     * @param qs The Query statement
     * @param value The object
     * @return A Scalar Expression
     */
    public abstract ScalarExpression newLiteral(QueryStatement qs,
                                                Object value);

    /**
     * Creates a expression from a field name/table.
     * e.g. tablename.fieldname; tablealias.fieldalias 
     * @param qs The Query statement
     * @param te the alias for the table
     * @return A Scalar Expression
     */ 
    public abstract ScalarExpression newScalarExpression(QueryStatement qs,
                                                         TableExpression te);
    /**
     * Utility to output any error message.
     * @param method The method that failed.
     **/
    protected String failureMessage(String method)
    {
        return LOCALISER.msg("Mapping.FailureMessage",getClass().getName(),method);
    }

    // ------------------- Accessors & Mutators for JDBC -----------------------

    /**
     * Convenience setter to provide a default value for this field.
     * If the field is nullable, a null is provided, otherwise the sample value.
     * @param pm The Persistence Manager
     * @param preparedStatement Prepared Statement
     * @param exprIndex The indices in the statement
     */
    public void setDefault(PersistenceManager pm, Object preparedStatement, int[] exprIndex)
    {
        // TODO Dont just use the first datastore mapping and parameter number
        getDataStoreMapping(0).setObject(preparedStatement, exprIndex[0], isNullable() ? null : getSampleValue());
    }

    public void setBoolean(PersistenceManager pm, Object preparedStatement, int[] exprIndex, boolean value)
    {
        throw new JDOFatalInternalException(failureMessage("setBoolean"));
    }

    public boolean getBoolean(PersistenceManager pm, Object resultSet, int[] exprIndex)
    {
        throw new JDOFatalInternalException(failureMessage("setBoolean"));
    }

    public void setChar(PersistenceManager pm, Object preparedStatement, int[] exprIndex, char value)
    {
        throw new JDOFatalInternalException(failureMessage("setChar"));
    }

    public char getChar(PersistenceManager pm, Object resultSet, int[] exprIndex)
    {
        throw new JDOFatalInternalException(failureMessage("getChar"));
    }

    public void setByte(PersistenceManager pm, Object preparedStatement, int[] exprIndex, byte value)
    {
        throw new JDOFatalInternalException(failureMessage("setByte"));
    }

    public byte getByte(PersistenceManager pm, Object resultSet, int[] exprIndex)
    {
        throw new JDOFatalInternalException(failureMessage("getByte"));
    }

    public void setShort(PersistenceManager pm, Object preparedStatement, int[] exprIndex, short value)
    {
        throw new JDOFatalInternalException(failureMessage("setShort"));
    }

    public short getShort(PersistenceManager pm, Object resultSet, int[] exprIndex)
    {
        throw new JDOFatalInternalException(failureMessage("getShort"));
    }

    public void setInt(PersistenceManager pm, Object preparedStatement, int[] exprIndex, int value)
    {
        throw new JDOFatalInternalException(failureMessage("setInt"));
    }

    public int getInt(PersistenceManager pm, Object resultSet, int[] exprIndex)
    {
        throw new JDOFatalInternalException(failureMessage("getInt"));
    }

    public void setLong(PersistenceManager pm, Object preparedStatement, int[] exprIndex, long value)
    {
        throw new JDOFatalInternalException(failureMessage("setLong"));
    }

    public long getLong(PersistenceManager pm, Object resultSet, int[] exprIndex)
    {
        throw new JDOFatalInternalException(failureMessage("getLong"));
    }

    public void setFloat(PersistenceManager pm, Object preparedStatement, int[] exprIndex, float value)
    {
        throw new JDOFatalInternalException(failureMessage("setFloat"));
    }

    public float getFloat(PersistenceManager pm, Object resultSet, int[] exprIndex)
    {
        throw new JDOFatalInternalException(failureMessage("getFloat"));
    }

    public void setDouble(PersistenceManager pm, Object preparedStatement, int[] exprIndex, double value)
    {
        throw new JDOFatalInternalException(failureMessage("setDouble"));
    }

    public double getDouble(PersistenceManager pm, Object resultSet, int[] exprIndex)
    {
        throw new JDOFatalInternalException(failureMessage("getDouble"));
    }

    public void setString(PersistenceManager pm, Object preparedStatement, int[] exprIndex, String value)
    {
        throw new JDOFatalInternalException(failureMessage("setString"));
    }

    public String getString(PersistenceManager pm, Object resultSet, int[] exprIndex)
    {
        throw new JDOFatalInternalException(failureMessage("getString"));
    }

    public void setObject(PersistenceManager pm, Object preparedStatement, int[] exprIndex, Object value)
    {
        throw new JDOFatalInternalException(failureMessage("setObject"));
    }

    public Object getObject(PersistenceManager pm, Object resultSet, int[] exprIndex)
    {
        throw new JDOFatalInternalException(failureMessage("getObject"));
    }
}