/**********************************************************************
Copyright (c) 2003 Mike Martin and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
    Andy Jefferson - coding standards
    Andy Jefferson - implementation of newScalarExpression, newLiteral
    ...
**********************************************************************/
package org.jpox.store.mapping;

import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOUserException;
import javax.jdo.spi.PersistenceCapable;

import org.jpox.StateManager;
import org.jpox.metadata.FieldMetaData;
import org.jpox.metadata.MapMetaData;
import org.jpox.sco.Map;
import org.jpox.sco.SCO;
import org.jpox.store.DatastoreAdapter;
import org.jpox.store.DatastoreClass;
import org.jpox.store.DatastoreObject;
import org.jpox.store.StoreManager;
import org.jpox.store.exceptions.NoDatastoreMappingException;
import org.jpox.store.expression.MapExpression;
import org.jpox.store.expression.MapLiteral;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.TableExpression;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.rdbms.scostore.InverseMapStore;
import org.jpox.store.scostore.MapStore;

/**
 * SCO Mapping for Map type.
 *
 * @version $Revision: 1.5 $
 **/
public class MapMapping extends JavaTypeMapping implements MappingCallbacks
{
    protected final String fieldName;
    protected final StoreManager storeMgr;
    protected final DatastoreClass ownerTable;

    private DatastoreObject mapTable;
    protected MapStore mapStore;

    /**
     * Constructor taking the database adapter representing the datastore being
     * used and the class of the mapping.
     *
     * @param dba Database adapter
     * @param type class mapped to the DB
     **/
    public MapMapping(DatastoreAdapter dba, Class type)
    {
        super(null, dba, type);
        this.fieldName           = null;
        this.storeMgr            = null;
        this.ownerTable          = null;
    }
    
    public MapMapping(DatastoreAdapter dba, FieldMetaData fmd, DatastoreClass table)
    {
        super(fmd, dba, fmd.getType());

        this.fieldName = fmd.getName();
        this.storeMgr = table.getStoreManager();
        this.ownerTable = table;

        MapMetaData mmd = fmd.getMap();
        if (mmd == null)
        {
            throw new JDOUserException("No map metadata found in " + fmd);
        }
        mapTable = storeMgr.newJoinDatastoreContainerObject(fmd);
        mapStore = null;
    }

    public synchronized MapStore getMapStore()
    {
        if (mapStore == null)
        {
            mapStore = (MapStore) storeMgr.newStore(mapTable,fmd);
        }

        return mapStore;
    }

    public int getNumberOfDatastoreFields()
    {
		return ownerTable.getIDMapping().getNumberOfDatastoreFields();
    }
    
    /**
     * Accessor for a datastore mapping
     * @param index The id of the mapping
     * @return The datastore mapping
     */
    public DatastoreMapping getDataStoreMapping(int index)
    {
        throw new NoDatastoreMappingException(fieldName);
    }
    
    /**
     * Accessor for the datastore mappings for this java type
     * @return The datastore mapping(s)
     */
    public List getDataStoreMappings()
    {
        throw new NoDatastoreMappingException(fieldName);
    }

    public boolean includeInFetchStatement()
    {
        return false;
    }

    /**
     * This mapping is included in the update statement.
     * @return Whether to include in update statement
     */
    public boolean includeInUpdateStatement()
    {
        return false;
    }

    /**
     * This mapping is included in the insert statement.
     * @return Whether to include in insert statement
     */
    public boolean includeInInsertStatement()
    {
        return false;
    }

    protected java.util.Map newMap(StateManager sm,String fieldName)
    {
        return new org.jpox.sco.Map(sm, fieldName);
    }

    public void deleteDependent(StateManager sm)
    {
        //makes sure field is loaded
        sm.isLoaded(sm.getObject(),fmd.getAbsoluteFieldNumber());
        java.util.Map value = (java.util.Map)sm.provideField(fmd.getAbsoluteFieldNumber());

        if( value != null && !value.isEmpty() )
        {
	        // if to delete key or values
	        if (fmd.getMap().isDependentValue() || fmd.getMap().isDependentKey())
	        {
		        java.util.Map m = new HashMap();
		        m.putAll(value);
		        //clear the collection
		        preDelete(sm);
		        for( Iterator it = m.entrySet().iterator(); it.hasNext(); )
		        {
		            Map.Entry entry = (Map.Entry) it.next();
		            Object val = entry.getValue();
		            Object key = entry.getKey();
		            if( fmd.getMap().isDependentKey() && key instanceof PersistenceCapable )
		            {
		                //clear reference to the key in value if inverse relationship
		                //if key field is non nullable, the value is deleted too
		                if( getMapStore() instanceof InverseMapStore )
		                {
		                    if( val instanceof PersistenceCapable )
		                    {
		                        ((InverseMapStore)getMapStore()).clearKeyOfValue(sm,key,val);
		                    }
		                }
		                sm.getPersistenceManager().deletePersistent(key);
		            }		            
		            if( fmd.getMap().isDependentValue() && val instanceof PersistenceCapable )
		            {
		                sm.getPersistenceManager().deletePersistent(val);
		            }
		        }        
	        }
	        else
	        {
		        //clear the collection
		        preDelete(sm);
	        }
        }
    }
    
    public void postInsert(StateManager sm)
    {
        java.util.Map value = (java.util.Map)sm.provideField(fmd.getAbsoluteFieldNumber());

        if (value == null)
        {
            //accept null mappings
            sm.replaceField(fmd.getAbsoluteFieldNumber(), newMap(sm, fieldName));
            return;
        }

        getMapStore().putAll(sm, value);

        sm.replaceField(fmd.getAbsoluteFieldNumber(), newMap(sm, fieldName));
    }

    public void postFetch(StateManager sm)
    {
        sm.replaceField(fmd.getAbsoluteFieldNumber(), newMap(sm, fieldName));
    }

    public void postUpdate(StateManager sm)
    {
        java.util.Map value = (java.util.Map)sm.provideField(fmd.getAbsoluteFieldNumber());

        if (value == null)
        {
            //accept null mappings
            sm.replaceField(fmd.getAbsoluteFieldNumber(), newMap(sm, fieldName));
            return;
        }

        if (value instanceof SCO)
        {
            SCO sco = (SCO)value;

            if (sm.getObject() == sco.getOwner() &&
                fieldName.equals(sco.getFieldName()))
            {
                sco.applyUpdates();
                return;
            }

            if (sco.getOwner() != null)
            {
                throw new JDOFatalInternalException("Owned second-class object was somehow assigned to a field other than its owner's");
            }
        }

        getMapStore().clear(sm);
        getMapStore().putAll(sm, value);

        sm.replaceField(fmd.getAbsoluteFieldNumber(), newMap(sm, fieldName));
    }

    /**
     * Method to be called before any delete of the owner class element,
     * if the field in the owner is dependent
     *
     * @param sm StateManager of the owner
     **/
    public void preDelete(StateManager sm)
    {
        if (fmd.isDependent())
        {
            getMapStore().clear(sm);
        }
    }

    public boolean equals(Object obj)
    {
        if (!super.equals(obj))
        {
            return false;
        }

        MapMapping mm = (MapMapping)obj;

        return fmd.equals(mm.fmd) && storeMgr.equals(mm.storeMgr);
    }

    public int hashCode()
    {
        return fmd.hashCode() ^ storeMgr.hashCode();
    }

    public Object getSampleValue()
    {
        return null;
    }

    public ScalarExpression newLiteral(QueryStatement qs, Object value)
    {
        return new MapLiteral(qs, this, (java.util.Map)value);
    }

    /**
     * Method to return a scalar expression for the Map.
     * @param qs The QueryStatement
     * @param te The TableExpression
     * @return The ScalarExpression 
     **/
    public ScalarExpression newScalarExpression(QueryStatement qs, TableExpression te)
    {
        return new MapExpression(qs,
                        ownerTable.getIDMapping(),
                        te,
                        getMapStore(),
                        fieldName);
    }

    /* (non-Javadoc)
     * @see org.jpox.store.mapping.JavaTypeMapping#getJavaType()
     */
    public Class getJavaType()
    {
        return Map.class;
    }

    /* (non-Javadoc)
     * @see org.jpox.store.mapping.JavaTypeMapping#getDatastoreClass()
     */
    public DatastoreClass getDatastoreClass()
    {
        return null;
    }
}
