/**********************************************************************
Copyright (c) 2003 Erik Bengtson and others. All rights reserved. 
This program and the accompanying materials are made available under
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - converted to use Reflection
2004 Andy Jefferson - output targetException when Invocation error
    ...
**********************************************************************/
package org.jpox.store.mapping;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;

import javax.jdo.JDOFatalInternalException;

import org.jpox.metadata.FieldMetaData;
import org.jpox.store.DatastoreAdapter;
import org.jpox.store.DatastoreClass;
import org.jpox.store.exceptions.UnsupportedDataTypeException;
import org.jpox.util.Localiser;

/**
 * Factory class for creating Mapping instances.
 * This is called to generate field mappings for the classes to be persisted.
 *
 * @version $Revision: 1.2 $ 
 */
public final class MappingFactory
{
    private static final Localiser LOCALISER = Localiser.getInstance("org.jpox.store.mapping.Localisation");

    /** Private constructor to prevent instantiation. */
    private MappingFactory()
    {
    }

    /**
     * Get a new instance of the Mapping using the DBA and type.
     * @param mappingClass the Mapping class to be created
     * @param dba The Database Adapter
     * @param type The type
     * @return The Mapping
     */
    public static JavaTypeMapping createMapping(Class mappingClass, DatastoreAdapter dba, Class type)
    {
        Object obj = null;
        Object[] args = new Object[]{dba, type};
        Class[] ctr_args_classes = new Class[]{DatastoreAdapter.class, Class.class};
        try
        {
            Constructor ctr = mappingClass.getConstructor(ctr_args_classes);
            try
            {
                obj = ctr.newInstance(args);
            }
            catch (InvocationTargetException e)
            {
                throw new JDOFatalInternalException(LOCALISER.msg("Mapping.ConstructionFailure", mappingClass.getName()), e.getTargetException());
            }
            catch (Exception e)
            {
                throw new JDOFatalInternalException(LOCALISER.msg("Mapping.ConstructionFailure", mappingClass.getName()), e);
            }
        }
        catch (NoSuchMethodException nsme)
        {
            throw new JDOFatalInternalException(LOCALISER.msg("Mapping.NoSuchConstructor", DatastoreAdapter.class.getName(), 
                Class.class.getName(), mappingClass.getName()));
        }
        return (JavaTypeMapping) obj;
    }

    /**
     * Get a new instance of the Mapping using the the DBA, field metadata, and the table managing the field.
     * @param mappingClass the Mapping class to be created
     * @param dba Datastore Adapter
     * @param fmd FieldMetaData for the field to be mapped
     * @param table The Table
     * @return The Mapping
     */
    public static JavaTypeMapping createMapping(Class mappingClass, DatastoreAdapter dba, FieldMetaData fmd, DatastoreClass table)
    {
        Object obj = null;
        Object[] args = new Object[]{dba, fmd, table};
        Class[] ctr_args_classes = new Class[]{DatastoreAdapter.class, FieldMetaData.class, DatastoreClass.class};
        try
        {
            Constructor ctr = mappingClass.getConstructor(ctr_args_classes);
            try
            {
                obj = ctr.newInstance(args);
            }
            catch (UnsupportedDataTypeException e)
            {
                throw new UnsupportedDataTypeException(e.getMessage());
            }
            catch (InvocationTargetException e)
            {
                throw new JDOFatalInternalException(LOCALISER.msg("Mapping.ConstructionFailure2", mappingClass.getName(), table, fmd.getName()), 
                                                    e.getTargetException());
            }
            catch (Exception e)
            {
                throw new JDOFatalInternalException(LOCALISER.msg("Mapping.ConstructionFailure2", mappingClass.getName(), table, fmd.getName()), e);
            }
        }
        catch (NoSuchMethodException nsme)
        {
            throw new JDOFatalInternalException(LOCALISER.msg("Mapping.NoSuchConstructor2", DatastoreAdapter.class.getName(), 
                FieldMetaData.class.getName(), DatastoreClass.class.getName(), mappingClass.getName()));
        }
        return (JavaTypeMapping) obj;
    }
}