/**********************************************************************
Copyright (c) 2003 Erik Bengtson and others. All rights reserved. 
This program and the accompanying materials are made available under
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - coding standards
    ...
**********************************************************************/
package org.jpox.store.mapping;

import org.jpox.store.StatementExpressionIndex;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.rdbms.sqlidentifier.TableIdentifier;

/**
 * Helper class for doing many tasks with mapping
 *
 * @version $Revision: 1.1 $ 
 */
public class Mappings
{
    /**
     * Add columns to a query. Columns added to the query are obtained from the
     * mappings included in the statementExpressionIndex passed as parameter.
     * After adding columns to the query, statementExpressionIndex is updated
     * with the index/position of the columns in the query.
     *  
     * @param qs the Query Statement
     * @param statementExpressionIndex the list of mapping/column indexes to be
     *                                 included in the query
     */
    public static void selectMapping(QueryStatement qs,
                            StatementExpressionIndex[] statementExpressionIndex)
    {
        if (statementExpressionIndex == null)
        {
            return;
        }
        for (int i = 0; i < statementExpressionIndex.length; ++i)
        {
            if (statementExpressionIndex[i] != null)
            {
                JavaTypeMapping m = statementExpressionIndex[i].getMapping();
                if (m != null)
                {
                    int[] columnNumbersByField = qs.select(m,true);
                    statementExpressionIndex[i].setExpressionIndex(columnNumbersByField);
                }
            }
        }
    }
    
    /**
     * Add columns to a query. Columns added to the query are obtained from the
     * mappings included in the statementExpressionIndex passed as parameter.
     * After adding columns to the query, statementExpressionIndex is updated
     * with the index/position of the columns in the query.
     *  
     * @param qs the Query Statement
     * @param tableIdentifier the Identifier for the table which the columns
     *                        are from
     * @param statementExpressionIndex the list of mapping/column indexes to be
     *                                 included in the query
     */
    public static void selectMapping(QueryStatement qs,
                            TableIdentifier tableIdentifier,
                            StatementExpressionIndex[] statementExpressionIndex)
    {
        for (int i = 0; i < statementExpressionIndex.length; ++i)
        {
            if (statementExpressionIndex[i] != null)
            {
                
                JavaTypeMapping m = statementExpressionIndex[i].getMapping();
    
                if (m != null)
                {
                    int[] columnNumbersByField = qs.select(tableIdentifier,m,true);
                    statementExpressionIndex[i].setExpressionIndex(columnNumbersByField);
                }
            }
        }
    }
    
    /**
     * Add columns to a query. Columns added to the query are obtained from the
     * mappings included in the statementExpressionIndex passed as parameter.
     * After adding columns to the query, statementExpressionIndex is updated
     * with the index/position of the columns in the query.
     *  
     * @param qs the Query Statement
     * @param fieldNumbers the field numbers to determine the columns to be
     * added to the query
     * @param statementExpressionIndex the list of mapping/column indexes
     *                                 to be included in the query
     */
    public static void selectMapping(QueryStatement qs,
                            int[] fieldNumbers,
                            StatementExpressionIndex[] statementExpressionIndex)
    {
        if (statementExpressionIndex == null || fieldNumbers == null)
        {
            return;
        }
        for (int i = 0; i < statementExpressionIndex.length; ++i)
        {
            boolean isInFieldNumberList = false;
            for (int j = 0; j < fieldNumbers.length; ++j)
            {
                if (i == fieldNumbers[j])
                {
                    isInFieldNumberList = true;
                    break;
                }                    
            }
            if (statementExpressionIndex[i] != null && isInFieldNumberList)
            {
                JavaTypeMapping m = statementExpressionIndex[i].getMapping();
                if (m != null)
                {
                    int[] columnNumbersByField = qs.select(m,true);
                    statementExpressionIndex[i].setExpressionIndex(columnNumbersByField);
                }
            }
        }
    }
    
    /**
     * Add columns to a query. Columns added to the query are obtained from the
     * mappings included in the statementExpressionIndex passed as parameter.
     * After adding columns to the query, statementExpressionIndex is updated
     * with the index/position of the columns in the query.
     *  
     * @param qs the Query Statement
     * @param fieldNumbers the field numbers to determine the columns to be
     *                     added to the query
     * @param tableIdentifier the Identifier for the table which the columns
     *                        are from
     * @param statementExpressionIndex the list of mapping/column indexes to
     *                                 be included in the query
     */
    public static void selectMapping(QueryStatement qs,
                            TableIdentifier tableIdentifier,
                            int[] fieldNumbers,
                            StatementExpressionIndex[] statementExpressionIndex)
    {
        if (statementExpressionIndex == null || fieldNumbers == null)
        {
            return;
        }
        for (int i = 0; i < statementExpressionIndex.length; ++i)
        {
            boolean isInFieldNumberList = false;
            for (int j = 0; j < fieldNumbers.length; ++j)
            {
                if (i == fieldNumbers[j])
                {
                    isInFieldNumberList = true;
                    break;
                }                    
            }
            if (statementExpressionIndex[i] != null && isInFieldNumberList)
            {
                
                JavaTypeMapping m = statementExpressionIndex[i].getMapping();
    
                if (m != null)
                {
                    int[] columnNumbersByField = qs.select(tableIdentifier,m,true);
                    statementExpressionIndex[i].setExpressionIndex(columnNumbersByField);
                }
            }
        }
    }    
    
    /**
     * Get the indexes/position for Mapping columns in a given the initial 
     * position. Used to set parameters in PreparedStatements
     * @param initialPosition the initialPosition
     * @param mapping the Mapping
     * @return an array containing indexes for parameters
     **/
    public static int[] getParametersIndex(int initialPosition, JavaTypeMapping mapping)
    {
        if (mapping.getNumberOfDatastoreFields() < 1)
        {
            return new int[]{initialPosition};
        }

        int parameter[] = new int[mapping.getNumberOfDatastoreFields()];
        for (int i=0; i<parameter.length; i++)
        {
            parameter[i] = initialPosition+i;
        }
        return parameter;
    }
}
