/**********************************************************************
Copyright (c) 2002 Kelly Grizzle (TJDO) and others. All rights reserved.
This program and the accompanying materials are made available under
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2002 Mike Martin (TJDO)
2003 Andy Jefferson - coding standards
    ...
**********************************************************************/
package org.jpox.store.mapping;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOObjectNotFoundException;

import oracle.sql.CLOB;

import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.metadata.FieldMetaData;
import org.jpox.store.DatastoreAdapter;
import org.jpox.store.DatastoreClass;
import org.jpox.store.FetchStatement;
import org.jpox.store.rdbms.RDBMSFetchStatement;
import org.jpox.util.JPOXLogger;

/**
 * Mapping for a String type.
 *
 * @version $Revision: 1.4 $
 **/
public class OracleStringMapping extends StringMapping implements MappingCallbacks
{
    public OracleStringMapping(DatastoreAdapter dba, Class type)
    {
        super(dba, type);
    }

    /**
     * Creates a new OracleStringMapping object.
     * @param dba Database Adapter
     * @param fmd FieldMetaData for the field
     * @param table Table for persisting this field
     */
    public OracleStringMapping(DatastoreAdapter dba, FieldMetaData fmd, DatastoreClass table)
    {
        super(dba, fmd, table);
    }

    /**
     * Retrieve the empty CLOB locator created by the insert statement
     * and write out the current CLOB field value to the Oracle CLOB object
     */
    public void postInsert(StateManager sm)
    {
        JavaTypeMapping idMapping = getDatastoreClass().getIDMapping();

        FetchStatement fetchStmt = new RDBMSFetchStatement(getDatastoreClass());
        //only one column for this mapping is supported
        fetchStmt.select(getDataStoreMapping(0).getDatastoreField());

        // and condition with columns from the id
        for (int i=0; i<idMapping.getNumberOfDatastoreFields(); i++)
        {
			fetchStmt.andCondition(fetchStmt.referenceColumn(idMapping.getDataStoreMapping(i).getDatastoreField()) + " = ?");
        }

        // Assumed to require locking of the rows
        String textStmt = fetchStmt.toString(true);

        String value = (String)sm.provideField(fmd.getAbsoluteFieldNumber());

        if (value == null)
        {
            value = "";
        }
        else if (value.length() == 0)
        {
            value = getDatabaseAdapter().getSurrogateForEmptyStrings();
        }

        PersistenceManager pm = sm.getPersistenceManager();

        try
        {
            Connection conn = pm.getConnection(true);

            try
            {
                PreparedStatement ps = conn.prepareStatement(textStmt);

                try
                {
                    Object pcObject = sm.getObject();
                    idMapping.setObject(pm, ps, Mappings.getParametersIndex(1,idMapping), pcObject);
                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(textStmt);
                    }

                    long startTime = System.currentTimeMillis();

                    ResultSet rs = ps.executeQuery();

                    try
                    {
                        if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                        {
                            JPOXLogger.RDBMS_SQL.debug(LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                        }

                        if (!rs.next())
                        {
                            throw new JDOObjectNotFoundException("No such database row", sm.getInternalObjectId());
                        }

                        CLOB clob = (CLOB)rs.getClob(1);

                        if (clob != null)
                        {
                            clob.putString(1, value);
                        }
                    }
                    finally
                    {
                        rs.close();
                    }
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                pm.releaseConnection(conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException("postInsert() failed: " + textStmt, e);
        }
        
    }
    
    public void deleteDependent(StateManager sm)
    {
    }

    public void postFetch(StateManager sm)
    {
    }

    public void postUpdate(StateManager sm)
    {
        postInsert(sm);
    }

    public void preDelete(StateManager sm)
    {
    }

}
