/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others. All rights reserved. 
This program and the accompanying materials are made available under
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Erik Bengtson - Important changes regarding the application identity
                     support when fetching PreparedStatements. Added an
                     inner class that should be moved away.
2004 Erik Bengtson - added commentary and Localisation
2004 Andy Jefferson - added capability to handle Abstract PC fields for
                      both SingleFieldIdentity and AID
    ...
**********************************************************************/
package org.jpox.store.mapping;

import java.lang.reflect.Field;
import java.lang.reflect.Modifier;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOObjectNotFoundException;
import javax.jdo.JDOUserException;
import javax.jdo.SingleFieldIdentity;
import javax.jdo.spi.PersistenceCapable;

import org.jpox.ClassLoaderResolver;
import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.metadata.ClassMetaData;
import org.jpox.metadata.FieldMetaData;
import org.jpox.metadata.IdentityType;
import org.jpox.store.DatastoreAdapter;
import org.jpox.store.DatastoreClass;
import org.jpox.store.FieldValues;
import org.jpox.store.OID;
import org.jpox.store.StatementExpressionIndex;
import org.jpox.store.exceptions.NotYetFlushedException;
import org.jpox.store.expression.ObjectExpression;
import org.jpox.store.expression.ObjectLiteral;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.TableExpression;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.rdbms.fieldmanager.ResultSetGetter;
import org.jpox.util.AIDUtils;
import org.jpox.util.JPOXLogger;

/**
 * Maps a java field to a PersistenceCapable class. For PersistenceCapable
 * classes using datastore identity, most of the necessary behaviour associated
 * are coded in the OIDMapping super class. 
 *
 * @version $Revision: 1.13 $ 
 **/
public class PersistenceCapableMapping extends OIDMapping
{
    private List javaTypeMappings;

    /**
     * Constructor.
     * @param dba Datastore Adapter
     * @param type The type of PersistenceCapable
     */
    public PersistenceCapableMapping(DatastoreAdapter dba, Class type)
    {
        super(dba, type);
    }

    /**
     * Creates a new PersistenceCapableMapping object.
     * @param dba Database Adapter
     * @param fmd FieldMetaData for the field
     * @param table Table for persisting this field
     */
    public PersistenceCapableMapping(DatastoreAdapter dba, FieldMetaData fmd, DatastoreClass table)
    {
        super(dba, fmd, table);
    }
   	
    /**
     * Add a new JavaTypeMapping
     * @param mapping the JavaTypeMapping
     */
	public void addJavaTypeMapping(JavaTypeMapping mapping)
	{
	    if (javaTypeMappings == null)
	    {
	        javaTypeMappings = new ArrayList();
	    }
	    if (mapping == null)
	    {
	        throw new JDOFatalInternalException("mapping argument in PersistenceCapableMapping.addJavaTypeMapping is null");
	    }
	    javaTypeMappings.add(mapping);	    
	}

	public int getNumberOfDatastoreFields()
    {
        int number = 0;
        for (int i=0; i<javaTypeMappings.size(); i++)
        {
            number += ((JavaTypeMapping)javaTypeMappings.get(i)).getNumberOfDatastoreFields();
        }
        return number;
    }

	/* (non-Javadoc)
	 * @see org.jpox.store.mapping.JavaTypeMapping#getDataStoreMapping(int)
	 */
	public DatastoreMapping getDataStoreMapping(int index)
	{
        List datastoreMappings = new ArrayList();
        for (int i=0; i<javaTypeMappings.size(); i++)
        {
            int number = ((JavaTypeMapping)javaTypeMappings.get(i)).getNumberOfDatastoreFields();
            for (int j=0; j<number; j++)
            {
                datastoreMappings.add(((JavaTypeMapping)javaTypeMappings.get(i)).getDataStoreMapping(j));
            }
        }
	    return (DatastoreMapping) datastoreMappings.get(index);
	}

	/**
	 * Method to set an object in the datastore.
	 * @param pm The Persistence Manager
	 * @param ps The Prepared Statement
	 * @param param The parameter ids in the statement
	 * @param value The value to put in the statement at these ids
	 * @throws NotYetFlushedException
	 */
    public void setObject(PersistenceManager pm, Object ps, int[] param, Object value)
    {
        Object id;

        if (value == null)
        {
            // Null out the PC object
            int n=0;
            for (int i=0; i<javaTypeMappings.size(); i++)
            {
                JavaTypeMapping mapping = ((JavaTypeMapping)javaTypeMappings.get(i));
                int[] posMapping = new int[mapping.getNumberOfDatastoreFields()];
                for (int j=0; j<posMapping.length; j++)
                {
                    posMapping[j] = param[n++];
                }
                mapping.setObject(pm, ps, posMapping, null);
            }
        }
        else
        {
            if (!(value instanceof PersistenceCapable)) 
            {
                throw new JDOFatalInternalException(LOCALISER.msg("Mapping.ObjectIsNotPersistenceCapable",value.getClass(),value));
            }
            else
            {
                PersistenceCapable pc = (PersistenceCapable) value;
                if (!pm.isInserting(pc))
                {
                    id = pc.jdoGetObjectId();

                    // Check if the PersistenceCapable exists in this datastore
                    boolean exists = true;
                    if (id == null)
                    {
                        exists = false;
                    }
                    else if (pc.jdoIsDetached())
                    {
                        // Any object that is detached will have an id, but may not be persistent in
                        // this datastore (may be detached from a different datastore).
                        try
                        {
                            // Check if the object exists in this datastore
                            pm.getObjectById(id, true, false);
                        }
                        catch (JDOObjectNotFoundException onfe)
                        {
                            exists = false;
                        }
                    }

                    if (!exists)
                    {
                        // Persist anything that doesnt exist
                        pm.makePersistent(value);
                        pm.flush();
                        id = pc.jdoGetObjectId();
                    }
                    else
                    {
                        // Cannot assign field to an object from a different PM, so throw exception
                        if (!pc.jdoIsDetached() && ((PersistenceManager)pc.jdoGetPersistenceManager()) != null)
                        {
                            if (pm.getPMHandle() != ((PersistenceManager)pc.jdoGetPersistenceManager()).getPMHandle())
                            {
                                throw new JDOUserException(LOCALISER.msg("Mapping.FieldNotAssignableIsFromAnotherPM"), id);
                            }
                        }
                    }
                    if (id instanceof OID)
                    {
                        super.setObject(pm, ps, param, id);
                    }
                    else
                    {
                        pc.jdoCopyKeyFieldsFromObjectId(new AppIDObjectIdFieldConsumer((PreparedStatement) ps, param), id);
                    }
                }
                else
                {
                    // bidirectional association
                    int n=0;
                    for (int i=0; i<javaTypeMappings.size(); i++)
                    {
                        JavaTypeMapping mapping = ((JavaTypeMapping)javaTypeMappings.get(i));
                        int[] posMapping = new int[mapping.getNumberOfDatastoreFields()];
                        for (int j=0; j<posMapping.length; j++)
                        {
                            posMapping[j] = param[n++];
                        }
                        mapping.setObject(pm, ps, posMapping, null);
                    }
                    throw new NotYetFlushedException(pc);
                }
            }
        }
    }

    /**
     * Returns a instance of a PersistenceCapable class.
     * Processes a FK field and converts the id stored firstly into an OID/AID
     * and then into the object that the FK id relates to.
     * @param pm The Persistence Manager
     * @param rs The ResultSet
     * @param param Array of parameter ids in the ResultSet to retrieve
     * @return The Persistence Capable object
     */
    public Object getObject(PersistenceManager pm, final Object rs, int[] param)
    {
        // Check for null FK
        try
        {
            if (((ResultSet) rs).getObject(param[0]) == null)
            {
                return null;
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(e.toString());
        }

        ClassMetaData cmd = pm.getMetaDataManager().getMetaDataForClass(this.type);
        if (cmd.getIdentityType() == IdentityType.DATASTORE)
        {
            // Datastore Identity - retrieve the OID for the class.
            // Note that this is a temporary OID that is simply formed from the type of base class in the relationship
            // and the id stored in the FK. The real OID for the object may be of a different class.
            // For that reason we get the object by checking the inheritance (final param in getObjectById())
            Object oid = super.getObject(pm, rs, param);
            return oid == null ? null : pm.getObjectById(oid, false, true);
        }
        else if (cmd.getIdentityType() == IdentityType.APPLICATION)
        {
            ClassLoaderResolver clr = pm.getClassLoaderResolver();

            // Application Identity - retrieve the AID for the class
            if (Modifier.isAbstract(this.type.getModifiers()) && cmd.getObjectidClass() != null)
            {
                // Abstract class, so we need to generate an AID before proceeding
                Class objectIdClass = clr.classForName(cmd.getObjectidClass());
                Class pcClass = clr.classForName(cmd.getFullClassName());
                if (AIDUtils.isSingleFieldIdentityClass(cmd.getObjectidClass()))
                {
                    // SingleFieldIdentity
                    List params = new ArrayList();
                    int totalFieldCount = cmd.getNoOfInheritedManagedFields() + cmd.getNoOfManagedFields();
                    int paramIndex = 0;
                    for (int i=0; i<totalFieldCount; ++i)
                    {
                        FieldMetaData fmd = cmd.getManagedFieldAbsolute(i);
                        if (fmd.isPrimaryKey())
                        {
                            JavaTypeMapping m = pm.getStoreManager().getDatastoreClass(cmd.getFullClassName()).getFieldMapping(fmd);
                            if (m.includeInFetchStatement())
                            {
                                for (int j = 0; j < m.getNumberOfDatastoreFields(); j++)
                                {
                                    params.add(new Integer(param[paramIndex++]));
                                }
                            }
                        }
                    }
                    int paramNumber = ((Integer)params.get(0)).intValue();
                    SingleFieldIdentity id = null;
                    try
                    {
                        id = AIDUtils.getNewSingleFieldIdentity(objectIdClass, pcClass, ((ResultSet)rs).getObject(paramNumber));
                    }
                    catch (SQLException sqe)
                    {
                        JPOXLogger.JDO.error(LOCALISER.msg("PersistenceCapableMapping.CantCreateAID",cmd.getObjectidClass(),sqe));
                        return null;
                    }
                    return pm.getObjectById(id, false, true);
                }
                else
                {
                    // Users own AID
                    Object id = null;
                    FieldMetaData fmd = null;
                    Object fieldValue = null;
                    try
                    {
                        // Create an AID
                        id = objectIdClass.newInstance();

                        // Set the fields of the AID
                        int totalFieldCount = cmd.getNoOfInheritedManagedFields() + cmd.getNoOfManagedFields();
                        int paramIndex = 0;
                        for (int i=0; i<totalFieldCount; ++i)
                        {
                            fmd = cmd.getManagedFieldAbsolute(i);
                            if (fmd.isPrimaryKey())
                            {
                                Field field = objectIdClass.getField(fmd.getName());

                                JavaTypeMapping m = pm.getStoreManager().getDatastoreClass(cmd.getFullClassName()).getFieldMapping(fmd);
                                if (m.includeInFetchStatement())
                                {
                                    // NOTE This assumes that each field has one datastore column.
                                    for (int j = 0; j < m.getNumberOfDatastoreFields(); j++)
                                    {
                                        Object obj = ((ResultSet)rs).getObject(param[paramIndex++]);
                                        // field with multiple columns should have values returned from db merged here
                                        fieldValue = obj;
                                    }
                                    field.set(id, fieldValue);
                                }
                            }
                        }
                    }
                    catch (Exception e)
                    {
                        JPOXLogger.JDO.error(LOCALISER.msg("PersistenceCapableMapping.CantCreateAIDWithReflection",cmd.getObjectidClass(),fmd == null ? null : fmd.getName(),fieldValue,e));
                        return null;
                    }

                    return pm.getObjectById(id, false, true);
                }
            }
            else
            {
                // Concrete class, so we can use pm.getObjectByAID() to avoid using reflection
                int totalFieldCount = cmd.getNoOfInheritedManagedFields() + cmd.getNoOfManagedFields();
                final StatementExpressionIndex[] statementExpressionIndex = new StatementExpressionIndex[totalFieldCount];
                List fields = new ArrayList();
                int paramIndex = 0;
                for (int i=0; i<totalFieldCount; ++i)
                {
                    FieldMetaData fmd = cmd.getManagedFieldAbsolute(i);
                    if (cmd.getManagedFieldAbsolute(i).isPrimaryKey())
                    {
                        JavaTypeMapping m = pm.getStoreManager().getDatastoreClass(cmd.getFullClassName()).getFieldMapping(fmd);
                        statementExpressionIndex[i] = new StatementExpressionIndex();
                        statementExpressionIndex[i].setMapping(m);
                        if (m.includeInFetchStatement())
                        {
                            fields.add(new Integer(i));
                            int expressionsIndex[] = new int[m.getNumberOfDatastoreFields()];
                            for (int j = 0; j < expressionsIndex.length; j++)
                            {
                                expressionsIndex[j] = param[paramIndex++];
                            }
                            statementExpressionIndex[i].setExpressionIndex(expressionsIndex);
                        }
                    }
                }

                final int[] fieldNumbers = new int[fields.size()];
                for (int i = 0; i < fieldNumbers.length; i++)
                {
                    fieldNumbers[i] = ((Integer) fields.get(i)).intValue();
                }

                return pm.getObjectByAID(clr.classForName(cmd.getFullClassName()),
                            new FieldValues()
                            {
    		                    // StateManager calls the fetchFields method
    		                    public void fetchFields(StateManager sm)
    		                    {
    		                        sm.replaceFields(fieldNumbers, new ResultSetGetter(sm, (ResultSet) rs, statementExpressionIndex));
    		                    }
                            }, false, true);
            }
        }
        else
        {
            // TODO Throw an exception since we dont persist any other PersistenceCapable!
            return null;
        }
    }

    public ScalarExpression newLiteral(QueryStatement qs, Object value)
    {
        ScalarExpression expr = new ObjectLiteral(qs, this, value,this.type);
        return expr;        
    }

    public ScalarExpression newScalarExpression(QueryStatement qs, TableExpression te)
    {
        ScalarExpression expr = new ObjectExpression(qs, this, te);
        return expr;        
    }
}