/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved.
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2004 Andy Jefferson - added javadocs
2004 Andy Jefferson - changed to use Column spec from MetaData
    ...
**********************************************************************/
package org.jpox.store.mapping;

import java.sql.Timestamp;

import javax.jdo.JDOException;

import org.jpox.PersistenceManager;
import org.jpox.metadata.ColumnMetaData;
import org.jpox.store.DatastoreAdapter;
import org.jpox.store.DatastoreField;
import org.jpox.store.Role;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.TableExpression;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.rdbms.sqlidentifier.SQLIdentifier;
import org.jpox.store.rdbms.table.ClassTable;

/**
 * Mapping class for mapping version state/timestamp columns in the database.
 *
 * @version $Revision: 1.2 $
 */
public class VersionMapping extends SingleFieldMapping implements SimpleDatastoreRepresentation
{
    private final SingleFieldMapping delegate;

    /**
     * Constructor.
     * TODO Remove ClassTable from here and replace with DatastoreClass
     * @param dba Datastore Adapter
     * @param table Datastore table
     * @param type Type to use for the column
     */
    public VersionMapping(DatastoreAdapter dba, ClassTable table, Class type)
    {
        super(dba, type);
        delegate = (SingleFieldMapping) dba.getMapping(type);

        // Currently we only use a single column mapping for versioning.
        // The MetaData supports multiple columns and so we could extend this in the future
        // to use all MetaData information.
        ColumnMetaData[] versionColumnMetaData = table.getVersionMetaData().getColumnMetaData();
        ColumnMetaData colmd;
        if (versionColumnMetaData.length == 0)
        {
            // user not defined the column name
            colmd = new ColumnMetaData(table.getVersionMetaData(), "opt_version");
            table.getVersionMetaData().addColumn(colmd);
        }
        else
        {
            colmd = versionColumnMetaData[0];
        }
        DatastoreField column = table.addDatastoreField(type,
            new SQLIdentifier(getDatabaseAdapter(), colmd.getName(), colmd.getName()),
            Role.CUSTOM, this, colmd);

        table.getStoreManager().getMappingManager().getDatastoreMapping(delegate, table.getStoreManager(), column, type);
    }
    
    /**
     * Accessor for a sample value for this column
     * @return The sample value
     */
    public Object getSampleValue()
    {
        return delegate.getSampleValue();
    }

    /**
     * Accessor for a new literal for this mapping.
     * @param qs The QueryStatement
     * @param value The value of the object
     * @return The new literal
     */
    public ScalarExpression newLiteral(QueryStatement qs, Object value)
    {
        return delegate.newLiteral(qs, value);
    }

    /**
     * Accessor for a new literal for this mapping.
     * @param qs The QueryStatement
     * @param te The table expression
     * @return The new literal
     */
    public ScalarExpression newScalarExpression(QueryStatement qs, TableExpression te)
    {
        return delegate.newScalarExpression(qs, te);
    }

    /**
     * Accessor for whether to include this column in any fetch statement
     * @return Whether to include the column when fetching.
     */
    public boolean includeInFetchStatement()
    {
        return false;
    }

    public int getNumberOfDatastoreFields()
    {
        return delegate.getNumberOfDatastoreFields();
    }

    public DatastoreMapping getDataStoreMapping(int index)
    {
        return delegate.getDataStoreMapping(index);
    }

    public void addDataStoreMapping(DatastoreMapping datastoreMapping)
    {
        datastoreMappings.add(datastoreMapping);
    }

    /**
     * Accessor for the next version of the object
     * @param actualVersion The actual version
     * @return The next version
     */
    public Object getNextVersion(Object actualVersion)
    {
        if (type == Long.class)
        {
            if (actualVersion == null)
            {
                return new Long(1);
            }
            return new Long(((Long)actualVersion).longValue()+1);
        }
        if (type == Timestamp.class)
        {
            return new Timestamp(System.currentTimeMillis());
        }
        throw new JDOException("Unsupported optimistic type");
    }
    
    /**
     * Compare the transaction vs datastore version.
     * @param versionDatastore
     * @param versionTransaction
     * @return true if transaction version == datastore version
     */
    public boolean compareVersion(Object versionDatastore, Object versionTransaction)
    {
        if( versionTransaction == null )
        {
            return true;
        }
        if (type == Long.class)
        {
            return ((Long)versionTransaction).longValue() == ((Long)versionDatastore).longValue();
        }
        if (type == Timestamp.class)
        {
            return ((Timestamp)versionTransaction).getTime() == ((Timestamp)versionDatastore).getTime();
        }
        throw new JDOException("Unsupported optimistic type");
    }

    /**
     * Accessor for the type represented here, returning the class itself
     * @return This class.
     */
    public Class getJavaType()
    {
        return VersionMapping.class;
    }

    /**
     * Mutator for the object in this column
     * @param pm The Persistence Manager
     * @param preparedStatement The statement
     * @param exprIndex The indexes
     * @param value The value to set it to
     */
    public void setObject(PersistenceManager pm, Object preparedStatement, int[] exprIndex, Object value)
    {
        delegate.setObject(pm, preparedStatement, exprIndex, value);
    }

    /**
     * Accessor for the object in this column
     * @param pm The Persistence manager
     * @param resultSet The ResultSet to get the value from
     * @param exprIndex The indexes
     * @return The object
     */
    public Object getObject(PersistenceManager pm, Object resultSet, int[] exprIndex)
    {
        return delegate.getObject(pm, resultSet, exprIndex);
    }
}