/**********************************************************************
Copyright (c) 2005 Andy Jefferson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
    ...
**********************************************************************/
package org.jpox.store.query;

import java.util.Collection;
import java.util.List;
import java.util.Map;

import javax.jdo.Extent;
import javax.jdo.JDOUserException;

import org.jpox.PersistenceManager;
import org.jpox.store.StatementExpressionIndex;
import org.jpox.util.JPOXLogger;

/**
 * Base definition of a query using SQL.
 *
 * @see Query
 * @version $Revision: 1.2 $
 */
public abstract class BaseSQLQuery extends Query
{
    protected transient final String sqlText;

    protected transient String jdbcSqlText = null;

    protected transient int[] fieldNumbers = null;
    protected transient List fieldColumnNames = null;
    protected transient StatementExpressionIndex[] statementExpressionIndex;

    /**
     * Constructs a new query instance having the same criteria as the given
     * query.
     *
     * @param pm The PersistenceManager
     * @param sql_text The SQL query string
     */
    public BaseSQLQuery(PersistenceManager pm, String sql_text)
    {
        super(pm);

        candidateClass = null;
        filter = null;
        imports = null;
        variables = null;
        parameters = null;
        ordering = null;

        if (sql_text == null)
        {
            throw new JDOUserException(LOCALISER.msg("SQL.StatementNullError"));
        }
        this.sqlText = sql_text.trim();
        if (!sqlText.startsWith("SELECT") && !sqlText.startsWith("select") && !sqlText.startsWith("Select"))
        {
            throw new JDOUserException(LOCALISER.msg("SQL.StatementNotSelectError", sqlText));
        }
    }

    /**
     * Utility to discard any compiled query.
     * @see org.jpox.store.query.Query#discardCompiled()
     */
    protected void discardCompiled()
    {
        super.discardCompiled();

        jdbcSqlText = null;
        fieldNumbers = null;
        statementExpressionIndex = null;
        fieldColumnNames = null;
    }

    /**
     * Set the candidate Extent to query.
     * <p>This implementation always throws a JDOUserException since this
     * concept doesn't apply to SQL queries.
     *
     * @param pcs   the Candidate Extent.
     * @exception JDOUserException  Always thrown.
     * @see javax.jdo.Query#setCandidates(javax.jdo.Extent)
     */
    public void setCandidates(Extent pcs)
    {
        throw new JDOUserException(LOCALISER.msg("SQL.CandidateExtentsInapplicableError"));
    }

    /**
     * Set the candidate Collection to query.
     *
     * <p>This implementation always throws a JDOUserException since this
     * concept doesn't apply to SQL queries.
     *
     * @param pcs the Candidate collection.
     * @exception JDOUserException  Always thrown.
     * @see javax.jdo.Query#setCandidates(java.util.Collection)
     */
    public void setCandidates(Collection pcs)
    {
        throw new JDOUserException(LOCALISER.msg("SQL.CandidateCollectionsInapplicableError"));
    }

    /**
     * Set the result for the results. The application might want to get results
     * from a query that are not instances of the candidate class. The results
     * might be fields of persistent instances, instances of classes other than
     * the candidate class, or aggregates of fields.
     * 
     * <p>This implementation always throws a JDOUserException since this
     * concept doesn't apply to SQL queries.
     * 
     * @param result The result parameter consists of the optional keyword
     * distinct followed by a commaseparated list of named result expressions or
     * a result class specification.
     * @see javax.jdo.Query#setResult
     * @since 1.1
     */
    public void setResult(String result)
    {
        throw new JDOUserException(LOCALISER.msg("SQL.ResultInapplicableError"));
    }

    /**
     * Set the range of the results. Not applicable to SQL/JPOXSQL queries.
     * @param fromIncl From element no (inclusive) to return
     * @param toExcl To element no (exclusive) to return
     * @see javax.jdo.Query#setRange
     * @since 1.1
     */
    public void setRange(int fromIncl, int toExcl)
    {
        throw new JDOUserException(LOCALISER.msg("SQL.RangeInapplicableError"));
    }

    /**
     * Method to set whether to use subclasses. This is not used with JPOXSQL.
     * @param subclasses Whether to use subclasses
     */
    public void setSubclasses(boolean subclasses)
    {
        throw new JDOUserException(LOCALISER.msg("SQL.CandidateExtentsInapplicableError"));
    }

    /**
     * Set the filter for the query.
     * <p>This implementation always throws a JDOUserException since this
     * concept doesn't apply to SQL queries.
     * @param filter the query filter.
     * @exception JDOUserException  Always thrown.
     * @see javax.jdo.Query#setFilter
     */
    public void setFilter(String filter)
    {
        throw new JDOUserException(LOCALISER.msg("SQL.FilterStringsInapplicableError"));
    }

    /**
     * Declare the unbound variables to be used in the query.
     * <p>This implementation always throws a JDOUserException since this
     * concept doesn't apply to SQL queries.
     * @param variables the variables separated by semicolons.
     * @exception JDOUserException  Always thrown.
     * @see javax.jdo.Query#declareVariables
     */
    public void declareVariables(String variables)
    {
        throw new JDOUserException(LOCALISER.msg("SQL.VariablesInapplicableError"));
    }

    /**
     * Set the grouping specification for the result Collection.
     * @param grouping the grouping specification.
     * @see javax.jdo.Query#setGrouping
     */
    public void setGrouping(String grouping)
    {
        throw new JDOUserException(LOCALISER.msg("SQL.GroupingNotExplicitError"));
    }

    /**
     * Set the ordering specification for the result Collection.
     * <p>This implementation always throws a JDOUserException since this
     * concept doesn't apply to SQL queries.
     * @param ordering  the ordering specification.
     * @exception JDOUserException  Always thrown.
     * @see javax.jdo.Query#setOrdering
     */
    public void setOrdering(String ordering)
    {
        throw new JDOUserException(LOCALISER.msg("SQL.OrderingNotExplicitError"));
    }

    /**
     * Verify the elements of the query and provide a hint to the query to
     * prepare and optimize an execution plan.
     * @see javax.jdo.Query#compile
     */
    public void compile()
    {
        if (!isCompiled)
        {
            super.compile();
            generateQueryStatement();
            if (JPOXLogger.JDO_QUERY.isDebugEnabled())
            {
                JPOXLogger.JDO_QUERY.debug(LOCALISER.msg("SQL.Statement.Summary", jdbcSqlText));
            }

            isCompiled = true;
        }
    }

    /**
     * Convenience method to take the user specified SQL statement, and align it with
     * field names for the candidate class etc.
     */
    protected abstract void generateQueryStatement();

    /**
     * Execute the query and return the filtered QueryResult.
     * @param parameters the Map containing all of the parameters.
     * @return the filtered QueryResult.
     */
    public abstract Collection performExecute(Map parameters);

    /**
     * Execute the query to delete persistent objects.
     * @param parameters the Map containing all of the parameters.
     * @return the filtered QueryResult of the deleted objects.
     */
    protected QueryResult performDeletePersistentAll(Map parameters)
    {
        throw new JDOUserException(LOCALISER.msg("Query.DeletePersistent.NotYetSupported"));
    }
}