/**********************************************************************
Copyright (c) 2003 Erik Bengtson and others. All rights reserved. 
This program and the accompanying materials are made available under
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2004 Andy Jefferson - coding standards
    ...
**********************************************************************/
package org.jpox.store.query;

import java.util.Collection;
import java.util.Iterator;

import javax.jdo.Extent;
import javax.jdo.JDOFatalInternalException;

import org.jpox.PersistenceManager;
import org.jpox.store.expression.BooleanExpression;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.query.Query.ResultObjectFactory;
import org.jpox.util.Localiser;

/**
 * Collection for candidates passed to the query by setCandidates(collection
 * user), and the user collection is not queryable
 * @version $Revision: 1.3 $
 */
public class CollectionCandidates implements Queryable
{
    protected static final Localiser LOCALISER=Localiser.getInstance("org.jpox.store.query.Localisation");

    private Collection userCandidates;

    private Extent extent;

    private final PersistenceManager pm;

    /**
     * Constructor.
     * @param pm Persistence Manager
     * @param candidateClass the Class candidate
     * @param candidates The candidates
     */
    public CollectionCandidates(PersistenceManager pm, Class candidateClass, Collection candidates)
    {
        this.userCandidates = candidates;
        this.pm = pm;

        if (candidates == null)
        {
            throw new JDOFatalInternalException(LOCALISER.msg("Candidates.CollectionCantBeNull"));
        }

        extent = pm.getExtent(candidateClass, true);
    }

    /**
     * @return Returns the userCandidates.
     */
    public Collection getUserCandidates()
    {
        return userCandidates;
    }
    
    /**
     * Accessor for a new Query statement.
     * @return The Query Statement
     */
    public QueryStatement newQueryStatement()
    {
        return ((Queryable) extent).newQueryStatement();
    }

	/**
     * Creates a QueryStatement The elements that are ALLOWED to be returned
     * after quering the database are the set of elements contained in the
     * candidate collection.
     * @see org.jpox.store.query.Queryable#newQueryStatement(java.lang.Class)
     */
    public QueryStatement newQueryStatement(Class candidateClass)
    {
        QueryStatement stmt = ((Queryable) extent).newQueryStatement(candidateClass);
        JavaTypeMapping m = pm.getStoreManager().getDatastoreClass(candidateClass.getName()).getIDMapping();

        /*
         * creates a query like WHERE ... AND (CANDIDATE_ID = ?1 OR CANDIDATE_ID =
         * ?2) or for classes with composite primary keys WHERE ... AND (
         * (CANDIDATE_IDa = ?1a AND CANDIDATE_IDb = ?1b) OR (CANDIDATE_IDa = ?2a
         * AND CANDIDATE_IDb = ?2b) )
         */
        BooleanExpression elementsExpr = null;
        for (Iterator it = userCandidates.iterator(); it.hasNext();)
        {
            Object candidateValue = it.next();
            BooleanExpression keyExpr = null;
            ScalarExpression expr = m.newScalarExpression(stmt, stmt.getDefaultTableExpression());
            if (keyExpr == null)
            {
                keyExpr = expr.eq(m.newLiteral(stmt, candidateValue));
            }
            else
            {
                keyExpr = keyExpr.and(expr.eq(m.newLiteral(stmt, candidateValue)));
            }
            if (elementsExpr == null)
            {
                elementsExpr = keyExpr;
            }
            else
            {
                elementsExpr = elementsExpr.ior(keyExpr);
            }
        }
        if (elementsExpr != null)
        {
            stmt.andCondition(elementsExpr, true);
        }
        return stmt;
    }

    /* (non-Javadoc)
     * @see org.jpox.store.query.Queryable#newResultObjectFactory(org.jpox.store.QueryStatement, boolean)
     */
    public ResultObjectFactory newResultObjectFactory(QueryStatement stmt, boolean ignoreCache,Class resultClass)
    {
        return ((Queryable) extent).newResultObjectFactory(stmt, ignoreCache, resultClass);
    }

    /**
     * Returns <tt>true</tt> if this collection contains no elements.<p>
     * @return <tt>true</tt> if this collection contains no elements.
     */
    public boolean isEmpty()
    {
        return userCandidates.isEmpty();
    }
}