/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Marco Schulze - replaced catch(NotPersistenceCapable) with TypeManager.isSupportedType()
2004 Andy Jefferson - added restriction on Discriminator, added check on subclass using same table
2004 Andy Jefferson - added comments. Added fix to inverse map with non primitive keys
2004 Andy Jefferson - added way of enforcing inclusion of "JPOXMETADATA"
2005 Andy Jefferson - catch "no-table" cases where user has a strange inheritance setup
    ...
**********************************************************************/
package org.jpox.store.query;

import java.util.Iterator;

import org.jpox.TypeManager;
import org.jpox.metadata.ClassMetaData;
import org.jpox.metadata.DiscriminatorMetaData;
import org.jpox.metadata.DiscriminatorStrategy;
import org.jpox.store.DatastoreAdapter;
import org.jpox.store.DatastoreClass;
import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.StoreManager;
import org.jpox.store.exceptions.IncompatibleQueryElementTypeException;
import org.jpox.store.expression.NullLiteral;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.TableExpression;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.rdbms.adapter.DatabaseAdapter;
import org.jpox.store.rdbms.sqlidentifier.TableIdentifier;
import org.jpox.store.rdbms.table.ClassTable;
import org.jpox.store.rdbms.table.SCOTable;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;
import org.jpox.util.StringUtils;

/**
 * Class to generate a QueryStatement for iterating through the elements
 * of a Set. This can be in an Extent/Query for example where the user has
 * selected a candidate class and to include/exclude subclasses. The generated
 * query statement typically contains UNIONs for each of the possible classes
 * involved.
 * <P>
 * To give an example, lets assume that we have class A which is the candidate
 * and this has a subclass B. We want to find all objects of the candidate type
 * and optionally its subclasses and we want information about what type the object
 * is (A or B). The query will be of the form
 * <PRE>
 * SELECT THIS.A_ID,'org.jpox.samples.A' as JPOXMETADATA, THIS.COLUMN1, THIS.COLUMN2 
 * FROM A THIS
 *   LEFT OUTER JOIN B SUBELEMENT0 ON SUBELEMENT0.B_ID = THIS.A_ID
 * WHERE SUBELEMENT0.B_ID IS NULL
 * 
 * UNION
 * 
 * SELECT THIS.A_ID,'org.jpox.samples.B' as JPOXMETADATA, THIS.COLUMN1, THIS.COLUMN2 
 * FROM A THIS
 *   INNER JOIN B 'ELEMENT' ON 'ELEMENT'.B_ID = THIS.A_ID
 * </PRE>
 * <H3>Extent</H3>
 * Here we use the above arrangement, with the "AssociationEnd" being the element
 * table and the mapping being the ID column mapping.
 * <H3>"Normal" Relationship</H3>
 * Here the join table is the candidate table, and the "AssociationEnd" would be 
 * the element table, and the mapping in the join table to join to this element table 
 * PK column.
 * <H3>"Inverse" Relationship</H3>
 * Here we use the same arrangement as the Extent case. The "AssociationEnd" is the
 * element table, and the mapping being the ID column mapping.
 * <H3>Map "Key" table</H3>
 * Here we have the key table as the candidate, and we join to the value table. So
 * the "AssociationEnd" is the value table, with the mapping being the mapping of the
 * key column in the value table.
 * <H3>Primitive key/value in SCO table</H3>
 * This also supports retrieval of keys or values of a map that are stored as a column
 * in a different table.
 * TODO Add full documentation for all combinations supported.
 *
 * @version $Revision: 1.10 $ 
 */
public class IteratorStatement
{
    /** Localisation facility. */
    private static final Localiser LOCALISER = Localiser.getInstance("org.jpox.store.query.Localisation");

    /** The type of object the iterator is to return */
    final Class candidateType;

    public static interface AssociationEnd
    {
        /** The element type. */
        Class getType();
        
        /** Mapping in the source table to get to join to the target. */
        JavaTypeMapping getMapping();

        /** The join table. Can be the same as the target table */
        DatastoreContainerObject getDatastoreContainerObject();

        /** Whether to use a join. Only used by map key tables. */
        boolean useJoin();
    }

    DatastoreClass candidateTable;
    StoreManager storeMgr;
    DatastoreAdapter dba;
    boolean includeSubClasses;

    /**
     * is the element embedded in the container table
     */
    boolean isEmbeddedElement = false;

    /**
     * the source association end
     */
    AssociationEnd source;

    /** Whether to include the JPOXMETADATA tag in the query to distinguish classes. */
    Boolean withMetadata = null;

    /**
     * Constructor. 
     * @param source the source for the association
     * @param candidateType the candidate is who are are looking to
     * @param storeMgr the store manager
     * @param dba the database adapter
     * @param includeSubClasses if the subclasses of the candidate should be
     *                          included in the result
     **/
    public IteratorStatement(
            AssociationEnd source,
            Class candidateType,
            StoreManager storeMgr,
            DatastoreAdapter dba,
            boolean includeSubClasses)
    {
        this.source = source;
        this.candidateType = candidateType;
        this.storeMgr = storeMgr;
        this.dba = dba;
        this.includeSubClasses = includeSubClasses;
        initialize();
    }

    /**
     * Constructor. 
     * @param source the source for the association
     * @param candidateType the candidate is who we are looking to
     * @param storeMgr the store manager
     * @param dba the database adapter
     * @param includeSubClasses if the subclasses of the candidate should be
     *                          included in the result
     * @param withMetadata if has subclasses, include the metadata in the query
     **/
    public IteratorStatement(
            AssociationEnd source,
            Class candidateType,
            StoreManager storeMgr,
            DatastoreAdapter dba,
            boolean includeSubClasses,
            boolean withMetadata)
    {
        this.source = source;
        this.candidateType = candidateType;
        this.storeMgr = storeMgr;
        this.dba = dba;
        this.includeSubClasses = includeSubClasses;
        this.withMetadata = (withMetadata ? Boolean.TRUE : Boolean.FALSE);
        initialize();
    }
    
    /**
     * Initialisation method.
     **/
    private void initialize()
    {
        // check if the element target is assignable from the candidate
        if (!source.getType().isAssignableFrom(candidateType))
        {
            throw new IncompatibleQueryElementTypeException(source.getType().getName(), candidateType.getName());
        }

        // if the candidate is a PersistenceCapable instance, it has a
        // ClassTable, otherwise the candidate type will be embedded.
        if (TypeManager.getTypeManager().isSupportedType(candidateType.getName()))
        {
            // non FCO elements are embedded
            isEmbeddedElement = true;
        }
        else
        {
            candidateTable = storeMgr.getDatastoreClass(candidateType.getName());
        }
    }

    /**
     * Method to generate a query for identifying the elements of this set.
     * @return The QueryStatement
     */
    public QueryStatement getQueryStatement()
    {
        if (isEmbeddedElement)
        {
            return getQueryForEmbeddedElement();
        }
        else
        {
            return getQueryStatementForFCO();
        }
    }

    /**
     * Method to generate a query when the (target) element is embedded in a table.
     * @return The QueryStatement
     */
    private QueryStatement getQueryForEmbeddedElement()
    {
        QueryStatement qs = ((DatabaseAdapter)dba).newQueryStatement(source.getDatastoreContainerObject());
        qs.select(source.getMapping());
        return qs;
    }

    /**
     * Method to generate a query when the (target) element has its own table.
     * @return The QueryStatement
     */
    public QueryStatement getQueryStatementForFCO()
    {
        // creates the main query
        QueryStatement stmt;
        stmt = getQueryForElement(candidateType.getName(), storeMgr.getDatastoreClass(candidateType.getName()));

        // Give warning if table is incorrect for candidate and 
        // subclasses not included in query.
        if (!source.getDatastoreContainerObject().equals(candidateTable) &&
            !includeSubClasses)
        {
            JPOXLogger.RDBMS.warn(LOCALISER.msg("Extent.NoSubclassesWarning", this, candidateType));
        }

        if (includeSubClasses)
        {
            // Add a select for each subclass elements, and union the selects
            Iterator iterator = storeMgr.getSubClassesForClass(candidateType.getName(), true).iterator();
            while (iterator.hasNext())
            {
                String subCandidateType = (String) iterator.next();
                DatastoreClass subCandidateTable = storeMgr.getDatastoreClass(subCandidateType);
                if (subCandidateTable != null)
                {
                    // Add UNION to any subclasses that have their own table
                    // This omits those using "no-table". Those will have their own subclasses where they
                    // store their objects.
                    QueryStatement stmt_subclass = getQueryForElement(subCandidateType, subCandidateTable);
                    stmt.union(stmt_subclass);
                }
            }
        }
        return stmt;
    }

    /**
     * Utility method to generate a select statement to find objects of
     * subclasses of the element class in the Set.
     *  <p>
     * In simple terms this creates a SELECT like
     * <pre>
     * SELECT THIS.A_ID,'org.jpox.samples.A' as JPOXMETADATA, THIS.COLUMN1, THIS.COLUMN2 
     * FROM A THIS
     *   LEFT OUTER JOIN B SUBELEMENT0 ON SUBELEMENT0.B_ID = THIS.A_ID
     * WHERE SUBELEMENT0.B_ID IS NULL
     * </pre>
     * The actual query will vary if this subclass has its own subclasses. If
     * there are no subclasses of this class then the LEFT OUTER JOIN parts will
     * not be present, nor will the SUBELEMENT_? parts of the WHERE clause.
     * </p>
     * <p>
     * If the target element shares the same table as the source element then
     * no (left outer) join will be made and it will be a simple select.
     * </p>
     * <p>
     * If a discriminator column is present on the table, it will be used to
     * separate the classes in that table rather than using the left outer join
     * strategy
     * </p>
     * @param targetElementType The Class for the element
     * @param targetElementTable The table for this class (will not be null)
     * @return QueryStatement
     **/
    private QueryStatement getQueryForElement(String targetElementType, DatastoreClass targetElementTable)
    {
        QueryStatement stmt;

        if (!(source.getDatastoreContainerObject() instanceof SCOTable))
        {
            // A). When creating a QueryStatement for FCO objects
            // B). When creating a QueryStatement for SCO objects using inverse relationaship (no join table)
            stmt = ((DatabaseAdapter)dba).newQueryStatement(candidateTable);
            if (withMetadata == null || withMetadata.booleanValue())
            {
                selectMetadata(stmt, targetElementType);
            }
            
            // Select the discriminator if there is one
            JavaTypeMapping discriminatorMapping = source.getDatastoreContainerObject().getDiscriminatorMapping();
            DiscriminatorMetaData dismd = source.getDatastoreContainerObject().getDiscriminatorMetaData();
            if (discriminatorMapping != null)
            {
                // Default to the "class-name" discriminator strategy
                String discriminatorValue = targetElementType;
                if (dismd.getStrategy() == DiscriminatorStrategy.VALUE_MAP)
                {
                    // Get the MetaData for the target class since that holds the "value"
                    ClassMetaData targetCmd = storeMgr.getMetaDataManager().getMetaDataForClass(targetElementType,storeMgr.getClassLoaderResolver());
                    discriminatorValue = targetCmd.getInheritanceMetaData().getDiscriminatorMetaData().getValue();
                }
                ScalarExpression discrExpr = discriminatorMapping.newScalarExpression(stmt, stmt.getDefaultTableExpression());
                ScalarExpression discrVal = discriminatorMapping.newLiteral(stmt, discriminatorValue);
                stmt.andCondition(discrExpr.eq(discrVal));
            }

            // in case of the elementType is a subClass of the element for the candidateTable
            // if the root (candidate) element type is not the same as the current target element type
            // joins the root to the current element
            if ((!targetElementTable.toString().equals(source.getDatastoreContainerObject().toString()) &&
                 !((ClassTable) candidateTable).getType().equals(targetElementType)) || 
                source.useJoin())
            {
                // Add inner joins to all classes above up to elementType and across to the join table
                if (source.useJoin())
                {
                    joinTargetToSourceElement(stmt, targetElementTable);
                }
                else
                {
                    joinSourceToTargetElement(stmt, targetElementTable);
                }
            }
            joinToExcludeTargetWhenSubElementsExists(stmt, ((ClassTable) candidateTable).getIDMapping(), targetElementType);
        }
        else
        {
            // When creating a QueryStatement for SCO objects using normal relationship (join table)
            stmt = ((DatabaseAdapter)dba).newQueryStatement(source.getDatastoreContainerObject());
            if (withMetadata == null || withMetadata.booleanValue())
            {
                selectMetadata(stmt, targetElementType);
            }

            // Add inner joins to all classes above up to elementType and across to the join table
            joinSourceToTargetElement(stmt, targetElementTable);
            joinToExcludeTargetWhenSubElementsExists(stmt, source.getMapping(), targetElementType);
        }
        return stmt;
    }

    /**
     * Joins the container table (source) to the element table (target).
     * This is used when the source mapping is from target to the source (normal
     * relationships where the target is SCO).
     * @param stmt The Query Statement
     * @param elementTargetTable the table that contains the elements
     **/
    private void joinSourceToTargetElement(QueryStatement stmt, DatastoreClass elementTargetTable)
    {
        ScalarExpression sourceExpr;
        ScalarExpression targetExpr;

        // get primary key columns
        sourceExpr = source.getMapping().newScalarExpression(stmt,stmt.getDefaultTableExpression());

        // table expression from the table identifier
        TableExpression teTargetElement = stmt.newTableExpression(elementTargetTable, new TableIdentifier(dba, "ELEMENT"));
        targetExpr = elementTargetTable.getIDMapping().newScalarExpression(stmt,teTargetElement);
        stmt.innerJoin(sourceExpr,targetExpr,teTargetElement);
    }

    /**
     * Joins the element table (target) to the container table (source).
     * This is used when the source mapping is from the source to the target.
     * @param stmt The Query Statement
     * @param elementTargetTable the table that contains the elements
     **/
    private void joinTargetToSourceElement(QueryStatement stmt, DatastoreClass elementTargetTable)
    {
        ScalarExpression sourceExpr;
        ScalarExpression targetExpr;

        TableExpression teSourceElement = stmt.newTableExpression(source.getDatastoreContainerObject(), new TableIdentifier(dba, "VALUE"));
        sourceExpr = source.getMapping().newScalarExpression(stmt, teSourceElement);
        targetExpr = elementTargetTable.getIDMapping().newScalarExpression(stmt, stmt.getDefaultTableExpression());

        stmt.innerJoin(sourceExpr,targetExpr,teSourceElement);
    }

    /**
     * Join to exclude targetElementType instances when exists sub elements
     * instances of targetElementType. This is made to avoid returning
     * targetElement when there are sub classes instances of targetElement
     * <pre>
     * ...
     * ELEMENT_TABLE TARGETELEMENT
     * LEFT OUTER JOIN SUB_TARGETELEMENT ON TARGETELEMENT.ELEMENT_ID = SUBCLASS_TARGETELEMENT.SUB_TARGETELEMENT_ID
     * WHERE SUB_TARGETELEMENT.SUB_TARGETELEMENT_ID IS NULL
     * ...
     * </pre>
     * <p>When the subelements are using the same table as the target element, no
     * exclusion is performed</p>
     *
     * @param stmt The Query Statement
     * @param targetElementQCL Query ColumnList for the target element for joining to
     * @param targetElementType The target element
     **/
    private void joinToExcludeTargetWhenSubElementsExists(
            QueryStatement stmt,
            JavaTypeMapping targetElementMapping,
            String targetElementType)
    {
        // Add LEFT OUTER JOIN to all direct subclasses of the targetRootElementQCL - if any
        Iterator iterTargetSubElementType = storeMgr.getSubClassesForClass(targetElementType, false).iterator();
        int subSequenceIdentifier = 0;
        while (iterTargetSubElementType.hasNext())
        {
            String targetSubElementType = (String) iterTargetSubElementType.next();

            QueryStatement targetQS;
            DatastoreClass cbtTargetSubElementType = storeMgr.getDatastoreClass(targetSubElementType);
            TableIdentifier tiTargetSubElementType;
            TableExpression teTargetSubElementType;
            ScalarExpression targetSubElementTypeExpr;

            DatastoreClass[] targetSubElementTypes = null;
            if (cbtTargetSubElementType == null)
            {
                ClassMetaData targetSubCmd = storeMgr.getMetaDataManager().getMetaDataForClass(targetSubElementType, storeMgr.getClassLoaderResolver());
                ClassMetaData[] targetSubCmds = storeMgr.getClassesManagingTableForClass(targetSubCmd);
                targetSubElementTypes = new DatastoreClass[targetSubCmds.length];
                for (int i=0;i<targetSubCmds.length;i++)
                {
                    targetSubElementTypes[i] = storeMgr.getDatastoreClass(targetSubCmds[i].getFullClassName());
                }
            }
            else
            {
                targetSubElementTypes = new DatastoreClass[1];
                targetSubElementTypes[0] = cbtTargetSubElementType;
            }

            for (int i=0;i<targetSubElementTypes.length;i++)
            {
                // Add leftOuterJoin to sub-target elements if they are not in the same table as the target
                if (!targetSubElementTypes[i].toString().equals(storeMgr.getDatastoreClass(targetElementType).toString()))
                {
                    tiTargetSubElementType = new TableIdentifier(dba, "SUBELEMENT" + (subSequenceIdentifier++));

                    // create a new statement
                    targetQS = new QueryStatement(targetSubElementTypes[i], tiTargetSubElementType);

                    // table expression from the table identifier
                    teTargetSubElementType = targetQS.newTableExpression(targetSubElementTypes[i], tiTargetSubElementType);

                    JavaTypeMapping targetMapping = targetSubElementTypes[i].getIDMapping();
                    targetSubElementTypeExpr = targetMapping.newScalarExpression(stmt,teTargetSubElementType);
                    ScalarExpression targetElementExpr = targetElementMapping.newScalarExpression(stmt,stmt.getDefaultTableExpression());
                    stmt.leftOuterJoin(targetElementExpr,targetSubElementTypeExpr, teTargetSubElementType);

                    ScalarExpression seTargetSubElementType = targetMapping.newScalarExpression(targetQS, teTargetSubElementType);
                    stmt.andCondition(new NullLiteral(stmt).eq(seTargetSubElementType));
                }
            }
        }
    }

    /**
     * Adds an expression to the statement. The expression is a metadata that
     * identifies an element in a hierarchy level of classes. This will appear
     * in the SELECT as "'org.jpox.samples.MyClass     ' AS JPOXMETADATA".
     * 
     * @param stmt Query Statement
     * @param pcClass The Persistence Capable class
     **/
    private void selectMetadata(QueryStatement stmt, String pcClass)
    {
        // If we are looking for subclasses, we need to add the class to the
        // select, and unfortunately some RDBMS set an odd limit on the first
        // row in the output set column for the length of the name column.
        // This fix looks through the classes concerned in our query and takes
        // the longest name and pads all other names to this length. This is
        // used in the "'package.classname' AS JPOXMETADATA" parts of the query.
        int maxClassNameLen = candidateType.getName().length();
        boolean hasSubClasses = false;
        Iterator iterator = storeMgr.getSubClassesForClass(candidateType.getName(), true).iterator();
        if (iterator.hasNext())
        {
            hasSubClasses = true;
            while (iterator.hasNext())
            {
                String nextClass = (String) iterator.next();
                if (nextClass.length() > maxClassNameLen)
                {
                    maxClassNameLen = nextClass.length();
                }
            }
        }

        // Include metadata if explicitly told to include it, or if subclasses present
        if ((withMetadata == null && hasSubClasses) ||
            (withMetadata != null && withMetadata.booleanValue()))
        {
            String classname = pcClass;
            if (maxClassNameLen > classname.length())
            {
                classname = StringUtils.leftAlignedPaddedString(pcClass, maxClassNameLen);
            }
            JavaTypeMapping m = dba.getMapping(String.class);
            stmt.selectScalarExpression(m.newLiteral(stmt,classname).as("JPOXMETADATA"));
        }
    }
}