/**********************************************************************
Copyright (c) 2003 Mike Martin (TJDO) and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Erik Bengtson - removed unused import
2003 Andy Jefferson - commented and localised
2004 Andy Jefferson - converted to use Logger
2004 Andy Jefferson - added support for uniqueness of query
2004 Andy Jefferson - added Single-String query support
2005 Andy Jefferson - added timeout support
    ...
**********************************************************************/
package org.jpox.store.query;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.Map;
import java.util.StringTokenizer;

import javax.jdo.Extent;
import javax.jdo.JDODataStoreException;
import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOUnsupportedOptionException;
import javax.jdo.JDOUserException;

import org.jpox.PersistenceManager;
import org.jpox.Transaction;
import org.jpox.store.expression.AggregateExpression;
import org.jpox.store.expression.BooleanExpression;
import org.jpox.store.expression.JDOHelperExpression;
import org.jpox.store.expression.MathExpression;
import org.jpox.store.expression.NullLiteral;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.UnboundVariable;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.util.JPOXLogger;
import org.jpox.util.StringUtils;

/**
 * A JDO query that uses the JQOQL language.
 * The query can be specified via method calls, or via a single-string form.
 *
 * @see Query
 * @version $Revision: 1.13 $
 */
public class JDOQLQuery extends Query
{
    /** Keywords used in single-string JDOQL. Uppercase variants specified here, but we allow the lowercase form */
    protected static final String[] SINGLE_STRING_KEYWORDS = {
            "SELECT", "UNIQUE", "INTO", "FROM", "EXCLUDE", "SUBCLASSES", "WHERE",
			"VARIABLES", "PARAMETERS", "IMPORTS", "GROUP", "ORDER", "BY", "RANGE", "TO"
			};
    
    protected transient Queryable candidates = null;
    private transient QueryStatement queryStmt = null;
    private transient Query.ResultObjectFactory rof = null;
    
    /**
     * Results from a query that are not instances of the candidate class. The
     * results might be fields of persistent instances, instances of classes
     * other than the candidate class, or aggregates of fields.
     */
    protected String result=null;    

    /**
     * Constructs a new query instance that uses the given persistence manager.
     * @param pm the associated persistence manager for this query.
     */
    public JDOQLQuery(PersistenceManager pm)
    {
        super(pm);

        candidateClass = null;
        filter = null;
        imports = null;
        variables = null;
        parameters = null;
        ordering = null;
        result = null;
        resultClass = null;
    }

    /**
     * Constructs a new query instance having the same criteria as the given query.
     * @param pm The PersistenceManager
     * @param q The query from which to copy criteria.
     */
    public JDOQLQuery(PersistenceManager pm, JDOQLQuery q)
    {
        super(pm);

        if (q == null)
        {
            candidateClass = null;
            filter = null;
            imports = null;
            variables = null;
            parameters = null;
            ordering = null;
            result = null;
            resultClass = null;
        }
        else
        {
            candidateClass = q.candidateClass;
            filter = q.filter;
            imports = q.imports;
            variables = q.variables;
            parameters = q.parameters;
            ordering = q.ordering;
            result = q.result;
            resultClass = q.resultClass;
            
            // JDO 1.0.1, Sect 12.6.3 : IgnoreCache is preserved when using other query baseline
            ignoreCache = q.ignoreCache;
        }
    }
    
    /**
     * Constructor for a JDOQL query where the query is specified using the "Single-String" format.
     * @param pm The persistence manager
     * @param query The query string
     */
    public JDOQLQuery(PersistenceManager pm, String query)
    {
        super(pm);
        
        new SingleStringParser(query).parse();
    }

    /**
     * Inner class to parse the single string query into the constituent variables.
     * @version $Revision: 1.13 $
     */
    class SingleStringParser
    {
        private String query;
        private StringTokenizer tokenizer;
        
        /**
         * Constructor for the Single-String parser.
         * @param query The query
         */
        public SingleStringParser(String query)
        {
            JPOXLogger.JDO_QUERY.debug("JDOQL Single-String with \"" + query + "\"");
            this.query = query;
        }
        
        /**
         * Method to parse the Single-String query
         */
        public void parse()
        {
            tokenizer = new StringTokenizer(query);
            if (StringUtils.isWhitespace(query))
            {
                // No query specified!
                throw new JDOUserException(LOCALISER.msg("JDOQL.SingleString.NoSelectStart"));
            }

            String firstToken = tokenizer.nextToken();
            if (!firstToken.equals("SELECT") && !firstToken.equals("select"))
            {
                throw new JDOUserException(LOCALISER.msg("JDOQL.SingleString.NoSelectStart"));
            }

            String latest_content="";
            String latest_keyword="SELECT";
            while (tokenizer.hasMoreTokens())
            {
                String token = tokenizer.nextToken();
                if (isKeyword(token))
                {
                    // Catch all double word keywords
                    if (latest_content.length() == 0 && latest_keyword.equals("ORDER") && 
                        (token.equals("BY") || token.equals("by")))
                    {
                        latest_keyword = "ORDER BY";
                    }
                    else if (latest_content.length() == 0 && latest_keyword.equals("GROUP") && 
                             (token.equals("BY") || token.equals("by")))
                    {
                        latest_keyword = "GROUP BY";
                    }
                    else if (latest_content.length() == 0 && latest_keyword.equals("EXCLUDE") &&
                             (token.equals("SUBCLASSES") || token.equals("subclasses")))
                    {
                        latest_keyword = "EXCLUDE SUBCLASSES";
                    }
                    else if (latest_keyword.length() > 0) // Don't process if no previous keyword
                    {
                        // Process the previous keyword
                        processKeywordWithContent(latest_keyword, latest_content);
                        
                        // Start on the next keyword
                        latest_keyword = token.toUpperCase();
                        latest_content = "";
                    }
                }
                else
                {
                    // Add this to the content
                    if (latest_content.length() > 0)
                    {
                        latest_content += " ";
                    }
                    latest_content += token;
                }
            }
            
            // Process any trailing keyword
            if (latest_keyword.length() > 0)
            {
                processKeywordWithContent(latest_keyword,latest_content);
            }
            
        }
        
        /**
         * Method to process the keyword and its optional content. In the query the keywords often have
         * content values following them that represent the constituent parts of the query. This takes the keyword
         * and sets the constituent part accordingly. The keywords pass in are assumed in uppercase here.
         * @param keyword The keyword (e.g UNIQUE)
         * @param content The optional content for that keyword
         */
        private void processKeywordWithContent(String keyword, String content)
        {
            if (keyword.equals("SELECT"))
            {
                /*if (content.length() > 0)
                {
                    throw new JDOUserException("<result> not yet supported by JDOQL Single-String query");
                    // TODO Process "result" since that will come through here
                }*/
            }
            else if (keyword.equals("UNIQUE"))
            {
                setUnique(true);
                
                /*if (content.length() > 0)
                {
                    throw new JDOUserException("<result> not yet supported by JDOQL Single-String query");
                    // TODO Process "result" since that will come through here
                }*/
            }
            else if (keyword.equals("INTO"))
            {
                if (content.length() == 0)
                {
                    throw new JDOUserException(LOCALISER.msg("JDOQL.SingleString.KeywordWithoutValue", keyword, "<result class>"));
                }
                
                Class cls=pm.getClassLoaderResolver().classForName(content, true);
                if (cls == null)
                {
                    throw new JDOUserException(LOCALISER.msg("JDOQL.SingleString.ClassNotFound", content));
                }
                else
                {
                    setResultClass(cls);
                }
            }
            else if (keyword.equals("FROM"))
            {
                if (content.length() == 0)
                {
                    throw new JDOUserException(LOCALISER.msg("JDOQL.SingleString.KeywordWithoutValue", keyword, "<candidate class>"));
                }
                
                Class cls=pm.getClassLoaderResolver().classForName(content, true);
                if (cls == null)
                {
                    throw new JDOUserException(LOCALISER.msg("JDOQL.SingleString.ClassNotFound", content));
                }
                else
                {
                    setClass(cls);
                }
            }
            else if (keyword.equals("EXCLUDE SUBCLASSES"))
            {
                if (content.length() > 0)
                {
                    throw new JDOUserException(LOCALISER.msg("JDOQL.SingleString.KeywordWithIllegalValue", keyword, content));
                }
                setSubclasses(false);
            }
            else if (keyword.equals("WHERE"))
            {
                if (content.length() == 0)
                {
                    throw new JDOUserException(LOCALISER.msg("JDOQL.SingleString.KeywordWithoutValue", keyword, "<filter>"));
                }
                setFilter(content);
            }
            else if (keyword.equals("VARIABLES"))
            {
                if (content.length() == 0)
                {
                    throw new JDOUserException(LOCALISER.msg("JDOQL.SingleString.KeywordWithoutValue", keyword, "<variable declarations>"));
                }
                declareVariables(content);
            }
            else if (keyword.equals("PARAMETERS"))
            {
                if (content.length() == 0)
                {
                    throw new JDOUserException(LOCALISER.msg("JDOQL.SingleString.KeywordWithoutValue", keyword, "<parameter declarations>"));
                }
                declareParameters(content);
            }
            else if (keyword.equals("IMPORTS"))
            {
                if (content.length() == 0)
                {
                    throw new JDOUserException(LOCALISER.msg("JDOQL.SingleString.KeywordWithoutValue", keyword, "<import declarations>"));
                }
                declareImports(content);
            }
            else if (keyword.equals("GROUP BY"))
            {
/*                if (content.length() == 0)
                {
                    throw new JDOUserException(LOCALISER.msg("JDOQL.SingleString.KeywordWithoutValue", keyword, "<grouping>"));
                }
                // TODO Set the grouping from the single-query
                throw  new JDOUserException("GROUP BY not yet supported in JDOQL Single-Query");*/
                setGrouping(content);
            }
            else if (keyword.equals("ORDER BY"))
            {
                if (content.length() == 0)
                {
                    throw new JDOUserException(LOCALISER.msg("JDOQL.SingleString.KeywordWithoutValue", keyword, "<ordering>"));
                }
                setOrdering(content);
            }
            else if (keyword.equals("RANGE"))
            {
                if (content.length() == 0)
                {
                    throw new JDOUserException(LOCALISER.msg("JDOQL.SingleString.KeywordWithoutValue", keyword, "<start>"));
                }
                fromInclNo = (new Integer(content)).intValue();
            }
            else if (keyword.equals("TO"))
            {
                if (content.length() == 0)
                {
                    throw new JDOUserException(LOCALISER.msg("JDOQL.SingleString.KeywordWithoutValue", keyword, "<end>"));
                }
                toExclNo = (new Integer(content)).intValue();
            }
            else if (keyword.length() == 0 && content.length() > 0)
            {
                // Should be nothing ending up down here
            }
        }

        /**
         * Method to detect whether this token is a keyword for JDOQL Single-String
         * @param token The token
         * @return Whether it is a keyword
         */
        private boolean isKeyword(String token)
        {
            // Compare the passed token against the provided keyword list, or their lowercase form
            for (int i=0;i<SINGLE_STRING_KEYWORDS.length;i++)
            {
                if (token.equals(SINGLE_STRING_KEYWORDS[i]) || token.equals(SINGLE_STRING_KEYWORDS[i].toLowerCase()))
                {
                    return true;
                }
            }
            return false;
        }
    }

    /**
     * Equality operator for JDOQL.
     * @see java.lang.Object#equals(java.lang.Object)
     */
    public boolean equals(Object obj)
    {
        if (obj == this)
        {
            return true;
        }

        if (!(obj instanceof JDOQLQuery) || !super.equals(obj))
        {
            return false;
        }

        return true;
    }

    /**
     * Set the candidate Extent to query.
     *
     * @param pcs the Candidate Extent.
     * @see javax.jdo.Query#setCandidates(javax.jdo.Extent)
     */
    public void setCandidates(Extent pcs)
    {
        if (pcs == null)
        {
            JPOXLogger.JDO_QUERY.warn(LOCALISER.msg("Candidates.ExtentCantBeNull"));
            return;
        }
        if (!(pcs instanceof Queryable))
        {
            throw new JDOUnsupportedOptionException(LOCALISER.msg("JDOQL.ExtentNotQueryableError",pcs.getClass().getName()));
        }

        discardCompiled();
        assertIsUnmodifiable();

        subclasses = pcs.hasSubclasses();
        candidates = (Queryable)pcs;
    }

    /**
     * Set the candidate Collection to query.
     *
     * @param pcs the Candidate collection.
     * @see javax.jdo.Query#setCandidates(java.util.Collection)
     */
    public void setCandidates(Collection pcs)
    {
        discardCompiled();
        assertIsUnmodifiable();

        if (pcs == null)
        {
            JPOXLogger.JDO_QUERY.warn(LOCALISER.msg("Candidates.CollectionCantBeNull"));
            return;
        }
        if (!(pcs instanceof Queryable))
        {
            candidates = new CollectionCandidates(pm,candidateClass,pcs);
        }
        else
        {
            candidates = (Queryable)pcs;
        }
    }

    /**
     * Set the result for the results. The application might want to get results
     * from a query that are not instances of the candidate class. The results
     * might be fields of persistent instances, instances of classes other than
     * the candidate class, or aggregates of fields.
     * @param result The result parameter consists of the optional keyword
     * distinct followed by a commaseparated list of named result expressions or
     * a result class specification.
     * @see javax.jdo.Query#setResult
     * @since 1.1
     */
    public void setResult(String result)
    {
        discardCompiled();
        assertIsUnmodifiable();

        this.result = result;
    }
    
    /**
     * Mutator for whether to include subclasses.
     * @see org.jpox.store.query.Query#setSubclasses(boolean)
     */
    public void setSubclasses(boolean subclasses)
    {
        super.setSubclasses(subclasses);
        assertIsUnmodifiable();
        
        if (candidateClass != null)
        {
            Extent ext = pm.getExtent(candidateClass,subclasses);
            setCandidates(ext);
        }
    }
    
    /**
     * Method to take the defined parameters for the query and form a single string.
     * This is used to print out the query for logging.
     * @return The single string
     */
    private String getSingleStringQuery()
    {
        StringBuffer str=new StringBuffer("SELECT ");
        
        if (unique)
        {
            str.append("UNIQUE ");
        }
        if (result != null)
        {
            str.append(result + " ");
        }
        if (resultClass != null)
        {
            str.append("INTO " + resultClass.getName() + " ");
        }
        if (candidateClass != null)
        {
            str.append("FROM " + candidateClass.getName() + " ");
        }
        if (!subclasses)
        {
            str.append("EXCLUDE SUBCLASSES ");
        }
        if (filter != null)
        {
            str.append("WHERE " + filter + " ");
        }
        if (parameters != null || variables != null || imports != null)
        {
            if (variables != null)
            {
                str.append("VARIABLES " + variables + " ");
            }
            if (parameters != null)
            {
                str.append("PARAMETERS " + parameters + " ");
            }
            if (imports != null)
            {
                str.append("IMPORTS " + imports + " ");
            }
        }
        if (grouping != null)
        {
            str.append("GROUP BY " + grouping + " ");
        }
        if (ordering != null)
        {
            str.append("ORDER BY " + ordering + " ");
        }
        if (fromInclNo > 0 || toExclNo != Integer.MAX_VALUE)
        {
            str.append("RANGE " + fromInclNo + " ");
            if (toExclNo != Integer.MAX_VALUE)
            {
                str.append("TO " + toExclNo + " ");
            }
        }
        return str.toString();
    }
    
    /**
     * Verify the elements of the query and provide a hint to the query to
     * prepare and optimize an execution plan.
     *
     * @see javax.jdo.Query#compile
     */
    public void compile()
    {
        super.compile();

        // TODO better compile filters... this is done here just to verify the filter
        if (isPreCompile)
        {
            new Compiler(null).preCompile();
        }
        isCompiled = true;
    }
    
    /**
     * Method to compile the query.
     * @param parameters The input parameters to the query.
     */
    private void compile(Map parameters)
    {
        discardCompiled();
        super.compile();

        try
        {
            if (JPOXLogger.JDO_QUERY.isDebugEnabled())
            {
                JPOXLogger.JDO_QUERY.debug(LOCALISER.msg("JDOQL.Statement.Summary", getSingleStringQuery()));
            }

            long startTime = System.currentTimeMillis();
            Compiler c = new Compiler(parameters);
            queryStmt = c.compileQueryStatement();
            if (result != null)
            {
                if (result.toLowerCase().startsWith("distinct"))
                {
                    ((ResultExpressionsQueryable)candidates).setDistinctResults(queryStmt,true);
                    result = result.substring(8);
                }
                ScalarExpression[] exprs = c.compileResultExpressions();

                ((ResultExpressionsQueryable)candidates).setResultExpressions(exprs);
            }
            rof = candidates.newResultObjectFactory(queryStmt, getIgnoreCache(), resultClass);

            if (JPOXLogger.JDO_QUERY.isDebugEnabled())
            {
                JPOXLogger.JDO_QUERY.debug(LOCALISER.msg("JDOQL.Statement.CompileTime",(System.currentTimeMillis() - startTime)));
            }

            isCompiled = true;
        }
        finally
        {
            if (!isCompiled)
            {
                discardCompiled();
            }
        }
    }
    
    /**
     * Method to discard our current compiled query due to changes.
     * @see org.jpox.store.query.Query#discardCompiled()
     */
    protected void discardCompiled()
    {
        super.discardCompiled();

        queryStmt = null;
        rof = null;
    }
    
    /**
     * Execute the query and return the filtered QueryResult.
     * @param parameters the Map containing all of the parameters.
     * @return the filtered QueryResult.
     */
    protected Collection performExecute(Map parameters)
    {
        isPreCompile = false;
        compile(parameters);

        if (candidates.isEmpty())
        {
            // if the candidates are empty, we don't have to go to the database
            // to find out that we have no elements to return
            return new ArrayList();
        }

        isPreCompile = true;

        QueryResult qr = null;
        try
        {
            Connection conn = pm.getConnection(false);

            try
            {
                Transaction tx = (Transaction) pm.currentTransaction();
                PreparedStatement ps = queryStmt.toStatementText(tx.useUpdateLockOnFetch()).prepareStatement(pm, conn);

                try
                {
                    // Apply any user-specified timeout
                    if (timeout > 0)
                    {
                        ps.setQueryTimeout(timeout);
                    }

                    // Apply any fetch size
                    ps.setFetchSize(pm.getPMFContext().getPmfConfiguration().getQueryFetchSize());

                    // Apply any fetch direction
                    String fetchDir = pm.getPMFContext().getPmfConfiguration().getQueryFetchDirection();
                    if (fetchDir.equals("reverse"))
                    {
                        ps.setFetchDirection(ResultSet.FETCH_REVERSE);
                    }
                    else if (fetchDir.equals("unknown"))
                    {
                        ps.setFetchDirection(ResultSet.FETCH_UNKNOWN);
                    }

                    // Add a limit on the number of rows to include the maximum we may need
                    if (toExclNo != 0 && toExclNo != Integer.MAX_VALUE)
                    {
                        ps.setMaxRows(toExclNo);
                    }

                    ResultSet rs = executeQuery(queryStmt.toString(tx.useUpdateLockOnFetch()), ps);
                    try
                    {
                        qr = new QueryResult(queryStmt, this, rof, rs);
                    }
                    finally
                    {
                        if (qr == null)
                        {
                            rs.close();
                        }
                    }
                }
                finally
                {
                    if (qr == null)
                    {
                        ps.close();
                    }
                }
            }
            finally
            {
                pm.releaseConnection(conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("JDOQL.ExecutionError",queryStmt.toStatementText(((Transaction)pm.currentTransaction()).useUpdateLockOnFetch())), e);
        }

        queryResults.add(qr);

        return qr;
    }

    /**
     * Execute the query to delete persistent objects.
     * @param parameters the Map containing all of the parameters.
     * @return the filtered QueryResult of the deleted objects.
     */
    protected QueryResult performDeletePersistentAll(Map parameters)
    {
        throw new JDOUserException(LOCALISER.msg("Query.DeletePersistent.NotYetSupported"));

        // TODO Implement this
    }

    /**
     * Inner class to compile the query.
     **/
    public class Compiler
    {
        private final Map parameters;
        private final HashMap expressionsByVariableName = new HashMap();
        private QueryStatement qs = null;
        private Parser p = null;

        public Compiler(Map parameters)
        {
            if (parameters != null &&
                parameters.size() != parameterNames.size())
            {
                throw new JDOUserException(LOCALISER.msg("JDOQL.IncorrectNumberOfParametersError","" + parameterNames.size(),"" + parameters.size()));
            }

            this.parameters = parameters;
        }

        public void bindVariable(String name, ScalarExpression expr)
        {
            ScalarExpression previousExpr = (ScalarExpression)expressionsByVariableName.put(name, expr);
            if (previousExpr != null)
            {
                throw new JDOFatalInternalException(LOCALISER.msg("JDOQL.VariableBindError",name,expr,previousExpr));
            }
        }

        /**
         * Method to provide pre-compilation checks to catch errors.
         */
        public void preCompile()
        {
            if (candidates == null)
            {
                if (candidateClass == null)
                {
                    throw new JDOUserException(LOCALISER.msg("JDOQL.NoCandidateClassError"));
                }
                candidates = (Queryable)pm.getExtent(candidateClass, true);
            }

            qs = candidates.newQueryStatement(candidateClass);

            if (filter != null && filter.length() > 0)
            {
                p = new Parser(filter, parsedImports);
                try
                {
                    ScalarExpression expr = compileExpression();
                    if (!p.parseEOS())
                    {
                        throw new ExpressionSyntaxException(LOCALISER.msg("JDOQL.InvalidExpressionError"));
                    }
    
                    if (!(expr instanceof BooleanExpression))
                    {
                        throw new JDOUserException(LOCALISER.msg("JDOQL.FilterExpressionNotBooleanError",filter));               
                    }
                }
                catch (NullPointerException e)
                {
                    // TODO : Analyse what to do here
                }
            }           
        }

        /**
         * Method to compile the query.
         */
        public QueryStatement compileQueryStatement()
        {
            if (result != null)
            {
                if (candidates != null && candidates instanceof CollectionCandidates)
                {
                    candidates = new ResultExpressionsQueryable(pm, candidateClass, 
                        ((CollectionCandidates)candidates).getUserCandidates(), subclasses);
                }
                else
                {
                    candidates = new ResultExpressionsQueryable(pm, candidateClass, subclasses);
                }
            }
            if (candidates == null)
            {
                if (candidateClass == null)
                {
                    throw new JDOUserException(LOCALISER.msg("JDOQL.NoCandidateClassError"));
                }
                candidates = (Queryable)pm.getExtent(candidateClass, true);
            }

            qs = candidates.newQueryStatement(candidateClass);

            if (filter != null && filter.length() > 0)
            {
                p = new Parser(filter, parsedImports);
                ScalarExpression expr = compileExpression();

                if (!p.parseEOS())
                {
                    throw new ExpressionSyntaxException(LOCALISER.msg("JDOQL.InvalidExpressionError"));
                }

                if (!(expr instanceof BooleanExpression))
                {
                    throw new JDOUserException(LOCALISER.msg("JDOQL.FilterExpressionNotBooleanError",filter));               
                }

                qs.andCondition((BooleanExpression)expr,true);
            }

            if (ordering != null && ordering.length() > 0)
            {
                StringTokenizer t1 = new StringTokenizer(ordering, ",");

                int n = t1.countTokens();
                ScalarExpression[] orderExprs = new ScalarExpression[n];
                boolean[] descending = new boolean[n];

                for (n = 0; t1.hasMoreTokens(); ++n)
                {
                    String orderExpression = t1.nextToken().trim(); 

                    // Allow all sensible forms of direction specification (not just JDOQL standard)
                    if (orderExpression.endsWith("ascending") )
                    {
                        descending[n] = false;
                        orderExpression = orderExpression.substring(0,orderExpression.length()-"ascending".length());
                    }
                    else if (orderExpression.endsWith("asc") )
                    {
                        descending[n] = false;
                        orderExpression = orderExpression.substring(0,orderExpression.length()-"asc".length());
                    }
                    else if (orderExpression.endsWith("ASCENDING") )
                    {
                        descending[n] = false;
                        orderExpression = orderExpression.substring(0,orderExpression.length()-"ASCENDING".length());
                    }
                    else if (orderExpression.endsWith("ASC") )
                    {
                        descending[n] = false;
                        orderExpression = orderExpression.substring(0,orderExpression.length()-"ASC".length());
                    }
                    else if (orderExpression.endsWith("descending") )
                    {
                        descending[n] = true;
                        orderExpression = orderExpression.substring(0,orderExpression.length()-"descending".length());
                    }
                    else if (orderExpression.endsWith("desc") )
                    {
                        descending[n] = true;
                        orderExpression = orderExpression.substring(0,orderExpression.length()-"desc".length());
                    }
                    else if (orderExpression.endsWith("DESCENDING") )
                    {
                        descending[n] = true;
                        orderExpression = orderExpression.substring(0,orderExpression.length()-"DESCENDING".length());
                    }
                    else if (orderExpression.endsWith("DESC") )
                    {
                        descending[n] = true;
                        orderExpression = orderExpression.substring(0,orderExpression.length()-"DESC".length());
                    }
                    else
                    {
                        throw new JDOUserException(LOCALISER.msg("JDOQL.InvalidOrderDirectionError",ordering));
                    }

                    p = new Parser(orderExpression, parsedImports);
                    orderExprs[n] = compileExpression();

                    if (!p.parseEOS())
                    {
                        throw new ExpressionSyntaxException(LOCALISER.msg("JDOQL.InvalidExpressionError"));
                    }
                }

                qs.setOrdering(orderExprs, descending);
            }

            // Add any range constraint
            if (fromInclNo > 0 || toExclNo != Integer.MAX_VALUE)
            {
                qs.setRangeConstraint(fromInclNo, toExclNo != Integer.MAX_VALUE ? toExclNo-fromInclNo : -1);
            }

            return qs;
        }

        public ScalarExpression[] compileResultExpressions()
        {
            if( result != null )
            {
                StringTokenizer t1 = new StringTokenizer(result, ",");
                
                int n = t1.countTokens();
                ScalarExpression[] resultExprs = new ScalarExpression[n];
                
                for (n = 0; t1.hasMoreTokens(); ++n)
                {
                    String resultExpression = t1.nextToken().trim(); 
                    
                    p = new Parser(resultExpression, parsedImports);
                    resultExprs[n] = compileExpression();
                    
                    if (!p.parseEOS())
                    {
                        throw new ExpressionSyntaxException(LOCALISER.msg("JDOQL.InvalidExpressionError"));
                    }
                }
                return resultExprs;
            }
            return null;
        }
                                
        private ScalarExpression compileExpression()
        {
            return compileConditionalOrExpression();
            /*
             * OR ("||")
             * AND ("&&")
             * Bitwise OR ("|")
             * Bitwise XOR ("^")
             * Bitwise AND ("&")
             * Equality ("==") ("=!")
             * Relational (">=") (">") ("<=") ("<")
             * Additive ("+") ("-")
             * Multiplicative ("*") ("/") ("%")
             * Unary ("+") ("-")
             * Unary ("~") ("!")
             * Cast
             */
        }

        /**
         * This method deals with the OR condition
         * A condition specifies a combination of one or more expressions and logical (Boolean) operators and returns a value of TRUE, FALSE, or unknown
         */
        private ScalarExpression compileConditionalOrExpression()
        {
            ScalarExpression expr = compileConditionalAndExpression();

            while (p.parseString("||"))
            {
                expr = expr.ior(compileConditionalAndExpression());
            }

            return expr;
        }

        /**
         * This method deals with the AND condition
         * A condition specifies a combination of one or more expressions and
         * logical (Boolean) operators and returns a value of TRUE, FALSE, or 
         * unknown
         */
        private ScalarExpression compileConditionalAndExpression()
        {
            ScalarExpression expr = compileInclusiveOrExpression();

            while (p.parseString("&&"))
            {
                expr = expr.and(compileInclusiveOrExpression());
            }

            return expr;
        }

        private ScalarExpression compileInclusiveOrExpression()
        {
            ScalarExpression expr = compileExclusiveOrExpression();

            while (p.parseChar('|', '|'))
            {
                expr = expr.ior(compileExclusiveOrExpression());
            }

            return expr;
        }

        private ScalarExpression compileExclusiveOrExpression()
        {
            ScalarExpression expr = compileAndExpression();

            while (p.parseChar('^'))
            {
                expr = expr.eor(compileExclusiveOrExpression());
            }

            return expr;
        }

        private ScalarExpression compileAndExpression()
        {
            ScalarExpression expr = compileEqualityExpression();

            while (p.parseChar('&', '&'))
            {
                expr = expr.and(compileEqualityExpression());
            }

            return expr;
        }

        private ScalarExpression compileEqualityExpression()
        {
            ScalarExpression expr = compileRelationalExpression();

            for (;;)
            {
                if (p.parseString("=="))
                {
                    expr = expr.eq(compileRelationalExpression());
                }
                else if (p.parseString("!="))
                {
                    expr = expr.noteq(compileRelationalExpression());
                }
                else
                {
                    break;
                }
            }

            return expr;
        }

        private ScalarExpression compileRelationalExpression()
        {
            ScalarExpression expr = compileAdditiveExpression();

            for (;;)
            {
                if (p.parseString("<="))
                {
                    expr = expr.lteq(compileAdditiveExpression());
                }
                else if (p.parseString(">="))
                {
                    expr = expr.gteq(compileAdditiveExpression());
                }
                else if (p.parseChar('<'))
                {
                    expr = expr.lt(compileAdditiveExpression());
                }
                else if (p.parseChar('>'))
                {
                    expr = expr.gt(compileAdditiveExpression());
                }
                else
                {
                    break;
                }
            }

            return expr;
        }

        private ScalarExpression compileAdditiveExpression()
        {
            ScalarExpression expr = compileMultiplicativeExpression();

            for (;;)
            {
                if (p.parseChar('+'))
                {
                    expr = expr.add(compileMultiplicativeExpression());
                }
                else if (p.parseChar('-'))
                {
                    expr = expr.sub(compileMultiplicativeExpression());
                }
                else
                {
                    break;
                }
            }

            return expr;
        }

        private ScalarExpression compileMultiplicativeExpression()
        {
            ScalarExpression expr = compileUnaryExpression();

            for (;;)
            {
                if (p.parseChar('*'))
                {
                    expr = expr.mul(compileUnaryExpression());
                }
                else if (p.parseChar('/'))
                {
                    expr = expr.div(compileUnaryExpression());
                }
                else if (p.parseChar('%'))
                {
                    expr = expr.mod(compileUnaryExpression());
                }
                else
                {
                    break;
                }
            }

            return expr;
        }

        private ScalarExpression compileUnaryExpression()
        {
            ScalarExpression expr;
            if (p.parseString("++"))
            {
                throw new JDOUserException("Unsupported operator '++'");
            }
            else if (p.parseString("--"))
            {
                throw new JDOUserException("Unsupported operator '--'");
            }               

            if (p.parseChar('+'))
            {
                expr = compileUnaryExpression();
            }
            else if (p.parseChar('-'))
            {
                expr = compileUnaryExpression().neg();
            }
            else
            {
                expr = compileUnaryExpressionNotPlusMinus();
            }
            return expr;
        }

        private ScalarExpression compileUnaryExpressionNotPlusMinus()
        {
            ScalarExpression expr;

            if (p.parseChar('~'))
            {
                expr = compileUnaryExpression().com();
            }
            else if (p.parseChar('!'))
            {
                expr = compileUnaryExpression().not();
            }
            else if ((expr = compileCastExpression()) == null)
            {
                expr = compilePrimary();
            }

            return expr;
        }

        private ScalarExpression compileCastExpression()
        {
            Class type;
            if ((type = p.parseCast(pm, getCandidateLoader())) == null)
            {
                return null;
            }

            return compileUnaryExpression().cast(type);
        }

        /**
         * this compiles a primary. First look for a literal (e.g. "text"), then
         * an identifier(e.g. variable) In the next step, call a function, if
         * executing a function, on the literal or the identifier found.
         * 
         * @return Scalar Expression for the primary compiled expression
         */
        private ScalarExpression compilePrimary()
        {
            // try to find a literal
            ScalarExpression expr = compileLiteral();

            // if expr == null, literal not found...
            if (expr == null)
            {
                //try to find an expression
                if (p.parseChar('('))
                {
                    expr = compileExpression();

                    if (!p.parseChar(')'))
                    {
                        throw new ExpressionSyntaxException("')' expected");
                    }
                    expr.encloseWithInParentheses();
                }
                else
                {
                    String methodId = p.parseMethod();
                    
                    
                    if( methodId == null )
                    {
	                    /*
	                     * We will have an identifier
	                     * an identifier can be an variable, parameter or an field in the candidate class
	                     */
	                    expr = compileIdentifier();
                    }
                    else
                    {
                        if (p.parseChar('('))
                        {
                            ArrayList args = new ArrayList();

                            if (!p.parseChar(')'))
                            {
                                do
                                {
                                    args.add(compileExpression());
                                } while (p.parseChar(','));
         
                                if (!p.parseChar(')'))
                                {
                                    throw new ExpressionSyntaxException("')' expected");
                                }
                            }
                            expr = new AggregateExpression(qs);
                            expr = expr.callMethod(methodId, args);
                        }
                    }
                }
            }

            /*
             * run function on literals or identifiers
             * e.g. "primary.runMethod(arg)"
             */
            while (p.parseChar('.'))
            {
                String id = p.parseIdentifier();
                if (id == null)
                {
                    throw new ExpressionSyntaxException("Identifier expected");
                }

                if (p.parseChar('('))
                {
                    ArrayList args = new ArrayList();

                    if (!p.parseChar(')'))
                    {
                        do
                        {
                            args.add(compileExpression());
                        } while (p.parseChar(','));
 
                        if (!p.parseChar(')'))
                        {
                            throw new ExpressionSyntaxException("')' expected");
                        }
                    }

                    expr = expr.callMethod(id, args);
                }
                else
                {
                    /*
                     * access the field from an identifier
                     */
                    expr = expr.accessField(id, false);
                }
            }
            return expr;
        }

        /**
         * A literal is one value of any type.
         * Supported literals are of types String, Floating Point, Integer,
         * Character, Boolean and null e.g. 'J', "String", 1, 1.8, true, false, null.
         * @return The compiled literal
         */
        private ScalarExpression compileLiteral()
        {
            Class  litType;
            Object litValue;

            String     sLiteral;
            BigDecimal fLiteral;
            BigInteger iLiteral;
            Boolean    bLiteral;

            boolean single_quote_next = p.nextIsSingleQuote();
            if ((sLiteral = p.parseStringLiteral()) != null)
            {
                // Both String and Character are allowed to use single-quotes
                // so we need to check if it was single-quoted and
                // use CharacterLiteral if length is 1.
                if (sLiteral.length() == 1 && single_quote_next)
                {
                    litType = Character.class;
                    litValue = new Character(sLiteral.charAt(0));
                }
                else
                {
                    litType  = String.class;
                    litValue = sLiteral;
                }
            }
            else if ((fLiteral = p.parseFloatingPointLiteral()) != null)
            {
                litType  = BigDecimal.class;
                litValue = fLiteral;
            }
            else if ((iLiteral = p.parseIntegerLiteral()) != null)
            {
                litType  = Long.class;
                litValue = new Long(iLiteral.longValue());
            }
            else if ((bLiteral = p.parseBooleanLiteral()) != null)
            {
                litType  = Boolean.class;
                litValue = bLiteral;
            }
            else if (p.parseNullLiteral())
            {
                return new NullLiteral(qs);
            }
            else
            {
                return null;
            }

            JavaTypeMapping m = dba.getMapping(litType, qs.getStoreManager());
            return m.newLiteral(qs, litValue);
        }

        /**
         * Checks whether the identifier is a Math (reserved word) expression
         * @param identifier the identitifer
         * @return
         */
        private boolean isMathExpression(String identifier)
        {
            return identifier.equals("Math");
        }
        
        /**
         * Checks whether the identifier is a JDOHelper (reserved word) expression
         * @param identifier the identitifer
         * @return
         */
        private boolean isJDOHelperExpression(String identifier)
        {
            return identifier.equals("JDOHelper");
        }        
        
        /**
         * An identifier always designates a reference to a single value.
         * A single value can be one collection, one field.
         * @return The compiled identifier
         */
        private ScalarExpression compileIdentifier()
        {
            ScalarExpression expr;

            String id = p.parseIdentifier();
	        if (id == null)
	        {
	            throw new ExpressionSyntaxException("Identifier expected");
	        }
            // if its a Math expression, like Math.abs(XXX)
            if( isMathExpression(id) )
            {
                expr = new MathExpression(qs);
            }
            // if its a JDOHelper expression, like JDOHelper.getObjectId(XXX)
            else if( isJDOHelperExpression(id) )
            {
                expr = new JDOHelperExpression(qs);
            }            
            /*
             * verifies if the query contain a parameter that is associated
             * to this identifier 
             */
            else if (parameterNames.contains(id))
            {
                // gets a mapping associted to the parameter type
                JavaTypeMapping m = dba.getMapping((Class)parameterTypesByName.get(id), qs.getStoreManager());

                if (isPreCompile)
                {
                    expr = m.newLiteral(qs, m.getSampleValue());
                }               
                else if (parameters == null ||
                    !parameters.containsKey(id))
                {
                    throw new JDOUserException(LOCALISER.msg("JDOQL.RequiredParameterNotProvidedError",id));
                }
                else
                {
                    Object parameterValue = parameters.get(id);

                    if (parameterValue == null)
                    {
                        expr = new NullLiteral(qs);
                    }
                    else
                    {
                        expr = m.newLiteral(qs, parameterValue);
                    }
                }
            }
            else if (variableNames.contains(id))
            {
                    expr = (ScalarExpression)expressionsByVariableName.get(id);
                    if (expr == null)
                    {
                        expr = new UnboundVariable(qs, id, (Class)variableTypesByName.get(id), this);
                    }
            }
            else
            {
                expr = qs.getDefaultTableExpression().newFieldExpression(id);
            }

            return expr;
        }

        /**
         * Inner class representing an exception in the syntax handling.
         **/
        private class ExpressionSyntaxException extends JDOUserException
        {
            public ExpressionSyntaxException(String msg)
            {
                super(msg + " at character " + (p.getIndex() + 1) + " in \"" + p.getInput() + '"');
            }
        }
    }

    /**
     * Stringifier method
     * @return Single-string form of this JDOQL query.
     */
    public String toString()
    {
        return getSingleStringQuery();
    }
}