/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - commented and localised
2005 Andy Jefferson - added timeout support
2005 Andy Jefferson - added support for use without candidate class
    ...
**********************************************************************/
package org.jpox.store.query;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOFatalUserException;
import javax.jdo.JDOUserException;

import org.jpox.PersistenceManager;
import org.jpox.exceptions.ClassNotPersistenceCapableException;
import org.jpox.metadata.ClassMetaData;
import org.jpox.metadata.FieldMetaData;
import org.jpox.metadata.FieldPersistenceModifier;
import org.jpox.metadata.IdentityType;
import org.jpox.store.Role;
import org.jpox.store.StatementExpressionIndex;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.MappingCallbacks;
import org.jpox.store.mapping.Mappings;
import org.jpox.store.exceptions.PersistentSuperclassNotAllowedException;
import org.jpox.store.rdbms.sqlidentifier.ColumnIdentifier;
import org.jpox.util.MacroString;

/**
 * A Query using SQL, and allowing specification of parameter names and imports,
 * known internally as "JPOXSQL", an extension to JDO.
 * @see Query
 * @version $Revision: 1.11 $
 */
public class JPOXSQLQuery extends BaseSQLQuery
{
    protected transient List parameterOccurrences = null;

    /**
     * Constructs a new query instance having the same criteria as the given
     * query.
     * @param pm The PersistenceManager
     * @param sql_text The JPOX SQL query string
     */
    public JPOXSQLQuery(PersistenceManager pm, String sql_text)
    {
        super(pm, sql_text);
    }

    protected void discardCompiled()
    {
        super.discardCompiled();
        parameterOccurrences = null;
    }

    /**
     * Equality operator
     * @param obj The object to compare against
     * @return Whether they are equal
     */
    public boolean equals(Object obj)
    {
        if (obj == this)
        {
            return true;
        }

        if (!(obj instanceof JPOXSQLQuery) || !super.equals(obj))
        {
            return false;
        }

        return sqlText.equals(((JPOXSQLQuery)obj).sqlText);
    }

    /**
     * Method to perform any necessary pre-processing on the users query statement
     * before we execute it.
     */
    protected void generateQueryStatement()
    {
        final ClassMetaData candidateCmd;
        if (candidateClass == null)
        {
            // We have no candidate so don't need to bother about aligning the return columns to our mappings.
            fieldColumnNames = new ArrayList(); // Empty
            candidateCmd = null;
        }
        else
        {
            candidateCmd = pm.getMetaDataManager().getMetaDataForClass(candidateClass);
            if (candidateCmd == null)
            {
                throw new ClassNotPersistenceCapableException(candidateClass.getName());
            }
            if (candidateCmd.getPersistenceCapableSuperclass() != null)
            {
                throw new PersistentSuperclassNotAllowedException(candidateClass.getName());
            }
            if (candidateCmd.isRequiresExtent())
            {
                throw new JDOUserException(LOCALISER.msg("JPOXSQL.InvalidCandidateClassExtentError",candidateClass.getName()));
            }
            if (candidateCmd.getIdentityType() != IdentityType.NONDURABLE)
            {
                throw new JDOUserException(LOCALISER.msg("JPOXSQL.InvalidCandidateClassNondurableError",candidateClass.getName()));
            }
            
            int fieldCount = candidateCmd.getNoOfManagedFields();
            int[] fn = new int[fieldCount];
            statementExpressionIndex = new StatementExpressionIndex[fieldCount];
            fieldColumnNames = new ArrayList(fieldCount);
            int n = 0;
            
            for (int fieldNumber = 0; fieldNumber < fieldCount; ++fieldNumber)
            {
                statementExpressionIndex[fieldNumber] = new StatementExpressionIndex();
                FieldMetaData fmd = candidateCmd.getManagedField(fieldNumber);
                String fieldName = fmd.getName();
                Class fieldType = fmd.getType();
                
                if (fmd.getPersistenceModifier() == FieldPersistenceModifier.PERSISTENT )
                {
                    JavaTypeMapping m = dba.getMapping(fieldType, getStoreManager());
                    if (!m.includeInFetchStatement() || m instanceof MappingCallbacks)
                    {
                        throw new JDOFatalUserException("Mapping " + m +
                            " not suitable for a JPOXSQL result column, field = " + fieldName);
                    }

                    statementExpressionIndex[fieldNumber].setMapping(m);
                    fn[n++] = fieldNumber;

                    String columnName = null;
                    if (fmd.getColumnMetaData() != null && fmd.getColumnMetaData().length > 0)
                    {
                        columnName = fmd.getColumnMetaData()[0].getName();
                    }
                    else
                    {
                        columnName = new ColumnIdentifier(dba, fieldName, fieldType, Role.NONE).getSQLIdentifier();
                    }
                    fieldColumnNames.add(columnName);
                }
                else if (fmd.getPersistenceModifier() != FieldPersistenceModifier.TRANSACTIONAL )
                {
                    throw new JDOFatalInternalException("Invalid persistence modifier on field " + fieldName);
                }
            }

            if (n == 0)
            {
                throw new JDOFatalUserException(
                    "View class has no persistent fields: " + candidateClass.getName());
            }

            fieldNumbers = new int[n];
            System.arraycopy(fn, 0, fieldNumbers, 0, n);
        }

        // Generate the actual JDBC SQL text by processing the embedded parameter/field macros
        // in the provided JPOXSQL text.
        parameterOccurrences = new ArrayList();
        MacroString ms = new MacroString(candidateClass != null ? candidateClass.getName() : null, 
                                         candidateClass != null ? imports : null, sqlText);
        jdbcSqlText = ms.substituteMacros(new MacroString.MacroHandler()
            {
            public void onIdentifierMacro(MacroString.IdentifierMacro im)
            {
                // Only process identifier macros when the candidate class is known
                if (candidateClass != null)
                {
                    if (im.className.equals(candidateClass.getName()))
                    {
                        if (im.fieldName == null)
                        {
                            throw new JDOUserException(LOCALISER.msg("JPOXSQL.ResultClassesHaveNoTableError",im));
                        }
                        if (im.subfieldName != null)
                        {
                            throw new JDOUserException(LOCALISER.msg("JPOXSQL.NoSuchFieldInResultClassError",im.className,im));
                        }
                        int fieldNumber = candidateCmd.getFieldNumber(im.fieldName);
                        
                        if (fieldNumber < 0)
                        {
                            throw new JDOUserException(LOCALISER.msg("JPOXSQL.NoSuchFieldInResultClassError",im.className,im));
                        }
                        im.value = (String) fieldColumnNames.get(fieldNumber);
                    }
                    else
                    {
                        getStoreManager().resolveIdentifierMacro(im);
                    }
                }
            }
            
            public void onParameterMacro(MacroString.ParameterMacro pm)
            {
                parameterOccurrences.add(pm.parameterName);
            }
            }, pm.getClassLoaderResolver());
    }

    /**
     * Execute the query and return the filtered QueryResult.
     * @param parameters the Map containing all of the parameters.
     * @return the filtered QueryResult
     */
    public Collection performExecute(Map parameters)
    {
        compile();

        if (parameters.size() != parameterNames.size())
        {
            throw new JDOUserException(LOCALISER.msg("JPOXSQL.IncorrectNumberOfParametersError","" + parameterNames.size(),"" + parameters.size()));
        }

        QueryResult qr = null;
        try
        {
            Connection conn = pm.getConnection(false);

            try
            {
                PreparedStatement ps = conn.prepareStatement(jdbcSqlText);
                try
                {
                    Iterator iter = parameterOccurrences.iterator();
                    int stmtParamNum = 1;

                    while (iter.hasNext())
                    {
                        String paramName = (String) iter.next();
                        Class paramType = (Class) parameterTypesByName.get(paramName);

                        if (!parameters.containsKey(paramName))
                        {
                            throw new JDOUserException(LOCALISER.msg("JPOXSQL.ParameterNotProvidedError",paramName));
                        }
                        if (paramType == null)
                        {
                            throw new JDOUserException(LOCALISER.msg("JPOXSQL.UndeclaredParameterError",paramName));
                        }

                        JavaTypeMapping mapping = dba.getMapping(paramType, getStoreManager());
                        Object paramValue = parameters.get(paramName);

                        mapping.setObject(pm, ps, Mappings.getParametersIndex(stmtParamNum,mapping), paramValue);
                        if( mapping.getNumberOfDatastoreFields()==0 )
                        {
                            stmtParamNum++;
                        }
                        else
                        {
                            stmtParamNum+=mapping.getNumberOfDatastoreFields();
                        }
                    }

                    // Apply any user-specified timeout
                    if (timeout > 0)
                    {
                        ps.setQueryTimeout(timeout);
                    }

                    // Apply any fetch size
                    ps.setFetchSize(pm.getPMFContext().getPmfConfiguration().getQueryFetchSize());

                    // Apply any fetch direction
                    String fetchDir = pm.getPMFContext().getPmfConfiguration().getQueryFetchDirection();
                    if (fetchDir.equals("reverse"))
                    {
                        ps.setFetchDirection(ResultSet.FETCH_REVERSE);
                    }
                    else if (fetchDir.equals("unknown"))
                    {
                        ps.setFetchDirection(ResultSet.FETCH_UNKNOWN);
                    }

                    ResultSet rs = executeQuery(jdbcSqlText, ps);
                    try
                    {
                        if (candidateClass != null)
                        {
                            ResultSetMetaData rsmd = rs.getMetaData();
                            HashSet remainingColumnNames = new HashSet(fieldColumnNames);

                            int colCount = rsmd.getColumnCount();
                            for (int colNum = 1; colNum <= colCount; ++colNum)
                            {
                                String colName = rsmd.getColumnName(colNum);
                                int fieldNumber = fieldColumnNames.indexOf(colName);
                                if (fieldNumber >= 0)
                                {
                                    statementExpressionIndex[fieldNumber].setExpressionIndex(new int[]{colNum});
                                    remainingColumnNames.remove(colName);
                                }
                            }
                            
                            if (!remainingColumnNames.isEmpty())
                            {
                                throw new JDOUserException(LOCALISER.msg("JPOXSQL.ResultSetMissingColumnsError",remainingColumnNames));
                            }

                            if (resultClass != null)
                            {
                                qr = new QueryResult(null, this, new ResultClassROF(resultClass, statementExpressionIndex), rs);
                            }
                            else
                            {
                                qr = new QueryResult(null, this, new TransientIDROF(candidateClass, fieldNumbers, statementExpressionIndex), rs);
                            }
                        }
                        else
                        {
                            // No candidate class, so use resultClass or Object/Object[]
                            Class requiredResultClass = resultClass;
                            int numberOfColumns = 0;
                            String[] resultFieldNames = null;
                            try
                            {
                                ResultSetMetaData rsmd = rs.getMetaData();
                                numberOfColumns = rsmd.getColumnCount();
                                if (requiredResultClass == null)
                                {
                                    if (numberOfColumns == 1)
                                    {
                                        requiredResultClass = Object.class;
                                    }
                                    else
                                    {
                                        requiredResultClass = Object[].class;
                                    }
                                }

                                // Generate names to use for the fields based on the column names
                                resultFieldNames = new String[numberOfColumns];
                                for (int i=0;i<numberOfColumns;i++)
                                {
                                    resultFieldNames[i] = rsmd.getColumnName(i+1).toLowerCase();
                                }
                            }
                            catch (SQLException sqe)
                            {
                                // Do nothing
                            }

                            qr = new QueryResult(null, this, new ResultClassROF(requiredResultClass, resultFieldNames), rs);
                        }
                    }
                    finally
                    {
                        if (qr == null)
                        {
                            rs.close();
                        }
                    }
                }
                finally
                {
                    if (qr == null)
                    {
                        ps.close();
                    }
                }
            }
            finally
            {
                pm.releaseConnection(conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("JPOXSQL.ExecutionError",jdbcSqlText),e);
        }

        queryResults.add(qr);

        return qr;
    }
}