/**********************************************************************
Copyright (c) 2002 Kelly Grizzle (TJDO) and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - commented and localised
2003 Andy Jefferson - coding standards
2004 Andy Jefferson - addition of setUnique() and setRange()
2005 Andy Jefferson - added timeout support.
    ...
**********************************************************************/
package org.jpox.store.query;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.StringTokenizer;

import javax.jdo.Extent;
import javax.jdo.FetchPlan;
import javax.jdo.JDOFatalUserException;
import javax.jdo.JDOUserException;

import org.jpox.FetchPlanImpl;
import org.jpox.PersistenceManager;
import org.jpox.store.DatastoreAdapter;
import org.jpox.store.StoreManager;
import org.jpox.store.exceptions.QueryNotUniqueException;
import org.jpox.store.rdbms.adapter.DatabaseAdapter;
import org.jpox.util.Imports;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;
import org.jpox.util.StringUtils;

/**
 * The Query interface allows applications to obtain persistent instances
 * from the data store.
 * TODO Document the parser/compiler
 *
 * @see javax.jdo.Query
 * @version $Revision: 1.18 $
 */
public abstract class Query implements javax.jdo.Query
{
    protected static final Localiser LOCALISER=Localiser.getInstance("org.jpox.store.query.Localisation");

    protected final transient PersistenceManager pm;
    protected final transient DatastoreAdapter dba;

    protected Class candidateClass;
    protected boolean subclasses=true;
    protected String filter;
    protected String imports;
    protected String variables;
    protected String parameters;
    protected String ordering;
    protected String grouping;

    /** Whether the query can be modified */
    protected boolean unmodifiable=false;

    /** Whether to return single value, or collection from the query. */
    protected boolean unique=false;

    /** "From" index to return (inclusive). */
    protected int fromInclNo=0;

    /** "To" index to return (exclusive). */
    protected int toExclNo=Integer.MAX_VALUE;

    /** User-defined class that best represents the results of a query. */
    protected Class resultClass=null;

    /** Whether to ignore dirty instances in the query. */
    protected boolean ignoreCache=false;

    /** Fetch Plan to use for the query. */
    protected final FetchPlan fetchPlan;

    /** State variable for the compilation state */
    protected transient boolean isCompiled = false;

    /** The imports definition. */
    protected transient Imports parsedImports = null;

    /** The parameter names. */
    protected transient List parameterNames = null;

    /** Look-up for the parameter types, keyed by the name. */
    protected transient Map parameterTypesByName = null;

    /** The variables names. */
    protected transient List variableNames = null;

    /** Look-up for the variables types, keyed by the name. */
    protected transient Map variableTypesByName = null;

    /**
     * All query results obtained from this query.
     * This is required because the query can be executed multiple times changing
     * the input slightly each time for example.
     */
    protected transient HashSet queryResults = new HashSet();

    protected transient boolean isPreCompile = true;

    /** The timeout for this query (in seconds) */
    protected transient int timeout = 0;

    /**
     * Constructs a new query instance that uses the given persistence manager.
     * @param pm The Persistence Manager for this query.
     */
    public Query(PersistenceManager pm)
    {
        this.pm = pm;

        // PM should always be provided so throw exception if null
        if (pm == null)
        {
            throw new JDOUserException(LOCALISER.msg("Quuery.NoPMSpecified"));
        }

        this.fetchPlan = ((FetchPlanImpl)pm.getFetchPlan()).getCopy();

        // Inherit IgnoreCache from PersistenceManager (JDO 1.0 $12.6.3)
        this.ignoreCache = pm.getIgnoreCache();

        dba = pm.getStoreManager() == null ? null : (DatabaseAdapter)pm.getStoreManager().getDatastoreAdapter();
    }

    /**
     * Returns the class loader of the candidade class 
     * @return the ClassLoader for the candidate class
     */
    protected ClassLoader getCandidateLoader()
    {
        return candidateClass == null ? null : candidateClass.getClassLoader();
    }

    /**
     * Utility to remove any previous compilation of this Query.
     **/
    protected void discardCompiled()
    {
        isCompiled = false;
        parsedImports = null;
        parameterNames = null;
        parameterTypesByName = null;
        variableNames = null;
        variableTypesByName = null;
    }

    /**
     * Equality operator.
     * @param obj Object to compare against
     * @return Whether this and the other object are equal.
     **/
    public boolean equals(Object obj)
    {
        if (obj == this)
        {
            return true;
        }

        if (!(obj instanceof Query))
        {
            return false;
        }

        Query q = (Query)obj;

        if (candidateClass == null) 
        {
            if (q.candidateClass != null)
            {
                return false;
            }
        }
        else if (!candidateClass.equals(q.candidateClass))
        {
            return false;
        }

        if (filter == null) 
        {
            if (q.filter != null)
            {
                return false;
            }
        }
        else if (!filter.equals(q.filter))
        {
            return false;
        }

        if (imports == null)
        {
            if (q.imports != null)
            {
                return false;
            }
        }
        else if (!imports.equals(q.imports)) 
        {
            return false;
        }

        if (parameters == null)
        {
            if (q.parameters != null)
            {
                return false;
            }
        }
        else if (!parameters.equals(q.parameters))
        {
            return false;
        }

        if (variables == null)
        {
            if (q.variables != null)
            {
                return false;
            }
        }
        else if (!variables.equals(q.variables))
        {
            return false;
        }

        if (unique != q.unique)
        {
            return false;
        }

        if (unmodifiable != q.unmodifiable)
        {
            return false;
        }

        if (resultClass != q.resultClass)
        {
            return false;
        }

        if (grouping == null)
        {
            if (q.grouping != null)
            {
                return false;
            }
        }
        else if (!grouping.equals(q.grouping))
        {
            return false;
        }

        if (ordering == null)
        {
            if (q.ordering != null)
            {
                return false;
            }
        }
        else if (!ordering.equals(q.ordering))
        {
            return false;
        }

        return true;
    }

    /**
     * Hashcode generator.
     * @return The Hashcode for this object.
     */
    public int hashCode()
    {
        return (candidateClass == null ? 0 : candidateClass.hashCode()) ^ 
        	(filter == null ? 0 : filter.hashCode()) ^ 
        	(imports == null ? 0 : imports.hashCode()) ^ 
        	(parameters == null ? 0 : parameters.hashCode()) ^ 
        	(variables == null ? 0 : variables.hashCode()) ^
        	(resultClass == null ? 0 : resultClass.hashCode()) ^
        	(grouping == null ? 0 : grouping.hashCode()) ^
        	(ordering == null ? 0 : ordering.hashCode());
    }

    /**
     * Accessor for the StoreManager associated with this Query.
     * @return the StoreManager associated with this Query.
     */
    public StoreManager getStoreManager()
    {
        return pm.getStoreManager();
    }

    /**
     * Accessor for the PersistenceManager associated with this Query.
     * @return the PersistenceManager associated with this Query.
     * @see javax.jdo.Query#getPersistenceManager
     */
    public javax.jdo.PersistenceManager getPersistenceManager()
    {
        return pm;
    }

    /**
     * This method retrieves the fetch plan associated with the Query. It always
     * returns the identical instance for the same Query instance. Any change
     * made to the fetch plan affects subsequent query execution. Fetch plan is
     * described in Section 12.7
     * @return the FetchPlan
     */
    public FetchPlan getFetchPlan()
    {
        return fetchPlan;
    }
    
    /**
     * Accessor for the class of the candidate instances of the query.
     * @return the Class of the candidate instances.
     * @see javax.jdo.Query#setClass
     */
    public Class getCandidateClass()
    {
        return candidateClass;
    }

    /**
     * Mutator for the class of the candidate instances of the query.
     * @param candidateClass the Class of the candidate instances.
     * @see javax.jdo.Query#setClass
     */
    public void setClass(Class candidateClass)
    {
        discardCompiled();
        assertIsUnmodifiable();

        this.candidateClass = candidateClass;
    }

    /**
     * Set the candidate Extent to query. To be implemented by extensions.
     * @param pcs the Candidate Extent.
     * @see javax.jdo.Query#setCandidates(javax.jdo.Extent)
     */
    public abstract void setCandidates(Extent pcs);

    /**
     * Set the candidate Collection to query. To be implemented by extensions.
     * @param pcs the Candidate collection.
     * @see javax.jdo.Query#setCandidates(java.util.Collection)
     */
    public abstract void setCandidates(Collection pcs);

    /**
     * Set the filter for the query.
     * @param filter the query filter.
     * @see javax.jdo.Query#setFilter
     */
    public void setFilter(String filter)
    {
        discardCompiled();
        assertIsUnmodifiable();

        this.filter = filter;
    }

    /**
     * Set the import statements to be used to identify the fully qualified name
     * of variables or parameters.
     * @param imports import statements separated by semicolons.
     * @see javax.jdo.Query#declareImports
     */
    public void declareImports(String imports)
    {
        discardCompiled();
        assertIsUnmodifiable();

        this.imports = imports;
    }

    /**
     * Declare the list of parameters query execution.
     * @param parameters the list of parameters separated by commas.
     * @see javax.jdo.Query#declareParameters
     */
    public void declareParameters(String parameters)
    {
        discardCompiled();
        assertIsUnmodifiable();

        this.parameters = parameters;
    }

    /**
     * Declare the unbound variables to be used in the query.
     * @param variables the variables separated by semicolons.
     * @see javax.jdo.Query#declareVariables
     */
    public void declareVariables(String variables)
    {
        discardCompiled();
        assertIsUnmodifiable();

        this.variables = variables;
    }

    /**
     * Set the ordering specification for the result Collection.
     * @param ordering the ordering specification.
     * @see javax.jdo.Query#setOrdering
     */
    public void setOrdering(String ordering)
    {
        discardCompiled();
        assertIsUnmodifiable();

        this.ordering = ordering;
    }

    /**
     * Set the grouping specification for the result Collection.
     * @param grouping the grouping specification.
     * @see javax.jdo.Query#setGrouping
     */
    public void setGrouping(String grouping)
    {
        discardCompiled();
        assertIsUnmodifiable();

//        this.grouping = grouping;
        throw new JDOUserException("setGrouping not yet supported");
    }

    /**
     * Set the uniqueness of the results. A value of true will return a single
     * value (or null) where the application knows that there are 0 or 1
     * objects to be returned. See JDO 2.0 specification section 14.6
     * @param unique whether the result is unique
     * @see javax.jdo.Query#setUnique
     * @since 1.1 
     */
    public void setUnique(boolean unique)
    {
        discardCompiled();
        assertIsUnmodifiable();

        this.unique = unique;
    }

    /**
     * Set the range of the results. By default all results are returned but
     * this allows specification of a range of elements required. See JDO 2.0
     * specification section 14.6.8
     * @param fromIncl From element no (inclusive) to return
     * @param toExcl To element no (exclusive) to return
     * @see javax.jdo.Query#setRange
     * @since 1.1
     */
    public void setRange(int fromIncl, int toExcl)
    {
        discardCompiled();
        assertIsUnmodifiable();

        // TODO Do error checking on invalid input
        this.fromInclNo = fromIncl;
        this.toExclNo = toExcl;
    }

    /**
     * Set the result class for the results.
     * The result class must obey various things as per the JDO 2.0 spec 14.6.12.
     * @param result_cls The result class
     * @see javax.jdo.Query#setResultClass
     * @since 1.1
     */
    public void setResultClass(Class result_cls)
    {
        discardCompiled();
        assertIsUnmodifiable();

        this.resultClass = result_cls;
    }

    /**
     * Set the ignoreCache option. Currently this simply stores the ignoreCache
     * value, and doesn't necessarily use it. The parameter is a "hint" to the
     * query engine. TODO : Implement this fully.
     * @param ignoreCache the setting of the ignoreCache option.
     * @see javax.jdo.Query#setIgnoreCache
     * @see javax.jdo.PersistenceManager#setIgnoreCache
     */
    public void setIgnoreCache(boolean ignoreCache)
    {
        discardCompiled();
        assertIsUnmodifiable();
        
        this.ignoreCache = ignoreCache;
    }

    /**
     * Accessor for the ignoreCache option setting.
     * @return the ignoreCache option setting
     * @see #setIgnoreCache
     * @see javax.jdo.Query#getIgnoreCache
     * @see javax.jdo.PersistenceManager#getIgnoreCache
     */
    public boolean getIgnoreCache()
    {
        return ignoreCache;
    }

    /**
     * Accessor for whether this query includes subclasses
     * @return Returns whether the query includes subclasses.
     */
    public boolean isSubclasses()
    {
        return subclasses;
    }

    /**
     * Mutator for whether this query includes subclasses
     * @param subclasses Where subclasses of the candidate class are to be included.
     */
    public void setSubclasses(boolean subclasses)
    {
        discardCompiled();
        assertIsUnmodifiable();

        this.subclasses = subclasses;

        // Set the candidates also if the candidate has been set
        if (candidateClass != null)
        {
            setCandidates(pm.getExtent(candidateClass, subclasses));
        }
    }

    /**
     * Accessor for unmodifiable.
     * @return Returns the unmodifiable.
     */
    public boolean isUnmodifiable()
    {
        return unmodifiable;
    }

    /**
     * Method to throw a JDOUserException if the query is currently
     * not modifiable.
     * @throws JDOUserException Thrown when it is unmodifiable
     */
    protected void assertIsUnmodifiable()
    {
        if (unmodifiable)
        {
            throw new JDOUserException(LOCALISER.msg("Query.Unmodifiable"));
        }
    }

    /**
     * Mutator for unmodifiable.
     * @param unmodifiable The unmodifiable to set.
     */
    public void setUnmodifiable(boolean unmodifiable)
    {
        this.unmodifiable = unmodifiable;
    }

    /**
     * Mutator for the timeout for this query.
     * <B>This is a JPOX extension.</B>
     * @param seconds The timeout to apply in seconds.
     */
    public void setTimeout(int seconds)
    {
        this.timeout = seconds;
    }

    /**
     * Accessor for the timeout for this query.
     * <B>This is a JPOX extension.</B>
     * @return The timeout in seconds.
     */
    public int getTimeout()
    {
        return timeout;
    }

    /**
     * Verify the elements of the query and provide a hint to the query to
     * prepare and optimize an execution plan. Processes the imports, variableNames and
     * parameterNames.
     * @see javax.jdo.Query#compile
     */
    public void compile()
    {
        if (isCompiled)
        {
            return;
        }

        if (pm == null)
        {
            throw new JDOUserException(LOCALISER.msg("Query.NoPersistenceManagerError"));
        }

        boolean done = false;

        try
        {
            // Parse imports list
            parsedImports = new Imports();
            if (candidateClass != null)
            {
                parsedImports.importPackage(candidateClass.getName());
            }
            if (imports != null)
            {
                parsedImports.parseImports(imports);
            }

            // Parse parameters list
            parameterNames = new ArrayList();
            parameterTypesByName = new HashMap();

            if (parameters != null && parameters.length() > 0)
            {
                StringTokenizer t1 = new StringTokenizer(parameters, ",");

                while (t1.hasMoreTokens())
                {
                    StringTokenizer t2 = new StringTokenizer(t1.nextToken(), " ");

                    if (t2.countTokens() != 2)
                    {
                        throw new JDOUserException(LOCALISER.msg("Query.InvalidParameterListError",parameters));
                    }

                    String classDecl     = t2.nextToken();
                    String parameterName = t2.nextToken();

                    if (!StringUtils.isValidJavaIdentifier(parameterName))
                    {
                        throw new JDOUserException(LOCALISER.msg("Query.IllegalParameterNameError",parameterName));
                    }

                    if (parameterNames.contains(parameterName))
                    {
                        throw new JDOUserException(LOCALISER.msg("Query.DuplicateParameterNameError",parameterName));
                    }

                    parameterNames.add(parameterName);
                    parameterTypesByName.put(parameterName, resolveClassDeclaration(classDecl));
                }
            }

            // Parse variables list
            variableNames = new ArrayList();
            variableTypesByName = new HashMap();

            if (variables != null && variables.length() > 0)
            {
                StringTokenizer t1 = new StringTokenizer(variables, ";");

                while (t1.hasMoreTokens())
                {
                    StringTokenizer t2 = new StringTokenizer(t1.nextToken(), " ");

                    if (t2.countTokens() != 2)
                    {
                        throw new JDOUserException(LOCALISER.msg("Query.InvalidVariableListError",variables));
                    }

                    String classDecl    = t2.nextToken();
                    String variableName = t2.nextToken();

                    if (!StringUtils.isValidJavaIdentifier(variableName))
                    {
                        throw new JDOUserException(LOCALISER.msg("Query.IllegalVariableName",variableName));
                    }

                    if (parameterNames.contains(variableName))
                    {
                        throw new JDOUserException(LOCALISER.msg("Query.ConflictingVariableName",variableName));
                    }

                    if (variableNames.contains(variableName))
                    {
                        throw new JDOUserException(LOCALISER.msg("Query.DuplicateVariableName",variableName));
                    }

                    variableNames.add(variableName);
                    variableTypesByName.put(variableName, resolveClassDeclaration(classDecl));
                }
            }

            done = true;
        }
        finally
        {
            if (!done)
            {
                discardCompiled();
            }
        }
    }

    /**
     * Utility to resolve the declaration to a particular class.
     * @param classDecl The declaration
     * @return The class it resolves to (if any)
     */
    public Class resolveClassDeclaration(String classDecl)
    {
        try
        {
            return parsedImports.resolveClassDeclaration(classDecl, pm.getClassLoaderResolver(), getCandidateLoader());
        }
        catch (ClassNotFoundException e)
        {
            throw new JDOUserException("Declared parameter or variable type not found: " + classDecl);
        }
    }

    // ----------------------------- Execute -----------------------------------

    /**
     * Execute the query and return the filtered List.
     * @return the filtered List.
     * @see javax.jdo.Query#execute()
     * @see #executeWithArray(Object[] parameters)
     */
    public Object execute()
    {
        return executeWithArray(new Object[0]);
    }

    /**
     * Execute the query and return the filtered List.
     *
     * @param p1 the value of the first parameter declared.
     * @return the filtered List.
     * @see javax.jdo.Query#execute(Object)
     * @see #executeWithArray(Object[] parameters)
     */
    public Object execute(Object p1)
    {
        return executeWithArray(new Object[] { p1 });
    }

    /**
     * Execute the query and return the filtered List.
     *
     * @param p1 the value of the first parameter declared.
     * @param p2 the value of the second parameter declared.
     * @return the filtered List.
     * @see javax.jdo.Query#execute(Object,Object)
     * @see #executeWithArray(Object[] parameters)
     */
    public Object execute(Object p1, Object p2)
    {
        return executeWithArray(new Object[] { p1, p2 });
    }

    /**
     * Execute the query and return the filtered List.
     *
     * @param p1 the value of the first parameter declared.
     * @param p2 the value of the second parameter declared.
     * @param p3 the value of the third parameter declared.
     * @return the filtered List.
     * @see javax.jdo.Query#execute(Object,Object,Object)
     * @see #executeWithArray(Object[] parameters)
     */
    public Object execute(Object p1, Object p2, Object p3)
    {
        return executeWithArray(new Object[] { p1, p2, p3 });
    }

    /**
     * Execute the query and return the filtered List.
     *
     * @param parameters the Object array with all of the parameters.
     * @return the filtered List.
     * @see javax.jdo.Query#executeWithArray(Object[])
     */
    public Object executeWithArray(Object[] parameters)
    {
        isPreCompile = false;
        compile();

        if (parameters.length != parameterNames.size())
        {
            throw new JDOUserException(LOCALISER.msg("Query.IncorrectNumberOfParametersError","" + parameterNames.size(),"" + parameters.length));
        }

        HashMap parameterMap = new HashMap();
        for (int i = 0; i < parameters.length; ++i)
        {
            parameterMap.put(parameterNames.get(i), parameters[i]);
        }

        Object theReturn = executeWithMap(parameterMap);
        isPreCompile = true;
        return theReturn; 
    }

    /**
     * Execute the query and return the filtered result(s).
     * @param parameters the Map containing all of the parameters.
     * @return the filtered results (List, or Object)
     * @see javax.jdo.Query#executeWithMap(Map)
     * @see #executeWithArray(Object[] parameters)
     */
    public Object executeWithMap(Map parameters)
    {
        // JDO 2 $14.6.1 : If PM is closed, throw JDOFatalUserException
        if (pm.isClosed())
        {
            throw new JDOFatalUserException(LOCALISER.msg("Query.PMIsClosed"));
        }

        Collection qr=performExecute(parameters);

        // If unique required, return single object
        if (unique)
        {
            if (qr == null || qr.size() == 0)
            {
                return null;
            }
            else
            {
                Iterator qr_iter=qr.iterator();
                Object uniqueValue = qr_iter.next();
                if (qr_iter.hasNext())
                {
                    throw new QueryNotUniqueException();
                }
                return uniqueValue;
            }
        }

        // Process any specified range
        // Use LIMIT where supported by the RDBMS, and otherwise just throw away ones outside the range
        boolean range_specified = (fromInclNo >= 0 && toExclNo >= 0 && (fromInclNo != 0 || toExclNo != Integer.MAX_VALUE));
        boolean using_limit_clause = (((DatabaseAdapter) dba).getRangeByLimitClause(fromInclNo, toExclNo).length() > 0);
        boolean using_rownum = (((DatabaseAdapter) dba).getRangeByRowNumberColumn().length() > 0);
        if (range_specified && !using_limit_clause && !using_rownum)
        {
            int i=0;
            Iterator qr_iter=qr.iterator();
            Collection res=new ArrayList();
            while (qr_iter.hasNext())
            {
                Object obj=qr_iter.next();
                if (i >= fromInclNo && i < toExclNo)
                {
                    // Accept the item if within range
                    res.add(obj);
                }
                i++;
            }
            return res;
        }
        else
        {
            return qr;
        }
    }

    /**
     * Method to actually execute the query. To be implemented by extending
     * classes for the particular query language.
     * @param parameters Map containing the parameters.
     * @return The filtered QueryResult.
     */
    protected abstract Collection performExecute(Map parameters);

    /**
     * Convenience wrapper to executing JDBC queries.
     * Provides logging of the query and its execution time.
     * @param statement The JDBC statement text
     * @param ps The PreparedStatement
     * @return The ResultSet for this query
     * @throws SQLException Thrown if an error occurs in the execution
     */
    protected ResultSet executeQuery(String statement, PreparedStatement ps)
    throws SQLException
    {
        if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
        {
            JPOXLogger.RDBMS_SQL.debug(statement);
        }
        long startTime = System.currentTimeMillis();
        ResultSet rs = ps.executeQuery();
        if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
        {
            JPOXLogger.RDBMS_SQL.debug(LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
        }

        return rs;
    }

    // ------------------------- Delete Persistent -----------------------------
 
    /**
     * Method to delete all objects found by this query, without instantiating
     * them into memory.
     * @return The deleted objects.
     * @see javax.jdo.Query#deletePersistentAll()
     */
    public Object deletePersistentAll()
    {
        return deletePersistentAll(new Object[0]);
    }

    /**
     * Method to delete all objects found by this query, without instantiating
     * them into memory.
     * @param parameters the Object array with all of the parameters.
     * @return the filtered Collection.
     * @see javax.jdo.Query#deletePersistentAll(Object[])
     */
    public Object deletePersistentAll(Object[] parameters)
    {
        isPreCompile = false;
        compile();

        if (parameters.length != parameterNames.size())
        {
            throw new JDOUserException(LOCALISER.msg("Query.IncorrectNumberOfParametersError","" + parameterNames.size(),"" + parameters.length));
        }

        HashMap parameterMap = new HashMap();
        for (int i=0; i<parameters.length; ++i)
        {
            parameterMap.put(parameterNames.get(i), parameters[i]);
        }

        Object theReturn = deletePersistentAll(parameterMap);
        isPreCompile = true;
        return theReturn;
    }

    /**
     * Method to delete all objects found by this query, without instantiating
     * them into memory.
     * @param parameters Map of parameters for the query
     * @return the filtered Collection of deleted objects.
     * @see javax.jdo.Query#deletePersistentAll(Map)
     */
    public Object deletePersistentAll(Map parameters)
    {
        // Check for specification of any illegal attributes
        if (resultClass != null)
        {
            throw new JDOUserException(LOCALISER.msg("Query.DeletePersistent.ResultClassSpecifiedError"));
        }
        if (ordering != null)
        {
            throw new JDOUserException(LOCALISER.msg("Query.DeletePersistent.OrderingSpecifiedError"));
        }
        if (fromInclNo >= 0 && toExclNo >= 0 &&
            (fromInclNo != 0 || toExclNo != Integer.MAX_VALUE))
        {
            throw new JDOUserException(LOCALISER.msg("Query.DeletePersistent.ResultRangeSpecifiedError"));
        }

        QueryResult qr=performDeletePersistentAll(parameters);

        return qr;
    }

    /**
     * Method to actually execute the deletion of objects.
     * To be implemented by extending classes.
     * @param parameters Map containing the parameters.
     * @return The filtered QueryResult.
     */
    protected abstract QueryResult performDeletePersistentAll(Map parameters);

    // -------------------------------------------------------------------------

    /**
     * Close a query result and release any resources associated with it.
     * @param queryResult the result of execute(...) on this Query instance.
     * @see javax.jdo.Query#close
     */
    public void close(Object queryResult)
    {
        if (queryResult != null && queryResult instanceof QueryResult)
        {
            ((QueryResult)queryResult).close();
            queryResults.remove(queryResult);
        }
    }

    /**
     * Close all query results associated with this Query instance,
     * and release all resources associated with them.
     * @see javax.jdo.Query#closeAll
     */
    public void closeAll()
    {
        QueryResult[] qrs = (QueryResult[])queryResults.toArray(new QueryResult[queryResults.size()]);

        for (int i = 0; i < qrs.length; ++i)
        {
            close(qrs[i]);
        }
    }

    /**
     * An object that reads result set rows and returns corresponding persistent
     * objects from them.  Different queries accomplish this in different ways,
     * so a query supplies a suitable ResultObjectFactory to each QueryResult
     * when it is executed.  The QueryResult only uses it to turn ResultSet rows
     * into objects and otherwise manages the ResultSet itself.
     * @see QueryResult
     */
    public static interface ResultObjectFactory
    {
        /**
         * Instantiates a persistent object instance from the current row of
         * the given result set.
         *
         * @param pm    The PersistenceManager.
         * @param rs    The result set.  The contents of the current row are
         *              used to locate or create a corresponding persistent
         *              object instance.
         * @param pcClass The PersistenceCapable class
         * @return  A persistent object instance.
         */
        Object getObject(PersistenceManager pm, ResultSet rs, Class pcClass);
    }
}