/**********************************************************************
Copyright (c) 2002 Kelly Grizzle (TJDO) and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2002 Mike Martin (TJDO)
2003 Erik Bengtson - fixed bug [833915] implements interface Queryable.
                    It is now able to be used as "candidates" for
                    another query
2003 Andy Jefferson - coding standards
2004 Andy Jefferson - added resultClass
2004 Andy Jefferson - added fix for optimistic tx, reading result set
    ...
**********************************************************************/
package org.jpox.store.query;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.AbstractList;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.List;
import java.util.NoSuchElementException;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOUserException;

import org.jpox.PersistenceManager;
import org.jpox.util.Localiser;
import org.jpox.util.SQLWarnings;

/**
 * Lazy collection results from a Query.
 * The actual result elements are only loaded when accessed.
 *
 * @version $Revision: 1.5 $ 
 */
public final class QueryResult extends AbstractList
    implements Queryable, java.io.Serializable
{
    protected static final Localiser LOCALISER=Localiser.getInstance("org.jpox.store.Localisation");

    private final Query query;
    private final Query.ResultObjectFactory rof;
    private ResultSet rs;
    private QueryStatement qs;

    protected boolean moreResultSetRows;
    protected List resultObjs = new ArrayList();

    /**
     * Constructor of the result from a Query.
     * @param qs The Query Statement
     * @param query The Query
     * @param rof The factory to retrieve results from
     * @param rs The ResultSet from the Query Statement
     */
    public QueryResult(QueryStatement qs, 
                       Query query,
                       Query.ResultObjectFactory rof,
                       ResultSet rs)
    throws SQLException
    {
        this.query = query;
        this.rof = rof;
        this.rs = rs;
        this.qs = qs;

        if (!(moreResultSetRows = rs.next()))
        {
            // ResultSet is empty, so just close it
            closeResultSet();
        }
        else if (!query.getPersistenceManager().currentTransaction().isActive() ||
                 query.getPersistenceManager().currentTransaction().getOptimistic())
        {
            // Called from either non-transactional context, or optimistic tx
            // in which case we should read the ResultSet now while we have the chance.
            // TODO We could adopt a different strategy for optimistic cases here and just
            // pin the optimistic connection. Needs more thought
            advanceToEndOfResultSet();
        }
    }

    /**
     * Internal method to advance to the end of the ResultSet.
     */
    private void advanceToEndOfResultSet()
    {
        // Advance through any remaining result set rows
        while (moreResultSetRows)
        {
            nextResultSetElement();
        }
    }

    /**
     * Accessor whether the results are open.
     * @return Whether it is open.
     */
    protected boolean isOpen()
    {
        return resultObjs != null;
    }

    /**
     * Internal method to throw a JDOUserException if the ResultSet is open.
     */
    private void assertIsOpen()
    {
        if (!isOpen())
        {
            throw new JDOUserException(LOCALISER.msg("QueryResult.Closed"));
        }
    }

    /**
     * Accessor for the next object from the ResultSet.
     * @return The next element from the ResultSet. 
     */
    protected Object nextResultSetElement()
    {
        Object nextElement = rof.getObject((PersistenceManager)query.getPersistenceManager(), rs, query.getCandidateClass());

        SQLWarnings.log(rs);

        resultObjs.add(nextElement);

        try
        {
            if (!(moreResultSetRows = rs.next()))
            {
                closeResultSet();
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("QueryResult.ReadError",e));
        }

        return nextElement;
    }

    /**
     * Accessor for an iterator for the results.
     * @return The iterator
     */
    public Iterator iterator()
    {
        assertIsOpen();

        return new QueryResultIterator();
    }

    /**
     * An Iterator results of a pm.query.execute().iterator()
     */
    private class QueryResultIterator implements Iterator
    {
        private int nextRowNum = 0;

        public QueryResultIterator()
        {
        }

        public boolean hasNext()
        {
            synchronized (QueryResult.this)
            {
                if (! isOpen())
                {
                    return false;
                }

                return nextRowNum < resultObjs.size() ? true : moreResultSetRows;
            }
        }

        public Object next()
        {
            synchronized (QueryResult.this)
            {
                if (! isOpen())
                {
                    throw new NoSuchElementException(LOCALISER.msg("QueryResult.Closed"));
                }

                Object nextElement = null;
                if (nextRowNum < resultObjs.size())
                {
                    nextElement = resultObjs.get(nextRowNum);
                }
                else if (moreResultSetRows)
                {
                    nextElement = nextResultSetElement();
                }
                else
                {
                    throw new NoSuchElementException(LOCALISER.msg("QueryResult.NoMoreElements"));
                }

                ++nextRowNum;

                return nextElement;
            }
        }

        public void remove()
        {
            throw new UnsupportedOperationException(LOCALISER.msg("QueryResult.NotModifiable"));
        }
    }

    /**
     * Method to check if the specified object is contained in this result.
     * @param o The object
     * @return Whether it is contained here.
     */
    public synchronized boolean contains(Object o)
    {
        assertIsOpen();
        advanceToEndOfResultSet();

        return resultObjs.contains(o);
    }

    /**
     * Method to check if all of the specified objects are contained here.
     * @param c The collection of objects
     * @return Whether they are all contained here.
     */
    public synchronized boolean containsAll(Collection c)
    {
        assertIsOpen();
        advanceToEndOfResultSet();

        return resultObjs.containsAll(c);
    }

    /**
     * Equality operator for QueryResults.
     * Overrides the AbstractList implementation since that uses 
     * size() and iterator() and that would cause problems when closed.
     * @param o The object to compare against
     * @return Whether they are equal
     */
    public boolean equals(Object o)
    {
        if (o == null || !(o instanceof QueryResult))
        {
            return false;
        }
        QueryResult other = (QueryResult)o;
        
        // Just base equality on the QueryStatement for now
        // NB For SQL queries the QueryStatement will be null so use Query
        if (qs != null)
        {
            return other.qs == qs;
        }
        else
        {
            return other.query == query;
        }
    }

    /**
     * Accessor for the hashcode of this object
     * @return The hash code
     */
    public int hashCode()
    {
        // Just base equality on the QueryStatement for now
        // NB For SQL queries the QueryStatement will be null so use Query
        if (qs != null)
        {
            return qs.hashCode();
        }
        else
        {
            return query.hashCode();
        }
    }

    /**
     * Method to retrieve a particular element from the list.
     * @param index The index of the element
     * @return The element at index
     */
    public synchronized Object get(int index)
    {
        assertIsOpen();
        advanceToEndOfResultSet();

        return resultObjs.get(index);
    }

    /**
     * Returns <tt>true</tt> if this collection contains no elements.<p>
     *
     * This implementation check if the result is empty.
     *
     * @return <tt>true</tt> if this collection contains no elements.
     */
    public synchronized boolean isEmpty()
    {
        assertIsOpen();

        return resultObjs.isEmpty() && !moreResultSetRows;
    }

    /**
     * Method to return the size of the result.
     * @return The size of the result.
     */
    public synchronized int size()
    {
        assertIsOpen();
        advanceToEndOfResultSet();

        return resultObjs.size();
    }

    /**
     * Method to return the results as an array.
     * @return The array.
     */
    public synchronized Object[] toArray()
    {
        assertIsOpen();
        advanceToEndOfResultSet();

        return resultObjs.toArray();
    }

    /**
     * Method to return the results as an array.
     * @param a The array to copy into. 
     * @return The array.
     */
    public synchronized Object[] toArray(Object[] a)
    {
        assertIsOpen();
        advanceToEndOfResultSet();

        return resultObjs.toArray(a);
    }

    /**
     * Internal method to close the ResultSet.
     */
    private void closeResultSet()
    {
        if (rs != null)
        {
            try
            {
                try
                {
                    rs.getStatement().close();
                }
                catch (SQLException e)
                {
                    throw new JDODataStoreException(LOCALISER.msg("QueryResult.CloseError",e));
                }
            }
            finally
            {
                rs = null;
            }
        }
    }

    /**
     * Method to close the results.
     */
    public synchronized void close()
    {
        resultObjs = null;
        moreResultSetRows = false;

        closeResultSet();
    }

    /* (non-Javadoc)
     * @see org.jpox.store.query.Queryable#newQueryStatement()
     */
    public QueryStatement newQueryStatement()
    {
        return qs;
    }

    /* (non-Javadoc)
     * @see org.jpox.store.query.Queryable#newQueryStatement(java.lang.Class)
     */
    public QueryStatement newQueryStatement(Class candidateClass)
    {
        return qs;
    }

    /* (non-Javadoc)
     * @see org.jpox.store.query.Queryable#newResultObjectFactory(org.jpox.store.QueryStatement)
     */
    public Query.ResultObjectFactory newResultObjectFactory(QueryStatement stmt,boolean ignoreCache,Class resultClass)
    {
        return rof;
    }
}