/**********************************************************************
Copyright (c) 2002 Kelly Grizzle (TJDO) and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2002 Mike Martin (TJDO)
2003 Erik Bengtson - added UNION operator
2003 Erik Bengtson - fixed andCondition to queries in UNION
2003 Erik Bengtson - fixed number of column positions in sql returned
                     on select method
2004 Erik Bengtson - removed ununsed imports
2004 Erik Bengtson - block duplicate joins
2004 Erik Bengtson - fixed bug [833915] QueryResult passed as parameter
                     for another Query
2004 Erik Bengtson - fixed bug [832484] Query - boolean expressions                  
2004 Andy Jefferson - commented
2004 Andy Jefferson - added selectAlias methods and support for DatabaseAdapters
                      that need the ORDER BY columns aliased.
2004 Erik Bengtson - added DELETE queries                      
    ...
**********************************************************************/
package org.jpox.store.query;

import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import javax.jdo.JDOFatalInternalException;

import org.jpox.store.DatastoreAdapter;
import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.DatastoreField;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.StatementText;
import org.jpox.store.StoreManager;
import org.jpox.store.expression.AggregateExpression;
import org.jpox.store.expression.BooleanExpression;
import org.jpox.store.expression.BooleanLiteral;
import org.jpox.store.expression.Literal;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.TableExpression;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.rdbms.adapter.DatabaseAdapter;
import org.jpox.store.rdbms.sqlidentifier.TableIdentifier;
import org.jpox.util.Localiser;


/**
 * Representation of a statement for a Query.
 * The SQL SELECT statement is built up from a series of components, and then
 * assembled into the final statement at the end. Supports UNIONs of
 * statements.
 *
 * The SQL DELETE statement is made of a FROM and WHERE clauses.
 * 
 * @version $Revision: 1.12 $
 **/
public class QueryStatement
{
    protected static final Localiser LOCALISER=Localiser.getInstance("org.jpox.store.Localisation");

    protected final StoreManager storeMgr;
    protected final DatastoreAdapter dba;
    protected final DatastoreIdentifier defaultRangeVar;
    protected final TableExpression initialTableExpr;

    protected Map     tableExprsByRangeVar  = new HashMap();
    protected boolean distinctResults       = false;
    protected List         		selected    = new ArrayList();
    protected List         		joins       = new ArrayList();
    protected List         	    union       = new ArrayList();
    protected BooleanExpression whereExpr   = null;
    protected BooleanExpression havingExpr  = null;
    protected StatementText     orderByList = null;
    protected StatementText     stmtText    = null;
    
    protected char stmtFromTablesSeparator  = ' ';
    protected char stmtAliasTablesSeparator = ',';

    /** non aggregate expressions **/
    protected List   nonAggregateExpression = new ArrayList();
    protected boolean hasAggregateExpression = false;

    /** ensures that we don't have duplicate JOINs From vs To **/ 
    protected List              joinsTo     = new ArrayList();

    /** add tables/alias but no joins to they, will be output as FROM TABLE1,TABLE2,TABLE3 **/ 
    protected List              alias       = new ArrayList();
    
    /** Whether this query is to be used as a as set for the Exists function. <PRE>e.g WHERE EXISTS(QUERY)</PRE>. */
    protected boolean isExistsSubQuery      = false;

    protected long rangeOffset = -1;
    protected long rangeCount = -1;

    public QueryStatement(DatastoreContainerObject initialTable)
    {
        this.storeMgr = initialTable.getStoreManager();
        this.dba      = storeMgr.getDatastoreAdapter();

        defaultRangeVar = new TableIdentifier(dba, "this");

        initialTableExpr = newTableExpression(initialTable, defaultRangeVar);
        tableExprsByRangeVar.put(defaultRangeVar, initialTableExpr);
        
    }

    public QueryStatement(DatastoreContainerObject initialTable, DatastoreIdentifier initialRangeVar)
    {
        this.storeMgr = initialTable.getStoreManager();
        this.dba      = storeMgr.getDatastoreAdapter();

        defaultRangeVar = new TableIdentifier(dba, "this");

        initialTableExpr = newTableExpression(initialTable, initialRangeVar);
        tableExprsByRangeVar.put(initialRangeVar, initialTableExpr);
    }

    protected void assertNotFrozen()
    {
        if (stmtText != null)
        {
            throw new JDOFatalInternalException("A query statement cannot be modified after being output");
        }
    }

    /**
     * Sort Join instances making sure that joining tables where previously
     * declared before the search condition ON.
     * @param toSort list of Join to sort
     * @return a list of sorted Join instances
     */
    private List sortJoins(List toSort)
    {
        List temp = new ArrayList(toSort);
        for( int i=0; i<temp.size(); i++)
        {
            Join join0 = (Join) temp.get(i);
            int index = i;
            boolean changed = false;
		    if( !getInitialTableExpr().getRangeVariable().equals(join0.expr.getTe().getRangeVariable()) )
		    {
		        for( int j=0; j<temp.size(); j++)
		        {
		            Join join1 = (Join) temp.get(j);
				    if( join1 != join0 && join1.tblExpr.getRangeVariable().equals(join0.expr.getTe().getRangeVariable()) )
				    {
				        if( !changed )
				        {
					        index = j;
					        changed = true;
				        }
				    }
		        }
		    }
		    if( !getInitialTableExpr().getRangeVariable().equals(join0.expr2.getTe().getRangeVariable()) )
		    {
		        for( int j=0; j<temp.size(); j++)
		        {
		            Join join1 = (Join) temp.get(j);
				    if( join1 != join0 && join1.tblExpr.getRangeVariable().equals(join0.expr2.getTe().getRangeVariable()) )
				    {
				        if( !changed || index < j)
				        {
					        index = j;
					        changed = true;
				        }
				    }
		        }
		    }
	        if( index != i )
	        {
	            temp.remove(i);
	            temp.add(( i < index ) ? index : index+1,join0);
	        }
        }
        return temp;
    }

    /**
     * @return Returns the initialTableExpr.
     */
    public TableExpression getInitialTableExpr()
    {
        return initialTableExpr;
    }
    
    public int getNoOfSelected()
    {
        return selected.size();
    }

    public List getSelected()
    {
        return selected;
    }

    public StoreManager getStoreManager()
    {
        return storeMgr;
    }

    public TableExpression getTableExpression(DatastoreIdentifier rangeVar)
    {
        return (TableExpression)tableExprsByRangeVar.get(rangeVar);
    }

    public TableExpression getDefaultTableExpression()
    {
        return getTableExpression(defaultRangeVar);
    }

    public TableExpression newTableExpression(DatastoreContainerObject mainTable, DatastoreIdentifier rangeVar)
    {
        assertNotFrozen();

        TableExpression te = (TableExpression)tableExprsByRangeVar.get(rangeVar);

        if (te == null)
        {
            te = ((DatabaseAdapter)dba).newTableExpression(this, mainTable, rangeVar);

            tableExprsByRangeVar.put(rangeVar, te);
        }
        else
        {
            if (!te.getMainTable().equals(mainTable))
            {
                throw new JDOFatalInternalException(LOCALISER.msg("QueryStatement.RangeVariableInUse",rangeVar,this.toStatementText(false)));
            }
        }

        return te;
    }

    public boolean getDistinctResults()
    {
        return distinctResults;
    }

    public void setDistinctResults(boolean distinctResults)
    {
        assertNotFrozen();

        this.distinctResults = distinctResults;
    }

    // ------------------------- SELECT CRITERIA -------------------------------
    /**
     * Add a column to the SELECT step
     * @param col The column to add.
     * @return The index of the column in the select
     **/
    private synchronized int select(DatastoreField col)
    {
        return selectQueryExpression( getQueryExpression(defaultRangeVar, col) );
    }

    /**
     * Select the columns for a mapping
     * @param mapping The mapping
     * @return The index of the columns in the select
     **/
    public synchronized int[] select(JavaTypeMapping mapping)
    {
        int[] index = new int[mapping.getNumberOfDatastoreFields()];
        for (int i=0; i<index.length; i++)
        {
            index[i] = select(mapping.getDataStoreMapping(i).getDatastoreField());
        }
        return index;
    }    
    
    /**
     * select a new column, add to union queries, if unionQueries is true
     * @param mapping The mapping
     * @param unionQueries Whether to add to any union
     * @return The index of the columns in the select
     */
    public synchronized int[] select(JavaTypeMapping mapping,boolean unionQueries)
    {
        int[] index = select(mapping);
        if (unionQueries)
        {
            for (int i=0; i<union.size(); i++)
            {
                QueryStatement qs = ((QueryStatement)union.get(i));
                qs.select(mapping);
            }
        }
        return index;        
    }
    
    /**
     * select an expression; eg: "'Text' as alias"
     * @param expr The expression to add to the select statement 
     * @return The index of the expression in the select
     */
    public int selectScalarExpression(ScalarExpression expr)
    {
        assertNotFrozen();

        if( !(expr instanceof AggregateExpression) )
        {
            nonAggregateExpression.add(expr);
        }
        else
        {
            hasAggregateExpression = true;
        }

        String expression = expr.toStatementText().toString();
        
        if (!selected.contains(expression))
        {
            selected.add(selected.size(),expression);
        }

        return selected.indexOf(expression)+1;
    }    
    
    /**
     * select an expression; eg: "'Text' as alias"
     * @param expr The expression to add to the select statement 
     * @return The index of the expression in the select
     */
    public int selectScalarExpression(ScalarExpression expr,boolean unionQueries)
    {
        assertNotFrozen();

        if (unionQueries)
        {
            for (int i=0; i<union.size(); i++)
            {
                QueryStatement qs = ((QueryStatement)union.get(i));
                qs.selectScalarExpression(expr);
            }
        }
        
        return selectScalarExpression(expr);
    }     
    
    /**
     * select a new column, add to union queries, if unionQueries is true
     * @param rangeVar The range variable
     * @param col The column to add.
     * @return The index of the column in the select
     */
    private synchronized int selectQueryExpression(QueryExpression queryExpr)
    {
        assertNotFrozen();

        String expr = queryExpr.toString();

        if (!selected.contains(expr))
        {
            selected.add(selected.size(),expr);
        }

        return selected.indexOf(expr)+1;
    }
    
    /**
     * Select columns, add to union queries, if unionQueries is true
     * @param rangeVar The range variable
     * @param mapping The mapping
     * @return The index of the columns in the select
     */
    public synchronized int[] select(DatastoreIdentifier rangeVar, JavaTypeMapping mapping)
    {
        assertNotFrozen();

        int      index[] = new int[mapping.getNumberOfDatastoreFields()];
        for (int i=0; i<index.length; i++)
        {
            index[i] = selectQueryExpression(getQueryExpression(rangeVar,mapping.getDataStoreMapping(i).getDatastoreField()));
        }
        return index;
    }
    
    /**
     * select columns, add to union queries, if unionQueries is true
     * @param rangeVar The range variable
     * @param mapping The mapping
     * @param unionQueries Whether to add to any union
     * @return The index of the column in the select
     */
    public synchronized int[] select(DatastoreIdentifier rangeVar, JavaTypeMapping mapping,
                                     boolean unionQueries)
    {
        int[] index = select(rangeVar, mapping);
        if (unionQueries)
        {
            for (int i=0; i<union.size(); i++)
            {
                QueryStatement qs = ((QueryStatement)union.get(i));
                qs.select(rangeVar, mapping);
            }
        }
        return index;   
    }
    
    /**
     * Accessor for the number of columns selected.
     * @return The number of columns in the SELECT
     **/
    public int columnsSelected()
    {
        return selected.size();
    }

    // ------------------------- QUERY COLUMNS ---------------------------------
      
    private QueryExpression getQueryExpression(DatastoreIdentifier rangeVar, DatastoreField col)
    {
        TableExpression te = (TableExpression)tableExprsByRangeVar.get(rangeVar);

        if (te == null)
        {
            throw new JDOFatalInternalException(LOCALISER.msg("QueryStatement.RangeVariableError",rangeVar));
        }

        return getQueryExpression(te, col);
    }   

    private QueryExpression getQueryExpression(TableExpression te, DatastoreField field)
    {
        return new QueryExpression(te, field);
    }   

    /**
     * Method to add tables/alias but no joins to they, will be output as FROM TABLE1,TABLE2,TABLE3
     * @param tableExpr table expression
     **/
    private void addAlias(TableExpression tableExpr)
    {
        assertNotFrozen();

        // if we already have this alias, return
        if (joinsTo.contains(tableExpr))
        {
            return;
        }
        alias.add(tableExpr);
    }
    
    /**
     * Method to add tables/alias but no joins to they, will be output as FROM TABLE1,TABLE2,TABLE3
     * @param tableExpr table expression
     * @param unionQueries Whether to apply the alias to unions of this query.
     **/
    public void addAlias(TableExpression tableExpr, boolean unionQueries)
    {
        if (unionQueries)
        {
            Iterator i = union.iterator();
            while (i.hasNext())
            {
                QueryStatement qs = ((QueryStatement)i.next());
                qs.addAlias(tableExpr);
            }
        }
        addAlias(tableExpr);
    }    
    
    /**
     * Method to do an inner join to another table, and optionally apply it to
     * any unions for this query.
     * @param expr the boolean expression
     * @param unionQueries Whether to apply to unions of this query.
     **/
    public synchronized void innerJoin(ScalarExpression expr, ScalarExpression expr2, 
            						   TableExpression tblExpr,
                                       boolean unionQueries)
    {
        if (unionQueries)
        {
            Iterator i = union.iterator();
            while (i.hasNext())
            {
                QueryStatement qs = ((QueryStatement)i.next());
                qs.innerJoin(expr, expr2, tblExpr, unionQueries);
            }
        }
        innerJoin(expr, expr2, tblExpr);
    }

    /**
     * Method to do an inner join to another table.
     * @param expr the boolean expression
     **/
    public void innerJoin(ScalarExpression expr, ScalarExpression expr2, TableExpression tblExpr)
    {
        assertNotFrozen();

        Join join = new Join("INNER JOIN", expr, expr2, tblExpr);

        if( tblExpr.equals(getInitialTableExpr()) )
        {
            andCondition(expr.eq(expr2));
            return;
        }
        if( alias.contains(tblExpr) )
        {
            alias.remove(tblExpr);
        }
        // if we already have this join, return
        if (joinsTo.contains(tblExpr.getRangeVariable()))
        {
            return;
        }
        joinsTo.add(tblExpr.getRangeVariable());

        joins.add(join);
    }

    /**
     * Method to do a left outer join to another table, and optionally apply it
     * to any unions for this query.
     * @param expr the boolean expression
     * @param unionQueries Whether to apply to unions of this query.
     **/
    public synchronized void leftOuterJoin(ScalarExpression expr, ScalarExpression expr2, 
            							   TableExpression tblExpr,
                                           boolean unionQueries)
    {
        if (unionQueries)
        {
            Iterator i = union.iterator();
            while (i.hasNext())
            {
                QueryStatement qs = ((QueryStatement)i.next());
                qs.leftOuterJoin(expr,expr2,tblExpr,unionQueries);
            }
        }
        leftOuterJoin(expr, expr2, tblExpr);
    }

    /**
     * Method to do a left outer join to another table.
     * @param expr the boolean expression
     **/
    public void leftOuterJoin(ScalarExpression expr, ScalarExpression expr2, TableExpression tblExpr)
    {
        assertNotFrozen();

        Join join = new Join("LEFT OUTER JOIN", expr, expr2, tblExpr);

        // if we already have this join, return
        if (joinsTo.contains(tblExpr.getRangeVariable()))
        {
            return;
        }
        joinsTo.add(tblExpr.getRangeVariable());

        joins.add(join);
    }

    /**
     * Method to do a right outer join to another table, and optionally apply it
     * to any unions for this query.
     * @param expr the boolean expression
     * @param unionQueries Whether to apply to unions of this query.
     **/
    public synchronized void rightOuterJoin(ScalarExpression expr, ScalarExpression expr2, 
            								TableExpression tblExpr,
                                            boolean unionQueries)
    {
        if (unionQueries)
        {
            Iterator i = union.iterator();
            while (i.hasNext())
            {
                QueryStatement qs = ((QueryStatement)i.next());
                qs.rightOuterJoin(expr,expr2,tblExpr,unionQueries);
            }
        }
        rightOuterJoin(expr,expr2,tblExpr);
    }

    /**
     * Method to do a right outer join to another table.
     * @param expr the boolean expression
     **/
    public void rightOuterJoin(ScalarExpression expr, ScalarExpression expr2, TableExpression tblExpr)
    {
        assertNotFrozen();

        Join join = new Join("RIGHT OUTER JOIN", expr, expr2, tblExpr);

        // if we already have this join, return
        if (joinsTo.contains(tblExpr.getRangeVariable()))
        {
            return;
        }
        joinsTo.add(tblExpr.getRangeVariable());

        joins.add(join);
    }

    public void union(QueryStatement qt)
    {
        assertNotFrozen();

        union.add(qt);
    }

    public void andCondition(BooleanExpression condition)
    {
        assertNotFrozen();

        // if condition.toString() == "TRUE" or "FALSE", doesnt make sense or doesnt work pass it to the sql query 
        if (condition instanceof BooleanLiteral)
        {
            if( ((Boolean)((BooleanLiteral)condition).getValue()).booleanValue() )
            {
                return;
            }       
            JavaTypeMapping m = dba.getMapping(Integer.class);
            
            condition = m.newLiteral(this,new Integer("1")).eq(m.newLiteral(this,new Integer("0")));
        }
             
        if (whereExpr == null)
        {
            whereExpr = condition;
        }
        else
        {
            whereExpr = whereExpr.and(condition);
        }
    }

    /**
     * add an condition to the query and queries involved in the union if unionQuery is true
     */
    public void andCondition(BooleanExpression condition,boolean unionQuery)
    {
        assertNotFrozen();

        andCondition(condition);
        
        if (!unionQuery)
        {
            return;
        }

        // add condition to union
        Iterator i = union.iterator();
        while (i.hasNext())
        {
            QueryStatement qs = ((QueryStatement)i.next());
            qs.andCondition(condition);
        }
    }

    public void iorCondition(BooleanExpression condition)
    {
        assertNotFrozen();

        // if condition.toString() == "TRUE" or "FALSE", doesnt make sense or doesnt work pass it to the sql query 
        if (condition instanceof BooleanLiteral)
        {
            if( condition.toStatementText().equals("TRUE") )
            {
                return;
            }       
            JavaTypeMapping m = dba.getMapping(Integer.class);
            
            condition = m.newLiteral(this,new Integer("1")).eq(m.newLiteral(this,new Integer("0")));
        }
             
        if (whereExpr == null)
        {
            whereExpr = condition;
        }
        else
        {
            whereExpr = whereExpr.ior(condition);
        }
    }

    /**
     * add an condition to the query and queries involved in the union if unionQuery is true
     */
    public void iorCondition(BooleanExpression condition,boolean unionQuery)
    {
        assertNotFrozen();

        iorCondition(condition);
        
        if (!unionQuery)
        {
            return;
        }

        // add condition to union
        Iterator i = union.iterator();
        while (i.hasNext())
        {
            QueryStatement qs = ((QueryStatement)i.next());
            qs.iorCondition(condition);
        }
    }    
    /**
     * Mutator for the ordering criteria.
     * @param exprs The expressions to order by
     * @param descending Whether each expression is ascending/descending
     **/
    public void setOrdering(ScalarExpression[] exprs, boolean[] descending)
    {
        assertNotFrozen();

        boolean needsSelect = ((DatabaseAdapter)dba).includeOrderByColumnsInSelect();
        
        orderByList = new StatementText();

        for (int i=0; i<exprs.length; ++i)
        {
            String orderExpr = "JPOXORDER" + i;
            if (i > 0)
            {
                orderByList.append(',');
            }

            if (needsSelect)
            {
                orderByList.append(orderExpr);
            }
            else
            {
                orderByList.append(exprs[i]);
            }

            if (descending[i])
            {
                orderByList.append(" DESC");
            }

            if (needsSelect)
            {
	            // go thought union
	            Iterator iterator = union.iterator();
	            while (iterator.hasNext())
	            {
	                QueryStatement qs = ((QueryStatement)iterator.next());
	                
	                qs.selectScalarExpression(exprs[i].as(orderExpr));
	            }
	            selectScalarExpression(exprs[i].as(orderExpr));
            }
        }
    }

    /**
     * Method to add a range constraint on any SELECT.
     * This typically will use LIMIT/OFFSET where they are supported by
     * the underlying RDBMS.
     * @param offset The offset to start from
     * @param count The number of records to return
     */
    public void setRangeConstraint(long offset, long count)
    {
        this.rangeOffset = offset;
        this.rangeCount = count;
    }

    /**
     * Set this query is to be used as a as set for the Exists function.
     * example WHERE EXISTS( QUERY )
     * @param isExistsSubQuery The isExistsSubQuery to set.
     */
    public void setExistsSubQuery(boolean isExistsSubQuery)
    {
        this.isExistsSubQuery = isExistsSubQuery;
    }
    
    /**
     * Method to convert the criteria into a delete statement text.
     * @return The StatementText
     **/
    public StatementText toDeleteStatementText()
    {
        if (stmtText == null)
        {
            stmtText = new StatementText("DELETE ");

            stmtText.append(" FROM ").append(initialTableExpr);

            Iterator iterator = alias.iterator();

            while (iterator.hasNext())
            {
                stmtText.append(stmtAliasTablesSeparator).append(iterator.next());
            }            
            
            iterator = joins.iterator();

            while (iterator.hasNext())
            {
                stmtText.append(stmtFromTablesSeparator).append(iterator.next());
            }

            if (whereExpr != null)
            {
                stmtText.append(" WHERE ").append(whereExpr);
            }

            if (rangeOffset > -1 || rangeCount > -1)
            {
                // Add a LIMIT clause if it is supported
                stmtText.append(" " + ((DatabaseAdapter)dba).getRangeByLimitClause(rangeOffset, rangeCount));
            }
        }

        return stmtText;
    }
    
    /**
     * Method to convert the criteria into the statement text.
     * @param lock whether to lock the instances using this statement
     * @return The StatementText
     **/
    public StatementText toStatementText(boolean lock)
    {
        if (stmtText == null)
        {
            stmtText = new StatementText("SELECT ");

            if (!isExistsSubQuery)
            {
	            if (distinctResults)
	            {
	                stmtText.append("DISTINCT ");
	            }

	            Iterator iterator = selected.iterator();
	            while (iterator.hasNext())
	            {
	                stmtText.append(iterator.next());
	
	                if (iterator.hasNext())
	                {
	                    stmtText.append(',');
	                }
	            }
                if ((rangeOffset > -1 || rangeCount > -1) && ((DatabaseAdapter)dba).getRangeByRowNumberColumn().length() > 0)
                {
                    // Add a ROW NUMBER column if supported as the means of handling ranges by the RDBMS
                    stmtText.append(',').append(((DatabaseAdapter)dba).getRangeByRowNumberColumn());
                }
            }
            else
            {
                JavaTypeMapping m = dba.getMapping(Integer.class);
                stmtText.append(m.newLiteral(this, new Integer("1")));
            }

            // FROM condition(s)
            stmtText.append(" FROM ").append(initialTableExpr);
            Iterator iterator = alias.iterator();
            while (iterator.hasNext())
            {
                stmtText.append(stmtAliasTablesSeparator).append(iterator.next());
            }

            List sorted = sortJoins(joins);
            Join[] sortedJoins = (Join[]) sorted.toArray(new Join[sorted.size()]);
            for( int i=0; i<sortedJoins.length; i++)
            {
                stmtText.append(stmtFromTablesSeparator).append(sortedJoins[i]);
            }

            // WHERE condition(s)
            if (whereExpr != null)
            {
                stmtText.append(" WHERE ").append(whereExpr);
            }
            if ((rangeOffset > -1 || rangeCount > -1) && ((DatabaseAdapter)dba).getRangeByRowNumberColumn().length() > 0)
            {
                // Add a condition on the row number where the RDBMS supports that means of range specification
                if (whereExpr != null)
                {
                    stmtText.append(" AND ");
                    if (rangeOffset > -1)
                    {
                        stmtText.append(((DatabaseAdapter)dba).getRangeByRowNumberColumn()).append(">=").append("" + rangeOffset);
                    }
                    if (rangeCount > -1)
                    {
                        if (rangeOffset > -1)
                        {
                            stmtText.append(" AND ");
                        }
                        stmtText.append(((DatabaseAdapter)dba).getRangeByRowNumberColumn()).append("<").append("" + (rangeCount - rangeOffset));
                    }
                }
                else
                {
                    stmtText.append(" WHERE ");
                    if (rangeOffset > -1)
                    {
                        stmtText.append(((DatabaseAdapter)dba).getRangeByRowNumberColumn()).append(">=").append("" + rangeOffset);
                    }
                    if (rangeCount > -1)
                    {
                        if (rangeOffset > -1)
                        {
                            stmtText.append(" AND ");
                        }
                        stmtText.append(((DatabaseAdapter)dba).getRangeByRowNumberColumn()).append("<").append("" + (rangeCount - rangeOffset));
                    }
                }
            }

            // GROUP BY clause(s)
            List groupBy = new ArrayList();
            for( int i=0; i<nonAggregateExpression.size(); i++)
            {
                ScalarExpression expr = (ScalarExpression)nonAggregateExpression.get(i);
                //literals are not part of grouping expressions
                if (!(expr instanceof Literal))
                {
                    String exprText;
                    // A column alias that is defined in the select list cannot be used to specify a grouping column
                    if (expr.getAlias() != null)
                    {
                        exprText = expr.getNonAliasExpression();
                    }
                    else
                    {
                        exprText = expr.toStatementText().toString();
                    }
                    if (!groupBy.contains(exprText))
                    {
                        groupBy.add(exprText);
                    }
                }
            }
            if (groupBy.size()>0 && hasAggregateExpression)
            {
                stmtText.append(" GROUP BY ");
                for (int i=0; i<groupBy.size(); i++)
                {
                    stmtText.append((String)groupBy.get(i));
            	    if (i<groupBy.size()-1)
            	    {
            	        stmtText.append(',');
            	    }
                }
            }
            while (iterator.hasNext())
        	{
                stmtText.append(iterator.next());
        	    if (iterator.hasNext())
        	    {
        	        stmtText.append(',');
        	    }
        	}

            // HAVING clause
            if (havingExpr != null)
            {
                stmtText.append(" HAVING ").append(havingExpr);
            }

            // UNION to other query statements
            iterator = union.iterator();
            while (iterator.hasNext())
            {
                if (((DatabaseAdapter)dba).useUnionAll())
                {
                    stmtText.append(" UNION ALL ");
                }
                else
                {
                    stmtText.append(" UNION ");
                }
                QueryStatement qs = ((QueryStatement)iterator.next());
                //do not use for update in other union queries. is this correct for all databases?
                stmtText.append(qs.toStatementText(false));
            }

            // ORDER clause
            if (!isExistsSubQuery)
            {
	            if (orderByList != null)
	            {
	                stmtText.append(" ORDER BY ").append(orderByList);
	            }
            }

            if (rangeOffset > -1 || rangeCount > -1)
            {
                // Add a LIMIT clause if it is supported
                stmtText.append(" " + ((DatabaseAdapter)dba).getRangeByLimitClause(rangeOffset, rangeCount));
            }

            // Add any required locking based on the RDBMS capability
            if (lock && ((DatabaseAdapter)dba).supportsLockWithSelectForUpdate())
            {
                stmtText.append(" FOR UPDATE");
            }
            else if (lock && ((DatabaseAdapter)dba).getSelectWithLockOption() != null)
            {
                stmtText.append(" WITH " + ((DatabaseAdapter)dba).getSelectWithLockOption());
            }
        }

        return stmtText;
    }

    public String toString()
    {
        throw new JDOFatalInternalException("This method must not be called");
    }

    /**
     * Stringifier
     * @param lock Whether to lock the rows found by this SELECT
     * @return The string form of the statement
     */
    public String toString(boolean lock)
    {
        return toStatementText(lock).toString();
    }

    /**
     * A join in a query.
     **/
    protected static class Join
    {
        private final String type;
        private final TableExpression tblExpr;
        private final ScalarExpression expr;
        private final ScalarExpression expr2;

        public Join(String type, ScalarExpression expr, ScalarExpression expr2, TableExpression tblExpr)
        {
            this.type = type;
            this.expr = expr;
            this.expr2 = expr2;
            this.tblExpr = tblExpr;
        }
        
        public String toString()
        {
            return type + " " + tblExpr + " ON " + expr.eq(expr2).toStatementText();
        }
    }

    /**
     * An expression in a query. In RBDMS datastores, an expression 
     * might be a column.
     **/
    public class QueryExpression
    {
        public final TableExpression te;
        public final DatastoreField datastoreField;

        protected QueryExpression(TableExpression te, DatastoreField datastoreField)
        {
            this.te = te;
            this.datastoreField = datastoreField;
        }

        public int hashCode()
        {
            return te.hashCode() ^ datastoreField.hashCode();
        }

        public boolean equals(Object o)
        {
            if (o == this)
            {
                return true;
            }

            if (!(o instanceof QueryExpression))
            {
                return false;
            }

            QueryExpression qsc = (QueryExpression)o;

            return te.equals(qsc.te) && datastoreField.equals(qsc.datastoreField);
        }

        public String toString()
        {
            return te.referenceColumn(datastoreField);
        }
    }
}
