/**********************************************************************
Copyright (c) 2005 Andy Jefferson and others. All rights reserved.
This program and the accompanying materials are made available under
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
    ...
**********************************************************************/
package org.jpox.store.query;

import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.math.BigDecimal;
import java.math.BigInteger;
import java.sql.ResultSet;
import java.sql.SQLException;

import javax.jdo.JDOUserException;

import org.jpox.PersistenceManager;
import org.jpox.store.StatementExpressionIndex;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;
import org.jpox.util.StringUtils;

/**
 * Take a ResultSet, and for each row retrieves an object of a specified type.
 * Follows the rules in JDO2 spec section 14.6.12 regarding the result class.
 * <P>
 * The <B>resultClass</B> will be used to create objects of that type when calling
 * <I>getObject()</I>. The <B>resultClass</B> can be one of the following
 * </P>
 * <UL>
 * <LI>Simple type - String, Long, Integer, Float, Boolean, Byte, Character, Double,
 * Short, BigDecimal, BigInteger, java.util.Date, java.sql.Date, java.sql.Time,
 * java.sql.Timestamp</LI>
 * <LI>Object[]</LI>
 * <LI>User defined type with either a constructor taking the result set fields,
 * or a put(Object,Object) method or setXXX methods, or public fields</LI>
 * </UL>
 * </P>
 * <P>
 * Objects of this class are created in 2 distinct situations. The first is where a
 * candidate class is available, and consequently field position mappings are
 * available. The second is where no candidate class is available and so
 * only the field names are available, and the results are taken in ResultSet order.
 * These 2 modes have their own constructor.
 *
 * @version $Revision: 1.7 $
 */
public class ResultClassROF implements Query.ResultObjectFactory
{
    protected static final Localiser LOCALISER=Localiser.getInstance("org.jpox.store.query.Localisation");

    /** The result class that we should use. */
    private final Class resultClass;

    /** The index of fields position to mapping type. */
    private final StatementExpressionIndex[] statementExpressionIndex;

    /** Names of the result fields. */
    private final String[] resultFieldNames;

    /**
     * Constructor for cases where we have a candidate class and so have mapping
     * information to base field positions on.
     * @param resultClass The result class to use
     * @param statementExpressionIndex The mapping information for field positions etc
     */
    public ResultClassROF(Class resultClass, StatementExpressionIndex[] statementExpressionIndex)
    {
        this.resultClass = resultClass;
        this.statementExpressionIndex = statementExpressionIndex;
        this.resultFieldNames = new String[statementExpressionIndex.length];
        for (int i=0;i<statementExpressionIndex.length;i++)
        {
            if (statementExpressionIndex[i].getMapping().getFieldMetaData() != null)
            {
                resultFieldNames[i] = statementExpressionIndex[i].getMapping().getFieldMetaData().getName();
            }
            else
            {
                resultFieldNames[i] = null;
            }
        }

        if (resultClassIsSimple() && statementExpressionIndex.length != 1)
        {
            String msg = LOCALISER.msg("Query.ResultClassSimpleButMultipleFields", resultClass.getName());
            JPOXLogger.JDO_QUERY.error(msg);
            throw new JDOUserException(msg);
        }
    }

    /**
     * Constructor for cases where we have no candidate class and so have no mapping
     * information to base field positions on. The fields will be retrieved in the
     * ResultSet order.
     * @param resultClass The result class to use
     * @param resultFieldNames Names for the result fields
     */
    public ResultClassROF(Class resultClass, String[] resultFieldNames)
    {
        this.resultClass = resultClass;
        this.statementExpressionIndex = null;
        if (resultFieldNames == null)
        {
            // We use the size of the array, so just allocate a 0-length array
            this.resultFieldNames = new String[0];
        }
        else
        {
            this.resultFieldNames = resultFieldNames;
        }
    }

    /**
     * Method to convert the ResultSet row into an Object.
     * @param rs The ResultSet from the Query.
     * @param pcClass Not used.
     * @return The (persisted) object.
     **/
    public Object getObject(final PersistenceManager pm, final ResultSet rs, Class pcClass)
    {
        // Retrieve the fields from the ResultSet
        Object[] object = null;
        if (statementExpressionIndex != null)
        {
            // Field mapping information available so use it to allocate our results
            object = new Object[statementExpressionIndex.length];
            for (int i=0; i<statementExpressionIndex.length; i++)
            {
                object[i] = statementExpressionIndex[i].getMapping().getObject(pm, rs, statementExpressionIndex[i].getExpressionIndex());
            }
        }
        else
        {
            // Allocate our results in the ResultSet parameter order.
            try
            {
                object = new Object[resultFieldNames.length];
                for (int i=0; i<object.length; i++)
                {
                    object[i] = rs.getObject(i+1);
                }
            }
            catch (SQLException sqe)
            {
                String msg = LOCALISER.msg("SQL.ResultExtractionError", sqe.getMessage());
                JPOXLogger.JDO_QUERY.error(msg);
                throw new JDOUserException(msg);
            }
        }

        // If the user requires Object[] then just give them what we have
        if (resultClass == Object[].class)
        {
            return object;
        }

        // If the user wants a single field and it is the correct type just give them the field
        if (object.length == 1 && (object[0] == null || resultClass.isAssignableFrom(object[0].getClass())))
        {
            return object[0];
        }
        else if (object.length == 1 && !resultClass.isAssignableFrom(object[0].getClass()))
        {
            // Simple object is not assignable to the ResultClass so throw an error
            String msg = LOCALISER.msg("Query.ResultClassSimpleButWrongType",
                resultClass.getName(), object[0].getClass().getName());
            JPOXLogger.JDO_QUERY.error(msg);
            throw new JDOUserException(msg);
        }

        // User requires creation of one of his own type of objects
        if (!resultClassIsSimple())
        {
            // A. Find a constructor with the correct constructor arguments
            Object obj = null;
            try
            {
                Constructor[] constructors = resultClass.getConstructors();
                if (constructors != null)
                {
                    // For each constructor check the parameter types and find one
                    // that aligns with the result set columns (allowing for primitives
                    // and their associated wrappers).
                    for (int i=0;i<constructors.length;i++)
                    {
                        Class[] ctrParams = constructors[i].getParameterTypes();
                        boolean ctrIsValid = true;
                        // Discard any constructor with a different number of params
                        if (ctrParams != null && ctrParams.length == object.length)
                        {
                            for (int j=0;j<ctrParams.length;j++)
                            {
                                // Compare the type with the object or any primitive
                                Class primType = getPrimitiveTypeForType(object[j].getClass());
                                if (ctrParams[j] != object[j].getClass() &&
                                    (primType == null ||
                                     primType != null && ctrParams[j] != primType))
                                {
                                    ctrIsValid = false;
                                    break;
                                }
                            }
                        }
                        else
                        {
                            ctrIsValid = false;
                        }

                        if (ctrIsValid)
                        {
                            try
                            {
                                obj = constructors[i].newInstance(object);
                                return obj;
                            }
                            catch (Exception e)
                            {
                            }
                        }
                    }
                }
            }
            catch (SecurityException se)
            {
                // Cant access the constructors
            }
            if (obj == null)
            {
                Class[] ctr_arg_types = new Class[resultFieldNames.length];
                for (int i=0;i<resultFieldNames.length;i++)
                {
                    ctr_arg_types[i] = object[i].getClass();
                }
                JPOXLogger.JDO_QUERY.debug(LOCALISER.msg("Query.ResultClass.ConstructorNotAvailable",
                    resultClass.getName(), StringUtils.objectArrayToString(ctr_arg_types)));
            }

            // B. No constructor exists so create an object and use put/set methods or update public fields
            try
            {
                // Create the object
                obj = resultClass.newInstance();
            }
            catch (Exception e)
            {
                String msg = LOCALISER.msg("Query.ResultClassNotCreatable",resultClass.getName());
                JPOXLogger.JDO_QUERY.error(msg);
                throw new JDOUserException(msg);
            }
            for (int i=0;i<object.length;i++)
            {
                // Update the fields of our object with the ResultSet values
                if (!setFieldForObject(obj, resultFieldNames[i], object[i]))
                {
                    String msg = LOCALISER.msg("Query.ResultClassHasUnsettableField", resultClass.getName(), resultFieldNames[i]);
                    JPOXLogger.JDO_QUERY.error(msg);
                    throw new JDOUserException(msg);
                }
            }
            return obj;
        }

        // Impossible to satisfy the resultClass requirements so throw exception
        String msg = LOCALISER.msg("Query.ResultClassInvalid",resultClass.getName());
        JPOXLogger.JDO_QUERY.error(msg);
        throw new JDOUserException(msg);
    }

    /**
     * Utility to return if the resultClass is a simple object with a single field.
     * @return Whether it is simple
     */
    private boolean resultClassIsSimple()
    {
        if (resultClass == Boolean.class ||
            resultClass == Byte.class ||
            resultClass == Character.class ||
            resultClass == Double.class ||
            resultClass == Float.class ||
            resultClass == Integer.class ||
            resultClass == Long.class ||
            resultClass == Short.class ||
            resultClass == String.class ||
            resultClass == BigDecimal.class ||
            resultClass == BigInteger.class ||
            resultClass == java.util.Date.class ||
            resultClass == java.sql.Date.class ||
            resultClass == java.sql.Time.class ||
            resultClass == java.sql.Timestamp.class ||
            resultClass == Object.class)
        {
            return true;
        }
        else
        {
            return false;
        }
    }

    /**
     * Method to return the primitive equivalent of the specified type (if any)
     * @param type The type
     * @return The primitive equivalent.
     */
    private Class getPrimitiveTypeForType(Class type)
    {
        if (type == Boolean.class)
        {
            return boolean.class;
        }
        else if (type == Byte.class)
        {
            return byte.class;
        }
        else if (type == Character.class)
        {
            return char.class;
        }
        else if (type == Double.class)
        {
            return double.class;
        }
        else if (type == Float.class)
        {
            return float.class;
        }
        else if (type == Integer.class)
        {
            return int.class;
        }
        else if (type == Long.class)
        {
            return long.class;
        }
        else if (type == Short.class)
        {
            return short.class;
        }
        else
        {
            return null;
        }
    }

    /**
     * Method to set a field of an object, using set/put methods, or via a public field.
     * @param obj The object to update the field for
     * @param fieldName Name of the field
     * @param value The value to apply
     * @return Whether it was updated
     */
    private boolean setFieldForObject(Object obj, String fieldName, Object value)
    {
        boolean fieldSet = false;

        // Try setMethod()
        if (!fieldSet)
        {
            String setMethodName = "set" + fieldName.substring(0,1).toUpperCase() + fieldName.substring(1);
            try
            {
                Method m = obj.getClass().getDeclaredMethod(setMethodName, new Class[]{value.getClass()});
                m.invoke(obj, new Object[]{value});
                fieldSet = true;
            }
            catch (Exception e)
            {
                Class primitive = getPrimitiveTypeForType(value.getClass());
                if (primitive != null)
                {
                    try
                    {
                        Method m = obj.getClass().getDeclaredMethod(setMethodName, new Class[]{primitive});
                        m.invoke(obj, new Object[]{value});
                        fieldSet = true;
                    }
                    catch (Exception e2)
                    {
                    }
                }
                if (!fieldSet)
                {
                    JPOXLogger.JDO_QUERY.debug(LOCALISER.msg("Query.ResultClass.SetMethodNotAvailable", 
                        resultClass.getName(), setMethodName, value.getClass().getName()));
                }
            }
        }

        // Try putMethod()
        if (!fieldSet)
        {
            String putMethodName = "put";
            try
            {
                Method m = obj.getClass().getDeclaredMethod(putMethodName, new Class[]{Object.class, Object.class});
                m.invoke(obj, new Object[]{fieldName, value});
                fieldSet = true;
            }
            catch (Exception e)
            {
                if (!fieldSet)
                {
                    JPOXLogger.JDO_QUERY.debug(LOCALISER.msg("Query.ResultClass.PutMethodNotAvailable", 
                        resultClass.getName(), putMethodName));
                }
            }
        }

        // Try setting the (public) field directly
        if (!fieldSet)
        {
            try
            {
                Field f = obj.getClass().getDeclaredField(fieldName);
                f.set(obj, value);
                fieldSet = true;
            }
            catch (Exception e)
            {
                JPOXLogger.JDO_QUERY.debug(LOCALISER.msg("Query.ResultClass.PublicFieldNotAvailable", 
                    resultClass.getName(), fieldName));
            }
        }

        return fieldSet;
    }
}