/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2005 Andy Jefferson - changed to use ResultClassROF
    ...
**********************************************************************/
package org.jpox.store.query;

import java.util.Collection;
import java.util.Iterator;

import javax.jdo.JDOFatalInternalException;

import org.jpox.PersistenceManager;
import org.jpox.store.DatastoreClass;
import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.StatementExpressionIndex;
import org.jpox.store.StoreManager;
import org.jpox.store.expression.BooleanExpression;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.query.Query.ResultObjectFactory;
import org.jpox.util.Localiser;

/**
 * Result expressions to be returned by the query. The results might be fields
 * of persistent instances, instances of classes other than the candidate class,
 * or aggregates of fields.
 * @version $Revision: 1.7 $
 */
public class ResultExpressionsQueryable implements Queryable
{
    protected static final Localiser LOCALISER=Localiser.getInstance("org.jpox.store.query.Localisation");
    
    private final Class candidateClass;
    
    private final Collection userCandidates;

    private ScalarExpression[] expressions;
    private final StoreManager storeMgr;
    private final boolean subclasses;

    /**
     * Constructor.
     * @param pm Persistence Manager
     * @param candidateClass The candidate
     * @param subclasses Whether to include subclasses
     */
    public ResultExpressionsQueryable(PersistenceManager pm, Class candidateClass, boolean subclasses)
    {
        this.candidateClass = candidateClass;
        this.storeMgr = pm.getStoreManager();
        this.subclasses = subclasses;
        this.userCandidates = null;
    }

    /**
     * Constructor.
     * @param pm Persistence Manager
     * @param candidateClass The candidate
     * @param userCandidates The users candidates
     * @param subclasses Whether to include subclasses
     */
    public ResultExpressionsQueryable(PersistenceManager pm, Class candidateClass, Collection userCandidates, boolean subclasses)
    {
        this.userCandidates = userCandidates;
        this.candidateClass = candidateClass;
        this.storeMgr = pm.getStoreManager();
        this.subclasses = subclasses;
    }
    
    /**
     * Create a new query to search for the candidate classes etc.
     * @return The new QueryStatement.
     */
    public QueryStatement newQueryStatement()
    {
        return newQueryStatement(candidateClass);
    }

    /**
     * @param expressions The expressions to set.
     */
    public void setResultExpressions(ScalarExpression[] expressions)
    {
        this.expressions = expressions;
    }
    
    /**
     * Create a query to search for the candidateClass and subclasses if true
     * @param candidateClass The candidate
     * @return The new QueryStatement
     */
    public QueryStatement newQueryStatement(Class candidateClass)
    {
        final DatastoreClass datastoreClass = storeMgr.getDatastoreClass(candidateClass.getName());
        final JavaTypeMapping m = datastoreClass.getIDMapping();
        final Class type = candidateClass;
        final boolean subclasses = this.subclasses;
        QueryStatement stmt = new IteratorStatement(
                new IteratorStatement.AssociationEnd()
                {
                    public JavaTypeMapping getMapping()
                    {
                        return m;
                    }
                    public Class getType()
                    {
                        return type;
                    }
                    public DatastoreContainerObject getDatastoreContainerObject()
                    {
                        return datastoreClass;
                    }
                    public boolean useJoin()
                    {
                        return false;
                    }                    
                },
                candidateClass,
                storeMgr,
                storeMgr.getDatastoreAdapter(),
                subclasses,
                false).getQueryStatement();
        
        if (userCandidates != null)
        {
            /*
             * creates a query like WHERE ... AND (CANDIDATE_ID = ?1 OR CANDIDATE_ID =
             * ?2) or for classes with composite primary keys WHERE ... AND (
             * (CANDIDATE_IDa = ?1a AND CANDIDATE_IDb = ?1b) OR (CANDIDATE_IDa = ?2a
             * AND CANDIDATE_IDb = ?2b) )
             */
            BooleanExpression elementsExpr = null;
            for (Iterator it = userCandidates.iterator(); it.hasNext();)
            {
                Object candidateValue = it.next();
                BooleanExpression keyExpr = null;
                ScalarExpression expr = m.newScalarExpression(stmt, stmt.getDefaultTableExpression());
                if (keyExpr == null)
                {
                    keyExpr = expr.eq(m.newLiteral(stmt, candidateValue));
                }
                else
                {
                    keyExpr = keyExpr.and(expr.eq(m.newLiteral(stmt, candidateValue)));
                }
                if (elementsExpr == null)
                {
                    elementsExpr = keyExpr;
                }
                else
                {
                    elementsExpr = elementsExpr.ior(keyExpr);
                }
            }
            if (elementsExpr != null)
            {
                stmt.andCondition(elementsExpr, true);
            }            
        }
        return stmt;
    }

    /* (non-Javadoc)
     * @see org.jpox.store.query.Queryable#newResultObjectFactory(org.jpox.store.QueryStatement, boolean)
     */
    public ResultObjectFactory newResultObjectFactory(QueryStatement stmt, boolean ignoreCache, Class resultClass)
    {
        StatementExpressionIndex[] statementExpressionIndex = new StatementExpressionIndex[expressions.length];
        for (int i=0; i<expressions.length; i++)
        {
            if (expressions[i].getMapping() == null)
            {
                throw new JDOFatalInternalException(LOCALISER.msg("ResultExpressions.UnsupportedExpressionInResult",expressions[i]));
            }
            statementExpressionIndex[i] = new StatementExpressionIndex();
            ScalarExpression[] expr = expressions[i].getExpressionList().toArray();
            int[] exprIndex = new int[expr.length];
            for(int j=0; j<expr.length; j++)
            {
                exprIndex[j] = stmt.selectScalarExpression(expr[j],true);
            }
            statementExpressionIndex[i].setExpressionIndex(exprIndex);
            statementExpressionIndex[i].setMapping(expressions[i].getMapping());
        }

        // Provide a ResultObjectFactory to convert to the required type
        if (resultClass != null)
        {
            return new ResultClassROF(resultClass, statementExpressionIndex);
        }
        else if (statementExpressionIndex.length == 1)
        {
            return new ResultClassROF(Object.class, statementExpressionIndex);
        }
        else
        {
            return new ResultClassROF(Object[].class, statementExpressionIndex);
        }
    }

    /**
     * Whether to retrieve distinct values 
     * @param stmt The Query Statement
     * @param distinct The distinct to set.
     */
    public void setDistinctResults(QueryStatement stmt, boolean distinct)
    {
        stmt.setDistinctResults(true);
    }
    
    /**
     * Returns <tt>true</tt> if this collection contains no elements.<p>
     * @return <tt>true</tt> if this collection contains no elements.
     */
    public boolean isEmpty()
    {
        if (userCandidates != null)
        {
            return userCandidates.isEmpty();            
        }
        return false;
    }
}