/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2004 Andy Jefferson - commented and localised
2004 Andy Jefferson - changed to allow for null candidate class
2005 Andy Jefferson - changed to use JDBC parameters
2005 Andy Jefferson - added timeout support
    ...
**********************************************************************/
package org.jpox.store.query;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.ResultSetMetaData;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOFatalUserException;
import javax.jdo.JDOUserException;

import org.jpox.PersistenceManager;
import org.jpox.exceptions.ClassNotPersistenceCapableException;
import org.jpox.metadata.ClassMetaData;
import org.jpox.metadata.FieldMetaData;
import org.jpox.metadata.FieldPersistenceModifier;
import org.jpox.metadata.IdentityType;
import org.jpox.store.Role;
import org.jpox.store.StatementExpressionIndex;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.MappingCallbacks;
import org.jpox.store.rdbms.sqlidentifier.ColumnIdentifier;
import org.jpox.store.exceptions.PersistentSuperclassNotAllowedException;

/**
 * A Query using SQL, and keeping to the JDO2 definition of a SQL query.
 * The majority of this has to be specified in the query filter itself.
 * Parameters are embodied in the query and are unnamed/untyped. There
 * are no variables/imports. Ordering/grouping is explicit in the query.
 *
 * @see Query
 * @version $Revision: 1.21 $
 */
public final class SQLQuery extends BaseSQLQuery
{
    /**
     * Constructs a new query instance having the same criteria as the given
     * query.
     *
     * @param pm The PersistenceManager
     * @param sql_text The JPOX SQL query string
     */
    public SQLQuery(PersistenceManager pm, String sql_text)
    {
        super(pm, sql_text);
    }

    /**
     * Equality operator
     * @param obj The object to compare against
     * @return Whether they are equal
     **/
    public boolean equals(Object obj)
    {
        if (obj == this)
        {
            return true;
        }

        if (!(obj instanceof SQLQuery) || !super.equals(obj))
        {
            return false;
        }

        return sqlText.equals(((SQLQuery)obj).sqlText);
    }

    /**
     * Declare the parameters to be used in the query.
     * <p>This implementation always throws a JDOUserException since this
     * concept doesn't apply to SQL queries.
     *
     * @param parameters the parameters separated by semicolons.
     * @exception JDOUserException Always thrown.
     * @see javax.jdo.Query#declareVariables
     */
    public void declareParameters(String parameters)
    {
        throw new JDOUserException(LOCALISER.msg("JDOSQL.ParametersInapplicableError"));
    }

    /**
     * Set the import statements to be used to identify the fully qualified
     * name of variables or parameters.
     * <p>This implementation always throws a JDOUserException since this
     * concept doesn't apply to SQL queries.
     * 
     * @param imports   import statements separated by semicolons.
     * @exception JDOUserException  Always thrown.
     * @see javax.jdo.Query#declareImports
     */
    public void declareImports(String imports)
    {
        throw new JDOUserException(LOCALISER.msg("JDOSQL.ImportInapplicableError"));
    }

    /**
     * Method to perform any necessary pre-processing on the users query statement
     * before we execute it. SQL queries are not modified in any way, as per JDO2 spec
     * section 14.7.
     */
    protected void generateQueryStatement()
    {
        jdbcSqlText = sqlText; // Just take the users SQL query direct - no substitution of params etc.

        // Perform any sanity checking on the users input
        if (candidateClass != null)
        {
            ClassMetaData cmd = pm.getMetaDataManager().getMetaDataForClass(candidateClass);
            if (cmd == null)
            {
                throw new ClassNotPersistenceCapableException(candidateClass.getName());
            }
            if (cmd.getPersistenceCapableSuperclass() != null)
            {
                throw new PersistentSuperclassNotAllowedException(candidateClass.getName());
            }
        }
    }

    /**
     * Execute the query and return the filtered List.
     * Overrides the version in Query so that we can allow for not having
     * parameter names in SQL.
     * @param parameters the Object array with all of the parameters.
     * @return the filtered List.
     * @see javax.jdo.Query#executeWithArray(Object[])
     */
    public Object executeWithArray(Object[] parameters)
    {
        isPreCompile = false;
        compile();

        // Insert dummy parameter names that are just "1", "2", etc
        HashMap parameterMap = new HashMap();
        if (parameters != null)
        {
            for (int i = 0; i < parameters.length; ++i)
            {
                String name = "" + (i+1);
                parameterNames.add(name);
                parameterMap.put(name, parameters[i]);
            }
        }

        Object theReturn = executeWithMap(parameterMap);
        isPreCompile = true;
        return theReturn; 
    }

    /**
     * Execute the query and return the filtered QueryResult.
     * @param parameters the Map containing all of the parameters.
     * @return the filtered QueryResult.
     */
    public Collection performExecute(Map parameters)
    {
        compile();

        if (parameters.size() != parameterNames.size())
        {
            throw new JDOUserException(LOCALISER.msg("JDOSQL.IncorrectNumberOfParametersError","" + parameterNames.size(),"" + parameters.size()));
        }

        QueryResult qr = null;
        try
        {
            Connection conn = pm.getConnection(false);

            try
            {
                PreparedStatement ps = conn.prepareStatement(jdbcSqlText);
                try
                {
                    // Prepare the field-column mappings in the statement
                    fieldColumnNames = new ArrayList();

                    // Where we have a candidate class we can set up the link between column name and field.
                    if (candidateClass != null)
                    {
                        // Candidate class should be any PC class, and the query should contain the
                        // PK fields as a minimum.
                        final ClassMetaData candidateCmd = pm.getMetaDataManager().getMetaDataForClass(candidateClass);

                        // TODO Remove this - not of any relevance
                        if (candidateCmd.isRequiresExtent())
                        {
                            throw new JDOUserException(LOCALISER.msg("JDOSQL.InvalidCandidateClassExtentError",candidateClass.getName()));
                        }
                        // TODO Remove this - should allow ANY identity type
                        if (candidateCmd.getIdentityType() != IdentityType.NONDURABLE)
                        {
                            throw new JDOUserException(LOCALISER.msg("JDOSQL.InvalidCandidateClassNondurableError",candidateClass.getName()));
                        }

                        // Create an index listing for ALL fields in the result class. This is wrong
                        // since the user may want to only populate some of them
                        int fieldCount = candidateCmd.getNoOfManagedFields();
                        int[] fn = new int[fieldCount];
                        statementExpressionIndex = new StatementExpressionIndex[fieldCount];
                        int n = 0;

                        for (int fieldNumber = 0; fieldNumber < fieldCount; ++fieldNumber)
                        {
                            statementExpressionIndex[fieldNumber] = new StatementExpressionIndex();
                            FieldMetaData fmd = candidateCmd.getManagedField(fieldNumber);
                            String fieldName = fmd.getName();
                            Class fieldType = fmd.getType();

                            // TODO Candidate can be any PC class, so this check should be removed
                            if (fmd.getPersistenceModifier() != FieldPersistenceModifier.PERSISTENT)
                            {
                                JavaTypeMapping m = dba.getMapping(fieldType, getStoreManager());
                                if (!m.includeInFetchStatement() || m instanceof MappingCallbacks)
                                {
                                    throw new JDOFatalUserException("Mapping " + m +
                                        " not suitable for a SQL result column, field = " + fieldName);
                                }

                                statementExpressionIndex[fieldNumber].setMapping(m);
                                fn[n++] = fieldNumber;

                                String columnName = null;
                                if (fmd.getColumnMetaData() != null && fmd.getColumnMetaData().length > 0)
                                {
                                    columnName = fmd.getColumnMetaData()[0].getName();
                                }
                                else
                                {
                                    columnName = new ColumnIdentifier(dba, fieldName, fieldType, Role.NONE).getSQLIdentifier();
                                }
                                fieldColumnNames.add(columnName);
                            }
                            else if (fmd.getPersistenceModifier() != FieldPersistenceModifier.TRANSACTIONAL)
                            {
                                throw new JDOFatalInternalException("Invalid persistence modifier on field " + fieldName);
                            }
                        }

                        if (n == 0)
                        {
                            throw new JDOFatalUserException("View class has no persistent fields: " + candidateClass.getName());
                        }

                        fieldNumbers = new int[n];
                        System.arraycopy(fn, 0, fieldNumbers, 0, n);
                    }

                    // Set the values of any parameters
                    if (parameters != null)
                    {
                        for (int i=0;i<parameters.size();i++)
                        {
                            Object obj = parameters.get("" +(i+1));
                            ps.setObject((i+1), obj);
                        }
                    }

                    // Apply any user-specified timeout
                    if (timeout > 0)
                    {
                        ps.setQueryTimeout(timeout);
                    }

                    // Apply any fetch size
                    ps.setFetchSize(pm.getPMFContext().getPmfConfiguration().getQueryFetchSize());

                    // Apply any fetch direction
                    String fetchDir = pm.getPMFContext().getPmfConfiguration().getQueryFetchDirection();
                    if (fetchDir.equals("reverse"))
                    {
                        ps.setFetchDirection(ResultSet.FETCH_REVERSE);
                    }
                    else if (fetchDir.equals("unknown"))
                    {
                        ps.setFetchDirection(ResultSet.FETCH_UNKNOWN);
                    }

                    ResultSet rs = executeQuery(jdbcSqlText, ps);
                    try
                    {
                        if (candidateClass != null)
                        {
                            // User has specified a candidate class so map the fields to the ResultSet columns
                            ResultSetMetaData rsmd = rs.getMetaData();
                            HashSet remainingColumnNames = new HashSet(fieldColumnNames);
                            int colCount = rsmd.getColumnCount();
                            for (int colNum = 1; colNum <= colCount; ++colNum)
                            {
                                String colName = rsmd.getColumnName(colNum);
                                int fieldNumber = fieldColumnNames.indexOf(colName);
                                if (fieldNumber >= 0)
                                {
                                    statementExpressionIndex[fieldNumber].setExpressionIndex(new int[]{colNum});
                                    remainingColumnNames.remove(colName);
                                }
                            }

                            if (resultClass != null)
                            {
                                qr = new QueryResult(null, this, new ResultClassROF(resultClass, statementExpressionIndex), rs);
                            }
                            else
                            {
                                if (!remainingColumnNames.isEmpty())
                                {
                                    throw new JDOUserException(LOCALISER.msg("JDOSQL.ResultSetMissingColumnsError",remainingColumnNames));
                                }
                                qr = new QueryResult(null, this, new TransientIDROF(candidateClass, fieldNumbers, statementExpressionIndex), rs);
                            }
                        }
                        else
                        {
                            // No candidate class, so use resultClass or Object/Object[]
                            Class requiredResultClass = resultClass;
                            int numberOfColumns = 0;
                            String[] resultFieldNames = null;
                            try
                            {
                                ResultSetMetaData rsmd = rs.getMetaData();
                                numberOfColumns = rsmd.getColumnCount();
                                if (requiredResultClass == null)
                                {
                                    if (numberOfColumns == 1)
                                    {
                                        requiredResultClass = Object.class;
                                    }
                                    else
                                    {
                                        requiredResultClass = Object[].class;
                                    }
                                }

                                // Generate names to use for the fields based on the column names
                                resultFieldNames = new String[numberOfColumns];
                                for (int i=0;i<numberOfColumns;i++)
                                {
                                    resultFieldNames[i] = rsmd.getColumnName(i+1).toLowerCase();
                                }
                            }
                            catch (SQLException sqe)
                            {
                                // Do nothing
                            }

                            qr = new QueryResult(null, this, new ResultClassROF(requiredResultClass, resultFieldNames), rs);
                        }
                    }
                    finally
                    {
                        if (qr == null)
                        {
                            rs.close();
                        }
                    }
                }
                finally
                {
                    if (qr == null)
                    {
                        ps.close();
                    }
                }
            }
            finally
            {
                pm.releaseConnection(conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("JDOSQL.ExecutionError",jdbcSqlText),e);
        }

        queryResults.add(qr);

        return qr;
    }
}