/**********************************************************************
Copyright (c) 2002 Kelly Grizzle (TJDO) and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
2003 Erik Bengtson - added a replaceAll operation in substitution for
                    java.lang.String.replaceAll of J2SDK 1.4. Now we
                    can compile with J2SDK 1.3.1
2003 Erik Bengtson - moved replaceAll to StringUtils class
2003 Andy Jefferson - updated the DEFAULT_MAX_STRING_LENGTH to be controllable
                     via a property.
2004 Andy Jefferson - updated the NULL/NOT NULL so that we always specify the 
                     nullability of a column on creation
2004 Erik Bengtson - added columnOptions getters that can be read by
                    the SetTable or MapTable (Collection). columnOptions is
                    used for create an equivalent column on the set table
2004 Andy Jefferson - added Auto, Unique, Default, options to column SQL options
    ...
**********************************************************************/
package org.jpox.store.rdbms;

import java.sql.DatabaseMetaData;
import java.util.StringTokenizer;

import javax.jdo.JDOUserException;

import org.jpox.metadata.ColumnDefinitionException;
import org.jpox.metadata.ColumnMetaData;
import org.jpox.store.DatastoreAdapter;
import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.DatastoreField;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.StoreManager;
import org.jpox.store.exceptions.IncompatibleDataTypeException;
import org.jpox.store.exceptions.WrongPrecisionException;
import org.jpox.store.exceptions.WrongScaleException;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.rdbms.adapter.DatabaseAdapter;
import org.jpox.store.rdbms.columninfo.ColumnInfo;
import org.jpox.store.rdbms.sqlidentifier.SQLIdentifier;
import org.jpox.store.rdbms.table.Table;
import org.jpox.store.rdbms.table.TableImpl;
import org.jpox.store.rdbms.typeinfo.TypeInfo;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;
import org.jpox.util.StringUtils;

/**
 * Definition of a Column in the datastore. Contains the full definition of the
 * column, its type, size, whether it is autoincrement, nullable, etc.
 * The SQL column definition is generated here.
 *
 * @version $Revision: 1.13 $
 **/ 
public class Column implements DatastoreField
{
    private static final Localiser LOCALISER=Localiser.getInstance("org.jpox.store.rdbms.Localisation");
	private static final Localiser LOCALISEREXC=Localiser.getInstance("org.jpox.store.exceptions.Localisation");
 
    public static final int LENGTH_NOT_SET    = 0;
    public static final int FIXED_LENGTH      = 1;
    public static final int MAXIMUM_LENGTH    = 2;
    public static final int UNLIMITED_LENGTH  = 3;

    private static final int LENGTH_TYPE_MASK = 3;
    private static final int PRIMARY_KEY_PART = 4;
    private static final int EXACT_PRECISION  = 8;
    private static final int NULLABLE         = 16;
    private static final int UNIQUE           = 32;

    /** store the columnOptions for further use */
    private ColumnMetaData columnOptions;

    protected final Table table;
    protected final Class type;
    protected final SQLIdentifier name;
    protected final StoreManager storeMgr;
    protected final DatastoreAdapter dba;
    protected final JavaTypeMapping mapping;

    protected TypeInfo typeInfo;
    protected String constraints;
    protected int flags;

    /** Default value to use when constructing the column, in the DEFAULT clause */
    protected String columnDefaultValue;

    /** Default value accepted by the datastore for this column, from DatabaseMetaData. */
    protected Object defaultValue;

    /** Whether this column should be defaulted on inserts. */
    protected boolean defaultable = false;

    /** Whether this is to use AUTOINCREMENT. */
    protected boolean autoIncrement = false;

    /*
     * a function wrapping the column 
     * example: SQRT(COLUMN) 
     */
    protected String[] wrapperFunction;

    public static final int WRAPPER_FUNCTION_SELECT = 0;
    public static final int WRAPPER_FUNCTION_INSERT = 1;
    public static final int WRAPPER_FUNCTION_UPDATE = 2;

    /**
     * Constructor.
     * @param table The table in the datastore that this column belongs to.
     * @param type The type of data being stored in this column
     * @param name The name of the column (in the datastore).
     **/
    public Column(Table table, Class type, SQLIdentifier name,JavaTypeMapping mapping,ColumnMetaData co)
    {
        this.table    = table;
        this.type     = type;
        this.name     = name;
        this.storeMgr = table.getStoreManager();
        this.dba      = storeMgr.getDatastoreAdapter();
		this.mapping  = mapping;
		typeInfo    = null;
		constraints = null;
		flags       = 0;
		setColumnMetaData(co);
		wrapperFunction = new String[3];
		wrapperFunction[WRAPPER_FUNCTION_SELECT]= "?";
		wrapperFunction[WRAPPER_FUNCTION_INSERT]= "?";
		wrapperFunction[WRAPPER_FUNCTION_UPDATE]= "?";
    }

    /**
     * Accessor for the DatastoreObject. The table owning this Column.
     * @return The DatastoreObject( The table )
     */
    public DatastoreContainerObject getDatastoreContainerObject()   
    {
        return table;
    }

    /**
     * Accessor for the name of the datastore column.
     * @return The name of the column in the datastore.
     **/
    public DatastoreIdentifier getName()
    {
        return name;
    }

    /**
     * Accessor for the type of data stored in this column.
     * @return The type of data in the column
     **/
    public Class getType()
    {
        return type;
    }

    /**
     * Accessor for the precision of data in the column.
     * @return The precision of data in the column
     **/
    private int getSQLPrecision()
    {
        int sqlPrecision = - 1;

        if (getLengthType() == UNLIMITED_LENGTH)
        {
            int ulpv = ((DatabaseAdapter) dba).getUnlimitedLengthPrecisionValue(typeInfo);

            if (ulpv > 0)
            {
                sqlPrecision = ulpv;
            }
        }
        else
        {
            if (columnOptions.getPrecision() != null)
            {
                sqlPrecision = columnOptions.getPrecision().intValue();
            }
            else if (columnOptions.getLength() != null)
            {
                sqlPrecision = columnOptions.getLength().intValue();
            }
        }

        /*
         * Databases like Cloudscape that use BIT types for binary need to
         * have the length expressed in bits, not bytes.
         */
        if (typeInfo.typeName.toLowerCase().startsWith("bit"))
        {
            return sqlPrecision * 8;
        }
        else
        {
            return sqlPrecision;
        }
    }

    /**
     * Accessor for the SQL definition of this column.
     * @return The SQL definition of the column
     **/
    public String getSQLDefinition()
    {
        StringBuffer def = new StringBuffer(name.toString());
        StringBuffer typeSpec = new StringBuffer(typeInfo.typeName);

        // Add type specification.
        // Parse and append createParams to the typeName if it looks like it's
        // supposed to be appended, i.e. if it contains parentheses, and the type name
        // itself doesn't. createParams is mighty ill-defined by the JDBC spec, but attempt to
        // interpret it.
        if (typeInfo.createParams != null && typeInfo.createParams.indexOf('(') >= 0 && typeInfo.typeName.indexOf('(') < 0)
        {
            StringTokenizer toks = new StringTokenizer(typeInfo.createParams);

            while (toks.hasMoreTokens())
            {
                String tok = toks.nextToken();

                if (tok.startsWith("[") && tok.endsWith("]"))
                {
                    // The brackets look like they indicate an optional param so
                    // skip
                    continue;
                }

                typeSpec.append(" " + tok);
            }
        }

        // Add any precision - note that there is no obvious flag in the JDBC typeinfo to
        // tell us whether the type allows this specification or not, and many JDBC just
        // return crap anyway. We use the allowsPrecisionSpec flag for this
        StringBuffer precSpec = new StringBuffer();
        int sqlPrecision = getSQLPrecision();
        if (sqlPrecision > 0 && typeInfo.allowsPrecisionSpec)
        {
            precSpec.append(sqlPrecision);
            if (columnOptions.getScale() != null)
            {
                precSpec.append("," + columnOptions.getScale());
            }
        }
        else if (sqlPrecision > 0 && !typeInfo.allowsPrecisionSpec)
        {
            JPOXLogger.RDBMS_DDL.warn(LOCALISER.msg("Column.FieldPrecisionInvalid", this.toString()));
        }
        int lParenIdx = typeSpec.toString().indexOf('(');
        int rParenIdx = typeSpec.toString().indexOf(')', lParenIdx);
        if (lParenIdx > 0 && rParenIdx > 0)
        {
            // Some databases (like DB2) give you typeNames with ()'s already
            // present ready for you to insert the values instead of appending them.
            if (precSpec.length() > 0)
            {
                typeSpec.replace(lParenIdx + 1, rParenIdx, precSpec.toString());
            }
            else if (rParenIdx == lParenIdx + 1)
            {
                throw new ColumnDefinitionException(LOCALISER.msg("Column.PrecisionError", this.toString()));
            }
        }
        else if (precSpec.length() > 0)
        {
            typeSpec.append('(');
            typeSpec.append(precSpec.toString());
            typeSpec.append(')');
        }
        def.append(" " + typeSpec.toString());

        // DEFAULT to impose on column
        // TODO Allow for user specification of "null" to symbolise "DEFAULT NULL"
        DatabaseAdapter adapter = (DatabaseAdapter) dba;
        if (columnDefaultValue != null && adapter.supportsDefaultKeywordInColumnOptions())
        {
            // Quote any character types (CHAR, VARCHAR, BLOB, CLOB)
            if (typeInfo.typeName.toUpperCase().indexOf("CHAR") >= 0 ||
                typeInfo.typeName.toUpperCase().indexOf("LOB") >= 0)
            {
                // NOTE We use single quote here but would be better to take
                // some character from the DatabaseMetaData. The "identifierQuoteString"
                // does not work for string quoting here for Postgres
                def.append(" DEFAULT '" + columnDefaultValue + "'");
            }
            else
            {
                def.append(" DEFAULT " + columnDefaultValue);
            }
        }

        // Nullability
        if (adapter.supportsAutoIncrementKeys() && autoIncrement && !adapter.supportsAutoIncrementKeysNullSpecification())
        {
            // Do nothing since the adapter doesn't allow NULL specifications with autoincrement/identity
        }
        else
        {
            if (!isNullable())
            {
                if( columnDefaultValue == null || adapter.supportsDefaultKeywordWithNotNullInColumnOptions() )
                {
                    def.append(" NOT NULL");
                }
            }
            else if (typeInfo.nullable == DatabaseMetaData.columnNullable)
            {
                if (adapter.supportsNullsKeywordInColumnOptions())
                {
                    def.append(" NULL");
                }
            }
        }

        // Constraints checks
        if (adapter.supportsCheckInCreateStatements() && constraints != null)
        {
            def.append(" " + constraints.toString());
        }

        // Auto Increment
        if (adapter.supportsAutoIncrementKeys() && autoIncrement)
        {
            def.append(" " + adapter.getAutoIncrementKeyword());
        }

        // Uniqueness
        if (isUnique() && !adapter.supportsUniqueConstraintsInEndCreateStatements())
        {
            def.append(" UNIQUE");
        }

        return def.toString();
    }

    /**
     * Accessor for the SQL Column Namedefinition of this column.
     * @return The SQL definition of the column
     */
    public String getSQLColumnNameDefinition()
    {
        StringBuffer def = new StringBuffer(name.toString());
        return def.toString();
    }

    /**
     * Initialize the default column value
     * @param ci The column information
     */
    public void initializeDefault(ColumnInfo ci)
    {
        String column_default = ci.getColumnDef();
        if (getType() == String.class)
        {
            if (column_default != null)
            {
                setDefault(StringUtils.replaceAll(
                    StringUtils.replaceAll(
                        StringUtils.replaceAll(
                            StringUtils.replaceAll(column_default, "'", ""), "\"", ""),
                    ")", ""), "(", ""));
            }
        }
        else
        {
            if (column_default != null)
            {
                setDefault(StringUtils.replaceAll(
                    StringUtils.replaceAll(
                        StringUtils.replaceAll(
                            StringUtils.replaceAll(column_default, "'", ""), "\"", ""),
                    ")", ""), "(", ""));
            }
        }
        // TODO Make sure that this lines up with the defaultValue when set.
    }

    /**
     * Method to validate the contents of the column. This method can throw
     * IncompatibleDataTypeException, WrongScaleException,
     * WrongPrecisionException, IsNullableException if the data in the column is
     * not compatible with the supplied ColumnInfo.
     * @param ci The column information
     */
    public void validate(ColumnInfo ci)
    {
        if (!typeInfo.isCompatibleWith(ci))
        {
            throw new IncompatibleDataTypeException(this, typeInfo.dataType, ci.getDataType());
        }
        if (!typeInfo.isToValidate(ci))
        {
            return;
        }
        if (table instanceof TableImpl)
        {
            // Only validate precision/scale/nullability for tables, not for
            // views
            int actualPrecision = ci.getColumnSize();
            int actualScale = ci.getDecimalDigits();
            String actualIsNullable = ci.getIsNullable();

            int sqlPrecision = getSQLPrecision();

            if (sqlPrecision > 0 && actualPrecision > 0)
            {
                if (sqlPrecision != actualPrecision)
                {
                    throw new WrongPrecisionException(this.toString(), sqlPrecision, actualPrecision);
                }
            }

            if (columnOptions.getScale() != null && actualScale >= 0)
            {
                if (columnOptions.getScale().intValue() != actualScale)
                {
                    throw new WrongScaleException(this.toString(), columnOptions.getScale().intValue(), actualScale);
                }
            }

            if (actualIsNullable.length() > 0)
            {
                switch (Character.toUpperCase(actualIsNullable.charAt(0)))
                {
                    case 'Y' :
                        if (!isNullable())
                        {
                            JPOXLogger.RDBMS.warn(LOCALISEREXC.msg("Exception.IsNullable", this));
                        }
                        break;

                    case 'N' :
                        // TODO convert to default value somewhere at runtime if
                        // column has a default value and "default"
                        break;

                    default :
                        break;
                }
            }
        }
    }

    /**
     * Method to set the options of this column. This allows setting of the
     * length, precision, scale etc.
     * @param co The column options containing the length etc.
     */
    public void setColumnMetaData(ColumnMetaData co)
    {
        if (co == null)
        {
            co = new ColumnMetaData(null, null);
        }
        else
        {
            co = new ColumnMetaData(
                null, 
                co.getName(), 
                co.getTarget(), 
                co.getTargetField(), 
                co.getJdbcType(), 
                co.getSqlType(),
                co.getLength() == null ? null : String.valueOf(co.getLength()), 
                co.getScale() == null ? null : String.valueOf(co.getScale()), 
                co.getPrecision() == null ? null : String.valueOf(co.getPrecision()), 
                String.valueOf(co.isAllowsNull()), co.getDefaultValue());
        }
        this.columnOptions = co;
        String option = co.getJdbcType();
        if (option != null)
        {
            if (option.toLowerCase().indexOf("lob") > 0)
            {
                setUnlimitedLength();
            }
        }

        // Set the column default if there is one
        if (co.getDefaultValue() != null)
        {
            columnDefaultValue = co.getDefaultValue();
        }
    }

    /**
     * Mutator for the type information of the column.
     * @param typeInfo The type info
     * @return The column with the updated info
     */
    public final Column setTypeInfo(TypeInfo typeInfo)
    {
		if (this.typeInfo == null)
		{
			this.typeInfo = typeInfo;
		}
        return this;
    }

    /**
     * Mutator for the constraints of the column.
     *
     * @param constraints The constraints
     * @return The column with the updated info
     **/
    public final Column setConstraints(String constraints)
    {
        this.constraints = constraints;
        return this;
    }

    /**
     * Mutator for the length type of the column.
     *
     * @param type The length type
     * @return The column with the updated info
     **/
    private void setLengthType(int type)
    {
        flags &= ~LENGTH_TYPE_MASK;
        flags |= type;
    }

    /**
     * Mutator for the length of the column, setting it to unlimited length.
     *
     * @return The column with the updated info
     **/
    public final Column setUnlimitedLength()
    {
        setLengthType(UNLIMITED_LENGTH);
        return this;
    }

    /**
     * Mutator to make the column the primary key
     **/
    public final void setAsPrimaryKey()
    {
        flags |= PRIMARY_KEY_PART;
    }

    /**
     * Mutator for the nullability of the column.
     *
     * @return The column with the updated info
     **/
    public final DatastoreField setNullable()
    {
        flags |= NULLABLE;
        return this;
    }

    /**
     * Mutator for the defaultability of the column.
     *
     * @return The column with the updated info
     **/
    public final DatastoreField setDefaultable()
    {
        defaultable = true;
        return this;
    }

    /**
     * Mutator for the default value of the column.
     *
     * @param value The default value
     * @return The column with the updated info
     **/
    public final Column setDefault(Object value)
    {
        defaultValue = value;
        return this;
    }

    /**
     * Mutator for the uniqueness of the column.
     *
     * @return The column with the updated info
     **/
    public final Column setUnique()
    {
        flags |= UNIQUE;
        return this;
    }

    /**
     * Accessor for the length type of the column.
     *
     * @return The length type
     **/
    public int getLengthType()
    {
        return (flags & LENGTH_TYPE_MASK);
    }

    /**
     * Accessor for whether the column is the primary key
     *
     * @return whether the column is (part of) the primary key
     **/
    public final boolean isPrimaryKey()
    {
        return ((flags & PRIMARY_KEY_PART) != 0);
    }

    /**
     * Accessor for whether the column has exact precision
     *
     * @return whether the column has exact precision
     **/
    public final boolean isExactPrecision()
    {
        return ((flags & EXACT_PRECISION) != 0);
    }

    /**
     * Accessor for whether the column is nullable
     *
     * @return whether the column is nullable
     **/
    public final boolean isNullable()
    {
        return ((flags & NULLABLE) != 0);
    }

    /**
     * Accessor for whether the column is defaultable
     *
     * @return whether the column is defaultable
     **/
    public final boolean isDefaultable()
    {
        return (defaultable);
    }

    /**
     * Accessor for whether the column is unique
     *
     * @return whether the column is unique
     **/
    public final boolean isUnique()
    {
        return ((flags & UNIQUE) != 0);
    }

    public boolean equals(Object obj)
    {
       if (obj == this)
        {
            return true;
        }

        if (!(obj instanceof Column))
        {
            return false;
        }

        Column col = (Column)obj;

        return table.equals(col.table) && name.equals(col.name);
    }

    public int hashCode()
    {
        return table.hashCode() ^ name.hashCode();
    }

    public String toString()
    {
        return table.toString() + "." + name;
    }
    
    /**
     * Wraps the column name with a FUNCTION.
     * <PRE>example: SQRT(?) generates: SQRT(columnName)</PRE>
     * @param replacementValue the replacement to ?. Probably it's a column name, that may be fully qualified name or not
     * @return a String with function taking as parameter the replacementValue
     */
    public String applySelectFunction(String replacementValue)
    {
        if( replacementValue == null )
        {
            return wrapperFunction[WRAPPER_FUNCTION_SELECT];
        }
        if( wrapperFunction[WRAPPER_FUNCTION_SELECT] != null )
        {
            return StringUtils.replaceAll(wrapperFunction[WRAPPER_FUNCTION_SELECT],"?",replacementValue);
        }
        return replacementValue;
    }    

    /**
     * Accessor for the default Value
     * @return the default value
     */
    public Object getDefaultValue()
    {
        return defaultValue;
    }

    /**
     * Mutator for the default Value
     * @param object default value
     */
    public void setDefaultValue(Object object)
    {
        defaultValue = object;
    }

    /**
     * Accessor for the columnOptions
     * @return The Column Options (length, scale, precision etc)
     */
    public ColumnMetaData getColumnOptions()
    {
        return columnOptions;
    }
    
    /**
     * @return Returns the constraints.
     */
    public String getConstraints()
    {
        return constraints;
    }
    
    /**
     * Accessor for the AutoIncrement
     * @return true if column is autoincrement
     */
    public boolean isAutoIncrement()
    {
        return autoIncrement;
    }

    /**
     * Mutator for the AutoIncrement.
     * @param auto_increment True if column is autoincrement
     */
    public void setAutoIncrement(boolean auto_increment)
    {
        autoIncrement = auto_increment;
    }

    /**
     * Accessor for the Mapping. 
     * @return The Mapping
     */
    public JavaTypeMapping getMapping()
    {
        return mapping;
    }

    public final void checkPrimitive() throws ColumnDefinitionException
    {
    }

    public final void checkInteger() throws ColumnDefinitionException
    {
    }

    public final void checkDecimal() throws ColumnDefinitionException
    {
    }

    public final void checkString() throws ColumnDefinitionException
    {
        if( columnOptions.getJdbcType() == null )
        {
            columnOptions.setJdbcType("VARCHAR");
        }
        if( columnOptions.getLength() == null )
        {
            // Use the default string length
            columnOptions.setLength(storeMgr.getPMFContext().getPmfConfiguration().getStringDefaultLength());
        }
    }
    
	public void copyConfigurationTo(DatastoreField field)
	{
	    Column col = (Column) field;
		col.typeInfo = this.typeInfo;
		col.flags |= this.flags;
		col.flags &= ~PRIMARY_KEY_PART;
		col.flags &= ~UNIQUE;
		col.flags &= ~NULLABLE;
		col.defaultValue = this.defaultValue;
		col.columnOptions = this.columnOptions;
		col.wrapperFunction = this.wrapperFunction;
	}
    
    /**
     * Sets a function to wrap the column. The wrapper function String must use ?
     * to be replaced later by the column name
     * 
     * example: SQRT(?) generates: SQRT(COLUMN)
     * 
     * @param wrapperFunction The wrapperFunction to set.
     * @param wrapperMode whether select, insert or update
     */
    public void setWrapperFunction(String wrapperFunction, int wrapperMode)
    {
        if( wrapperFunction != null && wrapperMode == WRAPPER_FUNCTION_SELECT && wrapperFunction.indexOf("?")<0 )
        {
            throw new JDOUserException("Wrapping function must have one ? (question mark). e.g. SQRT(?)");
        }
        this.wrapperFunction[wrapperMode] = wrapperFunction;
    }
    
    /**
     * Gets the wrapper for parameters
     * @param wrapperMode whether select, insert or update
     * @return Returns the wrapperFunction.
     */
    public String getWrapperFunction(int wrapperMode)
    {
        return wrapperFunction[wrapperMode];
    }
    
    /**
     * Accessor for the Store Manager.
     * @return Store Manager
     **/
    public StoreManager getStoreManager()
    {
        return storeMgr;
    }    
}