/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Erik Bengtson - fixed raised java.lang.NoSuchMethodError on
                    toString when appending an StringBuffer to
                    another StringBuffer
2003 Andy Jefferson - coding standards
2004 Andy Jefferson - changed to use fully qualified table names
2005 Andy Jefferson - sorted toString() to not modify fields.
    ...
**********************************************************************/
package org.jpox.store.rdbms;

import java.util.ArrayList;
import java.util.Iterator;

import org.jpox.store.DatastoreClass;
import org.jpox.store.DatastoreField;
import org.jpox.store.DatastoreObject;
import org.jpox.store.FetchStatement;
import org.jpox.store.rdbms.adapter.DatabaseAdapter;

/**
 * Representation of a Fetch statement for RDBMS datastores.
 * Takes the traditional SQL form of "SELECT .... FROM ... WHERE ... " and
 * can have an optional locking on the selected rows where required.
 *
 * @version $Revision: 1.1 $
 **/ 
public class RDBMSFetchStatement implements FetchStatement
{
    protected final DatastoreClass mainTable;
    protected final ArrayList selected = new ArrayList();
    protected final ArrayList supertables = new ArrayList();
    protected StringBuffer conditionList = new StringBuffer();

    /**
     * Constructor.
     * @param mainTable The table to select from
     */
    public RDBMSFetchStatement(DatastoreClass mainTable)
    {
        this.mainTable = mainTable;
    }

    /**
     * Add a column to the SELECT clause.
     * @param col The column to add
     * @return The parameter position of this column in the statement.
     */
    public int select(DatastoreField col)
    {
        String columnID = referenceColumn(col);

        if (!selected.contains(columnID))
        {
            selected.add(columnID);
        }

        return selected.indexOf(columnID) + 1;
    }

    /**
     * Adds an expression to the SELECT clause.
     * @param expr the expr to add to the select clause
     * @return The parameter position of this expression in the statement.
     */
    public int select(String expr)
    {
        if (!selected.contains(expr))
        {
            selected.add(expr);
        }

        return selected.indexOf(expr) + 1;
    }

    /**
     * Method to specify a column to be referenced.
     * @param col The column
     * @return The column statement text
     */
    public String referenceColumn(DatastoreField col)
    {
        DatastoreObject t = col.getDatastoreContainerObject();

        if (!t.equals(mainTable) && !supertables.contains(t))
        {
            supertables.add(t);
        }

        return col.applySelectFunction(t.toString() + '.' + col.getName().toString());
    }

    /**
     * Method to add an extra condition to the WHERE clause.
     * @param condition The condition
     */
    public void andCondition(String condition)
    {
        if (conditionList.length() > 0)
        {
            conditionList.append(" AND ");
        }

        conditionList.append('(').append(condition).append(')');
    }

    /**
     * Method to return the statement assuming no locking of rows.
     * @return The statement
     */
    public String toString()
    {
        return toString(false);
    }

    /**
     * Method to return the statement.
     * @param lock Whether to lock the rows found by this SELECT
     * @return The statement
     */
    public String toString(boolean lock)
    {
        // Build the SELECT list
        StringBuffer stmt = new StringBuffer("SELECT ");
        Iterator i = selected.iterator();
        while (i.hasNext())
        {
            stmt.append(i.next());

            if (i.hasNext())
            {
                stmt.append(',');
            }
        }

        // Generate the FROM clause
        String[] mainTableColumnID = new String[mainTable.getIDMapping().getNumberOfDatastoreFields()];
        for (int j = 0; j < mainTable.getIDMapping().getNumberOfDatastoreFields(); j++)
        {
            mainTableColumnID[j] = referenceColumn(mainTable.getIDMapping().getDataStoreMapping(j).getDatastoreField());
        }
        stmt.append(" FROM ").append(mainTable.toString());

        StringBuffer joinConditions = new StringBuffer();
        i = supertables.iterator();
        while (i.hasNext())
        {
            DatastoreClass supertable = (DatastoreClass) i.next();
            String[] supertableColumnID = new String[mainTable.getIDMapping().getNumberOfDatastoreFields()];
            for (int j = 0; j < supertable.getIDMapping().getNumberOfDatastoreFields(); j++)
            {
                supertableColumnID[j] = referenceColumn(supertable.getIDMapping().getDataStoreMapping(j).getDatastoreField());
            }

            stmt.append(',').append(supertable.toString());

            for (int j = 0; j < supertable.getIDMapping().getNumberOfDatastoreFields(); j++)
            {
                if (joinConditions.length() > 0)
                {
                    joinConditions.append(" AND ");
                }
                joinConditions.append(mainTableColumnID[j]).append(" = ").append(supertableColumnID[j]);
            }
        }

        StringBuffer conditionList = new StringBuffer(this.conditionList.toString());
        if (joinConditions.length() > 0)
        {
            if (conditionList.length() > 0)
            {
                conditionList = joinConditions.append(" AND (").append(conditionList.toString()).append(')');
            }
            else
            {
                conditionList = joinConditions;
            }
        }

        // Add the WHERE clause
        if (conditionList.length() > 0)
        {
            stmt.append(" WHERE ").append(conditionList.toString());
        }

        // Add any locking of the rows
        if (lock)
        {
            if (((DatabaseAdapter) mainTable.getStoreManager().getDatastoreAdapter()).supportsLockWithSelectForUpdate())
            {
                stmt.append(" FOR UPDATE");
            }
            else if (((DatabaseAdapter) mainTable.getStoreManager().getDatastoreAdapter()).getSelectWithLockOption() != null)
            {
                stmt.append(" WITH ").append(((DatabaseAdapter) mainTable.getStoreManager().getDatastoreAdapter()).getSelectWithLockOption());
            }
        }

        return stmt.toString();
    }
}