/**********************************************************************
Copyright (c) 2003 Mike Martin (TJDO) and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Kelly Grizzle (TJDO)
2003 Erik Bengtson  - removed exist() operation
2003 Erik Bengtson  - refactored the persistent id generator System property
2003 Andy Jefferson - added localiser
2003 Andy Jefferson - updated exception handling with SchemaTable
2003 Andy Jefferson - restructured to remove SchemaTable, and add StoreManagerHelper
2003 Andy Jefferson - updated getSubClassesForClass to recurse
2004 Erik Bengtson  - removed unused method and variables 
2004 Erik Bengtson  - fixed problem with getObjectById for App ID in getClassForOID
2004 Andy Jefferson - re-emergence of SchemaTable. Addition of addClass().
2004 Andy Jefferson - Addition of AutoStartMechanism interface
2004 Andy Jefferson - Update to use Logger
2004 Andy Jefferson - Addition of Catalog name to accompany Schema name
2004 Marco Schulze  - replaced catch(NotPersistenceCapableException ...)
                  by advance-check via TypeManager.isSupportedType(...)
2004 Andy Jefferson - split StoreData into superclass.
2004 Andy Jefferson - added support for other inheritance types
2004 Andy Jefferson - added capability to dynamically add columns
    ...
**********************************************************************/
package org.jpox.store.rdbms;

import java.net.URL;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Comparator;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.Map;
import java.util.Properties;
import java.util.Set;
import java.util.TreeSet;
import java.util.WeakHashMap;

import javax.jdo.Extent;
import javax.jdo.JDODataStoreException;
import javax.jdo.JDOException;
import javax.jdo.JDOFatalException;
import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOUserException;
import javax.jdo.SingleFieldIdentity;
import javax.jdo.spi.PersistenceCapable;

import org.jpox.AbstractPersistenceManagerFactory;
import org.jpox.ClassLoaderResolver;
import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.TypeManager;
import org.jpox.exceptions.ClassNotPersistenceCapableException;
import org.jpox.metadata.ClassMetaData;
import org.jpox.metadata.ExtensionMetaData;
import org.jpox.metadata.FieldMetaData;
import org.jpox.metadata.IdentityStrategy;
import org.jpox.metadata.IdentityType;
import org.jpox.metadata.InheritanceMetaData;
import org.jpox.metadata.InheritanceStrategy;
import org.jpox.poid.Poid;
import org.jpox.poid.PoidJDBCConnectionProvider;
import org.jpox.poid.PoidManager;
import org.jpox.store.DatastoreClass;
import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.DatastoreObject;
import org.jpox.store.OID;
import org.jpox.store.SCOID;
import org.jpox.store.StoreManager;
import org.jpox.store.XMLAutoStarter;
import org.jpox.store.exceptions.DatastorePermissionException;
import org.jpox.store.exceptions.NoExtentException;
import org.jpox.store.exceptions.NoTableManagedException;
import org.jpox.store.exceptions.ViewNotSupportedException;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.rdbms.adapter.AdapterManager;
import org.jpox.store.rdbms.adapter.DatabaseAdapter;
import org.jpox.store.rdbms.extent.ClassTableExtent;
import org.jpox.store.rdbms.extent.ClassViewExtent;
import org.jpox.store.rdbms.request.DeleteRequest;
import org.jpox.store.rdbms.request.FetchRequest;
import org.jpox.store.rdbms.request.InsertRequest;
import org.jpox.store.rdbms.request.LocateRequest;
import org.jpox.store.rdbms.request.UpdateRequest;
import org.jpox.store.rdbms.scostore.InverseListStore;
import org.jpox.store.rdbms.scostore.InverseMapStore;
import org.jpox.store.rdbms.scostore.InverseSetStore;
import org.jpox.store.rdbms.scostore.NormalListStore;
import org.jpox.store.rdbms.scostore.NormalMapStore;
import org.jpox.store.rdbms.scostore.NormalSetStore;
import org.jpox.store.rdbms.sqlidentifier.SQLIdentifier;
import org.jpox.store.rdbms.table.AbstractTable;
import org.jpox.store.rdbms.table.ClassTable;
import org.jpox.store.rdbms.table.ClassView;
import org.jpox.store.rdbms.table.FCOTable;
import org.jpox.store.rdbms.table.JDOTable;
import org.jpox.store.rdbms.table.ListTable;
import org.jpox.store.rdbms.table.MapTable;
import org.jpox.store.rdbms.table.ProbeTable;
import org.jpox.store.rdbms.table.SetTable;
import org.jpox.store.rdbms.table.Table;
import org.jpox.store.rdbms.table.TableImpl;
import org.jpox.store.rdbms.table.ViewImpl;
import org.jpox.store.scostore.Store;
import org.jpox.util.ClassUtils;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;
import org.jpox.util.MacroString;
import org.jpox.util.MultiMap;
import org.jpox.util.SoftValueMap;
import org.jpox.util.TypeConversionHelper;

/**
 * Manages the contents of a database schema on behalf of a particular PersistenceManagerFactory 
 * and all of its persistent instances.
 * <p>
 * The RDBMS manager's responsibilities extend those for Storemanager to add :
 * <ul>
 * <li>implements insert(), fetch(), update(), delete() in the interface to the StateManager.</li>
 * <li>Providing cached access to JDBC database metadata.</li>
 * <li>Resolving SQL identifier macros to actual SQL identifiers.</li>
 * </ul>
 *
 * @version $Revision: 1.41 $
 */
public class RDBMSManager extends StoreManager
{
    protected static final Localiser LOCALISER_RDBMS = Localiser.getInstance("org.jpox.store.rdbms.Localisation");

    /**
     * The amount of time before we expire any cached column info. Hardcoded to five minutes.
     */
    private static final int COLUMN_INFO_EXPIRATION_MS = 5 * 60 * 1000;

    /** Catalog name for the RDBMS database */
    private String catalogName = null;

    /** Schema name for the RDBMS database */
    private String schemaName = null;

    private Map columnInfoByTableName = new HashMap();
    private long columnInfoReadTimestamp = -1;

    /**
     * The cache of database requests. Access is synchronized on the map object
     * itself.
     */
    private Map requestsByID = Collections.synchronizedMap(new SoftValueMap());

    /**
     * The active class adder transaction, if any. Some RDBMSManager methods are
     * called recursively in the course of adding new classes. This field allows
     * such methods to coordinate with the active ClassAdder transaction.
     * Recursive methods include:
     * <ul>
     * <li>addClasses()</li>
     * <li>addSetTable()</li>
     * <li>addMapTable()</li>
     * <li>addListTable()</li>
     * </ul>
     * Access is synchronized on the RDBMSManager itself. 
     * Invariant: classAdder == null if RDBMSManager is unlocked.
     */
    private ClassAdder classAdder = null;

    /**
     * Constructs a new RDBMSManager. On successful return the new RDBMSManager
     * will have successfully connected to the database with the given
     * credentials and determined the schema name, but will not have inspected
     * the schema contents any further. The contents (tables, views, etc.) will
     * be subsequently created and/or validated on-demand as the application
     * accesses persistent classes.
     * <p>
     * To avoid creating unnecessary redundant RDBMSManagers, new RDBMSManagers
     * should always be obtained from the {@link RDBMSManagerFactory}, rather
     * than constructed directly.
     * 
     * @param clr the ClassLoaderResolver
     * @param pmf The corresponding PersistenceManagerFactory. This factory's
     *            non-transactional data source will be used to get database
     *            connections as needed to perform management functions.
     * @param userName The database user name.
     * @param password The database user's password.
     * @exception JDODataStoreException
     *                If the database could not be accessed or the name of the
     *                schema could not be determined.
     * @see RDBMSManagerFactory
     */
    RDBMSManager(ClassLoaderResolver clr,
                 AbstractPersistenceManagerFactory pmf, 
                 String userName,
                 String password)
    {
        super(clr, pmf, userName, password);

        // Retrieve the Database Adapter for this datastore
        try
        {
            Connection conn;
            if (userName == null)
            {
                conn = ds.getConnection();
            }
            else
            {
                conn = ds.getConnection(userName, password);
            }
            try
            {
                dba = AdapterManager.getAdapterManager().getInstance(conn);
                dba.setDictionary(dictionary);

                // Check if the user has specified DefaultCatalogName and
                // whether the DBA allows it
                if (pmf.getPMFConfiguration().getDefaultCatalogName() != null)
                {
                    if (!((DatabaseAdapter)dba).supportsCatalogsInTableDefinitions())
                    {
                        JPOXLogger.RDBMS.warn(LOCALISER_RDBMS.msg("RDBMSManager.DefaultCatalogNameUnusable",pmf.getPMFConfiguration().getDefaultCatalogName()));
                    }
                }
                // Check if the user has specified DefaultSchemaName and
                // whether the DBA allows it
                if (pmf.getPMFConfiguration().getDefaultSchemaName() != null)
                {
                    if (!((DatabaseAdapter)dba).supportsSchemasInTableDefinitions())
                    {
                        JPOXLogger.RDBMS.warn(LOCALISER_RDBMS.msg("RDBMSManager.DefaultSchemaNameUnusable",pmf.getPMFConfiguration().getDefaultSchemaName()));
                    }
                }

                // Initialise the Schema "Name", and "Data"
                initialiseSchema(pmf.getPMFConfiguration().getAutoStartMechanism(), 
                    			 pmf.getPMFConfiguration().getAutoStartMechanismMode(), 
                    			 pmf.getPMFConfiguration().getAutoStartMechanismXmlFile(), conn);
            }
            finally
            {
                conn.close();
            }
        }
        catch (JDOException jdoex)
        {
            JPOXLogger.RDBMS_SCHEMA.error(LOCALISER_RDBMS.msg("RDBMSManager.DatabaseInitialisationError"));
            JPOXLogger.RDBMS_SCHEMA.error(jdoex);
            throw jdoex;
        }
        catch (Exception e1)
        {
            String msg=LOCALISER_RDBMS.msg("RDBMSManager.DatabaseInitialisationError") + ' ' + LOCALISER_RDBMS.msg("RDBMSManager.DatabaseInitialisationAdvice");
            JPOXLogger.RDBMS_SCHEMA.error(msg);
            JPOXLogger.RDBMS_SCHEMA.error(e1);
            throw new JDODataStoreException(msg, e1);
        }
    }

    /**
     * Utility to return a Connection to the Datastore. Is simply a wrapper to
     * the DatabaseAdapter getConnection method.
     * @param isolation_level The transaction isolation scheme to use e.g
     * Connection.TRANSACTION_NONE
     * @return The Connection to the datastore
     */
    public Connection getConnection(int isolation_level) 
    throws SQLException
    {
        return ((DatabaseAdapter) dba).getConnection(ds, userName, password,
                isolation_level);
    }

    /**
     * Utility to close a Connection for the Datastore. Is simply a wrapper to
     * the DatabaseAdapter closeConnection method.
     * @param conn The Connection to the datastore
     */
    public void closeConnection(Connection conn)
    throws SQLException
    {
        ((DatabaseAdapter) dba).closeConnection(conn);
    }

    /**
     * Initialises the schema name for the datastore, and (optionally) the
     * schema table (and associated initial schema data).
     * @param auto_start_mechanism Name of AutoStart mechanism
     * @param auto_start_mechanism_mode Mode of AutoStart mechanism
     * @param auto_start_mechanism_xml_file Name of XML file if using XML output
     * @param conn A connection to the database
     */
    private void initialiseSchema(String auto_start_mechanism,
                                  String auto_start_mechanism_mode,
                                  String auto_start_mechanism_xml_file,
                                  Connection conn)
    throws Exception
    {
        // Initialise the Catalog/Schema names
        try
        {
            try
            {
                catalogName = ((DatabaseAdapter) dba).getCatalogName(conn);
                schemaName = ((DatabaseAdapter) dba).getSchemaName(conn);
            }
            catch (UnsupportedOperationException e)
            {
                if (!readOnlySchema && !fixedSchema)
                {
                    // If we aren't a read-only datastore, try to create a table and then 
                    // retrieve its details, so as to obtain the catalog, schema. 
                    ProbeTable pt = new ProbeTable(this);
                    pt.initialize();
                    pt.create(conn);
                    try
                    {
                        String[] schema_details = pt.findSchemaDetails(conn);
                        if (schema_details != null)
                        {
                            catalogName = schema_details[0];
                            schemaName = schema_details[1];
                        }
                    }
                    finally
                    {
                        pt.drop(conn);
                    }
                }
            }
        }
        catch (SQLException e)
        {
            String msg=LOCALISER_RDBMS.msg("RDBMSManager.DatabaseConnectionError") + ' ' + LOCALISER_RDBMS.msg("RDBMSManager.DatabaseInitialisationAdvice");
            JPOXLogger.RDBMS_SCHEMA.error(msg);
            throw new JDODataStoreException(msg, e);
        }

        if ((readOnlySchema || fixedSchema) && auto_start_mechanism.equals("SchemaTable"))
        {
            // TODO Change this to XML as the default
            auto_start_mechanism = "None";
        }

        if (JPOXLogger.RDBMS_SCHEMA.isInfoEnabled())
        {
            JPOXLogger.RDBMS_SCHEMA.info(LOCALISER_RDBMS.msg("RDBMSManager.SchemaDataInitialise", catalogName, schemaName, auto_start_mechanism));
        }

        // AutoStarter - Initialise
        if (auto_start_mechanism != null)
        {
            if (auto_start_mechanism.equalsIgnoreCase("SchemaTable"))
            {
                starter = new SchemaAutoStarter(this);
                starter.setMode(auto_start_mechanism_mode);
            }
            else if (auto_start_mechanism.equalsIgnoreCase("XML"))
            {
                starter = new XMLAutoStarter(new URL("file:" + auto_start_mechanism_xml_file));
                starter.setMode(auto_start_mechanism_mode);
            }
            else
            {
                // Provide alternatives
            }
        }

        // AutoStarter - Load up any startup class names
        if (starter != null)
        {
            initialiseAutoStart();
        }
        
        if (JPOXLogger.RDBMS_SCHEMA.isInfoEnabled())
        {
            if (readOnlySchema)
            {
                JPOXLogger.RDBMS_SCHEMA.info(LOCALISER_RDBMS.msg("RDBMSManager.SchemaDataInitialisedReadOnly",
                    catalogName, schemaName, "" + storeDataByClass.size()));
            }
            else if (fixedSchema)
            {
                JPOXLogger.RDBMS_SCHEMA.info(LOCALISER_RDBMS.msg("RDBMSManager.SchemaDataInitialisedFixed",
                        catalogName, schemaName, "" + storeDataByClass.size()));
            }
            else
            {
                JPOXLogger.RDBMS_SCHEMA.info(LOCALISER_RDBMS.msg("RDBMSManager.SchemaDataInitialised",
                        catalogName, schemaName, "" + storeDataByClass.size()));
            }
        }
    }

    /**
     * Clears all knowledge of tables, cached requests, metadata, etc and resets
     * the store manager to its initial state.
     */
    private void clearSchemaData()
    {
        deregisterAllStoreData();

        columnInfoByTableName.clear();
        columnInfoReadTimestamp = -1;
        requestsByID.clear();
    }

    /**
     * Returns the database adapter used by this store manager.
     * Simply provides a convenience cast over the equivalent method
     * in StoreManager.
     * TODO Remove this when getDatastoreAdapter() is adequate
     * @return The database adapter used by this store manager.
     */
    public DatabaseAdapter getDatabaseAdapter()
    {
        return (DatabaseAdapter) dba;
    }

    /**
     * Accessor for the RDBMS catalog name.
     * 
     * @return The catalog name.
     */
    public String getCatalogName()
    {
        return catalogName;
    }

    /**
     * Accessor for the RDBMS schema name.
     * 
     * @return The schema name.
     */
    public String getSchemaName()
    {
        return schemaName;
    }

    // ----------------------------- Class Management -------------------------------

    /**
     * Method to add several (PersistenceCapable) classes to the store manager's
     * set of supported classes. This will create any necessary database objects
     * (tables, views, constraints, indexes etc). This will cause the addition
     * of any related classes.
     * 
     * @param classNames Name of the class(es) to be added.
     */
    public void addClasses(String[] classNames)
    {
        synchronized (this)
        {
            if (classAdder != null)
            {
                // addClasses() has been recursively re-entered: just add table
                // objects for the requested classes and return.
                classAdder.addClasses(classNames, clr);
                return;
            }
        }
        new ClassAdder(classNames, clr).execute(clr);
    }

    /**
     * Method to add a (PersistenceCapable) class to the store manager's set of
     * supported classes. This will create any necessary database objects. This
     * will cause the addition of any related classes.
     * 
     * @param className Name of the class to be added.
     */
    public void addClass(String className)
    {
        addClasses(new String[] {className});
    }

    /**
     * Called by Mapping objects in the midst of RDBMSManager.addClasses() to
     * request the creation of a set table.
     * @param fmd The field metadata describing the set field.
     */
    public synchronized SetTable addSetTable(FieldMetaData fmd)
    {
        if (classAdder == null)
        {
            throw new IllegalStateException(LOCALISER_RDBMS.msg("RDBMSManager.SetTableCreationError"));
        }
        return classAdder.addSetTable(fmd);
    }

    /**
     * Called by Mapping objects in the midst of RDBMSManager.addClasses() to
     * request the creation of a map table.
     * @param fmd The field metadata describing the map field.
     */
    public synchronized MapTable addMapTable(FieldMetaData fmd)
    {
        if (classAdder == null)
        {
            throw new IllegalStateException(LOCALISER_RDBMS.msg("RDBMSManager.MapTableCreationError"));
        }
        return classAdder.addMapTable(fmd);
    }

    /**
     * Called by Mapping objects to request the creation of a DatastoreObject (table).
     * @param fmd The field metadata describing the field.
     */
    public synchronized DatastoreContainerObject newJoinDatastoreContainerObject(FieldMetaData fmd)
    {
        if( fmd.getJoinMetaData() == null )
        {
            return null;
        }
        if( TypeManager.getTypeManager().isSCOMap(fmd.getField().getType()) )
        {
            return addMapTable(fmd);
        }
        else if( TypeManager.getTypeManager().isSCOList(fmd.getField().getType()) )
        {
            return addListTable(fmd);
        }
        return addSetTable(fmd);
    }
    
    /**
     * Called by Mapping objects to get Store objects
     * @param table The join table
     * @param fmd The field metadata describing the field.
     */
    public Store newStore(DatastoreObject table, FieldMetaData fmd)
    {
        // If a join table is present we need a "Normal"
        if( table == null )
        {
            if( TypeManager.getTypeManager().isSCOMap(fmd.getField().getType()) )
            {
                return new InverseMapStore(fmd,this);
            }
            else if( TypeManager.getTypeManager().isSCOList(fmd.getField().getType()) )
            {
                return new InverseListStore(fmd,this);
            }
            return new InverseSetStore(fmd,this);
        }
        // otherwise we need an "Inverse". Where the owner column resides in the element table
        else
        {
            if( TypeManager.getTypeManager().isSCOMap(fmd.getField().getType()) )
            {
                return new NormalMapStore((MapTable)table);
            }
            else if( TypeManager.getTypeManager().isSCOList(fmd.getField().getType()) )
            {
                return new NormalListStore((ListTable)table);
            }
            return new NormalSetStore((SetTable)table);
        }
    }
    
    /**
     * Called by Mapping objects in the midst of RDBMSManager.addClasses() to
     * request the creation of a list table.
     * @param fmd The field metadata describing the list field.
     */
    public synchronized ListTable addListTable(FieldMetaData fmd)
    {
        if (classAdder == null)
        {
            throw new IllegalStateException(LOCALISER_RDBMS.msg("RDBMSManager.ListTableCreationError"));
        }
        return classAdder.addListTable(fmd);
    }

    /**
     * Utility to remove all classes that we are managing.
     */
    public void removeAllClasses()
    {
        MgmtTransaction mtx = new MgmtTransaction(Connection.TRANSACTION_READ_COMMITTED)
        {
            public String toString()
            {
                return LOCALISER_RDBMS.msg("RDBMSManager.RemoveAllClassesFromSchema", catalogName, schemaName);
            }

            protected void execute(Connection conn, ClassLoaderResolver clr)
            throws SQLException
            {
                synchronized (RDBMSManager.this)
                {
                    boolean success = true;
                    try
                    {
                        JPOXLogger.RDBMS_SCHEMA.info(LOCALISER_RDBMS.msg("RDBMSManager.RemoveAllClassesFromSchema", 
                            catalogName, schemaName));

                        Map baseTablesByName = new HashMap();
                        Map viewsByName = new HashMap();
                        for (Iterator i = storeDataByClass.values().iterator(); i.hasNext();)
                        {
                            RDBMSStoreData data = (RDBMSStoreData) i.next();
                            if (JPOXLogger.RDBMS_SCHEMA.isInfoEnabled())
                            {
                                JPOXLogger.RDBMS_SCHEMA.info(LOCALISER_RDBMS.msg("RDBMSManager.RemoveClass",data.getName()));
                            }
                            
                            // If the class has a table/view to remove, add it to the list
                            if (data.hasTable())
                            {
                                if (data.mapsToView())
                                {
                                    viewsByName.put(data.getTableIdentifier(), 
                                        data.getTable());
                                }
                                else
                                {
                                    baseTablesByName.put(data.getTableIdentifier(),
                                        data.getTable());
                                }
                            }
                        }
                        
                        // Remove tables, table constraints and views in
                        // reverse order from which they were created
                        Iterator viewsIter = viewsByName.values().iterator();
                        while (viewsIter.hasNext())
                        {
                            ((ViewImpl) viewsIter.next()).drop(conn);
                        }

                        Iterator tablesIter = baseTablesByName.values().iterator();
                        while (tablesIter.hasNext())
                        {
                            ((TableImpl) tablesIter.next()).dropConstraints(conn);
                        }
                        tablesIter = baseTablesByName.values().iterator();
                        while (tablesIter.hasNext())
                        {
                            ((TableImpl) tablesIter.next()).drop(conn);
                        }
                    }
                    catch (Exception e)
                    {
                        success = false;
                        JPOXLogger.RDBMS_SCHEMA.error(LOCALISER_RDBMS.msg(
                                "RDBMSManager.RemoveClassesExceptionError", e));
                    }
                    finally
                    {
                        // Only empty our schema data if we succeeded in
                        // dropping everything with no error.
                        if (success)
                        {
                            clearSchemaData();
                        }
                    }
                }
            }
        };
        mtx.execute(clr);
    }

    // ---------------------------------------------------------------------------------------

    /**
     * Returns the JDO table having the given SQL identifier.
     * Returns 'null' if no such table is (yet) known to the store manager.
     * @param name The identifier name of the table.
     * @return The corresponding JDO table, or 'null'
     */
    public synchronized JDOTable getTable(SQLIdentifier name)
    {
        Iterator iterator = storeDataByClass.values().iterator();
        while (iterator.hasNext())
        {
            RDBMSStoreData sd = (RDBMSStoreData) iterator.next();
            if (sd.hasTable() && sd.getTableIdentifier().equals(name))
            {
                return sd.getTable();
            }
        }
        return null;
    }

    /**
     * Returns the JDO table for the specified field. 
     * Returns 'null' if the field is not (yet) known to the store manager.
     * @param fmd The metadata for the field.
     * @return The corresponding JDO table, or 'null'.
     */
    public synchronized JDOTable getTable(FieldMetaData fmd)
    {
        RDBMSStoreData sd = (RDBMSStoreData) storeDataByClass.get(fmd.getField());
        if (sd != null)
        {
            return sd.getTable();
        }
        else
        {
            return null;
        }
    }

    /**
     * Returns the primary table serving as backing for the given class. If the
     * class is not yet known to the store manager, {@link #addClass}is called
     * to add it.
     * @param className Name of the class whose table is be returned.
     * @return The associated table.
     * @exception NoTableManagedException If the given class has no table
     * managed in the database.
     */
    public FCOTable getTable(String className)
    {
        FCOTable ct = null;
        if (className == null)
        {
            JPOXLogger.RDBMS_SCHEMA.error(LOCALISER_RDBMS.msg("RDBMSManager.TableRetrievalError"));
            return null;
        }
        boolean classKnown = false;
        synchronized (this)
        {
            RDBMSStoreData sd = (RDBMSStoreData) storeDataByClass.get(className);
            if (sd != null)
            {
                classKnown = true;
                ct = (FCOTable) sd.getTable();
            }
        }

        if (!classKnown)
        {
            // Add the class to our supported list
            addClass(className);

            // Retry
            synchronized (this)
            {
                RDBMSStoreData sd = (RDBMSStoreData) storeDataByClass.get(className);
                if (sd != null)
                {
                    ct = (FCOTable) sd.getTable();
                }
            }
        }

        // Throw an exception if class still not known and no table
        // Note : "no-table" inheritance strategies will return null from this method
        if (!classKnown && ct == null)
        {
            throw new NoTableManagedException(className);
        }

        return ct;
    }

    /**
     * Returns the primary table serving as backing for the given class. If the
     * class is not yet known to the store manager, {@link #addClass}is called
     * to add it. Classes which have inheritance strategy of "new-table" and
     * "superclass-table" will return a table here, whereas "no-table" will
     * return null since it doesn't have a table as such.
     * <p>
     * This method is functionally equivalent to {@link #getTable(String)}
     * except that it will only return base tables (not views).
     * @param className Name of the class whose table is be returned.
     * @return The corresponding class table.
     * @exception NoTableManagedException If the given class has no table
     * managed in the database.
     * @exception ViewNotSupportedException If the given class is backed by a
     * view.
     */
    public DatastoreClass getDatastoreClass(String className)
    {
        FCOTable t = getTable(className);
        if (t != null && t instanceof ClassView)
        {
            throw new ViewNotSupportedException(className); 
        }
        return (ClassTable) t;
    }

    /**
     * Utility to navigate the inheritance hierarchy to find the base class that defines the primary keys
     * for this tree.
     * @param cmd ClassMetaData for this class
     * @return The ClassMetaData for the class defining the primary keys
     */
    public ClassMetaData getClassWithPrimaryKeyForClass(ClassMetaData cmd)
    {
        if (cmd == null)
        {
            return null;
        }

        // Base class will have primary key fields
        if (cmd.getSuperclassMetaData() == null)
        {
            return cmd;
        }
        // Class has its own table so has the PK fields already
        else if (getDatastoreClass(cmd.getFullClassName()) != null)
        {
            return cmd;
        }

        return cmd.getSuperclassMetaData();
    }

    /**
     * Returns the class corresponding to the given object JDO ID. If the object
     * is an OID, return the PersistenceCapable class. If the object is a SCOID,
     * return the SCO class. If the object is an AppID PK, return the PersistenceCapable
     * class that the id relates to. If SingleFieldIdentity return the associated PC class.
     * @param pm Persistence Manager
     * @param id The JDO identity of some object.
     * @return For datastore identity, return the class of the corresponding
     * object. For application identity, return the class of the corresponding
     * object or null if object does not exist.
     * @exception ClassCastException If the type of ID is not recognized (
     * {@link OID}or {@link SCOID}).
     */
    public String getClassNameForObjectID(PersistenceManager pm, Object id)
    {
        // Object is an OID
        if (id instanceof OID)
        {
            // Using SingleFieldIdentity so can assume that object is of the target class or a subclass
            List subclasses = new ArrayList(this.getSubClassesForClass(((OID) id).getPcClass(), true));
            if (subclasses.size() < 1)
            {
                // No subclasses so must be an instance of the OID class
                return ((OID) id).getPcClass();
            }

            // Multiple subclasses so we need to find which one using a query
            // Only include classes which have a table.
            ArrayList classOptions = new ArrayList();
            RDBMSStoreData data = (RDBMSStoreData)storeDataByClass.get(((OID) id).getPcClass());
            if (data.hasTable())
            {
                classOptions.add(data);
            }
            Iterator subclassesIter = subclasses.iterator();
            while (subclassesIter.hasNext())
            {
                String subclassName = (String)subclassesIter.next();
                RDBMSStoreData subclassData = (RDBMSStoreData)storeDataByClass.get(subclassName);
                if (subclassData.hasTable())
                {
                    classOptions.add(subclassData);
                }
            }
            
            if (classOptions.size() == 0)
            {
                return ((OID)id).getPcClass();
            }
            else if (classOptions.size() == 1)
            {
                return ((RDBMSStoreData)classOptions.get(0)).getName();
            }
            else
            {
                RDBMSStoreData classOption = RDBMSStoreHelper.getClassForIdKey(pm, this, id, classOptions);
                if (classOption == null)
                {
                    return ((OID)id).getPcClass();
                }
                return classOption.getName();
            }
        }
        // Object is a SCOID
        else if (id instanceof SCOID)
        {
            return ((SCOID) id).getSCOClass();
        }
        else if (id instanceof SingleFieldIdentity)
        {
            // Using SingleFieldIdentity so can assume that object is of the target class or a subclass
            String targetClassName = ((SingleFieldIdentity)id).getTargetClassName();

            ArrayList classOptions = new ArrayList();
            RDBMSStoreData data = (RDBMSStoreData)storeDataByClass.get(targetClassName);
            if (data.hasTable())
            {
                classOptions.add(data);
            }

            HashSet subclasses = getSubClassesForClass(targetClassName, true);
            Iterator subclassesIter = subclasses.iterator();
            while (subclassesIter.hasNext())
            {
                String subclassName = (String)subclassesIter.next();
                RDBMSStoreData subclassData = (RDBMSStoreData)storeDataByClass.get(subclassName);
                if (subclassData.hasTable())
                {
                    classOptions.add(subclassData);
                }
            }

            if (classOptions.size() == 0)
            {
                return null;
            }
            else if (classOptions.size() == 1)
            {
                return ((RDBMSStoreData)classOptions.get(0)).getName();
            }
            else
            {
                RDBMSStoreData classOption = RDBMSStoreHelper.getClassForIdKey(pm, this, id, classOptions);
                if (classOption == null)
                {
                    return null;
                }
                return classOption.getName();
            }
        }
        else
        {
            // Check if it is a persisted class
            ArrayList classOptions = new ArrayList();
            
            // Find all of the application identity PK classes of this type
            Collection c = (Collection) storeDataByAppIdClass.get(id.getClass().getName());
            if (c != null)
            {
                Iterator iter = c.iterator();
                while (iter.hasNext())
                {
                    RDBMSStoreData store_data = (RDBMSStoreData)iter.next();
                    if (store_data.hasTable())
                    {
                        classOptions.add(store_data);
                    }
                }
                
                if (classOptions.size() == 0)
                {
                    return null;
                }
                else if (classOptions.size() == 1)
                {
                    return ((RDBMSStoreData)classOptions.get(0)).getName();
                }
                else
                {
                    RDBMSStoreData classOption = RDBMSStoreHelper.getClassForIdKey(pm, this, id, classOptions);
                    if (classOption == null)
                    {
                        return null;
                    }
                    return classOption.getName();
                }
            }
        }
        return null;
    }

    /**
     * Interface to getting an Extent for a class.
     * @param pm The Persistence Manager
     * @param c The class requiring the Extent
     * @param subclasses Whether to include subclasses of 'c'
     * @return The Extent
     * @exception NoExtentException if an extent is not managed for the
     * specified class
     */
    public Extent getExtent(PersistenceManager pm, Class c, boolean subclasses)
    {
        ClassMetaData cmd = getMetaDataManager().getMetaDataForClass(c);
        if (!cmd.isRequiresExtent())
        {
            throw new NoExtentException(c.getName());
        }
        FCOTable t = getTable(c.getName());
        if (t instanceof ClassView)
        {
            return new ClassViewExtent(pm, (ClassView)t, subclasses);
        }
        else if (t instanceof ClassTable)
        {
            return new ClassTableExtent(pm, (ClassTable)t, c.getName(), subclasses);
        }
        else
        {
            // No specific table, so either has no table, or multiple
            // This happens with the "no-table" inheritance strategy
            ClassMetaData[] cmds = getClassesManagingTableForClass(cmd);
            if (cmds != null)
            {
                ClassTable[] tables = new ClassTable[cmds.length];
                for (int i=0;i<cmds.length;i++)
                {
                    tables[i] = (ClassTable)getTable(cmds[i].getFullClassName());
                }
                return new ClassTableExtent(pm, tables, c.getName(), subclasses);
            }
            else
            {
                // TODO Throw an exception, since should never happen
                return null;
            }
        }
    }

    /**
     * Accessor for whether this query language is supported.
     * @param language The language
     * @return Whether it is supported.
     * @see org.jpox.store.StoreManager#supportsQueryLanguage(java.lang.String)
     */
    public boolean supportsQueryLanguage(String language)
    {
        if (language == null)
        {
            return false;
        }
        else if (language.equals("javax.jdo.query.JDOQL") ||
                 language.equals("javax.jdo.query.JPOXSQL") ||
                 language.equals("javax.jdo.query.SQL"))
        {
            return true;
        }
        return false;
    }

    /**
     * Returns a new, unique ID for an object of the given class.
     * @param className Name of the class of the object.
     * @param pc The persistenceCapable object. Useful for AppID
     * @return A new object ID.
     */
    public Object newObjectID(StateManager sm, String className, PersistenceCapable pc)
    {
        ClassMetaData cmd = getMetaDataManager().getMetaDataForClass(className, clr); 
        if (cmd.getIdentityType() != IdentityType.NONDURABLE)
        {
            FCOTable t = getTable(className);
            if (t == null && cmd.getInheritanceMetaData().getStrategyValue() == InheritanceStrategy.NO_TABLE)
            {
                throw new JDOUserException(LOCALISER_RDBMS.msg("RDBMSManager.NoTableForClass", className));
            }
            if (t instanceof ClassTable)
            {
                // Go up to overall superclass to find id for that class.
                ClassTable ct = (ClassTable) t;
                boolean has_superclass = true;
                while (has_superclass)
                {
                    ClassTable superct = ct.getSupertable();
                    if (superct != null)
                    {
                        ct = superct;
                    }
                    else
                    {
                        has_superclass = false;
                    }
                }
                Object oid = ct.newOID(sm, pc);
                if (oid instanceof OID)
                {
                    ((OID) oid).setPcClass(className);
                }
                return oid;
            }
            else
            {
                return t.newOID(sm, pc);
            }
        }
        else
        {
            return new SCOID(className);
        }
    }

    /**
     * Method to return the value for the specified strategy.
     * Supports the non-datastore strategies of JDO2. That is,
     * does not support "autoassign" and "identity" which are done elsewhere.
     * @param pm the PersistenceManager requesting an ID for the object
     * @param table The datastore table
     * @param cmd ClassMetaData for the class containing this field
     * @param absoluteFieldNumber Absolute number of the field (-1 means datastore identity field)
     * @param strategy The strategy to use for defining the value
     * @param sequence Name of a sequence if the strategy is sequence
     * @param extensions Any extensions to control the value generation process.
     *                   For value-strategy tags this is in the field extension.
     *                   For datastore-identity tags this is in the datastore-identity extension.
     * @return The value for this strategy
     * @exception JDODataStoreException Thrown if an error occurs getting the value
     */
    public Object getStrategyValue(final PersistenceManager pm,
                                   final DatastoreClass table,
                                   final ClassMetaData cmd,
                                   final int absoluteFieldNumber,
                                   final IdentityStrategy strategy,
                                   final String sequence,
                                   final ExtensionMetaData[] extensions)
    {
        if (storeDataByClass.size() == 0)
        {
            String msg = LOCALISER_RDBMS.msg("RDBMSManager.SchemaDataUninitialisedForSchema",catalogName, schemaName);
            JPOXLogger.RDBMS_SCHEMA.error(msg);
            throw new JDOFatalInternalException(msg);
        }

        if (strategy == null || strategy == IdentityStrategy.AUTOASSIGN || strategy == IdentityStrategy.IDENTITY)
        {
            throw new JDOFatalInternalException("RDBMSManager.getStrategyValue called for invalid strategy \"" + strategy + "\"");
        }

        // Choose the required POID generator
        String poidGeneratorClassName = defaultPoidGeneratorClass;
        Properties properties = new Properties();
        if (strategy == IdentityStrategy.SEQUENCE)
        {
            poidGeneratorClassName = "org.jpox.poid.SequencePoidGenerator";
            properties.put("sequence-name", sequence);
        }
        else if (strategy == IdentityStrategy.INCREMENT)
        {
            poidGeneratorClassName = "org.jpox.poid.SequenceTablePoidGenerator";
        }
        else if (strategy == IdentityStrategy.UUIDSTRING)
        {
            // TODO Add (16-character) UUID String generator when available
            throw new JDOUserException("datastore-identity strategy \"uuid-string\" specified for class \"" + 
                cmd.getFullClassName() + "\" is not yet supported. Try \"uuid-hex\" or \"auid\" until this is available");
        }
        else if (strategy == IdentityStrategy.UUIDHEX)
        {
            poidGeneratorClassName = "org.jpox.poid.UUIDHexPoidGenerator";

            if (extensions != null)
            {
	            for (int i=0;i<extensions.length;i++)
	            {
	                if (extensions[i].getKey().equalsIgnoreCase("datastore-assign-uuid") && extensions[i].getValue().equalsIgnoreCase("true"))
	                {
	                    poidGeneratorClassName = "org.jpox.poid.DatastoreUUIDHexPoidGenerator";
	                    break;
	                }
	            }
            }
        }
        else if (strategy == IdentityStrategy.MAX)
        {
            // JPOX extension "max" generator
            poidGeneratorClassName = "org.jpox.poid.MaxPoidGenerator";
        }
        else if (strategy == IdentityStrategy.AUID)
        {
            poidGeneratorClassName = "org.jpox.poid.AUIDPoidGenerator";
        }
        else if (strategy == IdentityStrategy.CUSTOM)
        {
            // Users custom generator class
            if (extensions != null)
            {
                for (int i=0;i<extensions.length;i++)
                {
                    if (extensions[i].getKey().equalsIgnoreCase("custom-generator-class"))
                    {
                        poidGeneratorClassName = extensions[i].getValue();
                        break;
                    }
                }
                if (poidGeneratorClassName == null)
                {
                    throw new JDOUserException("Class " + cmd.getFullClassName() + " has custom generator specified yet has no \"custom-generator-class\" extension specified!");
                }
            }
        }
        else
        {
            // Find the best alternative - sequence, otherwise SequenceTable
            if (getDatabaseAdapter().supportsSequences() && sequence != null)
            {
                poidGeneratorClassName = "org.jpox.poid.SequencePoidGenerator";
                properties.put("sequence-name", sequence);
            }
            else
            {
                poidGeneratorClassName = "org.jpox.poid.SequenceTablePoidGenerator";
            }
        }

        // Apply any id extension properties
        if (extensions != null)
        {
            for (int i=0;i<extensions.length;i++)
            {
                properties.put(extensions[i].getKey(), extensions[i].getValue());
            }
        }

        // Retrieve the table name for use in the POID generators that need it
        // e.g MaxPoidGenerator, SequenceTablePoidGenerator
        properties.put("table-name", new SQLIdentifier(dba, table.toString()).toString());

        // Retrieve the column name(s) for this field for use in POID generators that need it
        // e.g MaxPoidGenerator
        StringBuffer columnsName = new StringBuffer();
        JavaTypeMapping m = null;
        if (absoluteFieldNumber >= 0)
        {
            m = table.getFieldMapping(cmd.getManagedFieldAbsolute(absoluteFieldNumber));
        }
        else
        {
            m = table.getIDMapping();
        }
        for (int i = 0; i < m.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                columnsName.append(",");
            }
            columnsName.append(((SQLIdentifier) m.getDataStoreMapping(i).getDatastoreField().getName()).getSQLIdentifier());
        }
        properties.put("column-name", columnsName.toString());

        // Get a PoidManager to generate the value
        // TODO Fix this. This creates a new manager and then throws it away if it already exists!
        PoidManager poidManager = new PoidManager();
        poidManager.setPoidGeneratorName(poidGeneratorClassName);
        poidManager.setProperties(properties);
        if (poidManagers.get(String.valueOf(poidManager.hashCode())) != null)
        {
            poidManager = (PoidManager) poidManagers.get(String.valueOf(poidManager.hashCode()));
        }
        else
        {
            poidManagers.put(String.valueOf(poidManager.hashCode()), poidManager);
        }
        poidManager.setStoreManager(this);
        poidManager.setJDBCConnectionProvider(new PoidJDBCConnectionProvider()
        {
            Connection con;

            public Connection getPoidJDBCConnection()
            {
                try
                {
                    if (pm.getPMFContext().getPmfConfiguration().getPoidTransactionAttribute().equalsIgnoreCase("UsePM"))
                    {
                        this.con = pm.getConnection(false);
                    }
                    else
                    {
                        this.con = ((DatabaseAdapter) dba).getConnection(ds,
                                userName, password, 
                                pm.getPMFContext().getPmfConfiguration().getPoidTransactionIsolationLevel());
                    }
                }
                catch (SQLException e)
                {
                    String msg=LOCALISER_RDBMS.msg("RDBMSManager.POIDConnectionOpenError", e);
                    JPOXLogger.RDBMS_SCHEMA.error(msg);
                    throw new JDODataStoreException(msg,e);
                }
                return con;
            }

            public void releaseConnection()
            {
                try
                {
                    if (pm.getPMFContext().getPmfConfiguration().getPoidTransactionAttribute().equalsIgnoreCase("UsePM"))
                    {
                        pm.releaseConnection(con);
                    }
                    else
                    {
                        boolean success = false;
                        try
                        {
                            con.commit();
                            success = true;
                        }
                        finally
                        {
                            if (!success)
                            {
                                con.rollback();
                            }
                        }
                        ((DatabaseAdapter) dba).closeConnection(con);
                    }
                    con = null;
                }
                catch (SQLException e)
                {
                    String msg=LOCALISER_RDBMS.msg("RDBMSManager.POIDConnectionCloseError", e);
                    JPOXLogger.RDBMS_SCHEMA.error(msg);
                    throw new JDODataStoreException(msg);
                }
            }
        });

        Poid poid = poidManager.getPoid();

        FieldMetaData fmd = cmd.getManagedFieldAbsolute(absoluteFieldNumber);

        Object oid = poid.getOid();
        
        if( fmd != null )
        {
            //replace the value of the id, but before convert the value to the field type if needed
            Object convertedValue = TypeConversionHelper.convertTo(oid,fmd.getType());

            if( convertedValue == null )
            {
                throw new JDOFatalException(LOCALISER_RDBMS.msg("RDBMSManager.InvalidGeneratedId",fmd.getFullFieldName(),oid));
            }
            
            oid = convertedValue;
        }
        
        if (JPOXLogger.RDBMS_SCHEMA.isDebugEnabled())
        {
            String fieldName;
            if (fmd != null)
            {
                fieldName = fmd.getFullFieldName();
            }
            else
            {
                fieldName = cmd.getFullClassName() + " (datastore id)";
            }
            JPOXLogger.RDBMS_SCHEMA.debug(LOCALISER_RDBMS.msg("RDBMSManager.StrategyValueGenerated", fieldName, strategy,
                poidGeneratorClassName, oid));
        }

        return oid;
    }

    /**
     * Inserts a persistent object into the database.
     * @param sm The state manager of the object to be inserted.
     */
    public void insert(StateManager sm)
    {
        if (readOnlySchema)
        {
            throw new DatastorePermissionException(LOCALISER_RDBMS.msg("RDBMSManager.InsertObjectWhileReadOnlyError"));
        }
        
        DatastoreClass dc = getDatastoreClass(sm.getObject().getClass().getName());
        if( dc == null )
        {
            ClassMetaData cmd = getMetaDataManager().getMetaDataForClass(sm.getObject().getClass()); 
            if (cmd.getInheritanceMetaData().getStrategyValue() == InheritanceStrategy.NO_TABLE)
            {
                throw new JDOUserException(LOCALISER_RDBMS.msg("RDBMSManager.NoTableForClass", sm.getObject().getClass().getName()));
            }
            throw new JDOFatalInternalException(LOCALISER_RDBMS.msg("RDBMSManager.CantFindTableForClass", sm.getObject().getClass().getName(), cmd.getInheritanceMetaData().getStrategyValue()));
        }
        dc.insert(sm);
    }

    /**
     * Fetches a persistent object from the database.
     * @param sm The state manager of the object to be fetched.
     * @param fieldNumbers The numbers of the fields to be fetched.
     */
    public void fetch(StateManager sm, int fieldNumbers[])
    {
        getDatastoreClass(sm.getObject().getClass().getName()).fetch(sm, fieldNumbers);
    }

    /**
     * Updates a persistent object in the database.
     * @param sm The state manager of the object to be updated.
     * @param fieldNumbers The numbers of the fields to be updated.
     */
    public void update(StateManager sm, int fieldNumbers[])
    {
        if (readOnlySchema)
        {
            throw new DatastorePermissionException(LOCALISER_RDBMS.msg("RDBMSManager.UpdateObjectWhileReadOnlyError"));
        }
        
        getDatastoreClass(sm.getObject().getClass().getName()).update(sm, fieldNumbers);
    }

    /**
     * Deletes a persistent object from the database.
     * @param sm The state manager of the object to be deleted.
     */
    public void delete(StateManager sm)
    {
        if (readOnlySchema)
        {
            throw new DatastorePermissionException(LOCALISER_RDBMS.msg("RDBMSManager.DeleteObjectWhileReadOnlyError"));
        }
        
        getDatastoreClass(sm.getObject().getClass().getName()).delete(sm);
    }

    /**
     * Locates this object in the datastore. If the instance does not exist in
     * the datastore, this method will fail raising a
     * javax.jdo.JDOObjectNotFoundException.
     */
    public void locate(StateManager sm)
    {
        getDatastoreClass(sm.getObject().getClass().getName()).locate(sm);
    }
    
    /**
     * Delete dependent objects of a persistent object.
     * @since 1.1
     * @param sm StateManager for the object being deleted.
     * @param fieldNumbers the fields to delete.
     */
    public void deleteDependent(StateManager sm, int[] fieldNumbers)
    {
        if (readOnlySchema)
        {
            throw new DatastorePermissionException(LOCALISER_RDBMS.msg("RDBMSManager.DeleteDependentObjectWhileReadOnlyError"));
        }
        
        getDatastoreClass(sm.getObject().getClass().getName()).deleteDependent(sm, fieldNumbers);
    }

    /**
     * Returns a request object that will insert a row in the given table. The
     * store manager will cache the request object for re-use by subsequent
     * requests to the same table.
     * 
     * @param ct The table into which to insert.
     * @param className the class name of the object of the request
     * @return An insertion request object.
     */
    public InsertRequest getInsertRequest(ClassTable ct, String className)
    {
        RequestIdentifier reqID = new RequestIdentifier(ct, null, RequestIdentifier.Type.INSERT,className);

        InsertRequest req = (InsertRequest) requestsByID.get(reqID);
        if (req == null)
        {
            req = new InsertRequest(ct);
            requestsByID.put(reqID, req);
        }
        return req;
    }

    /**
     * Returns a request object that will fetch a row from the given table. The
     * store manager will cache the request object for re-use by subsequent
     * requests to the same table.
     * 
     * @param ct The table from which to fetch.
     * @param fieldNumbers
     *            The field numbers corresponding to the columns to be fetched.
     *            Field numbers whose columns exist in supertables will be
     *            ignored.
     * @param className the class name of the object of the request
     * @return A fetch request object.
     */
    public FetchRequest getFetchRequest(ClassTable ct, int[] fieldNumbers, String className)
    {
        RequestIdentifier reqID = new RequestIdentifier(ct,fieldNumbers,RequestIdentifier.Type.FETCH,className);

        FetchRequest req = (FetchRequest) requestsByID.get(reqID);
        if (req == null)
        {
            req = new FetchRequest(ct, fieldNumbers);
            requestsByID.put(reqID, req);
        }
        return req;
    }

    /**
     * Returns a request object that will update a row in the given table. The
     * store manager will cache the request object for re-use by subsequent
     * requests to the same table.
     * @param ct The table in which to update.
     * @param fieldNumbers The field numbers corresponding to the columns to be
     * updated. Field numbers whose columns exist in supertables will be
     * ignored.
     * @param className the class name of the object of the request
     * @return An update request object.
     */
    public UpdateRequest getUpdateRequest(ClassTable ct,int[] fieldNumbers, String className)
    {
        RequestIdentifier reqID = new RequestIdentifier(ct, fieldNumbers,
                RequestIdentifier.Type.UPDATE, className);
        UpdateRequest req = (UpdateRequest) requestsByID.get(reqID);
        if (req == null)
        {
            req = new UpdateRequest(ct, fieldNumbers);
            requestsByID.put(reqID, req);
        }
        return req;
    }

    /**
     * Returns a request object that will locate a row from the given table. The
     * store manager will cache the request object for re-use by subsequent
     * requests to the same table.
     * 
     * @param ct The table from which to locate.
     * @param className the class name of the object of the request
     * @return A locate request object.
     */
    public LocateRequest getLocateRequest(ClassTable ct, String className)
    {
        RequestIdentifier reqID = new RequestIdentifier(ct, null,
                RequestIdentifier.Type.LOCATE,className);
        LocateRequest req = (LocateRequest) requestsByID.get(reqID);
        if (req == null)
        {
            req = new LocateRequest(ct);
            requestsByID.put(reqID, req);
        }
        return req;
    }
    
    /**
     * Returns a request object that will delete a row from the given table. The
     * store manager will cache the request object for re-use by subsequent
     * requests to the same table.
     * 
     * @param ct The table from which to delete.
     * @param className the class name of the object of the request
     * @return A deletion request object.
     */
    public DeleteRequest getDeleteRequest(ClassTable ct, String className)
    {
        RequestIdentifier reqID = new RequestIdentifier(ct, null,
                RequestIdentifier.Type.DELETE,className);
        DeleteRequest req = (DeleteRequest) requestsByID.get(reqID);
        if (req == null)
        {
            req = new DeleteRequest(ct);
            requestsByID.put(reqID, req);
        }
        return req;
    }

    /**
     * Returns the column info for a database table. This should be used instead
     * of making direct calls to DatabaseMetaData.getColumns().
     * <p>
     * Where possible, this method loads and caches column info for more than
     * just the table being requested, improving performance by reducing the
     * overall number of calls made to DatabaseMetaData.getColumns() (each of
     * which usually results in one or more database queries).
     * </p>
     * @param table The table (or view).
     * @param conn A JDBC connection to the database.
     * @return A list of ColumnInfo objects describing the columns of the table.
     * The list is in the same order as was supplied by getColumns(). If no
     * column info is found for the given table, an empty list is returned.
     * @see org.jpox.store.rdbms.columninfo.ColumnInfo
     */
    public List getColumnInfo(Table table, Connection conn)
    throws SQLException
    {
        SQLIdentifier tableName = (SQLIdentifier) table.getName();
        List cols = null;
        if (storeDataByClass.size() == 0)
        {
            // No SchemaData yet so go to the database directly to find any
            // info for this table.
            cols = RDBMSStoreHelper.getColumnInfo(this, table, conn);
        }
        else
        {
            synchronized (this)
            {
                long now = System.currentTimeMillis();
                /*
                 * If we have cached column info that hasn't expired yet, see if
                 * it contains info for the given table.
                 */
                if (now >= columnInfoReadTimestamp && 
                    now < columnInfoReadTimestamp + COLUMN_INFO_EXPIRATION_MS)
                {
                    cols = (List) columnInfoByTableName.get(tableName);
                }
                /*
                 * If we have no info for that table, or stale info overall,
                 * refresh the column info cache.
                 */
                if (cols == null)
                {
                    // Extract a list of known JDO table names.
                    Set table_keys = new HashSet();
                    Set validated_tables = new HashSet();
                    for (Iterator i = storeDataByClass.values().iterator(); i.hasNext();)
                    {
                        RDBMSStoreData sd = (RDBMSStoreData) i.next();
                        table_keys.add(sd.getTableIdentifier());
                        if (sd.hasTable() && sd.getTable().isValidated())
                        {
                            validated_tables.add(sd.getTableIdentifier());
                        }
                    }

                    // Query for column info on all tables in catalog/schema,
                    // discarding info for any tables that aren't JDO tables,
                    // or that are JDO tables but are already validated.
                    HashMap cim = new HashMap();
                    DatabaseMetaData dmd = conn.getMetaData();
                    ResultSet rs = dmd.getColumns(table.getCatalogName(), table.getSchemaName(), null, null);

                    try
                    {
                        while (rs.next())
                        {
                            String tableIdentifierCatalogName = rs.getString(1);
                            if (rs.wasNull() ||
                                (tableIdentifierCatalogName != null && tableIdentifierCatalogName.length() < 1))
                            {
                                tableIdentifierCatalogName = null;
                            }
                            String tableIdentifierSchemaName = rs.getString(2);
                            if (rs.wasNull() ||
                                (tableIdentifierSchemaName != null && tableIdentifierSchemaName.length() < 1))
                            {
                                tableIdentifierSchemaName = null;
                            }
                            String tableIdentifierTableName = rs.getString(3);
                            if (rs.wasNull() ||
                                (tableIdentifierTableName != null && tableIdentifierTableName.length() < 1))
                            {
                                tableIdentifierTableName = null;
                            }
                            SQLIdentifier tblName = new SQLIdentifier(
                                    dba,
                                    tableIdentifierTableName);
                            if (!table_keys.contains(tblName) &&
                                this.getPMFContext().getPmfConfiguration().getFullyQualifiedNames())
                            {
                                if (tableIdentifierSchemaName != null &&
                                    tableIdentifierCatalogName == null)
                                {
                                    tblName = new SQLIdentifier(dba,
                                            tableIdentifierSchemaName +
                                            ((DatabaseAdapter) dba).getCatalogSeparator() +
                                            tableIdentifierTableName);
                                }
                                else if (tableIdentifierCatalogName != null &&
                                         tableIdentifierSchemaName != null)
                                {
                                    tblName = new SQLIdentifier(dba,
                                            tableIdentifierCatalogName +
                                            ((DatabaseAdapter) dba).getCatalogSeparator() +
                                            tableIdentifierSchemaName +
                                            ((DatabaseAdapter) dba).getCatalogSeparator() +
                                            tableIdentifierTableName);
                                }
                                else if (tableIdentifierCatalogName != null &&
                                         tableIdentifierSchemaName == null)
                                {
                                    tblName = new SQLIdentifier(dba,
                                            tableIdentifierCatalogName +
                                            ((DatabaseAdapter) dba).getCatalogSeparator() +
                                            tableIdentifierTableName);
                                }
                            }

                            if (table_keys.contains(tblName))
                            {
                                List l = (List) cim.get(tblName);
                                if (l == null)
                                {
                                    l = new ArrayList();
                                    cim.put(tblName, l);
                                }
                                l.add(((DatabaseAdapter)dba).newColumnInfo(rs));
                            }
                        }
                    }
                    finally
                    {
                        rs.close();
                    }

                    if (JPOXLogger.RDBMS_SCHEMA.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SCHEMA.debug(LOCALISER_RDBMS.msg(
                                "RDBMSManager.ColumnInfoLoaded", catalogName,
                                schemaName, "" + cim.size(), "" +
                                (System.currentTimeMillis() - now)));
                    }

                    // Replace the old cache (if any) with the new one
                    columnInfoByTableName = cim;
                    columnInfoReadTimestamp = now;

                    // Finally, lookup info for the desired table in the new
                    // cache.
                    cols = (List) columnInfoByTableName.get(tableName);
                    if (cols == null)
                    {
                        cols = Collections.EMPTY_LIST;
                        JPOXLogger.RDBMS_SCHEMA.warn(LOCALISER_RDBMS.msg("RDBMSManager.ColumnInfoError", table));
                    }
                }
            }
        }
        return cols;
    }

    /**
     * Resolves an identifier macro. The public fields <var>clazz </var>,
     * <var>fieldName </var>, and <var>subfieldName </var> of the given macro
     * are taken as inputs, and the public <var>value </var> field is set to the
     * SQL identifier of the corresponding database table or column.
     * @param im The macro to resolve.
     */
    public void resolveIdentifierMacro(MacroString.IdentifierMacro im)
    {
        FCOTable ct = getTable(im.className);
        if (im.fieldName == null)
        {
            im.value = ct.getName().toString();
            return;
        }

        JavaTypeMapping m;
        if (im.fieldName.equals("this"))
        {
            if (!(ct instanceof ClassTable))
            {
                throw new JDOUserException(LOCALISER_RDBMS.msg("RDBMSManager.TableIDColumnError", im.className));
            }

            if (im.subfieldName != null)
            {
                throw new JDOUserException(LOCALISER_RDBMS.msg("RDBMSManager.FieldTableError", im.className, im.fieldName, im.subfieldName));
            }
            m = ((ClassTable) ct).getIDMapping();
        }
        else
        {
            FieldMetaData fmd = getMetaDataManager().getMetaDataForField(im.className, im.fieldName, clr);
            m = ct.getFieldMapping(fmd);
            JDOTable t = getTable(fmd);
            if (im.subfieldName == null)
            {
                if (t != null)
                {
                    im.value = t.getName().toString();
                    return;
                }
            }
            else
            {
                if (t instanceof SetTable)
                {
                    SetTable st = (SetTable) t;
                    if (im.subfieldName.equals("owner"))
                    {
                        m = st.getOwnerMapping();
                    }
                    else if (im.subfieldName.equals("element"))
                    {
                        m = st.getElementMapping();
                    }
                    else
                    {
                        throw new JDOUserException(LOCALISER_RDBMS.msg(
                                "RDBMSManager.InvalidPseudoFieldNameForSet",
                                im.subfieldName, im));
                    }
                }
                else if (t instanceof ListTable)
                {
                    ListTable lt = (ListTable) t;
                    if (im.subfieldName.equals("owner"))
                    {
                        m = lt.getOwnerMapping();
                    }
                    else if (im.subfieldName.equals("element"))
                    {
                        m = lt.getElementMapping();
                    }
                    else if (im.subfieldName.equals("index"))
                    {
                        m = lt.getIndexMapping();
                    }
                    else
                    {
                        throw new JDOUserException(LOCALISER_RDBMS.msg(
                                "RDBMSManager.InvalidPseudoFieldNameForSet",
                                im.subfieldName, im));
                    }
                }
                else if (t instanceof MapTable)
                {
                    MapTable mt = (MapTable) t;
                    if (im.subfieldName.equals("owner"))
                    {
                        m = mt.getOwnerMapping();
                    }
                    else if (im.subfieldName.equals("key"))
                    {
                        m = mt.getKeyMapping();
                    }
                    else if (im.subfieldName.equals("value"))
                    {
                        m = mt.getValueMapping();
                    }
                    else
                    {
                        throw new JDOUserException(LOCALISER_RDBMS.msg(
                                "RDBMSManager.InvalidPseudoFieldNameForMap",
                                im.subfieldName, im));
                    }
                }
                else
                {
                    throw new JDOUserException(LOCALISER_RDBMS.msg(
                            "RDBMSManager.FieldTableError", im.className,
                            im.fieldName, im.subfieldName));
                }
            }
        }
        im.value = ((Column)m.getDataStoreMapping(0).getDatastoreField()).getName().toString();
    }

    /**
     * Accessor for the SQL identifier for a table.
     * Consults the StoreData for the class. If the class isn't known
     * it generates the table name.
     * @param cmd ClassMetaData for the class
     * @return The SQL identifier for the table
     */
    private SQLIdentifier getTableIdentifier(ClassMetaData cmd)
    {
        if (storeDataByClass.containsKey(cmd.getFullClassName()))
        {
            RDBMSStoreData sd = (RDBMSStoreData) storeDataByClass.get(cmd.getFullClassName());
            SQLIdentifier tableName = new SQLIdentifier(dba, sd.getTableName());
            return tableName;
        }
        else
        {
            SQLIdentifier tableName = RDBMSStoreHelper.generateTableNameForMetaData((DatabaseAdapter) dba, cmd);
            return tableName;
        }
    }

    /**
     * Accessor for the SQL identifier for a table.
     * Consults the StoreData for the field. If the field isn't known
     * it generates the table name.
     * @param fmd FieldMetaData for the class
     * @return The SQL identifier for the table
     */
    private SQLIdentifier getTableIdentifier(FieldMetaData fmd)
    {
        if (storeDataByClass.containsKey(fmd.getField()))
        {
            RDBMSStoreData sd = (RDBMSStoreData) storeDataByClass.get(fmd.getField());
            SQLIdentifier tableName = new SQLIdentifier(dba, sd.getTableName());
            return tableName;
        }
        else
        {
            SQLIdentifier tableName = RDBMSStoreHelper.generateTableNameForMetaData((DatabaseAdapter) dba, fmd);
            return tableName;
        }
    }

    // ============================ Inner classes ==============================

    /**
     * An abstract base class for RDBMSManager transactions that perform some
     * management function on the database.
     * <p>
     * Management transactions may be retried in the face of SQL exceptions to
     * work around failures caused by transient conditions, such as DB
     * deadlocks.
     * </p>
     */
    private abstract class MgmtTransaction
    {
        public static final String MAX_RETRIES_PROPERTY = "org.jpox.store.maxRetries";
        protected final int isolationLevel;
        protected final int maxRetries;

        /**
         * Constructs a new management transaction having the given isolation
         * level.
         * @param isolationLevel One of the isolation level constants from
         * java.sql.Connection.
         */
        private MgmtTransaction(int isolationLevel)
        {
            this.isolationLevel = isolationLevel;
            String s = System.getProperty(MAX_RETRIES_PROPERTY, "3");

            int mr;
            try
            {
                mr = Integer.parseInt(s);
            }
            catch (NumberFormatException e)
            {
                JPOXLogger.RDBMS_SCHEMA.warn("Failed parsing " +
                        MAX_RETRIES_PROPERTY + " property, value was " + s);
                mr = 3;
            }
            maxRetries = mr;
        }

        /**
         * Returns a description of the management transaction. Subclasses
         * should override this method so that transaction failures are given an
         * appropriate exception message.
         * 
         * @return A description of the management transaction.
         */
        public abstract String toString();

        /**
         * Implements the body of the transaction.
         * @param conn A connection to the database. If the selected isolation
         * level is Connection.TRANSACTION_NONE, the connection has auto-commit
         * set to true, otherwise auto-commit is false.
         * @param clr the ClassLoaderResolver
         * @exception SQLException Thrown if the transaction fails due to a
         * database error that should allow the entire transaction to be
         * retried.
         */
        protected abstract void execute(Connection conn,ClassLoaderResolver clr)
        throws SQLException;

        /**
         * Executes the transaction.
         * <p>
         * A database connection is acquired and the
         * {@link #execute(Connection,ClassLoaderResolver)}method is invoked.
         * If the selected isolation level is not Connection.TRANSACTION_NONE,
         * then commit() or rollback() is called on the connection according to
         * whether the invocation succeeded or not. If the invocation failed the
         * sequence is repeated, up to a maximum of <var>maxRetries </var>
         * times, configurable by the system property org.jpox.store.maxRetries.
         * @param clr the ClassLoaderResolver
         * @exception JDODataStoreException If a SQL exception occurred even
         * after <var>maxRetries </var> attempts.
         */
        public final void execute(ClassLoaderResolver clr)
        {
            int attempts = 0;
            for (;;)
            {
                try
                {
                    Connection conn = ((DatabaseAdapter) dba).getConnection(ds,
                            userName, password, isolationLevel);
                    try
                    {
                        boolean succeeded = false;
                        try
                        {
                            execute(conn, clr);
                            succeeded = true;
                        }
                        finally
                        {
                            if (isolationLevel != Connection.TRANSACTION_NONE)
                            {
                                if (succeeded)
                                {
                                    conn.commit();
                                }
                                else
                                {
                                    conn.rollback();
                                }
                            }
                        }
                    }
                    finally
                    {
                        ((DatabaseAdapter) dba).closeConnection(conn);
                    }
                    break;
                }
                catch (SQLException e)
                {
                    if (++attempts >= maxRetries)
                    {
                        throw new JDODataStoreException("SQL exception: " + this, e);
                    }
                }
            }
        }
    }

    /**
     * A management transaction that adds a set of classes to the RDBMSManager,
     * making them usable for persistence.
     * <p>
     * This class embodies the work necessary to activate a persistent class and
     * ready it for storage management. It is the primary mutator of a
     * RDBMSManager.
     * </p>
     * <p>
     * Adding classes is an involved process that includes the creation and/or
     * validation in the database of tables, views, and table constraints, and
     * their corresponding Java objects maintained by the RDBMSManager. Since
     * it's a management transaction, the entire process is subject to retry on
     * SQL exceptions. It is responsible for ensuring that the procedure either
     * adds <i>all </i> of the requested classes successfully, or adds none of
     * them and preserves the previous state of the RDBMSManager exactly as it
     * was.
     * </p>
     */
    private class ClassAdder extends MgmtTransaction
    {
        private HashSet schemaDataAdded = new HashSet();
        private final String[] classNames;
        private Connection schemaConnection = null;
        private TypeManager typeManager = TypeManager.getTypeManager();

        private String[] filterSimpleClasses(String[] classNames, ClassLoaderResolver clr)
        {
            int filteredClasses = 0;
            for (int i = 0; i < classNames.length; ++i)
            {
                if (typeManager.isSupportedType(classNames[i], clr))
                {
                    classNames[i] = null;
                    ++filteredClasses;
                }
            }
            if (filteredClasses == 0)
            {
                return classNames;
            }
            String[] restClasses = new String[classNames.length - filteredClasses];
            int m = 0;
            for (int i = 0; i < classNames.length; ++i)
            {
                if (classNames[i] != null)
                {
                    restClasses[m++] = classNames[i];
                }
            }
            return restClasses;
        }

        /**
         * Constructs a new class adder transaction that will add the given
         * classes to the RDBMSManager.
         * 
         * @param classNames Names of the class(es) to be added.
         */
        private ClassAdder(String[] classNames, ClassLoaderResolver clr)
        {
            super(Connection.TRANSACTION_SERIALIZABLE);
            this.classNames = filterSimpleClasses(classNames, clr);
        }

        /**
         * Method to give a string version of this object.
         * @return The String version of this object.
         */
        public String toString()
        {
            return LOCALISER_RDBMS.msg("RDBMSManager.AddClassesToSchema", catalogName, schemaName);
        }

        /**
         * Method to perform the action using the specified connection to the
         * datastore.
         * @param conn Connection to the datastore.
         * @param clr the ClassLoaderResolver
         * @throws SQLException Thrown if an error occurs in execution.
         */
        protected void execute(Connection conn, ClassLoaderResolver clr)
        throws SQLException
        {
            // I think, it is not necessary to put this into the synchronized
            // block, because classes doesn't change after the constructor.
            if (classNames.length == 0)
            {
                return;
            }
            synchronized (RDBMSManager.this)
            {
                classAdder = this;
                schemaConnection = conn;
                try
                {
                    addClassTablesAndValidate(classNames, clr);
                }
                finally
                {
                    schemaConnection = null;
                    classAdder = null;
                }
            }
        }

        /**
         * Called by RDBMSManager.addClasses() when it has been recursively
         * re-entered. This just adds table objects for the requested classes
         * and returns.
         * @param classNames Names of the class(es) to be added.
         * @param clr the ClassLoaderResolver
         */
        private void addClasses(String[] classNames, ClassLoaderResolver clr)
        {
            if (schemaConnection == null)
            {
                throw new IllegalStateException(LOCALISER_RDBMS.msg("RDBMSManager.AddClassesError"));
            }

            // Not sure, whether filterSimpleClasses(...) is necessary here.
            classNames = filterSimpleClasses(classNames, clr);
            if (classNames.length == 0)
            {
                return;
            }
            addClassTables(classNames, clr);
        }

        /**
         * Adds a new table object (ie ClassTable or ClassView) for every
         * class in the given list. These classes
         * <ol>
         * <li>requires an extent</li>
         * <li>does not yet have an extent (ie table) initialized in the store
         * manager.</li>
         * </ol>
         * <p>
         * This doesn't initialize or validate the tables, it just adds the
         * table objects to the RDBMSManager's internal data structures.
         * </p>
         * 
         * @param classNames Names of class(es) whose tables are to be added.
         * @param clr the ClassLoaderResolver
         */
        private void addClassTables(String[] classNames,
                                    ClassLoaderResolver clr)
        {
            Iterator i = getReferencedClasses(classNames, clr).iterator();
            try
            {
                if (starter != null && starterInitialised && !starter.isOpen())
                {
                    starter.open();
                }

                // Pass through the classes and create necessary tables
	            while (i.hasNext())
	            {
	                ClassMetaData cmd = (ClassMetaData) i.next();
	                RDBMSStoreData sd = (RDBMSStoreData) storeDataByClass.get(cmd.getFullClassName());
	                if ((sd == null) &&
	                    (cmd.getIdentityType() != IdentityType.NONDURABLE))
	                {
	                    // For application-identity classes with user-defined identities we
	                    // check for use of the objectid-class in different inheritance trees
	                    // We prevent this to avoid problems later on.
	                    // The JDO2 builtin objectid-classes are allowed to be duplicated.
	                    if (cmd.getIdentityType() == IdentityType.APPLICATION)
	                    {
	                        String objectid_class_name=cmd.getObjectidClass();
	                        if (objectid_class_name.indexOf("javax.jdo.") != 0)
	                        {
	                            Collection c=(Collection)storeDataByAppIdClass.get(cmd.getObjectidClass());
	                            if (c != null)
	                            {
	                                boolean in_same_tree=false;
	                                String sample_class_in_other_tree=null;
	                                
	                                Iterator id_iter=c.iterator();
	                                while (id_iter.hasNext())
	                                {
	                                    RDBMSStoreData id_sd=(RDBMSStoreData)id_iter.next();
	                                    sample_class_in_other_tree=id_sd.getName();
	                                    if (ClassUtils.classesAreDescendents(clr,cmd.getFullClassName(),id_sd.getName()))
	                                    {
	                                        in_same_tree = true;
	                                        break;
	                                    }
	                                }
	                                
	                                if (!in_same_tree)
	                                {
	                                    String error_msg=LOCALISER_RDBMS.msg("RDBMSManager.ObjectIdClassDuplicated",
	                                        cmd.getFullClassName(),
	                                        objectid_class_name,
	                                        sample_class_in_other_tree);
	                                    JPOXLogger.GENERAL.error(error_msg);
	                                    throw new JDOUserException(error_msg);
	                                }
	                            }
	                        }
	                    }

                        if (cmd.isEmbeddedOnly())
                        {
                            JPOXLogger.RDBMS.info(LOCALISER.msg("RDBMSManager.NoTableForEmbeddedClass", cmd.getFullClassName()));
                            // Nothing to do. Only persisted as SCO.
                        }
                        else
                        {
                            InheritanceMetaData imd=cmd.getInheritanceMetaData();
                            RDBMSStoreData sdNew=null;
                            if (imd.getStrategyValue() == InheritanceStrategy.NO_TABLE)
                            {
                                // Just add the SchemaData entry with no table - managed by subclass
                                sdNew = new RDBMSStoreData(cmd, null, false);
                                registerStoreData(sdNew);
                            }
                            else if (imd.getStrategyValue() == InheritanceStrategy.NEW_TABLE)
                            {
                                // Get the associated table for this class
                                SQLIdentifier tableName = getTableIdentifier(cmd);
                                
                                // Check that this table isn't already in use
                                RDBMSStoreData existing_sd = (RDBMSStoreData)getStoreDataForTable(tableName.getSQLIdentifier());
                                if (existing_sd != null &&
                                        cmd.getFullClassName() != existing_sd.getName())
                                {
                                    String msg=LOCALISER_RDBMS.msg("RDBMSManager.TableAlreadyInUseError",cmd.getFullClassName(),
                                        tableName.getSQLIdentifier(),existing_sd.getName());
                                    JPOXLogger.RDBMS.warn(msg);
                                }
                                
                                FCOTable t;
                                boolean hasViewDef = false;
                                if (((DatabaseAdapter) dba).getVendorID() != null)
                                {
                                    hasViewDef = cmd.hasExtension("view-definition" + '-' + ((DatabaseAdapter) dba).getVendorID());
                                }
                                if (!hasViewDef)
                                {
                                    hasViewDef = cmd.hasExtension("view-definition");
                                }
                                if (hasViewDef)
                                {
                                    t = new ClassView(tableName, cmd, RDBMSManager.this);
                                }
                                else
                                {
                                    t = new ClassTable(tableName, cmd, RDBMSManager.this);
                                }
                                
                                sdNew = new RDBMSStoreData(cmd, t, true);
                                registerStoreData(sdNew);
                            }
                            else if (imd.getStrategyValue() == InheritanceStrategy.SUPERCLASS_TABLE)
                            {
                                ClassMetaData[] managingCmds = getClassesManagingTableForClass(cmd);
                                
                                // Find the superclass - should have been created first
                                ClassTable superTable=null;
                                if (managingCmds != null && managingCmds.length == 1)
                                {
                                    RDBMSStoreData superData = (RDBMSStoreData)storeDataByClass.get(managingCmds[0].getFullClassName());
                                    if (superData != null)
                                    {
                                        // Specify the table if it already exists
                                        superTable = (ClassTable)superData.getTable();
                                    }
                                    sdNew = new RDBMSStoreData(cmd, superTable, false);
                                    registerStoreData(sdNew);
                                }
                                else
                                {
                                    String msg = LOCALISER_RDBMS.msg("RDBMSManager.UnableToFindSuperclassTable", cmd.getFullClassName());
                                    JPOXLogger.JDO.error(msg);
                                    throw new JDOUserException(msg);
                                }
                            }
                            schemaDataAdded.add(sdNew);
                        }
	                }
	            }

                // For data where the table wasn't defined, make a second pass
                // This is necessary where a subclass uses "superclass-table" and the superclass' table
                // hadn't been defined at the point of adding this class
                Iterator addedIter = schemaDataAdded.iterator();
                while (addedIter.hasNext())
                {
                    RDBMSStoreData data = (RDBMSStoreData)addedIter.next();
                    if (data.getTable() == null && data.isFCO())
                    {
                        ClassMetaData cmd = (ClassMetaData)data.getMetaData();
                        InheritanceMetaData imd = cmd.getInheritanceMetaData();
                        if (imd.getStrategyValue() == InheritanceStrategy.SUPERCLASS_TABLE)
                        {
                            ClassMetaData[] managingCmds = getClassesManagingTableForClass(cmd);
                            ClassTable superTable=null;
                            if (managingCmds != null && managingCmds.length == 1)
                            {
                                RDBMSStoreData superData = (RDBMSStoreData)storeDataByClass.get(managingCmds[0].getFullClassName());
                                if (superData == null)
                                {
                                    String msg = LOCALISER_RDBMS.msg("RDBMSManager.UnableToFindSuperclassTable", cmd.getFullClassName());
                                    JPOXLogger.JDO.error(msg);
                                    throw new JDOUserException(msg);
                                }
                                superTable = (ClassTable)superData.getTable();
                                data.setTable(superTable);
                            }
                        }
                    }
                }
            }
            finally
            {
                if (starter != null && starterInitialised && starter.isOpen())
                {
                    starter.close();
                }
            }
        }

        /**
         * Returns a List of {@link ClassMetaData}objects representing the set
         * of classes consisting of the requested classes plus any and all other
         * classes they may reference, directly or indirectly. The returned list
         * is ordered by dependency.
         * @param classNames Array of names of persistence-capable classes.
         * @param clr the ClassLoaderResolver
         * @return A List of <tt>ClassMetaData</tt> objects.
         * @exception ClassNotPersistenceCapableException If any of the given
         * classes is not persistence-capable.
         */
        private List getReferencedClasses(String[] classNames,
                ClassLoaderResolver clr)
        {
            List cmds = new ArrayList();
            for (int i = 0; i < classNames.length; ++i)
            {
                ClassMetaData cmd = getMetaDataManager().getMetaDataForClass(
                        classNames[i], clr);
                
                Class clazz = clr.classForName(classNames[i]);
                if(clazz == null || !clazz.isInterface())
                {
	                if (cmd == null)
	                {
	                    JPOXLogger.RDBMS_SCHEMA.warn(LOCALISER_RDBMS.msg(
	                            "RDBMSManager.AddClassesInvalidClassError",
	                            classNames[i]));
	                    throw new ClassNotPersistenceCapableException(classNames[i]);
	                }
	                cmds.addAll(getMetaDataManager().getReferencedClassMetaData(
	                        cmd, ((DatabaseAdapter) dba).getVendorID(), clr));
                }
            }
            return cmds;
        }

        private List removeDuplicateTables(List newTables)
        {
            List result = new ArrayList();
            Set uniqueTables = new TreeSet(new Comparator()
            {
                /*
                 * (non-Javadsoc)
                 * @see java.util.Comparator#compare(java.lang.Object,
                 * java.lang.Object)
                 */
                public int compare(Object o1, Object o2)
                {
                    JDOTable t1 = (JDOTable) o1;
                    JDOTable t2 = (JDOTable) o2;
                    return ((SQLIdentifier) t1.getName()).getSQLIdentifier().compareTo(((SQLIdentifier) t2.getName()).getSQLIdentifier());
                }
            });
            uniqueTables.addAll(newTables);
            result.addAll(uniqueTables);
            return result;
        }
        
        /**
         * Adds a new table object (ie ClassTable or ClassView) for every class
         * in the given list that 1) requires an extent and 2) does not yet have
         * an extent (ie table) initialized in the store manager.
         * <p>
         * After all of the table objects, including any other tables they might
         * reference, have been added, each table is initialized and validated
         * in the database.
         * </p>
         * <p>
         * If any error occurs along the way, any table(s) that were created are
         * dropped and the state of the RDBMSManager is rolled back to the point
         * at which this method was called.
         * </p>
         * @param classNames The class(es) whose tables are to be added.
         * @param clr the ClassLoaderResolver
         */
        private void addClassTablesAndValidate(String[] classNames,
                ClassLoaderResolver clr)
        {
            Map schemaDataByClassPrev = new WeakHashMap(storeDataByClass);
            MultiMap schemaDataByClassAppIDPrev = new MultiMap(storeDataByAppIdClass);
            List baseTablesCreated = new ArrayList();
            List baseTableConstraintsCreated = new ArrayList();
            List viewsCreated = new ArrayList();
            boolean completed = false;

            try
            {
                // Add SchemaData entries and Table's for the requested classes
                addClassTables(classNames, clr);

                // Initialisation of tables. Updates the internal representation of the
                // table to match what is required for the class(es).
                // Each time a table object is initialized, it may cause other
                // associated table objects to be added (via callbacks to addClasses())
                // so the loop isrepeated until no more initialisation is needed.
                List newBaseTables = new ArrayList();
                List newViews = new ArrayList();
                boolean someNeededInitialization;
                do
                {
                    someNeededInitialization = false;
                    Iterator i = new ArrayList(storeDataByClass.values()).iterator();
                    while (i.hasNext())
                    {
                        RDBMSStoreData sd = (RDBMSStoreData) i.next();
                        if (sd.hasTable())
                        {
                            JDOTable t = sd.getTable();

                            // Any classes managed by their own table needing initialising
                            if (!t.isInitialized())
                            {
                                t.initialize();
                                if (t instanceof ViewImpl)
                                {
                                    newViews.add(t);
                                }
                                else
                                {
                                    newBaseTables.add(t);
                                }
                                someNeededInitialization = true;
                            }

                            // Any classes that are managed by other tables needing initialising
                            if (!sd.isTableOwner() && !((ClassTable)t).managesClass(sd.getName()))
                            {
                                ((ClassTable)t).manageClass((ClassMetaData)sd.getMetaData());
                                if (!newBaseTables.contains(t))
                                {
                                    newBaseTables.add(t);
                                }
                                someNeededInitialization = true;
                            }
                        }
                        else
                        {
                            // Nothing to do for cases without their own table ("no-table" strategy) since
                            // the table is initialised to contain those fields by the subclass.
                        }
                    }
                }
                while (someNeededInitialization);

                if (AbstractTable.isDumpingDdl())
                {
                    // we have duplicate JDOTable objects in newBaseTables for
                    // bidirectional associations, each created
                    // when traversing from one side of the association. When
                    // autocreating tables,
                    // the second one will validate itself against the DB and
                    // normally find that it
                    // already exists. When the DDL is dumped into a file, the
                    // second table won't find itself in the DB
                    // and then generate the same DDL again, which we don't
                    // want.
                    // So remove all duplicate JDOTables before autocreating.
                    newBaseTables = removeDuplicateTables(newBaseTables);
                }

                // Table existence and validation.
                // a). Check for existence of the table
                // b). If autocreate, create the table if necessary
                // c). If validate, validate the table
                Iterator i = newBaseTables.iterator();
                while (i.hasNext())
                {
                    TableImpl t = (TableImpl) i.next();

                    // initialize the default value for columns
                    t.initializeDefault(schemaConnection);
                    if (checkExistTablesOrViews)
                    {
	                    if (t.exists(schemaConnection, autoCreateTables))
	                    {
	                        baseTablesCreated.add(t);
	                    }
	                    else
	                    {
	                        // Table wasn't just created, so do any autocreate of columns necessary
	                        if (t.isInitializedModified() || autoCreateColumns)
	                        {
	                            t.validateColumns(schemaConnection, autoCreateColumns);
	                        }
	                    }
                    }
                    if (validateTables)
                    {
                        t.validate(schemaConnection, false);
                    }
                    // Discard any cached column info used to validate the table
                    columnInfoByTableName.remove(t.getName());
                }

                // Table constraint existence and validation
                // a). Check for existence of the constraint
                // b). If autocreate, create the constraint if necessary
                // c). If validate, validate the constraint
                // Constraint processing is done as a separate step from table processing
                // since the constraints are dependent on tables being available
                i = newBaseTables.iterator();
                while (i.hasNext())
                {
                    TableImpl t = (TableImpl) i.next();
                    if (validateConstraints || autoCreateConstraints)
                    {
                        // TODO : split this method into checkExistsConstraints
                        // and validateConstraints
                        if (t.validateConstraints(schemaConnection, autoCreateConstraints))
                        {
                            baseTableConstraintsCreated.add(t);
                        }
                    }
                }

                // View existence and validation.
                // a). Check for existence of the view
                // b). If autocreate, create the view if necessary
                // c). If validate, validate the view
                i = newViews.iterator();
                while (i.hasNext())
                {
                    ViewImpl v = (ViewImpl) i.next();
                    if (checkExistTablesOrViews)
                    {
	                    if (v.exists(schemaConnection, autoCreateTables))
	                    {
	                        viewsCreated.add(v);
	                    }
                    }
                    if (validateTables)
                    {
                        v.validate(schemaConnection, false);
                    }

                    // Discard any cached column info used to validate the view
                    columnInfoByTableName.remove(v.getName());
                }

                completed = true;
            }
            catch (SQLException sqle)
            {
                String msg=LOCALISER_RDBMS.msg("RDBMSManager.AddClassesExceptionError", sqle);
                JPOXLogger.RDBMS_SCHEMA.error(msg);
                
                // Put exception in the log
                JPOXLogger.RDBMS_SCHEMA.error(sqle);
                throw new JDODataStoreException(msg, sqle);
            }
            catch (Exception e)
            {
                if (JDOException.class.isAssignableFrom(e.getClass()))
                {
                    throw (JDOException) e;
                }
                else
                {
                    JPOXLogger.RDBMS_SCHEMA.error(LOCALISER_RDBMS.msg("RDBMSManager.AddClassesExceptionError", e));
                }
                throw new JDOFatalInternalException(e.toString(), e);
            }
            finally
            {
                // If something went wrong, roll things back to the way they
                // were before we started.
                // This may not restore the database 100% of the time
                // (if DDL statements are not transactional) but it will always
                // put the RDBMSManager's internal structures back the way they
                // were.
                if (!completed)
                {
                    if (JPOXLogger.RDBMS_SCHEMA.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SCHEMA.debug(LOCALISER_RDBMS.msg("RDBMSManager.AddTablesRollback"));
                    }
                    storeDataByClass = schemaDataByClassPrev;
                    storeDataByAppIdClass = schemaDataByClassAppIDPrev;

                    // Tables, table constraints, and views get removed in the
                    // reverse order from which they were created.
                    try
                    {
                        ListIterator li = viewsCreated
                                .listIterator(viewsCreated.size());
                        while (li.hasPrevious())
                        {
                            ((ViewImpl) li.previous()).drop(schemaConnection);
                        }
                        li = baseTableConstraintsCreated.listIterator(baseTableConstraintsCreated.size());
                        while (li.hasPrevious())
                        {
                            ((TableImpl) li.previous()).dropConstraints(schemaConnection);
                        }
                        li = baseTablesCreated.listIterator(baseTablesCreated.size());
                        while (li.hasPrevious())
                        {
                            ((TableImpl) li.previous()).drop(schemaConnection);
                        }
                    }
                    catch (Exception e)
                    {
                        JPOXLogger.RDBMS_SCHEMA.warn(LOCALISER_RDBMS.msg(
                            "RDBMSManager.AddTablesRollbackError", e.toString()));
                    }

                    // AutoStarter - Remove all classes from the supported list
                    // that were added in this pass.
                    if (starter != null && starterInitialised)
                    {
                        try
                        {
                            if (!starter.isOpen())
                            {
                                starter.open();
                            }
                            Iterator schema_added_iter = schemaDataAdded.iterator();
                            while (schema_added_iter.hasNext())
                            {
                                RDBMSStoreData sd=(RDBMSStoreData)schema_added_iter.next();
                                starter.deleteClass(sd.getName());
                            }                            
                        }
                        finally
                        {
	                        if (starter.isOpen())
                            {
                                starter.close();
                            }
                        }
                    }
                }
                schemaDataAdded.clear();
            }
        }

        /**
         * Called by Mapping objects in the midst of RDBMSManager.addClasses()
         * to request the creation of a set table.
         * @param fmd The field metadata describing the set field.
         */
        private SetTable addSetTable(FieldMetaData fmd)
        {
            SQLIdentifier tableName = getTableIdentifier(fmd);
            SetTable st = new SetTable(tableName, fmd, RDBMSManager.this);
            RDBMSStoreData data;
            try
            {
                if (starter != null && starterInitialised && !starter.isOpen())
                {
                    starter.open();
                }

                data = new RDBMSStoreData(fmd, st);
                registerStoreData(data);
            }
            finally
            {
                if (starter != null && starterInitialised && starter.isOpen())
                {
                    starter.close();
                }
            }

            schemaDataAdded.add(data);
            return st;
        }

        /**
         * Called by Mapping objects in the midst of RDBMSManager.addClasses()
         * to request the creation of a map table.
         * @param fmd The field metadata describing the map field.
         */
        private MapTable addMapTable(FieldMetaData fmd)
        {
            SQLIdentifier tableName = getTableIdentifier(fmd);
            MapTable mt = new MapTable(tableName, fmd, RDBMSManager.this);
            RDBMSStoreData data;
            try
            {
                if (starter != null && starterInitialised && !starter.isOpen())
                {
                    starter.open();
                }
                
                data = new RDBMSStoreData(fmd, mt);
                registerStoreData(data);
            }
            finally
            {
                if (starter != null && starterInitialised && starter.isOpen())
                {
                    starter.close();
                }
            }

            schemaDataAdded.add(data);
            return mt;
        }

        /**
         * Called by Mapping objects in the midst of RDBMSManager.addClasses()
         * to request the creation of a list table.
         * @param fmd The field metadata describing the List field.
         */
        private ListTable addListTable(FieldMetaData fmd)
        {
            SQLIdentifier tableName = getTableIdentifier(fmd);
            ListTable lt = new ListTable(tableName, fmd, RDBMSManager.this);
            RDBMSStoreData data;
            try
            {
                if (starter != null && starterInitialised && !starter.isOpen())
                {
                    starter.open();
                }
                
                data = new RDBMSStoreData(fmd, lt);
                registerStoreData(data);
            }
            finally
            {
                if (starter != null && starterInitialised && starter.isOpen())
                {
                    starter.close();
                }
            }

            schemaDataAdded.add(data);
            return lt;
        }
    }
}