/**********************************************************************
Copyright (c) 2003 Andy Jefferson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2004 David Ezzio - fix the use of the default package for the persistent class
2004 Erik Bengtson - removed Connection parameter for generateTableNameForMetaData method 
2004 Andy Jefferson - added getClassForAppIdKey()
2004 Andy Jefferson - moved collection/map join table naming to generateTableName
    ...
**********************************************************************/
package org.jpox.store.rdbms;

import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;
import java.util.ListIterator;
import java.util.StringTokenizer;

import javax.jdo.JDODataStoreException;

import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.Transaction;
import org.jpox.metadata.ClassMetaData;
import org.jpox.metadata.DiscriminatorMetaData;
import org.jpox.metadata.DiscriminatorStrategy;
import org.jpox.metadata.ExtendableMetaData;
import org.jpox.metadata.FieldMetaData;
import org.jpox.metadata.InheritanceStrategy;
import org.jpox.state.StateManagerImpl;
import org.jpox.store.DatastoreClass;
import org.jpox.store.expression.NullLiteral;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.TableExpression;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.rdbms.adapter.DatabaseAdapter;
import org.jpox.store.rdbms.foreignkey.ForeignKeyInfo;
import org.jpox.store.rdbms.sqlidentifier.SQLIdentifier;
import org.jpox.store.rdbms.sqlidentifier.TableIdentifier;
import org.jpox.store.rdbms.table.ClassTable;
import org.jpox.store.rdbms.table.Table;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;
import org.jpox.util.StringUtils;

/**
 * Provides a series of uilities assisting in the datastore management process
 * for RDBMS datastores.
 *
 * @version $Revision: 1.6 $
 **/
public class RDBMSStoreHelper
{
    private static final Localiser LOCALISER=Localiser.getInstance("org.jpox.store.rdbms.Localisation");
    
    private static final int TABLE_IDENTIFIER_CATALOG = 0;
    private static final int TABLE_IDENTIFIER_SCHEMA = 1;
    private static final int TABLE_IDENTIFIER_TABLE = 2;
    
    /**
     * Method to return the table name for a class/field.
     * @param dba  Database Adapter
     * @param md   Meta data for the class/field
     * @return     The name of the table
     **/
    public static SQLIdentifier generateTableNameForMetaData(DatabaseAdapter dba,
                                                      ExtendableMetaData md)
    {
        ArrayList       name_parts=new ArrayList();
        StringTokenizer tokens;
        if (md instanceof ClassMetaData)
        {
            tokens=new StringTokenizer(((ClassMetaData)md).getFullClassName(),".");
        }
        else if (md instanceof FieldMetaData)
        {
            tokens=new StringTokenizer(((FieldMetaData)md).getFullFieldName(),".");
        }
        else
        {
            return null;
        }
        while (tokens.hasMoreTokens())
        {
            name_parts.add(tokens.nextToken());
        }

        ListIterator li=name_parts.listIterator(name_parts.size());
        
        // generates an unique name from a java class name
        String unique_name=(String)li.previous();
        String full_name  =(li.hasPrevious() ? (li.previous() + "_") : "") + unique_name;

        String sql_identifier=full_name;
        if (md instanceof ClassMetaData)
        {
            sql_identifier = ((ClassMetaData)md).getTable();

            // Fall back default
            if (sql_identifier == null)
            {
                sql_identifier = unique_name;
            }
        }
        else if (md instanceof FieldMetaData)
        {
            FieldMetaData fmd=(FieldMetaData)md;
            if( fmd.getColumnMetaData().length > 0 && fmd.getColumnMetaData()[0].getName() != null)
            {
                sql_identifier = fmd.getColumnMetaData()[0].getName();
            }

            // TODO Should we check the Join table here or the Field table ?
            // Normal Collection (join table name)
            if (fmd.getContainer() != null && fmd.getTable() != null)
            {
                sql_identifier = fmd.getTable();
            }
            // TODO Should we check the Join table here or the Field table ?
            // Normal Map (join table name)
            else if (fmd.getContainer() != null && fmd.getTable() != null)
            {
                sql_identifier = fmd.getTable();
            }

            // Fall back default
            if (sql_identifier == null)
            {
                sql_identifier = full_name;
            }
        }
        else
        {
            if (sql_identifier == null)
            {
                sql_identifier = unique_name;
            }
        }

        // Compare length of identifier with DB max table length
        int max_table_name_length = dba.getMaxTableNameLength();
        if (sql_identifier.length() > max_table_name_length)
        {
            JPOXLogger.RDBMS.warn(LOCALISER.msg("RDBMSManager.TableNameLengthError",sql_identifier,"" + max_table_name_length));
            //sql_identifier = sql_identifier.substring(0,max_table_name_length);
        }

        return new TableIdentifier(dba,unique_name,sql_identifier);
    }

    /**
     * Returns the type of a database table.
     * Uses the DatabaseMetaData to extract this information. 
     *
     * @param store_mgr   The RDBMSManager
     * @param table_name  The name of the table/view in the datastore
     * @param conn        Connection to the database.
     * @return            One of the TABLE_TYPE_* values from {@link Table}.
     * @see Table
     */
    public static int getTableType(RDBMSManager store_mgr,
                                   String table_name,
                                   Connection conn)
    throws SQLException
    {
        String table_type=null;
        DatabaseMetaData dmd=conn.getMetaData();
        String[] c=splitTableIdentifierName(store_mgr.getDatabaseAdapter().getCatalogSeparator(), table_name);
        ResultSet rs=dmd.getTables(c[TABLE_IDENTIFIER_CATALOG],
                                   c[TABLE_IDENTIFIER_SCHEMA] == null ? store_mgr.getSchemaName() : c[TABLE_IDENTIFIER_SCHEMA],
                                   c[TABLE_IDENTIFIER_TABLE],null);

        try
        {
            while (rs.next())
            {
                if (c[TABLE_IDENTIFIER_TABLE].equalsIgnoreCase(rs.getString(3)))
                {
                    table_type = rs.getString(4).toUpperCase();
                    break;
                }
            }
        }
        finally
        {
            rs.close();
        }

        if (table_type == null)
        {
            return Table.TABLE_TYPE_MISSING;
        }
        else if (table_type.equals("TABLE"))
        {
            return Table.TABLE_TYPE_TABLE;
        }
        else if (table_type.equals("VIEW"))
        {
            return Table.TABLE_TYPE_VIEW;
        }
        else
        {
            return Table.TABLE_TYPE_UNKNOWN;
        }
    }

    /**
     * Returns the column info for a database table.
     * Uses the DatabaseMetaData to extract this information. 
     *
     * @param store_mgr   The RDBMSManager
     * @param table       The table/view in the datastore
     * @param conn        Connection to the datastore
     * @return  A list of ColumnInfo objects describing the columns of the
     *          table.  The list is in the same order as was supplied by
     *          getColumns().  If no column info is found for the given table,
     *          an empty list is returned.
     * @see org.jpox.store.rdbms.columninfo.ColumnInfo
     */
    public static List getColumnInfo(RDBMSManager store_mgr,
                                     Table table,
                                     Connection conn)
    throws SQLException
    {
        String tableName = ((SQLIdentifier) table.getName()).getSQLIdentifier();

        List cols=new ArrayList();
        DatabaseMetaData dmd=conn.getMetaData();
        String[] c=splitTableIdentifierName(store_mgr.getDatabaseAdapter().getCatalogSeparator(), tableName);
        ResultSet rs=dmd.getColumns(c[TABLE_IDENTIFIER_CATALOG],
                     c[TABLE_IDENTIFIER_SCHEMA] == null ? store_mgr.getSchemaName() : c[TABLE_IDENTIFIER_SCHEMA],
                     c[TABLE_IDENTIFIER_TABLE],null);

        try
        {
            while (rs.next())
            {
                cols.add(store_mgr.getDatabaseAdapter().newColumnInfo(rs));
            }
        }
        finally
        {
            rs.close();
        }

        return cols;
    }

    /**
     * Returns the foreign key info for a database table.
     * Uses the DatabaseMetaData to extract this information. 
     *
     * @param store_mgr   The RDBMSManager
     * @param table_name  The name of the table/view in the datastore
     * @param conn        Connection to the datastore
     * @return  A list of ForeignKeyInfo objects describing the columns of the
     *          table's foreign keys. The list is in the same order as was
     *          supplied by get??portedKeys(). If no column info is found for
     *          the given table, an empty list is returned.
     * @see ForeignKeyInfo
     */
    public static List getForeignKeyInfo(RDBMSManager store_mgr,
                                         String table_name,
                                         Connection conn)
    throws SQLException
    {
        List             fk_cols=new ArrayList();
        DatabaseMetaData dmd=conn.getMetaData();
        String[]         c=splitTableIdentifierName(store_mgr.getDatabaseAdapter().getCatalogSeparator(), table_name);
        ResultSet        rs=dmd.getImportedKeys(c[TABLE_IDENTIFIER_CATALOG],
                     c[TABLE_IDENTIFIER_SCHEMA] == null ? store_mgr.getSchemaName() : c[TABLE_IDENTIFIER_SCHEMA],
                     c[TABLE_IDENTIFIER_TABLE]);

        try
        {
            while (rs.next())
            {
                ForeignKeyInfo fki=store_mgr.getDatabaseAdapter().newForeignKeyInfo(rs);

                // This is necessary because PostgreSQL returns duplicate rows
                // from getImportedKeys().
                if (!fk_cols.contains(fki))
                {
                    fk_cols.add(fki);
                }
            }
        }
        finally
        {
            rs.close();
        }

        return fk_cols;
    }
 
    /**
     * Method to split a fully-qualified database table name into its
     * constituent parts (CATALOG.SCHEMA.TABLE).
     * @param separator Separator character
     * @param name The fully qualified name.
     * @return The separated parts of the name
     **/
    public static String[] splitTableIdentifierName(String separator, String name)
    {
        String[] result = new String[3];

        int p = name.indexOf(separator);
        if (p < 0)
        {
            // Table name specified
            result[TABLE_IDENTIFIER_TABLE] = name;
        }
        else
        {
            int p1 = name.indexOf(separator, p + separator.length());
            if (p1 < 0)
            {
                // Schema and Table name specified
                result[TABLE_IDENTIFIER_SCHEMA] = name.substring(0, p);
                result[TABLE_IDENTIFIER_TABLE] = name.substring(p + separator.length());
            }
            else
            {
                // Catalog, Schema and Table name specified
                result[TABLE_IDENTIFIER_CATALOG] = name.substring(0, p);
                result[TABLE_IDENTIFIER_SCHEMA] = name.substring(p + separator.length(), p1);
                result[TABLE_IDENTIFIER_TABLE] = name.substring(p1 + separator.length());
            }
        }
        if (result[TABLE_IDENTIFIER_SCHEMA] != null &&
            result[TABLE_IDENTIFIER_SCHEMA].length() < 1)
        {
            result[TABLE_IDENTIFIER_SCHEMA] = null;
        }
        if (result[TABLE_IDENTIFIER_CATALOG] != null &&
            result[TABLE_IDENTIFIER_CATALOG].length() < 1)
        {
            result[TABLE_IDENTIFIER_CATALOG] = null;
        }
        return result;
    }

    /**
     * Utility that takes an id and a list of possible class RDBMSStoreData
     * and finds which of these classes contains the object with that id.
     * Works via a query to the datastore. Operates for all types of identity.
     * @param pm Persistence Manager
     * @param storeMgr Store Manager
     * @param id The id
     * @param schemaDataOptions List of possible RDBMSStoreData
     * @return RDBMSStoreData for the class with this key, or null if
     * no object was found for the passed classes with this id.
     **/
    public static RDBMSStoreData getClassForIdKey(PersistenceManager pm,
                                                  RDBMSManager storeMgr,
                                                  Object id,
                                                  List schemaDataOptions)
    {
        RDBMSStoreData data=null;

        // Check for input error
        if (schemaDataOptions == null || schemaDataOptions.size() == 0 ||
            id == null)
        {
            return null;
        }

        // Calculate max length of class name for identifier
        int metadata_id_len = 0;
        Iterator optionsIter = schemaDataOptions.iterator();
        while (optionsIter.hasNext())
        {
            RDBMSStoreData schemaDataOption = (RDBMSStoreData)optionsIter.next();
            if (schemaDataOption.getName().length() > metadata_id_len)
            {
                metadata_id_len = schemaDataOption.getName().length();
            }
        }

        // Form the query to find which one of these classes has the instance with this id
        QueryStatement qs_base = null;
        optionsIter = schemaDataOptions.iterator();
        boolean first = true;
        while (optionsIter.hasNext())
        {
            RDBMSStoreData schemaDataOption = (RDBMSStoreData)optionsIter.next();
            ClassMetaData cmd = (ClassMetaData)schemaDataOption.getMetaData();

            QueryStatement qs = storeMgr.getDatabaseAdapter().newQueryStatement(schemaDataOption.getTable());
            String classname = StringUtils.leftAlignedPaddedString(schemaDataOption.getName(), metadata_id_len);
            JavaTypeMapping m = storeMgr.getDatastoreAdapter().getMapping(String.class);
            qs.selectScalarExpression(m.newLiteral(qs,classname).as("JPOXMETADATA"));

            // LEFT OUTER JOIN to all direct subclasses
            Iterator subclass_iter = storeMgr.getSubClassesForClass(schemaDataOption.getName(),false).iterator();
            int subclasses_seq_id = 0;
            while (subclass_iter.hasNext())
            {
                String subclass = (String)subclass_iter.next();
                DatastoreClass subclassTable = storeMgr.getDatastoreClass(subclass);

                // No need to LEFT OUTER JOIN for "no-table" and "superclass-table" cases
                // "no-table" objects dont exist on their own
                // "superclass-table" are excluded using the discriminator clause
                if (subclassTable != null && !subclassTable.getName().equals(schemaDataOption.getTable().getName()))
                {
                    TableIdentifier subclassTableIdentifier = new TableIdentifier(storeMgr.getDatabaseAdapter(),"SUBCLASS" + (subclasses_seq_id++));
                    QueryStatement st = new QueryStatement(subclassTable, subclassTableIdentifier);
                    TableExpression table_expr_sub = st.newTableExpression(storeMgr.getDatastoreClass(subclass), subclassTableIdentifier);
                    st.select(subclassTableIdentifier, storeMgr.getDatastoreClass(subclass).getIDMapping());
                    JavaTypeMapping subMapping = storeMgr.getDatastoreClass(subclass).getIDMapping();

                    ScalarExpression subExpr = subMapping.newScalarExpression(qs,table_expr_sub);
                    ScalarExpression schExpr = (((ClassTable)schemaDataOption.getTable()).getIDMapping()).newScalarExpression(qs,qs.getDefaultTableExpression());
                    qs.leftOuterJoin(subExpr,schExpr, table_expr_sub);
                    qs.andCondition(new NullLiteral(qs).eq(subExpr));
                }
            }

            // WHERE (object id) = ?
            JavaTypeMapping idMapping = ((ClassTable)schemaDataOption.getTable()).getIDMapping();

            // We have to create a StateManager here just to map fields from the AppId key object
            // to the table fields. Really the table should have some way of doing this. TODO : Refactor this
            Class pc_class = pm.getClassLoaderResolver().classForName(schemaDataOption.getName());
            StateManager sm = new StateManagerImpl(pm,pc_class,id);
            ScalarExpression fieldExpr = idMapping.newScalarExpression(qs, qs.getDefaultTableExpression());
            ScalarExpression fieldValue = idMapping.newLiteral(qs, sm.getObject());
            qs.andCondition(fieldExpr.eq(fieldValue), true);
 
            // Discriminator for this class
            JavaTypeMapping discrimMapping = schemaDataOption.getTable().getDiscriminatorMapping();
            DiscriminatorMetaData discrimMetaData = cmd.getInheritanceMetaData().getDiscriminatorMetaData();
            if (discrimMapping != null)
            {
                ScalarExpression discrimExpr = discrimMapping.newScalarExpression(qs, qs.getDefaultTableExpression());
                Object value = null;
                if (cmd.getDiscriminatorStrategy() == DiscriminatorStrategy.CLASS_NAME)
                {
                    value = schemaDataOption.getName();
                }
                else if (cmd.getDiscriminatorStrategy() == DiscriminatorStrategy.VALUE_MAP)
                {
                    value = discrimMetaData.getValue();
                }
                ScalarExpression discrimValue = discrimMapping.newLiteral(qs, value);
                qs.andCondition(discrimExpr.eq(discrimValue), true);
            }

            if (first)
            {
                qs_base = qs;
            }
            else
            {
                qs_base.union(qs);
            }
            first = false;
        }

        // Perform the query
        try
        {
            boolean isConnectionOpen = pm.isConnectionOpen();
            Connection conn = pm.getConnection(false,isConnectionOpen);
            try
            {
                Transaction tx = (Transaction) pm.currentTransaction();
                String statement = qs_base.toStatementText(tx.useUpdateLockOnFetch()).toString();
                PreparedStatement ps = qs_base.toStatementText(tx.useUpdateLockOnFetch()).prepareStatement(pm,conn);
                try
                {
                    if (JPOXLogger.RDBMS.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS.debug(statement);
                    }

                    long startTime = System.currentTimeMillis();
                    ResultSet rs = ps.executeQuery();
                    try
                    {
                        if (JPOXLogger.RDBMS.isDebugEnabled())
                        {
                            JPOXLogger.RDBMS.debug(LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                        }

                        if (rs != null)
                        {
                            while (rs.next())
                            {
                                String class_name=rs.getString(1);
                                optionsIter = schemaDataOptions.iterator();
                                while (optionsIter.hasNext())
                                {
                                    RDBMSStoreData schemaDataOption = (RDBMSStoreData)optionsIter.next();
                                    if (schemaDataOption.getName().equals(class_name.trim()))
                                    {
                                        data = schemaDataOption;
                                    }
                                }
                            }
                        }
                    }
                    finally
                    {
                        rs.close();
                    }
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                if (!isConnectionOpen)
                {
                    pm.releaseConnection(conn);
                }
            }
        }
        catch (SQLException sqe)
        {
            JPOXLogger.RDBMS.error(sqe);
            throw new JDODataStoreException(sqe.toString());
        }

        return data;
    }

    /**
     * Utility to find the ClassMetaData for the class that manages the table where the supplied
     * class stores its fields. This is for use where the current class uses an inheritance strategy
     * of "superclass-table" and we want to find the table where we should add our fields.
     * @param cmd The meta data for this class.
     * @return The metadata for the class that manages the table
     */
    public static ClassMetaData getClassMetaDataManagingTableForClass(ClassMetaData cmd)
    {
        // We manage our own table
        if (cmd.getInheritanceMetaData().getStrategyValue() == InheritanceStrategy.NEW_TABLE)
        {
            return cmd;
        }

        // Go to the superclass
        ClassMetaData superCmd = cmd.getSuperclassMetaData();
        if (superCmd != null)
        {
            if (superCmd.getInheritanceMetaData().getStrategyValue() == InheritanceStrategy.NEW_TABLE)
            {
                return superCmd;
            }
            else if (superCmd.getInheritanceMetaData().getStrategyValue() == InheritanceStrategy.NO_TABLE)
            {
                return null;
            }
            else
            {
                // Go to the superclass and try again
                return getClassMetaDataManagingTableForClass(superCmd);
            }
        }
        
        // No class found that manages our fields!
        return null;
    }
}