/**********************************************************************
Copyright (c) 2002 Kelly Grizzle (TJDO) and others.
All rights reserved.   This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
2002 Mike Martin (TJDO)
2003 Andy Jefferson - coding standards
    ...
**********************************************************************/
package org.jpox.store.rdbms;

import java.util.Arrays;

import org.jpox.store.rdbms.table.ClassTable;

/**
 * Representation of a request id.
 *
 * @version $Revision: 1.2 $
 **/
class RequestIdentifier
{
    private final ClassTable table;
    private final int[] fields;
    private final Type type;
    private final int hashCode;
    private final String className;


    public RequestIdentifier(ClassTable table, int[] fields, Type type, String className)
    {
        this.table = table;
        this.type = type;

        if (fields == null)
        {
            this.fields = null;
        }
        else
        {
            this.fields = new int[fields.length];
            System.arraycopy(fields, 0, this.fields, 0, fields.length);

            // The key uniqueness is dependent on fields being sorted
            Arrays.sort(this.fields);
        }
        this.className = className;
        
        /*
         * Since we are an immutable object, pre-compute the hash code for
         * improved performance in equals().
         */
        int h = table.hashCode() ^ type.hashCode() ^ className.hashCode();

        if (this.fields != null)
        {
            for (int i = 0; i < this.fields.length; ++i)
            {
                h ^= this.fields[i];
            }
        }

        hashCode = h;
    }

    /**
     * Accessor for the hashcode
     * @return The hashcode
     **/
    public int hashCode()
    {
        return hashCode;
    }

    /**
     * Equality operator
     * @param o Object to compare with
     * @return Whether the objects are equal
     **/
    public boolean equals(Object o)
    {
        if (o == this)
        {
            return true;
        }

        if (!(o instanceof RequestIdentifier))
        {
            return false;
        }

        RequestIdentifier ri = (RequestIdentifier)o;

        if (hashCode != ri.hashCode)
        {
            return false;
        }

        return table.equals(ri.table) &&
            type.equals(ri.type) &&
            Arrays.equals(fields, ri.fields) &&
            className.equals(ri.className);
    }

    /**
     * Inner class representing the Type of request
     **/
    public static class Type
    {
        private int typeId;

        public static final Type INSERT        = new Type(0);
        public static final Type FETCH         = new Type(1);
        public static final Type UPDATE        = new Type(2);
        public static final Type DELETE        = new Type(3);
        public static final Type LOCATE        = new Type(4);

        private static final String[] TYPE_NAMES =
        {
            "Insert",
            "Fetch",
            "Update",
            "Delete",
            "Locate"
        };

        private Type(int typeId)
        {
            this.typeId = typeId;
        }

        public String toString()
        {
            return TYPE_NAMES[typeId];
        }
    }
}
