/**********************************************************************
Copyright (c) 2003 Andy Jefferson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2004 Andy Jefferson - added type, version
2004 Andy Jefferson - changed to use StoreData. Added owner
    ...
**********************************************************************/
package org.jpox.store.rdbms;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.Collection;
import javax.jdo.JDODataStoreException;
import javax.jdo.JDOFatalInternalException;

import org.jpox.store.exceptions.DatastoreInitialisationException;
import org.jpox.store.exceptions.MissingTableException;
import org.jpox.store.rdbms.table.SchemaTable;
import org.jpox.store.AutoStartMechanism;
import org.jpox.store.StoreData;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;

/**
 * Implementation of an Auto-Start Mechanism for JPOX. This implementation
 * stores the classes supported in a table in the datastore. It is initialised
 * and read at startup of JPOX, and is continually updated during the lifetime
 * of the calling application.
 *
 * @version $Revision: 1.2 $ 
 */
public class SchemaAutoStarter implements AutoStartMechanism
{
    public static final String SCHEMA_TABLE="SchemaTable";
    
    private static final Localiser LOCALISER=Localiser.getInstance("org.jpox.store.rdbms.Localisation");

    protected String        mode=null;
    protected SchemaTable   schemaTable=null;
    protected RDBMSManager  storeMgr=null;
    protected Connection    conn;


    /**
     * Constructor.
     * @param store_mgr The RDBMSManager managing the store that we are
     * auto-starting.
     **/
    public SchemaAutoStarter(RDBMSManager store_mgr)
    {
        storeMgr = store_mgr;

        schemaTable = new SchemaTable(store_mgr);
        schemaTable.initialize();

        // We need to autocreate the table and validate it being correct.
        try
        {
            Connection conn=storeMgr.getConnection(Connection.TRANSACTION_NONE);
            try
            {
                // Create the table if it doesnt exist, and validate it
                schemaTable.exists(conn,true);
                if (SchemaTable.isDumpingDdl())
                {
                    // when DDL is only written to a file instead of being sent
                    // to DB, validating it
                    // will throw a MissingTableException
                    try
                    {
                        schemaTable.validate(conn, false);
                    }
                    catch (MissingTableException mte)
                    {
                        // if table had not existed, the DDL for creating it has
                        // been written above in
                        // schemaTable.checkExists(conn,true),
                        // any other validation exception will be handled below
                    }
                }
                else
                {
                    schemaTable.validate(conn, false);
                }
            }
            catch (Exception e)
            {
                // Validation error - table is different to what we expect
                JPOXLogger.RDBMS_SCHEMA.error(LOCALISER.msg("SchemaAutoStarter.ValidationError",storeMgr.getSchemaName()));
                if (JPOXLogger.RDBMS_SCHEMA.isDebugEnabled())
                {
                    // TODO Remove this printStackTrace
                    e.printStackTrace();
                    JPOXLogger.RDBMS_SCHEMA.debug(e);
                }

                try
                {
                    if (JPOXLogger.RDBMS_SCHEMA.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SCHEMA.debug(LOCALISER.msg("SchemaAutoStarter.DropAndCreate"));
                    }
                    
                    // Drop the table in case it had incorrect structure before
                    try
                    {
                        schemaTable.drop(conn);
                    }
                    catch (SQLException sqe)
                    {
                        // Do nothing. Maybe the table didn't exist in the first place
                    }

                    // Create the table
                    schemaTable.exists(conn, true);
                    schemaTable.validate(conn, false);
                }
                catch (Exception e2)
                {
                    JPOXLogger.RDBMS_SCHEMA.error(LOCALISER.msg("SchemaAutoStarter.ValidationError",storeMgr.getSchemaName()));
                    if (JPOXLogger.RDBMS_SCHEMA.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SCHEMA.debug(e);
                    }
                }
            }
            finally
            {
                storeMgr.closeConnection(conn);
            }
        }
        catch (SQLException sqe)
        {
            JPOXLogger.RDBMS_SCHEMA.error(LOCALISER.msg("RDBMSManager.DatabaseConnectionError"));
            if (JPOXLogger.RDBMS_SCHEMA.isDebugEnabled())
            {
                JPOXLogger.RDBMS_SCHEMA.debug(sqe);
            }
        }
    }

    /**
     * Accessor for the mode of operation.
     * @return Mode of operation
     **/
    public String getMode()
    {
        return mode;
    }

    /**
     * Mutator for the mode of operation.
     * @param mode Mode of operation
     **/
    public void setMode(String mode)
    {
        this.mode = mode;
    }

    /**
     * Accessor for the data for the classes supported.
     * @return Collection of classes supported.
     **/
    public Collection getAllClassData()
    throws DatastoreInitialisationException
    {
        try
        {
	        assertIsOpen();
	        Collection data=null;
	        try
	        {
	            data = schemaTable.getAllClasses(conn);
	        }
	        catch (SQLException sqe2)
	        {
	            JPOXLogger.RDBMS_SCHEMA.error(LOCALISER.msg("SchemaAutoStarter.ClassesReadError"));
	            //TODO in case of exception should we throw a Fatal Exception???
	            if (JPOXLogger.RDBMS_SCHEMA.isDebugEnabled())
	            {
	                JPOXLogger.RDBMS_SCHEMA.debug(sqe2);
	            }
	        }
	        return data;
        }
        catch( Exception e )
        {
            throw new DatastoreInitialisationException(LOCALISER.msg("SchemaAutoStarter.DatastoreInitialisationException",e));
        }
    }

    /**
     * Assert that the mechanism is open from writings
     */
    private void assertIsOpen()
    {
        if( conn == null )
        {
            throw new JDOFatalInternalException(LOCALISER.msg("SchemaAutoStarter.AssertIsOpen"));
        }
    }

    /**
     * Assert that the mechanism is closed from writings
     */
    private void assertIsClosed()
    {
        if( conn != null )
        {
            throw new JDOFatalInternalException(LOCALISER.msg("SchemaAutoStarter.AssertIsClosed"));
        }
    }
    
    /**
     * Starts a transaction for writting (add/delete) classes to the auto start
     * mechanism 
     */
    public void open()
    {
        assertIsClosed();
        try
        {
            conn = storeMgr.getConnection(Connection.TRANSACTION_NONE);
        }
        catch (SQLException sqe2)
        {
            JPOXLogger.RDBMS_SCHEMA.error(LOCALISER.msg("RDBMSManager.DatabaseConnectionError"));
            if (JPOXLogger.RDBMS_SCHEMA.isDebugEnabled())
            {
                JPOXLogger.RDBMS_SCHEMA.debug(sqe2);
            }
        }
    }
    
    /**
     * Closes a transaction for writting (add/delete) classes to the auto start
     * mechanism 
     */
    public void close()
    {
        assertIsOpen();
        try
        {
            storeMgr.closeConnection(conn);
            conn = null;
        }
        catch (SQLException sqe2)
        {
            JPOXLogger.RDBMS_SCHEMA.error(LOCALISER.msg("RDBMSManager.DatabaseConnectionError"));
            if (JPOXLogger.RDBMS_SCHEMA.isDebugEnabled())
            {
                JPOXLogger.RDBMS_SCHEMA.debug(sqe2);
            }
        }
    }    
    
    /**
     * Whether it's open for writting (add/delete) classes to the auto start
     * mechanism
     * @return whether this is open for writting 
     */
    public boolean isOpen()
    {
        return conn != null;
    }
    
    /**
     * Method to add a class to the supported list.
     * @param data Data for the class to add.
     **/
    public void addClass(StoreData data)
    {
        assertIsOpen();
        try
        {
            schemaTable.addClass(data,conn);
        }
        catch (SQLException sqe2)
        {
            JPOXLogger.RDBMS_SCHEMA.error(LOCALISER.msg("SchemaAutoStarter.AddClassError",data.getName(),sqe2));
            if (JPOXLogger.RDBMS_SCHEMA.isDebugEnabled())
            {
                JPOXLogger.RDBMS_SCHEMA.debug(sqe2);
            }
            throw new JDODataStoreException(LOCALISER.msg("SchemaAutoStarter.AddClassError",data.getName(),sqe2));
        }
    }

    /**
     * Method to drop support for a class.
     * @param class_name The class 
     **/
    public void deleteClass(String class_name)
    {
        assertIsOpen();
        try
        {
            schemaTable.deleteClass(class_name,conn);
        }
        catch (SQLException sqe2)
        {
            JPOXLogger.RDBMS_SCHEMA.error(LOCALISER.msg("SchemaAutoStarter.DeleteClassError",class_name,sqe2));
            if (JPOXLogger.RDBMS_SCHEMA.isDebugEnabled())
            {
                JPOXLogger.RDBMS_SCHEMA.debug(sqe2);
            }
        }
    }

    /**
     * Method to drop support for all current classes.
     **/
    public void deleteAllClasses()
    {
        assertIsOpen();
        try
        {
            schemaTable.deleteAllClasses(conn);
        }
        catch (SQLException sqe2)
        {
            JPOXLogger.RDBMS_SCHEMA.error(LOCALISER.msg("SchemaAutoStarter.DeleteClassesError",sqe2));
            if (JPOXLogger.RDBMS_SCHEMA.isDebugEnabled())
            {
                JPOXLogger.RDBMS_SCHEMA.debug(sqe2);
            }
        }
    }

    /**
     * Utility to output the storage description for this mechanism.
     * @return The storage description
     **/
    public String getStorageDescription()
    {
        return LOCALISER.msg("SchemaAutoStarter.StorageDescription");
    }
}