/**********************************************************************
Copyright (c) 2003 Andy Jefferson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
    ...
**********************************************************************/
package org.jpox.store.rdbms.adapter;

import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.SQLException;
import java.util.HashMap;
import java.util.Set;
import java.util.Iterator;

import java.lang.reflect.Constructor;
import java.lang.reflect.InvocationTargetException;

import javax.jdo.JDODataStoreException;

import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;

/**
 * Manager of Adapters for JPOX. Acts as a registry of adapters to RDBMS
 * that can be selected for use in JPOX. Uses a singleton pattern, and the
 * manager can be accessed using the getAdapterManager() method.
 * <P>
 * An adapter can be added using the addAdapter method. This tags a partial
 * product naming string against the adapter class, so that whenever an adapter
 * is needed for a particular product name, if the product name includes this
 * partial product name, it will be returned as providing support for that
 * product.
 * </P>
 * <P>
 * The getInstance method provides easy access to the adapter given a
 * connection to a datastore.
 * </P> 
 *
 * @version $Revision: 1.9 $ 
 */
public class AdapterManager
{
    private static final Localiser LOCALISER=Localiser.getInstance("org.jpox.store.rdbms.adapter.Localisation");

    /** The adapter classes, mapped by naming string. */
    HashMap adapters=new HashMap();

    /** Singleton instance */
    static AdapterManager adapterMgr=new AdapterManager();

    /** Cache of Adapters in use, keyed by their product naming string. */
    static HashMap currentAdapters=new HashMap();

    /**
     * Static initialisation block to load up support for inbuilt adapters.
     **/
    static
    {
        adapterMgr.addAdapter("derby",org.jpox.store.rdbms.adapter.CloudscapeAdapter.class);
        adapterMgr.addAdapter("Apache Derby",org.jpox.store.rdbms.adapter.CloudscapeAdapter.class);
        adapterMgr.addAdapter("cloudscape",org.jpox.store.rdbms.adapter.CloudscapeAdapter.class);
        adapterMgr.addAdapter("db2",org.jpox.store.rdbms.adapter.DB2Adapter.class);
        adapterMgr.addAdapter("firebird",org.jpox.store.rdbms.adapter.FirebirdAdapter.class);
        adapterMgr.addAdapter("interbase",org.jpox.store.rdbms.adapter.FirebirdAdapter.class);
        adapterMgr.addAdapter("microsoft",org.jpox.store.rdbms.adapter.MSSQLServerAdapter.class);
        adapterMgr.addAdapter("hsql",org.jpox.store.rdbms.adapter.HSQLAdapter.class);
        adapterMgr.addAdapter("mckoi",org.jpox.store.rdbms.adapter.McKoiAdapter.class);
        adapterMgr.addAdapter("mysql",org.jpox.store.rdbms.adapter.MySQLAdapter.class);
        adapterMgr.addAdapter("Adaptive Server Anywhere",org.jpox.store.rdbms.adapter.SybaseAdapter.class);
        adapterMgr.addAdapter("Adaptive Server Enterprise",org.jpox.store.rdbms.adapter.SybaseAdapter.class);
        adapterMgr.addAdapter("Adaptive Server IQ",org.jpox.store.rdbms.adapter.SybaseAdapter.class);
        adapterMgr.addAdapter("sybase",org.jpox.store.rdbms.adapter.SybaseAdapter.class);
        adapterMgr.addAdapter("oracle",org.jpox.store.rdbms.adapter.OracleAdapter.class);
        adapterMgr.addAdapter("pointbase",org.jpox.store.rdbms.adapter.PointbaseAdapter.class);
        adapterMgr.addAdapter("postgresql",org.jpox.store.rdbms.adapter.PostgreSQLAdapter.class);
        adapterMgr.addAdapter("sapdb",org.jpox.store.rdbms.adapter.SAPDBAdapter.class);
        adapterMgr.addAdapter("sap db",org.jpox.store.rdbms.adapter.SAPDBAdapter.class);
        adapterMgr.addAdapter("informix",org.jpox.store.rdbms.adapter.InformixAdapter.class);
    }

    /**
     * Accessor for the AdapterManager for JPOX (singleton)
     * @return The manager of type information
     */
    public static AdapterManager getAdapterManager()
    {
        return adapterMgr;
    }

    /**
     * Protected constructor to prevent outside instantiation
     */
    protected AdapterManager()
    {
    }

    /**
     * Method to add support for a RDBMS product via an adapter.
     * @param naming_string Naming string
     * @param adapter_class The adapter class to use
     */
    public void addAdapter(String naming_string,Class adapter_class)
    {
        if (naming_string == null || adapter_class == null)
        {
            return;
        }

        adapters.put(naming_string,adapter_class);
        if (JPOXLogger.RDBMS.isDebugEnabled())
        {
            JPOXLogger.RDBMS.debug(LOCALISER.msg("Adapter.AddSupport",naming_string,adapter_class.getName()));
        }
    }

    /**
     * Accessor for an adapter, given a Connection to the datastore.
     * @param conn The Connection
     * @return The database adapter for this connection.
     * @throws SQLException Thrown if a DB error occurs.
     **/
    public DatabaseAdapter getInstance(Connection conn)
    throws SQLException
    {
        DatabaseMetaData metadata = conn.getMetaData();
        String id = metadata.getDatabaseProductName() + ", " +
                    metadata.getDatabaseProductVersion() + ", " +
                    metadata.getDriverName() + ", " +
                    metadata.getDriverVersion();

        // Check existing adapters
        DatabaseAdapter adapter=(DatabaseAdapter)currentAdapters.get(id);
        if (adapter == null)
        {
            adapter = getDatabaseAdapterForProduct(metadata);
            if (adapter == null)
            {
                // Nothing suitable found so warn the user and continue with the generic adapter
                JPOXLogger.RDBMS.warn(LOCALISER.msg("Adapter.NoAdapterFound"));
                adapter = new DatabaseAdapter(metadata);
            }

            currentAdapters.put(id,adapter);

            if (JPOXLogger.RDBMS.isInfoEnabled())
            {
                JPOXLogger.RDBMS.info(LOCALISER.msg("Adapter.Initialised",adapter));
            }
        }

        return adapter;
    }

    /**
     * Accessor for the adapter for a specified product.
     * @param metadata Database MetaData for the RDBMS
     * @return Instance of the database adapter
     */
    protected DatabaseAdapter getDatabaseAdapterForProduct(DatabaseMetaData metadata)
    {
        if (metadata == null)
        {
            return null;
        }

        String product_name=null;
        try
        {
            product_name = metadata.getDatabaseProductName();
            if (product_name == null)
            {
                JPOXLogger.RDBMS.error(LOCALISER.msg("Adapter.JDBC.NoProductNameError"));
                return null;
            }
        }
        catch (SQLException sqe)
        {
            JPOXLogger.RDBMS.error(LOCALISER.msg("Adapter.JDBC.ProductNameAccessError",sqe));
            return null;
        }

        Set adapter_keys=adapters.keySet();
        if (adapter_keys != null)
        {
            Iterator keys_iter=adapter_keys.iterator();
            while (keys_iter.hasNext())
            {
                String key=(String)keys_iter.next();
                if (product_name.toLowerCase().indexOf(key.toLowerCase()) >= 0)
                {
                    Object obj=adapters.get(key);
                    Class  adapter_class=(Class)obj;

                    Object adapter_obj=null;
                    Object[] ctr_args=new Object[] {metadata};
                    Class[] ctr_args_classes=new Class[] 
                        {
                            DatabaseMetaData.class
                        };
                    try
                    {
                        Constructor ctr=adapter_class.getConstructor(ctr_args_classes);
                        try
                        {
                            adapter_obj = ctr.newInstance(ctr_args);
                        }
                        catch (InvocationTargetException ite)
                        {
                            if (ite.getTargetException()!= null && ite.getTargetException() instanceof JDODataStoreException)
                            {
                                throw (JDODataStoreException) ite.getTargetException();
                            }
                        }
                        catch (Exception e)
                        {
                            JPOXLogger.RDBMS.error(LOCALISER.msg("Adapter.InstantiationError",product_name,e));
                            return null;
                        }
                    }
                    catch (NoSuchMethodException nsme)
                    {
                        return null;
                    }

                    return (DatabaseAdapter)adapter_obj;
                }
            }
        }
        return null;
    }
}
