/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others. All rights reserved.
This program and the accompanying materials are made available under
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - coding standards
    ...
**********************************************************************/
package org.jpox.store.rdbms.adapter;

import java.math.BigInteger;
import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.util.ArrayList;

import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.expression.BooleanExpression;
import org.jpox.store.expression.NumericExpression;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.StringExpression;
import org.jpox.store.expression.TableExprAsJoins;
import org.jpox.store.expression.TableExprAsSubquery;
import org.jpox.store.expression.TableExpression;
import org.jpox.store.mapping.MappingManager;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.rdbms.mapping.CloudscapeRDBMSMappingManager;
import org.jpox.store.rdbms.table.TableImpl;
import org.jpox.store.rdbms.typeinfo.CloudscapeTypeInfo;
import org.jpox.store.rdbms.typeinfo.TypeInfo;

/**
 * Provides methods for adapting SQL language elements to the Cloudscape/Derby
 * database.
 *
 * @see DatabaseAdapter
 * @version $Revision: 1.8 $ 
 */
class CloudscapeAdapter extends DatabaseAdapter
{
    /**
     * Cloudscape 10.0 beta reserved words, includes SQL92 reserved words
     */
    private static final String CLOUDSCAPE_RESERVED_WORDS =
        "ADD,ALL,ALLOCATE,ALTER,AND,ANY,ARE,AS," +
        "ASC,ASSERTION,AT,AUTHORIZATION,AVG,BEGIN,BETWEEN,BIT," +
        "BIT_LENGTH,BOOLEAN,BOTH,BY,CALL,CASCADE,CASCADED,CASE," +
        "CAST,CHAR,CHARACTER,CHARACTER_LENGTH,CHAR_LENGTH,CHECK,CLOSE,COLLATE," +
        "COLLATION,COLUMN,COMMIT,CONNECT,CONNECTION,CONSTRAINT,CONSTRAINTS,CONTINUE," +
        "CONVERT,CORRESPONDING,COUNT,CREATE,CROSS,CURRENT,CURRENT_DATE,CURRENT_TIME," +
        "CURRENT_TIMESTAMP,CURRENT_USER,CURSOR,DEALLOCATE,DEC,DECIMAL,DECLARE,DEFERRABLE," +
        "DEFERRED,DELETE,DESC,DESCRIBE,DIAGNOSTICS,DISCONNECT,DISTINCT,DOUBLE," +
        "DROP,ELSE,END,ENDEXEC,ESCAPE,EXCEPT,EXCEPTION,EXEC," +
        "EXECUTE,EXISTS,EXPLAIN,EXTERNAL,EXTRACT,FALSE,FETCH,FIRST," +
        "FLOAT,FOR,FOREIGN,FOUND,FROM,FULL,FUNCTION,GET," +
        "GET_CURRENT_CONNECTION,GLOBAL,GO,GOTO,GRANT,GROUP,HAVING,HOUR," +
        "IDENTITY,IMMEDIATE,IN,INDICATOR,INITIALLY,INNER,INOUT,INPUT," +
        "INSENSITIVE,INSERT,INT,INTEGER,INTERSECT,INTO,IS,ISOLATION," +
        "JOIN,KEY,LAST,LEADING,LEFT,LIKE,LOCAL,LONGINT," +
        "LOWER,LTRIM,MATCH,MAX,MIN,MINUTE,NATIONAL,NATURAL," +
        "NCHAR,NVARCHAR,NEXT,NO,NOT,NULL,NULLIF,NUMERIC," +
        "OCTET_LENGTH,OF,ON,ONLY,OPEN,OPTION,OR,ORDER," +
        "OUT,OUTER,OUTPUT,OVERLAPS,PAD,PARTIAL,PREPARE,PRESERVE," +
        "PRIMARY,PRIOR,PRIVILEGES,PROCEDURE,PUBLIC,READ,REAL,REFERENCES," +
        "RELATIVE,RESTRICT,REVOKE,RIGHT,ROLLBACK,ROWS,RTRIM,RUNTIMESTATISTICS," +
        "SCHEMA,SCROLL,SECOND,SELECT,SESSION_USER,SET,SMALLINT,SOME," +
        "SPACE,SQL,SQLCODE,SQLERROR,SQLSTATE,SUBSTR,SUBSTRING,SUM," +
        "SYSTEM_USER,TABLE,TEMPORARY,TIMEZONE_HOUR,TIMEZONE_MINUTE,TINYINT,TO,TRAILING," +
        "TRANSACTION,TRANSLATE,TRANSLATION,TRIM,TRUE,UNION,UNIQUE,UNKNOWN," +
        "UPDATE,UPPER,USER,USING,VALUES,VARCHAR,VARYING,VIEW," +
        "WHENEVER,WHERE,WITH,WORK,WRITE,YEAR";
    
    /**
     * Constructs an Cloudscape adapter based on the given JDBC metadata.
     *
     * @param   metadata    the database metadata.
     */
    public CloudscapeAdapter(DatabaseMetaData metadata)
    {
        super(metadata);
        
        reservedKeywords.addAll(parseKeywordList(CLOUDSCAPE_RESERVED_WORDS));
    }

    public String getVendorID()
    {
        return "cloudscape";
    }

    /**
     * Method to add a new type.
     * @param rs ResultSet containing the type info.
     */
    public TypeInfo newTypeInfo(ResultSet rs)
    {
        TypeInfo ti = new CloudscapeTypeInfo(rs);
        return ti;
    }

    /**
     * Accessor for whether Cloudscape supports deferred constraints.
     * @return Whether it supports deferred constraints
     */
    public boolean supportsDeferredConstraints()
    {
        return false;
    }

    public TableExpression newTableExpression(QueryStatement qs, DatastoreContainerObject table, DatastoreIdentifier rangeVar)
    {
        /*
         * Syntactically Cloudscape can also handle TableExprAsJoins,
         * but when I tried that style with Cloudscape 4.0 certain queries
         * (like those involving EXISTS) failed with a message stating that
         * no execution plan could be found, or some such.
         */
        if (datastoreMajorVersion >= 10)
        {    
            return new TableExprAsJoins(qs, table, rangeVar);
        }
        else
        {
            return new TableExprAsSubquery(qs, table, rangeVar);
        }
    }
    
    public String getDropTableStatement(TableImpl table)
    {
        return "DROP TABLE " + table.toString();
    }
    
    /**
     * Whether the database support NULLs in the column options for table creation.
     * By default all columns are nullable in DB2, and the keyword NULL in the column options are not supported
     * @return whether the database support NULLs in the column options for table creation.
     **/
    public boolean supportsNullsKeywordInColumnOptions()
    {
        if (datastoreMajorVersion >= 10)
        {    
            return false;
        }
        return true;
    }    
    
    /**
     * Accessor for the maximum table name length permitted on this
     * datastore.
     * @return Max table name length
     **/
    public int getMaxTableNameLength()
    {
        if (datastoreMajorVersion >= 10)
        {    
            return 30;
        }        
        return maxTableNameLength;
    }
    
    /**
     * Accessor for the maximum constraint name length permitted on this
     * datastore.
     * @return Max constraint name length
     **/
    public int getMaxConstraintNameLength()
    {
        if (datastoreMajorVersion >= 10)
        {    
            return 18;
        }        
        return maxConstraintNameLength;
    }    
    
    /**
     * Accessor for the maximum index name length permitted on this datastore.
     * @return Max index name length
     **/
    public int getMaxIndexNameLength()
    {
        if (datastoreMajorVersion >= 10)
        {    
            return 18;
        }         
        return maxIndexNameLength;
    }

    /**
     * Accessor for the maximum column name length permitted on this datastore.
     * @return Max column name length
     **/
    public int getMaxColumnNameLength()
    {
        if (datastoreMajorVersion >= 10)
        {    
            return 30;
        }         
        return maxColumnNameLength;
    }    
    
    public NumericExpression lengthMethod(StringExpression str)
    {
        ArrayList args = new ArrayList();
        args.add(str);

        return new NumericExpression("LENGTH", args);
    }
    
    public StringExpression substringMethod(StringExpression str,
            NumericExpression begin)
    {
		ArrayList args = new ArrayList();
		args.add(str);
		args.add(begin.add(getMapping(BigInteger.class).newLiteral(str.getQueryStatement(), BigInteger.ONE)));
		//Cloudscape 10.0
		//SUBSTR( string, start )
		return new StringExpression("SUBSTR", args);
	}
	
	public StringExpression substringMethod(StringExpression str,
	               NumericExpression begin,
	               NumericExpression end)
	{
		ArrayList args = new ArrayList();
		args.add(str);
		args.add(begin.add(getMapping(BigInteger.class).newLiteral(str.getQueryStatement(), BigInteger.ONE)));
		args.add(end.sub(begin));
		//Cloudscape 10.0
		//SUBSTR( string, start, length )
		return new StringExpression("SUBSTR", args);
	} 
	
    /**
     * Returns the appropriate SQL expression for the JDOQL
     * String.trim(str) method.
     * It should return something like:
     * <p>
     * <pre>
     * TRIM(str)
     * </pre>
     *
     * @param str The first argument to the trim() method.
     * @return  The text of the SQL expression.
     */
    public StringExpression trimMethod(StringExpression str)
    {
        ArrayList args = new ArrayList();
        args.add(str);

        return new StringExpression("RTRIM", args);
    }
    
    /**
     * Method to handle the starts with operation.
     * @param str The string to compare against 
     * @return The expression.
     **/
    public BooleanExpression startsWithMethod(ScalarExpression source, ScalarExpression str)
    {
        
        ScalarExpression integerLiteral = getMapping(BigInteger.class).newLiteral(source.getQueryStatement(), BigInteger.ONE);
        ArrayList args = new ArrayList();
        args.add(str);
        args.add(source);
		//Cloudscape 10.0
        //LOCATE( stringSearched, SearchString, [StartPosition] )
        return new BooleanExpression(new StringExpression("LOCATE", args),ScalarExpression.OP_EQ,integerLiteral);
    }
    
	/**
	 * Accessor for the auto-increment sql statement for this datastore.
	 * @return The statement for getting the latest auto-increment key
	 **/
	public String getAutoIncrementStmt()
	{
		return "VALUES IDENTITY_VAL_LOCAL()";
	}

	/**
	 * Accessor for the auto-increment keyword for generating DDLs (CREATE TABLEs...).
	 * @return The keyword for a column using auto-increment
	 **/
	public String getAutoIncrementKeyword()
	{
		return "generated always as identity (start with 1)";
	}
	
	/**
	 * Whether we support auto-increment in keys.
	 * @return whether we support auto-increment in keys.
	 **/
	public boolean supportsAutoIncrementKeys()
	{
		return true;
	}
	
    /**
     * Whether we support DEFAULT tag together with NOT NULL in CREATE TABLE statements
     * 
     * <pre>
     * CREATE TABLE X ( MEMORY_SIZE BIGINT DEFAULT 0 NOT NULL )
     * </pre>
     * 
     * Some databases only support <i>DEFAULT {ConstantExpression | NULL}</i>
     * 
     * @return whether we support the DEFAULT tag together with NOT NULL in CREATE TABLE
     **/
    public boolean supportsDefaultKeywordWithNotNullInColumnOptions()
    {
        return false;
    }    
	
    /**
     * Whether this datastore supports SELECT ... FOR UPDATE.
     * @return whether we support SELECT ... FOR UPDATE.
     **/
    public boolean supportsLockWithSelectForUpdate()
    {
        return true;
    }
    
    /**
     * Acessor for the MappingManager
     * @return the MappingManager
     */
    public MappingManager getMappingManager()
    {
        if (mappingManager == null)
        {
            mappingManager = new CloudscapeRDBMSMappingManager();
        }
        return mappingManager;
    }
}