/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others. All rights reserved.
This program and the accompanying materials are made available under
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - coding standards
    ...
**********************************************************************/
package org.jpox.store.rdbms.adapter;

import java.math.BigInteger;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;

import javax.jdo.JDOFatalDataStoreException;

import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.expression.BooleanExpression;
import org.jpox.store.expression.ConvertExpression;
import org.jpox.store.expression.NumericExpression;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.StringExpression;
import org.jpox.store.expression.StringLiteral;
import org.jpox.store.expression.TableExprAsJoins;
import org.jpox.store.expression.TableExprAsSubquery;
import org.jpox.store.expression.TableExpression;
import org.jpox.store.mapping.MappingManager;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.rdbms.columninfo.ColumnInfo;
import org.jpox.store.rdbms.columninfo.DB2ColumnInfo;
import org.jpox.store.rdbms.mapping.DB2RDBMSMappingManager;
import org.jpox.store.rdbms.table.TableImpl;
import org.jpox.store.rdbms.typeinfo.DB2TypeInfo;
import org.jpox.store.rdbms.typeinfo.TypeInfo;

/**
 * Provides methods for adapting SQL language elements to the DB2 database.
 *
 * @see DatabaseAdapter
 * @version $Revision: 1.8 $ 
 */
class DB2Adapter extends DatabaseAdapter
{
    /**
     * A string containing the list of DB2 keywords 
     * This list is normally obtained dynamically from the driver using
     * DatabaseMetaData.getSQLKeywords()
     * 
     * Based on database DB2 version 7
     */
    public static final String DB2_RESERVED_WORDS =
        "ACCESS,ALIAS,ALLOW,ASUTIME,AUDIT,AUX,AUXILIARY,BUFFERPOOL," +
        "CAPTURE,CCSID,CLUSTER,COLLECTION,COLLID,COMMENT,CONCAT," +
        "CONTAINS,COUNT_BIG,CURRENT_LC_PATH,CURRENT_SERVER," +
        "CURRENT_TIMEZONE,DATABASE,DAYS,DB2GENERAL,DB2SQL,DBA," +
        "DBINFO,DBSPACE,DISALLOW,DSSIZE,EDITPROC,ERASE,EXCLUSIVE," +
        "EXPLAIN,FENCED,FIELDPROC,FILE,FINAL,GENERATED,GRAPHIC,HOURS," +
        "IDENTIFIED,INDEX,INTEGRITY,ISOBID,JAVA,LABEL,LC_CTYPE,LINKTYPE," +
        "LOCALE,LOCATORS,LOCK,LOCKSIZE,LONG,MICROSECOND,MICROSECONDS," +
        "MINUTES,MODE,MONTHS,NAME,NAMED,NHEADER,NODENAME,NODENUMBER," +
        "NULLS,NUMPARTS,OBID,OPTIMIZATION,OPTIMIZE,PACKAGE,PAGE," +
        "PAGES,PART,PCTFREE,PCTINDEX,PIECESIZE,PLAN,PRIQTY,PRIVATE," +
        "PROGRAM,PSID,QYERYNO,RECOVERY,RENAME,RESET,RESOURCE,RRN,RUN," +
        "SCHEDULE,SCRATCHPAD,SECONDS,SECQTY,SECURITY,SHARE,SIMPLE," +
        "SOURCE,STANDARD,STATISTICS,STAY,STOGROUP,STORES,STORPOOL," +
        "STYLE,SUBPAGES,SYNONYM,TABLESPACE,TYPE,VALIDPROC,VARIABLE," +
        "VARIANT,VCAT,VOLUMES,WLM,YEARS";
    
    /**
     * Constructs a DB2 adapter based on the given JDBC metadata.
     *
     * @param   metadata    the database metadata.
     */
    public DB2Adapter(DatabaseMetaData metadata)
    {
        super(metadata);

        reservedKeywords.addAll(parseKeywordList(DB2_RESERVED_WORDS));
        
        maxConstraintNameLength = 18;
        maxIndexNameLength = 18;
    }

    public String getVendorID()
    {
        return "db2";
    }
    
    public String getSchemaName(Connection conn) throws SQLException
    {
        Statement stmt = conn.createStatement();
        
        try
        {
            String stmtText = "VALUES (CURRENT SCHEMA)";
            ResultSet rs = stmt.executeQuery(stmtText);

            try
            {
                if (!rs.next())
                {
                    throw new JDOFatalDataStoreException("No result returned from " + stmtText);
                }

                return rs.getString(1).trim();
            }
            finally
            {
                rs.close();
            }
        }
        finally
        {
            stmt.close();
        }
    }

    /**
     * Method to close a Connection to the datastore.
     * @param conn The connection
     * @throws SQLException Thrown if error occurs on the close.
     **/
    public void closeConnection(Connection conn)
    throws SQLException
    {
        try
        {
            conn.rollback();
        }
        catch( Exception ex)
        {
            //nothing to do here
        }
        finally
        {
            conn.close();
        }
    }

    /**
     * Whether this datastore supports locking using SELECT ... FOR UPDATE.
     * @return whether we support locking using SELECT ... FOR UPDATE.
     **/
    public boolean supportsLockWithSelectForUpdate()
    {
        return true;
    }

    public boolean supportsBooleanComparison()
    {
        return false;
    }

    public boolean supportsDeferredConstraints()
    {
        return false;
    }

    public boolean supportsNullsInCandidateKeys()
    {
        return false;
    }

    public TypeInfo newTypeInfo(ResultSet rs)
    {
        TypeInfo ti = new DB2TypeInfo(rs);

        return ti;
    }
    
    public ColumnInfo newColumnInfo(ResultSet rs)
    {
        return new DB2ColumnInfo(rs);
    }

    public TableExpression newTableExpression(QueryStatement qs, DatastoreContainerObject table, DatastoreIdentifier rangeVar)
    {
        // there is no real reason, why only on or after version 8, but the version I'm testing against
        if (datastoreMajorVersion >= 8)
        {    
            return new TableExprAsJoins(qs, table, rangeVar);
        }
        else
        {
            return new TableExprAsSubquery(qs, table, rangeVar);
        }
    }

    public int getUnlimitedLengthPrecisionValue(TypeInfo typeInfo)
    {
        if (typeInfo.dataType == java.sql.Types.BLOB || typeInfo.dataType == java.sql.Types.CLOB)
        {
            return 1 << 30;
        }
        else
        {
            return super.getUnlimitedLengthPrecisionValue(typeInfo);
        }
    }

    public String getDropTableStatement(TableImpl table)
    {
        return "DROP TABLE " + table.toString();
    }

    public NumericExpression lengthMethod(StringExpression str)
    {
        ArrayList args = new ArrayList();
        args.add(str);

        return new NumericExpression("LENGTH", args);
    }

    public StringExpression substringMethod(StringExpression str,
                                            NumericExpression begin)
    {
        ArrayList args = new ArrayList();
        if( str instanceof StringLiteral )
        {
            args.add(new ConvertExpression(str.getQueryStatement(),str,"VARCHAR(32672)"));
        }
        else
        {
            args.add(str);
        }
        args.add(begin.add(getMapping(BigInteger.class).newLiteral(str.getQueryStatement(), BigInteger.ONE)));
		//DB2 8.0 manual
        //SUBSTR( string, start )
        return new StringExpression("SUBSTR", args);
    }

    public StringExpression substringMethod(StringExpression str,
                                               NumericExpression begin,
                                               NumericExpression end)
    {
        ArrayList args = new ArrayList();
        if( str instanceof StringLiteral )
        {
            args.add(new ConvertExpression(str.getQueryStatement(),str,"VARCHAR(32672)"));
        }
        else
        {
            args.add(str);
        }
        args.add(begin.add(getMapping(BigInteger.class).newLiteral(str.getQueryStatement(), BigInteger.ONE)));
        args.add(end.sub(begin));
		//DB2 8.0
        //SUBSTR( string, start, length )
        return new StringExpression("SUBSTR", args);
    }
    
    /**
     * Returns the appropriate SQL expression for the JDOQL
     * String.trim(str) method.
     * It should return something like:
     * <p>
     * <pre>
     * TRIM(str)
     * </pre>
     *
     * @param str The first argument to the trim() method.
     * @return  The text of the SQL expression.
     */
    public StringExpression trimMethod(StringExpression str)
    {
        ArrayList args = new ArrayList();
        args.add(str);

        return new StringExpression("RTRIM", args);
    }
    
    /**
     * Whether the database support NULLs in the column options for table creation.
     * By default all columns are nullable in DB2, and the keyword NULL in the column options are not supported
     * @return whether the database support NULLs in the column options for table creation.
     **/
    public boolean supportsNullsKeywordInColumnOptions()
    {
        return false;
    }
    
	/**
	 * Accessor for the auto-increment sql statement for this datastore.
	 * @return The statement for getting the latest auto-increment key
	 **/
	public String getAutoIncrementStmt()
	{
		return "VALUES IDENTITY_VAL_LOCAL()";
	}

	/**
	 * Accessor for the auto-increment keyword for generating DDLs (CREATE TABLEs...).
	 * @return The keyword for a column using auto-increment
	 **/
	public String getAutoIncrementKeyword()
	{
		return "generated always as identity (start with 1)";
	}
	
	/**
	 * Whether we support auto-increment in keys.
	 * @return whether we support auto-increment in keys.
	 **/
	public boolean supportsAutoIncrementKeys()
	{
		return true;
	}    

    /**
     * Continuation string to use where the SQL statement goes over more than 1
     * line. DB2 doesn't convert newlines into continuation characters and so
     * we just provide a space so that it accepts the statement.
     * @return Continuation string.
     **/
    public String getContinuationString()
    {
        return " ";
    }
    
    // ---------------------------- Sequence Support ---------------------------
    
    /**
     * Whether we support sequences.
     * @return whether we support sequences.
     **/
    public boolean supportsSequences()
    {
        return true;
    }
    
    /**
     * Accessor for the sequence statement to create the sequence.
     * @param sequence_name Name of the sequence 
     * @param min Minimum value for the sequence
     * @param max Maximum value for the sequence
     * @param start Start value for the sequence
     * @param increment Increment value for the sequence
     * @param cache_size Cache size for the sequence
     * @return The statement for getting the next id from the sequence
     **/
    public String getSequenceCreateStmt(String sequence_name,
            String min,String max,
            String start,String increment,
            String cache_size)
    {
        if (sequence_name == null)
        {
            throw new UnsupportedOperationException("Adapter.SequenceNameNullNotSupported");
        }
        
        StringBuffer stmt = new StringBuffer("CREATE SEQUENCE ");
        stmt.append(sequence_name);
        stmt.append(" AS INTEGER ");

        if (start != null && start.length() > 0)
        {
            stmt.append(" START WITH " + start);
        }
        if (increment != null && increment.length() > 0)
        {
            stmt.append(" INCREMENT BY " + increment);
        }
        if (min != null && min.length() > 0)
        {
            stmt.append(" MINVALUE " + min);
        }
        if (max != null && max.length() > 0)
        {
            stmt.append(" MAXVALUE " + max);
        }        
        if (cache_size != null && cache_size.length() > 0)
        {
            stmt.append(" CACHE " + cache_size);
        }
        else
        {
            stmt.append(" NOCACHE");
        }
        
        return stmt.toString();
    }        
    
    /**
     * Accessor for the statement for getting the next id from the sequence
     * for this datastore.
     * @param sequence_name Name of the sequence 
     * @return The statement for getting the next id for the sequence
     **/
    public String getSequenceNextStmt(String sequence_name)
    {
        if (sequence_name == null)
        {
            throw new UnsupportedOperationException("Adapter.SequenceNameNullNotSupported");
        }
        StringBuffer stmt=new StringBuffer("VALUES NEXTVAL FOR ");
        stmt.append(sequence_name);

        return stmt.toString();
    }

    /**
     * Method to handle the starts with operation.
     * @param str The string to compare against 
     * @return The expression.
     **/
    public BooleanExpression startsWithMethod(ScalarExpression source, ScalarExpression str)
    {
        
        ScalarExpression integerLiteral = getMapping(BigInteger.class).newLiteral(source.getQueryStatement(), BigInteger.ONE);
        ArrayList args = new ArrayList();
        if( str instanceof StringLiteral )
        {
            args.add(new ConvertExpression(str.getQueryStatement(),str,"VARCHAR(4000)"));
        }
        else
        {
            args.add(str);
        }
        if( source instanceof StringLiteral )
        {
            args.add(new ConvertExpression(source.getQueryStatement(),source,"VARCHAR(4000)"));
        }
        else
        {
            args.add(source);
        }
		//DB2 8 manual
        //a VARCHAR the maximum length is 4 000 bytes 
        //LOCATE( CAST(stringSearched AS VARCHAR(4000)), SearchString, [StartPosition] )
        return new BooleanExpression(new StringExpression("LOCATE", args),ScalarExpression.OP_EQ,integerLiteral);
    }	
    
    /**
     * Acessor for the MappingManager
     * @return the MappingManager
     */
    public MappingManager getMappingManager()
    {
        if (mappingManager == null)
        {
            mappingManager = new DB2RDBMSMappingManager();
        }
        return mappingManager;
    }    
}
