/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others. All rights reserved.
This program and the accompanying materials are made available under
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2002 Kelly Grizzle (TJDO)
2002 Christopher Walk (TJDO)
2003 Andy Jefferson - commented and javadocs
2003 Andy Jefferson - added localiser
2003 Andy Jefferson - added sequence methods
2004 Erik Bengtson - added auto increment
2004 Erik Bengtson - added query operators, sql expressions
2004 Andy Jefferson - removed IndexMapping/OptimisticMapping 
    ...
**********************************************************************/
package org.jpox.store.rdbms.adapter;

import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;
import java.util.StringTokenizer;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOUserException;
import javax.sql.DataSource;

import org.jpox.store.DatastoreAdapter;
import org.jpox.store.DatastoreClass;
import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.StoreManager;
import org.jpox.store.exceptions.NoTableManagedException;
import org.jpox.store.exceptions.UnsupportedDataTypeException;
import org.jpox.store.expression.BooleanExpression;
import org.jpox.store.expression.ConcatOperatorExpression;
import org.jpox.store.expression.ConvertExpression;
import org.jpox.store.expression.NumericExpression;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.StringExpression;
import org.jpox.store.expression.StringLiteral;
import org.jpox.store.expression.SubstringExpression;
import org.jpox.store.expression.TableExprAsSubjoins;
import org.jpox.store.expression.TableExpression;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.MappingFactory;
import org.jpox.store.mapping.MappingManager;
import org.jpox.store.mapping.OIDMapping;
import org.jpox.store.mapping.SimpleDatastoreRepresentation;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.columninfo.ColumnInfo;
import org.jpox.store.rdbms.foreignkey.ExportedKeyInfo;
import org.jpox.store.rdbms.foreignkey.ForeignKeyInfo;
import org.jpox.store.rdbms.key.CandidateKey;
import org.jpox.store.rdbms.key.ForeignKey;
import org.jpox.store.rdbms.key.Index;
import org.jpox.store.rdbms.key.PrimaryKey;
import org.jpox.store.rdbms.mapping.RDBMSMappingManager;
import org.jpox.store.rdbms.sqlidentifier.SQLIdentifier;
import org.jpox.store.rdbms.table.Table;
import org.jpox.store.rdbms.table.TableImpl;
import org.jpox.store.rdbms.table.ViewImpl;
import org.jpox.store.rdbms.typeinfo.TypeInfo;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;
import org.jpox.util.StringUtils;

/**
 * Provides methods for adapting SQL language elements to a specific vendor's
 * database.  A database adapter is primarily used to map generic JDBC data
 * types and SQL identifiers to specific types/identifiers suitable for the
 * database in use.
 *
 * <p>Each database adapter corresponds to a particular combination of database,
 * database version, driver, and driver version, as provided by the driver's
 * own metadata.  Database adapters cannot be constructed directly, but must be
 * obtained using the {@link AdapterManager#getInstance} method.</p>
 *
 * @see AdapterManager
 * @see java.sql.DatabaseMetaData
 * @version $Revision: 1.23 $ 
 */
public class DatabaseAdapter extends DatastoreAdapter
{
    protected static final Localiser LOCALISER=Localiser.getInstance("org.jpox.store.rdbms.adapter.Localisation");

    /** the JDBC driver name **/
    protected String driverName;
    
    /** the JDBC driver version **/
    protected String driverVersion;
    
    /**
     * The major version number of the underlying driver.
     */
    protected int driverMajorVersion;

    /**
     * The minor version number of the underlying driver.
     */
    protected int driverMinorVersion;
    
    /** The maximum length to be used for a table name. */
    protected int maxTableNameLength;

    /** The maximum length to be used for a table constraint name. */
    protected int maxConstraintNameLength;

    /** The maximum length to be used for an index name. */
    protected int maxIndexNameLength;

    /** The maximum length to be used for a column name. */
    protected int maxColumnNameLength;

    /** <tt>true</tt> if the database supports catalogs in table definition */
    protected boolean supportsCatalogsInTableDefinitions;

    /** <tt>true</tt> if the database supports schemas in table definition */
    protected boolean supportsSchemasInTableDefinitions;

    /** The String used to separate catalog and table name. */
    protected String catalogSeparator;

    /** JDBC types, keyed by the type number. Allowing for overriding by RDBMS. */
    protected final HashMap typesByTypeNumber = new HashMap();

    /**
     * Constructs a database adapter based on the given JDBC metadata.
     * @param metadata the database metadata.
     */
    protected DatabaseAdapter(DatabaseMetaData metadata)
    {
        reservedKeywords.addAll(parseKeywordList(SQLConstants.SQL92_RESERVED_WORDS));
        reservedKeywords.addAll(parseKeywordList(SQLConstants.SQL99_RESERVED_WORDS));
        reservedKeywords.addAll(parseKeywordList(SQLConstants.SQL2003_RESERVED_WORDS));
        reservedKeywords.addAll(parseKeywordList(SQLConstants.NONRESERVED_WORDS));

        try
        {
            reservedKeywords.addAll(parseKeywordList(metadata.getSQLKeywords()));

            driverMinorVersion      = metadata.getDriverMinorVersion();
            driverMajorVersion      = metadata.getDriverMajorVersion();
            driverName 				= metadata.getDriverName();
            driverVersion			= metadata.getDriverVersion();
            datastoreProductName    = metadata.getDatabaseProductName();
            datastoreProductVersion = metadata.getDatabaseProductVersion();

            // Try to convert the "version" string into a W.X.Y.Z version string
            StringBuffer strippedProductVersion=new StringBuffer();
            char previousChar = ' ';
            for (int i=0; i<datastoreProductVersion.length(); ++i)
            {
                char c = datastoreProductVersion.charAt(i);
                if (Character.isDigit(c) || c == '.')
                {
                    // Only update the stripped version when we have "X."
                    if (previousChar != ' ')
                    {
                        if (strippedProductVersion.length() == 0)
                        {
                            strippedProductVersion.append(previousChar);
                        }
                        strippedProductVersion.append(c);
                    }
                    previousChar = c;
                }
                else
                {
                    previousChar = ' ';
                }
            }

            try
            {
                Class mdc = metadata.getClass();

                datastoreMajorVersion = ((Integer)mdc.getMethod("getDatabaseMajorVersion", null).invoke(metadata, null)).intValue();
                datastoreMinorVersion = ((Integer)mdc.getMethod("getDatabaseMinorVersion", null).invoke(metadata, null)).intValue();

                // Retrieve the Revision version if it is accessible
                StringTokenizer parts = new StringTokenizer(strippedProductVersion.toString(), ".");
                if (parts.hasMoreTokens())
                {
                    // Major Version - ignore it
                    parts.nextToken();
                }
                if (parts.hasMoreTokens())
                {
                    // Minor Version - ignore it
                    parts.nextToken();
                }
                if (parts.hasMoreTokens())
                {
                    // Revision Version
                    try
                    {
                        datastoreRevisionVersion = Integer.parseInt(parts.nextToken());
                    }
                    catch (Exception e)
                    {
                        datastoreRevisionVersion = -1; //unknown
                    }
                }
            }
            catch (Throwable t)
            {
                /*
                 * The driver doesn't support JDBC 3.  Try to parse major and
                 * minor version numbers out of the product version string.
                 * We do this by stripping out everything but digits and periods
                 * and hoping we get something that looks like <major>.<minor>.
                 */
                StringTokenizer parts = new StringTokenizer(strippedProductVersion.toString(), ".");
                if (parts.hasMoreTokens())
                {
                    try
                    {
                        datastoreMajorVersion = Integer.parseInt(parts.nextToken());
                    }
                    catch (Exception e)
                    {
                        datastoreMajorVersion = -1; //unknown
                    }
                }
                if (parts.hasMoreTokens())
                {
                    try
                    {
                        datastoreMinorVersion = Integer.parseInt(parts.nextToken());
                    }
                    catch (Exception e)
                    {
                        datastoreMajorVersion = -1; //unknown
                    }
                }
                if (parts.hasMoreTokens())
                {
                    try
                    {
                        datastoreRevisionVersion = Integer.parseInt(parts.nextToken());
                    }
                    catch (Exception e)
                    {
                        datastoreRevisionVersion = -1; //unknown
                    }
                }
            }

            // Extract attributes of the Database adapter
            maxTableNameLength      = metadata.getMaxTableNameLength();
            maxConstraintNameLength = metadata.getMaxTableNameLength();
            maxIndexNameLength      = metadata.getMaxTableNameLength();
            maxColumnNameLength     = metadata.getMaxColumnNameLength();
            supportsCatalogsInTableDefinitions = metadata.supportsCatalogsInTableDefinitions();
            supportsSchemasInTableDefinitions = metadata.supportsSchemasInTableDefinitions();
            storesLowerCaseIdentifiers = metadata.storesLowerCaseIdentifiers();
            storesMixedCaseIdentifiers = metadata.storesMixedCaseIdentifiers();
            storesUpperCaseIdentifiers = metadata.storesUpperCaseIdentifiers();
            storesLowerCaseQuotedIdentifiers = metadata.storesLowerCaseQuotedIdentifiers();
            storesMixedCaseQuotedIdentifiers = metadata.storesMixedCaseQuotedIdentifiers();
            storesUpperCaseQuotedIdentifiers = metadata.storesUpperCaseQuotedIdentifiers();

            // Fix to case where the JDBC driver is screwy (e.g Sybase JConnect
            // 5.5) and returns false for all of these (give us a clue!)
            if (storesLowerCaseIdentifiers == false &&
                storesMixedCaseIdentifiers == false &&
                storesUpperCaseIdentifiers == false &&
                storesLowerCaseQuotedIdentifiers == false &&
                storesMixedCaseQuotedIdentifiers == false &&
                storesUpperCaseQuotedIdentifiers == false)
            {
                storesUpperCaseIdentifiers = true;
            }

            // Retrieve the catalog separator string (default = ".")
            catalogSeparator = metadata.getCatalogSeparator();
            catalogSeparator =
                ((catalogSeparator == null) || (catalogSeparator.trim().length() < 1)) ? "." : catalogSeparator;

            // Retrieve the identifier quote string (default = "")
            identifierQuoteString = metadata.getIdentifierQuoteString();
            identifierQuoteString =
                ((null == identifierQuoteString) || (identifierQuoteString.trim().length() < 1)) ?
                "\"" : identifierQuoteString;

            // Create TypeInfo objects for all of the data types and index them
            // in a HashMap by their JDBC type number.
            createTypeInfo(metadata);

            if (supportsCatalogsInTableDefinitions)
            {
                JPOXLogger.RDBMS_SCHEMA.debug(LOCALISER.msg("Adapter.CatalogsInTableDefinitions"));
            }
            if (supportsSchemasInTableDefinitions)
            {
                JPOXLogger.RDBMS_SCHEMA.debug(LOCALISER.msg("Adapter.SchemasInTableDefinitions"));
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("Adapter.MetaDataAccessError"),e);
        }
    }

    /**
     * Accessor for the maximum table name length permitted on this
     * datastore.
     * @return Max table name length
     **/
    public int getMaxTableNameLength()
    {
        return maxTableNameLength;
    }

    /**
     * Accessor for the maximum constraint name length permitted on this
     * datastore.
     * @return Max constraint name length
     **/
    public int getMaxConstraintNameLength()
    {
        return maxConstraintNameLength;
    }

    /**
     * Accessor for the maximum index name length permitted on this datastore.
     * @return Max index name length
     **/
    public int getMaxIndexNameLength()
    {
        return maxIndexNameLength;
    }

    /**
     * Accessor for the maximum column name length permitted on this datastore.
     * @return Max column name length
     **/
    public int getMaxColumnNameLength()
    {
        return maxColumnNameLength;
    }

    /**
     * Iterator for the TypeInfo objects constructed from the method
     * DataBaseMetaData.getTypeInfo
     * @return an Iterator with org.jpox.store.rdbms.typeinfo.TypeInfo
     */
    public Iterator iteratorTypeInfo()
    {
        return typesByTypeNumber.values().iterator();
    }

    /**
     * Iterator for the reserved words constructed from the method
     * DataBaseMetaData.getSQLKeywords + standard SQL reserved words
     * @return an Iterator with a set of reserved words
     */
    public Iterator iteratorReservedWords()
    {
        return reservedKeywords.iterator();
    }
    
    /**
     * Creates TypeInfo objects for all of the data types and indexes them
     * in the typesByTypeNumber map by their JDBC data type number.
     * @param metadata The MetaData for this datastore
     * @throws SQLException Thrown if error occurs in the creation. 
     */
    protected void createTypeInfo(DatabaseMetaData metadata)
    throws SQLException
    {
        ResultSet rs = metadata.getTypeInfo();

        try
        {
            while (rs.next())
            {
                TypeInfo ti = newTypeInfo(rs);

                if (ti != null)
                {
                    Integer key = new Integer(ti.dataType);

                    if (typesByTypeNumber.get(key) == null)
                    {
                        typesByTypeNumber.put(key, ti);
                    }
                }
            }
        }
        finally
        {
            rs.close();
        }
    }

    /**
     * A factory for TypeInfo objects.  This method should always be used
     * instead of directly constructing TypeInfo objects in order to give the
     * DatabaseAdapter an opportunity to modify and/or correct the metadata
     * obtained from the JDBC driver.
     *
     * The type information object is constructed from the current row of the
     * given result set.  The {@link ResultSet} object passed must have been
     * obtained from a call to DatabaseMetaData.getTypeInfo().
     *
     * <p>The constructor only retrieves the values from the current row; the
     * caller is required to advance to the next row with {@link ResultSet#next}
     *
     * @param rs    The result set returned from DatabaseMetaData.getTypeInfo().
     *
     * @return
     *      A TypeInfo object constructed from the current result set row, or
     *      <code>null</code> if the type indicated by this row should be
     *      excluded from use.
     */
    protected TypeInfo newTypeInfo(ResultSet rs)
    {
        return new TypeInfo(rs);
    }

    /**
     * A factory for ColumnInfo objects.  This method should always be used
     * instead of directly constructing ColumnInfo objects in order to give the
     * DatabaseAdapter an opportunity to modify and/or correct the metadata
     * obtained from the JDBC driver.
     *
     * The column information object is constructed from the current row of the
     * given result set.  The {@link ResultSet} object passed must have been
     * obtained from a call to DatabaseMetaData.getColumns().
     *
     * <p>The constructor only retrieves the values from the current row; the
     * caller is required to advance to the next row with {@link ResultSet#next}
     *
     * @param rs    The result set returned from DatabaseMetaData.getColumns().
     * @return The column info 
     */
    public ColumnInfo newColumnInfo(ResultSet rs)
    {
        return new ColumnInfo(rs);
    }

    /**
     * A factory for ForeignKeyInfo objects.  This method should always be used
     * instead of directly constructing ForeignKeyInfo objects in order to give
     * the DatabaseAdapter an opportunity to modify and/or correct the metadata
     * obtained from the JDBC driver.
     *
     * The column information object is constructed from the current row of the
     * given result set.  The {@link ResultSet} object passed must have been
     * obtained from a call to DatabaseMetaData.getImportedKeys() or
     * DatabaseMetaData.getExportedKeys().
     *
     * <p>The constructor only retrieves the values from the current row; the
     * caller is required to advance to the next row with {@link ResultSet#next}
     *
     * @param rs The result set returned from DatabaseMetaData.get??portedKeys()
     * @return The foreign key info 
     */
    public ForeignKeyInfo newForeignKeyInfo(ResultSet rs)
    {
        return new ForeignKeyInfo(rs);
    }

    /**
     * A factory for ForeignKeyInfo objects.  This method should always be used
     * instead of directly constructing ForeignKeyInfo objects in order to give
     * the DatabaseAdapter an opportunity to modify and/or correct the metadata
     * obtained from the JDBC driver.
     *
     * The column information object is constructed from the current row of the
     * given result set.  The {@link ResultSet} object passed must have been
     * obtained from a call to DatabaseMetaData.getExportedKeys().
     *
     * <p>The constructor only retrieves the values from the current row; the
     * caller is required to advance to the next row with {@link ResultSet#next}
     *
     * @param rs The result set returned from DatabaseMetaData.getExportedKeys()
     */
    public ExportedKeyInfo newExportedKeyInfo(ResultSet rs)
    {
        return new ExportedKeyInfo(rs);
    }

    /**
     * Utility method to parse a list of keywords and split them out into
     * words.
     * @param list The comma-separated list of keywords.
     * @return Set of keywords.
     **/
    protected Set parseKeywordList(String list)
    {
        StringTokenizer tokens = new StringTokenizer(list, ",");
        HashSet words = new HashSet();

        while (tokens.hasMoreTokens())
        {
            words.add(tokens.nextToken().trim().toUpperCase());
        }

        return words;
    }

    /**
     * Tests if a given string is a SQL key word.
     * <p>
     * The list of key words tested against is defined to contain all SQL/92 key
     * words, plus any additional key words reported by the JDBC driver for this
     * adapter via <code>DatabaseMetaData.getSQLKeywords()</code>.
     * <p>
     * In general, use of a SQL key word as an identifier should be avoided.
     * SQL/92 key words are divided into reserved and non-reserved words. If a
     * reserved word is used as an identifier it must be quoted with double
     * quotes. Strictly speaking, the same is not true of non-reserved words.
     * However, as C.J. Date writes in <u>A Guide To The SQL Standard </u>:
     * <blockquote>The rule by which it is determined within the standard that
     * one key word needs to be reserved while another need not is not clear to
     * this writer. In practice, it is probably wise to treat all key words as
     * reserved. </blockquote>
     * @param word The word to test.
     * @return <code>true</code> if <var>word </var> is a SQL key word for
     * this DBMS. The comparison is case-insensitive.
     * @see SQLConstants
     */
    public boolean isSQLKeyword(String word)
    {
        return isReservedKeyword(word.toUpperCase());
    }

    /**
     * Returns type information for the database type that best implements the
     * given JDBC type.
     * @param dataType JDBC type number of the data type.
     * @return type information for the best matching type.
     * @throws UnsupportedDataTypeException Thrown if the specified datatype is
     * not supported.
     */
    public TypeInfo getTypeInfo(int dataType)
    throws UnsupportedDataTypeException
    {
        TypeInfo ti = (TypeInfo)typesByTypeNumber.get(new Integer(dataType));

        if (ti == null)
        {
            throw new UnsupportedDataTypeException(LOCALISER.msg("Adapter.JDBCTypeNotSupported",TypeInfo.getJDBCTypeName(dataType)));
        }

        return ti;
    }

    /**
     * Returns type information for the first one of the given candidate JDBC
     * data types supported by this database.
     *
     * @param   candidateDataTypes
     *              array of JDBC type numbers of the data types to be checked
     *              in order of preference.
     * @return  type information for the first supported type.
     * @throws UnsupportedDataTypeException Thrown if one of the datatypes is
     * not supported. 
     */
    public TypeInfo getTypeInfo(int[] candidateDataTypes)
    throws UnsupportedDataTypeException
    {
        for (int i = 0; i < candidateDataTypes.length; i++)
        {
            TypeInfo ti = (TypeInfo)typesByTypeNumber.get(new Integer(candidateDataTypes[i]));

            if (ti != null)
            {
                return ti;
            }
        }

        JPOXLogger.RDBMS_SCHEMA.error(LOCALISER.msg("Adapter.JDBCTypesNotSupported"));
        for (int i=0;i<candidateDataTypes.length;i++)
        {
            JPOXLogger.RDBMS_SCHEMA.error(" type=" + TypeInfo.getJDBCTypeName(candidateDataTypes[i]));
        }

        throw new UnsupportedDataTypeException(LOCALISER.msg("Adapter.JDBCTypesNotSupported"));
    }

    /**
     * Returns the precision value to be used when creating string columns of
     * "unlimited" length.  Usually, if this value is needed it is provided in
     * the database metadata ({@link TypeInfo#precision}).  However, for some
     * types in some databases the value must be computed specially.
     *
     * @param typeInfo  the typeInfo object for which the precision value is
     *                  needed.
     *
     * @return  the precision value to be used when creating the column, or -1
     *          if no value should be used.
     */

    public int getUnlimitedLengthPrecisionValue(TypeInfo typeInfo)
    {
        if (typeInfo.createParams != null && typeInfo.createParams.length() > 0)
        {
            return typeInfo.precision;
        }
        else
        {
            return -1;
        }
    }

    /**
     * Accessor for whether the class is an embedded type.
     * @param c Class to query
     * @return Whether the class is an embedded type
     **/
    public boolean isEmbeddedType(Class c)
    {
        return !OIDMapping.class.isAssignableFrom(getMappingManager().getMappingClass(c));
    }
    
    /**
     * Some databases, Oracle, treats an empty string (0 length) equals null
     * @return whether the database treats an empty string as null
     */
    public boolean isNullEqualsEmptyStrings()
    {
        return false;
    }

    /**
     * Accessor for whether a JDBC Types.BIT is really mapped in the datastore
     * to a boolean field.
     * @return Whether it is really a boolean
     */
    public boolean isBitReallyBoolean()
    {
        return false;
    }

    /**
     * Some databases, Oracle, treats an empty string (0 length) equals null
     * @return returns a surrogate to replace the empty string in the database
     * otherwise it would be treated as null
     */
    public String getSurrogateForEmptyStrings()
    {
        return null;
    }

    /**
     * Acessor for the MappingManager
     * @return the MappingManager
     */
    public MappingManager getMappingManager()
    {
        if (mappingManager == null)
        {
            mappingManager = new RDBMSMappingManager();
        }
        return mappingManager;
    }

    /**
     * Accessor for the mapping for the specified class.
     * @param c Class to query
     * @return The mapping for the class.
     **/
    public JavaTypeMapping getMapping(Class c)
    {
        Class mc = getMappingManager().getMappingClass(c);

        JavaTypeMapping m = MappingFactory.createMapping(mc,this,c);
        if (m == null)
        {
            String name = mc.getName();
            name = name.substring(name.lastIndexOf('.') + 1);

            throw new JDOUserException(LOCALISER_BASE.msg("DatastoreAdapter.MappingFieldError",name));
        }
        return m;
    }

    /**
     * Accessor for the mapping for the specified class.
     * @param c Class to query
     * @param storeMgr The Store manager
     * @return The mapping for the class.
     **/
    public JavaTypeMapping getMapping(Class c, StoreManager storeMgr)
    {
        JavaTypeMapping m;
        try
        {
            DatastoreClass datastoreClass = storeMgr.getDatastoreClass(c.getName());
            m = datastoreClass.getIDMapping();
        }
        catch (NoTableManagedException ex)
        {
            Class mc = getMappingManager().getMappingClass(c);
            m = MappingFactory.createMapping(mc, this, c);
            if (m == null)
            {
                String name = mc.getName();
                name = name.substring(name.lastIndexOf('.') + 1);

                throw new JDOUserException(LOCALISER_BASE.msg("DatastoreAdapter.MappingFieldError", name));
            }
            if (m instanceof SimpleDatastoreRepresentation)
            {
                getMappingManager().getDatastoreMapping(m, storeMgr, null, m.getJavaType());
            }
        }
        return m;
    }

    /**
     * Accessor for whether the adapter supports the transaction isolation level
     * @return Whether the transaction isolation level setting is supported.
     */
    public boolean supportsTransactionIsolationLevel(int isolationLevel)
    {
        return true;
    }

    /**
     * Accessor for a Connection to the datastore.
     * @param ds The data source
     * @param userName The username for the datastore
     * @param password The password for the datastore
     * @param isolationLevel The level of transaction isolation
     * @return The Connection
     * @throws SQLException Thrown when an error occurs in the creation.
     **/
    public Connection getConnection(DataSource ds, String userName, String password, int isolationLevel)
    throws SQLException
    {
        Connection conn;

        if (userName == null)
        {
            conn = ds.getConnection();
        }
        else
        {
            conn = ds.getConnection(userName, password);
        }

        boolean succeeded = false;

        try
        {
            if (isolationLevel == Connection.TRANSACTION_NONE)
            {
                conn.setAutoCommit(true);
            }
            else
            {
                conn.setAutoCommit(false);
                if (supportsTransactionIsolationLevel(isolationLevel))
                {
                    conn.setTransactionIsolation(isolationLevel);
                }
                else
                {
                    JPOXLogger.RDBMS.warn(LOCALISER.msg("Adapter.TransactionIsolationUnsupported", isolationLevel));
                }
            }

            succeeded = true;
        }
        finally
        {
            if (!succeeded)
            {
                conn.close();
            }
        }

        return conn;
    }

    /**
     * Method to close a Connection to the datastore.
     * @param conn The connection
     * @throws SQLException Thrown if error occurs on the close.
     **/
    public void closeConnection(Connection conn)
    throws SQLException
    {
        conn.close();
    }

    /**
     * Accessor for the Catalog Name for this datastore.
     * @param conn Connection to the datastore
     * @return The catalog name
     * @throws SQLException Thrown if error occurs in determining the
     * catalog name.
     **/
    public String getCatalogName(Connection conn)
    throws SQLException
    {
        throw new UnsupportedOperationException(LOCALISER.msg("Adapter.CatalogNameError",datastoreProductName,datastoreProductVersion));
    }

    /**
     * Accessor for the Schema Name for this datastore.
     * @param conn Connection to the datastore
     * @return The schema name
     * @throws SQLException Thrown if error occurs in determining the
     * schema name.
     **/
    public String getSchemaName(Connection conn)
    throws SQLException
    {
        throw new UnsupportedOperationException(LOCALISER.msg("Adapter.SchemaNameError",datastoreProductName,datastoreProductVersion));
    }

    /**
     * Accessor for the catalog separator.
     * @return Catalog separator string.
     **/
    public String getCatalogSeparator()
    {
        return catalogSeparator;
    }

    /**
     * Whether this datastore supports the use of the escape expression
     * in like predicates
     * @return whether we can use the escape expression in like predicates
     **/
    public boolean supportsEscapeExpressionInLikePredicate()
    {
        return true;
    }
    
    /**
     * Whether this datastore supports the use of the catalog name in SQL
     * table definitions (DDL).
     * @return whether we can use the catalog name in DDL
     **/
    public boolean supportsCatalogsInTableDefinitions()
    {
        return supportsCatalogsInTableDefinitions;
    }

    /**
     * Whether this datastore supports the use of the schema name in SQL
     * table definitions (DDL).
     * @return whether we can use the schema name in DDL
     **/
    public boolean supportsSchemasInTableDefinitions()
    {
        return supportsSchemasInTableDefinitions;
    }

    /**
     * Accessor for an identifier quote string.
     * @return Identifier quote string.
     **/
    public String getIdentifierQuoteString()
    {
        return identifierQuoteString;
    }

    /**
     * The pattern string for representing one character.
     * Most of databases will use the underscore character.
     * @return the pattern string.
     **/
    public ScalarExpression getEscapedPatternExpression(ScalarExpression patternExpression)
    {
        if (patternExpression instanceof StringLiteral)
        {
	        String value = (String) ((StringLiteral)patternExpression).getValue();
	        value = StringUtils.replaceAll(value,"\\","\\\\");
	        value = StringUtils.replaceAll(value,"%","\\%");
	        value = StringUtils.replaceAll(value,"_","\\_");
	        return getMapping(String.class).newLiteral(patternExpression.getQueryStatement(),value);
        }
        else
        {
            return patternExpression;
        }
    }

    /**
     * The pattern string for representing one character.
     * Most of databases will use the underscore character.
     * @return the pattern string.
     **/
    public String getPatternExpressionAnyCharacter()
    {
        return "_";
    }
    
    /**
     * The pattern string for representing zero or more characters.
     * Most of databases will use the percent sign character.
     * @return the pattern string.
     **/
    public String getPatternExpressionZeroMoreCharacters()
    {
        return "%";
    }
    
    /**
     * The character for escaping patterns.
     * @return the character.
     **/
    public String getEscapePatternExpression()
    {
        return "ESCAPE '\\'";
    }
    
    /**
     * Continuation string to use where the SQL statement goes over more than 1
     * line. Some JDBC adapters (e.g DB2) don't do conversion.
     * @return Continuation string.
     **/
    public String getContinuationString()
    {
        return "\n";
    }
 
    /**
     * Whether to create indexes before foreign keys.
     * @return Whether to create indexes before foreign keys
     **/
    public boolean createIndexesBeforeForeignKeys()
    {
        return false;
    }

    /**
     * Whether to include ORDER BY columns in a SELECT.
     * @return Whether to include ORDER BY in SELECTs
     **/
    public boolean includeOrderByColumnsInSelect()
    {
        return true;
    }

    /**
     * Whether to specify ORDER BY columns in SELECTS using aliases.
     * @return Whether to include ORDER BY in SELECTs using aliases
     **/
    public boolean includeOrderByColumnsInSelectUsingAlias()
    {
        return false;
    }
    
    /**
     * Union combines the results of two or more queries into a single result
     * set. Union include only distinct rows and Union all may include
     * duplicates.
     * @return Whether to use UNION ALL instead of UNION
     */
    public boolean useUnionAll()
    {
        return false;
    }

    /**
     * Whether this datastore supports ALTER TABLE DROP constraints
     * @return whether we support ALTER TABLE DROP constraints
     **/
    public boolean supportsAlterTableDropConstraint()
    {
        return true;
    }

    /**
     * Whether this datastore supports ALTER TABLE DROP FOREIGN KEY constraints
     * @return whether we support ALTER TABLE DROP FOREIGN KEY constraints
     **/
    public boolean supportsAlterTableDropForeignKeyConstraint()
    {
        return false;
    }

    /**
     * Whether this datastore supports deferred constraints.
     * @return whether we support deferred constraints.
     **/
    public boolean supportsDeferredConstraints()
    {
        return true;
    }

    /**
     * Whether this datastore supports SELECT ... FOR UPDATE.
     * @return whether we support SELECT ... FOR UPDATE.
     **/
    public boolean supportsLockWithSelectForUpdate()
    {
        return false;
    }

    /**
     * The option to specify in "SELECT ... FROM TABLE ... WITH (option)" to lock instances
     * Null if not supported.
     * @return The option to specify with "SELECT ... FROM TABLE ... WITH (option)"
     **/
    public String getSelectWithLockOption()
    {
        return null;
    }

    /**
     * The function to creates a unique value of type uniqueidentifier.
     * @return The function. e.g. "SELECT NEWID()"
     **/
    public String getSelectNewUUIDStmt()
    {
        return null;
    }
    
    /**
     * The function to creates a unique value of type uniqueidentifier.
     * @return The function. e.g. "NEWID()"
     **/
    public String getNewUUIDFunction()
    {
        return null;
    }
    
    /**
     * Whether this datastore supports the specified foreign key update action
     * @param action The update action
     * @return Whether it is supported
     */
    public boolean supportsForeignKeyUpdateAction(ForeignKey.FKAction action)
    {
        return true;
    }

    /**
     * Whether this datastore supports the specified foreign key delete action
     * @param action The delete action
     * @return Whether it is supported
     */
    public boolean supportsForeignKeyDeleteAction(ForeignKey.FKAction action)
    {
        return true;
    }

    /**
     * Whether the datastore supports specification of the primary key in
     * CREATE TABLE statements.
     * @return Whetehr it allows "PRIMARY KEY ..."
     */
    public boolean supportsPrimaryKeyInCreateStatements()
    {
        return false;
    }

    /**
     * Whether this datastore supports the use of CHECK in CREATE TABLE
     * statements (DDL).
     * @return whether we can use CHECK in CREATE TABLE.
     **/
    public boolean supportsCheckInCreateStatements()
    {
        return true;
    }

    /**
     * Whether this datastore supports the use of CHECK after the column
     * definitions in CREATE TABLE statements (DDL).
     * e.g.<PRE>
     * CREATE TABLE XXX
     * (
     *     COL_A int,
     *     COL_B char(1),
     *     PRIMARY KEY (COL_A),
     *     CHECK (COL_B IN ('Y','N'))
     * )
     * </PRE>
     * @return whether we can use CHECK after the column
     * definitions CREATE TABLE.
     **/
    public boolean supportsCheckConstraintsInEndCreateStatements()
    {
        return false;
    }

    /**
     * Whether this datastore supports the use of UNIQUE after the column
     * definitions in CREATE TABLE statements (DDL).
     * e.g.<PRE>
     * CREATE TABLE XXX
     * (
     *     COL_A int,
     *     COL_B char(1),
     *     PRIMARY KEY (COL_A),
     *     UNIQUE (COL_B ...)
     * )
     * </PRE> 
     * @return whether we can use UNIQUE after the column definitions CREATE TABLE.
     **/
    public boolean supportsUniqueConstraintsInEndCreateStatements()
    {
        return false;
    }
    
    /**
     * Whether we support Boolean comparisons.
     * @return whether we support Boolean comparisons.
     **/
    public boolean supportsBooleanComparison()
    {
        return true;
    }

    /**
     * Whether we support NULLs in candidate keys.
     * @return whether we support NULLs in candidate keys.
     **/
    public boolean supportsNullsInCandidateKeys()
    {
        return true;
    }

    /**
     * Whether the database support NULLs in the column options for table creation.
     * @return whether the database support NULLs in the column options for table creation.
     **/
    public boolean supportsNullsKeywordInColumnOptions()
    {
        return true;
    }
    

    /**
     * Whether we support DEFAULT tag in CREATE TABLE statements
     * @return whether we support the DEFAULT tag in CREATE TABLE
     **/
    public boolean supportsDefaultKeywordInColumnOptions()
    {
        return true;
    }
    
    /**
     * Whether we support DEFAULT tag together with NOT NULL in CREATE TABLE statements.
     * <pre>
     * CREATE TABLE X ( MEMORY_SIZE BIGINT DEFAULT 0 NOT NULL )
     * </pre>
     * Some databases only support <i>DEFAULT {ConstantExpression | NULL}</i>
     * 
     * @return whether we support the DEFAULT tag together with NOT NULL in CREATE TABLE
     **/
    public boolean supportsDefaultKeywordWithNotNullInColumnOptions()
    {
        return true;
    }

    // ---------------------------- AutoIncrement Support ----------------------
 
    /**
     * Whether we support auto-increment/identity in keys.
     * @return whether we support auto-increment/identity in keys.
     **/
    public boolean supportsAutoIncrementKeys()
    {
        return false;
    }
 
    /**
     * Accessor for the auto-increment/identity sql statement for this datastore.
     * @return The statement for getting the latest auto-increment/identity key
     **/
    public String getAutoIncrementStmt()
    {
        throw new UnsupportedOperationException(LOCALISER.msg("Adapter.AutoIncrementNotSupported"));
    }

    /**
     * Accessor for the auto-increment/identity keyword for generating DDLs.
     * (CREATE TABLEs...).
     * @return The keyword for a column using auto-increment/identity
     **/
    public String getAutoIncrementKeyword()
    {
        throw new UnsupportedOperationException(LOCALISER.msg("Adapter.AutoIncrementNotSupported"));
    }

    /**
     * Whether we support auto-increment/identity keys with nullability specification.
     * @return whether we support auto-increment keys with nullability spec.
     **/
    public boolean supportsAutoIncrementKeysNullSpecification()
    {
        return true;
    }

    // ---------------------------- Sequence Support ---------------------------
 
    /**
     * Whether we support sequences.
     * @return whether we support sequences.
     **/
    public boolean supportsSequences()
    {
        return false;
    }

    /**
     * Accessor for the sequence create statement for this datastore.
     * TODO Change param types to int.
     * @param sequence_name Name of the sequence 
     * @param min Minimum value for the sequence
     * @param max Maximum value for the sequence
     * @param start Start value for the sequence
     * @param increment Increment value for the sequence
     * @param cache_size Cache size for the sequence
     * @return The statement for getting the next id from the sequence
     **/
    public String getSequenceCreateStmt(String sequence_name,
                                        String min,String max,
                                        String start,String increment,
                                        String cache_size)
    {
        throw new UnsupportedOperationException(LOCALISER.msg("Adapter.SequenceNotSupported"));
    }

    /**
     * Accessor for the sequence statement to get the next id for this 
     * datastore.
     * @param sequence_name Name of the sequence 
     * @return The statement for getting the next id for the sequence
     **/
    public String getSequenceNextStmt(String sequence_name)
    {
        throw new UnsupportedOperationException(LOCALISER.msg("Adapter.SequenceNotSupported"));
    }

    /**
     * Provide the existing indexes in the database for the table
     * @param conn the JDBC connection
     * @param dmd the DatabaseMetaData
     * @param catalog the catalog name
     * @param schema the schema name
     * @param table the table name
     * @return a ResultSet with the format @see DatabaseMetaData#getIndexInfo(java.lang.String, java.lang.String, java.lang.String, boolean, boolean)
     * @throws SQLException
     */
    public ResultSet getExistingIndexes(Connection conn, DatabaseMetaData dmd, String catalog, String schema, String table) throws SQLException
    {
        return dmd.getIndexInfo(catalog, schema, table, false, true);        
    }
    
    // ------------------------------- Queries ---------------------------------
    
    /**
     * Accessor for a new query statement.
     * @param table The table to query 
     * @return The Query Statement
     **/
    public QueryStatement newQueryStatement(DatastoreContainerObject table)
    {
        return new QueryStatement(table);
    }

    /**
     * Accessor for a new query statement.
     * @param table The table to query 
     * @param rangeVar A range variable for the query
     * @return The Query Statement
     **/
    public QueryStatement newQueryStatement(DatastoreContainerObject table, DatastoreIdentifier rangeVar)
    {
        return new QueryStatement(table, rangeVar);
    }

    /**
     * Returns a new TableExpression object appropriate for this DBMS.
     * This should be an instance of one of the three built-in styles of table
     * expression:
     * <ul>
     *   <li>TableExprAsJoins</li>
     *   <li>TableExprAsSubjoins</li>
     *   <li>TableExprAsSubquery</li>
     * </ul>
     * TableExprAsSubjoins is the default, which arguably produces the most
     * readable SQL but doesn't work on all DBMS's.  TableExprAsSubjoins
     * should work anywhere, but may be less efficient.
     *
     * @param qs        The query statement in which the table expression will
     *                  be included.
     * @param table     The main table in the expression.
     * @param rangeVar  The SQL alias, or "range variable", to assign to the
     *                  expression or to the main table.
     * @return The expression
     */
    public TableExpression newTableExpression(QueryStatement qs, DatastoreContainerObject table, DatastoreIdentifier rangeVar)
    {
        return new TableExprAsSubjoins(qs, table, rangeVar);
    }

    /**
     * Returns the appropriate SQL to create the given table having the given
     * columns. No column constraints or key definitions should be included.
     * It should return something like:
     * <p>
     * <pre>
     * CREATE TABLE FOO ( BAR VARCHAR(30), BAZ INTEGER )
     * </pre>
     *
     * @param table The table to create.
     * @param columns The columns of the table.
     * @return The text of the SQL statement.
     */
    public String getCreateTableStatement(TableImpl table, Column[] columns)
    {
        StringBuffer createStmt = new StringBuffer();

        // CREATE TABLE with column specifiers
        createStmt.append("CREATE TABLE ").append(table.toString())
                  .append(getContinuationString())
                  .append("(")
                  .append(getContinuationString());
        for (int i = 0; i < columns.length; ++i)
        {
            if (i > 0)
            {
                createStmt.append(",").append(getContinuationString());
            }

            createStmt.append("    ").append(columns[i].getSQLDefinition());
        }

        // PRIMARY KEY(col[,col])
        if (supportsPrimaryKeyInCreateStatements())
        {
            StringBuffer pkConstraintStmt = new StringBuffer();
            for (int i = 0; i < columns.length; ++i)
            {
                if (columns[i].isPrimaryKey())
                {
                    if (pkConstraintStmt.length() < 1)
                    {
                        pkConstraintStmt.append(",").append(getContinuationString());
                        pkConstraintStmt.append("    PRIMARY KEY (");
                    }
                    else
                    {
                        pkConstraintStmt.append(",");
                    }
                    pkConstraintStmt.append(columns[i].getSQLColumnNameDefinition());
                }
            }
            if (pkConstraintStmt.length() > 1)
            {
                pkConstraintStmt.append(")");
                createStmt.append(pkConstraintStmt.toString());
            }
        }
        
        // UNIQUE( col [,col] )
        if (supportsUniqueConstraintsInEndCreateStatements())
        {
            StringBuffer uniqueConstraintStmt = new StringBuffer();
            for (int i = 0; i < columns.length; ++i)
            {
                if (columns[i].isUnique())
                {
                    if (uniqueConstraintStmt.length() < 1)
                    {
                        uniqueConstraintStmt.append(",").append(getContinuationString());
                        uniqueConstraintStmt.append(" UNIQUE (");
                    }
                    else
                    {
                        uniqueConstraintStmt.append(",");
                    }
                    uniqueConstraintStmt.append(columns[i].getSQLColumnNameDefinition());
                }
            }       
            if (uniqueConstraintStmt.length() > 1)
            {
                uniqueConstraintStmt.append(")");
                createStmt.append(uniqueConstraintStmt.toString());
            }
        }
        
        // CHECK (column_identifier IN (literal[,literal]))
        if (supportsCheckConstraintsInEndCreateStatements())
        {
            StringBuffer checkConstraintStmt = new StringBuffer();
	        for (int i = 0; i < columns.length; ++i)
	        {
	            if (columns[i].getConstraints() != null)
	            {
	                checkConstraintStmt.append(",").append(getContinuationString());
	                checkConstraintStmt.append(columns[i].getConstraints());
	            }
	        }
	        if (checkConstraintStmt.length() > 1)
	        {
	            createStmt.append(checkConstraintStmt.toString());
	        }
        }

        createStmt.append(getContinuationString()).append(")");

        return createStmt.toString();
    }

    /**
     * Returns the appropriate SQL to add a primary key to its table.
     * It should return something like:
     * <p>
     * <pre>
     * ALTER TABLE FOO ADD CONSTRAINT FOO_PK PRIMARY KEY (BAR)
     * </pre>
     *
     * @param pkName The name of the primary key to add.
     * @param pk An object describing the primary key.
     * @return The text of the SQL statement.
     */
    public String getAddPrimaryKeyStatement(SQLIdentifier pkName, PrimaryKey pk)
    {
        return "ALTER TABLE " + pk.getDatastoreContainerObject().toString() + " ADD CONSTRAINT " + pkName + ' ' + pk;
    }

    /**
     * Returns the appropriate SQL to add a candidate key to its table.
     * It should return something like:
     * <p>
     * <pre>
     * ALTER TABLE FOO ADD CONSTRAINT FOO_CK UNIQUE (BAZ)
     * </pre>
     *
     * @param ck An object describing the candidate key.
     * @return  The text of the SQL statement.
     */
    public String getAddCandidateKeyStatement(CandidateKey ck)
    {
        return "ALTER TABLE " + ck.getDatastoreContainerObject().toString() + " ADD CONSTRAINT " + ck.getName() + ' ' + ck;
    }

    /**
     * Returns the appropriate SQL to add a foreign key to its table.
     * It should return something like:
     * <p>
     * <pre>
     * ALTER TABLE FOO ADD CONSTRAINT FOO_FK1 FOREIGN KEY (BAR, BAZ) REFERENCES ABC (COL1, COL2)
     * </pre>
     * @param fk An object describing the foreign key.
     * @return  The text of the SQL statement.
     */
    public String getAddForeignKeyStatement(ForeignKey fk)
    {
        return "ALTER TABLE " + fk.getDatastoreContainerObject().toString() + " ADD CONSTRAINT " + fk.getName() + ' ' + fk;
    }

    /**
     * Accessor for the SQL statement to add a column to a table.
     * @param table The table
     * @param col The column
     * @return The SQL necessary to add the column
     */
    public String getAddColumnStatement(Table table, Column col)
    {
        return "ALTER TABLE " + table.toString() + " ADD " + col.getSQLDefinition();
    }
    
    /**
     * Returns the appropriate SQL to add an index to its table.
     * It should return something like:
     * <p>
     * <pre>
     * CREATE INDEX FOO_N1 ON FOO (BAR,BAZ)
     * CREATE UNIQUE INDEX FOO_U1 ON FOO (BAR,BAZ)
     * </pre>
     *
     * @param idx An object describing the index.
     * @return The text of the SQL statement.
     */
    public String getCreateIndexStatement(Index idx)
    {
        return "CREATE " + (idx.getUnique() ? "UNIQUE " : "") + "INDEX " + idx.getName() + " ON " + idx.getDatastoreContainerObject().toString() + ' ' + idx;
    }

    /**
     * Returns the appropriate SQL to drop the given table.
     * It should return something like:
     * <p>
     * <pre>
     * DROP TABLE FOO CASCADE
     * </pre>
     *
     * @param table The table to drop.
     * @return The text of the SQL statement.
     */
    public String getDropTableStatement(TableImpl table)
    {
        return "DROP TABLE " + table.toString() + " CASCADE";
    }

    /**
     * Returns the appropriate SQL to drop the given view.
     * It should return something like:
     * <p>
     * <pre>
     * DROP VIEW FOO
     * </pre>
     *
     * @param view The view to drop.
     * @return The text of the SQL statement.
     */
    public String getDropViewStatement(ViewImpl view)
    {
        return "DROP VIEW " + view.toString();
    }

    /**
     * Method to return the SQL to append to a SELECT statement to handle
     * restriction of ranges using the LIMUT keyword. Defaults to an empty
     * string (not supported).
     * @param offset The offset to return from
     * @param count The number of items to return
     * @return The SQL to append to allow for ranges using LIMIT.
     */
    public String getRangeByLimitClause(long offset, long count)
    {
        return "";
    }

    /**
     * Method to return the column name to use when handling ranges via
     * a rownumber on the select. Defaults to an empty string (not supported).
     * @return The row number column.
     */
    public String getRangeByRowNumberColumn()
    {
        return "";
    }

    /**
     * Returns the appropriate SQL expression for the JDOQL Math.abs(EXPRESSION)
     * method.
     * It should return something like:
     * <p>
     * <blockquote><pre>
     * ABS(str)
     * </pre></blockquote>
     *
     * @param expr   The argument to the abs() method.
     * @return  The text of the SQL expression.
     */
    public NumericExpression absMethod(ScalarExpression expr)
    {
        ArrayList args = new ArrayList();
        args.add(expr);

        return new NumericExpression("ABS", args);
    }
    
    /**
     * Returns the appropriate SQL expression for the JDOQL Math.sqrt(EXPRESSION)
     * method.
     * It should return something like:
     * <p>
     * <blockquote><pre>
     * SQRT(str)
     * </pre></blockquote>
     *
     * @param expr   The argument to the sqrt() method.
     * @return  The text of the SQL expression.
     */
    public NumericExpression sqrtMethod(ScalarExpression expr)
    {
        ArrayList args = new ArrayList();
        args.add(expr);

        return new NumericExpression("SQRT", args);
    }    
    
    /**
     * Returns the appropriate SQL expression for the JDOQL String.length()
     * method.
     * It should return something like:
     * <p>
     * <blockquote><pre>
     * CHAR_LENGTH(str)
     * </pre></blockquote>
     *
     * @param str   The argument to the length() method.
     * @return  The text of the SQL expression.
     */
    public NumericExpression lengthMethod(StringExpression str)
    {
        ArrayList args = new ArrayList();
        args.add(str);

        return new NumericExpression("CHAR_LENGTH", args);
    }

    /**
     * Converts a numeric expression into a StringExpression
     * @param expr the expression to be converted
     * @return String form of the numeric expression
     */
    public ConvertExpression convertExprToString(NumericExpression expr)
    {
        return new ConvertExpression(expr.getQueryStatement(), expr, "VARCHAR(4000)");
    }

    /**
     * Returns the appropriate SQL expression for the JDOQL
     * String.substring(str,begin) method.
     * It should return something like:
     * <p>
     * <pre>
     * SUBSTRING(str FROM begin)
     * </pre>
     * Note that the value of <var>begin</var> is base 0(Java-style), while most
     * SQL string functions use base 1.
     *
     * @param str The first argument to the substring() method.
     * @param begin The second argument to the substring() method.
     * @return  The text of the SQL expression.
     */
    public StringExpression substringMethod(StringExpression str,
                                            NumericExpression begin)
    {
        return new SubstringExpression(str, begin);
    }

    /**
     * Returns the appropriate SQL expression for the JDOQL
     * String.trim(str) method.
     * It should return something like:
     * <p>
     * <pre>
     * TRIM(str)
     * </pre>
     *
     * @param str The first argument to the trim() method.
     * @return  The text of the SQL expression.
     */
    public StringExpression trimMethod(StringExpression str)
    {
        ArrayList args = new ArrayList();
        args.add(str);

        return new StringExpression("TRIM", args);
    }

    /**
     * Returns the appropriate SQL expression for the JDOQL
     * String.substring(str,begin,end) method.
     * It should return something like:
     * <p>
     * <blockquote><pre>
     * SUBSTRING(str FROM begin FOR len)
     * </pre></blockquote>
     * Note that the value of <var>begin</var> is base 0 (Java-style), while most
     * SQL string functions use base 1.
     * Note also that an end position is given, while most SQL substring
     * functions take a length.
     *
     * @param str   The first argument to the substring() method.
     * @param begin The second argument to the substring() method.
     * @param end   The third argument to the substring() method.
     * @return  The text of the SQL expression.
     */
    public StringExpression substringMethod(StringExpression str,
                                            NumericExpression begin,
                                            NumericExpression end)
    {
        return new SubstringExpression(str, begin, end);
    }

    /**
     * Method to handle the starts with operation.
     * @param str The string to compare against 
     * @return The expression.
     **/
    public BooleanExpression startsWithMethod(ScalarExpression source, ScalarExpression str)
    {
        ScalarExpression pct = getMapping(String.class).newLiteral(source.getQueryStatement(), "%");

        return new BooleanExpression(source, ScalarExpression.OP_LIKE, getEscapedPatternExpression(str).add(pct));
    }
 
    /**
     * An operator in a string expression that concatenates two or more
     * character or binary strings, columns, or a combination of strings and
     * column names into one expression (a string operator).
     * 
     * @return the operator SQL String
     */
    public String getOperatorConcat()
    {
        return "||";
    }

    /**
     * Concatenates two or more
     * character or binary strings, columns, or a combination of strings and
     * column names into one expression (a string operator).
     * 
     * @return the operator SQL String
     */
    public ScalarExpression concatenateExpressions(ScalarExpression operand1, ScalarExpression operand2)
    {
        return new ConcatOperatorExpression(operand1, operand2);
    }

    /**
     * Overridden toString() method adding more information.
     * @return The string form of this object.
     */
    public String toString()
    {
        StringBuffer str=new StringBuffer(super.toString());
        str.append("\n");
        str.append("Driver name="+ driverName + ", version=" + driverVersion + ", major=" + driverMajorVersion + ", minor=" + driverMinorVersion + "\n");
        str.append("Identifier Max Lengths : Table=" + maxTableNameLength + 
            "  Constraints=" + maxConstraintNameLength + 
            "  Index=" + maxIndexNameLength + 
            "  Column=" + maxColumnNameLength +
            "  Delimeters=" + identifierQuoteString + "\n");
        str.append("Identifier Support in DDL : catalog=" + supportsCatalogsInTableDefinitions + 
            "  schema=" + supportsSchemasInTableDefinitions);

        return str.toString();
    }
}