/**********************************************************************
Copyright (c) 2003 Erik Bengtson and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
2004 Andy Jefferson - updated to specify the transaction isolation level limitation
2004 Andy Jefferson - changed to use CREATE TABLE of DatabaseAdapter
2004 Andy Jefferson - added MappingManager
    ...
**********************************************************************/
package org.jpox.store.rdbms.adapter;

import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.SQLException;

import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.expression.TableExprAsJoins;
import org.jpox.store.expression.TableExpression;
import org.jpox.store.mapping.MappingManager;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.key.ForeignKey;
import org.jpox.store.rdbms.key.PrimaryKey;
import org.jpox.store.rdbms.mapping.HSQLRDBMSMappingManager;
import org.jpox.store.rdbms.sqlidentifier.SQLIdentifier;
import org.jpox.store.rdbms.table.Table;
import org.jpox.store.rdbms.table.TableImpl;
import org.jpox.store.rdbms.typeinfo.HSQLTypeInfo;
import org.jpox.store.rdbms.typeinfo.TypeInfo;

/**
 * Provides methods for adapting SQL language elements to the Hypersonic SQL
 * Server database.
 *
 * @version $Revision: 1.8 $ 
 */
class HSQLAdapter extends DatabaseAdapter
{
    /**
     * Constructs a Hypersonic SQL adapter based on the given JDBC
     * metadata.
     *
     * @param metadata the database metadata.
     */
    public HSQLAdapter(DatabaseMetaData metadata)
    {
        super(metadata);
    }

    /**
     * Accessor for the vendor ID for this adapter.
     * @return The vendor ID
     */
    public String getVendorID()
    {
        return "hsql";
    }

    /**
     * Accessor for the maximum table name length permitted on this
     * datastore.
     * @return Max table name length
     **/
    public int getMaxTableNameLength()
    {
        return SQLConstants.MAX_IDENTIFIER_LENGTH;
    }

    /**
     * Accessor for the maximum constraint name length permitted on this
     * datastore.
     * @return Max constraint name length
     **/
    public int getMaxConstraintNameLength()
    {
        return SQLConstants.MAX_IDENTIFIER_LENGTH;
    }

    /**
     * Accessor for the maximum index name length permitted on this datastore.
     * @return Max index name length
     **/
    public int getMaxIndexNameLength()
    {
        return SQLConstants.MAX_IDENTIFIER_LENGTH;
    }

    /**
     * Accessor for the maximum column name length permitted on this datastore.
     * @return Max column name length
     **/
    public int getMaxColumnNameLength()
    {
        return SQLConstants.MAX_IDENTIFIER_LENGTH;
    }

    /**
     * Accessor for the SQL statement to add a column to a table.
     * @param table The table
     * @param col The column
     * @return The SQL necessary to add the column
     */
    public String getAddColumnStatement(Table table, Column col)
    {
        return "ALTER TABLE " + table.toString() + " ADD COLUMN " + col.getSQLDefinition();
    }

    /**
     * Method to return the SQL to append to a SELECT statement to handle
     * restriction of ranges using the LIMUT keyword.
     * @param offset The offset to return from
     * @param count The number of items to return
     * @return The SQL to append to allow for ranges using LIMIT.
     */
    public String getRangeByLimitClause(long offset, long count)
    {
        if (offset >= 0 && count > 0)
        {
            return "LIMIT " + offset + "," + count;
        }
        else if (offset <= 0 && count > 0)
        {
            return "LIMIT 0 " + count;
        }
        else
        {
            return "";
        }
    }

    /**
     * Factory for TypeInfo objects.
     * @param rs The ResultSet from DatabaseMetaData.getTypeInfo(). 
     * @return A TypeInfo object.
     **/
    public TypeInfo newTypeInfo(ResultSet rs)
    {
        return new HSQLTypeInfo(rs);
    }

    /**
     * Accessor for whether the adapter supports the transaction isolation level
     * HSQL 1.7.0 does not support Connection.setTransactionIsolation(isolationLevel);
     * Up to and including 1.7.1, HSQLDB supports only Connection.TRANSACTION_READ_UNCOMMITTED. 
     * @return Whether the transaction isolation level setting is supported.
     */
    public boolean supportsTransactionIsolationLevel(int isolationLevel)
    {
        if (isolationLevel == Connection.TRANSACTION_READ_UNCOMMITTED)
        {
            return true;
        }

        return false;
    }

    /**
     * Whether the datastore supports specification of the primary key in CREATE
     * TABLE statements.
     * @return Whetehr it allows "PRIMARY KEY ..."
     */
    public boolean supportsPrimaryKeyInCreateStatements()
    {
        return true;
    }

    /**
     * Accessor for the Schema Name for this datastore.
     * HSQL 1.7.0 does not support schemas (catalog)
     * 
     * @param conn Connection to the datastore
     * @return The schema name
     * @throws SQLException Thrown if error occurs in determining the
     * schema name.
     **/
    public String getSchemaName(Connection conn)
    throws SQLException
    {
        return "";
    }

    /**
     * HSQL 1.7.0 does not support ALTER TABLE to define a primary key
     * 
     * Returns the appropriate SQL to add a primary key to its table.
     * It should return something like:
     * <p>
     * <blockquote><pre>
     * ALTER TABLE FOO ADD CONSTRAINT FOO_PK PRIMARY KEY (BAR)
     * </pre></blockquote>
     *
     * @param pkName    The name of the primary key to add.
     * @param pk        An object describing the primary key.
     * @return  null
     */
    public String getAddPrimaryKeyStatement(SQLIdentifier pkName, PrimaryKey pk)
    {
        return null;
    }

    /**
     * Returns the appropriate SQL to drop the given table.
     * It should return something like:
     * <p>
     * <blockquote><pre>
     * DROP TABLE FOO
     * </pre></blockquote>
     *
     * @param table     The table to drop.
     * @return  The text of the SQL statement.
     */
    public String getDropTableStatement(TableImpl table)
    {
        return "DROP TABLE " + table.toString();
    }

    /**
     * Return a new TableExpression appropriate to MySQL. MySQL does not
     * support the TableExprAsSubjoins so instead we use TableExprAsJoins.
     * @param qs The QueryStatement to add the expression to
     * @param table The table in the expression
     * @param rangeVar range variable to assign to the expression.
     * @return The expression.
     **/ 
    public TableExpression newTableExpression(QueryStatement qs, DatastoreContainerObject table, DatastoreIdentifier rangeVar)
    {
        return new TableExprAsJoins(qs, table, rangeVar);
    }

    /**
     * Whether we support deferred constraints in keys.
     * @return whether we support deferred constraints in keys.
     **/
    public boolean supportsDeferredConstraints()
    {
        return false;
    }

    /**
     * Whether we support auto-increment in keys.
     * @return whether we support auto-increment in keys.
     **/
    public boolean supportsAutoIncrementKeys()
    {
        return true;
    }
    
    /**
     * Whether this datastore supports the specified foreign key update action
     * @param action The update action
     * @return Whether it is supported
     */
    public boolean supportsForeignKeyUpdateAction(ForeignKey.FKAction action)
    {
        // HSQL before 1.7.* doesn't support foreign keys
        if (datastoreMajorVersion < 1 || (datastoreMajorVersion == 1 && datastoreMinorVersion < 7))
        {
            return false;
        }
        
        if (action.equals(ForeignKey.CASCADE_ACTION) ||
            action.equals(ForeignKey.DEFAULT_ACTION) ||
            action.equals(ForeignKey.NULL_ACTION))
        {
            return true;
        }
        return false;
    }
    
    /**
     * Whether this datastore supports the specified foreign key delete action
     * @param action The delete action
     * @return Whether it is supported
     */
    public boolean supportsForeignKeyDeleteAction(ForeignKey.FKAction action)
    {
        // HSQL before 1.7.* doesn't support foreign keys
        if (datastoreMajorVersion < 1 || (datastoreMajorVersion == 1 && datastoreMinorVersion < 7))
        {
            return false;
        }
        
        if (action.equals(ForeignKey.CASCADE_ACTION) ||
            action.equals(ForeignKey.DEFAULT_ACTION) ||
            action.equals(ForeignKey.NULL_ACTION))
        {
            return true;
        }
        return false;
    }

    /**
     * Accessor for the auto-increment sql statement for this datastore.
     * @return The statement for getting the latest auto-increment key
     **/
    public String getAutoIncrementStmt()
    {
        return "CALL IDENTITY()";
    }

    /**
     * Accessor for the auto-increment keyword for generating DDLs (CREATE TABLEs...).
     * @return The keyword for a column using auto-increment
     **/
    public String getAutoIncrementKeyword()
    {
        return "IDENTITY";
    }

    /**
     * Whether to allow Unique statements in the section of CREATE TABLE after the
     * column definitions.
     * @see org.jpox.store.rdbms.adapter.DatabaseAdapter#supportsUniqueConstraintsInEndCreateStatements()
     */
    public boolean supportsUniqueConstraintsInEndCreateStatements()
    {
        return true;
    }

    /**
     * Whether this datastore supports the use of CHECK in CREATE TABLE
     * statements (DDL). HSQL doesn't seem to support this.
     * @return whether we can use CHECK in CREATE TABLE.
     **/
    public boolean supportsCheckInCreateStatements()
    {
        return false;
    }

    /**
     * Whether this datastore supports the use of CHECK after the column
     * definitions in CREATE TABLE statements (DDL).
     * e.g.
     * CREATE TABLE XXX
     * (
     *   COL_A int,
     *   COL_B char(1),
     *   PRIMARY KEY (COL_A),
     *   CHECK (COL_B IN ('Y','N'))
     * )
     * @return whether we can use CHECK after the column definitions in CREATE TABLE.
     **/
    public boolean supportsCheckContraintsInEndCreateStatements()
    {
        // HSQL Introduced in v 1.7.2 the use of CHECK, but ONLY as statements
        // at the end of the CREATE TABLE statement. We currently don't support
        // anything there other than primary key.
        // This was introduced in version 1.7.2
        if (datastoreMajorVersion < 1 ||
            (datastoreMajorVersion == 1 && datastoreMinorVersion < 7) ||
            (datastoreMajorVersion == 1 && datastoreMinorVersion == 7 && datastoreRevisionVersion < 2))
        {
            return false;
        }
        else
        {
            return true;
        }
    }

    /**
     * Acessor for the MappingManager
     * @return the MappingManager
     */
    public MappingManager getMappingManager()
    {
        if (mappingManager == null)
        {
            mappingManager = new HSQLRDBMSMappingManager();
        }
        return mappingManager;
    }
}