/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others. All rights reserved.
This program and the accompanying materials are made available under
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - coding standards
    ...
**********************************************************************/
package org.jpox.store.rdbms.adapter;

import java.math.BigInteger;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.sql.Types;
import java.util.ArrayList;

import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.expression.BooleanExpression;
import org.jpox.store.expression.NumericExpression;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.StringExpression;
import org.jpox.store.expression.TableExprAsJoins;
import org.jpox.store.expression.TableExpression;
import org.jpox.store.mapping.MappingManager;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.rdbms.columninfo.ColumnInfo;
import org.jpox.store.rdbms.columninfo.MSSQLServerColumnInfo;
import org.jpox.store.rdbms.key.ForeignKey;
import org.jpox.store.rdbms.mapping.MSSQLRDBMSMappingManager;
import org.jpox.store.rdbms.table.TableImpl;
import org.jpox.store.rdbms.typeinfo.MSSQLTypeInfo;
import org.jpox.store.rdbms.typeinfo.TypeInfo;


/**
 * Provides methods for adapting SQL language elements to the Microsoft SQL
 * Server database.
 *
 * @see DatabaseAdapter
 * @version $Revision: 1.10 $ 
 */
class MSSQLServerAdapter extends DatabaseAdapter
{
    /**
     * Microsoft SQL Server 2000 uses reserved keywords for defining,
     * manipulating, and accessing databases. Reserved keywords are part of the
     * grammar of the Transact-SQL language used by SQL Server to parse and
     * understand Transact-SQL statements and batches. Although it is
     * syntactically possible to use SQL Server reserved keywords as identifiers
     * and object names in Transact-SQL scripts, this can be done only using
     * delimited identifiers.
     */
    private static final String MSSQL_RESERVED_WORDS =
        "ADD,ALL,ALTER,AND,ANY,AS," +
        "ASC,AUTHORIZATION,BACKUP,BEGIN,BETWEEN,BREAK," +
        "BROWSE,BULK,BY,CASCADE,CASE,CHECK," +
        "CHECKPOINT,CLOSE,CLUSTERED,COALESCE,COLLATE,COLUMN," +
        "COMMIT,COMPUTE,CONSTRAINT,CONTAINS,CONTAINSTABLE,CONTINUE," +
        "CONVERT,CREATE,CROSS,CURRENT,CURRENT_DATE,CURRENT_TIME," +
        "CURRENT_TIMESTAMP,CURRENT_USER,CURSOR,DBCC,DEALLOCATE,DECLARE," +
        "DEFAULT,DELETE,DENY,DESC,DISK,DISTINCT," +
        "DISTRIBUTED,DOUBLE,DROP,DUMMY,DUMP,ELSE," +
        "END,ERRLVL,ESCAPE,EXCEPT,EXEC,EXECUTE," +
        "EXISTS,EXIT,FETCH,FILE,FILLFACTOR,FOR," +
        "FOREIGN,FREETEXT,FREETEXTTABLE,FROM,FULL,FUNCTION," +
        "GOTO,GRANT,GROUP,HAVING,HOLDLOCK,IDENTITY," +
        "IDENTITY_INSERT,IDENTITYCOL,IF,IN,INDEX,INNER," +
        "INSERT,INTERSECT,INTO,IS,JOIN,KEY," +
        "KILL,LEFT,LIKE,LINENO,LOAD,NATIONAL," +
        "NOCHECK,NONCLUSTERED,NOT,NULL,NULLIF,OF," +
        "OFF,OFFSETS,ON,OPEN,OPENDATASOURCE,OPENQUERY," +
        "OPENROWSET,OPENXML,OPTION,OR,ORDER,OUTER," +
        "OVER,PERCENT,PLAN,PRECISION,PRIMARY,PRINT," +
        "PROC,PROCEDURE,PUBLIC,RAISERROR,READ,READTEXT," +
        "RECONFIGURE,REFERENCES,REPLICATION,RESTORE,RESTRICT,RETURN," +
        "REVOKE,RIGHT,ROLLBACK,ROWCOUNT,ROWGUIDCOL,RULE," +
        "SAVE,SCHEMA,SELECT,SESSION_USER,SET,SETUSER," +
        "SHUTDOWN,SOME,STATISTICS,SYSTEM_USER,TABLE,TEXTSIZE," +
        "THEN,TO,TOP,TRAN,DATABASE,TRANSACTION,TRIGGER," +
        "TRUNCATE,TSEQUAL,UNION,UNIQUE,UPDATE,UPDATETEXT," +
        "USE,USER,VALUES,VARYING,VIEW,WAITFOR," +
        "WHEN,WHERE,WHILE,WITH,WRITETEXT";
    
    /**
     * Constructs a Microsoft SQL Server adapter based on the given JDBC
     * metadata.
     *
     * @param   metadata    the database metadata.
     */
    public MSSQLServerAdapter(DatabaseMetaData metadata)
    {
        super(metadata);
        
        reservedKeywords.addAll(parseKeywordList(MSSQL_RESERVED_WORDS));
    }

    public String getVendorID()
    {
        return "sqlserver";
    }

    public String getSchemaName(Connection conn) throws SQLException
    {
        /*
         * As of version 7 there was no equivalent to the concept of "schema"
         * in SQL Server.  For DatabaseMetaData functions that include
         * SCHEMA_NAME drivers usually return the user name that owns the table.
         *
         * So the default ProbeTable method for determining the current schema
         * just ends up returning the current user.  If we then use that name in
         * performing metadata queries our results may get filtered down to just
         * objects owned by that user.  So instead we report the schema name as
         * null which should cause those queries to return everything.
         */
        return null;
    }

    /**
     * The function to creates a unique value of type uniqueidentifier.
     * @return The function. e.g. "SELECT NEWID()"
     **/
    public String getSelectNewUUIDStmt()
    {
        return "SELECT NEWID()";
    }

    /**
     * The function to creates a unique value of type uniqueidentifier.
     * @return The function. e.g. "NEWID()"
     **/
    public String getNewUUIDFunction()
    {
        return "NEWID()";
    }
    
    public boolean supportsBooleanComparison()
    {
        return false;
    }

    public boolean supportsDeferredConstraints()
    {
        return false;
    }

    /**
     * Whether this datastore supports the specified foreign key update action
     * @param action The update action
     * @return Whether it is supported
     */
    public boolean supportsForeignKeyUpdateAction(ForeignKey.FKAction action)
    {
        if( action == ForeignKey.CASCADE_ACTION )
        {
            return true;
        }
        return false;
    }

    /**
     * Whether this datastore supports the specified foreign key delete action
     * @param action The delete action
     * @return Whether it is supported
     */
    public boolean supportsForeignKeyDeleteAction(ForeignKey.FKAction action)
    {
        if( action == ForeignKey.CASCADE_ACTION )
        {
            return true;
        }
        return false;
    }
    
    public ColumnInfo newColumnInfo(ResultSet rs)
    {
        return new MSSQLServerColumnInfo(rs);
    }

    public TableExpression newTableExpression(QueryStatement qs, DatastoreContainerObject table, DatastoreIdentifier rangeVar)
    {
        return new TableExprAsJoins(qs, table, rangeVar);
    }

    public TypeInfo newTypeInfo(ResultSet rs)
    {
        TypeInfo ti = new MSSQLTypeInfo(rs);

        /*
         * Discard the tinyint type because it doesn't support negative values.
         */
        if (ti.typeName.toLowerCase().startsWith("tinyint"))
        {
            return null;
        }

        return ti;
    }
    
    protected void createTypeInfo(DatabaseMetaData metadata) throws SQLException
    {
        super.createTypeInfo(metadata);

        /*
         * If the MSSQL JDBC driver does not provide info for the UNIQUEIDENTIFER type,
         * spoof some and add it.
         */
        Integer key = new Integer(MSSQLTypeInfo.UNIQUEIDENTIFIER);
        if (typesByTypeNumber.get(key) == null)
        {
            TypeInfo ti = new TypeInfo("UNIQUEIDENTIFIER",
                                       (short)Types.CHAR,
                                        36,
                                        "'",
                                        "'",
                                        null,
                                        1,
                                        false,
                                        (short)2,
                                        false,
                                        false,
                                        false,
                                        "UNIQUEIDENTIFIER",
                                        (short)0,
                                        (short)0,
                                        10);

            typesByTypeNumber.put(key, ti);
        }
    }    

    public String getDropTableStatement(TableImpl table)
    {
        return "DROP TABLE " + table.toString();
    }

    public NumericExpression lengthMethod(StringExpression str)
    {
        ArrayList args = new ArrayList();
        args.add(str);

        return new NumericExpression("LEN", args);
    }

    public StringExpression substringMethod(StringExpression str,
                                               NumericExpression begin)
    {
        ArrayList args = new ArrayList();
        args.add(str);
        args.add(begin.add(getMapping(BigInteger.class).newLiteral(str.getQueryStatement(), BigInteger.ONE)));
        args.add(lengthMethod(str).sub(begin));

        return new StringExpression("SUBSTRING", args);
    }

    public StringExpression substringMethod(StringExpression str,
                                               NumericExpression begin,
                                               NumericExpression end)
    {
        ArrayList args = new ArrayList();
        args.add(str);
        args.add(begin.add(getMapping(BigInteger.class).newLiteral(str.getQueryStatement(), BigInteger.ONE)));
        args.add(end.sub(begin));

        return new StringExpression("SUBSTRING", args);
    }
    
    /**
     * Returns the appropriate SQL expression for the JDOQL
     * String.trim(str) method.
     * It should return something like:
     * <p>
     * <pre>
     * TRIM(str)
     * </pre>
     *
     * @param str The first argument to the trim() method.
     * @return  The text of the SQL expression.
     */
    public StringExpression trimMethod(StringExpression str)
    {
        ArrayList args = new ArrayList();
        args.add(str);

        return new StringExpression("RTRIM", args);
    }
    
    /**
     * Method to handle the starts with operation.
     * @param str The string to compare against 
     * @return The expression.
     **/
    public BooleanExpression startsWithMethod(ScalarExpression source, ScalarExpression str)
    {
        
        ScalarExpression integerLiteral = getMapping(BigInteger.class).newLiteral(source.getQueryStatement(), BigInteger.ONE);
        ArrayList args = new ArrayList();
        args.add(str);
        args.add(source);
        return new BooleanExpression(new StringExpression("CHARINDEX", args),ScalarExpression.OP_EQ,integerLiteral);
    }    

	public long getAdapterTime(Timestamp time)
	{
		long timestamp = getTime(time.getTime(),time.getNanos());
		int ms = getMiliseconds(time.getNanos());
		if (time.getNanos() == 0)
        {
			return timestamp+ms;
        }

		// MSSQL driver is losing date milliseconds if time.getTime() represents a number greater than 9 digits
		boolean losing = (String.valueOf(time.getTime())).length()>9;
		if (losing)
		{
			return timestamp+ms-1;
		}
		return timestamp+ms+1;
	} 

	/**
	 * Whether we support auto-increment in keys.
	 * @return whether we support auto-increment in keys.
	 **/
	public boolean supportsAutoIncrementKeys()
	{
		return true;
	}
	
	/**
	 * Accessor for the auto-increment sql statement for this datastore.
	 * @return The statement for getting the latest auto-increment key
	 **/
	public String getAutoIncrementStmt()
	{
		return "SELECT @@IDENTITY";
	}

	/**
	 * Accessor for the auto-increment keyword for generating DDLs (CREATE TABLEs...).
	 * @return The keyword for a column using auto-increment
	 **/
	public String getAutoIncrementKeyword()
	{
		return "IDENTITY";
	}

    /**
     * An operator in a string expression that concatenates two or more
     * character or binary strings, columns, or a combination of strings and
     * column names into one expression (a string operator).
     * 
     * @return the operator SQL String
     */
    public String getOperatorConcat()
    {
        return "+";
    }
    
    /**
     * Acessor for the MappingManager
     * @return the MappingManager
     */
    public MappingManager getMappingManager()
    {
        if (mappingManager == null)
        {
            mappingManager = new MSSQLRDBMSMappingManager();
        }
        return mappingManager;
    }
}