/**********************************************************************
Copyright (c) 2002 Kelly Grizzle (TJDO) and others. All rights reserved.
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2002 Mike Martin (TJDO)
2003 Erik Bengtson - Oracle 9 support the ISO 99 outer join syntax
                    so I have changed newQueryStatement to use
                    QueryStatement instead of OracleQueryStatement
                    when Oracle version is 9
2003 Andy Jefferson - coding standards
2004 Andy Jefferson - added sequence methods
2004 Andy Jefferson - updated Oracle version checks to use majorVersion
    ...
**********************************************************************/
package org.jpox.store.rdbms.adapter;

import java.math.BigInteger;
import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.sql.Types;
import java.util.ArrayList;

import javax.jdo.JDOFatalDataStoreException;

import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.expression.NumericExpression;
import org.jpox.store.expression.StringExpression;
import org.jpox.store.expression.TableExprAsJoins;
import org.jpox.store.expression.TableExprAsSubquery;
import org.jpox.store.expression.TableExpression;
import org.jpox.store.mapping.MappingManager;
import org.jpox.store.mapping.OIDMapping;
import org.jpox.store.mapping.OracleStringMapping;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.rdbms.key.ForeignKey;
import org.jpox.store.rdbms.mapping.OracleRDBMSMappingManager;
import org.jpox.store.rdbms.query.Oracle99QueryStatement;
import org.jpox.store.rdbms.query.OracleQueryStatement;
import org.jpox.store.rdbms.table.TableImpl;
import org.jpox.store.rdbms.typeinfo.TypeInfo;

/**
 * Provides methods for adapting SQL language elements to the Oracle database.
 * <P>
 * There may be issues about the DatabaseMajorVersion. Oracle 10g returns 1010 when
 * it should be 10. Maybe other versions of Oracle return other such silliness ?
 * Anyway, we make the assumption that we can use the major version number to
 * detect the feature set.
 * 
 * @see DatabaseAdapter
 * @version $Revision: 1.14 $
 */
class OracleAdapter extends DatabaseAdapter
{
    /**
     * A string containing the list of Oracle keywords 
     * This list is normally obtained dynamically from the driver using
     * DatabaseMetaData.getSQLKeywords()
     * 
     * Based on database Oracle8
     */
    public static final String ORACLE_8_RESERVED_WORDS =
        "ACCESS,AUDIT,CLUSTER,COMMENT,COMPRESS,EXCLUSIVE,FILE,IDENTIFIED," +
        "INCREMENT,INDEX,INITIAL,LOCK,LONG,MAXEXTENTS,MINUS,MLSLABEL,MODE," +
        "MODIFY,NOAUDIT,NOCOMPRESS,NOWAIT,NUMBER,OFFLINE,ONLINE,PCTFREE,RAW," +
        "RENAME,RESOURCE,ROWID,ROWNUM,SHARE,SUCCESSFUL,SYNONYM,SYSDATE,UID," +
        "VALIDATE,VARCHAR2,VALIDATE,VARCHAR2";

    /**
     * A string containing the list of Oracle keywords 
     * This list is normally obtained dynamically from the driver using
     * DatabaseMetaData.getSQLKeywords()
     * 
     * Based on database Oracle9i
     */
    public static final String ORACLE_9_RESERVED_WORDS =
        "ACCESS,CHAR,DEFAULT,ADD,CHECK,DELETE,ALL,CLUSTER,DESC,ALTER,COLUMN," +
        "DISTINCT,AND,COMMENT,DROP,ANY,COMPRESS,ELSE,AS,CONNECT,EXCLUSIVE,ASC," +
        "CREATE,EXISTS,AUDIT,CURRENT,FILE,BETWEEN,DATE,FLOAT,BY,DECIMAL,FOR,FROM," +
        "NOT,SHARE,GRANT,NOWAIT,SIZE,GROUP,NULL,SMALLINT,HAVING,NUMBER,START," +
        "IDENTIFIED,OF,SUCCESSFUL,IMMEDIATE,OFFLINE,SYNONYM,IN,ON,SYSDATE," +
        "INCREMENT,ONLINE,TABLE,INDEX,OPTION,THEN,INITIAL,OR,TO,INSERT,ORDER," +
        "TRIGGER,INTEGER,PCTFREE,UID,INTERSECT,PRIOR,UNION,INTO,PRIVILEGES,UNIQUE," +
        "IS,PUBLIC,UPDATE,LEVEL,RAW,USER,LIKE,RENAME,VALIDATE,LOCK,RESOURCE,VALUES," +
        "LONG,REVOKE,VARCHAR,MAXEXTENTS,ROW,VARCHAR2,MINUS,ROWID,VIEW,MLSLABEL," +
        "ROWNUM,WHENEVER,MODE,ROWS,WHERE,MODIFY,SELECT,WITH,NOAUDIT,SESSION," +
        "NOCOMPRESS,SET";
    
    /**
     * A string containing the list of Oracle keywords 
     * This list is normally obtained dynamically from the driver using
     * DatabaseMetaData.getSQLKeywords()
     * 
     * Based on database Oracle10g
     */
    public static final String ORACLE_10_RESERVED_WORDS =
        "ACCESS,ADD,ALL,ALTER,AND,ANY,AS,ASC,AUDIT,BETWEEN,BY,CHAR,CHECK,CLUSTER," +
        "COLUMN,COMMENT,COMPRESS,CONNECT,CREATE,CURRENT,DATE,DECIMAL,DEFAULT,DELETE," +
        "DESC,DISTINCT,DROP,ELSE,EXCLUSIVE,EXISTS,FILE,FLOAT,FOR,FROM,GRANT,GROUP," +
        "HAVING,IDENTIFIED,IMMEDIATE,IN,INCREMENT,INDEX,INITIAL,INSERT,INTEGER," +
        "INTERSECT,INTO,IS,LEVEL,LIKE,LOCK,LONG,MAXEXTENTS,MINUS,MLSLABEL,MODE," +
        "MODIFY,NOAUDIT,NOCOMPRESS,NOT,NOWAIT,NULL,NUMBER,OF,OFFLINE,ON,ONLINE," +
        "OPTION,OR,ORDER,PCTFREE,PRIOR,PRIVILEGES,PUBLIC,RAW,RENAME,RESOURCE," +
        "REVOKE,ROW,ROWID,ROWNUM,ROWS,SELECT,SESSION,SET,SHARE,SIZE,SMALLINT," +
        "START,SUCCESSFUL,SYNONYM,SYSDATE,TABLE,THEN,TO,TRIGGER,UID,UNION," +
        "UNIQUE,UPDATE,USER,VALIDATE,VALUES,VARCHAR,VARCHAR2,VIEW,WHENEVER," +
        "WHERE,WITH";

    /**
     * Constructs an Oracle adapter based on the given JDBC metadata.
     * @param metadata the database metadata.
     */
    public OracleAdapter(DatabaseMetaData metadata)
    {
        super(metadata);
        
        if(datastoreMajorVersion<=8)
        {
            reservedKeywords.addAll(parseKeywordList(ORACLE_8_RESERVED_WORDS));
        }
        else if(datastoreMajorVersion==9)
        {
            reservedKeywords.addAll(parseKeywordList(ORACLE_9_RESERVED_WORDS));
        }        
        else if(datastoreMajorVersion>=10)
        {
            reservedKeywords.addAll(parseKeywordList(ORACLE_10_RESERVED_WORDS));
        }
    }

    /**
     * Accessor for the vendor id
     * @return The Oracle vendor id
     */
    public String getVendorID()
    {
        return "oracle";
    }

    protected void createTypeInfo(DatabaseMetaData metadata) throws SQLException
    {
        super.createTypeInfo(metadata);

        /*
         * If the Oracle JDBC driver does not provide info for the CLOB type,
         * spoof some and add it.
         */
        Integer clobKey = new Integer(Types.CLOB);
        if (typesByTypeNumber.get(clobKey) == null)
        {
            TypeInfo ti = new TypeInfo("CLOB",
                                       (short)Types.CLOB,
                                        1073741823,
                                        "'",
                                        "'",
                                        null,
                                        1,
                                        true,
                                        (short)0,
                                        false,
                                        false,
                                        false,
                                        "CLOB",
                                        (short)0,
                                        (short)0,
                                        10);

            typesByTypeNumber.put(clobKey, ti);
        }
    }

    /**
     * Accessor for whether the specified class is an embedded type
     * @param c The class
     * @return Whether the class is an embedded type
     */
    public boolean isEmbeddedType(Class c)
    {
        if (c.equals(String.class))
        {
            return !OIDMapping.class.isAssignableFrom(OracleStringMapping.class);
        }
        else
        {
            return super.isEmbeddedType(c);
        }
    }

    /**
     * Some databases, Oracle, treats an empty string (0 length) equals null
     * @return whether the database treats an empty string as null
     */
    public boolean isNullEqualsEmptyStrings()
    {
        return true;
    }    

    /**
     * Some databases, Oracle, treats an empty string (0 length) equals null
     * @return returns a surrogate to replace the empty string in the database
     * otherwise it would be treated as null
     */
    public String getSurrogateForEmptyStrings()
    {
        return "\u0001";
    }
    
    public String getSchemaName(Connection conn) throws SQLException
    {
        Statement stmt = conn.createStatement();
        
        try
        {
            String stmtText = "SELECT SYS_CONTEXT('USERENV', 'CURRENT_SCHEMA') FROM DUAL";
            ResultSet rs = stmt.executeQuery(stmtText);

            try
            {
                if (!rs.next())
                {
                    throw new JDOFatalDataStoreException("No result returned from " + stmtText);
                }

                return rs.getString(1);
            }
            finally
            {
                rs.close();
            }
        }
        finally
        {
            stmt.close();
        }
    }

    /**
     * Whether this datastore supports locking using SELECT ... FOR UPDATE.
     * @return whether we support locking using SELECT ... FOR UPDATE.
     **/
    public boolean supportsLockWithSelectForUpdate()
    {
        return true;
    }

    /**
     * Whether this datastore supports the specified foreign key delete action
     * @param action The delete action
     * @return Whether it is supported
     */
    public boolean supportsForeignKeyDeleteAction(ForeignKey.FKAction action)
    {
        if( action == ForeignKey.CASCADE_ACTION )
        {
            return true;
        }
        if( action == ForeignKey.NULL_ACTION )
        {
            return true;
        }
        return false;
    }
    
    /**
     * Whether this datastore supports the specified foreign key update action
     * @param action The update action
     * @return Whether it is supported
     */
    public boolean supportsForeignKeyUpdateAction(ForeignKey.FKAction action)
    {
        return false;
    }    
    
    public boolean supportsBooleanComparison()
    {
        return false;
    }

    /**
     * Provide the existing indexes in the database for the table
     * @param conn the JDBC connection
     * @param dmd the DatabaseMetaData
     * @param catalog the catalog name
     * @param schema the schema name
     * @param table the table name
     * @return a ResultSet with the format @see DatabaseMetaData#getIndexInfo(java.lang.String, java.lang.String, java.lang.String, boolean, boolean)
     * @throws SQLException
     */
    public ResultSet getExistingIndexes(Connection conn, DatabaseMetaData dmd, String catalog, String schema, String table) throws SQLException
    {
        String GET_INDEXES_STMT =
            "select null as table_cat, "+
            "owner as table_schem, "+
            "table_name, "+
            "0 as NON_UNIQUE, "+
            "null as index_qualifier, "+
            "null as index_name, 0 as type, "+
            "0 as ordinal_position, null as column_name, "+
            "null as asc_or_desc, "+
            "num_rows as cardinality, "+
            "blocks as pages, "+
            "null as filter_condition "+
            "from all_tables "+
            "where table_name = ? "+
            "and owner = ? "+
            "union "+
            "select null as table_cat, "+
            "i.owner as table_schem, "+
            "i.table_name, "+
            "decode (i.uniqueness, 'UNIQUE', 0, 1), "+
            "null as index_qualifier, "+
            "i.index_name, "+
            "1 as type, "+
            "c.column_position as ordinal_position, "+
            "c.column_name, "+
            "null as asc_or_desc, "+
            "i.distinct_keys as cardinality, "+
            "i.leaf_blocks as pages, "+
            "null as filter_condition "+
            "from all_indexes i, all_ind_columns c "+
            "where i.table_name = ? "+
            "and i.owner = ? "+
            "and i.index_name = c.index_name "+
            "and i.table_owner = c.table_owner "+
            "and i.table_name = c.table_name "+
            "and i.owner = c.index_owner "+
            "order by non_unique, type, index_name, ordinal_position";

        if( datastoreMajorVersion==10)
        {
            //THIS IS A FIX to the JDBC driver. Already noticed in previous versions, but this fix was only tested with Oracle 10g, and all drivers versions
            PreparedStatement stmt = conn.prepareStatement(GET_INDEXES_STMT);
	        stmt.setString(1,table);
	        stmt.setString(2,schema);
	        stmt.setString(3,table);
	        stmt.setString(4,schema);
	        return stmt.executeQuery();
        }
        //the above was used and tested with Oracle 10 to fix a driver bug when the table identifier is a reserved word
        return super.getExistingIndexes(conn,dmd,catalog,schema,table);
    }
    
    /**
     * Method to return the column name to use when handling ranges via
     * a rownumber on the select.
     * @return The row number column.
     */
    /*public String getRangeByRowNumberColumn()
    {
        // This is commented out since it won't work for ranges
        return "ROWNUM";
    } */

   
    public QueryStatement newQueryStatement(DatastoreContainerObject table)
    {
        // Oracle 9 or upper supports the ISO 99 outer join syntax
		if (datastoreMajorVersion >= 9)
        {
            return new Oracle99QueryStatement(table);
        }
        else
        {
            return new OracleQueryStatement(table);
        }
    }

    public QueryStatement newQueryStatement(DatastoreContainerObject table, DatastoreIdentifier rangeVar)
    {
		// Oracle 9 or upper supports the ISO 99 outer join syntax
        if (datastoreMajorVersion >= 9)
        {
            return new Oracle99QueryStatement(table, rangeVar);
        }
        else
        {
            return new OracleQueryStatement(table, rangeVar);
        }
    }

    public TableExpression newTableExpression(QueryStatement qs, DatastoreContainerObject table, DatastoreIdentifier rangeVar)
    {
		// Oracle 9 or upper supports the ISO 99 outer join syntax
        if (datastoreMajorVersion >= 9)
        {
        	return new TableExprAsJoins(qs, table, rangeVar);
        }
        else
        {
        	return new TableExprAsSubquery(qs, table, rangeVar);
        }
    }

    public String getDropTableStatement(TableImpl table)
    {
        return "DROP TABLE " + table.toString() + " CASCADE CONSTRAINTS";
    }

    public NumericExpression lengthMethod(StringExpression str)
    {
        ArrayList args = new ArrayList();
        args.add(str);

        return new NumericExpression("LENGTH", args);
    }

    public StringExpression substringMethod(StringExpression str,
                                               NumericExpression begin)
    {
        ArrayList args = new ArrayList();
        args.add(str);
        args.add(begin.add(getMapping(BigInteger.class).newLiteral(str.getQueryStatement(), BigInteger.ONE)));

        return new StringExpression("SUBSTR", args);
    }

    public StringExpression substringMethod(StringExpression str,
                                               NumericExpression begin,
                                               NumericExpression end)
    {
        ArrayList args = new ArrayList();
        args.add(str);
        args.add(begin.add(getMapping(BigInteger.class).newLiteral(str.getQueryStatement(), BigInteger.ONE)));
        args.add(end.sub(begin));

        return new StringExpression("SUBSTR", args);
    }

    // ---------------------------- Sequence Support ---------------------------
 
    /**
     * Whether we support sequences.
     * @return whether we support sequences.
     **/
    public boolean supportsSequences()
    {
        return true;
    }

    /**
     * Accessor for the sequence statement to create the sequence.
     * @param sequence_name Name of the sequence 
     * @param min Minimum value for the sequence
     * @param max Maximum value for the sequence
     * @param start Start value for the sequence
     * @param increment Increment value for the sequence
     * @param cache_size Cache size for the sequence
     * @return The statement for getting the next id from the sequence
     **/
    public String getSequenceCreateStmt(String sequence_name,
                                        String min,String max,
                                        String start,String increment,
                                        String cache_size)
    {
        if (sequence_name == null)
        {
            throw new UnsupportedOperationException("Adapter.SequenceNameNullNotSupported");
        }

        StringBuffer stmt = new StringBuffer("CREATE SEQUENCE ");
        stmt.append(sequence_name);
        if (min != null && min.length() > 0)
        {
            stmt.append(" MINVALUE " + min);
        }
        if (max != null && max.length() > 0)
        {
            stmt.append(" MAXVALUE " + max);
        }
        if (start != null && start.length() > 0)
        {
            stmt.append(" START WITH " + start);
        }
        if (increment != null && increment.length() > 0)
        {
            stmt.append(" INCREMENT BY " + increment);
        }
        if (cache_size != null && cache_size.length() > 0)
        {
            stmt.append(" CACHE " + cache_size);
        }
        else
        {
            stmt.append(" NOCACHE");
        }

        return stmt.toString();
    }

    /**
     * Accessor for the statement for getting the next id from the sequence
     * for this datastore.
     * @param sequence_name Name of the sequence 
     * @return The statement for getting the next id for the sequence
     **/
    public String getSequenceNextStmt(String sequence_name)
    {
        if (sequence_name == null)
        {
            throw new UnsupportedOperationException("Adapter.SequenceNameNullNotSupported");
        }
        StringBuffer stmt=new StringBuffer("SELECT ");
        stmt.append(sequence_name);
        stmt.append(".NEXTVAL from dual ");

        return stmt.toString();
    }
    
    /**
     * Acessor for the MappingManager
     * @return the MappingManager
     */
    public MappingManager getMappingManager()
    {
        if (mappingManager == null)
        {
            mappingManager = new OracleRDBMSMappingManager();
        }
        return mappingManager;
    }    
}