/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - coding standards
2004 Andy Jefferson - added sequence methods
2004 Andy Jefferson - patch from Rey Amarego for "escape" method
2004 Andy Jefferson - update to cater for bitReallyBoolean (Tibor Kiss)
    ...
**********************************************************************/
package org.jpox.store.rdbms.adapter;

import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.Types;
import java.util.Hashtable;

import javax.jdo.JDODataStoreException;

import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.expression.TableExprAsJoins;
import org.jpox.store.expression.TableExpression;
import org.jpox.store.mapping.MappingManager;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.columninfo.ColumnInfo;
import org.jpox.store.rdbms.columninfo.PostgreSQLColumnInfo;
import org.jpox.store.rdbms.foreignkey.ForeignKeyInfo;
import org.jpox.store.rdbms.foreignkey.PostgreSQLForeignKeyInfo;
import org.jpox.store.rdbms.key.PrimaryKey;
import org.jpox.store.rdbms.mapping.PostgresqlRDBMSMappingManager;
import org.jpox.store.rdbms.sqlidentifier.SQLIdentifier;
import org.jpox.store.rdbms.table.Table;
import org.jpox.store.rdbms.table.TableImpl;
import org.jpox.store.rdbms.typeinfo.PostgreSQLTypeInfo;
import org.jpox.store.rdbms.typeinfo.TypeInfo;
import org.jpox.util.JPOXLogger;

/**
 * Provides methods for adapting SQL language elements to the PostgreSQL
 * database.
 *
 * @version $Revision: 1.11 $ 
 */
class PostgreSQLAdapter extends DatabaseAdapter
{
    /** List of Postgresql keywords that aren't in SQL92, SQL99 */
    public static final String POSTGRESQL_RESERVED_WORDS =
        "ALL,ANALYSE,ANALYZE,DO,FREEZE,ILIKE,ISNULL,OFFSET,PLACING,VERBOSE";
        
    protected Hashtable psqlTypes;

    public PostgreSQLAdapter(DatabaseMetaData metadata)
    {
        super(metadata);

        if (datastoreMajorVersion < 7)
        {
            throw new JDODataStoreException("PostgreSQL version is " + datastoreMajorVersion + '.' + datastoreMinorVersion + ", 7.0 or later required");
        }
        else if (datastoreMajorVersion == 7)
        {
            if (datastoreMinorVersion <= 2)
            {
                // The driver correctly reports the max table name length as 32.
                // However, constraint names are apparently limited to 31.  In
                // this case we get better looking names by simply treating them
                // all as limited to 31.
                --maxTableNameLength;
                --maxConstraintNameLength;
                --maxIndexNameLength;
            }
        }
        
        reservedKeywords.addAll(parseKeywordList(POSTGRESQL_RESERVED_WORDS));
    }

    /**
     * Accessor for the vendor id.
     * @return The vendor id.
     **/
    public String getVendorID()
    {
        return "postgresql";
    }

    /**
     * Method to initialise the PostgreSQL types that we will use for the
     * various JDBC types.
     **/
    protected void initialisePostgresqlTypes()
    {
        psqlTypes = new Hashtable();

        psqlTypes.put("" + Types.BIT,"bool");
        psqlTypes.put("" + Types.TIMESTAMP,"timestamptz");
        psqlTypes.put("" + Types.BIGINT,"int8");
        psqlTypes.put("" + Types.CHAR,"char");
        psqlTypes.put("" + Types.DATE,"date");
        psqlTypes.put("" + Types.DOUBLE,"float8");
        psqlTypes.put("" + Types.INTEGER,"int4");
        psqlTypes.put("" + Types.LONGVARCHAR,"text");
        psqlTypes.put("" + Types.NUMERIC,"numeric");
        psqlTypes.put("" + Types.REAL,"float4");
        psqlTypes.put("" + Types.SMALLINT,"int2");
        psqlTypes.put("" + Types.TIME,"time");
        psqlTypes.put("" + Types.VARCHAR,"varchar");
        psqlTypes.put("" + Types.OTHER,"***TOTALRUBBISH***");

        // PostgreSQL (7.3, 7.4) doesn't provide a type to map to these JDBC types
//      psqlTypes.put("" + Types.FLOAT,"");
//      psqlTypes.put("" + Types.DECIMAL,"");
//      psqlTypes.put("" + Types.BLOB,"");
//      psqlTypes.put("" + Types.CLOB,"");
//      psqlTypes.put("" + Types.BOOLEAN,"");
    }

    /**
     * Accessor for whether a JDBC Types.BIT is really mapped in the datastore
     * to a boolean field. Postgresql treats Types.BIT as boolean
     * @return Whether it is really a boolean
     */
    public boolean isBitReallyBoolean()
    {
        return true;
    }

    public TableExpression newTableExpression(QueryStatement qs, DatastoreContainerObject table, DatastoreIdentifier rangeVar)
    {
        return new TableExprAsJoins(qs, table, rangeVar);
    }

    public TypeInfo newTypeInfo(ResultSet rs)
    {
        TypeInfo ti = new PostgreSQLTypeInfo(rs);

        /*
         * Since PostgreSQL supports many user defined data types and uses
         * many type aliases the default methods have trouble finding the
         * right associations between JDBC and PostgreSQL data types.  We
         * filter the returned type info to be sure we use the appropriate
         * base PostgreSQL types for the important JDBC types.
         */
        if (psqlTypes == null)
        {
            initialisePostgresqlTypes();
        }
        Object obj=psqlTypes.get("" + ti.dataType);
        if (obj != null)
        {
            String  psql_type_name=(String)obj;

            // We don't support this JDBC type using *this* PostgreSQL type
            if (!ti.typeName.equalsIgnoreCase(psql_type_name))
            {
                JPOXLogger.RDBMS_SCHEMA.info(LOCALISER.msg("Adapter.JDBCTypeDiscarded", ti.typeName, TypeInfo.getJDBCTypeName(ti.dataType)));
                return null;
            }
        }

        return ti;
    }

    public ColumnInfo newColumnInfo(ResultSet rs)
    {
        return new PostgreSQLColumnInfo(rs);
    }

    public ForeignKeyInfo newForeignKeyInfo(ResultSet rs)
    {
        return new PostgreSQLForeignKeyInfo(rs);
    }

    /**
     * Accessor for the SQL statement to add a column to a table.
     * @param table The table
     * @param col The column
     * @return The SQL necessary to add the column
     */
    public String getAddColumnStatement(Table table, Column col)
    {
        return "ALTER TABLE " + table.toString() + " ADD COLUMN " + col.getSQLDefinition();
    }

    /**
     * Method to return the SQL to append to a SELECT statement to handle
     * restriction of ranges using the LIMUT keyword.
     * @param offset The offset to return from
     * @param count The number of items to return
     * @return The SQL to append to allow for ranges using LIMIT.
     */
    public String getRangeByLimitClause(long offset, long count)
    {
        String str = "";
        if (count > 0)
        {
            str += "LIMIT " + count;
        }
        if (offset >= 0)
        {
            str += " OFFSET " + offset;
        }
        return str;
    }

    /**
     * Whether this datastore supports locking using SELECT ... FOR UPDATE.
     * @return whether we support locking using SELECT ... FOR UPDATE.
     **/
    public boolean supportsLockWithSelectForUpdate()
    {
        return true;
    }

    /**
     * PostgreSQL has supported the ALTER TABLE tbl-name DROP CONSTRAINT ...
     * since 7.2
     * @return Whether it supports the ALTER TABLE DROP CONSTRAINT syntax 
     **/
    public boolean supportsAlterTableDropConstraint()
    {
        if (datastoreMajorVersion < 7 ||
            (datastoreMajorVersion == 7 && datastoreMinorVersion < 2))
        {
            return false;
        }
        else
        {
            return true;
        }
    }

    /**
     * PostgreSQL needs any ORDER BY columns to be aliases.
     * @return true
     **/
    public boolean includeOrderByColumnsInSelectUsingAlias()
    {
        return true;
    }

    /**
     * PostgreSQL allows specification of PRIMARY KEY in the CREATE TABLE, so
     * we need nothing here.
     * 
     * @param pkName    The name of the primary key to add.
     * @param pk        An object describing the primary key.
     * @return  null
     */
    public String getAddPrimaryKeyStatement(SQLIdentifier pkName, PrimaryKey pk)
    {
        return null;
    }
    
    /**
     * Whether the datastore supports specification of the primary key in
     * CREATE TABLE statements.
     * @return Whetehr it allows "PRIMARY KEY ..."
     */
    public boolean supportsPrimaryKeyInCreateStatements()
    {
        return true;
    }

    /**
     * Accessor for the statement for dropping a table.
     * PostgreSQL has supported the DROP TABLE tbl-name CASCADE since 7.3.
     * @return The statement for dropping a table.
     **/
    public String getDropTableStatement(TableImpl table)
    {
        /* DROP TABLE t CASCADE is supported beginning in 7.3 */
        if (datastoreMajorVersion < 7 ||
            (datastoreMajorVersion == 7 && datastoreMinorVersion < 3))
        {
            return "DROP TABLE " + table.toString();
        }
        else
        {
            return "DROP TABLE " + table.toString() + " CASCADE";
        }
    }

    // ---------------------------- Sequence Support ---------------------------
 
    /**
     * Whether we support sequences.
     * @return whether we support sequences.
     **/
    public boolean supportsSequences()
    {
        return true;
    }

    /**
     * Accessor for the sequence statement to create the sequence.
     * @param sequence_name Name of the sequence 
     * @param min Minimum value for the sequence
     * @param max Maximum value for the sequence
     * @param start Start value for the sequence
     * @param increment Increment value for the sequence
     * @param cache_size Cache size for the sequence
     * @return The statement for getting the next id from the sequence
     **/
    public String getSequenceCreateStmt(String sequence_name,
                                        String min,String max,
                                        String start,String increment,
                                        String cache_size)
    {
        if (sequence_name == null)
        {
            throw new UnsupportedOperationException("Adapter.SequenceNameNullNotSupported");
        }

        StringBuffer stmt = new StringBuffer("CREATE SEQUENCE ");
        stmt.append(sequence_name);
        if (min != null && min.length() > 0)
        {
            stmt.append(" MINVALUE " + min);
        }
        if (max != null && max.length() > 0)
        {
            stmt.append(" MAXVALUE " + max);
        }
        if (start != null && start.length() > 0)
        {
            stmt.append(" START WITH " + start);
        }
        if (increment != null && increment.length() > 0)
        {
            stmt.append(" INCREMENT BY " + increment);
        }
        if (cache_size != null && cache_size.length() > 0)
        {
            stmt.append(" CACHE " + cache_size);
        }
        else
        {
            stmt.append(" CACHE 1");
        }

        return stmt.toString();
    }

    /**
     * Accessor for the statement for getting the next id from the sequence
     * for this datastore.
     * @param sequence_name Name of the sequence 
     * @return The statement for getting the next id for the sequence
     **/
    public String getSequenceNextStmt(String sequence_name)
    {
        if (sequence_name == null)
        {
            throw new UnsupportedOperationException("Adapter.SequenceNameNullNotSupported");
        }

        StringBuffer stmt=new StringBuffer("SELECT nextval('");
        stmt.append(sequence_name);
        stmt.append("')");

        return stmt.toString();
    }

    /**
     * The character for escaping patterns.
     * @return Escape character(s)
     **/
    public String getEscapePatternExpression()
    {
        return "ESCAPE '\\\\'";
    }
    
    /**
     * Acessor for the MappingManager
     * @return the MappingManager
     */
    public MappingManager getMappingManager()
    {
        if (mappingManager == null)
        {
            mappingManager = new PostgresqlRDBMSMappingManager();
        }
        return mappingManager;
    }

    /* TODO support for serial
     * The data types serial and bigserial are not true types, but merely a notational convenience for setting up unique identifier columns (similar to the AUTO_INCREMENT property supported by some other databases). In the current implementation, specifying 

		CREATE TABLE tablename (
		    colname SERIAL
		);
		is equivalent to specifying: 
		
		CREATE SEQUENCE tablename_colname_seq;
		CREATE TABLE tablename (
		    colname integer DEFAULT nextval('tablename_colname_seq') NOT NULL
		);
		
		To see what the value is in a serial column you can just use a regular select.
		
		If you want information on the sequence itself you can query it like:
		
		select * from foo_fooid_seq;  (using your example names)
		
		There is also the nextval function that can be called on a sequence.  More info is available in the PostgreSQL docs.
		*/
}