/**********************************************************************
Copyright (c) 2003 Andy Jefferson and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
    ...
**********************************************************************/
package org.jpox.store.rdbms.adapter;

import java.math.BigInteger;
import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.util.ArrayList;

import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.expression.NumericExpression;
import org.jpox.store.expression.StringExpression;
import org.jpox.store.expression.TableExprAsJoins;
import org.jpox.store.expression.TableExpression;
import org.jpox.store.mapping.MappingManager;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.rdbms.columninfo.ColumnInfo;
import org.jpox.store.rdbms.columninfo.SybaseColumnInfo;
import org.jpox.store.rdbms.mapping.SybaseRDBMSMappingManager;
import org.jpox.store.rdbms.table.TableImpl;
import org.jpox.store.rdbms.typeinfo.TypeInfo;

/**
 * Provides methods for adapting SQL language elements to the Sybase
 * database.
 *
 * @version $Revision: 1.7 $ 
 */
class SybaseAdapter extends DatabaseAdapter
{
    public SybaseAdapter(DatabaseMetaData metadata)
    {
        super(metadata);
    }

    public String getVendorID()
    {
        return "sybase";
    }

    /**
     * Accessor for the DROP TABLE statement for Sybase.
     * Sybase doesnt support CASCADE CONSTRAINTS so we just return a simple
     * DROP TABLE table-name
     * @return The DROP TABLE statement
     **/
	public String getDropTableStatement(TableImpl table)
	{
		return "DROP TABLE " + table.toString();
	}

    /**
     * Accessor for TableExpression for Sybase. Sybase (TransactSQL) accepts
     * the TableExprAsJoins style. The TableExprAsSubjoins doesnt work here.
     * @param qs The query statement
     * @param table The table in question
     * @param rangeVar range variable (?) 
     * @return The Table Expression 
     **/
    public TableExpression newTableExpression(QueryStatement qs, DatastoreContainerObject table, DatastoreIdentifier rangeVar)
    {
        return new TableExprAsJoins(qs, table, rangeVar);
    }

    /**
     * Accessor for whether Sybase supports deferred constraints.
     * Sybase doesnt support INITIALLY DEFERRED keywords so we return false.
     * @return Whether Sybase supports deferred constraints 
     **/
    public boolean supportsDeferredConstraints()
    {
        return false;
    }

    /**
     * Accessor for whether Sybase supports Boolean comparisons.
     * @return Whether we support Boolean comparisons
     **/
	public boolean supportsBooleanComparison()
	{
		return false;
	}

    /**
     * Whether this datastore supports locking using SELECT ... FOR UPDATE.
     * @return whether we support locking using SELECT ... FOR UPDATE.
     **/
    public boolean supportsLockWithSelectForUpdate()
    {
        return true;
    }

    public ColumnInfo newColumnInfo(ResultSet rs)
    {
        return new SybaseColumnInfo(rs);
    }

    public TypeInfo newTypeInfo(ResultSet rs)
    {
        TypeInfo ti = new TypeInfo(rs);

        /*
         * Discard the tinyint type because it doesn't support negative values.
         */
        if (ti.typeName.toLowerCase().startsWith("tinyint"))
        {
            return null;
        }

        return ti;
    }

    public StringExpression substringMethod(StringExpression str,
                                               NumericExpression begin)
    {
        ArrayList args = new ArrayList();
        args.add(str);
        args.add(begin.add(getMapping(Integer.class).newLiteral(str.getQueryStatement(), BigInteger.ONE)));
        args.add(lengthMethod(str).sub(begin));

        return new StringExpression("SUBSTRING", args);
    }

    public StringExpression substringMethod(StringExpression str,
                                               NumericExpression begin,
                                               NumericExpression end)
    {
        ArrayList args = new ArrayList();
        args.add(str);
        args.add(begin.add(getMapping(BigInteger.class).newLiteral(str.getQueryStatement(), BigInteger.ONE)));
        args.add(end.sub(begin));

        return new StringExpression("SUBSTRING", args);
    }
    
	/**
	 * Accessor for the auto-increment sql statement for this datastore.
	 * @return The statement for getting the latest auto-increment key
	 **/
	public String getAutoIncrementStmt()
	{
		return "SELECT @@IDENTITY";
	}

	/**
	 * Accessor for the auto-increment keyword for generating DDLs (CREATE TABLEs...).
	 * @return The keyword for a column using auto-increment
	 **/
	public String getAutoIncrementKeyword()
	{
		return "IDENTITY";
	}
	
	/**
	 * Whether we support auto-increment in keys.
	 * @return whether we support auto-increment in keys.
	 **/
	public boolean supportsAutoIncrementKeys()
	{
		return true;
	}

    /**
     * Whether we support auto-increment/identity keys with nullability spec.
     * @return whether we support auto-increment keys with nullability spec.
     **/
    public boolean supportsAutoIncrementKeysNullSpecification()
    {
        return false;
    }

    /**
     * Acessor for the MappingManager
     * @return the MappingManager
     */
    public MappingManager getMappingManager()
    {
        if (mappingManager == null)
        {
            mappingManager = new SybaseRDBMSMappingManager();
        }
        return mappingManager;
    }	
}
