/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - coding standards
    ...
**********************************************************************/
package org.jpox.store.rdbms.columninfo;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Types;

import javax.jdo.JDOFatalDataStoreException;

/**
 * Represents the metadata of a specific table column.  This class is
 * basically a data structure that makes accessing the JDBC column metadata
 * easier.  Each of the items returned by 
 * {@link java.sql.DatabaseMetaData#getColumns(String,String,String,String)}
 * is represented by a public field in this class.
 *
 * Subclasses of ColumnInfo can be created on a per-DBMS basis to supply missing
 * metadata or correct faulty metadata obtained from that DBMS's JDBC driver(s).
 *
 * @see org.jpox.store.rdbms.RDBMSManager#getColumnInfo
 * @see org.jpox.store.rdbms.adapter.DatabaseAdapter#newColumnInfo
 * @version $Revision: 1.1 $ 
 */
public class ColumnInfo
{
    /**
     * The table catalog, which may be <tt>null</tt>.
     */
    protected String tableCat;

    /**
     * The table schema, which may be <tt>null</tt>.
     */
    protected String tableSchem;

    /**
     * The table name.
     */
    protected String tableName;

    /**
     * The column name.
     */
    protected String columnName;

    /**
     * Indicates the JDBC (SQL) data type from {@link java.sql.Types}.
     */
    protected short dataType;

    /**
     * The local type name used by the data source.
     */
    protected String typeName;

    /**
     * Indicates the column size.  For char or date types, this is the maximum
     * number of characters; for numeric/ decimal types, this is the precision.
     */
    protected int columnSize;

    /**
     * Indicates the number of fractional digits.
     */
    protected int decimalDigits;

    /**
     * Indicates the radix, which is typically either 10 or 2.
     */
    protected int numPrecRadix;

    /**
     * Indicates whether the column can be NULL.
     */
    protected int nullable;

    /**
     * An explanatory comment on the column; may be <tt>null</tt>.
     */
    protected String remarks;

    /**
     * The default value for the column; may be <tt>null</tt>.
     */
    protected String columnDef;

    /**
     * Indicates the maximum number of bytes in the column (for char types
     * only).
     */
    protected int charOctetLength;

    /**
     * Indicates the index of the column in its table; the first column is 1,
     * the second column is 2, and so on.
     */
    protected int ordinalPosition;

    /**
     * Either "NO" indicating that the column definitely does not allow
     * <tt>null</tt> values, "YES" indicating that the column might allow
     * <tt>null</tt> values, or an empty string ("") indicating that nullability
     * is unknown.
     */
    protected String isNullable;

    private int hash = 0;

    /**
     * Accessor for no of fractional digits
     * @return The no of fractional digits
     **/
    public int getDecimalDigits()
    {
        return decimalDigits;
    }

    /**
     * Accessor for nullable attribute.
     * @return The nullability value
     **/
    public String getIsNullable()
    {
        return isNullable;
    }

    /**
     * Accessor for whether the column can be null
     * @return Whether the column can be null
     **/
    public int getNullable()
    {
        return nullable;
    }

    /**
     * Accessor for the column size.
     * @return The column size.
     **/
    public int getColumnSize()
    {
        return columnSize;
    }

    /**
     * Accessor for the JDBC data type.
     * @return The JDBC data type
     **/
    public short getDataType()
    {
        return dataType;
    }

    /**
     * Accessor for the radix
     * @return The radix
     **/
    public int getNumPrecRadix()
    {
        return numPrecRadix;
    }

    /**
     * Accessor for the maximum no of bytes in a column
     * @return The maximum number of bytes in a column
     **/
    public int getCharOctetLength()
    {
        return charOctetLength;
    }

    /**
     * Accessor for the ordinal position.
     * @return The ordinal position (index of column in table).
     **/
    public int getOrdinalPosition()
    {
        return ordinalPosition;
    }

    /**
     * Accessor for default value for column.
     * @return The default value for column
     **/
    public String getColumnDef()
    {
        return columnDef;
    }

    /**
     * Accessor for column remarks
     * @return The remarks on the column
     **/
    public String getRemarks()
    {
        return remarks;
    }

    /**
     * Accessor for type name
     * @return The type name for the column
     **/
    public String getTypeName()
    {
        return typeName;
    }

    /**
     * Accessor for table name
     * @return The table name for the column
     **/
    public String getTableName()
    {
        return tableName;
    }

    /**
     * Accessor for column name
     * @return The column name
     **/
    public String getColumnName()
    {
        return columnName;
    }

    /**
     * Accessor for table catalog
     * @return The table catalog
     **/
    public String getTableCat()
    {
        return tableCat;
    }

    /**
     * Accessor for table schema
     * @return The table schema
     **/
    public String getTableSchem()
    {
        return tableSchem;
    }

    /**
     * Constructs a column information object from the current row of the given
     * result set.  The {@link ResultSet} object passed must have been obtained
     * from a call to java.sql.DatabaseMetaData.getColumns().
     *
     * <p>This method only retrieves the values from the current row; the caller
     * is required to advance to the next row with {@link ResultSet#next}.
     *
     * @param rs    The result set returned from 
     *              java.sql.DatabaseMetaData.getColumns().
     *
     * @exception JDOFatalDataStoreException
     *      if a column of column information could not be retrieved from the
     *      result set.
     */
    public ColumnInfo(ResultSet rs) throws JDOFatalDataStoreException
    {
        try
        {
            tableCat        = rs.getString(1);
            tableSchem      = rs.getString(2);
            tableName       = rs.getString(3);
            columnName      = rs.getString(4);
            dataType        = rs.getShort(5);
            typeName        = rs.getString(6);
            columnSize      = rs.getInt(7);
            decimalDigits   = rs.getInt(9);
            numPrecRadix    = rs.getInt(10);
            nullable        = rs.getInt(11);
            remarks         = rs.getString(12);
            columnDef       = rs.getString(13);
            charOctetLength = rs.getInt(16);
            ordinalPosition = rs.getInt(17);
            isNullable      = rs.getString(18);
        }
        catch (SQLException e)
        {
            throw new JDOFatalDataStoreException("Can't read JDBC metadata from result set", e);
        }

        if (dataType == Types.OTHER)
        {
            /*
             * This nonsense is for drivers (like Oracle's) that inexplicably
             * declare some columns to be of type Types.OTHER when they should
             * obviously know better.
             */
            String upperTypeName = typeName.toUpperCase();

            if (upperTypeName.equals("BLOB"))
            {
                dataType = Types.BLOB;
            }
            else if (upperTypeName.equals("CLOB"))
            {
                dataType = Types.CLOB;
            }
            else if (upperTypeName.equals("FLOAT"))
            {
                dataType = Types.FLOAT;
            }
            else if (upperTypeName.endsWith("LONGVARBINARY"))
            {
                dataType = Types.LONGVARBINARY;
            }
            else if (upperTypeName.endsWith("LONGVARCHAR"))
            {
                dataType = Types.LONGVARCHAR;
            }
        }
    }

    /**
     * Indicates whether some object is "equal to" this one.  Two <tt>ColumnInfo
     * </tt> objects are considered equal if their catalog, schema, table, and
     * column names are all equal.
     *
     * @param   obj     the reference object with which to compare
     *
     * @return  <tt>true</tt> if this object is equal to the obj argument;
     *          <tt>false</tt> otherwise.
     */
    public final boolean equals(Object obj)
    {
        if (obj == this)
        {
            return true;
        }

        if (!(obj instanceof ColumnInfo))
        {
            return false;
        }

        ColumnInfo ci = (ColumnInfo)obj;

        return (tableCat   == null ? ci.tableCat   == null : tableCat.equals(ci.tableCat)) &&
            (tableSchem == null ? ci.tableSchem == null : tableSchem.equals(ci.tableSchem)) &&
            tableName.equals(ci.tableName) &&
            columnName.equals(ci.columnName);
    }

    /**
     * Returns a hash code value for this object.
     *
     * @return  a hash code value for this object.
     */
    public final int hashCode()
    {
        if (hash == 0)
        {
            hash = (tableCat   == null ? 0 : tableCat.hashCode()) ^
                 (tableSchem == null ? 0 : tableSchem.hashCode()) ^
                 tableName.hashCode() ^
                 columnName.hashCode();
        }

        return hash;
    }

    /**
     * Returns the string representation of this object.
     *
     * @return  string representation of this object.
     */
    public String toString()
    {
        StringBuffer str=new StringBuffer();
        str.append(this.getClass().getName() + "\n");
        str.append("  tableCat        = " + tableCat + "\n");
        str.append("  tableSchem      = " + tableSchem + "\n");
        str.append("  tableName       = " + tableName + "\n");
        str.append("  columnName      = " + columnName + "\n");
        str.append("  dataType        = " + dataType + "\n");
        str.append("  typeName        = " + typeName + "\n");
        str.append("  columnSize      = " + columnSize + "\n");
        str.append("  decimalDigits   = " + decimalDigits + "\n");
        str.append("  numPrecRadix    = " + numPrecRadix + "\n");
        str.append("  nullable        = " + nullable + "\n");
        str.append("  remarks         = " + remarks + "\n");
        str.append("  columnDef       = " + columnDef + "\n");
        str.append("  charOctetLength = " + charOctetLength + "\n");
        str.append("  ordinalPosition = " + ordinalPosition + "\n");
        str.append("  isNullable      = " + isNullable + "\n");

        return str.toString();
    }
}
