/**********************************************************************
Copyright (c) 2003 Erik Bengtson and others. All rights reserved.
This program and the accompanying materials are made available under
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2004 Andy Jefferson - coding standards
    ...
**********************************************************************/
package org.jpox.store.rdbms.datatype;

import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.OutputStream;
import java.io.StreamCorruptedException;
import java.sql.Blob;
import java.sql.SQLException;

import javax.jdo.JDOFatalUserException;

import org.jpox.util.JPOXLogger;

/**
 * <p>
 * The representation (mapping) in the Java <sup><font size=-2>TM </font>
 * </sup> programming language of an SQL <code>BLOB</code> value. An SQL
 * <code>BLOB</code> is a built-in type that stores a Binary Large Object as a
 * column value in a row of a database table. The driver implements
 * <code>Blob</code> using an SQL <code>locator(BLOB)</code>, which means
 * that a <code>Blob</code> object contains a logical pointer to the SQL
 * <code>BLOB</code> data rather than the data itself. A <code>Blob</code>
 * object is valid for the duration of the transaction in which is was created.
 * <P>
 * Methods in the interfaces {@link java.sql.ResultSet},
 * {@link java.sql.CallableStatement}, and {@link java.sql.PreparedStatement},
 * such as <code>getBlob</code> and <code>setBlob</code> allow a programmer
 * to access an SQL <code>BLOB</code> value. The <code>Blob</code> interface
 * provides methods for getting the length of an SQL <code>BLOB</code> (Binary
 * Large Object) value, for materializing a <code>BLOB</code> value on the
 * client, and for determining the position of a pattern of bytes within a
 * <code>BLOB</code> value.
 * <P>
 * This class is new in the JDBC 2.0 API.
 * @version $Revision: 1.2 $
 */
public class BlobImpl implements Blob
{
    private InputStream stream;

    private int length;

    private byte bytes[];

    /**
     * Constructor taking a serialised object.
     * @param obj The serialised object.
     * @throws IOException
     **/
    public BlobImpl(Object obj) throws IOException
    {
        ByteArrayOutputStream baos = new ByteArrayOutputStream();
        ObjectOutputStream oos = new ObjectOutputStream(baos);
        oos.writeObject(obj);
        bytes = baos.toByteArray();
        this.stream = new ByteArrayInputStream(bytes);
        this.length = bytes.length;
    }

    /**
     * Constructor taking a byte array.
     * @param bytes The byte array
     * @throws IOException
     **/
    public BlobImpl(byte[] bytes) throws IOException
    {
        this.bytes = bytes;
        this.stream = new ByteArrayInputStream(bytes);
        this.length = bytes.length;
    }

    /**
     * Constructor taking an InputStream.
     * @param stream The InputStream
     * @throws IOException
     **/
    public BlobImpl(InputStream stream) throws IOException
    {
        this.stream = stream;
        this.length = bytes.length;
    }

    /**
     * Accessor for the Object.
     * @return The object.
     **/
    public Object getObject()
    {
        ByteArrayInputStream bais = new ByteArrayInputStream(bytes);
        ObjectInputStream ois;
        try
        {
            ois = new ObjectInputStream(bais);
            return ois.readObject();
        }
        catch (StreamCorruptedException e)
        {
            String msg = "StreamCorruptedException: object is corrupted";
            JPOXLogger.RDBMS.error(msg);
            throw new JDOFatalUserException(msg, e);
        }
        catch (IOException e)
        {
            String msg = "IOException: error when reading object";
            JPOXLogger.RDBMS.error(msg);
            throw new JDOFatalUserException(msg, e);
        }
        catch (ClassNotFoundException e)
        {
            String msg = "ClassNotFoundException: error when creating object";
            JPOXLogger.RDBMS.error(msg);
            throw new JDOFatalUserException(msg, e);
        }
    }

    /* (non-Javadoc)
     * @see java.sql.Blob#length()
     */
    public long length() throws SQLException
    {
        return length;
    }

    /* (non-Javadoc)
     * @see java.sql.Blob#getBytes(long, int)
     */
    public byte[] getBytes(long pos, int length) throws SQLException
    {
        byte bytesToReturn[] = new byte[length];
        for (int i = 0; i < length; i++)
        {
            bytesToReturn[i] = bytes[(int) pos + i];
        }
        return bytesToReturn;
    }

    public void setBytes(byte[] bytes, long pos, int length) throws SQLException
    {
    }

    public int setBytes(long value, byte[] bytes, int pos, int length) throws SQLException
    {
        return -1;
    }

    public void truncate(long value)
    {
    }

    public int setBytes(long value, byte[] bytes) throws SQLException
    {
        return -1;
    }

    /* (non-Javadoc)
     * @see java.sql.Blob#getBinaryStream()
     */
    public InputStream getBinaryStream() throws SQLException
    {
        return stream;
    }

    public OutputStream setBinaryStream(long value) throws SQLException
    {
        return null;
    }

    /* (non-Javadoc)
     * @see java.sql.Blob#position(byte[], long)
     */
    public long position(byte[] pattern, long start) throws SQLException
    {
        throw new UnsupportedOperationException("[BlobImpl.position] may not be called");
    }

    /* (non-Javadoc)
     * @see java.sql.Blob#position(java.sql.Blob, long)
     */
    public long position(Blob pattern, long start) throws SQLException
    {
        throw new UnsupportedOperationException("[BlobImpl.position] may not be called");
    }
}