/**********************************************************************
Copyright (c) 2003 Mike Martin (TJDO) and others. All rights reserved. 
This program and the accompanying materials are made available under
the terms of the JPOX License v1.0 which accompanies this distribution.

Contributors:
2003 Erik Bengtson - changed newQueryStatement to support JPOX ID
2003 Erik Bengtson - fixed BUG [800693]
2003 Erik Bengtson - refactored and commented newQueryStatement method
2003 Andy Jefferson - fixed BUG [804902]
2003 Andy Jefferson - updated logic for getSubClasses and added localiser
2004 Erik Bengtson - removed unused import
2004 Andy Jefferson - added capability to handle multiple level inheritance
2004 Andy Jefferson - split out AbstractRDBMSExtent
    ...
**********************************************************************/
package org.jpox.store.rdbms.extent;

import org.jpox.ClassLoaderResolver;
import org.jpox.PersistenceManager;
import org.jpox.FetchPlanImpl.FetchPlanForClass;
import org.jpox.metadata.ClassMetaData;
import org.jpox.metadata.IdentityType;
import org.jpox.store.DatastoreClass;
import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.StatementExpressionIndex;
import org.jpox.store.extent.AbstractExtent;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.MappingCallbacks;
import org.jpox.store.mapping.Mappings;
import org.jpox.store.query.IteratorStatement;
import org.jpox.store.query.PersistentIDROF;
import org.jpox.store.query.Query;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.query.Queryable;
import org.jpox.store.query.ResultClassROF;
import org.jpox.store.rdbms.sqlidentifier.TableIdentifier;
import org.jpox.store.rdbms.table.ClassTable;
import org.jpox.util.Localiser;

/**
 * An Extent of all persistent objects backed by a class table.
 * There are 3 inheritance strategies, and each is handled by this class but
 * in different ways.
 * <H3>no-table Strategy</H3>
 * In this case a class does not have its own table and its fields are
 * persisted to the table of a subclass. If the class requiring the Extent
 * is the "no-table" class, then there will be an array of tables where objects
 * of this type can be persisted (as the base table). For this reason the caller
 * passes in an array of tables.
 * <H3>new-table Strategy</H3>
 * In this case the class has its own table, and so there is only one base
 * table for the class. The array of tables passed in will be of size 1.
 * <H3>superclass-table Strategy</H3>
 * In this case the class doesn't have its own table, BUT will be persisted
 * into the table of its superclass. In this case the array of tables passed
 * in will be of size 1 (the superclass' table).
 * 
 * @version $Revision: 1.9 $
 */
public class ClassTableExtent extends AbstractExtent implements Queryable
{
    /** Localised messages */
    protected static final Localiser LOCALISER_RDBMS = Localiser.getInstance("org.jpox.store.rdbms.extent.Localisation");

    /** Table(s) required for this Extent. */
    private final DatastoreClass[] tables;
    
    private final boolean multipleTableCase;

    /**
     * Constructor. Used for classes that have "new-table" or "superclass-table" inheritance
     * strategies.
     * @param pm PersistenceManager
     * @param table Table representing the class whose extent we want
     * @param className Name of the class whose extent we want
     * @param subclasses Whether to include subclasses of the class.
     */
    public ClassTableExtent(PersistenceManager pm, DatastoreClass table, String className, boolean subclasses)
    {
        super(pm, className, subclasses);

        this.tables = new ClassTable[1];
        tables[0] = table;
        multipleTableCase = false;
    }

    /**
     * Constructor. Used for classes that have "no-table" inheritance strategy, and hence
     * can be persisted into multiple tables.
     * @param pm PersistenceManager
     * @param tables Tables representing the class whose extent we want
     * @param className Name of the class whose extent we want
     * @param subclasses Whether to include subclasses of the class.
     */
    public ClassTableExtent(PersistenceManager pm, DatastoreClass[] tables, String className, boolean subclasses)
    {
        super(pm, className, subclasses);

        this.tables = tables;
        multipleTableCase = true;
    }

    /**
     * Create a new query to search for the candidate classes etc.
     * @return The new QueryStatement.
     */
    public QueryStatement newQueryStatement()
    {
        return newQueryStatement(getCandidateClass());
    }

    /**
     * Create a query to search for the candidateClass and subclasses if true
     * @param candidateClass The candidate class
     * @return The new QueryStatement
     */
    public QueryStatement newQueryStatement(Class candidateClass)
    {
        final ClassLoaderResolver clr=storeMgr.getClassLoaderResolver();
        QueryStatement query = null;

        if (tables == null)
        {
            return null;
        }

        // In the multiple table case we pass in the class of the table rather than the candidate class since
        // the instances of that table can only be of the tables type ("no-table" strategy)
        for (int i = 0; i < tables.length; i++)
        {
            final int tableNo = i;
            QueryStatement query_table = new IteratorStatement(new IteratorStatement.AssociationEnd()
                {
                	public JavaTypeMapping getMapping()
                	{
                	    return tables[tableNo].getIDMapping();
                	}
                	public Class getType()
                	{
                	    return clr.classForName(tables[tableNo].getType());
                	}
                	public DatastoreContainerObject getDatastoreContainerObject()
                	{
                	    return tables[tableNo];
                	}
                	public boolean useJoin()
                	{
                	    return false;
                	}
                },
                multipleTableCase ? clr.classForName(tables[tableNo].getType()) : candidateClass, 
                storeMgr, storeMgr.getDatastoreAdapter(), subclasses, true).getQueryStatement();

            if (query != null)
            {
                query.union(query_table);
            }
            else
            {
                query = query_table;
            }
        }

        return query;
    }

    /**
     * Create a new factory for objects from the ResultSet.
     * @param stmt The Query Statement
     * @param ignoreCache Whether to ignore dirty objects
     * @param resultClass Whether to create objects of a particular class
     * @return The result object factory
     */
    public Query.ResultObjectFactory newResultObjectFactory(QueryStatement stmt, boolean ignoreCache, Class resultClass)
    {
        int[] prefetchFieldNumbers;
        StatementExpressionIndex[] statementExpressionIndex;
        
        ClassMetaData cmd = storeMgr.getMetaDataManager().getMetaDataForClass(getCandidateClass());
        fetchPlan.manageFetchPlanForClass(cmd);
        FetchPlanForClass fpc = fetchPlan.getFetchPlanForClass(cmd);
        int fieldNumbers[] = fpc.getFieldsInActualFetchPlan();
        int fn[] = new int[fieldNumbers.length];
        int prefetchFieldCount = 0;
        int fieldCount = cmd.getNoOfInheritedManagedFields() + cmd.getNoOfManagedFields();
        
        statementExpressionIndex = new StatementExpressionIndex[fieldCount];
        
        for (int i=0; i<fieldNumbers.length; ++i)
        {
            JavaTypeMapping m = tables[0].getFieldMapping(cmd.getManagedFieldAbsolute(fieldNumbers[i]));
	            
            if( m != null ) // field is not stored in the table, e.g. List, Set, etc or is transactional
            {
	            if (m.includeInFetchStatement() && !(m instanceof MappingCallbacks))
	            {
	                statementExpressionIndex[fieldNumbers[i]] = new StatementExpressionIndex();
	                statementExpressionIndex[fieldNumbers[i]].setMapping(m);
	                fn[prefetchFieldCount++] = fieldNumbers[i];
	            }
            }
        }
        
        prefetchFieldNumbers = new int[prefetchFieldCount];
        System.arraycopy(fn, 0, prefetchFieldNumbers, 0, prefetchFieldCount);                

        int[] datastoreIndex = null;

        // following the actual fetch plan, we select the fields to be fetched from the database
        if (stmt.getTableExpression(new TableIdentifier(storeMgr.getDatastoreAdapter(),"ELEMENT")) != null)
        {
            if( ((ClassTable) tables[0]).getIdentityType() == IdentityType.DATASTORE )
            {
                datastoreIndex = stmt.select(new TableIdentifier(storeMgr.getDatastoreAdapter(), "ELEMENT"),tables[0].getDataStoreObjectIdMapping(),true);
            }
            Mappings.selectMapping(stmt, new TableIdentifier(storeMgr.getDatastoreAdapter(), "ELEMENT"), statementExpressionIndex);
        }
        else
        {
            if( ((ClassTable) tables[0]).getIdentityType() == IdentityType.DATASTORE )
            {
                datastoreIndex = stmt.select(new TableIdentifier(storeMgr.getDatastoreAdapter(), "this"),tables[0].getDataStoreObjectIdMapping(),true);
            }
            Mappings.selectMapping(stmt,statementExpressionIndex);
        }

        if (resultClass == null)
        {
            return new PersistentIDROF(tables[0],
                prefetchFieldNumbers,
                statementExpressionIndex,
                datastoreIndex,
                ignoreCache);
        }
        else
        {
            return new ResultClassROF(resultClass, statementExpressionIndex);
        }
    }

    /**
     * Returns <tt>true</tt> if this collection contains no elements.<p>
     * @return <tt>true</tt> if this collection contains no elements.
     */
    public boolean isEmpty()
    {
        // we don't know if we are empty until it calls the datastore
        return false;
    }
}