/**********************************************************************
Copyright (c) 2003 Mike Martin and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
    Andy Jefferson - coding standards
    Andy Jefferson - split out AbstractRDBMSExtent
    ...
**********************************************************************/
package org.jpox.store.rdbms.extent;

import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOFatalUserException;

import org.jpox.PersistenceManager;
import org.jpox.store.StatementExpressionIndex;
import org.jpox.store.exceptions.IncompatibleQueryElementTypeException;
import org.jpox.store.extent.AbstractExtent;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.MappingCallbacks;
import org.jpox.store.mapping.Mappings;
import org.jpox.store.query.Query;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.query.Queryable;
import org.jpox.store.query.TransientIDROF;
import org.jpox.store.rdbms.adapter.DatabaseAdapter;
import org.jpox.store.rdbms.table.ClassView;
import org.jpox.util.Localiser;

/**
 * An Extent of all persistent objects backed by a view.
 * @version $Revision: 1.3 $
 */
public class ClassViewExtent extends AbstractExtent implements Queryable
{
    /** Localised messages */
    protected static final Localiser LOCALISER_RDBMS = Localiser.getInstance("org.jpox.store.rdbms.extent.Localisation");

    private final ClassView view;

    private final int fieldCount;
    private final int[] prefetchFieldNumbers;
	private final StatementExpressionIndex[] statementExpressionIndex;

    /**
     * Constructor.
     * @param pm Persistence Manager
     * @param view The view
     * @param subclasses Whether to include subclasses
     */
    public ClassViewExtent(PersistenceManager pm, ClassView view, boolean subclasses)
    {
        super(pm, view.getType(), subclasses);

        this.view = view;

        // Set up list of fields required for the candidate class
        fieldCount = cmd.getNoOfManagedFields();
        int[] fn = new int[fieldCount];
		statementExpressionIndex = new StatementExpressionIndex[fieldCount];
        int prefetchFieldCount = 0;

        for (int i=0; i<fieldCount; ++i)
        {
            if (view.isFieldPersistent(i))
            {
                JavaTypeMapping m = view.getFieldMapping(cmd.getManagedFieldAbsolute(i));
                if (!m.includeInFetchStatement() || m instanceof MappingCallbacks)
                {
                    throw new JDOFatalInternalException(LOCALISER_RDBMS.msg("Extent.MappingInvalidForView",m, className));
                }
                
				statementExpressionIndex[i] = new StatementExpressionIndex();
				statementExpressionIndex[i].setMapping(m);
                fn[prefetchFieldCount++] = i;
            }
        }

        if (prefetchFieldCount == 0)
        {
            throw new JDOFatalUserException(LOCALISER_RDBMS.msg("Extent.ClassHasNoPersistentFields", className));
        }

        prefetchFieldNumbers = new int[prefetchFieldCount];
        System.arraycopy(fn, 0, prefetchFieldNumbers, 0, prefetchFieldCount);
    }

    /**
     * Create a new query to search for the candidate classes etc.
     * @return The new QueryStatement.
     */
    public QueryStatement newQueryStatement()
    {
        return ((DatabaseAdapter) storeMgr.getDatastoreAdapter()).newQueryStatement(view);
    }

    /**
     * Create a query to search for the candidateClass and subclasses if true
     * @param candidateClass
     * @return The new QueryStatement
     */
    public QueryStatement newQueryStatement(Class candidateClass)
    {
        String extentType = view.getType();

        if (!extentType.equals(className))
        {
            throw new IncompatibleQueryElementTypeException(extentType, className);
        }

        return ((DatabaseAdapter) storeMgr.getDatastoreAdapter()).newQueryStatement(view);
    }

    /**
     * Create a new factory for objects from the ResultSet.
     * @param stmt The Query Statement
     * @param ignoreCache Whether to ignore dirty objects
     * @param resultClass Whether to create objects of a particular class
     * @return The result object factory
     */
    public Query.ResultObjectFactory newResultObjectFactory(QueryStatement stmt,boolean ignoreCache,Class resultClass)
    {
        Mappings.selectMapping(stmt,statementExpressionIndex);

        return new TransientIDROF(getCandidateClass(), prefetchFieldNumbers, statementExpressionIndex);
    }

    /**
     * Returns <tt>true</tt> if this collection contains no elements.<p>
     * @return <tt>true</tt> if this collection contains no elements.
     */
    public boolean isEmpty()
    {
        // we don't know if we are empty until it calls the datastore
        return false;
    }
}