/**********************************************************************
Copyright (c) 2003 Mike Martin and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
    Andy Jefferson - updated to use StoreManagerHelper reference
    Andy Jefferson - coding standards
    ...
**********************************************************************/
package org.jpox.store.rdbms.foreignkey;

import java.sql.ResultSet;
import java.sql.SQLException;

import javax.jdo.JDOFatalDataStoreException;


/**
 * Represents the metadata of a specific foreign key column.  This class is
 * basically a data structure that makes accessing the JDBC foreign key metadata
 * easier.  Each of the items returned by
 * {@link java.sql.DatabaseMetaData#getImportedKeys(String,String,String)} or
 * {@link java.sql.DatabaseMetaData#getExportedKeys(String,String,String)}
 * is represented by a public field in this class.
 *
 * Subclasses of ForeignKeyInfo can be created on a per-DBMS basis to supply
 * missing metadata or correct faulty metadata obtained from that DBMS's JDBC
 * driver(s).
 *
 * @see org.jpox.store.rdbms.RDBMSStoreHelper#getForeignKeyInfo
 * @see org.jpox.store.rdbms.adapter.DatabaseAdapter#newForeignKeyInfo
 * @version $Revision: 1.1 $ 
 */
public class ForeignKeyInfo
{
    /**
     * The primary key table catalog, which may be <tt>null</tt>.
     */
    public String pkTableCat;

    /**
     * The primary key table schema, which may be <tt>null</tt>.
     */
    public String pkTableSchem;

    /**
     * The primary key table name.
     */
    public String pkTableName;

    /**
     * The primary key column name.
     */
    public String pkColumnName;

    /**
     * The foreign key table catalog, which may be <tt>null</tt>.
     */
    public String fkTableCat;

    /**
     * The foreign key table schema, which may be <tt>null</tt>.
     */
    public String fkTableSchem;

    /**
     * The foreign key table name.
     */
    public String fkTableName;

    /**
     * The foreign key column name.
     */
    public String fkColumnName;

    /**
     * The sequence number within the foreign key, base 1.
     */
    public short keySeq;

    /**
     * What happens to a foreign key when the primary key is updated.
     *
     * @see java.sql.DatabaseMetaData#importedKeyNoAction
     * @see java.sql.DatabaseMetaData#importedKeyCascade
     * @see java.sql.DatabaseMetaData#importedKeySetNull
     * @see java.sql.DatabaseMetaData#importedKeySetDefault
     * @see java.sql.DatabaseMetaData#importedKeyRestrict
     */
    public short updateRule;

    /**
     * What happens to a foreign key when the primary key is deleted.
     *
     * @see java.sql.DatabaseMetaData#importedKeyNoAction
     * @see java.sql.DatabaseMetaData#importedKeyCascade
     * @see java.sql.DatabaseMetaData#importedKeySetNull
     * @see java.sql.DatabaseMetaData#importedKeySetDefault
     * @see java.sql.DatabaseMetaData#importedKeyRestrict
     */
    public short deleteRule;

    /**
     * The foreign key name.
     */
    public String fkName;

    /**
     * The primary key name.
     */
    public String pkName;

    /**
     * Indicates whether the evaluation of the foreign key constraint can be
     * deferred until commit time.
     *
     * @see java.sql.DatabaseMetaData#importedKeyInitiallyDeferred
     * @see java.sql.DatabaseMetaData#importedKeyInitiallyImmediate
     * @see java.sql.DatabaseMetaData#importedKeyNotDeferrable
     */
    public short deferrability;

    private int hash = 0;

    /**
     * Constructs a foreign key information object from the current row of the
     * given result set.
     * The {@link ResultSet} object passed must have been obtained from a call
     * to java.sql.DatabaseMetaData.getImportedKeys() or
     * java.sql.DatabaseMetaData.getImportedKeys().
     *
     * <p>This method only retrieves the values from the current row; the caller
     * is required to advance to the next row with {@link ResultSet#next}.
     *
     * @param rs    The result set returned from
     *              java.sql.DatabaseMetaData.getImportedKeys() or
     *              java.sql.DatabaseMetaData.getExportedKeys().
     *
     * @exception JDOFatalDataStoreException
     *      if a column of foreign key information could not be retrieved from
     *      the result set.
     */
    public ForeignKeyInfo(ResultSet rs) throws JDOFatalDataStoreException
    {
        try
        {
            pkTableCat    = rs.getString(1);
            pkTableSchem  = rs.getString(2);
            pkTableName   = rs.getString(3);
            pkColumnName  = rs.getString(4);
            fkTableCat    = rs.getString(5);
            fkTableSchem  = rs.getString(6);
            fkTableName   = rs.getString(7);
            fkColumnName  = rs.getString(8);
            keySeq        = rs.getShort(9);
            updateRule    = rs.getShort(10);
            deleteRule    = rs.getShort(11);
            fkName        = rs.getString(12);
            pkName        = rs.getString(13);
            deferrability = rs.getShort(14);
        }
        catch (SQLException e)
        {
            throw new JDOFatalDataStoreException("Can't read JDBC metadata from result set", e);
        }
    }

    /**
     * Constructs a foreign key information object from its individual
     * attributes.
     *
     * <p>This can be useful to subclasses and/or custom DatabaseAdapters that
     * need to modify and/or correct the metadata returned by the JDBC driver.
     */
    public ForeignKeyInfo(String pkTableCat,
                          String pkTableSchem,
                          String pkTableName,
                          String pkColumnName,
                          String fkTableCat,
                          String fkTableSchem,
                          String fkTableName,
                          String fkColumnName,
                          short keySeq,
                          short updateRule,
                          short deleteRule,
                          String fkName,
                          String pkName,
                          short deferrability)
    {
        this.pkTableCat     = pkTableCat;
        this.pkTableSchem   = pkTableSchem;
        this.pkTableName    = pkTableName;
        this.pkColumnName   = pkColumnName;
        this.fkTableCat     = fkTableCat;
        this.fkTableSchem   = fkTableSchem;
        this.fkTableName    = fkTableName;
        this.fkColumnName   = fkColumnName;
        this.keySeq         = keySeq;
        this.updateRule     = updateRule;
        this.deleteRule     = deleteRule;
        this.fkName         = fkName;
        this.pkName         = pkName;
        this.deferrability  = deferrability;
    }

    /**
     * Indicates whether some object is "equal to" this one.
     * Two <tt>ForeignKeyInfo</tt> objects are considered equal if their
     * catalog, schema, table, and column names, both primary and foreign, are
     * all equal.
     *
     * @param   obj     the reference object with which to compare
     *
     * @return  <tt>true</tt> if this object is equal to the obj argument;
     *          <tt>false</tt> otherwise.
     */
    public final boolean equals(Object obj)
    {
        if (obj == this)
        {
            return true;
        }

        if (!(obj instanceof ForeignKeyInfo))
        {
            return false;
        }

        ForeignKeyInfo ci = (ForeignKeyInfo)obj;

        return (pkTableCat   == null ? ci.pkTableCat   == null : pkTableCat.equals(ci.pkTableCat)) &&
            (pkTableSchem == null ? ci.pkTableSchem == null : pkTableSchem.equals(ci.pkTableSchem)) &&
            pkTableName.equals(ci.pkTableName) &&
            pkColumnName.equals(ci.pkColumnName) &&
            (fkTableCat   == null ? ci.fkTableCat   == null : fkTableCat.equals(ci.fkTableCat)) &&
            (fkTableSchem == null ? ci.fkTableSchem == null : fkTableSchem.equals(ci.fkTableSchem)) &&
            fkTableName.equals(ci.fkTableName) &&
            fkColumnName.equals(ci.fkColumnName);
    }

    /**
     * Returns a hash code value for this object.
     *
     * @return  a hash code value for this object.
     */
    public final int hashCode()
    {
        if (hash == 0)
        {
            hash = (pkTableCat   == null ? 0 : pkTableCat.hashCode()) ^
                    (pkTableSchem == null ? 0 : pkTableSchem.hashCode()) ^
                    pkTableName.hashCode() ^
                    pkColumnName.hashCode() ^
                    (fkTableCat   == null ? 0 : fkTableCat.hashCode()) ^
                    (fkTableSchem == null ? 0 : fkTableSchem.hashCode()) ^
                    fkTableName.hashCode() ^
                    fkColumnName.hashCode();
        }

        return hash;
    }

    /**
     * Returns the string representation of this object.
     *
     * @return  string representation of this object.
     */
    public String toString()
    {
        StringBuffer str= new StringBuffer();

        str.append(this.getClass().getName() + "\n");
        str.append("  pkTableCat    = " + pkTableCat + "\n");
        str.append("  pkTableSchem  = " + pkTableSchem + "\n");
        str.append("  pkTableName   = " + pkTableName + "\n");
        str.append("  pkColumnName  = " + pkColumnName + "\n");
        str.append("  fkTableCat    = " + fkTableCat + "\n");
        str.append("  fkTableSchem  = " + fkTableSchem + "\n");
        str.append("  fkTableName   = " + fkTableName + "\n");
        str.append("  fkColumnName  = " + fkColumnName + "\n");
        str.append("  keySeq        = " + keySeq + "\n");
        str.append("  updateRule    = " + updateRule + "\n");
        str.append("  deleteRule    = " + deleteRule + "\n");
        str.append("  fkName        = " + fkName + "\n");
        str.append("  pkName        = " + pkName + "\n");
        str.append("  deferrability = " + deferrability + "\n");

        return str.toString();
    }
}
