/**********************************************************************
Copyright (c) 2003 Mike Martin (TJDO) and others. All rights reserved.
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - equality operator
2004 Andy Jefferson - deleteAction, updateAction, MetaData interface
    ...
**********************************************************************/
package org.jpox.store.rdbms.key;

import java.util.ArrayList;

import javax.jdo.JDOFatalException;
import javax.jdo.JDOFatalInternalException;

import org.jpox.metadata.ForeignKeyDeleteAction;
import org.jpox.metadata.ForeignKeyMetaData;
import org.jpox.metadata.ForeignKeyUpdateAction;
import org.jpox.store.DatastoreAdapter;
import org.jpox.store.DatastoreClass;
import org.jpox.store.DatastoreField;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.rdbms.adapter.DatabaseAdapter;

/**
 * Representation of a foreign key to another table.
 *
 * @version $Revision: 1.5 $
 **/
public class ForeignKey extends Key
{
    /** Constant representing that we should cascade the action. */
    public static final FKAction CASCADE_ACTION = new FKAction(1, "CASCADE");
    
    /** Constant representing that we should restrict the action. */
    public static final FKAction RESTRICT_ACTION = new FKAction(2, "RESTRICT");
    
    /** Constant representing that we should null the foreign key */
    public static final FKAction NULL_ACTION = new FKAction(3, "SET NULL");
    
    /** Constant representing that we should set the FK to default value. */
    public static final FKAction DEFAULT_ACTION = new FKAction(4, "SET DEFAULT");
    
    /** Inner class representing an action on the FK. */
    public static class FKAction
    {
        int type;
        String keyword;
        
        public FKAction(int type, String keyword)
        {
            this.type = type;
            this.keyword = keyword;
        }
        
        public int getType()
        {
            return type;
        }
        
        public String getKeyword()
        {
            return keyword;
        }
    }
    
    private boolean initiallyDeferred;
    private DatastoreClass refTable;
    private DatastoreAdapter dba;
    
    /** Action to perform on update */
    private FKAction updateAction;
    
    /** Action to perform on delete */
    private FKAction deleteAction;

    private ArrayList refColumns = new ArrayList();

    /**
     * Constructor.
     * @param initiallyDeferred Whether the constraints are deferred
     */
    public ForeignKey(boolean initiallyDeferred)
    {
        super(null);

        this.initiallyDeferred = initiallyDeferred;
        this.refTable = null;
        this.dba = null;
    }

    /**
     * Constructor.
     * @param mapping The type mapping for this Foreign-key field
     * @param dba Datastore adapter
     * @param refTable Referred to table
     * @param initiallyDeferred Whether they are deferred
     */
    public ForeignKey(JavaTypeMapping mapping, DatabaseAdapter dba, DatastoreClass refTable, boolean initiallyDeferred)
    {
        super(mapping.getDatastoreClass());
        this.initiallyDeferred = initiallyDeferred;
        this.refTable = refTable;
        this.dba = dba;

        if (refTable.getIDMapping() == null)
        {
            throw new JDOFatalException("ForeignKey ID mapping is not initilized for "+mapping+". Table referenced: " + refTable.toString());
        }

        for (int i=0; i<refTable.getIDMapping().getNumberOfDatastoreFields(); i++)
        {
            setDatastoreField(i, mapping.getDataStoreMapping(i).getDatastoreField(), refTable.getIDMapping().getDataStoreMapping(i).getDatastoreField()); 
        }
    }
    
    /**
     * Convenience mutator for setting the specification based on MetaData
     * @param fkmd ForeignKey MetaData definition
     */
    public void setForMetaData(ForeignKeyMetaData fkmd)
    {
        if (fkmd == null)
        {
            return;
        }
        
        if (fkmd.getName() != null)
        {
            setName(fkmd.getName());
        }

        ForeignKeyDeleteAction deleteAction = fkmd.getDeleteAction();
        if (deleteAction != null)
        {
            if (deleteAction.equals(ForeignKeyDeleteAction.CASCADE))
            {
                setDeleteAction(ForeignKey.CASCADE_ACTION);
            }
            else if (deleteAction.equals(ForeignKeyDeleteAction.RESTRICT))
            {
                setDeleteAction(ForeignKey.RESTRICT_ACTION);
            }
            else if (deleteAction.equals(ForeignKeyDeleteAction.NULL))
            {
                setDeleteAction(ForeignKey.NULL_ACTION);
            }
            else if (deleteAction.equals(ForeignKeyDeleteAction.DEFAULT))
            {
                setDeleteAction(ForeignKey.DEFAULT_ACTION);
            }
        }

        ForeignKeyUpdateAction updateAction = fkmd.getUpdateAction();
        if (updateAction != null)
        {
            if (updateAction.equals(ForeignKeyUpdateAction.CASCADE))
            {
                setUpdateAction(ForeignKey.CASCADE_ACTION);
            }
            else if (updateAction.equals(ForeignKeyUpdateAction.RESTRICT))
            {
                setUpdateAction(ForeignKey.RESTRICT_ACTION);
            }
            else if (updateAction.equals(ForeignKeyUpdateAction.NULL))
            {
                setUpdateAction(ForeignKey.NULL_ACTION);
            }
            else if (updateAction.equals(ForeignKeyUpdateAction.DEFAULT))
            {
                setUpdateAction(ForeignKey.DEFAULT_ACTION);
            }
        }

        if (fkmd.isDeferred())
        {
            initiallyDeferred = true;
        }
    }

    /**
     * Accessor for deleteAction.
     * @return Returns the deleteAction.
     */
    public FKAction getDeleteAction()
    {
        return deleteAction;
    }
    
    /**
     * Mutator for deleteAction.
     * @param deleteAction The deleteAction to set.
     */
    public void setDeleteAction(FKAction deleteAction)
    {
        this.deleteAction = deleteAction;
    }
    
    /**
     * Accessor for updateAction.
     * @return Returns the updateAction.
     */
    public FKAction getUpdateAction()
    {
        return updateAction;
    }
    
    /**
     * Mutator for updateAction.
     * @param updateAction The updateAction to set.
     */
    public void setUpdateAction(FKAction updateAction)
    {
        this.updateAction = updateAction;
    }

    /**
     * Method to add a Column.
     * @param col The column to add
     * @param refCol The column to reference 
     **/
    public void addDatastoreField(DatastoreField col, DatastoreField refCol)
    {
        setDatastoreField(columns.size(), col, refCol);
    }

    public void setDatastoreField(int seq, DatastoreField col, DatastoreField refCol)
    {
        if (table == null)
        {
            table = col.getDatastoreContainerObject();
            refTable = (DatastoreClass) refCol.getDatastoreContainerObject();
            dba = table.getStoreManager().getDatastoreAdapter();
        }
        else
        {
            if (!table.equals(col.getDatastoreContainerObject()))
            {
                throw new JDOFatalInternalException("Cannot add " + col + " as FK column for " + table);
            }
            if (!refTable.equals(refCol.getDatastoreContainerObject()))
            {
                throw new JDOFatalInternalException("Cannot add " + refCol + " as referenced FK column for " + refTable);
            }
        }

        setMinSize(columns, seq + 1);
        setMinSize(refColumns, seq + 1);

        columns.set(seq, col);
        refColumns.set(seq, refCol);
    }

    /**
     * Hashcode operator.
     * @return The hashcode
     **/
    public int hashCode()
    {
        return super.hashCode() ^ refColumns.hashCode();
    }

    /**
     * Equality operator.
     * @param obj Object to compare against
     * @return Whether they are equal.
     **/
    public boolean equals(Object obj)
    {
        if (obj == this)
        {
            return true;
        }
        if (!(obj instanceof ForeignKey))
        {
            return false;
        }

        ForeignKey fk = (ForeignKey)obj;
        if (!refColumns.equals(fk.refColumns))
        {
            return false;
        }
        
        // TODO Add update-action, delete-action here

        return super.equals(obj);
    }

    /**
     * Stringify method. Generates the foreign key statement ready for use in an SQL call.
     * @return String version of this object.
     **/
    public String toString()
    {
        StringBuffer s = new StringBuffer("FOREIGN KEY ");
        s.append(getColumnList(columns));

        // Referenced table
        // Include the referenced column list because some RDBMS require it (e.g MySQL)
        s.append(" REFERENCES ");
        s.append(refTable.toString());
        s.append(" ").append(getColumnList(refColumns));

        // Delete action
        if (deleteAction != null)
        {
            if (((DatabaseAdapter)dba).supportsForeignKeyDeleteAction(deleteAction))
            {
                s.append(" ON DELETE ").append(deleteAction.getKeyword());
            }
        }

        // Update action
        if (updateAction != null)
        {
            if (((DatabaseAdapter)dba).supportsForeignKeyUpdateAction(updateAction))
            {
                s.append(" ON UPDATE ").append(updateAction.getKeyword());
            }
        }

        // Deferral of constraints
        if (initiallyDeferred && ((DatabaseAdapter)dba).supportsDeferredConstraints())
        {
            s.append(" INITIALLY DEFERRED");
        }
        
        s.append(" ");

        return s.toString();
    }
}