/******************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved. 
This program and the accompanying materials are made available under
the terms of the JPOX License v1.0 which accompanies this distribution.

Contributors:
2004 Erik Bengtson - addition of JDBC, SQL type maps
2004 Andy Jefferson - fixed use of SQL/JDBC to find the one applicable to
                      the specified field and type.
    ...
*****************************************************************/
package org.jpox.store.rdbms.mapping;

import java.io.Serializable;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;

import javax.jdo.JDOFatalException;

import org.jpox.metadata.FieldMetaData;
import org.jpox.store.DatastoreField;
import org.jpox.store.StoreManager;
import org.jpox.store.mapping.AbstractMappingManager;
import org.jpox.store.mapping.DatastoreMapping;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.util.Localiser;
import org.jpox.util.MultiMap;

/**
 * Mapping manager for RDBMS datastores.
 * Provides mappings from standard Java types defined in org.jpox.store.mapping to
 * datastore mappings for JDBC types.
 * 
 * @version $Revision: 1.5 $
 */
class AbstractRDBMSMappingManager extends AbstractMappingManager
{
    protected static final Localiser LOCALISER=Localiser.getInstance("org.jpox.store.rdbms.mapping.Localisation");

    MultiMap datastoreMappingsByJDBCType;
    MultiMap datastoreMappingsBySQLType;
    MultiMap datastoreMappingsByJavaType;

    /**
     * Constructor.
     */
    public AbstractRDBMSMappingManager()
    {
        super();
    }

    /**
     * Initialisation method, allocating map storage for lookups.
     */
    public void initialise()
    {
        datastoreMappingsByJDBCType = new MultiMap();
        datastoreMappingsBySQLType = new MultiMap();
        datastoreMappingsByJavaType = new MultiMap();
        super.initialise();
    }

    /**
     * Utility to register a datastore mapping for a java type, and the SQL/JDBC types it can be mapped to.
     * The datastore type will be made the default for this java type.
     * @param javaTypeName Name of the java type
     * @param datastoreMappingType The datastore mapping
     * @param jdbcType The JDBC type that can be used
     * @param sqlType The SQL type that can be used
     */
    protected void registerDatastoreMapping(String javaTypeName, Class datastoreMappingType, String jdbcType, String sqlType)
    {
        registerDatastoreMapping(javaTypeName, datastoreMappingType, jdbcType, sqlType, true);
    }

    /**
     * Utility to register a datastore mapping for a java type, and the SQL/JDBC types it can be mapped to.
     * @param javaTypeName Name of the java type
     * @param datastoreMappingType The datastore mapping
     * @param jdbcType The JDBC type that can be used
     * @param sqlType The SQL type that can be used
     * @param dflt Whether this type should be used as the default mapping for this Java type
     */
    protected void registerDatastoreMapping(String javaTypeName, Class datastoreMappingType, String jdbcType, String sqlType, boolean dflt)
    {
        RDBMSTypeMapping mapping = new RDBMSTypeMapping(datastoreMappingType, dflt, javaTypeName, jdbcType, sqlType);
        datastoreMappingsByJDBCType.put(jdbcType,mapping);
        datastoreMappingsBySQLType.put(sqlType,mapping);
        datastoreMappingsByJavaType.put(javaTypeName,mapping);

        if (dflt)
        {
            // TODO Update all other mappings for this Java type to not be the default
        }
    }

    /**
     * Utility to deregister all mappings for a java type.
     * @param javaTypeName Java type name
     */
    protected void deregisterDatastoreMappingsForJavaType(String javaTypeName)
    {
        // Find the mappings that we have for this java type
        Collection mappings=new HashSet();
        Collection coll = (Collection)datastoreMappingsByJavaType.get(javaTypeName);
        if (coll != null)
        {
            Iterator collIter = coll.iterator();
            while (collIter.hasNext())
            {
                RDBMSTypeMapping mapping = (RDBMSTypeMapping)collIter.next();
                mappings.add(mapping);
            }
        }

        // Delete the mappings
        Iterator iter = mappings.iterator();
        while (iter.hasNext())
        {
            RDBMSTypeMapping mapping = (RDBMSTypeMapping)iter.next();

            // Remove the JDBC type entry
            datastoreMappingsByJDBCType.remove(mapping.getJdbcType(), mapping);

            // Remove the SQL type entry
            datastoreMappingsBySQLType.remove(mapping.getSqlType(), mapping);

            // Remove the java type entry
            datastoreMappingsByJavaType.remove(javaTypeName, mapping);
        }
    }

    /**
     * Accessor for a datastore mapping class.
     * @param fmd The field MetaData for the field
     * @param index Index of the column
     * @return The mapping class
     */
    protected Class getDatastoreMappingClass(FieldMetaData fmd, int index)
    {
        RDBMSTypeMapping datastoreMapping = null;

        // Check for any user specification of the datastore type to use
        if (fmd.getColumnMetaData().length > 0)
        {
            // Firstly look for "sql-type"
            if (fmd.getColumnMetaData()[index].getSqlType() != null)
            {
                if (datastoreMappingsBySQLType.get(fmd.getColumnMetaData()[index].getSqlType().toUpperCase()) == null)
                {
                    throw new JDOFatalException(LOCALISER.msg("AbstractRDBMSMappingManager.UnknownSQLType", fmd.getFullFieldName(),
                        fmd.getColumnMetaData()[index].getSqlType()));
                }

                // Find if this sql-type has been defined for this field
                Iterator sqlTypeIter = ((Collection) datastoreMappingsBySQLType.get(fmd.getColumnMetaData()[index].getSqlType().toUpperCase())).iterator();
                while (sqlTypeIter.hasNext())
                {
                    RDBMSTypeMapping sqlTypeMapping = (RDBMSTypeMapping)sqlTypeIter.next();
                    if (sqlTypeMapping.getJavaType().equals(fmd.getTypeName()))
                    {
                        datastoreMapping = sqlTypeMapping;
                        break;
                    }
                }
            }
            // Secondly look for "jdbc-type"
            else if (fmd.getColumnMetaData()[index].getJdbcType() != null)
            {
                if (datastoreMappingsByJDBCType.get(fmd.getColumnMetaData()[index].getJdbcType().toUpperCase()) == null)
                {
                    throw new JDOFatalException(LOCALISER.msg("AbstractRDBMSMappingManager.UnknownJDBCType", fmd.getFullFieldName(),
                        fmd.getColumnMetaData()[index].getJdbcType()));
                }

                // Find if this jdbc-type has been defined for this field
                Iterator jdbcTypeIter = ((Collection) datastoreMappingsByJDBCType.get(fmd.getColumnMetaData()[index].getJdbcType().toUpperCase())).iterator();
                while (jdbcTypeIter.hasNext())
                {
                    RDBMSTypeMapping jdbcTypeMapping = (RDBMSTypeMapping)jdbcTypeIter.next();
                    if (jdbcTypeMapping.getJavaType().equals(fmd.getTypeName()))
                    {
                        datastoreMapping = jdbcTypeMapping;
                        break;
                    }
                }
            }
        }

        // No user specification, so find one from our supported list
        if (datastoreMapping == null)
        {
            // None supported so try a Serializable mapping
            if (datastoreMappingsByJavaType.get(fmd.getType().getName()) == null)
            {
                for (int i = 0; i < fmd.getType().getInterfaces().length; i++)
                {
                    if (fmd.getType().getInterfaces()[i].getClass().getName().equals(Serializable.class.getName()))
                    {
                        break;
                    }
                }
                datastoreMapping = (RDBMSTypeMapping) ((Collection) datastoreMappingsByJavaType.get(Serializable.class.getName())).iterator().next();
            }
            else
            {
                // Find the mapping from the ones we've been given
                Collection mappings = (Collection) datastoreMappingsByJavaType.get(fmd.getTypeName());
                if (mappings != null)
                {
                    if (mappings.size() == 1)
                    {
                        datastoreMapping = (RDBMSTypeMapping) mappings.iterator().next();
                    }
                    else
                    {
                        // More than 1 so take the default
                        Iterator mappingsIter = mappings.iterator();
                        while (mappingsIter.hasNext())
                        {
                            RDBMSTypeMapping mapping = (RDBMSTypeMapping) mappingsIter.next();
                            if (mapping.isDefault())
                            {
                                datastoreMapping = mapping;
                                break;
                            }
                        }
                        // No default set, so use the first one
                        if (datastoreMapping == null)
                        {
                            datastoreMapping = (RDBMSTypeMapping) mappings.iterator().next();
                        }
                    }
                }
            }
        }

        // Still no mapping, so throw exception
        if (datastoreMapping == null)
        {
            throw new JDOFatalException(LOCALISER.msg("AbstractRDBMSMappingManager.UnsupportedType", fmd.getFullFieldName(), fmd.getTypeName()));
        }
        
        return datastoreMapping.getMappingType();
    }

    protected Class getDatastoreMappingClass(Class javaType)
    {
        RDBMSTypeMapping datastoreMapping = null;
        datastoreMapping = (RDBMSTypeMapping) ((Collection) datastoreMappingsByJavaType.get(javaType.getName())).iterator().next();
        return datastoreMapping.getMappingType();
    }

    protected class RDBMSTypeMapping extends TypeMapping
    {
        private String javaType;
        private String jdbcType;
        private String sqlType;

        /**
         * @param mappingType Datastroe mapping type
         * @param isDefault Whether is the default for this java type
         */
        public RDBMSTypeMapping(Class mappingType, boolean isDefault, String javaType, String jdbcType, String sqlType)
        {
            super(mappingType, isDefault);
            this.javaType = javaType;
            this.jdbcType = jdbcType;
            this.sqlType = sqlType;
        }

        /**
         * @return Returns the javaType.
         */
        public String getJavaType()
        {
            return javaType;
        }

        /**
         * @return Returns the jdbcType.
         */
        public String getJdbcType()
        {
            return jdbcType;
        }

        /**
         * @return Returns the sqlType.
         */
        public String getSqlType()
        {
            return sqlType;
        }
    }

    /*
     * @see org.jpox.store.mapping.MappingManager#getDatastoreMapping(org.jpox.store.mapping.JavaTypeMapping, org.jpox.metadata.FieldMetaData, int, org.jpox.store.DatastoreAdapter, org.jpox.store.DatastoreField)
     */
    public DatastoreMapping getDatastoreMapping(JavaTypeMapping mapping, FieldMetaData fmd, int index, StoreManager storeMgr, DatastoreField column)
    {
        return RDBMSMappingFactory.createMapping(getDatastoreMappingClass(fmd, index), mapping, storeMgr, column);
    }

    /* (non-Javadoc)
     * @see org.jpox.store.mapping.MappingManager#getDatastoreMappingInstance(org.jpox.metadata.FieldMetaData, int, java.lang.String)
     */
    public DatastoreMapping getDatastoreMapping(JavaTypeMapping mapping, StoreManager storeMgr, DatastoreField column, Class javaType)
    {
        return RDBMSMappingFactory.createMapping(getDatastoreMappingClass(javaType), mapping, storeMgr, column);
    }
}