/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
2004 Andy Jefferson - localised messages
2004 Andy Jefferson - fixed getObject() to cater for decimal values
    ...
**********************************************************************/
package org.jpox.store.rdbms.mapping;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Types;

import javax.jdo.JDODataStoreException;

import org.jpox.store.DatastoreField;
import org.jpox.store.OID;
import org.jpox.store.StoreManager;
import org.jpox.store.exceptions.NullValueException;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.typeinfo.TypeInfo;

/**
 * Mapping of a Big Integer RDBMS type.
 * 
 * @version $Revision: 1.7 $
 */
public class BigIntRDBMSMapping extends ColumnMapping
{
    /**
     * @param storeMgr Store Manager
     * @param mapping The java mapping
     */
    protected BigIntRDBMSMapping(StoreManager storeMgr, JavaTypeMapping mapping)
    {
        super(storeMgr, mapping);
    }
    
    public BigIntRDBMSMapping(JavaTypeMapping mapping, StoreManager storeMgr, DatastoreField field)
    {
		super(storeMgr, mapping);
		column = (Column) field;
		initialize();
	}

    private void initialize()
    {
        if( column != null )
        {
            column.checkPrimitive();
        }
		initTypeInfo();
    }
    
    public TypeInfo getTypeInfo()
    {
        return getDatabaseAdapter().getTypeInfo(Types.BIGINT);
    }
    
    public DatastoreField getDatastoreField()
    {
        return column;
    }
    
    public void setInt(Object ps, int param, int value)
    {
        try
        {
            ((PreparedStatement) ps).setLong(param, value);
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSMapping.UnableToSetParam", "int", "" + value), e);
        }
    }

    public int getInt(Object rs, int param)
    {
        int value;

        try
        {
            value = (int)((ResultSet) rs).getLong(param);

            if (((ResultSet) rs).wasNull())
            {
                throw new NullValueException(LOCALISER.msg("RDBMSMapping.IllegalNullValueInColumn",column));
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSMapping.UnableToGetParam", "int", "" + param), e);
        }

        return value;
    }
    
    public void setLong(Object ps, int param, long value)
    {
        try
        {
            ((PreparedStatement) ps).setLong(param, value);
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSMapping.UnableToSetParam", "long", "" + value), e);
        }
    }

    public long getLong(Object rs, int param)
    {
        long value;

        try
        {
            value = ((ResultSet) rs).getLong(param);

            if (((ResultSet) rs).wasNull())
            {
                throw new NullValueException(LOCALISER.msg("RDBMSMapping.IllegalNullValueInColumn",column));
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSMapping.UnableToGetParam", "long", "" + param), e);
        }

        return value;
    }

    /**
     * Setter for a parameter in a PreparedStatement
     * @param ps The PreparedStatement
     * @param param The parameter number to set
     * @param value The value to set it to.
     */
    public void setObject(Object ps, int param, Object value)
    {
        try
        {
            if (value == null)
            {
                if (column.isDefaultable() && column.getDefaultValue() != null)
                {
                    ((PreparedStatement) ps).setLong(param, Long.valueOf(column.getDefaultValue().toString().trim()).longValue());
                }
                else
                {
                    ((PreparedStatement) ps).setNull(param, getTypeInfo().dataType);
                }
            }
            else
            {
                if (value instanceof OID)
                {
                    ((PreparedStatement) ps).setLong(param, Long.parseLong(((OID)value).keyValue()));
                }
                else
                {
                    ((PreparedStatement) ps).setLong(param, ((Long)value).longValue());
                }
            }
        }
        catch (SQLException e)
        {
            String msg = LOCALISER.msg("RDBMSMapping.UnableToSetParam", "Long", "" + value);
            throw new JDODataStoreException(msg, e);
        }
    }

    /**
     * Method to retrieve a Big int from a ResultSet.
     * @param rs ResultSet
     * @param param The Parameter number in the result set
     * @return The BIGINT object
     */
    public Object getObject(Object rs, int param)
    {
        Object value;
        try
        {
            // Read the object as a String since that is the most DB independent
            // type we can use and should always get us something.
            String str = ((ResultSet) rs).getString(param);
            if (((ResultSet)rs).wasNull())
            {
                value = null;
            }
            else
            {
                // Some RDBMS (e.g PostgreSQL) can return a long as a double
                // so cater for this, and generate a Long :-)
                try
                {
                    // Try it as a long
                    value = new Long(str);
                }
                catch (NumberFormatException nfe)
                {
                    // Must be a double precision, so cast it
                    value = new Long((new Double(str)).longValue());
                }
            }
        }
        catch (SQLException e)
        {
            String msg = LOCALISER.msg("RDBMSMapping.UnableToGetParam", "Long", "" + param);
            throw new JDODataStoreException(msg, e);
        }

        return value;
    }
}