/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved.
This program and the accompanying materials are made available under
the terms of the JPOX License v1.0 which accompanies this distribution.

Contributors:
2004 Andy Jefferson - localised messages
2004 Andy Jefferson - added Currency
2005 Andy Jefferson - added control over TimeZone
    ...
**********************************************************************/
package org.jpox.store.rdbms.mapping;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Timestamp;
import java.sql.Types;
import java.util.Calendar;
import java.util.Currency;
import java.util.GregorianCalendar;
import java.util.Locale;
import java.util.TimeZone;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOUserException;

import org.jpox.store.DatastoreField;
import org.jpox.store.OID;
import org.jpox.store.StoreManager;
import org.jpox.store.exceptions.NullValueException;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.typeinfo.TypeInfo;
import org.jpox.util.I18nUtils;

/**
 * Mapping of a CHAR RDBMS type.
 * 
 * @version $Revision: 1.10 $
 */
public class CharRDBMSMapping extends ColumnMapping
{
    private static final int LOCALE_LENGTH = 20;

    /* Based on JDBC date escape format: "hh:mm:ss". */
    private static final int TIME_STRING_LENGTH = 8;

    /* Based on JDBC date escape format: "YYYY-MM-DD". */
    private static final int DATE_STRING_LENGTH = 10;

    /*
     * Maximum length of a timestamp string, based on JDBC timestamp escape
     * format: "YYYY-MM-DD HH:MM:SS.FFFFFFFFF"
     */
    private static final int TIMESTAMP_STRING_LENGTH = 29;

    /** Length of a Timezone identifier. */
    private static final int TIMEZONE_LENGTH = 30;

    /** Length of an ISO4217 currency code */
    private static final int CURRENCY_LENGTH = 3;

    /**
     * Constructor.
     * @param storeMgr Store Manager
     * @param mapping Java type mapping
     */
    protected CharRDBMSMapping(StoreManager storeMgr, JavaTypeMapping mapping)
    {
        super(storeMgr, mapping);
    }
    
    /**
     * Constructor.
     * @param mapping Java type mapping
     * @param storeMgr Store Manager
     * @param field Field to be mapped
     */
    public CharRDBMSMapping(JavaTypeMapping mapping, StoreManager storeMgr, DatastoreField field)
    {
		super(storeMgr, mapping);
		column = (Column) field;
		initialize();
	}

    private void initialize()
    {
        if (column != null)
        {
	        if (getJavaTypeMapping().getJavaType().getName().equals(Boolean.class.getName()))
			{
			    column.getColumnOptions().setLength(Integer.valueOf("1"));
			    column.getColumnOptions().setJdbcType("CHAR");
	            column.checkString();
	
	            StringBuffer constraints = new StringBuffer("CHECK (" + column.getName() + " IN ('Y','N')");
	            if (column.isNullable()) 
	            {
	                constraints.append(" OR " + column.getName() + " IS NULL");
	            }
	            constraints.append(')');
	            column.setConstraints(constraints.toString());		    
			}
		    /*
		     * Locale should be stored in colums from 2 to 20 characters. Normaly, we will have a string no longer than
		     * 5 characters, but variants, in general, are vendor specific and can be longer than expected. 
		     * The Variant codes are vendor and browser-specific. For example, use WIN for Windows, MAC for Macintosh,
		     * and POSIX for POSIX. Where there are two variants, separate them with an underscore, and put the most
		     * important one first. For example, a Traditional Spanish collation might construct a locale with
		     * parameters for language, country and variant as: "es", "ES", "Traditional_WIN".
		     *  
		     * language_country_variant
		     * 
		     * Examples: "en", "de_DE", "_GB", "en_US_WIN", "de__POSIX", "fr_MAC"
		     * 
		     * @see java.util.Locale
		     */
			else if (getJavaTypeMapping().getJavaType().getName().equals(Locale.class.getName()))
			{
				column.getColumnOptions().setLength(LOCALE_LENGTH);
			}
			else if (getJavaTypeMapping().getJavaType().getName().equals(java.sql.Time.class.getName()))
			{
				column.getColumnOptions().setLength(TIME_STRING_LENGTH);
				column.getColumnOptions().setJdbcType("CHAR");
				column.checkString();
			}
			else if (getJavaTypeMapping().getJavaType().getName().equals(java.sql.Date.class.getName()))
			{
				column.getColumnOptions().setLength(DATE_STRING_LENGTH);
				column.getColumnOptions().setJdbcType("CHAR");
				column.checkString();
			}
			else if (getJavaTypeMapping().getJavaType().getName().equals(Timestamp.class.getName()))
			{
			    column.getColumnOptions().setLength(TIMESTAMP_STRING_LENGTH);
			    column.getColumnOptions().setJdbcType("CHAR");
			    column.checkString();		
			}
			else if (getJavaTypeMapping().getJavaType().getName().equals(TimeZone.class.getName()))
			{
			    column.getColumnOptions().setLength(TIMEZONE_LENGTH);
				column.getColumnOptions().setJdbcType("CHAR");
				column.checkString();		
            }
            else if (getJavaTypeMapping().getJavaType().getName().equals(Currency.class.getName()))
            {
                column.getColumnOptions().setLength(CURRENCY_LENGTH);
                column.getColumnOptions().setJdbcType("CHAR");
                column.checkString();       
            }
			else
			{
		        int maxlength = getTypeInfo().precision;
				column.checkString();		
				if (column.getColumnOptions().getLength().intValue() <= 0 || column.getColumnOptions().getLength().intValue() > maxlength)
				{
					throw new JDOUserException("String max length of " + column.getColumnOptions().getLength() +
						" is outside the acceptable range [0, " + maxlength + "] for column \"" + column.getName() + "\"");
				}
			}
        }
		initTypeInfo();
    }
    
    public TypeInfo getTypeInfo()
    {
        return getDatabaseAdapter().getTypeInfo(Types.CHAR);
    }
    
    public DatastoreField getDatastoreField()
    {
        return column;
    }
    
    /**
     * Mutator for the TimeZone object.
     * @param ps The statement
     * @param param The parameter id
     * @param value The object value
     **/
    public void setTimeZone(Object ps, 
                            int param, 
                            TimeZone value)
    {
        try
        {
            ((PreparedStatement) ps).setString(param, value.getID());
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSMapping.UnableToSetParam","Timezone","" + value), e);
        }
    }

    /**
     * Accessor for the TimeZone.
     * @param rs The Result Set
     * @param param The parameter id
     * @return The TimeZone
     **/
    public TimeZone getTimeZone(Object rs, int param)
    {
        TimeZone value;

        try
        {
            value = TimeZone.getTimeZone(((ResultSet) rs).getString(param));

            if (((ResultSet) rs).wasNull())
            {
                throw new NullValueException(LOCALISER.msg("RDBMSMapping.IllegalNullValueInColumn",column));
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSManager.UnableToGetParam","Timezone","" + param), e);
        }

        return value;
    }

    /**
     * Mutator for the Currency object.
     * @param ps The statement
     * @param param The parameter id
     * @param value The object value
     **/
    public void setCurrency(Object ps, 
                            int param, 
                            Currency value)
    {
        try
        {
            ((PreparedStatement) ps).setString(param, value.toString());
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSMapping.UnableToSetParam","Currency","" + value), e);
        }
    }

    /**
     * Accessor for the Currency.
     * @param rs The Result Set
     * @param param The parameter id
     * @return The Currency
     **/
    public Currency getCurrency(Object rs, int param)
    {
        Currency value;

        try
        {
            value = Currency.getInstance(((ResultSet) rs).getString(param));

            if (((ResultSet) rs).wasNull())
            {
                throw new NullValueException(LOCALISER.msg("RDBMSMapping.IllegalNullValueInColumn",column));
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSManager.UnableToGetParam","Currency","" + param), e);
        }

        return value;
    }

    public void setChar(Object ps, int param, char value)
    {
        try
        {
            ((PreparedStatement) ps).setString(param,new Character(value).toString());                 
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSManager.UnableToSetParam","char","" + value), e);
        }
    }

    public char getChar(Object rs, int param)
    {
        char value;

        try
        {
            value = ((ResultSet) rs).getString(param).charAt(0);
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSManager.UnableToGetParam","char","" + param), e);
        }
        return value;
    }
    
    public void setLocale(Object ps, int param, Locale value)
    {
        try
        {
            ((PreparedStatement) ps).setString(param, value.toString());
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSManager.UnableToSetParam","Locale","" + value), e);
        }
    }

    public Locale getLocale(Object rs, int param)
    {
        Locale value;

        try
        {
            value = I18nUtils.getLocaleFromString(((ResultSet) rs).getString(param));

            if (((ResultSet) rs).wasNull())
            {
                throw new NullValueException(LOCALISER.msg("RDBMSMapping.IllegalNullValueInColumn",column));
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSManager.UnableToGetParam","Locale","" + param), e);
        }

        return value;
    }
    
    public void setString(Object ps, int param, String value)
    {
        try
        {
            if (value == null)
            {
                if (column.isDefaultable() && column.getDefaultValue() != null)
                {
                    ((PreparedStatement) ps).setString(param,column.getDefaultValue().toString().trim());
                }
                else
                {
                    ((PreparedStatement) ps).setNull(param,getTypeInfo().dataType);
                }
            }
            else if( getDatabaseAdapter().isNullEqualsEmptyStrings() )
            {
                if (value != null && value.length() == 0)
                {
                    value = getDatabaseAdapter().getSurrogateForEmptyStrings();
                }                
                ((PreparedStatement) ps).setString(param, value);
            }
            else
            {
                ((PreparedStatement) ps).setString(param, value);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSManager.UnableToSetParam","String","" + value), e);
        }
    }

    public String getString(Object rs, int param)
    {
        String value;

        try
        {
            value = ((ResultSet) rs).getString(param);
            
            if( getDatabaseAdapter().isNullEqualsEmptyStrings() )
            {
                if (value != null && value.equals(getDatabaseAdapter().getSurrogateForEmptyStrings()))
                {
                    value = "";
                }
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSManager.UnableToGetParam","String","" + param), e);
        }

        return value;
    }
    
    public void setBoolean(Object ps, int param, boolean value)
    {
        try
        {
            ((PreparedStatement) ps).setString(param, value ? "Y" : "N");
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSManager.UnableToSetParam","boolean","" + value), e);
        }
    }

    public boolean getBoolean(Object rs, int param)
    {
        boolean value;

        try
        {
                String s = ((ResultSet) rs).getString(param);

                if (s == null)
                {
                    throw new NullValueException(LOCALISER.msg("RDBMSMapping.IllegalNullValueInColumn",column));
                }

                if (s.equals("Y"))
                {
                    value = true;
                }
                else if (s.equals("N"))
                {
                    value = false;
                }
                else
                {
                    throw new JDODataStoreException(LOCALISER.msg("RDBMSMapping.IllegalNullValueInColumn",column));
                }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSManager.UnableToGetParam","boolean","" + param), e);
        }

        return value;
    }
    
    public void setObject(Object ps, int param, Object value)
    {
        try
        {
            if (value == null)
            {
                ((PreparedStatement) ps).setNull(param, getTypeInfo().dataType);
            }
            else
            {
                if (value instanceof Boolean)
                {
                    ((PreparedStatement) ps).setString(param, ((Boolean) value).booleanValue() ? "Y" : "N");
                }
                else if (value instanceof OID)
                {
                    ((PreparedStatement) ps).setString(param, ((OID) value).keyValue());
                }
                else if (value instanceof Locale)
                {
                    ((PreparedStatement) ps).setString(param, ((Locale) value).toString());
                }
                else if (value instanceof java.sql.Time)
                {
                    ((PreparedStatement) ps).setString(param, ((java.sql.Time) value).toString());
                }
                else if (value instanceof java.sql.Date)
                {
                    ((PreparedStatement) ps).setString(param, ((java.sql.Date) value).toString());
                }
                else if (value instanceof java.sql.Timestamp)
                {
                    Calendar cal = new GregorianCalendar(storeMgr.getPMFContext().getPmfConfiguration().DATE_TIMEZONE);

                    // pass the calendar to oracle makes it loses milliseconds
                    ((PreparedStatement) ps).setTimestamp(param, (Timestamp) value, cal);
                }
                else if (value instanceof java.util.TimeZone)
                {
                    ((PreparedStatement) ps).setString(param, ((TimeZone) value).getID());
                }
                else if (value instanceof java.util.Currency)
                {
                    ((PreparedStatement) ps).setString(param, ((Currency) value).toString());
                }
                else if (!(value instanceof String))
                {
                    // This caters for all non-string types. If any more need specific treatment, split them out above.
                    ((PreparedStatement) ps).setString(param, value.toString());
                }
                else
                {
                    ((PreparedStatement) ps).setString(param, ((String) value));
                }
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSManager.UnableToSetParam","Object","" + value), e);
        }
    }

    public Object getObject(Object rs, int param)
    {
        Object value;

        try
        {
            String s = ((ResultSet) rs).getString(param);

            if (s == null)
            {
                value = null;
            }
            else
            {
                if (getJavaTypeMapping().getJavaType().getName().equals(Boolean.class.getName()) )
                {
                    if (s.equals("Y"))
                    {
                        value = Boolean.TRUE;
                    }
                    else if (s.equals("N"))
                    {
                        value = Boolean.FALSE;
                    }
                    else
                    {
                        throw new JDODataStoreException(LOCALISER.msg("RDBMSMapping.IllegalNullValueInColumn",column));
                    }
                }  
                else if (getJavaTypeMapping().getJavaType().getName().equals(java.sql.Time.class.getName()))
                {
                    value = java.sql.Time.valueOf(s);
                }
                else if (getJavaTypeMapping().getJavaType().getName().equals(java.sql.Date.class.getName()))
                {
                    value = java.sql.Date.valueOf(s);
                }
                else if (getJavaTypeMapping().getJavaType().getName().equals(Locale.class.getName()))
                {
                    value = I18nUtils.getLocaleFromString(s);
                }
                else if (getJavaTypeMapping().getJavaType().getName().equals(Timestamp.class.getName()))
                {
                    Calendar cal = new GregorianCalendar(storeMgr.getPMFContext().getPmfConfiguration().DATE_TIMEZONE);

                    value = TimestampRDBMSMapping.stringToTimestamp(s, cal);
                }
                else if (getJavaTypeMapping().getJavaType().getName().equals(TimeZone.class.getName()))
                {
                    value = TimeZone.getTimeZone(s.trim());
                }
                else if (getJavaTypeMapping().getJavaType().getName().equals(Currency.class.getName()))
                {
                    value = Currency.getInstance(s);
                }
                else
                {
    	            value = s;
                }
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSManager.UnableToGetParam","Object","" + param), e);
        }

        return value;
    }    
}