/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution.

Contributors:
2004 Andy Jefferson - coding standards and localisation of messages
    ...
**********************************************************************/
package org.jpox.store.rdbms.mapping;

import org.jpox.store.DatastoreField;
import org.jpox.store.StoreManager;
import org.jpox.store.exceptions.UnsupportedDataTypeException;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.typeinfo.TypeInfo;

/**
 * Column Mapping.
 *
 * @version $Revision: 1.4 $
 **/
public abstract class ColumnMapping extends RDBMSMapping
{
    protected Column column;

    public ColumnMapping(StoreManager storeMgr, JavaTypeMapping mapping)
    {
        super(storeMgr, mapping);
    }

    public boolean isNullable()
    {
        if (column != null)
        {
            return column.isNullable();
        }
        return true;
    }

    public DatastoreField getDatastoreField()
    {
        return column;
    }

    public void setDatastoreField(DatastoreField field)
    {
        this.column = (Column)field;
    }

	/**
	 * Returns the TypeInfo (JDBC SQL type) for columns.
     * This method is capable of returning only one TypeInfo (SQL Type).
     * In that way, it seems adequate for using in 1 column Mapping or many
	 * columns that share the same TypeInfo.
	 * However adequate in most of uses, Mappings using two or more columns
     * should overwrite the initTypeInfo method to appropriate set differents
     * TypeInfo (SQL type) for all the columns
	 * @return The TypeInfo
	 */
    public abstract TypeInfo getTypeInfo();

	/**
	 * Sets the TypeInfo for the columns of the Mapping.
	 * Mappings using two or more columns using different TypeInfo(s) should
     * overwrite this method to appropriate set the TypeInfo (SQL type) for
     * all the columns
	 */
    protected void initTypeInfo()
    {
        TypeInfo typeInfo = getTypeInfo();

        if (typeInfo == null)
        {
            throw new UnsupportedDataTypeException(LOCALISER.msg("RDBMSMapping.NoTypeFoundForColumn",column));
        }

        if( column != null )
        {
            column.setTypeInfo(typeInfo);
        }
    }

    
    /**
     * @return Returns the column.
     */
    public Column getColumn()
    {
        return column;
    }
    
    /* (non-Javadoc)
     * @see org.jpox.store.rdbms.mapping.RDBMSMapping#includeInFetchStatement()
     */
    public boolean includeInFetchStatement()
    {
        return true;
    }

    /* (non-Javadoc)
     * @see org.jpox.store.rdbms.mapping.RDBMSMapping#getInsertionInputParameter()
     */
    public String getInsertionInputParameter()
    {
        return column.getWrapperFunction(Column.WRAPPER_FUNCTION_INSERT);
    }

    /* (non-Javadoc)
     * @see org.jpox.store.rdbms.mapping.RDBMSMapping#getUpdateInputParameter()
     */
    public String getUpdateInputParameter()
    {
        return column.getWrapperFunction(Column.WRAPPER_FUNCTION_UPDATE);
    }

    public boolean equals(Object obj)
    {
        if (this == obj)
        {
            return true;
        }

        if (!(obj instanceof ColumnMapping))
        {
            return false;
        }

        ColumnMapping cm = (ColumnMapping)obj;

        return getClass().equals(cm.getClass()) &&
            storeMgr.equals(cm.storeMgr) &&
            (column == null ? cm.column == null : column.equals(cm.column));
    }

    public int hashCode()
    {
        return storeMgr.hashCode() ^
             (column == null ? 0 : column.hashCode());
    }

    public void addColumn(Column col)
    {
        column = col;
    }
}