/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2004 Andy Jefferson - localised messages
    ...
**********************************************************************/
package org.jpox.store.rdbms.mapping;

import java.math.BigDecimal;
import java.math.BigInteger;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Types;

import javax.jdo.JDODataStoreException;

import org.jpox.store.DatastoreField;
import org.jpox.store.OID;
import org.jpox.store.StoreManager;
import org.jpox.store.exceptions.NullValueException;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.typeinfo.TypeInfo;
import org.jpox.util.JPOXLogger;

/**
 * Mapping of a Decimal RDBMS type.
 * 
 * @version $Revision: 1.5 $
 */
public class DecimalRDBMSMapping extends ColumnMapping
{
    private static final int INT_MAX_DECIMAL_DIGITS = 10;
    private static final int LONG_MAX_DECIMAL_DIGITS = 19;

    /**
     * @param storeMgr Store Manager
     * @param mapping The java type mapping
     */
    protected DecimalRDBMSMapping(StoreManager storeMgr, JavaTypeMapping mapping)
    {
        super(storeMgr, mapping);
    }

    /**
     * Constructor.
     * @param mapping Java type mapping
     * @param storeMgr Store Manager
     * @param field Field to be mapped
     */
    public DecimalRDBMSMapping(JavaTypeMapping mapping, StoreManager storeMgr, DatastoreField field)
    {
		super(storeMgr, mapping);
		column = (Column) field;
		initialize();
	}

    /**
     * Initialise the mapping, setting any default precision.
     */
    private void initialize()
    {
        // If the column has no precision specified, set its size.
        // If the user has already set their precision we do nothing here since they
        // want to control it.
        if (column != null && column.getColumnOptions().getPrecision() == null)
        {
            // In case the default DECIMAL precision is less than the number of
            // digits we need, set it manually
            if (getJavaTypeMapping().getJavaType().getName().equals(Integer.class.getName()))
            {
                column.getColumnOptions().setPrecision(new Integer(INT_MAX_DECIMAL_DIGITS));
                column.checkDecimal();
            }
            else
            {
                column.getColumnOptions().setPrecision(new Integer(Math.min(getTypeInfo().precision, LONG_MAX_DECIMAL_DIGITS)));
                column.checkDecimal();
            }
        }
        initTypeInfo();
    }

    public TypeInfo getTypeInfo()
    {
        return getDatabaseAdapter().getTypeInfo(Types.DECIMAL);
    }
    
    public DatastoreField getDatastoreField()
    {
        return column;
    }
    
    public void setDouble(Object ps, int param, double value)
    {
        try
        {
            ((PreparedStatement) ps).setDouble(param, value);
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSManager.UnableToSetParam","double","" + value), e);
        }
    }
    
    public void setFloat(Object ps, int param, float value)
    {
        try
        {
            ((PreparedStatement) ps).setDouble(param, value);
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSManager.UnableToGetParam","float","" + value), e);
        }
    }    
    
    public void setInt(Object ps, int param, int value)
    {
        try
        {
            ((PreparedStatement) ps).setInt(param, value);
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSMapping.UnableToSetParam","int","" + value), e);
        }
    }

    public double getDouble(Object rs, int param)
    {
        double value;

        try
        {
            value = ((ResultSet) rs).getDouble(param);

            if (((ResultSet) rs).wasNull())
            {
                throw new NullValueException(LOCALISER.msg("RDBMSMapping.IllegalNullValueInColumn",column));
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSManager.UnableToGetParam","double","" + param), e);
        }

        return value;
    }
    
    public float getFloat(Object rs, int param)
    {
        float value;

        try
        {
            value = (float) ((ResultSet) rs).getDouble(param);

            if (((ResultSet) rs).wasNull())
            {
                throw new NullValueException(LOCALISER.msg("RDBMSMapping.IllegalNullValueInColumn",column));
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSManager.UnableToSetParam","float","" + param), e);
        }

        return value;
    }
    
    public int getInt(Object rs, int param)
    {
        int value;

        try
        {
            value = ((ResultSet) rs).getInt(param);

            if (((ResultSet) rs).wasNull())
            {
                throw new NullValueException(LOCALISER.msg("RDBMSMapping.IllegalNullValueInColumn",column));
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSMapping.UnableToGetParam","int","" + param), e);
        }

        return value;
    }

    public void setLong(Object ps, int param, long value)
    {
        try
        {
            ((PreparedStatement) ps).setLong(param, value);
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSMapping.UnableToSetParam","long","" + value), e);
        }
    }

    public long getLong(Object rs, int param)
    {
        long value;

        try
        {
            value = ((ResultSet) rs).getLong(param);

            if (((ResultSet) rs).wasNull())
            {
                throw new NullValueException(LOCALISER.msg("RDBMSMapping.IllegalNullValueInColumn",column));
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSMapping.UnableToGetParam","long","" + param), e);
        }

        return value;
    }
    
    public void setObject(Object ps, int param, Object value)
    {
        try
        {
            if (value == null)
            {
                if (column.isDefaultable() && column.getDefaultValue() != null)
                {
                    ((PreparedStatement) ps).setInt(param, Integer.valueOf(column.getDefaultValue().toString()).intValue());
                }
                else
                {
                    ((PreparedStatement) ps).setNull(param, getTypeInfo().dataType);
                }
            }
            else
            {
                if (value instanceof OID)
                {
                    JPOXLogger.RDBMS.debug(((OID)value).keyValue());
                    ((PreparedStatement) ps).setInt(param, (new Long(((OID)value).keyValue())).intValue());
                }
                else if( value instanceof Integer) 
                {
                    ((PreparedStatement) ps).setBigDecimal(param, new BigDecimal(((Integer)value).intValue()));
                }
                else if( value instanceof Long) 
                {
                    ((PreparedStatement) ps).setBigDecimal(param, new BigDecimal(((Long)value).longValue()));
                }
                else if( value instanceof BigDecimal) 
                {
                    ((PreparedStatement) ps).setBigDecimal(param, (BigDecimal)value);
                }
                else if( value instanceof Float )
                {
                    ((PreparedStatement) ps).setDouble(param, ((Float)value).doubleValue());
                }            
                else if( value instanceof Double )
                {
                    ((PreparedStatement) ps).setDouble(param, ((Double)value).doubleValue());
                }                
                else if( value instanceof BigInteger) 
                {
                    ((PreparedStatement) ps).setBigDecimal(param, new BigDecimal((BigInteger)value));
                }                
                else
                {
                    ((PreparedStatement) ps).setInt(param, ((Integer)value).intValue());
                }
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSMapping.UnableToSetParam","Object","" + value), e);
        }
    }

    public Object getObject(Object rs, int param)
    {
        Object value;

        try
        {
            if( getJavaTypeMapping().getJavaType().getName().equals(Integer.class.getName()) )
            {
                value = ((ResultSet) rs).getBigDecimal(param);                
	            value = value == null ? null : new Integer(((BigDecimal)value).toBigInteger().intValue());
            }
            else if( getJavaTypeMapping().getJavaType().getName().equals(Long.class.getName()) )
            {
                value = ((ResultSet) rs).getBigDecimal(param);                
	            value = value == null ? null : new Long(((BigDecimal)value).toBigInteger().longValue());
            }
            else if( getJavaTypeMapping().getJavaType().getName().equals(BigInteger.class.getName()) )
            {
                value = ((ResultSet) rs).getBigDecimal(param);                
	            value = value == null ? null : ((BigDecimal)value).toBigInteger();
            }
            else if( getJavaTypeMapping().getJavaType().getName().equals(BigDecimal.class.getName()) )
            {
                value = ((ResultSet) rs).getBigDecimal(param);                
            }            
            else if( getJavaTypeMapping().getJavaType().getName().equals(Float.class.getName()) )
            {
                double d = ((ResultSet) rs).getDouble(param);                
                value = ((ResultSet) rs).wasNull() ? null : new Float(d);
            }
            else if( getJavaTypeMapping().getJavaType().getName().equals(Double.class.getName()) )
            {
                double d = ((ResultSet) rs).getDouble(param);                
                value = ((ResultSet) rs).wasNull() ? null : new Double(d);
            }            
            else
            {
	            int i = ((ResultSet) rs).getInt(param);
	            value = ((ResultSet) rs).wasNull() ? null : new Integer(i);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSMapping.UnableToGetParam","Object","" + param), e);
        }

        return value;
    }    
}