/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2004 Andy Jefferson - localised messages
    ...
**********************************************************************/
package org.jpox.store.rdbms.mapping;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Types;

import javax.jdo.JDODataStoreException;

import org.jpox.store.DatastoreField;
import org.jpox.store.StoreManager;
import org.jpox.store.exceptions.NullValueException;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.typeinfo.TypeInfo;

/**
 * Mapping of a DOUBLE RDBMS type.
 * 
 * @version $Revision: 1.3 $
 */
public class DoubleRDBMSMapping extends ColumnMapping
{
    /**
     * @param storeMgr Store Manager
     * @param mapping The java type mapping
     */
    protected DoubleRDBMSMapping(StoreManager storeMgr, JavaTypeMapping mapping)
    {
        super(storeMgr, mapping);
    }
    
    /**
     * Constructor.
     * @param mapping The java type mapping
     * @param storeMgr Store Manager
     * @param field Field being mapped
     */
    public DoubleRDBMSMapping(JavaTypeMapping mapping, StoreManager storeMgr, DatastoreField field)
    {
		super(storeMgr, mapping);
		column = (Column) field;
		initialize();
	}

    private void initialize()
    {
        if (column != null)
        {
            column.checkPrimitive();
        }
		initTypeInfo();
    }
    
    public TypeInfo getTypeInfo()
    {
        return getDatabaseAdapter().getTypeInfo(Types.DOUBLE);
    }
    
    public DatastoreField getDatastoreField()
    {
        return column;
    }
    
    public void setInt(Object ps, int param, int value)
    {
        try
        {
            ((PreparedStatement) ps).setDouble(param, value);
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSManager.UnableToSetParam","int","" + value), e);
        }
    }

    public int getInt(Object rs, int param)
    {
        int value;

        try
        {
            value = (int)((ResultSet) rs).getDouble(param);

            if (((ResultSet) rs).wasNull())
            {
                throw new NullValueException(LOCALISER.msg("RDBMSMapping.IllegalNullValueInColumn",column));
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSManager.UnableToGetParam","int","" + param), e);
        }

        return value;
    }
    
    public void setLong(Object ps, int param, long value)
    {
        try
        {
            ((PreparedStatement) ps).setLong(param, value);
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSManager.UnableToSetParam","long","" + value), e);
        }
    }

    public long getLong(Object rs, int param)
    {
        long value;

        try
        {
            value = ((ResultSet) rs).getLong(param);

            if (((ResultSet) rs).wasNull())
            {
                throw new NullValueException(LOCALISER.msg("RDBMSMapping.IllegalNullValueInColumn",column));
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSManager.UnableToGetParam","long","" + param), e);
        }

        return value;
    }
    
    public void setDouble(Object ps, int param, double value)
    {
        try
        {
            ((PreparedStatement) ps).setDouble(param, value);
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSManager.UnableToSetParam","double","" + value), e);
        }
    }

    public double getDouble(Object rs, int param)
    {
        double value;

        try
        {
            value = ((ResultSet) rs).getDouble(param);

            if (((ResultSet) rs).wasNull())
            {
                throw new NullValueException(LOCALISER.msg("RDBMSMapping.IllegalNullValueInColumn",column));
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSManager.UnableToGetParam","double","" + param), e);
        }

        return value;
    }

    public float getFloat(Object rs, int param)
    {
        float value;

        try
        {
            value = (float) ((ResultSet) rs).getDouble(param);

            if (((ResultSet) rs).wasNull())
            {
                throw new NullValueException(LOCALISER.msg("RDBMSMapping.IllegalNullValueInColumn",column));
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSManager.UnableToSetParam","float","" + param), e);
        }

        return value;
    }
    
    public void setFloat(Object ps, int param, float value)
    {
        try
        {
            ((PreparedStatement) ps).setDouble(param, value);
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSManager.UnableToGetParam","float","" + value), e);
        }
    }
    
    public void setObject(Object ps, int param, Object value)
    {
        try
        {
            if (value == null)
            {
                ((PreparedStatement) ps).setNull(param, getTypeInfo().dataType);
            }
            else if( value instanceof Integer )
            {
                ((PreparedStatement) ps).setDouble(param, ((Integer)value).doubleValue());
            }            
            else if( value instanceof Long )
            {
                ((PreparedStatement) ps).setDouble(param, ((Long)value).doubleValue());
            }            
            else if( value instanceof Float )
            {
                ((PreparedStatement) ps).setDouble(param, ((Float)value).doubleValue());
            }            
            else
            {
                ((PreparedStatement) ps).setDouble(param, ((Double)value).doubleValue());
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSManager.UnableToSetParam","Object","" + value), e);
        }
    }

    public Object getObject(Object rs, int param)
    {
        Object value;

        try
        {
            double d = ((ResultSet) rs).getDouble(param);
            if( getJavaTypeMapping().getJavaType().getName().equals(Integer.class.getName()) )
            {
                value = ((ResultSet) rs).wasNull() ? null : new Integer((int)d);
            }
            else if( getJavaTypeMapping().getJavaType().getName().equals(Long.class.getName()) )
            {
                value = ((ResultSet) rs).wasNull() ? null : new Long((long)d);
            }
            else if( getJavaTypeMapping().getJavaType().getName().equals(Float.class.getName()) )
            {
                value = ((ResultSet) rs).wasNull() ? null : new Float(d);
            }
            else
            {
                value = ((ResultSet) rs).wasNull() ? null : new Double(d);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSManager.UnableToGetParam","Object","" + param), e);
        }

        return value;
    }    
}