/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2004 Brendan de Beer - primitive array support
2004 Andy Jefferson - localised messages
    ...
**********************************************************************/
package org.jpox.store.rdbms.mapping;

import java.io.IOException;
import java.io.Serializable;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Types;

import javax.jdo.JDODataStoreException;

import org.jpox.store.DatastoreField;
import org.jpox.store.StoreManager;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.datatype.BlobImpl;
import org.jpox.store.rdbms.typeinfo.TypeInfo;
import org.jpox.util.TypeConversionHelper;

/**
 * Mapping of a LONGVARBINARY RDBMS type.
 * 
 * @version $Revision: 1.4 $
 */
public class LongVarBinaryRDBMSMapping extends ColumnMapping
{
    /**
     * @param storeMgr Store Manager
     * @param mapping Java type mapping
     */
    protected LongVarBinaryRDBMSMapping(StoreManager storeMgr, JavaTypeMapping mapping)
    {
        super(storeMgr, mapping);
    }
    
    /**
     * Constructor.
     * @param mapping Java type mapping
     * @param storeMgr Store Manager
     * @param field Field to be mapped
     */
    public LongVarBinaryRDBMSMapping(JavaTypeMapping mapping, StoreManager storeMgr, DatastoreField field)
    {
		super(storeMgr, mapping);
		column = (Column) field;
		initialize();
	}

    private void initialize()
    {
		initTypeInfo();
    }
    
    public TypeInfo getTypeInfo()
    {
        return getDatabaseAdapter().getTypeInfo(Types.LONGVARBINARY);
    }
    
    public void setObject(Object ps, int param, Object value)
    {
        if (value == null)
        {
            try
            {
                ((PreparedStatement) ps).setBytes(param,null);
            }
            catch (SQLException sqle)
            {
				throw new JDODataStoreException(LOCALISER.msg("RDBMSManager.UnableToSetParam","Object","" + value),sqle);
            }
        }
        else
        {
            try
            {
                if( value instanceof boolean[] )
                {
	                byte[] data = TypeConversionHelper.getByteArrayFromBooleanArray(value);
	                ((PreparedStatement) ps).setBytes(param, data);
                }
                else if( value instanceof char[] )
                {
	                byte[] data = TypeConversionHelper.getByteArrayFromCharArray(value);
	                ((PreparedStatement) ps).setBytes(param, data);
                }                
                else if( value instanceof double[] )
                {
	                byte[] data = TypeConversionHelper.getByteArrayFromDoubleArray(value);
	                ((PreparedStatement) ps).setBytes(param, data);
                }                
                else if( value instanceof float[] )
                {
	                byte[] data = TypeConversionHelper.getByteArrayFromFloatArray(value);
	                ((PreparedStatement) ps).setBytes(param, data);
                }                
                else if( value instanceof int[] )
                {
	                byte[] data = TypeConversionHelper.getByteArrayFromIntArray(value);
	                ((PreparedStatement) ps).setBytes(param, data);
                }                
                else if( value instanceof long[] )
                {
	                byte[] data = TypeConversionHelper.getByteArrayFromLongArray(value);
	                ((PreparedStatement) ps).setBytes(param, data);
                }                
                else if( value instanceof short[] )
                {
	                byte[] data = TypeConversionHelper.getByteArrayFromShortArray(value);
	                ((PreparedStatement) ps).setBytes(param, data);
                }                
                else if( value instanceof byte[] )
                {
	                ((PreparedStatement) ps).setBytes(param, (byte[]) value);
                }
                else
                {
	                if (! (value instanceof Serializable))
	                {
	                    throw new JDODataStoreException(LOCALISER.msg("RDBMSMapping.NotSerializable",value.getClass().getName()));
	                }
	                BlobImpl b = new BlobImpl(value);
	                ((PreparedStatement) ps).setBytes(param,b.getBytes(0,(int)b.length()));
                }
            }
            catch (Exception e)
            {
				throw new JDODataStoreException(LOCALISER.msg("RDBMSManager.UnableToSetParam","Object","" + value),e);
            }
        }
    }

    public Object getObject(Object rs, int param)
    {
        byte[] bytes = null;
        try
        {
            bytes = ((ResultSet) rs).getBytes(param);
        }
        catch (SQLException sqle)
        {
			throw new JDODataStoreException(LOCALISER.msg("RDBMSMapping.UnableToGetParam","Object", "" + param),sqle);
        }
        if (bytes == null)
        {
            return null;
        }
 
        if( getJavaTypeMapping().getJavaType().getName().equals(boolean[].class.getName()))
        {
            return TypeConversionHelper.getBooleanArrayfromByteArray(bytes);
        }
        else if( getJavaTypeMapping().getJavaType().getName().equals(char[].class.getName()))
        {
            return TypeConversionHelper.getCharArrayFromByteArray(bytes);
        }
        else if( getJavaTypeMapping().getJavaType().getName().equals(double[].class.getName()))
        {
            return TypeConversionHelper.getDoubleArrayFromByteArray(bytes);
        }
        else if( getJavaTypeMapping().getJavaType().getName().equals(float[].class.getName()))
        {
            return TypeConversionHelper.getFloatArrayFromByteArray(bytes);
        }
        else if( getJavaTypeMapping().getJavaType().getName().equals(int[].class.getName()))
        {
            return TypeConversionHelper.getIntArrayFromByteArray(bytes);
        }
        else if( getJavaTypeMapping().getJavaType().getName().equals(long[].class.getName()))
        {
            return TypeConversionHelper.getLongArrayFromByteArray(bytes);
        }
        else if( getJavaTypeMapping().getJavaType().getName().equals(short[].class.getName()))
        {
            return TypeConversionHelper.getShortArrayFromByteArray(bytes);
        }
        else if( getJavaTypeMapping().getJavaType().getName().equals(byte[].class.getName()))
        {
            return bytes;
        }        
        else
        {
	        BlobImpl blob;
	        try
	        {
	            blob = new BlobImpl(bytes);
	            return blob.getObject();
	        }
	        catch (IOException ioe)
	        {
				throw new JDODataStoreException(LOCALISER.msg("RDBMSMapping.UnableToGetParam","Object","" + param),ioe);
	        }
        }
    }
}