/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2004 Andy Jefferson - localised messages
2005 Andy Jefferson - added control over timezone to use
    ...
**********************************************************************/
package org.jpox.store.rdbms.mapping;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Time;
import java.sql.Timestamp;
import java.sql.Types;
import java.util.Calendar;
import java.util.Date;
import java.util.GregorianCalendar;

import javax.jdo.JDODataStoreException;

import org.jpox.store.DatastoreField;
import org.jpox.store.StoreManager;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.typeinfo.TypeInfo;

/**
 * Mapping of a TIMESTAMP RDBMS type.
 * 
 * @version $Revision: 1.5 $
 */
public class TimestampRDBMSMapping extends ColumnMapping
{
    /** Used to zero-fill the fractional seconds to nine digits. */
    private static final String ZEROES = "000000000";

    /**
     * Constructor
     * @param storeMgr Store Manager
     * @param mapping Java type mapping
     */
    protected TimestampRDBMSMapping(StoreManager storeMgr, JavaTypeMapping mapping)
    {
        super(storeMgr, mapping);
    }

    /**
     * Constructor.
     * @param mapping Java type mapping
     * @param storeMgr Store Manager
     * @param field Field to be mapped
     */
    public TimestampRDBMSMapping(JavaTypeMapping mapping, StoreManager storeMgr, DatastoreField field)
    {
        super(storeMgr, mapping);
        column = (Column) field;
        initialize();
    }

    private void initialize()
    {
        if (column != null)
        {
            column.checkPrimitive();
        }
        initTypeInfo();
    }

    public TypeInfo getTypeInfo()
    {
        return getDatabaseAdapter().getTypeInfo(Types.TIMESTAMP);
    }

    public DatastoreField getDatastoreField()
    {
        return column;
    }

    public void setObject(Object ps, int param, Object value)
    {
        try
        {
            Calendar cal = new GregorianCalendar(storeMgr.getPMFContext().getPmfConfiguration().DATE_TIMEZONE);

            if (value == null)
            {
                ((PreparedStatement) ps).setNull(param, getTypeInfo().dataType);
            }
            else if (value instanceof java.util.Date)
            {
                //pass the calendar to oracle makes it loses milliseconds
                //ps.setTimestamp(param, (Timestamp)value,cal);
                ((PreparedStatement) ps).setTimestamp(param, new Timestamp(((java.util.Date) value).getTime()), cal);
            }
            else if (value instanceof Time)
            {
                //pass the calendar to oracle makes it loses milliseconds
                //ps.setTimestamp(param, (Timestamp)value,cal);
                ((PreparedStatement) ps).setTimestamp(param, new Timestamp(((Time) value).getTime()), cal);
            }
            else if (value instanceof java.sql.Date)
            {
                //pass the calendar to oracle makes it loses milliseconds
                //ps.setTimestamp(param, (Timestamp)value,cal);
                ((PreparedStatement) ps).setTimestamp(param, new Timestamp(((java.sql.Date) value).getTime()), cal);
            }
            else
            {
                //pass the calendar to oracle makes it loses milliseconds
                //ps.setTimestamp(param, (Timestamp)value,cal);
                ((PreparedStatement) ps).setTimestamp(param, (Timestamp) value, cal);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RDBMSMapping.UnableToSetParam", "Timestamp", "" + value), e);
        }
    }

    protected Timestamp getTimestamp(Object rs, int param)
    {
        Timestamp value;

        Calendar cal = new GregorianCalendar(storeMgr.getPMFContext().getPmfConfiguration().DATE_TIMEZONE);

        try
        {
            //pass the calendar to oracle makes it loses milliseconds
            //value = rs.getTimestamp(param,cal);
            value = ((ResultSet) rs).getTimestamp(param, cal);
        }
        catch (SQLException e)
        {
            try
            {
                String s = ((ResultSet) rs).getString(param);
                if (((ResultSet) rs).wasNull())
                {
                    value = null;
                }
                else
                {
                    value = s == null ? null : stringToTimestamp(s, cal);
                }
            }
            catch (SQLException nestedEx)
            {
                throw new JDODataStoreException(LOCALISER.msg("RDBMSMapping.UnableToGetParam", "Timestamp", "" + param), nestedEx);
            }
        }

        return value;
    }

    public Object getObject(Object rs, int param)
    {
        Timestamp value = getTimestamp(rs, param);

        if (value == null)
        {
            return null;
        }
        else if (getJavaTypeMapping().getJavaType().getName().equals(java.util.Date.class.getName()))
        {
            return new Date(getDatabaseAdapter().getAdapterTime(value));
        }
        else if (getJavaTypeMapping().getJavaType().getName().equals(java.sql.Date.class.getName()))
        {
            return new java.sql.Date(getDatabaseAdapter().getAdapterTime(value));
        }
        else if (getJavaTypeMapping().getJavaType().getName().equals(java.sql.Time.class.getName()))
        {
            return new Time(getDatabaseAdapter().getAdapterTime(value));
        }
        else
        {
            return new Timestamp(getDatabaseAdapter().getAdapterTime(value));
        }
    }

    /**
     * Converts a string in JDBC timestamp escape format to a
     * <tt>java.sql.Timestamp</tt> object using the "UTC" time zone.
     *
     * @param s     Timestamp string in format
     *              <tt>yyyy-mm-dd hh:mm:ss.fffffffff</tt>.
     * @return  Corresponding <tt>java.sql.Timestamp</tt> value.
     * @exception java.lang.IllegalArgumentException
     *              If the given argument does not have the format
     *              <tt>yyyy-mm-dd hh:mm:ss.fffffffff</tt>.
     * @see java.sql.Timestamp
     */
    public static Timestamp stringToTimestamp(String s, Calendar cal)
    {
        String formatError = "Bad timestamp format: \"" + s + "\", must be yyyy-mm-dd hh:mm:ss.fffffffff";

        /* Split the string into date and time components. */
        s = s.trim();
        int dividingSpace = s.indexOf(' ');

        if (dividingSpace <= 0)
        {
            throw new java.lang.IllegalArgumentException(formatError);
        }

        String date_s = s.substring(0, dividingSpace);
        String time_s = s.substring(dividingSpace + 1);

        /* Parse the date. */
        int firstDash = date_s.indexOf('-');
        int secondDash = date_s.indexOf('-', firstDash + 1);
        int firstColon = time_s.indexOf(':');
        int secondColon = time_s.indexOf(':', firstColon + 1);
        int period = time_s.indexOf('.', secondColon + 1);

        /* Convert the date. */
        if (firstDash <= 0 || secondDash <= 0 || secondDash >= date_s.length() - 1)
        {
            throw new java.lang.IllegalArgumentException(formatError);
        }

        int year = Integer.parseInt(date_s.substring(0, firstDash));
        int month = Integer.parseInt(date_s.substring(firstDash + 1, secondDash));
        int day = Integer.parseInt(date_s.substring(secondDash + 1));

        /* Convert the time; default missing nanos. */
        if (firstColon <= 0 || secondColon <= 0 || secondColon >= time_s.length() - 1)
        {
            throw new java.lang.IllegalArgumentException(formatError);
        }

        int hour = Integer.parseInt(time_s.substring(0, firstColon));
        int minute = Integer.parseInt(time_s.substring(firstColon + 1, secondColon));
        int second;
        int nanos = 0;

        if (period < 0)
        {
            second = Integer.parseInt(time_s.substring(secondColon + 1));
        }
        else if (period == 0 || period >= time_s.length() - 1)
        {
            throw new java.lang.IllegalArgumentException(formatError);
        }
        else
        {
            second = Integer.parseInt(time_s.substring(secondColon + 1, period));
            String nanos_s = time_s.substring(period + 1);

            if (nanos_s.length() > ZEROES.length())
            {
                throw new java.lang.IllegalArgumentException(formatError);
            }

            nanos_s += ZEROES.substring(0, ZEROES.length() - nanos_s.length());
            nanos = Integer.parseInt(nanos_s);
        }

        cal.set(Calendar.ERA, GregorianCalendar.AD);
        cal.set(Calendar.YEAR, year);
        cal.set(Calendar.MONTH, month - 1); // Months are zero based in Calendar
        cal.set(Calendar.DATE, day);
        cal.set(Calendar.HOUR_OF_DAY, hour);
        cal.set(Calendar.MINUTE, minute);
        cal.set(Calendar.SECOND, second);

        Timestamp ts = new Timestamp(cal.getTime().getTime());
        ts.setNanos(nanos);

        return ts;
    }

    /**
     * Formats a timestamp in JDBC timestamp escape format using the "UTC" time
     * zone.
     * @param ts    The timestamp to be formatted.
     * @return  A String in <tt>yyyy-mm-dd hh:mm:ss.fffffffff</tt> format.
     * @see java.sql.Timestamp
     */
    public static String timestampToString(Timestamp ts, Calendar cal)
    {
        cal.setTime(ts);

        int year = cal.get(Calendar.YEAR);
        int month = cal.get(Calendar.MONTH) + 1; // Months are zero based in Calendar
        int day = cal.get(Calendar.DATE);
        int hour = cal.get(Calendar.HOUR_OF_DAY);
        int minute = cal.get(Calendar.MINUTE);
        int second = cal.get(Calendar.SECOND);

        String yearString = Integer.toString(year);
        String monthString = month < 10 ? "0" + month : Integer.toString(month);
        String dayString = day < 10 ? "0" + day : Integer.toString(day);
        String hourString = hour < 10 ? "0" + hour : Integer.toString(hour);
        String minuteString = minute < 10 ? "0" + minute : Integer.toString(minute);
        String secondString = second < 10 ? "0" + second : Integer.toString(second);
        String nanosString = Integer.toString(ts.getNanos());

        if (ts.getNanos() != 0)
        {
            // Add leading zeroes
            nanosString = ZEROES.substring(0, ZEROES.length() - nanosString.length()) + nanosString;

            // Truncate trailing zeroes
            int truncIndex = nanosString.length() - 1;
            while (nanosString.charAt(truncIndex) == '0')
            {
                --truncIndex;
            }

            nanosString = nanosString.substring(0, truncIndex + 1);
        }

        return (yearString + "-" + monthString + "-" + dayString + " " + hourString + ":" + minuteString + ":" + secondString + "." + nanosString);
    }
}