/**********************************************************************
Copyright (c) 12/February/2004 Erik Bengtson and others.
All rights reserved.   This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
    ...
**********************************************************************/
package org.jpox.store.rdbms.query;

import java.util.Iterator;

import javax.jdo.JDOFatalInternalException;

import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.StatementText;
import org.jpox.store.expression.BooleanCharColumnExpression;
import org.jpox.store.expression.BooleanExpression;
import org.jpox.store.expression.CharacterExpression;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.rdbms.adapter.DatabaseAdapter;

/**
 * Representation of a Query Statement in Oracle 9 or upper.
 *
 * @version $Revision: 1.4 $
 **/ 
public class Oracle99QueryStatement extends QueryStatement
{
    /**
     * The system property that selects the "linguistic definition" to be used
     * for native language sorting of String fields.  This is the string
     * "org.jpox.oracle.nlsSortOrder".  The default value is "LATIN".
     * A value of "BINARY" disables native language sorting.
     */
    public static final String NLS_SORT_ORDER_PROPERTY = "org.jpox.oracle.nlsSortOrder";

    private final String nlsSortOrder = System.getProperty(NLS_SORT_ORDER_PROPERTY, "LATIN").toUpperCase();

    public Oracle99QueryStatement(DatastoreContainerObject initialTable)
    {
        super(initialTable);
    }

    public Oracle99QueryStatement(DatastoreContainerObject initialTable, DatastoreIdentifier initialRangeVar)
    {
        super(initialTable, initialRangeVar);
    }

    /**
     * Method to set the ordering for the statement.
     **/
    public void setOrdering(ScalarExpression[] exprs, boolean[] descending)
    {
        assertNotFrozen();

        boolean needsSelect = ((DatabaseAdapter)dba).includeOrderByColumnsInSelect();

        orderByList = new StatementText();

        for (int i = 0; i < exprs.length; ++i)
        {
            if (i > 0)
            {
                orderByList.append(',');
            }

            if (exprs[i] instanceof CharacterExpression && !nlsSortOrder.equals("BINARY"))
            {
                orderByList.append("NLSSORT(").append(exprs[i]).append(", 'NLS_SORT = ").append(nlsSortOrder).append("')");
            }
            else
            {
                if (exprs[i] instanceof BooleanExpression)
                {
                    if (false == (exprs[i] instanceof BooleanCharColumnExpression))
                    {
                        throw new JDOFatalInternalException(LOCALISER.msg("QueryStatement.OracleBooleanExpression"));
                    }

					orderByList.append(stripTruthTest((BooleanCharColumnExpression)exprs[i]));
                }
				else
				{
					orderByList.append(removeTableAlias(exprs[i]));
				}
            }

            if (descending[i])
            {
                orderByList.append(" DESC");
            }

            if (needsSelect)
            {
	            // go thought union
	            Iterator iterator = union.iterator();
	            while (iterator.hasNext())
	            {
	                QueryStatement qs = ((QueryStatement)iterator.next());
	                
	                qs.selectScalarExpression(exprs[i]);
	            }
	            selectScalarExpression(exprs[i]);
            }            
        }
    }

	/**
	 * Return the StatementText as a String with the truth table alias truncanted.
	 * @param   expr  The ScalarExpression to remove the table alias
	 * @return  The StatementText as a String with the alias removed.
	 */
	private static String removeTableAlias(ScalarExpression expr)
	{
		String text = expr.toStatementText().toString();

		if (null != text)
		{
			text = text.substring(text.lastIndexOf(".")+1);
		}

		return text;
	}
    /**
     * Return the StatementText as a String with the truth test stripped off.
     * This is a hack to account for the fact that Oracle does't support boolean
     * expressions in ORDER BY clauses.
     * @param   expr  The BooleanCharColumnExpression to strip to truth test
     * @return  The StatementText as a String with the truth test stripped off.
     */
    private static String stripTruthTest(BooleanCharColumnExpression expr)
    {
        String text = expr.toStatementText().toString();

        if (null != text)
        {
            text = text.substring(0, text.lastIndexOf(" = '"));
        }

        return text;
    }
}
