/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - coding standards
2004 Andy Jefferson - conversion to use Logger
    ...
**********************************************************************/
package org.jpox.store.rdbms.request;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Iterator;
import java.util.List;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOOptimisticVerificationException;

import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.metadata.ClassMetaData;
import org.jpox.metadata.FieldMetaData;
import org.jpox.metadata.IdentityType;
import org.jpox.store.OID;
import org.jpox.store.StatementExpressionIndex;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.MappingCallbacks;
import org.jpox.store.mapping.MappingConsumer;
import org.jpox.store.mapping.VersionMapping;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.fieldmanager.ParameterSetter;
import org.jpox.store.rdbms.mapping.RDBMSMapping;
import org.jpox.store.rdbms.table.ClassTable;
import org.jpox.util.JPOXLogger;

/**
 * Class to provide a means of deletion of records from a data store.
 * Extends basic request class implementing the execute method to do a JDBC
 * delete operation.
 *
 * @version $Revision: 1.6 $ 
 **/
public class DeleteRequest extends Request
{
    private static final int IDPARAMNUMBER = 1;
    
    private MappingCallbacks[] callbacks;

    private String versionStmt;
    
    private String deleteStmt;

    /** one StatementExpressionIndex for each field **/
    private StatementExpressionIndex[] statementExpressionIndex;
       
    private int[] pkfieldsToBeProvided;
    
    private ClassMetaData cmd;

    /**
     * Constructor, taking the table. Uses the structure of the datastore
     * table to build a basic query.
     * @param table The Class Table representing the datastore table to
     *              delete.
     **/
    public DeleteRequest(ClassTable table)
    {
        super(table);
    }

    private void initalize(StateManager sm)
    {
        cmd = sm.getPersistenceManager().getPMFContext().getMetaDataManager().getMetaDataForClass(sm.getObject().getClass());        

        int paramIndex = IDPARAMNUMBER;
        
        // Construct the statement for the Delete
        StringBuffer where = new StringBuffer();

        if (table.getIdentityType() == IdentityType.DATASTORE)
        {
            Iterator iterator = key.getColumns().iterator();
            where.append(((Column)iterator.next()).getName().toString());
            where.append(" = ");             
            where.append("?");           
            paramIndex++;
        }


        DeleteMappingConsumer dmconsumer = new DeleteMappingConsumer(where,paramIndex);
        table.providePrimaryKeyMappings(dmconsumer);
        table.provideDatastoreIdMappings(dmconsumer);
        table.provideNonPrimaryKeyMappings(dmconsumer); // to compute callbacks
        pkfieldsToBeProvided = dmconsumer.getPrimaryKeysToBeProvided();
        statementExpressionIndex = dmconsumer.getStatementExpressionIndex();
        callbacks = (MappingCallbacks[])dmconsumer.getMappingCallBacks().toArray(new MappingCallbacks[dmconsumer.getMappingCallBacks().size()]);

        deleteStmt = "DELETE FROM " + table.toString() + " WHERE " + where;

        // version query
        if (table.getVersionMapping() != null)
        {
            versionStmt = 
                "SELECT " +
                table.getVersionMapping().getDataStoreMapping(0).getDatastoreField() +
                " FROM " + table.toString() +   " WHERE " + where;
        }
        else
        {
            versionStmt = null;
        }
    }

    /**
     * Method performing the deletion of the record from the datastore. Takes
     * the constructed deletion query and populates with the specific record
     * information.
     *
     * @param sm The state manager for the record to be deleted.
     **/
    public void execute(StateManager sm)
    {
        if( deleteStmt == null )
        {
            initalize(sm);
        }
        // Execute any mapping actions before the delete (things like clearing
        // the links to other objects that we have relationships with).
        for (int i = 0; i < callbacks.length; ++i)
        {
            callbacks[i].preDelete(sm);
        }

        PersistenceManager pm = sm.getPersistenceManager();
        try
        {
            Connection conn = pm.getConnection(true);

            try
            {
                PreparedStatement ps;
                if (versionStmt != null)
                {
                    ps = conn.prepareStatement(versionStmt);
                    try
                    {
                        // version query
                        if (pm.currentTransaction().getOptimistic() && 
                            table.getVersionMapping() != null)
                        {
                            // datastore identity
                            if (sm.getInternalObjectId() instanceof OID)
                            {
                                int[] paramNumber = { IDPARAMNUMBER };
                                                        
                                table.getDataStoreObjectIdMapping().setObject(pm, ps, paramNumber, sm.getInternalObjectId());
                            }
                            // application identity
                            else
                            {
                                sm.provideFields(pkfieldsToBeProvided, new ParameterSetter(sm, ps, statementExpressionIndex, false));
                            }           

                            ResultSet rs = executeQuery(versionStmt, ps);
                            try
                            {
                                if (rs.next())
                                {
                                    Object datastoreVersion = table.getVersionMapping().getObject(pm, rs, new int[]{1});
                                    if (!((VersionMapping)table.getVersionMapping()).compareVersion(datastoreVersion, sm.getOptimisticDatastoreVersion()))
                                    {
                                        String msg = LOCALISER.msg("Request.OptimisticVersionMismatch", 
                                            sm.getInternalObjectId(), table.toString(), datastoreVersion, sm.getOptimisticDatastoreVersion());
                                        JPOXLogger.RDBMS.error(msg);
                                        throw new JDOOptimisticVerificationException(msg);
                                    }
                                }
                            }
                            finally
                            {
                                rs.close();
                            }
                        }
                    }
                    finally
                    {
                        ps.close();
                    }
                }

                ps = conn.prepareStatement(deleteStmt);
                try
                {
                    // provide primary key fields
 
                    //datastore identity
                    if (sm.getInternalObjectId() instanceof OID)
                    {
                        int[] paramNumber = { IDPARAMNUMBER };
                                                
                        table.getDataStoreObjectIdMapping().setObject(pm, ps, paramNumber, sm.getInternalObjectId());
                    }
                    //application identity
                    else
                    {
                        sm.provideFields(pkfieldsToBeProvided, new ParameterSetter(sm, ps, statementExpressionIndex, false));
                    }           

                    executeUpdate(deleteStmt, ps);
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                pm.releaseConnection(conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException("Delete request failed: " + deleteStmt, e);
        }
    }
    
    private class DeleteMappingConsumer implements MappingConsumer
    {
        
        private final StringBuffer where;
        private int paramIndex;
        /** one StatementExpressionIndex for each field **/
        private StatementExpressionIndex[] statementExpressionIndex;
        
        /**
         * All the pk fields that must be provided by the StateManager.
         * The StateManager will call the PC class to provide fields
         */         
        private List pkbp = new ArrayList();
        
        
        private List mc = new ArrayList();


        public DeleteMappingConsumer(StringBuffer where, int initialParamIndex)
        {
            this.where = where;
            this.paramIndex = initialParamIndex;
        }
        public void preConsumeMapping(int highestFieldNumberIndex)
        {
            if( statementExpressionIndex == null )
            {
                statementExpressionIndex = new StatementExpressionIndex[highestFieldNumberIndex];
            }
        }
        public void consumeMapping(JavaTypeMapping m, FieldMetaData fmd)
        {
            if( !fmd.getClassMetaData().isSameOrAncestorOf(cmd) )
            {
                return;
            }            
            if (m.includeInUpdateStatement() && fmd.isPrimaryKey())
            {
                statementExpressionIndex[fmd.getAbsoluteFieldNumber()] = new StatementExpressionIndex();
                statementExpressionIndex[fmd.getAbsoluteFieldNumber()].setMapping(m);
                int parametersIndex[] = new int[m.getNumberOfDatastoreFields()];
                for (int j = 0; j < parametersIndex.length; j++)
                {
                    if (where.length() > 0)
                    {
                        where.append(" AND ");
                    }
                    
                    where.append(((Column)m.getDataStoreMapping(j).getDatastoreField()).getName());
                    where.append(" = ");
                    where.append(((RDBMSMapping)m.getDataStoreMapping(j)).getUpdateInputParameter());
                    
                    pkbp.add(new Integer(fmd.getAbsoluteFieldNumber()));
                        
                    parametersIndex[j] = paramIndex++;
                }
                statementExpressionIndex[fmd.getAbsoluteFieldNumber()].setParameterIndex(parametersIndex);                     
            }
            
            /*
             * Build up list of mappings callbacks for the fields of this class.
             * The Mapping callback called delete is the preDelete
             */
            if (m instanceof MappingCallbacks)
            {
                mc.add(m);
            }
        }

        /**
         * Consumes a mapping not associated to a field
         * @param m
         * @param mappingType the Mapping type
         */
        public void consumeMapping(JavaTypeMapping m, int mappingType)
        {
        }

        /**
         * All the pk fields that must be provided by the StateManager.
         * The StateManager will call the PC class to provide fields
         */        
        public int[] getPrimaryKeysToBeProvided()
        {
            int []pkfieldsToBeProvided = new int[pkbp.size()];
            for (int i = 0; i < pkbp.size(); i++)
            {
                pkfieldsToBeProvided[i] = ((Integer)pkbp.get(i)).intValue();
            }     
            return pkfieldsToBeProvided;
        }
        
        public List getMappingCallBacks()
        {
            return mc;
        }
        
        public StatementExpressionIndex[] getStatementExpressionIndex()
        {
            return statementExpressionIndex;
        }
    }
}