/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Erik Bengtson - throws a JDOObjectNotFoundException
2003 Erik Bengtson - removed unused variable and import
2003 Andy Jefferson - coding standards
2003 Andy Jefferson - conversion to use Logger
2003 Erik Bengtson - added optimistic
2004 Erik Bengtson - added consumer
2004 Andy Jefferson - added discriminator
    ...
**********************************************************************/
package org.jpox.store.rdbms.request;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOObjectNotFoundException;
import javax.jdo.JDOOptimisticVerificationException;

import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.Transaction;
import org.jpox.metadata.ClassMetaData;
import org.jpox.metadata.FieldMetaData;
import org.jpox.store.FetchStatement;
import org.jpox.store.OID;
import org.jpox.store.StatementExpressionIndex;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.MappingCallbacks;
import org.jpox.store.mapping.MappingConsumer;
import org.jpox.store.mapping.VersionMapping;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.RDBMSFetchStatement;
import org.jpox.store.rdbms.fieldmanager.ParameterSetter;
import org.jpox.store.rdbms.fieldmanager.ResultSetGetter;
import org.jpox.store.rdbms.table.ClassTable;
import org.jpox.util.JPOXLogger;

/**
 * Class to provide a means of selection of records from a data store.
 * Extends basic request class implementing the execute method to do a JDBC
 * select operation.
 *
 * @version $Revision: 1.12 $ 
 **/
public class FetchRequest extends Request
{
    private MappingCallbacks[] callbacks;

    private FetchMappingConsumer consumer;

    private MappingStatementIndex mappingStatementIndex;

    private FetchStatement fetchStatement;

    private final int[] fieldNumbers;

    private ClassMetaData cmd;

    /**
     * Constructor, taking the table. Uses the structure of the datastore
     * table to build a basic query.
     * @param table The Class Table representing the datastore table to retrieve
     * @param fieldNumbers The numbers of the fields to retrieve
     **/
    public FetchRequest(ClassTable table, int[] fieldNumbers)
    {
        super(table);

        this.fieldNumbers = fieldNumbers;
    }

    public void initialize(StateManager sm)
    {
        cmd = sm.getPersistenceManager().getPMFContext().getMetaDataManager().getMetaDataForClass(sm.getObject().getClass());        

        mappingStatementIndex = new MappingStatementIndex();
        fetchStatement = new RDBMSFetchStatement(table);

        // Extract the fields to be retrieved
        consumer = new FetchMappingConsumer(fetchStatement);
        table.provideMappingsForFields(consumer, fieldNumbers, cmd);
        table.provideDatastoreIdMappings(consumer);
        table.provideDiscriminatorMappings(consumer);

        callbacks = (MappingCallbacks[]) consumer.getMappingCallbacks().toArray(new MappingCallbacks[consumer.getMappingCallbacks().size()]);

        // Construct the statement for the Fetch (the WHERE conditions)
        if (consumer.getFieldsToBeReplaced().length > 0)
        {
            table.provideVersionMappings(consumer);
            table.providePrimaryKeyMappings(consumer);
        }
        else
        {
            fetchStatement = null;
        }
    }

    /**
     * Method performing the retrieval of the record from the datastore. Takes
     * the constructed retrieval query and populates with the specific record
     * information.
     *
     * @param sm The state manager for the record to be retrieved
     **/
    public void execute(StateManager sm)
    {
        if (fetchStatement == null)
        {
            initialize(sm);
        }
        if (fetchStatement != null)
        {
            PersistenceManager pm = sm.getPersistenceManager();
            Transaction tx = (Transaction) pm.currentTransaction();
            String stmt = fetchStatement.toString(tx.useUpdateLockOnFetch());

            try
            {
                Connection conn = tx.getConnection(false);

                try
                {
                    PreparedStatement ps = conn.prepareStatement(stmt);
                    try
                    {
                        //datastore identity
                        if (sm.getInternalObjectId() instanceof OID)
                        {
                            // provide primary key fields
                            table.getDataStoreObjectIdMapping().setObject(pm, ps, mappingStatementIndex.getDatastoreId().getParameterIndex(), sm.getInternalObjectId());
                        }
                        //application identity
                        else
                        {
                            sm.provideFields(consumer.getPrimaryKeyFieldsToBeProvided(), new ParameterSetter(sm, ps, mappingStatementIndex.getFields(), false));
                        }
                        ResultSet rs = executeQuery(stmt, ps);
                        try
                        {
                            if (!rs.next())
                            {
                                JPOXLogger.RDBMS_SQL.warn(LOCALISER.msg("RDBMSManager.ObjectWithIdNotFound",sm.getInternalObjectId()));
                                throw new JDOObjectNotFoundException("No such database row", sm.getInternalObjectId());
                            }
                            sm.replaceFields(consumer.getFieldsToBeReplaced(), new ResultSetGetter(sm, rs, mappingStatementIndex.getFields()));

                            if (mappingStatementIndex.getVersion().getExpressionIndex() != null)
                            {
                                Object datastoreVersion = table.getVersionMapping().getObject(pm, rs,  mappingStatementIndex.getVersion().getExpressionIndex());
                                if (sm.getPersistenceManager().currentTransaction().getOptimistic() &&
                                    !((VersionMapping)table.getVersionMapping()).compareVersion(datastoreVersion, sm.getOptimisticDatastoreVersion()))
                                {
                                    String msg = LOCALISER.msg("Request.OptimisticVersionMismatch", 
                                        sm.getInternalObjectId(), table.toString(), datastoreVersion, sm.getOptimisticDatastoreVersion());
                                    JPOXLogger.RDBMS.error(msg);
                                    throw new JDOOptimisticVerificationException(msg);
                                }
                                sm.setOptimisticDatastoreVersion(datastoreVersion);
                            }
                        }
                        finally
                        {
                            rs.close();
                        }
                    }
                    finally
                    {
                        ps.close();
                    }
                }
                finally
                {
                    tx.releaseConnection(conn);
                }
            }
            catch (SQLException e)
            {
                throw new JDODataStoreException("Fetch request failed: " + stmt, e);
            }
        }

        // Execute any mapping actions now that we have fetched the element
        for (int i = 0; i < callbacks.length; ++i)
        {
            callbacks[i].postFetch(sm);
        }
    }

    private class FetchMappingConsumer implements MappingConsumer
    {
        private int paramIndex = 1;

        final FetchStatement fetchStmt;

        /** fields to be replaced */
        List ftbr = new ArrayList();

        /** pk fields to be provided */
        List pkbp = new ArrayList();

        List mappingCallbacks = new ArrayList();

        boolean initialized = false;

        /**
         * @param fetchStmt The Fetch Statement
         */
        public FetchMappingConsumer(final FetchStatement fetchStmt)
        {
            super();
            this.fetchStmt = fetchStmt;
        }

        public void preConsumeMapping(int highestFieldNumber)
        {
            if (!initialized)
            {
                mappingStatementIndex.setPrimaryKeys(new StatementExpressionIndex[highestFieldNumber]);
                mappingStatementIndex.setFields(new StatementExpressionIndex[highestFieldNumber]);
                initialized = true;
            }
        }

        public void consumeMapping(JavaTypeMapping m, FieldMetaData fmd)
        {
            if (!fmd.getClassMetaData().isSameOrAncestorOf(cmd))
            {
                return;
            }
            if (m.includeInFetchStatement())
            {
                mappingStatementIndex.getFields()[fmd.getAbsoluteFieldNumber()] = new StatementExpressionIndex();
                mappingStatementIndex.getFields()[fmd.getAbsoluteFieldNumber()].setMapping(m);

                //create the expressions index (columns index)
                int expressionsIndex[] = new int[m.getNumberOfDatastoreFields()];
                for (int j = 0; j < expressionsIndex.length; j++)
                {
                    expressionsIndex[j] = fetchStmt.select(m.getDataStoreMapping(j).getDatastoreField());
                }
                mappingStatementIndex.getFields()[fmd.getAbsoluteFieldNumber()].setExpressionIndex(expressionsIndex);

                /*
                 * check if the our column is primary key. if the
                 * first column for the mapping is primary key,
                 * all columns for the same mapping must be primary key
                 */
                Integer abs_field_num = new Integer(fmd.getAbsoluteFieldNumber());
                if (m.getFieldMetaData().isPrimaryKey())
                {
                    int parametersIndex[] = new int[m.getNumberOfDatastoreFields()];
                    for (int j = 0; j < parametersIndex.length; j++)
                    {
                        fetchStmt.andCondition(fetchStmt.referenceColumn(m.getDataStoreMapping(j).getDatastoreField()) + " = ?");

                        if (!pkbp.contains(abs_field_num))
                        {
                            pkbp.add(abs_field_num);
                        }

                        parametersIndex[j] = paramIndex++;
                    }
                    mappingStatementIndex.getFields()[fmd.getAbsoluteFieldNumber()].setParameterIndex(parametersIndex);
                }
                else
                {
                    if (!ftbr.contains(abs_field_num))
                    {
                        ftbr.add(abs_field_num);
                    }
                }
            }

            if (m instanceof MappingCallbacks)
            {
                mappingCallbacks.add(m);
            }
        }

        /**
         * Consumes a mapping not associated to a field
         * @param m The mapping
         * @param mappingType the Mapping type
         */
        public void consumeMapping(JavaTypeMapping m, int mappingType)
        {
            if (mappingType == MappingConsumer.MAPPING_TYPE_VERSION)
            {
                // version column
                mappingStatementIndex.getVersion().setMapping(table.getVersionMapping());
                //create the expressions index (columns index)
                int expressionsIndex[] = new int[table.getVersionMapping().getNumberOfDatastoreFields()];
                for (int j = 0; j < expressionsIndex.length; j++)
                {
                    expressionsIndex[j] = fetchStmt.select(table.getVersionMapping().getDataStoreMapping(j).getDatastoreField());
                }
                mappingStatementIndex.getVersion().setExpressionIndex(expressionsIndex);
            }
            else if (mappingType == MappingConsumer.MAPPING_TYPE_DATASTORE_ID)
            {
                fetchStmt.andCondition(fetchStmt.referenceColumn((Column) key.getColumns().get(0)) + " = ?");
                mappingStatementIndex.getDatastoreId().setParameterIndex(new int[]{paramIndex});
                paramIndex++;
            }
        }

        /**
         * @return Returns the mappingCallbacks.
         */
        public List getMappingCallbacks()
        {
            return mappingCallbacks;
        }

        public int[] getFieldsToBeReplaced()
        {
            int[] fieldsToBeReplaced = new int[ftbr.size()];
            for (int i = 0; i < ftbr.size(); ++i)
            {
                fieldsToBeReplaced[i] = ((Integer) ftbr.get(i)).intValue();
            }
            return fieldsToBeReplaced;
        }

        public int[] getPrimaryKeyFieldsToBeProvided()
        {
            int[] pkfieldsToBeProvided = new int[pkbp.size()];
            for (int i = 0; i < pkbp.size(); i++)
            {
                pkfieldsToBeProvided[i] = ((Integer) pkbp.get(i)).intValue();
            }
            return pkfieldsToBeProvided;
        }
    }
}