/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others. All rights reserved.
This program and the accompanying materials are made available under
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Erik Bengtson - the fields to fetch are better managed for application
                     identity. Their selection were moved from execute
                     method to the constructor
2003 Andy Jefferson - coding standards
2004 Andy Jefferson - conversion to use Logger
2004 Erik Bengtson - changed to use mapping consumer
2004 Andy Jefferson - added discriminator support
    ...
**********************************************************************/
package org.jpox.store.rdbms.request;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOFatalException;

import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.metadata.ClassMetaData;
import org.jpox.metadata.DiscriminatorMetaData;
import org.jpox.metadata.DiscriminatorStrategy;
import org.jpox.metadata.FieldMetaData;
import org.jpox.metadata.IdentityType;
import org.jpox.state.ActivityState;
import org.jpox.state.StateManagerImpl;
import org.jpox.store.StatementExpressionIndex;
import org.jpox.store.StoreManager;
import org.jpox.store.exceptions.NotYetFlushedException;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.MappingCallbacks;
import org.jpox.store.mapping.MappingConsumer;
import org.jpox.store.mapping.VersionMapping;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.store.rdbms.fieldmanager.ParameterSetter;
import org.jpox.store.rdbms.mapping.RDBMSMapping;
import org.jpox.store.rdbms.table.ClassTable;

/**
 * Class to provide a means of insertion of records to a data store.
 * Extends basic request class implementing the execute method to do a JDBC
 * insert operation.
 *
 * @version $Revision: 1.12 $ 
 **/
public class InsertRequest extends Request
{
    private static final int IDPARAMNUMBER = 1;

    private MappingCallbacks[] callbacks;
    private int[] fieldsToBeProvided;
    private int[] pkfieldsToBeProvided;
    private String textStmt;
    private boolean hasAutoIncrementColumn = false;

    /** one StatementExpressionIndex for each field **/
    private StatementExpressionIndex[] statementExpressionIndex;

    /** StatementExpressionIndex for version **/
    private StatementExpressionIndex versionStatementExpressionIndex;

    /** StatementExpressionIndex for discriminator **/
    private StatementExpressionIndex discriminatorStatementExpressionIndex;

    private ClassMetaData cmd;
    
    /**
     * Constructor, taking the table. Uses the structure of the datastore
     * table to build a basic query.
     * @param table The Class Table representing the datastore table to insert.
     **/
    public InsertRequest(ClassTable table)
    {
        super(table);
    }

    private void initalize(StateManager sm)
    {
        cmd = sm.getPersistenceManager().getPMFContext().getMetaDataManager().getMetaDataForClass(sm.getObject().getClass());        

        InsertMappingConsumer consumer = new InsertMappingConsumer(IDPARAMNUMBER);
        table.provideDatastoreIdMappings(consumer);
        table.provideNonPrimaryKeyMappings(consumer);
        table.providePrimaryKeyMappings(consumer);
        table.provideVersionMappings(consumer);
        table.provideDiscriminatorMappings(consumer);

        callbacks = (MappingCallbacks[])consumer.getMappingCallbacks().toArray(new MappingCallbacks[consumer.getMappingCallbacks().size()]);

        statementExpressionIndex = consumer.getStatementExpressionIndex();
        versionStatementExpressionIndex = consumer.getVersionStatementExpressionIndex();
        discriminatorStatementExpressionIndex = consumer.getDiscriminatorStatementExpressionIndex();

        fieldsToBeProvided = consumer.getFieldsToBeProvided();
        pkfieldsToBeProvided = consumer.getPrimaryKeyFieldsToBeProvided();
        if (table.getIdentityType() == IdentityType.APPLICATION)
        {
	        if (pkfieldsToBeProvided.length<1 && !hasAutoIncrementColumn)
	        {
	            throw new JDOFatalException("It was expected at least one primary key column.");
	        }
        }
        textStmt = consumer.getInsertStmt();
    }

    /**
     * Method performing the insertion of the record from the datastore. Takes
     * the constructed insert query and populates with the specific record
     * information.
     * @param sm The state manager for the record to be inserted
     **/
    public void execute(StateManager sm)
    {
        if (textStmt == null)
        {
            initalize(sm);
        }
        PersistenceManager pm = sm.getPersistenceManager();
        try
        {
            Connection conn = pm.getConnection(true);
            try
            {
                PreparedStatement ps = conn.prepareStatement(textStmt);
                try
                {
                    // datastore identity
                    if (table.getIdentityType() == IdentityType.DATASTORE)
                    {
                        if (!table.isObjectIDDatabaseAttributed() || !table.isRootTable())
                        {
                            int[] paramNumber = {IDPARAMNUMBER};
                            table.getDataStoreObjectIdMapping().setObject(pm, ps, paramNumber, sm.getInternalObjectId());
                        }
                    }
                    // application identity
                    else if (table.getIdentityType() == IdentityType.APPLICATION)
                    {
                        sm.provideFields(pkfieldsToBeProvided, new ParameterSetter(sm, ps, statementExpressionIndex, true));
                    }

                    // Provide all non-key fields needed for the insert
                    // This makes persistent any instances of PC classes thats
                    // are not yet persistent
                    int numberOfFieldsToProvide = 0;
                    for (int i = 0; i < fieldsToBeProvided.length; i++)
                    {
                        if (fieldsToBeProvided[i] < sm.getHighestFieldNumber())
                        {
                            numberOfFieldsToProvide++;
                        }
                    }
                    int j = 0;
                    int[] classFieldNumbers = new int[numberOfFieldsToProvide];
                    for (int i = 0; i < fieldsToBeProvided.length; i++)
                    {
                        if (fieldsToBeProvided[i] < sm.getHighestFieldNumber())
                        {
                            classFieldNumbers[j++] = fieldsToBeProvided[i];
                        }
                        else
                        {
                            // Any fields out of range for the class, we default
                            // or null.
                            // This happens when we insert a base class object
                            // and the sub-class is persisted to the
                            // same table ("superclass-table").
                            StatementExpressionIndex stmtExprIndex = statementExpressionIndex[fieldsToBeProvided[i]];

                            // TODO Insert a null if the column allows it
                            stmtExprIndex.getMapping().setDefault(pm, ps, stmtExprIndex.getParameterIndex());
                        }
                    }

                    sm.provideFields(classFieldNumbers, new ParameterSetter(sm, ps, statementExpressionIndex, true));

                    // Version mapping (optional)
                    if (table.getVersionMapping() != null)
                    {
                        Object nextOptimisticVersion = ((VersionMapping) table.getVersionMapping()).getNextVersion(sm.getOptimisticDatastoreVersion());
                        table.getVersionMapping().setObject(pm, ps, versionStatementExpressionIndex.getParameterIndex(), nextOptimisticVersion);
                        sm.setOptimisticTransactionalVersion(nextOptimisticVersion);
                    }

                    // Discriminator mapping (optional)
                    if (table.getDiscriminatorMapping() != null)
                    {
                        DiscriminatorMetaData dismd = table.getDiscriminatorMetaData();
                        if (dismd.getStrategy() == DiscriminatorStrategy.CLASS_NAME)
                        {
                            table.getDiscriminatorMapping().setObject(pm, ps, discriminatorStatementExpressionIndex.getParameterIndex(),
                                sm.getObject().getClass().getName());
                        }
                        else if (dismd.getStrategy() == DiscriminatorStrategy.VALUE_MAP)
                        {
                            // Use Discriminator info for the actual class
                            dismd = ((StateManagerImpl) sm).getClassMetaData().getInheritanceMetaData().getDiscriminatorMetaData();
                            table.getDiscriminatorMapping().setObject(pm, ps, discriminatorStatementExpressionIndex.getParameterIndex(), dismd.getValue());
                        }
                    }

                    executeUpdate(textStmt, ps);

                    // gets the key for auto-increment keys
                    if (hasAutoIncrementColumn)
                    {
                        String autoIncStmt = ((RDBMSManager) pm.getStoreManager()).getDatabaseAdapter().getAutoIncrementStmt();
                        PreparedStatement psAutoIncrement = conn.prepareStatement(autoIncStmt);

                        try
                        {
                            ResultSet rs = executeQuery(autoIncStmt, psAutoIncrement);
                            if (rs.next())
                            {
                                sm.setPostStoreNewObjectId(rs.getObject(1));
                            }
                        }
                        finally
                        {
                            psAutoIncrement.close();
                        }
                    }
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                pm.releaseConnection(conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException("Insert request failed: " + textStmt, e);
        }

        sm.changeActivityState(ActivityState.INSERTING_CALLBACKS);

        // Execute any mapping actions now that we have inserted the element
        // (things like inserting any association parent-child).
        for (int i = 0; i < callbacks.length; ++i)
        {
            try
            {
                callbacks[i].postInsert(sm);
            }
            catch (NotYetFlushedException e)
            {
                sm.updateFieldAfterInsert(e.getPc(), ((JavaTypeMapping) callbacks[i]).getFieldMetaData().getAbsoluteFieldNumber());
            }
        }
    }

    /**
     * Internal class to provide mapping consumption. TODO Think about making
     * this a InsertStatement (copying FetchStatement)
     */
    private class InsertMappingConsumer implements MappingConsumer
    {
        /** fields to be provided */
        ArrayList ftbp = new ArrayList();

        /** primary key fields to be provided */
        ArrayList pkbp = new ArrayList();

        StringBuffer columnNames = new StringBuffer();
        StringBuffer columnValues = new StringBuffer();

        HashMap assignedColumns = new HashMap();

        ArrayList mc = new ArrayList();

        boolean initialized = false;

        int paramIndex;

        /** one StatementExpressionIndex for each field **/
        private StatementExpressionIndex[] statementExpressionIndex;

        /** StatementExpressionIndex for version **/
        private StatementExpressionIndex versionStatementExpressionIndex;

        /** StatementExpressionIndex for discriminator **/
        private StatementExpressionIndex discriminatorStatementExpressionIndex;

        /**
         * @param initialParamIndex
         */
        public InsertMappingConsumer(int initialParamIndex)
        {
            super();
            this.paramIndex = initialParamIndex;
        }

        public void preConsumeMapping(int highestFieldNumber)
        {
            if (!initialized)
            {
                statementExpressionIndex = new StatementExpressionIndex[highestFieldNumber];
                initialized = true;
            }
        }

        public void consumeMapping(JavaTypeMapping m, FieldMetaData fmd)
        {
            if (!fmd.getClassMetaData().isSameOrAncestorOf(cmd))
            {
                return;
            }
            statementExpressionIndex[fmd.getAbsoluteFieldNumber()] = new StatementExpressionIndex();
            statementExpressionIndex[fmd.getAbsoluteFieldNumber()].setMapping(m);
            if (m.includeInInsertStatement())
            {
                //create the expressions index (columns index)
                int parametersIndex[] = new int[m.getNumberOfDatastoreFields()];
                for (int j = 0; j < parametersIndex.length; j++)
                {
                    // check if the column was not already assigned         
                    if (!assignedColumns.containsKey(((Column)m.getDataStoreMapping(j).getDatastoreField()).getName().toString()))
                    {
                        // application identity with no autoincrement, the auto-increment is declared inside the field
                        if (!StoreManager.isStrategyDatastoreAttributed(fmd.getValueStrategy()) ||
                            !table.isRootTable())
                        {
                            if (columnNames.length() > 0)
                            {
                                columnNames.append(',');
                                columnValues.append(',');
                            }
                            String param = ((RDBMSMapping)m.getDataStoreMapping(j)).getInsertionInputParameter();

                            columnNames.append(((Column)m.getDataStoreMapping(j).getDatastoreField()).getName());
                            columnValues.append(param);

                            if (param.indexOf("?") > -1) // only add fields to be replaced by the real values only if the param value has ?
                            {
                                Integer abs_field_num = new Integer(fmd.getAbsoluteFieldNumber());
	                            if (fmd.isPrimaryKey())
	                            {
                                    if (!pkbp.contains(abs_field_num))
                                    {
	                                    pkbp.add(abs_field_num);
                                    }
	                            }
	                            else
	                            {
                                    if (!ftbp.contains(abs_field_num))
                                    {
	                                    ftbp.add(abs_field_num);
                                    }
	                            }
	                            parametersIndex[j] = paramIndex++;
                            }
                            assignedColumns.put(((Column)m.getDataStoreMapping(j).getDatastoreField()).getName().toString(),new Integer(fmd.getAbsoluteFieldNumber()));
                        }
                        else
                        {
                            hasAutoIncrementColumn = true;
                        }
                    }
                    else
                    {
                        parametersIndex[j] = ((Integer)assignedColumns.get(((Column)m.getDataStoreMapping(j).getDatastoreField()).getName().toString())).intValue();
                    }
                    statementExpressionIndex[fmd.getAbsoluteFieldNumber()].setParameterIndex(parametersIndex);
                }
            }
            if (m instanceof MappingCallbacks)
            {
                mc.add(m);
            }
        }

        /**
         * Consumes a mapping not associated to a field.
         * @param m The mapping.
         * @param mappingType the Mapping type
         */
        public void consumeMapping(JavaTypeMapping m, int mappingType)
        {
            if (mappingType == MappingConsumer.MAPPING_TYPE_VERSION)
            {
                // version column
                if (table.getVersionMapping() != null)
                {
                    String val = ((RDBMSMapping)table.getVersionMapping().getDataStoreMapping(0)).getUpdateInputParameter();
                    if (columnNames.length() > 0)
                    {
                        columnNames.append(',');
                        columnValues.append(',');
                    }
                    columnNames.append(((Column)table.getVersionMapping().getDataStoreMapping(0).getDatastoreField()).getName());
                    columnValues.append(val);

                    versionStatementExpressionIndex = new StatementExpressionIndex();
                    versionStatementExpressionIndex.setMapping(table.getVersionMapping());
                    int[] param = { paramIndex++ };
                    versionStatementExpressionIndex.setParameterIndex(param);
                }
                else
                {
                    versionStatementExpressionIndex = null;
                }
            }
            else if (mappingType == MappingConsumer.MAPPING_TYPE_DISCRIMINATOR)
            {
                // discriminator column
                if (table.getDiscriminatorMapping() != null)
                {
                    String val = ((RDBMSMapping)table.getDiscriminatorMapping().getDataStoreMapping(0)).getUpdateInputParameter();

                    if (columnNames.length() > 0)
                    {
                        columnNames.append(',');
                        columnValues.append(',');
                    }
                    columnNames.append(((Column)table.getDiscriminatorMapping().getDataStoreMapping(0).getDatastoreField()).getName());
                    columnValues.append(val);
                    discriminatorStatementExpressionIndex = new StatementExpressionIndex();
                    discriminatorStatementExpressionIndex.setMapping(table.getDiscriminatorMapping());
                    int[] param = { paramIndex++ };
                    discriminatorStatementExpressionIndex.setParameterIndex(param);
                }
                else
                {
                    discriminatorStatementExpressionIndex = null;
                }
            }
            else if (mappingType == MappingConsumer.MAPPING_TYPE_DATASTORE_ID)
            {
                if (table.getIdentityType() == IdentityType.DATASTORE)
                {
                    if (!table.isObjectIDDatabaseAttributed() || !table.isRootTable())
                    {
                        Iterator iterator = key.getColumns().iterator();
                        if (columnNames.length() > 0)
                        {
                            columnNames.append(',');
                            columnValues.append(',');
                        }
                        columnNames.append(((Column) iterator.next()).getName().toString());
                        columnValues.append("?");
                        paramIndex++;
                    }
                    else
                    {
                        hasAutoIncrementColumn = true;
                    }
                }
            }
        }

        /**
         * @return Returns the mappingCallbacks.
         */
        public List getMappingCallbacks()
        {
            return mc;
        }

        public int[] getFieldsToBeProvided()
        {
            int[] fieldsToBeProvided = new int[ftbp.size()];
            for (int i = 0; i < ftbp.size(); ++i)
            {
                fieldsToBeProvided[i] = ((Integer) ftbp.get(i)).intValue();
            }
            return fieldsToBeProvided;
        }

        public int[] getPrimaryKeyFieldsToBeProvided()
        {
            int[] pkfieldsToBeProvided = new int[pkbp.size()];
            for (int i = 0; i < pkbp.size(); i++)
            {
                pkfieldsToBeProvided[i] = ((Integer) pkbp.get(i)).intValue();
            }
            return pkfieldsToBeProvided;
        }

        public StatementExpressionIndex[] getStatementExpressionIndex()
        {
            return statementExpressionIndex;
        }

        public StatementExpressionIndex getVersionStatementExpressionIndex()
        {
            return versionStatementExpressionIndex;
        }

        public StatementExpressionIndex getDiscriminatorStatementExpressionIndex()
        {
            return discriminatorStatementExpressionIndex;
        }

        public String getInsertStmt()
        {
            // Construct the statement for the Insert
            return "INSERT INTO " + table.toString() + " (" + columnNames + ") VALUES (" + columnValues + ")";
        }
    }
}