/**********************************************************************
Copyright (c) Oct 12, 2004 Erik Bengtson and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
    ...
**********************************************************************/
package org.jpox.store.rdbms.request;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.List;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOObjectNotFoundException;

import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.Transaction;
import org.jpox.metadata.FieldMetaData;
import org.jpox.store.FetchStatement;
import org.jpox.store.OID;
import org.jpox.store.StatementExpressionIndex;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.MappingConsumer;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.RDBMSFetchStatement;
import org.jpox.store.rdbms.fieldmanager.ParameterSetter;
import org.jpox.store.rdbms.table.ClassTable;
import org.jpox.util.JPOXLogger;

/**
 * Class to locate a record in the data store.
 * Extends basic request class implementing the execute method to do a JDBC
 * select operation.
 *
 * @version $Revision: 1.5 $ 
 **/
public class LocateRequest extends Request
{
    private final FetchMappingConsumer consumer;
    
    private final MappingStatementIndex mappingStatementIndex;

    private final String textStmt;

    /**
     * Constructor, taking the table. Uses the structure of the datastore
     * table to build a basic query.
     * @param table The Class Table representing the datastore table to retrieve
     **/
    public LocateRequest(ClassTable table)
    {
        super(table);

        mappingStatementIndex = new MappingStatementIndex();
        FetchStatement fetchStmt = new RDBMSFetchStatement(table);

        consumer = new FetchMappingConsumer(fetchStmt);

        fetchStmt.select("1");
        table.provideDatastoreIdMappings(consumer);
        table.providePrimaryKeyMappings(consumer);
        textStmt = fetchStmt.toString();
    }

    /**
     * Method performing the retrieval of the record from the datastore. Takes
     * the constructed retrieval query and populates with the specific record
     * information.
     *
     * @param sm The state manager for the record to be retrieved
     **/
    public void execute(StateManager sm)
    {
        if (textStmt != null)
        {
            PersistenceManager pm = sm.getPersistenceManager();
            Transaction tx = (Transaction) pm.currentTransaction();
            String stmt = textStmt;
            try
            {
                Connection conn = tx.getConnection(false);

                try
                {
                    PreparedStatement ps = conn.prepareStatement(stmt);
                    try
                    {
                        // datastore identity
                        if (sm.getInternalObjectId() instanceof OID)
                        {
                            table.getDataStoreObjectIdMapping().setObject(pm, ps, mappingStatementIndex.getDatastoreId().getParameterIndex(), sm.getInternalObjectId());
                        }
                        // application identity
                        else
                        {
                            // provide primary key fields
                            sm.provideFields(consumer.getPrimaryKeyFieldsToBeProvided(), new ParameterSetter(sm, ps, mappingStatementIndex.getFields(), false));
                        }

                        ResultSet rs = executeQuery(stmt, ps);
                        try
                        {
                            if (!rs.next())
                            {
                                JPOXLogger.RDBMS_SQL.warn(LOCALISER.msg("RDBMSManager.ObjectWithIdNotFound",sm.getInternalObjectId()));
                                throw new JDOObjectNotFoundException("No such database row", sm.getInternalObjectId());
                            }
                        }
                        finally
                        {
                            rs.close();
                        }
                    }
                    finally
                    {
                        ps.close();
                    }
                }
                finally
                {
                    tx.releaseConnection(conn);
                }
            }
            catch (SQLException e)
            {
                throw new JDODataStoreException("Locate request failed: " + stmt, e);
            }
        }
    }
    
    private class FetchMappingConsumer implements MappingConsumer
    {
        private int paramIndex = 1;
        final FetchStatement fetchStmt;
        /* pk fields to be provided */
        List pkbp = new ArrayList();
        
        boolean initialized = false;
        
        /**
         * @param fetchStmt
         * @param initialParamIndex
         */
        public FetchMappingConsumer(final FetchStatement fetchStmt)
        {
            super();
            this.fetchStmt = fetchStmt;
        }
        
        public void preConsumeMapping(int highestFieldNumber)
        {
            if( !initialized )
            {
                mappingStatementIndex.setPrimaryKeys(new StatementExpressionIndex[highestFieldNumber]);
                mappingStatementIndex.setFields(new StatementExpressionIndex[highestFieldNumber]);
                initialized = true;
            }
        }

        public void consumeMapping(JavaTypeMapping m, FieldMetaData fmd)
        {
            if (m.includeInFetchStatement())
            {
                mappingStatementIndex.getFields()[fmd.getAbsoluteFieldNumber()] = new StatementExpressionIndex();
                mappingStatementIndex.getFields()[fmd.getAbsoluteFieldNumber()].setMapping(m);

                /*
                 * check if the our column is primary key. if the
                 * first column for the mapping is primary key,
                 * all columns for the same mapping must be primary key
                 */
                if (m.getFieldMetaData().isPrimaryKey())
                {
                    
                    int parametersIndex[] = new int[m.getNumberOfDatastoreFields()];
                    for (int j = 0; j < parametersIndex.length; j++)
                    {
                        fetchStmt.andCondition(fetchStmt.referenceColumn(m.getDataStoreMapping(j).getDatastoreField()) + " = ?");
                        
                        pkbp.add(new Integer(fmd.getAbsoluteFieldNumber()));
                        
                        parametersIndex[j] = paramIndex++;
                    }
                    mappingStatementIndex.getFields()[fmd.getAbsoluteFieldNumber()].setParameterIndex(parametersIndex);
                }
            } 
          
        }
        
        /**
         * Consumes a mapping not associated to a field
         * @param m
         * @param mappingType the Mapping type
         */
        public void consumeMapping(JavaTypeMapping m, int mappingType)
        {
	        if( mappingType == MappingConsumer.MAPPING_TYPE_DATASTORE_ID)
	        {            
	            fetchStmt.andCondition(fetchStmt.referenceColumn((Column) key.getColumns().get(0)) + " = ?");
	            mappingStatementIndex.getDatastoreId().setParameterIndex(new int[] {paramIndex} );   
	            paramIndex++;
	        }
        }
        
        public int[] getPrimaryKeyFieldsToBeProvided()
        {
            int[] pkfieldsToBeProvided = new int[pkbp.size()];
            for (int i = 0; i < pkbp.size(); i++)
            {
                pkfieldsToBeProvided[i] = ((Integer) pkbp.get(i)).intValue();
            }
            return pkfieldsToBeProvided;
        }       
    }
}