/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others. All rights reserved.
This program and the accompanying materials are made available under
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - commented
    ...
**********************************************************************/
package org.jpox.store.rdbms.request;

import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;

import org.jpox.StateManager;
import org.jpox.store.rdbms.key.PrimaryKey;
import org.jpox.store.rdbms.table.ClassTable;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;

/**
 * Base class representing a request to perform an action on the datastore.
 * All requests have 2 methods - constructor and execute. Currently they build
 * an SQL statement and execute it.
 *
 * @version $Revision: 1.2 $ 
 **/
public abstract class Request
{
    /** Localisation of messages. */
    protected static final Localiser LOCALISER=Localiser.getInstance("org.jpox.store.rdbms.Localisation");

    protected final ClassTable table;
    protected final PrimaryKey key;

    /**
     * Constructor, taking the table to use for the request.
     * @param table The Table to use for the request.
     **/
    public Request(ClassTable table)
    {
        this.table = table;
        this.key   = table.getPrimaryKey();
    }

    /**
     * Method to execute the request - to be implemented by deriving classes.
     * @param sm The StateManager for the object in question. 
     **/
    public abstract void execute(StateManager sm);

    /**
     * Convenience wrapper for executing a JDBC update.
     * Provides logging of the statement and its execution time.
     * @param jdbcText The statement text
     * @param ps The PreparedStatement
     * @throws SQLException Thrown if an error occurs in the update
     */
    public void executeUpdate(String jdbcText, PreparedStatement ps)
    throws SQLException
    {
        if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
        {
            JPOXLogger.RDBMS_SQL.debug(jdbcText);
        }
        long startTime = System.currentTimeMillis();
        ps.executeUpdate();
        if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
        {
            JPOXLogger.RDBMS_SQL.debug(LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
        }
    }

    /**
     * Convenience wrapper for executing a JDBC update.
     * Provides logging of the statement and its execution time.
     * @param jdbcText The statement text
     * @param ps The PreparedStatement
     * @throws SQLException Thrown if an error occurs in the update
     */
    public ResultSet executeQuery(String jdbcText, PreparedStatement ps)
    throws SQLException
    {
        if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
        {
            JPOXLogger.RDBMS_SQL.debug(jdbcText);
        }
        long startTime = System.currentTimeMillis();
        ResultSet rs = ps.executeQuery();
        if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
        {
            JPOXLogger.RDBMS_SQL.debug(LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
        }

        return rs;
    }
}