/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Erik Bengtson  - optimistic transaction support
2004 Andy Jefferson - coding standards
2004 Andy Jefferson - conversion to use Logger
    ...
**********************************************************************/
package org.jpox.store.rdbms.request;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.List;
import java.util.Map;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOOptimisticVerificationException;

import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.metadata.ClassMetaData;
import org.jpox.metadata.FieldMetaData;
import org.jpox.metadata.IdentityType;
import org.jpox.store.FetchStatement;
import org.jpox.store.StatementExpressionIndex;
import org.jpox.store.exceptions.NotYetFlushedException;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.MappingCallbacks;
import org.jpox.store.mapping.MappingConsumer;
import org.jpox.store.mapping.VersionMapping;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.RDBMSFetchStatement;
import org.jpox.store.rdbms.fieldmanager.ParameterSetter;
import org.jpox.store.rdbms.mapping.RDBMSMapping;
import org.jpox.store.rdbms.table.ClassTable;
import org.jpox.util.JPOXLogger;

/**
 * Class to provide a means of update of records in a data store.
 * Extends basic request class implementing the execute method to do a JDBC
 * update operation.
 *
 * @version $Revision: 1.10 $ 
 **/
public class UpdateRequest extends Request
{
    /**
     * callback mappings will have their postUpdate method called after the update
     */
    private MappingCallbacks[] callbacks;
    
    /**
     * the sql update statement
     */
    private String updateStmt;

    /**
     * the sql version statement
     */
    private String versionStmt;
       
    /**
     * the index for the expression in the update sql statement.
     */    
    private MappingStatementIndex mappingStatementIndex = new MappingStatementIndex();
    
    /**
     * the index for the expression in the version sql statement.
     * for classes with datastore identity 
     */    
    private MappingStatementIndex versionMappingStatementIndex = new MappingStatementIndex();
    
    private UpdateMappingConsumer consumer = new UpdateMappingConsumer();

    private VersionMappingConsumer consumerVersion = new VersionMappingConsumer();
    
    private final int[] reqFieldNumbers;

    private ClassMetaData cmd;
    
    /**
     * Constructor, taking the table. Uses the structure of the datastore
     * table to build a basic query. The query will be of the form
     * <PRE>
     * UPDATE table-name
     * SET param1=?,param2=?
     * WHERE param3=? AND param4=?
     *</PRE> 
     *
     * @param table The Class Table representing the datastore table to
     *              update
     * @param reqFieldNumbers The numbers of the fields to update
     **/
    public UpdateRequest(ClassTable table, int[] reqFieldNumbers)
    {
        super(table);

        this.reqFieldNumbers = reqFieldNumbers;
    }
    
    private void initalize(StateManager sm)
    {
        cmd = sm.getPersistenceManager().getPMFContext().getMetaDataManager().getMetaDataForClass(sm.getObject().getClass());        
        table.provideMappingsForFields(consumer,reqFieldNumbers,cmd);
        table.provideVersionMappings(consumer);
        table.providePrimaryKeyMappings(consumer);
        table.provideDatastoreIdMappings(consumer);

        callbacks = (MappingCallbacks[])consumer.getMappingCallbacks().toArray(new MappingCallbacks[consumer.getMappingCallbacks().size()]);

        updateStmt = consumer.getStatement();

        table.provideVersionMappings(consumerVersion);
        table.providePrimaryKeyMappings(consumerVersion);
        table.provideDatastoreIdMappings(consumerVersion);
        
        versionStmt = consumerVersion.getStatement();
        
    }

    /**
     * Method performing the deletion of the record from the datastore. Takes
     * the constructed update query and populates with the specific record
     * information.
     *
     * @param sm The state manager for the record to be updated
     **/
    public void execute(StateManager sm)
    {
        if (updateStmt == null)
        {
            initalize(sm);
        }
        if (updateStmt != null)
        {
            PersistenceManager pm = sm.getPersistenceManager();
            try
            {
                Connection conn = pm.getConnection(true);

                try
                {
                    PreparedStatement ps;
                    if (pm.currentTransaction().getOptimistic() && versionStmt != null)
                    {
                        ps = conn.prepareStatement(versionStmt);
                        try
                        {
	                        // provide primary key fields for WHERE clause
	                        // a). datastore identity
	                        if (table.getIdentityType() == IdentityType.DATASTORE)
	                        {
	                            table.getDataStoreObjectIdMapping().setObject(pm, ps, versionMappingStatementIndex.getDatastoreId().getParameterIndex(), sm.getInternalObjectId());
	                        }
	                        // b). application identity
	                        else if (table.getIdentityType() == IdentityType.APPLICATION)
	                        {
	                            sm.provideFields(consumerVersion.getPrimaryKeyFieldsToBeProvided(), new ParameterSetter(sm, ps, versionMappingStatementIndex.getPrimaryKeys(), true));
	                        }

	                        ResultSet rs = executeQuery(versionStmt, ps);	                        
	                        try
	                        {
		                        if (rs.next())
		                        {
		                            Object datastoreVersion = table.getVersionMapping().getObject(pm, rs, versionMappingStatementIndex.getVersion().getExpressionIndex());
		                            if (!((VersionMapping)table.getVersionMapping()).compareVersion(datastoreVersion, sm.getOptimisticDatastoreVersion()))
		                            {
                                        String msg = LOCALISER.msg("Request.OptimisticVersionMismatch", 
                                            sm.getInternalObjectId(), table.toString(), datastoreVersion, sm.getOptimisticDatastoreVersion());
                                        JPOXLogger.RDBMS.error(msg);
		                                throw new JDOOptimisticVerificationException(msg);
		                            }
		                        }
		                        else
		                        {
                                    String msg = LOCALISER.msg("Request.OptimisticVersionMissing", sm.getInternalObjectId(), table.toString());
                                    JPOXLogger.RDBMS.error(msg);
	                                throw new JDOOptimisticVerificationException(msg);
		                        }
	                        }
	                        finally
	                        {
	                            rs.close();
	                        }
                        }
                        finally
                        {
                            ps.close();
                        }
                    }

                    ps = conn.prepareStatement(updateStmt);
                    try
                    {
                        // Set the required fields
                        if (consumer.getFieldsToBeProvided() != null)
                        {
                            sm.provideFields(consumer.getFieldsToBeProvided(), new ParameterSetter(sm, ps, mappingStatementIndex.getFields(),true));
                        }

                        // provide primary key fields for WHERE clause
                        // a). datastore identity
                        if (table.getIdentityType() == IdentityType.DATASTORE)
                        {
                            table.getDataStoreObjectIdMapping().setObject(pm, ps, mappingStatementIndex.getDatastoreId().getParameterIndex(), sm.getInternalObjectId());
                        }
                        // b). application identity
                        else if (table.getIdentityType() == IdentityType.APPLICATION)
                        {
                            sm.provideFields(consumer.getPrimaryKeyFieldsToBeProvided(), new ParameterSetter(sm, ps, mappingStatementIndex.getPrimaryKeys(),true));
                        }

                        if (versionStmt != null)
                        {
                            Object nextOptimisticVersion = ((VersionMapping)table.getVersionMapping()).getNextVersion(sm.getOptimisticDatastoreVersion());
                            table.getVersionMapping().setObject(pm, ps, mappingStatementIndex.getVersion().getParameterIndex(), nextOptimisticVersion );
                            sm.setOptimisticTransactionalVersion(nextOptimisticVersion);
                        }

                        executeUpdate(updateStmt, ps);
                    }
                    finally
                    {
                        ps.close();
                    }
                }
                finally
                {
                    pm.releaseConnection(conn);
                }
            }
            catch (SQLException e)
            {
                throw new JDODataStoreException("Update request failed: " + updateStmt, e);
            }
        }

        // Execute any mapping actions now that we have done the update
        for (int i=0; i<callbacks.length; ++i)
        {
            try
            {
                callbacks[i].postUpdate(sm);
            }
            catch (NotYetFlushedException e)
            {
                sm.updateFieldAfterInsert(e.getPc(),((JavaTypeMapping)callbacks[i]).getFieldMetaData().getAbsoluteFieldNumber());
            }
        }
    }

    private class UpdateMappingConsumer implements MappingConsumer
    {
        boolean initialized = false;

        int paramIndex = 1;

        List ftbp = new ArrayList();
        List mc = new ArrayList();
        List pkbp = new ArrayList();

        /** for UPDATE statement **/
        StringBuffer columnAssignments = new StringBuffer();

        Map assignedColumns = new HashMap();

        StringBuffer where = new StringBuffer();

        public UpdateMappingConsumer()
        {
            super();
        }

        public void preConsumeMapping(int highestFieldNumber)
        {
            if (!initialized)
            {
                mappingStatementIndex.setPrimaryKeys(new StatementExpressionIndex[highestFieldNumber]);
                mappingStatementIndex.setFields(new StatementExpressionIndex[highestFieldNumber]);
                initialized = true;
            }
        }

        public void consumeMapping(JavaTypeMapping m, FieldMetaData fmd)
        {
            if (!fmd.getClassMetaData().isSameOrAncestorOf(cmd))
            {
                return;
            }
            if (m.includeInUpdateStatement())
            {
                int parametersIndex[] = new int[m.getNumberOfDatastoreFields()];

                StatementExpressionIndex sei = new StatementExpressionIndex();
                sei.setMapping(m);
                sei.setParameterIndex(parametersIndex);

                if (m.getFieldMetaData().isPrimaryKey())
                {
                    //primary keys goes to the UPDATE SET COLUMN = ?, ....
                    mappingStatementIndex.getPrimaryKeys()[fmd.getAbsoluteFieldNumber()] = sei; 
                    for (int j=0; j<parametersIndex.length; j++)
                    {
                        if (where.length() > 0)
                        {
                            where.append(" AND ");
                        }
                        String condition = ((Column)m.getDataStoreMapping(j).getDatastoreField()).getName() + " = " + ((RDBMSMapping)m.getDataStoreMapping(j)).getUpdateInputParameter();
                        where.append(condition);

                        Integer abs_field_num = new Integer(fmd.getAbsoluteFieldNumber());
                        if (!pkbp.contains(abs_field_num))
                        {
                            pkbp.add(abs_field_num);
                        }
                        parametersIndex[j] = paramIndex++;
                    }
                }
                else
                {
                    //non primary keys goes to the WHERE COLUMN = ?, ....
                    mappingStatementIndex.getFields()[fmd.getAbsoluteFieldNumber()] = sei;

                    for (int j = 0; j < parametersIndex.length; j++)
	                {       
	                    // check if the column was not already assigned         
	                    if (!assignedColumns.containsKey(((Column)m.getDataStoreMapping(j).getDatastoreField()).getName().toString()))
	                    {
	                        if (columnAssignments.length() > 0)
	                        {
	                            columnAssignments.append(", ");
	                        }

                            String param = ((RDBMSMapping)m.getDataStoreMapping(j)).getUpdateInputParameter();

	                        columnAssignments.append(((Column)m.getDataStoreMapping(j).getDatastoreField()).getName()).append(" = ").append(param);

                            if (param.indexOf("?") > -1) // only add fields to be replaced by the real values only if the param value has ?
                            {
                                Integer abs_field_num = new Integer(fmd.getAbsoluteFieldNumber());
                                if (!ftbp.contains(abs_field_num))
                                {
	                                ftbp.add(abs_field_num);
                                }
	                            parametersIndex[j] = paramIndex++;
                            }
	                        assignedColumns.put(((Column)m.getDataStoreMapping(j).getDatastoreField()).getName().toString(),new Integer(fmd.getAbsoluteFieldNumber()));
	                    }
	                    else
	                    {
	                        parametersIndex[j] = ((Integer)assignedColumns.get(((Column)m.getDataStoreMapping(j).getDatastoreField()).getName().toString())).intValue();
	                    }
	                }
                }
            }

            if (m instanceof MappingCallbacks)
            {
                mc.add(m);
            }
        }

        public void consumeMapping(JavaTypeMapping m, int mappingType)
        {
            if (mappingType == MappingConsumer.MAPPING_TYPE_VERSION)
            {
                String val = ((RDBMSMapping)m.getDataStoreMapping(0)).getUpdateInputParameter();
                if (columnAssignments.length() > 0)
                {
                    columnAssignments.append(", ");
                }
                String condition = ((Column)m.getDataStoreMapping(0).getDatastoreField()).getName() + " = " + val;
                columnAssignments.append(condition);

                mappingStatementIndex.getVersion().setMapping(m);
                int[] param = { paramIndex++ };
                mappingStatementIndex.getVersion().setParameterIndex(param);
            }
            else if (mappingType == MappingConsumer.MAPPING_TYPE_DATASTORE_ID)
            {
                String condition = ((Column)key.getColumns().get(0)).getName() + " = ? ";
                where.append(condition);

                int[] param = { paramIndex++ };
                mappingStatementIndex.getDatastoreId().setParameterIndex(param);
            }         
        }

        /**
         * @return Returns the mappingCallbacks.
         */
        public List getMappingCallbacks()
        {
            return mc;
        }

        /**
         * All non-primary key fields to be provided to set the values in the sql statement.
         * The StateManager will call the Persistent capable class instance to provide fields
         */
        public int[] getFieldsToBeProvided()
        {
            int[] fieldsToBeProvided = new int[ftbp.size()];
            for (int i = 0; i < ftbp.size(); ++i)
            {
                fieldsToBeProvided[i] = ((Integer) ftbp.get(i)).intValue();
            }
            return fieldsToBeProvided;
        }

        /**
         * All primary key fields to be provided to set the values in the sql statement.
         * The StateManager will call the Persistent capable class instance to provide fields
         */
        public int[] getPrimaryKeyFieldsToBeProvided()
        {
            int[] pkfieldsToBeProvided = new int[pkbp.size()];
            for (int i = 0; i < pkbp.size(); i++)
            {
                pkfieldsToBeProvided[i] = ((Integer) pkbp.get(i)).intValue();
            }
            return pkfieldsToBeProvided;
        }

        public String getStatement()
        {
            if (columnAssignments.length() < 1)
            {
                return null;
            }
            /*
             * create the UPDATE SQL
             * e.g.
             * UPDATE TABLE SET COL1 = ?, COL2 = ? WHERE COL3 = ? AND COL4 = ? 
             */
            return "UPDATE " + table.toString() + " SET " + columnAssignments + " WHERE " + where;
        }
    }

    private class VersionMappingConsumer implements MappingConsumer
    {

        boolean initialized = false;
        boolean hasVersionMapping = false;
        int paramIndex = 1;
        int exprIndex = 1;
        
        List pkbp = new ArrayList();

        /** for SELECT statement **/
        FetchStatement fetchStmt = new RDBMSFetchStatement(table);
        
        public VersionMappingConsumer()
        {
            super();
        }
        
        public void preConsumeMapping(int highestFieldNumber)
        {
            if( !initialized )
            {
                versionMappingStatementIndex.setPrimaryKeys(new StatementExpressionIndex[highestFieldNumber]);
                initialized = true;
            }            
        }        

        public void consumeMapping(JavaTypeMapping m, FieldMetaData fmd)
        {
            if( !fmd.getClassMetaData().isSameOrAncestorOf(cmd) )
            {
                return;
            }            
            if (m.includeInUpdateStatement())
            {
                int parametersIndex[] = new int[m.getNumberOfDatastoreFields()];

                StatementExpressionIndex sei = new StatementExpressionIndex();
                sei.setMapping(m);
                sei.setParameterIndex(parametersIndex);

                if (m.getFieldMetaData().isPrimaryKey())
                {
                    versionMappingStatementIndex.getPrimaryKeys()[fmd.getAbsoluteFieldNumber()] = sei; 
                    for (int j=0; j<parametersIndex.length; j++)
                    {                   
                        String condition = ((Column)m.getDataStoreMapping(j).getDatastoreField()).getName() + " = " + ((RDBMSMapping)m.getDataStoreMapping(j)).getUpdateInputParameter();
                        
                        fetchStmt.andCondition(condition);

                        pkbp.add(new Integer(fmd.getAbsoluteFieldNumber()));
                        parametersIndex[j] = paramIndex++;
                    }                       
                }
            }
        }

        public void consumeMapping(JavaTypeMapping m, int mappingType)
        {
            if( mappingType == MappingConsumer.MAPPING_TYPE_VERSION)
            {
                hasVersionMapping = true;

                fetchStmt.select(m.getDataStoreMapping(0).getDatastoreField());
                
                versionMappingStatementIndex.getVersion().setMapping(m);
                versionMappingStatementIndex.getVersion().setExpressionIndex(new int[]{ exprIndex++ });
            }             
            else if( mappingType == MappingConsumer.MAPPING_TYPE_DATASTORE_ID)
            {
                String condition = ((Column)key.getColumns().get(0)).getName() + " = ? ";
                fetchStmt.andCondition(condition);
                
                int[] param = { paramIndex++ };
                versionMappingStatementIndex.getDatastoreId().setParameterIndex(param);
            }            
        }
        
        /**
         * All primary key fields to be provided to set the values in the sql statement.
         * The StateManager will call the Persistent capable class instance to provide fields
         */        
        public int[] getPrimaryKeyFieldsToBeProvided()
        {
            int[] pkfieldsToBeProvided = new int[pkbp.size()];
            for (int i = 0; i < pkbp.size(); i++)
            {
                pkfieldsToBeProvided[i] = ((Integer) pkbp.get(i)).intValue();
            }
            return pkfieldsToBeProvided;
        }       
               
        public String getStatement()
        {
            if( !hasVersionMapping )
            {
                return null;
            }
            return fetchStmt.toString();
        }        
    }
}