/**********************************************************************
Copyright (c) 2003 David Jencks and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - coding standards
2003 Andy Jefferson - updated to support inherited objects
2003 Andy Jefferson - revised logging
2004 Andy Jefferson - merged IteratorStmt and GetStmt into GetRangeStmt
    ...
**********************************************************************/
package org.jpox.store.rdbms.scostore;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Collections;
import java.util.Iterator;
import java.util.ListIterator;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOHelper;
import javax.jdo.JDOObjectNotFoundException;
import javax.jdo.JDOUserException;
import javax.jdo.spi.PersistenceCapable;

import org.jpox.ClassLoaderResolver;
import org.jpox.FetchPlanImpl;
import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.Transaction;
import org.jpox.FetchPlanImpl.FetchPlanForClass;
import org.jpox.metadata.ClassMetaData;
import org.jpox.metadata.DiscriminatorStrategy;
import org.jpox.metadata.IdentityType;
import org.jpox.store.DatastoreAdapter;
import org.jpox.store.StatementExpressionIndex;
import org.jpox.store.StoreManager;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.TableExpression;
import org.jpox.store.mapping.InterfaceMapping;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.MappingCallbacks;
import org.jpox.store.mapping.Mappings;
import org.jpox.store.mapping.OIDMapping;
import org.jpox.store.query.PersistentIDROF;
import org.jpox.store.query.Query;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.adapter.DatabaseAdapter;
import org.jpox.store.rdbms.mapping.RDBMSMapping;
import org.jpox.store.rdbms.sqlidentifier.SQLIdentifier;
import org.jpox.store.rdbms.sqlidentifier.TableIdentifier;
import org.jpox.store.rdbms.table.ClassTable;
import org.jpox.store.rdbms.table.Table;
import org.jpox.store.scostore.ListStore;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;
import org.jpox.util.SQLWarnings;

/**
 * Abstract representation of a backing store for a List.
 *
 * @version $Revision: 1.9 $
 **/
abstract class AbstractListStore implements ListStore
{
    protected static final Localiser LOCALISER=Localiser.getInstance("org.jpox.store.rdbms.scostore.Localisation");
    protected static final Localiser STORE_LOCALISER=Localiser.getInstance("org.jpox.store.rdbms.Localisation");

    protected Table listTable;
    protected ClassTable elementTable;

    protected String listName;
    protected StoreManager storeMgr;
    protected DatastoreAdapter dba;
    protected JavaTypeMapping ownerMapping;
    protected JavaTypeMapping elementMapping;
    protected JavaTypeMapping elementDiscriminatorMapping;
    protected JavaTypeMapping indexMapping;
    protected String elementType;
    protected boolean elementsAreEmbedded;

    // Statements for operations
    protected String addStmt;
    protected String sizeStmt;
    protected String containsStmt;
    protected String clearStmt;
    protected String removeAtStmt;
    protected String setStmt;
    protected String shiftStmt;
    protected String indexOfStmt;
    protected String lastIndexOfStmt;

    protected ClassMetaData emd;

    private Query.ResultObjectFactory rof = null;

    /**
     * Constructor - protected to prevent instantiation.
     **/
    protected AbstractListStore()
    {
    }

    /**
     * Method to initialise the statements being used.
     * Subclasses should override the getXXXStmt() if they want to use an
     * alternative statement.
     **/
    protected void initialiseStatements()
    {
        // Our statements use listTable as the place to select/insert/delete
        addStmt         = getAddStmt();
        containsStmt    = getContainsStmt();
        clearStmt       = getClearStmt();
        indexOfStmt     = getIndexOfStmt();
        lastIndexOfStmt = getLastIndexOfStmt();
        removeAtStmt    = getRemoveAtStmt();
        setStmt         = getSetStmt();
        shiftStmt       = getShiftStmt();
        sizeStmt        = getSizeStmt();
    }

    /**
     * Accessor for the RDBMSManager.
     * @return The RDBMSManager.
     **/
    public StoreManager getStoreManager()
    {
        return storeMgr;
    }

    /**
     * Accessor for the element type stored in this List.
     * @return The element type.
     **/
    public String getElementType()
    {
        return elementType;
    }

    /**
     * Accessor for the owner mapping.
     * @return The Owner table mapping.
     **/
    public JavaTypeMapping getOwnerMapping()
    {
        return ownerMapping;    
    }

    // -------------------------- List Method implementations ------------------
 
    /**
     * Accessor for an iterator through the list elements.
     * @param sm State Manager for the container.
     * @return The Iterator
     **/
    public Iterator iterator(StateManager sm)
    {
        return listIterator(sm);
    }

    /**
     * Accessor for an iterator through the list elements.
     * @param sm State Manager for the container.
     * @return The List Iterator
     **/
    public ListIterator listIterator(StateManager sm)
    {
        return listIterator(sm, 0);
    }

    /**
     * Accessor for an iterator through the list elements.
     * @param sm State Manager for the container.
     * @param start The start point in the list.
     * @return The List Iterator
     **/
    public ListIterator listIterator(StateManager sm, int start)
    {
        ListIterator i;
        PersistenceManager pm = sm.getPersistenceManager();

        QueryStatement stmt=null;
        if (start >= 0)
        {
            stmt = getGetRangeStatement(sm,start,-1);
        }
        else
        {
            stmt = getGetRangeStatement(sm,-1,-1);
        }
        rof = newResultObjectFactory(sm,stmt,false);

        Transaction tx = (Transaction) pm.currentTransaction();
        String statement = stmt.toStatementText(tx.useUpdateLockOnFetch()).toString();

        try
        {
            Connection conn = pm.getConnection(false);

            try
            {
                PreparedStatement ps = stmt.toStatementText(tx.useUpdateLockOnFetch()).prepareStatement(pm,conn);
                try
                {
                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(statement);
                    }
                    long startTime = System.currentTimeMillis();
                    ResultSet rs = ps.executeQuery();
                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(STORE_LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                    }

                    try
                    {
                        i = new JPOXListIterator(sm, rs);
                    }
                    finally
                    {
                        rs.close();
                    }
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                pm.releaseConnection(conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("SCO.IteratorRequestFailed",statement),e);
        }

        return i;
    }

    /**
     * Inner class providing iterator.
     **/
    private class JPOXListIterator implements ListIterator
    {
        private final StateManager sm;
        private final PersistenceManager pm;
        private final ListIterator delegate;
        private Object lastElement = null;
        private int currentIndex = -1;

        public JPOXListIterator(StateManager sm, ResultSet rs) 
        throws SQLException
        {
            this.sm = sm;
            this.pm = sm.getPersistenceManager();

            ArrayList results = new ArrayList();

            if (rs != null)
            {
                while (rs.next())
                {
                    Object nextElement;
                    if (!isEmbeddedMapping(elementMapping))
                    {
                        // OID type - retrieve the object
                        nextElement = rof.getObject(pm, rs, pm.getClassLoaderResolver().classForName(elementType));
                    }
                    else
                    {
                        // Primitive type - read the object direct from mapping
                        int param[] = new int[elementMapping.getNumberOfDatastoreFields()];
                        for (int i = 0; i < param.length; ++i)
                        {
                            param[i] = i + 1;
                        }
                        nextElement = elementMapping.getObject(pm,rs,param);
                    }
                    
                    results.add(nextElement);
                }
            }
            delegate = results.listIterator();
        }

        public void add(Object o)
        {
            currentIndex = delegate.nextIndex();
            AbstractListStore.this.add(sm, currentIndex, o);
            delegate.add(o);
            lastElement = null;
        }

        public boolean hasNext()
        {
            return delegate.hasNext();
        }

        public boolean hasPrevious()
        {
            return delegate.hasPrevious();
        }

        public Object next()
        {
            currentIndex = delegate.nextIndex();
            lastElement = delegate.next();

            return lastElement;
        }

        public int nextIndex()
        {
            return delegate.nextIndex();
        }

        public Object previous()
        {
            currentIndex = delegate.previousIndex();
            lastElement = delegate.previous();

            return lastElement;
        }

        public int previousIndex()
        {
            return delegate.previousIndex();
        }

        public synchronized void remove()
        {
            if (lastElement == null)
            {
                throw new IllegalStateException("No entry to remove");
            }

            AbstractListStore.this.remove(sm, currentIndex);
            delegate.remove();

            lastElement = null;
            currentIndex = -1;
        }

        public synchronized void set(Object o)
        {
            if (lastElement == null)
            {
                throw new IllegalStateException("No entry to replace");
            }

            AbstractListStore.this.set(sm, currentIndex, o);
            delegate.set(o);

            lastElement = o;
        }
    }

    /**
     * Method to add an element to the List.
     * @param sm The state manager
     * @param element The element to remove
     * @return Whether it was added successfully.
     **/
    public boolean add(StateManager sm, Object element)
    {
        return internalAdd(sm, 0, true, Collections.singleton(element));
    }

    /**
     * Method to add an element to the List.
     * @param sm The state manager.
     * @param index The location to add at
     * @param element The element to add.
     **/
    public void add(StateManager sm, int index, Object element)
    {
        internalAdd(sm, index, false, Collections.singleton(element));
    }

    /**
     * Method to add a collection of elements to the List.
     * @param sm The state manager
     * @param elements The elements to remove
     * @return Whether they were added successfully.
     **/
    public boolean addAll(StateManager sm, Collection elements)
    {
        return internalAdd(sm, 0, true, elements);
    }

    /**
     * Method to add all elements from a Collection to the List.
     * @param sm The state manager
     * @param index The location to add at
     * @param c The collection
     * @return Whether it was successful
     **/
    public boolean addAll(StateManager sm, int index, Collection c)
    {
        return internalAdd(sm, index, false, c);
    }

    /**
     * Method to clear the List.
     * @param sm The state manager
     **/
    public void clear(StateManager sm)
    {
        internalClear(sm, clearStmt);
    }

    /**
     * Method to check for existence of an element in the List.
     * @param sm The state manager
     * @param element The element
     * @return Whether it was found.
     **/
    public boolean contains(StateManager sm, Object element)
    {
        if (!validateElementForReading(sm, element))
        {
            return false;
        }

        boolean retval;
        PersistenceManager pm = sm.getPersistenceManager();

        try
        {
            Connection conn = pm.getConnection(false);

            try
            {
                PreparedStatement ps = conn.prepareStatement(containsStmt);
                try
                {
                    int jdbc_position = 1;
                    ownerMapping.setObject(pm, ps, Mappings.getParametersIndex(jdbc_position,ownerMapping), sm.getObject());
                    jdbc_position += ownerMapping.getNumberOfDatastoreFields();
                    elementMapping.setObject(pm, ps, Mappings.getParametersIndex(jdbc_position,elementMapping), element);
                    jdbc_position += elementMapping.getNumberOfDatastoreFields();
                    if (elementDiscriminatorMapping != null)
                    {
                        if (emd.getDiscriminatorStrategy() == DiscriminatorStrategy.CLASS_NAME)
                        {
                            elementDiscriminatorMapping.setObject(pm, ps, Mappings.getParametersIndex(jdbc_position, elementDiscriminatorMapping), elementType);
                        }
                        else if (emd.getDiscriminatorStrategy() == DiscriminatorStrategy.VALUE_MAP)
                        {
                            elementDiscriminatorMapping.setObject(pm, ps, Mappings.getParametersIndex(jdbc_position, elementDiscriminatorMapping),
                                emd.getInheritanceMetaData().getDiscriminatorMetaData().getValue());
                        }
                    }

                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(containsStmt);
                    }
                    long startTime = System.currentTimeMillis();
                    ResultSet rs = ps.executeQuery();
                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(STORE_LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                    }

                    try
                    {
                        retval = rs.next();
                        SQLWarnings.log(rs);
                    }
                    finally
                    {
                        rs.close();
                    }
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                pm.releaseConnection(conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("SCO.ContainsRequestFailed",containsStmt),e);
        }

        return retval;
    }

    /**
     * Method to retrieve an element from the List.
     * @param sm The state manager
     * @param index The location of the element.
     * @return The object
     **/
    public Object get(StateManager sm, int index)
    {
        java.util.List list=internalGetRange(sm,index,index);

        if (list == null || list.size() == 0)
        {
            return null;
        }

        return list.get(0);
    }

    /**
     * Accessor for the indexOf an object in the List.
     * @param sm The state manager
     * @param element The element.
     * @return The index
     **/
    public int indexOf(StateManager sm, Object element)
    {
        return internalIndexOf(sm, element, indexOfStmt);
    }

    /**
     * Method to retrieve the last index of an object in the list.
     * @param sm The state manager.
     * @param element The object
     * @return The last index
     **/
    public int lastIndexOf(StateManager sm, Object element)
    {
        return internalIndexOf(sm, element, lastIndexOfStmt);
    }

    /**
     * Method to remove an element from the List.
     * @param sm The state manager
     * @param element The element to remove.
     * @return Whether it was removed successfully.
     **/
    public boolean remove(StateManager sm, Object element)
    {
        if (!validateElementForReading(sm, element))
        {
            return false;
        }

        int index = indexOf(sm, element);
        if (index == -1)
        {
            return false;
        }
        removeAt(sm, index);
        return true;
    }

    /**
     * Remove all elements from a collection from the association owner vs
     * elements.
     * TODO : Change the query to do all in one go for efficiency. Currently
     * removes an element and shuffles the indexes, then removes an element
     * and shuffles the indexes, then removes an element and shuffles the
     * indexes etc ... a bit inefficient !!!
     * @param sm State Manager for the container
     * @param elements Collection of elements to remove 
     * @return Whether the database was updated 
     */
    public boolean removeAll(StateManager sm, Collection elements)
    {
        if (elements == null || elements.size() == 0)
        {
            return false;
        }

        boolean modified = false;

        // Get the indices of the elements to remove
        int[] indices=getIndicesOf(sm,elements);

        // Remove each element in turn, doing the shifting of indexes each time
        // TODO : Change this to remove all in one go and then shift once 
        for (int i=indices.length-1;i>=0;i--)
        {
            removeAt(sm,indices[i]);
            modified = true;
        }

        return modified;
    }

    /**
     * Method to remove an object at a location in the List.
     * @param sm The state manager.
     * @param index The location
     * @return The object that was removed. 
     **/
    public Object remove(StateManager sm, int index)
    {
        Object o = get(sm, index);
        removeAt(sm, index);
        return o;
    }

    /**
     * Method to set an object in the List.
     * @param sm The state manager
     * @param index The item index
     * @param element What to set it to.
     * @return The value before setting.
     **/
    public Object set(StateManager sm, int index, Object element)
    {
        validateElementForWriting(sm, element);
        Object o = get(sm, index);
        PersistenceManager pm = sm.getPersistenceManager();

        try
        {
            Connection conn = pm.getConnection(true);
            try
            {
                PreparedStatement ps = conn.prepareStatement(setStmt);
                try
                {
                    elementMapping.setObject(pm, ps, org.jpox.store.mapping.Mappings.getParametersIndex(1,elementMapping), element);
                    ownerMapping.setObject(pm, ps, org.jpox.store.mapping.Mappings.getParametersIndex(elementMapping.getNumberOfDatastoreFields()+1,ownerMapping), sm.getObject());
                    indexMapping.setObject(pm, ps, org.jpox.store.mapping.Mappings.getParametersIndex(elementMapping.getNumberOfDatastoreFields()+ownerMapping.getNumberOfDatastoreFields()+1,indexMapping), new Integer(index));

                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(setStmt);
                    }
                    long startTime = System.currentTimeMillis();
                    ps.executeUpdate();
                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(STORE_LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                    }
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                pm.releaseConnection(conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("SCO.SetRequestFailed",setStmt),e);
        }
        return o;
    }

    /**
     * Method to return the size of the List.
     * @param sm The state manager.
     * @return The size.
     **/
    public int size(StateManager sm)
    {
        int numRows;
        PersistenceManager pm = sm.getPersistenceManager();

        try
        {
            Connection conn = pm.getConnection(false);
            try
            {
                PreparedStatement ps = conn.prepareStatement(sizeStmt);
                try
                {
                    int jdbc_position = 1;
                    ownerMapping.setObject(pm, ps, Mappings.getParametersIndex(jdbc_position,ownerMapping), sm.getObject());
                    jdbc_position += ownerMapping.getNumberOfDatastoreFields();
                    if (elementDiscriminatorMapping != null)
                    {
                        if (emd.getDiscriminatorStrategy() == DiscriminatorStrategy.CLASS_NAME)
                        {
                            elementDiscriminatorMapping.setObject(pm, ps, Mappings.getParametersIndex(jdbc_position, elementDiscriminatorMapping), elementType);
                        }
                        else if (emd.getDiscriminatorStrategy() == DiscriminatorStrategy.VALUE_MAP)
                        {
                            elementDiscriminatorMapping.setObject(pm, ps, Mappings.getParametersIndex(jdbc_position, elementDiscriminatorMapping),
                                emd.getInheritanceMetaData().getDiscriminatorMetaData().getValue());
                        }
                    }

                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(sizeStmt);
                    }
                    long startTime = System.currentTimeMillis();
                    ResultSet rs = ps.executeQuery();
                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(STORE_LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                    }

                    try
                    {
                        if (!rs.next())
                        {
                            throw new JDODataStoreException(LOCALISER.msg("SCO.SizeRequestFailed",sizeStmt));
                        }

                        numRows = rs.getInt(1);
                        SQLWarnings.log(rs);
                    }
                    finally
                    {
                        rs.close();
                    }
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                pm.releaseConnection(conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("SCO.SizeRequestFailed",sizeStmt),e);
        }

        return numRows;
    }

    /**
     * Method to retrieve a list of elements in a range.
     * @param sm The state manager.
     * @param start From index (inclusive)
     * @param end To index (exclusive)
     * @return Sub List of elements in this range.
     **/
    public java.util.List subList(StateManager sm, int start, int end)
    {
        return internalGetRange(sm,start,end);
    }

    // -------------------------- Common Statements ----------------------------

    /**
     * Generates the statement for adding items.
     * <PRE> 
     * INSERT INTO LISTTABLE (OWNERCOL,ELEMENTCOL,INDEXCOL) 
     * VALUES (?,?,?)
     * </PRE> 
     * @return The Statement for adding an item
     */
    private String getAddStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("INSERT INTO ");
        stmt.append(listTable.toString());
        stmt.append(" (");
        stmt.append(((Column)ownerMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
        for (int i=1; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(",");
            stmt.append(((Column)ownerMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
        }
        for (int i=0; i<elementMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(",");
            stmt.append(((Column)elementMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
        }
        for (int i=0; i<indexMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(",");
            stmt.append(((Column)indexMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
        }
        stmt.append(") VALUES (");
        stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(0)).getInsertionInputParameter());
        for (int i=1; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(",");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getInsertionInputParameter());
        }
        for (int i=0; i<elementMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(",");
            stmt.append(((RDBMSMapping)elementMapping.getDataStoreMapping(0)).getInsertionInputParameter());
        }
        for (int i=0; i<indexMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(",");
            stmt.append(((RDBMSMapping)indexMapping.getDataStoreMapping(0)).getInsertionInputParameter());
        }
        stmt.append(") ");
        return stmt.toString();
    }

    /**
     * Generates the statement for clearing the List.
     * <PRE> 
     * DELETE FROM LISTTABLE
     * WHERE OWNERCOL = ?
     * </PRE>
     * @return The Statement for clearing the List
     */
    protected String getClearStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("DELETE FROM ");
        stmt.append(listTable.toString());
        stmt.append(" WHERE ");
        stmt.append(((Column)ownerMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
        stmt.append(" = ");
        stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(0)).getUpdateInputParameter());
        for (int i=1; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(((Column)ownerMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        return stmt.toString();
    }

    /**
     * Generates the statement to check contains.
     * The discriminator part is only present when the element type has
     * inheritance strategy of "superclass-table" and is Inverse.
     * <PRE> 
     * SELECT OWNERCOL FROM LISTTABLE 
     * WHERE OWNERCOL=? 
     * AND ELEMENTCOL=?
     * [AND DISCRIMINATOR = ?]
     *</PRE> 
     * @return The Statement for checking if an item is contained in the List 
     */
    protected String getContainsStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("SELECT ");
        stmt.append(((Column)ownerMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
        for (int i=1; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(",");
            stmt.append(((Column)ownerMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
        }
        stmt.append(" FROM ");
        stmt.append(listTable.toString());
        stmt.append(" WHERE ");
        stmt.append(((Column)ownerMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
        stmt.append(" = ");
        stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(0)).getUpdateInputParameter());
        for (int i=1; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(((Column)ownerMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        for (int i=0; i<elementMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(((Column)elementMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)elementMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }

        // Only for inverse cases with a discriminator
        if (elementDiscriminatorMapping != null)
        {
            for (int i=0;i<elementDiscriminatorMapping.getNumberOfDatastoreFields();i++)
            {
                stmt.append(" AND ");
                stmt.append(((Column)elementDiscriminatorMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
                stmt.append(" = ");
                stmt.append(((RDBMSMapping)elementDiscriminatorMapping.getDataStoreMapping(i)).getUpdateInputParameter());
            }
        }

        return stmt.toString();
    }

    /**
     * Generate statement for getting the index of an item.
     * <PRE>
     * SELECT INDEXCOL FROM LISTTABLE
     * WHERE OWNERCOL=? 
     * AND ELEMENTCOL=? 
     * ORDER BY INDEXCOL
     * </PRE>
     * @return The Statement for getting the index of an item 
     */
    protected String getIndexOfStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("SELECT ");
        stmt.append(((Column)indexMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
        for (int i=1; i<indexMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(",");
            stmt.append(((Column)indexMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
        }
        stmt.append(" FROM ");
        stmt.append(listTable.toString());
        stmt.append(" WHERE ");
        stmt.append(((Column)ownerMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
        stmt.append(" = ");
        stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(0)).getUpdateInputParameter());
        for (int i=1; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(((Column)ownerMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        for (int i=0; i<elementMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(((Column)elementMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)elementMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        stmt.append(" ORDER BY ");
        stmt.append(((Column)indexMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
        for (int i=1; i<indexMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(",");
            stmt.append(((Column)indexMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
        }
        return stmt.toString();
    }

    /**
     * Generates the statement for getting the index of the last item.
     * <PRE> 
     * SELECT INDEXCOL FROM LISTTABLE 
     * WHERE OWNERCOL=? 
     * AND ELEMENTCOL=? 
     * ORDER BY INDEXCOL DESC
     * </PRE>
     * @return The Statement for getting the last item 
     */
    protected String getLastIndexOfStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("SELECT ");
        stmt.append(((Column)indexMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
        for (int i=1; i<indexMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(",");
            stmt.append(((Column)indexMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
        }
        stmt.append(" FROM ");
        stmt.append(listTable.toString());
        stmt.append(" WHERE ");
        stmt.append(((Column)ownerMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
        stmt.append(" = ");
        stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(0)).getUpdateInputParameter());
        for (int i=1; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(((Column)ownerMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        for (int i=0; i<elementMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(((Column)elementMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)elementMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        stmt.append(" ORDER BY ");
        stmt.append(((Column)indexMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
        stmt.append(" DESC ");
        for (int i=1; i<indexMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(",");
            stmt.append(((Column)indexMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" DESC ");
        }
        return stmt.toString();
    }

    /**
     * Generates the statement for removing an item.
     * <PRE> 
     * DELETE FROM LISTTABLE 
     * WHERE OWNERCOL = ? 
     * AND INDEXCOL = ?
     * </PRE>
     * @return The Statement for removing an item from a position
     */
    protected String getRemoveAtStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("DELETE FROM ");
        stmt.append(listTable.toString());
        stmt.append(" WHERE ");
        stmt.append(((Column)ownerMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
        stmt.append(" = ");
        stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(0)).getUpdateInputParameter());
        for (int i=1; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(((Column)ownerMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        for (int i=0; i<indexMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(((Column)indexMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)indexMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        return stmt.toString();
    }

    /**
     * Generates the statement for setting an item.
     * <PRE> 
     * UPDATE LISTTABLE SET ELEMENTCOL = ?
     * WHERE OWNERCOL = ? 
     * AND INDEXCOL = ?
     * </PRE>
     * @return The Statement for setting an item
     */
    protected String getSetStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("UPDATE ");
        stmt.append(listTable.toString());
        stmt.append(" SET ");
        stmt.append(((Column)elementMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
        stmt.append(" = ");
        stmt.append(((RDBMSMapping)elementMapping.getDataStoreMapping(0)).getUpdateInputParameter());
        for (int i=1; i<elementMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(",");
            stmt.append(((Column)elementMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)elementMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        stmt.append(" WHERE ");
        stmt.append(((Column)ownerMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
        stmt.append(" = ");
        stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(0)).getUpdateInputParameter());
        for (int i=1; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(((Column)ownerMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        for (int i=0; i<indexMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(((Column)indexMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)indexMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }       
        return stmt.toString();
    }

    /**
     * Generates the statement for shifting items.
     * <PRE> 
     * UPDATE LISTTABLE SET ELEMENTCOL = ? 
     * WHERE OWNERCOL = ? 
     * AND INDEXCOL = ?
     * </PRE>
     * @return The Statement for shifting elements
     */
    protected String getShiftStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("UPDATE ");
        stmt.append(listTable.toString());
        stmt.append(" SET ");
        stmt.append(((Column)indexMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
        stmt.append(" = ");
        stmt.append(((RDBMSMapping)indexMapping.getDataStoreMapping(0)).getUpdateInputParameter());
        stmt.append(" + ");
        stmt.append(((Column)indexMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
        for (int i=1; i<indexMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(",");
            stmt.append(((Column)indexMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)indexMapping.getDataStoreMapping(i)).getUpdateInputParameter());
            stmt.append(" + ");
            stmt.append(((Column)indexMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
        }
        stmt.append(" WHERE ");
        stmt.append(((Column)ownerMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
        stmt.append(" = ");
        stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(0)).getUpdateInputParameter());
        for (int i=1; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(((Column)ownerMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        for (int i=0; i<indexMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(((Column)indexMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)indexMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        return stmt.toString();
    }

    /**
     * Generate statement for getting the size of the list.
     * The discriminator part is only present when the element type has
     * inheritance strategy of "superclass-table" and is Inverse.
     * <PRE>
     * SELECT COUNT(*) FROM LISTTABLE
     * WHERE OWNERCOL=?
     * AND INDEXCOL IS NOT NULL
     * [AND DISCRIMINATOR=?]
     * </PRE> 
     * @return The Statement returning the size of the List
     */
    private String getSizeStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("SELECT COUNT(*) FROM ");
        stmt.append(listTable.toString());
        stmt.append(" WHERE ");
        stmt.append(((Column)ownerMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
        stmt.append(" = ");
        stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(0)).getUpdateInputParameter());
        for (int i=1; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(((Column)ownerMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        // We need to restrict to items where the index is not null to
        // eliminate records that are added but not positioned yet.
        for (int i=0; i<indexMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(((Column)indexMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" IS NOT NULL");
        }

        // Only for inverse cases with a discriminator
        if (elementDiscriminatorMapping != null)
        {
            for (int i=0;i<elementDiscriminatorMapping.getNumberOfDatastoreFields();i++)
            {
                stmt.append(" AND ");
                stmt.append(((Column)elementDiscriminatorMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
                stmt.append(" = ");
                stmt.append(((RDBMSMapping)elementDiscriminatorMapping.getDataStoreMapping(i)).getUpdateInputParameter());
            }
        }
        return stmt.toString();
    }

    /**
     * Generates the statement for getting the indices of a collection of
     * element. Ordered into ascending order, so they bear no resemblance to
     * the order of the elements in the Collection.
     * <PRE> 
     * SELECT INDEXCOL FROM LISTTABLE 
     * WHERE (OWNERCOL=? AND ELEMENT_COL=?) OR
     *       (OWNERCOL=? AND ELEMENT_COL=?) OR
     *       (OWNERCOL=? AND ELEMENT_COL=?) 
     * ORDER BY INDEXCOL
     * </PRE>
     * @param elements The elements to retrieve the indices for. 
     * @return The Statement for getting the indices of the collection.
     */
    protected String getIndicesOfStmt(Collection elements)
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("SELECT ");
        stmt.append(((Column)indexMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
        for (int i=1; i<indexMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(",");
            stmt.append(((Column)indexMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
        }
        stmt.append(" FROM ");
        stmt.append(listTable.toString());
        stmt.append(" WHERE ");
        Iterator iter=elements.iterator();
        boolean first_element=true;
        while (iter.hasNext())
        {
            iter.next(); // Move to next element

            if (!first_element)
            {
                stmt.append(" OR (");
            }
            else
            {
                stmt.append("(");
            }

            stmt.append(((Column)ownerMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(0)).getUpdateInputParameter());
            for (int i=1; i<ownerMapping.getNumberOfDatastoreFields(); i++)
            {
                stmt.append(" AND ");
                stmt.append(((Column)ownerMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
                stmt.append(" = ");
                stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
            }
            for (int i=0; i<elementMapping.getNumberOfDatastoreFields(); i++)
            {
                stmt.append(" AND ");
                stmt.append(((Column)elementMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
                stmt.append(" = ");
                stmt.append(((RDBMSMapping)elementMapping.getDataStoreMapping(i)).getUpdateInputParameter());
            }
            stmt.append(")");
            first_element = false;
        }
        stmt.append(" ORDER BY ");
        stmt.append(((Column)indexMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
        for (int i=1; i<indexMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(",");
            stmt.append(((Column)indexMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" DESC");
        }

        return stmt.toString();
    }

    /**
     * Method to prepare the indicesOf statement for use. Populates the various
     * parameters. This is required because the query is built dynamically
     * depending on the number of elements to retrieve the indices for.
     * @param sm State Manager of the container.
     * @param ps The Prepared Statement
     * @param elements Collection of elements
     **/
    protected void prepareIndicesOfStmt(StateManager sm,
                                        PreparedStatement ps,
                                        Collection elements)
    {
        if (elements == null || elements.size() == 0)
        {
            return;
        }

        PersistenceManager pm = sm.getPersistenceManager();
        Iterator iter=elements.iterator();
        int jdbc_position=1;
        while (iter.hasNext())
        {
            Object element=iter.next();

            ownerMapping.setObject(pm, ps, org.jpox.store.mapping.Mappings.getParametersIndex(jdbc_position,ownerMapping), sm.getObject());
            jdbc_position += ownerMapping.getNumberOfDatastoreFields();
            elementMapping.setObject(pm, ps, org.jpox.store.mapping.Mappings.getParametersIndex(jdbc_position,elementMapping), element);
            jdbc_position += elementMapping.getNumberOfDatastoreFields();
        }
    }

    /**
     * Method to remove an element from the specified position
     * @param sm The State Manager for the list
     * @param index The index of the element
     */
    protected void removeAt(StateManager sm, int index)
    {
        internalRemoveAt(sm, index, removeAtStmt);
    }
    
    // -------------------------- Internal Methods -----------------------------
 
    /**
     * Internal method to remove an object at a location in the List.
     * @param sm The state manager.
     * @param index The location
     * @param stmt The statement to remove the element from the List
     **/
    protected void internalRemoveAt(StateManager sm, int index, String stmt)
    {
        int size = size(sm);
        PersistenceManager pm = sm.getPersistenceManager();

        try
        {
            Connection conn = pm.getConnection(true);
            try
            {
                PreparedStatement ps = conn.prepareStatement(stmt);
                try
                {
                    ownerMapping.setObject(pm, ps, org.jpox.store.mapping.Mappings.getParametersIndex(1,ownerMapping), sm.getObject());
                    indexMapping.setObject(pm, ps, org.jpox.store.mapping.Mappings.getParametersIndex(ownerMapping.getNumberOfDatastoreFields()+1,indexMapping), new Integer(index));

                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(stmt);
                    }
                    long startTime = System.currentTimeMillis();
                    int rowsDeleted = ps.executeUpdate();
                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(STORE_LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                    }

                    if (rowsDeleted == 0)
                    {
                        //?? throw exception??
                    }
                }
                finally
                {
                    ps.close();
                }

                //shift down
                if (index != size - 1)
                {
                    ps = conn.prepareStatement(shiftStmt);

                    try
                    {
                        for (int i=index + 1; i<size; i++)
                        {
                            indexMapping.setObject(pm, ps, org.jpox.store.mapping.Mappings.getParametersIndex(1,indexMapping), new Integer(-1));
                            ownerMapping.setObject(pm, ps, org.jpox.store.mapping.Mappings.getParametersIndex(indexMapping.getNumberOfDatastoreFields()+1,ownerMapping), sm.getObject());
                            indexMapping.setObject(pm, ps, org.jpox.store.mapping.Mappings.getParametersIndex(indexMapping.getNumberOfDatastoreFields()+ownerMapping.getNumberOfDatastoreFields()+1,indexMapping), new Integer(i));

                            if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                            {
                                JPOXLogger.RDBMS_SQL.debug(shiftStmt);
                            }
                            long startTime = System.currentTimeMillis();
                            ps.executeUpdate();
                            if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                            {
                                JPOXLogger.RDBMS_SQL.debug(STORE_LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                            }
                        }
                    }
                    finally
                    {
                        ps.close();
                    }
                }
            }
            finally
            {
                pm.releaseConnection(conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("RemoveRequestFailed",stmt),e);
        }
    }

    /**
     * Internal method to add element(s) to the List.
     * @param sm The state manager
     * @param start The start location (if required)
     * @param atEnd Whether to add the element at the end
     * @param c The collection of objects to add. 
     * @return Whether it was successful
     **/ 
    protected boolean internalAdd(StateManager sm, int start, boolean atEnd,  Collection c)
    {
        if (c == null)
        {
            return true;
        }
        int shift = c.size();
        if (shift == 0)
        {
            return false;
        }

        //check all elements
        for (Iterator i = c.iterator(); i.hasNext();)
        {
            validateElementForWriting(sm, i.next());
        }

        int size = size(sm);

        PersistenceManager pm = sm.getPersistenceManager();

        try
        {
            Connection conn = pm.getConnection(true);
            try
            {
                //shift up by c.size
                if (!atEnd && start != size)
                {
                    PreparedStatement ps = conn.prepareStatement(shiftStmt);
                    try
                    {
                        for (int i = size - 1; i >= start; i--)
                        {
                            indexMapping.setObject(pm, ps, org.jpox.store.mapping.Mappings.getParametersIndex(1,indexMapping), new Integer(shift));
                            ownerMapping.setObject(pm, ps, org.jpox.store.mapping.Mappings.getParametersIndex(indexMapping.getNumberOfDatastoreFields()+1,ownerMapping), sm.getObject());
                            indexMapping.setObject(pm, ps, org.jpox.store.mapping.Mappings.getParametersIndex(indexMapping.getNumberOfDatastoreFields()+ownerMapping.getNumberOfDatastoreFields()+1,indexMapping), new Integer(i));

                            if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                            {
                                JPOXLogger.RDBMS_SQL.debug(shiftStmt);
                            }
                            long startTime = System.currentTimeMillis();
                            ps.executeUpdate();
                            if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                            {
                                JPOXLogger.RDBMS_SQL.debug(STORE_LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                            }
                        }
                    }
                    finally
                    {
                        ps.close();
                    }
                }
                else
                {
                    start = size;
                }

                //insert at correct location.
                PreparedStatement ps = conn.prepareStatement(addStmt);
                try
                {
                    for (Iterator i = c.iterator(); i.hasNext(); )
                    {
                        ownerMapping.setObject(pm, ps, org.jpox.store.mapping.Mappings.getParametersIndex(1,ownerMapping), sm.getObject());
                        elementMapping.setObject(pm, ps, org.jpox.store.mapping.Mappings.getParametersIndex(ownerMapping.getNumberOfDatastoreFields()+1,elementMapping), i.next());
                        indexMapping.setObject(pm, ps, org.jpox.store.mapping.Mappings.getParametersIndex(elementMapping.getNumberOfDatastoreFields()+ownerMapping.getNumberOfDatastoreFields()+1,indexMapping), new Integer(start));
                        start++;

                        if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                        {
                            JPOXLogger.RDBMS_SQL.debug(addStmt);
                        }
                        long startTime = System.currentTimeMillis();
                        ps.executeUpdate();
                        if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                        {
                            JPOXLogger.RDBMS_SQL.debug(STORE_LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                        }
                    }
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                pm.releaseConnection(conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("SCO.AddRequestFailed",addStmt),e);
        }
        return true;
    }
 
    /**
     * Internal method to find the index of an element.
     * @param sm The state manager.
     * @param element The element
     * @param stmt The statement to find the element.
     * @return The index of the element in the List.
     **/
    private int internalIndexOf(StateManager sm, Object element, String stmt)
    {
        validateElementForReading(sm, element);

        PersistenceManager pm = sm.getPersistenceManager();

        try
        {
            Connection conn = pm.getConnection(false);
            try
            {
                PreparedStatement ps = conn.prepareStatement(stmt);
                try
                {
                    ownerMapping.setObject(pm, ps, org.jpox.store.mapping.Mappings.getParametersIndex(1,ownerMapping), sm.getObject());
                    elementMapping.setObject(pm, ps, org.jpox.store.mapping.Mappings.getParametersIndex(ownerMapping.getNumberOfDatastoreFields()+1,elementMapping), element);

                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(stmt);
                    }
                    long startTime = System.currentTimeMillis();
                    ResultSet rs = ps.executeQuery();
                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(STORE_LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                    }
                    try
                    {
                        boolean found = rs.next();
                        if (!found)
                        {
                            SQLWarnings.log(rs);
                            return -1;
                        }
                        int index = rs.getInt(1);
                        SQLWarnings.log(rs);
                        return index;
                    }
                    finally
                    {
                        rs.close();
                    }
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                pm.releaseConnection(conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("SCO.IndexofRequestFailed",stmt),e);
        }
    }

    /**
     * Utility to find the indices of a collection of elements.
     * The returned list are not in any special order - they don't relate to
     * the order of the elements in the Collection. 
     * @param sm The state manager.
     * @param elements The elements
     * @return The indices of the elements in the List.
     **/
    private int[] getIndicesOf(StateManager sm, Collection elements)
    {
        if (elements == null || elements.size() == 0)
        {
            return null;
        }

        Iterator iter=elements.iterator();
        while (iter.hasNext())
        {
            Object element=iter.next();
            validateElementForReading(sm, element);
        }

        String stmt = getIndicesOfStmt(elements);
        PersistenceManager pm = sm.getPersistenceManager();
        int[] indices=new int[elements.size()];
        try
        {
            Connection conn = pm.getConnection(false);
            try
            {
                PreparedStatement ps = conn.prepareStatement(stmt);
                try
                {
                    prepareIndicesOfStmt(sm,ps,elements);

                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(stmt);
                    }
                    long startTime = System.currentTimeMillis();
                    ResultSet rs = ps.executeQuery();
                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(STORE_LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                    }

                    try
                    {
                        int i=0;
                        while (rs.next())
                        {
                            indices[i++] = rs.getInt(1);
                        }

                        if (i < elements.size())
                        {
                            throw new JDODataStoreException(LOCALISER.msg("SCO.GetIndicesFailed.InvalidElements",stmt));
                        }
                        SQLWarnings.log(rs);
                    }
                    finally
                    {
                        rs.close();
                    }
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                pm.releaseConnection(conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("SCO.IndicesofRequestFailed",stmt),e);
        }

        return indices;
    }
    
    /**
     * Internal method to perform the clearing of the List using the specified statement.
     * @param sm The state manager
     * @param stmt The statement
     **/
    protected void internalClear(StateManager sm, String stmt)
    {
        PersistenceManager pm = sm.getPersistenceManager();

        try
        {
            Connection conn = pm.getConnection(true);
            try
            {
                PreparedStatement ps = conn.prepareStatement(stmt);
                try
                {
                    ownerMapping.setObject(pm, ps, org.jpox.store.mapping.Mappings.getParametersIndex(1,ownerMapping), sm.getObject());

                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(stmt);
                    }
                    long startTime = System.currentTimeMillis();
                    ps.executeUpdate();
                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(STORE_LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                    }
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                pm.releaseConnection(conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("SCO.ClearRequestFailed",stmt),e);
        }
    }

    /**
     * Accessor for the GetRange statement. This has 5 modes
     * <UL>
     * <LI>start and end have the same value and &ge;0 then get that
     * element</LI>
     * <LI>start and end have values &ge;0 and are different then get a
     * range</LI>
     * <LI>start is &ge;0 and end is -1 the get all from start (inclusive).</LI>
     * <LI>start is -1 and end is &ge;0 the get all up to end(exclusive).</LI> 
     * <LI>start and end values are both -1 then get all elements</LI> 
     * </UL> 
     * @param start_index The start position in the List. 
     * @param end_index The end position in the List. 
     * @return The QueryStatement.
     **/
    protected QueryStatement getGetRangeStatement(StateManager sm,int start_index,int end_index)
    {
        return null;
    }

    /**
     * Utility to retrieve (as a List) the elements for the List in a range
     * depending on the specified start/end positions.
     * @param sm State Manager for the container
     * @param start Start position in the List.
     * @param end End position in the List.
     * @return The List of elements matching the range specification.
     **/
    protected java.util.List internalGetRange(StateManager sm,int start,int end)
    {
        java.util.List list=new java.util.ArrayList();

        PersistenceManager pm=sm.getPersistenceManager();
        QueryStatement     stmt=getGetRangeStatement(sm,start,end);
        Query.ResultObjectFactory get_rof=newResultObjectFactory(sm,stmt,false);
        Transaction tx = (Transaction) pm.currentTransaction();
        String statement = stmt.toStatementText(tx.useUpdateLockOnFetch()).toString();

        try
        {
            Connection conn = pm.getConnection(false);
            try
            {
                PreparedStatement ps = stmt.toStatementText(tx.useUpdateLockOnFetch()).prepareStatement(pm,conn);
                try
                {
                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(statement);
                    }
                    long startTime = System.currentTimeMillis();
                    ResultSet rs = ps.executeQuery();
                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(STORE_LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                    }

                    try
                    {
                        while (rs.next())
                        {
                            Object obj=null;
                            if (!isEmbeddedMapping(elementMapping))
                            {
                                // OID type - retrieve the object
                                obj = get_rof.getObject(pm, rs, pm.getClassLoaderResolver().classForName(elementType));
                            }
                            else
                            {
                                // Primitive type - read direct from mapping
                                int param[] = new int[elementMapping.getNumberOfDatastoreFields()];
                                for (int i = 0; i < param.length; ++i)
                                {
                                    param[i] = i + 1;
                                }
                                obj = elementMapping.getObject(pm, rs, param);
                            }

                            list.add(obj);
                        }

                        SQLWarnings.log(rs);
                    }
                    finally
                    {
                        rs.close();
                    }
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                pm.releaseConnection(conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("SCO.GetRequestFailed",statement),e);
        }

        return list;
    }

    /**
     * Internal method to check the element type to make sure that it is the
     * element type or a subclass and so can be stored here.
     * @param element The object
     * @return Whether it can be stored here.
     **/
    protected boolean validateElementType(ClassLoaderResolver clr, Object element)
    {
        return element == null || clr.isAssignableFrom(elementType,element.getClass());
    }

    protected boolean validateElementForReading(StateManager sm, Object element)
    {
        if (!validateElementType(sm.getPersistenceManager().getClassLoaderResolver(), element))
        {
            return false;
        }

        if (element != null && !elementsAreEmbedded)
        {
            PersistenceManager pm = sm.getPersistenceManager();

            if (!JDOHelper.isPersistent(element) ||
                pm != JDOHelper.getPersistenceManager(element))
            {
                return false;
            }
        }

        return true;
    }

    protected void validateElementForWriting(StateManager sm, Object element)
    {
        if (!validateElementType(sm.getPersistenceManager().getClassLoaderResolver(), element))
        {
            throw new ClassCastException(LOCALISER.msg("SCO.List.ElementIsInvalid",element.getClass().getName(),elementType));
        }

        if (element != null && !elementsAreEmbedded)
        {
            PersistenceManager pm = sm.getPersistenceManager();

            if (!JDOHelper.isPersistent(element))
            {
                boolean exists = false;
                if (JDOHelper.isDetached(element))
                {
                    try
                    {
                        // Check if the object exists in this datastore
                        pm.getObjectById(((PersistenceCapable)element).jdoGetObjectId(), true, false);
                        exists = true;
                    }
                    catch (JDOObjectNotFoundException onfe)
                    {
                        exists = false;
                    }
                }
                if (!exists)
                {
                    pm.makePersistent(element);
                }
            }
            else if (pm != JDOHelper.getPersistenceManager(element))
            {
                throw new JDOUserException(LOCALISER.msg("SCO.List.WriteInvalidWithDifferentPM"),JDOHelper.getObjectId(element));
            }
        }
    }

    // ---------------------------- Query Methods ------------------------------

    public QueryStatement getExistsSubquery(JavaTypeMapping mapping,
            								TableExpression ownerTe,
                                            SQLIdentifier listRangeVar)
    {
        QueryStatement stmt = ((DatabaseAdapter) dba).newQueryStatement(listTable, listRangeVar);

        ScalarExpression ownerExpr      = mapping.newScalarExpression(stmt, ownerTe);
        ScalarExpression ownerInSetExpr = ownerMapping.newScalarExpression(stmt, stmt.getTableExpression(listRangeVar));

        stmt.andCondition(ownerExpr.eq(ownerInSetExpr));

        stmt.select(listRangeVar, elementMapping);

        return stmt;
    }

    public Query.ResultObjectFactory newResultObjectFactory(StateManager sm, QueryStatement stmt, boolean ignoreCache)
    {
        int[] prefetchFieldNumbers;
        StatementExpressionIndex[] statementExpressionIndex;
        
        if (emd == null)
        {
            return new PersistentIDROF(this.elementTable, null, null, null, ignoreCache);           
        }
        FetchPlanImpl fp =((FetchPlanImpl)sm.getPersistenceManager().getFetchPlan());
        fp.manageFetchPlanForClass(emd);
        FetchPlanForClass fpc = fp.getFetchPlanForClass(emd);
        int fieldNumbers[] = fpc.getFieldsInActualFetchPlan();
        int fn[] = new int[fieldNumbers.length];
        int prefetchFieldCount = 0;
        int fieldCount = emd.getNoOfInheritedManagedFields() + emd.getNoOfManagedFields();
        
        statementExpressionIndex = new StatementExpressionIndex[fieldCount];
        
        for (int i=0; i<fieldNumbers.length; ++i)
        {
            JavaTypeMapping m = elementTable.getFieldMapping(emd.getManagedFieldAbsolute(fieldNumbers[i]));
            
            if( m != null ) // field is not stored in the elementTable, e.g. List, Set, etc or is transactional
            {
	            if (m.includeInFetchStatement() && !(m instanceof MappingCallbacks))
	            {
	                statementExpressionIndex[fieldNumbers[i]] = new StatementExpressionIndex();
	                statementExpressionIndex[fieldNumbers[i]].setMapping(m);
	                fn[prefetchFieldCount++] = fieldNumbers[i];
	            }
            }
        }
        
        prefetchFieldNumbers = new int[prefetchFieldCount];
        System.arraycopy(fn, 0, prefetchFieldNumbers, 0, prefetchFieldCount);                
        
        int[] datastoreIndex = null;

        // following the actual fetch plan, we select the fields to be fetched from the database
        if (stmt.getTableExpression(new TableIdentifier(storeMgr.getDatastoreAdapter(),"ELEMENT")) != null)
        {
            if( elementTable.getIdentityType() == IdentityType.DATASTORE )
            {
                datastoreIndex = stmt.select(new TableIdentifier(storeMgr.getDatastoreAdapter(), "ELEMENT"),elementTable.getDataStoreObjectIdMapping(),true);
            }
            Mappings.selectMapping(stmt, new TableIdentifier(storeMgr.getDatastoreAdapter(), "ELEMENT"), statementExpressionIndex);
        }
        else
        {
            if( elementTable.getIdentityType() == IdentityType.DATASTORE )
            {
                datastoreIndex = stmt.select(new TableIdentifier(storeMgr.getDatastoreAdapter(), "this"),elementTable.getDataStoreObjectIdMapping(),true);
            }
            Mappings.selectMapping(stmt,statementExpressionIndex);
        }        
        
        return new PersistentIDROF(
            this.getStoreManager().getDatastoreClass(getElementType()),
            prefetchFieldNumbers,
            statementExpressionIndex,
            datastoreIndex,
            ignoreCache);
    }

    /**
     * Check if the mapping correspond to a non pc object or embedded field
     * @param mapping the mapping
     * @return true if the field is embedded into one column
     */
    protected boolean isEmbeddedMapping(JavaTypeMapping mapping)
    {
        return !InterfaceMapping.class.isAssignableFrom(mapping.getClass()) &&
        	   !OIDMapping.class.isAssignableFrom(mapping.getClass());
    }
}
