/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others. All rights reserved. 
This program and the accompanying materials are made available under
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Erik Bengtson - removed unused import
2003 Andy Jefferson - coding standards
2003 Andy Jefferson - addition of getGetStatement for inherited values
2004 Andy Jefferson - addition of query methods
2004 Andy Jefferson - moved statements from subclasses
    ...
**********************************************************************/
package org.jpox.store.rdbms.scostore;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Iterator;
import java.util.Map;
import java.util.NoSuchElementException;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOHelper;
import javax.jdo.JDOObjectNotFoundException;
import javax.jdo.JDOUserException;
import javax.jdo.spi.PersistenceCapable;

import org.jpox.ClassLoaderResolver;
import org.jpox.FetchPlanImpl;
import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.Transaction;
import org.jpox.FetchPlanImpl.FetchPlanForClass;
import org.jpox.metadata.ClassMetaData;
import org.jpox.metadata.IdentityType;
import org.jpox.store.DatastoreAdapter;
import org.jpox.store.StatementExpressionIndex;
import org.jpox.store.StoreManager;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.TableExpression;
import org.jpox.store.mapping.InterfaceMapping;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.MappingCallbacks;
import org.jpox.store.mapping.Mappings;
import org.jpox.store.mapping.OIDMapping;
import org.jpox.store.query.PersistentIDROF;
import org.jpox.store.query.Query;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.adapter.DatabaseAdapter;
import org.jpox.store.rdbms.mapping.RDBMSMapping;
import org.jpox.store.rdbms.sqlidentifier.SQLIdentifier;
import org.jpox.store.rdbms.sqlidentifier.TableIdentifier;
import org.jpox.store.rdbms.table.ClassTable;
import org.jpox.store.rdbms.table.Table;
import org.jpox.store.scostore.MapStore;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;
import org.jpox.util.SQLWarnings;

/**
 * Abstract representation of the backing store for a Map.
 *
 * @version $Revision: 1.8 $
 **/
abstract class AbstractMapStore implements MapStore
{
    protected static final Localiser LOCALISER=Localiser.getInstance("org.jpox.store.rdbms.scostore.Localisation");
    protected static final Localiser STORE_LOCALISER=Localiser.getInstance("org.jpox.store.rdbms.Localisation");

    protected StoreManager storeMgr;
    protected DatastoreAdapter dba;

    protected Table mapTable;
    protected ClassTable valueTable;

    protected JavaTypeMapping ownerMapping;
    protected JavaTypeMapping keyMapping;
    protected JavaTypeMapping valueMapping;
    protected String  keyType;
    protected String  valueType;
    protected boolean keysAreEmbedded;
    protected boolean valuesAreEmbedded;

    protected String  containsValueStmt;
    protected ClassMetaData vmd;

    private Query.ResultObjectFactory rof = null;

    /**
     * Default Constructor
     **/
    public AbstractMapStore()
    {
    }

    /**
     * Method to initialise the statements being used.
     * Subclasses should override the getXXXStmt() if they want to use an
     * alternative statement.
     **/
    protected void initialiseStatements()
    {
        // Our statements use mapTable as the place to select/insert/delete
        containsValueStmt = getContainsValueStmt();
    }

    /**
     * Accessor for the RDBMSManager.
     * @return The RDBMSManager
     **/
    public StoreManager getStoreManager()
    {
        return storeMgr;
    }

    /**
     * Accessor for the key type for storing in this Map.
     * @return The type of the key
     **/
    public String getKeyType()
    {
        return keyType;
    }

    /**
     * Accessor for the value type for storing in this Map.
     * Types inherited from this (base) type can also be stored as values. 
     * @return The type of the value (base class).
     **/
    public String getValueType()
    {
        return valueType;
    }

    // ------------------------- Public Methods --------------------------------
 
    /**
     * Method to check if a key exists in the Map.
     * @param sm State Manager for the map
     * @param key The key to check for.
     * @return Whether the key exists in the Map.
     **/
    public boolean containsKey(StateManager sm, Object key)
    {
        try
        {
            getValue(sm, key);
            return true;
        }
        catch (NoSuchElementException e)
        {
            return false;
        }
    }

    /**
     * Method to check if a value exists in the Map.
     * @param sm State Manager for the map
     * @param value The value to check for.
     * @return Whether the value exists in the Map.
     **/
    public boolean containsValue(StateManager sm, Object value)
    {
        if (!validateValueForReading(sm, value))
        {
            return false;
        }

        boolean exists = false;
        PersistenceManager pm = sm.getPersistenceManager();

        try
        {
            Connection conn = pm.getConnection(false);
            try
            {
                PreparedStatement ps = conn.prepareStatement(containsValueStmt);
                try
                {
					ownerMapping.setObject(pm, ps, Mappings.getParametersIndex(1,ownerMapping), sm.getObject());
					valueMapping.setObject(pm, ps, Mappings.getParametersIndex(ownerMapping.getNumberOfDatastoreFields()+1,valueMapping), value);

                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(containsValueStmt);
                    }
                    long startTime = System.currentTimeMillis();
                    ResultSet rs = ps.executeQuery();
                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(STORE_LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                    }

                    try
                    {
                        if (rs.next())
                        {
                            exists = true;
                        }

                        SQLWarnings.log(rs);
                    }
                    finally
                    {
                        rs.close();
                    }
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                pm.releaseConnection(conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("SCO.ContainsValueRequestFailed",containsValueStmt),e);
        }

        return exists;
    }

    /**
     * Method to return the value for a key.
     * @param sm State Manager for the Map.
     * @param key The key of the object to retrieve.
     * @return The value for this key.
     **/
    public Object get(StateManager sm, Object key)
    {
        try
        {
            return getValue(sm, key);
        }
        catch (NoSuchElementException e)
        {
            return null;
        }
    }

    /**
     * Method to put all elements from a Map into our Map.
     * @param sm State Manager for the Map
     * @param m The Map to add
     **/
    public void putAll(StateManager sm, Map m)
    {
        Iterator i = m.entrySet().iterator();

        while (i.hasNext())
        {
            Map.Entry e = (Map.Entry)i.next();
            put(sm, e.getKey(), e.getValue());
        }
    }

    // --------------------------- Utility Methods -----------------------------
 
    /**
     * Generate statement to check if a value is contained in the Map.
     * <PRE>
     * SELECT OWNERCOL
     * FROM MAPTABLE
     * WHERE OWNERCOL=? AND VALUECOL = ?
     * </PRE>
     * @return Statement to check if a value is contained in the Map. 
     */
    private String getContainsValueStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("SELECT ");
        stmt.append(((Column)ownerMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
        for (int i=1; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(",");
            stmt.append(((Column)ownerMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
        }
        stmt.append(" FROM ");
        stmt.append(mapTable.toString());
        stmt.append(" WHERE ");
        stmt.append(((Column)ownerMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
        stmt.append(" = ");
        stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(0)).getUpdateInputParameter());
        for (int i=1; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(((Column)ownerMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        for (int i=0; i<valueMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(((Column)valueMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)valueMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        return stmt.toString();
    }

    /**
     * Utility to validate the type of a key for storing in the Map.
     * @param key The key to check.
     **/
    protected void validateKeyType(ClassLoaderResolver clr, Object key)
    {
        if (key == null)
        {
            throw new NullPointerException(LOCALISER.msg("SCO.Map.NullKeysNotAllowed"));
        }

        if (!clr.isAssignableFrom(keyType,key.getClass()))
        {
            throw new ClassCastException(LOCALISER.msg("SCO.Map.KeyIsInvalid",key.getClass().getName(),keyType));
        }
    }

    /**
     * Utility to validate the type of a value for storing in the Map.
     * @param value The value to check.
     **/
    protected void validateValueType(ClassLoaderResolver clr, Object value)
    {
        if (value != null && !clr.isAssignableFrom(valueType,value.getClass()))
        {
            throw new ClassCastException(LOCALISER.msg("SCO.Map.ValueIsInvalid",value.getClass().getName(),valueType));
        }
    }

    /**
     * Utility to validate a key is ok for reading.
     * @param sm State Manager for the map.
     * @param key The key to check.
     * @return Whether it is validated. 
     **/
    protected boolean validateKeyForReading(StateManager sm, Object key)
    {
        validateKeyType(sm.getPersistenceManager().getClassLoaderResolver(), key);

        if (!keysAreEmbedded)
        {
            PersistenceManager pm = sm.getPersistenceManager();
            if (!JDOHelper.isPersistent(key) ||
                pm != JDOHelper.getPersistenceManager(key))
            {
                return false;
            }
        }

        return true;
    }

    /**
     * Utility to validate a value is ok for reading.
     * @param sm State Manager for the map.
     * @param value The value to check.
     * @return Whether it is validated. 
     **/
    protected boolean validateValueForReading(StateManager sm, Object value)
    {
        validateValueType(sm.getPersistenceManager().getClassLoaderResolver(), value);

        if (!valuesAreEmbedded)
        {
            PersistenceManager pm = sm.getPersistenceManager();

            if (value!=null && (!JDOHelper.isPersistent(value) ||
                pm != JDOHelper.getPersistenceManager(value)))
            {
                return false;
            }
        }

        return true;
    }

    /**
     * Utility to validate a key is ok for writing.
     * @param sm State Manager for the map.
     * @param key The key to check.
     **/
    protected void validateKeyForWriting(StateManager sm, Object key)
    {
        validateKeyType(sm.getPersistenceManager().getClassLoaderResolver(), key);

        if (!keysAreEmbedded)
        {
            PersistenceManager pm = sm.getPersistenceManager();

            if (!JDOHelper.isPersistent(key))
            {
                boolean exists = false;
                if (JDOHelper.isDetached(key))
                {
                    try
                    {
                        // Check if the object exists in this datastore
                        pm.getObjectById(((PersistenceCapable)key).jdoGetObjectId(), true, false);
                        exists = true;
                    }
                    catch (JDOObjectNotFoundException onfe)
                    {
                        exists = false;
                    }
                }
                if (!exists)
                {
                    pm.makePersistent(key);
                }
            }
            else if (pm != JDOHelper.getPersistenceManager(key))
            {
                throw new JDOUserException(LOCALISER.msg("SCO.Map.WriteKeyInvalidWithDifferentPM"),JDOHelper.getObjectId(key));
            }
        }
    }

    /**
     * Utility to validate a value is ok for writing.
     * @param sm State Manager for the map.
     * @param value The value to check.
     **/
    protected void validateValueForWriting(StateManager sm, Object value)
    {
        validateValueType(sm.getPersistenceManager().getClassLoaderResolver(), value);

        if (!valuesAreEmbedded)
        {
            PersistenceManager pm = sm.getPersistenceManager();

            if (!JDOHelper.isPersistent(value))
            {
                boolean exists = false;
                if (JDOHelper.isDetached(value))
                {
                    try
                    {
                        // Check if the object exists in this datastore
                        pm.getObjectById(((PersistenceCapable)value).jdoGetObjectId(), true, false);
                        exists = true;
                    }
                    catch (JDOObjectNotFoundException onfe)
                    {
                        exists = false;
                    }
                }
                if (!exists)
                {
                    pm.makePersistent(value);
                }
            }
            else if (pm != JDOHelper.getPersistenceManager(value))
            {
                throw new JDOUserException(LOCALISER.msg("SCO.Map.WriteValueInvalidWithDifferentPM"),JDOHelper.getObjectId(value));
            }
        }
    }

    /**
     * Accessor for the Get statement.
     * @return The QueryStatement.
     **/
    protected  QueryStatement getGetStatement(StateManager sm, Object key)
    {
        return null;
    }

    /**
     * Method to retrieve a value from the Map given the key.
     * @param sm State Manager for the map.
     * @param key The key to retrieve the value for.
     * @return The value for this key
     **/
    protected Object getValue(StateManager sm, Object key)
    throws NoSuchElementException
    {
        if (!validateKeyForReading(sm, key))
        {
            return null;
        }

        PersistenceManager pm = sm.getPersistenceManager();

        QueryStatement stmt=getGetStatement(sm,key);
        rof = newResultObjectFactory(sm,stmt,false);

        Transaction tx = (Transaction) pm.currentTransaction();
        String statement = stmt.toStatementText(tx.useUpdateLockOnFetch()).toString();
 
        Object value=null;
        try
        {
            Connection conn = pm.getConnection(false);
            try
            {
                PreparedStatement ps = stmt.toStatementText(tx.useUpdateLockOnFetch()).prepareStatement(pm,conn);
                try
                {
                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(statement);
                    }
                    long startTime = System.currentTimeMillis();
                    ResultSet rs = ps.executeQuery();
                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(STORE_LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                    }

                    try
                    {
                        boolean found=rs.next();
                        if (!found)
                        {
                            throw new NoSuchElementException();
                        }

                        if (!isEmbeddedMapping(valueMapping))
                        {
                            // OID type - retrieve the object
                            value = rof.getObject(pm, rs, pm.getClassLoaderResolver().classForName(valueType));
                        }
                        else
                        {
                            // Primitive type - read direct from mapping
                            int param[] = new int[valueMapping.getNumberOfDatastoreFields()];
                            for (int i = 0; i < param.length; ++i)
                            {
                                param[i] = i + 1;
                            }
                            value = valueMapping.getObject(pm,rs,param);
                        }

                        SQLWarnings.log(rs);
                    }
                    finally
                    {
                        rs.close();
                    }
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                pm.releaseConnection(conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("SCO.GetRequestFailed",statement),e);
        }

        return value;
    }

    public Query.ResultObjectFactory newResultObjectFactory(StateManager sm, QueryStatement stmt, boolean ignoreCache)
    {
        int[] prefetchFieldNumbers;
        StatementExpressionIndex[] statementExpressionIndex;
        
        if (vmd == null)
        {
            return new PersistentIDROF(this.valueTable, null, null, null, ignoreCache);           
        }
        FetchPlanImpl fp =((FetchPlanImpl)sm.getPersistenceManager().getFetchPlan());
        fp.manageFetchPlanForClass(vmd);
        FetchPlanForClass fpc = fp.getFetchPlanForClass(vmd);
        int fieldNumbers[] = fpc.getFieldsInActualFetchPlan();
        int fn[] = new int[fieldNumbers.length];
        int prefetchFieldCount = 0;
        int fieldCount = vmd.getNoOfInheritedManagedFields() + vmd.getNoOfManagedFields();
        
        statementExpressionIndex = new StatementExpressionIndex[fieldCount];
        
        for (int i=0; i<fieldNumbers.length; ++i)
        {
            JavaTypeMapping m = valueTable.getFieldMapping(vmd.getManagedFieldAbsolute(fieldNumbers[i]));
            
            if( m != null ) // field is not stored in the elementTable, e.g. List, Set, etc or is transactional
            {
	            if (m.includeInFetchStatement() && !(m instanceof MappingCallbacks))
	            {
	                statementExpressionIndex[fieldNumbers[i]] = new StatementExpressionIndex();
	                statementExpressionIndex[fieldNumbers[i]].setMapping(m);
	                fn[prefetchFieldCount++] = fieldNumbers[i];
	            }
            }
        }
        
        prefetchFieldNumbers = new int[prefetchFieldCount];
        System.arraycopy(fn, 0, prefetchFieldNumbers, 0, prefetchFieldCount);                
        
        int[] datastoreIndex = null;
        
        // following the actual fetch plan, we select the fields to be fetched from the database
        if (stmt.getTableExpression(new TableIdentifier(storeMgr.getDatastoreAdapter(),"ELEMENT")) != null)
        {
            if( valueTable.getIdentityType() == IdentityType.DATASTORE )
            {
                datastoreIndex = stmt.select(new TableIdentifier(storeMgr.getDatastoreAdapter(), "ELEMENT"),valueTable.getDataStoreObjectIdMapping(),true);
            }            
            Mappings.selectMapping(stmt, new TableIdentifier(storeMgr.getDatastoreAdapter(), "ELEMENT"), statementExpressionIndex);
        }
        else
        {
            if( valueTable.getIdentityType() == IdentityType.DATASTORE )
            {
                datastoreIndex = stmt.select(new TableIdentifier(storeMgr.getDatastoreAdapter(), "this"),valueTable.getDataStoreObjectIdMapping(),true);
            }
            
            Mappings.selectMapping(stmt,statementExpressionIndex);
        }        
        
        return new PersistentIDROF(
            this.getStoreManager().getDatastoreClass(getValueType()),
            prefetchFieldNumbers,
            statementExpressionIndex,
            datastoreIndex,
            ignoreCache);
    }    

    /**
     * Method used where a Query uses map.isEmpty().
     * @param mapping The mapping of the java type
     * @param ownerTe The owner table expression
     * @param mapRangeVar The range variable for the "Map" table.
     * @return A subquery
     **/
    public QueryStatement getExistsSubquery(JavaTypeMapping mapping,
                            				TableExpression ownerTe,
                            				SQLIdentifier mapRangeVar)
    {
        QueryStatement stmt = ((DatabaseAdapter) dba).newQueryStatement(mapTable, mapRangeVar);

        ScalarExpression ownerExpr      = mapping.newScalarExpression(stmt, ownerTe);
        ScalarExpression ownerInMapExpr = ownerMapping.newScalarExpression(stmt, stmt.getTableExpression(mapRangeVar));

        stmt.andCondition(ownerExpr.eq(ownerInMapExpr));

        stmt.select(mapRangeVar, valueMapping);

        return stmt;
    }
    
    /**
     * Used as part of the Querying of Maps where a get(Key) is used.
     * @param stmt The Query Statement to apply the join
     * @param ownerMapping Mapping for the owner. 
     * @param te Table Expression for the owner
     * @param mapRangeVar  The SQL alias, or "range variable", to assign to the
     *                     expression or to the main table.
     * @param filteredKeyType The Class Type for the filtered key
     * @param keyRangeVar  The SQL alias, or "range variable", to assign to the
     *                     expression or to the key table.
     * @param valueRangeVar The SQL alias, or "range variable", to assign to the
     *                      expression or to the value table.
     * @return an array with 2 elements of QueryColumnList. The first element
     * contains the columns from the key mapping and the second element the
     * columns from the value mapping
     */
    public ScalarExpression[] joinKeysToGet(
                                QueryStatement stmt,
                                JavaTypeMapping ownerMapping,
                                TableExpression te,
                                SQLIdentifier mapRangeVar,
                                Class filteredKeyType,
                                SQLIdentifier keyRangeVar,
    							SQLIdentifier valueRangeVar)
    {
        ClassLoaderResolver clr=storeMgr.getClassLoaderResolver();
        Class value_class=clr.classForName(valueType);
        return joinKeysValuesTo(stmt,ownerMapping,te,mapRangeVar,filteredKeyType,value_class,keyRangeVar,valueRangeVar);
    }
    
    /**
     * Check if the mapping correspond to a non pc object or embedded field
     * @param mapping the mapping
     * @return true if the field is embedded into one column
     */
    protected boolean isEmbeddedMapping(JavaTypeMapping mapping)
    {
        return !InterfaceMapping.class.isAssignableFrom(mapping.getClass()) &&
        	   !OIDMapping.class.isAssignableFrom(mapping.getClass());
    }    
}
