/**********************************************************************
Copyright (c) 2002 Kelly Grizzle (TJDO) and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
2002 Mike Martin (TJDO)
2003 Andy Jefferson - coding standards
2003 Andy Jefferson - changed to use Logger
2004 Andy Jefferson - moved statements from subclasses to this class.
    ...
**********************************************************************/
package org.jpox.store.rdbms.scostore;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOHelper;
import javax.jdo.JDOObjectNotFoundException;
import javax.jdo.JDOUserException;
import javax.jdo.spi.PersistenceCapable;

import org.jpox.ClassLoaderResolver;
import org.jpox.FetchPlanImpl;
import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.Transaction;
import org.jpox.FetchPlanImpl.FetchPlanForClass;
import org.jpox.metadata.ClassMetaData;
import org.jpox.metadata.DiscriminatorStrategy;
import org.jpox.metadata.IdentityType;
import org.jpox.store.DatastoreAdapter;
import org.jpox.store.StatementExpressionIndex;
import org.jpox.store.StoreManager;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.TableExpression;
import org.jpox.store.mapping.InterfaceMapping;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.MappingCallbacks;
import org.jpox.store.mapping.Mappings;
import org.jpox.store.mapping.OIDMapping;
import org.jpox.store.query.PersistentIDROF;
import org.jpox.store.query.Query;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.adapter.DatabaseAdapter;
import org.jpox.store.rdbms.mapping.RDBMSMapping;
import org.jpox.store.rdbms.sqlidentifier.SQLIdentifier;
import org.jpox.store.rdbms.sqlidentifier.TableIdentifier;
import org.jpox.store.rdbms.table.ClassTable;
import org.jpox.store.rdbms.table.Table;
import org.jpox.store.scostore.SetStore;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;
import org.jpox.util.SQLWarnings;

/**
 * Abstract representation of the backing store for a Set.
 *
 * @version $Revision: 1.9 $
 **/
abstract class AbstractSetStore implements SetStore
{
    protected static final Localiser STORE_LOCALISER=Localiser.getInstance("org.jpox.store.rdbms.Localisation");
    protected static final Localiser LOCALISER=Localiser.getInstance("org.jpox.store.rdbms.scostore.Localisation");

    protected Table setTable;
    protected String setName;
    protected StoreManager storeMgr;
    protected DatastoreAdapter dba;
    protected JavaTypeMapping ownerMapping;
    protected JavaTypeMapping elementMapping;
    protected JavaTypeMapping elementDiscriminatorMapping;
    protected String elementType;
    protected boolean elementsAreEmbedded;
    protected String sizeStmt;
    protected String containsStmt;
    protected String addStmt;
    protected String removeStmt;
    protected String clearStmt;
    protected ClassTable elementTable;
    protected ClassMetaData emd;
    
    private Query.ResultObjectFactory rof = null;

    protected AbstractSetStore()
    {
    }

    /**
     * Method to initialise the statements being used.
     * Subclasses should override the getXXXStmt() if they want to use an
     * alternative statement.
     **/
    protected void initialiseStatements()
    {
        // Our statements use setTable as the place to select/insert/delete
        containsStmt = getContainsStmt();
        sizeStmt     = getSizeStmt();
        clearStmt    = getClearStmt();
        removeStmt   = getRemoveStmt();
    }

    /**
     * Accessor for the Store Manager.
     * @return Store Manager.
     **/
    public StoreManager getStoreManager()
    {
        return storeMgr;
    }

    /**
     * Accessor for the element type.
     * @return Element type.
     **/
    public String getElementType()
    {
        return elementType;
    }

    /**
     * Accessor for the owner mapping.
     * @return Owner mapping.
     **/
    public JavaTypeMapping getOwnerMapping()
    {
        return ownerMapping;
    }

    /**
     * Method to validate an element against the accepted type.
     * @param element The element to validate
     * @return Whether it is valid.
     **/ 
    protected boolean validateElementType(ClassLoaderResolver clr, Object element)
    {
        return element == null || clr.isAssignableFrom(elementType,element.getClass());
    }


    /**
     * Method to check if an element is already persistent or is persistent but managed by 
     * a different persistence manager.
     * @param sm The state manager of this collection
     * @param element The element
     * @return Whether it is valid for reading.
     */
    protected boolean validateElementForReading(StateManager sm, Object element)
    {
        if (!validateElementType(sm.getPersistenceManager().getClassLoaderResolver(), element))
        {
            return false;
        }

        if (element != null && !elementsAreEmbedded)
        {
            PersistenceManager pm = sm.getPersistenceManager();
            if (!JDOHelper.isPersistent(element) || pm != JDOHelper.getPersistenceManager(element))
            {
                return false;
            }
        }

        return true;
    }

    /**
     * Method to check if an element is already persistent, or is managed by a different
     * Persistencemanager. If not persistent, this will persist it.
     * @param sm The state manager of this collection
     * @param element The element
     */
    protected void validateElementForWriting(StateManager sm, Object element)
    {
        if (!validateElementType(sm.getPersistenceManager().getClassLoaderResolver(), element))
        {
            throw new ClassCastException(LOCALISER.msg("SCO.Set.ElementIsInvalid",element.getClass().getName(),elementType));
        }

        if (element != null && !elementsAreEmbedded)
        {
            PersistenceManager pm = sm.getPersistenceManager();

            if (!JDOHelper.isPersistent(element))
            {
                boolean exists = false;
                if (JDOHelper.isDetached(element))
                {
                    try
                    {
                        // Check if the object exists in this datastore
                        pm.getObjectById(((PersistenceCapable)element).jdoGetObjectId(), true, false);
                        exists = true;
                    }
                    catch (JDOObjectNotFoundException onfe)
                    {
                        exists = false;
                    }
                }
                if (!exists)
                {
                    pm.makePersistent(element);
                }
            }
            else if (pm != JDOHelper.getPersistenceManager(element))
            {
                throw new JDOUserException(LOCALISER.msg("SCO.Set.WriteInvalidWithDifferentPM"),JDOHelper.getObjectId(element));
            }
        }
    }

    /**
     * Accessor for the statement for the iterator.
     * @return The iterator Query Statement.
     **/
    protected QueryStatement getIteratorStatement(StateManager sm)
    {
        return null;
    }
 
    /**
     * Accessor for the iterator of the Set.
     * @param sm State Manager for the container.
     **/
    public Iterator iterator(StateManager sm)
    {
        Iterator it = null;
        PersistenceManager pm = sm.getPersistenceManager();
        QueryStatement stmt = getIteratorStatement(sm);
        if (stmt == null) 
        {
            throw new JDOFatalInternalException(LOCALISER.msg("SCO.IteratorStatementIsNull"));
        }
        
        rof = newResultObjectFactory(sm,stmt,false);
        
        Transaction tx = (Transaction) pm.currentTransaction();
        String statement = stmt.toStatementText(tx.useUpdateLockOnFetch()).toString();

        // this happens when collection is null
        if (statement == null)
        {   
            throw new JDOFatalInternalException(LOCALISER.msg("SCO.IteratorStatementIsNull"));
        }

        try
        {
            Connection conn = pm.getConnection(false);
            try
            {
                PreparedStatement ps = stmt.toStatementText(tx.useUpdateLockOnFetch()).prepareStatement(pm,conn);
                try
                {
                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(statement);
                    }
                    long startTime = System.currentTimeMillis();
                    ResultSet rs = ps.executeQuery();
                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(STORE_LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                    }

                    try
                    {
                        it = new SetIterator(sm, rs);
                    }
                    finally
                    {
                        rs.close();
                    }
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                pm.releaseConnection(conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("SCO.IteratorRequestFailed",statement),e);
        }
        return it;
    }

    /**
     * Inner class represting an iterator of the Set.
     **/
    private class SetIterator implements Iterator
    {
        private final StateManager sm;
        private final PersistenceManager pm;
        private final Iterator delegate;
        private Object lastElement = null;

        public SetIterator(StateManager sm, ResultSet rs) throws SQLException
        {
            this.sm = sm;
            this.pm = sm.getPersistenceManager();

            ArrayList results = new ArrayList();
            if (rs != null)
            {
                while(rs.next())
                {
                    Object nextElement;
                    if (!isEmbeddedMapping(elementMapping)) 
                    {
                        nextElement = rof.getObject(pm, rs, pm.getClassLoaderResolver().classForName(elementType));
                    }
                    else
                    {
                        // primitive type - read the object directly from the
                        // mapping. Should be same number of columns in the
                        // rs as in the mapping
                        int param[] = new int[elementMapping.getNumberOfDatastoreFields()];
                        for (int i = 0; i < param.length; ++i)
                        {
                            param[i] = i + 1;
                        }
                        nextElement = elementMapping.getObject(pm,rs,param);
                    }
 
                    results.add(nextElement);
                }
            }
            delegate = results.iterator();
        }

        public boolean hasNext()
        {
            return delegate.hasNext();
        }

        public Object next()
        {
            lastElement = delegate.next();

            return lastElement;
        }

        public synchronized void remove()
        {
            if (lastElement == null)
            {
                throw new IllegalStateException("No entry to remove");
            }

            AbstractSetStore.this.remove(sm, lastElement);
            delegate.remove();

            lastElement = null;
        }
    }

    /**
     * Accessor for the size of the container.
     * @return The size of the Set. 
     */
    public int size(StateManager sm)
    {
        int numRows;
        PersistenceManager pm = sm.getPersistenceManager();

        try
        {
            Connection conn = pm.getConnection(false);
            try
            {
                PreparedStatement ps = conn.prepareStatement(sizeStmt);
                try
                {
                    int jdbc_position = 1;
                    ownerMapping.setObject(pm, ps, Mappings.getParametersIndex(jdbc_position,ownerMapping), sm.getObject());
                    jdbc_position += ownerMapping.getNumberOfDatastoreFields();
                    if (elementDiscriminatorMapping != null)
                    {
                        if (emd.getDiscriminatorStrategy() == DiscriminatorStrategy.CLASS_NAME)
                        {
                            elementDiscriminatorMapping.setObject(pm, ps, Mappings.getParametersIndex(jdbc_position, elementDiscriminatorMapping), elementType);
                        }
                        else if (emd.getDiscriminatorStrategy() == DiscriminatorStrategy.VALUE_MAP)
                        {
                            elementDiscriminatorMapping.setObject(pm, ps, Mappings.getParametersIndex(jdbc_position, elementDiscriminatorMapping),
                                emd.getInheritanceMetaData().getDiscriminatorMetaData().getValue());
                        }
                    }

                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(sizeStmt);
                    }
                    long startTime = System.currentTimeMillis();
                    ResultSet rs = ps.executeQuery();
                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(STORE_LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                    }

                    try
                    {
                        if (!rs.next())
                        {
                            throw new JDODataStoreException(LOCALISER.msg("SCO.SizeRequestFailed",sizeStmt));
                        }

                        numRows = rs.getInt(1);

                        SQLWarnings.log(rs);
                    }
                    finally
                    {
                        rs.close();
                    }
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                pm.releaseConnection(conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("SCO.SizeRequestFailed",sizeStmt),e);
        }

        return numRows;
    }

    /**
     * Method to verify if the association owner vs elements contains
     * an specific element in the association 
     * @param sm The StateManager
     * @param element The element
     * @return Whether it contains the element 
     */
    public boolean contains(StateManager sm, Object element)
    {
        if (!validateElementForReading(sm, element))
        {
            return false;
        }

        boolean retval;
        PersistenceManager pm = sm.getPersistenceManager();

        try
        {
            Connection conn = pm.getConnection(false);
            try
            {
                PreparedStatement ps = conn.prepareStatement(containsStmt);
                try
                {
                    int jdbc_position = 1;
                    ownerMapping.setObject(pm, ps, Mappings.getParametersIndex(jdbc_position,ownerMapping), sm.getObject());
                    jdbc_position += ownerMapping.getNumberOfDatastoreFields();
                    elementMapping.setObject(pm, ps, Mappings.getParametersIndex(jdbc_position,elementMapping), element);
                    jdbc_position += elementMapping.getNumberOfDatastoreFields();
                    if (elementDiscriminatorMapping != null)
                    {
                        if (emd.getDiscriminatorStrategy() == DiscriminatorStrategy.CLASS_NAME)
                        {
                            elementDiscriminatorMapping.setObject(pm, ps, Mappings.getParametersIndex(jdbc_position, elementDiscriminatorMapping), elementType);
                        }
                        else if (emd.getDiscriminatorStrategy() == DiscriminatorStrategy.VALUE_MAP)
                        {
                            elementDiscriminatorMapping.setObject(pm, ps, Mappings.getParametersIndex(jdbc_position, elementDiscriminatorMapping),
                                emd.getInheritanceMetaData().getDiscriminatorMetaData().getValue());
                        }
                    }

                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(containsStmt);
                    }
                    long startTime = System.currentTimeMillis();
                    ResultSet rs = ps.executeQuery();
                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(STORE_LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                    }

                    try
                    {
                        retval = rs.next();

                        SQLWarnings.log(rs);
                    }
                    finally
                    {
                        rs.close();
                    }
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                pm.releaseConnection(conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("SCO.ContainsRequestFailed",containsStmt),e);
        }

        return retval;
    }

    /**
     * Adds one element to the association owner vs elements
     * @param sm State Manager for the container
     * @param element The element to add
     * @return Whether it was successful 
     */
    public boolean add(StateManager sm, Object element)
    {
        validateElementForWriting(sm, element);

        boolean modified = false;
        PersistenceManager pm = sm.getPersistenceManager();

        try
        {
            Connection conn = pm.getConnection(true);

            try
            {
                PreparedStatement ps = conn.prepareStatement(addStmt);
                try
                {
                    ownerMapping.setObject(pm, ps, Mappings.getParametersIndex(1,ownerMapping), sm.getObject());
                    elementMapping.setObject(pm, ps, Mappings.getParametersIndex(ownerMapping.getNumberOfDatastoreFields()+1,elementMapping), element);

                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(addStmt);
                    }
                    long startTime = System.currentTimeMillis();
                    ps.executeUpdate();
                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(STORE_LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                    }

                    modified = true;
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                pm.releaseConnection(conn);
            }
        }
        catch (SQLException e)
        {
            if (!contains(sm, element))
            {
                throw new JDODataStoreException(LOCALISER.msg("SCO.AddRequestFailed",addStmt),e);
            }
        }

        return modified;
    }
    
    /**
     * Adds all elements from a collection to the association owner vs elements
     * @param sm State Manager for the container
     * @param elements The elements to add
     * @return Whether it was successful 
     **/
    public boolean addAll(StateManager sm, Collection elements)
    {
        boolean modified = false;
        Iterator i = elements.iterator();

        if (i.hasNext())
        {
            PersistenceManager pm = sm.getPersistenceManager();
            Object element = null;
            try
            {
                Connection conn = pm.getConnection(true);
                PreparedStatement ps = conn.prepareStatement(addStmt);
                try
                {
                    while (i.hasNext())
                    {
                        element = i.next();
                        validateElementForWriting(sm, element);
 
                        ownerMapping.setObject(pm, ps, Mappings.getParametersIndex(1,ownerMapping), sm.getObject());
                        elementMapping.setObject(pm, ps, Mappings.getParametersIndex(ownerMapping.getNumberOfDatastoreFields()+1,elementMapping), element);

                        if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                        {
                            JPOXLogger.RDBMS_SQL.debug(addStmt);
                        }
                        long startTime = System.currentTimeMillis();
                        ps.executeUpdate();
                        if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                        {
                            JPOXLogger.RDBMS_SQL.debug(STORE_LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                        }

                        modified = true;
                    }
                }
                finally
                {
                    ps.close();
                    pm.releaseConnection(conn);
                }
            }
            catch (SQLException e)
            {
                throw new JDODataStoreException(LOCALISER.msg("SCO.AddRequestFailed",addStmt),e);
            }
        }

        return modified;
    }

    /**
     * Removes the association to one element
     * @param sm State Manager for the container
     * @param element Element to remove
     * @return Whether it was successful 
     */
    public boolean remove(StateManager sm, Object element)
    {
        if (!validateElementForReading(sm, element))
        {
            JPOXLogger.RDBMS.debug("AbstractSetStore::remove element=" + element + " doesn't exist in this Set.");
            return false;
        }

        boolean modified = false;
        PersistenceManager pm = sm.getPersistenceManager();

        try
        {
            Connection conn = pm.getConnection(true);
            try
            {
                PreparedStatement ps = conn.prepareStatement(removeStmt);
                try
                {
                    int jdbc_position = 1;
                    ownerMapping.setObject(pm, ps, Mappings.getParametersIndex(jdbc_position,ownerMapping), sm.getObject());
                    jdbc_position += ownerMapping.getNumberOfDatastoreFields();
                    elementMapping.setObject(pm, ps, Mappings.getParametersIndex(jdbc_position,elementMapping), element);

                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(removeStmt);
                    }
                    long startTime = System.currentTimeMillis();
                    int rowsDeleted = ps.executeUpdate();
                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(STORE_LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                    }

                    modified = rowsDeleted == 1;
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                pm.releaseConnection(conn);
            }
        }
        catch (SQLException e)
        {
            if (contains(sm, element))
            {
                throw new JDODataStoreException(LOCALISER.msg("SCO.RemoveRequestFailed",removeStmt),e);
            }
        }

        return modified;
    }

    /**
     * Remove all elements from a collection from the association owner vs
     * elements. This implementation iterates around the remove() method doing
     * each element 1 at a time. Please refer to the NormalSetStore and
     * InverseSetStore for the variations used there.
     * @param sm State Manager for the container
     * @param elements Collection of elements to remove 
     * @return Whether the database was updated
     */
    public boolean removeAll(StateManager sm, Collection elements)
    {
        if (elements == null || elements.size() == 0)
        {
            return false;
        }

        boolean modified = false;
        try
        {
            PersistenceManager pm = sm.getPersistenceManager();
            Connection conn = pm.getConnection(true);
            PreparedStatement ps = conn.prepareStatement(removeStmt);
            try
            {
                Object element = null;
                Iterator iter=elements.iterator();
                while (iter.hasNext())
                {
                    element = iter.next();
                    validateElementForReading(sm, element);
 
                    // Populate statement for this element
                    int jdbc_position = 1;
                    ownerMapping.setObject(pm, ps, Mappings.getParametersIndex(jdbc_position,ownerMapping), sm.getObject());
                    jdbc_position += ownerMapping.getNumberOfDatastoreFields();
                    elementMapping.setObject(pm, ps, Mappings.getParametersIndex(jdbc_position,elementMapping), element);

                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(removeStmt);
                    }
                    long startTime = System.currentTimeMillis();
                    int rows_deleted=ps.executeUpdate();
                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(STORE_LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                    }

                    // Update modification indicator if we've not updated yet
                    if (!modified)
                    {
                        modified = (rows_deleted == 1);
                    }
                }
            }
            finally
            {
                ps.close();
                pm.releaseConnection(conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("SCO.RemoveRequestFailed",removeStmt),e);
        }

        return modified;
    }

    /**
     * Clear the association from owner to all elements
     * @param sm State Manager for the container. 
     */
    public void clear(StateManager sm)
    {
        PersistenceManager pm = sm.getPersistenceManager();

        try
        {
            Connection conn = pm.getConnection(true);
            try
            {
                PreparedStatement ps = conn.prepareStatement(clearStmt);
                try
                {
                    int jdbc_position = 1;
                    ownerMapping.setObject(pm, ps, Mappings.getParametersIndex(jdbc_position,ownerMapping), sm.getObject());

                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(clearStmt);
                    }
                    long startTime = System.currentTimeMillis();
                    ps.executeUpdate();
                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(STORE_LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                    }
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                pm.releaseConnection(conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("SCO.ClearRequestFailed",clearStmt),e);
        }
    }

    // ---------------------------- Internal Methods ---------------------------

    /**
     * Generate statement returning the size of the Set.
     * The discriminator part is only present when the element type has
     * inheritance strategy of "superclass-table" and is Inverse.
     * <PRE> 
     * SELECT COUNT(ELEMENTCOL) FROM SETTABLE 
     * WHERE OWNERCOL=?
     * [AND DISCRIMINATOR=?]
     * </PRE> 
     * @return Statement returning the size of the Set.
     */
    protected String getSizeStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("SELECT COUNT(*) FROM ");
        stmt.append(setTable.toString());
        stmt.append(" WHERE ");
        stmt.append(((Column)ownerMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
        stmt.append(" = ");
        stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(0)).getUpdateInputParameter());
        for (int i=1; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(((Column)ownerMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        // Only for inverse cases with a discriminator
        if (elementDiscriminatorMapping != null)
        {
            for (int i=0;i<elementDiscriminatorMapping.getNumberOfDatastoreFields();i++)
            {
                stmt.append(" AND ");
                stmt.append(((Column)elementDiscriminatorMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
                stmt.append(" = ");
                stmt.append(((RDBMSMapping)elementDiscriminatorMapping.getDataStoreMapping(i)).getUpdateInputParameter());
            }
        }

        return stmt.toString();
    }

    /**
     * Generate statement for retrieving the contents of the Set.
     * The discriminator part is only present when the element type has
     * inheritance strategy of "superclass-table" and is Inverse.
     * <PRE> 
     * SELECT OWNERCOL FROM SETTABLE 
     * WHERE OWNERCOL=?
     * AND ELEMENTCOL = ?
     * [AND DISCRIMINATOR = ?]
     * </PRE>
     * @return Statement for retrieving the contents of the set. 
     */
    protected String getContainsStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("SELECT ");
        stmt.append(((Column)ownerMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
        for (int i=1; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(",");
            stmt.append(((Column)ownerMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
        }
        stmt.append(" FROM ");
        stmt.append(setTable.toString());
        stmt.append(" WHERE ");
        stmt.append(((Column)ownerMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
        stmt.append(" = ");
        stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(0)).getUpdateInputParameter());
        for (int i=1; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(((Column)ownerMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        for (int i=0; i<elementMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(((Column)elementMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)elementMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }

        // Only for inverse cases with a discriminator
        if (elementDiscriminatorMapping != null)
        {
            for (int i=0;i<elementDiscriminatorMapping.getNumberOfDatastoreFields();i++)
            {
                stmt.append(" AND ");
                stmt.append(((Column)elementDiscriminatorMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
                stmt.append(" = ");
                stmt.append(((RDBMSMapping)elementDiscriminatorMapping.getDataStoreMapping(i)).getUpdateInputParameter());
            }
        }
        return stmt.toString();
    }

    /**
     * Generate statement for clearing the Set.
     * <PRE>
     * DELETE FROM SETTABLE
     * WHERE OWNERCOL = ?
     * </PRE>
     * @return Statement for clearing the Set
     */
    protected String getClearStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("DELETE FROM ");
        stmt.append(setTable.toString());
        stmt.append(" WHERE ");
        stmt.append(((Column)ownerMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
        stmt.append(" = ");
        stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(0)).getUpdateInputParameter());
        for (int i=1; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(((Column)ownerMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        return stmt.toString();
    }

    /**
     * Generate statement for deleting items from the Set.
     * <PRE>
     * DELETE FROM SETTABLE 
     * WHERE OWNERCOL=?
     * AND ELEMENTCOL = ?
     * </PRE> 
     * @return Statement for deleting items from the Set.
     */
    protected String getRemoveStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("DELETE FROM ");
        stmt.append(setTable.toString());
        stmt.append(" WHERE ");
        stmt.append(((Column)ownerMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
        stmt.append(" = ");
        stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(0)).getUpdateInputParameter());
        for (int i=1; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(((Column)ownerMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        for (int i=0; i<elementMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(((Column)elementMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)elementMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        return stmt.toString();
    }

    // ----------------------------- Query Methods -----------------------------

    public Query.ResultObjectFactory newResultObjectFactory(StateManager sm, QueryStatement stmt, boolean ignoreCache)
    {
        int[] prefetchFieldNumbers;
        StatementExpressionIndex[] statementExpressionIndex;
        
        if (emd == null)
        {
            return new PersistentIDROF(this.elementTable, null, null, null, ignoreCache);           
        }
        FetchPlanImpl fp =((FetchPlanImpl)sm.getPersistenceManager().getFetchPlan());
        fp.manageFetchPlanForClass(emd);
        FetchPlanForClass fpc = fp.getFetchPlanForClass(emd);
        int fieldNumbers[] = fpc.getFieldsInActualFetchPlan();
        int fn[] = new int[fieldNumbers.length];
        int prefetchFieldCount = 0;
        int fieldCount = emd.getNoOfInheritedManagedFields() + emd.getNoOfManagedFields();
        
        statementExpressionIndex = new StatementExpressionIndex[fieldCount];
        
        for (int i=0; i<fieldNumbers.length; ++i)
        {
            JavaTypeMapping m = elementTable.getFieldMapping(emd.getManagedFieldAbsolute(fieldNumbers[i]));
            
            if( m != null ) // field is not stored in the elementTable, e.g. List, Set, etc or is transactional
            {
	            if (m.includeInFetchStatement() && !(m instanceof MappingCallbacks))
	            {
	                statementExpressionIndex[fieldNumbers[i]] = new StatementExpressionIndex();
	                statementExpressionIndex[fieldNumbers[i]].setMapping(m);
	                fn[prefetchFieldCount++] = fieldNumbers[i];
	            }
            }
        }
        
        prefetchFieldNumbers = new int[prefetchFieldCount];
        System.arraycopy(fn, 0, prefetchFieldNumbers, 0, prefetchFieldCount);                

        int[] datastoreIndex = null;

        // following the actual fetch plan, we select the fields to be fetched from the database
        if (stmt.getTableExpression(new TableIdentifier(storeMgr.getDatastoreAdapter(),"ELEMENT")) != null)
        {
            if( elementTable.getIdentityType() == IdentityType.DATASTORE )
            {
                datastoreIndex = stmt.select(new TableIdentifier(storeMgr.getDatastoreAdapter(), "ELEMENT"),elementTable.getDataStoreObjectIdMapping(),true);
            }
            Mappings.selectMapping(stmt, new TableIdentifier(storeMgr.getDatastoreAdapter(), "ELEMENT"), statementExpressionIndex);
        }
        else
        {
            if( elementTable.getIdentityType() == IdentityType.DATASTORE )
            {
                datastoreIndex = stmt.select(new TableIdentifier(storeMgr.getDatastoreAdapter(), "this"),elementTable.getDataStoreObjectIdMapping(),true);
            }
            Mappings.selectMapping(stmt,statementExpressionIndex);
        }        
        
        return new PersistentIDROF(
            this.getStoreManager().getDatastoreClass(getElementType()),
            prefetchFieldNumbers,
            statementExpressionIndex,
            datastoreIndex,
            ignoreCache);
    }


    public QueryStatement getExistsSubquery(JavaTypeMapping mapping,
                                			TableExpression ownerTe,
                                			SQLIdentifier setRangeVar)
    {
        QueryStatement stmt = ((DatabaseAdapter) dba).newQueryStatement(setTable, setRangeVar);

        ScalarExpression ownerExpr      = mapping.newScalarExpression(stmt, ownerTe);
        ScalarExpression ownerInSetExpr = ownerMapping.newScalarExpression(stmt, stmt.getTableExpression(setRangeVar));

        stmt.andCondition(ownerExpr.eq(ownerInSetExpr));

        stmt.select(setRangeVar, elementMapping);

        return stmt;
    }
    
    /**
     * Check if the mapping correspond to a non pc object or embedded field
     * @param mapping the mapping
     * @return true if the field is embedded into one column
     */
    protected boolean isEmbeddedMapping(JavaTypeMapping mapping)
    {
        return !InterfaceMapping.class.isAssignableFrom(mapping.getClass()) &&
        	   !OIDMapping.class.isAssignableFrom(mapping.getClass());
    }    
}
