/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - updated retrieval of ColumnList's
2003 Andy Jefferson - updated to support inherited elements
2003 Andy Jefferson - fixed sizeStmt to ignore records with no index value
2004 Andy Jefferson - merged IteratorStmt and GetStmt into GetRangeStmt.
2004 Andy Jefferson - moved majority statement construction to AbstractListStore.
2004 Andy Jefferson - fixed use of shift statement
2004 Andy Jefferson - added support for inverse 1-N unidirectional
    ...
**********************************************************************/
package org.jpox.store.rdbms.scostore;

import java.lang.reflect.Field;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.util.Collection;
import java.util.Iterator;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOHelper;
import javax.jdo.JDOUserException;
import javax.jdo.spi.PersistenceCapable;

import org.jpox.ClassLoaderResolver;
import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.metadata.CollectionMetaData;
import org.jpox.metadata.DiscriminatorStrategy;
import org.jpox.metadata.FieldMetaData;
import org.jpox.metadata.InheritanceStrategy;
import org.jpox.store.DatastoreClass;
import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.FieldValues;
import org.jpox.store.exceptions.IncompatibleQueryElementTypeException;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.TableExpression;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.Mappings;
import org.jpox.store.query.IteratorStatement;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.store.rdbms.adapter.DatabaseAdapter;
import org.jpox.store.rdbms.mapping.RDBMSMapping;
import org.jpox.store.rdbms.table.ClassTable;
import org.jpox.util.JPOXLogger;

/**
 * Representation of an Inverse List as part of a relationship. This class is
 * used where you have a 1-N and the tables are not joined via a join table.
 * There is an owner table and an element table, and the element table has a
 * column being the id of the owner table. This is in contrast to
 * NormalListStore which represents 1-N relationships using a join table.
 * There are 2 possible uses here
 * <UL>
 * <LI><B>bidirectional</B> - where the owner has a List of elements, and
 * the element has an owner. In this case the element class will have an owner field
 * and index which can be updated directly</LI>
 * <LI><B>unidirectional</B> - where the owner has a List of elements, but
 * the element knows nothing about the owner. In this case the element class has no
 * owner field. In this case the owner/index columns in the element table have to be
 * updated directly.</LI>
 * </UL>
 *
 * @version $Revision: 1.10 $
 **/
public class InverseListStore extends AbstractListStore
{
    private final boolean hasOwnerFieldInElement;
    private final int ownerFieldNumber;

    private final String removeAtNullifyStmt;
    private final String clearNullifyStmt;

    /** Statement for updating a foreign key in a 1-N unidirectional */
    private final String updateFkStmt;

    /**
     * Constructor.
     * @param fmd Field MetaData for the field that this represents
     * @param storeMgr The Store Manager in use
     */
    public InverseListStore(FieldMetaData fmd, RDBMSManager storeMgr)
    {
        this.storeMgr = storeMgr;
        dba = storeMgr.getDatabaseAdapter();

        Field field = fmd.getField();
        CollectionMetaData colmd = (CollectionMetaData)fmd.getContainer();
        if (colmd == null)
        {
            throw new JDOUserException("No collection metadata found in " + fmd);
        }

        // Load the element class
        elementType = colmd.getElementType();
        ClassLoaderResolver clr=storeMgr.getClassLoaderResolver();
        Class element_class = clr.classForName(elementType);

        // Check that the element class has MetaData
        emd = storeMgr.getMetaDataManager().getMetaDataForClass(element_class);
        if (emd == null)
        {
            throw new JDOUserException("The MetaData for the class " + element_class.getName() + " was not found." +
                " This class is declared to be used as an element for the List in the field: " + fmd.getFullFieldName());
        }

        elementTable = (ClassTable) storeMgr.getDatastoreClass(elementType);
        elementMapping = elementTable.getIDMapping();
        elementsAreEmbedded = false;
        if (emd != null)
        {
            if (emd.getInheritanceMetaData().getStrategyValue() == InheritanceStrategy.SUPERCLASS_TABLE &&
                emd.getDiscriminatorStrategy() != DiscriminatorStrategy.NONE)
            {
                elementDiscriminatorMapping = elementTable.getDiscriminatorMapping();
            }
        }

        // Get the field in the element table (if any)
        String mappedByFieldName = fmd.getMappedBy();
        if (mappedByFieldName != null)
        {
            // Inverse 1-N bidirectional
            // The element class has a field for the owner.
            FieldMetaData eofmd = storeMgr.getMetaDataManager().getMetaDataForField(element_class, mappedByFieldName);
            if (eofmd == null)
            {
                throw new JDOUserException("The field " + mappedByFieldName + " was not found in the class " + element_class.getName());
            }

            // Check that the type of the element "mapped-by" field is consistent with the owner type
            if (!clr.isAssignableFrom(eofmd.getField().getType(), fmd.getClassMetaData().getFullClassName()))
            {
                throw new JDOUserException("Can't assign class " + fmd.getClassMetaData().getFullClassName() +
                    " to the field " + eofmd.getFullFieldName() + " declared as " + eofmd.getField().getType().getName());
            }

            String ownerFieldName = eofmd.getName();
            ownerFieldNumber = emd.getFieldNumberAbsolute(ownerFieldName);
            if (!elementTable.isFieldPersistent(ownerFieldNumber))
            {
                throw new JDOUserException("Non-persistent field type for collection owner field: " + 
                    field.getDeclaringClass().getName() + "." + field.getName());
            }

            // Find field mapping for owner field
            JavaTypeMapping om = elementTable.getFieldMapping(emd.getManagedFieldAbsolute(ownerFieldNumber));
            if (isEmbeddedMapping(om))
            {
                throw new JDOUserException("Invalid field type for collection owner field: " + 
                    field.getDeclaringClass().getName() + "." + field.getName());
            }
            ownerMapping = om;
            indexMapping = elementTable.getOrderMapping(eofmd);
            hasOwnerFieldInElement = true;
        }
        else
        {
            // Inverse 1-N unidirectional
            // The element class knows nothing about the owner
            JavaTypeMapping em = elementTable.getExternalFkMapping(fmd);
            if (em == null)
            {
                throw new JDOUserException("Class " + fmd.getClassMetaData().getFullClassName() + " has collection field " + fmd.getName() + 
                    " and this has no mapping in the table for the element class " + elementType);
            }
            ownerMapping = em;
            JavaTypeMapping om = elementTable.getOrderMapping(fmd);
            if (om == null)
            {
                throw new JDOUserException("Class " + fmd.getClassMetaData().getFullClassName() + " has collection field " + fmd.getName() + 
                    " and this has no order mapping in the table for the element class " + elementType);
            }
            indexMapping = om;
            ownerFieldNumber = -1;
            hasOwnerFieldInElement = false;
        }

        listTable = elementTable;
        listName = "inverseList";

        // Generate the statements
        initialiseStatements();
        addStmt = getAddStmt();
        removeAtNullifyStmt = getRemoveAtNullifyStmt();
        clearNullifyStmt = getClearNullifyStmt();
        updateFkStmt = getUpdateFkStmt();
    }

    /**
     * Utility to update a foreign-key in the element in the case of
     * a unidirectional 1-N relationship.
     * @param sm StateManager for the owner
     * @param element The element to update
     * @param owner The owner object to set in the FK
     * @param index The index position
     * @return Whether it was performed successfully
     */
    private boolean updateElementFk(StateManager sm, Object element, Object owner, int index)
    {
        if (element == null)
        {
            return false;
        }
        validateElementForWriting(sm, element);

        boolean retval;
        PersistenceManager pm = sm.getPersistenceManager();
        try
        {
            Connection conn = pm.getConnection(false);
            try
            {
                PreparedStatement ps = conn.prepareStatement(updateFkStmt);
                try
                {
                    int jdbcPosition = 1;
                    if (owner == null)
                    {
                        ownerMapping.setObject(pm, ps, Mappings.getParametersIndex(jdbcPosition,ownerMapping), null);
                    }
                    else
                    {
                        ownerMapping.setObject(pm, ps, Mappings.getParametersIndex(jdbcPosition,ownerMapping), sm.getObject());
                    }

                    jdbcPosition += ownerMapping.getNumberOfDatastoreFields();
                    indexMapping.setObject(pm, ps, Mappings.getParametersIndex(jdbcPosition,indexMapping), new Integer(index));

                    jdbcPosition += indexMapping.getNumberOfDatastoreFields();
                    elementMapping.setObject(pm, ps, Mappings.getParametersIndex(jdbcPosition,elementMapping), element);

                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(updateFkStmt);
                    }
                    long startTime = System.currentTimeMillis();
                    ps.executeUpdate();
                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(STORE_LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                    }

                    retval = true;
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                pm.releaseConnection(conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("SCO.UpdateFkRequestFailed",updateFkStmt),e);
        }

        return retval;
    }

    /**
     * Generate statement for updating the owner, index columns in an inverse 1-N.
     * <PRE>
     * UPDATE ELEMENTTABLE SET FK_COL_1 = ?, FK_COL_2 = ?, FK_IDX = ?
     * WHERE ELEMENT_ID = ?
     * </PRE>
     * @return Statement for updating the owner/index in an inverse 1-N
     */
    private String getUpdateFkStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("UPDATE ");
        stmt.append(elementTable.toString());
        stmt.append(" SET ");
        for (int i=0; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(",");
            }
            stmt.append(((Column)ownerMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        for (int i=0; i<indexMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(",");
            stmt.append(((Column)indexMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)indexMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        stmt.append(" WHERE ");
        for (int i=0; i<elementMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(" AND ");
            }
            stmt.append(((Column)elementMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)elementMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }

        return stmt.toString();
    }

    /**
     * Generate statement for adding an item to the List. Overrides the method
     * in AbstractListStore because we just update the index on an existing
     * element record.
     * <PRE>
     * UPDATE ELEMENTTABLE SET INDEXCOL = ?
     * WHERE ELEMENTCOL=?
     * </PRE>
     * @return The Statement for adding an item
     */
    private String getAddStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("UPDATE ");
        stmt.append(elementTable.toString());
        stmt.append(" SET ");
        for (int i=0; i<indexMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(",");
            }
            stmt.append(((Column)indexMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)indexMapping.getDataStoreMapping(0)).getUpdateInputParameter());
        }
        stmt.append(" WHERE ");
        for (int i=0; i<elementMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(" AND ");
            }
            stmt.append(((Column)elementMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)elementMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        return stmt.toString();
    }

    // ------------------------- GetRange Statement ----------------------------

    /**
     * Accessor for the GetRange statement to retrieve element(s) in a range 
     * from the List. This has 5 modes
     * <UL>
     * <LI>start and end have the same value and &ge;0 then get that
     * element</LI>
     * <LI>start and end have values &ge;0 and are different then get a
     * range</LI>
     * <LI>start is &ge;0 and end is -1 the get all from start (inclusive).</LI>
     * <LI>start is -1 and end is &ge;0 the get all up to end(exclusive).</LI>
     * <LI>start and end values are both -1 then get all elements</LI> 
     * </UL>
     * @param start_index The start position in the List.
     * @param end_index The end position in the List.
     * @return The QueryStatement.
     **/
    protected QueryStatement getGetRangeStatement(StateManager sm,int start_index,int end_index)
    {
        final ClassLoaderResolver clr=storeMgr.getClassLoaderResolver();
        QueryStatement stmt = new IteratorStatement(
                new IteratorStatement.AssociationEnd() 
                {
                    public JavaTypeMapping getMapping()
                    {
                        return elementMapping;
                    }
                    public Class getType()
                    {
                        return clr.classForName(elementType);
                    }
                    public DatastoreContainerObject getDatastoreContainerObject()
                    {
                        return listTable;
                    }
                    public boolean useJoin()
                    {
                        return false;
                    }                    
                },
                clr.classForName(this.elementType),
                this.storeMgr,
                this.dba,
                true).getQueryStatement();
 
        // Apply condition on join-table owner field to filter by owner
        ScalarExpression ownerExpr = ownerMapping.newScalarExpression(stmt, stmt.getDefaultTableExpression());
        ScalarExpression ownerVal = ownerMapping.newLiteral(stmt, sm.getObject());
        stmt.andCondition(ownerExpr.eq(ownerVal), true);

        boolean returning_range=false;
        if (start_index == -1 && end_index == -1)
        {
            // Add no more constraints. We want all elements.
            returning_range = true; 
        }
        else if (start_index == end_index && start_index >= 0)
        {
            // Apply condition on join-table owner field to filter by index
            ScalarExpression indexExpr = indexMapping.newScalarExpression(stmt,stmt.getDefaultTableExpression());
            ScalarExpression indexVal = indexMapping.newLiteral(stmt, new Integer(start_index));
            
            stmt.andCondition(indexExpr.eq(indexVal), true);     
        }
        else
        {
            returning_range = true; 

            // Apply filter to retrieve indexes in the range
            // a). start_index (inclusive) constraint
            if (start_index >= 0)
            { 
                // Apply condition on join-table index field to filter by index
                ScalarExpression indexExpr = indexMapping.newScalarExpression(stmt,stmt.getDefaultTableExpression());
                ScalarExpression indexVal = indexMapping.newLiteral(stmt, new Integer(start_index));
                stmt.andCondition(indexExpr.gteq(indexVal), true);   
            }
            // b). end_index (exclusive) constraint
            if (end_index >= 0)
            {
                // Apply condition on join-table index field to filter by index
                ScalarExpression indexExpr = indexMapping.newScalarExpression(stmt,stmt.getDefaultTableExpression());
                ScalarExpression indexVal = indexMapping.newLiteral(stmt, new Integer(end_index));
                stmt.andCondition(indexExpr.lt(indexVal), true); 
            }
        }

		// If returning a range, add ordering by index
        if (returning_range)
        {
		    ScalarExpression exprIndex[] = new ScalarExpression[indexMapping.getNumberOfDatastoreFields()];
		    boolean descendingOrder[] = new boolean[indexMapping.getNumberOfDatastoreFields()];
            exprIndex = indexMapping.newScalarExpression(stmt,stmt.getDefaultTableExpression()).getExpressionList().toArray();
		    stmt.setOrdering(exprIndex, descendingOrder);
        }

        return stmt;
    }

    /**
     * Internal method for adding an item to the List.
     * @param sm The state manager
     * @param startAt The start position
     * @param atEnd Whether to add at the end
     * @param c The Collection of elements to add.
     * @return Whether it was successful
     **/
    protected boolean internalAdd(StateManager sm, 
                                  int startAt,
                                  boolean atEnd,
                                  Collection c)
    {
        if (c == null || c.size() == 0)
        {
            return true;
        }
        int shift = c.size();

        // Check all elements to be inserted
        for (Iterator i = c.iterator(); i.hasNext();)
        {
            validateElementForWriting(sm, i.next());
        }

        // Check what we have already
        int size = size(sm);

        PersistenceManager pm = sm.getPersistenceManager();

        try
        {
            Connection conn = pm.getConnection(true);

            try
            {
                // shift up by c.size
                if (!atEnd && startAt != size)
                {
                    PreparedStatement ps = conn.prepareStatement(shiftStmt);

                    try
                    {
                        for (int i=size-1; i>=startAt; i--)
                        {
                            indexMapping.setObject(pm, ps, Mappings.getParametersIndex(1,indexMapping), new Integer(shift));
                            ownerMapping.setObject(pm, ps, Mappings.getParametersIndex(indexMapping.getNumberOfDatastoreFields()+1,ownerMapping), sm.getObject());
                            indexMapping.setObject(pm, ps, Mappings.getParametersIndex(indexMapping.getNumberOfDatastoreFields()+ownerMapping.getNumberOfDatastoreFields()+1,indexMapping), new Integer(i));

                            if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                            {
                                JPOXLogger.RDBMS_SQL.debug(shiftStmt);
                            }
                            long startTime = System.currentTimeMillis();
                            ps.executeUpdate();
                            if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                            {
                                JPOXLogger.RDBMS_SQL.debug(STORE_LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                            }
                        }
                    }
                    finally
                    {
                        ps.close();
                    }
                }
                else
                {
                    startAt = size;
                }

                // Insert at correct location. This just updates the index
                // id to the right value since the element records are already
                // there.
                PreparedStatement ps = conn.prepareStatement(addStmt);
                try
                {
                    for (Iterator i=c.iterator(); i.hasNext(); )
                    {
                        indexMapping.setObject(pm, ps, Mappings.getParametersIndex(1,indexMapping), new Integer(startAt));
                        elementMapping.setObject(pm, ps, Mappings.getParametersIndex(indexMapping.getNumberOfDatastoreFields()+1,elementMapping), i.next());
                        startAt++;

                        if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                        {
                            JPOXLogger.RDBMS_SQL.debug(shiftStmt);
                        }
                        long startTime = System.currentTimeMillis();
                        ps.executeUpdate();
                        if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                        {
                            JPOXLogger.RDBMS_SQL.debug(STORE_LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                        }
                    }
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                pm.releaseConnection(conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("SCO.AddRequestFailed",addStmt),e);
        }
        return true;
    }

    /**
     * Generates the statement for removing an item by nulling it out.
     * <PRE>
     * UPDATE LISTTABLE SET OWNERCOL = NULL, INDEXCOL = NULL
     * WHERE OWNERCOL = ?
     * AND INDEXCOL = ?
     * </PRE>
     * @return The Statement for removing an item from a position
     */
    protected String getRemoveAtNullifyStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("UPDATE ");
        stmt.append(listTable.toString());
        stmt.append(" SET ");
        for (int i=0; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(", ");
            }
            stmt.append(((Column)ownerMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString() + " = NULL");
        }
        for (int i=0; i<indexMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(", ");
            stmt.append(((Column)indexMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString() + " = NULL");
        }
        
        stmt.append(" WHERE ");
        for (int i=0; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(" AND ");
            }
            stmt.append(((Column)ownerMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        for (int i=0; i<indexMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(((Column)indexMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)indexMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        return stmt.toString();
    }

    /**
     * Internal method to remove an object at a location in the List.
     * Differs from the Normal List in that it nulls out the owner FK. 
     * @param sm The state manager.
     * @param index The location
     **/
    protected void removeAt(StateManager sm, int index)
    {
        String stmt = removeAtStmt;
        if (ownerMapping.isNullable() && indexMapping.isNullable())
        {
            JPOXLogger.RDBMS.debug(LOCALISER.msg("SCO.List.RemoveElementFromInverseNull"));
            stmt = removeAtNullifyStmt;
        }
        else
        {
            JPOXLogger.RDBMS.debug(LOCALISER.msg("SCO.List.RemoveElementFromInverseDelete"));
        }
        internalRemoveAt(sm, index, stmt);
    }

    /**
     * Generates the statement for clearing items by nulling the owner link out.
     * <PRE> 
     * UPDATE LISTTABLE 
     * SET OWNERCOL = NULL, INDEXCOL = NULL
     * WHERE OWNERCOL = ?
     * </PRE>
     * @return The Statement for clearing items for the owner.
     */
    protected String getClearNullifyStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("UPDATE ");
        stmt.append(listTable.toString());
        stmt.append(" SET ");
        for (int i=0; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(", ");
            }
            stmt.append(((Column)ownerMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString() + " = NULL");
        }
        for (int i=0; i<indexMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(", ");
            stmt.append(((Column)indexMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString() + " = NULL");
        }

        stmt.append(" WHERE ");
        for (int i=0; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(" AND ");
            }
            stmt.append(((Column)ownerMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        return stmt.toString();
    }
    
    /**
     * Method to clear the List.
     * @param sm The state manager
     **/
    public void clear(StateManager sm)
    {
        String stmt = clearStmt;
        if (ownerMapping.isNullable() && indexMapping.isNullable())
        {
            JPOXLogger.RDBMS.debug(LOCALISER.msg("SCO.List.ClearElementsFromInverseNull"));
            stmt = clearNullifyStmt;
        }
        else
        {
            JPOXLogger.RDBMS.debug(LOCALISER.msg("SCO.List.ClearElementsFromInverseDelete"));
        }
        internalClear(sm, stmt);
    }

    /**
     * Method to validate that an element is valid for writing to the datastore.
     * @param sm StateManager for the List
     * @param element The element to validate
     */
    protected void validateElementForWriting(StateManager sm, Object element)
    {
        if (!validateElementType(sm.getPersistenceManager().getClassLoaderResolver(), element))
        {
            throw new ClassCastException(LOCALISER.msg("SCO.List.ElementIsInvalid",element.getClass().getName(),elementType));
        }

        if (element != null && !elementsAreEmbedded)
        {
            PersistenceManager pm = sm.getPersistenceManager();
    
            final Object newOwner = sm.getObject();
            if (!JDOHelper.isPersistent(element))
            {
                /*
                 * The new element is not yet persistent.
                 * Update its owner field to the appropriate value and *then*
                 * make it persistent. Making the change before DB insertion
                 * avoids an unnecessary UPDATE and allows the owner field to
                 * be non-nullable.
                 */
                pm.makePersistent(element, new FieldValues()
                {
                    public void fetchFields(StateManager esm)
                    {
                        if (hasOwnerFieldInElement)
                        {
                            esm.replaceField(ownerFieldNumber, newOwner);
                        }
                    }
                });

                // Unidirectional 1-N, so set the owner
                // TODO Consider doing this as part of makePersistent to reduce number of DB calls
                if (!hasOwnerFieldInElement)
                {
                    // Just put in index 0 for now
                    updateElementFk(sm, element, newOwner, 0);
                }
            }
            else
            {
                PersistenceCapable pcElement = (PersistenceCapable) element;
                /*
                 * The new element is already persistent.
                 * "Add" the new value to the set by updating its owner field
                 * to the appropriate value. This is done with the same methods
                 * the PC itself would use if the application code modified the
                 * field. It should result in no actual database activity if 
                 * the field was already set to the right value.
                 */
                if (pm != pcElement.jdoGetPersistenceManager())
                {
                    throw new JDOUserException(LOCALISER.msg("SCO.List.WriteInvalidWithDifferentPM"),pcElement.jdoGetObjectId());
                }

                StateManager esm = pm.findStateManager(pcElement);

                if (hasOwnerFieldInElement)
                {
                    // Ensure the current owner value is loaded.
                    esm.isLoaded(pcElement, ownerFieldNumber);

                    // Get the current value.
                    Object oldOwner = esm.provideField(ownerFieldNumber);

                    // Replace with new value.
                    esm.setObjectField(pcElement, ownerFieldNumber, oldOwner, newOwner);
                }
            }
        }
    }

    // ---------------------------- Query Methods ------------------------------

    /**
     * Utility method to return a new QueryStatement for retrieval of the
     * elements of this List.
     * @param sm StateManager for this object
     * @param candidateClass Class for the element end of the link.
     * @return The QueryStatement
     **/
    public QueryStatement newQueryStatement(StateManager sm, String candidateClass)
    {
        if (!sm.getPersistenceManager().getClassLoaderResolver().isAssignableFrom(elementType,candidateClass))
        {
            throw new IncompatibleQueryElementTypeException(elementType, candidateClass);
        }

        DatastoreClass candidateTable = storeMgr.getDatastoreClass(candidateClass);
        QueryStatement stmt = ((DatabaseAdapter) dba).newQueryStatement(candidateTable);

        ScalarExpression ownerExpr = ownerMapping.newScalarExpression(stmt, stmt.getDefaultTableExpression());
        ScalarExpression ownerVal = ownerMapping.newLiteral(stmt, sm.getObject());
        stmt.andCondition(ownerExpr.eq(ownerVal));

        stmt.select(elementMapping);

        return stmt;
    }

    /**
     * Method used during Queries where contains() is called.
     **/
    public ScalarExpression joinElementsTo(
            		    QueryStatement stmt,
    					JavaTypeMapping ownerMapping,
    					TableExpression te,
    					DatastoreIdentifier setRangeVar,
                        Class filteredElementType,
                        TableExpression elmte,
                        DatastoreIdentifier elementRangeVar)
    {
        ClassLoaderResolver clr=storeMgr.getClassLoaderResolver();
        if (!clr.isAssignableFrom(elementType,filteredElementType) &&
            !clr.isAssignableFrom(filteredElementType,elementType))
        {
            throw new IncompatibleQueryElementTypeException(elementType, filteredElementType.getName());
        }

        // Join the element table on the owner ID column.
        ClassTable filteredElementTable = (ClassTable) storeMgr.getDatastoreClass(filteredElementType.getName());
        stmt.newTableExpression(filteredElementTable, elementRangeVar);

        ScalarExpression ownerExpr = ownerMapping.newScalarExpression(stmt,te);
        ScalarExpression ownerListExpr = this.ownerMapping.newScalarExpression(stmt,stmt.getTableExpression(elementRangeVar));
        stmt.innerJoin(ownerExpr,ownerListExpr,stmt.getTableExpression(elementRangeVar), true);
        
        JavaTypeMapping elementTableID = filteredElementTable.getIDMapping();
        return elementTableID.newScalarExpression(stmt,stmt.getTableExpression(elementRangeVar));
    }
}