/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others. All rights reserved.
This program and the accompanying materials are made available under
the terms of the JPOX License v1.0 which accompanies this distribution.

Contributors:
2003 Erik Bengtson - removed unused variable
2003 Andy Jefferson - coding standards
2003 Andy Jefferson - addition of getGetStatement for inherited values
2004 Andy Jefferson - addition of query methods
2004 Marco Schulze  - replaced catch(NotPersistenceCapableException ...) by
                      advance-check via TypeManager.isSupportedType(...)
2004 Andy Jefferson - moved statements to AbstractMapStore
2004 Andy Jefferson - added support for inverse 1-N unidirectional
    ...
**********************************************************************/
package org.jpox.store.rdbms.scostore;

import java.lang.reflect.Field;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.util.Iterator;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOUserException;
import javax.jdo.spi.PersistenceCapable;

import org.jpox.ClassLoaderResolver;
import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.TypeManager;
import org.jpox.metadata.FieldMetaData;
import org.jpox.metadata.MapMetaData;
import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.FieldValues;
import org.jpox.store.exceptions.ClassDefinitionException;
import org.jpox.store.exceptions.IncompatibleQueryElementTypeException;
import org.jpox.store.expression.BooleanExpression;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.TableExpression;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.Mappings;
import org.jpox.store.mapping.ObjectMapping;
import org.jpox.store.mapping.SimpleDatastoreRepresentation;
import org.jpox.store.query.IteratorStatement;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.store.rdbms.adapter.DatabaseAdapter;
import org.jpox.store.rdbms.mapping.RDBMSMapping;
import org.jpox.store.rdbms.sqlidentifier.SQLIdentifier;
import org.jpox.store.rdbms.table.ClassTable;
import org.jpox.store.scostore.SetStore;
import org.jpox.util.JPOXLogger;

/**
 * Representation of an "inverse" backing for a Map. This class is
 * used where you have a 1-N and the tables are not joined via a link table.
 * There are 2 possible uses here
 * <UL>
 * <LI><B>bidirectional</B> - where the owner has a Map of (key,value)s, and
 * the value has an owner. In this case the value class will have an owner field
 * which can be updated directly</LI>
 * <LI><B>unidirectional</B> - where the owner has a Map of (key,value), but
 * the value knows nothing about the owner. In this case the value class has no
 * owner field. In this case the column in the value table has to be updated
 * directly.</LI>
 * </UL>
 *
 * @version $Revision: 1.9 $
 **/
public class InverseMapStore extends AbstractMapStore
{
    private final boolean hasOwnerFieldInValue;
    private final int ownerFieldNumber;
    private final int keyFieldNumber;

    /** Statement for updating a foreign key in a 1-N unidirectional */
    private final String updateFkStmt;

    private SetStore keySetStore = null;
    private SetStore valueSetStore = null;
    private SetStore entrySetStore = null;

    /**
     * Constructor for an Inverse Map.
     * @param fmd Field Meta-Data for the Map field.
     * @param store_mgr The Store Manager we are using.
     **/
    public InverseMapStore(FieldMetaData fmd, RDBMSManager store_mgr)
    {
        storeMgr = store_mgr;
        dba = storeMgr.getDatastoreAdapter();

        Field field = fmd.getField();
        MapMetaData mmd = (MapMetaData)fmd.getContainer();
        if (mmd == null)
        {
            throw new JDOUserException("No map metadata found in " + fmd);
        }

        // Load the key and value classes
        keyType = mmd.getKeyType();
        valueType = mmd.getValueType();
        ClassLoaderResolver clr = storeMgr.getClassLoaderResolver();
        Class value_class = clr.classForName(valueType);

        // Check that the value class has MetaData
        vmd = storeMgr.getMetaDataManager().getMetaDataForClass(value_class);
        if (vmd == null)
        {
            throw new JDOUserException("The MetaData for the class " + value_class.getName() + " was not found." +
                " This class is declared to be used as a value for the Map in the field: " + fmd.getFullFieldName());
        }

        valueTable = (ClassTable) storeMgr.getDatastoreClass(valueType);
        valueMapping  = storeMgr.getDatastoreClass(valueType).getIDMapping();
        valuesAreEmbedded = false;

        // Get the field in the value table (if any)
        String mappedByFieldName = fmd.getMappedBy();
        if (mappedByFieldName != null)
        {
            // Inverse 1-N bidirectional
            // The value class has a field for the owner.
            FieldMetaData vofmd = vmd.getField(mappedByFieldName);
            if (vofmd == null)
            {
                throw new JDOUserException("The field " + mappedByFieldName + " was not found in the class " + value_class.getName());
            }
            String ownerFieldName = vofmd.getName();

            // Check that the type of the value "mapped-by" field is consistent with the owner type
            if (!clr.isAssignableFrom(vofmd.getField().getType(), fmd.getClassMetaData().getFullClassName()))
            {
                throw new JDOUserException("Can't assign class " + fmd.getClassMetaData().getFullClassName() +
                    " to the field " + vofmd.getFullFieldName() + " declared as " + vofmd.getField().getType().getName());
            }

            ownerFieldNumber = vmd.getFieldNumberAbsolute(ownerFieldName);
            if (!valueTable.isFieldPersistent(ownerFieldNumber))
            {
                throw new JDOUserException("Non-persistent field type for map owner field: " + 
                    field.getDeclaringClass().getName() + "." + field.getName());
            }

            // Find field mapping for owner field
            JavaTypeMapping om = valueTable.getFieldMapping(vmd.getManagedFieldAbsolute(ownerFieldNumber));
            if (isEmbeddedMapping(om))
            {
                throw new JDOUserException("Invalid field type for map owner field: " + 
                    field.getDeclaringClass().getName() + "." + field.getName());
            }
            ownerMapping = om;
            hasOwnerFieldInValue = true;
        }
        else
        {
            // Inverse 1-N Unidirectional
            // The value class knows nothing about the owner
            JavaTypeMapping om = valueTable.getExternalFkMapping(fmd);
            if (om == null)
            {
                throw new JDOUserException("Class " + fmd.getClassMetaData().getFullClassName() + " has map field " + fmd.getName() + 
                    " and this has no mapping in the table for the value class " + valueType);
            }
            ownerMapping = om;
            ownerFieldNumber = -1;
            hasOwnerFieldInValue = false;
        }

        // Load the "key-field" tag
        FieldMetaData vkfmd = null;
        String key_field_name = mmd.getValueForExtension("key-field");
        if (key_field_name != null)
        {
            // check if key field exists in the ClassMetaData for the element-value type
            vkfmd = storeMgr.getMetaDataManager().getMetaDataForField(value_class, key_field_name);
            if (vkfmd == null)
            {
                throw new JDOUserException("The field " + key_field_name + " was not found in the class " + value_class.getName());
            }
        }
        if (vkfmd == null)
        {
            throw new ClassDefinitionException("Missing map \"key-field\" in " + fmd);
        }
        if (!clr.isAssignableFrom(vkfmd.getField().getType(), keyType))
        {
            throw new JDOUserException("Can't assign class " + keyType + " to the field " + 
                vkfmd.getFullFieldName() + " declared as " + vkfmd.getField().getType().getName());
        }

        // Set up key field
        String keyFieldName = vkfmd.getName();
        keyFieldNumber = vmd.getFieldNumberAbsolute(keyFieldName);
        if (!valueTable.isFieldPersistent(keyFieldNumber))
        {
            throw new JDOUserException("Non-persistent field type for map key field: " + field.getDeclaringClass().getName() + "." + field.getName());
        }

        // Extract key and value mappings
        JavaTypeMapping km = valueTable.getFieldMapping(vmd.getManagedFieldAbsolute(keyFieldNumber));
        if (!SimpleDatastoreRepresentation.class.isAssignableFrom(km.getClass()))
        {
            throw new JDOUserException("Invalid field type for map key field: " + field.getDeclaringClass().getName() + "." + field.getName());
        }
        keyMapping = km;
        keysAreEmbedded = isEmbeddedMapping(keyMapping);

        mapTable = valueTable;

        // Generate the statements
        initialiseStatements();
        updateFkStmt = getUpdateFkStmt();
    }

    /**
     * Generate statement for updating a Foreign Key from value to owner in an inverse 1-N.
     * <PRE>
     * UPDATE ELEMENTTABLE SET FK_COL_1 = ?, FK_COL_2 = ?
     * WHERE ELEMENT_ID = ?
     * </PRE>
     * @return Statement for updating the FK in an inverse 1-N
     */
    private String getUpdateFkStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("UPDATE ");
        stmt.append(mapTable.toString());
        stmt.append(" SET ");
        for (int i=0; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(",");
            }
            stmt.append(((Column)ownerMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        stmt.append(" WHERE ");
        for (int i=0; i<valueMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(" AND ");
            }
            stmt.append(((Column)valueMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)valueMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }

        return stmt.toString();
    }

    /**
     * Utility to update a foreign-key in the value in the case of
     * a unidirectional 1-N relationship.
     * @param sm StateManager for the owner
     * @param element The value to update
     * @param owner The owner object to set in the FK
     * @return Whether it was performed successfully
     */
    private boolean updateValueFk(StateManager sm, Object element, Object owner)
    {
        if (element == null)
        {
            return false;
        }
        validateValueForWriting(sm, element);

        boolean retval;
        PersistenceManager pm = sm.getPersistenceManager();
        try
        {
            Connection conn = pm.getConnection(false);
            try
            {
                PreparedStatement ps = conn.prepareStatement(updateFkStmt);
                try
                {
                    if (owner == null)
                    {
                        ownerMapping.setObject(pm, ps, Mappings.getParametersIndex(1,ownerMapping), null);
                    }
                    else
                    {
                        ownerMapping.setObject(pm, ps, Mappings.getParametersIndex(1,ownerMapping), sm.getObject());
                    }
                    valueMapping.setObject(pm, ps, Mappings.getParametersIndex(ownerMapping.getNumberOfDatastoreFields()+1, valueMapping), element);

                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(updateFkStmt);
                    }
                    long startTime = System.currentTimeMillis();
                    ps.executeUpdate();
                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(STORE_LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                    }

                    retval = true;
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                pm.releaseConnection(conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("SCO.UpdateFkRequestFailed",updateFkStmt),e);
        }

        return retval;
    }

    // ---------------------------- Statement Methods ------------------------------

    /**
     * Accessor for the Get statement to retrieve an value from the Map.
     * @return The QueryStatement.
     **/
    protected QueryStatement getGetStatement(StateManager sm, Object key)
    {
        final ClassLoaderResolver clr=storeMgr.getClassLoaderResolver();
        QueryStatement stmt = new IteratorStatement(
                new IteratorStatement.AssociationEnd()
                {
                    public JavaTypeMapping getMapping()
                    {
                        return valueMapping;
                    }
                    public Class getType()
                    {
                        return clr.classForName(valueType);
                    }
                    public DatastoreContainerObject getDatastoreContainerObject()
                    {
                        return valueTable;
                    }
                    public boolean useJoin()
                    {
                        return false;
                    }                    
                },
                clr.classForName(this.valueType),
                this.storeMgr,
                this.dba,
                true).getQueryStatement();

        // Apply condition on join-table owner field to filter by owner
        ScalarExpression ownerExpr = ownerMapping.newScalarExpression(stmt, stmt.getDefaultTableExpression());
        ScalarExpression ownerVal = ownerMapping.newLiteral(stmt, sm.getObject());
        stmt.andCondition(ownerExpr.eq(ownerVal), true);        

        //TODO
        //this is a database hack. :-)
        //if the keyMapping contains a BLOB column (or any other column not supported by the database
        //as primary key), uses like instead of the operator OP_EQ (=)
        //in future do not check if the keyMapping is of ObjectMapping, but use the database 
        //adapter to check the data types not supported as primary key
        //
        //if object mapping (BLOB) use like
        if (keyMapping instanceof ObjectMapping)
        {
            // Apply condition on join-table owner field to filter by owner
            ScalarExpression keyExpr = keyMapping.newScalarExpression(stmt,stmt.getDefaultTableExpression());
            ScalarExpression keyVal = keyMapping.newLiteral(stmt, key).add(dba.getMapping(String.class).newLiteral(stmt,"%"));
            stmt.andCondition(new BooleanExpression(keyExpr,ScalarExpression.OP_LIKE,keyVal), true);             
        }
        else
        {
            // Apply condition on join-table owner field to filter by owner
            ScalarExpression keyExpr = keyMapping.newScalarExpression(stmt,stmt.getDefaultTableExpression());
            ScalarExpression keyVal = keyMapping.newLiteral(stmt, key);
            stmt.andCondition(keyExpr.eq(keyVal), true);
        }
        return stmt;
    }

    /**
     * Utility to validate the type of a value for storing in the Map.
     * @param value The value to check.
     **/
    protected void validateValueType(ClassLoaderResolver clr, Object value)
    {
        if (value == null)
        {
            throw new NullPointerException(LOCALISER.msg("SCO.Map.NullValuesNotAllowed"));
        }

        super.validateValueType(clr, value);
    }

    /**
     * Method to put an item in the Map.
     * @param sm State Manager for the map.
     * @param newKey The key to store the value against
     * @param newValue The value to store.
     * @return The value stored.
     **/
    public Object put(StateManager sm, final Object newKey, Object newValue)
    {
        validateKeyForWriting(sm, newKey);
        validateValueType(sm.getPersistenceManager().getClassLoaderResolver(), newValue);

        // Check if there's an existing value having the key.  
        // If so, remove that object from the map by nulling out its key field.
        Object oldValue = get(sm, newKey);
        if (oldValue != newValue)
        {
            if (oldValue != null)
            {
                removeValue(sm, newKey, oldValue);
            }

            PersistenceManager pm = sm.getPersistenceManager();
            PersistenceCapable pcNewValue = (PersistenceCapable)newValue;
            final Object newOwner = sm.getObject();

            if (pcNewValue.jdoIsPersistent())
            {
                /*
                 * The new value is already persistent.
                 *
                 * "Put" the new value in the map by updating its owner and key
                 * fields to the appropriate values.  This is done with the same
                 * methods the PC itself would use if the application code
                 * modified the fields.  It should result in no actual database
                 * activity if the fields were already set to the right values.
                 */
                if (pm != pcNewValue.jdoGetPersistenceManager())
                {
                    throw new JDOUserException(LOCALISER.msg("SCO.Map.WriteValudInvalidWithDifferentPM"),pcNewValue.jdoGetObjectId());
                }

                StateManager vsm = pm.findStateManager(pcNewValue);

                // Ensure the current owner field is loaded, and replace with new value
                if (hasOwnerFieldInValue)
                {
                    vsm.isLoaded(pcNewValue, ownerFieldNumber);
                    Object oldOwner = vsm.provideField(ownerFieldNumber);
                    vsm.setObjectField(pcNewValue, ownerFieldNumber, oldOwner, newOwner);
                }
                else
                {
                    updateValueFk(sm, pcNewValue, newOwner);
                }

                // Ensure the current key field is loaded, and replace with new value
                vsm.isLoaded(pcNewValue, keyFieldNumber);
                Object oldKey = vsm.provideField(keyFieldNumber);
                vsm.setObjectField(pcNewValue, keyFieldNumber, oldKey, newKey);
            }
            else
            {
                /*
                 * The new value is not yet persistent.
                 *
                 * Update its owner and key fields to the appropriate values and
                 * *then* make it persistent.  Making the changes before DB
                 * insertion avoids an unnecessary UPDATE allows the owner
                 * and/or key fields to be non-nullable.
                 */
                pm.makePersistent(newValue, new FieldValues()
                    {
                        public void fetchFields(StateManager vsm)
                        {
                            if (hasOwnerFieldInValue)
                            {
                                vsm.replaceField(ownerFieldNumber, newOwner);
                            }
                            vsm.replaceField(keyFieldNumber, newKey);
                        }
                    }
                );
                
                if (!hasOwnerFieldInValue)
                {
                    updateValueFk(sm, pcNewValue, newOwner);
                }
            }
        }

        return oldValue;
    }

    /**
     * Method to remove an item from the map.
     * @param sm State Manager for the map.
     * @param key Key of the item to remove.
     * @return The value that was removed.
     **/
    public Object remove(StateManager sm, Object key)
    {
        Object oldValue = get(sm, key);
        if (oldValue != null)
        {
            removeValue(sm, key, oldValue);
        }

        return oldValue;
    }

    /**
     * Utility to remove a value from the Map.
     * @param sm State Manager for the map.
     * @param key Key of the object
     * @param oldValue Value to remove
     **/
    private void removeValue(StateManager sm, Object key, Object oldValue)
    {
        PersistenceManager pm = sm.getPersistenceManager();

        // Null out the key and owner fields if they are nullable
        if (keyMapping.isNullable())
        {
            PersistenceCapable pcOldValue = (PersistenceCapable)oldValue;
            StateManager vsm = pm.findStateManager(pcOldValue);

            // Null the key field
            vsm.setObjectField(pcOldValue, keyFieldNumber, key, null);
            vsm.replaceField(keyFieldNumber, null);

            // Null the owner field
            if (hasOwnerFieldInValue)
            {
                Object oldOwner = vsm.provideField(ownerFieldNumber);
                vsm.setObjectField(pcOldValue, ownerFieldNumber, oldOwner, null);
                vsm.replaceField(ownerFieldNumber, null);
            }
            else
            {
                updateValueFk(sm, pcOldValue, null);
            }
        }
        // otherwise just delete the item
        else
        {
            pm.deletePersistent(oldValue);
        }
    }

    /**
     * Method to clear the map of all values.
     * @param sm State Manager for the map.
     **/
    public void clear(StateManager sm)
    {
        // TODO Fix this. Should not be retrieving objects only to remove them since they
        // may be cached in the org.jpox.sco object.
        Iterator iter = keySetStore().iterator(sm);
        while (iter.hasNext())
        {
            remove(sm, iter.next());
        }
    }

    /**
     * Utility to clear the key of a value from the Map.
     * If the key is non nullable, delete the value.
     * @param sm State Manager for the map.
     * @param key Key of the object
     * @param oldValue Value to remove
     **/
    public void clearKeyOfValue(StateManager sm, Object key, Object oldValue)
    {
        PersistenceManager pm = sm.getPersistenceManager();

        // Null out the key and owner fields if they are nullable
        if (keyMapping.isNullable())
        {
            PersistenceCapable pcOldValue = (PersistenceCapable)oldValue;
            StateManager vsm = pm.findStateManager(pcOldValue);

            // Null the key field
            vsm.setObjectField(pcOldValue, keyFieldNumber, key, null);
            vsm.replaceField(keyFieldNumber, null);
            vsm.makeDirty(keyFieldNumber);
        }
        // otherwise just delete the item
        else
        {
            pm.deletePersistent(oldValue);
        }
    }

    /**
     * Accessor for the keys in the Map.
     * @return The keys
     **/
    public synchronized SetStore keySetStore()
    {
        return (keySetStore == null) ? new MapKeySetStore(valueTable, ownerMapping, keyMapping) : keySetStore;
    }

    /**
     * Accessor for the values in the Map.
     * @return The values.
     **/
    public synchronized SetStore valueSetStore()
    {
        return (valueSetStore == null) ? new MapValueSetStore(valueTable,ownerMapping,this) : valueSetStore;
    }

    /**
     * Accessor for the map entries in the Map.
     * @return The map entries.
     **/
    public synchronized SetStore entrySetStore()
    {
        return (entrySetStore == null) ? new MapEntrySetStore(valueTable, ownerMapping, keyMapping, this) : entrySetStore;
    }

    // ------------------------------- Query Methods ---------------------------
 
     /**
     * Utility method to return a new QueryStatement for retrieval of the
     * elements of this Map.
     *
     * @param sm StateManager for this object
     * @param candidateClass Class for the value end of the link.
     * @return The QueryStatement
     **/
    public QueryStatement newQueryStatement(StateManager sm,
                                            String candidateClass)
    {
        if (!sm.getPersistenceManager().getClassLoaderResolver().isAssignableFrom(valueType, candidateClass))
        {
            throw new IncompatibleQueryElementTypeException(valueType,candidateClass);
        }

        ClassTable candidateTable=(ClassTable) storeMgr.getDatastoreClass(candidateClass);
        QueryStatement stmt=((DatabaseAdapter) dba).newQueryStatement(candidateTable);

        ScalarExpression ownerExpr=ownerMapping.newScalarExpression(stmt, stmt.getDefaultTableExpression());
        ScalarExpression ownerVal=ownerMapping.newLiteral(stmt, sm.getObject());
        stmt.andCondition(ownerExpr.eq(ownerVal));

        stmt.select(valueMapping);
        return stmt;
    }

    /**
     * Utility to create a join for keys to be used in a containsKey() query.
     * @param stmt The Query Statement 
     * @param ownerMapping Mapping for the owner
     * @param te Table Expression for the owner
     * @param mapRangeVar   The range variable for the "Map" table. 
     * @param filteredKeyType The Class Type for the filtered key
     * @param keyRangeVar  The SQL alias, or "range variable", to assign to the
     *                     expression or to the key table.
     * @return QueryColumnList with the columns from the key mapping 
     **/
    public ScalarExpression joinKeysTo(
                                QueryStatement stmt,
            					JavaTypeMapping ownerMapping,
            					TableExpression te,
                                SQLIdentifier mapRangeVar,
                                Class filteredKeyType,
                                SQLIdentifier keyRangeVar)
    {
        ClassLoaderResolver clr=storeMgr.getClassLoaderResolver();
        if (!clr.isAssignableFrom(keyType,filteredKeyType) &&
            !clr.isAssignableFrom(filteredKeyType,keyType))
        {
            throw new IncompatibleQueryElementTypeException(keyType, filteredKeyType.getName());
        }

        // Join the map table on the owner ID column
        stmt.newTableExpression(mapTable, mapRangeVar);

        ScalarExpression ownerExpr = ownerMapping.newScalarExpression(stmt,te);
        ScalarExpression ownerMapExpr = this.ownerMapping.newScalarExpression(stmt,stmt.getTableExpression(mapRangeVar));
        stmt.innerJoin(ownerExpr,ownerMapExpr,stmt.getTableExpression(mapRangeVar), true);
        
        // This depends on whether we have a separate key table or not
        if (!TypeManager.getTypeManager().isSupportedType(filteredKeyType.getName()))
        {
            // Join the key table on the key ID column (if theres a key table)
            ClassTable filteredKeyTable = (ClassTable) storeMgr.getDatastoreClass(filteredKeyType.getName());
            JavaTypeMapping keyTableID = filteredKeyTable.getIDMapping();
            stmt.newTableExpression(filteredKeyTable, keyRangeVar);

            ScalarExpression keyMapExpr = keyMapping.newScalarExpression(stmt,stmt.getTableExpression(mapRangeVar));
            ScalarExpression keyExpr = keyTableID.newScalarExpression(stmt, stmt.getTableExpression(keyRangeVar));
            stmt.innerJoin(keyExpr,keyMapExpr,stmt.getTableExpression(keyRangeVar), true);

            return keyTableID.newScalarExpression(stmt, stmt.getTableExpression(keyRangeVar));
        }
        else
        {
            // Key is embedded in the map table
            return keyMapping.newScalarExpression(stmt, stmt.getTableExpression(mapRangeVar));
        }
    }

    /**
     * Utility to create a join to be used in a containsValue() query.
     * @param stmt The Query Statement 
     * @param ownerMapping Mapping for the owner
     * @param te Table Expression for the owner
     * @param mapRangeVar   The range variable for the "Map" table. 
     * @param filteredValueType The Class Type for the filtered value
     * @param valueRangeVar The SQL alias, or "range variable", to assign to the
     *                      expression or to the value table.
     * @return QueryColumnList with the columns from the value mapping 
     **/
    public ScalarExpression joinValuesTo(
                                QueryStatement stmt,
            					JavaTypeMapping ownerMapping,
            					TableExpression te,
                                SQLIdentifier mapRangeVar,
                                Class filteredValueType,
                                SQLIdentifier valueRangeVar)
    {
        ClassLoaderResolver clr=storeMgr.getClassLoaderResolver();
        if (!clr.isAssignableFrom(valueType,filteredValueType) &&
            !clr.isAssignableFrom(filteredValueType,valueType))
        {
            throw new IncompatibleQueryElementTypeException(valueType,filteredValueType.getName());
        }

        // Join the value table on the owner ID column.
        ClassTable filteredValueTable = (ClassTable) storeMgr.getDatastoreClass(filteredValueType.getName());
        stmt.newTableExpression(filteredValueTable, valueRangeVar);

        ScalarExpression ownerExpr = ownerMapping.newScalarExpression(stmt,te);
        ScalarExpression ownerValueExpr = this.ownerMapping.newScalarExpression(stmt,stmt.getTableExpression(valueRangeVar));
        stmt.innerJoin(ownerExpr,ownerValueExpr, stmt.getTableExpression(valueRangeVar),true);

        JavaTypeMapping valueTableID = filteredValueTable.getIDMapping();
        return valueTableID.newScalarExpression(stmt,stmt.getTableExpression(valueRangeVar));
    }
    
    /**
     * Utility to create a join for keys and values to be used in ai
     * containsEntry() query.
     * @param stmt The Query Statement to apply the join
     * @param ownerMapping Mapping for the owner
     * @param te Table Expression for the owner
     * @param mapRangeVar  The SQL alias, or "range variable", to assign to the
     *                     expression or to the main table.
     * @param filteredKeyType The Class Type for the filtered key
     * @param filteredValueType The Class Type for the filtered value
     * @param keyRangeVar  The SQL alias, or "range variable", to assign to the
     *                     expression or to the key table.
     * @param valueRangeVar The SQL alias, or "range variable", to assign to the
     *                      expression or to the value table.
     * @return an array with 2 elements of QueryColumnList. The first element
     * contains the columns from the key mapping and the second element the
     * columns from the value mapping
     */
    public ScalarExpression[] joinKeysValuesTo(
                                QueryStatement stmt,
            					JavaTypeMapping ownerMapping,
            					TableExpression te,
                                SQLIdentifier mapRangeVar,
                                Class filteredKeyType,
                                Class filteredValueType,
                                SQLIdentifier keyRangeVar,
    							SQLIdentifier valueRangeVar)
    {
        ClassLoaderResolver clr=storeMgr.getClassLoaderResolver();
        if (!clr.isAssignableFrom(keyType,filteredKeyType) &&
            !clr.isAssignableFrom(filteredKeyType,keyType))
        {
            throw new IncompatibleQueryElementTypeException(keyType, filteredKeyType.getName());
        }

        // Join the map table on the owner ID column
        stmt.newTableExpression(mapTable, mapRangeVar);

        ScalarExpression ownerExpr = ownerMapping.newScalarExpression(stmt,te);
        ScalarExpression ownerMapExpr = this.ownerMapping.newScalarExpression(stmt,stmt.getTableExpression(mapRangeVar));
        stmt.innerJoin(ownerExpr,ownerMapExpr,stmt.getTableExpression(mapRangeVar), true);

        if (!clr.isAssignableFrom(valueType,filteredValueType) &&
            !clr.isAssignableFrom(filteredValueType,valueType))
        {
            throw new IncompatibleQueryElementTypeException(valueType, filteredValueType.getName());
        }

        ScalarExpression[] qclKeyValues = new ScalarExpression[2];

        // This depends on whether we have a separate key table or not
        if (!TypeManager.getTypeManager().isSupportedType(filteredKeyType.getName()))
        {
            // Join the key table on the key ID column (if theres a key table)
            ClassTable filteredKeyTable = (ClassTable) storeMgr.getDatastoreClass(filteredKeyType.getName());
            JavaTypeMapping keyTableID = filteredKeyTable.getIDMapping();
            stmt.newTableExpression(filteredKeyTable, keyRangeVar);

            ScalarExpression keyMapExpr = keyMapping.newScalarExpression(stmt,stmt.getTableExpression(mapRangeVar));
            ScalarExpression keyExpr = keyTableID.newScalarExpression(stmt, stmt.getTableExpression(keyRangeVar));
            stmt.innerJoin(keyExpr,keyMapExpr,stmt.getTableExpression(keyRangeVar), true);

            qclKeyValues[0] = keyTableID.newScalarExpression(stmt,stmt.getTableExpression(keyRangeVar));
        }
        else
        {
            // Key is embedded in the map table
            qclKeyValues[0] = keyMapping.newScalarExpression(stmt,stmt.getTableExpression(mapRangeVar));
        }

        // Join the value table on the owner ID column.
        ClassTable filteredValueTable = (ClassTable) storeMgr.getDatastoreClass(filteredValueType.getName());
        stmt.newTableExpression(filteredValueTable, valueRangeVar);

        ScalarExpression ownerValueExpr = this.ownerMapping.newScalarExpression(stmt,stmt.getTableExpression(valueRangeVar));
        stmt.innerJoin(ownerExpr,ownerValueExpr, stmt.getTableExpression(valueRangeVar),true);
        JavaTypeMapping valueTableID = filteredValueTable.getIDMapping();

        qclKeyValues[1] = valueTableID.newScalarExpression(stmt,stmt.getTableExpression(valueRangeVar));

        return qclKeyValues;
    }
}