/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - added comments
2003 Erik Bengtson - removed unused variable and import
2003 Erik Bengtson - fixed bug [832635] Imcompatiblities in casting
                    inherited classes in query
2003 Andy Jefferson - coding standards
2004 Andy Jefferson - updated retrieval of ColumnLists
2004 Andy Jefferson - addition of removeAll()
2004 Andy Jefferson - moved statements to AbstractSetStore
2004 Andy Jefferson - added support for inverse 1-N unidirectional
    ...
**********************************************************************/
package org.jpox.store.rdbms.scostore;

import java.lang.reflect.Field;
import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.SQLException;
import java.util.Collection;
import java.util.Iterator;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOUserException;
import javax.jdo.spi.PersistenceCapable;

import org.jpox.ClassLoaderResolver;
import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.metadata.CollectionMetaData;
import org.jpox.metadata.DiscriminatorStrategy;
import org.jpox.metadata.FieldMetaData;
import org.jpox.metadata.InheritanceStrategy;
import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.FieldValues;
import org.jpox.store.exceptions.IncompatibleQueryElementTypeException;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.TableExpression;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.Mappings;
import org.jpox.store.query.IteratorStatement;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.store.rdbms.adapter.DatabaseAdapter;
import org.jpox.store.rdbms.mapping.RDBMSMapping;
import org.jpox.store.rdbms.table.ClassTable;
import org.jpox.util.JPOXLogger;

/**
 * Representation of an Inverse Set as part of a relationship. This class is
 * used where you have a 1-N and the tables are not joined via a link table.
 * That is there is an owner table, and a collection table, and the collection
 * table has a column being the id of the owner table. This is in contrast to
 * NormalSetStore which represents 1-N relationships using a link table.
 * There are 2 possible uses here
 * <UL>
 * <LI><B>bidirectional</B> - where the owner has a Collection of elements, and
 * the element has an owner. In this case the element class will have an owner field
 * which can be updated directly</LI>
 * <LI><B>unidirectional</B> - where the owner has a Collection of elements, but
 * the element knows nothing about the owner. In this case the element class has no
 * owner field. In this case the column in the element table has to be updated
 * directly.</LI>
 * </UL>
 *
 * @version $Revision: 1.11 $
 **/
public class InverseSetStore extends AbstractSetStore
{
    private final boolean hasOwnerFieldInElement;
    private final int ownerFieldNumber;

    /** Statement for updating a foreign key in a 1-N unidirectional */
    private final String updateFkStmt;

    /**
     * Constructor for the relationship representation.
     *
     * @param fmd The MetaData for the field that this represents
     * @param storeMgr The RDBMSManager managing the associated datastore.
     **/
    public InverseSetStore(FieldMetaData fmd, RDBMSManager storeMgr)
    {
        this.storeMgr = storeMgr;
        dba = storeMgr.getDatabaseAdapter();

        Field field = fmd.getField();
        CollectionMetaData colmd = (CollectionMetaData)fmd.getContainer();
        if (colmd == null)
        {
            throw new JDOUserException("No collection metadata found in " + fmd);
        }

        // Load the element class
        elementType = colmd.getElementType();
        ClassLoaderResolver clr=storeMgr.getClassLoaderResolver();
        Class element_class = clr.classForName(elementType);

        // Check that the element class has MetaData
        emd = storeMgr.getMetaDataManager().getMetaDataForClass(element_class);
        if (emd == null)
        {
            throw new JDOUserException("The metadata for the class " + element_class.getName() + " was not found." +
                " This class is declared to be used as an element for the Set in the field: " + fmd.getFullFieldName());
        }

        elementTable = (ClassTable) storeMgr.getDatastoreClass(elementType);
        elementMapping = elementTable.getIDMapping();
        elementsAreEmbedded = false;
        if (emd != null)
        {
            if (emd.getInheritanceMetaData().getStrategyValue() == InheritanceStrategy.SUPERCLASS_TABLE &&
                emd.getDiscriminatorStrategy() != DiscriminatorStrategy.NONE)
            {
                elementDiscriminatorMapping = elementTable.getDiscriminatorMapping();
            }
        }

        // Get the field in the element table (if any)
        String mappedByFieldName = fmd.getMappedBy();
        if (mappedByFieldName != null)
        {
            // Inverse 1-N bidirectional
            // The element class has a field for the owner.
            FieldMetaData eofmd = emd.getField(mappedByFieldName);
            if (eofmd == null)
            {
                throw new JDOUserException("The field (" + mappedByFieldName + 
                    ") marked as \"mapped-by\" on the owner class was not found in the class " + element_class.getName());
            }

            // Check that the type of the element "mapped-by" field is consistent with the owner type
            if (!clr.isAssignableFrom(eofmd.getField().getType(), fmd.getClassMetaData().getFullClassName()))
            {
                throw new JDOUserException("Can't assign class " + fmd.getClassMetaData().getFullClassName() + 
                    " to the field " + eofmd.getFullFieldName() + " declared as " + eofmd.getField().getType().getName());
            }

            String ownerFieldName = eofmd.getName();
            ownerFieldNumber = emd.getFieldNumberAbsolute(ownerFieldName);
            if (!elementTable.isFieldPersistent(ownerFieldNumber))
            {
                throw new JDOUserException("Non-persistent field type for collection owner field: " + 
                    field.getDeclaringClass().getName() + "." + field.getName());
            }

            // Find field mapping for owner field
            JavaTypeMapping om = elementTable.getFieldMapping(emd.getManagedFieldAbsolute(ownerFieldNumber));
            if (isEmbeddedMapping(om))
            {
                throw new JDOUserException("Invalid field type for collection owner field: " + 
                    field.getDeclaringClass().getName() + "." + field.getName());
            }
            ownerMapping = om;
            hasOwnerFieldInElement = true;
        }
        else
        {
            // Inverse 1-N unidirectional
            // The element class knows nothing about the owner
            JavaTypeMapping om = elementTable.getExternalFkMapping(fmd);
            if (om == null)
            {
                throw new JDOUserException("Class " + fmd.getClassMetaData().getFullClassName() + " has collection field " + fmd.getName() + 
                    " and this has no mapping in the table for the element class " + elementType);
            }
            ownerMapping = om;
            ownerFieldNumber = -1;
            hasOwnerFieldInElement = false;
        }

        setTable = elementTable;
        setName = "inverseSet";

        // Generate the statements
        initialiseStatements();
        addStmt = null;
        removeStmt = null;
        clearStmt = null;
        updateFkStmt = getUpdateFkStmt();
    }

    /**
     * Generate statement for updating a Foreign Key in an inverse 1-N.
     * <PRE>
     * UPDATE ELEMENTTABLE SET FK_COL_1 = ?, FK_COL_2 = ? 
     * WHERE ELEMENT_ID = ?
     * </PRE>
     * @return Statement for updating the FK in an inverse 1-N
     */
    private String getUpdateFkStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("UPDATE ");
        stmt.append(setTable.toString());
        stmt.append(" SET ");
        for (int i=0; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(",");
            }
            stmt.append(((Column)ownerMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        stmt.append(" WHERE ");
        for (int i=0; i<elementMapping.getNumberOfDatastoreFields(); i++)
        {
            if (i > 0)
            {
                stmt.append(" AND ");
            }
            stmt.append(((Column)elementMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)elementMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }

        return stmt.toString();
    }

    /**
     * Utility to update a foreign-key in the element in the case of
     * a unidirectional 1-N relationship.
     * @param sm StateManager for the owner
     * @param element The element to update
     * @param owner The owner object to set in the FK
     * @return Whether it was performed successfully
     */
    private boolean updateElementFk(StateManager sm, Object element, Object owner)
    {
        if (element == null)
        {
            return false;
        }
        validateElementForWriting(sm, element);

        boolean retval;
        PersistenceManager pm = sm.getPersistenceManager();
        try
        {
            Connection conn = pm.getConnection(false);
            try
            {
                PreparedStatement ps = conn.prepareStatement(updateFkStmt);
                try
                {
                    if (owner == null)
                    {
                        ownerMapping.setObject(pm, ps, Mappings.getParametersIndex(1,ownerMapping), null);
                    }
                    else
                    {
                        ownerMapping.setObject(pm, ps, Mappings.getParametersIndex(1,ownerMapping), sm.getObject());
                    }
                    elementMapping.setObject(pm, ps, Mappings.getParametersIndex(ownerMapping.getNumberOfDatastoreFields()+1,elementMapping), element);

                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(updateFkStmt);
                    }
                    long startTime = System.currentTimeMillis();
                    ps.executeUpdate();
                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(STORE_LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                    }

                    retval = true;
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                pm.releaseConnection(conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("SCO.UpdateFkRequestFailed",updateFkStmt),e);
        }

        return retval;
    }

    /**
     * Method to add an object to the relationship at the collection end.
     * @param sm StateManager of the Set
     * @param element Element to be added
     * @return Success indicator
     **/
    public boolean add(StateManager sm, Object element)
    {
        if (element == null)
        {
            throw new JDOUserException(LOCALISER.msg("SCO.Set.NullElementsNotAllowed"));
        }
        if (!sm.getPersistenceManager().getClassLoaderResolver().isAssignableFrom(elementType,element.getClass()))
        {
            throw new ClassCastException(LOCALISER.msg("SCO.Set.ElementIsInvalid",element.getClass().getName(),elementType));
        }

        PersistenceManager pm = sm.getPersistenceManager();
        PersistenceCapable pcElement = (PersistenceCapable) element;
        final Object newOwner = sm.getObject();

        if (pcElement.jdoIsPersistent())
        {
            /*
             * The new element is already persistent.
             *
             * "Add" the new value to the set by updating its owner field to the
             * appropriate value.  This is done with the same methods the PC
             * itself would use if the application code modified the field.  It
             * should result in no actual database activity if the field was
             * already set to the right value.
             */
            if (pm != pcElement.jdoGetPersistenceManager())
            {
                throw new JDOUserException(LOCALISER.msg("SCO.Set.WriteInvalidWithDifferentPM"),pcElement.jdoGetObjectId());
            }
            StateManager esm = pm.findStateManager(pcElement);

            // Bidirectional 1-N, so update the owner
            if (hasOwnerFieldInElement)
            {
                // Ensure the current owner value is loaded.
                esm.isLoaded(pcElement, ownerFieldNumber);

                // Get the current value.
                Object oldOwner = esm.provideField(ownerFieldNumber);

                // Replace with new value.
                esm.setObjectField(pcElement, ownerFieldNumber, oldOwner, newOwner);

                return oldOwner != newOwner;
            }
            // Unidirectional 1-N, so set the owner
            else
            {
                return updateElementFk(sm, pcElement, newOwner);
            }
        }
        else
        {
            /*
             * The new element is not yet persistent.
             *
             * Update its owner field to the appropriate value and *then* make
             * it persistent.  Making the change before DB insertion avoids an
             * unnecessary UPDATE and allows the owner field to be non-nullable.
             */
            pm.makePersistent(element, new FieldValues()
            {
                public void fetchFields(StateManager esm)
                {
                    if (hasOwnerFieldInElement)
                    {
                        esm.replaceField(ownerFieldNumber, newOwner);
                    }
                }
            });

            // Unidirectional 1-N, so set the owner
            // TODO Consider doing this as part of makePersistent to reduce number of DB calls
            if (!hasOwnerFieldInElement)
            {
                updateElementFk(sm, element, newOwner);
            }

            return true;
        }
    }
 
    /**
     * Method to add a collection of object to the relationship at the
     * collection end.
     * @param sm StateManager of the Set
     * @param elements Elements to be added
     * @return Success indicator
     **/
    public boolean addAll(StateManager sm, Collection elements)
    {
        if (elements == null || elements.size() == 0)
        {
            return false;
        }

        boolean success = false;

        Iterator iter = elements.iterator();
        while (iter.hasNext())
        {
            if (add(sm, iter.next()))
            {
                success = true;
            }
        }

        return success;
    }

    /**
     * Method to remove the link to the collection object specified.
     * Depending on the column characteristics in the collection table, the id
     * of the owner field may be NULLed, or the record may be deleted
     * completely (as per cascade-delete in EJB).
     *
     * @param sm The StateManager of the Set
     * @param element The element of the collection to be deleted.
     * @return A success indicator.
     **/
    public boolean remove(StateManager sm, Object element)
    {
        if (element == null)
        {
            return false;
        }
        if (!validateElementForReading(sm, element))
        {
            return false;
        }

        PersistenceManager pm = sm.getPersistenceManager();
        PersistenceCapable pcElement = (PersistenceCapable) element;
        StateManager esm = pm.findStateManager(pcElement);

        Object oldOwner = null;
        if (hasOwnerFieldInElement)
        {
            esm.isLoaded(pcElement, ownerFieldNumber);
            oldOwner = esm.provideField(ownerFieldNumber);
        }
        else
        {
            // TODO Check if the element is managed by a different owner now
        }

        // Owner of the element has been changed
        if (hasOwnerFieldInElement && oldOwner != sm.getObject())
        {
            return false;
        }

        // If the column is nullable, null it out
        if (ownerMapping.isNullable())
        {
            // Bidirectional 1-N, so nullify the owner
            if (hasOwnerFieldInElement)
            {
                esm.setObjectField(pcElement, ownerFieldNumber, oldOwner, null);
            }
            // Unidirectional 1-N, so nullify the owner
            else
            {
                updateElementFk(sm, pcElement, null);
            }
        }
        // otherwise remove the element
        else
        {
            // deletePersistent() is used to take advantage of dependent fields
            pm.deletePersistent(element);
        }

        return true;
    }

    /**
     * Method to remove the links to a collection of elements specified.
     * Depending on the column characteristics in the collection table, the id
     * of the owner fields may be NULLed, or the records may be deleted
     * completely (as per cascade-delete in EJB).
     *
     * @param sm The StateManager of the Set
     * @param elements The elements of the collection to be deleted.
     * @return A success indicator.
     **/
    public boolean removeAll(StateManager sm, Collection elements)
    {
        if (elements == null || elements.size() == 0)
        {
            return false;
        }

        // Check the first element for whether we can null the column or
        // whether we have to delete
        boolean success = true;

        Iterator iter=elements.iterator();
        while (iter.hasNext())
        {
            if (remove(sm, iter.next()))
            {
                success = false;
            }
        }

        return success;
    }

    /**
     * Method to allow the Set relationship to be cleared out. This goes
     * through all collection elements of the Set and removes them from the
     * relationship.
     * @param sm StateManager of the Set
     **/
    public void clear(StateManager sm)
    {
        boolean success = true;

        // This will retrieve the elements and then remove them.
        // Really should be removed and the clear() method in the org.jpox.sco classes
        // should call the removeAll() method.
        Iterator iter = iterator(sm);
        while (iter.hasNext())
        {
            if (!remove(sm, iter.next()))
            {
                success = false;
            }
        }

        if (!success)
        {
            JPOXLogger.RDBMS.error("Error clearing elements of field " + sm.getExternalObjectId(sm.getObject()));
        }
    }

    /**
     * Accessor for a QueryStatement to retrieve the elements of the set.
     * @param sm StateManager of the Set.
     * @return The QueryStatement to retrieve the elements
     */
    protected QueryStatement getIteratorStatement(StateManager sm)
    {
        if (elementTable == null)
        {
            return null;
        }
        final ClassLoaderResolver clr=storeMgr.getClassLoaderResolver();
        QueryStatement stmt = new IteratorStatement(
                new IteratorStatement.AssociationEnd() {
                    public JavaTypeMapping getMapping()
                    {
                        return elementMapping;
                    }
                    public Class getType()
                    {
                        return clr.classForName(elementType);
                    }
                    public DatastoreContainerObject getDatastoreContainerObject()
                    {
                        return elementTable;
                    }
                    public boolean useJoin()
                    {
                        return false;
                    }
                },
                clr.classForName(this.elementType),
                this.storeMgr,
                this.dba,
                true).getQueryStatement();

        // Apply condition on join-table owner field to filter by owner
        ScalarExpression ownerExpr = ownerMapping.newScalarExpression(stmt, stmt.getDefaultTableExpression());
        ScalarExpression ownerVal = ownerMapping.newLiteral(stmt, sm.getObject());
        stmt.andCondition(ownerExpr.eq(ownerVal), true);

        return stmt;
    }

    // --------------------------- Query Methods -------------------------------

    /**
     * Utility method to return a new QueryStatement for retrieval of the
     * elements of this Set.
     *
     * @param sm StateManager for this object
     * @param candidateClass Class for the element end of the link.
     * @return The QueryStatement
     **/
    public QueryStatement newQueryStatement(StateManager sm, String candidateClass)
    {
        if (!sm.getPersistenceManager().getClassLoaderResolver().isAssignableFrom(elementType,candidateClass))
        {
            throw new IncompatibleQueryElementTypeException(elementType, candidateClass);
        }

        ClassTable candidateTable = (ClassTable) storeMgr.getDatastoreClass(candidateClass);
        QueryStatement stmt = ((DatabaseAdapter) dba).newQueryStatement(candidateTable);

        ScalarExpression ownerExpr = ownerMapping.newScalarExpression(stmt, stmt.getDefaultTableExpression());
        ScalarExpression ownerVal = ownerMapping.newLiteral(stmt, sm.getObject());
        stmt.andCondition(ownerExpr.eq(ownerVal));

        stmt.select(elementMapping);

        return stmt;
    }

    /**
     * Utility for use in building a query, joining the element table and the
     * owner table.
     *
     * @param stmt The QueryStatement
     * @param ownerMapping Mapping for the owner
     * @param te Table Expression for the owner
     * @param setRangeVar The range variable for the "element" table
     * @param filteredElementType ?
     * @param elementRangeVar ?
     * @return A subquery for the provided query that joins the element table
     * to the owner table.
     * */
    public ScalarExpression joinElementsTo(
        QueryStatement stmt,
		JavaTypeMapping ownerMapping,
		TableExpression te,
		DatastoreIdentifier setRangeVar,
        Class filteredElementType,
        TableExpression elmte,
        DatastoreIdentifier elementRangeVar)
    {
        ClassLoaderResolver clr=storeMgr.getClassLoaderResolver();
        if (!clr.isAssignableFrom(elementType,filteredElementType) &&
            !clr.isAssignableFrom(filteredElementType,elementType))
        {
            throw new IncompatibleQueryElementTypeException(elementType, filteredElementType.getName());
        }

        // Join the element table on the owner ID column.
        ClassTable filteredElementTable = (ClassTable) storeMgr.getDatastoreClass(filteredElementType.getName());
        stmt.newTableExpression(filteredElementTable, elementRangeVar);

        ScalarExpression ownerExpr = ownerMapping.newScalarExpression(stmt,te);
        ScalarExpression ownerSetExpr = this.ownerMapping.newScalarExpression(stmt,stmt.getTableExpression(elementRangeVar));
        stmt.innerJoin(ownerExpr,ownerSetExpr,stmt.getTableExpression(elementRangeVar), true);

        JavaTypeMapping elementTableID = filteredElementTable.getIDMapping();
        return elementTableID.newScalarExpression(stmt,stmt.getTableExpression(elementRangeVar));
    }
}