/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others. All rights reserved.
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - coding standards
    ...
**********************************************************************/
package org.jpox.store.rdbms.scostore;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.Map.Entry;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOUnsupportedOptionException;

import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.store.DatastoreAdapter;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.StoreManager;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.TableExpression;
import org.jpox.store.mapping.InterfaceMapping;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.Mappings;
import org.jpox.store.mapping.OIDMapping;
import org.jpox.store.query.Query;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.mapping.RDBMSMapping;
import org.jpox.store.rdbms.sqlidentifier.SQLIdentifier;
import org.jpox.store.rdbms.table.ClassTable;
import org.jpox.store.rdbms.table.MapTable;
import org.jpox.store.rdbms.table.Table;
import org.jpox.store.scostore.MapStore;
import org.jpox.store.scostore.SetStore;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;
import org.jpox.util.SQLWarnings;

/**
 * Representation of backing store for a MapEntry Set.
 *
 * @version $Revision: 1.4 $
 **/
public class MapEntrySetStore implements SetStore
{
    protected static final Localiser LOCALISER=Localiser.getInstance("org.jpox.store.rdbms.Localisation");

    protected Table setTable;
    protected MapStore mapStore;
    protected StoreManager storeMgr;
    protected DatastoreAdapter dba;
    protected JavaTypeMapping ownerMapping;
    protected JavaTypeMapping keyMapping;
    protected JavaTypeMapping valueMapping;
    protected String keyType;
    protected String valueType;
    protected boolean keysAreEmbedded;
    protected boolean valuesAreEmbedded;
    protected String iteratorStmt;
    protected String sizeStmt;


    public MapEntrySetStore(MapTable mapTable, MapStore mapStore)
    {
        this.setTable = mapTable;
        this.mapStore = mapStore;

        storeMgr = mapTable.getStoreManager();
        dba      = storeMgr.getDatastoreAdapter();

        keyType   = mapStore.getKeyType();
        valueType = mapStore.getValueType();

        ownerMapping = mapTable.getOwnerMapping();
        keyMapping   = mapTable.getKeyMapping();
        valueMapping = mapTable.getValueMapping();

        initialize();
    }

    public MapEntrySetStore(ClassTable valueTable,
                            JavaTypeMapping ownerMapping,
                            JavaTypeMapping keyMapping,
                            MapStore mapStore)
    {
        this.setTable = valueTable;
        this.mapStore = mapStore;

        storeMgr = valueTable.getStoreManager();
        dba      = storeMgr.getDatastoreAdapter();

        keyType   = mapStore.getKeyType();
        valueType = mapStore.getValueType();

        this.ownerMapping = ownerMapping;
        this.keyMapping   = keyMapping;
        this.valueMapping = valueTable.getIDMapping();

        initialize();
    }

    private void initialize()
    {
        keysAreEmbedded   = isEmbeddedMapping(keyMapping);
        valuesAreEmbedded = isEmbeddedMapping(valueMapping);

        iteratorStmt = getIteratorStmt();
        sizeStmt     = getSizeStmt();
    }

	/**
     * Method to return a size statement. 
     * <PRE> 
	 * SELECT COUNT(*) FROM SETTABLE WHERE OWNERCOL=?
     * </PRE> 
	 * @return The size statement
	 */
	private String getSizeStmt()
	{
		StringBuffer stmt = new StringBuffer();
		stmt.append("SELECT COUNT(*) FROM ");
		stmt.append(setTable.toString());
		stmt.append(" WHERE ");
		stmt.append(((Column)ownerMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
		stmt.append(" = ");
		stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(0)).getUpdateInputParameter());
		for( int i=1; i<ownerMapping.getNumberOfDatastoreFields(); i++)
		{
			stmt.append(" AND ");
			stmt.append(((Column)ownerMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
			stmt.append(" = ");
			stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
		}
		return stmt.toString();
	}

	/**
     * Method to return a statement to get the objects via an iterator.
     * <PRE> 
	 * SELECT KEYCOL, VALUECOL FROM SETTABLE WHERE OWNERCOL=?
     * </PRE> 
	 * @return The iterator statement
	 */
	private String getIteratorStmt()
	{
		StringBuffer stmt = new StringBuffer();
		stmt.append("SELECT ");
		stmt.append(((Column)keyMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
		for( int i=1; i<keyMapping.getNumberOfDatastoreFields(); i++)
		{
			stmt.append(",");
			stmt.append(((Column)keyMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
		}
		for( int i=0; i<valueMapping.getNumberOfDatastoreFields(); i++)
		{
			stmt.append(",");
			stmt.append(((Column)valueMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
		}		
		stmt.append(" FROM ");
		stmt.append(setTable.toString());
		stmt.append(" WHERE ");
		stmt.append(((Column)ownerMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
		stmt.append(" = ");
		stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(0)).getUpdateInputParameter());
		for( int i=1; i<ownerMapping.getNumberOfDatastoreFields(); i++)
		{
			stmt.append(" AND ");
			stmt.append(((Column)ownerMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
			stmt.append(" = ");
			stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
		}
		return stmt.toString();
	}

    /**
     * Accessor for the Store Manager.
     * @return The Store Manager.
     **/
    public StoreManager getStoreManager()
    {
        return storeMgr;
    }

    /**
     * Accessor for the element type.
     * @return The element type.
     **/
    public String getElementType()
    {
        return EntryImpl.class.getName();
    }

    public JavaTypeMapping getOwnerMapping()
    {
        return ownerMapping;
    }

    protected boolean validateElementType(Object element)
    {
        return element instanceof Entry;
    }

    public Iterator iterator(StateManager sm)
    {
        Iterator i;
        PersistenceManager pm = sm.getPersistenceManager();

        try
        {
            Connection conn = pm.getConnection(false);

            try
            {
                PreparedStatement ps = conn.prepareStatement(iteratorStmt);

                try
                {
                    ownerMapping.setObject(pm, ps, Mappings.getParametersIndex(1,ownerMapping), sm.getObject());
                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(iteratorStmt);
                    }

                    long startTime = System.currentTimeMillis();
                    ResultSet rs = ps.executeQuery();
                    try
                    {
                        if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                        {
                            JPOXLogger.RDBMS_SQL.debug(LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                        }

                        i = new SetIterator(sm, rs);
                    }
                    finally
                    {
                        rs.close();
                    }
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                pm.releaseConnection(conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException("Iteration request failed: " + iteratorStmt, e);
        }

        return i;
    }

    /**
     * Inner class representing an iterator for the Set.
     **/
    private class SetIterator implements Iterator
    {
        private final StateManager sm;
        private final PersistenceManager pm;
        private final Iterator delegate;
        private Entry lastElement = null;

        public SetIterator(StateManager sm, ResultSet rs) throws SQLException
        {
            this.sm = sm;
            this.pm = sm.getPersistenceManager();

            ArrayList results = new ArrayList();

            while (rs.next())
            {
                results.add(new EntryImpl(sm, keyMapping.getObject(pm, rs, Mappings.getParametersIndex(1,keyMapping)), valueMapping.getObject(pm, rs, Mappings.getParametersIndex(keyMapping.getNumberOfDatastoreFields()+1,valueMapping))));
            }

            delegate = results.iterator();
        }

        public boolean hasNext()
        {
            return delegate.hasNext();
        }

        public Object next()
        {
            lastElement = (Entry)delegate.next();

            return lastElement;
        }

        public synchronized void remove()
        {
            if (lastElement == null)
            {
                throw new IllegalStateException("No entry to remove");
            }

            mapStore.remove(sm, lastElement.getKey());
            delegate.remove();

            lastElement = null;
        }
    }

    /**
     * Accessor for the size of the Map.
     * @param sm The state manager
     * @return The size
     **/
    public int size(StateManager sm)
    {
        int numRows;
        PersistenceManager pm = sm.getPersistenceManager();

        try
        {
            Connection conn = pm.getConnection(false);

            try
            {
                PreparedStatement ps = conn.prepareStatement(sizeStmt);

                try
                {
                    ownerMapping.setObject(pm, ps, Mappings.getParametersIndex(1,ownerMapping), sm.getObject());
                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(sizeStmt);
                    }

                    long startTime = System.currentTimeMillis();
                    ResultSet rs = ps.executeQuery();
                    try
                    {
                        if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                        {
                            JPOXLogger.RDBMS_SQL.debug(LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                        }

                        if (!rs.next())
                        {
                            throw new JDODataStoreException("Size request returned no result row: " + sizeStmt);
                        }

                        numRows = rs.getInt(1);

                        SQLWarnings.log(rs);
                    }
                    finally
                    {
                        rs.close();
                    }
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                pm.releaseConnection(conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException("Size request failed: " + sizeStmt, e);
        }

        return numRows;
    }


    public boolean contains(StateManager sm, Object element)
    {
        if (!validateElementType(element))
        {
            return false;
        }

        Entry entry = (Entry)element;

        return mapStore.containsKey(sm, entry.getKey());
    }


    public boolean add(StateManager sm, Object element)
    {
        throw new UnsupportedOperationException("Cannot add to a map through its entry set");
    }


    public boolean addAll(StateManager sm, Collection elements)
    {
        throw new UnsupportedOperationException("Cannot add to a map through its entry set");
    }


    public boolean remove(StateManager sm, Object element)
    {
        if (!validateElementType(element))
        {
            return false;
        }

        Entry entry = (Entry)element;

        Object removed = mapStore.remove(sm, entry.getKey());

        /*
         * NOTE: this may not return an accurate result if a null value is being
         * removed.
         */
        return removed == null ? entry.getValue() == null : removed.equals(entry.getValue());
    }

    public boolean removeAll(StateManager sm, Collection elements)
    {
        if (elements == null || elements.size() == 0)
        {
            return false;
        }

        Iterator iter=elements.iterator();
        boolean modified=false;
        while (iter.hasNext())
        {
            Object element=iter.next();
            Entry entry = (Entry)element;

            Object removed = mapStore.remove(sm, entry.getKey());

            /*
             * NOTE: this may not return an accurate result if a null value is
             * being removed.
             */
            modified = removed == null ? entry.getValue() == null : removed.equals(entry.getValue());
        }

        return modified;
    }


    public void clear(StateManager sm)
    {
        mapStore.clear(sm);
    }

    /**
     * Inner class representing the entry.
     **/
    private class EntryImpl implements Entry
    {
        private final StateManager sm;
        private final Object key;
        private Object value;

        public EntryImpl(StateManager sm, Object key, Object value)
        {
            this.sm = sm;
            this.key = key;
            this.value = value;
        }

        public int hashCode()
        {
            return (key == null   ? 0 : key.hashCode()) ^
                   (value == null ? 0 : value.hashCode());
        }

        public boolean equals(Object o)
        {
            if (o == this)
            {
                return true;
            }

            if (!(o instanceof Entry))
            {
                return false;
            }

            Entry e = (Entry)o;

            return (key == null ? e.getKey() == null : key.equals(e.getKey())) &&
                   (value == null ? e.getValue() == null : value.equals(e.getValue()));
        }

        /**
         * Accessor for the Key.
         * @return The Key.
         **/
        public Object getKey()
        {
            return key;
        }

        /**
         * Accessor for the Value.
         * @return The Value.
         **/
        public Object getValue()
        {
            return value;
        }

        /**
         * Mutator for the Value.
         * @param value The Value.
         **/
        public Object setValue(Object value)
        {
            return mapStore.put(sm, key, value);
        }
    }

    // ----------------------------- Query Methods -----------------------------
 
    /**
     * Method to return a new QueryStatement.
     * @param sm StateManager for the collection.
     * @param candidateClass The class to query
     * @return The new QueryStatement.
     */
    public QueryStatement newQueryStatement(StateManager sm, String candidateClass)
    {
        throw new JDOUnsupportedOptionException("Cannot query sets obtained by Map.entrySet()");
    }

    /**
     * Method to return a ResultObjectFactory for the store.
     */
    public Query.ResultObjectFactory newResultObjectFactory(StateManager sm, QueryStatement stmt, boolean ignoreCache)
    {
        throw new JDOUnsupportedOptionException("Cannot query sets obtained by Map.entrySet()");
    }

    /**
     * Method to add an existence check to a Query.
     */
    public QueryStatement getExistsSubquery(JavaTypeMapping mapping,
            							    TableExpression te,
                                            SQLIdentifier setRangeVar)
    {
        throw new JDOUnsupportedOptionException("Cannot query sets obtained by Map.entrySet()");
    }

    /**
     * Method to join elements of the collection in a Query.
     */
    public ScalarExpression
        joinElementsTo(QueryStatement stmt,
                	   JavaTypeMapping ownerMapping,
 		   			   TableExpression te,
 		   			   DatastoreIdentifier setRangeVar,
                       Class filteredElementType,
                       TableExpression elmte,
                       DatastoreIdentifier elementRangeVar)
    {
        throw new JDOUnsupportedOptionException("Cannot query sets obtained by Map.entrySet()");
    }
    
    /**
     * Check if the mapping correspond to a non pc object or embedded field
     * @param mapping the mapping
     * @return true if the field is embedded into one column
     */
    protected boolean isEmbeddedMapping(JavaTypeMapping mapping)
    {
        return !InterfaceMapping.class.isAssignableFrom(mapping.getClass()) &&
        	   !OIDMapping.class.isAssignableFrom(mapping.getClass());
    }    
}

