/**********************************************************************
Copyright (c) 2002 Kelly Grizzle (TJDO) and others. All rights reserved.
This program and the accompanying materials are made available under
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2002 Mike Martin (TJDO)
2003 Erik Bengtson - removed unused variable and import
2003 Andy Jefferson - coding standards
2004 Andy Jefferson - moved statements to AbstractSetStore
    ...
**********************************************************************/
package org.jpox.store.rdbms.scostore;

import javax.jdo.JDOUnsupportedOptionException;

import org.jpox.ClassLoaderResolver;
import org.jpox.StateManager;
import org.jpox.TypeManager;
import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.TableExpression;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.query.IteratorStatement;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.rdbms.sqlidentifier.TableIdentifier;
import org.jpox.store.rdbms.table.ClassTable;
import org.jpox.store.rdbms.table.MapTable;
import org.jpox.store.rdbms.table.SCOTable;

/**
 * Representation of the backing store for a MapKey Set.
 * The keys for a Map can be stored in several ways. There are four
 * possibilities
 * <UL>
 * <LI><B>Primitive key, Normal Map</B> - The container has its FCO table,
 * and there is a join table (SCO) that will contain the key object.</LI>
 * <LI><B>Primitive key, Inverse Map</B> - The container has its FCO table,
 * and the key is stored in the FCO table of the element</LI>
 * <LI><B>Non-primitive key, Normal Map</B> - The container has its FCO table,
 * and the key has its own FCO table. There is an SCO join table, and this join
 * table has a FK (..._ID_KID) to the id of the key FCO table.</LI>
 * <LI><B>Non-primitive key, Inverse Map</B> - The container has its FCO table,
 * and the key has its own FCO table. The value is stored in a table, and this
 * table has a FK (KEY_...._ID_OID) to the key FCO table.</LI>
 * </UL>
 *
 * @version $Revision: 1.6 $
 **/
public class MapKeySetStore extends AbstractSetStore
{
    /**
     * Constructor.
     * @param mapTable The table for the map (join table)
     */
    public MapKeySetStore(MapTable mapTable)
    {
        setTable = mapTable;
        setName  = "keys";
        storeMgr = mapTable.getStoreManager();
        dba      = storeMgr.getDatastoreAdapter();

        ownerMapping   = mapTable.getOwnerMapping();
        elementMapping = mapTable.getKeyMapping();

        initialize();
    }

    /**
     * Constructor.
     * @param valueTable The table for the "value" of the map
     * @param ownerMapping The mapping for this table to the owner table
     * @param keyMapping The mapping for this table to the key table
     */
    public MapKeySetStore(ClassTable valueTable,
            				JavaTypeMapping ownerMapping,
            				JavaTypeMapping keyMapping)
    {
        setTable = valueTable;
        setName  = "keys";
        storeMgr = valueTable.getStoreManager();
        dba      = storeMgr.getDatastoreAdapter();

        this.ownerMapping   = ownerMapping;
        this.elementMapping = keyMapping;

        initialize();

        /* Set is read-only. */
        removeStmt = null;
        clearStmt  = null;
    }

    /**
     * Initialisation method.
     */
    private void initialize()
    {
        elementType = elementMapping.getType().getName();
        elementsAreEmbedded = isEmbeddedMapping(elementMapping);

        // Generate the statements
        initialiseStatements();

        // Load the element class
        ClassLoaderResolver clr=storeMgr.getClassLoaderResolver();
        Class element_class=clr.classForName(elementType);

        if (element_class.isInterface())
        {
            emd = storeMgr.getMetaDataManager().getMetaDataForInterface(element_class,null);
        }
        else
        {
            emd = storeMgr.getMetaDataManager().getMetaDataForClass(element_class);
        }
        if (emd != null)
        {
            elementType = emd.getFullClassName();
            elementTable = (ClassTable) storeMgr.getDatastoreClass(elementType);
        }
    }

    /**
     * Method to create an iterator for the map keys.
     * @param sm StateManager for the map
     */
    protected QueryStatement getIteratorStatement(StateManager sm)
    {
        final ClassLoaderResolver clr=storeMgr.getClassLoaderResolver();
        QueryStatement stmt = new IteratorStatement(
                new IteratorStatement.AssociationEnd()
                {
                    public JavaTypeMapping getMapping()
                    {
                        return elementMapping;
                    }
                    public Class getType()
                    {
                        return clr.classForName(elementType);
                    }
                    public DatastoreContainerObject getDatastoreContainerObject()
                    {
                        return setTable;
                    }
                    public boolean useJoin()
                    {
                        return true;
                    }
                },
                clr.classForName(this.elementType),
                this.storeMgr,
                this.dba,
                true).getQueryStatement();

        // SCO cases/primitiveKeys just provide the owner id on their table
        if (setTable instanceof SCOTable || TypeManager.getTypeManager().isSupportedType(elementType))
        {
            ScalarExpression ownerExpr = ownerMapping.newScalarExpression(stmt, stmt.getDefaultTableExpression());
            ScalarExpression ownerVal = ownerMapping.newLiteral(stmt, sm.getObject());

            stmt.andCondition(ownerExpr.eq(ownerVal), true);
        }
        // FCO cases/non-primitiveKeys provide the owner id on the value table
        else
        {
            ScalarExpression ownerExpr = ownerMapping.newScalarExpression(stmt, stmt.getTableExpression(new TableIdentifier(dba, "VALUE")));
            ScalarExpression ownerVal = ownerMapping.newLiteral(stmt, sm.getObject());

            stmt.andCondition(ownerExpr.eq(ownerVal), true);
        }

        return stmt;
    }

    /**
     * Method to add an element. Overridden because we want to prevent it.
     * @param sm State Manager of collection.
     * @param element Element to add.
     * @return Whether it was successful
     **/
    public boolean add(StateManager sm, Object element)
    {
        throw new UnsupportedOperationException("Cannot add to a map through its key set");
    }

    /**
     * Method to remove an element. Overridden because we want to prevent it.
     * @param sm State Manager of collection.
     * @param element Element to remove.
     * @return Whether it was successful
     **/
    public boolean remove(StateManager sm, Object element)
    {
        if (removeStmt == null)
        {
            throw new UnsupportedOperationException("Cannot remove from an inverse map through its key set");
        }

        return super.remove(sm, element);
    }

    /**
     * Method to clear the collection. Overridden because we want to prevent it.
     * @param sm State Manager of collection.
     **/
    public void clear(StateManager sm)
    {
        if (clearStmt == null)
        {
            throw new UnsupportedOperationException("Cannot clear an inverse map through its key set");
        }

        super.clear(sm);
    }

    // ------------------------------- Query Methods ---------------------------

    public ScalarExpression
        joinElementsTo(QueryStatement stmt,
                	   JavaTypeMapping ownerMapping,
                	   TableExpression te,
                	   DatastoreIdentifier setRangeVar,
                       Class filteredElementType,
                       TableExpression elmte,
                       DatastoreIdentifier elementRangeVar)
    {
        throw new JDOUnsupportedOptionException("Cannot query sets obtained by Map.keySet()");
    }

    public QueryStatement newQueryStatement(StateManager sm, String candidateClass)
    {
        throw new JDOUnsupportedOptionException("Cannot query sets obtained by Map.keySet()");
    }
}