/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others. All rights reserved.
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Erik Bengtson - removed unused import
2003 Andy Jefferson - coding standards
2004 Andy Jefferson - moved statements to AbstractSetStore
    ...
**********************************************************************/
package org.jpox.store.rdbms.scostore;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOUnsupportedOptionException;

import org.jpox.ClassLoaderResolver;
import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.TableExpression;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.Mappings;
import org.jpox.store.query.IteratorStatement;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.mapping.RDBMSMapping;
import org.jpox.store.rdbms.table.ClassTable;
import org.jpox.store.rdbms.table.MapTable;
import org.jpox.store.scostore.MapStore;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;
import org.jpox.util.SQLWarnings;

/**
 * Representation of the backing store for a MapValue Set.
 *
 * @version $Revision: 1.6 $
 **/
public class MapValueSetStore extends AbstractSetStore
{
    protected static final Localiser LOCALISER=Localiser.getInstance("org.jpox.store.rdbms.Localisation");

    private final MapStore mapStore;
    private final JavaTypeMapping keyMapping;
    private final String findKeyStmt;

    public MapValueSetStore(MapTable mapTable, MapStore mapStore)
    {
        this.setTable = mapTable;
        this.mapStore = mapStore;

        setName  = "values";
        storeMgr = mapTable.getStoreManager();
        dba      = storeMgr.getDatastoreAdapter();

        ownerMapping   = mapTable.getOwnerMapping();
        keyMapping     = mapTable.getKeyMapping();
        elementMapping = mapTable.getValueMapping();

        elementType = elementMapping.getType().getName();
        elementsAreEmbedded = isEmbeddedMapping(elementMapping);

        initialize();

        findKeyStmt  = getFindKeyStmt();
    }

    public MapValueSetStore(ClassTable valueTable, JavaTypeMapping ownerMapping, MapStore mapStore)
    {
        this.setTable = valueTable;
        this.mapStore = mapStore;

        setName  = "values";
        storeMgr = valueTable.getStoreManager();
        dba      = storeMgr.getDatastoreAdapter();

        elementType = valueTable.getType();

        this.ownerMapping   = ownerMapping;
        this.keyMapping     = null;
        this.elementMapping = valueTable.getIDMapping();

        elementsAreEmbedded = false;

        initialize();

        /* Set is read-only. */
        clearStmt    = null;
        findKeyStmt  = null;
    }

    /**
     * Initialise Method.
     **/
    private void initialize()
    {
        // Generate the statements
        initialiseStatements();

        ClassLoaderResolver clr=storeMgr.getClassLoaderResolver();
        Class element_class=clr.classForName(elementType);
        if (element_class.isInterface())
        {
            emd = storeMgr.getMetaDataManager().getMetaDataForInterface(element_class,null);
        }
        else
        {
            emd = storeMgr.getMetaDataManager().getMetaDataForClass(element_class);
        }
        
        if (emd != null)
        {
            elementType = emd.getFullClassName();
            elementTable = (ClassTable) storeMgr.getDatastoreClass(elementType);
        }

    }
 
    /**
     * Generate statement to find keys in the Map.
     * <PRE> 
     * SELECT KEYCOL FROM SETTABLE 
     * WHERE OWNERCOL=? 
     * AND ELEMENTCOL = ?
     * </PRE>
     * @return Statement to find keys in the Map. 
     */
    private String getFindKeyStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("SELECT ");
        stmt.append(((Column)keyMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
        for (int i=1; i<keyMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(",");
            stmt.append(((Column)keyMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
        }
        stmt.append(" FROM ");
        stmt.append(setTable.toString());
        stmt.append(" WHERE ");
        stmt.append(((Column)ownerMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
        stmt.append(" = ");
        stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(0)).getUpdateInputParameter());
        for (int i=1; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(((Column)ownerMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        for (int i=0; i<elementMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(((Column)elementMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)elementMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }   
        return stmt.toString();
    }
 
    protected QueryStatement getIteratorStatement(StateManager sm)
    {
        final ClassLoaderResolver clr=storeMgr.getClassLoaderResolver();
        QueryStatement stmt = new IteratorStatement(
                new IteratorStatement.AssociationEnd()
                {
                    public JavaTypeMapping getMapping()
                    {
                        return elementMapping;
                    }
                    public Class getType()
                    {
                        return clr.classForName(elementType);
                    }
                    public DatastoreContainerObject getDatastoreContainerObject()
                    {
                        return setTable;
                    }
                    public boolean useJoin()
                    {
                        return false;
                    }                    
                },
                clr.classForName(this.elementType),
                this.storeMgr,
                this.dba,
                true).getQueryStatement();

        /*
         * Apply condition on join-table owner field to filter by owner
         */
        ScalarExpression ownerExpr = ownerMapping.newScalarExpression(stmt, stmt.getDefaultTableExpression());
        ScalarExpression ownerVal = ownerMapping.newLiteral(stmt, sm.getObject());

        stmt.andCondition(ownerExpr.eq(ownerVal), true); 
        
        return stmt;
    }
    
    public boolean add(StateManager sm, Object element)
    {
        throw new UnsupportedOperationException("Cannot add to a map through its value set");
    }

    public boolean remove(StateManager sm, Object element)
    {
        if (findKeyStmt == null)
        {
            throw new UnsupportedOperationException("Cannot remove from an inverse map through its value set");
        }

        if (!validateElementForReading(sm, element))
        {
            return false;
        }

        Object key = null;
        boolean keyExists = false;
        PersistenceManager pm = sm.getPersistenceManager();

        try
        {
            Connection conn = pm.getConnection(true);

            try
            {
                PreparedStatement ps = conn.prepareStatement(findKeyStmt);

                try
                {
                    ownerMapping.setObject(pm, ps, Mappings.getParametersIndex(1,ownerMapping), sm.getObject());
                    elementMapping.setObject(pm, ps, Mappings.getParametersIndex(ownerMapping.getNumberOfDatastoreFields()+1,elementMapping), element);
                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(findKeyStmt);
                    }

                    long startTime = System.currentTimeMillis();
                    ResultSet rs = ps.executeQuery();
                    try
                    {
                        if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                        {
                            JPOXLogger.RDBMS_SQL.debug(LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                        }

                        if (rs.next())
                        {
                            key = keyMapping.getObject(pm, rs, Mappings.getParametersIndex(1,keyMapping));

                            keyExists = true;
                        }

                        SQLWarnings.log(rs);
                    }
                    finally
                    {
                        rs.close();
                    }
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                pm.releaseConnection(conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException("Request failed to check if set contains an element: " + findKeyStmt, e);
        }

        if (keyExists)
        {
            mapStore.remove(sm, key);
            return true;
        }
        else
        {
            return false;
        }
    }

    public void clear(StateManager sm)
    {
        if (clearStmt == null)
        {
            throw new UnsupportedOperationException("Cannot clear an inverse map through its value set");
        }

        super.clear(sm);
    }
 
    // --------------------------- Query Methods -------------------------------

    public QueryStatement newQueryStatement(StateManager sm, String candidateClass)
    {
        throw new JDOUnsupportedOptionException("Cannot query sets obtained by Map.values()");
    }

    public ScalarExpression
        joinElementsTo(QueryStatement stmt,
                	   JavaTypeMapping ownerMapping,
 		   			   TableExpression te,
 		   			   DatastoreIdentifier setRangeVar,
                       Class filteredElementType,
                       TableExpression elmte,
                       DatastoreIdentifier elementRangeVar)
    {
        throw new JDOUnsupportedOptionException("Cannot query sets obtained by Map.values()");
    }
}
