/**********************************************************************
Copyright (c) 2003 David Jencks and others. All rights reserved.
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Erik Bengtson  - removed unused import
2003 Erik Bengtson  - fixed bug [832635] Imcompatiblities in casting
                     inherited classes in query
2003 Andy Jefferson - coding standards
2003 Andy Jefferson - added getSizeStmt
2003 Andy Jefferson - updated retrieval of ColumnList's
2004 Andy Jefferson - updated to support inherited objects
2004 Marco Schulze  - replaced catch(NotPersistenceCapableException ...) 
                     by advance-check via TypeManager.isSupportedType(...)
2004 Andy Jefferson - merged IteratorStmt and GetStmt into GetRangeStmt.
2004 Andy Jefferson - added removeAll
2004 Andy Jefferson - moved all statement construction to AbstractListStore
    ...
**********************************************************************/
package org.jpox.store.rdbms.scostore;

import javax.jdo.JDOUnsupportedOptionException;

import org.jpox.ClassLoaderResolver;
import org.jpox.StateManager;
import org.jpox.TypeManager;
import org.jpox.store.DatastoreClass;
import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.exceptions.IncompatibleQueryElementTypeException;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.TableExpression;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.query.IteratorStatement;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.rdbms.adapter.DatabaseAdapter;
import org.jpox.store.rdbms.sqlidentifier.TableIdentifier;
import org.jpox.store.rdbms.table.ClassTable;
import org.jpox.store.rdbms.table.ListTable;
import org.jpox.store.rdbms.table.SCOTable;
import org.jpox.util.JPOXLogger;

/**
 * Representation of the backing store for a Normal List.
 * Uses a Join table, so we use 3 tables - owner table, join table and element
 * table. 
 *
 * @version $Revision: 1.8 $
 **/
public class NormalListStore extends AbstractListStore
{
    public NormalListStore(ListTable listTable)
    {
        this.listTable = listTable;
        listName       = "list";

        storeMgr = listTable.getStoreManager();
        dba      = storeMgr.getDatastoreAdapter();

        ownerMapping   = listTable.getOwnerMapping();
        elementMapping = listTable.getElementMapping();
        indexMapping   = listTable.getIndexMapping();

        // Generate the statements
        initialiseStatements(); 

        elementType = elementMapping.getType().getName();
        elementsAreEmbedded = isEmbeddedMapping(elementMapping);

        ClassLoaderResolver clr=storeMgr.getClassLoaderResolver();
        Class element_class=clr.classForName(elementType);
        if (element_class.isInterface())
        {
            JPOXLogger.JDO.warn(LOCALISER.msg("SCO.List.InterfaceElementNotSupported", element_class.getName()));
            emd = storeMgr.getMetaDataManager().getMetaDataForInterface(element_class,null);
            if (emd != null)
            {
                elementType = element_class.getName();
                // TODO This currently just grabs the cmd of the first implementation. It needs to
                // get the cmds for all implementations, so we can have a handle to all possible elements.
                // This would mean changing the SCO classes to have multiple elementTable/elementMapping.
                elementTable = (ClassTable) storeMgr.getDatastoreClass(emd.getFullClassName());
            }
        }
        else
        {
            emd = storeMgr.getMetaDataManager().getMetaDataForClass(element_class);
            if (emd != null)
            {
                elementType  = emd.getFullClassName();
                elementTable = (ClassTable) storeMgr.getDatastoreClass(elementType);
            }
        }
    }

    // ------------------------- GetRange Statement ----------------------------
 
    /**
     * Accessor for the GetRange statement to retrieve element(s) in a range
     * from the List. This has 5 modes
     * <UL>
     * <LI>start and end have the same value and &ge;0 then get that
     * element</LI>
     * <LI>start and end have values &ge;0 and are different then get a
     * range</LI>
     * <LI>start is &ge;0 and end is -1 the get all from start (inclusive).</LI>
     * <LI>start is -1 and end is &ge;0 the get all up to end(exclusive).</LI> 
     * <LI>start and end values are both -1 then get all elements</LI> 
     * </UL>
     * @param start_index The start position in the List. 
     * @param end_index The end position in the List. 
     * @return The QueryStatement.
     **/
    protected QueryStatement getGetRangeStatement(StateManager sm,int start_index,int end_index)
    {
        final ClassLoaderResolver clr=storeMgr.getClassLoaderResolver();
        QueryStatement stmt = new IteratorStatement(
                new IteratorStatement.AssociationEnd()
                {
                    public JavaTypeMapping getMapping()
                    {
                        return elementMapping;
                    }
                    public Class getType()
                    {
                        return clr.classForName(elementType);
                    }
                    public DatastoreContainerObject getDatastoreContainerObject()
                    {
                        return listTable;
                    }
                    public boolean useJoin()
                    {
                        return false;
                    }                    
                },
                clr.classForName(this.elementType), 
                this.storeMgr, 
                this.dba, 
                true).getQueryStatement();
 
        JavaTypeMapping ownerMapping=((SCOTable)listTable).getOwnerMapping();

        /*
         * Apply condition on join-table owner field to filter by owner
         */
        ScalarExpression ownerExpr = ownerMapping.newScalarExpression(stmt, stmt.getDefaultTableExpression());
        ScalarExpression ownerVal = ownerMapping.newLiteral(stmt, sm.getObject());

        stmt.andCondition(ownerExpr.eq(ownerVal), true);        

        boolean returning_range=false;
        if (start_index == -1 && end_index == -1)
        {
            // Add no more constraints. We want all elements.
            returning_range = true; 
        }
        else if (start_index == end_index && start_index >= 0)
        {
            /*
             * Apply condition on join-table owner field to filter by index
             */
            ScalarExpression indexExpr = indexMapping.newScalarExpression(stmt,stmt.getDefaultTableExpression());
            ScalarExpression indexVal = indexMapping.newLiteral(stmt, new Integer(start_index));
            
            stmt.andCondition(indexExpr.eq(indexVal), true);        
        }
        else
        {
            // Apply filter to retrieve indexes in the range
            // a). >= start_index 
            if (start_index >= 0)
            {
                /*
                 * Apply condition on join-table index field to filter by index
                 */
                ScalarExpression indexExpr = indexMapping.newScalarExpression(stmt,stmt.getDefaultTableExpression());
                ScalarExpression indexVal = indexMapping.newLiteral(stmt, new Integer(start_index));
                
                stmt.andCondition(indexExpr.gteq(indexVal), true);                     
            }
            // b). < end_index 
            if (end_index >= 0)
            {
                /*
                 * Apply condition on join-table index field to filter by index
                 */
                ScalarExpression indexExpr = indexMapping.newScalarExpression(stmt,stmt.getDefaultTableExpression());
                ScalarExpression indexVal = indexMapping.newLiteral(stmt, new Integer(end_index));
                
                stmt.andCondition(indexExpr.lt(indexVal), true);                 
            }
            returning_range = true; 
        }

        // If returning a range, add ordering by index
        if (returning_range)
        { 
            ScalarExpression exprIndex[] = new ScalarExpression[indexMapping.getNumberOfDatastoreFields()];
            boolean descendingOrder[] = new boolean[indexMapping.getNumberOfDatastoreFields()];
            exprIndex = indexMapping.newScalarExpression(stmt,stmt.getDefaultTableExpression()).getExpressionList().toArray();
            stmt.setOrdering(exprIndex, descendingOrder);
        }
 
        return stmt;
    }

    // ---------------------------- Query Methods ------------------------------
 
    /**
     * Utility to return a new QueryStatement.
     * @param sm The state manager
     * @param candidateClass The base class
     * @return The QueryStatement
     **/
    public QueryStatement newQueryStatement(StateManager sm,
                                            String candidateClass)
    {
        if (elementsAreEmbedded)
        {
            throw new JDOUnsupportedOptionException(LOCALISER.msg("SCO.QueryOverCollectionImpossible"));
        }

        TableIdentifier listRangeVar    = new TableIdentifier(dba, listName);
        TableIdentifier elementRangeVar = new TableIdentifier(dba, "this");

        QueryStatement stmt = ((DatabaseAdapter) dba).newQueryStatement(listTable, listRangeVar);

        ScalarExpression ownerExpr = ownerMapping.newScalarExpression(stmt, stmt.getTableExpression(listRangeVar));
        ScalarExpression ownerVal = ownerMapping.newLiteral(stmt, sm.getObject());
        stmt.andCondition(ownerExpr.eq(ownerVal), true);

        if (!sm.getPersistenceManager().getClassLoaderResolver().isAssignableFrom(elementType,candidateClass))
        {
            throw new IncompatibleQueryElementTypeException(elementType, candidateClass);
        }

        if (!TypeManager.getTypeManager().isSupportedType(candidateClass))
        {
            DatastoreClass candidateTable=storeMgr.getDatastoreClass(candidateClass);
            JavaTypeMapping elementTableID=candidateTable.getIDMapping();

            stmt.newTableExpression(candidateTable, elementRangeVar);

            ScalarExpression elmListExpr = elementMapping.newScalarExpression(stmt,stmt.getTableExpression(listRangeVar));
            ScalarExpression elmExpr = elementTableID.newScalarExpression(stmt,stmt.getTableExpression(elementRangeVar));

            stmt.innerJoin(elmExpr,elmListExpr,stmt.getTableExpression(elementRangeVar), true);
            
            stmt.select(elementTableID);
        }
        else
        {
            stmt.select(listRangeVar, elementMapping);
        }

        return stmt;
    }

    /**
     * Used as part of the Querying of Lists where a contains() is used.
     **/
    public ScalarExpression joinElementsTo(
            		   QueryStatement stmt,
            		   JavaTypeMapping ownerMapping,
            		   TableExpression te,
            		   DatastoreIdentifier listRangeVar,
                       Class filteredElementType,
                       TableExpression elmte,
                       DatastoreIdentifier elementRangeVar)
    {
        ClassLoaderResolver clr=storeMgr.getClassLoaderResolver();
        if (!clr.isAssignableFrom(elementType,filteredElementType) &&
            !clr.isAssignableFrom(filteredElementType,elementType))
        {
            throw new IncompatibleQueryElementTypeException(elementType, filteredElementType.getName());
        }

        // Join the list table on the owner ID column
        stmt.newTableExpression(listTable, listRangeVar);

        ScalarExpression ownerExpr = ownerMapping.newScalarExpression(stmt,te);
        ScalarExpression ownerListExpr = this.ownerMapping.newScalarExpression(stmt,stmt.getTableExpression(listRangeVar));

        stmt.innerJoin(ownerExpr,ownerListExpr,stmt.getTableExpression(listRangeVar), true);

        if (!TypeManager.getTypeManager().isSupportedType(filteredElementType.getName()))
        {
            DatastoreClass elementTable = storeMgr.getDatastoreClass(filteredElementType.getName());
            DatastoreClass joiningClass = (elmte == null ? elementTable : (DatastoreClass)elmte.getMainTable());
            // Join the element table on the element ID column.
            JavaTypeMapping elementTableID = joiningClass.getIDMapping();

            if( stmt.getTableExpression(elementRangeVar) == null )
            {
                stmt.newTableExpression(joiningClass, elementRangeVar);
            }
            
            ScalarExpression elmListExpr = elementMapping.newScalarExpression(stmt,stmt.getTableExpression(listRangeVar));
            ScalarExpression elmExpr = elementTableID.newScalarExpression(stmt,stmt.getTableExpression(elementRangeVar));

            stmt.innerJoin(elmExpr,elmListExpr,stmt.getTableExpression(elementRangeVar), true);
            
            return elementTableID.newScalarExpression(stmt,stmt.getTableExpression(elementRangeVar));
        }
        else
        {
            return elementMapping.newScalarExpression(stmt,stmt.getTableExpression(listRangeVar));
        }
    }
}
