/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others. All rights reserved.
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - coding standards
2004 Andy Jefferson - addition of getGetStatement for inherited values
2004 Andy Jefferson - addition of query methods
2004 Marco Schulze  - replaced catch(NotPersistenceCapableException ...) by
                     advance-check via TypeManager.isSupportedType(...)
2004 Andy Jefferson - moved statements to AbstractMapStore
    ...
**********************************************************************/
package org.jpox.store.rdbms.scostore;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.NoSuchElementException;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOUnsupportedOptionException;

import org.jpox.ClassLoaderResolver;
import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.TypeManager;
import org.jpox.store.DatastoreClass;
import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.exceptions.IncompatibleQueryElementTypeException;
import org.jpox.store.expression.BooleanExpression;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.TableExpression;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.Mappings;
import org.jpox.store.mapping.ObjectMapping;
import org.jpox.store.query.IteratorStatement;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.adapter.DatabaseAdapter;
import org.jpox.store.rdbms.mapping.RDBMSMapping;
import org.jpox.store.rdbms.sqlidentifier.SQLIdentifier;
import org.jpox.store.rdbms.sqlidentifier.TableIdentifier;
import org.jpox.store.rdbms.table.ClassTable;
import org.jpox.store.rdbms.table.MapTable;
import org.jpox.store.scostore.SetStore;
import org.jpox.util.JPOXLogger;
import org.jpox.util.SQLWarnings;

/**
 * Representation of the backing store for a Map.
 * Uses a join table to link the owning container object with the values/keys
 * contained in the map. 
 *
 * @version $Revision: 1.9 $
 **/
public class NormalMapStore extends AbstractMapStore
{
    private final String putStmt;
    private final String updateStmt;
    private final String removeStmt;
    private final String clearStmt;

    private SetStore keySetStore = null;
    private SetStore valueSetStore = null;
    private SetStore entrySetStore = null;
    
    /**
     * when the element mappings columns can't be part of the primary key
     * by datastore limitations like BLOB types.
     * An adapter mapping is used to be a kind of "index"
     */
    protected final JavaTypeMapping adapterMapping;    

    /**
     * Constructor for an Inverse Map.
     * @param mapTable Join table for the Map
     **/
    public NormalMapStore(MapTable mapTable)
    {
        this.mapTable = mapTable;

        storeMgr = mapTable.getStoreManager();
        dba      = storeMgr.getDatastoreAdapter();

        ownerMapping   = mapTable.getOwnerMapping();
        keyMapping     = mapTable.getKeyMapping();
        valueMapping   = mapTable.getValueMapping();
        adapterMapping = mapTable.getAdapterMapping();

        keyType = keyMapping.getType().getName();
        valueType = valueMapping.getType().getName();

        keysAreEmbedded   = isEmbeddedMapping(keyMapping);
        valuesAreEmbedded = isEmbeddedMapping(valueMapping);

        // Generate the statements
        initialiseStatements(); 
        putStmt           = getPutStmt();
        updateStmt        = getUpdateStmt();
        removeStmt        = getRemoveStmt(); 
        clearStmt         = getClearStmt();

        ClassLoaderResolver clr=storeMgr.getClassLoaderResolver();
        Class value_class=clr.classForName(valueType);
        if (value_class.isInterface())
        {
            JPOXLogger.JDO.warn(LOCALISER.msg("SCO.MapInterfaceValueNotSupported", value_class.getName()));
            vmd = storeMgr.getMetaDataManager().getMetaDataForInterface(value_class,null);
            if (vmd != null)
            {
                valueType = value_class.getName();
                // TODO This currently just grabs the cmd of the first implementation. It needs to
                // get the cmds for all implementations, so we can have a handle to all possible elements.
                // This would mean changing the SCO classes to have multiple elementTable/elementMapping.
                valueTable = (ClassTable) storeMgr.getDatastoreClass(vmd.getFullClassName());
            }
        }
        else
        {
            vmd = storeMgr.getMetaDataManager().getMetaDataForClass(value_class);
            if (vmd != null)
            {
                valueType = vmd.getFullClassName();
                valueTable = (ClassTable) storeMgr.getDatastoreClass(valueType);
            }
        }
    }

    /**
     * Generate statement to add an item to the Map.
     * Adds a row to the link table, linking container with value object.
     * <PRE>
     * INSERT INTO MAPTABLE (VALUECOL, OWNERCOL, KEYCOL)
     * VALUES (?, ?, ?)
     * </PRE>
     * @return Statement to add an item to the Map.
     */
    private String getPutStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("INSERT INTO ");
        stmt.append(mapTable.toString());
        stmt.append(" (");
        stmt.append(((Column)valueMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
        for (int i=1; i<valueMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(",");
            stmt.append(((Column)valueMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
        }
        for (int i=0; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(",");
            stmt.append(((Column)ownerMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
        }
        if (adapterMapping != null)
        {
            for (int i=0; i<adapterMapping.getNumberOfDatastoreFields(); i++)
            {
                stmt.append(",");
                stmt.append(((Column)adapterMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            }
        }        
        for (int i=0; i<keyMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(",");
            stmt.append(((Column)keyMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
        }
        stmt.append(") VALUES (");
        stmt.append(((RDBMSMapping)valueMapping.getDataStoreMapping(0)).getInsertionInputParameter());
        for (int i=1; i<valueMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(",");
            stmt.append(((RDBMSMapping)valueMapping.getDataStoreMapping(i)).getInsertionInputParameter());
        }
        for (int i=0; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(",");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getInsertionInputParameter());
        }
        if (adapterMapping != null)
        {
            for (int i=0; i<adapterMapping.getNumberOfDatastoreFields(); i++)
            {
                stmt.append(",");
                stmt.append(((RDBMSMapping)adapterMapping.getDataStoreMapping(i)).getInsertionInputParameter());
            }
        }        
        for (int i=0; i<keyMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(",");
            stmt.append(((RDBMSMapping)keyMapping.getDataStoreMapping(i)).getInsertionInputParameter());
        }
        stmt.append(") ");
        return stmt.toString();
    }

    /**
     * Generate statement to update an item in the Map.
     * Updates the link table row, changing the value object for this key. 
     * <PRE>
     * UPDATE MAPTABLE 
     * SET VALUECOL=? 
     * WHERE OWNERCOL=? 
     * AND KEYCOL=?
     * </PRE>
     * @return Statement to update an item in the Map.
     */
    private String getUpdateStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("UPDATE ");
        stmt.append(mapTable.toString());
        stmt.append(" SET ");
        stmt.append(((Column)valueMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
        stmt.append(" = ");
        stmt.append(((RDBMSMapping)valueMapping.getDataStoreMapping(0)).getUpdateInputParameter());
        for (int i=1; i<valueMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(",");
            stmt.append(((Column)valueMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)valueMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        stmt.append(" WHERE ");
        stmt.append(((Column)ownerMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
        stmt.append(" = ");
        stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(0)).getUpdateInputParameter());
        for (int i=1; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(((Column)ownerMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        for (int i=0; i<keyMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(((Column)keyMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)keyMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        return stmt.toString();
    }

    /**
     * Generate statement to remove an item from the Map.
     * Deletes the link from the join table, leaving the value object in its
     * own table.
     * <PRE>
     * DELETE FROM MAPTABLE 
     * WHERE OWNERCOL=? 
     * AND KEYCOL=?
     * </PRE>
     * @return Return an item from the Map.
     */
    private String getRemoveStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("DELETE FROM ");
        stmt.append(mapTable.toString());
        stmt.append(" WHERE ");
        stmt.append(((Column)ownerMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
        stmt.append(" = ");
        stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(0)).getUpdateInputParameter());
        for (int i=1; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(((Column)ownerMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        for (int i=0; i<keyMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(((Column)keyMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)keyMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        return stmt.toString();
    }

    /**
     * Generate statement to clear the Map.
     * Deletes the links from the join table for this Map, leaving the value
     * objects in their own table(s).
     * <PRE>
     * DELETE FROM MAPTABLE 
     * WHERE OWNERCOL=?
     * </PRE>
     * @return Statement to clear the Map.
     */
    private String getClearStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("DELETE FROM ");
        stmt.append(mapTable.toString());
        stmt.append(" WHERE ");
        stmt.append(((Column)ownerMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
        stmt.append(" = ");
        stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(0)).getUpdateInputParameter());
        for (int i=1; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(((Column)ownerMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        return stmt.toString();
    }

    /**
     * Accessor for the higher id when elements primary key can't be part of
     * the primary key by datastore limitations like BLOB types can't be
     * primary keys.
     * @param sm State Manager for container 
     * @return The next id 
     */
    private int getNextIDForAdapterColumn(StateManager sm)
    {
        int nextID;
        PersistenceManager pm = sm.getPersistenceManager();

        try
        {
            Connection conn = pm.getConnection(false);

            try
            {
                PreparedStatement ps = conn.prepareStatement(getMaxAdapterColumnIdStmt());

                try
                {
                    ownerMapping.setObject(pm, ps, Mappings.getParametersIndex(1,ownerMapping), sm.getObject());

                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(getMaxAdapterColumnIdStmt());
                    }
                    long startTime = System.currentTimeMillis();
                    ResultSet rs = ps.executeQuery();
                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(STORE_LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                    }

                    try
                    {
                        if (!rs.next())
                        {
                            nextID = 1;
                        }
                        else
                        {
                            nextID = rs.getInt(1)+1;
                        }

                        SQLWarnings.log(rs);
                    }
                    finally
                    {
                        rs.close();
                    }
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                pm.releaseConnection(conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("SCO.MaxAdapterColumnIdRequestFailed",getMaxAdapterColumnIdStmt()),e);
        }

        return nextID;
    }
    
    /**
     * Method to put an item in the Map.
     * @param sm State Manager for the map.
     * @param key The key to store the value against
     * @param value The value to store.
     * @return The value stored.
     **/
    public Object put(StateManager sm, Object key, Object value)
    {
        validateKeyForWriting(sm, key);
        validateValueForWriting(sm, value);

        PersistenceManager pm = sm.getPersistenceManager();
        Object oldValue;
        String stmt;

        try
        {
            oldValue = getValue(sm, key);
            stmt = updateStmt;
        }
        catch (NoSuchElementException e)
        {
            oldValue = null;
            stmt = putStmt;
        }

        if (oldValue != value)
        {
            try
            {
                Connection conn = pm.getConnection(true);

                try
                {
                    PreparedStatement ps = conn.prepareStatement(stmt);

                    try
                    {
                        valueMapping.setObject(pm, ps, Mappings.getParametersIndex(1,valueMapping), value);
                        ownerMapping.setObject(pm, ps, Mappings.getParametersIndex(valueMapping.getNumberOfDatastoreFields()+1,ownerMapping), sm.getObject());
                        //only sets the adapterMapping column s if the a new object (oldValue=null)
                        //for updates, we dont need to update the adapterMapping columns
                        if (oldValue == null && adapterMapping != null)
                        {
                            long nextIDAdapter = getNextIDForAdapterColumn(sm);
                            adapterMapping.setObject(pm, ps, Mappings.getParametersIndex(valueMapping.getNumberOfDatastoreFields()+ownerMapping.getNumberOfDatastoreFields()+1,adapterMapping), new Long(nextIDAdapter));
                            keyMapping.setObject(pm, ps, Mappings.getParametersIndex(valueMapping.getNumberOfDatastoreFields()+ownerMapping.getNumberOfDatastoreFields()+2,keyMapping), key);
                        }
                        else
                        {
                            keyMapping.setObject(pm, ps, Mappings.getParametersIndex(valueMapping.getNumberOfDatastoreFields()+ownerMapping.getNumberOfDatastoreFields()+1,keyMapping), key);
                        }

                        if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                        {
                            JPOXLogger.RDBMS_SQL.debug(stmt);
                        }
                        long startTime = System.currentTimeMillis();
                        ps.executeUpdate();
                        if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                        {
                            JPOXLogger.RDBMS_SQL.debug(STORE_LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                        }
                    }
                    finally
                    {
                        ps.close();
                    }
                }
                finally
                {
                    pm.releaseConnection(conn);
                }
            }
            catch (SQLException e)
            {
                throw new JDODataStoreException(LOCALISER.msg("SCO.PutRequestFailed",stmt),e);
            }
        }

        return oldValue;
    }

    /**
     * Method to remove an item from the map.
     * @param sm State Manager for the map.
     * @param key Key of the item to remove.
     * @return The value that was removed.
     **/
    public Object remove(StateManager sm, Object key)
    {
        if (!validateKeyForReading(sm, key))
        {
            return null;
        }

        PersistenceManager pm = sm.getPersistenceManager();
        Object oldValue;
        boolean exists;

        try
        {
            oldValue = getValue(sm, key);
            exists = true;
        }
        catch (NoSuchElementException e)
        {
            oldValue = null;
            exists = false;
        }

        if (exists)
        {
            try
            {
                Connection conn = pm.getConnection(true);

                try
                {
                    PreparedStatement ps = conn.prepareStatement(removeStmt);

                    try
                    {
                        ownerMapping.setObject(pm, ps, Mappings.getParametersIndex(1,ownerMapping), sm.getObject());
                        keyMapping.setObject(pm, ps, Mappings.getParametersIndex(ownerMapping.getNumberOfDatastoreFields()+1,keyMapping), key);

                        if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                        {
                            JPOXLogger.RDBMS_SQL.debug(removeStmt);
                        }
                        long startTime = System.currentTimeMillis();
                        ps.executeUpdate();
                        if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                        {
                            JPOXLogger.RDBMS_SQL.debug(STORE_LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                        }
                    }
                    finally
                    {
                        ps.close();
                    }
                }
                finally
                {
                    pm.releaseConnection(conn);
                }
            }
            catch (SQLException e)
            {
                throw new JDODataStoreException(LOCALISER.msg("SCO.RemoveRequestFailed",removeStmt),e);
            }
        }

        return oldValue;
    }

    /**
     * Method to clear the map of all values.
     * @param sm State Manager for the map.
     **/
    public void clear(StateManager sm)
    {
        PersistenceManager pm = sm.getPersistenceManager();

        try
        {
            Connection conn = pm.getConnection(true);

            try
            {
                PreparedStatement ps = conn.prepareStatement(clearStmt);

                try
                {
                    ownerMapping.setObject(pm, ps, Mappings.getParametersIndex(1,ownerMapping), sm.getObject());

                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(clearStmt);
                    }
                    long startTime = System.currentTimeMillis();
                    ps.executeUpdate();
                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(STORE_LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                    }
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                pm.releaseConnection(conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(LOCALISER.msg("SCO.ClearRequestFailed",clearStmt),e);
        }
    }

    /**
     * Accessor for the keys in the Map.
     * @return The keys
     **/
    public synchronized SetStore keySetStore()
    {
        if (keySetStore == null)
        {
            keySetStore = new MapKeySetStore((MapTable)mapTable);
        }

        return keySetStore;
    }

    /**
     * Accessor for the values in the Map.
     * @return The values.
     **/
    public synchronized SetStore valueSetStore()
    {
        if (valueSetStore == null)
        {
            valueSetStore = new MapValueSetStore((MapTable)mapTable, this);
        }

        return valueSetStore;
    }

    /**
     * Accessor for the map entries in the Map.
     * @return The map entries.
     **/
    public synchronized SetStore entrySetStore()
    {
        if (entrySetStore == null)
        {
            entrySetStore = new MapEntrySetStore((MapTable)mapTable, this);
        }

        return entrySetStore;
    }

    /**
     * Accessor for the Get statement to retrieve a value from the Map.
     * @return The QueryStatement.
     **/
    protected QueryStatement getGetStatement(StateManager sm, Object key)
    {
        final ClassLoaderResolver clr=storeMgr.getClassLoaderResolver();
        QueryStatement stmt = new IteratorStatement(
                new IteratorStatement.AssociationEnd()
                {
                    public JavaTypeMapping getMapping()
                    {
                        return valueMapping;
                    }
                    public Class getType()
                    {
                        return clr.classForName(valueType);
                    }
                    public DatastoreContainerObject getDatastoreContainerObject()
                    {
                        return mapTable;
                    }
                    public boolean useJoin()
                    {
                        return false;
                    }                    
                },
                clr.classForName(this.valueType),
                this.storeMgr,
                this.dba,
                true).getQueryStatement();

        /*
         * Apply condition on join-table owner field to filter by owner
         */
        ScalarExpression ownerExpr = ownerMapping.newScalarExpression(stmt, stmt.getDefaultTableExpression());
        ScalarExpression ownerVal = ownerMapping.newLiteral(stmt, sm.getObject());

        stmt.andCondition(ownerExpr.eq(ownerVal), true);        

        //TODO
        //this is a database hack. :-)
        //if the keyMapping contains a BLOB column (or any other column not supported by the database
        //as primary key), uses like instead of the operator OP_EQ (=)
        //in future do not check if the keyMapping is of ObjectMapping, but use the database 
        //adapter to check the data types not supported as primary key
        //
        //if object mapping (BLOB) use like
        if (keyMapping instanceof ObjectMapping)
        {
            /*
             * Apply condition on join-table owner field to filter by owner
             */
            ScalarExpression keyExpr = keyMapping.newScalarExpression(stmt,stmt.getDefaultTableExpression());
            ScalarExpression keyVal = keyMapping.newLiteral(stmt, key).add(dba.getMapping(String.class).newLiteral(stmt,"%"));
            stmt.andCondition(new BooleanExpression(keyExpr,ScalarExpression.OP_LIKE,keyVal), true);             
        }
        else
        {
            /*
             * Apply condition on join-table owner field to filter by owner
             */
            ScalarExpression keyExpr = keyMapping.newScalarExpression(stmt,stmt.getDefaultTableExpression());
            ScalarExpression keyVal = keyMapping.newLiteral(stmt, key);

            stmt.andCondition(keyExpr.eq(keyVal), true);             
        }
        return stmt;
    }
 
    /**
     * Generate statement for obtaining the maximum id.
     * <PRE>
     * SELECT MAX(SCOID) FROM MAPTABLE 
     * WHERE OWNERCOL=?
     * </PRE> 
     * @return The Statement returning the higher id
     */
    private String getMaxAdapterColumnIdStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("SELECT ");
        stmt.append(" MAX(" + ((Column)adapterMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString() + ") ");
        stmt.append(" FROM ");
        stmt.append(mapTable.toString());
        stmt.append(" WHERE ");
        stmt.append(((Column)ownerMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
        stmt.append(" = ");
        stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(0)).getUpdateInputParameter());
        for (int i=1; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(((Column)ownerMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        return stmt.toString();
    }

    // ---------------------------- Query Methods ------------------------------
 
    /**
     * Utility to return a new QueryStatement.
     * @param sm The state manager
     * @param candidateClass The base value class
     * @return The QueryStatement
     **/
    public QueryStatement newQueryStatement(StateManager sm,
                                            String candidateClass)
    {
        if (valuesAreEmbedded)
        {
            throw new JDOUnsupportedOptionException(LOCALISER.msg("SCO.QueryOverMapImpossible"));
        }

        TableIdentifier mapRangeVar   = new TableIdentifier(dba, "map");
        TableIdentifier valueRangeVar = new TableIdentifier(dba, "this");

        QueryStatement stmt = ((DatabaseAdapter) dba).newQueryStatement(mapTable, mapRangeVar);

        ScalarExpression ownerExpr = ownerMapping.newScalarExpression(stmt, stmt.getDefaultTableExpression());
        ScalarExpression ownerVal = ownerMapping.newLiteral(stmt, sm.getObject());
        stmt.andCondition(ownerExpr.eq(ownerVal), true);
        
        if (!sm.getPersistenceManager().getClassLoaderResolver().isAssignableFrom(valueType,candidateClass))
        {
            throw new IncompatibleQueryElementTypeException(valueType, candidateClass);
        }

        if (!TypeManager.getTypeManager().isSupportedType(candidateClass))
        {
            DatastoreClass candidateTable=storeMgr.getDatastoreClass(candidateClass);
            JavaTypeMapping valueTableID=candidateTable.getIDMapping();

            stmt.newTableExpression(candidateTable, valueRangeVar);

            ScalarExpression valueMapExpr = valueMapping.newScalarExpression(stmt,stmt.getTableExpression(mapRangeVar));
            ScalarExpression valueExpr = valueTableID.newScalarExpression(stmt, stmt.getTableExpression(valueRangeVar));

            stmt.innerJoin(valueExpr,valueMapExpr,stmt.getTableExpression(valueRangeVar), true);

            stmt.select(valueTableID);
        }
        else
        {
            stmt.select(mapRangeVar, valueMapping);
        }

        return stmt;
    }

    /**
     * Utility to create a join for keys to be used in a containsKey() query.
     * @param stmt The Query Statement 
     * @param ownerMapping Mapping for the owner
     * @param te Table Expression for the owner
     * @param mapRangeVar   The range variable for the "Map" table. 
     * @param filteredKeyType The Class Type for the filtered key
     * @param keyRangeVar  The SQL alias, or "range variable", to assign to the
     *                     expression or to the key table.
     * @return QueryColumnList with the columns from the key mapping 
     **/
    public ScalarExpression joinKeysTo(
                                QueryStatement stmt,
                                JavaTypeMapping ownerMapping,
                                TableExpression te,
                                SQLIdentifier mapRangeVar,
                                Class filteredKeyType,
                                SQLIdentifier keyRangeVar)
    {
        ClassLoaderResolver clr=storeMgr.getClassLoaderResolver();
        if (!clr.isAssignableFrom(keyType,filteredKeyType) &&
            !clr.isAssignableFrom(filteredKeyType,keyType))
        {
            throw new IncompatibleQueryElementTypeException(keyType, filteredKeyType.getName());
        }

        // Join the map table on the owner ID column (apply to any unions)
        stmt.newTableExpression(mapTable, mapRangeVar);

        ScalarExpression ownerMapExpr = this.ownerMapping.newScalarExpression(stmt,stmt.getTableExpression(mapRangeVar));
        ScalarExpression ownerExpr = ownerMapping.newScalarExpression(stmt, te);

        stmt.innerJoin(ownerExpr,ownerMapExpr,stmt.getTableExpression(mapRangeVar), true);
        
        // if the key is a pc class, then join the "join table" to the table holding the key
        if (!TypeManager.getTypeManager().isSupportedType(filteredKeyType.getName()))
        {
            // Join the key table on the key ID column
            DatastoreClass keyTable=storeMgr.getDatastoreClass(filteredKeyType.getName());
            JavaTypeMapping keyTableID=keyTable.getIDMapping();

            stmt.newTableExpression(keyTable,keyRangeVar);

            ScalarExpression keyMapExpr = keyMapping.newScalarExpression(stmt,stmt.getTableExpression(mapRangeVar));
            ScalarExpression keyExpr = keyTableID.newScalarExpression(stmt, stmt.getTableExpression(keyRangeVar));

            stmt.innerJoin(keyExpr,keyMapExpr,stmt.getTableExpression(keyRangeVar), true);

            return keyTableID.newScalarExpression(stmt,stmt.getTableExpression(keyRangeVar));
        }
        else
        {
            return keyMapping.newScalarExpression(stmt,stmt.getTableExpression(mapRangeVar));
        }
    }

    /**
     * Utility to create a join for keys and values to be used in ai
     * containsEntry() query.
     * @param stmt The Query Statement to apply the join
     * @param ownerMapping Mapping for the owner
     * @param te Table Expression for the owner
     * @param mapRangeVar  The SQL alias, or "range variable", to assign to the
     *                     expression or to the main table.
     * @param filteredKeyType The Class Type for the filtered key
     * @param filteredValueType The Class Type for the filtered value
     * @param keyRangeVar  The SQL alias, or "range variable", to assign to the
     *                     expression or to the key table.
     * @param valueRangeVar The SQL alias, or "range variable", to assign to the
     *                      expression or to the value table.
     * @return an array with 2 elements of QueryColumnList. The first element
     * contains the columns from the key mapping and the second element the
     * columns from the value mapping
     */
    public ScalarExpression[] joinKeysValuesTo(
                                QueryStatement stmt,
                                JavaTypeMapping ownerMapping,
                                TableExpression te,
                                SQLIdentifier mapRangeVar,
                                Class filteredKeyType,
                                Class filteredValueType,
                                SQLIdentifier keyRangeVar,
    							SQLIdentifier valueRangeVar)
    {
        ClassLoaderResolver clr=storeMgr.getClassLoaderResolver();
        if (!clr.isAssignableFrom(keyType,filteredKeyType) &&
            !clr.isAssignableFrom(filteredKeyType,keyType))
        {
            throw new IncompatibleQueryElementTypeException(keyType, filteredKeyType.getName());
        }

        if (!clr.isAssignableFrom(valueType,filteredValueType) &&
            !clr.isAssignableFrom(filteredValueType,valueType))
        {
            throw new IncompatibleQueryElementTypeException(valueType, filteredValueType.getName());
        }
        
        // Join the map table on the owner ID column (apply to any unions)
        stmt.newTableExpression(mapTable, mapRangeVar);
        
        ScalarExpression ownerExpr = ownerMapping.newScalarExpression(stmt,te);
        ScalarExpression ownerMapExpr = this.ownerMapping.newScalarExpression(stmt, stmt.getTableExpression(mapRangeVar));

        stmt.innerJoin(ownerExpr,ownerMapExpr,stmt.getTableExpression(mapRangeVar), true);
        
        ScalarExpression[] qclKeyValues = new ScalarExpression[2];
        
        TypeManager tm = TypeManager.getTypeManager();
        if (!tm.isSupportedType(filteredKeyType.getName()))
        {
            // Join the key table on the key ID column
            DatastoreClass keyTable=storeMgr.getDatastoreClass(filteredKeyType.getName());
            JavaTypeMapping keyTableID=keyTable.getIDMapping();

            stmt.newTableExpression(keyTable,keyRangeVar);

            ScalarExpression keyMapExpr = keyMapping.newScalarExpression(stmt,stmt.getTableExpression(mapRangeVar));
            ScalarExpression keyExpr = keyTableID.newScalarExpression(stmt, stmt.getTableExpression(keyRangeVar));

            stmt.innerJoin(keyExpr,keyMapExpr,stmt.getTableExpression(keyRangeVar), true);

            qclKeyValues[0] = keyTableID.newScalarExpression(stmt,stmt.getTableExpression(keyRangeVar));
        }
        else
        {
            qclKeyValues[0] = keyMapping.newScalarExpression(stmt,stmt.getTableExpression(mapRangeVar));
        }
        if (!tm.isSupportedType(filteredValueType.getName()))
        {
            // Join the value table on the value ID column
            DatastoreClass valueTable=storeMgr.getDatastoreClass(filteredValueType.getName());
            JavaTypeMapping valueTableID=valueTable.getIDMapping();

            stmt.newTableExpression(valueTable,valueRangeVar);

            ScalarExpression valueMapExpr = valueMapping.newScalarExpression(stmt,stmt.getTableExpression(mapRangeVar));
            ScalarExpression valueExpr = valueTableID.newScalarExpression(stmt, stmt.getTableExpression(valueRangeVar));

            stmt.innerJoin(valueExpr,valueMapExpr,stmt.getTableExpression(valueRangeVar), true);
            
            qclKeyValues[1] = valueTableID.newScalarExpression(stmt,stmt.getTableExpression(valueRangeVar));
        }
        else
        {
            qclKeyValues[1] = valueMapping.newScalarExpression(stmt,stmt.getTableExpression(mapRangeVar));
        }
        return qclKeyValues;
    }
    
    /**
     * Used as part of the Querying of Maps where a containsValue() is used.
     * @param stmt The Query Statement 
     * @param ownerMapping Mapping for the owner
     * @param te Table Expression for the owner
     * @param mapRangeVar   The range variable for the "Map" table. 
     * @param filteredValueType The Class Type for the filtered value
     * @param valueRangeVar The SQL alias, or "range variable", to assign to the
     *                      expression or to the value table.
     * @return QueryColumnList with the columns from the value mapping 
     **/
    public ScalarExpression joinValuesTo(
                                QueryStatement stmt,
                                JavaTypeMapping ownerMapping,
                                TableExpression te,
                                SQLIdentifier mapRangeVar,
                                Class filteredValueType,
                                SQLIdentifier valueRangeVar)
    {
        ClassLoaderResolver clr=storeMgr.getClassLoaderResolver();
        if (!clr.isAssignableFrom(valueType,filteredValueType) &&
            !clr.isAssignableFrom(filteredValueType,valueType))
        {
            throw new IncompatibleQueryElementTypeException(valueType, filteredValueType.getName());
        }

        // Join the map table on the owner ID column
        stmt.newTableExpression(mapTable, mapRangeVar);

        ScalarExpression ownerExpr = ownerMapping.newScalarExpression(stmt,te);
        ScalarExpression ownerMapExpr = this.ownerMapping.newScalarExpression(stmt, stmt.getTableExpression(mapRangeVar));

        stmt.innerJoin(ownerExpr,ownerMapExpr,stmt.getTableExpression(mapRangeVar), true);

        if (!TypeManager.getTypeManager().isSupportedType(filteredValueType.getName()))
        {
            // Join the value table on the value ID column
            DatastoreClass valueTable=storeMgr.getDatastoreClass(filteredValueType.getName());
            JavaTypeMapping valueTableID=valueTable.getIDMapping();

            stmt.newTableExpression(valueTable,valueRangeVar);

            ScalarExpression valueMapExpr = valueMapping.newScalarExpression(stmt,stmt.getTableExpression(mapRangeVar));
            ScalarExpression valueExpr = valueTableID.newScalarExpression(stmt, stmt.getTableExpression(valueRangeVar));

            stmt.innerJoin(valueExpr,valueMapExpr,stmt.getTableExpression(valueRangeVar), true);

            return valueTableID.newScalarExpression(stmt,stmt.getTableExpression(valueRangeVar));
        }
        else
        {
            return valueMapping.newScalarExpression(stmt,stmt.getTableExpression(mapRangeVar));
        }
    }
       
}
