/**********************************************************************
Copyright (c) 2002 Kelly Grizzle (TJDO) and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2002 Mike Martin (TJDO)
2003 Andy Jefferson - added comments
2003 Erik Bengtson - removed unused import
2003 Erik Bengtson - fixed bug [832635] Imcompatiblities in casting
                    inherited classes in query
2003 Andy Jefferson - coding standards
2004 Andy Jefferson - updated retrieval of ColumnLists
2004 Marco Schulze  - replaced catch(NotPersistenceCapableException ...) by
                     advance-check via TypeManager.isSupportedType(...)
2004 Andy Jefferson - moved statements to AbstractSetStore
    ...
**********************************************************************/
package org.jpox.store.rdbms.scostore;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.Collection;
import java.util.Iterator;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOUnsupportedOptionException;

import org.jpox.ClassLoaderResolver;
import org.jpox.PersistenceManager;
import org.jpox.StateManager;
import org.jpox.TypeManager;
import org.jpox.store.DatastoreClass;
import org.jpox.store.DatastoreContainerObject;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.exceptions.IncompatibleQueryElementTypeException;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.TableExpression;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.Mappings;
import org.jpox.store.query.IteratorStatement;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.adapter.DatabaseAdapter;
import org.jpox.store.rdbms.mapping.RDBMSMapping;
import org.jpox.store.rdbms.sqlidentifier.SQLIdentifier;
import org.jpox.store.rdbms.sqlidentifier.TableIdentifier;
import org.jpox.store.rdbms.table.ClassTable;
import org.jpox.store.rdbms.table.SetTable;
import org.jpox.util.JPOXLogger;
import org.jpox.util.SQLWarnings;

/**
 * Representation of a Normal Set as part of a relationship. This class is
 * used where you have a 1-N and the tables are joined via a link table.
 * That is one table is the owner, and it has a link table to another table,
 * with the link table having 2 columns - the ids of the 2 tables.
 * This is in contrast to InverseSetStore which represents 1-N relationships
 * without using a link table (using an id in the other table).
 * <p>
 * For sets of primitive types (eg Date,String etc), the NormalSetStore is used,
 * but the 'link' table contains the id of the owner and the field(s) 
 * representing the primitive type.
 *
 * @version $Revision: 1.11 $
 **/
public class NormalSetStore extends AbstractSetStore
{
    /**
     * when the element mappings columns can't be part of the primary key
     * by datastore limitations like BLOB types.
     * An adapter mapping is used to be a kind of "index"
     */
    protected final JavaTypeMapping adapterMapping;

    /**
     * Constructor for the relationship representation.
     *
     * @param setTable The table for the link
     **/
    public NormalSetStore(SetTable setTable)
    {
        this.setTable = setTable;

        setName = "set";
        storeMgr = setTable.getStoreManager();
        dba = storeMgr.getDatastoreAdapter();

        ownerMapping = setTable.getOwnerMapping();
        elementMapping = setTable.getElementMapping();
        adapterMapping = setTable.getAdapterMapping();

        elementType = elementMapping.getType().getName();
        elementsAreEmbedded = isEmbeddedMapping(elementMapping);

        ClassLoaderResolver clr=storeMgr.getClassLoaderResolver();
        Class element_class=clr.classForName(elementType);
        if (element_class.isInterface())
        {
            JPOXLogger.JDO.warn(LOCALISER.msg("SCO.Set.InterfaceElementNotSupported", element_class.getName()));
            emd = storeMgr.getMetaDataManager().getMetaDataForInterface(element_class,null);
            if (emd != null)
            {
                elementType = element_class.getName();
                // TODO This currently just grabs the cmd of the first implementation. It needs to
                // get the cmds for all implementations, so we can have a handle to all possible elements.
                // This would mean changing the SCO classes to have multiple elementTable/elementMapping.
                elementTable = (ClassTable) storeMgr.getDatastoreClass(emd.getFullClassName());
            }
        }
        else
        {
            // Check if the element is FCO
            emd = storeMgr.getMetaDataManager().getMetaDataForClass(element_class);
            if (emd != null)
            {
                elementType = emd.getFullClassName();
                elementTable = (ClassTable) storeMgr.getDatastoreClass(elementType);
                elementDiscriminatorMapping = null; // We dont need this for Normal sets
            }
        }

        // Generate the statements
        initialiseStatements();
        addStmt = getAddStmt();
    }

    /**
     * Generate statement for obtaining the maximum id.
     * <PRE>
     * SELECT MAX(SCOID) FROM SETTABLE 
     * WHERE OWNERCOL=?
     * </PRE> 
     * @return The Statement returning the higher id
     */
    private String getMaxAdapterColumnIdStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("SELECT ");
        stmt.append(" MAX(" + ((Column)adapterMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString() + ") ");
        stmt.append(" FROM ");
        stmt.append(setTable.toString());
        stmt.append(" WHERE ");
        stmt.append(((Column)ownerMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
        stmt.append(" = ");
        stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(0)).getUpdateInputParameter());
        for (int i=1; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(" AND ");
            stmt.append(((Column)ownerMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            stmt.append(" = ");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getUpdateInputParameter());
        }
        return stmt.toString();
    }

    /**
     * Generate statement for adding an item to the Set.
     * <PRE> 
     * INSERT INTO SETTABLE (OWNERCOL, ELEMENTCOL) VALUES (?, ?)
     * </PRE> 
     * @return Statement for adding an item to the Set.
     */
    private String getAddStmt()
    {
        StringBuffer stmt = new StringBuffer();
        stmt.append("INSERT INTO ");
        stmt.append(setTable.toString());
        stmt.append(" (");
        stmt.append(((Column)ownerMapping.getDataStoreMapping(0).getDatastoreField()).getName().toString());
        for (int i=1; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(",");
            stmt.append(((Column)ownerMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
        }
        if (adapterMapping != null)
        {
            for (int i=0; i<adapterMapping.getNumberOfDatastoreFields(); i++)
            {
                stmt.append(",");
                stmt.append(((Column)adapterMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
            }
        }
        for (int i=0; i<elementMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(",");
            stmt.append(((Column)elementMapping.getDataStoreMapping(i).getDatastoreField()).getName().toString());
        }
        stmt.append(") VALUES (");
        stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(0)).getInsertionInputParameter());
        for (int i=1; i<ownerMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(",");
            stmt.append(((RDBMSMapping)ownerMapping.getDataStoreMapping(i)).getInsertionInputParameter());
        }
        if (adapterMapping != null)
        {
            for (int i=0; i<adapterMapping.getNumberOfDatastoreFields(); i++)
            {
                stmt.append(",");
                stmt.append(((RDBMSMapping)adapterMapping.getDataStoreMapping(i)).getInsertionInputParameter());
            }
        }        
        for (int i=0; i<elementMapping.getNumberOfDatastoreFields(); i++)
        {
            stmt.append(",");
            stmt.append(((RDBMSMapping)elementMapping.getDataStoreMapping(i)).getInsertionInputParameter());
        }
        stmt.append(") ");
        return stmt.toString();
    }

    /**
     * Generate statement for removing a collection of items from the Set.
     * <PRE>
     * DELETE FROM SETTABLE 
     * WHERE (OWNERCOL=? AND ELEMENTCOL=?) OR
     *      (OWNERCOL=? AND ELEMENTCOL=?) OR
     *      (OWNERCOL=? AND ELEMENTCOL=?)
     * </PRE>
     * @param elements Collection of elements to remove
     * @return Statement for deleting items from the Set.
     **/
    protected QueryStatement getRemoveAllStmt(StateManager sm, Collection elements)
    {
        if (elements == null || elements.size() == 0)
        {
            // TODO : throw an exception when null collection passed
            return null;
        }
        
        QueryStatement stmt = new QueryStatement(setTable, new SQLIdentifier(dba,""));
        
        Iterator iter=elements.iterator();
        while (iter.hasNext())
        {
            Object element = iter.next(); // Move to next element

            /*
             * Apply condition on join-table owner field to filter by owner
             */
            ScalarExpression ownerExpr = ownerMapping.newScalarExpression(stmt, stmt.getTableExpression(new SQLIdentifier(dba,"")));
            ScalarExpression ownerVal = ownerMapping.newLiteral(stmt, sm.getObject());

            /*
             * Apply condition on join-table owner field to filter by element
             */
            ScalarExpression elementExpr = elementMapping.newScalarExpression(stmt, stmt.getTableExpression(new SQLIdentifier(dba,"")));
            ScalarExpression elementVal = elementMapping.newLiteral(stmt, element);

            stmt.iorCondition(ownerExpr.eq(ownerVal).and(elementExpr.eq(elementVal).encloseWithInParentheses()), true);     
        }
        return stmt;
    }

    // -------------------- Overridden SetStore methods ------------------------
 
    /**
     * Remove all elements from a collection from the association owner vs
     * elements.
     * TODO : Change the query to do all in one go for efficiency.
     * @param sm State Manager for the container
     * @param elements Collection of elements to remove 
     * @return Whether the database was updated 
     */
    public boolean removeAll(StateManager sm, Collection elements)
    {
        if (elements == null || elements.size() == 0)
        {
            return false;
        }

        boolean modified = false;

        QueryStatement removeAllStmt=getRemoveAllStmt(sm,elements);

        try
        {
            // Validate elements for writing
            Iterator iter=elements.iterator();
            while (iter.hasNext())
            {
                Object element = iter.next();
                validateElementForWriting(sm, element);
            }

            PersistenceManager pm = sm.getPersistenceManager();
            Connection conn = pm.getConnection(true);
            PreparedStatement ps = removeAllStmt.toDeleteStatementText().prepareStatement(pm,conn);
            try
            {
                if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                {
                    JPOXLogger.RDBMS_SQL.debug(removeAllStmt.toDeleteStatementText().toString());
                }
                long startTime = System.currentTimeMillis();
                int rows_deleted = ps.executeUpdate();
                if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                {
                    JPOXLogger.RDBMS_SQL.debug(STORE_LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                }

                modified = (rows_deleted == 1);
            }
            finally
            {
                ps.close();
                pm.releaseConnection(conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(STORE_LOCALISER.msg("SCO.RemoveRequestFailed",removeAllStmt.toDeleteStatementText().toString()),e);
        }

        return modified;
    }

    /**
     * Adds one element to the association owner vs elements.
     * @param sm State Manager for the container.
     * @param element Element to add
     * @return Whether it was successful
     */
    public boolean add(StateManager sm, Object element)
    {
        validateElementForWriting(sm, element);

        boolean modified = false;
        PersistenceManager pm = sm.getPersistenceManager();

        try
        {
            Connection conn = pm.getConnection(true);
            try
            {
                PreparedStatement ps = conn.prepareStatement(addStmt);
                try
                {
                    ownerMapping.setObject(pm, ps, Mappings.getParametersIndex(1,ownerMapping), sm.getObject());
                    if (adapterMapping != null)
                    {
                        long nextIDAdapter = getNextIDForAdapterColumn(sm);
                        adapterMapping.setObject(pm, ps, Mappings.getParametersIndex(ownerMapping.getNumberOfDatastoreFields()+1,adapterMapping), new Long(nextIDAdapter));
                        elementMapping.setObject(pm, ps, Mappings.getParametersIndex(ownerMapping.getNumberOfDatastoreFields()+2,elementMapping), element);
                    }
                    else
                    {
                        elementMapping.setObject(pm, ps, Mappings.getParametersIndex(ownerMapping.getNumberOfDatastoreFields()+1,elementMapping), element);
                    }

                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(addStmt);
                    }
                    long startTime = System.currentTimeMillis();
                    ps.executeUpdate();
                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(STORE_LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                    }

                    modified = true;
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                pm.releaseConnection(conn);
            }
        }
        catch (SQLException e)
        {
            if (!contains(sm, element))
            {
                throw new JDODataStoreException(STORE_LOCALISER.msg("SCO.AddRequestFailed",addStmt),e);
            }
        }

        return modified;
    }
    
    /**
     * Adds all elements from a collection to the association container.
     * @param sm State Manager for the container.
     * @param elements Collection of elements to add
     * @return Whether it was successful
     */
    public boolean addAll(StateManager sm, Collection elements)
    {
        boolean modified = false;
        Iterator i = elements.iterator();

        if (i.hasNext())
        {
            PersistenceManager pm = sm.getPersistenceManager();
            Object element = null;
            try
            {
                Connection conn = pm.getConnection(true);
                PreparedStatement ps = conn.prepareStatement(addStmt);

                try
                {
                    long nextIDAdapter = 0;
                    if (adapterMapping != null)
                    {
                        nextIDAdapter = getNextIDForAdapterColumn(sm);
                    }
                    while (i.hasNext())
                    {
                        element = i.next();
                        validateElementForWriting(sm, element);
                        
                        ownerMapping.setObject(pm, ps, Mappings.getParametersIndex(1,ownerMapping), sm.getObject());
                        if (adapterMapping != null)
                        {
                            adapterMapping.setObject(pm, ps, Mappings.getParametersIndex(ownerMapping.getNumberOfDatastoreFields()+1,adapterMapping), new Long(nextIDAdapter));
                            elementMapping.setObject(pm, ps, Mappings.getParametersIndex(ownerMapping.getNumberOfDatastoreFields()+2,elementMapping), element);
                            nextIDAdapter++;
                        }
                        else
                        {
                            elementMapping.setObject(pm, ps, Mappings.getParametersIndex(ownerMapping.getNumberOfDatastoreFields()+1,elementMapping), element);
                        }

                        if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                        {
                            JPOXLogger.RDBMS_SQL.debug(addStmt);
                        }
                        long startTime = System.currentTimeMillis();
                        try
						{
                        	ps.executeUpdate();
						}
                        catch (SQLException e)
			            {
			                //TODO for a while silent ignore. The exception is expected for elements that are already
                        	//in the database, so we should check if the object is in the database before trying to insert
			            }
                        if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                        {
                            JPOXLogger.RDBMS_SQL.debug(STORE_LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime))+element);
                        }

                        modified = true;
                    }
                }
                finally
                {
                    ps.close();
                    pm.releaseConnection(conn);
                }
            }
            catch (SQLException e)
            {
                throw new JDODataStoreException(STORE_LOCALISER.msg("SCO.AddRequestFailed",addStmt),e);
            }
        }

        return modified;
    }

    // ---------------------------- Query Methods ------------------------------
 
    /**
     * Utility method to return a new QueryStatement for retrieval of the
     * elements of this Set.
     * @param sm StateManager for this object
     * @param candidateClass Class for the element end of the link.
     * @return The QueryStatement
     **/
    public QueryStatement newQueryStatement(StateManager sm, String candidateClass)
    {
        if (elementsAreEmbedded)
        {
            throw new JDOUnsupportedOptionException(STORE_LOCALISER.msg("SCO.QueryOverCollectionImpossible"));
        }

        TableIdentifier setRangeVar = new TableIdentifier(dba, setName);

        QueryStatement stmt = ((DatabaseAdapter) dba).newQueryStatement(setTable, setRangeVar);

         ScalarExpression ownerExpr = ownerMapping.newScalarExpression(stmt, stmt.getTableExpression(setRangeVar));
         ScalarExpression ownerVal = ownerMapping.newLiteral(stmt, sm.getObject());
         stmt.andCondition(ownerExpr.eq(ownerVal));

        if (!sm.getPersistenceManager().getClassLoaderResolver().isAssignableFrom(elementType,candidateClass))
        {
            throw new IncompatibleQueryElementTypeException(elementType, candidateClass);
        }

        /* 
         * Try to join to the element table and select the element ID. 
         * 
         * If the element class does not have a mapping to a table (e.g. - 
         * the Set has the element defined as Object or some other super 
         * class or interface that is not PersistenceCapable, but has 
         * PersistenceCapable implementations), just select the element 
         * ID from the join table.  The actual element table will be joined 
         * in later if it is casted to in the Query. 
         * Because the ClassNotPersistenceCapableException is not thrown anymore
         * with types that are natively supported by TypeManager, we check for 
         * it here.
         */
        if (!TypeManager.getTypeManager().isSupportedType(candidateClass))
        {
            TableIdentifier elementRangeVar = new TableIdentifier(dba, "this");

            DatastoreClass candidateTable = storeMgr.getDatastoreClass(candidateClass);
            JavaTypeMapping elementTableID = candidateTable.getIDMapping();

            stmt.newTableExpression(candidateTable, elementRangeVar);

            ScalarExpression elmExpr = elementMapping.newScalarExpression(stmt,stmt.getTableExpression(setRangeVar));
            ScalarExpression elmTblExpr = elementTableID.newScalarExpression(stmt,stmt.getTableExpression(elementRangeVar));

            stmt.innerJoin(elmExpr,elmTblExpr,stmt.getTableExpression(elementRangeVar), true);

            stmt.select(elementTableID);
        }
        else
        {
            stmt.select(setRangeVar, elementMapping);
        }

        return stmt;
    }

    /**
     * Utility for use in building a query, joining the element table and the
     * owner table.
     *
     * @param stmt The QueryStatement
     * @param ownerMapping Mapping for the owner
     * @param te Table Expression for the owner
     * @param setRangeVar The range variable for the "element" table
     * @param filteredElementType ?
     * @param elementRangeVar ?
     * @return A subquery for the provided query that joins the element table
     * to the owner table.
     * */
    public ScalarExpression joinElementsTo(
        QueryStatement stmt,
        JavaTypeMapping ownerMapping,
        TableExpression te,
        DatastoreIdentifier setRangeVar,
        Class filteredElementType,
        TableExpression elmte,
        DatastoreIdentifier elementRangeVar)
    {
        ClassLoaderResolver clr=storeMgr.getClassLoaderResolver();
        if (!clr.isAssignableFrom(elementType,filteredElementType) &&
            !clr.isAssignableFrom(filteredElementType,elementType))
        {
            throw new IncompatibleQueryElementTypeException(elementType, filteredElementType.getName());
        }

        // create a Join from the owner to the set table based on the owner ID column.
        /*
         * FROM OWNER THIS
         * INNER JOIN OWNER_ELEMENT THIS_SETTABLE 
         * ON THIS.OWNER_ID = THIS_SETTABLE.OWNER_OID 
         */
        stmt.newTableExpression(setTable, setRangeVar);

        ScalarExpression ownerExpr = ownerMapping.newScalarExpression(stmt,te);
        ScalarExpression ownerSetExpr = this.ownerMapping.newScalarExpression(stmt, stmt.getTableExpression(setRangeVar));

        stmt.innerJoin(ownerExpr,ownerSetExpr,stmt.getTableExpression(setRangeVar), true);

        if (!TypeManager.getTypeManager().isSupportedType(filteredElementType.getName()))
        {
            DatastoreClass elementTable = storeMgr.getDatastoreClass(filteredElementType.getName());
            DatastoreClass joiningClass = (elmte == null ? elementTable : (DatastoreClass)elmte.getMainTable());
            // Join the element table on the element ID column.
            JavaTypeMapping elementTableID = joiningClass.getIDMapping();
            /*
             * if elementType == filteredElementType
             *    INNER JOIN ELEMENT THIS_FILTER_ELEMENT_VARIABLE
             *    ON THIS_SETTABLE.ELEMENT_EID = THIS_FILTER_ELEMENT_VARIABLE.ELEMENT_ID
             * 
             * else if elementType != filteredElementType but filteredElementType is a superclass of elementType
             * 
             *    INNER JOIN FILTER_ELEMENT_TYPE THIS_FILTER_ELEMENT_VARIABLE
             *    ON THIS_SETTABLE.ELEMENT_EID = THIS_FILTER_ELEMENT_VARIABLE.FILTER_ELEMENT_TYPE_ID
             */
            stmt.newTableExpression(elementTable, elementRangeVar);
            
            ScalarExpression elmSetExpr = elementMapping.newScalarExpression(stmt,stmt.getTableExpression(setRangeVar));
            ScalarExpression elmExpr = elementTableID.newScalarExpression(stmt, stmt.getTableExpression(elementRangeVar));

            stmt.innerJoin(elmSetExpr,elmExpr,stmt.getTableExpression(elementRangeVar), true);
            
            return elementTableID.newScalarExpression(stmt,stmt.getTableExpression(elementRangeVar));
        }
        else
        {
            return elementMapping.newScalarExpression(stmt,stmt.getTableExpression(setRangeVar));
        }
    }

    protected QueryStatement getIteratorStatement(StateManager sm)
    {
        final ClassLoaderResolver clr=storeMgr.getClassLoaderResolver();
        QueryStatement stmt = new IteratorStatement(
                new IteratorStatement.AssociationEnd()
                {
                    public JavaTypeMapping getMapping()
                    {
                        return elementMapping;
                    }
                    public Class getType()
                    {
                        return clr.classForName(elementType);
                    }
                    public DatastoreContainerObject getDatastoreContainerObject()
                    {
                        return setTable;
                    }
                    public boolean useJoin()
                    {
                        return false;
                    }                    
                },              
                clr.classForName(this.elementType),
                this.storeMgr,
                this.dba,
                true).getQueryStatement();

        /*
         * Apply condition on join-table owner field to filter by owner
         */
        ScalarExpression ownerExpr = ownerMapping.newScalarExpression(stmt, stmt.getDefaultTableExpression());
        ScalarExpression ownerVal = ownerMapping.newLiteral(stmt, sm.getObject());

        stmt.andCondition(ownerExpr.eq(ownerVal), true);

        return stmt;
    }

    /**
     * Accessor for the higher id when elements primary key can't be part of
     * the primary key by datastore limitations like BLOB types can't be
     * primary keys.
     */
    private int getNextIDForAdapterColumn(StateManager sm)
    {
        int nextID;
        PersistenceManager pm = sm.getPersistenceManager();

        try
        {
            Connection conn = pm.getConnection(false);

            try
            {
                PreparedStatement ps = conn.prepareStatement(getMaxAdapterColumnIdStmt());

                try
                {
                    ownerMapping.setObject(pm, ps, Mappings.getParametersIndex(1,ownerMapping), sm.getObject());

                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(getMaxAdapterColumnIdStmt());
                    }
                    long startTime = System.currentTimeMillis();
                    ResultSet rs = ps.executeQuery();
                    if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_SQL.debug(STORE_LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
                    }

                    try
                    {
                        if (!rs.next())
                        {
                            nextID = 1;
                        }
                        else
                        {
                            nextID = rs.getInt(1)+1;
                        }

                        SQLWarnings.log(rs);
                    }
                    finally
                    {
                        rs.close();
                    }
                }
                finally
                {
                    ps.close();
                }
            }
            finally
            {
                pm.releaseConnection(conn);
            }
        }
        catch (SQLException e)
        {
            throw new JDODataStoreException(STORE_LOCALISER.msg("SCO.MaxAdapterColumnIdRequestFailed",getMaxAdapterColumnIdStmt()),e);
        }

        return nextID;
    }
}
