/**********************************************************************
Copyright (c) 2002 TJDO and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
    Erik Bengtson - refactored
    Andy Jefferson - coding standards
    Andy Jefferson - fix to case of UpperCase requested, but not supported
    Andy Jefferson - fix to conversion of names in setJavaName() when
                     user has selected PreserveCase
    Andy Jefferson - updates for use of quoted identifiers
    ...
**********************************************************************/
package org.jpox.store.rdbms.sqlidentifier;

import javax.jdo.JDOFatalInternalException;

import org.jpox.store.DatastoreAdapter;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.Dictionary;
import org.jpox.store.rdbms.adapter.DatabaseAdapter;
import org.jpox.store.rdbms.adapter.SQLConstants;

/**
 * Representation of an identifier in SQL. This can be something
 * like a table name, or column name etc.
 *
 * @version $Revision: 1.2 $
 **/
public class SQLIdentifier implements DatastoreIdentifier
{
    protected final DatastoreAdapter dba;

    protected String javaName;
    protected String sqlIdentifier;
    protected String identifierQuoteString;

    protected SQLIdentifier(DatastoreAdapter dba)
    {
        this.dba = dba;
        this.javaName = null;
        this.sqlIdentifier = null;
        this.identifierQuoteString = ((DatabaseAdapter)dba).getIdentifierQuoteString();
    }


    public SQLIdentifier(DatastoreAdapter dba, String sqlIdentifier)
    {
        this.dba = dba;
        this.javaName = null;
        this.identifierQuoteString = ((DatabaseAdapter)dba).getIdentifierQuoteString();
        setSQLIdentifier(sqlIdentifier);
    }

    public SQLIdentifier(DatastoreAdapter dba, String javaName,String sqlIdentifier)
    {
        this.dba = dba;
        this.identifierQuoteString = ((DatabaseAdapter)dba).getIdentifierQuoteString();
        
        // a defined sql identifier
        if (sqlIdentifier != null)
        {
            setJavaName(javaName);
            setSQLIdentifier(sqlIdentifier);
        }
        else
        {
            setJavaName(javaName);
        }
    }

    protected int getMaxLength()
    {
        return SQLConstants.MAX_IDENTIFIER_LENGTH;
    }

    private static final int HASH_RANGE = Character.MAX_RADIX * Character.MAX_RADIX / 2;

    /**
     * Sets the Java identifier to the given value and computes a corresponding
     * SQL identifier.
     *
     * <p>Conversion consists of breaking the identifier into words, converting
     * each word to upper-case, and separating each one with an underscore "_".
     * Words are identified by a leading upper-case character.
     * Any leading or trailing underscores are removed.
     *
     * <p>The resulting SQL identifier is "generic" and not suitable for direct
     * use in a SQL statement; the toString() method should be used for that.
     *
     * <p>The resulting SQL identifier also truncated to getMaxLength().
     * The excess characters plus the previous two characters are removed and
     * replaced with the base 36 representation of their hash value mod 1296 (36
     * * 36).
     *
     * @param javaName  the Java identifier.
     */
    protected void setJavaName(String javaName)
    {
        this.javaName = javaName;
        if (javaName == null)
        {
            return;
        }
        StringBuffer s = new StringBuffer();
        char prev = '\0';

        for (int i = 0; i < javaName.length(); ++i)
        {
            char c = javaName.charAt(i);

            if (c >= 'A' && c <= 'Z' &&
                dba.getDictionary().getCaseIdentifier() != Dictionary.CaseIdentifier.PRESERVECASE)
            {
                if (prev >= 'a' && prev <= 'z')
                {
                    s.append('_');
                }

                s.append(c);
            }
            else if (c >= 'A' && c <= 'Z' &&
                dba.getDictionary().getCaseIdentifier() == Dictionary.CaseIdentifier.PRESERVECASE)
            {
                s.append(c);
            }
            else if (c >= 'a' && c <= 'z' &&
                 dba.getDictionary().getCaseIdentifier() == Dictionary.CaseIdentifier.PRESERVECASE)
            {
                s.append(c);
            }
            else if (c >= 'a' && c <= 'z' &&
                 dba.getDictionary().getCaseIdentifier() != Dictionary.CaseIdentifier.PRESERVECASE)
            {
                s.append((char)(c - ('a' - 'A')));
            }
            else if (c >= '0' && c <= '9' || c == '_')
            {
                s.append(c);
            }
            else if (c == '.')
            {
                s.append('_');
            }
            else
            {
                String cval = "000" + Integer.toHexString(c);

                s.append(cval.substring(cval.length() - (c > 0xff ? 4 : 2)));
            }

            prev = c;
        }

        /* Remove leading and trailing underscores. */
        while (s.length() > 0 && s.charAt(0) == '_')
        {
            s.deleteCharAt(0);
        }
        while (s.length() > 0 && s.charAt(s.length() - 1) == '_')
        {
            s.deleteCharAt(s.length() - 1);
        }

        if (s.length() == 0)
        {
            throw new IllegalArgumentException("Illegal Java identifier: " + javaName);
        }

        setSQLIdentifier(truncate(s.toString(), getMaxLength()));
    }

    /**
     * Mutator for the sql identifier. Takes account of the DB policy on upper,
     * lower and mixed case identifiers.
     * @param sqlIdentifier The SQL Identifier
     **/
    protected void setSQLIdentifier(String sqlIdentifier)
    {
        sqlIdentifier = truncate(sqlIdentifier, getMaxLength());

        if (dba.getDictionary().getCaseIdentifier() == Dictionary.CaseIdentifier.LOWERCASE)
        {
            if (dba.storesLowerCaseIdentifiers() ||
                dba.storesLowerCaseQuotedIdentifiers())
            {
                this.sqlIdentifier = sqlIdentifier.toLowerCase();
            }
            else
            {
                this.sqlIdentifier = sqlIdentifier.toUpperCase();
            }
        }
        else if (dba.getDictionary().getCaseIdentifier() == Dictionary.CaseIdentifier.UPPERCASE)
        {
            if (dba.storesUpperCaseIdentifiers() || dba.storesMixedCaseIdentifiers() ||
                dba.storesUpperCaseQuotedIdentifiers() || dba.storesMixedCaseQuotedIdentifiers())
            {
                this.sqlIdentifier = sqlIdentifier.toUpperCase();
            }
            else if (dba.storesLowerCaseIdentifiers() ||
                     dba.storesLowerCaseQuotedIdentifiers())
            {
                this.sqlIdentifier = sqlIdentifier.toLowerCase();
            }
        }
        else if (dba.getDictionary().getCaseIdentifier() == Dictionary.CaseIdentifier.SENTENCECASE)
        {
            //TODO
            throw new JDOFatalInternalException("Sentence case is not yet implemented");
        }
        else if (dba.getDictionary().getCaseIdentifier() == Dictionary.CaseIdentifier.CAPITALIZECASE)
        {
            //TODO
            throw new JDOFatalInternalException("Capitalize case is not yet implemented");
        }
        else if (dba.getDictionary().getCaseIdentifier() == Dictionary.CaseIdentifier.PRESERVECASE)
        {
            if (dba.storesLowerCaseIdentifiers() ||
                dba.storesLowerCaseQuotedIdentifiers())
            {
                this.sqlIdentifier = sqlIdentifier.toLowerCase();
            }
            else if (dba.storesMixedCaseIdentifiers() ||
                     dba.storesMixedCaseQuotedIdentifiers())
            {
                this.sqlIdentifier = sqlIdentifier;
            }
            else
            {
                this.sqlIdentifier = sqlIdentifier.toUpperCase();
            }
        }
        
    }


    protected static String truncate(String sqlIdentifier, int length)
    {
        if (sqlIdentifier.length() > length)
        {
            int tailIndex = length - 2;
            int tailHash = sqlIdentifier.substring(tailIndex).hashCode();

            /* Scale the hash code down to the range 0 - 1295. */
            if (tailHash < 0)
            {
                tailHash = tailHash % HASH_RANGE + (HASH_RANGE - 1);
            }
            else
            {
                tailHash = tailHash % HASH_RANGE + HASH_RANGE;
            }

            String suffix = "0" + Integer.toString(tailHash, Character.MAX_RADIX);

            return sqlIdentifier.substring(0, tailIndex) + suffix.substring(suffix.length() - 2);
        }
        else
        {
            return sqlIdentifier;
        }
    }


    public String getJavaName()
    {
        if (javaName == null)
        {
            return getSQLIdentifier().toLowerCase();
        }
        return javaName;
    }


    public String getSQLIdentifier()
    {
        return sqlIdentifier;
    }


    public int hashCode()
    {
        return sqlIdentifier.hashCode();
    }


    public boolean equals(Object obj)
    {
        if (obj == this)
        {
            return true;
        }

        if (!(obj instanceof SQLIdentifier))
        {
            return false;
        }

        SQLIdentifier id = (SQLIdentifier)obj;
        return sqlIdentifier.equals(id.sqlIdentifier);
    }

    /**
     * Method to output the name of the identifier. This will be quoted where necessary.
     * @return The identifier name
     */
    public String toString()
    {
        // Note that by adding on the quotes here we assume that the limits on the size of
        // the SQLIdentifier are not including the quotes. This may or may not be correct
        if (((DatabaseAdapter)dba).isSQLKeyword(sqlIdentifier))
        {
            return identifierQuoteString + sqlIdentifier + identifierQuoteString;
        }
        else
        {
            if (dba.getDictionary().getCaseIdentifier() == Dictionary.CaseIdentifier.PRESERVECASE &&
                dba.storesMixedCaseQuotedIdentifiers())
            {
                return identifierQuoteString + sqlIdentifier + identifierQuoteString;
            }
            else if (dba.getDictionary().getCaseIdentifier() == Dictionary.CaseIdentifier.UPPERCASE &&
                dba.storesUpperCaseQuotedIdentifiers())
            {
                return identifierQuoteString + sqlIdentifier + identifierQuoteString;
            }
            else if (dba.getDictionary().getCaseIdentifier() == Dictionary.CaseIdentifier.LOWERCASE &&
                dba.storesLowerCaseQuotedIdentifiers())
            {
                return identifierQuoteString + sqlIdentifier + identifierQuoteString;
            }
            return sqlIdentifier;
        }
    }


    /* (non-Javadoc)
     * @see org.jpox.store.DatastoreIdentifier#getNewJavaName(java.lang.String)
     */
    public String getNewJavaName(String suffix)
    {
        return getJavaName()+'.'+suffix;
    }

    /* (non-Javadoc)
     * @see org.jpox.store.DatastoreIdentifier#getNewIdentifier(java.lang.String)
     */
    public DatastoreIdentifier getNewIdentifier(String suffix)
    {
        return new SQLIdentifier(dba, getSQLIdentifier() + '_' + suffix);
    }
    
}