/**********************************************************************
Copyright (c) 2002 Mike Martin and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution.

Contributors:
2002 Kelly Grizzle - unknown changes
2003 Andy Jefferson - added localiser
2003 Andy Jefferson - added output of logging when an exception is thrown
2004 Andy Jefferson - allowed updateable catalogName, schemaName
2004 Andy Jefferson - rewritten to remove levels of inheritance
2004 Andy Jefferson - added capability to add columns after initialisation.
    ...
**********************************************************************/
package org.jpox.store.rdbms.table;

import java.io.FileWriter;
import java.io.IOException;
import java.sql.Connection;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import org.jpox.PMFConfiguration;
import org.jpox.metadata.ColumnMetaData;
import org.jpox.metadata.ExtendableMetaData;
import org.jpox.store.DatastoreField;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.Role;
import org.jpox.store.StoreManager;
import org.jpox.store.exceptions.DuplicateColumnNameException;
import org.jpox.store.exceptions.MissingTableException;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.store.rdbms.RDBMSStoreHelper;
import org.jpox.store.rdbms.adapter.DatabaseAdapter;
import org.jpox.store.rdbms.sqlidentifier.ColumnIdentifier;
import org.jpox.store.rdbms.sqlidentifier.SQLIdentifier;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;
import org.jpox.util.SQLWarnings;

/**
 * Abstract implementation of a table in the datastore.
 * The table exists in various states.
 * After initialisation it can be created in the datastore by calling <B>create</B>. 
 * At any point after initialisation it can be modified, but only by addition of columns.
 * The table can be dropped from the datastore by calling <B>drop</B>.
 *
 * @version $Revision: 1.7 $
 **/
public abstract class AbstractTable implements Table
{
    /** Localiser for messages. */
    protected static final Localiser LOCALISER=Localiser.getInstance("org.jpox.store.rdbms.table.Localisation");

    /** Manager for this table. */
    protected final RDBMSManager storeMgr;

    /** Database Adapter being used. */
    protected final DatabaseAdapter dba;

    /** Catalog name for the table. */
    protected String catalogName;

    /** Schema name for the table. */
    protected String schemaName;

    /** Name for the table. */
    protected SQLIdentifier name;

    /** State of the table */
    protected int state = TABLE_STATE_NEW;

    /** Columns for this table. */
    protected List columns = new ArrayList();

    /** Index to the columns, keyed by name. */
    protected HashMap columnsByName = new HashMap();

    /** Writer to output any DDL. TODO Refactor this out and provide cleaner interface to DDL */
    protected static FileWriter ddlDumpWriter = null;

    /**
     * Constructor.
     * @param storeMgr The Store Manager.
     **/
    public AbstractTable(RDBMSManager storeMgr)
    {
        this(null, storeMgr);
    }

    /**
     * Constructor taking the table name and the RDBMSManager managing this
     * table.
     * @param name Name of the table
     * @param storeMgr The RDBMS Manager
     */
    public AbstractTable(SQLIdentifier name, RDBMSManager storeMgr)
    {
        this.storeMgr = storeMgr;
        this.dba = storeMgr.getDatabaseAdapter();
        this.name = name;
    }

    /**
     * Accessor for whether the table is initialised.
     * @return Whether it is initialised
     */
    public boolean isInitialized()
    {
        // All of the states from initialised onwards imply that it has (at some time) been initialised
        return state >= TABLE_STATE_INITIALIZED;
    }

    /**
     * Accessor for whether the primary key of the table is initialised.
     * @return Whether the primary key of the table is initialised
     */
    public boolean isPKInitialized()
    {
        return state >= TABLE_STATE_PK_INITIALIZED;
    }

    /**
     * Accessor for whether the table is validated.
     * @return Whether it is validated.
     */
    public boolean isValidated()
    {
        return state == TABLE_STATE_VALIDATED;
    }
    
    /**
     * Accessor for whether the table has been modified since initialisation.
     * @return Whether it is modified since initialisation.
     */
    public boolean isInitializedModified()
    {
        return state == TABLE_STATE_INITIALIZED_MODIFIED;
    }

    /**
     * Accessor for the Store Manager.
     * @return Store Manager
     **/
    public StoreManager getStoreManager()
    {
        return storeMgr;
    }

    /**
     * Accessor for the Catalog Name.
     * @return Catalog Name
     **/
    public String getCatalogName()
    {
        if (catalogName == null)
        {
            // Nothing set, so try "default" catalog name
            PMFConfiguration pmfConfig=storeMgr.getPMFContext().getPmfConfiguration();
            if (pmfConfig.getDefaultCatalogName() != null)
            {
                catalogName = pmfConfig.getDefaultCatalogName();
            }
            // Next try the store managers catalog name
            else
            {
                catalogName = storeMgr.getCatalogName();
            }
        }
        if (catalogName != null)
        {
            return catalogName;
            //dba.getIdentifierInAdapterCase(catalogName);
        }
        return null;
    }

    /**
     * Accessor for the Schema Name.
     * @return Schema Name
     **/
    public String getSchemaName()
    {
        if (schemaName == null)
        {
            // Nothing set, so try "default" schema name
            PMFConfiguration pmfConfig=storeMgr.getPMFContext().getPmfConfiguration();
            if (pmfConfig.getDefaultSchemaName() != null)
            {
                schemaName = pmfConfig.getDefaultSchemaName();
            }
            // Next try the store managers schema name
            else
            {
                schemaName = storeMgr.getSchemaName();
            }
        }
        if (schemaName != null)
        {
            return schemaName;
            //dba.getIdentifierInAdapterCase(schemaName);
        }
        return null;
    }

    /**
     * Accessor for the SQL identifier (the table name).
     * @return The name
     **/
    public DatastoreIdentifier getName()
    {
        return name;
    }

    /**
     * Creates a new column.
     * @param type the type of the field that this represents.
     * @param name the SQL identifier
     * @param role the Role for column (e.g FK?)
     * @param mapping the mapping for this column
     * @param colmd the ExtendableMetadata (ColumnMetaData for RDBMS)
     * @return the new Column
     * @throws DuplicateColumnNameException if a column already exists with same name
     */
    public synchronized DatastoreField addDatastoreField(Class type, DatastoreIdentifier name,
                                		 Role role, JavaTypeMapping mapping, ExtendableMetaData colmd)
    {
        // Create the column, and add to the internal representation
        ColumnIdentifier columnName = new ColumnIdentifier(dba, ((SQLIdentifier) name).getSQLIdentifier(), type, role);
        Column col = new Column(this, type, columnName, mapping, (ColumnMetaData) colmd);
        addColumnInternal(col);

        if (isInitialized())
        {
            // Set state to modified
            state = TABLE_STATE_INITIALIZED_MODIFIED;
        }

        return col;
    }

    /**
     * Method to create this table.
     * @param conn Connection to the datastore.
     * @throws SQLException Thrown if an error occurs creating the table.
     **/
    public void create(Connection conn)
    throws SQLException
    {
        assertIsInitialized();

        if (JPOXLogger.RDBMS.isInfoEnabled())
        {
            JPOXLogger.RDBMS.info(LOCALISER.msg("Table.Creating", this));
        }

        executeDdlStatementList(getSQLCreateStatements(), conn);
    }

    /**
     * Method to drop this table.
     * @param conn Connection to the datastore.
     * @throws SQLException Thrown if an error occurs dropping the table.
     **/
    public void drop(Connection conn)
    throws SQLException
    {
        assertIsInitialized();

        if (JPOXLogger.RDBMS.isInfoEnabled())
        {
            JPOXLogger.RDBMS.info(LOCALISER.msg("Table.Dropping", this));
        }

        executeDdlStatementList(getSQLDropStatements(), conn);
    }

    /**
     * Method to check the existence of the table/view, optionally auto creating it
     * where required. If it doesn't exist and auto creation isn't specified this
     * throws a MissingTableException.
     * @param conn The JDBC Connection
     * @param auto_create Whether to auto create the table if not existing
     * @return Whether the table was added
     * @throws SQLException Thrown when an error occurs in the JDBC calls
     */
    public boolean exists(Connection conn,boolean auto_create) 
    throws SQLException
    {
        assertIsInitialized();

        int tableType = RDBMSStoreHelper.getTableType(storeMgr,name.getSQLIdentifier(),conn);
        if (tableType == TABLE_TYPE_MISSING)
        {
            if (!auto_create)
            {
                throw new MissingTableException(this.toString());
            }

            create(conn);
            state = TABLE_STATE_VALIDATED;
            return true;
        }

        return false;
    }

    /**
     * Equality operator.
     * @param obj The object to compare against
     * @return Whether the objects are equal
     **/
    public final boolean equals(Object obj)
    {
        if (obj == this)
        {
            return true;
        }

        if (!(obj instanceof AbstractTable))
        {
            return false;
        }

        AbstractTable t = (AbstractTable)obj;

        return getClass().equals(t.getClass()) && name.equals(t.name) && storeMgr.equals(t.storeMgr);
    }

    /**
     * Accessor for the hascode of this table.
     * @return The hash code.
     **/
    public final int hashCode()
    {
        return name.hashCode() ^ storeMgr.hashCode();
    }

    /**
     * Method to return a string version of this table. This name is the
     * fully-qualified name of the table (including catalog/schema names where
     * appropriate).
     * @return String name of the table (catalog.schema.table)
     */
    public final String toString()
    {
        // Check whether the adapter supports the fully qualified name
        boolean supports_catalog_name=storeMgr.getDatabaseAdapter().supportsCatalogsInTableDefinitions();
        boolean supports_schema_name=storeMgr.getDatabaseAdapter().supportsSchemasInTableDefinitions();

        String cat_name=getCatalogName();
        String sch_name=getSchemaName();
        String separator=storeMgr.getDatabaseAdapter().getCatalogSeparator();

        StringBuffer fullQualifiedName = new StringBuffer();

        // Qualify the table name with the catalog/schema depending on the users PMF
        String default_catalog_name=storeMgr.getPMFContext().getPmfConfiguration().getDefaultCatalogName();
        String default_schema_name=storeMgr.getPMFContext().getPmfConfiguration().getDefaultSchemaName();

        if ((default_schema_name != null && supports_schema_name) ||
            (default_catalog_name != null && supports_catalog_name))
        {
            // Default schema/catalog names supplied
            // User-supplied Catalog name
            if (supports_catalog_name && default_catalog_name != null)
            {
                fullQualifiedName.append(default_catalog_name);
                fullQualifiedName.append(separator);
            }
            // User-supplied Schema name
            if (supports_schema_name && default_schema_name != null)
            {
                fullQualifiedName.append(default_schema_name);
                fullQualifiedName.append(separator);
            }
        }
        else if (storeMgr.getPMFContext().getPmfConfiguration().getFullyQualifiedNames())
        {
            // No defaults, but Fully Qualified names are required
            // Catalog name
	        if (supports_catalog_name &&
                cat_name != null && cat_name.length() > 0)
	        {
	            fullQualifiedName.append(cat_name);
	            fullQualifiedName.append(separator);
	        }
            // Schema name
	        if (supports_schema_name &&
                sch_name != null && sch_name.length() > 0)
	        {
	            fullQualifiedName.append(sch_name);
	            fullQualifiedName.append(separator);
	        }
        }

        fullQualifiedName.append(name.toString());

        return fullQualifiedName.toString();
    }

    // -------------------------------- Internal Implementation ---------------------------
    
    /**
     * Utility method to add a column to the internal representation
     * @param col The column
     */
    protected synchronized void addColumnInternal(Column col)
    {
        DatastoreIdentifier colName = col.getName();
        if (hasColumnName(colName))
        {
            throw new DuplicateColumnNameException(this.toString(), col.getName());
        }

        columns.add(col);
        columnsByName.put(col.getName(), col);
    }
    
    /**
     * Utility to return if a column of this name exists.
     * @param colName The column name
     * @return Whether the column of this name exists
     */
    protected boolean hasColumnName(DatastoreIdentifier colName)
    {
        return columnsByName.get(colName) != null;
    }

    /**
     * Accessor for the SQL create statements.
     * @return The SQL Create statements
     **/
    protected abstract List getSQLCreateStatements();

    /**
     * Accessor for the SQL drop statements.
     * @return The SQL Drop statements
     **/
    protected abstract List getSQLDropStatements();

    protected void assertIsPKUninitialized()
    {
        if (isPKInitialized())
        {
            throw new IllegalStateException(LOCALISER.msg("Table.AlreadyInitialisedError",this));
        }
    }

    protected void assertIsUninitialized()
    {
        if (isInitialized())
        {
            throw new IllegalStateException(LOCALISER.msg("Table.AlreadyInitialisedError",this));
        }
    }

    protected void assertIsInitialized()
    {
        if (!isInitialized())
        {
            throw new IllegalStateException(LOCALISER.msg("Table.UninitialisedError",this));
        }
    }

    protected void assertIsInitializedModified()
    {
        if (!isInitializedModified())
        {
            throw new IllegalStateException(LOCALISER.msg("Table.UnmodifiedError",this));
        }
    }
    
    protected void assertIsPKInitialized()
    {
        if (!isPKInitialized())
        {
            throw new IllegalStateException(LOCALISER.msg("Table.UninitialisedError",this));
        }
    }

    protected void assertIsValidated()
    {
        if (!isValidated())
        {
            throw new IllegalStateException(LOCALISER.msg("Table.NotValidatedError",this));
        }
    }
    
    /**
     * Determine whether we or our concrete class allow DDL to be written into a file instead of
     * sending it to the DB. Defaults to true.
     * @return Whether it allows DDL dumping
     */
    protected boolean allowDdlDump()
    {
        return true;
    }
    
    public static boolean isDumpingDdl()
    {
        return ddlDumpWriter != null;
    }

    /**
     * Set the file where any subsequent DDL SQL commands will be written to
     * instead of issuing them to the database. The default null means normal
     * operation, i.e. DDL SQL commands will be sent to the DB.
     * @param fileWriter The File writer to use for outputting DDL
     */
    public static void setDdlDumpWriter(FileWriter fileWriter)
    {
        ddlDumpWriter = fileWriter;
    }

    /**
     * Method to perform the required SQL statements.
     * @param stmts A List of statements
     * @param conn The Connection to the datastore
     * @throws SQLException Any exceptions thrown by the statements 
     **/
    protected void executeDdlStatementList(List stmts, Connection conn)
    throws SQLException
    {
        Statement stmt = conn.createStatement();

        String stmtText=null;
        try
        {
            Iterator i = stmts.iterator();
            while (i.hasNext())
            {
                stmtText = (String)i.next();
                executeDdlStatement(stmt, stmtText);
            }
        }
        catch (SQLException sqe)
        {
            JPOXLogger.RDBMS.error(LOCALISER.msg("Table.Error",stmtText,sqe));
            throw sqe;
        }
        finally
        {
            stmt.close();
        }
    }

    /**
     * Execute a single DDL SQL statement with appropriate logging. 
     * If ddlDumpWriter is set, do not actually execute the SQL but write it to that FileWriter.
     * @param stmt The JDBC Statement object to execute on
     * @param stmtText The actual SQL statement text
     * @throws SQLException Thrown if an error occurs
     * @see #setDdlDumpWriter(FileWriter)
     */
    protected void executeDdlStatement(Statement stmt, String stmtText) 
    throws SQLException
    {
        if (JPOXLogger.RDBMS_DDL.isDebugEnabled())
        {
            JPOXLogger.RDBMS_DDL.debug(stmtText);
        }
        long startTime = System.currentTimeMillis();

        if (ddlDumpWriter != null && allowDdlDump())
        {
            try
            {
                // we shall write the SQL to a file instead of issuing to DB
                ddlDumpWriter.write(stmtText + ";\n\n");
            }
            catch (IOException e)
            {
                JPOXLogger.RDBMS_DDL.error("error writing DDL into file", e);
            }
        }
        else
        {
            stmt.execute(stmtText);
        }

        if (JPOXLogger.RDBMS_DDL.isDebugEnabled())
        {
            JPOXLogger.RDBMS_DDL.debug(LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
        }

        SQLWarnings.log(stmt);
    }
}