/**********************************************************************
Copyright (c) 2002 Kelly Grizzle (TJDO) and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - added localiser
2003 Andy Jefferson - replaced TableMetadata with identifier.
2003 Erik Bengtson - refactored the datastore identity together with the SQLIdentifier class
2003 Erik Bengtson - added OptimisticMapping
2003 Andy Jefferson - coding standards
2004 Andy Jefferson - merged with JDOBaseTable
2004 Andy Jefferson - changed to store map of fieldMappings keyed by absolute field num. Added start point for inheritance strategy handling
2004 Andy Jefferson - changed consumer to set highest field number based on actual highest (to allow for inheritance strategies)
2004 Andy Jefferson - added DiscriminatorMapping
2004 Andy Jefferson - enabled use of Long/String datastore identity column
2004 Andy Jefferson - added capability to handle 1-N inverse unidirectional FKs
2004 Andy Jefferson - removed the majority of the value-strategy code - done elsewhere
 	...
 **********************************************************************/
package org.jpox.store.rdbms.table;

import java.sql.Timestamp;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOFatalUserException;
import javax.jdo.JDOUnsupportedOptionException;
import javax.jdo.JDOUserException;
import javax.jdo.spi.PersistenceCapable;

import org.jpox.PMFConfiguration;
import org.jpox.StateManager;
import org.jpox.TypeManager;
import org.jpox.metadata.ClassMetaData;
import org.jpox.metadata.ColumnMetaData;
import org.jpox.metadata.DiscriminatorMetaData;
import org.jpox.metadata.DiscriminatorStrategy;
import org.jpox.metadata.EmbeddedMetaData;
import org.jpox.metadata.FieldMetaData;
import org.jpox.metadata.FieldPersistenceModifier;
import org.jpox.metadata.ForeignKeyDeleteAction;
import org.jpox.metadata.ForeignKeyMetaData;
import org.jpox.metadata.IdentityMetaData;
import org.jpox.metadata.IdentityStrategy;
import org.jpox.metadata.IdentityType;
import org.jpox.metadata.IndexMetaData;
import org.jpox.metadata.InheritanceMetaData;
import org.jpox.metadata.InheritanceStrategy;
import org.jpox.metadata.JoinMetaData;
import org.jpox.metadata.NullValue;
import org.jpox.metadata.OrderMetaData;
import org.jpox.metadata.UniqueMetaData;
import org.jpox.metadata.VersionMetaData;
import org.jpox.metadata.VersionStrategy;
import org.jpox.store.DatastoreClass;
import org.jpox.store.DatastoreField;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.OID;
import org.jpox.store.Role;
import org.jpox.store.StoreManager;
import org.jpox.store.exceptions.ClassDefinitionException;
import org.jpox.store.exceptions.NoSuchPersistentFieldException;
import org.jpox.store.mapping.DatastoreMapping;
import org.jpox.store.mapping.DiscriminatorMapping;
import org.jpox.store.mapping.EmbeddedPCMapping;
import org.jpox.store.mapping.IndexMapping;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.MappingCallbacks;
import org.jpox.store.mapping.MappingConsumer;
import org.jpox.store.mapping.OIDMapping;
import org.jpox.store.mapping.VersionMapping;
import org.jpox.store.mapping.PersistenceCapableMapping;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.store.rdbms.key.CandidateKey;
import org.jpox.store.rdbms.key.ForeignKey;
import org.jpox.store.rdbms.key.Index;
import org.jpox.store.rdbms.key.PrimaryKey;
import org.jpox.store.rdbms.sqlidentifier.CandidateKeyIdentifier;
import org.jpox.store.rdbms.sqlidentifier.ColumnIdentifier;
import org.jpox.store.rdbms.sqlidentifier.SQLIdentifier;
import org.jpox.util.AIDUtils;
import org.jpox.util.JPOXLogger;
import org.jpox.util.MultiMap;

/**
 * Table representing a Java class (or classes) as a first class object (FCO).
 * Uses the inheritance strategy to control whether this represents multiple classes
 * or just the one class.
 * <H3>Mappings</H3>
 * A Table is built from a series of field mappings. Each Java class has a series of
 * fields and each of these has an associated JavaTypeMapping. Each JavaTypeMapping has
 * related DatastoreMapping(s). These are used in mapping the Java class to the table,
 * and are used when populating the table, and when retrieving data from the table back
 * to the object. There are several categories of mappings
 * <UL>
 * <LI><B>fieldMappingsMap</B> - the set of mappings relating to the fields in the
 * class. The mappings are keyed by the FieldMetaData of the field. Any embedded field
 * will have a single mapping here of type EmbeddedPCMapping, with a set of datastore
 * mappings attached.</LI>
 * <LI><B>datastoreIDMapping</B> - the Identity mapping when using "datastore identity"</LI>
 * <LI><B>pkMappings</B> - the mappings for the primary key column(s).</LI>
 * <LI><B>externalFkMappings</B> - any mappings for Collections that have no associated
 * field in this class</LI>
 * <LI><B>orderMappings</B> - mappings for any ordering column used by Lists for ordering
 * elements of this class</LI>
 * <LI><B>discriminatorMapping</B> - mapping for any discriminator column. This is only
 * used where classes share this table and some of them use "superclass-table" strategy</LI>
 * <LI><B>versionMapping</B> - mapping for any versioning column</LI>
 * </UL>
 * <H3>Classes</H3>
 * A table can represent multiple classes. It has a nominal owner which is the class
 * that has an inheritance strategy of "new-table". All classes that utilise this table
 * have their MetaData stored in this object.
 *
 * @version $Revision: 1.54 $
 **/
public class ClassTable extends TableImpl implements FCOTable, DatastoreClass 
{
    /** 
     * MetaData for the principal class being stored here.
     * In inheritance situations where multiple classes share the same table, this will
     * be the class which uses "new-table" strategy.
     */
    private final ClassMetaData cmd;

    /** MetaData for all classes being managed here. */
    private final Collection managedClassMetaData = new HashSet();

    /** Table above this table storing superclass information (if any). */
    private ClassTable supertable;

    /**
     * Mappings for fields mapped to this table. 
     * Field mappings are keyed by the absolute field number.
     */
    private Map fieldMappingsMap = new HashMap();

    /** Mapping for datastore identity (optional). */
    private JavaTypeMapping datastoreIDMapping;

    /** Mappings for application identity (optional). */
    private JavaTypeMapping[] pkMappings;

    /**
     * Mappings for Collections/Lists not managed by this class.
     * Keyed by the FieldMetaData of the owner field.
     */
    private HashMap externalFkMappings;

    /**
     * Mappings for (Inverse) Lists managed by this class.
     * Keyed by the FieldMetaData of the owner field.
     */
    private HashMap orderMappings;

    /** Mapping for any version/timestamp column. */
    private JavaTypeMapping versionMapping;

    /** Mapping for any discriminator column. */
    private JavaTypeMapping discriminatorMapping;

    /** Highest absolute field number managed by this table */
    private int highestFieldNumber=0;

    /** All callbacks for class tables waiting to be performed. */
    private static MultiMap callbacks = new MultiMap();

    HashMap candidateKeysByMapField = new HashMap();

    /**
     * Constructor.
     * @param tableName Table name SQL identifier
     * @param cmd MetaData for the class.
     * @param storeMgr Store Manager to manage this table
     */
    public ClassTable(SQLIdentifier tableName, ClassMetaData cmd, RDBMSManager storeMgr)
    {
        super(tableName, storeMgr);

        this.cmd = cmd;

        // Check for valid class to map to a table
        if (cmd.getInheritanceMetaData().getStrategyValue() != InheritanceStrategy.NEW_TABLE)
        {
            throw new JDOFatalUserException(LOCALISER.msg("Table.ClassHasNoTableError", cmd.getFullClassName(),
                cmd.getInheritanceMetaData().getStrategyValue().toString()));
        }
        else if (cmd.getIdentityType() == IdentityType.NONDURABLE)
        {
            throw new JDOUnsupportedOptionException(LOCALISER.msg("Table.NondurableIdentityUnsupported", cmd.getFullClassName()));
        }
        else if (cmd.getIdentityType() != IdentityType.APPLICATION && cmd.getIdentityType() != IdentityType.DATASTORE)
        {
            throw new JDOFatalInternalException(LOCALISER.msg("Table.InvalidIdentityType", cmd.getFullClassName()));
        }

        highestFieldNumber = cmd.getNoOfManagedFields() + cmd.getNoOfInheritedManagedFields();
    }

    /**
     * Method to initialise the table. This adds the columns based on the
     * MetaData representation for the class being represented by this table.
     */
    public void initialize()
    {
        assertIsUninitialized();

        // Find the super table in any inheritance hierarchy (navigates to find any table above this)
        supertable = getSupertable(cmd);

        // Initialise the PK field(s)
        if (!isPKInitialized())
        {
            initializePK();
        }

        // Add the fields for this class (and any other superclasses that we need to manage the
        // fields for (inheritance-strategy="no-table" in the superclass)
        initializeForClass(cmd);

        // Add Version where specified in MetaData
        VersionMetaData vmd = cmd.getVersionMetaData();
        if (vmd != null)
        {
            if (vmd.getVersionStrategy() == VersionStrategy.VERSION_NUMBER)
            {
                versionMapping = new VersionMapping(dba, this, Long.class);
            }
            else if (vmd.getVersionStrategy() == VersionStrategy.DATE_TIME)
            {
                versionMapping = new VersionMapping(dba, this, Timestamp.class);
            }
        }

        // Add Discriminator where specified in MetaData
        InheritanceMetaData imd = cmd.getInheritanceMetaData();
        if (imd != null && imd.getStrategyValue() == InheritanceStrategy.NEW_TABLE)
        {
            DiscriminatorMetaData dismd = imd.getDiscriminatorMetaData();
            if (dismd != null)
            {
                if (dismd.getStrategy() == DiscriminatorStrategy.CLASS_NAME)
                {
                    discriminatorMapping = new DiscriminatorMapping(dba, this, String.class);
                }
                else if (dismd.getStrategy() == DiscriminatorStrategy.VALUE_MAP)
                {
                    ColumnMetaData disColmd = dismd.getColumnMetaData();
                    if (disColmd != null && disColmd.getJdbcType() != null)
                    {
                        if (disColmd.getJdbcType().equals("INTEGER") ||
                            disColmd.getJdbcType().equals("BIGINT") ||
                            disColmd.getJdbcType().equals("NUMERIC"))
                        {
                            discriminatorMapping = new DiscriminatorMapping(dba, this, long.class);
                        }
                        else
                        {
                            discriminatorMapping = new DiscriminatorMapping(dba, this, String.class);
                        }
                    }
                    else
                    {
                        discriminatorMapping = new DiscriminatorMapping(dba, this, String.class);
                    }
                }
            }
        }

        state = TABLE_STATE_INITIALIZED;

        runCallBacks();
    }

    /**
     * Method to give management of the persistence of the fields of the specified class
     * to this table.
     * @param theCmd ClassMetaData for the class to be managed
     */
    public void manageClass(ClassMetaData theCmd)
    {
        JPOXLogger.RDBMS.debug(LOCALISER.msg("Table.ManagingClassFields", toString(), theCmd.getFullClassName(), 
            theCmd.getInheritanceMetaData().getStrategyValue().toString()));
        managedClassMetaData.add(theCmd);

        // Go through the fields for this class and add columns for them.
        int fieldCount = theCmd.getNoOfManagedFields();
        for (int relativeFieldNumber=0; relativeFieldNumber<fieldCount; ++relativeFieldNumber)
        {
            FieldMetaData fmd = theCmd.getManagedField(relativeFieldNumber);
            // Primary key fields are added by the initialisePK method.
            if (!fmd.isPrimaryKey())
            {
                if (fmd.getPersistenceModifier() == FieldPersistenceModifier.PERSISTENT)
                {
                    if (fmd.getMappedBy() != null)
                    {
                        // 1-1 bidirectional with FK at other side
                    }
                    JavaTypeMapping fieldMapping = dba.getMappingManager().getMapping(this, fmd, dba);
                    int absoluteFieldNumber = theCmd.getAbsoluteFieldNumberForRelativeFieldNumber(relativeFieldNumber);
                    addFieldMapping(absoluteFieldNumber, fieldMapping);
                }
                else if (fmd.getPersistenceModifier() != FieldPersistenceModifier.TRANSACTIONAL)
                {
                    throw new JDOFatalInternalException(LOCALISER.msg("Table.InvalidPersistenceModifierForField",fmd.getName()));
                }

                // Update the other side of any relationship that links back with a FK
                if (fmd.getContainer() != null && fmd.getJoinMetaData() == null)
                {
                    // 1-N uni/bidirectional using FK
                    if (fmd.getCollection() != null)
                    {
                        ClassMetaData elementCmd = storeMgr.getMetaDataManager().getMetaDataForClass(fmd.getCollection().getElementType(), 
                                                                                                     storeMgr.getClassLoaderResolver());
                        if (elementCmd == null)
                        {
                            // Interface elements will come through here and java.lang.String and others as well
                            JPOXLogger.METADATA.warn(LOCALISER.msg("Table.Collection.ElementHasNoMetaData", 
                                theCmd.getFullClassName(), fmd.getCollection().getElementType()));
                        }
                        else
                        {
	                        ClassMetaData[] elementCmds = null;
	                        // TODO : Cater for interface elements, and get the metadata for the implementation classes here
	                        if (elementCmd.getInheritanceMetaData().getStrategyValue() == InheritanceStrategy.NO_TABLE)
	                        {
	                            elementCmds = storeMgr.getClassesManagingTableForClass(elementCmd);
	                        }
	                        else
	                        {
	                            elementCmds = new ClassMetaData[1];
	                            elementCmds[0] = elementCmd;
	                        }
	
	                        // Run callbacks for each of the element classes.
	                        for (int i=0;i<elementCmds.length;i++)
	                        {
	                            callbacks.put(elementCmds[i].getFullClassName(), new CallBack(fmd));
	                            DatastoreClass dc = storeMgr.getDatastoreClass(elementCmds[i].getFullClassName());
	                            ClassTable ct = (ClassTable) dc;
	                            if (ct.isInitialized())
	                            {
	                                // if the target table is already initialized, run the callbacks
	                                ct.runCallBacks();
	                            }
	                        }
                        }
                    }
                    else if (fmd.getMap() != null)
                    {
                        ClassMetaData valueCmd = storeMgr.getMetaDataManager().getMetaDataForClass(fmd.getMap().getValueType(), 
                                                                                                   storeMgr.getClassLoaderResolver());
                        if (valueCmd == null)
                        {
                            // Interface elements will come through here and java.lang.String and others as well
                            JPOXLogger.METADATA.warn(LOCALISER.msg("Table.Map.ValueHasNoMetaData", 
                                theCmd.getFullClassName(), fmd.getMap().getValueType()));
                        }
                        else
                        {
	                        ClassMetaData[] valueCmds = null;
	                        // TODO : Cater for interface values, and get the metadata for the implementation classes here
	                        if (valueCmd.getInheritanceMetaData().getStrategyValue() == InheritanceStrategy.NO_TABLE)
	                        {
	                            valueCmds = storeMgr.getClassesManagingTableForClass(valueCmd);
	                        }
	                        else
	                        {
	                            valueCmds = new ClassMetaData[1];
	                            valueCmds[0] = valueCmd;
	                        }
	
	                        // Run callbacks for each of the value classes.
	                        for (int i=0;i<valueCmds.length;i++)
	                        {
	                            callbacks.put(valueCmds[i].getFullClassName(), new CallBack(fmd));
	                            DatastoreClass dc = storeMgr.getDatastoreClass(valueCmds[i].getFullClassName());
	                            ClassTable ct = (ClassTable) dc;
	                            if (ct.isInitialized())
	                            {
	                                // if the target table is already initialized, run the callbacks
	                                ct.runCallBacks();
	                            }
	                        }
                        }
                    }
                }
                else if (fmd.getContainer() != null && fmd.getJoinMetaData() != null)
                {
                    // 1-N relationship uni/bidirectional using join table
                    // Do nothing for now
                }
            }
        }
    }

    /**
     * Accessor for whether this table manages the specified class
     * @param className Name of the class
     * @return Whether it is managed by this table
     */
    public boolean managesClass(String className)
    {
        if (className == null)
        {
            return false;
        }

        Iterator iter = managedClassMetaData.iterator();
        while (iter.hasNext())
        {
            ClassMetaData managedCmd = (ClassMetaData)iter.next();
            if (managedCmd.getFullClassName().equals(className))
            {
                return true;
            }
        }

        return false;
    }

    /**
     * Method to initialise the table primary key field(s).
     **/
    private void initializePK()
    {
        assertIsPKUninitialized();

        // Initialise Primary Key mappings for application id with PK fields in this class
        pkMappings = new JavaTypeMapping[cmd.getNoOfPrimaryKeyFields()];
        int pkFieldNum=0;
        int fieldCount = cmd.getNoOfManagedFields();
        for (int relativeFieldNumber = 0; relativeFieldNumber < fieldCount; ++relativeFieldNumber)
        {
            FieldMetaData fmd = cmd.getManagedField(relativeFieldNumber);
            if (fmd.isPrimaryKey())
            {
                if (fmd.getPersistenceModifier() == FieldPersistenceModifier.PERSISTENT)
                {
                    JavaTypeMapping fieldMapping = dba.getMappingManager().getMapping(this, fmd, dba);
                    int absoluteFieldNumber = cmd.getAbsoluteFieldNumberForRelativeFieldNumber(relativeFieldNumber);
                    addFieldMapping(absoluteFieldNumber, fieldMapping);
                    pkMappings[pkFieldNum++] = fieldMapping;
                }
                else if (fmd.getPersistenceModifier() != FieldPersistenceModifier.TRANSACTIONAL)
                {
                    throw new JDOFatalInternalException(LOCALISER.msg("Table.InvalidPersistenceModifierForField",fmd.getName()));
                }

                // Check if auto-increment and that it is supported by this RDBMS
                // The database adapter uses "auto-increment" only, but this refers to EITHER
                // "AUTOINCREMENT" or "IDENTITY" SQL keywords. If there is an RDBMS that supports both
                // of these we'll have to change this!
                if ((fmd.getValueStrategy() == IdentityStrategy.AUTOASSIGN ||
                     fmd.getValueStrategy() == IdentityStrategy.IDENTITY) &&
                    !dba.supportsAutoIncrementKeys())
                {
                    throw new JDOFatalInternalException(LOCALISER.msg("Table.AutoIncrementNotSupportedByRDBMS", 
                        cmd.getFullClassName(), fmd.getName()));
                }
            }
        }

        // No Primary Key defined, so search for superclass or handle datastore id
        PrimaryKey primaryKey = getPrimaryKey();
        if (primaryKey.size() < 1)
        {
            // application identity will inherit PK columns from super class
            if (cmd.getIdentityType() == IdentityType.APPLICATION)
            {
                // Check if the superclass has a table and, if so, copy its PK mappings
                DatastoreClass elementCT = storeMgr.getDatastoreClass(cmd.getPersistenceCapableSuperclass());
                if (elementCT != null)
                {
                    // Check if the user has provided column namings for the copied column mappings
                    ColumnMetaData[] userdefinedCols = null;
                    JoinMetaData inhjoinmd = cmd.getInheritanceMetaData() != null ? cmd.getInheritanceMetaData().getJoinMetaData() : null;
                    int nextUserdefinedCol = 0;
                    if (inhjoinmd != null)
                    {
                        userdefinedCols = inhjoinmd.getColumnMetaData();
                    }

                    JavaTypeMapping referenceMapping = elementCT.getIDMapping();
                    pkMappings = new JavaTypeMapping[elementCT.getIDMapping().getNumberOfDatastoreFields()];
                    for (int i=0; i<elementCT.getIDMapping().getNumberOfDatastoreFields(); i++)
                    {
                        Column superIdColumn = (Column)referenceMapping.getDataStoreMapping(i).getDatastoreField();
                        JavaTypeMapping mapping = dba.getMapping(superIdColumn.getType());
                        ColumnMetaData userdefinedColumn = null;
                        if (userdefinedCols != null)
                        {
                            for (int j=0;j<userdefinedCols.length;j++)
                            {
                                if (superIdColumn.getName().toString().equals(userdefinedCols[j].getTarget()))
                                {
                                    userdefinedColumn = userdefinedCols[j];
                                    break;
                                }
                            }
                            if (userdefinedColumn == null && nextUserdefinedCol < userdefinedCols.length)
                            {
                                userdefinedColumn = userdefinedCols[nextUserdefinedCol++];
                            }
                        }

                        // Add this application identity column
                        DatastoreField refColumn = null;
                        if (userdefinedColumn != null)
                        {
                            // User has provided a name for this column
                            // Currently we only use the column namings from the users definition but we could easily
                            // take more of their details.
                            refColumn = addDatastoreField(superIdColumn.getType(), new SQLIdentifier(dba, userdefinedColumn.getName()), Role.CUSTOM,
                                mapping, superIdColumn.getColumnOptions());
                        }
                        else
                        {
                            // No name provided so take same as superclass
                            refColumn = addDatastoreField(superIdColumn.getType(), superIdColumn.getName(), Role.CUSTOM,
                                mapping,superIdColumn.getColumnOptions());
                        }
                        if (referenceMapping != null &&
                            ((Column)referenceMapping.getDataStoreMapping(i).getDatastoreField()).getColumnOptions() != null)
                        {
                            ((Column)referenceMapping.getDataStoreMapping(i).getDatastoreField()).copyConfigurationTo(refColumn);
                        }
                        refColumn.setAsPrimaryKey();
                        getPrimaryKey().addInheritedDatastoreField(refColumn);

                        pkMappings[i] = mapping;
                        getStoreManager().getMappingManager().getDatastoreMapping(mapping, storeMgr, refColumn, superIdColumn.getType());
                        JPOXLogger.RDBMS.debug(LOCALISER.msg("Table.NewColumnPK", refColumn));
                    }
                }
                else
                {
                    // Superclass has no table, new fields here using its fields to create the mappings
                    ClassMetaData pkCmd = storeMgr.getClassWithPrimaryKeyForClass(cmd.getSuperclassMetaData());
                    if (pkCmd != null)
                    {
                        pkMappings = new JavaTypeMapping[pkCmd.getNoOfPrimaryKeyFields()];
                        pkFieldNum = 0;
                        fieldCount = pkCmd.getNoOfManagedFields();
                        for (int relativeFieldNumber = 0; relativeFieldNumber < fieldCount; ++relativeFieldNumber)
                        {
                            FieldMetaData fmd = pkCmd.getManagedField(relativeFieldNumber);
                            if (fmd.isPrimaryKey())
                            {
                                if (fmd.getPersistenceModifier() == FieldPersistenceModifier.PERSISTENT)
                                {
                                    JavaTypeMapping fieldMapping = dba.getMappingManager().getMapping(this, fmd, dba);
                                    int absoluteFieldNumber = pkCmd.getAbsoluteFieldNumberForRelativeFieldNumber(relativeFieldNumber);
                                    addFieldMapping(absoluteFieldNumber, fieldMapping);
                                    pkMappings[pkFieldNum++] = fieldMapping;
                                }
                                else if (fmd.getPersistenceModifier() != FieldPersistenceModifier.TRANSACTIONAL)
                                {
                                    throw new JDOFatalInternalException(LOCALISER.msg("Table.InvalidPersistenceModifierForField",fmd.getName()));
                                }
                            }
                        }
                    }
                }
            }
            // datastore identity
            else if (cmd.getIdentityType() == IdentityType.DATASTORE)
            {
                Column idColumn;
                Class pc_class = storeMgr.getClassLoaderResolver().classForName(cmd.getFullClassName());
                datastoreIDMapping = new OIDMapping(dba,pc_class);

                // If no identity column specified add one (use any column name specified on datastore-identity element)
                if (cmd.getIdentityMetaData().getColumnMetaData().length < 1)
                {
                    cmd.getIdentityMetaData().addColumn(
                        new ColumnMetaData(cmd.getIdentityMetaData(), cmd.getIdentityMetaData().getColumn()));
                }
                ColumnMetaData icmd = cmd.getIdentityMetaData().getColumnMetaData()[0];
                if (icmd.getName() == null)
                {
                    icmd.setName(new ColumnIdentifier(dba, name, OID.class, Role.NONE).getSQLIdentifier());
                }

                // Add the datastore identity column
                idColumn = (Column) addDatastoreField(OID.class, new SQLIdentifier(dba, icmd.getName()),
                                     Role.CUSTOM, datastoreIDMapping, null);

                // Set the identity column type based on the IdentityStrategy
                // TODO Cater for user custom generators and allow them to set the type also
                if (cmd.getIdentityMetaData().getStrategyValue() == IdentityStrategy.UUIDSTRING ||
                    cmd.getIdentityMetaData().getStrategyValue() == IdentityStrategy.UUIDHEX ||
                    cmd.getIdentityMetaData().getStrategyValue() == IdentityStrategy.AUID)
                {
                    dba.getMappingManager().getDatastoreMapping(datastoreIDMapping, storeMgr, idColumn, String.class);
                }
                else
                {
                    dba.getMappingManager().getDatastoreMapping(datastoreIDMapping, storeMgr, idColumn, long.class);
                }

                idColumn.setAsPrimaryKey();
                if (cmd.getIdentityMetaData().getStrategyValue() != null &&
                    StoreManager.isStrategyDatastoreAttributed(cmd.getIdentityMetaData().getStrategyValue()))
                {
                    idColumn.setAutoIncrement(true);
                }

                // Check if auto-increment and that it is supported by this RDBMS
                if (idColumn.isAutoIncrement() && !dba.supportsAutoIncrementKeys())
                {
                    throw new JDOFatalInternalException(LOCALISER.msg("Table.AutoIncrementNotSupportedByRDBMS", 
                        cmd.getFullClassName(), "datastore-identity"));
                }

                getPrimaryKey().addDatastoreField(idColumn);
                JPOXLogger.RDBMS.debug(LOCALISER.msg("Table.NewColumnPK", idColumn));
            }
            // nondurable identity
            else
            {
                // Do nothing. We have no id
            }
        }
        state = TABLE_STATE_PK_INITIALIZED;
    }

    /**
     * Method to initialise this table to include all fields in the specified class.
     * This is used to recurse up the hierarchy so that we include all immediate superclasses
     * that have "no-table" specified as their inheritance strategy. If we encounter the parent of
     * this class with other than "no-table" we stop the process.
     * @param theCmd The ClassMetaData for the class
     */
    private void initializeForClass(ClassMetaData theCmd)
    {
        // Manage the fields for this class
        manageClass(theCmd);

        // Now check the superclass of this class
        // This caters for superclasses which have their fields persisted with the subclass ("no-table" strategy)
        ClassMetaData parentCmd = theCmd.getSuperclassMetaData();
        if (parentCmd != null && parentCmd.getInheritanceMetaData().getStrategyValue() == InheritanceStrategy.NO_TABLE)
        {
            initializeForClass(parentCmd);
        }
    }

    /**
     * Utility to add the mapping for a field to the managed list.
     * @param absoluteFieldNumber The field number (absolute)
     * @param fieldMapping The mapping for the field
     */
    private void addFieldMapping(int absoluteFieldNumber, JavaTypeMapping fieldMapping)
    {
        fieldMappingsMap.put(fieldMapping.getFieldMetaData(), fieldMapping);
        if (absoluteFieldNumber > highestFieldNumber)
        {
            highestFieldNumber = absoluteFieldNumber;
        }
    }

    /**
     * Create a newOID for a class.
     * Works for datastore identity, or application identity (inc composite keys)
     * @param pcClass The persistence capable class
     * @return The new OID
     */
    public synchronized final Object newOID(StateManager sm, PersistenceCapable pcClass)
    {
        Object id = null;
        ClassMetaData cmd = ((ClassTable)this.getStoreManager().getDatastoreClass(pcClass.getClass().getName())).getClassMetaData();
        if (cmd.getIdentityType() == IdentityType.APPLICATION)
        {
            // All values will have been populated by this point
            id = AIDUtils.getNewObjectId(pcClass, cmd);
        }
        else if (cmd.getIdentityType() == IdentityType.DATASTORE)
        {
            // Populate any strategy value for the "datastore-identity" element
            IdentityMetaData imd = cmd.getIdentityMetaData();
            Object nextIdentifier = storeMgr.getStrategyValue(sm.getPersistenceManager(), this, cmd, -1, imd.getStrategyValue(), imd.getSequence(),
                imd.getExtensions());
            if (nextIdentifier instanceof Number)
            {
                Long value = new Long(nextIdentifier.toString());
                id = new OID(cmd.getFullClassName(), value);
                ((OID) id).turnDataStoreIdentityAsNumeric();
            }
            else
            {
                // Use String
                id = new OID(cmd.getFullClassName(), nextIdentifier.toString());
                ((OID) id).turnDataStoreIdentityAsString();
            }
        }

        return id;
    }

    /**
     * Execute the callbacks for the classes that this table maps to.
     */
    private void runCallBacks()
    {
        Collection c = (Collection)callbacks.remove(cmd.getFullClassName());
        if (c != null)
        {
            for (Iterator it = c.iterator(); it.hasNext();)
            {
                CallBack callback = (CallBack) it.next();

                // 1-N Inverse relationship (no "join")
                if (callback.fmd.getJoinMetaData() == null)
                {
                    // Bidirectional
                    if (callback.fmd.getMappedBy() != null)
                    {
                        // Check that the "mapped-by" field in the other class actually exists
                        FieldMetaData fmd = cmd.getField(callback.fmd.getMappedBy());
                        if (fmd == null)
                        {
                            throw new JDOUserException(LOCALISER.msg("Table.RelationshipMappedByFieldNotFound",
                                callback.fmd.getMappedBy(), cmd.getFullClassName(), callback.fmd.getFullFieldName()));
                        }
                        
                        if (callback.fmd.getMap() != null)
                        {
                            initializeSQLAddUniqueConstraintsStatements(callback.fmd);
                        }
                        
                        // if the field is list type, add index column
                        if (TypeManager.getTypeManager().isSCOList(callback.fmd.getType()))
                        {
                            state = TABLE_STATE_NEW;
                            JavaTypeMapping orderMapping = this.addOrderColumn(fmd);
                            getOrderMappings().put(fmd, orderMapping);
                            state = TABLE_STATE_INITIALIZED;
                        }
                    }
                    // Unidirectional
                    else if (callback.fmd.getMappedBy() == null)
                    {
                        FieldMetaData ownerFmd = callback.fmd;
                        String ownerClassName = ownerFmd.getClassMetaData().getFullClassName();
                        Class ownerType = storeMgr.getClassLoaderResolver().classForName(ownerClassName);
                        JavaTypeMapping fkMapping = new PersistenceCapableMapping(dba, ownerType);
                        
                        // Get the owner id mapping of the "1" end
                        JavaTypeMapping ownerIdMapping = storeMgr.getDatastoreClass(ownerClassName).getIDMapping();
                        CorrespondentColumnsMapping correspondentColumnsMapping =
                            new CorrespondentColumnsMapping(ownerFmd.getElementMetaData() != null ? ownerFmd.getElementMetaData().getColumnMetaData() : null, ownerIdMapping);
                        for (int i=0; i<ownerIdMapping.getNumberOfDatastoreFields(); i++)
                        {
                            DatastoreMapping datastoreMapping = ownerIdMapping.getDataStoreMapping(i);
                            JavaTypeMapping mapping = dba.getMapping(datastoreMapping.getJavaTypeMapping().getJavaType());
                            String userDefinedColumnName = correspondentColumnsMapping.getColumNameByIdentifier(((Column)datastoreMapping.getDatastoreField()).getName());
                            if (userDefinedColumnName == null)
                            {
                                throw new JDOFatalUserException(LOCALISER.msg("Table.PrimaryKeyColumnNotMapped",
                                    ((Column)datastoreMapping.getDatastoreField()).getName(), toString()));
                            }

                            // TODO Check if this column already exists since we're maybe reusing it across 1-N relationships
                            DatastoreField refColumn;
                            ColumnMetaData colmd = null;
                            if (ownerFmd.getElementMetaData() != null &&
                                    ownerFmd.getElementMetaData().getColumnMetaData().length > 0)
                            {
                                colmd = ownerFmd.getElementMetaData().getColumnMetaData()[0];
                            }
                            if (userDefinedColumnName.length() < 1)
                            {
                                SQLIdentifier columnId = new SQLIdentifier(dba, callback.fmd.getName()+((SQLIdentifier) datastoreMapping.getDatastoreField().getName()).getSQLIdentifier(),null);
                                refColumn = addDatastoreField(mapping.getJavaType(), columnId, Role.OWNER, mapping, colmd);
                            }
                            else
                            {
                                SQLIdentifier columnId = new SQLIdentifier(dba, ownerFmd.getName(), userDefinedColumnName);
                                refColumn = addDatastoreField(mapping.getJavaType(), columnId, Role.CUSTOM, mapping, colmd);
                            }
                            ((Column)datastoreMapping.getDatastoreField()).copyConfigurationTo(refColumn);
                            
                            // Foreign key needs to accept nulls since this table knows nothing about it!
                            refColumn.setNullable();
                            
                            fkMapping.addDataStoreMapping(getStoreManager().getMappingManager().getDatastoreMapping(mapping, storeMgr, refColumn, datastoreMapping.getJavaTypeMapping().getJavaType()));
                            ((PersistenceCapableMapping)fkMapping).addJavaTypeMapping(mapping);
                            if (JPOXLogger.RDBMS.isDebugEnabled())
                            {
                                JPOXLogger.RDBMS.debug(LOCALISER.msg("Table.NewColumnReference", refColumn));
                            }
                        }
                        
                        // Save the external FK
                        getExternalFkMappings().put(ownerFmd, fkMapping);
                        
                        // if the field is list type, add index column
                        if (TypeManager.getTypeManager().isSCOList(callback.fmd.getType()))
                        {
                            state = TABLE_STATE_NEW;
                            JavaTypeMapping orderMapping = this.addOrderColumn(ownerFmd);
                            getOrderMappings().put(ownerFmd, orderMapping);
                            state = TABLE_STATE_INITIALIZED;
                        }
                    }
                }
            }
        }
    }

    /**
     * Utility to return the FieldMetaData for a particular field defined by its 
     * absolute field number. Provides the handling for the cases where this table
     * is managing multiple ClassMetaData, and the field is in a subclass to the
     * main class (and hence is not found by a simple call to the ClassMetaData).
     * @param absoluteFieldNumber The absolute field number
     * @return The FieldMetaData for that field
     */
    protected FieldMetaData getFieldMetaDataForField(int absoluteFieldNumber)
    {
        FieldMetaData fmd = cmd.getManagedFieldAbsolute(absoluteFieldNumber);
        if (fmd == null)
        {
            Iterator iter = managedClassMetaData.iterator();
            while (iter.hasNext())
            {
                ClassMetaData theCmd = (ClassMetaData)iter.next();
                fmd = theCmd.getManagedFieldAbsolute(absoluteFieldNumber);
                if (fmd != null)
                {
                    break;
                }
            }
        }
        return fmd;
    }
    
    /**
     * Utility to return the FieldMetaData for a particular field defined by its 
     * absolute field number. Provides the handling for the cases where this table
     * is managing multiple ClassMetaData, and the field is in a subclass to the
     * main class (and hence is not found by a simple call to the ClassMetaData).
     * @param absoluteFieldNumber The absolute field number
     * @param className The class to find the field 
     * @return The FieldMetaData for that field
     */
    protected FieldMetaData getFieldMetaDataForField(int absoluteFieldNumber,String className)
    {
        FieldMetaData fmd = cmd.getManagedFieldAbsolute(absoluteFieldNumber);
        if (fmd == null)
        {
            Iterator iter = managedClassMetaData.iterator();
            while (iter.hasNext())
            {
                ClassMetaData theCmd = (ClassMetaData)iter.next();
                if(theCmd.getFullClassName().equals(className))
                {
	                fmd = theCmd.getManagedFieldAbsolute(absoluteFieldNumber);
	                if (fmd != null)
	                {
	                    break;
	                }
                }
            }
        }
        return fmd;
    }    

    /**
     * Accessor for the main class represented.
     * @return The name of the class
     **/
    public String getType()
    {
        return cmd.getFullClassName();
    }

    /**
     * Accessor for the identity-type. 
     * @return identity-type tag value
     */    
    public IdentityType getIdentityType()
    {
        return cmd.getIdentityType();
    }

    /**
     * Accessor for versionMetaData
     * @return Returns the versionMetaData.
     */
    public final VersionMetaData getVersionMetaData()
    {
        return cmd.getVersionMetaData();
    }

    /**
     * Accessor for Discriminator MetaData
     * @return Returns the Discriminator MetaData.
     */
    public final DiscriminatorMetaData getDiscriminatorMetaData()
    {
        return cmd.getInheritanceMetaData().getDiscriminatorMetaData();
    }

    /**
     * Whether this table or super table has id (primary key) attributed by the datastore
     * @return true if the id attributed by the datastore
     */
    public boolean isObjectIDDatabaseAttributed()
    {
        return StoreManager.isStrategyDatastoreAttributed(cmd.getIdentityMetaData().getStrategyValue());
    }

    /**
     * Whether this table is the root table and does not inherit primary key
     * values from any super table.
     * @return true if this table is a root table
     */
    public boolean isRootTable()
    {
        return supertable == null ? true : false;
    }

    /**
     * Accessor for the supertable for this table.
     * @return The supertable
     **/
    public ClassTable getSupertable()
    {
        assertIsInitialized();

        return supertable;
    }

    /**
     * Utility to find the table above this one. Will recurse to cater for inheritance
     * strategies where fields are handed up to the super class, or down to this class.
     * @param theCmd ClassMetaData of the class to find the supertable for.
     * @return The table above this one in any inheritance hierarchy
     */
    private ClassTable getSupertable(ClassMetaData theCmd)
    {
        ClassMetaData superCmd = theCmd.getSuperclassMetaData();
        if (superCmd != null)
        {
            if (superCmd.getInheritanceMetaData().getStrategyValue() == InheritanceStrategy.NEW_TABLE)
            {
            	// This class has its own table, so return it.
                return (ClassTable) storeMgr.getDatastoreClass(superCmd.getFullClassName());
            }
            else if (superCmd.getInheritanceMetaData().getStrategyValue() == InheritanceStrategy.NO_TABLE)
            {
            	// This class is mapped to the same table, so go up another level.
            	return getSupertable(superCmd);
            }
            else
            {
            	// This class is mapped to its superclass table, so go up to that and try again.
                return getSupertable(superCmd);
            }
        }
        return null;
    }

    /**
     * Convenience accessor for the base table for this table which has the specified field.
     * @param fmd Field MetaData for this field
     * @return The base table which has the field specified
     */
    public ClassTable getBaseTableWithField(FieldMetaData fmd)
    {
        if (fieldMappingsMap.get(fmd) != null)
        {
            // We have this field so return this table
            return this;
        }
        else if (getSupertable() == null)
        {
            // We don't have the field, but have no superclass, so return null
            return this;
        }
        else
        {
            // Return the superclass sicne we don't have it
            return getSupertable().getBaseTableWithField(fmd);
        }
    }

    /**
     * Accessor for the class MetaData.
     * @return The class MetaData
     **/
    private ClassMetaData getClassMetaData()
    {
        return cmd;
    }

    /**
     * Accessor for the indices for this table. This includes both the
     * user-defined indices (via MetaData), and the ones required by foreign
     * keys (required by relationships).
     * @return The indices
     */
    protected Set getExpectedIndices()
    {
        // The indices required by foreign keys (BaseTable)
        Set indices=super.getExpectedIndices();

        // Add on any user-required indices for the fields
        Set fieldNumbersSet = fieldMappingsMap.keySet();
        Iterator iter = fieldNumbersSet.iterator();
        while (iter.hasNext())
        {
            FieldMetaData fmd=(FieldMetaData) iter.next();
            IndexMetaData imd=fmd.getIndexMetaData();
            if (!fmd.isPrimaryKey() && imd != null)
            {
                Index index = getIndexForIndexMetaData(imd, fmd);
                if (index != null)
                {
                    indices.add(index);
                }
            }
        }

        // Check if any version column needs indexing
        if (versionMapping != null)
        {
            VersionMetaData vermd = getVersionMetaData();
            IndexMetaData idxmd = vermd.getIndexMetaData();
            if (idxmd != null)
            {
                Index index = new Index(this, idxmd.isUnique() != null ? idxmd.isUnique().booleanValue() : false);
                if (idxmd.getName() != null)
                {
                    index.setName(idxmd.getName());
                }
                for (int i=0; i<versionMapping.getNumberOfDatastoreFields(); i++)
                {
                    index.addDatastoreField(versionMapping.getDataStoreMapping(i).getDatastoreField());
                }
                indices.add(index);
            }
        }

        // Check if any discriminator column needs indexing
        if (discriminatorMapping != null)
        {
            DiscriminatorMetaData dismd = getDiscriminatorMetaData();
            IndexMetaData idxmd = dismd.getIndexMetaData();
            if (idxmd != null)
            {
                Index index = new Index(this, idxmd.isUnique() != null ? idxmd.isUnique().booleanValue() : false);
                if (idxmd.getName() != null)
                {
                    index.setName(idxmd.getName());
                }
                for (int i=0; i<discriminatorMapping.getNumberOfDatastoreFields(); i++)
                {
                    index.addDatastoreField(discriminatorMapping.getDataStoreMapping(i).getDatastoreField());
                }
                indices.add(index);
            }
        }

        // Add on any order fields (for lists) that need indexing
        Set orderMappingsKeys = getOrderMappings().keySet();
        Iterator orderMappingsKeyIter = orderMappingsKeys.iterator();
        while (orderMappingsKeyIter.hasNext())
        {
            FieldMetaData fmd = (FieldMetaData)orderMappingsKeyIter.next();
            OrderMetaData omd = fmd.getOrderMetaData();
            if (omd != null && omd.getIndexMetaData() != null)
            {
                Index index = getIndexForIndexMetaData(omd.getIndexMetaData());
                if (index != null)
                {
                    indices.add(index);
                }
            }
        }

        // Add on any user-required indices for the class(es) as a whole (subelement of <class>)
        Iterator cmdIter = managedClassMetaData.iterator();
        while (cmdIter.hasNext())
        {
            ClassMetaData thisCmd = (ClassMetaData)cmdIter.next();
            IndexMetaData[] classIndices = thisCmd.getIndexMetaData();
            if (classIndices != null)
            {
                for (int i=0;i<classIndices.length;i++)
                {
                    Index index = getIndexForIndexMetaData(classIndices[i]);
                    if (index != null)
                    {
                        indices.add(index);
                    }
                }
            }
        }

        return indices;
    }

    /**
     * Convenience method to convert an IndexMetaData into an Index.
     * @param imd The Index MetaData
     * @return The Index
     */
    private Index getIndexForIndexMetaData(IndexMetaData imd)
    {
        // Verify if a unique index is needed
        boolean unique=false;
        if (imd.isUnique() != null && imd.isUnique().booleanValue())
        {
            unique = true;
        }

        Index index = new Index(this, unique);

        // Set the index name if required
        if (imd.getName() != null)
        {
            index.setName(imd.getName());
        }

        // Set the column(s) to index
        // Class-level index so use its column definition
        ColumnMetaData[] colmds = imd.getColumnMetaData();
        FieldMetaData[] fmds = imd.getFieldMetaData();
        // a). Columns specified directly
        if (colmds != null && colmds.length > 0)
        {
            for (int i=0;i<colmds.length;i++)
            {
                SQLIdentifier colName = new SQLIdentifier(dba, colmds[i].getName());
                Column col = (Column)columnsByName.get(colName);
                if (col == null)
                {
                    JPOXLogger.RDBMS.warn(LOCALISER.msg("Table.Index.ColumnNotFound", 
                        toString(), index.getName(), colmds[i].getName()));
                    break;
                }
                else
                {
                    index.addDatastoreField(col);
                }
            }
        }
        // b). Columns specified using fields
        else if (fmds != null && fmds.length > 0)
        {
            for (int i=0;i<fmds.length;i++)
            {
                JavaTypeMapping fieldMapping = (JavaTypeMapping)fieldMappingsMap.get(fmds[i]);
                for (int j=0; j<fieldMapping.getNumberOfDatastoreFields(); j++)
                {
                    index.addDatastoreField(fieldMapping.getDataStoreMapping(j).getDatastoreField());
                }
            }
        }
        else
        {
            // We can't have an index of no columns
            JPOXLogger.RDBMS.warn(LOCALISER.msg("Table.Index.NoColumnsSpecified", 
                toString(), index.getName()));
            return null;
        }

        return index;
    }

    /**
     * Convenience method to convert an IndexMetaData into an Index.
     * @param imd The Index MetaData
     * @param fmd the FieldMetaData 
     * @return The Index
     */
    private Index getIndexForIndexMetaData(IndexMetaData imd, FieldMetaData fmd)
    {
        // Verify if a unique index is needed
        boolean unique=false;
        if (imd.isUnique() != null && imd.isUnique().booleanValue())
        {
            unique = true;
        }

        Index index = new Index(this, unique);

        // Set the index name if required
        if (imd.getName() != null)
        {
            index.setName(imd.getName());
        }

        // Field-level index so use all columns for the field
        JavaTypeMapping fieldMapping = (JavaTypeMapping)fieldMappingsMap.get(fmd);
        for (int j=0; j<fieldMapping.getNumberOfDatastoreFields(); j++)
        {
            index.addDatastoreField(fieldMapping.getDataStoreMapping(j).getDatastoreField());
        }

        return index;
    }

    /**
     * Accessor for the expected foreign keys for this table.
     * @return The expected foreign keys.
     **/
    protected List getExpectedForeignKeys()
    {
        assertIsInitialized();

        // Auto mode allows JPOX to decide which FKs are needed as well as using
        // what is in the users MetaData.
        boolean autoMode = false;
        if (storeMgr.getPMFContext().getPmfConfiguration().getForeignKeyCreateMode() == PMFConfiguration.FOREIGN_KEY_AUTO_MODE)
        {
            autoMode = true;
        }

        ArrayList foreignKeys = new ArrayList();

        // Check each field for FK requirements (user-defined, or required)
        Set fieldNumbersSet = fieldMappingsMap.keySet();
        Iterator iter = fieldNumbersSet.iterator();
        while (iter.hasNext())
        {
            FieldMetaData fmd = (FieldMetaData) iter.next();

            // If the field is of a PC class and is not embedded, look to add a FK to the PC class table
            if (storeMgr.getMetaDataManager().getMetaDataForClass(fmd.getType()) != null && fmd.getEmbeddedMetaData() == null)
            {
                ClassTable referencedTable = (ClassTable)storeMgr.getDatastoreClass(fmd.getTypeName());
                if (referencedTable != null)
                {
                    ForeignKeyMetaData fkmd = fmd.getForeignKeyMetaData();
                    if ((fkmd != null && fkmd.getDeleteAction() != ForeignKeyDeleteAction.NONE) || autoMode)
                    {
                        // Either has been specified by user, or using autoMode, so add FK
                        JavaTypeMapping fieldMapping = (JavaTypeMapping)fieldMappingsMap.get(fmd);
                        ForeignKey fk = new ForeignKey(fieldMapping, dba, referencedTable, true);
                        fk.setForMetaData(fkmd); // Does nothing when no FK MetaData
                        foreignKeys.add(fk);
                    }
                }
            }
        }

        // When in auto mode add FK to any superclass table
        if (autoMode && supertable != null)
        {
            if (cmd.getIdentityType() == IdentityType.DATASTORE)
            {
                foreignKeys.add(0, new ForeignKey(getIDMapping(), dba, supertable, false));
            }
            else if (cmd.getIdentityType() == IdentityType.APPLICATION)
            {
                // TODO Add FK for application id cases to superclass table
            }
        }
        
        // Add any FKs to owner Collection tables
        HashMap externalFks = getExternalFkMappings();
        if (!externalFks.isEmpty())
        {
            Collection externalFkKeys = externalFks.keySet();
            Iterator externalFkKeysIter = externalFkKeys.iterator();
            while (externalFkKeysIter.hasNext())
            {
                FieldMetaData fmd = (FieldMetaData)externalFkKeysIter.next();
                ClassTable referencedTable = (ClassTable)storeMgr.getDatastoreClass(fmd.getClassMetaData().getFullClassName());
                if (referencedTable != null)
                {
                    ForeignKeyMetaData fkmd = fmd.getElementMetaData() != null ? fmd.getElementMetaData().getForeignKeyMetaData() : null;
                    if ((fkmd != null && fkmd.getDeleteAction() != ForeignKeyDeleteAction.NONE) || autoMode)
                    {
                        // Either has been specified by user, or using autoMode, so add FK
                        JavaTypeMapping fkMapping = (JavaTypeMapping)externalFks.get(fmd);
                        ForeignKey fk = new ForeignKey(fkMapping, dba, referencedTable, true);
                        fk.setForMetaData(fkmd); // Does nothing when no FK MetaData
                        foreignKeys.add(fk);
                    }
                }
            }
        }

        // TODO Add any user-required FKs for the class as a whole

        return foreignKeys;
    }

    /**
     * Accessor for the expected candidate keys for this table.
     * @return The expected candidate keys.
     **/
    protected List getExpectedCandidateKeys()
    {
        assertIsInitialized();

        // The candidate keys required by the basic table
        List candidateKeys = super.getExpectedCandidateKeys();

        // Add on any user-required candidate keys for the fields
        Set fieldNumbersSet = fieldMappingsMap.keySet();
        Iterator iter = fieldNumbersSet.iterator();
        while (iter.hasNext())
        {
            FieldMetaData fmd=(FieldMetaData) iter.next();
            UniqueMetaData umd = fmd.getUniqueMetaData();
            if (umd != null)
            {
                CandidateKey ck = getCandidateKeyForUniqueMetaData(umd, fmd);
                if (ck != null)
                {
                    candidateKeys.add(ck);
                }
            }
        }

        // Add on any user-required candidate keys for the class(es) as a whole (composite keys)
        Iterator cmdIter = managedClassMetaData.iterator();
        while (cmdIter.hasNext())
        {
            ClassMetaData thisCmd = (ClassMetaData)cmdIter.next();
            UniqueMetaData[] classCKs = thisCmd.getUniqueMetaData();
            if (classCKs != null)
            {
                for (int i=0;i<classCKs.length;i++)
                {
                    CandidateKey ck = getCandidateKeyForUniqueMetaData(classCKs[i]);
                    if (ck != null)
                    {
                        candidateKeys.add(ck);
                    }
                }
            }
        }

        return candidateKeys;
    }

    /**
     * Convenience method to convert a UniqueMetaData into a CandidateKey.
     * @param umd The Unique MetaData
     * @param fmd The FieldMetaData
     * @return The Candidate Key
     */
    private CandidateKey getCandidateKeyForUniqueMetaData(UniqueMetaData umd, FieldMetaData fmd)
    {
        CandidateKey ck = new CandidateKey(this);

        // Set the key name if required
        if (umd.getName() != null)
        {
            ck.setName(umd.getName());
        }

        // Set the column(s) to key

        // Field-level index so use all columns for the field
        JavaTypeMapping fieldMapping = (JavaTypeMapping)fieldMappingsMap.get(fmd);
        for (int j=0; j<fieldMapping.getNumberOfDatastoreFields(); j++)
        {
            ck.addDatastoreField(fieldMapping.getDataStoreMapping(j).getDatastoreField());
        }


        return ck;
    }

    /**
     * Convenience method to convert a UniqueMetaData into a CandidateKey.
     * @param umd The Unique MetaData
     * @return The Candidate Key
     */
    private CandidateKey getCandidateKeyForUniqueMetaData(UniqueMetaData umd)
    {
        CandidateKey ck = new CandidateKey(this);

        // Set the key name if required
        if (umd.getName() != null)
        {
            ck.setName(umd.getName());
        }

        // Class-level index so use its column definition
        ColumnMetaData[] colmds = umd.getColumnMetaData();
        FieldMetaData[] fmds = umd.getFieldMetaData();
        // a). Columns specified directly
        if (colmds != null && colmds.length > 0)
        {
            for (int i=0;i<colmds.length;i++)
            {
                SQLIdentifier colName = new SQLIdentifier(dba, colmds[i].getName());
                Column col = (Column)columnsByName.get(colName);
                if (col == null)
                {
                    JPOXLogger.RDBMS.warn(LOCALISER.msg("Table.CandidateKey.ColumnNotFound", 
                        toString(), ck.getName(), colmds[i].getName()));
                    break;
                }
                else
                {
                    ck.addDatastoreField(col);
                }
            }
        }
        // b). Columns specified using fields
        else if (fmds != null && fmds.length > 0)
        {
            for (int i=0;i<fmds.length;i++)
            {
                JavaTypeMapping fieldMapping = (JavaTypeMapping)fieldMappingsMap.get(fmds[i]);
                for (int j=0; j<fieldMapping.getNumberOfDatastoreFields(); j++)
                {
                    ck.addDatastoreField(fieldMapping.getDataStoreMapping(j).getDatastoreField());
                }
            }
        }
        else
        {
            // We can't have an index of no columns
            JPOXLogger.RDBMS.warn(LOCALISER.msg("Table.CandidateKey.NoColumnsSpecified", 
                toString(), ck.getName()));
            return null;
        }

        return ck;
    }
    
    protected List getSQLCreateStatements()
    {
        List stmts = super.getSQLCreateStatements();
        stmts.addAll(getSQLAddUniqueConstraintsStatements());

        return stmts;
    }

    private List getSQLAddUniqueConstraintsStatements()
    {
        ArrayList stmts = new ArrayList();
        Iterator cks = candidateKeysByMapField.values().iterator();
        int ckNum = 0;

        while (cks.hasNext())
        {
            CandidateKeyIdentifier ckName = new CandidateKeyIdentifier(this, ++ckNum);
            CandidateKey ck = (CandidateKey)cks.next();
            ck.setName(ckName.getSQLIdentifier());

            stmts.add(dba.getAddCandidateKeyStatement(ck));
        }

        return stmts;
    }

    private void initializeSQLAddUniqueConstraintsStatements(FieldMetaData ownerfmd)
    {
        //if unique constraints creation for inverse maps are not enabled, just return
        if (!storeMgr.getPMFContext().getPmfConfiguration().getUniqueConstraintsMapInverse())
        {
            return;
        }

        // Load "mapped-by"
        FieldMetaData mfmd=null;
        String map_field_name=ownerfmd.getMappedBy();
        if (map_field_name != null)
        {
            if (ownerfmd.getMap() != null)
            {
                mfmd = cmd.getField(map_field_name);
                if (mfmd == null)
                {
                    throw new JDOUserException(LOCALISER.msg("Table.RelationshipMappedByFieldNotFound",
                                               map_field_name, cmd.getFullClassName(), ownerfmd.getFullFieldName()));                       
                }
            }
        }

        if (mfmd != null)
        {
            if (ownerfmd.getJoinMetaData() == null)
            {
                // Load "key-field"
                FieldMetaData kmd = null;
                String key_field_name = ownerfmd.getMap().getValueForExtension("key-field");
                if (key_field_name != null)
                {
                    kmd = cmd.getField(key_field_name);
                }
                if (kmd == null)
                {
                    throw new ClassDefinitionException(LOCALISER.msg("Table.MissingMapKeyField",mfmd));
                }

                JavaTypeMapping ownerMapping = getFieldMapping(map_field_name);
                JavaTypeMapping keyMapping = getFieldMapping(kmd.getName());

                if (dba.supportsNullsInCandidateKeys() || (!ownerMapping.isNullable() && !keyMapping.isNullable()))
                {
                    // If the owner and key fields are represented in this table then we can impose
                    // a unique constraint on them. If the key field is in a superclass then we
                    // cannot do this so just omit it.
                    if (keyMapping.getDatastoreClass() == this &&
                        ownerMapping.getDatastoreClass() == this)
                    {
                        CandidateKey ck = new CandidateKey(this);

                        // This HashSet is to avoid duplicate adding of columns.
                        HashSet addedColumns = new HashSet();

                        // Add columns for the owner field
                        for (int i = 0; i < ownerMapping.getNumberOfDatastoreFields(); i++)
                        {
                            Column col = (Column) ownerMapping.getDataStoreMapping(i).getDatastoreField();
                            addedColumns.add(col);
                            ck.addDatastoreField(col);
                        }

                        // Add columns for the key field
                        for (int i = 0; i < keyMapping.getNumberOfDatastoreFields(); i++)
                        {
                            Column col = (Column) keyMapping.getDataStoreMapping(i).getDatastoreField();
                            if (!addedColumns.contains(col))
                            {
                                addedColumns.add(col);
                                ck.addDatastoreField(col);
                            }
                            else
                            {
                                JPOXLogger.RDBMS.warn(LOCALISER.msg("Table.KeyTypeMapColumnOverwriteValueTypeMapColumn", ownerfmd.getName()));
                            }
                        }

                        if (candidateKeysByMapField.put(mfmd, ck) != null)
                        {
                            throw new ClassDefinitionException(LOCALISER.msg("Table.DuplicateMapOwnerField", mfmd));
                        }
                    }
                }
            }
        }
    }

    /**
     * Accessor for a mapping for the ID (PersistenceCapable) for this table.
     * @return The (PersistenceCapable) ID mapping.
     **/
    public JavaTypeMapping getIDMapping()
    {
        if (!isPKInitialized())
        {
            initializePK();
        }

        Class pc_class = storeMgr.getClassLoaderResolver().classForName(cmd.getFullClassName());
        PersistenceCapableMapping mapping = new PersistenceCapableMapping(getStoreManager().getDatastoreAdapter(),pc_class);
        if (getIdentityType() == IdentityType.DATASTORE)
        {
            mapping.addJavaTypeMapping(datastoreIDMapping);
        }
        else if (getIdentityType() == IdentityType.APPLICATION)
        {
            for (int i = 0; i < pkMappings.length; i++)
            {
                mapping.addJavaTypeMapping(pkMappings[i]);
            }
        }
        else
        {
            // Nothing to do for nondurable since no identity
        }
        return mapping;
    }

    /**
     * Accessor for a mapping for the datastore ID (OID) for this table.
     * @return The (OID) mapping.
     **/
    public JavaTypeMapping getDataStoreObjectIdMapping()
    {
        assertIsInitialized();

        return datastoreIDMapping;
    }

    /**
     * Accessor for the version mapping specified .
     * @return The version mapping.
     **/
    public JavaTypeMapping getVersionMapping()
    {
        return versionMapping;
    }

    /**
     * Accessor for the discriminator mapping specified .
     * @return The mapping for the discriminator column
     **/
    public JavaTypeMapping getDiscriminatorMapping()
    {
        return discriminatorMapping;
    }

    /**
     * Accessor for all of the order mappings (used by Inverse Lists)
     * @return The mappings for the order columns.
     **/
    private HashMap getOrderMappings()
    {
        if (orderMappings == null)
        {       
            orderMappings = new HashMap();
        }
        return orderMappings;
    }

    /**
     * Accessor for the index mapping specified by the field.
     * @param fmd The metadata for the List field.
     * @return The index mapping for the specified field.
     **/
    public JavaTypeMapping getOrderMapping(FieldMetaData fmd)
    {
        return (JavaTypeMapping) getOrderMappings().get(fmd);
    }

    /**
     * Accessor for all of the external FK mappings.
     * @return The mappings for external FKs
     **/
    public HashMap getExternalFkMappings()
    {
        if (externalFkMappings == null)
        {       
            externalFkMappings = new HashMap();
        }
        return externalFkMappings;
    }

    /**
     * Accessor for the external FK mapping for an owner field.
     * @param fmd The metadata for the field.
     * @return The external mapping for the specified field.
     **/
    public JavaTypeMapping getExternalFkMapping(FieldMetaData fmd)
    {
        return (JavaTypeMapping) getExternalFkMappings().get(fmd);
    }

    /**
     * Accessor for whether a field is persistent. Caters for the field being
     * persisted in the superclasses table.
     * @param absoluteFieldNumber The absolute field number
     * @return Whether it is persistent
     */
    public boolean isFieldPersistent(int absoluteFieldNumber)
    {
        return getFieldMapping(getFieldMetaDataForField(absoluteFieldNumber)) != null;
    }

    /**
     * Acessor for the field Mapping. This method handle superclasses
     * @param fmd Field MetaData for this field
     * @return the Mapping for the field
     */
    public JavaTypeMapping getFieldMapping(FieldMetaData fmd)
    {
        assertIsInitialized();

        // Check if we manage this field
        JavaTypeMapping m = (JavaTypeMapping)fieldMappingsMap.get(fmd);
        if (m != null)
        {
            return m;
        }

        // Check supertable
        int ifc = cmd.getNoOfInheritedManagedFields();
        if (fmd.getAbsoluteFieldNumber() < ifc)
        {
            if (supertable != null)
            {
                return supertable.getFieldMapping(fmd);
            }
        }

        return null;
    }

    /**
     * Accessor for the field mapping for the named field.
     * The field may exist in a parent table.
     * @param fieldName Name of field
     * @return The field mapping.
     */
    public JavaTypeMapping getFieldMapping(String fieldName)
    {
        assertIsInitialized();

        JavaTypeMapping m = null;
        ClassTable ct = this;

        do
        {
            Iterator cmdIter = ct.managedClassMetaData.iterator();
            while (cmdIter.hasNext())
            {
                // Try all of the ClassMetaData being managed by this class
                ClassMetaData theCmd = (ClassMetaData)cmdIter.next();
                FieldMetaData fmd = theCmd.getField(fieldName);
                if( fmd != null )
                {
                    m = (JavaTypeMapping)ct.fieldMappingsMap.get(fmd);
                    break;
                }
            }

            if (m != null)
            {
                break;
            }

            // Go up to the superclass table
            ct = ct.supertable;
        } while (ct != null);

        if (m == null)
        {
            throw new NoSuchPersistentFieldException(cmd.getFullClassName(), fieldName);
        }

        return m;
    }

    /**
     * Acessor for the FieldMetaData.
     * @param fieldName the field name
     * @return the FieldMetaData
     */
    public FieldMetaData getFieldMetaData(String fieldName)
    {
        return getFieldMetaDataForField(cmd.getFieldNumberAbsolute(fieldName));
    }
    
    /**
     * Utility to throw an exception if the object is not PersistenceCapable.
     * @param sm The StateManager for the object
     **/ 
    private void assertPCClass(StateManager sm)
    {
        Class c = sm.getObject().getClass();

        if (!sm.getPersistenceManager().getClassLoaderResolver().isAssignableFrom(cmd.getFullClassName(),c))
        {
            throw new JDOFatalInternalException(LOCALISER.msg("Table.ClassObjectInfo",cmd.getFullClassName(),c));
        }
    }

    /**
     * Method to fetch an insert into this table. Will insert any superclass
     * records as well (recursive).
     * @param sm StateManager for the object to insert.
     **/
    public void insert(StateManager sm)
    {
        assertIsInitialized();
        assertPCClass(sm);

        if (supertable != null)
        {
            supertable.insert(sm);
        }

        storeMgr.getInsertRequest(this,sm.getObject().getClass().getName()).execute(sm);
    }

    /**
     * Method to fetch an object in this table. Will fetch any superclass
     * records as well (recursive).
     * @param sm StateManager for the object to fetch.
     * @param fieldNumbers The field numbers to be fetched
     */
    public void fetch(StateManager sm, int[] fieldNumbers)
    {
        assertIsInitialized();
        assertPCClass(sm);

        if (supertable != null)
        {
            supertable.fetch(sm, fieldNumbers);
        }

        storeMgr.getFetchRequest(this, fieldNumbers, sm.getObject().getClass().getName()).execute(sm);
    }

    /**
     * Method to update an object in this table. Will update any superclass
     * records as well (recursive).
     * @param sm StateManager for the object to update.
     * @param fieldNumbers the field numbers to be updated
     */
    public void update(StateManager sm, int[] fieldNumbers)
    {
        assertIsInitialized();
        assertPCClass(sm);

        if (supertable != null)
        {
            supertable.update(sm, fieldNumbers);
        }

        storeMgr.getUpdateRequest(this, fieldNumbers,sm.getObject().getClass().getName()).execute(sm);
    }

    /**
     * Method to delete an object from this table. Will delete any superclass
     * records as well (recursive).
     * @param sm StateManager for the object to delete.
     **/
    public void delete(StateManager sm)
    {
        assertIsInitialized();
        assertPCClass(sm);

        storeMgr.getDeleteRequest(this, sm.getObject().getClass().getName()).execute(sm);

        if (supertable != null)
        {
            supertable.delete(sm);
        }
    }

    /**
     * Locates this object in the datastore. If the instance does not exist in
     * the datastore, this method will fail raising a
     * javax.jdo.JDOObjectNotFoundException.
     * @param sm StateManager for the object to locate
     */
    public void locate(StateManager sm)
    {
        assertIsInitialized();
        assertPCClass(sm);

        storeMgr.getLocateRequest(this,sm.getObject().getClass().getName()).execute(sm);
    }

    /**
     * Delete dependent objects of a persistent object.
     *  
     * @since 1.1
     * @param sm StateManager for the object being deleted.
     * @param fieldNumbers the fields to delete.
     */
    public void deleteDependent(StateManager sm,int[] fieldNumbers)
    {
        assertIsInitialized();
        assertPCClass(sm);
 
        for (int i = 0; i < fieldNumbers.length; ++i)
        {
            int fieldNumber = fieldNumbers[i];        

            FieldMetaData fmd = getFieldMetaDataForField(fieldNumber);
            if (fmd.isDependent())
            {
                JavaTypeMapping m = getFieldMapping(fmd);
	            if (m instanceof MappingCallbacks)
	            {
	               ((MappingCallbacks)m).deleteDependent(sm);
	            }
	            else if (m instanceof PersistenceCapableMapping)
	            {
	                //makes sure field is loaded
	                sm.isLoaded(sm.getObject(),fieldNumber);
	                PersistenceCapable pc = (PersistenceCapable) sm.provideField(fieldNumber);
	                if (pc != null)
	                {
	                    //clear reference to object
	                    sm.replaceField(fieldNumber,null);
	                    sm.makeDirty(fieldNumber);
	                    //delete object
	                    sm.getPersistenceManager().deletePersistent(pc);
	                }
	            }
            }
        }
    }

    /**
     * Adds an ordering column to the element table (this) in inverse list relationships.
     * Used to store the position of the element in the List.
     * @param fmd The FieldMetaData for the column to map to
     * @return The Mapping for the order column
     */
    private JavaTypeMapping addOrderColumn(FieldMetaData fmd)
    {
        Class indexType = Integer.class;
        JavaTypeMapping indexMapping = new IndexMapping(dba, indexType);
        SQLIdentifier indexColumnName;

        // Allow for any user definition in OrderMetaData
        OrderMetaData omd = fmd.getOrderMetaData();
        boolean userDefinedName = false;
        if (omd != null)
        {
            String colName = omd.getColumn();
            if (colName == null && omd.getColumnMetaData() != null)
            {
                colName = omd.getColumnMetaData()[0].getName();
            }
            indexColumnName = new SQLIdentifier(dba, colName);
            userDefinedName = true;
        }
        else
        {
            indexColumnName = (SQLIdentifier) getDefaultNameForType(indexType.getName());
            indexColumnName = new SQLIdentifier(dba, fmd.getName() + indexColumnName.getSQLIdentifier());
        }

        ColumnMetaData colmd = (fmd.getColumnMetaData().length > 0 ? fmd.getColumnMetaData()[0] : null);
        DatastoreField column = addDatastoreField(indexType, indexColumnName, userDefinedName ? Role.CUSTOM : Role.INDEX, indexMapping, colmd).setNullable();

        getStoreManager().getMappingManager().getDatastoreMapping(indexMapping, storeMgr, column, indexType);

        return indexMapping;
    }

    /**
     * Accessor for the column name for the specified class.
     * @param className Name of the class for which the column is required.
     * @return The SQL identifier of the column. 
     **/
    protected DatastoreIdentifier getDefaultNameForType(String className)
    {
        if (TypeManager.getTypeManager().isSupportedType(className)) {
            String javaName = className.substring(className.lastIndexOf('.') + 1);
            return new ColumnIdentifier(dba, javaName);
        }
        else
        {
        	return storeMgr.getDatastoreClass(className).getName();
        }
    }

    /**
     * Callbacks is used for inverse relationships to run some operation
     * in the target table. The operation is creation of columns, indexes, or whatever needed. 
     */
    private static class CallBack
    {
        final FieldMetaData fmd;

        /**
         * Default constructor
         */
        public CallBack(FieldMetaData fmd)
        {
            this.fmd = fmd;
        }
    }


    /**
     * Provide the mappings to the consumer for all primary-key fields mapped to
     * this table.
     * @param consumer Consumer for the mappings
     */
    public void providePrimaryKeyMappings(MappingConsumer consumer)
    {
        consumer.preConsumeMapping(highestFieldNumber + 1);

        int[] primaryKeyFieldNumbers = cmd.getPrimaryKeyFieldNumbers();
        for (int i = 0; i < cmd.getNoOfPrimaryKeyFields(); i++)
        {
            consumer.consumeMapping(getFieldMapping(getFieldMetaDataForField(primaryKeyFieldNumbers[i])),
                					getFieldMetaDataForField(primaryKeyFieldNumbers[i]));
        }
    }

    /**
     * Provide the mappings to the consumer for all non primary-key fields
     * mapped to this table.
     * @param consumer Consumer for the mappings
     */
    public void provideNonPrimaryKeyMappings(MappingConsumer consumer)
    {
        consumer.preConsumeMapping(highestFieldNumber + 1);

        Set fieldNumbersSet = fieldMappingsMap.keySet();
        Iterator iter = fieldNumbersSet.iterator();
        while (iter.hasNext())
        {
            FieldMetaData fmd = (FieldMetaData) iter.next();
            JavaTypeMapping fieldMapping = (JavaTypeMapping)fieldMappingsMap.get(fmd);
            if (fieldMapping != null)
            {
                if (!fmd.isPrimaryKey())
                {
                    consumer.consumeMapping(fieldMapping, fmd);
                }
            }
        }
    }

    /**
     * Provide the mappings to the consumer for all absolute field Numbers in this table
     * that are container in the fieldNumbers parameter.
     * @param consumer Consumer for the mappings
     * @param fieldNumbers Absolute numbers of the fields to provide mappings for
     */
    public void provideMappingsForFields(MappingConsumer consumer, int[] fieldNumbers, ClassMetaData cmd)
    {
        consumer.preConsumeMapping(highestFieldNumber + 1);

        for (int i = 0; i < fieldNumbers.length; i++)
        {
            FieldMetaData fmd = cmd.getManagedFieldAbsolute(fieldNumbers[i]);
            JavaTypeMapping fieldMapping = (JavaTypeMapping)fieldMappingsMap.get(fmd);
            if (fieldMapping != null)
            {
                if (!fmd.isPrimaryKey())
                {
                    consumer.consumeMapping(fieldMapping, fmd);
                }
            }
        }
    }

    /**
     * Provide the mappings to version mappings
     * @param consumer Consumer for the version mappings
     */
    public void provideVersionMappings(MappingConsumer consumer)
    {
        consumer.preConsumeMapping(highestFieldNumber + 1);

        if (getVersionMapping() != null)
        {
            consumer.consumeMapping(getVersionMapping(), MappingConsumer.MAPPING_TYPE_VERSION);
        }
    }

    /**
     * Provide the mappings to discriminator mappings
     * @param consumer Consumer for the mappings
     */
    public void provideDiscriminatorMappings(MappingConsumer consumer)
    {
        consumer.preConsumeMapping(highestFieldNumber + 1);

        if (getDiscriminatorMapping() != null)
        {
            consumer.consumeMapping(getDiscriminatorMapping(), MappingConsumer.MAPPING_TYPE_DISCRIMINATOR);
        }
    }

    /**
     * Provide the mappings to datastore identity mappings
     * @param consumer Consumer for the mappings
     */
    public void provideDatastoreIdMappings(MappingConsumer consumer)
    {
        consumer.preConsumeMapping(highestFieldNumber + 1);

        if (getIdentityType() == IdentityType.DATASTORE)
        {
            consumer.consumeMapping(getDataStoreObjectIdMapping(), MappingConsumer.MAPPING_TYPE_DATASTORE_ID);
        }
    }

    /**
     * Method to prepare a field mapping for use in the datastore.
     * This creates the column in the table.
     * @param m The mapping for the field
     * @param fmd The MetaData for the field
     */
    public void prepareDatastoreMapping(JavaTypeMapping m, FieldMetaData fmd)
    {
        Class fieldType = fmd.getType();

        if (m instanceof PersistenceCapableMapping)
        {
            JavaTypeMapping referenceMapping = storeMgr.getDatastoreClass(fieldType.getName()).getIDMapping();
            CorrespondentColumnsMapping correspondentColumnsMapping = new CorrespondentColumnsMapping(fmd.getColumnMetaData(), referenceMapping);
            for (int i=0; i<referenceMapping.getNumberOfDatastoreFields(); i++)
            {
                DatastoreMapping datastoreMapping = referenceMapping.getDataStoreMapping(i);
                JavaTypeMapping mapping = dba.getMapping(datastoreMapping.getJavaTypeMapping().getJavaType());

                String userDefinedColumnName = correspondentColumnsMapping.getColumNameByIdentifier(((Column)datastoreMapping.getDatastoreField()).getName());
                if (userDefinedColumnName == null)
                {
                    throw new JDOFatalUserException(LOCALISER.msg("Table.PrimaryKeyColumnNotMapped", 
                        ((Column)datastoreMapping.getDatastoreField()).getName(), toString()));
                }
                DatastoreField refColumn;
                ColumnMetaData colmd = (fmd.getColumnMetaData().length > 0 ? fmd.getColumnMetaData()[0] : null);
                if (userDefinedColumnName.length() < 1)
                {
                    SQLIdentifier columnId = new SQLIdentifier(dba, fmd.getName()+((SQLIdentifier) datastoreMapping.getDatastoreField().getName()).getSQLIdentifier(), null);
                    refColumn = addDatastoreField(fieldType, columnId, Role.OWNER, mapping, colmd);
                }
                else
                {
                    SQLIdentifier columnId = new SQLIdentifier(dba, fmd.getName(), userDefinedColumnName);
                    refColumn = addDatastoreField(fieldType, columnId, Role.CUSTOM, mapping, colmd);
                }
                datastoreMapping.getDatastoreField().copyConfigurationTo(refColumn);
                
                setDatastoreFieldNullability(fmd,fieldType,colmd,refColumn);

                if (fmd.getNullValue() == NullValue.DEFAULT)
                {
                    refColumn.setDefaultable();
                }
                if (fmd.isPrimaryKey())
                {
                    refColumn.setAsPrimaryKey();
                    getPrimaryKey().addDatastoreField(refColumn);
                }

                m.addDataStoreMapping(getStoreManager().getMappingManager().getDatastoreMapping(mapping, storeMgr, refColumn, datastoreMapping.getJavaTypeMapping().getJavaType()));
                ((PersistenceCapableMapping)m).addJavaTypeMapping(mapping);
                if (JPOXLogger.RDBMS.isDebugEnabled())
                {
                    JPOXLogger.RDBMS.debug(LOCALISER.msg("Table.NewColumnReference", refColumn));
                }
            }
        }
        else if (m instanceof EmbeddedPCMapping)
        {
            EmbeddedMetaData emd = fmd.getEmbeddedMetaData();
            FieldMetaData[] embFmds = emd.getFieldMetaData();
            for (int i=0;i<embFmds.length;i++)
            {
                // We cant rely on there being a table for this (SCO) class so create the types here
                JavaTypeMapping embFmdMapping = dba.getMappingManager().getMapping(this, embFmds[i], dba);
                ((EmbeddedPCMapping)m).addJavaTypeMapping(embFmdMapping);
                for (int j=0; j<embFmdMapping.getNumberOfDatastoreFields(); j++)
                {
                    DatastoreMapping datastoreMapping = embFmdMapping.getDataStoreMapping(j);
                    m.addDataStoreMapping(datastoreMapping);
                }
            }
        }
        else
        {
            DatastoreField col = addDatastoreField(fmd, m);
            getStoreManager().getMappingManager().getDatastoreMapping(m, fmd, 0, getStoreManager(), col);

            if (JPOXLogger.RDBMS.isDebugEnabled())
            {
                JPOXLogger.RDBMS.debug(LOCALISER.msg("Table.NewColumn", col));
            }
        }
    }

    /**
     * Convenience method that takes a FieldMetaData and adds a Column to this table, 
     * returning the Column to the caller.
     * Currently assumes that the field requires only a single column in the datastore.
     * @param fmd The FieldMetaData
     * @param m JavaTypeMapping for the field
     * @return The Column
     */
    private DatastoreField addDatastoreField(FieldMetaData fmd, JavaTypeMapping m)
    {
        Class fieldType = fmd.getType();

        Column col;
        ColumnMetaData colmd = null;
        if (fmd.getColumnMetaData().length > 0)
        {
            colmd = fmd.getColumnMetaData()[0];
        }
        else
        {
            // If column specified add one (use any column name specified on field element)
            colmd = new ColumnMetaData(fmd, fmd.getColumn());
            fmd.addColumn(colmd);
        }
        if (colmd.getName() == null)
        {
            colmd.setName(new SQLIdentifier(dba,fmd.getName(),null).getSQLIdentifier());
        }
        col = (Column) addDatastoreField(fieldType, new SQLIdentifier(dba,fmd.getName(),fmd.getColumnMetaData()[0].getName()), Role.CUSTOM, m, colmd);

        if (fmd.isPrimaryKey())
        {
            col.setAsPrimaryKey();
            getPrimaryKey().addDatastoreField(col);
        }

        if (StoreManager.isStrategyDatastoreAttributed(fmd.getValueStrategy()))
        {
            col.setAutoIncrement(true);
        }

        if (fmd.getValueForExtension("select-function") != null)
        {
            col.setWrapperFunction(fmd.getValueForExtension("select-function"),Column.WRAPPER_FUNCTION_SELECT);
        }
        if (fmd.getValueForExtension("insert-function") != null)
        {
            col.setWrapperFunction(fmd.getValueForExtension("insert-function"),Column.WRAPPER_FUNCTION_INSERT);
        }
        if (fmd.getValueForExtension("update-function") != null)
        {
            col.setWrapperFunction(fmd.getValueForExtension("update-function"),Column.WRAPPER_FUNCTION_UPDATE);
        }

        setDatastoreFieldNullability(fmd,fieldType,colmd,col);

        if (fmd.getNullValue() == NullValue.DEFAULT)
        {
            col.setDefaultable();
        }
        
        return col;
    }
    
    /**
     * Sets the DatastoreField nullability based on metadata configuration.
     * 
     * configuration is taken in this order:
     * 
     * - ColumnMetaData (allows-null)
     * - FieldMetadata (null-value)
     * - Field type (primitive does not allows null)
     * 
     * @param fmd the FieldMetaData
     * @param fieldType the Class type for the Field
     * @param colmd the ColumnMetaData
     * @param field the DatastoreField
     */
    private void setDatastoreFieldNullability(FieldMetaData fmd, Class fieldType, ColumnMetaData colmd, DatastoreField field)
    {
        // Provide defaults for "allows-null" in ColumnMetaData
        if (colmd != null && !colmd.isAllowsNullSet())
        {
            if (fmd.isPrimaryKey())
            {
                colmd.setAllowsNull(new Boolean(false));
            }
            else if (!fieldType.isPrimitive() &&
                 fmd.getNullValue() != NullValue.EXCEPTION)
            {
                colmd.setAllowsNull(new Boolean(true));
            }
            else
            {
                colmd.setAllowsNull(new Boolean(false));
            }
            if( colmd.isAllowsNull() )
            {
                field.setNullable();
            }
        }
        // Set the nullability of the column in the datastore
        else if (colmd != null && colmd.isAllowsNullSet())
        {
            if( colmd.isAllowsNull() )
            {
                field.setNullable();
            }
        }
        else if (fmd.isPrimaryKey())
        {
            // field is never nullable
        }
        else if (!fieldType.isPrimitive() &&
             fmd.getNullValue() != NullValue.EXCEPTION)
        {
            field.setNullable();
        }
        
    }
}