/**********************************************************************
Copyright (c) 2002 Kelly Grizzle and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution.

Contributors:
2002 Mike Martin - unknown changes
2003 Andy Jefferson - replaced TableMetadata with identifier and java name
2003 Andy Jefferson - coding standards
2004 Andy Jefferson - merged with JDOView
2004 Andy Jefferson - split out CorrespondentColumnsMapping
    ...
**********************************************************************/
package org.jpox.store.rdbms.table;

import java.util.ArrayList;
import java.util.List;
import java.util.StringTokenizer;

import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOFatalUserException;
import javax.jdo.JDOUnsupportedOptionException;
import javax.jdo.JDOUserException;
import javax.jdo.spi.PersistenceCapable;

import org.jpox.StateManager;
import org.jpox.metadata.ClassMetaData;
import org.jpox.metadata.ColumnMetaData;
import org.jpox.metadata.DiscriminatorMetaData;
import org.jpox.metadata.FieldMetaData;
import org.jpox.metadata.FieldPersistenceModifier;
import org.jpox.metadata.IdentityType;
import org.jpox.metadata.NullValue;
import org.jpox.store.DatastoreClass;
import org.jpox.store.DatastoreField;
import org.jpox.store.OID;
import org.jpox.store.Role;
import org.jpox.store.exceptions.NoSuchPersistentFieldException;
import org.jpox.store.exceptions.PersistentSuperclassNotAllowedException;
import org.jpox.store.exceptions.ViewDefinitionException;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.PersistenceCapableMapping;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.store.rdbms.sqlidentifier.SQLIdentifier;
import org.jpox.util.MacroString;

/**
 * Representation of a View for a Class.
 *
 * @version $Revision: 1.4 $
 **/
public class ClassView extends ViewImpl implements FCOTable
{
    /** Ids for the identifier */
    private int nextHiValue = -1;
    private int nextLoValue = -1;
    
    private final ClassMetaData cmd;
    private final MacroString viewDef;

    private JavaTypeMapping[] fieldMappings;

    /**
     * Constructor for class view.
     * @param tableName The name of the view.
     * @param cmd The metadata for the class represented by this view.
     * @param storeMgr The RDBMS manager managing this view
     */
    public ClassView(SQLIdentifier tableName,ClassMetaData cmd,RDBMSManager storeMgr)
    {
        super(tableName, storeMgr);

        this.cmd = cmd;

        if (cmd.getIdentityType() == IdentityType.NONDURABLE)
        {
        }
        else if (cmd.getIdentityType() == IdentityType.APPLICATION)
        {
            throw new JDOUnsupportedOptionException("Application identity not supported: " + cmd.getFullClassName());
        }
        else if (cmd.getIdentityType() == IdentityType.DATASTORE)
        {
            throw new JDOUnsupportedOptionException("Datastore identity not supported: " + cmd.getFullClassName());
        }
        else
        {
            throw new JDOFatalInternalException("Invalid identity type on class " + cmd.getFullClassName());
        }

        // Extract the view definition from MetaData
        String viewImpStr = cmd.getValueForExtension("view-imports");
        String viewDefStr = null;
        if (dba.getVendorID() != null)
        {
            viewDefStr = cmd.getValueForExtension("view-definition" + '-' + dba.getVendorID());
        }
        if (viewDefStr == null)
        {
            viewDefStr = cmd.getValueForExtension("view-definition");
        }
        if (viewDefStr == null)
        {
            throw new ViewDefinitionException(cmd.getFullClassName(), viewDefStr);
        }

        viewDef = new MacroString(cmd.getFullClassName(), viewImpStr, viewDefStr);
    }


    public void initialize()
    {
        assertIsUninitialized();

        if (cmd.getPersistenceCapableSuperclass() != null)
        {
            throw new PersistentSuperclassNotAllowedException(cmd.getFullClassName());
        }

        int fieldCount = cmd.getNoOfManagedFields();
        fieldMappings = new JavaTypeMapping[fieldCount];

        for (int fieldNumber = 0; fieldNumber < fieldCount; ++fieldNumber)
        {
            FieldMetaData fmd = cmd.getManagedField(fieldNumber);

            if( fmd.getPersistenceModifier() == FieldPersistenceModifier.PERSISTENT )
            {
                fieldMappings[fieldNumber] = newColumn(fmd);
            }
            else if( fmd.getPersistenceModifier() != FieldPersistenceModifier.TRANSACTIONAL )
            {
                throw new JDOFatalInternalException("Invalid persistence modifier on field " + fmd.getName());
            }            
        }

        state = TABLE_STATE_INITIALIZED;
    }

    /**
     * Method to get the next OID for this class-view.
     * TODO Move to StoreManager in some way
     */
    public synchronized final Object newOID(StateManager sm,PersistenceCapable pc)
    {
        if (nextHiValue < 0)
        {
            nextHiValue = 0;
            nextLoValue = 0;
        }
        OID id = new OID(cmd.getFullClassName(), new Long(nextLoValue)).turnDataStoreIdentityAsNumeric();

        if (nextLoValue == Long.MIN_VALUE)
        {
            nextHiValue = nextLoValue = -1;
        }
        else
        {
            ++nextLoValue;
        }

        return id;
    }
    
    /**
     * Method to get the next AID for this class-view.
     * TODO Move to StoreManager in some way
     */
	public synchronized final Object newAID(StateManager sm, String pcClassName, Object idValue)
	{
		throw new JDOFatalInternalException("Cannot create an Application ID.");
	}

    /**
     * Accessor for the type of the class being represented by this view.
     * @return The name of the class being represented here
     */
    public String getType()
    {
        return cmd.getFullClassName();
    }

    /**
     * Method to add a new column to the view
     * @param fmd The field meta data for the field
     * @return the mapping The Java type mapping for this field
     */
    protected JavaTypeMapping newColumn(FieldMetaData fmd)
    {
        Class type = fmd.getType();

        // Cater for specified column names
        DatastoreField col;
        JavaTypeMapping mapping = dba.getMapping(type);
		if( mapping instanceof PersistenceCapableMapping )
		{
	        if (mapping instanceof PersistenceCapableMapping)
	        {
	            DatastoreClass cbt = storeMgr.getDatastoreClass(type.getName());

	            JavaTypeMapping referenceMapping = cbt.getIDMapping();
	            CorrespondentColumnsMapping correspondentColumnsMapping = new CorrespondentColumnsMapping(fmd.getColumnMetaData(), referenceMapping);
	            for (int i=0; i<referenceMapping.getNumberOfDatastoreFields(); i++)
	            {
	                String userDefinedColumnName = correspondentColumnsMapping.getColumNameByIdentifier(((Column)referenceMapping.getDataStoreMapping(i).getDatastoreField()).getName());
	                if (userDefinedColumnName == null)
	                {
	                    throw new JDOFatalUserException("Primary Key Column " + ((Column)referenceMapping.getDataStoreMapping(i).getDatastoreField()).getName() + " is not mapped");
	                }
	                DatastoreField refColumn;
	                ColumnMetaData colmd = null;
	                if( fmd.getColumnMetaData().length > 0 )
	                {
	                    colmd = fmd.getColumnMetaData()[0];
	                } 	                
	                if (userDefinedColumnName.length()<1)
	                {
	                    refColumn = addDatastoreField(type, new SQLIdentifier(dba,fmd.getName()+((Column)referenceMapping.getDataStoreMapping(i).getDatastoreField()).getName(),null), Role.OWNER,mapping,colmd);
	                }
	                else
	                {
	                    refColumn = addDatastoreField(type, new SQLIdentifier(dba,fmd.getName(),userDefinedColumnName), Role.CUSTOM,mapping,colmd);
	                }
	                ((Column)referenceMapping.getDataStoreMapping(i).getDatastoreField()).copyConfigurationTo(refColumn);
	                if (!fmd.isPrimaryKey() &&
	                    fmd.getNullValue() != NullValue.EXCEPTION &&
	                    !type.isPrimitive())
	                {
	                    refColumn.setNullable();
	                }
	                if (fmd.getNullValue() == NullValue.DEFAULT)
	                {
	                    refColumn.setDefaultable();
	                }

	                columns.add(refColumn);             
	            }            
	        }
		}
		else
		{
	        ColumnMetaData colmd = null;
	        if( fmd.getColumnMetaData().length > 0 )
	        {
	            colmd = fmd.getColumnMetaData()[0];
	        } 		    
	        if (colmd != null)
	        {
	            col = addDatastoreField(type, new SQLIdentifier(dba,fmd.getName(),fmd.getColumnMetaData()[0].getName()),Role.CUSTOM,mapping,colmd);
	        }
	        else
	        {
	            col = addDatastoreField(type, new SQLIdentifier(dba,fmd.getName(),null),Role.NONE,mapping,colmd);
	        }
	
	        if (fmd.getNullValue() != NullValue.EXCEPTION && !type.isPrimitive())
	        {
	            col.setNullable();  
	        }
	        if (fmd.getNullValue() == NullValue.DEFAULT)
	        {
	            col.setDefaultable();
	        }
		}
        return mapping;
    }

    /**
     * Accessor for whether the specified field is persistent
     * @param fieldNumber The absolute field number of the field
     * @return Whether the field is persistent
     */
    public boolean isFieldPersistent(int fieldNumber)
    {
        assertIsInitialized();

        return fieldMappings[fieldNumber] != null;
    }

    /**
     * Accessor for the mapping for the specified FieldMetaData.
     * @param fmd the FieldMetaData
     * @return The Mapping for the field.
     */
    public JavaTypeMapping getFieldMapping(FieldMetaData fmd)
    {
        assertIsInitialized();

        JavaTypeMapping m = fieldMappings[fmd.getAbsoluteFieldNumber()];

        if (m == null)
        {
            throw new NoSuchPersistentFieldException(cmd.getFullClassName(), fmd.getAbsoluteFieldNumber());
        }

        return m;
    }

    /**
     * Accessor for the field mapping for the specified field name.
     * @param fieldName Name of the field
     * @return The Java type mapping for the field
     */
    public JavaTypeMapping getFieldMapping(String fieldName)
    {
        assertIsInitialized();

        int rfn = cmd.getFieldNumber(fieldName);

        if (rfn < 0)
        {
            throw new NoSuchPersistentFieldException(cmd.getFullClassName(), fieldName);
        }

        return getFieldMapping(cmd.getManagedField(rfn));
    }

    /**
     * Method to return the necessary SQL create statements for this table.
     * @return The SQL create statements.
     */
    protected List getSQLCreateStatements()
    {
        assertIsInitialized();

        String stmt = viewDef.substituteMacros(new MacroString.MacroHandler()
            {
                public void onIdentifierMacro(MacroString.IdentifierMacro im)
                {
                    storeMgr.resolveIdentifierMacro(im);
                }

                public void onParameterMacro(MacroString.ParameterMacro pm)
                {
                    throw new JDOUserException("Parameter macros not allowed in view definitions: " + pm);
                }
            }, storeMgr.getClassLoaderResolver()
        );

        ArrayList stmts = new ArrayList();
        StringTokenizer tokens = new StringTokenizer(stmt, ";");

        while (tokens.hasMoreTokens())
        {
            stmts.add(tokens.nextToken());
        }

        return stmts;
    }
    
    /**
     * Accessor for Discriminator MetaData
     * @return Returns the Discriminator MetaData.
     */
    public final DiscriminatorMetaData getDiscriminatorMetaData()
    {
        throw new JDOFatalInternalException("Unsupported discriminator in views");
    }
    
    /**
     * Accessor for the discriminator mapping specified .
     * @return The mapping for the discriminator datastore field
     **/
    public JavaTypeMapping getDiscriminatorMapping()
    {
        throw new JDOFatalInternalException("Unsupported discriminator in views");
    }
}