/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2004 Andy Jefferson - Updated to use CorrespondentColumnsMapping to match up column specs.
2004 Andy Jefferson - changed class.forName to use ClassLoaderResolver.classForName
2005 Andy Jefferson - fix for use of "no-table" owners.
    ...
**********************************************************************/
package org.jpox.store.rdbms.table;

import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;

import javax.jdo.JDOFatalUserException;
import javax.jdo.JDOUserException;

import org.jpox.TypeManager;
import org.jpox.metadata.ClassMetaData;
import org.jpox.metadata.ColumnMetaData;
import org.jpox.metadata.FieldMetaData;
import org.jpox.store.DatastoreClass;
import org.jpox.store.DatastoreField;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.Role;
import org.jpox.store.StoreManager;
import org.jpox.store.mapping.InterfaceMapping;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.PersistenceCapableMapping;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.store.rdbms.key.PrimaryKey;
import org.jpox.store.rdbms.sqlidentifier.ColumnIdentifier;
import org.jpox.store.rdbms.sqlidentifier.InterfaceIdentifier;
import org.jpox.store.rdbms.sqlidentifier.SQLIdentifier;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;

/**
 * Class to create columns.
 * Takes the definition provided and adds columns to the associated table.
 *
 * @version $Revision: 1.7 $
 */
public final class ColumnCreator
{
    /** Localiser for messages. */
    protected static final Localiser LOCALISER=Localiser.getInstance("org.jpox.store.rdbms.table.Localisation");
    
    /** Type Manager for JPOX. */
    private static final TypeManager TM=TypeManager.getTypeManager();
    
    private ColumnCreator()
    {
    }

    /**
     * Create columns for interface fields.
     * @param mapping the InterfaceMapping
     * @param table the Table which will hold the columns
     * @param fmd the Field
     */
    public static void createColumnsForInterface(JavaTypeMapping mapping, Table table, FieldMetaData fmd)
    {       
        createColumnsForFieldInterface(mapping, table, fmd, false, true, null, Role.ELEMENT, fmd.getColumnMetaData(), null);
    }

    /**
     * Create columns for interface fields.
     * This call createColumnsForField for each implementation class
     * @param mapping the InterfaceMapping
     * @param table the Table which will hold the columns
     * @param fmd the Field
     */
    private static void createColumnsForFieldInterface(
            JavaTypeMapping mapping,
            Table table,
            FieldMetaData fmd,
            boolean isPrimaryKey,
            boolean isNullable,
            PrimaryKey primaryKey,
            Role role,
            ColumnMetaData[] columnMetaData,
            JavaTypeMapping mappingForAdaptiveColumn)
    {
        String[] implTypes;
        if (Role.KEY == role)
        {
            implTypes = fmd.getValuesForExtension("key-implementation-classes");
        }
        else if (Role.VALUE == role)
        {
            implTypes = fmd.getValuesForExtension("value-implementation-classes");
        }
        else
        {
            implTypes = fmd.getValuesForExtension("implementation-classes");
        }
        if (implTypes == null)
        {
            //TODO: Throw exception? For a while yes, maybe later we can
            //have a way of loading the interface metadata better
            throw new JDOUserException("No implementation classes specified for " + fmd.getName());
        }
        
        // use set to avoid duplicated declared classes
        Set implTypesSet = new HashSet();
        for (int i=0; i<implTypes.length; i++)
        {
            implTypesSet.add(implTypes[i]);
        }
        
        int colPos = 0;
        //create columns for all possible implementations
        for (Iterator it=implTypesSet.iterator(); it.hasNext();)
        {
            StoreManager storeMgr = table.getStoreManager();
            
            Class type = storeMgr.getClassLoaderResolver().classForName((String) it.next());
            if (type == null)
            {
                throw new JDOUserException("No implementation classes found.");
            }
            if (type.isInterface())
            {
                throw new JDOUserException("You can't have fields of interfaces with interfaces as concrete implementations." + fmd.getName());
            }
            JavaTypeMapping m = storeMgr.getDatastoreAdapter().getMapping(type);
            
            //obtain the column names for this type
            ColumnMetaData[] columnMetaDataForType = null;
            if (columnMetaData != null && columnMetaData.length > 0 && columnMetaData.length <= colPos+m.getNumberOfDatastoreFields())
            {
                columnMetaDataForType = new ColumnMetaData[m.getNumberOfDatastoreFields()];
                System.arraycopy(columnMetaData, colPos, columnMetaDataForType, 0, columnMetaDataForType.length);
                colPos += columnMetaDataForType.length;
            }
            createColumnsForField(type,
                mapping, 
                table, 
                storeMgr, 
                fmd, 
                isPrimaryKey, 
                isNullable, 
                primaryKey, 
                role, 
                mappingForAdaptiveColumn,
                columnMetaDataForType,
                new IdentifierBuilder()
                {
                public DatastoreIdentifier create(StoreManager storeMgr,
                        FieldMetaData fmd,
                        ClassMetaData cmd,
                        DatastoreIdentifier identifier)
                {
                    SQLIdentifier identiferType = new InterfaceIdentifier(storeMgr.getDatastoreAdapter(), fmd, cmd);
                    SQLIdentifier identiferTypeColumn; 
                    identiferTypeColumn = new SQLIdentifier(
                        storeMgr.getDatastoreAdapter(),
                        identiferType.getSQLIdentifier()+
                        "_"+
                        ((SQLIdentifier) identifier).getSQLIdentifier());
                    return identiferTypeColumn;
                }
                }
            );

            JPOXLogger.RDBMS.info("Added column for implementation type [" + type + "] for field [" + fmd.getName() + "]");
        }
    }

    private static JavaTypeMapping createColumnsForField(
            Class type,
            JavaTypeMapping mapping,
            Table table, 
            StoreManager storeMgr,
            FieldMetaData fmd, 
            boolean isPrimaryKey,
            boolean isNullable,
            PrimaryKey primaryKey,
            Role role,
            JavaTypeMapping mappingAdapterColumn,
            ColumnMetaData[] columnMetaData,
            IdentifierBuilder identifierBuilder)
    {
        if (!type.isInterface() && !TM.isSupportedType(type.getName()))
        {
            if (mapping == null)
            {
                mapping = storeMgr.getDatastoreAdapter().getMapping(type);
            }
            JavaTypeMapping container = mapping;
            if (mapping instanceof InterfaceMapping)
            {
                container = storeMgr.getDatastoreAdapter().getMapping(type);
                ((InterfaceMapping) mapping).addJavaTypeMapping(container);
            }

            DatastoreClass ownerTable = storeMgr.getDatastoreClass(type.getName());
            if (ownerTable == null)
            {
                // Maybe the owner hasn't got its own table (e.g "no-table" inheritance strategy)
                ClassMetaData ownerCmd = storeMgr.getMetaDataManager().getMetaDataForClass(type.getName(), storeMgr.getClassLoaderResolver());
                ClassMetaData[] ownerCmds = storeMgr.getClassesManagingTableForClass(ownerCmd);
                if (ownerCmds == null)
                {
                    throw new JDOFatalUserException(LOCALISER.msg("Table.OwnerHasNoTable", type.getName()));
                }
                // Use the first one for now
                // TODO Really ought to use all owner subclasses.
                ownerTable = storeMgr.getDatastoreClass(ownerCmds[0].getFullClassName());
            }
            if (ownerTable != null)
            {
                JavaTypeMapping m = ownerTable.getIDMapping();

                // For each datastore mapping, add a column.
                CorrespondentColumnsMapping correspondentColumnsMapping = new CorrespondentColumnsMapping(columnMetaData, m);
                for (int i=0; i<m.getNumberOfDatastoreFields(); i++)
                {
                    DatastoreField column;
                    JavaTypeMapping datastoreMapping = storeMgr.getDatastoreAdapter().getMapping(m.getDataStoreMapping(i).getJavaTypeMapping().getJavaType());
                    String userDefinedColumnName = correspondentColumnsMapping.getColumNameByIdentifier(((Column)m.getDataStoreMapping(i).getDatastoreField()).getName());
                    if (userDefinedColumnName == null || userDefinedColumnName.length() == 0)
                    {
                        // User hasn't provided a name, so we use default naming
                        ColumnMetaData colmd = null;
                        if (columnMetaData != null && columnMetaData.length > 0)
                        {
                            colmd = columnMetaData[0];
                        }
                        column = table.addDatastoreField(type, identifierBuilder.create(storeMgr, fmd, 
                            storeMgr.getMetaDataManager().getMetaDataForClass(type), 
                            ((Column) m.getDataStoreMapping(i).getDatastoreField()).getName()), role, datastoreMapping, colmd);
                    }
                    else
                    {
                        // User defined name, so we use that.
                        ColumnMetaData colmd = null;
                        if (columnMetaData != null && columnMetaData.length > 0)
                        {
                            colmd = columnMetaData[0];
                        }
                        column = table.addDatastoreField(type, new SQLIdentifier(storeMgr.getDatastoreAdapter(), userDefinedColumnName), Role.CUSTOM, 
                            datastoreMapping, colmd);
                    }

                    ((Column) m.getDataStoreMapping(i).getDatastoreField()).copyConfigurationTo(column);
                    if (isPrimaryKey)
                    {
                        column.setAsPrimaryKey();
                        primaryKey.addDatastoreField(column);
                    }
                    if (isNullable)
                    {
                        column.setNullable();
                    }
                    storeMgr.getMappingManager().getDatastoreMapping(datastoreMapping, storeMgr, column,
                        m.getDataStoreMapping(i).getJavaTypeMapping().getJavaType());
                    ((PersistenceCapableMapping) container).addJavaTypeMapping(datastoreMapping);
                }
            }
        }
        else if (type.isInterface())
        {
            mapping = storeMgr.getDatastoreAdapter().getMapping(type);
            createColumnsForFieldInterface(mapping, table, fmd, isPrimaryKey, isNullable, primaryKey, role, columnMetaData, mappingAdapterColumn);
        }
        else
        {
            if (mapping == null)
            {
                mapping = storeMgr.getDatastoreAdapter().getMapping(type);
            }

            // Add column for the field
            ColumnMetaData colmd = null;
            if (columnMetaData != null && columnMetaData.length > 0)
            {
                colmd = columnMetaData[0];
            }

            DatastoreField column = null;
            if (colmd != null && colmd.getName() != null)
            {
                column = table.addDatastoreField(type, new SQLIdentifier(storeMgr.getDatastoreAdapter(), colmd.getName()), Role.CUSTOM, mapping, colmd);
            }
            else
            {
                column = table.addDatastoreField(type, getDefaultNameForType(storeMgr, type.getName()), role, mapping, colmd);
            }
            storeMgr.getMappingManager().getDatastoreMapping(mapping, storeMgr, column, type);

            if (isPrimaryKey)
            {
                if (mappingAdapterColumn != null)
                {
                    // Add an adapter column to allow duplicate elements in any Collection
                    colmd = null;
                    if (fmd.getColumnMetaData() != null && fmd.getColumnMetaData().length > 0)
                    {
                        colmd = fmd.getColumnMetaData()[0];
                    }

                    DatastoreField adapterColumn = null;
                    if (fmd.hasExtension("adapter-column-name"))
                    {
                        adapterColumn = table.addDatastoreField(mappingAdapterColumn.getType(), 
                            new SQLIdentifier(storeMgr.getDatastoreAdapter(), fmd.getValueForExtension("adapter-column-name")), 
                            Role.CUSTOM, mappingAdapterColumn, colmd);
                    }
                    else
                    {
                        adapterColumn = table.addDatastoreField(mappingAdapterColumn.getType(), 
                            new ColumnIdentifier(storeMgr.getDatastoreAdapter(), "ADPT_PK"),
                            Role.INDEX, mappingAdapterColumn, colmd);
                    }
                    storeMgr.getMappingManager().getDatastoreMapping(mappingAdapterColumn, storeMgr, adapterColumn,
                        mappingAdapterColumn.getJavaType());
                    adapterColumn.setAsPrimaryKey();
                    primaryKey.addDatastoreField(adapterColumn);
                }
                else
                {
                    column.setAsPrimaryKey();
                    primaryKey.addDatastoreField(column);
                }
            }
            if (isNullable)
            {
                column.setNullable();
            }
        }
        return mapping;
    }

    /**
     * create columns
     * @param co the ColumnMetaData
     * @return the mapping for the columns created
     */
    public static JavaTypeMapping createColumns(ColumnOptions co)
    {
        return createColumnsForField(
                co.type,
                null,
                co.table,
                co.storeMgr, 
                co.fmd, 
                co.isPrimaryKey, 
                co.isNullable, 
                co.primaryKey, 
                co.role, 
                co.mappingAdapterColumn, 
                co.columnMetaData,
                new IdentifierBuilder()
                {
                    public DatastoreIdentifier create(StoreManager storeMgr, FieldMetaData fmd, ClassMetaData cmd,
                            DatastoreIdentifier identifier)
                    {
                        return identifier;
                    }
                }
        );
    }

    /**
     * Inner class defining the meta data for a column.
     **/
    public static final class ColumnOptions
    {       
        /**
         * the column options included in the jdo metadata file.
         */
        final ColumnMetaData[] columnMetaData;
        
        /**
         * Is the column a primary key.
         * If the column is a primary key, can't be nullable
         * true of is part of the primary key
         */
        final boolean isPrimaryKey;

        /**
         * Gets the PrimaryKey that the column is part of. The newly created
         * column is added to the PrimaryKey.
         * Required only if the column is part of the primary key
         * the PrimaryKey or null if the column is not part of the primary key
         */
        final PrimaryKey primaryKey;

        /**
         * A mapping for an adapter column. When the newly created column is
         * part of the primary key, but the data type for column created is
         * is not supported as primary key in the database, an adapter column
         * is created to be used as primary key (a kind of index), and the
         * original column is not set as primary key.
         * Required only if the column is primary key is can fits in the above
         * case null if an adaptive column is not supported or not needed.
         */
        final JavaTypeMapping mappingAdapterColumn;

        /**
         * Is the column is nullable.
         * A scenario mandates the column be nullable:
         * - if the type is a java Interface, and it has many implementations,
         *   must be nullable.
         * A scenario mandates the column be non-nullable:
         * - if the field is a primary key.
         */
        boolean isNullable;

        /**
         * The role to apply when creating identifiers
         */
        final Role role;

        /**
         * Gets the mapping type for the column.
         * Some conditions apply for creating columns caused by database rules:
         * - A primary key can't have a null value in the database, and if the
         *   type is a java Interface, it can have only one implementation class
         * - if the type is a java Interface, and it has many implementations,
         *   columns must be nullable.
         */
        final Class type;

        /**
         * The RDBMSManager that manages the table
         */
        final RDBMSManager storeMgr;

        /**
         * The FieldMetaData which has launched the process of creating this
         * column
         */
        final FieldMetaData fmd;

        /**
         * The table which the column is part of.
         */
        final Table table;
 
        /**
         * Constructor.
         * @param isPrimaryKey Whether the column is a primary key column
         * @param primaryKey The primary key info
         * @param isNullable Whether the column is nullable
         * @param role The security role
         * @param type Type of field stored in this column
         * @param storeMgr The store manager
         * @param fmd The FieldMetaData for the column
         * @param table The table to add the column to.
         * @param columnMetaData MetaData of the column
         * @param mappingAdapterColumn 
         */
        public ColumnOptions(
                final boolean isPrimaryKey, final PrimaryKey primaryKey,
                boolean isNullable, final Role role, final Class type,
                final RDBMSManager storeMgr, final FieldMetaData fmd,
                final Table table,
                final ColumnMetaData[] columnMetaData,
                final JavaTypeMapping mappingAdapterColumn)
        {
            super();
            this.isPrimaryKey = isPrimaryKey;
            this.primaryKey = primaryKey;
            this.isNullable = isNullable;
            this.role = role;
            this.type = type;
            this.storeMgr = storeMgr;
            this.fmd = fmd;
            this.table = table;
            this.columnMetaData = columnMetaData;
            this.mappingAdapterColumn = mappingAdapterColumn;
        }
    }
 
    /**
     * Accessor for the column name for the specified class.
     * @param storeMgr the RDBMSManager
     * @param class_name Name of class for which the column is required.
     * @return The SQL identifier of the column. 
     **/
    private static DatastoreIdentifier getDefaultNameForType(StoreManager storeMgr,
                                                          String class_name)
    {
        if (TM.isSupportedType(class_name))
        {
            String javaName = class_name.substring(class_name.lastIndexOf('.') + 1);
            return new ColumnIdentifier(storeMgr.getDatastoreAdapter(), javaName);
        }
        else
        {
            return storeMgr.getDatastoreClass(class_name).getName();
        }
    }
 
    /**
     * Inner class providing a builder for an identifier based on the metadata.
     **/
    private static interface IdentifierBuilder
    {
        /**
         * Creates an identifier.
         * @param storeMgr The Store Manager
         * @param fmd The field meta data
         * @param cmd The class meta data
         * @param identifier The identifier
         * @return The SQL Identifier
         */
        DatastoreIdentifier create(StoreManager storeMgr, FieldMetaData fmd, ClassMetaData cmd, DatastoreIdentifier identifier);
    }
}
