/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
2004 Erik Bengtson - original version in ClassTable
    ...
**********************************************************************/
package org.jpox.store.rdbms.table;

import java.util.HashMap;
import java.util.Map;

import javax.jdo.JDOFatalUserException;

import org.jpox.metadata.ColumnMetaData;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.rdbms.sqlidentifier.SQLIdentifier;
import org.jpox.util.Localiser;

/**
 * Class to make correspondence between columns in one side of an
 * association to the mapping at the other side. The 2 sides of the association are
 * referred to as "sideA" and "sideB". The JDO 2 metadata allows definition
 * of the correspondence using the 
 * <column name="{column-name}" target="{target-name}"/> syntax.
 * <P>
 * This means that the column specified on sideA will be mapped to the specified "target"
 * column on sideB. If no target is provided then the first available sideB column
 * is used for the mapping. Where no columns are defined on sideA, then they will be
 * created to match those on sideB. Checks are made for consistency of the sideA data.
 * 
 * @version $Revision: 1.2 $
 */
public class CorrespondentColumnsMapping
{
    /** Localiser for messages. */
    protected static final Localiser LOCALISER=Localiser.getInstance("org.jpox.store.rdbms.table.Localisation");

    /** Map of the column names in side A, keyed by the side B column identifier. */
    private final Map columnNamesSideA_BySideBIdentifiers = new HashMap();

    private final String columnsName;

    /**
     * Constructor. Takes the sideB definition and the side A definition and matches them up as
     * defined by the user, and if not defined by the user matches them as best it can.
     * @param colmds the column definitions for side A
     * @param mappingSideB the mapping in the side B
     */
    public CorrespondentColumnsMapping(ColumnMetaData[] colmds, JavaTypeMapping mappingSideB)
    {
        // Go through the user-defined columns and allocate them as required
        if (colmds != null)
        {
            StringBuffer str=new StringBuffer("Columns [");
            for (int i=0;i<colmds.length;i++)
            {
                str.append(colmds[i].getName());
                if (i < colmds.length-1)
                {
                    str.append(", ");
                }
            }
            str.append("]");
            columnsName = str.toString();

            // Check if too many columns have been defined
            if (colmds.length > mappingSideB.getNumberOfDatastoreFields())
            {
                throw new JDOFatalUserException(LOCALISER.msg("CorrespondentColumnsMapping.ColumnNumbersMismatch", 
                    columnsName, "" + colmds.length, "" + mappingSideB.getNumberOfDatastoreFields()));
            }

            // Retrieve sideB column names
            DatastoreIdentifier[] sideBidentifiers = new SQLIdentifier[mappingSideB.getNumberOfDatastoreFields()];
            boolean[] sideButilised = new boolean[mappingSideB.getNumberOfDatastoreFields()];
            for (int i = 0; i < mappingSideB.getNumberOfDatastoreFields(); i++)
            {
                sideBidentifiers[i] = mappingSideB.getDataStoreMapping(i).getDatastoreField().getName();
                sideButilised[i] = false;
            }

            // Allocate the user-defined columns using the sideB list where target column has been defined
            for (int i = 0; i < colmds.length; i++)
            {
                String userDefinedColumnName = colmds[i].getName();
                String targetColumnName = colmds[i].getTarget();

                // Find the target on sideB
                if (targetColumnName != null)
                {
                    boolean targetExists = false;
                    for (int j = 0; j < sideBidentifiers.length; j++)
                    {
                        if (((SQLIdentifier) sideBidentifiers[j]).getSQLIdentifier().equals(targetColumnName) &&
                            !sideButilised[j])
                        {
                            putColumnName(sideBidentifiers[j], userDefinedColumnName);
                            sideButilised[j] = true;
                            targetExists = true;
                            
                            break;
                        }
                    }
                    
                    // Check for invalid sideB column
                    if (!targetExists)
                    {
                        throw new JDOFatalUserException(LOCALISER.msg("CorrespondentColumnsMapping.ColumnTargetDoesntExist", 
                            columnsName, userDefinedColumnName, targetColumnName));
                    }
                }
            }
            
            // Allocate the user defined columns using the sideB list where target column has not been defined
            for (int i = 0; i < colmds.length; i++)
            {
                String userDefinedColumnName = colmds[i].getName();
                String targetColumnName = colmds[i].getTarget();

                if (targetColumnName == null)
                {
                    // Find the next unutilised column on sideB
                    for (int j = 0; j < sideBidentifiers.length; j++)
                    {
                        if (!sideButilised[j])
                        {
                            putColumnName(sideBidentifiers[j], userDefinedColumnName);
                            sideButilised[j] = true;
                            break;
                        }
                    }
                }
            }

            // Allocate any missing columns
            for (int i = colmds.length; i < mappingSideB.getNumberOfDatastoreFields(); i++)
            {
                // Find next unallocated sideB column
                DatastoreIdentifier sideBidentifier = null;
                for (int j=0; j < sideBidentifiers.length; j++)
                {
                    if (!sideButilised[j])
                    {
                        sideBidentifier = sideBidentifiers[j];
                        sideButilised[j] = true;
                        break;
                    }
                }
                if (sideBidentifier == null)
                {
                    throw new JDOFatalUserException(LOCALISER.msg("CorrespondentColumnsMapping.NoRemainingColumnsToMapTo", 
                        columnsName, "" + i));
                }
                putColumnName(sideBidentifier, "");
            }
        }
        else
        {
            columnsName = null;
            for (int i = 0; i < mappingSideB.getNumberOfDatastoreFields(); i++)
            {
                final DatastoreIdentifier sideBidentifier;
                final String sideAcolumnName;
                sideBidentifier = mappingSideB.getDataStoreMapping(i).getDatastoreField().getName();
                sideAcolumnName = ""; // no column name defined by the user
                putColumnName(sideBidentifier, sideAcolumnName);
            }
        }
    }

    /**
     * Accessor for the column name in side A given the side B identifier.
     * @param name side B identifier
     * @return Column name in side A.
     */
    public String getColumNameByIdentifier(DatastoreIdentifier name)
    {
        return (String)columnNamesSideA_BySideBIdentifiers.get(name);
    }

    /**
     * Method to associate a sideB identifier with a sideA column
     * @param identifier side B identifier
     * @param columnName side A column
     */
    private void putColumnName(DatastoreIdentifier identifier, String columnName)
    {
        if (columnNamesSideA_BySideBIdentifiers.put(identifier, columnName) != null)
        {
            throw new JDOFatalUserException(LOCALISER.msg("CorrespondentColumnsMapping.ColumnMappedTwice", identifier, columnsName));
        }
    }
}