/**********************************************************************
Copyright (c) 2002 Kelly Grizzle (TJDO) and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - replaced TableMetadata with SQLIdentifier and java name
2004 Marco Schulze - added advance-check via TypeManager.isSupportedType(...)
     ...
 **********************************************************************/
package org.jpox.store.rdbms.table;

import java.util.ArrayList;
import java.util.List;

import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOUserException;

import org.jpox.ClassLoaderResolver;
import org.jpox.PMFConfiguration;
import org.jpox.metadata.CollectionMetaData;
import org.jpox.metadata.DiscriminatorMetaData;
import org.jpox.metadata.FieldMetaData;
import org.jpox.metadata.ForeignKeyMetaData;
import org.jpox.store.Role;
import org.jpox.store.exceptions.NoTableManagedException;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.store.rdbms.key.ForeignKey;
import org.jpox.store.rdbms.sqlidentifier.SQLIdentifier;

/**
 * Table representation of a list (join table).
 * <P>There are 2 ways of handling a relationship between 2 tables :-
 * <OL>
 * <LI>by having a foreign key to the other table on one table</LI>
 * <LI>by using a join table containing the primary keys of the 2 tables</LI>
 * </OL>
 * The 2 tables at the two ends of the relationship are referred to as
 * "owner" and "element", with the "owner" being the '1' end of the 1-N
 * relationship.
 * 
 * @version $Revision: 1.5 $
 */
public class ListTable extends FieldTable implements SCOTable
{
    /** Mapping of an owner (OID) column back to the owner table PK */
    private JavaTypeMapping ownerMapping;

    /** Mapping of an element (EID) column back to the element table PK */
    private JavaTypeMapping elementMapping;

    /** Mapping of an index column to retain list ordering info. */
    private JavaTypeMapping indexMapping;

    /**
     * Constructor.
     * @param tableName Identifier name of the table
     * @param fmd MetaData for the field of the owner
     * @param storeMgr The Store Manager managing these tables.
     */
    public ListTable(SQLIdentifier tableName, 
                     FieldMetaData fmd,
                     RDBMSManager storeMgr)
    {
        super(tableName, fmd, storeMgr);
    }

    /**
     * Method to initialise the table definition.
     */
    public void initialize()
    {
        assertIsUninitialized();
        CollectionMetaData colmd = fmd.getCollection();
        if (colmd == null)
        {
            throw new JDOUserException(LOCALISER.msg("Table.CollectionMetaDataError", fmd));
        }
        ClassLoaderResolver clr = storeMgr.getClassLoaderResolver();
        Class owner_class = null;
        Class element_class = null;
        Class index_class = Integer.class;
        owner_class = clr.classForName(ownerType);
        element_class = clr.classForName(colmd.getElementType());

        // This assumes that the column mapping is defined on the owner field
        // TODO We need to allow for bidirectional relationships where it can be defined
        // on the element field.
        ownerMapping = 
            ColumnCreator.createColumns(new ColumnCreator.ColumnOptions(true,
                getPrimaryKey(), false, Role.OWNER, owner_class, storeMgr, fmd, this, 
                fmd.getJoinMetaData() == null ? null : fmd.getJoinMetaData().getColumnMetaData(),
                null));
        indexMapping = 
            ColumnCreator.createColumns(new ColumnCreator.ColumnOptions(true,
                getPrimaryKey(), false, Role.INDEX, index_class, storeMgr, fmd, this, 
                fmd.getOrderMetaData() == null ? null : fmd.getOrderMetaData().getColumnMetaData(),
                null));
        elementMapping = 
            ColumnCreator.createColumns(new ColumnCreator.ColumnOptions(false, null,
                true, Role.ELEMENT, element_class, storeMgr, fmd, this,
                fmd.getElementMetaData() == null ? null : fmd.getElementMetaData().getColumnMetaData(), null));

        state = TABLE_STATE_INITIALIZED;
    }

    /**
     * Accessor for the expected foreign keys for this table.
     * @return The expected foreign keys.
     */
    protected List getExpectedForeignKeys()
    {
        assertIsInitialized();

        boolean autoMode = false;
        if (storeMgr.getPMFContext().getPmfConfiguration().getForeignKeyCreateMode() == PMFConfiguration.FOREIGN_KEY_AUTO_MODE)
        {
            autoMode = true;
        }

        ArrayList foreignKeys = new ArrayList();
        try
        {
            ClassTable referencedTable = (ClassTable) storeMgr.getTable(fmd.getClassName(true));
            ForeignKeyMetaData fkmd = fmd.getForeignKeyMetaData();
            if (referencedTable != null)
            {
                if (fkmd != null || autoMode)
                {
                    ForeignKey fk = new ForeignKey(ownerMapping, dba, referencedTable, true);
                    fk.setForMetaData(fkmd);
                    foreignKeys.add(fk);
                }
            }
            referencedTable = (ClassTable) storeMgr.getTable(((CollectionMetaData) fmd.getContainer()).getElementType());
            if (referencedTable != null)
            {
                if (fkmd != null || autoMode)
                {
                    ForeignKey fk = new ForeignKey(elementMapping, dba, referencedTable, true);
                    fk.setForMetaData(fkmd);
                    foreignKeys.add(fk);
                }
            }
        }
        catch (NoTableManagedException e)
        {
            //expected when no table exists
        }
        return foreignKeys;
    }

    /**
     * Accessor for the "owner" mapping end of the relationship. This will be
     * the primary key of the owner table.
     * @return The column mapping for the owner.
     */
    public JavaTypeMapping getOwnerMapping()
    {
        assertIsInitialized();
        return ownerMapping;
    }

    /**
     * Accessor for the "element" mapping end of the relationship. This will be
     * the primary key of the element table.
     * @return The column mapping for the element.
     */
    public JavaTypeMapping getElementMapping()
    {
        assertIsInitialized();
        return elementMapping;
    }

    /**
     * Accessor for the "index" mapping end of the relationship.
     * @return The column mapping for the index.
     */
    public JavaTypeMapping getIndexMapping()
    {
        assertIsInitialized();
        return indexMapping;
    }

    /**
     * Accessor for Discriminator MetaData
     * @return Returns the Discriminator MetaData.
     */
    public final DiscriminatorMetaData getDiscriminatorMetaData()
    {
        throw new JDOFatalInternalException("Unsupported discriminator in lists");
    }
    /**
     * Accessor for the discriminator mapping specified .
     * @return The mapping for the discriminator datastore field
     **/
    public JavaTypeMapping getDiscriminatorMapping()
    {
        throw new JDOFatalInternalException("Unsupported discriminator in lists");
    }
    
}