/**********************************************************************
Copyright (c) 2002 Kelly Grizzle and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
2002 Mike Martin - unknown changes
2003 Andy Jefferson - added localiser
2003 Andy Jefferson - replaced TableMetadata with identifier
2004 Marco Schulze - added advance-check via TypeManager.isSupportedType(...)
    ...
**********************************************************************/
package org.jpox.store.rdbms.table;

import java.util.ArrayList;
import java.util.List;

import javax.jdo.JDOFatalInternalException;
import javax.jdo.JDOUserException;

import org.jpox.ClassLoaderResolver;
import org.jpox.PMFConfiguration;
import org.jpox.metadata.DiscriminatorMetaData;
import org.jpox.metadata.FieldMetaData;
import org.jpox.metadata.ForeignKeyMetaData;
import org.jpox.metadata.MapMetaData;
import org.jpox.store.Role;
import org.jpox.store.exceptions.NoTableManagedException;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.store.rdbms.key.ForeignKey;
import org.jpox.store.rdbms.sqlidentifier.SQLIdentifier;

/**
 * Representation of a join table for a Map.
 * There are 2 ways of handling a relationship between 2 tables, firstly by
 * having a foreign key to the other table on one table, and secondly by using
 * a join table containing the primary keys of the 2 tables. This represents
 * the second of these. The 2 tables are referred to as "owner", "key", and
 * "value" with the "owner" being the '1' end of the 1-N relationship.
 *
 * @version $Revision: 1.4 $
 **/
public class MapTable extends FieldTable implements SCOTable
{
    private JavaTypeMapping ownerMapping;
    private JavaTypeMapping keyMapping;
    private JavaTypeMapping valueMapping;

    /**
     * when the key mapping columns can't be part of the primary key
     * by datastore limitations like BLOB types.
     * An adapter mapping is used to be a kind of "index"
     */
    private JavaTypeMapping adapterMapping;

    /**
     * Constructor
     *
     * @param tableName Identifier name of the table
     * @param fmd       MetaData for the field of the owner
     * @param storeMgr  The Store Manager managing these tables.
     **/
    public MapTable(SQLIdentifier tableName,FieldMetaData fmd,RDBMSManager storeMgr)
    {
        super(tableName,fmd,storeMgr);
    }

    /**
     * Method to initialise the table definition.
     **/
    public void initialize()
    {
        assertIsUninitialized();

        MapMetaData mmd = fmd.getMap();
        if (mmd == null)
        {
            throw new JDOUserException(LOCALISER.msg("Table.MapMetaDataError",fmd));
        }

        // Load key and value classes
        ClassLoaderResolver clr=storeMgr.getClassLoaderResolver();
        Class owner_class=null;
        Class key_class=null;
        Class value_class=null;
        owner_class = clr.classForName(ownerType);
        key_class = clr.classForName(mmd.getKeyType());
        value_class = clr.classForName(mmd.getValueType());

        ownerMapping = ColumnCreator.createColumns(
            new ColumnCreator.ColumnOptions(true, getPrimaryKey(), false, Role.OWNER,
                owner_class, storeMgr, fmd, this,
                fmd.getJoinMetaData() == null ? null : fmd.getJoinMetaData().getColumnMetaData(),
                null));

        adapterMapping = dba.getMapping(long.class);

        keyMapping = ColumnCreator.createColumns(
            new ColumnCreator.ColumnOptions(true, getPrimaryKey(), false, Role.KEY,
                key_class, storeMgr, fmd, this,
                fmd.getKeyMetaData() == null ? null : fmd.getKeyMetaData().getColumnMetaData(),
                adapterMapping));

        if (adapterMapping.getNumberOfDatastoreFields() < 1)
        {
            adapterMapping = null;
        }
 
        valueMapping = ColumnCreator.createColumns(
            new ColumnCreator.ColumnOptions(false, null, true, Role.VALUE,
                value_class, storeMgr, fmd, this,
                fmd.getValueMetaData() == null ? null : fmd.getValueMetaData().getColumnMetaData(),
                null));

        state = TABLE_STATE_INITIALIZED;
    }

    /**
     * Accessor for the "owner" mapping end of the relationship. This will be
     * the primary key of the owner table.
     * @return The column mapping for the owner.
     **/
    public JavaTypeMapping getOwnerMapping()
    {
        assertIsInitialized();

        return ownerMapping;
    }

    /**
     * Accessor for the "key" mapping end of the relationship.
     * @return The column mapping for the element.
     **/
    public JavaTypeMapping getKeyMapping()
    {
        assertIsInitialized();

        return keyMapping;
    }

    /**
     * Accessor for the "value" mapping end of the relationship.
     * @return The column mapping for the element.
     **/
    public JavaTypeMapping getValueMapping()
    {
        assertIsInitialized();

        return valueMapping;
    }

    /**
     * Accessor for adapeter mapping.
     * The columns in this mapping are part of the primary key.
     * @return The column mapping for objects without identity or not
     *         supported as part of the primary keys
     **/
    public JavaTypeMapping getAdapterMapping()
    {
        assertIsInitialized();

        return adapterMapping;
    }

    /**
     * Accessor for the "element" mapping end of the relationship. This will be
     * the primary key of the element table.
     * @return The column mapping for the element.
     **/
    public JavaTypeMapping getElementMapping()
    {
        assertIsInitialized();

        return valueMapping;
    }
    
    /**
     * Accessor for the expected foreign keys for this table.
     * 
     * @return The expected foreign keys.
     */
    protected List getExpectedForeignKeys()
    {
        assertIsInitialized();

        boolean autoMode = false;
        if (storeMgr.getPMFContext().getPmfConfiguration().getForeignKeyCreateMode() == PMFConfiguration.FOREIGN_KEY_AUTO_MODE)
        {
            autoMode = true;
        }

        ArrayList foreignKeys = new ArrayList();
        try
        {
            ClassTable referencedTable = (ClassTable) storeMgr.getTable(fmd.getClassName(true));
            ForeignKeyMetaData fkmd = fmd.getForeignKeyMetaData();
            if (referencedTable != null)
            {
                if (fkmd != null || autoMode)
                {
                    ForeignKey fk = new ForeignKey(ownerMapping,dba,referencedTable, true);
                    fk.setForMetaData(fkmd);
                    foreignKeys.add(fk);
                }
            }
            referencedTable = (ClassTable) storeMgr.getTable(fmd.getMap().getValueType());
            if (referencedTable != null)
            {
                if (fkmd != null || autoMode)
                {
                    ForeignKey fk = new ForeignKey(valueMapping,dba,referencedTable, true);
                    fk.setForMetaData(fkmd);
                    foreignKeys.add(fk);
                }
            }
            referencedTable = (ClassTable) storeMgr.getTable(fmd.getMap().getKeyType());
            if (referencedTable != null)
            {
                if (fkmd != null || autoMode)
                {
                    ForeignKey fk = new ForeignKey(keyMapping,dba,referencedTable, true);
                    fk.setForMetaData(fkmd);
                    foreignKeys.add(fk);
                }
            }
        }
        catch (NoTableManagedException e)
        {
            //expected when no table exists
        }
        return foreignKeys;
    }
    
    /**
     * Accessor for Discriminator MetaData
     * @return Returns the Discriminator MetaData.
     */
    public final DiscriminatorMetaData getDiscriminatorMetaData()
    {
        throw new JDOFatalInternalException("Unsupported discriminator in map");
    }
    /**
     * Accessor for the discriminator mapping specified .
     * @return The mapping for the discriminator datastore field
     **/
    public JavaTypeMapping getDiscriminatorMapping()
    {
        throw new JDOFatalInternalException("Unsupported discriminator in map");
    }
    
}