/**********************************************************************
Copyright (c) 2002 Kelly Grizzle and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
2002 Mike Martin - unknown changes
2003 Andy Jefferson - added localiser
2003 Andy Jefferson - added catalog name
2003 Andy Jefferson - added default catalog/schema
    ...
**********************************************************************/
package org.jpox.store.rdbms.table;

import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.SQLException;

import javax.jdo.JDODataStoreException;
import javax.jdo.JDOFatalInternalException;

import org.jpox.PMFConfiguration;
import org.jpox.metadata.DiscriminatorMetaData;
import org.jpox.store.DatastoreField;
import org.jpox.store.Role;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.store.rdbms.sqlidentifier.SQLIdentifier;
import org.jpox.store.rdbms.sqlidentifier.TableIdentifier;
import org.jpox.util.JPOXLogger;

/**
 * Utility class used for detecting database schema existence and provides
 * means of determining the schema name.
 *
 * @version $Revision: 1.4 $
 **/ 
public class ProbeTable extends TableImpl
{
    /**
     * Constructor
     * @param storeMgr The RDBMSManager for this datastore
     **/
    public ProbeTable(RDBMSManager storeMgr)
    {
        super(storeMgr);

        name = new TableIdentifier(dba, "deleteMe" + System.currentTimeMillis());
    }

    /**
     * Method to initialise the table.
     **/
    public void initialize()
    {
        assertIsUninitialized();

		JavaTypeMapping mapping = dba.getMapping(int.class);
		DatastoreField column = addDatastoreField(int.class, new SQLIdentifier(dba,"unused",null),Role.NONE,mapping,null);
		getStoreManager().getMappingManager().getDatastoreMapping(mapping, storeMgr, column, int.class);

        state = TABLE_STATE_INITIALIZED;
    }

    /**
     * Accessor for the Schema details. This will return a String array with 2
     * elements. The first is the Catalog name, and the second the Schema name.
     * @param conn Connection for this datastore.
     * @return The Schema details
     * @throws SQLException Thrown when an error occurs in the process.
     **/
    public String[] findSchemaDetails(Connection conn)
    throws SQLException
    {
        String[] schemaDetails = new String[2];

        DatabaseMetaData dmd = conn.getMetaData();

        // Make sure the table name is in the correct case.
        // This is required by RDBMS such as PostgreSQL which allow creation in one format yet
        // actually store it in another.
        String table_name=name.getSQLIdentifier();
        if (dba.storesMixedCaseIdentifiers() || dba.storesMixedCaseQuotedIdentifiers())
        {
            // Nothing to do.
        }
        else if (dmd.storesUpperCaseIdentifiers() || dba.storesUpperCaseQuotedIdentifiers())
        {
            table_name = table_name.toUpperCase();
        }
        else if (dmd.storesLowerCaseIdentifiers() || dba.storesLowerCaseQuotedIdentifiers())
        {
            table_name = table_name.toLowerCase();
        }

        // Utilise default catalog/schema if available and applicable
        PMFConfiguration pmfConfig=storeMgr.getPMFContext().getPmfConfiguration();
        String catalog_name=pmfConfig.getDefaultCatalogName();
        String schema_name=pmfConfig.getDefaultSchemaName();
        if (!dba.supportsCatalogsInTableDefinitions())
        {
            catalog_name = null;
        }
        if (!dba.supportsSchemasInTableDefinitions())
        {
            schema_name = null;
        }

        // Find the schema details
        ResultSet rs = dmd.getTables(catalog_name,schema_name,table_name,null);
        try
        {
            if (!rs.next())
            {
                throw new JDODataStoreException(LOCALISER.msg("Table.ProbeTableError",name));
            }

            schemaDetails[0] = rs.getString(1);
            schemaDetails[1] = rs.getString(2);
        }
        finally
        {
            rs.close();
        }

        // Log any failures in this process due to the database adapter
        if (schemaDetails[0] == null)
        {
            JPOXLogger.RDBMS_SCHEMA.info(LOCALISER.msg("Table.CatalogNameUndeterminable"));
        }
        if (schemaDetails[1] == null)
        {
            JPOXLogger.RDBMS_SCHEMA.info(LOCALISER.msg("Table.SchemaNameUndeterminable"));
        }

        return schemaDetails;
    }

    /**
     * Override to always really create ProbeTables in the DB. Needed to
     * determine schema name.
     * @see org.jpox.store.rdbms.table.AbstractTable#allowDdlDump()
     */
    protected boolean allowDdlDump()
    {
        return false;
    }
    
    /**
     * Accessor for Discriminator MetaData
     * @return Returns the Discriminator MetaData.
     */
    public final DiscriminatorMetaData getDiscriminatorMetaData()
    {
        throw new JDOFatalInternalException("Unsupported discriminator in probe table");
    }
    
    /**
     * Accessor for the discriminator mapping specified .
     * @return The mapping for the discriminator datastore field
     **/
    public JavaTypeMapping getDiscriminatorMapping()
    {
        throw new JDOFatalInternalException("Unsupported discriminator in proble table");
    }
    
}