/**********************************************************************
Copyright (c) 2003 Andy Jefferson and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - reinstated SchemaTable, but with different purpose
                     The new SchemaTable supports insert, deleteAll and
                     fetchAll initially.
2004 Andy Jefferson - changed to use Logger
2004 Andy Jefferson - Added type and version columns
2004 Andy Jefferson - Changed to use StoreData. Added owner column
    ...
**********************************************************************/
package org.jpox.store.rdbms.table;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.HashSet;

import javax.jdo.JDOFatalInternalException;

import org.jpox.metadata.DiscriminatorMetaData;
import org.jpox.store.Role;
import org.jpox.store.StoreData;
import org.jpox.store.exceptions.MissingTableException;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.mapping.Mappings;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.store.rdbms.sqlidentifier.SQLIdentifier;
import org.jpox.store.rdbms.sqlidentifier.TableIdentifier;
import org.jpox.util.JPOXLogger;

/**
 * Class defining JPOX schema definition tables. Represents a table in the
 * datastore storing the class and table mappings. This table is used when
 * restarting a JPOX system so that it is 'aware' of what classes were
 * supported the previous time this datastore was used. It uses this
 * information to pre-populate the RDBMSManager with the classes stored in this
 * table. The table names are not used as such, other than as a record of what
 * table a class maps to - because it goes off and finds the MetaData for the
 * class which, with the JPOX naming scheme, defines the table name anyway.
 *
 * @version $Revision: 1.5 $
 **/ 
public class SchemaTable extends TableImpl
{
    private JavaTypeMapping classMapping=null;
    private JavaTypeMapping tableMapping=null;
    private JavaTypeMapping typeMapping=null;
    private JavaTypeMapping ownerMapping=null;
    private JavaTypeMapping versionMapping=null;

    private String insertStmt=null;
    private String deleteStmt=null;
    private String deleteAllStmt=null;
    private String fetchAllStmt=null;
    private String fetchStmt=null;

    /**
     * Constructor
     *
     * @param storeMgr The RDBMSManager for this datastore
     **/
    public SchemaTable(RDBMSManager storeMgr)
    {
        super(storeMgr);

        name = new TableIdentifier(dba,"JPOX_TABLES");
    }

    /**
     * Method to initialise the table.
     **/
    public void initialize()
    {
        assertIsUninitialized();

		classMapping = dba.getMapping(String.class);
        Column class_column=(Column) addDatastoreField(String.class,new SQLIdentifier(dba,"CLASS_NAME",null),Role.NONE,classMapping,null);
		getStoreManager().getMappingManager().getDatastoreMapping(classMapping, storeMgr, class_column, String.class);
        class_column.getColumnOptions().setLength(128);
        class_column.getColumnOptions().setJdbcType("VARCHAR");
        class_column.setUnique();

		tableMapping = dba.getMapping(String.class);
        Column table_column=(Column) addDatastoreField(String.class,new SQLIdentifier(dba,"TABLE_NAME",null),Role.NONE,tableMapping,null);
		getStoreManager().getMappingManager().getDatastoreMapping(tableMapping, storeMgr, table_column, String.class);
        table_column.getColumnOptions().setLength(128);
        table_column.getColumnOptions().setJdbcType("VARCHAR");

		typeMapping = dba.getMapping(String.class);
        Column type_column=(Column) addDatastoreField(String.class,new SQLIdentifier(dba,"TYPE",null),Role.NONE,typeMapping,null);
		getStoreManager().getMappingManager().getDatastoreMapping(typeMapping, storeMgr, type_column, String.class);
        type_column.getColumnOptions().setLength(4);
        type_column.getColumnOptions().setJdbcType("VARCHAR");

        // TODO Change type to SMALLINT/BIT
        ownerMapping = dba.getMapping(String.class);
        Column owner_column=(Column) addDatastoreField(String.class,new SQLIdentifier(dba,"OWNER",null),Role.NONE,ownerMapping,null);
		getStoreManager().getMappingManager().getDatastoreMapping(ownerMapping, storeMgr, owner_column, String.class);
        owner_column.getColumnOptions().setLength(2);
        owner_column.getColumnOptions().setJdbcType("VARCHAR");

		versionMapping = dba.getMapping(String.class);
        Column version_column=(Column) addDatastoreField(String.class,new SQLIdentifier(dba,"VERSION",null),Role.NONE,versionMapping,null);
		getStoreManager().getMappingManager().getDatastoreMapping(versionMapping, storeMgr, version_column, String.class);
        version_column.getColumnOptions().setLength(20);
        version_column.getColumnOptions().setJdbcType("VARCHAR");

        // Set up JDBC statements for supported operations
        insertStmt = "INSERT INTO " + toString() + " (" + class_column.getName() + "," + table_column.getName() + "," + type_column.getName() + "," + 
        	owner_column.getName() + "," + version_column.getName() + ") VALUES (?,?,?,?,?)";
        deleteStmt = "DELETE FROM " + toString() + " WHERE CLASS_NAME=?";
        deleteAllStmt = "DELETE FROM " + toString();
        fetchAllStmt = "SELECT " + class_column.getName() + "," + table_column.getName() + "," + type_column.getName() + "," + 
        	owner_column.getName() + "," + version_column.getName() + " FROM " + toString() + " ORDER BY " + table_column.getName();
        fetchStmt = "SELECT 1 FROM " + toString() + " WHERE CLASS_NAME = ? ";

        state = TABLE_STATE_INITIALIZED;
    }
    
    /**
     * Accessor for the classes already supported by this Schema Table.
     * @param conn Connection for this datastore.
     * @return The HashSet of class names
     * @throws SQLException Thrown when an error occurs in the process.
     **/
    public HashSet getAllClasses(Connection conn)
    throws SQLException
    {
        HashSet schema_data=new HashSet();
        
        if (isDumpingDdl() && !tableExists(conn)) {
            // do not query non-existing schema table when DDL is only written to file
            return schema_data;
        } else {
	        PreparedStatement ps=conn.prepareStatement(fetchAllStmt);
	        try
	        {
	            if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
	            {
	                JPOXLogger.RDBMS_SQL.debug(fetchAllStmt);
	            }
	
	            long start_time=System.currentTimeMillis();
	            ResultSet rs=ps.executeQuery();
	            try
	            {
		            if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
		            {
		                JPOXLogger.RDBMS_SQL.debug(LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - start_time)));
		            }
		
		            while (rs.next())
		            {
		                StoreData data=new StoreData(rs.getString(1),rs.getString(2),
		                    rs.getString(4).equals("1") ? true : false,
		                    rs.getString(3).equals("FCO") ? StoreData.FCO_TYPE : StoreData.SCO_TYPE);
		                schema_data.add(data);
		            }
	            }
	            finally
	            {
	                rs.close();
	            }
	        }
	        finally
	        {
	            ps.close();
	        }
	
	        return schema_data;
        }
    }

    /**
     * Method to insert a row in the SchemaTable. This is called when JPOX is
     * now supporting a new class (and hence DB table).
     * @param data Data for the class
     * @param conn Connection to the datastore
     * @throws SQLException Thrown when an error occurs inserting the schema. 
     **/
    public void addClass(StoreData data, Connection conn)
    throws SQLException
    {
        if (hasClass(data, conn) || isDumpingDdl())
        {
            return;
        }
        PreparedStatement ps = conn.prepareStatement(insertStmt);
        try
        {
            int jdbc_id = 1;

            classMapping.setString(null, ps, Mappings.getParametersIndex(jdbc_id, classMapping), data.getName());
            jdbc_id += 1;

            tableMapping.setString(null, ps, Mappings.getParametersIndex(jdbc_id, tableMapping), data.hasTable() ? data.getTableName() : "");
            jdbc_id += classMapping.getNumberOfDatastoreFields();

            typeMapping.setString(null, ps, Mappings.getParametersIndex(jdbc_id, typeMapping), data.isFCO() ? "FCO" : "SCO");
            jdbc_id += tableMapping.getNumberOfDatastoreFields();

            ownerMapping.setString(null, ps, Mappings.getParametersIndex(jdbc_id, typeMapping), data.isTableOwner() ? "1" : "0");
            jdbc_id += tableMapping.getNumberOfDatastoreFields();
            
            // TODO Sort out version
            versionMapping.setString(null, ps, Mappings.getParametersIndex(jdbc_id, versionMapping), "JPOX");
            jdbc_id += typeMapping.getNumberOfDatastoreFields();

            if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
            {
                JPOXLogger.RDBMS_SQL.debug(insertStmt);
            }

            long start_time = System.currentTimeMillis();
            ps.executeUpdate();
            if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
            {
                JPOXLogger.RDBMS_SQL.debug(LOCALISER.msg("JDBC.Statement.ExecutionTime", (System.currentTimeMillis() - start_time)));
            }

            // TODO : handle any warning messages
        }
        finally
        {
            ps.close();
        }
    }

    /**
     * Method to verify the a class is already stored in the table.
     * @param data Data for the class
     * @param conn Connection to the datastore
     * @return if the SchemaTable already has the class
     * @throws SQLException Thrown when an error occurs inserting the schema. 
     **/
    private boolean hasClass(StoreData data, Connection conn)
    throws SQLException
    {
        if (isDumpingDdl() && !tableExists(conn))
        {
            return false;
        }
        else
        {
            PreparedStatement ps = conn.prepareStatement(fetchStmt);
            try
            {
                int jdbc_id = 1;

                tableMapping.setString(null, ps, Mappings.getParametersIndex(jdbc_id, tableMapping), data.getName());

                if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
                {
                    JPOXLogger.RDBMS_SQL.debug(insertStmt);
                }

                long start_time = System.currentTimeMillis();
                ResultSet rs = ps.executeQuery();
                try
                {
	                if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
	                {
	                    JPOXLogger.RDBMS_SQL.debug(LOCALISER.msg("JDBC.Statement.ExecutionTime", (System.currentTimeMillis() - start_time)));
	                }
	                if (rs.next())
	                {
	                    return true;
	                }
                }
                finally
                {
                    rs.close();
                }
            }
            finally
            {
                ps.close();
            }
            return false;
        }
    }
    
    /**
     * Method to delete a class from the SchemaTable.
     * This is called when JPOX is required to clean out support for a
     * particular class.
     *
     * @param class_name Name of class to delete
     * @param conn Connection to the datastore
     * @throws SQLException Thrown when an error occurs deleting the schema. 
     **/
    public void deleteClass(String class_name,Connection conn)
    throws SQLException
    {
        PreparedStatement ps=conn.prepareStatement(deleteStmt);
        try
        {
            ps.setString(1,class_name);
            if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
            {
                JPOXLogger.RDBMS_SQL.debug(deleteStmt);
            }

            long start_time=System.currentTimeMillis();
            ps.executeUpdate();
            if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
            {
                JPOXLogger.RDBMS_SQL.debug(LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - start_time)));
            }

            // TODO : handle any warning messages
        }
        finally
        {
            ps.close();
        }
    }

    /**
     * Method to delete all classes from the SchemaTable.
     * This is called when JPOX is required to clean out its supported classes
     * (and hence DB table).
     *
     * @param conn Connection to the datastore
     * @throws SQLException Thrown when an error occurs deleting the schema. 
     **/
    public void deleteAllClasses(Connection conn)
    throws SQLException
    {
        PreparedStatement ps=conn.prepareStatement(deleteAllStmt);
        try
        {
            if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
            {
                JPOXLogger.RDBMS_SQL.debug(deleteAllStmt);
            }

            long start_time=System.currentTimeMillis();
            ps.executeUpdate();
            if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
            {
                JPOXLogger.RDBMS_SQL.debug(LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - start_time)));
            }

            // TODO : handle any warning messages
        }
        finally
        {
            ps.close();
        }
    }

    /**
     * Convenience existence checker. This shouldn't be needed since exists already does this!
     * @param conn Connection to use
     * @return Whether it exists
     * @throws SQLException
     */
    private boolean tableExists(Connection conn)
    throws SQLException
    {
        try
        {
            return exists(conn, false);
        }
        catch (MissingTableException mte)
        {
            return false;
        }
    }
    
    /**
     * Accessor for Discriminator MetaData
     * @return Returns the Discriminator MetaData.
     */
    public final DiscriminatorMetaData getDiscriminatorMetaData()
    {
        throw new JDOFatalInternalException("Unsupported discriminator in schema table");
    }
    /**
     * Accessor for the discriminator mapping specified .
     * @return The mapping for the discriminator datastore field
     **/
    public JavaTypeMapping getDiscriminatorMapping()
    {
        throw new JDOFatalInternalException("Unsupported discriminator in schema table");
    }
    
}