/**********************************************************************
Copyright (c) 2004 Erik Bengtson and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
2004 Andy Jefferson - changed to pass Properties rather than MetaData
    ...
**********************************************************************/
package org.jpox.store.rdbms.table;

import java.sql.Connection;
import java.sql.PreparedStatement;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.util.HashSet;
import java.util.Properties;

import javax.jdo.JDOFatalInternalException;

import org.jpox.metadata.DiscriminatorMetaData;
import org.jpox.poid.PoidException;
import org.jpox.store.DatastoreField;
import org.jpox.store.Role;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.store.rdbms.sqlidentifier.SQLIdentifier;
import org.jpox.store.rdbms.sqlidentifier.TableIdentifier;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;

/**
 * Class defining a table sequence for unique identifiers
 *  
 * @version $Revision: 1.6 $
 **/ 
public class SequenceTable extends TableImpl
{
    private static final Localiser LOCALISER_POID=Localiser.getInstance("org.jpox.poid.Localisation");
    private JavaTypeMapping sequenceNameMapping=null;
    private JavaTypeMapping nextValMapping=null;

    private String insertStmt=null;
    private String incrementByStmt=null;
    private String deleteStmt=null;
    private String deleteAllStmt=null;
    private String fetchAllStmt=null;
    private String fetchStmt=null;

    /**
     * the sequence table name, defaults to SEQUENCE_TABLE
     */
    private String tableName = "SEQUENCE_TABLE";

    /**
     * the sequence column name, defaults to SEQUENCE_NAME
     */
    private String sequenceNameColumnName = "SEQUENCE_NAME";

    /**
     * the sequence NextVal column name, defaults to NEXT_VAL
     */
    private String nextValColumnName = "NEXT_VAL";
    
    /**
     * Constructor
     *
     * @param storeMgr The RDBMSManager for this datastore
     **/
    public SequenceTable(RDBMSManager storeMgr,Properties props)
    {
        super(storeMgr);

        if (props.getProperty("sequence-table-name") != null)
        {
            tableName = props.getProperty("sequence-table-name");
        }
        if (props.getProperty("sequence-name-column-name") != null)
        {
            sequenceNameColumnName = props.getProperty("sequence-name-column-name");
        }
        if (props.getProperty("sequence-nextval-column-name") != null)
        {
            nextValColumnName = props.getProperty("sequence-nextval-column-name");
        }
        name = new TableIdentifier(dba,tableName);
    }

    /**
     * Method to initialise the table.
     **/
    public void initialize()
    {
        assertIsUninitialized();

        sequenceNameMapping = dba.getMapping(String.class);
        Column colSequenceName=(Column) addDatastoreField(String.class,new SQLIdentifier(dba,sequenceNameColumnName,null),Role.CUSTOM,sequenceNameMapping,null);
        colSequenceName.getColumnOptions().setLength(Integer.valueOf("128"));
        colSequenceName.getColumnOptions().setJdbcType("VARCHAR");
		getStoreManager().getMappingManager().getDatastoreMapping(sequenceNameMapping, storeMgr, colSequenceName, String.class);

        nextValMapping = dba.getMapping(Long.class);
        DatastoreField colNextVal=addDatastoreField(Long.class,new SQLIdentifier(dba,nextValColumnName,null),Role.CUSTOM,nextValMapping,null);
		getStoreManager().getMappingManager().getDatastoreMapping(nextValMapping, storeMgr, colNextVal, Long.class);

        // Set up JDBC statements for supported operations
        insertStmt = "INSERT INTO " + toString() + " (" + colSequenceName.getName() + "," + colNextVal.getName() + ") VALUES (?,?)";
        incrementByStmt = "UPDATE " + toString() + " SET "+colNextVal.getName() + "=("+colNextVal.getName()+"+?) WHERE "+colSequenceName.getName()+"=?";
        deleteStmt = "DELETE FROM " + toString();
        deleteAllStmt = "DELETE FROM " + toString();
        fetchStmt = "SELECT " + colNextVal.getName() + " FROM " + toString() + " WHERE "+colSequenceName.getName()+"=?";
        fetchAllStmt = "SELECT " + colNextVal.getName() + "," + colSequenceName.getName() + " FROM " + toString() + " ORDER BY " + colSequenceName.getName();

        state = TABLE_STATE_INITIALIZED;
    }

    /**
     * Accessor for the sequences
     * @param conn Connection for this datastore.
     * @return The HashSet of Sequence names
     * @throws SQLException Thrown when an error occurs in the process.
     **/
    public HashSet getFetchAllSequences(Connection conn)
    throws SQLException
    {
        HashSet sequenceNames=new HashSet();

        PreparedStatement ps=conn.prepareStatement(fetchAllStmt);
        try
        {
            if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
            {
                JPOXLogger.RDBMS_SQL.debug(fetchAllStmt);
            }

            long start_time=System.currentTimeMillis();
            ResultSet rs=ps.executeQuery();
            try
            {
	            if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
	            {
	                JPOXLogger.RDBMS_SQL.debug(LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - start_time)));
	            }
	
	            while (rs.next())
	            {
	                sequenceNames.add(rs.getString(2));
	            }
            }
            finally
            {
                rs.close();
            }
        }
        finally
        {
            ps.close();
        }

        return sequenceNames;
    }

    /**
     * Accessor for the nextval of a sequence
     * @param conn Connection for this datastore.
     * @return The HashSet of Sequence names
     * @throws SQLException Thrown when an error occurs in the process.
     **/
    public Long getNextVal(String sequenceName, Connection conn, int incrementBy)
    throws SQLException
    {
        PreparedStatement ps=conn.prepareStatement(fetchStmt);
        Long nextVal = null;
        try
        {
            if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
            {
                JPOXLogger.RDBMS_SQL.debug(fetchStmt);
            }

            sequenceNameMapping.setString(null, ps, new int[] {1}, sequenceName);

            long start_time=System.currentTimeMillis();
            ResultSet rs=ps.executeQuery();
            try
            {
	            if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
	            {
	                JPOXLogger.RDBMS_SQL.debug(LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - start_time)));
	            }
	
	            if (! rs.next())
	            {
	                addSequence(sequenceName,new Long(incrementBy+1),conn);
	                nextVal = new Long(1);
	            }
	            else
	            {
	                nextVal = new Long(rs.getLong(1));
	                incrementSequence(sequenceName,incrementBy,conn);
	            }
            }
            finally
            {
                rs.close();
            }
        }
        catch (SQLException e)
        {
            throw new PoidException(LOCALISER_POID.msg("SequencePoidGenerator.CantObtainSequenceId", e.getMessage()));
        }
        finally
        {
            ps.close();
        }

        return nextVal;
    }

    /**
     * Method to increment a sequence
     * @param conn Connection to the datastore
     * @throws SQLException Thrown when an error occurs incrementing the sequence. 
     **/
    private void incrementSequence(String sequenceName, long incrementBy, Connection conn)
    throws SQLException
    {
        PreparedStatement ps=conn.prepareStatement(incrementByStmt);
        try
        {
            nextValMapping.setLong(null, ps, new int[] {1}, incrementBy);

            sequenceNameMapping.setString(null, ps, new int[] {2}, sequenceName);

            if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
            {
                JPOXLogger.RDBMS_SQL.debug(insertStmt);
            }

            long start_time=System.currentTimeMillis();
            ps.executeUpdate();
            if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
            {
                JPOXLogger.RDBMS_SQL.debug(LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - start_time)));
            }

            // TODO : handle any warning messages
        }
        finally
        {
            ps.close();
        }
    }

    /**
     * Method to insert a row in the SequenceTable
     * @param conn Connection to the datastore
     * @throws SQLException Thrown when an error occurs inserting the sequence. 
     **/
    private void addSequence(String sequenceName, Long nextVal, Connection conn)
    throws SQLException
    {
        PreparedStatement ps=conn.prepareStatement(insertStmt);
        try
        {

            sequenceNameMapping.setString(null, ps, new int[] {1}, sequenceName);
            nextValMapping.setLong(null, ps, new int[] {2}, nextVal.longValue());

            if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
            {
                JPOXLogger.RDBMS_SQL.debug(insertStmt);
            }

            long start_time=System.currentTimeMillis();
            ps.executeUpdate();
            if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
            {
                JPOXLogger.RDBMS_SQL.debug(LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - start_time)));
            }

            // TODO : handle any warning messages
        }
        finally
        {
            ps.close();
        }
    }

    /**
     * Method to delete a sequence
     *
     * @param sequenceName
     * @param conn Connection to the datastore
     * @throws SQLException Thrown when an error occurs deleting the schema. 
     **/
    public void deleteSequence(String sequenceName, Connection conn)
    throws SQLException
    {
        PreparedStatement ps=conn.prepareStatement(deleteStmt);
        try
        {
            ps.setString(1,sequenceName);
            if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
            {
                JPOXLogger.RDBMS_SQL.debug(deleteStmt);
            }

            long start_time=System.currentTimeMillis();
            ps.executeUpdate();
            if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
            {
                JPOXLogger.RDBMS_SQL.debug(LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - start_time)));
            }

            // TODO : handle any warning messages
        }
        finally
        {
            ps.close();
        }
    }

    /**
     * Method to delete all sequences
     *
     * @param conn Connection to the datastore
     * @throws SQLException Thrown when an error occurs deleting. 
     **/
    public void deleteAllSequences(Connection conn)
    throws SQLException
    {
        PreparedStatement ps=conn.prepareStatement(deleteAllStmt);
        try
        {
            if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
            {
                JPOXLogger.RDBMS_SQL.debug(deleteAllStmt);
            }

            long start_time=System.currentTimeMillis();
            ps.executeUpdate();
            if (JPOXLogger.RDBMS_SQL.isDebugEnabled())
            {
                JPOXLogger.RDBMS_SQL.debug(LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - start_time)));
            }

            // TODO : handle any warning messages
        }
        finally
        {
            ps.close();
        }
    }
    
    /**
     * Accessor for Discriminator MetaData
     * @return Returns the Discriminator MetaData.
     */
    public final DiscriminatorMetaData getDiscriminatorMetaData()
    {
        throw new JDOFatalInternalException("Unsupported discriminator in sequence table");
    }    
    /**
     * Accessor for the discriminator mapping specified .
     * @return The mapping for the discriminator datastore field
     **/
    public JavaTypeMapping getDiscriminatorMapping()
    {
        throw new JDOFatalInternalException("Unsupported discriminator in sequence table");
    }
    
}
