/**********************************************************************
Copyright (c) 2003 Mike Martin and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - Added getCatalogName()
2004 Andy Jefferson - Moved state constants here
2004 Andy Jefferson - Added documentation
    ...
**********************************************************************/
package org.jpox.store.rdbms.table;

import java.sql.Connection;
import java.sql.SQLException;

import org.jpox.store.DatastoreContainerObject;

/**
 * Representation of a table in a datastore.
 * <P>
 * There are 2 aspects to a table. The first is the internal representation, provided here.
 * This has a state. The second aspect to the table is its external (datastore) representation.
 * This reflects whether it exists, or whether it has been deleted, etc.
 * </P>
 * <P>
 * This interface provides some methods for mapping from the internal representation to
 * the external representation. These are the methods
 * <UL>
 * <LI>create() - to create the table in the datastore</LI>
 * <LI>drop() - to drop the table from the datastore</LI>
 * <LI>validate() - to compare the internal and external representations.</LI>
 * <LI>exists() - whether the external representation exists</LI>
 * </UL>
 *
 * @version $Revision: 1.3 $
 **/
public interface Table extends DatastoreContainerObject
{
    /** Returned by RDBMSStoreHelper.getTableType() if the table does not exist. */
    static final int TABLE_TYPE_MISSING = -1;

    /** Returned by RDBMSStoreHelper.getTableType() if the table type is not recognized. */
    static final int TABLE_TYPE_UNKNOWN = 0;

    /** Returned by RDBMSStoreHelper.getTableType() if the table is a table. */
    static final int TABLE_TYPE_TABLE = 1;

    /** Returned by RDBMSStoreHelper.getTableType() if the table is a view. */
    static final int TABLE_TYPE_VIEW = 2;

    /** Table object has just been created. */
    static final int TABLE_STATE_NEW = 0;

    /** Table object is created and PK initialised. */
    static final int TABLE_STATE_PK_INITIALIZED = 1;

    /** Table object has been initialised. */
    static final int TABLE_STATE_INITIALIZED = 2;

    /** Table object has been initialized but has had structural modifications since. */
    static final int TABLE_STATE_INITIALIZED_MODIFIED = 3;

    /** Table object has been validated. */
    static final int TABLE_STATE_VALIDATED = 4;

    /**
     * Accessor for the Catalog name for this table.
     * @return The Catalog name.
     **/
    String getCatalogName();

    /**
     * Accessor for the Schema name for this table.
     * @return The Schema name.
     **/
    String getSchemaName();

    /**
     * Method to initialise the table.
     */
    void initialize();

    /**
     * Accessor for whether the table has been initialised.
     * @return Whether it is initialised.
     */
    boolean isInitialized();

    /**
     * Accessor for whether the table has been modified after being initialised.
     * @return Whether it has been modified after being initialised.
     */
    boolean isInitializedModified();

    // ---------------------------- Comparison with External ------------------------------------

    /**
     * Method to validate the table against what is in the datastore
     * @param conn The connection
     * @param update Whether to update the table to fix any errors.
     * @return Whether it  validates successfully
     * @throws SQLException Thrown if an error occurrs in the validation
     */
    boolean validate(Connection conn, boolean update) throws SQLException;

    /**
     * Accessor for whether the table is validated.
     * @return Whether it is validated.
     */
    boolean isValidated();

    /**
     * Accessor for whether the table exists in the datastore. 
     * Will throw a MissingTableException if the table doesn't exist.
     * @param conn The connecton to use to verify it
     * @param create Whether to create it if it doesn't exist
     * @return Whether the table was added.
     * @throws SQLException Thrown if an error occurs in the check
     */
    boolean exists(Connection conn, boolean create) throws SQLException;

    /**
     * Method to create the table in the datastore representation.
     * @param conn The connection to use
     * @throws SQLException Thrown if an error occurs
     */
    void create(Connection conn) throws SQLException;
    
    /**
     * Method to drop the table from the datastore representation.
     * @param conn The connection to use
     * @throws SQLException Thrown if an error occurs
     */
    void drop(Connection conn) throws SQLException;
}