/**********************************************************************
Copyright (c) 2002 Mike Martin and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution.

Contributors:
2003 Erik Bengtson  - removed unused variable
2003 Andy Jefferson - localised
2003 Andy Jefferson - addition of checkExists, made getExpectedPK protected
2004 Andy Jefferson - rewritten to remove many levels of inheritance
2004 Andy Jefferson - added validation of columns, and auto create option
2004 Andy Jefferson - separated out validatePK
    ...
**********************************************************************/
package org.jpox.store.rdbms.table;

import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;
import java.util.Map;
import java.util.Set;

import org.jpox.store.DatastoreField;
import org.jpox.store.exceptions.MissingColumnException;
import org.jpox.store.exceptions.MissingTableException;
import org.jpox.store.exceptions.NoTableManagedException;
import org.jpox.store.exceptions.NotATableException;
import org.jpox.store.exceptions.UnexpectedColumnException;
import org.jpox.store.exceptions.WrongPrimaryKeyException;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.store.rdbms.RDBMSStoreHelper;
import org.jpox.store.rdbms.columninfo.ColumnInfo;
import org.jpox.store.rdbms.foreignkey.ForeignKeyInfo;
import org.jpox.store.rdbms.key.CandidateKey;
import org.jpox.store.rdbms.key.ForeignKey;
import org.jpox.store.rdbms.key.Index;
import org.jpox.store.rdbms.key.PrimaryKey;
import org.jpox.store.rdbms.sqlidentifier.CandidateKeyIdentifier;
import org.jpox.store.rdbms.sqlidentifier.ForeignKeyIdentifier;
import org.jpox.store.rdbms.sqlidentifier.IndexIdentifier;
import org.jpox.store.rdbms.sqlidentifier.PrimaryKeyIdentifier;
import org.jpox.store.rdbms.sqlidentifier.SQLIdentifier;
import org.jpox.util.JPOXLogger;
import org.jpox.util.SQLWarnings;
import org.jpox.util.StringUtils;

/**
 * Class representing a table in a datastore (RDBMS).
 * Provides a series of methods for validating the aspects of the table, namely
 * <UL>
 * <LI>validate - Validate the table</LI>
 * <LI>validateColumns - Validate the columns in the table</LI>
 * <LI>validatePrimaryKey - Validate the primary key</LI>
 * <LI>validateIndices - Validate the indices on the table</LI>
 * <LI>validateForeignKeys - Validate all FKs for the table</LI>
 * <LI>validateConstraints - Validate the indices and FKs.</LI>
 * </UL>
 *
 * @version $Revision: 1.10 $
 */
public abstract class TableImpl extends AbstractTable
{
    /**
     * Constructor.
     * @param storeMgr The StoreManager for this table.
     */
    public TableImpl(RDBMSManager storeMgr)
    {
        super(storeMgr);
    }

    /**
     * Constructor.
     * @param name The name of the table (in SQL).
     * @param storeMgr The StoreManager for this table.
     */
    public TableImpl(SQLIdentifier name, RDBMSManager storeMgr)
    {
        super(name, storeMgr);
    }

    /**
     * Initialize the default value for columns, if is null
     * @param conn The JDBC Connection
     * @throws SQLException
     */
    public void initializeDefault(Connection conn)
    throws SQLException
    {
        HashMap columns = new HashMap(columnsByName);
        Iterator i = storeMgr.getColumnInfo(this, conn).iterator();
        while (i.hasNext())
        {
            ColumnInfo ci = (ColumnInfo) i.next();
            SQLIdentifier colName = new SQLIdentifier(dba, ci.getColumnName());
            Column col = (Column) columns.get(colName);
            if (col != null)
            {
                col.initializeDefault(ci);
            }
        }
    }

    /**
     * Accessor for the primary key for this table. 
     * Will always return a PrimaryKey but if we have defined no columns,
     * the pk.size() will be 0.
     * @return The primary key.
     */
    public PrimaryKey getPrimaryKey()
    {
        PrimaryKey pk = new PrimaryKey(this);
        Iterator i = columns.iterator();
        while (i.hasNext())
        {
            Column col = (Column) i.next();
            if (col.isPrimaryKey())
            {
                pk.addDatastoreField(col);
            }
        }

        return pk;
    }

    /**
     * Method to validate the table in the datastore.
     * @param conn The JDBC Connection
     * @param update Whether to update the table to fix any validation errors. Only applies to missing columns.
     * @return Whether the database was modified
     * @throws SQLException Thrown when an error occurs in the JDBC calls
     */
    public boolean validate(Connection conn, boolean update)
    throws SQLException
    {
        assertIsInitialized();

        // Check existence
        int tableType = RDBMSStoreHelper.getTableType(storeMgr, name.getSQLIdentifier(), conn);
        if (tableType == TABLE_TYPE_MISSING)
        {
            throw new MissingTableException(this.toString());
        }

        long startTime = System.currentTimeMillis();
        if (JPOXLogger.RDBMS.isDebugEnabled())
        {
            JPOXLogger.RDBMS.debug(LOCALISER.msg("Table.Validating", this));
        }

        // Check type
        if (tableType != TABLE_TYPE_TABLE)
        {
            throw new NotATableException(this.toString());
        }

        // Validate the column(s)
        validateColumns(conn, update);

        // Validate the primary key(s)
        validatePrimaryKey(conn);

        state = TABLE_STATE_VALIDATED;
        if (JPOXLogger.RDBMS.isDebugEnabled())
        {
            JPOXLogger.RDBMS.debug(LOCALISER.msg("JDBC.Statement.ExecutionTime",
                                   (System.currentTimeMillis() - startTime)));
        }

        return false;
    }

    /**
     * Utility to validate the columns of the table.
     * Will throw a MissingColumnException if a column is not found (and
     * is not required to auto create it)
     * @param conn Connection to use for validation
     * @param auto_create Whether to auto create any missing columns
     * @return Whether it validates
     * @throws SQLException Thrown if an error occurs in the validation process
     */
    public boolean validateColumns(Connection conn, boolean auto_create)
    throws SQLException
    {
        HashMap unvalidated = new HashMap(columnsByName);
        Iterator i = storeMgr.getColumnInfo(this, conn).iterator();
        while (i.hasNext())
        {
            ColumnInfo ci = (ColumnInfo) i.next();
            SQLIdentifier colName = new SQLIdentifier(dba, ci.getColumnName());
            Column col = (Column) unvalidated.get(colName);
            if (col != null)
            {
                col.validate(ci);
                unvalidated.remove(colName);
            }
        }

        if (unvalidated.size() > 0)
        {
            if (auto_create)
            {
                // Add all missing columns
                List stmts = new ArrayList();
                Set columnKeys = unvalidated.keySet();
                Iterator columnKeysIter = columnKeys.iterator();
                while (columnKeysIter.hasNext())
                {
                    SQLIdentifier colName = (SQLIdentifier)columnKeysIter.next();

                    Column col = (Column)(unvalidated.get(colName));
                    String addColStmt = dba.getAddColumnStatement(this, col);
                    stmts.add(addColStmt);
                    JPOXLogger.RDBMS_DDL.info(LOCALISER.msg("Table.CreatingColumn", col.getName(), this.toString()));
                }
                executeDdlStatementList(stmts, conn);
            }
            else
            {
                throw new MissingColumnException(this, unvalidated.values());
            }
        }
        state = TABLE_STATE_VALIDATED;

        return true;
    }

    /**
     * Utility method to validate the primary key of the table.
     * Will throw a WrongPrimaryKeyException if the PK is incorrect.
     * TODO Add an auto_create parameter on this
     * @param conn Connection to use
     * @return Whether it validates
     * @throws SQLException When an error occurs in the valdiation
     */
    protected boolean validatePrimaryKey(Connection conn)
    throws SQLException
    {
        Map actualPKs = getExistingPrimaryKeys(conn);
        PrimaryKey expectedPK = getPrimaryKey();
        if (expectedPK.size() == 0)
        {
            if (!actualPKs.isEmpty())
            {
                throw new WrongPrimaryKeyException(this.toString(), expectedPK.toString(), StringUtils.collectionToString(actualPKs.values()));
            }
        }
        else
        {
            if (actualPKs.size() != 1 ||
                !actualPKs.values().contains(expectedPK))
            {
                throw new WrongPrimaryKeyException(this.toString(), expectedPK.toString(), StringUtils.collectionToString(actualPKs.values()));
            }
        }

        return true;
    }

    /**
     * Method to validate any constraints, and auto create them if required.
     * @param conn The JDBC Connection
     * @param auto_create Whether to auto create the constraints if not existing
     * @return Whether the database was modified
     * @throws SQLException Thrown when an error occurs in the JDBC calls
     */
    public boolean validateConstraints(Connection conn, boolean auto_create)
    throws SQLException
    {
        assertIsInitialized();

        boolean cksWereModified;
        boolean fksWereModified;
        boolean idxsWereModified;
        if (dba.createIndexesBeforeForeignKeys())
        {
            idxsWereModified = validateIndices(conn, auto_create);
            fksWereModified = validateForeignKeys(conn, auto_create);
            cksWereModified = validateCandidateKeys(conn, auto_create);
        }
        else
        {
            cksWereModified = validateCandidateKeys(conn, auto_create);
            fksWereModified = validateForeignKeys(conn, auto_create);
            idxsWereModified = validateIndices(conn, auto_create);
        }

        return fksWereModified || idxsWereModified || cksWereModified;
    }

    /**
     * Method to validate any foreign keys on this table in the datastore, and
     * auto create any that are missing where required.
     * 
     * @param conn The JDBC Connection
     * @param auto_create Whether to auto create the FKs if not existing
     * @return Whether the database was modified
     * @throws SQLException Thrown when an error occurs in the JDBC calls
     */
    private boolean validateForeignKeys(Connection conn, boolean auto_create)
    throws SQLException
    {
        boolean dbWasModified = false;

        // Validate and/or create all foreign keys.
        Map actualForeignKeysByName = getExistingForeignKeys(conn);
        int numActualFKs = actualForeignKeysByName.size();
        Map stmtsByFKName = getSQLAddFKStatements(actualForeignKeysByName);
        if (stmtsByFKName.isEmpty())
        {
            if (numActualFKs > 0)
            {
                if (JPOXLogger.RDBMS.isInfoEnabled())
                {
                    JPOXLogger.RDBMS.info(LOCALISER.msg("Table.ForeignKey.Validated", 
                        "" + numActualFKs,this));
                }
            }
        }
        else
        {
            // Auto Create any missing foreign keys
            if (auto_create)
            {
                Statement stmt = conn.createStatement();
                try
                {
                    Iterator i = stmtsByFKName.entrySet().iterator();
                    while (i.hasNext())
                    {
                        Map.Entry e = (Map.Entry) i.next();
                        String fkName = (String) e.getKey();
                        String stmtText = (String) e.getValue();
                        if (JPOXLogger.RDBMS.isInfoEnabled())
                        {
                            JPOXLogger.RDBMS.info(LOCALISER.msg("Table.ForeignKey.Creating",
                                getSchemaName(), fkName));
                        }

						executeDdlStatement(stmt, stmtText);                    
					}
                }
                finally
                {
                    stmt.close();
                }
                dbWasModified = true;
            }
            else
            {
                // JPOX supports existing schemas so don't raise an exception.
                JPOXLogger.RDBMS.warn(LOCALISER.msg("Table.ForeignKey.MissingForeignKeys",
                        this, stmtsByFKName.values()));
            }
        }
        return dbWasModified;
    }

    /**
     * Method to validate any indices for this table, and auto create any
     * missing ones where required.
     * @param conn The JDBC Connection
     * @param auto_create Whether to auto create any missing indices
     * @return Whether the database was changed
     * @throws SQLException Thrown when an error occurs in the JDBC calls
     */
    private boolean validateIndices(Connection conn, boolean auto_create)
    throws SQLException
    {
        boolean dbWasModified = false;

        // Compute the number of existing indices *created by us*, which are
        // recognized by having a name that starts with the table name. This
        // number is only used to control what gets logged below. There are
        // often other system-generated indices.
        Map actualIndicesByName = getExistingIndices(conn);
        int numActualIdxs = 0;

        Iterator names = actualIndicesByName.keySet().iterator();
        while (names.hasNext())
        {
            SQLIdentifier idxName = (SQLIdentifier) names.next();
            if (idxName.toString().startsWith(name.toString()))
            {
                ++numActualIdxs;
            }
        }

        Map stmtsByIdxName = getSQLCreateIndexStatements(actualIndicesByName);
        if (stmtsByIdxName.isEmpty())
        {
            if (numActualIdxs > 0)
            {
                if (JPOXLogger.RDBMS.isInfoEnabled())
                {
                    JPOXLogger.RDBMS.info(LOCALISER.msg("Table.Index.Validated", "" + numActualIdxs, this));
                }
            }
        }
        else
        {
            // Auto Create any missing indices
            if (auto_create)
            {
                Statement stmt = conn.createStatement();
                try
                {
                    Iterator i = stmtsByIdxName.entrySet().iterator();
                    while (i.hasNext())
                    {
                        Map.Entry e = (Map.Entry) i.next();
                        String idxName = (String) e.getKey();
                        String stmtText = (String) e.getValue();
                        if (JPOXLogger.RDBMS.isInfoEnabled())
                        {
                            JPOXLogger.RDBMS.info(LOCALISER.msg("Table.Index.Creating", 
                                getSchemaName(), idxName));
                        }
						executeDdlStatement(stmt, stmtText);                    
					}
                }
                finally
                {
                    stmt.close();
                }
                dbWasModified = true;
            }
            else
            {
                // JPOX supports existing schemas so don't raise an exception.
                JPOXLogger.RDBMS.warn(LOCALISER.msg("Table.Index.MissingIndexes",
                        this, stmtsByIdxName.values()));
            }
        }
        return dbWasModified;
    }

    /**
     * Method to validate any Candidate keys on this table in the datastore, and
     * auto create any that are missing where required.
     * 
     * @param conn The JDBC Connection
     * @param auto_create Whether to auto create the Candidate Keys if not existing
     * @return Whether the database was modified
     * @throws SQLException Thrown when an error occurs in the JDBC calls
     */
    private boolean validateCandidateKeys(Connection conn, boolean auto_create)
    throws SQLException
    {
        boolean dbWasModified = false;

        // Validate and/or create all candidate keys.
        Map actualCandidateKeysByName = getExistingCandidateKeys(conn);
        int numActualCKs = actualCandidateKeysByName.size();
        Map stmtsByCKName = getSQLAddCandidateKeyStatements(actualCandidateKeysByName);
        if (stmtsByCKName.isEmpty())
        {
            if (numActualCKs > 0)
            {
                if (JPOXLogger.RDBMS.isInfoEnabled())
                {
                    JPOXLogger.RDBMS.info(LOCALISER.msg("Table.CandidateKey.Validated", 
                        "" + numActualCKs,this));
                }
            }
        }
        else
        {
            // Auto Create any missing candidate keys
            if (auto_create)
            {
                Statement stmt = conn.createStatement();
                try
                {
                    Iterator i = stmtsByCKName.entrySet().iterator();
                    while (i.hasNext())
                    {
                        Map.Entry e = (Map.Entry) i.next();
                        String ckName = (String) e.getKey();
                        String stmtText = (String) e.getValue();
                        if (JPOXLogger.RDBMS.isInfoEnabled())
                        {
                            JPOXLogger.RDBMS.info(LOCALISER.msg("Table.CandidateKey.Creating",
                                getSchemaName(), ckName));
                        }

                        executeDdlStatement(stmt, stmtText);                    
                    }
                }
                finally
                {
                    stmt.close();
                }
                dbWasModified = true;
            }
            else
            {
                // JPOX supports existing schemas so don't raise an exception.
                JPOXLogger.RDBMS.warn(LOCALISER.msg("Table.CandidateKey.MissingForeignKeys",
                        this, stmtsByCKName.values()));
            }
        }

        return dbWasModified;
    }

    /**
     * Method to drop the constraints for the table from the datastore.
     * @param conn The JDBC Connection
     * @throws SQLException Thrown when an error occurs in the JDBC call.
     */
    public void dropConstraints(Connection conn)
    throws SQLException
    {
        assertIsInitialized();

        boolean drop_using_constraint = dba.supportsAlterTableDropConstraint();
        boolean drop_using_foreign_key = dba.supportsAlterTableDropForeignKeyConstraint();
        if (!drop_using_constraint && !drop_using_foreign_key)
        {
            return;
        }

        /*
         * There's no need to drop indices; we assume they'll go away quietly
         * when the table is dropped.
         */
        HashSet fkNames = new HashSet();
        Iterator i = RDBMSStoreHelper.getForeignKeyInfo(storeMgr,
                name.getSQLIdentifier(), conn).iterator();
        while (i.hasNext())
        {
            ForeignKeyInfo fki = (ForeignKeyInfo) i.next();
            /*
             * Legally, JDBC drivers are allowed to return null names for
             * foreign keys. If they do, we simply have to skip DROP CONSTRAINT.
             */
            if (fki.fkName != null)
            {
                fkNames.add(fki.fkName);
            }
        }
        int numFKs = fkNames.size();
        if (numFKs > 0)
        {
            if (JPOXLogger.RDBMS.isInfoEnabled())
            {
                JPOXLogger.RDBMS.info(LOCALISER.msg("Table.ForeignKey.Dropping",
                    "" + numFKs, this));
            }
            i = fkNames.iterator();
            Statement stmt = conn.createStatement();
            try
            {
                while (i.hasNext())
                {
                    String constraintName = (String) i.next();
                    String stmtText = null;
                    if (drop_using_constraint)
                    {
                        stmtText = "ALTER TABLE " + name + " DROP CONSTRAINT " + constraintName;
                    }
                    else
                    {
                        stmtText = "ALTER TABLE " + name + " DROP FOREIGN KEY " + constraintName;
                    }
                    if (JPOXLogger.RDBMS_DDL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_DDL.debug(stmtText);
                    }
                    long startTime = System.currentTimeMillis();
                    stmt.execute(stmtText);
                    if (JPOXLogger.RDBMS_DDL.isDebugEnabled())
                    {
                        JPOXLogger.RDBMS_DDL.debug(LOCALISER.msg(
                                "JDBC.Statement.ExecutionTime",
                                (System.currentTimeMillis() - startTime)));
                    }
                    SQLWarnings.log(stmt);
                }
            }
            finally
            {
                stmt.close();
            }
        }
    }

    // ----------------------- Internal Implementation Methods -----------------

    /**
     * Accessor for the expected foreign keys for this table in the datastore.
     * Currently only checks the columns for referenced tables (i.e
     * relationships) and returns those.
     * @return List of foreign keys.
     */
    protected List getExpectedForeignKeys()
    {
        assertIsInitialized();

        /*
         * The following HashSet is to avoid the duplicate usage of columns that
         * have already been used in conjunction with another column
         */
        Set colsInFKs = new HashSet();
        ArrayList foreignKeys = new ArrayList();
        Iterator i = columns.iterator();
        while (i.hasNext())
        {
            Column col = (Column) i.next();
            if (!colsInFKs.contains(col))
            {
                try
                {
                    ClassTable referencedTable = (ClassTable) storeMgr.getTable(col.getType().getName());
                    if (referencedTable != null)
                    {
                        for (int j = 0; j < col.getMapping().getNumberOfDatastoreFields(); j++)
                        {
                            colsInFKs.add(col.getMapping().getDataStoreMapping(j).getDatastoreField());
                        }
                        if (referencedTable != null)
                        {
                            ForeignKey fk = new ForeignKey(col.getMapping(), dba, referencedTable, true);
                            foreignKeys.add(fk);
                        }
                    }
                }
                catch (NoTableManagedException e)
                {
                    //expected when no table exists
                }
            }
        }
        return foreignKeys;
    }


    /**
     * Accessor for the expected candidate keys for this table in the datastore.
     * Currently returns an empty list.
     * @return List of candidate keys.
     */
    protected List getExpectedCandidateKeys()
    {
        assertIsInitialized();

        ArrayList candidateKeys = new ArrayList();
        return candidateKeys;
    }

    /**
     * Accessor for the indices for this table in the datastore.
     * @return Set of indices expected.
     */
    protected Set getExpectedIndices()
    {
        assertIsInitialized();

        HashSet indices = new HashSet();

        /*
         * For each foreign key, add to the list an index made up of the "from"
         * column(s) of the key, *unless* those columns also happen to be the
         * leading columns of the primary key. If they are, we're assuming that
         * they're going to be indexed anyway, regardless of whether that fact
         * gets reflected by getIndexInfo().
         */
        PrimaryKey pk = getPrimaryKey();
        Iterator i = getExpectedForeignKeys().iterator();
        while (i.hasNext())
        {
            ForeignKey fk = (ForeignKey) i.next();
            if (!pk.startsWith(fk))
            {
                indices.add(new Index(fk));
            }
        }

        return indices;
    }

    /**
     * Accessor for the primary keys for this table in the datastore.
     * @param conn The JDBC Connection
     * @return Map of primary keys
     * @throws SQLException Thrown when an error occurs in the JDBC call.
     */
    private Map getExistingPrimaryKeys(Connection conn)
    throws SQLException
    {
        DatabaseMetaData dmd = conn.getMetaData();
        String[] c = RDBMSStoreHelper.splitTableIdentifierName(storeMgr.getDatabaseAdapter().getCatalogSeparator(), name.getSQLIdentifier());

        HashMap primaryKeysByName = new HashMap();

        // Note : the table name has no quotes here.
        ResultSet rs = dmd.getPrimaryKeys(c[0], c[1] == null ? getSchemaName() : c[1], c[2]);

        try
        {
            while (rs.next())
            {
                SQLIdentifier pkName;
                String s = rs.getString(6);
                if (s == null)
                {
                    pkName = new PrimaryKeyIdentifier(this);
                }
                else
                {
                    pkName = new SQLIdentifier(dba, s);
                }
                PrimaryKey pk = (PrimaryKey) primaryKeysByName.get(pkName);
                if (pk == null)
                {
                    pk = new PrimaryKey(this);
                    primaryKeysByName.put(pkName, pk);
                }

                int keySeq = rs.getInt(5) - 1;
                SQLIdentifier colName = new SQLIdentifier(dba, rs.getString(4));
                Column col = (Column) columnsByName.get(colName);

                /*
                 * should not happens
                 */
                /*
                 * if (col == null) { col =
                 * newColumn(Long.class,colName).setAsPrimaryKey();
                 * getPrimaryKey().addColumn(col);
                 * 
                 * if (JPOXLogger.RDBMS.isInfoEnabled())
                 * JPOXLogger.RDBMS.info(LOCALISER.msg("Table.NewPKColumnAdded",colName)); }
                 */
                if (col == null)
                {
                    throw new UnexpectedColumnException(this.toString(), colName.getSQLIdentifier());
                }
                pk.setDatastoreField(keySeq, col);
            }
        }
        finally
        {
            rs.close();
        }

        return primaryKeysByName;
    }

    /**
     * Accessor for the foreign keys for this table.
     * @param conn The JDBC Connection
     * @return Map of foreign keys
     * @throws SQLException Thrown when an error occurs in the JDBC call.
     */
    private Map getExistingForeignKeys(Connection conn)
    throws SQLException
    {
        HashMap foreignKeysByName = new HashMap();

        Iterator i = RDBMSStoreHelper.getForeignKeyInfo(storeMgr,
                name.getSQLIdentifier(), conn).iterator();
        while (i.hasNext())
        {
            SQLIdentifier fkName;
            ForeignKeyInfo fki = (ForeignKeyInfo) i.next();
            if (fki.fkName == null)
            {
                fkName = new ForeignKeyIdentifier(this, foreignKeysByName.size());
            }
            else
            {
                fkName = new SQLIdentifier(dba, fki.fkName);
            }

            boolean initiallyDeferred = fki.deferrability == DatabaseMetaData.importedKeyInitiallyDeferred;
            ForeignKey fk = (ForeignKey) foreignKeysByName.get(fkName);
            if (fk == null)
            {
                fk = new ForeignKey(initiallyDeferred);
                fk.setName(fkName.getSQLIdentifier());
                foreignKeysByName.put(fkName, fk);
            }

            TableImpl refTable = (TableImpl) storeMgr.getTable(new SQLIdentifier(dba, fki.pkTableName));
            if (refTable != null)
            {
                SQLIdentifier colName = new SQLIdentifier(dba, fki.fkColumnName);
                SQLIdentifier refColName = new SQLIdentifier(dba,fki.pkColumnName);
                DatastoreField col = (DatastoreField) columnsByName.get(colName);
                DatastoreField refCol = (DatastoreField) refTable.columnsByName.get(refColName);
                if (col != null && refCol != null)
                {
                    fk.addDatastoreField(col, refCol);
                }
                else
                {
                    //TODO throw exception?
                }
            }
        }

        return foreignKeysByName;
    }

    /**
     * Accessor for the candidate keys for this table.
     * @param conn The JDBC Connection
     * @return Map of candidate keys
     * @throws SQLException Thrown when an error occurs in the JDBC call.
     */
    private Map getExistingCandidateKeys(Connection conn)
    throws SQLException
    {
        DatabaseMetaData dmd = conn.getMetaData();
        HashMap candidateKeysByName = new HashMap();
        String[] c = RDBMSStoreHelper.splitTableIdentifierName(storeMgr.getDatabaseAdapter().getCatalogSeparator(), name.getSQLIdentifier());

        // Check for table existence first - a prerequisite
        // Note : The table name here has no quotes on it.
        ResultSet rs = dmd.getTables(c[0], c[1] == null ? getSchemaName() : c[1], c[2], null);
        try
        {
            if (rs.next()) // proceed if rs contains a table description for our table
            {
                // Note : The table name has no quottes on it here. Some JDBC drivers (e.g Oracle 8i) fail here when
                // the table name is an SQL keyword. This is likely an error in the JDBC driver
                ResultSet rsIndexes = storeMgr.getDatabaseAdapter().getExistingIndexes(conn,dmd,c[0], c[1] == null ? getSchemaName() : c[1], c[2]);
                try
                {
	                while (rsIndexes.next())
		            {
	                    boolean isUnique = !rsIndexes.getBoolean(4);
	                    if( isUnique )
	                    {
			                short idxType = rsIndexes.getShort(7);
			                if (idxType == DatabaseMetaData.tableIndexStatistic)
			                {
			                    continue;
			                }
		
			                String keyName=rsIndexes.getString(6);
			                SQLIdentifier idxName = new SQLIdentifier(dba, keyName);
			                CandidateKey can = (CandidateKey) candidateKeysByName.get(idxName);
			                if (can == null)
			                {
			                    can = new CandidateKey(this);
			                    can.setName(keyName);
			
			                    candidateKeysByName.put(keyName, can);
			                }
			
			                // Set the column
			                int colSeq = rsIndexes.getShort(8) - 1;
			                SQLIdentifier colName = new SQLIdentifier(dba, rsIndexes.getString(9));
			                Column col = (Column) columnsByName.get(colName);
			
			                /*
			                 * if (col == null) throw new
			                 * UnexpectedColumnException(this.toString(),
			                 * colName.toSQLIdentifier());
			                 * 
			                 * idx.setColumn(colSeq, col);
			                 */
			                if (col != null)
			                {
			                    can.setDatastoreField(colSeq, col);
			                }
	                    }
		            } 
                }
                finally
                {
                    rsIndexes.close();
                }
            }
        }
        finally
        {
            rs.close();
        }

        return candidateKeysByName;
    }

    
    /**
     * Accessor for indices on the actual table.
     * @param conn The JDBC Connection
     * @return Map of indices (keyed by the index name)
     * @throws SQLException Thrown when an error occurs in the JDBC call.
     */
    private Map getExistingIndices(Connection conn)
    throws SQLException
    {
        DatabaseMetaData dmd = conn.getMetaData();
        HashMap indicesByName = new HashMap();
        String[] c = RDBMSStoreHelper.splitTableIdentifierName(storeMgr.getDatabaseAdapter().getCatalogSeparator(), name.getSQLIdentifier());
        // Check for table existence first - a prerequisite
        // Note : The table name here has no quotes on it.
        ResultSet rs = dmd.getTables(c[0], c[1] == null ? getSchemaName() : c[1], c[2], null);
        try
        {
            if (rs.next()) // proceed if rs contains a table description for our table
            {
                // Note : The table name has no quottes on it here. Some JDBC drivers (e.g Oracle 9i, 10g) fail here when
                // the table name is an SQL keyword. This is likely an error in the JDBC driver
                ResultSet rsIndexes = storeMgr.getDatabaseAdapter().getExistingIndexes(conn,dmd,c[0], c[1] == null ? getSchemaName() : c[1], c[2]);
                try
                {
		            while (rsIndexes.next())
		            {
		                short idxType = rsIndexes.getShort(7);
		                if (idxType == DatabaseMetaData.tableIndexStatistic)
		                {
		                    continue;
		                }
	
		                String indexName=rsIndexes.getString(6);
		                SQLIdentifier idxName = new SQLIdentifier(dba, indexName);
		                Index idx = (Index) indicesByName.get(idxName);
		                if (idx == null)
		                {
		                    boolean isUnique = !rsIndexes.getBoolean(4);
		
		                    idx = new Index(this, isUnique);
		                    idx.setName(indexName);
		
		                    indicesByName.put(idxName, idx);
		                }
		
		                // Set the column
		                int colSeq = rsIndexes.getShort(8) - 1;
		                SQLIdentifier colName = new SQLIdentifier(dba, rsIndexes.getString(9));
		                Column col = (Column) columnsByName.get(colName);
		
		                /*
		                 * if (col == null) throw new
		                 * UnexpectedColumnException(this.toString(),
		                 * colName.toSQLIdentifier());
		                 * 
		                 * idx.setColumn(colSeq, col);
		                 */
		                if (col != null)
		                {
		                    idx.setColumn(colSeq, col);
		                }
		            }
                }
                finally
                {
                    rsIndexes.close();
                }
            }
        }
        finally
        {
            rs.close();
        }

        return indicesByName;
    }

    /**
     * Accessor for the SQL CREATE statements for this table.
     * @return List of statements.
     */
    protected List getSQLCreateStatements()
    {
        assertIsInitialized();

        ArrayList stmts = new ArrayList();
        stmts.add(dba.getCreateTableStatement(this, (Column[]) columns.toArray(new Column[columns.size()])));

        PrimaryKey pk = getPrimaryKey();
        if (pk.size() > 0)
        {
            // Some databases define the primary key on the create table
            // statement so we don't have a Statement for the primary key here.
            String pkStmt = dba.getAddPrimaryKeyStatement(new PrimaryKeyIdentifier(this), pk);
            if (pkStmt != null)
            {
                stmts.add(pkStmt);
            }
        }

        return stmts;
    }

    /**
     * Get SQL statements to add expected Foreign Keys that are not yet at the
     * table. If the returned Map is empty, the current FK setup is correct.
     * @param actualForeignKeysByName Actual Map of foreign keys
     * @return a Map with the SQL statements
     */
    protected Map getSQLAddFKStatements(Map actualForeignKeysByName)
    {
        assertIsInitialized();

        HashMap stmtsByFKName = new HashMap();
        List expectedForeignKeys = getExpectedForeignKeys();
        Iterator i = expectedForeignKeys.iterator();
        int n = 1;
        while (i.hasNext())
        {
            ForeignKey fk = (ForeignKey) i.next();
            if (!actualForeignKeysByName.containsValue(fk))
            {
                // If no name assigned, make one up
                if (fk.getName() == null)
                {
                    // Use the ForeignKeyIdentifier to generate the name
                    ForeignKeyIdentifier fkName;
                    do
                    {
                        fkName = new ForeignKeyIdentifier(this, n++);
                    }
                    while (actualForeignKeysByName.containsKey(fkName));
                    fk.setName(fkName.getSQLIdentifier());
                }
                String stmtText = dba.getAddForeignKeyStatement(fk);
                stmtsByFKName.put(fk.getName(), stmtText);
            }
        }

        return stmtsByFKName;
    }

    /**
     * Get SQL statements to add expected Candidate Keys that are not yet on the
     * table. If the returned Map is empty, the current Candidate Key setup is correct.
     * @param actualCandidateKeysByName Actual Map of candidate keys
     * @return a Map with the SQL statements
     */
    protected Map getSQLAddCandidateKeyStatements(Map actualCandidateKeysByName)
    {
        assertIsInitialized();

        HashMap stmtsByCKName = new HashMap();
        List expectedCandidateKeys = getExpectedCandidateKeys();
        Iterator i = expectedCandidateKeys.iterator();
        int n = 1;
        while (i.hasNext())
        {
            CandidateKey ck = (CandidateKey) i.next();
            if (!actualCandidateKeysByName.containsValue(ck))
            {
                // If no name assigned, make one up
                if (ck.getName() == null)
                {
                    // Use the CandidateKeyIdentifier to generate the name
                    CandidateKeyIdentifier ckName;
                    do
                    {
                        ckName = new CandidateKeyIdentifier(this, n++);
                    }
                    while (actualCandidateKeysByName.containsKey(ckName));
                    ck.setName(ckName.getSQLIdentifier());
                }
                String stmtText = dba.getAddCandidateKeyStatement(ck);
                stmtsByCKName.put(ck.getName(), stmtText);
            }
        }

        return stmtsByCKName;
    }

    /**
     * Utility to check if an index is necessary.
     * @param requiredIdx The index
     * @param actualIndices The actual indexes
     * @return Whether the index is needed.
     */
    private boolean isIndexReallyNeeded(Index requiredIdx,
                                        Collection actualIndices)
    {
        Iterator i = actualIndices.iterator();
        while (i.hasNext())
        {
            Index actualIdx = (Index) i.next();
            if (actualIdx.startsWith(requiredIdx))
            {
                return false;
            }
        }
        return true;
    }

    /**
     * Accessor for the CREATE INDEX statements for this table.
     * @param actualIndicesByName Map of actual indexes
     * @return Map of statements
     */
    protected Map getSQLCreateIndexStatements(Map actualIndicesByName)
    {
        assertIsInitialized();
        HashMap stmtsByIdxName = new HashMap();
        Set expectedIndices = getExpectedIndices();

        int n = 1;
        Iterator i = expectedIndices.iterator();
        while (i.hasNext())
        {
            Index idx = (Index) i.next();
            if (isIndexReallyNeeded(idx, actualIndicesByName.values()))
            {
                // If no name assigned, make one up
                if (idx.getName() == null)
                {
                    // Use IndexIdentifier to generate the name.
                    IndexIdentifier idxName;
                    do
                    {
                        idxName = new IndexIdentifier(this, idx.getUnique(), n++);
                        idx.setName(idxName.getSQLIdentifier());
                    }
                    while (actualIndicesByName.containsKey(idxName));
                }

                String stmtText = dba.getCreateIndexStatement(idx);
                stmtsByIdxName.put(idx.getName(), stmtText);
            }
        }
        return stmtsByIdxName;
    }

    /**
     * Accessor for the DROP statements for this table.
     * @return List of statements
     */
    protected List getSQLDropStatements()
    {
        assertIsInitialized();
        ArrayList stmts = new ArrayList();
        stmts.add(dba.getDropTableStatement(this));
        return stmts;
    }
}