/**********************************************************************
Copyright (c) 2002 Mike Martin and others.
All rights reserved. This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
2004 Andy Jefferson - rewritten to remove many levels of inheritance
    ...
**********************************************************************/
package org.jpox.store.rdbms.table;

import java.sql.Connection;
import java.sql.SQLException;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;

import org.jpox.store.exceptions.MissingColumnException;
import org.jpox.store.exceptions.MissingTableException;
import org.jpox.store.exceptions.NotAViewException;
import org.jpox.store.exceptions.PrimaryKeyColumnNotAllowedException;
import org.jpox.store.exceptions.UnexpectedColumnException;
import org.jpox.store.exceptions.ViewNotSupportedException;
import org.jpox.store.rdbms.Column;
import org.jpox.store.rdbms.RDBMSManager;
import org.jpox.store.rdbms.RDBMSStoreHelper;
import org.jpox.store.rdbms.columninfo.ColumnInfo;
import org.jpox.store.rdbms.sqlidentifier.SQLIdentifier;
import org.jpox.util.JPOXLogger;

/**
 * Representation of a View in a datastore (RDBMS).
 *
 * @version $Revision: 1.1 $
 **/
public abstract class ViewImpl extends AbstractTable
{
    /**
     * Constructor, taking just the StoreManager.
     * @param storeMgr The Store Manager
     **/
    public ViewImpl(RDBMSManager storeMgr)
    {
        super(storeMgr);
    }

    /**
     * Constructor, taking the table identifier.
     * @param name The identifier for the table.
     * @param storeMgr The Store Manager
     **/
    public ViewImpl(SQLIdentifier name, RDBMSManager storeMgr)
    {
        super(name, storeMgr);
    }
 
    /**
     * Method to validate the view in the datastore. Validates the existence of
     * the table, and then the specifications of the Columns.
     * @param conn The JDBC Connection
     * @param update Whether to update the view to fix errors (not used).
     * @return Whether the database was modified
     * @throws SQLException Thrown when an error occurs in the JDBC calls 
     */
    public boolean validate(Connection conn, boolean update)
    throws SQLException
    {
        assertIsInitialized();

        // Check existence
        int tableType = RDBMSStoreHelper.getTableType(storeMgr, name.getSQLIdentifier(), conn);
        if (tableType == TABLE_TYPE_MISSING)
        {
            throw new MissingTableException(this.toString());
        }
 
        long startTime = System.currentTimeMillis();
        if (JPOXLogger.RDBMS.isDebugEnabled())
        {
            JPOXLogger.RDBMS.debug(LOCALISER.msg("View.Validating",this));
        }
 
        // Check type
        if (tableType != TABLE_TYPE_VIEW)
        {
            throw new NotAViewException(this.toString());
        }

        // Validate the column(s)
        HashMap unvalidated = new HashMap(columnsByName);
        Iterator i = storeMgr.getColumnInfo(this, conn).iterator();
        while (i.hasNext())
        {
            ColumnInfo ci = (ColumnInfo)i.next();
            SQLIdentifier colName = new SQLIdentifier(dba, ci.getColumnName());
            Column col = (Column)unvalidated.get(colName);
            if (col == null)
            {
                if (!hasColumnName(colName))
                {
                    throw new UnexpectedColumnException(this.toString(),colName.getSQLIdentifier());
                }
 
                /*
                 * Otherwise it's a duplicate column name in the
                 * metadata and we ignore it.  Cloudscape is known to
                 * do this, although I think that's probably a bug.
                 */
            }
            else
            {
                col.validate(ci);
                unvalidated.remove(colName);
            }
        }
        if (unvalidated.size() > 0)
        {
            throw new MissingColumnException(this, unvalidated.values());
        }

        state = TABLE_STATE_VALIDATED;
        if (JPOXLogger.RDBMS.isDebugEnabled())
        {
            JPOXLogger.RDBMS.debug(LOCALISER.msg("JDBC.Statement.ExecutionTime",(System.currentTimeMillis() - startTime)));
        }

        return false;
    }

    /**
     * Method to override the default behaviour of modify() since we can't modify a View.
     * @param conn The Connection to use
     * @throws SQLException Never thrown
     */
    public void modify(Connection conn)
    throws SQLException
    {
        throw new ViewNotSupportedException("Modifications to views are impossible");
    }

    /**
     * Internal method to generate the SQL statements for dropping the view.
     * @return The List of SQL statements.
     **/
    protected List getSQLDropStatements()
    {
        assertIsInitialized();

        ArrayList stmts = new ArrayList();
        stmts.add(dba.getDropViewStatement(this));

        return stmts;
    }

    /**
     * Method to add a Column to the View.
     * @param col The column
     **/
    protected synchronized void addColumnInternal(Column col)
    {
        if (col.isPrimaryKey())
        {
            throw new PrimaryKeyColumnNotAllowedException(this.toString(), col.getName().toString());
        }

        super.addColumnInternal(col);
    }
}