/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others. All rights reserved. 
This program and the accompanying materials are made available under
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - coding standards
    ...
**********************************************************************/
package org.jpox.store.rdbms.typeinfo;

import java.sql.ResultSet;
import java.sql.SQLException;
import java.sql.Types;

import javax.jdo.JDOFatalDataStoreException;

import org.jpox.store.rdbms.columninfo.ColumnInfo;
import org.jpox.util.JPOXLogger;
import org.jpox.util.Localiser;

/**
 * Represents the metadata of a specific JDBC data type.  This class is
 * basically a data structure that makes accessing the JDBC type metadata
 * easier.  Each of the columns of information returned by
 * {@link java.sql.DatabaseMetaData#getTypeInfo()} is represented by a
 * public field in this class.
 * <p>
 * Subclasses of TypeInfo can be created on a per-DBMS basis to supply missing
 * metadata or correct faulty metadata obtained from that DBMS's JDBC driver(s).
 *</p> 
 *
 * @see org.jpox.store.rdbms.adapter.DatabaseAdapter#getTypeInfo
 * @see org.jpox.store.rdbms.adapter.DatabaseAdapter#newTypeInfo
 * @see java.sql.DatabaseMetaData#getTypeInfo
 * @version $Revision: 1.2 $ 
 */
public class TypeInfo
{
    private static final Localiser LOCALISER=Localiser.getInstance("org.jpox.store.rdbms.Localisation");

    /**
     * The DBMS-specific name for this data type.
     */
    public String typeName;

    /**
     * The JDBC (SQL) data type number of this data type.
     */
    public short dataType;

    /**
     * The maximum precision/length allowed for this data type.
     */
    public int precision;

    /**
     * The prefix used to quote a literal of this data type; may be
     * <tt>null</tt>.
     */
    public String literalPrefix;

    /**
     * The suffix used to quote a literal of this data type; may be
     * <tt>null</tt>.
     */
    public String literalSuffix;

    /**
     * Indicates the parameters used in defining columns of this type.
     */
    public String createParams;

    /**
     * Indicates whether null values are allowed for this data type.
     *
     * @see java.sql.DatabaseMetaData#typeNoNulls
     * @see java.sql.DatabaseMetaData#typeNullable
     * @see java.sql.DatabaseMetaData#typeNullableUnknown
     */
    public int nullable;

    /**
     * <tt>true</tt> indicates this data type is case-sensitive in comparisons,
     * <tt>false</tt> otherwise.
     */
    public boolean caseSensitive;

    /**
     * Indicates searchability of this data type in terms of the kinds of SQL
     * WHERE clauses that are allowed.
     *
     * @see java.sql.DatabaseMetaData#typePredNone
     * @see java.sql.DatabaseMetaData#typePredChar
     * @see java.sql.DatabaseMetaData#typePredBasic
     * @see java.sql.DatabaseMetaData#typeSearchable
     */
    public short searchable;

    /**
     * <tt>true</tt> indicates the type is unsigned, <tt>false</tt> otherwise.
     */
    public boolean unsignedAttribute;

    /**
     * <tt>true</tt> indicates the type can be assigned a fixed scale value,
     * such as for decimal or currency types, <tt>false</tt> otherwise.
     */
    public boolean fixedPrecScale;

    /**
     * <tt>true</tt> indicates the type automatically increments for each new
     * row inserted, <tt>false</tt> otherwise.
     */
    public boolean autoIncrement;

    /**
     * Localized version of the DBMS-specific type name of this data type.
     */
    public String localTypeName;

    /**
     * The minimum supported scale value for this data type.
     */
    public short minimumScale;

    /**
     * The maximum supported scale value for this data type.
     */
    public short maximumScale;

    /**
     * Indicates the numeric radix of this data type, which is usually 2 or 10.
     */
    public int numPrecRadix;

    private int hash = 0;

    /**
     * This is not defined in the JDBC type info, and defines if the type allows
     * specification of the precision in parentheses after the type name.
     * Added so that we can define particular types as not allowing this.
     */
    public boolean allowsPrecisionSpec = true;

    /**
     * Constructs a type information objectd from the current row of the given
     * result set.  The {@link ResultSet} object passed must have been obtained
     * from a call to java.sql.DatabaseMetaData.getTypeInfo().
     * <p>
     * This method only retrieves the values from the current row; the caller
     * is required to advance to the next row with {@link ResultSet#next}.
     *
     * @param rs    The result set returned from
     *              java.sql.DatabaseMetaData.getTypeInfo().
     *
     * @exception JDOFatalDataStoreException
     *      if a column of type information could not be retrieved from the
     *      result set.
     */
    public TypeInfo(ResultSet rs) throws JDOFatalDataStoreException
    {
        try
        {
            typeName          = rs.getString(1);
            dataType          = rs.getShort(2);
            precision         = (int)rs.getLong(3);
            literalPrefix     = rs.getString(4);
            literalSuffix     = rs.getString(5);
            createParams      = rs.getString(6);
            nullable          = rs.getInt(7);
            caseSensitive     = rs.getBoolean(8);
            searchable        = rs.getShort(9);
            unsignedAttribute = rs.getBoolean(10);
            fixedPrecScale    = rs.getBoolean(11);
            autoIncrement     = rs.getBoolean(12);
            localTypeName     = rs.getString(13);
            minimumScale      = rs.getShort(14);
            maximumScale      = rs.getShort(15);
            numPrecRadix      = rs.getInt(18);
        }
        catch (SQLException e)
        {
            throw new JDOFatalDataStoreException("Can't read JDBC metadata from result set", e);
        }
    }

    /**
     * Constructs a type information object from its individual attributes.
     *
     * <p>This can be useful to subclasses and/or custom DatabaseAdapters that
     * need to modify and/or correct the metadata returned by the JDBC driver.
     */
    public TypeInfo(String typeName,
                    short dataType,
                    int precision,
                    String literalPrefix,
                    String literalSuffix,
                    String createParams,
                    int nullable,
                    boolean caseSensitive,
                    short searchable,
                    boolean unsignedAttribute,
                    boolean fixedPrecScale,
                    boolean autoIncrement,
                    String localTypeName,
                    short minimumScale,
                    short maximumScale,
                    int numPrecRadix)
    {
        this.typeName          = typeName;
        this.dataType          = dataType;
        this.precision         = precision;
        this.literalPrefix     = literalPrefix;
        this.literalSuffix     = literalSuffix;
        this.createParams      = createParams;
        this.nullable          = nullable;
        this.caseSensitive     = caseSensitive;
        this.searchable        = searchable;
        this.unsignedAttribute = unsignedAttribute;
        this.fixedPrecScale    = fixedPrecScale;
        this.autoIncrement     = autoIncrement;
        this.localTypeName     = localTypeName;
        this.minimumScale      = minimumScale;
        this.maximumScale      = maximumScale;
        this.numPrecRadix      = numPrecRadix;
    }

    /**
     * Indicates whether the type of a given database column is "compatible"
     * with this type.
     * @param ci The ColumnInfo
     * @return Whether they are compatible
     */
    public boolean isCompatibleWith(ColumnInfo ci)
    {
        return areCompatibleTypes(dataType, ci.getDataType());
    }

    /**
     * Indicates whether validate columns with certain types
     * @param ci The ColumnInfo
     * @return Whether it is to be validated
     */
    public boolean isToValidate(ColumnInfo ci)
    {
        if (ci.getDataType() == Types.OTHER)
        {
            return false;
        }
        return true;
    }
    
    /**
     * Indicates whether some object is "equal to" this one.  Two <tt>TypeInfo
     * </tt> objects are considered equal if their <tt>typeName</tt> and
     * <tt>dataType</tt> fields are equal.
     *
     * @param   obj     the reference object with which to compare
     * @return  <tt>true</tt> if this object is equal to the obj argument;
     *          <tt>false</tt> otherwise.
     */
    public boolean equals(Object obj)
    {
        if (obj == this)
        {
            return true;
        }

        if (!(obj instanceof TypeInfo))
        {
            return false;
        }

        TypeInfo ti = (TypeInfo)obj;

        return typeName.equals(ti.typeName) && dataType == ti.dataType;
    }

    /**
     * Returns a hash code value for this object.
     *
     * @return  a hash code value for this object.
     */
    public int hashCode()
    {
        if (hash == 0)
        {
            hash = typeName.hashCode() ^ dataType;
        }

        return hash;
    }

    /**
     * Returns the string representation of this object.
     *
     * @return  string representation of this object.
     */
    public String toString()
    {
        StringBuffer str=new StringBuffer();

        str.append(this.getClass().getName() + "\n");
        str.append("  typeName          = " + typeName + "\n");
        str.append("  jdbcTypeName      = " + getJDBCTypeName(dataType) + "\n");
        str.append("  dataType          = " + dataType + "\n");
        str.append("  precision         = " + precision + "\n");
        str.append("  literalPrefix     = " + literalPrefix + "\n");
        str.append("  literalSuffix     = " + literalSuffix + "\n");
        str.append("  createParams      = " + createParams + "\n");
        str.append("  nullable          = " + nullable + "\n");
        str.append("  caseSensitive     = " + caseSensitive + "\n");
        str.append("  searchable        = " + searchable + "\n");
        str.append("  unsignedAttribute = " + unsignedAttribute + "\n");
        str.append("  fixedPrecScale    = " + fixedPrecScale + "\n");
        str.append("  autoIncrement     = " + autoIncrement + "\n");
        str.append("  localTypeName     = " + localTypeName + "\n");
        str.append("  minimumScale      = " + minimumScale + "\n");
        str.append("  maximumScale      = " + maximumScale + "\n");
        str.append("  numPrecRadix      = " + numPrecRadix + "\n");

        return str.toString();
    }

    /**
     * Returns the string name of the Java constant that equals the given JDBC
     * type number <tt>type</tt>.  JDBC type numbers are defined in
     * {@link Types}.
     *
     * <p>If <tt>type</tt> is not equal to one of the known type numbers,
     * the string "Unrecognized type (<i>type</i>)" is returned.
     *
     * @param   type    the JDBC type number.
     *
     * @return  the string name of the corresponding constant from
     *          {@link Types}.
     */
    public static String getJDBCTypeName(int type)
    {
        switch (type)
        {
            case Types.BIT:
                return "Types.BIT";
            case Types.TINYINT:
                return "Types.TINYINT";
            case Types.SMALLINT:
                return "Types.SMALLINT";
            case Types.INTEGER:
                return "Types.INTEGER";
            case Types.BIGINT:
                return "Types.BIGINT";
            case Types.FLOAT:
                return "Types.FLOAT";
            case Types.REAL:
                return "Types.REAL";
            case Types.DOUBLE:
                return "Types.DOUBLE";
            case Types.NUMERIC:
                return "Types.NUMERIC";
            case Types.DECIMAL:
                return "Types.DECIMAL";
            case Types.CHAR:
                return "Types.CHAR";
            case Types.VARCHAR:
                return "Types.VARCHAR";
            case Types.LONGVARCHAR:
                return "Types.LONGVARCHAR";
            case Types.DATE:
                return "Types.DATE";
            case Types.TIME:
                return "Types.TIME";
            case Types.TIMESTAMP:
                return "Types.TIMESTAMP";
            case Types.BINARY:
                return "Types.BINARY";
            case Types.VARBINARY:
                return "Types.VARBINARY";
            case Types.LONGVARBINARY:
                return "Types.LONGVARBINARY";
            case Types.NULL:
                return "Types.NULL";
            case Types.OTHER:
                return "Types.OTHER";
            case Types.JAVA_OBJECT:
                return "Types.JAVA_OBJECT";
            case Types.DISTINCT:
                return "Types.DISTINCT";
            case Types.STRUCT:
                return "Types.STRUCT";
            case Types.ARRAY:
                return "Types.ARRAY";
            case Types.BLOB:
                return "Types.BLOB";
            case Types.CLOB:
                return "Types.CLOB";
            case Types.REF:
                return "Types.REF";
            default:
                return "Unrecognized type (" + type + ")";
        }
    }

    /**
     * Method to return if 2 types are compatible.
     * @param expected The expected
     * @param actual The actual
     * @return Whether they are compatible.
     **/
    public static boolean areCompatibleTypes(int expected, int actual)
    {
        if (actual == Types.OTHER)
        {
            JPOXLogger.RDBMS.warn(LOCALISER.msg("TypeInfo.UnrecognizedDatatype"));
            return true;
        }
        
        switch (expected)
        {
            case Types.TINYINT:
            case Types.SMALLINT:
            case Types.INTEGER:
            case Types.BIGINT:
                return isIntegerType(actual);

            case Types.FLOAT:
            case Types.REAL:
            case Types.DOUBLE:
                return isFloatingType(actual);

            case Types.NUMERIC:
            case Types.DECIMAL:
                return isNumericType(actual);

            case Types.VARCHAR:
            case Types.LONGVARCHAR:
                return isCharacterType(actual);

            case Types.TIMESTAMP:
            case Types.DATE:
            case Types.TIME:
                return isDateType(actual);
    
            case Types.OTHER:
            case Types.BIT:
            case Types.CHAR:
            case Types.BINARY:
            case Types.VARBINARY:
            case Types.LONGVARBINARY:
            case Types.NULL:
            case Types.JAVA_OBJECT:
            case Types.DISTINCT:
            case Types.STRUCT:
            case Types.ARRAY:
            case Types.BLOB:
            case Types.CLOB:
            case Types.REF:
            default:
                return expected == actual;
        }
    }

    private static boolean isIntegerType(int type)
    {
        switch (type)
        {
            case Types.TINYINT:
            case Types.SMALLINT:
            case Types.INTEGER:
            case Types.BIGINT:
                return true;

            default:
                return isNumericType(type);
        }
    }

    private static boolean isDateType(int type)
    {
        switch (type)
        {
            case Types.TIMESTAMP:
            case Types.TIME:
            case Types.DATE:
                return true;

            default:
                return false;
        }
    }    
    

    private static boolean isFloatingType(int type)
    {
        switch (type)
        {
            case Types.FLOAT:
            case Types.REAL:
            case Types.DOUBLE:
                return true;

            default:
                return isNumericType(type);
        }
    }

    private static boolean isNumericType(int type)
    {
        switch (type)
        {
            case Types.NUMERIC:
            case Types.DECIMAL:
                return true;

            default:
                return false;
        }
    }

    /**
     * Verify if its LONGVARCHAR or VARCHAR. JPOX added value
     * @param type The type
     * @return Whether the type is of a character type
     */
    private static boolean isCharacterType(int type)
    {
        switch (type)
        {
            case Types.VARCHAR:
            case Types.LONGVARCHAR:
                return true;

            default:
                return false;
        }
    }    
}
