/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
    ...
**********************************************************************/
package org.jpox.store.scostore;

import java.util.Collection;
import java.util.Iterator;

import org.jpox.StateManager;
import org.jpox.store.DatastoreIdentifier;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.TableExpression;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.query.Query;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.rdbms.sqlidentifier.SQLIdentifier;

/**
 * Interface representation of the backing store for a Collection.
 *
 * @version $Revision: 1.1 $
 **/
public interface CollectionStore extends Store
{
    // --------------------------- Accessor Methods ----------------------------
 
    /**
     * Accessor for the element type in the collection.
     * @return The element type.
     **/
    String getElementType();

    /**
     * Accessor for the owner mapping.
     * @return Owner mapping for the collection.
     **/
    JavaTypeMapping getOwnerMapping(); 
 
    // -------------------------- Collection Methods ---------------------------
 
    /**
     * Accessor for an iterator for the collection.
     * @param sm State Manager for the collection. 
     * @return Iterator for the collection.
     **/
    Iterator iterator(StateManager sm);

    /**
     * Accessor for the size of the collection.
     * @param sm State Manager for the collection. 
     * @return The size of the collection.
     **/
    int size(StateManager sm);

    /**
     * Method to check if an element exists in the collection.
     * @param sm State Manager for the collection. 
     * @param element Element to check
     * @return Whether the element exists in the collection.
     **/
    boolean contains(StateManager sm, Object element);

    /**
     * Method to add an element to the collection.
     * @param sm State Manager for the collection. 
     * @param element Element to add
     * @return Whether the element was added ok
     **/
    boolean add(StateManager sm, Object element);

    /**
     * Method to add a collection of elements to the collection.
     * @param sm State Manager for the collection. 
     * @param elements Elements to add
     * @return Whether the elements were added ok
     **/
    boolean addAll(StateManager sm, Collection elements);

    /**
     * Method to remove an element from the collection.
     * @param sm State Manager for the collection. 
     * @param element Element to remove
     * @return Whether the element was removed ok
     **/
    boolean remove(StateManager sm, Object element);

    /**
     * Method to remove a collection of elements from the collection.
     * @param sm State Manager for the collection. 
     * @param elements Element to remove
     * @return Whether the elements were removed ok
     **/
    boolean removeAll(StateManager sm, Collection elements);

    /**
     * Method to clear the collection.
     * @param sm State Manager for the collection. 
     **/
    void clear(StateManager sm);

    // -------------------------- Query Methods --------------------------------
 
    /**
     * Method to return a new Query Statement containing the candidate class.
     * @param sm State Manager for the collection.
     * @param candidateClass Class of the elements.
     * @return The Query Statement.
     **/
    QueryStatement newQueryStatement(StateManager sm, String candidateClass);

    /**
     * Method to return an object factory for processing query statements.
     * @param sm State Manager for the collection.
     * @param stmt The Query Statement.
     * @param ignoreCache Whether to ignore the cache.
     * @return The Persistent object factory.
     **/
    Query.ResultObjectFactory newResultObjectFactory(StateManager sm,
                                                 QueryStatement stmt,
                                                 boolean ignoreCache);

    /**
     * Create a subquery for the given query that joins a SetStore
     * element table to the owner table. This subquery can subsequently be used
     * in an EXISTS expression to determine whether a Collection is empty. 
     * @param collRangeVar   The range variable for the "Collection" table. 
     * @return  A subquery for the given query that joins a SetStore
     *          element table to the owner table. 
     */ 
    QueryStatement getExistsSubquery(
            				JavaTypeMapping mapping,
                            TableExpression ownerTe,
                            SQLIdentifier collRangeVar);

    /**
     * Method used in queries when contains() has been invoked.
     * @param stmt The Query Statement 
     * @param ownerMapping the mapping for the owner. 
     * @param collRangeVar   The range variable for the "Collection" table. 
     **/
    ScalarExpression joinElementsTo(
                                QueryStatement stmt,
                                JavaTypeMapping ownerMapping,
                                TableExpression te,
                                DatastoreIdentifier collRangeVar,
                                Class filteredElementType,
                                TableExpression elmte,
                                DatastoreIdentifier elementRangeVar);
}
