/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - coding standards
2004 Andy Jefferson - added query methods and comments
2004 Erik Bengtson - added joinKeysValuesTo
2004 Erik Bengtson - added joinKeysToGet
    ...
**********************************************************************/
package org.jpox.store.scostore;

import java.util.Map;

import org.jpox.StateManager;
import org.jpox.store.expression.ScalarExpression;
import org.jpox.store.expression.TableExpression;
import org.jpox.store.mapping.JavaTypeMapping;
import org.jpox.store.query.Query;
import org.jpox.store.query.QueryStatement;
import org.jpox.store.rdbms.sqlidentifier.SQLIdentifier;

/**
 * Interface representation of the backing store for a Map.
 *
 * @version $Revision: 1.1 $
 **/
public interface MapStore extends Store
{
    // --------------------------- Accessor Methods ----------------------------

    /**
     * Accessor for the key type in the collection.
     * @return The key type.
     **/
    String getKeyType();

    /**
     * Accessor for the value type in the collection.
     * @return The value type.
     **/
    String getValueType();

    // -------------------------------- Map Methods ----------------------------
 
    /**
     * Accessor for whether the Map contains this value.
     * @param ownerSM State Manager for the Map.
     * @param value The value to check
     * @return Whether it is contained.
     **/
    boolean containsValue(StateManager ownerSM, Object value);

    /**
     * Accessor for whether the Map contains this key.
     * @param ownerSM State Manager for the Map.
     * @param key The key to check
     * @return Whether it is contained.
     **/
    boolean containsKey(StateManager ownerSM, Object key);

    /**
     * Accessor for a value from the Map.
     * @param ownerSM State Manager for the Map. 
     * @param key Key for the value.
     * @return Value for this key.
     **/
    Object get(StateManager ownerSM, Object key);

    /**
     * Method to add a value to the Map against this key.
     * @param ownerSM State Manager for the Map. 
     * @param key The key.
     * @param value The value.
     * @return Value that was previously against this key.
     **/
    Object put(StateManager ownerSM, Object key, Object value);

    /**
     * Method to add a map of values to the Map.
     * @param ownerSM State Manager for the Map. 
     * @param m The map to add.
     **/ 
    void putAll(StateManager ownerSM, Map m);

    /**
     * Method to remove a value from the Map.
     * @param ownerSM State Manager for the Map. 
     * @param key Key whose value is to be removed.
     * @return Value that was removed.
     **/
    Object remove(StateManager ownerSM, Object key);

    /**
     * Method to clear the collection.
     * @param ownerSM State Manager for the collection. 
     **/
    void clear(StateManager ownerSM);

    /**
     * Accessor for the keys in the Map.
     * @return Keys for the Map.
     **/
    SetStore keySetStore();

    /**
     * Accessor for the values in the Map.
     * @return Values for the Map.
     **/
    SetStore valueSetStore();

    /**
     * Accessor for the entry set for the Map.
     * @return Entry set for the Map.
     **/
    SetStore entrySetStore();

    // -------------------------- Query Methods --------------------------------

    /**
     * Method to create a query statement for a Map with values of the supplied
     * candidate class.
     * @param ownerSM StateManager for the Map
     * @param candidateClass The class for the values
     * @return The Query Statement.
     **/
    QueryStatement newQueryStatement(StateManager ownerSM,String candidateClass);

    /**
     * Utility to create a Result Object Factory to process the results of
     * selects returning values.
     * @param ownerSM State Manager for the Map.
     * @param stmt The Query Statement
     * @param ignoreCache Whether to ignore the cache
     * @return The result object factory.
     **/
    Query.ResultObjectFactory newResultObjectFactory(StateManager ownerSM,
                                                     QueryStatement stmt,
                                                     boolean ignoreCache);

    /**
	 * Create a subquery for the given query that joins a MapStore
     * value table to the owner table. This subquery can subsequently be used
     * in an EXISTS expression to determine whether a Map is empty or not. 
     * @param mapping Mapping for the owner
     * @param te Table Expression for the owner
     * @param mapRangeVar   The range variable for the "Map" table. 
     * @return  A subquery for the given query that joins a MapStore value
     *          table to the owner table. 
     **/
    QueryStatement getExistsSubquery(
            				JavaTypeMapping mapping,
            				TableExpression te,
                            SQLIdentifier mapRangeVar);

    /**
     * Utility to create a join for keys and values to be used in ai
     * containsEntry() query.
     * @param stmt The Query Statement to apply the join
     * @param ownerMapping Mapping for the owner
     * @param te Table Expression for the owner
     * @param mapRangeVar  The SQL alias, or "range variable", to assign to the
     *                     expression or to the main table.
     * @param filteredKeyType The Class Type for the filtered key
     * @param filteredValueType The Class Type for the filtered value
     * @param keyRangeVar  The SQL alias, or "range variable", to assign to the
     *                     expression or to the key table.
     * @param valueRangeVar The SQL alias, or "range variable", to assign to the
     *                      expression or to the value table.
     * @return an array with 2 elements of QueryColumnList. The first element
     * contains the columns from the key mapping and the second element the
     * columns from the value mapping
     */
    ScalarExpression[] joinKeysValuesTo(
                                QueryStatement stmt,
                                JavaTypeMapping ownerMapping,
                                TableExpression te,
                                SQLIdentifier mapRangeVar,
                                Class filteredKeyType,
                                Class filteredValueType,
                                SQLIdentifier keyRangeVar,
    							SQLIdentifier valueRangeVar);
    
    /**
     * Utility to create a join for keys to be used in a containsKey() query.
     * @param stmt The Query Statement 
     * @param ownerMapping Mapping for the owner
     * @param te Table Expression for the owner
     * @param mapRangeVar   The range variable for the "Map" table. 
     * @param filteredKeyType The Class Type for the filtered key
     * @param keyRangeVar  The SQL alias, or "range variable", to assign to the
     *                     expression or to the key table.
     * @return QueryColumnList with the columns from the key mapping 
     **/
    ScalarExpression joinKeysTo(
                                QueryStatement stmt,
                                JavaTypeMapping ownerMapping,
                                TableExpression te,
                                SQLIdentifier mapRangeVar,
                                Class filteredKeyType,
                                SQLIdentifier keyRangeVar);

    /**
     * Utility to create a join to be used in a containsValue() query.
     * @param stmt The Query Statement 
     * @param ownerMapping Mapping for the owner
     * @param te Table Expression for the owner
     * @param mapRangeVar   The range variable for the "Map" table. 
     * @param filteredValueType The Class Type for the filtered value
     * @param valueRangeVar The SQL alias, or "range variable", to assign to the
     *                      expression or to the value table.
     * @return QueryColumnList with the columns from the value mapping 
     **/
    ScalarExpression joinValuesTo(
                                QueryStatement stmt,
                                JavaTypeMapping ownerMapping,
                                TableExpression te,
                                SQLIdentifier mapRangeVar,
                                Class filteredValueType,
                                SQLIdentifier valueRangeVar);
    
    /**
     * Used as part of the Querying of Maps where a get(Key) is used.
     * @param stmt The Query Statement to apply the join
     * @param ownerMapping Mapping for the owner
     * @param te Table Expression for the owner
     * @param mapRangeVar  The SQL alias, or "range variable", to assign to the
     *                     expression or to the main table.
     * @param filteredKeyType The Class Type for the filtered key
     * @param keyRangeVar  The SQL alias, or "range variable", to assign to the
     *                     expression or to the key table.
     * @param valueRangeVar The SQL alias, or "range variable", to assign to the
     *                      expression or to the value table.
     * @return an array with 2 elements of QueryColumnList. The first element
     * contains the columns from the key mapping and the second element the
     * columns from the value mapping
     */
    ScalarExpression[] joinKeysToGet(
                                QueryStatement stmt,
                                JavaTypeMapping ownerMapping,
                                TableExpression te,
                                SQLIdentifier mapRangeVar,
                                Class filteredKeyType,
                                SQLIdentifier keyRangeVar,
    							SQLIdentifier valueRangeVar);
}
