/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved. 
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution.

Contributors:
    ...
**********************************************************************/
package org.jpox.util;

import java.io.File;
import java.lang.reflect.Field;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.HashSet;
import java.util.Iterator;
import java.util.List;

import javax.jdo.spi.Detachable;
import javax.jdo.spi.PersistenceCapable;

import org.jpox.ClassLoaderResolver;

/**
 * Utilities for handling classes.
 * These are to supplement the methods provided by the Class object.
 *
 * @version $Revision: 1.7 $
 **/
public class ClassUtils
{
    /**
     * Utility method to check if the specified class is PersistenceCapable.
     * @param loader The loader to load the class from
     * @param className Name of the class
     * @return Whether it is PersistenceCapable
     */
    public static boolean isPersistenceCapableClass(ClassLoader loader, String className)
    {
        try
        {
            Class cls = loader.loadClass(className);
            return isPersistenceCapableClass(cls);
        }
        catch (ClassNotFoundException cnfe)
        {
            JPOXLogger.JDO.error(">> isPersistenceCapableClass : couldnt find class " + className);
            return false;
        }
    }

    /**
     * Utility method to check if the specified class is Persistence Capable.
     * Checks that it implements PersistenceCapable and that it has some of
     * the jdoXXX methods.
     * @param cls The class to check
     * @return Whether the class is Persistence Capable
     */
    public static boolean isPersistenceCapableClass(Class cls)
    {
        if (cls == null)
        {
            return false;
        }

        boolean hasInterface = false;
        Class interfaces[] = cls.getInterfaces();
        for (int i=0; i<interfaces.length; i++)
        {
            if (interfaces[i].getName().equals(PersistenceCapable.class.getName()))
            {
                hasInterface = true;
                break;
            }
        }
        if (!hasInterface)
        {
            return false;
        }

        boolean hasMethod = false;
        Method methods[] = cls.getMethods();
        if (methods != null)
        {
            for (int i = 0; i < methods.length; i++)
            {
                if (methods[i].getName().equals("jdoReplaceField"))
                {
                    hasMethod = true;
                    break;
                }
            }
        }
        if (!hasMethod)
        {
            return false;
        }

        return true;
    }

    /**
     * Utility method to check if the specified object is Persistence Capable.
     * @param object The object to check
     * @return Whether the object is Persistence Capable
     */
    public static boolean isPersistenceCapable(Object object)
    {
        if (object == null)
        {
            return false;
        }
        return isPersistenceCapableClass(object.getClass());
    }

    /**
     * Utility method to check if the specified object is Detachable.
     * @param object The object to check
     * @return Whether the object is Detachable
     */
    public static boolean isDetachable(Object object)
    {
        if (!(object instanceof Detachable))
        {
            return false;
        }
        return true;
    }

    /**
     * Utility to perform the same function as the JDK 1.4 method
     * Boolean.valueOf(boolean), so we have no real dependency on JDK 1.4.
     * Please remove when our minimum requirement is JDK1.4 or above.
     * @param value The value
     * @return The Boolean value
     */
    public static Boolean booleanValueOf(boolean value)
    {
        if (value)
        {
            return Boolean.TRUE;
        }
        else
        {
            return Boolean.FALSE;
        }
    }

    /**
     * Utility to get the Class name given the file name and the file name of
     * the root of the package structure.
     * @param filename The filename of the class file
     * @param rootfilename The filename of the root of the package structure
     * @return The Class name
     **/
    public static String getClassnameForFilename(String filename,
                                                 String rootfilename)
    {
        if (filename == null)
        {
            return null;
        }

        String classname=filename;

        // Omit the root part of the filename
        if (rootfilename != null)
        {
            classname = classname.substring(rootfilename.length());
        }

        // Remove the ".class" suffix
        classname = classname.substring(0,classname.length()-6);

        String file_separator=System.getProperty("file.separator");

        // Remove any leading separator
        if (classname.indexOf(file_separator) == 0)
        {
            classname = classname.substring(file_separator.length());
        }

        // Change all separator characters to "."
        classname = StringUtils.replaceAll(classname,file_separator,".");

        return classname;
    }

    /**
     * Method to return the class files below the specified directory.
     * @param dir The directory
     * @param normal_classes Whether to include normal classes
     * @param inner_classes Whether to include inner classes
     * @return The class files (Collection of File objects).
     **/
    public static Collection getClassFilesForDirectory(File dir,
                                        boolean normal_classes,
                                        boolean inner_classes)
    {
        if (dir == null)
        {
            return null;
        }

        Collection classes=new HashSet();
        File[] files=dir.listFiles();
        if (files != null)
        {
            for (int i=0;i<files.length;i++)
            {
                if (files[i].isFile())
                {
                    // If this is a class file, add it
                    if (files[i].getName().endsWith(".class"))
                    {
                        boolean is_inner_class=isInnerClass(files[i].getName());
                        if ((normal_classes && !is_inner_class) ||
                            (inner_classes && is_inner_class))
                        {
                            classes.add(files[i]);
                        }
                    }
                }
                else
                {
                    // Check for classes in subdirectories
                    Collection child_classes=getClassFilesForDirectory(files[i],normal_classes,inner_classes);
                    if (child_classes != null && child_classes.size() > 0)
                    {
                        classes.addAll(child_classes);
                    }
                }
            }
        }

        return classes;
    }

    /**
     * Method to check whether a classname is for an inner class.
     * Currently checks for the presence of $ in the name.
     * @param class_name The class name
     * @return Whether it is an inner class
     **/
    public static boolean isInnerClass(String class_name)
    {
        if (class_name == null)
        {
            return false;
        }
        else if (class_name.indexOf('$') >= 0)
        {
            return true;
        }
        return false;
    }

    /**
     * Method to check for a default constructor on a class.
     * Particular relevance for JDO is the requirement for a default
     * constructor on all Persistence-Capable classes. Doesn't check
     * superclasses for the default constructor.
     * @param the_class The class
     * @return Whether it has a default constructor
     **/
    public static boolean hasDefaultConstructor(Class the_class)
    {
        if (the_class == null)
        {
            return false;
        }
        try
        {
            the_class.getDeclaredConstructor(null);
        }
        catch (Exception e)
        {
            return false;
        }

        return true;
    }

    /**
     * Method to return the superclasses for a class.
     * The superclasses will be ordered.
     * @param the_class The class
     * @return The superclass of this class.
     **/
    public static Collection getSuperclasses(Class the_class)
    {
        List superclasses = new ArrayList();

        Class c = the_class;
        boolean more_superclasses = true;
        while (more_superclasses)
        {
            Class superclass = c.getSuperclass();
            if (superclass != null)
            {
                superclasses.add(superclass);
                c = superclass;
            }
            else
            {
                more_superclasses = false;
            }
        }
        return superclasses;
    }

    /**
     * Utility to create the full class name given the package and class name.
     * Some examples 
     * <PRE>
     * packageName=test className=Test, returns result=test.Test
     * packageName=test className=test1.Test, returns result=test1.Test
     * packageName=&lt;null&gt; className=Test, returns result=Test
     * packageName=&lt;null&gt; className=test1.Test, returns result=test1.Test
     * </PRE>
     * @param pkg_name package name.
     * @param cls_name class name.
     * @return generated full class name.
     */
    public static String createFullClassName(String pkg_name, String cls_name)
    {
        if (StringUtils.isWhitespace(cls_name))
        {
            throw new IllegalArgumentException("Class name not specified");
        }
        else if (StringUtils.isWhitespace(pkg_name))
        {
            return cls_name;
        }
        else if (cls_name.indexOf('.') >= 0)
        {
            return cls_name;
        }
        return pkg_name + "." + cls_name;
    }

    /**
     * Method to check if 2 classes are direct descendents. So one of them is a
     * superclass of the other.
     * @param clr ClassLoaderResolver for loading the classes 
     * @param class_name_1 Name of first class
     * @param class_name_2 Name of second class
     * @return Whether they are direct descendents.
     */
    public static boolean classesAreDescendents(ClassLoaderResolver clr,
                                                String class_name_1,
                                                String class_name_2)
    {
        Class class_1=clr.classForName(class_name_1);
        Class class_2=clr.classForName(class_name_2);
        if (class_1 == null || class_2 == null)
        {
            return false;
        }

        // Check for direct descendents
        if (class_1.isAssignableFrom(class_2) ||
            class_2.isAssignableFrom(class_1))
        {
            return true;
        }

        return false;
    }

    /**
     * Utility to use Reflection to dump out the details of a class.
     * Will list all superclasses, interfaces, methods and fields.
     * Can be used, for example, in checking the methods adding by the
     * enhancement process. The information is dumped out the JPOX GENERAL log.
     * @param cls The class to dump out to the log
     */
    public static void dumpClassInformation(Class cls)
    {
        JPOXLogger.GENERAL.info("----------------------------------------");
        JPOXLogger.GENERAL.info("Class Information for class " + cls.getName());

        // Superclasses
        Collection superclasses=ClassUtils.getSuperclasses(cls);
        Iterator superclass_iter=superclasses.iterator();
        while (superclass_iter.hasNext())
        {
            Class superclass=(Class)superclass_iter.next();
            JPOXLogger.GENERAL.info("    Superclass : " + superclass.getName());
        }

        // Interfaces
        Class[] interfaces=cls.getInterfaces();
        if (interfaces != null)
        {
            for (int i=0;i<interfaces.length;i++)
            {
                JPOXLogger.GENERAL.info("    Interface : " + interfaces[i].getName());
            }
        }

        // Methods
        try
        {
            Method[] methods=cls.getDeclaredMethods();
            for (int i=0;i<methods.length;i++)
            {
                JPOXLogger.GENERAL.info("    Method : " + methods[i].toString());
            }
        }
        catch (Exception e)
        {
        }

        // Fields
        try
        {
            Field[] fields=cls.getDeclaredFields();
            for (int i=0;i<fields.length;i++)
            {
                JPOXLogger.GENERAL.info("    Field : " + fields[i].toString());
            }
        }
        catch (Exception e)
        {
        }
        JPOXLogger.GENERAL.info("----------------------------------------");
    }
}