/**********************************************************************
Copyright (c) 2002 Mike Martin (TJDO) and others. All rights reserved.
This program and the accompanying materials are made available under
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2003 Andy Jefferson - commented
    ...
**********************************************************************/
package org.jpox.util;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.StringTokenizer;

import javax.jdo.JDOUserException;

import org.jpox.ClassLoaderResolver;

/**
 * Utility class handling Imports.
 *
 * @version $Revision: 1.2 $
 **/
public class Imports
{
    private static final Localiser LOCALISER=Localiser.getInstance("org.jpox.util.Localisation");

    private HashMap primitives = new HashMap();
    private HashMap importedClassesByName = new HashMap();
    private HashSet importedPackageNames = new HashSet();

    /**
     * Constructor.
     **/
    public Imports()
    {
        primitives.put("boolean", boolean.class);
        primitives.put("byte",    byte.class);
        primitives.put("char",    char.class);
        primitives.put("short",   short.class);
        primitives.put("int",     int.class);
        primitives.put("long",    long.class);
        primitives.put("float",   float.class);
        primitives.put("double",  double.class);
        importedPackageNames.add("java.lang");
    }

    /**
     * Method to import the package given by the specified class.
     * @param className The class name
     **/
    public void importPackage(String className)
    {
        int lastDot = className.lastIndexOf('.');

        if (lastDot > 0)
        {
            importedPackageNames.add(className.substring(0, lastDot));
        }
    }

    /**
     * Utility to parse the imports.
     * @param imports The Imports string
     **/
    public void parseImports(String imports)
    {
        StringTokenizer t1 = new StringTokenizer(imports, ";");

        while (t1.hasMoreTokens())
        {
            String importDecl = t1.nextToken().trim();

            if (importDecl.length() == 0 && !t1.hasMoreTokens())
            {
                break;
            }

            StringTokenizer t2 = new StringTokenizer(importDecl, " ");

            if (t2.countTokens() != 2 || !t2.nextToken().equals("import"))
            {
                throw new JDOUserException(LOCALISER.msg("Import.InvalidImportDeclaration",importDecl));
            }

            String importName = t2.nextToken();
            int lastDot = importName.lastIndexOf(".");
            String lastPart = importName.substring(lastDot + 1);

            if (lastPart.equals("*"))
            {
                if (lastDot < 1)
                {
                    throw new JDOUserException(LOCALISER.msg("Import.InvalidPackageImport",importName));
                }

                importedPackageNames.add(importName.substring(0, lastDot));
            }
            else
            {
                if (importedClassesByName.put(lastPart, importName) != null)
                {
                    throw new JDOUserException(LOCALISER.msg("Import.DuplicateClassImport",importName));
                }
            }
        }
    }

    /**
     * Utility to resolve a class declaration.
     * @param classDecl The class declaration
     * @param clr ClassLoaderResolver
     * @param primaryClassLoader The primary ClassLoader for the class 
     * @return The class 
     * @throws ClassNotFoundException
     **/
    public Class resolveClassDeclaration(String classDecl, 
                                         ClassLoaderResolver clr, 
                                         ClassLoader primaryClassLoader)
    throws ClassNotFoundException
    {
        if (classDecl.indexOf('[') >= 0)
        {
            throw new JDOUserException(LOCALISER.msg("Import.ArrayTypesIllegal",classDecl));
        }

        Class c;
        if (classDecl.indexOf('.') < 0)
        {
            c = (Class)primitives.get(classDecl);
            if (c == null)
            {
                String cd = (String)importedClassesByName.get(classDecl);
                if (cd != null)
                {
                    c = clr.classForName(cd, primaryClassLoader);
                }
            }

            if (c == null)
            {
                Iterator packageNames = importedPackageNames.iterator();
                while (packageNames.hasNext())
                {
                    String packageName = (String)packageNames.next();

                    try
                    {
                        Class c1 = clr.classForName(packageName + '.' + classDecl, primaryClassLoader);
                        if (c != null && c1 != null)
                        {
                            // Duplicate definition of type
                            throw new JDOUserException(LOCALISER.msg("Import.AmbiguousClassDeclaration",c.getName(),c1.getName()));
                        }

                        c = c1;
                    }
                    catch (ClassNotFoundException e)
                    {
                        // Do nothing
                    }
                }

                if (c == null)
                {
                    throw new ClassNotFoundException(classDecl);
                }
                
                JPOXLogger.GENERAL.info(LOCALISER.msg("Import.ClassResolved",classDecl, c.getName()));

                // Give warning that the performance could have been improved by explicit usage
                JPOXLogger.GENERAL.info(LOCALISER.msg("Import.ClassNotImportedExplicitly",c.getName()));
            }
        }
        else
        {
            c = clr.classForName(classDecl, primaryClassLoader);
        }
        return c;
    }
}
