/**********************************************************************
Copyright (c) 2003 Andy Jefferson
All rights reserved.   This program and the accompanying materials
are made available under the terms of the JPOX License v1.0
which accompanies this distribution. 

Contributors:
    ...
**********************************************************************/
package org.jpox.util;

// Standard packages
import  java.io.InputStream;
import  java.io.IOException;
import  java.util.Collections;
import  java.util.HashMap;
import  java.util.Map;
import  java.util.Properties;

// Extension packages
import  javax.jdo.JDOHelper;
import  javax.jdo.PersistenceManagerFactory;
import  javax.jdo.PersistenceManager;
import  javax.jdo.JDOFatalException;

/**
 * Helper class providing a mean of accessing a PersistenceManager factory
 * for a specified data source. The data-store name is used for lookup for a
 * properties file. The class can be used in cases where you have multiple data
 * sources and each has its own properties file defining the location etc.
 *
 * @version $Revision: 1.1 $  
 */
public class JDOFactory
{
    private static final Localiser LOCALISER=Localiser.getInstance("org.jpox.util.Localisation");

    /** 
     * Map of factories. Factories are mapped using their datasource name as
     * the key. 
     **/
    private Map    datastoreFactories=null;

    /** Singleton of this class. */
    private static JDOFactory   me=null;

    /** Private constructor to enforce singleton pattern. */
    private JDOFactory()
    {
        datastoreFactories = Collections.synchronizedMap(new HashMap());
		if (JPOXLogger.JDO.isDebugEnabled())
        {
	        JPOXLogger.JDO.debug(LOCALISER.msg("JDOFactory.Initialised"));
        }
    }

    /** Accessor for the singleton instance.
     * @return  The JDOFactory instance. */
    public static JDOFactory    getFactory()
    {
        if (JDOFactory.me == null)
        {
            JDOFactory.me = new JDOFactory();
        }

        return JDOFactory.me;
    }

    /**
     * Utility to read in the properties from an external file. This method can
     * throw a JDOFatalException if it cannot find the properties file.
     *
     * @param name Name of the file containing the properties
     * @return The properties
     **/
    public static Properties getDatasourcePropertiesFromFile(String name)
    {
        Properties props=new Properties();
        try
        {
            InputStream store_properties=ClassLoader.getSystemResourceAsStream(name + ".properties");
            if (store_properties != null)
            {
                props.load(store_properties);
            }
            else
            {
                throw new JDOFatalException("Couldn't load resource '" + name + ".properties'.");
            }
        }
        catch (IOException ioe)
        {
            throw new JDOFatalException("Couldn't load resource '" + name + ".properties'.",ioe);
        }

        return props;
    }

    /**
     * Accessor for the PersistenceManager for a data store. This method can
     * throw a JDOFatalException if it cannot find the properties file.
     *
     * @param   data_store_name Name of data store
     * @return  PersistenceManager factory for that data store.
     **/
    public PersistenceManagerFactory getPersistenceManagerFactory(String data_store_name)
    {
		if (JPOXLogger.JDO.isDebugEnabled())
        {
	        JPOXLogger.JDO.debug(LOCALISER.msg("JDOFactory.FindingPersistenceManager",data_store_name));
        }
        Object  obj=this.datastoreFactories.get(data_store_name);
        if (obj != null)
        {
			if (JPOXLogger.JDO.isDebugEnabled())
            {
		        JPOXLogger.JDO.debug(LOCALISER.msg("JDOFactory.AlreadyExists"));
            }
            return (PersistenceManagerFactory)obj;
        }
        else
        {
			if (JPOXLogger.JDO.isDebugEnabled())
            {
	            JPOXLogger.JDO.debug(LOCALISER.msg("JDOFactory.DoesntExist"));
            }
            Properties props=getDatasourcePropertiesFromFile(data_store_name);

            // Obtain the persistence manager factory
            PersistenceManagerFactory   factory=JDOHelper.getPersistenceManagerFactory(props);

            datastoreFactories.put(data_store_name,factory);

            return factory;
        }
    }

    /**
     * Convenience accessor for a PersistenceManager from the datastore with
     * the given name.
     *
     * @param data_store_name Name of datastore
     * @return The PersistenceManager 
     **/
    public static PersistenceManager getPersistenceManager(String data_store_name)
    {
        PersistenceManagerFactory pmf=getFactory().getPersistenceManagerFactory(data_store_name);
        if (pmf == null)
        {
            return null;
        }
        else
        {
            return pmf.getPersistenceManager();
        }
    }
}
