/**********************************************************************
Copyright (c) 2004 Andy Jefferson and others. All rights reserved.
This program and the accompanying materials are made available under 
the terms of the JPOX License v1.0 which accompanies this distribution.

Contributors:
    ...
**********************************************************************/
package org.jpox.util;

import java.io.Serializable;
import java.lang.reflect.Constructor;
import java.lang.reflect.Field;
import java.lang.reflect.Modifier;

import org.jpox.JDOClassLoaderResolver;
import org.jpox.metadata.ClassMetaData;
import org.jpox.metadata.FieldMetaData;
import org.jpox.metadata.IdentityType;
import org.jpox.metadata.MetaDataManager;

/**
 * Utilities for checking and manipulating JDO objects.
 * 
 * @version $Revision: 1.2 $
 */
public class JDOUtils
{
    private static final Localiser LOCALISER = Localiser.getInstance("org.jpox.util.Localisation");

    /**
     * Utility to check if a primary-key class is valid.
     * Will throw a InvalidPrimaryKeyException if it is invalid, otherwise returning true.
     * @param pkClass The Primary Key class
     * @param cmd ClassMetaData for the PersistenceCapable class
     * @param noOfPkFields Number of primary key fields
     * @return Whether it is valid
     */
    public static boolean isValidPrimaryKeyClass(Class pkClass, ClassMetaData cmd, int noOfPkFields)
    {
        // When using inner class, must be static
        if (ClassUtils.isInnerClass(pkClass.getName()) &&
            !Modifier.isStatic(pkClass.getModifiers()))
        {
            throw new InvalidPrimaryKeyException(LOCALISER, "PrimaryKey.InnerNotStaticError",
                    cmd.getFullClassName(), pkClass.getName());
        }

        // Must be public
        if (!Modifier.isPublic(pkClass.getModifiers()))
        {
            throw new InvalidPrimaryKeyException(LOCALISER, "PrimaryKey.NotPublicError",
                    cmd.getFullClassName(), pkClass.getName());
        }

        // Must implement Serializable
        if (!Serializable.class.isAssignableFrom(pkClass))
        {
            throw new InvalidPrimaryKeyException(LOCALISER, "PrimaryKey.NotSerializableError",
                    cmd.getFullClassName(), pkClass.getName());
        }

        // a). JDO's SingleFieldIdentity class
        if (AIDUtils.isSingleFieldIdentityClass(pkClass.getName()))
        {
            if (noOfPkFields != 1)
            {
                throw new InvalidPrimaryKeyException(LOCALISER, "PrimaryKey.SingleFieldIdMultipleFieldsError",
                    cmd.getFullClassName());
            }
        }
        // b). Users own primary key class
        else
        {
            // Must have public default constructor
            try
            {
                Constructor constructor = pkClass.getConstructor(new Class[0]);
                if (constructor == null ||
                    !Modifier.isPublic(constructor.getModifiers()))
                {
                    throw new InvalidPrimaryKeyException(LOCALISER, "PrimaryKey.DefaultConstructorError",
                        cmd.getFullClassName(), pkClass.getName());
                }
            }
            catch (NoSuchMethodException ex)
            {
                throw new InvalidPrimaryKeyException(LOCALISER, "PrimaryKey.DefaultConstructorError",
                    cmd.getFullClassName(), pkClass.getName());
            }

            // Must have public String arg constructor
            try
            {
                Constructor constructor=
                    pkClass.getConstructor(new Class[] {String.class});
                if (constructor == null ||
                    !Modifier.isPublic(constructor.getModifiers()))
                {
                    throw new InvalidPrimaryKeyException(LOCALISER, "PrimaryKey.StringConstructorError",
                        cmd.getFullClassName(), pkClass.getName());
                }
            }
            catch (NoSuchMethodException nsme)
            {
            }

            // Must override toString() method
            try
            {
                java.lang.reflect.Method method=pkClass.getMethod("toString",new Class[0]);
                if (method == null ||
                    !Modifier.isPublic(method.getModifiers()) ||
                    method.getDeclaringClass().equals(Object.class))
                {
                    throw new InvalidPrimaryKeyException(LOCALISER, "PrimaryKey.ToStringError",
                        cmd.getFullClassName(), pkClass.getName());
                }
            }
            catch (NoSuchMethodException nsme)
            {
            }

            // Check the field types of the objectid-class
            Field[] ref_fields=pkClass.getDeclaredFields();
            int no_objectid_fields=0;
            for (int j=0;j<ref_fields.length;j++)
            {
                if (!Modifier.isStatic(ref_fields[j].getModifiers()))
                {
                    // All non-static fields must be serializable
                    if (!ref_fields[j].getType().isPrimitive() &&
                        !(Serializable.class).isAssignableFrom(ref_fields[j].getType()))
                    {
                        throw new InvalidPrimaryKeyException(LOCALISER, "PrimaryKey.FieldNotSerializableError",
                            cmd.getFullClassName(), pkClass.getName(),
                            ref_fields[j].getName());
                    }

                    // All non-static fields must be public
                    if (!Modifier.isPublic(ref_fields[j].getModifiers()))
                    {
                        throw new InvalidPrimaryKeyException(LOCALISER, "PrimaryKey.FieldNotPublicError",
                            cmd.getFullClassName(), pkClass.getName(),
                            ref_fields[j].getName());
                    }

                    // non-static fields of objectid-class include
                    // persistence-capable object field
                    FieldMetaData the_field = cmd.getField(ref_fields[j].getName());
                    boolean found_field=false;
                    if (the_field == null)
                    {
                        throw new InvalidPrimaryKeyException(LOCALISER, "PrimaryKey.FieldNotFound",
                            cmd.getFullClassName(), pkClass.getName(),
                            ref_fields[j].getName());
                    }

                    // check if the field in objectid-class has the
                    // same Type as the Type declared in the
                    // PersistenceCapable class 
                    if (the_field.getTypeName().equals(ref_fields[j].getType().getName()))
                    {
                        found_field = true;
                    }

                    // if not the same type, check for primary key if
                    // identifying relationships
                    if (!found_field)
                    {
                        MetaDataManager mgr=cmd.getMetaDataManager();
                        if (mgr == null)
                        {
                            mgr = new MetaDataManager();
                        }
                        String ref_field_type=ref_fields[j].getType().getName();
                        ClassMetaData ref_cmd=mgr.getMetaDataForClass(the_field.getTypeName(),new JDOClassLoaderResolver());
                        if (ref_cmd != null && 
                            ref_field_type.equals(ref_cmd.getObjectidClass()))
                        {
                            found_field = true;
                        }
                    }
                    if (!found_field)
                    {
                        throw new InvalidPrimaryKeyException(LOCALISER, "PrimaryKey.FieldPersistenceCapableError",
                            cmd.getFullClassName(), pkClass.getName(),
                            ref_fields[j].getName());
                    }

                    no_objectid_fields++;
                }
            }

            // No of Primary Key fields and no of fields in
            // objectid-class must concur
            if (noOfPkFields != no_objectid_fields &&
                cmd.getIdentityType() == IdentityType.APPLICATION)
            {
                throw new InvalidPrimaryKeyException(LOCALISER, "PrimaryKey.IncorrectNumberOfPKFieldsError",
                    cmd.getFullClassName(), pkClass.getName(),
                    ""+noOfPkFields, ""+no_objectid_fields);
            }
        }

        return true;
    }
}