/**********************************************************************
Copyright (c) 2003 Andy Jefferson and others. All rights reserved. 
This program and the accompanying materials are made available under
the terms of the JPOX License v1.0 which accompanies this distribution. 

Contributors:
2004 Erik Bengtson - added exception methods
    ...
**********************************************************************/
package org.jpox.util;

import java.util.Hashtable;
import java.util.Locale;
import java.util.ResourceBundle;
import java.util.MissingResourceException;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.lang.reflect.InvocationTargetException;
import java.sql.SQLException;
import java.text.MessageFormat;

/**
 * Resource Bundle manager, providing simplified means of using MessageFormat
 * to localise the JPOX system.
 *
 * @version $Revision: 1.1 $
 */
public class Localiser
{
    private static Hashtable helpers=new Hashtable();

    private static Locale locale=Locale.getDefault();

    private ResourceBundle bundle=null;

    /**
     * Private constructor to prevent outside instantiation. Operates on a
     * principle of one helper per bundle.
     * @param bundle_name the name of the resource bundle
     * @param class_loader the class loader from which to load the resource
     * bundle
     */
    private Localiser(String bundle_name,ClassLoader class_loader)
    {
        // We can assume that the bundle hasn't been loaded since no helper has
        // been found for it
        try
        {
            bundle = ResourceBundle.getBundle(bundle_name,locale,class_loader);
        }
        catch (MissingResourceException mre)
        {
            JPOXLogger.GENERAL.error("ResourceBundle " + bundle_name + " for locale " + locale + " was not found!");
        }
    }
 
    /**
     * Accessor for a helper instance for a bundle.
     * @param bundle_name the name of the bundle
     * @return the helper instance bound to the bundle
     */
    public static Localiser getInstance(String bundle_name)
    {
        return getInstance(bundle_name,Localiser.class.getClassLoader());
    }

    /**
     * Accessor for a helper instance for a bundle.
     * @param bundle_name the name of the bundle
     * @param class_loader the class loader from which to load the resource
     * bundle
     * @return the helper instance bound to the bundle
     */
    public static Localiser getInstance(String bundle_name,
                                             ClassLoader class_loader)
    {
        Localiser localiser=(Localiser)helpers.get(bundle_name);
        if (localiser != null)
        {
            return localiser;
        }
        localiser = new Localiser(bundle_name,class_loader);
        helpers.put(bundle_name,localiser);

        return localiser;
    }

    /**
     * Message formatter for an internationalised message.
     * @param message_key the message key
     * @return the resolved message text
     */
    public String msg(String message_key)
    {
        return bundle.getString(message_key);
    }

    /**
     * Message formatter with one argument passed in that will be embedded in an
     * internationalised message.
     * @param message_key the message key
     * @param arg1 the first argument
     * @return the resolved message text
     */
    public String msg(String message_key,Object arg1)
    {
        Object[] args={arg1};
        return getMessage(bundle,message_key,args);
    }

    /**
     * Message formatter with a series of arguments passed in that will be
     * embedded in an internationalised message.
     * @param message_key the message key
     * @param arg1 the first argument
     * @param arg2 the second argument
     * @return the resolved message text
     */
    public String msg(String message_key,Object arg1,Object arg2)
    {
        Object[] args={arg1,arg2};
        return getMessage(bundle,message_key,args);
    }

    /**
     * Message formatter with a series of arguments passed in that will be
     * embedded in an internationalised message.
     * @param message_key the message key
     * @param arg1 the first argument
     * @param arg2 the second argument
     * @param arg3 the third argument
     * @return the resolved message text
     */
    public String msg(String message_key,Object arg1,Object arg2,Object arg3)
    {
        Object[] args={arg1,arg2,arg3};
        return getMessage(bundle,message_key,args);
    }

    /**
     * Message formatter with a series of arguments passed in that will be
     * embedded in an internationalised message.
     * @param message_key the message key
     * @param arg1 the first argument
     * @param arg2 the second argument
     * @param arg3 the third argument
     * @param arg4 the third argument
     * @return the resolved message text
     */
    public String msg(String message_key,Object arg1,Object arg2,Object arg3,Object arg4)
    {
        Object[] args={arg1,arg2,arg3,arg4};
        return getMessage(bundle,message_key,args);
    }

    /**
     * Message formatter with an array of arguments passed in that will be
     * embedded in an internationalised message.
     * @param message_key the message key
     * @param args the array of arguments
     * @return the resolved message text
     */
    public String msg(String message_key,Object[] args)
    {
        return getMessage(bundle,message_key,args);
    }

    /**
     * Message formatter with one argument passed in that will be embedded in an
     * internationalised message.
     * @param message_key the message key
     * @param arg the argument
     * @return the resolved message text
     */
    public String msg(String message_key,int arg)
    {
        Object[] args={new Integer(arg)};
        return getMessage(bundle,message_key,args);
    }
 
    /**
     * Message formatter with one argument passed in that will be embedded in an
     * internationalised message.
     * @param message_key the message key
     * @param arg the argument
     * @return the resolved message text
     */
    public String msg(String message_key,long arg)
    {
        Object[] args={new Long(arg)};
        return getMessage(bundle,message_key,args);
    }

    /**
     * Message formatter with one argument passed in that will be embedded in an
     * internationalised message.
     * @param message_key the message key
     * @param arg the argument
     * @return the resolved message text
     */
    public String msg(String message_key,boolean arg)
    {
        Object[] args={String.valueOf(arg)};
        return getMessage(bundle,message_key,args);
    }

    /**
     * Gets a String message from Exceptions. This method transforms nested
     * exceptions into printable messages.
     * @param exception to be read and transformed into a messsage to print
     * @return the message to output
     */
    private static String getStringFromException(java.lang.Throwable exception)
    {
        StringBuffer msg = new StringBuffer();
        if (exception != null)
        {
            StringWriter stringWriter = new StringWriter();
            PrintWriter printWriter = new PrintWriter(stringWriter);
            exception.printStackTrace(printWriter);
            printWriter.close();
            try
            {
                stringWriter.close();
            }
            catch(Exception e)
            {
                //do nothing
            }
            msg.append(exception.getMessage());
            msg.append('\n');
            msg.append(stringWriter.toString());
            
            // JDBC: SQLException
            if (exception instanceof SQLException)
            {
                if (((SQLException) exception).getNextException() != null)
                {
                    msg.append('\n');
                    msg.append(getStringFromException(((SQLException) exception).getNextException()));
                }
            }
            // Reflection: InvocationTargetException
            else if (exception instanceof InvocationTargetException)
            {
                if (((InvocationTargetException) exception).getTargetException() != null)
                {
                    msg.append('\n');
                    msg.append(getStringFromException(((InvocationTargetException) exception).getTargetException()));
                }
            }
            
            // Add more exceptions here, so we can provide a relly complete information at the log
        }
        return msg.toString();
    }
    
    /**
     * Method to provide the MessageFormat logic. All of the msg()
     * methods are wrappers to this one.
     * @param the_bundle the resource bundle
     * @param message_key the message key
     * @param msg_args an array of arguments to substitute into the message
     * @return the resolved message text
     */
    final private static String getMessage(ResourceBundle the_bundle,
                                           String message_key,
                                           Object msg_args[]) 
    {
        if (message_key == null)
        {
            JPOXLogger.GENERAL.error("Attempt to retrieve resource with NULL name !");
            return null;
        }

        for (int i=0; i<msg_args.length; i++)
        {
            if (msg_args[i] == null)
            {
                msg_args[i] = "";
            }
            if (Throwable.class.isAssignableFrom(msg_args[i].getClass()))
            { 
                msg_args[i] = getStringFromException((Throwable)msg_args[i]);
            }
        }
        try
        {
            MessageFormat formatter=new MessageFormat(the_bundle.getString(message_key));
            return formatter.format(msg_args);
        }
        catch (MissingResourceException mre)
        {
            JPOXLogger.GENERAL.error("Parameter " + message_key + " doesn't exist for bundle " + the_bundle);
        }

        return null;
    }
}
